/* Automatically @generated by tree-sitter v0.25.10 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 2127
#define LARGE_STATE_COUNT 507
#define SYMBOL_COUNT 415
#define ALIAS_COUNT 0
#define TOKEN_COUNT 233
#define EXTERNAL_TOKEN_COUNT 1
#define FIELD_COUNT 14
#define MAX_ALIAS_SEQUENCE_LENGTH 12
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 26
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym_comment = 1,
  sym_decimal_integer_literal = 2,
  sym_hexadecimal_integer_literal = 3,
  sym_real_literal = 4,
  aux_sym_expandable_string_literal_token1 = 5,
  aux_sym_expandable_string_literal_token2 = 6,
  aux_sym_expandable_string_literal_token3 = 7,
  aux_sym_expandable_string_literal_token4 = 8,
  anon_sym_DQUOTE_DQUOTE = 9,
  anon_sym_DOLLAR = 10,
  aux_sym_expandable_string_literal_token5 = 11,
  aux_sym_expandable_here_string_literal_token1 = 12,
  aux_sym_expandable_here_string_literal_token2 = 13,
  aux_sym_expandable_here_string_literal_token3 = 14,
  aux_sym_expandable_here_string_literal_token4 = 15,
  aux_sym_expandable_here_string_literal_token5 = 16,
  sym_verbatim_string_characters = 17,
  sym_verbatim_here_string_characters = 18,
  sym_simple_name = 19,
  sym_type_identifier = 20,
  anon_sym_DOT = 21,
  anon_sym_LBRACK = 22,
  anon_sym_EQ = 23,
  anon_sym_BANG_EQ = 24,
  anon_sym_PLUS_EQ = 25,
  anon_sym_STAR_EQ = 26,
  anon_sym_SLASH_EQ = 27,
  anon_sym_PERCENT_EQ = 28,
  anon_sym_DASH_EQ = 29,
  anon_sym_GT = 30,
  anon_sym_GT_GT = 31,
  anon_sym_2_GT = 32,
  anon_sym_2_GT_GT = 33,
  anon_sym_3_GT = 34,
  anon_sym_3_GT_GT = 35,
  anon_sym_4_GT = 36,
  anon_sym_4_GT_GT = 37,
  anon_sym_5_GT = 38,
  anon_sym_5_GT_GT = 39,
  anon_sym_6_GT = 40,
  anon_sym_6_GT_GT = 41,
  anon_sym_STAR_GT = 42,
  anon_sym_STAR_GT_GT = 43,
  anon_sym_LT = 44,
  anon_sym_STAR_GT_AMP1 = 45,
  anon_sym_2_GT_AMP1 = 46,
  anon_sym_3_GT_AMP1 = 47,
  anon_sym_4_GT_AMP1 = 48,
  anon_sym_5_GT_AMP1 = 49,
  anon_sym_6_GT_AMP1 = 50,
  anon_sym_STAR_GT_AMP2 = 51,
  anon_sym_1_GT_AMP2 = 52,
  anon_sym_3_GT_AMP2 = 53,
  anon_sym_4_GT_AMP2 = 54,
  anon_sym_5_GT_AMP2 = 55,
  anon_sym_6_GT_AMP2 = 56,
  aux_sym_comparison_operator_token1 = 57,
  aux_sym_comparison_operator_token2 = 58,
  aux_sym_comparison_operator_token3 = 59,
  aux_sym_comparison_operator_token4 = 60,
  aux_sym_comparison_operator_token5 = 61,
  aux_sym_comparison_operator_token6 = 62,
  aux_sym_comparison_operator_token7 = 63,
  aux_sym_comparison_operator_token8 = 64,
  aux_sym_comparison_operator_token9 = 65,
  aux_sym_comparison_operator_token10 = 66,
  aux_sym_comparison_operator_token11 = 67,
  aux_sym_comparison_operator_token12 = 68,
  aux_sym_comparison_operator_token13 = 69,
  aux_sym_comparison_operator_token14 = 70,
  aux_sym_comparison_operator_token15 = 71,
  aux_sym_comparison_operator_token16 = 72,
  aux_sym_comparison_operator_token17 = 73,
  aux_sym_comparison_operator_token18 = 74,
  aux_sym_comparison_operator_token19 = 75,
  aux_sym_comparison_operator_token20 = 76,
  aux_sym_comparison_operator_token21 = 77,
  aux_sym_comparison_operator_token22 = 78,
  aux_sym_comparison_operator_token23 = 79,
  aux_sym_comparison_operator_token24 = 80,
  aux_sym_comparison_operator_token25 = 81,
  aux_sym_comparison_operator_token26 = 82,
  aux_sym_comparison_operator_token27 = 83,
  aux_sym_comparison_operator_token28 = 84,
  aux_sym_comparison_operator_token29 = 85,
  aux_sym_comparison_operator_token30 = 86,
  aux_sym_comparison_operator_token31 = 87,
  aux_sym_comparison_operator_token32 = 88,
  aux_sym_comparison_operator_token33 = 89,
  aux_sym_comparison_operator_token34 = 90,
  aux_sym_comparison_operator_token35 = 91,
  aux_sym_comparison_operator_token36 = 92,
  aux_sym_comparison_operator_token37 = 93,
  aux_sym_comparison_operator_token38 = 94,
  aux_sym_comparison_operator_token39 = 95,
  aux_sym_comparison_operator_token40 = 96,
  aux_sym_comparison_operator_token41 = 97,
  aux_sym_comparison_operator_token42 = 98,
  aux_sym_comparison_operator_token43 = 99,
  aux_sym_comparison_operator_token44 = 100,
  aux_sym_comparison_operator_token45 = 101,
  aux_sym_comparison_operator_token46 = 102,
  aux_sym_comparison_operator_token47 = 103,
  aux_sym_comparison_operator_token48 = 104,
  aux_sym_comparison_operator_token49 = 105,
  aux_sym_comparison_operator_token50 = 106,
  aux_sym_format_operator_token1 = 107,
  anon_sym_DOLLAR_DOLLAR = 108,
  anon_sym_DOLLAR_CARET = 109,
  anon_sym_DOLLAR_QMARK = 110,
  anon_sym_DOLLAR_ = 111,
  aux_sym_variable_token1 = 112,
  aux_sym_variable_token2 = 113,
  sym_braced_variable = 114,
  sym_generic_token = 115,
  sym__command_token = 116,
  sym_command_parameter = 117,
  aux_sym__verbatim_command_argument_chars_token1 = 118,
  aux_sym__verbatim_command_argument_chars_token2 = 119,
  aux_sym__verbatim_command_argument_chars_token3 = 120,
  anon_sym_SEMI = 121,
  aux_sym_param_block_token1 = 122,
  anon_sym_LPAREN = 123,
  anon_sym_RPAREN = 124,
  anon_sym_COMMA = 125,
  aux_sym_block_name_token1 = 126,
  aux_sym_block_name_token2 = 127,
  aux_sym_block_name_token3 = 128,
  aux_sym_block_name_token4 = 129,
  anon_sym_LBRACE = 130,
  anon_sym_RBRACE = 131,
  aux_sym_if_statement_token1 = 132,
  aux_sym_elseif_clause_token1 = 133,
  aux_sym_else_clause_token1 = 134,
  aux_sym_switch_statement_token1 = 135,
  aux_sym_switch_parameter_token1 = 136,
  aux_sym_switch_parameter_token2 = 137,
  aux_sym_switch_parameter_token3 = 138,
  aux_sym_switch_parameter_token4 = 139,
  aux_sym_switch_parameter_token5 = 140,
  aux_sym_switch_condition_token1 = 141,
  aux_sym_foreach_statement_token1 = 142,
  aux_sym_foreach_statement_token2 = 143,
  aux_sym_for_statement_token1 = 144,
  aux_sym_for_statement_token2 = 145,
  aux_sym_while_statement_token1 = 146,
  aux_sym_do_statement_token1 = 147,
  aux_sym_do_statement_token2 = 148,
  aux_sym_function_statement_token1 = 149,
  aux_sym_function_statement_token2 = 150,
  aux_sym_function_statement_token3 = 151,
  aux_sym_flow_control_statement_token1 = 152,
  aux_sym_flow_control_statement_token2 = 153,
  aux_sym_flow_control_statement_token3 = 154,
  aux_sym_flow_control_statement_token4 = 155,
  aux_sym_flow_control_statement_token5 = 156,
  sym_label = 157,
  aux_sym_trap_statement_token1 = 158,
  aux_sym_try_statement_token1 = 159,
  aux_sym_catch_clause_token1 = 160,
  aux_sym_finally_clause_token1 = 161,
  aux_sym_data_statement_token1 = 162,
  aux_sym_data_commands_allowed_token1 = 163,
  aux_sym_inlinescript_statement_token1 = 164,
  aux_sym_parallel_statement_token1 = 165,
  aux_sym_sequence_statement_token1 = 166,
  anon_sym_AMP_AMP = 167,
  anon_sym_PIPE_PIPE = 168,
  anon_sym_PIPE = 169,
  anon_sym_AMP = 170,
  aux_sym__expandable_string_literal_immediate_token1 = 171,
  aux_sym__expandable_string_literal_immediate_token2 = 172,
  aux_sym__expandable_string_literal_immediate_token3 = 173,
  anon_sym_DQUOTE_DQUOTE2 = 174,
  anon_sym_DOLLAR2 = 175,
  anon_sym_DQUOTE = 176,
  aux_sym_command_name_token1 = 177,
  aux_sym_command_name_token2 = 178,
  aux_sym_command_name_token3 = 179,
  aux_sym_command_name_token4 = 180,
  aux_sym_command_name_token5 = 181,
  aux_sym_command_name_token6 = 182,
  aux_sym_command_name_token7 = 183,
  aux_sym_command_name_token8 = 184,
  aux_sym_command_name_token9 = 185,
  aux_sym_command_name_token10 = 186,
  aux_sym_command_name_token11 = 187,
  aux_sym_command_name_token12 = 188,
  aux_sym_command_name_token13 = 189,
  aux_sym_command_name_token14 = 190,
  aux_sym_command_name_token15 = 191,
  aux_sym_command_name_token16 = 192,
  aux_sym_command_name_token17 = 193,
  aux_sym_command_name_token18 = 194,
  aux_sym_command_name_token19 = 195,
  anon_sym_DQUOTE2 = 196,
  anon_sym_SQUOTE_SQUOTE = 197,
  sym_path_command_name_token = 198,
  sym_stop_parsing = 199,
  anon_sym_SPACE = 200,
  anon_sym_COLON = 201,
  anon_sym_DASH_DASH_PERCENT = 202,
  aux_sym_class_attribute_token1 = 203,
  aux_sym_class_attribute_token2 = 204,
  aux_sym_class_statement_token1 = 205,
  aux_sym_enum_statement_token1 = 206,
  aux_sym_logical_expression_token1 = 207,
  aux_sym_logical_expression_token2 = 208,
  aux_sym_logical_expression_token3 = 209,
  aux_sym_bitwise_expression_token1 = 210,
  aux_sym_bitwise_expression_token2 = 211,
  aux_sym_bitwise_expression_token3 = 212,
  anon_sym_PLUS = 213,
  anon_sym_DASH = 214,
  anon_sym_SLASH = 215,
  anon_sym_BSLASH = 216,
  anon_sym_PERCENT = 217,
  anon_sym_STAR = 218,
  anon_sym_DOT_DOT = 219,
  aux_sym_expression_with_unary_operator_token1 = 220,
  anon_sym_BANG = 221,
  aux_sym_expression_with_unary_operator_token2 = 222,
  anon_sym_PLUS_PLUS = 223,
  anon_sym_DASH_DASH = 224,
  anon_sym_DOLLAR_LPAREN = 225,
  anon_sym_AT_LPAREN = 226,
  anon_sym_AT_LBRACE = 227,
  anon_sym_DOT2 = 228,
  anon_sym_COLON_COLON = 229,
  anon_sym_RBRACK = 230,
  aux_sym_invokation_foreach_expression_token1 = 231,
  sym__statement_terminator = 232,
  sym_program = 233,
  sym__literal = 234,
  sym_integer_literal = 235,
  sym_string_literal = 236,
  sym_expandable_string_literal = 237,
  sym_expandable_here_string_literal = 238,
  sym_type_name = 239,
  sym_array_type_name = 240,
  sym_generic_type_name = 241,
  sym_assignement_operator = 242,
  sym_file_redirection_operator = 243,
  sym_merging_redirection_operator = 244,
  sym_comparison_operator = 245,
  sym_format_operator = 246,
  sym_variable = 247,
  aux_sym__verbatim_command_argument_chars = 248,
  sym_script_block = 249,
  sym_param_block = 250,
  sym_parameter_list = 251,
  sym_script_parameter = 252,
  sym_script_parameter_default = 253,
  sym_script_block_body = 254,
  sym_named_block_list = 255,
  sym_named_block = 256,
  sym_block_name = 257,
  sym_statement_block = 258,
  sym_statement_list = 259,
  sym__statement = 260,
  sym_empty_statement = 261,
  sym_if_statement = 262,
  sym_elseif_clauses = 263,
  sym_elseif_clause = 264,
  sym_else_clause = 265,
  sym__labeled_statement = 266,
  sym_switch_statement = 267,
  sym_switch_parameters = 268,
  sym_switch_parameter = 269,
  sym_switch_condition = 270,
  sym_switch_filename = 271,
  sym_switch_body = 272,
  sym_switch_clauses = 273,
  sym_switch_clause = 274,
  sym_switch_clause_condition = 275,
  sym_foreach_statement = 276,
  sym_foreach_parameter = 277,
  sym_for_statement = 278,
  sym_for_initializer = 279,
  sym_for_condition = 280,
  sym_for_iterator = 281,
  sym_while_statement = 282,
  sym_while_condition = 283,
  sym_do_statement = 284,
  sym_function_statement = 285,
  sym_function_name = 286,
  sym_function_parameter_declaration = 287,
  sym_flow_control_statement = 288,
  sym_label_expression = 289,
  sym_trap_statement = 290,
  sym_try_statement = 291,
  sym_catch_clauses = 292,
  sym_catch_clause = 293,
  sym_catch_type_list = 294,
  sym_finally_clause = 295,
  sym_data_statement = 296,
  sym_data_name = 297,
  sym_data_commands_allowed = 298,
  sym_data_commands_list = 299,
  sym_data_command = 300,
  sym_inlinescript_statement = 301,
  sym_parallel_statement = 302,
  sym_sequence_statement = 303,
  sym_pipeline = 304,
  sym_pipeline_chain = 305,
  sym_pipeline_chain_tail = 306,
  sym_left_assignment_expression = 307,
  sym_assignment_expression = 308,
  aux_sym__pipeline_tail = 309,
  sym_command = 310,
  sym_command_invokation_operator = 311,
  sym__expandable_string_literal_immediate = 312,
  sym_command_name = 313,
  sym_path_command_name = 314,
  sym_command_name_expr = 315,
  sym_command_elements = 316,
  sym__command_element = 317,
  sym_command_argument_sep = 318,
  sym__command_argument = 319,
  sym_verbatim_command_argument = 320,
  sym_redirections = 321,
  sym_redirection = 322,
  sym_redirected_file_name = 323,
  sym_class_attribute = 324,
  sym_class_property_definition = 325,
  sym_class_method_parameter = 326,
  sym_class_method_parameter_list = 327,
  sym_class_method_definition = 328,
  sym_class_statement = 329,
  sym_enum_statement = 330,
  sym_enum_member = 331,
  sym__expression = 332,
  sym_logical_expression = 333,
  sym_bitwise_expression = 334,
  sym_comparison_expression = 335,
  sym_additive_expression = 336,
  sym_multiplicative_expression = 337,
  sym_format_expression = 338,
  sym_range_expression = 339,
  sym_array_literal_expression = 340,
  sym_unary_expression = 341,
  sym_expression_with_unary_operator = 342,
  sym_pre_increment_expression = 343,
  sym_pre_decrement_expression = 344,
  sym_cast_expression = 345,
  sym__primary_expression = 346,
  sym__value = 347,
  sym_parenthesized_expression = 348,
  sym_sub_expression = 349,
  sym_array_expression = 350,
  sym_script_block_expression = 351,
  sym_hash_literal_expression = 352,
  sym_hash_literal_body = 353,
  sym_hash_entry = 354,
  sym_key_expression = 355,
  sym_post_increment_expression = 356,
  sym_post_decrement_expression = 357,
  sym_member_access = 358,
  sym_member_name = 359,
  sym_element_access = 360,
  sym_invokation_expression = 361,
  sym_invokation_foreach_expression = 362,
  sym_argument_list = 363,
  sym_argument_expression_list = 364,
  sym_argument_expression = 365,
  sym_logical_argument_expression = 366,
  sym_bitwise_argument_expression = 367,
  sym_comparison_argument_expression = 368,
  sym_additive_argument_expression = 369,
  sym_multiplicative_argument_expression = 370,
  sym_format_argument_expression = 371,
  sym_range_argument_expression = 372,
  sym_type_literal = 373,
  sym_type_spec = 374,
  sym_dimension = 375,
  sym_generic_type_arguments = 376,
  sym_attribute_list = 377,
  sym_attribute = 378,
  sym_attribute_name = 379,
  sym_attribute_arguments = 380,
  sym_attribute_argument = 381,
  aux_sym_expandable_string_literal_repeat1 = 382,
  aux_sym_expandable_string_literal_repeat2 = 383,
  aux_sym_expandable_here_string_literal_repeat1 = 384,
  aux_sym_script_block_repeat1 = 385,
  aux_sym_parameter_list_repeat1 = 386,
  aux_sym_named_block_list_repeat1 = 387,
  aux_sym_statement_list_repeat1 = 388,
  aux_sym_elseif_clauses_repeat1 = 389,
  aux_sym_switch_parameters_repeat1 = 390,
  aux_sym_switch_clauses_repeat1 = 391,
  aux_sym_catch_clauses_repeat1 = 392,
  aux_sym_catch_type_list_repeat1 = 393,
  aux_sym_data_commands_list_repeat1 = 394,
  aux_sym_pipeline_repeat1 = 395,
  aux_sym__expandable_string_literal_immediate_repeat1 = 396,
  aux_sym__expandable_string_literal_immediate_repeat2 = 397,
  aux_sym_command_name_repeat1 = 398,
  aux_sym_path_command_name_repeat1 = 399,
  aux_sym_command_elements_repeat1 = 400,
  aux_sym_command_argument_sep_repeat1 = 401,
  aux_sym_redirections_repeat1 = 402,
  aux_sym_class_property_definition_repeat1 = 403,
  aux_sym_class_method_parameter_list_repeat1 = 404,
  aux_sym_class_statement_repeat1 = 405,
  aux_sym_class_statement_repeat2 = 406,
  aux_sym_enum_statement_repeat1 = 407,
  aux_sym_array_literal_expression_repeat1 = 408,
  aux_sym_hash_literal_body_repeat1 = 409,
  aux_sym_argument_expression_list_repeat1 = 410,
  aux_sym_dimension_repeat1 = 411,
  aux_sym_generic_type_arguments_repeat1 = 412,
  aux_sym_attribute_list_repeat1 = 413,
  aux_sym_attribute_arguments_repeat1 = 414,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_comment] = "comment",
  [sym_decimal_integer_literal] = "decimal_integer_literal",
  [sym_hexadecimal_integer_literal] = "hexadecimal_integer_literal",
  [sym_real_literal] = "real_literal",
  [aux_sym_expandable_string_literal_token1] = "expandable_string_literal_token1",
  [aux_sym_expandable_string_literal_token2] = "expandable_string_literal_token2",
  [aux_sym_expandable_string_literal_token3] = "expandable_string_literal_token3",
  [aux_sym_expandable_string_literal_token4] = "expandable_string_literal_token4",
  [anon_sym_DQUOTE_DQUOTE] = "\"\"",
  [anon_sym_DOLLAR] = "$",
  [aux_sym_expandable_string_literal_token5] = "expandable_string_literal_token5",
  [aux_sym_expandable_here_string_literal_token1] = "expandable_here_string_literal_token1",
  [aux_sym_expandable_here_string_literal_token2] = "expandable_here_string_literal_token2",
  [aux_sym_expandable_here_string_literal_token3] = "expandable_here_string_literal_token3",
  [aux_sym_expandable_here_string_literal_token4] = "expandable_here_string_literal_token4",
  [aux_sym_expandable_here_string_literal_token5] = "expandable_here_string_literal_token5",
  [sym_verbatim_string_characters] = "verbatim_string_characters",
  [sym_verbatim_here_string_characters] = "verbatim_here_string_characters",
  [sym_simple_name] = "simple_name",
  [sym_type_identifier] = "type_identifier",
  [anon_sym_DOT] = ".",
  [anon_sym_LBRACK] = "[",
  [anon_sym_EQ] = "=",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_2_GT] = "2>",
  [anon_sym_2_GT_GT] = "2>>",
  [anon_sym_3_GT] = "3>",
  [anon_sym_3_GT_GT] = "3>>",
  [anon_sym_4_GT] = "4>",
  [anon_sym_4_GT_GT] = "4>>",
  [anon_sym_5_GT] = "5>",
  [anon_sym_5_GT_GT] = "5>>",
  [anon_sym_6_GT] = "6>",
  [anon_sym_6_GT_GT] = "6>>",
  [anon_sym_STAR_GT] = "*>",
  [anon_sym_STAR_GT_GT] = "*>>",
  [anon_sym_LT] = "<",
  [anon_sym_STAR_GT_AMP1] = "*>&1",
  [anon_sym_2_GT_AMP1] = "2>&1",
  [anon_sym_3_GT_AMP1] = "3>&1",
  [anon_sym_4_GT_AMP1] = "4>&1",
  [anon_sym_5_GT_AMP1] = "5>&1",
  [anon_sym_6_GT_AMP1] = "6>&1",
  [anon_sym_STAR_GT_AMP2] = "*>&2",
  [anon_sym_1_GT_AMP2] = "1>&2",
  [anon_sym_3_GT_AMP2] = "3>&2",
  [anon_sym_4_GT_AMP2] = "4>&2",
  [anon_sym_5_GT_AMP2] = "5>&2",
  [anon_sym_6_GT_AMP2] = "6>&2",
  [aux_sym_comparison_operator_token1] = "-as",
  [aux_sym_comparison_operator_token2] = "-ccontains",
  [aux_sym_comparison_operator_token3] = "-ceq",
  [aux_sym_comparison_operator_token4] = "-cge",
  [aux_sym_comparison_operator_token5] = "-cgt",
  [aux_sym_comparison_operator_token6] = "-cle",
  [aux_sym_comparison_operator_token7] = "-clike",
  [aux_sym_comparison_operator_token8] = "-clt",
  [aux_sym_comparison_operator_token9] = "-cmatch",
  [aux_sym_comparison_operator_token10] = "-cne",
  [aux_sym_comparison_operator_token11] = "-cnotcontains",
  [aux_sym_comparison_operator_token12] = "-cnotlike",
  [aux_sym_comparison_operator_token13] = "-cnotmatch",
  [aux_sym_comparison_operator_token14] = "-contains",
  [aux_sym_comparison_operator_token15] = "-creplace",
  [aux_sym_comparison_operator_token16] = "-csplit",
  [aux_sym_comparison_operator_token17] = "-eq",
  [aux_sym_comparison_operator_token18] = "-ge",
  [aux_sym_comparison_operator_token19] = "-gt",
  [aux_sym_comparison_operator_token20] = "-icontains",
  [aux_sym_comparison_operator_token21] = "-ieq",
  [aux_sym_comparison_operator_token22] = "-ige",
  [aux_sym_comparison_operator_token23] = "-igt",
  [aux_sym_comparison_operator_token24] = "-ile",
  [aux_sym_comparison_operator_token25] = "-ilike",
  [aux_sym_comparison_operator_token26] = "-ilt",
  [aux_sym_comparison_operator_token27] = "-imatch",
  [aux_sym_comparison_operator_token28] = "-in",
  [aux_sym_comparison_operator_token29] = "-ine",
  [aux_sym_comparison_operator_token30] = "-inotcontains",
  [aux_sym_comparison_operator_token31] = "-inotlike",
  [aux_sym_comparison_operator_token32] = "-inotmatch",
  [aux_sym_comparison_operator_token33] = "-ireplace",
  [aux_sym_comparison_operator_token34] = "-is",
  [aux_sym_comparison_operator_token35] = "-isnot",
  [aux_sym_comparison_operator_token36] = "-isplit",
  [aux_sym_comparison_operator_token37] = "-join",
  [aux_sym_comparison_operator_token38] = "-le",
  [aux_sym_comparison_operator_token39] = "-like",
  [aux_sym_comparison_operator_token40] = "-lt",
  [aux_sym_comparison_operator_token41] = "-match",
  [aux_sym_comparison_operator_token42] = "-ne",
  [aux_sym_comparison_operator_token43] = "-notcontains",
  [aux_sym_comparison_operator_token44] = "-notin",
  [aux_sym_comparison_operator_token45] = "-notlike",
  [aux_sym_comparison_operator_token46] = "-notmatch",
  [aux_sym_comparison_operator_token47] = "-replace",
  [aux_sym_comparison_operator_token48] = "-shl",
  [aux_sym_comparison_operator_token49] = "-shr",
  [aux_sym_comparison_operator_token50] = "-split",
  [aux_sym_format_operator_token1] = "-f",
  [anon_sym_DOLLAR_DOLLAR] = "$$",
  [anon_sym_DOLLAR_CARET] = "$^",
  [anon_sym_DOLLAR_QMARK] = "$\?",
  [anon_sym_DOLLAR_] = "$_",
  [aux_sym_variable_token1] = "variable_token1",
  [aux_sym_variable_token2] = "variable_token2",
  [sym_braced_variable] = "braced_variable",
  [sym_generic_token] = "generic_token",
  [sym__command_token] = "_command_token",
  [sym_command_parameter] = "command_parameter",
  [aux_sym__verbatim_command_argument_chars_token1] = "_verbatim_command_argument_chars_token1",
  [aux_sym__verbatim_command_argument_chars_token2] = "_verbatim_command_argument_chars_token2",
  [aux_sym__verbatim_command_argument_chars_token3] = "_verbatim_command_argument_chars_token3",
  [anon_sym_SEMI] = ";",
  [aux_sym_param_block_token1] = "param",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COMMA] = ",",
  [aux_sym_block_name_token1] = "dynamicparam",
  [aux_sym_block_name_token2] = "begin",
  [aux_sym_block_name_token3] = "process",
  [aux_sym_block_name_token4] = "end",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [aux_sym_if_statement_token1] = "if",
  [aux_sym_elseif_clause_token1] = "elseif",
  [aux_sym_else_clause_token1] = "else",
  [aux_sym_switch_statement_token1] = "switch",
  [aux_sym_switch_parameter_token1] = "-regex",
  [aux_sym_switch_parameter_token2] = "-wildcard",
  [aux_sym_switch_parameter_token3] = "-exact",
  [aux_sym_switch_parameter_token4] = "-casesensitive",
  [aux_sym_switch_parameter_token5] = "-parallel",
  [aux_sym_switch_condition_token1] = "-file",
  [aux_sym_foreach_statement_token1] = "foreach",
  [aux_sym_foreach_statement_token2] = "in",
  [aux_sym_for_statement_token1] = "for",
  [aux_sym_for_statement_token2] = "for_statement_token2",
  [aux_sym_while_statement_token1] = "while",
  [aux_sym_do_statement_token1] = "do",
  [aux_sym_do_statement_token2] = "until",
  [aux_sym_function_statement_token1] = "function",
  [aux_sym_function_statement_token2] = "filter",
  [aux_sym_function_statement_token3] = "workflow",
  [aux_sym_flow_control_statement_token1] = "break",
  [aux_sym_flow_control_statement_token2] = "continue",
  [aux_sym_flow_control_statement_token3] = "throw",
  [aux_sym_flow_control_statement_token4] = "return",
  [aux_sym_flow_control_statement_token5] = "exit",
  [sym_label] = "label",
  [aux_sym_trap_statement_token1] = "trap",
  [aux_sym_try_statement_token1] = "try",
  [aux_sym_catch_clause_token1] = "catch",
  [aux_sym_finally_clause_token1] = "finally",
  [aux_sym_data_statement_token1] = "data",
  [aux_sym_data_commands_allowed_token1] = "-supportedcommand",
  [aux_sym_inlinescript_statement_token1] = "inlinescript",
  [aux_sym_parallel_statement_token1] = "parallel",
  [aux_sym_sequence_statement_token1] = "sequence",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_PIPE] = "|",
  [anon_sym_AMP] = "&",
  [aux_sym__expandable_string_literal_immediate_token1] = "_expandable_string_literal_immediate_token1",
  [aux_sym__expandable_string_literal_immediate_token2] = "_expandable_string_literal_immediate_token2",
  [aux_sym__expandable_string_literal_immediate_token3] = "_expandable_string_literal_immediate_token3",
  [anon_sym_DQUOTE_DQUOTE2] = "\"\"",
  [anon_sym_DOLLAR2] = "$",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_command_name_token1] = "command_name_token1",
  [aux_sym_command_name_token2] = "command_name_token2",
  [aux_sym_command_name_token3] = "command_name_token3",
  [aux_sym_command_name_token4] = "command_name_token4",
  [aux_sym_command_name_token5] = "command_name_token5",
  [aux_sym_command_name_token6] = "command_name_token6",
  [aux_sym_command_name_token7] = "command_name_token7",
  [aux_sym_command_name_token8] = "command_name_token8",
  [aux_sym_command_name_token9] = "command_name_token9",
  [aux_sym_command_name_token10] = "command_name_token10",
  [aux_sym_command_name_token11] = "command_name_token11",
  [aux_sym_command_name_token12] = "command_name_token12",
  [aux_sym_command_name_token13] = "command_name_token13",
  [aux_sym_command_name_token14] = "command_name_token14",
  [aux_sym_command_name_token15] = "command_name_token15",
  [aux_sym_command_name_token16] = "command_name_token16",
  [aux_sym_command_name_token17] = "command_name_token17",
  [aux_sym_command_name_token18] = "command_name_token18",
  [aux_sym_command_name_token19] = "command_name_token19",
  [anon_sym_DQUOTE2] = "\"",
  [anon_sym_SQUOTE_SQUOTE] = "''",
  [sym_path_command_name_token] = "path_command_name_token",
  [sym_stop_parsing] = "stop_parsing",
  [anon_sym_SPACE] = " ",
  [anon_sym_COLON] = ":",
  [anon_sym_DASH_DASH_PERCENT] = "--%",
  [aux_sym_class_attribute_token1] = "class_attribute_token1",
  [aux_sym_class_attribute_token2] = "class_attribute_token2",
  [aux_sym_class_statement_token1] = "class_statement_token1",
  [aux_sym_enum_statement_token1] = "enum_statement_token1",
  [aux_sym_logical_expression_token1] = "-and",
  [aux_sym_logical_expression_token2] = "-or",
  [aux_sym_logical_expression_token3] = "-xor",
  [aux_sym_bitwise_expression_token1] = "-band",
  [aux_sym_bitwise_expression_token2] = "-bor",
  [aux_sym_bitwise_expression_token3] = "-bxor",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_SLASH] = "/",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_PERCENT] = "%",
  [anon_sym_STAR] = "*",
  [anon_sym_DOT_DOT] = "..",
  [aux_sym_expression_with_unary_operator_token1] = "-not",
  [anon_sym_BANG] = "!",
  [aux_sym_expression_with_unary_operator_token2] = "-bnot",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_DOLLAR_LPAREN] = "$(",
  [anon_sym_AT_LPAREN] = "@(",
  [anon_sym_AT_LBRACE] = "@{",
  [anon_sym_DOT2] = ".",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_RBRACK] = "]",
  [aux_sym_invokation_foreach_expression_token1] = "invokation_foreach_expression_token1",
  [sym__statement_terminator] = "_statement_terminator",
  [sym_program] = "program",
  [sym__literal] = "_literal",
  [sym_integer_literal] = "integer_literal",
  [sym_string_literal] = "string_literal",
  [sym_expandable_string_literal] = "expandable_string_literal",
  [sym_expandable_here_string_literal] = "expandable_here_string_literal",
  [sym_type_name] = "type_name",
  [sym_array_type_name] = "array_type_name",
  [sym_generic_type_name] = "generic_type_name",
  [sym_assignement_operator] = "assignement_operator",
  [sym_file_redirection_operator] = "file_redirection_operator",
  [sym_merging_redirection_operator] = "merging_redirection_operator",
  [sym_comparison_operator] = "comparison_operator",
  [sym_format_operator] = "format_operator",
  [sym_variable] = "variable",
  [aux_sym__verbatim_command_argument_chars] = "_verbatim_command_argument_chars",
  [sym_script_block] = "script_block",
  [sym_param_block] = "param_block",
  [sym_parameter_list] = "parameter_list",
  [sym_script_parameter] = "script_parameter",
  [sym_script_parameter_default] = "script_parameter_default",
  [sym_script_block_body] = "script_block_body",
  [sym_named_block_list] = "named_block_list",
  [sym_named_block] = "named_block",
  [sym_block_name] = "block_name",
  [sym_statement_block] = "statement_block",
  [sym_statement_list] = "statement_list",
  [sym__statement] = "_statement",
  [sym_empty_statement] = "empty_statement",
  [sym_if_statement] = "if_statement",
  [sym_elseif_clauses] = "elseif_clauses",
  [sym_elseif_clause] = "elseif_clause",
  [sym_else_clause] = "else_clause",
  [sym__labeled_statement] = "_labeled_statement",
  [sym_switch_statement] = "switch_statement",
  [sym_switch_parameters] = "switch_parameters",
  [sym_switch_parameter] = "switch_parameter",
  [sym_switch_condition] = "switch_condition",
  [sym_switch_filename] = "switch_filename",
  [sym_switch_body] = "switch_body",
  [sym_switch_clauses] = "switch_clauses",
  [sym_switch_clause] = "switch_clause",
  [sym_switch_clause_condition] = "switch_clause_condition",
  [sym_foreach_statement] = "foreach_statement",
  [sym_foreach_parameter] = "foreach_parameter",
  [sym_for_statement] = "for_statement",
  [sym_for_initializer] = "for_initializer",
  [sym_for_condition] = "for_condition",
  [sym_for_iterator] = "for_iterator",
  [sym_while_statement] = "while_statement",
  [sym_while_condition] = "while_condition",
  [sym_do_statement] = "do_statement",
  [sym_function_statement] = "function_statement",
  [sym_function_name] = "function_name",
  [sym_function_parameter_declaration] = "function_parameter_declaration",
  [sym_flow_control_statement] = "flow_control_statement",
  [sym_label_expression] = "label_expression",
  [sym_trap_statement] = "trap_statement",
  [sym_try_statement] = "try_statement",
  [sym_catch_clauses] = "catch_clauses",
  [sym_catch_clause] = "catch_clause",
  [sym_catch_type_list] = "catch_type_list",
  [sym_finally_clause] = "finally_clause",
  [sym_data_statement] = "data_statement",
  [sym_data_name] = "data_name",
  [sym_data_commands_allowed] = "data_commands_allowed",
  [sym_data_commands_list] = "data_commands_list",
  [sym_data_command] = "data_command",
  [sym_inlinescript_statement] = "inlinescript_statement",
  [sym_parallel_statement] = "parallel_statement",
  [sym_sequence_statement] = "sequence_statement",
  [sym_pipeline] = "pipeline",
  [sym_pipeline_chain] = "pipeline_chain",
  [sym_pipeline_chain_tail] = "pipeline_chain_tail",
  [sym_left_assignment_expression] = "left_assignment_expression",
  [sym_assignment_expression] = "assignment_expression",
  [aux_sym__pipeline_tail] = "_pipeline_tail",
  [sym_command] = "command",
  [sym_command_invokation_operator] = "command_invokation_operator",
  [sym__expandable_string_literal_immediate] = "_expandable_string_literal_immediate",
  [sym_command_name] = "command_name",
  [sym_path_command_name] = "path_command_name",
  [sym_command_name_expr] = "command_name_expr",
  [sym_command_elements] = "command_elements",
  [sym__command_element] = "_command_element",
  [sym_command_argument_sep] = "command_argument_sep",
  [sym__command_argument] = "_command_argument",
  [sym_verbatim_command_argument] = "verbatim_command_argument",
  [sym_redirections] = "redirections",
  [sym_redirection] = "redirection",
  [sym_redirected_file_name] = "redirected_file_name",
  [sym_class_attribute] = "class_attribute",
  [sym_class_property_definition] = "class_property_definition",
  [sym_class_method_parameter] = "class_method_parameter",
  [sym_class_method_parameter_list] = "class_method_parameter_list",
  [sym_class_method_definition] = "class_method_definition",
  [sym_class_statement] = "class_statement",
  [sym_enum_statement] = "enum_statement",
  [sym_enum_member] = "enum_member",
  [sym__expression] = "_expression",
  [sym_logical_expression] = "logical_expression",
  [sym_bitwise_expression] = "bitwise_expression",
  [sym_comparison_expression] = "comparison_expression",
  [sym_additive_expression] = "additive_expression",
  [sym_multiplicative_expression] = "multiplicative_expression",
  [sym_format_expression] = "format_expression",
  [sym_range_expression] = "range_expression",
  [sym_array_literal_expression] = "array_literal_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_expression_with_unary_operator] = "expression_with_unary_operator",
  [sym_pre_increment_expression] = "pre_increment_expression",
  [sym_pre_decrement_expression] = "pre_decrement_expression",
  [sym_cast_expression] = "cast_expression",
  [sym__primary_expression] = "_primary_expression",
  [sym__value] = "_value",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_sub_expression] = "sub_expression",
  [sym_array_expression] = "array_expression",
  [sym_script_block_expression] = "script_block_expression",
  [sym_hash_literal_expression] = "hash_literal_expression",
  [sym_hash_literal_body] = "hash_literal_body",
  [sym_hash_entry] = "hash_entry",
  [sym_key_expression] = "key_expression",
  [sym_post_increment_expression] = "post_increment_expression",
  [sym_post_decrement_expression] = "post_decrement_expression",
  [sym_member_access] = "member_access",
  [sym_member_name] = "member_name",
  [sym_element_access] = "element_access",
  [sym_invokation_expression] = "invokation_expression",
  [sym_invokation_foreach_expression] = "invokation_foreach_expression",
  [sym_argument_list] = "argument_list",
  [sym_argument_expression_list] = "argument_expression_list",
  [sym_argument_expression] = "argument_expression",
  [sym_logical_argument_expression] = "logical_argument_expression",
  [sym_bitwise_argument_expression] = "bitwise_argument_expression",
  [sym_comparison_argument_expression] = "comparison_argument_expression",
  [sym_additive_argument_expression] = "additive_argument_expression",
  [sym_multiplicative_argument_expression] = "multiplicative_argument_expression",
  [sym_format_argument_expression] = "format_argument_expression",
  [sym_range_argument_expression] = "range_argument_expression",
  [sym_type_literal] = "type_literal",
  [sym_type_spec] = "type_spec",
  [sym_dimension] = "dimension",
  [sym_generic_type_arguments] = "generic_type_arguments",
  [sym_attribute_list] = "attribute_list",
  [sym_attribute] = "attribute",
  [sym_attribute_name] = "attribute_name",
  [sym_attribute_arguments] = "attribute_arguments",
  [sym_attribute_argument] = "attribute_argument",
  [aux_sym_expandable_string_literal_repeat1] = "expandable_string_literal_repeat1",
  [aux_sym_expandable_string_literal_repeat2] = "expandable_string_literal_repeat2",
  [aux_sym_expandable_here_string_literal_repeat1] = "expandable_here_string_literal_repeat1",
  [aux_sym_script_block_repeat1] = "script_block_repeat1",
  [aux_sym_parameter_list_repeat1] = "parameter_list_repeat1",
  [aux_sym_named_block_list_repeat1] = "named_block_list_repeat1",
  [aux_sym_statement_list_repeat1] = "statement_list_repeat1",
  [aux_sym_elseif_clauses_repeat1] = "elseif_clauses_repeat1",
  [aux_sym_switch_parameters_repeat1] = "switch_parameters_repeat1",
  [aux_sym_switch_clauses_repeat1] = "switch_clauses_repeat1",
  [aux_sym_catch_clauses_repeat1] = "catch_clauses_repeat1",
  [aux_sym_catch_type_list_repeat1] = "catch_type_list_repeat1",
  [aux_sym_data_commands_list_repeat1] = "data_commands_list_repeat1",
  [aux_sym_pipeline_repeat1] = "pipeline_repeat1",
  [aux_sym__expandable_string_literal_immediate_repeat1] = "_expandable_string_literal_immediate_repeat1",
  [aux_sym__expandable_string_literal_immediate_repeat2] = "_expandable_string_literal_immediate_repeat2",
  [aux_sym_command_name_repeat1] = "command_name_repeat1",
  [aux_sym_path_command_name_repeat1] = "path_command_name_repeat1",
  [aux_sym_command_elements_repeat1] = "command_elements_repeat1",
  [aux_sym_command_argument_sep_repeat1] = "command_argument_sep_repeat1",
  [aux_sym_redirections_repeat1] = "redirections_repeat1",
  [aux_sym_class_property_definition_repeat1] = "class_property_definition_repeat1",
  [aux_sym_class_method_parameter_list_repeat1] = "class_method_parameter_list_repeat1",
  [aux_sym_class_statement_repeat1] = "class_statement_repeat1",
  [aux_sym_class_statement_repeat2] = "class_statement_repeat2",
  [aux_sym_enum_statement_repeat1] = "enum_statement_repeat1",
  [aux_sym_array_literal_expression_repeat1] = "array_literal_expression_repeat1",
  [aux_sym_hash_literal_body_repeat1] = "hash_literal_body_repeat1",
  [aux_sym_argument_expression_list_repeat1] = "argument_expression_list_repeat1",
  [aux_sym_dimension_repeat1] = "dimension_repeat1",
  [aux_sym_generic_type_arguments_repeat1] = "generic_type_arguments_repeat1",
  [aux_sym_attribute_list_repeat1] = "attribute_list_repeat1",
  [aux_sym_attribute_arguments_repeat1] = "attribute_arguments_repeat1",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_comment] = sym_comment,
  [sym_decimal_integer_literal] = sym_decimal_integer_literal,
  [sym_hexadecimal_integer_literal] = sym_hexadecimal_integer_literal,
  [sym_real_literal] = sym_real_literal,
  [aux_sym_expandable_string_literal_token1] = aux_sym_expandable_string_literal_token1,
  [aux_sym_expandable_string_literal_token2] = aux_sym_expandable_string_literal_token2,
  [aux_sym_expandable_string_literal_token3] = aux_sym_expandable_string_literal_token3,
  [aux_sym_expandable_string_literal_token4] = aux_sym_expandable_string_literal_token4,
  [anon_sym_DQUOTE_DQUOTE] = anon_sym_DQUOTE_DQUOTE,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [aux_sym_expandable_string_literal_token5] = aux_sym_expandable_string_literal_token5,
  [aux_sym_expandable_here_string_literal_token1] = aux_sym_expandable_here_string_literal_token1,
  [aux_sym_expandable_here_string_literal_token2] = aux_sym_expandable_here_string_literal_token2,
  [aux_sym_expandable_here_string_literal_token3] = aux_sym_expandable_here_string_literal_token3,
  [aux_sym_expandable_here_string_literal_token4] = aux_sym_expandable_here_string_literal_token4,
  [aux_sym_expandable_here_string_literal_token5] = aux_sym_expandable_here_string_literal_token5,
  [sym_verbatim_string_characters] = sym_verbatim_string_characters,
  [sym_verbatim_here_string_characters] = sym_verbatim_here_string_characters,
  [sym_simple_name] = sym_simple_name,
  [sym_type_identifier] = sym_type_identifier,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_2_GT] = anon_sym_2_GT,
  [anon_sym_2_GT_GT] = anon_sym_2_GT_GT,
  [anon_sym_3_GT] = anon_sym_3_GT,
  [anon_sym_3_GT_GT] = anon_sym_3_GT_GT,
  [anon_sym_4_GT] = anon_sym_4_GT,
  [anon_sym_4_GT_GT] = anon_sym_4_GT_GT,
  [anon_sym_5_GT] = anon_sym_5_GT,
  [anon_sym_5_GT_GT] = anon_sym_5_GT_GT,
  [anon_sym_6_GT] = anon_sym_6_GT,
  [anon_sym_6_GT_GT] = anon_sym_6_GT_GT,
  [anon_sym_STAR_GT] = anon_sym_STAR_GT,
  [anon_sym_STAR_GT_GT] = anon_sym_STAR_GT_GT,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_STAR_GT_AMP1] = anon_sym_STAR_GT_AMP1,
  [anon_sym_2_GT_AMP1] = anon_sym_2_GT_AMP1,
  [anon_sym_3_GT_AMP1] = anon_sym_3_GT_AMP1,
  [anon_sym_4_GT_AMP1] = anon_sym_4_GT_AMP1,
  [anon_sym_5_GT_AMP1] = anon_sym_5_GT_AMP1,
  [anon_sym_6_GT_AMP1] = anon_sym_6_GT_AMP1,
  [anon_sym_STAR_GT_AMP2] = anon_sym_STAR_GT_AMP2,
  [anon_sym_1_GT_AMP2] = anon_sym_1_GT_AMP2,
  [anon_sym_3_GT_AMP2] = anon_sym_3_GT_AMP2,
  [anon_sym_4_GT_AMP2] = anon_sym_4_GT_AMP2,
  [anon_sym_5_GT_AMP2] = anon_sym_5_GT_AMP2,
  [anon_sym_6_GT_AMP2] = anon_sym_6_GT_AMP2,
  [aux_sym_comparison_operator_token1] = aux_sym_comparison_operator_token1,
  [aux_sym_comparison_operator_token2] = aux_sym_comparison_operator_token2,
  [aux_sym_comparison_operator_token3] = aux_sym_comparison_operator_token3,
  [aux_sym_comparison_operator_token4] = aux_sym_comparison_operator_token4,
  [aux_sym_comparison_operator_token5] = aux_sym_comparison_operator_token5,
  [aux_sym_comparison_operator_token6] = aux_sym_comparison_operator_token6,
  [aux_sym_comparison_operator_token7] = aux_sym_comparison_operator_token7,
  [aux_sym_comparison_operator_token8] = aux_sym_comparison_operator_token8,
  [aux_sym_comparison_operator_token9] = aux_sym_comparison_operator_token9,
  [aux_sym_comparison_operator_token10] = aux_sym_comparison_operator_token10,
  [aux_sym_comparison_operator_token11] = aux_sym_comparison_operator_token11,
  [aux_sym_comparison_operator_token12] = aux_sym_comparison_operator_token12,
  [aux_sym_comparison_operator_token13] = aux_sym_comparison_operator_token13,
  [aux_sym_comparison_operator_token14] = aux_sym_comparison_operator_token14,
  [aux_sym_comparison_operator_token15] = aux_sym_comparison_operator_token15,
  [aux_sym_comparison_operator_token16] = aux_sym_comparison_operator_token16,
  [aux_sym_comparison_operator_token17] = aux_sym_comparison_operator_token17,
  [aux_sym_comparison_operator_token18] = aux_sym_comparison_operator_token18,
  [aux_sym_comparison_operator_token19] = aux_sym_comparison_operator_token19,
  [aux_sym_comparison_operator_token20] = aux_sym_comparison_operator_token20,
  [aux_sym_comparison_operator_token21] = aux_sym_comparison_operator_token21,
  [aux_sym_comparison_operator_token22] = aux_sym_comparison_operator_token22,
  [aux_sym_comparison_operator_token23] = aux_sym_comparison_operator_token23,
  [aux_sym_comparison_operator_token24] = aux_sym_comparison_operator_token24,
  [aux_sym_comparison_operator_token25] = aux_sym_comparison_operator_token25,
  [aux_sym_comparison_operator_token26] = aux_sym_comparison_operator_token26,
  [aux_sym_comparison_operator_token27] = aux_sym_comparison_operator_token27,
  [aux_sym_comparison_operator_token28] = aux_sym_comparison_operator_token28,
  [aux_sym_comparison_operator_token29] = aux_sym_comparison_operator_token29,
  [aux_sym_comparison_operator_token30] = aux_sym_comparison_operator_token30,
  [aux_sym_comparison_operator_token31] = aux_sym_comparison_operator_token31,
  [aux_sym_comparison_operator_token32] = aux_sym_comparison_operator_token32,
  [aux_sym_comparison_operator_token33] = aux_sym_comparison_operator_token33,
  [aux_sym_comparison_operator_token34] = aux_sym_comparison_operator_token34,
  [aux_sym_comparison_operator_token35] = aux_sym_comparison_operator_token35,
  [aux_sym_comparison_operator_token36] = aux_sym_comparison_operator_token36,
  [aux_sym_comparison_operator_token37] = aux_sym_comparison_operator_token37,
  [aux_sym_comparison_operator_token38] = aux_sym_comparison_operator_token38,
  [aux_sym_comparison_operator_token39] = aux_sym_comparison_operator_token39,
  [aux_sym_comparison_operator_token40] = aux_sym_comparison_operator_token40,
  [aux_sym_comparison_operator_token41] = aux_sym_comparison_operator_token41,
  [aux_sym_comparison_operator_token42] = aux_sym_comparison_operator_token42,
  [aux_sym_comparison_operator_token43] = aux_sym_comparison_operator_token43,
  [aux_sym_comparison_operator_token44] = aux_sym_comparison_operator_token44,
  [aux_sym_comparison_operator_token45] = aux_sym_comparison_operator_token45,
  [aux_sym_comparison_operator_token46] = aux_sym_comparison_operator_token46,
  [aux_sym_comparison_operator_token47] = aux_sym_comparison_operator_token47,
  [aux_sym_comparison_operator_token48] = aux_sym_comparison_operator_token48,
  [aux_sym_comparison_operator_token49] = aux_sym_comparison_operator_token49,
  [aux_sym_comparison_operator_token50] = aux_sym_comparison_operator_token50,
  [aux_sym_format_operator_token1] = aux_sym_format_operator_token1,
  [anon_sym_DOLLAR_DOLLAR] = anon_sym_DOLLAR_DOLLAR,
  [anon_sym_DOLLAR_CARET] = anon_sym_DOLLAR_CARET,
  [anon_sym_DOLLAR_QMARK] = anon_sym_DOLLAR_QMARK,
  [anon_sym_DOLLAR_] = anon_sym_DOLLAR_,
  [aux_sym_variable_token1] = aux_sym_variable_token1,
  [aux_sym_variable_token2] = aux_sym_variable_token2,
  [sym_braced_variable] = sym_braced_variable,
  [sym_generic_token] = sym_generic_token,
  [sym__command_token] = sym__command_token,
  [sym_command_parameter] = sym_command_parameter,
  [aux_sym__verbatim_command_argument_chars_token1] = aux_sym__verbatim_command_argument_chars_token1,
  [aux_sym__verbatim_command_argument_chars_token2] = aux_sym__verbatim_command_argument_chars_token2,
  [aux_sym__verbatim_command_argument_chars_token3] = aux_sym__verbatim_command_argument_chars_token3,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [aux_sym_param_block_token1] = aux_sym_param_block_token1,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [aux_sym_block_name_token1] = aux_sym_block_name_token1,
  [aux_sym_block_name_token2] = aux_sym_block_name_token2,
  [aux_sym_block_name_token3] = aux_sym_block_name_token3,
  [aux_sym_block_name_token4] = aux_sym_block_name_token4,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_if_statement_token1] = aux_sym_if_statement_token1,
  [aux_sym_elseif_clause_token1] = aux_sym_elseif_clause_token1,
  [aux_sym_else_clause_token1] = aux_sym_else_clause_token1,
  [aux_sym_switch_statement_token1] = aux_sym_switch_statement_token1,
  [aux_sym_switch_parameter_token1] = aux_sym_switch_parameter_token1,
  [aux_sym_switch_parameter_token2] = aux_sym_switch_parameter_token2,
  [aux_sym_switch_parameter_token3] = aux_sym_switch_parameter_token3,
  [aux_sym_switch_parameter_token4] = aux_sym_switch_parameter_token4,
  [aux_sym_switch_parameter_token5] = aux_sym_switch_parameter_token5,
  [aux_sym_switch_condition_token1] = aux_sym_switch_condition_token1,
  [aux_sym_foreach_statement_token1] = aux_sym_foreach_statement_token1,
  [aux_sym_foreach_statement_token2] = aux_sym_foreach_statement_token2,
  [aux_sym_for_statement_token1] = aux_sym_for_statement_token1,
  [aux_sym_for_statement_token2] = aux_sym_for_statement_token2,
  [aux_sym_while_statement_token1] = aux_sym_while_statement_token1,
  [aux_sym_do_statement_token1] = aux_sym_do_statement_token1,
  [aux_sym_do_statement_token2] = aux_sym_do_statement_token2,
  [aux_sym_function_statement_token1] = aux_sym_function_statement_token1,
  [aux_sym_function_statement_token2] = aux_sym_function_statement_token2,
  [aux_sym_function_statement_token3] = aux_sym_function_statement_token3,
  [aux_sym_flow_control_statement_token1] = aux_sym_flow_control_statement_token1,
  [aux_sym_flow_control_statement_token2] = aux_sym_flow_control_statement_token2,
  [aux_sym_flow_control_statement_token3] = aux_sym_flow_control_statement_token3,
  [aux_sym_flow_control_statement_token4] = aux_sym_flow_control_statement_token4,
  [aux_sym_flow_control_statement_token5] = aux_sym_flow_control_statement_token5,
  [sym_label] = sym_label,
  [aux_sym_trap_statement_token1] = aux_sym_trap_statement_token1,
  [aux_sym_try_statement_token1] = aux_sym_try_statement_token1,
  [aux_sym_catch_clause_token1] = aux_sym_catch_clause_token1,
  [aux_sym_finally_clause_token1] = aux_sym_finally_clause_token1,
  [aux_sym_data_statement_token1] = aux_sym_data_statement_token1,
  [aux_sym_data_commands_allowed_token1] = aux_sym_data_commands_allowed_token1,
  [aux_sym_inlinescript_statement_token1] = aux_sym_inlinescript_statement_token1,
  [aux_sym_parallel_statement_token1] = aux_sym_parallel_statement_token1,
  [aux_sym_sequence_statement_token1] = aux_sym_sequence_statement_token1,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_AMP] = anon_sym_AMP,
  [aux_sym__expandable_string_literal_immediate_token1] = aux_sym__expandable_string_literal_immediate_token1,
  [aux_sym__expandable_string_literal_immediate_token2] = aux_sym__expandable_string_literal_immediate_token2,
  [aux_sym__expandable_string_literal_immediate_token3] = aux_sym__expandable_string_literal_immediate_token3,
  [anon_sym_DQUOTE_DQUOTE2] = anon_sym_DQUOTE_DQUOTE,
  [anon_sym_DOLLAR2] = anon_sym_DOLLAR,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_command_name_token1] = aux_sym_command_name_token1,
  [aux_sym_command_name_token2] = aux_sym_command_name_token2,
  [aux_sym_command_name_token3] = aux_sym_command_name_token3,
  [aux_sym_command_name_token4] = aux_sym_command_name_token4,
  [aux_sym_command_name_token5] = aux_sym_command_name_token5,
  [aux_sym_command_name_token6] = aux_sym_command_name_token6,
  [aux_sym_command_name_token7] = aux_sym_command_name_token7,
  [aux_sym_command_name_token8] = aux_sym_command_name_token8,
  [aux_sym_command_name_token9] = aux_sym_command_name_token9,
  [aux_sym_command_name_token10] = aux_sym_command_name_token10,
  [aux_sym_command_name_token11] = aux_sym_command_name_token11,
  [aux_sym_command_name_token12] = aux_sym_command_name_token12,
  [aux_sym_command_name_token13] = aux_sym_command_name_token13,
  [aux_sym_command_name_token14] = aux_sym_command_name_token14,
  [aux_sym_command_name_token15] = aux_sym_command_name_token15,
  [aux_sym_command_name_token16] = aux_sym_command_name_token16,
  [aux_sym_command_name_token17] = aux_sym_command_name_token17,
  [aux_sym_command_name_token18] = aux_sym_command_name_token18,
  [aux_sym_command_name_token19] = aux_sym_command_name_token19,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [anon_sym_SQUOTE_SQUOTE] = anon_sym_SQUOTE_SQUOTE,
  [sym_path_command_name_token] = sym_path_command_name_token,
  [sym_stop_parsing] = sym_stop_parsing,
  [anon_sym_SPACE] = anon_sym_SPACE,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_DASH_DASH_PERCENT] = anon_sym_DASH_DASH_PERCENT,
  [aux_sym_class_attribute_token1] = aux_sym_class_attribute_token1,
  [aux_sym_class_attribute_token2] = aux_sym_class_attribute_token2,
  [aux_sym_class_statement_token1] = aux_sym_class_statement_token1,
  [aux_sym_enum_statement_token1] = aux_sym_enum_statement_token1,
  [aux_sym_logical_expression_token1] = aux_sym_logical_expression_token1,
  [aux_sym_logical_expression_token2] = aux_sym_logical_expression_token2,
  [aux_sym_logical_expression_token3] = aux_sym_logical_expression_token3,
  [aux_sym_bitwise_expression_token1] = aux_sym_bitwise_expression_token1,
  [aux_sym_bitwise_expression_token2] = aux_sym_bitwise_expression_token2,
  [aux_sym_bitwise_expression_token3] = aux_sym_bitwise_expression_token3,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [aux_sym_expression_with_unary_operator_token1] = aux_sym_expression_with_unary_operator_token1,
  [anon_sym_BANG] = anon_sym_BANG,
  [aux_sym_expression_with_unary_operator_token2] = aux_sym_expression_with_unary_operator_token2,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_DOLLAR_LPAREN] = anon_sym_DOLLAR_LPAREN,
  [anon_sym_AT_LPAREN] = anon_sym_AT_LPAREN,
  [anon_sym_AT_LBRACE] = anon_sym_AT_LBRACE,
  [anon_sym_DOT2] = anon_sym_DOT,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [aux_sym_invokation_foreach_expression_token1] = aux_sym_invokation_foreach_expression_token1,
  [sym__statement_terminator] = sym__statement_terminator,
  [sym_program] = sym_program,
  [sym__literal] = sym__literal,
  [sym_integer_literal] = sym_integer_literal,
  [sym_string_literal] = sym_string_literal,
  [sym_expandable_string_literal] = sym_expandable_string_literal,
  [sym_expandable_here_string_literal] = sym_expandable_here_string_literal,
  [sym_type_name] = sym_type_name,
  [sym_array_type_name] = sym_array_type_name,
  [sym_generic_type_name] = sym_generic_type_name,
  [sym_assignement_operator] = sym_assignement_operator,
  [sym_file_redirection_operator] = sym_file_redirection_operator,
  [sym_merging_redirection_operator] = sym_merging_redirection_operator,
  [sym_comparison_operator] = sym_comparison_operator,
  [sym_format_operator] = sym_format_operator,
  [sym_variable] = sym_variable,
  [aux_sym__verbatim_command_argument_chars] = aux_sym__verbatim_command_argument_chars,
  [sym_script_block] = sym_script_block,
  [sym_param_block] = sym_param_block,
  [sym_parameter_list] = sym_parameter_list,
  [sym_script_parameter] = sym_script_parameter,
  [sym_script_parameter_default] = sym_script_parameter_default,
  [sym_script_block_body] = sym_script_block_body,
  [sym_named_block_list] = sym_named_block_list,
  [sym_named_block] = sym_named_block,
  [sym_block_name] = sym_block_name,
  [sym_statement_block] = sym_statement_block,
  [sym_statement_list] = sym_statement_list,
  [sym__statement] = sym__statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym_if_statement] = sym_if_statement,
  [sym_elseif_clauses] = sym_elseif_clauses,
  [sym_elseif_clause] = sym_elseif_clause,
  [sym_else_clause] = sym_else_clause,
  [sym__labeled_statement] = sym__labeled_statement,
  [sym_switch_statement] = sym_switch_statement,
  [sym_switch_parameters] = sym_switch_parameters,
  [sym_switch_parameter] = sym_switch_parameter,
  [sym_switch_condition] = sym_switch_condition,
  [sym_switch_filename] = sym_switch_filename,
  [sym_switch_body] = sym_switch_body,
  [sym_switch_clauses] = sym_switch_clauses,
  [sym_switch_clause] = sym_switch_clause,
  [sym_switch_clause_condition] = sym_switch_clause_condition,
  [sym_foreach_statement] = sym_foreach_statement,
  [sym_foreach_parameter] = sym_foreach_parameter,
  [sym_for_statement] = sym_for_statement,
  [sym_for_initializer] = sym_for_initializer,
  [sym_for_condition] = sym_for_condition,
  [sym_for_iterator] = sym_for_iterator,
  [sym_while_statement] = sym_while_statement,
  [sym_while_condition] = sym_while_condition,
  [sym_do_statement] = sym_do_statement,
  [sym_function_statement] = sym_function_statement,
  [sym_function_name] = sym_function_name,
  [sym_function_parameter_declaration] = sym_function_parameter_declaration,
  [sym_flow_control_statement] = sym_flow_control_statement,
  [sym_label_expression] = sym_label_expression,
  [sym_trap_statement] = sym_trap_statement,
  [sym_try_statement] = sym_try_statement,
  [sym_catch_clauses] = sym_catch_clauses,
  [sym_catch_clause] = sym_catch_clause,
  [sym_catch_type_list] = sym_catch_type_list,
  [sym_finally_clause] = sym_finally_clause,
  [sym_data_statement] = sym_data_statement,
  [sym_data_name] = sym_data_name,
  [sym_data_commands_allowed] = sym_data_commands_allowed,
  [sym_data_commands_list] = sym_data_commands_list,
  [sym_data_command] = sym_data_command,
  [sym_inlinescript_statement] = sym_inlinescript_statement,
  [sym_parallel_statement] = sym_parallel_statement,
  [sym_sequence_statement] = sym_sequence_statement,
  [sym_pipeline] = sym_pipeline,
  [sym_pipeline_chain] = sym_pipeline_chain,
  [sym_pipeline_chain_tail] = sym_pipeline_chain_tail,
  [sym_left_assignment_expression] = sym_left_assignment_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [aux_sym__pipeline_tail] = aux_sym__pipeline_tail,
  [sym_command] = sym_command,
  [sym_command_invokation_operator] = sym_command_invokation_operator,
  [sym__expandable_string_literal_immediate] = sym__expandable_string_literal_immediate,
  [sym_command_name] = sym_command_name,
  [sym_path_command_name] = sym_path_command_name,
  [sym_command_name_expr] = sym_command_name_expr,
  [sym_command_elements] = sym_command_elements,
  [sym__command_element] = sym__command_element,
  [sym_command_argument_sep] = sym_command_argument_sep,
  [sym__command_argument] = sym__command_argument,
  [sym_verbatim_command_argument] = sym_verbatim_command_argument,
  [sym_redirections] = sym_redirections,
  [sym_redirection] = sym_redirection,
  [sym_redirected_file_name] = sym_redirected_file_name,
  [sym_class_attribute] = sym_class_attribute,
  [sym_class_property_definition] = sym_class_property_definition,
  [sym_class_method_parameter] = sym_class_method_parameter,
  [sym_class_method_parameter_list] = sym_class_method_parameter_list,
  [sym_class_method_definition] = sym_class_method_definition,
  [sym_class_statement] = sym_class_statement,
  [sym_enum_statement] = sym_enum_statement,
  [sym_enum_member] = sym_enum_member,
  [sym__expression] = sym__expression,
  [sym_logical_expression] = sym_logical_expression,
  [sym_bitwise_expression] = sym_bitwise_expression,
  [sym_comparison_expression] = sym_comparison_expression,
  [sym_additive_expression] = sym_additive_expression,
  [sym_multiplicative_expression] = sym_multiplicative_expression,
  [sym_format_expression] = sym_format_expression,
  [sym_range_expression] = sym_range_expression,
  [sym_array_literal_expression] = sym_array_literal_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_expression_with_unary_operator] = sym_expression_with_unary_operator,
  [sym_pre_increment_expression] = sym_pre_increment_expression,
  [sym_pre_decrement_expression] = sym_pre_decrement_expression,
  [sym_cast_expression] = sym_cast_expression,
  [sym__primary_expression] = sym__primary_expression,
  [sym__value] = sym__value,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_sub_expression] = sym_sub_expression,
  [sym_array_expression] = sym_array_expression,
  [sym_script_block_expression] = sym_script_block_expression,
  [sym_hash_literal_expression] = sym_hash_literal_expression,
  [sym_hash_literal_body] = sym_hash_literal_body,
  [sym_hash_entry] = sym_hash_entry,
  [sym_key_expression] = sym_key_expression,
  [sym_post_increment_expression] = sym_post_increment_expression,
  [sym_post_decrement_expression] = sym_post_decrement_expression,
  [sym_member_access] = sym_member_access,
  [sym_member_name] = sym_member_name,
  [sym_element_access] = sym_element_access,
  [sym_invokation_expression] = sym_invokation_expression,
  [sym_invokation_foreach_expression] = sym_invokation_foreach_expression,
  [sym_argument_list] = sym_argument_list,
  [sym_argument_expression_list] = sym_argument_expression_list,
  [sym_argument_expression] = sym_argument_expression,
  [sym_logical_argument_expression] = sym_logical_argument_expression,
  [sym_bitwise_argument_expression] = sym_bitwise_argument_expression,
  [sym_comparison_argument_expression] = sym_comparison_argument_expression,
  [sym_additive_argument_expression] = sym_additive_argument_expression,
  [sym_multiplicative_argument_expression] = sym_multiplicative_argument_expression,
  [sym_format_argument_expression] = sym_format_argument_expression,
  [sym_range_argument_expression] = sym_range_argument_expression,
  [sym_type_literal] = sym_type_literal,
  [sym_type_spec] = sym_type_spec,
  [sym_dimension] = sym_dimension,
  [sym_generic_type_arguments] = sym_generic_type_arguments,
  [sym_attribute_list] = sym_attribute_list,
  [sym_attribute] = sym_attribute,
  [sym_attribute_name] = sym_attribute_name,
  [sym_attribute_arguments] = sym_attribute_arguments,
  [sym_attribute_argument] = sym_attribute_argument,
  [aux_sym_expandable_string_literal_repeat1] = aux_sym_expandable_string_literal_repeat1,
  [aux_sym_expandable_string_literal_repeat2] = aux_sym_expandable_string_literal_repeat2,
  [aux_sym_expandable_here_string_literal_repeat1] = aux_sym_expandable_here_string_literal_repeat1,
  [aux_sym_script_block_repeat1] = aux_sym_script_block_repeat1,
  [aux_sym_parameter_list_repeat1] = aux_sym_parameter_list_repeat1,
  [aux_sym_named_block_list_repeat1] = aux_sym_named_block_list_repeat1,
  [aux_sym_statement_list_repeat1] = aux_sym_statement_list_repeat1,
  [aux_sym_elseif_clauses_repeat1] = aux_sym_elseif_clauses_repeat1,
  [aux_sym_switch_parameters_repeat1] = aux_sym_switch_parameters_repeat1,
  [aux_sym_switch_clauses_repeat1] = aux_sym_switch_clauses_repeat1,
  [aux_sym_catch_clauses_repeat1] = aux_sym_catch_clauses_repeat1,
  [aux_sym_catch_type_list_repeat1] = aux_sym_catch_type_list_repeat1,
  [aux_sym_data_commands_list_repeat1] = aux_sym_data_commands_list_repeat1,
  [aux_sym_pipeline_repeat1] = aux_sym_pipeline_repeat1,
  [aux_sym__expandable_string_literal_immediate_repeat1] = aux_sym__expandable_string_literal_immediate_repeat1,
  [aux_sym__expandable_string_literal_immediate_repeat2] = aux_sym__expandable_string_literal_immediate_repeat2,
  [aux_sym_command_name_repeat1] = aux_sym_command_name_repeat1,
  [aux_sym_path_command_name_repeat1] = aux_sym_path_command_name_repeat1,
  [aux_sym_command_elements_repeat1] = aux_sym_command_elements_repeat1,
  [aux_sym_command_argument_sep_repeat1] = aux_sym_command_argument_sep_repeat1,
  [aux_sym_redirections_repeat1] = aux_sym_redirections_repeat1,
  [aux_sym_class_property_definition_repeat1] = aux_sym_class_property_definition_repeat1,
  [aux_sym_class_method_parameter_list_repeat1] = aux_sym_class_method_parameter_list_repeat1,
  [aux_sym_class_statement_repeat1] = aux_sym_class_statement_repeat1,
  [aux_sym_class_statement_repeat2] = aux_sym_class_statement_repeat2,
  [aux_sym_enum_statement_repeat1] = aux_sym_enum_statement_repeat1,
  [aux_sym_array_literal_expression_repeat1] = aux_sym_array_literal_expression_repeat1,
  [aux_sym_hash_literal_body_repeat1] = aux_sym_hash_literal_body_repeat1,
  [aux_sym_argument_expression_list_repeat1] = aux_sym_argument_expression_list_repeat1,
  [aux_sym_dimension_repeat1] = aux_sym_dimension_repeat1,
  [aux_sym_generic_type_arguments_repeat1] = aux_sym_generic_type_arguments_repeat1,
  [aux_sym_attribute_list_repeat1] = aux_sym_attribute_list_repeat1,
  [aux_sym_attribute_arguments_repeat1] = aux_sym_attribute_arguments_repeat1,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_decimal_integer_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_hexadecimal_integer_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_real_literal] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_expandable_string_literal_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_string_literal_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_string_literal_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_string_literal_token4] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_expandable_string_literal_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_token5] = {
    .visible = false,
    .named = false,
  },
  [sym_verbatim_string_characters] = {
    .visible = true,
    .named = true,
  },
  [sym_verbatim_here_string_characters] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_name] = {
    .visible = true,
    .named = true,
  },
  [sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_2_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_2_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_3_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_3_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_4_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_4_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_5_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_5_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_6_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_6_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_2_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_3_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_4_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_5_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_6_GT_AMP1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_1_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_3_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_4_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_5_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_6_GT_AMP2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token4] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token5] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token6] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token7] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token8] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token9] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token10] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token11] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token12] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token13] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token14] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token15] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token16] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token17] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token18] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token19] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token20] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token21] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token22] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token23] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token24] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token25] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token26] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token27] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token28] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token29] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token30] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token31] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token32] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token33] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token34] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token35] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token36] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token37] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token38] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token39] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token40] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token41] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token42] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token43] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token44] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token45] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token46] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token47] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token48] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token49] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comparison_operator_token50] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_format_operator_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_variable_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variable_token2] = {
    .visible = false,
    .named = false,
  },
  [sym_braced_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_token] = {
    .visible = true,
    .named = true,
  },
  [sym__command_token] = {
    .visible = false,
    .named = true,
  },
  [sym_command_parameter] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__verbatim_command_argument_chars_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__verbatim_command_argument_chars_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__verbatim_command_argument_chars_token3] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_param_block_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_block_name_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_block_name_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_block_name_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_block_name_token4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_if_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_elseif_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_else_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_parameter_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_parameter_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_parameter_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_parameter_token4] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_parameter_token5] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_condition_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_while_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_do_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_do_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_function_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_function_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_function_statement_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_flow_control_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_flow_control_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_flow_control_statement_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_flow_control_statement_token4] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_flow_control_statement_token5] = {
    .visible = true,
    .named = false,
  },
  [sym_label] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_trap_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_try_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_catch_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_finally_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_data_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_data_commands_allowed_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_inlinescript_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_parallel_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_sequence_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__expandable_string_literal_immediate_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expandable_string_literal_immediate_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expandable_string_literal_immediate_token3] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_command_name_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token6] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token7] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token8] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token9] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token10] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token11] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token12] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token13] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token14] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token15] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token16] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token17] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token18] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_token19] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_path_command_name_token] = {
    .visible = true,
    .named = true,
  },
  [sym_stop_parsing] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SPACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_attribute_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_attribute_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_statement_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_statement_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_logical_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_logical_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_logical_expression_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_bitwise_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_bitwise_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_bitwise_expression_token3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_expression_with_unary_operator_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_expression_with_unary_operator_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_invokation_foreach_expression_token1] = {
    .visible = false,
    .named = false,
  },
  [sym__statement_terminator] = {
    .visible = false,
    .named = true,
  },
  [sym_program] = {
    .visible = true,
    .named = true,
  },
  [sym__literal] = {
    .visible = false,
    .named = true,
  },
  [sym_integer_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_expandable_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_expandable_here_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_assignement_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_file_redirection_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_merging_redirection_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_comparison_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_format_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_variable] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__verbatim_command_argument_chars] = {
    .visible = false,
    .named = false,
  },
  [sym_script_block] = {
    .visible = true,
    .named = true,
  },
  [sym_param_block] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_script_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_script_parameter_default] = {
    .visible = true,
    .named = true,
  },
  [sym_script_block_body] = {
    .visible = true,
    .named = true,
  },
  [sym_named_block_list] = {
    .visible = true,
    .named = true,
  },
  [sym_named_block] = {
    .visible = true,
    .named = true,
  },
  [sym_block_name] = {
    .visible = true,
    .named = true,
  },
  [sym_statement_block] = {
    .visible = true,
    .named = true,
  },
  [sym_statement_list] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_elseif_clauses] = {
    .visible = true,
    .named = true,
  },
  [sym_elseif_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__labeled_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_filename] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_body] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_clauses] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_clause_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_for_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_for_iterator] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_while_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_function_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_function_name] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_flow_control_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_label_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_trap_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clauses] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_type_list] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_data_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_data_name] = {
    .visible = true,
    .named = true,
  },
  [sym_data_commands_allowed] = {
    .visible = true,
    .named = true,
  },
  [sym_data_commands_list] = {
    .visible = true,
    .named = true,
  },
  [sym_data_command] = {
    .visible = true,
    .named = true,
  },
  [sym_inlinescript_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_parallel_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline_chain] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline_chain_tail] = {
    .visible = true,
    .named = true,
  },
  [sym_left_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__pipeline_tail] = {
    .visible = false,
    .named = false,
  },
  [sym_command] = {
    .visible = true,
    .named = true,
  },
  [sym_command_invokation_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__expandable_string_literal_immediate] = {
    .visible = false,
    .named = true,
  },
  [sym_command_name] = {
    .visible = true,
    .named = true,
  },
  [sym_path_command_name] = {
    .visible = true,
    .named = true,
  },
  [sym_command_name_expr] = {
    .visible = true,
    .named = true,
  },
  [sym_command_elements] = {
    .visible = true,
    .named = true,
  },
  [sym__command_element] = {
    .visible = false,
    .named = true,
  },
  [sym_command_argument_sep] = {
    .visible = true,
    .named = true,
  },
  [sym__command_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_verbatim_command_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_redirections] = {
    .visible = true,
    .named = true,
  },
  [sym_redirection] = {
    .visible = true,
    .named = true,
  },
  [sym_redirected_file_name] = {
    .visible = true,
    .named = true,
  },
  [sym_class_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_class_property_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_method_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_class_method_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_class_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_member] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym_logical_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_bitwise_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_comparison_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_additive_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_multiplicative_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_format_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_range_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_literal_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_with_unary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_pre_increment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_pre_decrement_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__primary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__value] = {
    .visible = false,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sub_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_script_block_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_literal_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_literal_body] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_entry] = {
    .visible = true,
    .named = true,
  },
  [sym_key_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_post_increment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_post_decrement_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_member_access] = {
    .visible = true,
    .named = true,
  },
  [sym_member_name] = {
    .visible = true,
    .named = true,
  },
  [sym_element_access] = {
    .visible = true,
    .named = true,
  },
  [sym_invokation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_invokation_foreach_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_expression_list] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_bitwise_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_comparison_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_additive_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_multiplicative_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_format_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_range_argument_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_type_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_dimension] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_list] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_name] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_argument] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_expandable_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_string_literal_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expandable_here_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_script_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_named_block_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_statement_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_elseif_clauses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_clauses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_catch_clauses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_catch_type_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_data_commands_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipeline_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expandable_string_literal_immediate_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expandable_string_literal_immediate_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_path_command_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_elements_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_argument_sep_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_redirections_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_property_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_method_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_statement_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_literal_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hash_literal_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_expression_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dimension_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_argument_expression_list = 1,
  field_command_elements = 2,
  field_command_name = 3,
  field_condition = 4,
  field_else_clause = 5,
  field_elseif_clauses = 6,
  field_for_condition = 7,
  field_for_initializer = 8,
  field_for_iterator = 9,
  field_named_block_list = 10,
  field_script_block_body = 11,
  field_statement_list = 12,
  field_statements = 13,
  field_value = 14,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_argument_expression_list] = "argument_expression_list",
  [field_command_elements] = "command_elements",
  [field_command_name] = "command_name",
  [field_condition] = "condition",
  [field_else_clause] = "else_clause",
  [field_elseif_clauses] = "elseif_clauses",
  [field_for_condition] = "for_condition",
  [field_for_initializer] = "for_initializer",
  [field_for_iterator] = "for_iterator",
  [field_named_block_list] = "named_block_list",
  [field_script_block_body] = "script_block_body",
  [field_statement_list] = "statement_list",
  [field_statements] = "statements",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [2] = {.index = 1, .length = 1},
  [3] = {.index = 2, .length = 1},
  [4] = {.index = 3, .length = 1},
  [5] = {.index = 4, .length = 1},
  [6] = {.index = 5, .length = 2},
  [7] = {.index = 7, .length = 1},
  [8] = {.index = 8, .length = 1},
  [9] = {.index = 9, .length = 2},
  [10] = {.index = 11, .length = 1},
  [11] = {.index = 12, .length = 1},
  [12] = {.index = 13, .length = 1},
  [13] = {.index = 14, .length = 1},
  [14] = {.index = 15, .length = 1},
  [15] = {.index = 16, .length = 2},
  [16] = {.index = 18, .length = 2},
  [17] = {.index = 20, .length = 1},
  [18] = {.index = 21, .length = 1},
  [19] = {.index = 22, .length = 3},
  [20] = {.index = 25, .length = 1},
  [21] = {.index = 26, .length = 1},
  [22] = {.index = 27, .length = 2},
  [23] = {.index = 29, .length = 2},
  [24] = {.index = 31, .length = 2},
  [25] = {.index = 33, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_command_name, 0},
  [1] =
    {field_script_block_body, 0},
  [2] =
    {field_named_block_list, 0},
  [3] =
    {field_statement_list, 0},
  [4] =
    {field_command_name, 1},
  [5] =
    {field_command_elements, 1},
    {field_command_name, 0},
  [7] =
    {field_statements, 1},
  [8] =
    {field_value, 2},
  [9] =
    {field_command_elements, 2},
    {field_command_name, 1},
  [11] =
    {field_script_block_body, 2},
  [12] =
    {field_statement_list, 1},
  [13] =
    {field_script_block_body, 3},
  [14] =
    {field_condition, 2},
  [15] =
    {field_argument_expression_list, 1},
  [16] =
    {field_condition, 2},
    {field_elseif_clauses, 5},
  [18] =
    {field_condition, 2},
    {field_else_clause, 5},
  [20] =
    {field_for_initializer, 2},
  [21] =
    {field_condition, 4},
  [22] =
    {field_condition, 2},
    {field_else_clause, 6},
    {field_elseif_clauses, 5},
  [25] =
    {field_for_condition, 3},
  [26] =
    {field_for_iterator, 4},
  [27] =
    {field_for_condition, 5},
    {field_for_initializer, 2},
  [29] =
    {field_for_condition, 3},
    {field_for_iterator, 6},
  [31] =
    {field_for_initializer, 2},
    {field_for_iterator, 6},
  [33] =
    {field_for_condition, 5},
    {field_for_initializer, 2},
    {field_for_iterator, 8},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 3,
  [5] = 3,
  [6] = 3,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 8,
  [16] = 9,
  [17] = 8,
  [18] = 9,
  [19] = 8,
  [20] = 9,
  [21] = 7,
  [22] = 7,
  [23] = 7,
  [24] = 7,
  [25] = 7,
  [26] = 7,
  [27] = 7,
  [28] = 7,
  [29] = 7,
  [30] = 7,
  [31] = 7,
  [32] = 32,
  [33] = 32,
  [34] = 32,
  [35] = 32,
  [36] = 32,
  [37] = 32,
  [38] = 32,
  [39] = 32,
  [40] = 32,
  [41] = 32,
  [42] = 32,
  [43] = 32,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 46,
  [50] = 47,
  [51] = 47,
  [52] = 47,
  [53] = 46,
  [54] = 46,
  [55] = 55,
  [56] = 48,
  [57] = 47,
  [58] = 48,
  [59] = 46,
  [60] = 47,
  [61] = 61,
  [62] = 47,
  [63] = 48,
  [64] = 46,
  [65] = 47,
  [66] = 48,
  [67] = 47,
  [68] = 48,
  [69] = 46,
  [70] = 47,
  [71] = 48,
  [72] = 46,
  [73] = 47,
  [74] = 48,
  [75] = 75,
  [76] = 55,
  [77] = 61,
  [78] = 78,
  [79] = 79,
  [80] = 79,
  [81] = 81,
  [82] = 81,
  [83] = 3,
  [84] = 3,
  [85] = 3,
  [86] = 3,
  [87] = 87,
  [88] = 88,
  [89] = 88,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 88,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 88,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 122,
  [123] = 123,
  [124] = 117,
  [125] = 101,
  [126] = 93,
  [127] = 112,
  [128] = 102,
  [129] = 90,
  [130] = 105,
  [131] = 123,
  [132] = 108,
  [133] = 109,
  [134] = 110,
  [135] = 104,
  [136] = 91,
  [137] = 92,
  [138] = 120,
  [139] = 106,
  [140] = 95,
  [141] = 96,
  [142] = 87,
  [143] = 97,
  [144] = 98,
  [145] = 103,
  [146] = 121,
  [147] = 100,
  [148] = 99,
  [149] = 107,
  [150] = 150,
  [151] = 150,
  [152] = 113,
  [153] = 114,
  [154] = 116,
  [155] = 118,
  [156] = 122,
  [157] = 119,
  [158] = 94,
  [159] = 150,
  [160] = 150,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 104,
  [166] = 105,
  [167] = 123,
  [168] = 168,
  [169] = 108,
  [170] = 109,
  [171] = 164,
  [172] = 172,
  [173] = 172,
  [174] = 110,
  [175] = 168,
  [176] = 104,
  [177] = 105,
  [178] = 108,
  [179] = 109,
  [180] = 161,
  [181] = 123,
  [182] = 110,
  [183] = 183,
  [184] = 162,
  [185] = 185,
  [186] = 183,
  [187] = 187,
  [188] = 185,
  [189] = 189,
  [190] = 189,
  [191] = 163,
  [192] = 187,
  [193] = 193,
  [194] = 194,
  [195] = 193,
  [196] = 194,
  [197] = 197,
  [198] = 197,
  [199] = 199,
  [200] = 199,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 201,
  [206] = 204,
  [207] = 203,
  [208] = 201,
  [209] = 204,
  [210] = 203,
  [211] = 211,
  [212] = 203,
  [213] = 204,
  [214] = 202,
  [215] = 81,
  [216] = 211,
  [217] = 201,
  [218] = 3,
  [219] = 81,
  [220] = 3,
  [221] = 3,
  [222] = 3,
  [223] = 223,
  [224] = 224,
  [225] = 223,
  [226] = 226,
  [227] = 224,
  [228] = 228,
  [229] = 229,
  [230] = 228,
  [231] = 229,
  [232] = 224,
  [233] = 223,
  [234] = 229,
  [235] = 3,
  [236] = 228,
  [237] = 3,
  [238] = 226,
  [239] = 223,
  [240] = 3,
  [241] = 3,
  [242] = 224,
  [243] = 229,
  [244] = 228,
  [245] = 245,
  [246] = 226,
  [247] = 226,
  [248] = 248,
  [249] = 245,
  [250] = 250,
  [251] = 245,
  [252] = 245,
  [253] = 250,
  [254] = 250,
  [255] = 250,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 257,
  [261] = 256,
  [262] = 258,
  [263] = 263,
  [264] = 258,
  [265] = 265,
  [266] = 258,
  [267] = 258,
  [268] = 258,
  [269] = 258,
  [270] = 258,
  [271] = 258,
  [272] = 258,
  [273] = 258,
  [274] = 258,
  [275] = 263,
  [276] = 259,
  [277] = 257,
  [278] = 263,
  [279] = 259,
  [280] = 257,
  [281] = 256,
  [282] = 263,
  [283] = 259,
  [284] = 256,
  [285] = 3,
  [286] = 3,
  [287] = 3,
  [288] = 3,
  [289] = 226,
  [290] = 226,
  [291] = 291,
  [292] = 291,
  [293] = 293,
  [294] = 294,
  [295] = 294,
  [296] = 293,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 305,
  [306] = 297,
  [307] = 298,
  [308] = 301,
  [309] = 303,
  [310] = 310,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 313,
  [317] = 317,
  [318] = 315,
  [319] = 317,
  [320] = 311,
  [321] = 310,
  [322] = 315,
  [323] = 313,
  [324] = 314,
  [325] = 312,
  [326] = 313,
  [327] = 315,
  [328] = 328,
  [329] = 329,
  [330] = 330,
  [331] = 331,
  [332] = 332,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 336,
  [337] = 337,
  [338] = 338,
  [339] = 339,
  [340] = 340,
  [341] = 341,
  [342] = 342,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 313,
  [354] = 354,
  [355] = 355,
  [356] = 356,
  [357] = 315,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 373,
  [374] = 374,
  [375] = 375,
  [376] = 376,
  [377] = 377,
  [378] = 365,
  [379] = 371,
  [380] = 354,
  [381] = 349,
  [382] = 373,
  [383] = 375,
  [384] = 376,
  [385] = 352,
  [386] = 344,
  [387] = 364,
  [388] = 337,
  [389] = 346,
  [390] = 347,
  [391] = 350,
  [392] = 351,
  [393] = 377,
  [394] = 359,
  [395] = 360,
  [396] = 366,
  [397] = 372,
  [398] = 329,
  [399] = 330,
  [400] = 331,
  [401] = 332,
  [402] = 333,
  [403] = 334,
  [404] = 335,
  [405] = 336,
  [406] = 315,
  [407] = 338,
  [408] = 339,
  [409] = 340,
  [410] = 342,
  [411] = 361,
  [412] = 362,
  [413] = 356,
  [414] = 367,
  [415] = 358,
  [416] = 341,
  [417] = 369,
  [418] = 374,
  [419] = 343,
  [420] = 345,
  [421] = 313,
  [422] = 348,
  [423] = 363,
  [424] = 328,
  [425] = 368,
  [426] = 370,
  [427] = 355,
  [428] = 299,
  [429] = 429,
  [430] = 300,
  [431] = 150,
  [432] = 304,
  [433] = 150,
  [434] = 150,
  [435] = 150,
  [436] = 81,
  [437] = 81,
  [438] = 438,
  [439] = 439,
  [440] = 168,
  [441] = 439,
  [442] = 442,
  [443] = 81,
  [444] = 81,
  [445] = 438,
  [446] = 442,
  [447] = 168,
  [448] = 448,
  [449] = 448,
  [450] = 164,
  [451] = 172,
  [452] = 164,
  [453] = 438,
  [454] = 164,
  [455] = 172,
  [456] = 172,
  [457] = 438,
  [458] = 439,
  [459] = 439,
  [460] = 183,
  [461] = 448,
  [462] = 187,
  [463] = 448,
  [464] = 464,
  [465] = 185,
  [466] = 189,
  [467] = 187,
  [468] = 468,
  [469] = 187,
  [470] = 468,
  [471] = 468,
  [472] = 468,
  [473] = 185,
  [474] = 468,
  [475] = 468,
  [476] = 476,
  [477] = 183,
  [478] = 189,
  [479] = 468,
  [480] = 185,
  [481] = 468,
  [482] = 189,
  [483] = 183,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 81,
  [489] = 81,
  [490] = 487,
  [491] = 438,
  [492] = 492,
  [493] = 493,
  [494] = 439,
  [495] = 487,
  [496] = 193,
  [497] = 487,
  [498] = 498,
  [499] = 194,
  [500] = 487,
  [501] = 487,
  [502] = 487,
  [503] = 487,
  [504] = 438,
  [505] = 439,
  [506] = 506,
  [507] = 507,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 193,
  [513] = 509,
  [514] = 514,
  [515] = 509,
  [516] = 509,
  [517] = 517,
  [518] = 518,
  [519] = 194,
  [520] = 194,
  [521] = 193,
  [522] = 522,
  [523] = 522,
  [524] = 522,
  [525] = 448,
  [526] = 522,
  [527] = 448,
  [528] = 522,
  [529] = 529,
  [530] = 530,
  [531] = 529,
  [532] = 529,
  [533] = 529,
  [534] = 534,
  [535] = 529,
  [536] = 534,
  [537] = 529,
  [538] = 538,
  [539] = 529,
  [540] = 534,
  [541] = 541,
  [542] = 534,
  [543] = 534,
  [544] = 529,
  [545] = 199,
  [546] = 197,
  [547] = 547,
  [548] = 548,
  [549] = 549,
  [550] = 549,
  [551] = 549,
  [552] = 549,
  [553] = 549,
  [554] = 554,
  [555] = 199,
  [556] = 3,
  [557] = 554,
  [558] = 197,
  [559] = 3,
  [560] = 202,
  [561] = 554,
  [562] = 554,
  [563] = 211,
  [564] = 199,
  [565] = 3,
  [566] = 197,
  [567] = 554,
  [568] = 3,
  [569] = 569,
  [570] = 570,
  [571] = 202,
  [572] = 572,
  [573] = 211,
  [574] = 570,
  [575] = 575,
  [576] = 570,
  [577] = 202,
  [578] = 570,
  [579] = 211,
  [580] = 570,
  [581] = 581,
  [582] = 582,
  [583] = 583,
  [584] = 583,
  [585] = 585,
  [586] = 583,
  [587] = 583,
  [588] = 583,
  [589] = 589,
  [590] = 589,
  [591] = 591,
  [592] = 589,
  [593] = 589,
  [594] = 594,
  [595] = 589,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 598,
  [600] = 600,
  [601] = 601,
  [602] = 598,
  [603] = 597,
  [604] = 601,
  [605] = 598,
  [606] = 597,
  [607] = 600,
  [608] = 600,
  [609] = 598,
  [610] = 597,
  [611] = 600,
  [612] = 600,
  [613] = 601,
  [614] = 598,
  [615] = 597,
  [616] = 600,
  [617] = 601,
  [618] = 598,
  [619] = 597,
  [620] = 600,
  [621] = 598,
  [622] = 597,
  [623] = 600,
  [624] = 598,
  [625] = 597,
  [626] = 600,
  [627] = 598,
  [628] = 597,
  [629] = 600,
  [630] = 598,
  [631] = 597,
  [632] = 600,
  [633] = 633,
  [634] = 598,
  [635] = 597,
  [636] = 600,
  [637] = 601,
  [638] = 598,
  [639] = 597,
  [640] = 600,
  [641] = 598,
  [642] = 597,
  [643] = 600,
  [644] = 601,
  [645] = 598,
  [646] = 597,
  [647] = 600,
  [648] = 601,
  [649] = 598,
  [650] = 597,
  [651] = 600,
  [652] = 598,
  [653] = 597,
  [654] = 600,
  [655] = 601,
  [656] = 598,
  [657] = 597,
  [658] = 600,
  [659] = 598,
  [660] = 597,
  [661] = 600,
  [662] = 601,
  [663] = 598,
  [664] = 597,
  [665] = 600,
  [666] = 601,
  [667] = 598,
  [668] = 597,
  [669] = 600,
  [670] = 598,
  [671] = 598,
  [672] = 597,
  [673] = 600,
  [674] = 597,
  [675] = 598,
  [676] = 597,
  [677] = 600,
  [678] = 600,
  [679] = 597,
  [680] = 680,
  [681] = 87,
  [682] = 87,
  [683] = 680,
  [684] = 87,
  [685] = 87,
  [686] = 680,
  [687] = 680,
  [688] = 688,
  [689] = 689,
  [690] = 689,
  [691] = 689,
  [692] = 689,
  [693] = 689,
  [694] = 689,
  [695] = 689,
  [696] = 689,
  [697] = 689,
  [698] = 689,
  [699] = 689,
  [700] = 689,
  [701] = 689,
  [702] = 702,
  [703] = 689,
  [704] = 689,
  [705] = 689,
  [706] = 689,
  [707] = 689,
  [708] = 689,
  [709] = 709,
  [710] = 689,
  [711] = 709,
  [712] = 702,
  [713] = 713,
  [714] = 702,
  [715] = 715,
  [716] = 713,
  [717] = 702,
  [718] = 709,
  [719] = 719,
  [720] = 719,
  [721] = 721,
  [722] = 709,
  [723] = 715,
  [724] = 721,
  [725] = 713,
  [726] = 713,
  [727] = 721,
  [728] = 721,
  [729] = 719,
  [730] = 719,
  [731] = 715,
  [732] = 715,
  [733] = 81,
  [734] = 734,
  [735] = 734,
  [736] = 88,
  [737] = 87,
  [738] = 738,
  [739] = 88,
  [740] = 734,
  [741] = 734,
  [742] = 734,
  [743] = 738,
  [744] = 734,
  [745] = 738,
  [746] = 738,
  [747] = 87,
  [748] = 87,
  [749] = 87,
  [750] = 112,
  [751] = 87,
  [752] = 99,
  [753] = 150,
  [754] = 120,
  [755] = 103,
  [756] = 121,
  [757] = 87,
  [758] = 91,
  [759] = 92,
  [760] = 99,
  [761] = 94,
  [762] = 107,
  [763] = 113,
  [764] = 114,
  [765] = 95,
  [766] = 96,
  [767] = 116,
  [768] = 118,
  [769] = 93,
  [770] = 112,
  [771] = 771,
  [772] = 91,
  [773] = 92,
  [774] = 97,
  [775] = 94,
  [776] = 95,
  [777] = 97,
  [778] = 98,
  [779] = 104,
  [780] = 98,
  [781] = 117,
  [782] = 119,
  [783] = 100,
  [784] = 113,
  [785] = 122,
  [786] = 106,
  [787] = 114,
  [788] = 101,
  [789] = 88,
  [790] = 96,
  [791] = 116,
  [792] = 118,
  [793] = 122,
  [794] = 150,
  [795] = 100,
  [796] = 120,
  [797] = 103,
  [798] = 121,
  [799] = 106,
  [800] = 101,
  [801] = 102,
  [802] = 90,
  [803] = 105,
  [804] = 123,
  [805] = 117,
  [806] = 108,
  [807] = 109,
  [808] = 110,
  [809] = 88,
  [810] = 104,
  [811] = 105,
  [812] = 123,
  [813] = 108,
  [814] = 109,
  [815] = 110,
  [816] = 93,
  [817] = 102,
  [818] = 90,
  [819] = 819,
  [820] = 119,
  [821] = 107,
  [822] = 94,
  [823] = 94,
  [824] = 95,
  [825] = 110,
  [826] = 826,
  [827] = 97,
  [828] = 98,
  [829] = 117,
  [830] = 119,
  [831] = 104,
  [832] = 87,
  [833] = 833,
  [834] = 81,
  [835] = 150,
  [836] = 87,
  [837] = 120,
  [838] = 103,
  [839] = 105,
  [840] = 121,
  [841] = 117,
  [842] = 119,
  [843] = 99,
  [844] = 107,
  [845] = 845,
  [846] = 113,
  [847] = 105,
  [848] = 114,
  [849] = 123,
  [850] = 96,
  [851] = 116,
  [852] = 123,
  [853] = 118,
  [854] = 854,
  [855] = 845,
  [856] = 854,
  [857] = 857,
  [858] = 108,
  [859] = 150,
  [860] = 93,
  [861] = 112,
  [862] = 91,
  [863] = 92,
  [864] = 109,
  [865] = 92,
  [866] = 109,
  [867] = 110,
  [868] = 120,
  [869] = 97,
  [870] = 90,
  [871] = 98,
  [872] = 103,
  [873] = 121,
  [874] = 104,
  [875] = 81,
  [876] = 99,
  [877] = 107,
  [878] = 113,
  [879] = 122,
  [880] = 106,
  [881] = 114,
  [882] = 96,
  [883] = 826,
  [884] = 116,
  [885] = 118,
  [886] = 886,
  [887] = 886,
  [888] = 857,
  [889] = 100,
  [890] = 833,
  [891] = 102,
  [892] = 101,
  [893] = 102,
  [894] = 108,
  [895] = 93,
  [896] = 90,
  [897] = 112,
  [898] = 100,
  [899] = 91,
  [900] = 122,
  [901] = 106,
  [902] = 101,
  [903] = 95,
  [904] = 904,
  [905] = 826,
  [906] = 857,
  [907] = 833,
  [908] = 857,
  [909] = 833,
  [910] = 826,
  [911] = 904,
  [912] = 912,
  [913] = 81,
  [914] = 81,
  [915] = 915,
  [916] = 912,
  [917] = 915,
  [918] = 904,
  [919] = 912,
  [920] = 88,
  [921] = 912,
  [922] = 915,
  [923] = 915,
  [924] = 904,
  [925] = 925,
  [926] = 926,
  [927] = 88,
  [928] = 928,
  [929] = 929,
  [930] = 928,
  [931] = 926,
  [932] = 925,
  [933] = 119,
  [934] = 92,
  [935] = 94,
  [936] = 95,
  [937] = 96,
  [938] = 926,
  [939] = 97,
  [940] = 98,
  [941] = 925,
  [942] = 99,
  [943] = 107,
  [944] = 928,
  [945] = 120,
  [946] = 113,
  [947] = 926,
  [948] = 925,
  [949] = 87,
  [950] = 100,
  [951] = 101,
  [952] = 114,
  [953] = 116,
  [954] = 102,
  [955] = 90,
  [956] = 104,
  [957] = 105,
  [958] = 123,
  [959] = 118,
  [960] = 108,
  [961] = 109,
  [962] = 110,
  [963] = 963,
  [964] = 122,
  [965] = 106,
  [966] = 928,
  [967] = 103,
  [968] = 121,
  [969] = 963,
  [970] = 93,
  [971] = 112,
  [972] = 117,
  [973] = 91,
  [974] = 164,
  [975] = 172,
  [976] = 168,
  [977] = 164,
  [978] = 172,
  [979] = 88,
  [980] = 168,
  [981] = 88,
  [982] = 105,
  [983] = 98,
  [984] = 87,
  [985] = 104,
  [986] = 105,
  [987] = 123,
  [988] = 108,
  [989] = 109,
  [990] = 110,
  [991] = 168,
  [992] = 110,
  [993] = 123,
  [994] = 101,
  [995] = 172,
  [996] = 168,
  [997] = 172,
  [998] = 104,
  [999] = 108,
  [1000] = 102,
  [1001] = 109,
  [1002] = 90,
  [1003] = 1003,
  [1004] = 100,
  [1005] = 164,
  [1006] = 1006,
  [1007] = 161,
  [1008] = 117,
  [1009] = 119,
  [1010] = 104,
  [1011] = 105,
  [1012] = 120,
  [1013] = 103,
  [1014] = 121,
  [1015] = 123,
  [1016] = 108,
  [1017] = 109,
  [1018] = 99,
  [1019] = 107,
  [1020] = 1003,
  [1021] = 113,
  [1022] = 114,
  [1023] = 110,
  [1024] = 116,
  [1025] = 118,
  [1026] = 122,
  [1027] = 106,
  [1028] = 93,
  [1029] = 112,
  [1030] = 161,
  [1031] = 91,
  [1032] = 92,
  [1033] = 94,
  [1034] = 95,
  [1035] = 96,
  [1036] = 150,
  [1037] = 150,
  [1038] = 97,
  [1039] = 164,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 1043,
  [1044] = 1044,
  [1045] = 1045,
  [1046] = 90,
  [1047] = 1044,
  [1048] = 1045,
  [1049] = 102,
  [1050] = 1003,
  [1051] = 90,
  [1052] = 1041,
  [1053] = 102,
  [1054] = 122,
  [1055] = 106,
  [1056] = 1056,
  [1057] = 96,
  [1058] = 104,
  [1059] = 161,
  [1060] = 96,
  [1061] = 1040,
  [1062] = 1003,
  [1063] = 122,
  [1064] = 104,
  [1065] = 105,
  [1066] = 123,
  [1067] = 108,
  [1068] = 106,
  [1069] = 109,
  [1070] = 110,
  [1071] = 150,
  [1072] = 150,
  [1073] = 161,
  [1074] = 1042,
  [1075] = 1043,
  [1076] = 105,
  [1077] = 123,
  [1078] = 108,
  [1079] = 109,
  [1080] = 110,
  [1081] = 1056,
  [1082] = 1042,
  [1083] = 122,
  [1084] = 106,
  [1085] = 102,
  [1086] = 826,
  [1087] = 1045,
  [1088] = 96,
  [1089] = 1042,
  [1090] = 1043,
  [1091] = 1056,
  [1092] = 1056,
  [1093] = 1040,
  [1094] = 833,
  [1095] = 1095,
  [1096] = 833,
  [1097] = 1040,
  [1098] = 1044,
  [1099] = 1043,
  [1100] = 826,
  [1101] = 1045,
  [1102] = 1044,
  [1103] = 1041,
  [1104] = 81,
  [1105] = 102,
  [1106] = 1041,
  [1107] = 90,
  [1108] = 90,
  [1109] = 122,
  [1110] = 106,
  [1111] = 81,
  [1112] = 96,
  [1113] = 1095,
  [1114] = 1114,
  [1115] = 1114,
  [1116] = 1116,
  [1117] = 1116,
  [1118] = 81,
  [1119] = 1119,
  [1120] = 302,
  [1121] = 1121,
  [1122] = 172,
  [1123] = 172,
  [1124] = 164,
  [1125] = 168,
  [1126] = 168,
  [1127] = 164,
  [1128] = 1044,
  [1129] = 1040,
  [1130] = 1043,
  [1131] = 1042,
  [1132] = 1040,
  [1133] = 1043,
  [1134] = 1042,
  [1135] = 1056,
  [1136] = 1056,
  [1137] = 1044,
  [1138] = 1138,
  [1139] = 1139,
  [1140] = 1140,
  [1141] = 1141,
  [1142] = 302,
  [1143] = 1143,
  [1144] = 1143,
  [1145] = 1145,
  [1146] = 1146,
  [1147] = 1147,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 1149,
  [1151] = 1151,
  [1152] = 1146,
  [1153] = 1148,
  [1154] = 1154,
  [1155] = 1155,
  [1156] = 1146,
  [1157] = 1151,
  [1158] = 1151,
  [1159] = 1146,
  [1160] = 1149,
  [1161] = 1151,
  [1162] = 1148,
  [1163] = 1148,
  [1164] = 1154,
  [1165] = 1155,
  [1166] = 1154,
  [1167] = 1154,
  [1168] = 1155,
  [1169] = 1155,
  [1170] = 1149,
  [1171] = 1171,
  [1172] = 1171,
  [1173] = 1171,
  [1174] = 1171,
  [1175] = 1171,
  [1176] = 1176,
  [1177] = 1177,
  [1178] = 1178,
  [1179] = 1176,
  [1180] = 1178,
  [1181] = 1177,
  [1182] = 1178,
  [1183] = 1176,
  [1184] = 1178,
  [1185] = 1176,
  [1186] = 1177,
  [1187] = 1177,
  [1188] = 1178,
  [1189] = 1176,
  [1190] = 1177,
  [1191] = 1176,
  [1192] = 1177,
  [1193] = 1176,
  [1194] = 1177,
  [1195] = 1177,
  [1196] = 1176,
  [1197] = 1197,
  [1198] = 1198,
  [1199] = 1197,
  [1200] = 1200,
  [1201] = 1198,
  [1202] = 1198,
  [1203] = 1198,
  [1204] = 1197,
  [1205] = 1198,
  [1206] = 1197,
  [1207] = 1197,
  [1208] = 1198,
  [1209] = 1198,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 1197,
  [1213] = 1197,
  [1214] = 1198,
  [1215] = 1197,
  [1216] = 1216,
  [1217] = 1217,
  [1218] = 1216,
  [1219] = 1216,
  [1220] = 1217,
  [1221] = 1217,
  [1222] = 1222,
  [1223] = 680,
  [1224] = 87,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 302,
  [1229] = 101,
  [1230] = 1230,
  [1231] = 87,
  [1232] = 1232,
  [1233] = 100,
  [1234] = 1234,
  [1235] = 87,
  [1236] = 101,
  [1237] = 1237,
  [1238] = 100,
  [1239] = 101,
  [1240] = 100,
  [1241] = 87,
  [1242] = 1242,
  [1243] = 1243,
  [1244] = 1244,
  [1245] = 1245,
  [1246] = 1246,
  [1247] = 1247,
  [1248] = 713,
  [1249] = 1249,
  [1250] = 715,
  [1251] = 1251,
  [1252] = 1252,
  [1253] = 1253,
  [1254] = 1254,
  [1255] = 1255,
  [1256] = 313,
  [1257] = 1257,
  [1258] = 1258,
  [1259] = 315,
  [1260] = 1260,
  [1261] = 1261,
  [1262] = 1260,
  [1263] = 1263,
  [1264] = 1264,
  [1265] = 1260,
  [1266] = 1266,
  [1267] = 1260,
  [1268] = 1268,
  [1269] = 1263,
  [1270] = 1263,
  [1271] = 1263,
  [1272] = 87,
  [1273] = 150,
  [1274] = 1274,
  [1275] = 1275,
  [1276] = 1276,
  [1277] = 1261,
  [1278] = 1266,
  [1279] = 1279,
  [1280] = 1280,
  [1281] = 1281,
  [1282] = 1281,
  [1283] = 1283,
  [1284] = 1284,
  [1285] = 857,
  [1286] = 886,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1284,
  [1291] = 1291,
  [1292] = 1287,
  [1293] = 854,
  [1294] = 1281,
  [1295] = 150,
  [1296] = 1296,
  [1297] = 1297,
  [1298] = 1283,
  [1299] = 1291,
  [1300] = 1281,
  [1301] = 1291,
  [1302] = 1291,
  [1303] = 912,
  [1304] = 1304,
  [1305] = 150,
  [1306] = 904,
  [1307] = 886,
  [1308] = 293,
  [1309] = 1309,
  [1310] = 294,
  [1311] = 854,
  [1312] = 1312,
  [1313] = 294,
  [1314] = 150,
  [1315] = 886,
  [1316] = 293,
  [1317] = 915,
  [1318] = 1318,
  [1319] = 1304,
  [1320] = 854,
  [1321] = 1321,
  [1322] = 1321,
  [1323] = 926,
  [1324] = 1324,
  [1325] = 1325,
  [1326] = 925,
  [1327] = 928,
  [1328] = 1321,
  [1329] = 1325,
  [1330] = 1321,
  [1331] = 1325,
  [1332] = 1332,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1333,
  [1336] = 1333,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1340,
  [1341] = 1341,
  [1342] = 1342,
  [1343] = 1340,
  [1344] = 1344,
  [1345] = 315,
  [1346] = 1340,
  [1347] = 1347,
  [1348] = 1342,
  [1349] = 1340,
  [1350] = 1350,
  [1351] = 1351,
  [1352] = 1352,
  [1353] = 1353,
  [1354] = 1340,
  [1355] = 1353,
  [1356] = 1340,
  [1357] = 1340,
  [1358] = 1340,
  [1359] = 1340,
  [1360] = 298,
  [1361] = 1340,
  [1362] = 1340,
  [1363] = 1340,
  [1364] = 1340,
  [1365] = 1340,
  [1366] = 1347,
  [1367] = 1341,
  [1368] = 1338,
  [1369] = 1340,
  [1370] = 1370,
  [1371] = 301,
  [1372] = 303,
  [1373] = 313,
  [1374] = 1374,
  [1375] = 1375,
  [1376] = 1351,
  [1377] = 1374,
  [1378] = 1375,
  [1379] = 963,
  [1380] = 303,
  [1381] = 1340,
  [1382] = 1382,
  [1383] = 297,
  [1384] = 297,
  [1385] = 1332,
  [1386] = 298,
  [1387] = 301,
  [1388] = 1340,
  [1389] = 1333,
  [1390] = 1340,
  [1391] = 1391,
  [1392] = 1392,
  [1393] = 1393,
  [1394] = 1394,
  [1395] = 1394,
  [1396] = 1396,
  [1397] = 1397,
  [1398] = 1392,
  [1399] = 963,
  [1400] = 1400,
  [1401] = 1401,
  [1402] = 1394,
  [1403] = 1393,
  [1404] = 1404,
  [1405] = 963,
  [1406] = 1397,
  [1407] = 1396,
  [1408] = 1408,
  [1409] = 1408,
  [1410] = 1396,
  [1411] = 1411,
  [1412] = 1397,
  [1413] = 1413,
  [1414] = 1396,
  [1415] = 1397,
  [1416] = 1408,
  [1417] = 1394,
  [1418] = 1408,
  [1419] = 1419,
  [1420] = 1420,
  [1421] = 1421,
  [1422] = 1422,
  [1423] = 1423,
  [1424] = 1424,
  [1425] = 1425,
  [1426] = 1425,
  [1427] = 1427,
  [1428] = 1428,
  [1429] = 1427,
  [1430] = 1430,
  [1431] = 1431,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1435,
  [1436] = 1436,
  [1437] = 1425,
  [1438] = 1438,
  [1439] = 1439,
  [1440] = 1440,
  [1441] = 1441,
  [1442] = 1421,
  [1443] = 1424,
  [1444] = 1425,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 314,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1450,
  [1454] = 1427,
  [1455] = 1455,
  [1456] = 1450,
  [1457] = 1451,
  [1458] = 1458,
  [1459] = 1432,
  [1460] = 1432,
  [1461] = 1434,
  [1462] = 1451,
  [1463] = 314,
  [1464] = 1434,
  [1465] = 1439,
  [1466] = 1425,
  [1467] = 1441,
  [1468] = 1421,
  [1469] = 1469,
  [1470] = 1424,
  [1471] = 1471,
  [1472] = 1427,
  [1473] = 1445,
  [1474] = 1427,
  [1475] = 1439,
  [1476] = 1455,
  [1477] = 317,
  [1478] = 1478,
  [1479] = 1479,
  [1480] = 1480,
  [1481] = 1425,
  [1482] = 1482,
  [1483] = 310,
  [1484] = 1425,
  [1485] = 1441,
  [1486] = 1486,
  [1487] = 1432,
  [1488] = 1434,
  [1489] = 1489,
  [1490] = 1439,
  [1491] = 1441,
  [1492] = 1421,
  [1493] = 317,
  [1494] = 1424,
  [1495] = 1445,
  [1496] = 1450,
  [1497] = 1497,
  [1498] = 311,
  [1499] = 1451,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 1445,
  [1503] = 1427,
  [1504] = 310,
  [1505] = 1505,
  [1506] = 311,
  [1507] = 1507,
  [1508] = 1451,
  [1509] = 312,
  [1510] = 1427,
  [1511] = 1427,
  [1512] = 1512,
  [1513] = 312,
  [1514] = 1425,
  [1515] = 1450,
  [1516] = 1479,
  [1517] = 1479,
  [1518] = 1479,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1522,
  [1529] = 1523,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 1534,
  [1535] = 1535,
  [1536] = 1536,
  [1537] = 1537,
  [1538] = 1538,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1543,
  [1544] = 1524,
  [1545] = 1530,
  [1546] = 1546,
  [1547] = 1547,
  [1548] = 1542,
  [1549] = 1549,
  [1550] = 1550,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1556,
  [1557] = 1557,
  [1558] = 1558,
  [1559] = 1559,
  [1560] = 1560,
  [1561] = 1561,
  [1562] = 1562,
  [1563] = 1563,
  [1564] = 1525,
  [1565] = 1565,
  [1566] = 1546,
  [1567] = 1540,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1541,
  [1572] = 1549,
  [1573] = 1526,
  [1574] = 1527,
  [1575] = 1522,
  [1576] = 1523,
  [1577] = 1577,
  [1578] = 1533,
  [1579] = 1535,
  [1580] = 1536,
  [1581] = 1537,
  [1582] = 1582,
  [1583] = 1538,
  [1584] = 1539,
  [1585] = 1520,
  [1586] = 1586,
  [1587] = 1540,
  [1588] = 1541,
  [1589] = 1543,
  [1590] = 1524,
  [1591] = 1530,
  [1592] = 1546,
  [1593] = 1542,
  [1594] = 1549,
  [1595] = 1550,
  [1596] = 1551,
  [1597] = 1552,
  [1598] = 1553,
  [1599] = 1554,
  [1600] = 1555,
  [1601] = 1556,
  [1602] = 1557,
  [1603] = 1603,
  [1604] = 1559,
  [1605] = 1560,
  [1606] = 1561,
  [1607] = 1562,
  [1608] = 1563,
  [1609] = 1525,
  [1610] = 1537,
  [1611] = 1554,
  [1612] = 1555,
  [1613] = 1556,
  [1614] = 1557,
  [1615] = 1535,
  [1616] = 1527,
  [1617] = 1617,
  [1618] = 1526,
  [1619] = 1527,
  [1620] = 1522,
  [1621] = 1523,
  [1622] = 1622,
  [1623] = 1537,
  [1624] = 1624,
  [1625] = 1532,
  [1626] = 1537,
  [1627] = 1533,
  [1628] = 1628,
  [1629] = 1535,
  [1630] = 1536,
  [1631] = 1631,
  [1632] = 1041,
  [1633] = 1537,
  [1634] = 1532,
  [1635] = 1635,
  [1636] = 1550,
  [1637] = 1536,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1537,
  [1641] = 1538,
  [1642] = 1539,
  [1643] = 1540,
  [1644] = 1541,
  [1645] = 1645,
  [1646] = 1559,
  [1647] = 1647,
  [1648] = 1639,
  [1649] = 1649,
  [1650] = 1543,
  [1651] = 1524,
  [1652] = 1530,
  [1653] = 1546,
  [1654] = 1542,
  [1655] = 1549,
  [1656] = 1550,
  [1657] = 1551,
  [1658] = 1639,
  [1659] = 1552,
  [1660] = 1553,
  [1661] = 1538,
  [1662] = 1554,
  [1663] = 1555,
  [1664] = 1556,
  [1665] = 1557,
  [1666] = 1559,
  [1667] = 1560,
  [1668] = 1561,
  [1669] = 1669,
  [1670] = 1562,
  [1671] = 1563,
  [1672] = 1672,
  [1673] = 1525,
  [1674] = 1551,
  [1675] = 1560,
  [1676] = 1561,
  [1677] = 1562,
  [1678] = 1678,
  [1679] = 1563,
  [1680] = 1639,
  [1681] = 1552,
  [1682] = 1526,
  [1683] = 1683,
  [1684] = 1553,
  [1685] = 1533,
  [1686] = 1539,
  [1687] = 1687,
  [1688] = 1638,
  [1689] = 1638,
  [1690] = 1638,
  [1691] = 1520,
  [1692] = 1520,
  [1693] = 1635,
  [1694] = 1635,
  [1695] = 1635,
  [1696] = 1543,
  [1697] = 1537,
  [1698] = 1698,
  [1699] = 1699,
  [1700] = 1700,
  [1701] = 368,
  [1702] = 1702,
  [1703] = 348,
  [1704] = 370,
  [1705] = 1705,
  [1706] = 1706,
  [1707] = 1707,
  [1708] = 339,
  [1709] = 1709,
  [1710] = 1710,
  [1711] = 1711,
  [1712] = 1706,
  [1713] = 1713,
  [1714] = 331,
  [1715] = 332,
  [1716] = 371,
  [1717] = 1717,
  [1718] = 1709,
  [1719] = 1710,
  [1720] = 1705,
  [1721] = 1721,
  [1722] = 372,
  [1723] = 1705,
  [1724] = 1724,
  [1725] = 340,
  [1726] = 342,
  [1727] = 1727,
  [1728] = 1728,
  [1729] = 1727,
  [1730] = 1709,
  [1731] = 1731,
  [1732] = 1732,
  [1733] = 349,
  [1734] = 1705,
  [1735] = 329,
  [1736] = 1736,
  [1737] = 1705,
  [1738] = 365,
  [1739] = 1728,
  [1740] = 1705,
  [1741] = 1741,
  [1742] = 1742,
  [1743] = 1705,
  [1744] = 333,
  [1745] = 1745,
  [1746] = 1705,
  [1747] = 1747,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1750,
  [1751] = 1751,
  [1752] = 1727,
  [1753] = 1728,
  [1754] = 1754,
  [1755] = 349,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1756,
  [1760] = 1757,
  [1761] = 1698,
  [1762] = 1749,
  [1763] = 1758,
  [1764] = 373,
  [1765] = 1700,
  [1766] = 1766,
  [1767] = 1767,
  [1768] = 1706,
  [1769] = 361,
  [1770] = 1770,
  [1771] = 1698,
  [1772] = 362,
  [1773] = 1705,
  [1774] = 367,
  [1775] = 1775,
  [1776] = 1776,
  [1777] = 375,
  [1778] = 1702,
  [1779] = 1727,
  [1780] = 1728,
  [1781] = 369,
  [1782] = 1756,
  [1783] = 1757,
  [1784] = 1758,
  [1785] = 334,
  [1786] = 1786,
  [1787] = 1698,
  [1788] = 1749,
  [1789] = 374,
  [1790] = 1766,
  [1791] = 1706,
  [1792] = 1792,
  [1793] = 1793,
  [1794] = 365,
  [1795] = 1795,
  [1796] = 1796,
  [1797] = 1797,
  [1798] = 1705,
  [1799] = 376,
  [1800] = 1727,
  [1801] = 1727,
  [1802] = 1728,
  [1803] = 1728,
  [1804] = 1756,
  [1805] = 1757,
  [1806] = 1758,
  [1807] = 1807,
  [1808] = 1808,
  [1809] = 1698,
  [1810] = 1749,
  [1811] = 1811,
  [1812] = 1766,
  [1813] = 1706,
  [1814] = 361,
  [1815] = 1815,
  [1816] = 1702,
  [1817] = 1817,
  [1818] = 1756,
  [1819] = 1757,
  [1820] = 1758,
  [1821] = 1821,
  [1822] = 1822,
  [1823] = 1727,
  [1824] = 1728,
  [1825] = 362,
  [1826] = 1756,
  [1827] = 1757,
  [1828] = 1758,
  [1829] = 1829,
  [1830] = 352,
  [1831] = 1698,
  [1832] = 1749,
  [1833] = 1766,
  [1834] = 1706,
  [1835] = 1835,
  [1836] = 1705,
  [1837] = 1727,
  [1838] = 1728,
  [1839] = 1756,
  [1840] = 1698,
  [1841] = 1717,
  [1842] = 1706,
  [1843] = 1705,
  [1844] = 1844,
  [1845] = 1727,
  [1846] = 1728,
  [1847] = 1756,
  [1848] = 1698,
  [1849] = 1706,
  [1850] = 1850,
  [1851] = 1710,
  [1852] = 1727,
  [1853] = 1728,
  [1854] = 1756,
  [1855] = 1698,
  [1856] = 1706,
  [1857] = 1857,
  [1858] = 1727,
  [1859] = 1728,
  [1860] = 1860,
  [1861] = 1706,
  [1862] = 1862,
  [1863] = 1863,
  [1864] = 1698,
  [1865] = 1865,
  [1866] = 1717,
  [1867] = 1698,
  [1868] = 1868,
  [1869] = 1863,
  [1870] = 1766,
  [1871] = 1871,
  [1872] = 1749,
  [1873] = 1873,
  [1874] = 1874,
  [1875] = 1705,
  [1876] = 1876,
  [1877] = 1713,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 344,
  [1884] = 364,
  [1885] = 1885,
  [1886] = 1886,
  [1887] = 341,
  [1888] = 1873,
  [1889] = 1705,
  [1890] = 1741,
  [1891] = 1891,
  [1892] = 1857,
  [1893] = 1756,
  [1894] = 1868,
  [1895] = 1707,
  [1896] = 1896,
  [1897] = 367,
  [1898] = 1766,
  [1899] = 1899,
  [1900] = 1900,
  [1901] = 1901,
  [1902] = 343,
  [1903] = 1705,
  [1904] = 1904,
  [1905] = 1905,
  [1906] = 1906,
  [1907] = 299,
  [1908] = 1749,
  [1909] = 369,
  [1910] = 1844,
  [1911] = 1865,
  [1912] = 1878,
  [1913] = 1700,
  [1914] = 1914,
  [1915] = 337,
  [1916] = 1916,
  [1917] = 341,
  [1918] = 1918,
  [1919] = 343,
  [1920] = 1920,
  [1921] = 345,
  [1922] = 1922,
  [1923] = 348,
  [1924] = 1924,
  [1925] = 1706,
  [1926] = 1713,
  [1927] = 1863,
  [1928] = 1918,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1709,
  [1932] = 1710,
  [1933] = 346,
  [1934] = 1742,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1873,
  [1938] = 1938,
  [1939] = 1741,
  [1940] = 1857,
  [1941] = 355,
  [1942] = 1868,
  [1943] = 1707,
  [1944] = 1702,
  [1945] = 347,
  [1946] = 356,
  [1947] = 1947,
  [1948] = 1901,
  [1949] = 358,
  [1950] = 1705,
  [1951] = 350,
  [1952] = 1905,
  [1953] = 1906,
  [1954] = 1727,
  [1955] = 1728,
  [1956] = 335,
  [1957] = 1844,
  [1958] = 1865,
  [1959] = 1878,
  [1960] = 1960,
  [1961] = 1914,
  [1962] = 1962,
  [1963] = 345,
  [1964] = 336,
  [1965] = 363,
  [1966] = 1756,
  [1967] = 1905,
  [1968] = 1757,
  [1969] = 1758,
  [1970] = 328,
  [1971] = 1900,
  [1972] = 1972,
  [1973] = 351,
  [1974] = 338,
  [1975] = 304,
  [1976] = 1976,
  [1977] = 1896,
  [1978] = 354,
  [1979] = 1906,
  [1980] = 1698,
  [1981] = 1873,
  [1982] = 355,
  [1983] = 1857,
  [1984] = 356,
  [1985] = 1868,
  [1986] = 1707,
  [1987] = 377,
  [1988] = 1766,
  [1989] = 358,
  [1990] = 1990,
  [1991] = 1901,
  [1992] = 1711,
  [1993] = 359,
  [1994] = 1705,
  [1995] = 1905,
  [1996] = 1906,
  [1997] = 363,
  [1998] = 328,
  [1999] = 360,
  [2000] = 1844,
  [2001] = 1865,
  [2002] = 1878,
  [2003] = 2003,
  [2004] = 1914,
  [2005] = 368,
  [2006] = 2006,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 370,
  [2010] = 371,
  [2011] = 2011,
  [2012] = 2012,
  [2013] = 2013,
  [2014] = 1749,
  [2015] = 1901,
  [2016] = 373,
  [2017] = 375,
  [2018] = 376,
  [2019] = 352,
  [2020] = 330,
  [2021] = 1904,
  [2022] = 344,
  [2023] = 364,
  [2024] = 2024,
  [2025] = 300,
  [2026] = 2026,
  [2027] = 337,
  [2028] = 1757,
  [2029] = 346,
  [2030] = 347,
  [2031] = 350,
  [2032] = 351,
  [2033] = 377,
  [2034] = 359,
  [2035] = 360,
  [2036] = 1776,
  [2037] = 1904,
  [2038] = 1916,
  [2039] = 1808,
  [2040] = 1829,
  [2041] = 1891,
  [2042] = 1914,
  [2043] = 366,
  [2044] = 2044,
  [2045] = 1962,
  [2046] = 1936,
  [2047] = 372,
  [2048] = 1935,
  [2049] = 329,
  [2050] = 1807,
  [2051] = 2051,
  [2052] = 330,
  [2053] = 1745,
  [2054] = 1751,
  [2055] = 1976,
  [2056] = 1776,
  [2057] = 1705,
  [2058] = 1916,
  [2059] = 1808,
  [2060] = 1829,
  [2061] = 331,
  [2062] = 332,
  [2063] = 2044,
  [2064] = 1962,
  [2065] = 1936,
  [2066] = 333,
  [2067] = 1935,
  [2068] = 334,
  [2069] = 1807,
  [2070] = 2051,
  [2071] = 335,
  [2072] = 1745,
  [2073] = 1751,
  [2074] = 1976,
  [2075] = 1776,
  [2076] = 2076,
  [2077] = 1916,
  [2078] = 1808,
  [2079] = 336,
  [2080] = 366,
  [2081] = 2044,
  [2082] = 1962,
  [2083] = 1936,
  [2084] = 338,
  [2085] = 1935,
  [2086] = 339,
  [2087] = 1807,
  [2088] = 2051,
  [2089] = 340,
  [2090] = 1745,
  [2091] = 1751,
  [2092] = 1976,
  [2093] = 1811,
  [2094] = 1821,
  [2095] = 2095,
  [2096] = 2013,
  [2097] = 1775,
  [2098] = 1792,
  [2099] = 1811,
  [2100] = 1821,
  [2101] = 342,
  [2102] = 2013,
  [2103] = 1775,
  [2104] = 1792,
  [2105] = 1811,
  [2106] = 1821,
  [2107] = 374,
  [2108] = 2013,
  [2109] = 1775,
  [2110] = 1792,
  [2111] = 1706,
  [2112] = 2051,
  [2113] = 2113,
  [2114] = 1766,
  [2115] = 2115,
  [2116] = 2116,
  [2117] = 1700,
  [2118] = 2044,
  [2119] = 1758,
  [2120] = 1904,
  [2121] = 1796,
  [2122] = 2122,
  [2123] = 1796,
  [2124] = 2124,
  [2125] = 1796,
  [2126] = 354,
};

static const TSCharacterRange sym_generic_token_character_set_1[] = {
  {0, 0x08}, {0x0e, 0x1f}, {'!', '!'}, {'#', '#'}, {'%', '%'}, {'*', ','}, {'.', '?'}, {'A', 'Z'},
  {'\\', '_'}, {'a', 'z'}, {'~', 0x10ffff},
};

static const TSCharacterRange aux_sym_command_name_token1_character_set_1[] = {
  {0, 0x08}, {0x0e, 0x1f}, {'#', '#'}, {'%', '%'}, {'.', '.'}, {'0', ':'}, {'=', '?'}, {'A', 'Z'},
  {'\\', '\\'}, {'^', '_'}, {'a', 'z'}, {'~', 0x10ffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1846,
        '"', 1772,
        '#', 523,
        '$', 654,
        '%', 1839,
        '&', 1372,
        '\'', 179,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1824,
        '.', 1870,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        ';', 1302,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(500);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(736);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(713);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(717);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(769);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(755);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(747);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(719);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(738);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(737);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(752);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(783);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(753);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(513);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(129);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(172);
      if (lookahead == '#') ADVANCE(171);
      if (lookahead == '`') ADVANCE(103);
      if (lookahead != 0) ADVANCE(172);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(130);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(130);
      if (lookahead == '\r') SKIP(2);
      END_STATE();
    case 4:
      if (lookahead == '\n') ADVANCE(658);
      END_STATE();
    case 5:
      if (lookahead == '\n') ADVANCE(658);
      if (lookahead == '\r') ADVANCE(4);
      if (lookahead == ' ') ADVANCE(5);
      END_STATE();
    case 6:
      if (lookahead == '\n') ADVANCE(9);
      if (lookahead == '\r') ADVANCE(6);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(6);
      END_STATE();
    case 7:
      if (lookahead == '\n') ADVANCE(8);
      if (lookahead == '\r') ADVANCE(6);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(6);
      END_STATE();
    case 8:
      if (lookahead == '\n') ADVANCE(8);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '\'') ADVANCE(73);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(9);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 9:
      if (lookahead == '\n') ADVANCE(8);
      if (lookahead == '\r') ADVANCE(7);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(9);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 10:
      if (lookahead == '\n') ADVANCE(85);
      END_STATE();
    case 11:
      if (lookahead == '\n') ADVANCE(85);
      if (lookahead == '\r') ADVANCE(10);
      if (lookahead == '\'') ADVANCE(73);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 12:
      if (lookahead == '\n') ADVANCE(85);
      if (lookahead == '\r') ADVANCE(10);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 13:
      if (lookahead == '\n') SKIP(139);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(139);
      if (lookahead == '\r') SKIP(13);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(140);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(140);
      if (lookahead == '\r') SKIP(15);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(144);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(144);
      if (lookahead == '\r') SKIP(17);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '\n', 1340,
        '\r', 21,
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(24);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1669);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1616);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1639);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1596);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1568);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1535);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1566);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1708);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1579);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1580);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(137);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1400);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 20:
      if (lookahead == '\n') ADVANCE(1340);
      if (lookahead == '\r') ADVANCE(22);
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == ')') ADVANCE(1306);
      if (lookahead == ';') ADVANCE(1302);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(70);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(166);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '\n', 1340,
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(24);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1669);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1616);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1639);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1596);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1568);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1535);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1566);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1708);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1579);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1580);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1400);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 22:
      if (lookahead == '\n') ADVANCE(1340);
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == ')') ADVANCE(1306);
      if (lookahead == ';') ADVANCE(1302);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(70);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(166);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(137);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(137);
      if (lookahead == '\r') SKIP(23);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(138);
      END_STATE();
    case 26:
      if (lookahead == '\n') SKIP(138);
      if (lookahead == '\r') SKIP(25);
      END_STATE();
    case 27:
      if (lookahead == '\n') SKIP(126);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(126);
      if (lookahead == '\r') SKIP(27);
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(145);
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(145);
      if (lookahead == '\r') SKIP(29);
      END_STATE();
    case 31:
      if (lookahead == '\n') SKIP(150);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(150);
      if (lookahead == '\r') SKIP(31);
      END_STATE();
    case 33:
      if (lookahead == '\n') SKIP(127);
      END_STATE();
    case 34:
      if (lookahead == '\n') SKIP(127);
      if (lookahead == '\r') SKIP(33);
      END_STATE();
    case 35:
      if (lookahead == '\n') SKIP(113);
      END_STATE();
    case 36:
      if (lookahead == '\n') SKIP(113);
      if (lookahead == '\r') SKIP(35);
      END_STATE();
    case 37:
      if (lookahead == '\n') SKIP(122);
      END_STATE();
    case 38:
      if (lookahead == '\n') SKIP(122);
      if (lookahead == '\r') SKIP(37);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(116);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(116);
      if (lookahead == '\r') SKIP(39);
      END_STATE();
    case 41:
      if (lookahead == '\n') SKIP(161);
      END_STATE();
    case 42:
      if (lookahead == '\n') SKIP(161);
      if (lookahead == '\r') SKIP(41);
      END_STATE();
    case 43:
      if (lookahead == '\n') SKIP(109);
      END_STATE();
    case 44:
      if (lookahead == '\n') SKIP(109);
      if (lookahead == '\r') SKIP(43);
      END_STATE();
    case 45:
      if (lookahead == '\n') SKIP(119);
      END_STATE();
    case 46:
      if (lookahead == '\n') SKIP(119);
      if (lookahead == '\r') SKIP(45);
      END_STATE();
    case 47:
      if (lookahead == '\n') SKIP(124);
      END_STATE();
    case 48:
      if (lookahead == '\n') SKIP(124);
      if (lookahead == '\r') SKIP(47);
      END_STATE();
    case 49:
      if (lookahead == '\n') SKIP(173);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(173);
      if (lookahead == '\r') SKIP(49);
      END_STATE();
    case 51:
      if (lookahead == '\n') ADVANCE(1385);
      if (lookahead == '\r') ADVANCE(1384);
      if (lookahead != 0) ADVANCE(1383);
      END_STATE();
    case 52:
      if (lookahead == '\n') ADVANCE(1381);
      END_STATE();
    case 53:
      if (lookahead == '\n') ADVANCE(649);
      if (lookahead == '\r') ADVANCE(650);
      if (lookahead != 0) ADVANCE(649);
      END_STATE();
    case 54:
      ADVANCE_MAP(
        '\n', 54,
        '\r', 56,
        '"', 213,
        '#', 704,
        '$', 702,
        '<', 701,
        '@', 703,
        '`', 700,
        '\t', 700,
        0x0b, 700,
        '\f', 700,
        ' ', 700,
        0xa0, 700,
        0x200b, 700,
        0x2060, 700,
        0xfeff, 700,
      );
      if (lookahead != 0) ADVANCE(700);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        '\n', 54,
        '\r', 56,
        '#', 697,
        '$', 654,
        '<', 660,
        '@', 695,
        '`', 53,
        '\t', 659,
        0x0b, 659,
        '\f', 659,
        ' ', 659,
        0xa0, 659,
        0x200b, 659,
        0x2060, 659,
        0xfeff, 659,
      );
      if (lookahead != 0) ADVANCE(699);
      END_STATE();
    case 56:
      if (lookahead == '\n') ADVANCE(54);
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == '$') ADVANCE(175);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '@') ADVANCE(208);
      if (lookahead == '`') SKIP(58);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      END_STATE();
    case 57:
      if (lookahead == '\n') SKIP(159);
      END_STATE();
    case 58:
      if (lookahead == '\n') SKIP(159);
      if (lookahead == '\r') SKIP(57);
      END_STATE();
    case 59:
      if (lookahead == '\n') SKIP(164);
      END_STATE();
    case 60:
      if (lookahead == '\n') SKIP(164);
      if (lookahead == '\r') SKIP(59);
      END_STATE();
    case 61:
      if (lookahead == '\n') SKIP(162);
      END_STATE();
    case 62:
      if (lookahead == '\n') SKIP(162);
      if (lookahead == '\r') SKIP(61);
      END_STATE();
    case 63:
      if (lookahead == '\n') SKIP(167);
      END_STATE();
    case 64:
      if (lookahead == '\n') SKIP(167);
      if (lookahead == '\r') SKIP(63);
      END_STATE();
    case 65:
      if (lookahead == '\n') SKIP(165);
      END_STATE();
    case 66:
      if (lookahead == '\n') SKIP(165);
      if (lookahead == '\r') SKIP(65);
      END_STATE();
    case 67:
      if (lookahead == '\n') SKIP(169);
      END_STATE();
    case 68:
      if (lookahead == '\n') SKIP(169);
      if (lookahead == '\r') SKIP(67);
      END_STATE();
    case 69:
      if (lookahead == '\n') SKIP(166);
      END_STATE();
    case 70:
      if (lookahead == '\n') SKIP(166);
      if (lookahead == '\r') SKIP(69);
      END_STATE();
    case 71:
      if (lookahead == '\n') SKIP(168);
      END_STATE();
    case 72:
      if (lookahead == '\n') SKIP(168);
      if (lookahead == '\r') SKIP(71);
      END_STATE();
    case 73:
      if (lookahead == '\n') ADVANCE(11);
      if (lookahead == '\r') ADVANCE(12);
      if (lookahead == '@') ADVANCE(711);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 74:
      if (lookahead == '\n') SKIP(132);
      END_STATE();
    case 75:
      if (lookahead == '\n') SKIP(132);
      if (lookahead == '\r') SKIP(74);
      END_STATE();
    case 76:
      if (lookahead == '\n') SKIP(143);
      END_STATE();
    case 77:
      if (lookahead == '\n') SKIP(143);
      if (lookahead == '\r') SKIP(76);
      END_STATE();
    case 78:
      if (lookahead == '\n') SKIP(105);
      END_STATE();
    case 79:
      if (lookahead == '\n') SKIP(105);
      if (lookahead == '\r') SKIP(78);
      END_STATE();
    case 80:
      if (lookahead == '\n') SKIP(133);
      END_STATE();
    case 81:
      if (lookahead == '\n') SKIP(133);
      if (lookahead == '\r') SKIP(80);
      END_STATE();
    case 82:
      if (lookahead == '\n') SKIP(135);
      END_STATE();
    case 83:
      if (lookahead == '\n') SKIP(135);
      if (lookahead == '\r') SKIP(82);
      END_STATE();
    case 84:
      if (lookahead == '\n') ADVANCE(86);
      END_STATE();
    case 85:
      if (lookahead == '\n') ADVANCE(86);
      if (lookahead == '\r') ADVANCE(84);
      if (lookahead == '\'') ADVANCE(212);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 86:
      if (lookahead == '\n') ADVANCE(86);
      if (lookahead == '\r') ADVANCE(84);
      if (lookahead == '\'') ADVANCE(211);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 87:
      if (lookahead == '\n') SKIP(128);
      END_STATE();
    case 88:
      if (lookahead == '\n') SKIP(128);
      if (lookahead == '\r') SKIP(87);
      END_STATE();
    case 89:
      if (lookahead == '\n') SKIP(115);
      END_STATE();
    case 90:
      if (lookahead == '\n') SKIP(115);
      if (lookahead == '\r') SKIP(89);
      END_STATE();
    case 91:
      if (lookahead == '\n') SKIP(117);
      END_STATE();
    case 92:
      if (lookahead == '\n') SKIP(117);
      if (lookahead == '\r') SKIP(91);
      END_STATE();
    case 93:
      if (lookahead == '\n') SKIP(121);
      END_STATE();
    case 94:
      if (lookahead == '\n') SKIP(121);
      if (lookahead == '\r') SKIP(93);
      END_STATE();
    case 95:
      if (lookahead == '\n') SKIP(125);
      END_STATE();
    case 96:
      if (lookahead == '\n') SKIP(125);
      if (lookahead == '\r') SKIP(95);
      END_STATE();
    case 97:
      if (lookahead == '\n') SKIP(107);
      END_STATE();
    case 98:
      if (lookahead == '\n') SKIP(107);
      if (lookahead == '\r') SKIP(97);
      END_STATE();
    case 99:
      if (lookahead == '\n') SKIP(123);
      END_STATE();
    case 100:
      if (lookahead == '\n') SKIP(123);
      if (lookahead == '\r') SKIP(99);
      END_STATE();
    case 101:
      if (lookahead == '\n') SKIP(141);
      END_STATE();
    case 102:
      if (lookahead == '\n') SKIP(141);
      if (lookahead == '\r') SKIP(101);
      END_STATE();
    case 103:
      if (lookahead == '\r') ADVANCE(1);
      if (lookahead != 0) ADVANCE(172);
      END_STATE();
    case 104:
      if (lookahead == '\r') ADVANCE(648);
      if (lookahead != 0) ADVANCE(647);
      END_STATE();
    case 105:
      ADVANCE_MAP(
        ' ', 1798,
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1829,
        '.', 495,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(79);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(105);
      END_STATE();
    case 106:
      ADVANCE_MAP(
        ' ', 1798,
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1829,
        '.', 1869,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(79);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(105);
      END_STATE();
    case 107:
      ADVANCE_MAP(
        ' ', 1798,
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1832,
        '.', 495,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(98);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(107);
      END_STATE();
    case 108:
      ADVANCE_MAP(
        ' ', 1798,
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1832,
        '.', 1869,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(98);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(107);
      END_STATE();
    case 109:
      ADVANCE_MAP(
        ' ', 1798,
        '"', 600,
        '#', 533,
        '$', 175,
        '\'', 180,
        '(', 1305,
        '.', 495,
        '0', 545,
        ':', 1802,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(44);
      if (lookahead == '{') ADVANCE(1318);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(109);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 110:
      ADVANCE_MAP(
        ' ', 1798,
        '"', 1773,
        '#', 1768,
        '&', 177,
        '\'', 179,
        '(', 1305,
        ')', 1306,
        '*', 1761,
        '-', 1758,
        '1', 1762,
        '2', 1763,
        '3', 1764,
        '4', 1765,
        '5', 1766,
        '6', 1767,
        ':', 1804,
        '<', 881,
        '>', 839,
        '`', 1771,
        '{', 1318,
        '|', 1370,
        0xa0, 1749,
        0x200b, 1749,
        0x2060, 1749,
        0xfeff, 1749,
      );
      if (('\t' <= lookahead && lookahead <= '\r')) SKIP(124);
      if (lookahead != 0 &&
          lookahead != ',' &&
          lookahead != '-' &&
          (lookahead < ':' || '<' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1771);
      END_STATE();
    case 111:
      ADVANCE_MAP(
        ' ', 1798,
        '"', 1773,
        '#', 1768,
        '&', 177,
        '\'', 179,
        '(', 1305,
        ')', 1306,
        '*', 1761,
        '-', 1760,
        '1', 1762,
        '2', 1763,
        '3', 1764,
        '4', 1765,
        '5', 1766,
        '6', 1767,
        ':', 1804,
        '<', 881,
        '>', 839,
        '`', 1771,
        '{', 1318,
        '|', 1370,
        0xa0, 1750,
        0x200b, 1750,
        0x2060, 1750,
        0xfeff, 1750,
      );
      if (('\t' <= lookahead && lookahead <= '\r')) SKIP(125);
      if (lookahead != 0 &&
          lookahead != ',' &&
          lookahead != '-' &&
          (lookahead < ':' || '<' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1771);
      END_STATE();
    case 112:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        '-', 1776,
        '.', 1867,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(36);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(113);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 113:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        '-', 1776,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(36);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(113);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 114:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        '-', 1780,
        '.', 1867,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(90);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(115);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 115:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        '-', 1780,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1803,
        '<', 879,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(90);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(115);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 116:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '-', 1777,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1802,
        '<', 879,
        '>', 837,
        '@', 208,
      );
      if (lookahead == '`') SKIP(40);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(116);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 117:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '-', 1781,
        '1', 1782,
        '2', 1783,
        '3', 1784,
        '4', 1785,
        '5', 1786,
        '6', 1787,
        ':', 1802,
        '<', 879,
        '>', 837,
        '@', 208,
      );
      if (lookahead == '`') SKIP(92);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 118:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 183,
        '.', 1868,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1803,
        '<', 879,
        '>', 837,
        '[', 827,
      );
      if (lookahead == '`') SKIP(46);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(119);
      END_STATE();
    case 119:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 183,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1803,
        '<', 879,
        '>', 837,
        '[', 827,
      );
      if (lookahead == '`') SKIP(46);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(119);
      END_STATE();
    case 120:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 185,
        '.', 1868,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1803,
        '<', 879,
        '>', 837,
        '[', 827,
      );
      if (lookahead == '`') SKIP(94);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(121);
      END_STATE();
    case 121:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 185,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1803,
        '<', 879,
        '>', 837,
        '[', 827,
      );
      if (lookahead == '`') SKIP(94);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(121);
      END_STATE();
    case 122:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        ',', 1308,
        '-', 184,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1802,
        '<', 879,
        '>', 837,
      );
      if (lookahead == '`') SKIP(38);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(122);
      END_STATE();
    case 123:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        ',', 1308,
        '-', 186,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1802,
        '<', 879,
        '>', 837,
      );
      if (lookahead == '`') SKIP(100);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(123);
      END_STATE();
    case 124:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '-', 184,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1802,
        '<', 879,
        '>', 837,
      );
      if (lookahead == '`') SKIP(48);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(124);
      END_STATE();
    case 125:
      ADVANCE_MAP(
        ' ', 1798,
        '#', 533,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '-', 186,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1802,
        '<', 879,
        '>', 837,
      );
      if (lookahead == '`') SKIP(96);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(125);
      END_STATE();
    case 126:
      ADVANCE_MAP(
        ' ', 1799,
        '!', 1847,
        '"', 600,
        '#', 527,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1829,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ':', 1805,
        '<', 880,
        '>', 838,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(28);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1199);
      if (('\t' <= lookahead && lookahead <= '\r')) SKIP(126);
      if (lookahead != 0 &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 127:
      ADVANCE_MAP(
        ' ', 1800,
        '!', 1847,
        '"', 600,
        '#', 527,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1828,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        '<', 880,
        '>', 838,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(34);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1200);
      if (('\t' <= lookahead && lookahead <= '\r')) SKIP(127);
      if (lookahead != 0 &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 128:
      ADVANCE_MAP(
        ' ', 1801,
        '!', 1847,
        '"', 600,
        '#', 527,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1832,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ':', 1805,
        '<', 880,
        '>', 838,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(88);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1201);
      if (('\t' <= lookahead && lookahead <= '\r')) SKIP(128);
      if (lookahead != 0 &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 129:
      ADVANCE_MAP(
        '!', 1846,
        '"', 1390,
        '#', 523,
        '$', 1389,
        '%', 1839,
        '&', 1372,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1824,
        '.', 824,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        ';', 1302,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(3);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(736);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(713);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(717);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(769);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(755);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(747);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(719);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(738);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(737);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(752);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(783);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(753);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(130);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(129);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 130:
      ADVANCE_MAP(
        '!', 1846,
        '"', 1390,
        '#', 533,
        '$', 1389,
        '%', 1839,
        '&', 1372,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1824,
        '.', 824,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        ';', 1302,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(3);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(736);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(713);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(717);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(769);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(755);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(747);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(719);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(738);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(737);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(752);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(783);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(753);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(130);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 131:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '%', 1838,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1840,
        '+', 1820,
        ',', 1308,
        '-', 1827,
        '.', 1866,
        '/', 1835,
        '0', 545,
        ':', 197,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(75);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(132);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 132:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '%', 1838,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1840,
        '+', 1820,
        ',', 1308,
        '-', 1827,
        '.', 188,
        '/', 1835,
        '0', 545,
        ':', 197,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(75);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(132);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 133:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 495,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 196,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(81);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      END_STATE();
    case 134:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 1869,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 196,
        '<', 879,
        '>', 837,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(81);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      END_STATE();
    case 135:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '\'', 180,
        '(', 1305,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 495,
        '0', 545,
        ':', 196,
        '<', 170,
        '=', 830,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(83);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(809);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(135);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 136:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 533,
        '$', 175,
        '\'', 180,
        '(', 1305,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 1869,
        '0', 545,
        ':', 196,
        '<', 170,
        '=', 830,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(83);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(809);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(135);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 137:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(24);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1669);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1616);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1639);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1596);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1568);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1535);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1566);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1708);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1579);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1580);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1400);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 138:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(26);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1669);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1616);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1639);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1596);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1568);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1535);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1566);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1708);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1579);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1580);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1400);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(138);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ';' &&
          lookahead != '<' &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 139:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(14);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1627);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1522);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1392);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 140:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(16);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1551);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1513);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1631);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1521);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1398);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(140);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 141:
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(102);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1551);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1513);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1631);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1532);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1399);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(141);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 142:
      ADVANCE_MAP(
        '!', 200,
        '"', 1390,
        '#', 533,
        '$', 1387,
        '%', 1839,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1826,
        '.', 1865,
        '/', 1836,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(77);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(389);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(143);
      END_STATE();
    case 143:
      ADVANCE_MAP(
        '!', 200,
        '"', 1390,
        '#', 533,
        '$', 1387,
        '%', 1839,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1826,
        '.', 187,
        '/', 1836,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(77);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(389);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(143);
      END_STATE();
    case 144:
      ADVANCE_MAP(
        '!', 200,
        '#', 533,
        '$', 176,
        '%', 1839,
        '&', 177,
        ')', 1306,
        '*', 1841,
        '+', 1823,
        ',', 1308,
        '-', 1834,
        '.', 187,
        '/', 1836,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 1802,
        ';', 1302,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 208,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(18);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(809);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 145:
      ADVANCE_MAP(
        '!', 1847,
        '"', 600,
        '#', 527,
        '$', 175,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1828,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        '<', 880,
        '>', 838,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(30);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1202);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(145);
      if (lookahead != 0 &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 146:
      ADVANCE_MAP(
        '"', 5,
        '\'', 6,
        '(', 1862,
        '?', 1063,
        '{', 1863,
        'G', 1108,
        'g', 1108,
        'L', 1114,
        'l', 1114,
        'P', 1116,
        'p', 1116,
        'S', 1099,
        's', 1099,
        'U', 1119,
        'u', 1119,
        'W', 1113,
        'w', 1113,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 147:
      if (lookahead == '"') ADVANCE(655);
      if (lookahead == '#') ADVANCE(523);
      if (lookahead == '$') ADVANCE(652);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(68);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(169);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(148);
      END_STATE();
    case 148:
      if (lookahead == '"') ADVANCE(655);
      if (lookahead == '#') ADVANCE(523);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(68);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(169);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(148);
      END_STATE();
    case 149:
      if (lookahead == '"') ADVANCE(655);
      if (lookahead == '#') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(149);
      END_STATE();
    case 150:
      ADVANCE_MAP(
        '"', 600,
        '#', 528,
        '$', 175,
        '\'', 180,
        '(', 1305,
        '-', 1796,
        '.', 1489,
        '0', 535,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(32);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1470);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1450);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1458);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1442);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1437);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1427);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1431);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1488);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1440);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1439);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1401);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(150);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(536);
      if (('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if (lookahead != 0 &&
          (lookahead < ' ' || '$' < lookahead) &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ';' &&
          lookahead != '<' &&
          (lookahead < '?' || ']' < lookahead) &&
          (lookahead < '_' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 151:
      ADVANCE_MAP(
        '"', 602,
        '#', 526,
        '$', 1256,
        '\'', 1257,
        '(', 1305,
        '.', 1266,
        '0', 557,
        ';', 1302,
        '<', 1253,
        '@', 1251,
        '[', 829,
        '`', 1269,
        '{', 1318,
        '}', 1320,
        0xa0, 1250,
        0x200b, 1250,
        0x2060, 1250,
        0xfeff, 1250,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(151);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(558);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead)) ADVANCE(1269);
      END_STATE();
    case 152:
      ADVANCE_MAP(
        '"', 1773,
        '#', 1768,
        '\'', 179,
        ',', 1308,
        '<', 1752,
        '`', 1771,
        '{', 1318,
        0xa0, 1751,
        0x200b, 1751,
        0x2060, 1751,
        0xfeff, 1751,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(167);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '<' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1771);
      END_STATE();
    case 153:
      if (lookahead == '"') ADVANCE(1391);
      if (lookahead == '#') ADVANCE(525);
      if (lookahead == '$') ADVANCE(1388);
      if (lookahead == '<') ADVANCE(1374);
      if (lookahead == '@') ADVANCE(1378);
      if (lookahead == '`') ADVANCE(51);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1373);
      if (lookahead != 0) ADVANCE(1380);
      END_STATE();
    case 154:
      ADVANCE_MAP(
        '"', 656,
        '#', 603,
        '$', 653,
        '<', 607,
        '@', 642,
        '`', 53,
        0xa0, 606,
        0x200b, 606,
        0x2060, 606,
        0xfeff, 606,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(605);
      if (lookahead != 0) ADVANCE(646);
      END_STATE();
    case 155:
      if (lookahead == '"') ADVANCE(1297);
      if (lookahead == '#') ADVANCE(524);
      if (lookahead == '&') ADVANCE(1291);
      if (lookahead == ')') ADVANCE(1307);
      if (lookahead == '<') ADVANCE(1298);
      if (lookahead == '`') ADVANCE(1301);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(155);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1294);
      if (lookahead != 0) ADVANCE(1301);
      END_STATE();
    case 156:
      if (lookahead == '"') ADVANCE(1297);
      if (lookahead == '#') ADVANCE(524);
      if (lookahead == '&') ADVANCE(1291);
      if (lookahead == '<') ADVANCE(1298);
      if (lookahead == '`') ADVANCE(1301);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(156);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1295);
      if (lookahead != 0) ADVANCE(1301);
      END_STATE();
    case 157:
      if (lookahead == '"') ADVANCE(1297);
      if (lookahead == '#') ADVANCE(524);
      if (lookahead == '&') ADVANCE(1290);
      if (lookahead == '<') ADVANCE(1298);
      if (lookahead == '`') ADVANCE(1301);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(157);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1296);
      if (lookahead != 0 &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 158:
      if (lookahead == '"') ADVANCE(1288);
      if (lookahead != 0) ADVANCE(158);
      END_STATE();
    case 159:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == '$') ADVANCE(175);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '@') ADVANCE(208);
      if (lookahead == '`') SKIP(58);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      END_STATE();
    case 160:
      ADVANCE_MAP(
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 182,
        '.', 1868,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(42);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(161);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 161:
      ADVANCE_MAP(
        '#', 533,
        '$', 176,
        '&', 177,
        '(', 1305,
        ')', 1306,
        '*', 201,
        '+', 181,
        ',', 1308,
        '-', 182,
        '1', 202,
        '2', 203,
        '3', 204,
        '4', 205,
        '5', 206,
        '6', 207,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(42);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(161);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 162:
      ADVANCE_MAP(
        '#', 533,
        '$', 176,
        '&', 177,
        ')', 1306,
        ',', 1308,
        '-', 1779,
        '<', 170,
        '@', 208,
      );
      if (lookahead == '`') SKIP(62);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(162);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 163:
      ADVANCE_MAP(
        '#', 533,
        '$', 176,
        '+', 181,
        ',', 1308,
        '-', 1778,
        '.', 1867,
        ':', 196,
        '<', 170,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(60);
      if (lookahead == '{') ADVANCE(1318);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(164);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 164:
      ADVANCE_MAP(
        '#', 533,
        '$', 176,
        '+', 181,
        ',', 1308,
        '-', 1778,
        ':', 196,
        '<', 170,
        '@', 208,
        '[', 827,
      );
      if (lookahead == '`') SKIP(60);
      if (lookahead == '{') ADVANCE(1318);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(164);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('?' <= lookahead && lookahead <= '\\') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 165:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == '(') ADVANCE(1305);
      if (lookahead == ',') ADVANCE(1308);
      if (lookahead == '.') ADVANCE(823);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '[') ADVANCE(827);
      if (lookahead == ']') ADVANCE(1872);
      if (lookahead == '`') SKIP(66);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(165);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(822);
      END_STATE();
    case 166:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == ')') ADVANCE(1306);
      if (lookahead == ';') ADVANCE(1302);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(70);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(166);
      END_STATE();
    case 167:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == ',') ADVANCE(1308);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(64);
      if (lookahead == '{') ADVANCE(1318);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(167);
      END_STATE();
    case 168:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == '0') ADVANCE(560);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(72);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(561);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(168);
      END_STATE();
    case 169:
      if (lookahead == '#') ADVANCE(533);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(68);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(169);
      END_STATE();
    case 170:
      if (lookahead == '#') ADVANCE(172);
      END_STATE();
    case 171:
      if (lookahead == '#') ADVANCE(171);
      if (lookahead == '>') ADVANCE(522);
      if (lookahead != 0) ADVANCE(172);
      END_STATE();
    case 172:
      if (lookahead == '#') ADVANCE(171);
      if (lookahead == '`') ADVANCE(103);
      if (lookahead != 0) ADVANCE(172);
      END_STATE();
    case 173:
      if (lookahead == '#') ADVANCE(528);
      if (lookahead == '&') ADVANCE(1371);
      if (lookahead == '.') ADVANCE(826);
      if (lookahead == '<') ADVANCE(170);
      if (lookahead == '`') SKIP(50);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1669);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1616);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1639);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1596);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1568);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1535);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1566);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1708);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1579);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1580);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1402);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(173);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 174:
      if (lookahead == '#') ADVANCE(526);
      if (lookahead == '<') ADVANCE(1253);
      if (lookahead == '`') ADVANCE(1269);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1252);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(174);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '<' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 175:
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 176:
      ADVANCE_MAP(
        '$', 952,
        '?', 958,
        '^', 955,
        '_', 961,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 177:
      if (lookahead == '&') ADVANCE(1367);
      END_STATE();
    case 178:
      if (lookahead == '&') ADVANCE(195);
      END_STATE();
    case 179:
      if (lookahead == '\'') ADVANCE(1774);
      END_STATE();
    case 180:
      if (lookahead == '\'') ADVANCE(707);
      if (lookahead != 0) ADVANCE(180);
      END_STATE();
    case 181:
      if (lookahead == '+') ADVANCE(1849);
      END_STATE();
    case 182:
      ADVANCE_MAP(
        '-', 1851,
        'C', 222,
        'c', 222,
        'E', 493,
        'e', 493,
        'F', 335,
        'f', 335,
        'P', 226,
        'p', 226,
        'R', 308,
        'r', 308,
        'S', 489,
        's', 489,
        'W', 338,
        'w', 338,
      );
      END_STATE();
    case 183:
      if (lookahead == '-') ADVANCE(1853);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 184:
      if (lookahead == '-') ADVANCE(1270);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 185:
      if (lookahead == '-') ADVANCE(1855);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 186:
      if (lookahead == '-') ADVANCE(1272);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 187:
      if (lookahead == '.') ADVANCE(1842);
      END_STATE();
    case 188:
      if (lookahead == '.') ADVANCE(1842);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 189:
      if (lookahead == '1') ADVANCE(882);
      if (lookahead == '2') ADVANCE(888);
      END_STATE();
    case 190:
      if (lookahead == '1') ADVANCE(883);
      END_STATE();
    case 191:
      if (lookahead == '1') ADVANCE(884);
      if (lookahead == '2') ADVANCE(890);
      END_STATE();
    case 192:
      if (lookahead == '1') ADVANCE(885);
      if (lookahead == '2') ADVANCE(891);
      END_STATE();
    case 193:
      if (lookahead == '1') ADVANCE(886);
      if (lookahead == '2') ADVANCE(892);
      END_STATE();
    case 194:
      if (lookahead == '1') ADVANCE(887);
      if (lookahead == '2') ADVANCE(893);
      END_STATE();
    case 195:
      if (lookahead == '2') ADVANCE(889);
      END_STATE();
    case 196:
      if (lookahead == ':') ADVANCE(1871);
      END_STATE();
    case 197:
      if (lookahead == ':') ADVANCE(1871);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1355);
      END_STATE();
    case 198:
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == '?') ADVANCE(964);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1058);
      END_STATE();
    case 199:
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == '?') ADVANCE(1063);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1188);
      END_STATE();
    case 200:
      if (lookahead == '=') ADVANCE(831);
      END_STATE();
    case 201:
      if (lookahead == '>') ADVANCE(873);
      END_STATE();
    case 202:
      if (lookahead == '>') ADVANCE(178);
      END_STATE();
    case 203:
      if (lookahead == '>') ADVANCE(843);
      END_STATE();
    case 204:
      if (lookahead == '>') ADVANCE(849);
      END_STATE();
    case 205:
      if (lookahead == '>') ADVANCE(855);
      END_STATE();
    case 206:
      if (lookahead == '>') ADVANCE(861);
      END_STATE();
    case 207:
      if (lookahead == '>') ADVANCE(867);
      END_STATE();
    case 208:
      ADVANCE_MAP(
        '?', 1063,
        'G', 1108,
        'g', 1108,
        'L', 1114,
        'l', 1114,
        'P', 1116,
        'p', 1116,
        'S', 1099,
        's', 1099,
        'U', 1119,
        'u', 1119,
        'W', 1113,
        'w', 1113,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 209:
      if (lookahead == '?') ADVANCE(1063);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1188);
      END_STATE();
    case 210:
      if (lookahead == '?') ADVANCE(964);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1058);
      END_STATE();
    case 211:
      if (lookahead == '@') ADVANCE(710);
      END_STATE();
    case 212:
      if (lookahead == '@') ADVANCE(710);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 213:
      if (lookahead == '@') ADVANCE(706);
      if (lookahead != 0) ADVANCE(705);
      END_STATE();
    case 214:
      if (lookahead == '`') ADVANCE(1382);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1381);
      END_STATE();
    case 215:
      if (lookahead == 'b') ADVANCE(534);
      END_STATE();
    case 216:
      if (lookahead == 'b') ADVANCE(583);
      END_STATE();
    case 217:
      if (lookahead == 'b') ADVANCE(568);
      END_STATE();
    case 218:
      if (lookahead == '}') ADVANCE(1195);
      if (lookahead != 0) ADVANCE(218);
      END_STATE();
    case 219:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(496);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(592);
      END_STATE();
    case 220:
      ADVANCE_MAP(
        'A', 391,
        'a', 391,
        'N', 413,
        'n', 413,
        'O', 439,
        'o', 439,
        'X', 415,
        'x', 415,
      );
      END_STATE();
    case 221:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(391);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(439);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(415);
      END_STATE();
    case 222:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(456);
      END_STATE();
    case 223:
      ADVANCE_MAP(
        'A', 456,
        'a', 456,
        'C', 421,
        'c', 421,
        'E', 436,
        'e', 436,
        'G', 288,
        'g', 288,
        'L', 289,
        'l', 289,
        'M', 237,
        'm', 237,
        'N', 290,
        'n', 290,
        'O', 398,
        'o', 398,
        'R', 313,
        'r', 313,
        'S', 431,
        's', 431,
      );
      END_STATE();
    case 224:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(471);
      END_STATE();
    case 225:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(261);
      END_STATE();
    case 226:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(443);
      END_STATE();
    case 227:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(385);
      END_STATE();
    case 228:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(381);
      END_STATE();
    case 229:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(375);
      END_STATE();
    case 230:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(382);
      END_STATE();
    case 231:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(267);
      END_STATE();
    case 232:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(444);
      END_STATE();
    case 233:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(372);
      END_STATE();
    case 234:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(265);
      END_STATE();
    case 235:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(266);
      END_STATE();
    case 236:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(396);
      END_STATE();
    case 237:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(472);
      END_STATE();
    case 238:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(339);
      END_STATE();
    case 239:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(445);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(420);
      END_STATE();
    case 240:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(268);
      END_STATE();
    case 241:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(473);
      END_STATE();
    case 242:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(340);
      END_STATE();
    case 243:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(446);
      END_STATE();
    case 244:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(269);
      END_STATE();
    case 245:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(476);
      END_STATE();
    case 246:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(341);
      END_STATE();
    case 247:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(477);
      END_STATE();
    case 248:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(344);
      END_STATE();
    case 249:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(478);
      END_STATE();
    case 250:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(346);
      END_STATE();
    case 251:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(479);
      END_STATE();
    case 252:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(348);
      END_STATE();
    case 253:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(324);
      END_STATE();
    case 254:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(325);
      END_STATE();
    case 255:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(326);
      END_STATE();
    case 256:
      ADVANCE_MAP(
        'C', 421,
        'c', 421,
        'E', 436,
        'e', 436,
        'G', 288,
        'g', 288,
        'L', 289,
        'l', 289,
        'M', 237,
        'm', 237,
        'N', 290,
        'n', 290,
        'O', 398,
        'o', 398,
        'R', 313,
        'r', 313,
        'S', 431,
        's', 431,
      );
      END_STATE();
    case 257:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(430);
      END_STATE();
    case 258:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(327);
      END_STATE();
    case 259:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(410);
      END_STATE();
    case 260:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(328);
      END_STATE();
    case 261:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(464);
      END_STATE();
    case 262:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(329);
      END_STATE();
    case 263:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(330);
      END_STATE();
    case 264:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(331);
      END_STATE();
    case 265:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(332);
      END_STATE();
    case 266:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(333);
      END_STATE();
    case 267:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(299);
      END_STATE();
    case 268:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(301);
      END_STATE();
    case 269:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(303);
      END_STATE();
    case 270:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(311);
      END_STATE();
    case 271:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(232);
      END_STATE();
    case 272:
      ADVANCE_MAP(
        'C', 422,
        'c', 422,
        'E', 437,
        'e', 437,
        'G', 291,
        'g', 291,
        'L', 292,
        'l', 292,
        'M', 241,
        'm', 241,
        'N', 921,
        'n', 921,
        'R', 317,
        'r', 317,
        'S', 927,
        's', 927,
      );
      END_STATE();
    case 273:
      ADVANCE_MAP(
        'C', 423,
        'c', 423,
        'I', 388,
        'i', 388,
        'L', 354,
        'l', 354,
        'M', 245,
        'm', 245,
      );
      END_STATE();
    case 274:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(424);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(355);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(247);
      END_STATE();
    case 275:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(425);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(356);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(249);
      END_STATE();
    case 276:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1814);
      END_STATE();
    case 277:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1817);
      END_STATE();
    case 278:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1330);
      END_STATE();
    case 279:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1363);
      END_STATE();
    case 280:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1316);
      END_STATE();
    case 281:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(271);
      END_STATE();
    case 282:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(259);
      END_STATE();
    case 283:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(911);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(912);
      END_STATE();
    case 284:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(934);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(358);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(936);
      END_STATE();
    case 285:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(938);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(461);
      END_STATE();
    case 286:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(938);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(470);
      END_STATE();
    case 287:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(320);
      END_STATE();
    case 288:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(897);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(898);
      END_STATE();
    case 289:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(899);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(359);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(901);
      END_STATE();
    case 290:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(903);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(463);
      END_STATE();
    case 291:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(915);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(916);
      END_STATE();
    case 292:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(917);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(360);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(919);
      END_STATE();
    case 293:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1334);
      END_STATE();
    case 294:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(935);
      END_STATE();
    case 295:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(492);
      END_STATE();
    case 296:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(900);
      END_STATE();
    case 297:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(918);
      END_STATE();
    case 298:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(941);
      END_STATE();
    case 299:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(943);
      END_STATE();
    case 300:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(905);
      END_STATE();
    case 301:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(908);
      END_STATE();
    case 302:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(924);
      END_STATE();
    case 303:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(926);
      END_STATE();
    case 304:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1332);
      END_STATE();
    case 305:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(321);
      END_STATE();
    case 306:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1324);
      END_STATE();
    case 307:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1341);
      END_STATE();
    case 308:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(319);
      END_STATE();
    case 309:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(392);
      END_STATE();
    case 310:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(365);
      END_STATE();
    case 311:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(458);
      END_STATE();
    case 312:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(457);
      END_STATE();
    case 313:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(432);
      END_STATE();
    case 314:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(282);
      END_STATE();
    case 315:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(235);
      END_STATE();
    case 316:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(428);
      END_STATE();
    case 317:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(433);
      END_STATE();
    case 318:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1322);
      END_STATE();
    case 319:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(295);
      END_STATE();
    case 320:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(295);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(371);
      END_STATE();
    case 321:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(352);
      END_STATE();
    case 322:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(364);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(368);
      END_STATE();
    case 323:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(364);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(368);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(427);
      END_STATE();
    case 324:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(937);
      END_STATE();
    case 325:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(902);
      END_STATE();
    case 326:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(920);
      END_STATE();
    case 327:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(942);
      END_STATE();
    case 328:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(906);
      END_STATE();
    case 329:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(925);
      END_STATE();
    case 330:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1358);
      END_STATE();
    case 331:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1327);
      END_STATE();
    case 332:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1335);
      END_STATE();
    case 333:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1873);
      END_STATE();
    case 334:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(353);
      END_STATE();
    case 335:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(370);
      END_STATE();
    case 336:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(490);
      END_STATE();
    case 337:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(387);
      END_STATE();
    case 338:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(369);
      END_STATE();
    case 339:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(394);
      END_STATE();
    case 340:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(395);
      END_STATE();
    case 341:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(397);
      END_STATE();
    case 342:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(366);
      END_STATE();
    case 343:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(466);
      END_STATE();
    case 344:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(399);
      END_STATE();
    case 345:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(467);
      END_STATE();
    case 346:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(400);
      END_STATE();
    case 347:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(468);
      END_STATE();
    case 348:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(402);
      END_STATE();
    case 349:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(475);
      END_STATE();
    case 350:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(404);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(442);
      END_STATE();
    case 351:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(257);
      END_STATE();
    case 352:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(390);
      END_STATE();
    case 353:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(374);
      END_STATE();
    case 354:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(361);
      END_STATE();
    case 355:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(362);
      END_STATE();
    case 356:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(363);
      END_STATE();
    case 357:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(480);
      END_STATE();
    case 358:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(294);
      END_STATE();
    case 359:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(296);
      END_STATE();
    case 360:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(297);
      END_STATE();
    case 361:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(298);
      END_STATE();
    case 362:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(300);
      END_STATE();
    case 363:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(302);
      END_STATE();
    case 364:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(944);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(945);
      END_STATE();
    case 365:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1333);
      END_STATE();
    case 366:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1345);
      END_STATE();
    case 367:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(494);
      END_STATE();
    case 368:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(343);
      END_STATE();
    case 369:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(281);
      END_STATE();
    case 370:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(293);
      END_STATE();
    case 371:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(231);
      END_STATE();
    case 372:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(367);
      END_STATE();
    case 373:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(310);
      END_STATE();
    case 374:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(307);
      END_STATE();
    case 375:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(373);
      END_STATE();
    case 376:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(345);
      END_STATE();
    case 377:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(459);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(280);
      END_STATE();
    case 378:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(347);
      END_STATE();
    case 379:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(240);
      END_STATE();
    case 380:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(244);
      END_STATE();
    case 381:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1303);
      END_STATE();
    case 382:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1310);
      END_STATE();
    case 383:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(384);
      END_STATE();
    case 384:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(236);
      END_STATE();
    case 385:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(351);
      END_STATE();
    case 386:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(276);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(894);
      END_STATE();
    case 387:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(930);
      END_STATE();
    case 388:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(940);
      END_STATE();
    case 389:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1337);
      END_STATE();
    case 390:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1312);
      END_STATE();
    case 391:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(277);
      END_STATE();
    case 392:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(460);
      END_STATE();
    case 393:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(413);
      END_STATE();
    case 394:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(449);
      END_STATE();
    case 395:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(450);
      END_STATE();
    case 396:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(279);
      END_STATE();
    case 397:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(451);
      END_STATE();
    case 398:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(474);
      END_STATE();
    case 399:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(452);
      END_STATE();
    case 400:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(453);
      END_STATE();
    case 401:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(227);
      END_STATE();
    case 402:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(454);
      END_STATE();
    case 403:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(482);
      END_STATE();
    case 404:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(233);
      END_STATE();
    case 405:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(483);
      END_STATE();
    case 406:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(484);
      END_STATE();
    case 407:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(485);
      END_STATE();
    case 408:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(486);
      END_STATE();
    case 409:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(487);
      END_STATE();
    case 410:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(383);
      END_STATE();
    case 411:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1343);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(401);
      END_STATE();
    case 412:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(337);
      END_STATE();
    case 413:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(462);
      END_STATE();
    case 414:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(440);
      END_STATE();
    case 415:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(441);
      END_STATE();
    case 416:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(447);
      END_STATE();
    case 417:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(465);
      END_STATE();
    case 418:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(448);
      END_STATE();
    case 419:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(469);
      END_STATE();
    case 420:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(270);
      END_STATE();
    case 421:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(405);
      END_STATE();
    case 422:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(406);
      END_STATE();
    case 423:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(407);
      END_STATE();
    case 424:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(408);
      END_STATE();
    case 425:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(409);
      END_STATE();
    case 426:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(368);
      END_STATE();
    case 427:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(429);
      END_STATE();
    case 428:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(371);
      END_STATE();
    case 429:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(416);
      END_STATE();
    case 430:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(243);
      END_STATE();
    case 431:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(376);
      END_STATE();
    case 432:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(379);
      END_STATE();
    case 433:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(380);
      END_STATE();
    case 434:
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(910);
      END_STATE();
    case 435:
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(910);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(225);
      END_STATE();
    case 436:
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(896);
      END_STATE();
    case 437:
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(914);
      END_STATE();
    case 438:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1815);
      END_STATE();
    case 439:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1818);
      END_STATE();
    case 440:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1816);
      END_STATE();
    case 441:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1819);
      END_STATE();
    case 442:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1339);
      END_STATE();
    case 443:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(229);
      END_STATE();
    case 444:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(278);
      END_STATE();
    case 445:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(228);
      END_STATE();
    case 446:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(230);
      END_STATE();
    case 447:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(481);
      END_STATE();
    case 448:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(315);
      END_STATE();
    case 449:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(907);
      END_STATE();
    case 450:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(895);
      END_STATE();
    case 451:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(913);
      END_STATE();
    case 452:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(939);
      END_STATE();
    case 453:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(904);
      END_STATE();
    case 454:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(923);
      END_STATE();
    case 455:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1314);
      END_STATE();
    case 456:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(312);
      END_STATE();
    case 457:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(309);
      END_STATE();
    case 458:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(455);
      END_STATE();
    case 459:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(306);
      END_STATE();
    case 460:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(349);
      END_STATE();
    case 461:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1844);
      END_STATE();
    case 462:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1848);
      END_STATE();
    case 463:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(274);
      END_STATE();
    case 464:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1331);
      END_STATE();
    case 465:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(928);
      END_STATE();
    case 466:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(946);
      END_STATE();
    case 467:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(909);
      END_STATE();
    case 468:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(929);
      END_STATE();
    case 469:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1843);
      END_STATE();
    case 470:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(273);
      END_STATE();
    case 471:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(253);
      END_STATE();
    case 472:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(254);
      END_STATE();
    case 473:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(255);
      END_STATE();
    case 474:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(238);
      END_STATE();
    case 475:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(336);
      END_STATE();
    case 476:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(258);
      END_STATE();
    case 477:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(260);
      END_STATE();
    case 478:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(262);
      END_STATE();
    case 479:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(263);
      END_STATE();
    case 480:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(264);
      END_STATE();
    case 481:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(314);
      END_STATE();
    case 482:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(342);
      END_STATE();
    case 483:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(242);
      END_STATE();
    case 484:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(246);
      END_STATE();
    case 485:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(248);
      END_STATE();
    case 486:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(250);
      END_STATE();
    case 487:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(252);
      END_STATE();
    case 488:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(275);
      END_STATE();
    case 489:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(427);
      END_STATE();
    case 490:
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(304);
      END_STATE();
    case 491:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(357);
      END_STATE();
    case 492:
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1329);
      END_STATE();
    case 493:
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(225);
      END_STATE();
    case 494:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1360);
      END_STATE();
    case 495:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 496:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(592);
      END_STATE();
    case 497:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(572);
      END_STATE();
    case 498:
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(218);
      END_STATE();
    case 499:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(513);
      END_STATE();
    case 500:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(513);
      if (lookahead == '\r') SKIP(499);
      END_STATE();
    case 501:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(516);
      END_STATE();
    case 502:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(516);
      if (lookahead == '\r') SKIP(501);
      END_STATE();
    case 503:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(515);
      END_STATE();
    case 504:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(515);
      if (lookahead == '\r') SKIP(503);
      END_STATE();
    case 505:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(517);
      END_STATE();
    case 506:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(517);
      if (lookahead == '\r') SKIP(505);
      END_STATE();
    case 507:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(518);
      END_STATE();
    case 508:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(518);
      if (lookahead == '\r') SKIP(507);
      END_STATE();
    case 509:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(519);
      END_STATE();
    case 510:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(519);
      if (lookahead == '\r') SKIP(509);
      END_STATE();
    case 511:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(520);
      END_STATE();
    case 512:
      if (eof) ADVANCE(521);
      if (lookahead == '\n') SKIP(520);
      if (lookahead == '\r') SKIP(511);
      END_STATE();
    case 513:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1846,
        '"', 1390,
        '#', 533,
        '$', 1389,
        '%', 1839,
        '&', 1372,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1824,
        '.', 824,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 1803,
        ';', 1302,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(500);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(736);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(713);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(717);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(769);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(755);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(756);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(747);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(719);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(738);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(737);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(752);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(783);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(753);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(513);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 514:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1846,
        '"', 600,
        '#', 533,
        '$', 175,
        '%', 1839,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1825,
        '.', 1866,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(504);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(305);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(251);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(411);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(377);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(350);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(389);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(239);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(491);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(403);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(334);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(515);
      END_STATE();
    case 515:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1846,
        '"', 600,
        '#', 533,
        '$', 175,
        '%', 1839,
        '&', 177,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '*', 1841,
        '+', 1821,
        ',', 1308,
        '-', 1825,
        '.', 188,
        '/', 1836,
        '0', 545,
        '1', 539,
        '2', 540,
        '3', 541,
        '4', 542,
        '5', 543,
        '6', 544,
        ':', 196,
        '<', 879,
        '=', 830,
        '>', 837,
        '@', 146,
        '[', 827,
        '\\', 1837,
        ']', 1872,
      );
      if (lookahead == '`') SKIP(504);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '|') ADVANCE(1370);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(305);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(251);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(411);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(377);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(350);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(389);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(239);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(491);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(403);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(334);
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(546);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(515);
      END_STATE();
    case 516:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(502);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1627);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1533);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1393);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(516);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 517:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(506);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1609);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1533);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1395);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(517);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 518:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(508);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1529);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1627);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1582);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1533);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1397);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(518);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 519:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(510);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1627);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1582);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1533);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1394);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(519);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 520:
      if (eof) ADVANCE(521);
      ADVANCE_MAP(
        '!', 1845,
        '"', 600,
        '#', 528,
        '$', 175,
        '&', 1371,
        '\'', 180,
        '(', 1305,
        ')', 1306,
        '+', 1820,
        ',', 1308,
        '-', 1828,
        '.', 825,
        '0', 547,
        ':', 1712,
        ';', 1302,
        '<', 170,
        '@', 146,
        '[', 827,
      );
      if (lookahead == '`') SKIP(512);
      if (lookahead == '{') ADVANCE(1318);
      if (lookahead == '}') ADVANCE(1320);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1660);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1604);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1514);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1620);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1581);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1567);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1533);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1554);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1547);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1577);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1576);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1396);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(520);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if (lookahead != 0 &&
          (lookahead < '&' || '<' < lookahead) &&
          lookahead != ']' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 521:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 522:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 523:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '"') ADVANCE(657);
      if (lookahead == '#') ADVANCE(523);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(523);
      if (lookahead != 0) ADVANCE(533);
      END_STATE();
    case 524:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '|') ADVANCE(533);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(524);
      END_STATE();
    case 525:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(1380);
      if (lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '`') ADVANCE(533);
      if (lookahead != 0) ADVANCE(525);
      END_STATE();
    case 526:
      ACCEPT_TOKEN(sym_comment);
      ADVANCE_MAP(
        '\t', 533,
        0x0b, 533,
        '\f', 533,
        ' ', 533,
        '&', 533,
        '(', 533,
        ')', 533,
        ';', 533,
        '{', 533,
        '}', 533,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(526);
      END_STATE();
    case 527:
      ACCEPT_TOKEN(sym_comment);
      ADVANCE_MAP(
        '\t', 533,
        0x0b, 533,
        '\f', 533,
        ' ', 533,
        '&', 533,
        '(', 533,
        ')', 533,
        ',', 533,
        ';', 533,
        '|', 533,
        '}', 533,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(527);
      END_STATE();
    case 528:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          (' ' <= lookahead && lookahead <= '"') ||
          lookahead == '$' ||
          ('&' <= lookahead && lookahead <= '-') ||
          lookahead == '/' ||
          lookahead == ';' ||
          lookahead == '<' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(533);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(528);
      END_STATE();
    case 529:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 530:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 531:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 532:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(1380);
      END_STATE();
    case 533:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(533);
      END_STATE();
    case 534:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      END_STATE();
    case 535:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '-', 1796,
        '.', 1489,
        'e', 1403,
        'g', 1421,
        'k', 1421,
        'm', 1421,
        'p', 1421,
        't', 1421,
        'x', 1490,
        'd', 537,
        'l', 537,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(536);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 536:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '-', 1796,
        '.', 1489,
        'e', 1403,
        'g', 1421,
        'k', 1421,
        'm', 1421,
        'p', 1421,
        't', 1421,
        'd', 537,
        'l', 537,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(536);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 537:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'g') ADVANCE(1421);
      if (lookahead == 'k') ADVANCE(1421);
      if (lookahead == 'm') ADVANCE(1421);
      if (lookahead == 'p') ADVANCE(1421);
      if (lookahead == 't') ADVANCE(1421);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 538:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 539:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 178,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 540:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 843,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 541:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 849,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 542:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 855,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 543:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 861,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 544:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        '>', 867,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 545:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'x', 497,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 546:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 495,
        'e', 219,
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(546);
      END_STATE();
    case 547:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1710,
        'e', 1512,
        'g', 1509,
        'k', 1509,
        'm', 1509,
        'p', 1509,
        't', 1509,
        'x', 1711,
        'd', 562,
        'l', 562,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 548:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1710,
        'e', 1512,
        'g', 1509,
        'k', 1509,
        'm', 1509,
        'p', 1509,
        't', 1509,
        'd', 562,
        'l', 562,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 549:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 1207,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 550:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 844,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 551:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 850,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 552:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 856,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 862,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 554:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        '>', 868,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'x', 1244,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 556:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1240,
        'e', 1219,
        'g', 1215,
        'k', 1215,
        'm', 1215,
        'p', 1215,
        't', 1215,
        'd', 563,
        'l', 563,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 557:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1266,
        'e', 1265,
        'g', 1262,
        'k', 1262,
        'm', 1262,
        'p', 1262,
        't', 1262,
        'x', 1268,
        'd', 564,
        'l', 564,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(558);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 558:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        '.', 1266,
        'e', 1265,
        'g', 1262,
        'k', 1262,
        'm', 1262,
        'p', 1262,
        't', 1262,
        'd', 564,
        'l', 564,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(558);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 559:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == 'g') ADVANCE(215);
      if (lookahead == 'k') ADVANCE(215);
      if (lookahead == 'm') ADVANCE(215);
      if (lookahead == 'p') ADVANCE(215);
      if (lookahead == 't') ADVANCE(215);
      END_STATE();
    case 560:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      ADVANCE_MAP(
        'g', 215,
        'k', 215,
        'm', 215,
        'p', 215,
        't', 215,
        'x', 497,
        'd', 559,
        'l', 559,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(561);
      END_STATE();
    case 561:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == 'g') ADVANCE(215);
      if (lookahead == 'k') ADVANCE(215);
      if (lookahead == 'm') ADVANCE(215);
      if (lookahead == 'p') ADVANCE(215);
      if (lookahead == 't') ADVANCE(215);
      if (lookahead == 'd' ||
          lookahead == 'l') ADVANCE(559);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(561);
      END_STATE();
    case 562:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1509);
      if (lookahead == 'k') ADVANCE(1509);
      if (lookahead == 'm') ADVANCE(1509);
      if (lookahead == 'p') ADVANCE(1509);
      if (lookahead == 't') ADVANCE(1509);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 563:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1215);
      if (lookahead == 'k') ADVANCE(1215);
      if (lookahead == 'm') ADVANCE(1215);
      if (lookahead == 'p') ADVANCE(1215);
      if (lookahead == 't') ADVANCE(1215);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1262);
      if (lookahead == 'k') ADVANCE(1262);
      if (lookahead == 'm') ADVANCE(1262);
      if (lookahead == 'p') ADVANCE(1262);
      if (lookahead == 't') ADVANCE(1262);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 565:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 566:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(sym_decimal_integer_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 568:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      END_STATE();
    case 569:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'g') ADVANCE(1423);
      if (lookahead == 'k') ADVANCE(1423);
      if (lookahead == 'l') ADVANCE(570);
      if (lookahead == 'm') ADVANCE(1423);
      if (lookahead == 'p') ADVANCE(1423);
      if (lookahead == 't') ADVANCE(1423);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(569);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('h' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 570:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'g') ADVANCE(1423);
      if (lookahead == 'k') ADVANCE(1423);
      if (lookahead == 'm') ADVANCE(1423);
      if (lookahead == 'p') ADVANCE(1423);
      if (lookahead == 't') ADVANCE(1423);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 571:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 572:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(217);
      if (lookahead == 'k') ADVANCE(217);
      if (lookahead == 'l') ADVANCE(573);
      if (lookahead == 'm') ADVANCE(217);
      if (lookahead == 'p') ADVANCE(217);
      if (lookahead == 't') ADVANCE(217);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(572);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(217);
      if (lookahead == 'k') ADVANCE(217);
      if (lookahead == 'm') ADVANCE(217);
      if (lookahead == 'p') ADVANCE(217);
      if (lookahead == 't') ADVANCE(217);
      END_STATE();
    case 574:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1511);
      if (lookahead == 'k') ADVANCE(1511);
      if (lookahead == 'l') ADVANCE(575);
      if (lookahead == 'm') ADVANCE(1511);
      if (lookahead == 'p') ADVANCE(1511);
      if (lookahead == 't') ADVANCE(1511);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(574);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 575:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1511);
      if (lookahead == 'k') ADVANCE(1511);
      if (lookahead == 'm') ADVANCE(1511);
      if (lookahead == 'p') ADVANCE(1511);
      if (lookahead == 't') ADVANCE(1511);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 576:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1217);
      if (lookahead == 'k') ADVANCE(1217);
      if (lookahead == 'l') ADVANCE(577);
      if (lookahead == 'm') ADVANCE(1217);
      if (lookahead == 'p') ADVANCE(1217);
      if (lookahead == 't') ADVANCE(1217);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(576);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 577:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1217);
      if (lookahead == 'k') ADVANCE(1217);
      if (lookahead == 'm') ADVANCE(1217);
      if (lookahead == 'p') ADVANCE(1217);
      if (lookahead == 't') ADVANCE(1217);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 578:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1264);
      if (lookahead == 'k') ADVANCE(1264);
      if (lookahead == 'l') ADVANCE(579);
      if (lookahead == 'm') ADVANCE(1264);
      if (lookahead == 'p') ADVANCE(1264);
      if (lookahead == 't') ADVANCE(1264);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(578);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 579:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead == 'g') ADVANCE(1264);
      if (lookahead == 'k') ADVANCE(1264);
      if (lookahead == 'm') ADVANCE(1264);
      if (lookahead == 'p') ADVANCE(1264);
      if (lookahead == 't') ADVANCE(1264);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 580:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 581:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 582:
      ACCEPT_TOKEN(sym_hexadecimal_integer_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(sym_real_literal);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'e') ADVANCE(1403);
      if (lookahead == 'g') ADVANCE(1422);
      if (lookahead == 'k') ADVANCE(1422);
      if (lookahead == 'm') ADVANCE(1422);
      if (lookahead == 'p') ADVANCE(1422);
      if (lookahead == 't') ADVANCE(1422);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(584);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 585:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'g') ADVANCE(1422);
      if (lookahead == 'k') ADVANCE(1422);
      if (lookahead == 'm') ADVANCE(1422);
      if (lookahead == 'p') ADVANCE(1422);
      if (lookahead == 't') ADVANCE(1422);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 586:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 587:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'e') ADVANCE(219);
      if (lookahead == 'g') ADVANCE(216);
      if (lookahead == 'k') ADVANCE(216);
      if (lookahead == 'm') ADVANCE(216);
      if (lookahead == 'p') ADVANCE(216);
      if (lookahead == 't') ADVANCE(216);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 588:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'e') ADVANCE(1512);
      if (lookahead == 'g') ADVANCE(1510);
      if (lookahead == 'k') ADVANCE(1510);
      if (lookahead == 'm') ADVANCE(1510);
      if (lookahead == 'p') ADVANCE(1510);
      if (lookahead == 't') ADVANCE(1510);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'e') ADVANCE(1219);
      if (lookahead == 'g') ADVANCE(1216);
      if (lookahead == 'k') ADVANCE(1216);
      if (lookahead == 'm') ADVANCE(1216);
      if (lookahead == 'p') ADVANCE(1216);
      if (lookahead == 't') ADVANCE(1216);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 590:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'e') ADVANCE(1265);
      if (lookahead == 'g') ADVANCE(1263);
      if (lookahead == 'k') ADVANCE(1263);
      if (lookahead == 'm') ADVANCE(1263);
      if (lookahead == 'p') ADVANCE(1263);
      if (lookahead == 't') ADVANCE(1263);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(590);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 591:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'g') ADVANCE(1788);
      if (lookahead == 'k') ADVANCE(1788);
      if (lookahead == 'm') ADVANCE(1788);
      if (lookahead == 'p') ADVANCE(1788);
      if (lookahead == 't') ADVANCE(1788);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(591);
      if (lookahead == '-' ||
          lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 592:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'g') ADVANCE(216);
      if (lookahead == 'k') ADVANCE(216);
      if (lookahead == 'm') ADVANCE(216);
      if (lookahead == 'p') ADVANCE(216);
      if (lookahead == 't') ADVANCE(216);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(592);
      END_STATE();
    case 593:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'g') ADVANCE(1510);
      if (lookahead == 'k') ADVANCE(1510);
      if (lookahead == 'm') ADVANCE(1510);
      if (lookahead == 'p') ADVANCE(1510);
      if (lookahead == 't') ADVANCE(1510);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(593);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 594:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'g') ADVANCE(1216);
      if (lookahead == 'k') ADVANCE(1216);
      if (lookahead == 'm') ADVANCE(1216);
      if (lookahead == 'p') ADVANCE(1216);
      if (lookahead == 't') ADVANCE(1216);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(594);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 595:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == 'g') ADVANCE(1263);
      if (lookahead == 'k') ADVANCE(1263);
      if (lookahead == 'm') ADVANCE(1263);
      if (lookahead == 'p') ADVANCE(1263);
      if (lookahead == 't') ADVANCE(1263);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(595);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 596:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(sym_real_literal);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 598:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 599:
      ACCEPT_TOKEN(sym_real_literal);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 600:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token1);
      if (lookahead == '#') ADVANCE(600);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(600);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token1);
      if (lookahead == '#') ADVANCE(601);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(600);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token1);
      if (lookahead == '#') ADVANCE(602);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(600);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      ADVANCE_MAP(
        '"', 657,
        '#', 603,
        '\n', 604,
        '\r', 604,
        '$', 533,
        '`', 533,
        '\t', 603,
        0x0b, 603,
        '\f', 603,
        ' ', 603,
      );
      if (lookahead != 0) ADVANCE(644);
      END_STATE();
    case 604:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '"') ADVANCE(655);
      if (lookahead == '#') ADVANCE(604);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(604);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          lookahead != '`') ADVANCE(646);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '#') ADVANCE(603);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '@') ADVANCE(642);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(606);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(605);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          lookahead != '`') ADVANCE(646);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '#') ADVANCE(644);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '@') ADVANCE(642);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(606);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          lookahead != '`') ADVANCE(646);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '#') ADVANCE(609);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          lookahead != '`') ADVANCE(646);
      END_STATE();
    case 608:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '#') ADVANCE(608);
      if (lookahead == '>') ADVANCE(646);
      if (('"' <= lookahead && lookahead <= '$') ||
          lookahead == '`') ADVANCE(172);
      if (lookahead != 0) ADVANCE(609);
      END_STATE();
    case 609:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '#') ADVANCE(608);
      if (lookahead == '`') ADVANCE(103);
      if (('"' <= lookahead && lookahead <= '$')) ADVANCE(172);
      if (lookahead != 0) ADVANCE(609);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == '?') ADVANCE(646);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(645);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 611:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(636);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 612:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(623);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 613:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(612);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 614:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(612);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 615:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(634);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 616:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(641);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(625);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 618:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(641);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 619:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(638);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 620:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(626);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 621:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(631);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 622:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(617);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 623:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(641);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 624:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(627);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 625:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(628);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 626:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(618);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 627:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(613);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 628:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(639);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 629:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(633);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 630:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(614);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 631:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(637);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 632:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(619);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 633:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(622);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 634:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(621);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 635:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(620);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 636:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(616);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 637:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(641);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 638:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 639:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(641);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 640:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(643);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 641:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == ':') ADVANCE(610);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 642:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      ADVANCE_MAP(
        '?', 646,
        'G', 624,
        'g', 624,
        'L', 630,
        'l', 630,
        'P', 632,
        'p', 632,
        'S', 615,
        's', 615,
        'U', 635,
        'u', 635,
        'W', 629,
        'w', 629,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(640);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 643:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '?') ADVANCE(646);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(645);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 644:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(646);
      if (lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '`') ADVANCE(533);
      if (lookahead != 0) ADVANCE(644);
      END_STATE();
    case 645:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(645);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(646);
      END_STATE();
    case 646:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token2);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(646);
      END_STATE();
    case 647:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token3);
      END_STATE();
    case 648:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token3);
      if (lookahead == '\n') ADVANCE(647);
      END_STATE();
    case 649:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token4);
      END_STATE();
    case 650:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token4);
      if (lookahead == '\n') ADVANCE(649);
      END_STATE();
    case 651:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 653:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '`', 104,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '\\') ADVANCE(647);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 654:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 655:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token5);
      END_STATE();
    case 656:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token5);
      if (lookahead == '"') ADVANCE(651);
      END_STATE();
    case 657:
      ACCEPT_TOKEN(aux_sym_expandable_string_literal_token5);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(533);
      END_STATE();
    case 658:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token1);
      END_STATE();
    case 659:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      ADVANCE_MAP(
        '#', 697,
        '<', 660,
        '@', 695,
        '\t', 659,
        0x0b, 659,
        '\f', 659,
        ' ', 659,
        0xa0, 659,
        0x200b, 659,
        0x2060, 659,
        0xfeff, 659,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '#' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(699);
      END_STATE();
    case 660:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == '#') ADVANCE(662);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '#' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(699);
      END_STATE();
    case 661:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == '#') ADVANCE(661);
      if (lookahead == '>') ADVANCE(699);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '$' ||
          lookahead == '`') ADVANCE(172);
      if (lookahead != 0) ADVANCE(662);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == '#') ADVANCE(661);
      if (lookahead == '`') ADVANCE(103);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '$') ADVANCE(172);
      if (lookahead != 0) ADVANCE(662);
      END_STATE();
    case 663:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == '?') ADVANCE(699);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(698);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 664:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(689);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(676);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 666:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(665);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 667:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(665);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 668:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(687);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 669:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 670:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(678);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 671:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 672:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(691);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 673:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(679);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 674:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(684);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 675:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(670);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 676:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 677:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(680);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 678:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(681);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 679:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(671);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 680:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(666);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 681:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(692);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 682:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(686);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 683:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(667);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 684:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(690);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 685:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(672);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 686:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(675);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 687:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(674);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 688:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(673);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 689:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(669);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 690:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 691:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(664);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 692:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 693:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(696);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 694:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == ':') ADVANCE(663);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 695:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      ADVANCE_MAP(
        '?', 699,
        'G', 677,
        'g', 677,
        'L', 683,
        'l', 683,
        'P', 685,
        'p', 685,
        'S', 668,
        's', 668,
        'U', 688,
        'u', 688,
        'W', 682,
        'w', 682,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 696:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == '?') ADVANCE(699);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(698);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 697:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead == '$' ||
          lookahead == '`') ADVANCE(533);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(697);
      END_STATE();
    case 698:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(698);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(699);
      END_STATE();
    case 699:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token2);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(699);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token3);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token3);
      if (lookahead == '#') ADVANCE(172);
      END_STATE();
    case 702:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token3);
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 703:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token3);
      ADVANCE_MAP(
        '?', 1063,
        'G', 1108,
        'g', 1108,
        'L', 1114,
        'l', 1114,
        'P', 1116,
        'p', 1116,
        'S', 1099,
        's', 1099,
        'U', 1119,
        'u', 1119,
        'W', 1113,
        'w', 1113,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 704:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token3);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(533);
      END_STATE();
    case 705:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token4);
      END_STATE();
    case 706:
      ACCEPT_TOKEN(aux_sym_expandable_here_string_literal_token5);
      END_STATE();
    case 707:
      ACCEPT_TOKEN(sym_verbatim_string_characters);
      if (lookahead == '\'') ADVANCE(180);
      END_STATE();
    case 708:
      ACCEPT_TOKEN(sym_verbatim_string_characters);
      if (lookahead == '\'') ADVANCE(1208);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(sym_verbatim_string_characters);
      if (lookahead == '\'') ADVANCE(1257);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(sym_verbatim_here_string_characters);
      END_STATE();
    case 711:
      ACCEPT_TOKEN(sym_verbatim_here_string_characters);
      if (lookahead == '\n') ADVANCE(85);
      if (lookahead == '\r') ADVANCE(10);
      if (lookahead != 0) ADVANCE(12);
      END_STATE();
    case 712:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 713:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(807);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(718);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(787);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(726);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 715:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(791);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(776);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 717:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(811);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(821);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(784);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 718:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(804);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 719:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(798);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(789);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 720:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(779);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 721:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(772);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 722:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(766);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(777);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 724:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(799);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 725:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(814);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 726:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(751);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 727:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1811);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 728:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(792);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 729:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(746);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(735);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 731:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(815);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 732:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(821);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(778);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 733:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(734);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 734:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(741);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 735:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 736:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(750);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(739);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 737:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(793);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(725);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(754);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 738:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(808);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 739:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(722);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 740:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(768);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 741:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(781);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 742:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(794);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 743:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(786);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 744:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(758);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 745:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(714);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 746:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(805);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 747:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(821);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 748:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 749:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(771);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 750:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(761);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 751:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 752:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(715);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 753:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(757);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(795);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 754:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(807);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 755:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(773);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(800);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(782);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 756:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(733);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 757:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(774);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 758:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(748);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 759:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(790);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 760:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(768);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 761:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(780);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 762:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(727);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 763:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(785);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 764:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(806);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 765:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(728);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 766:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 767:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(749);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 768:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 769:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(803);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(732);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(764);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 770:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(820);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 771:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(788);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 772:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(770);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 773:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(810);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(721);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 774:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(735);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 775:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(740);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 776:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(775);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 777:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 778:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1813);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 779:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(765);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 780:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 781:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1810);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 782:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(731);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 783:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(812);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 784:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(720);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 785:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(817);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 786:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(730);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 787:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(813);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 788:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(819);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 789:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(729);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 790:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(780);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 791:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 792:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(724);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 793:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(818);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 794:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 795:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(767);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 796:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(788);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 797:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(780);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 798:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(716);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 799:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(723);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 800:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(745);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 801:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 802:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1812);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 803:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(744);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 804:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(802);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 805:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(801);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 806:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 807:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(726);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 808:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(816);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 809:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(725);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 810:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(742);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 811:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(712);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 812:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(760);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 813:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(763);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 814:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(762);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 815:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(759);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 816:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(797);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 817:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(735);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 818:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(743);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 819:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 820:
      ACCEPT_TOKEN(sym_simple_name);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 821:
      ACCEPT_TOKEN(sym_simple_name);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(821);
      END_STATE();
    case 822:
      ACCEPT_TOKEN(sym_type_identifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(822);
      END_STATE();
    case 823:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 824:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 825:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 826:
      ACCEPT_TOKEN(anon_sym_DOT);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 827:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 828:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 829:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 830:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 831:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 832:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 833:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 834:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 835:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 836:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 837:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '>') ADVANCE(840);
      END_STATE();
    case 838:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '>') ADVANCE(842);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 839:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '>') ADVANCE(841);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 840:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 841:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 842:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 843:
      ACCEPT_TOKEN(anon_sym_2_GT);
      if (lookahead == '&') ADVANCE(190);
      if (lookahead == '>') ADVANCE(846);
      END_STATE();
    case 844:
      ACCEPT_TOKEN(anon_sym_2_GT);
      if (lookahead == '&') ADVANCE(190);
      if (lookahead == '>') ADVANCE(848);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 845:
      ACCEPT_TOKEN(anon_sym_2_GT);
      if (lookahead == '&') ADVANCE(190);
      if (lookahead == '>') ADVANCE(847);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 846:
      ACCEPT_TOKEN(anon_sym_2_GT_GT);
      END_STATE();
    case 847:
      ACCEPT_TOKEN(anon_sym_2_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 848:
      ACCEPT_TOKEN(anon_sym_2_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 849:
      ACCEPT_TOKEN(anon_sym_3_GT);
      if (lookahead == '&') ADVANCE(191);
      if (lookahead == '>') ADVANCE(852);
      END_STATE();
    case 850:
      ACCEPT_TOKEN(anon_sym_3_GT);
      if (lookahead == '&') ADVANCE(191);
      if (lookahead == '>') ADVANCE(854);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 851:
      ACCEPT_TOKEN(anon_sym_3_GT);
      if (lookahead == '&') ADVANCE(191);
      if (lookahead == '>') ADVANCE(853);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 852:
      ACCEPT_TOKEN(anon_sym_3_GT_GT);
      END_STATE();
    case 853:
      ACCEPT_TOKEN(anon_sym_3_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 854:
      ACCEPT_TOKEN(anon_sym_3_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 855:
      ACCEPT_TOKEN(anon_sym_4_GT);
      if (lookahead == '&') ADVANCE(192);
      if (lookahead == '>') ADVANCE(858);
      END_STATE();
    case 856:
      ACCEPT_TOKEN(anon_sym_4_GT);
      if (lookahead == '&') ADVANCE(192);
      if (lookahead == '>') ADVANCE(860);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 857:
      ACCEPT_TOKEN(anon_sym_4_GT);
      if (lookahead == '&') ADVANCE(192);
      if (lookahead == '>') ADVANCE(859);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 858:
      ACCEPT_TOKEN(anon_sym_4_GT_GT);
      END_STATE();
    case 859:
      ACCEPT_TOKEN(anon_sym_4_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 860:
      ACCEPT_TOKEN(anon_sym_4_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 861:
      ACCEPT_TOKEN(anon_sym_5_GT);
      if (lookahead == '&') ADVANCE(193);
      if (lookahead == '>') ADVANCE(864);
      END_STATE();
    case 862:
      ACCEPT_TOKEN(anon_sym_5_GT);
      if (lookahead == '&') ADVANCE(193);
      if (lookahead == '>') ADVANCE(866);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 863:
      ACCEPT_TOKEN(anon_sym_5_GT);
      if (lookahead == '&') ADVANCE(193);
      if (lookahead == '>') ADVANCE(865);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 864:
      ACCEPT_TOKEN(anon_sym_5_GT_GT);
      END_STATE();
    case 865:
      ACCEPT_TOKEN(anon_sym_5_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 866:
      ACCEPT_TOKEN(anon_sym_5_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 867:
      ACCEPT_TOKEN(anon_sym_6_GT);
      if (lookahead == '&') ADVANCE(194);
      if (lookahead == '>') ADVANCE(870);
      END_STATE();
    case 868:
      ACCEPT_TOKEN(anon_sym_6_GT);
      if (lookahead == '&') ADVANCE(194);
      if (lookahead == '>') ADVANCE(872);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 869:
      ACCEPT_TOKEN(anon_sym_6_GT);
      if (lookahead == '&') ADVANCE(194);
      if (lookahead == '>') ADVANCE(871);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 870:
      ACCEPT_TOKEN(anon_sym_6_GT_GT);
      END_STATE();
    case 871:
      ACCEPT_TOKEN(anon_sym_6_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 872:
      ACCEPT_TOKEN(anon_sym_6_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 873:
      ACCEPT_TOKEN(anon_sym_STAR_GT);
      if (lookahead == '&') ADVANCE(189);
      if (lookahead == '>') ADVANCE(876);
      END_STATE();
    case 874:
      ACCEPT_TOKEN(anon_sym_STAR_GT);
      if (lookahead == '&') ADVANCE(189);
      if (lookahead == '>') ADVANCE(878);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 875:
      ACCEPT_TOKEN(anon_sym_STAR_GT);
      if (lookahead == '&') ADVANCE(189);
      if (lookahead == '>') ADVANCE(877);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 876:
      ACCEPT_TOKEN(anon_sym_STAR_GT_GT);
      END_STATE();
    case 877:
      ACCEPT_TOKEN(anon_sym_STAR_GT_GT);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 878:
      ACCEPT_TOKEN(anon_sym_STAR_GT_GT);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 879:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '#') ADVANCE(172);
      END_STATE();
    case 880:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '#') ADVANCE(1205);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 881:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '#') ADVANCE(1754);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 882:
      ACCEPT_TOKEN(anon_sym_STAR_GT_AMP1);
      END_STATE();
    case 883:
      ACCEPT_TOKEN(anon_sym_2_GT_AMP1);
      END_STATE();
    case 884:
      ACCEPT_TOKEN(anon_sym_3_GT_AMP1);
      END_STATE();
    case 885:
      ACCEPT_TOKEN(anon_sym_4_GT_AMP1);
      END_STATE();
    case 886:
      ACCEPT_TOKEN(anon_sym_5_GT_AMP1);
      END_STATE();
    case 887:
      ACCEPT_TOKEN(anon_sym_6_GT_AMP1);
      END_STATE();
    case 888:
      ACCEPT_TOKEN(anon_sym_STAR_GT_AMP2);
      END_STATE();
    case 889:
      ACCEPT_TOKEN(anon_sym_1_GT_AMP2);
      END_STATE();
    case 890:
      ACCEPT_TOKEN(anon_sym_3_GT_AMP2);
      END_STATE();
    case 891:
      ACCEPT_TOKEN(anon_sym_4_GT_AMP2);
      END_STATE();
    case 892:
      ACCEPT_TOKEN(anon_sym_5_GT_AMP2);
      END_STATE();
    case 893:
      ACCEPT_TOKEN(anon_sym_6_GT_AMP2);
      END_STATE();
    case 894:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token1);
      END_STATE();
    case 895:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token2);
      END_STATE();
    case 896:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token3);
      END_STATE();
    case 897:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token4);
      END_STATE();
    case 898:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token5);
      END_STATE();
    case 899:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token6);
      END_STATE();
    case 900:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token7);
      END_STATE();
    case 901:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token8);
      END_STATE();
    case 902:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token9);
      END_STATE();
    case 903:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token10);
      END_STATE();
    case 904:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token11);
      END_STATE();
    case 905:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token12);
      END_STATE();
    case 906:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token13);
      END_STATE();
    case 907:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token14);
      END_STATE();
    case 908:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token15);
      END_STATE();
    case 909:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token16);
      END_STATE();
    case 910:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token17);
      END_STATE();
    case 911:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token18);
      END_STATE();
    case 912:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token19);
      END_STATE();
    case 913:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token20);
      END_STATE();
    case 914:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token21);
      END_STATE();
    case 915:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token22);
      END_STATE();
    case 916:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token23);
      END_STATE();
    case 917:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token24);
      END_STATE();
    case 918:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token25);
      END_STATE();
    case 919:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token26);
      END_STATE();
    case 920:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token27);
      END_STATE();
    case 921:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token28);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(922);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(488);
      END_STATE();
    case 922:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token29);
      END_STATE();
    case 923:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token30);
      END_STATE();
    case 924:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token31);
      END_STATE();
    case 925:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token32);
      END_STATE();
    case 926:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token33);
      END_STATE();
    case 927:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token34);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(417);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(378);
      END_STATE();
    case 928:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token35);
      END_STATE();
    case 929:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token36);
      END_STATE();
    case 930:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token37);
      END_STATE();
    case 931:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token37);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 932:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token37);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 933:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token37);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 934:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token38);
      END_STATE();
    case 935:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token39);
      END_STATE();
    case 936:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token40);
      END_STATE();
    case 937:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token41);
      END_STATE();
    case 938:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token42);
      END_STATE();
    case 939:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token43);
      END_STATE();
    case 940:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token44);
      END_STATE();
    case 941:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token45);
      END_STATE();
    case 942:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token46);
      END_STATE();
    case 943:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token47);
      END_STATE();
    case 944:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token48);
      END_STATE();
    case 945:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token49);
      END_STATE();
    case 946:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token50);
      END_STATE();
    case 947:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token50);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 948:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token50);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 949:
      ACCEPT_TOKEN(aux_sym_comparison_operator_token50);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 950:
      ACCEPT_TOKEN(aux_sym_format_operator_token1);
      END_STATE();
    case 951:
      ACCEPT_TOKEN(aux_sym_format_operator_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(370);
      END_STATE();
    case 952:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DOLLAR);
      END_STATE();
    case 953:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DOLLAR);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 954:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DOLLAR);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 955:
      ACCEPT_TOKEN(anon_sym_DOLLAR_CARET);
      END_STATE();
    case 956:
      ACCEPT_TOKEN(anon_sym_DOLLAR_CARET);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 957:
      ACCEPT_TOKEN(anon_sym_DOLLAR_CARET);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 958:
      ACCEPT_TOKEN(anon_sym_DOLLAR_QMARK);
      END_STATE();
    case 959:
      ACCEPT_TOKEN(anon_sym_DOLLAR_QMARK);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 960:
      ACCEPT_TOKEN(anon_sym_DOLLAR_QMARK);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 961:
      ACCEPT_TOKEN(anon_sym_DOLLAR_);
      if (lookahead == ':') ADVANCE(210);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 962:
      ACCEPT_TOKEN(anon_sym_DOLLAR_);
      if (lookahead == ':') ADVANCE(1214);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 963:
      ACCEPT_TOKEN(anon_sym_DOLLAR_);
      if (lookahead == ':') ADVANCE(1261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 964:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      END_STATE();
    case 965:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(990);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 966:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(977);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 967:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(966);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 968:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(966);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 969:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(988);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 970:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(995);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 971:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(979);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 972:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(995);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 973:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(992);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 974:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(980);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 975:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(985);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 976:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(971);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 977:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(995);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 978:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(981);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 979:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(982);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 980:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(972);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 981:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(967);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 982:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(993);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 983:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(987);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 984:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(968);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 985:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(991);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 986:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(973);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 987:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(976);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 988:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(975);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 989:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 990:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(970);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 991:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(995);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 992:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(965);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 993:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(995);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 994:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(210);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 995:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(198);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 996:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1021);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 997:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1008);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 998:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(997);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 999:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(997);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1000:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1019);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1001:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1026);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1002:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1010);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1003:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1026);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1004:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1023);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1005:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1011);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1006:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1016);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1007:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1002);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1008:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1026);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1009:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1012);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1010:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1013);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1011:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1003);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1012:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(998);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1013:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1024);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1014:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1018);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1015:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(999);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1016:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1022);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1017:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1004);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1018:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1007);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1019:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1006);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1020:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1005);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1021:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1001);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1022:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1026);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1023:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(996);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1024:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1026);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1025:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1214);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1026:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1210);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1027:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1052);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1028:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1039);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1029:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1028);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1030:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1028);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1031:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1050);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1032:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1057);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1033:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1041);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1034:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1057);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1035:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1054);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1036:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1042);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1037:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1047);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1038:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1033);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1039:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1057);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1040:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1043);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1041:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1044);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1042:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1043:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1029);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1044:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1055);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1045:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1049);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1046:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1030);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1047:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1053);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1048:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1035);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1049:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1038);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1050:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1037);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1051:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1036);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1052:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1032);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1053:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1057);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1054:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1027);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1055:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1057);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1056:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1057:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead == ':') ADVANCE(1258);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1058:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1058);
      END_STATE();
    case 1059:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1059);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1060:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1060);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1061:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1062:
      ACCEPT_TOKEN(aux_sym_variable_token1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1063:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      END_STATE();
    case 1064:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1089);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1065:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1076);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1066:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1065);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1067:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1065);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1068:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1087);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1069:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1094);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1070:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1078);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1071:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1094);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1072:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1091);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1073:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1079);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1074:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1084);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1075:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1070);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1076:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1094);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1077:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1080);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1078:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1081);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1079:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1071);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1080:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1066);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1081:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1092);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1082:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1086);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1083:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1067);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1084:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1090);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1085:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1072);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1086:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1075);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1087:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1074);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1088:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1073);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1089:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1069);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1090:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1094);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1091:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1064);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1092:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1094);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1093:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1379);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1094:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1377);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1095:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1120);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1096:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1107);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1097:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1096);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1098:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1096);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1099:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1118);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1100:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1101:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1109);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1102:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1103:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1122);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1104:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1110);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1105:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1115);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1106:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1101);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1107:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1108:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1111);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1109:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1112);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1110:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1102);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1111:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1097);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1112:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1123);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1113:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1114:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1098);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1115:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1121);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1116:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1103);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1117:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1106);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1118:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1105);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1119:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1104);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1120:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1100);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1121:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1122:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1095);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1123:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1124:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(209);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1125:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(199);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1124);
      END_STATE();
    case 1126:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1151);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1127:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1138);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1128:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1127);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1129:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1127);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1130:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1149);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1131:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1132:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1133:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1134:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1153);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1135:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1141);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1136:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1146);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1137:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1132);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1138:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1139:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1142);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1140:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1143);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1141:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1133);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1142:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1128);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1143:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1154);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1144:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1148);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1145:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1129);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1146:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1152);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1147:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1134);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1148:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1149:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1136);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1150:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1135);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1151:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1131);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1152:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1153:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1126);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1154:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1155:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1213);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1156:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1211);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1157:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1182);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1158:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1169);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1159:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1158);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1160:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1158);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1161:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1180);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1162:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1163:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1171);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1164:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1165:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1184);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1166:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1172);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1167:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1177);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1168:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1163);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1169:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1170:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1173);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1171:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1174);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1172:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1164);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1173:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1174:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1185);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1175:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1179);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1176:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1160);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1177:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1178:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1165);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1179:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1168);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1180:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1167);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1181:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1166);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1182:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1162);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1183:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1184:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1157);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1185:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1186:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1260);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1187:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead == ':') ADVANCE(1259);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1188:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1188);
      END_STATE();
    case 1189:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1189);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1190:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1190);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1191:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1191);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1192:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1193:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1194:
      ACCEPT_TOKEN(aux_sym_variable_token2);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(1380);
      END_STATE();
    case 1195:
      ACCEPT_TOKEN(sym_braced_variable);
      END_STATE();
    case 1196:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '\n') ADVANCE(658);
      if (lookahead == '\r') ADVANCE(4);
      if (lookahead == ' ') ADVANCE(5);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1197:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '\n') ADVANCE(9);
      if (lookahead == '\r') ADVANCE(6);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(6);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1198:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '\r') ADVANCE(1);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1205);
      END_STATE();
    case 1199:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        ' ', 1799,
        '!', 1847,
        '"', 601,
        '#', 527,
        '$', 1206,
        '\'', 1208,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1830,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ':', 1805,
        ';', 1246,
        '<', 880,
        '>', 838,
        '@', 1203,
        '[', 828,
        '`', 1246,
        '{', 1319,
      );
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1199);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          (lookahead < '&' || '.' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 1200:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        ' ', 1800,
        '!', 1847,
        '"', 601,
        '#', 527,
        '$', 1206,
        '\'', 1208,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1831,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ';', 1246,
        '<', 880,
        '>', 838,
        '@', 1203,
        '[', 828,
        '`', 1246,
        '{', 1319,
      );
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1200);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          (lookahead < '&' || '.' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 1201:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        ' ', 1801,
        '!', 1847,
        '"', 601,
        '#', 527,
        '$', 1206,
        '\'', 1208,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1833,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ':', 1805,
        ';', 1246,
        '<', 880,
        '>', 838,
        '@', 1203,
        '[', 828,
        '`', 1246,
        '{', 1319,
      );
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1201);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          (lookahead < '&' || '.' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 1202:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        '!', 1847,
        '"', 601,
        '#', 527,
        '$', 1206,
        '\'', 1208,
        '*', 1212,
        '+', 1822,
        ',', 1309,
        '-', 1831,
        '.', 1240,
        '0', 555,
        '1', 549,
        '2', 550,
        '3', 551,
        '4', 552,
        '5', 553,
        '6', 554,
        ';', 1246,
        '<', 880,
        '>', 838,
        '@', 1203,
        '[', 828,
        '`', 1246,
        '{', 1319,
      );
      if (('7' <= lookahead && lookahead <= '9')) ADVANCE(556);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1202);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          (lookahead < ' ' || '$' < lookahead) &&
          (lookahead < '&' || '.' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 1203:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        '"', 1196,
        '\'', 1197,
        '(', 1862,
        '?', 1192,
        '{', 1864,
        'G', 1139,
        'g', 1139,
        'L', 1145,
        'l', 1145,
        'P', 1147,
        'p', 1147,
        'S', 1130,
        's', 1130,
        'U', 1150,
        'u', 1150,
        'W', 1144,
        'w', 1144,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1155);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ',' &&
          lookahead != ';' &&
          (lookahead < 'a' || '}' < lookahead)) ADVANCE(1246);
      END_STATE();
    case 1204:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '#') ADVANCE(1204);
      if (lookahead == '>') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1205);
      END_STATE();
    case 1205:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '#') ADVANCE(1204);
      if (lookahead == '`') ADVANCE(1198);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1205);
      END_STATE();
    case 1206:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        '$', 953,
        '(', 1861,
        '?', 959,
        '^', 956,
        '_', 962,
        '{', 1243,
        'G', 1009,
        'g', 1009,
        'L', 1015,
        'l', 1015,
        'P', 1017,
        'p', 1017,
        'S', 1000,
        's', 1000,
        'U', 1020,
        'u', 1020,
        'W', 1014,
        'w', 1014,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1025);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '\'' ||
          lookahead == '-' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1207:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '&') ADVANCE(195);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1208:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '\'') ADVANCE(708);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|' ||
          lookahead == '}') ADVANCE(180);
      if (lookahead != 0) ADVANCE(1208);
      END_STATE();
    case 1209:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '-') ADVANCE(1209);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1210:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == ':') ADVANCE(1214);
      if (lookahead == '?') ADVANCE(1061);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1059);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1211:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == ':') ADVANCE(1213);
      if (lookahead == '?') ADVANCE(1192);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1189);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          (lookahead < '0' || ';' < lookahead) &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1212:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '>') ADVANCE(874);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1213:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '?') ADVANCE(1192);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1189);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1214:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '?') ADVANCE(1061);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1059);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1215:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'b') ADVANCE(566);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1216:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'b') ADVANCE(598);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1217:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'b') ADVANCE(581);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1218:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '}') ADVANCE(1195);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|') ADVANCE(218);
      if (lookahead != 0) ADVANCE(1218);
      END_STATE();
    case 1219:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(1241);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(594);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '+' || '-' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1220:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1226);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1221:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1238);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1222:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1227);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1223:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1239);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1224:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1221);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1225:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1223);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1226:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(932);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1227:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(933);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1228:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1231);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1229:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1233);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1230:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1220);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1231:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1237);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1232:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1222);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1233:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1236);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1234:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1224);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1235:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1225);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1236:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1246);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1237:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1245);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1238:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(948);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1239:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(949);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1240:
      ACCEPT_TOKEN(sym_generic_token);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1241:
      ACCEPT_TOKEN(sym_generic_token);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(594);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1242:
      ACCEPT_TOKEN(sym_generic_token);
      ADVANCE_MAP(
        '\t', 1797,
        0x0b, 1797,
        '\f', 1797,
        ' ', 1797,
        '&', 1797,
        '(', 1797,
        ')', 1797,
        ',', 1797,
        ';', 1797,
        '|', 1797,
        '}', 1797,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1242);
      END_STATE();
    case 1243:
      ACCEPT_TOKEN(sym_generic_token);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '|') ADVANCE(218);
      if (lookahead != 0 &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1218);
      END_STATE();
    case 1244:
      ACCEPT_TOKEN(sym_generic_token);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(576);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1245:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1246:
      ACCEPT_TOKEN(sym_generic_token);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1247:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '\n') ADVANCE(658);
      if (lookahead == '\r') ADVANCE(4);
      if (lookahead == ' ') ADVANCE(5);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1248:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '\n') ADVANCE(9);
      if (lookahead == '\r') ADVANCE(6);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(6);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1249:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '\r') ADVANCE(1);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ';' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1255);
      END_STATE();
    case 1250:
      ACCEPT_TOKEN(sym__command_token);
      ADVANCE_MAP(
        '"', 602,
        '#', 526,
        '$', 1256,
        '\'', 1257,
        '.', 1266,
        '0', 557,
        '<', 1253,
        '@', 1251,
        '[', 829,
        '`', 1269,
        0xa0, 1250,
        0x200b, 1250,
        0x2060, 1250,
        0xfeff, 1250,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(558);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '<' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1251:
      ACCEPT_TOKEN(sym__command_token);
      ADVANCE_MAP(
        '"', 1247,
        '\'', 1248,
        '(', 1862,
        '?', 1193,
        '{', 1863,
        'G', 1170,
        'g', 1170,
        'L', 1176,
        'l', 1176,
        'P', 1178,
        'p', 1178,
        'S', 1161,
        's', 1161,
        'U', 1181,
        'u', 1181,
        'W', 1175,
        'w', 1175,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1186);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1252:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '#') ADVANCE(526);
      if (lookahead == '<') ADVANCE(1253);
      if (lookahead == '`') ADVANCE(1269);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1252);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          ('"' <= lookahead && lookahead <= '$') ||
          lookahead == '\'' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '|') &&
          lookahead != ';' &&
          lookahead != '<') ADVANCE(1269);
      END_STATE();
    case 1253:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '#') ADVANCE(1255);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1254:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '#') ADVANCE(1254);
      if (lookahead == '>') ADVANCE(530);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ';' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1255);
      END_STATE();
    case 1255:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '#') ADVANCE(1254);
      if (lookahead == '`') ADVANCE(1249);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '&' ||
          lookahead == '(' ||
          lookahead == ')' ||
          lookahead == ';' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1255);
      END_STATE();
    case 1256:
      ACCEPT_TOKEN(sym__command_token);
      ADVANCE_MAP(
        '$', 954,
        '(', 1861,
        '?', 960,
        '^', 957,
        '_', 963,
        '{', 498,
        'G', 1040,
        'g', 1040,
        'L', 1046,
        'l', 1046,
        'P', 1048,
        'p', 1048,
        'S', 1031,
        's', 1031,
        'U', 1051,
        'u', 1051,
        'W', 1045,
        'w', 1045,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1056);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1257:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '\'') ADVANCE(709);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ';' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(180);
      if (lookahead != 0) ADVANCE(1257);
      END_STATE();
    case 1258:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == ':') ADVANCE(1261);
      if (lookahead == '?') ADVANCE(1062);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1060);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1259:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == ':') ADVANCE(1260);
      if (lookahead == '?') ADVANCE(1193);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1190);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          (lookahead < '0' || ';' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1260:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '?') ADVANCE(1193);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1190);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1261:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '?') ADVANCE(1062);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1060);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < 'a' || '{' < lookahead) &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1262:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == 'b') ADVANCE(567);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1263:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == 'b') ADVANCE(599);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1264:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == 'b') ADVANCE(582);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1265:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(1267);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(595);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1266:
      ACCEPT_TOKEN(sym__command_token);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(590);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1267:
      ACCEPT_TOKEN(sym__command_token);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(595);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1268:
      ACCEPT_TOKEN(sym__command_token);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(578);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1269:
      ACCEPT_TOKEN(sym__command_token);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '{' &&
          lookahead != '}') ADVANCE(1269);
      END_STATE();
    case 1270:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '%') ADVANCE(1809);
      if (lookahead == '-') ADVANCE(1274);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1271:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '%') ADVANCE(1809);
      if (lookahead == '-') ADVANCE(1275);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1272:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '%') ADVANCE(1797);
      if (lookahead == '-') ADVANCE(1274);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1273:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '%') ADVANCE(1797);
      if (lookahead == '-') ADVANCE(1275);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1274:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '-') ADVANCE(1274);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1275:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '-') ADVANCE(1275);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1276:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1277:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1280);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1278:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1286);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1279:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1278);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1280:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(931);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1281:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1283);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1282:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1277);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1283:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1285);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1284:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1279);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1285:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1287);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1286:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(947);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1287:
      ACCEPT_TOKEN(sym_command_parameter);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1288:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token1);
      END_STATE();
    case 1289:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1290:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token2);
      if (lookahead == '&') ADVANCE(1301);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '|') ADVANCE(1292);
      if (lookahead != 0) ADVANCE(1290);
      END_STATE();
    case 1291:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token2);
      if (lookahead == '&') ADVANCE(1368);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '|') ADVANCE(1292);
      if (lookahead != 0) ADVANCE(1290);
      END_STATE();
    case 1292:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token2);
      if (lookahead != 0 &&
          lookahead != '&') ADVANCE(1292);
      END_STATE();
    case 1293:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead == '\r') ADVANCE(1);
      if (lookahead == '\n' ||
          lookahead == '|') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1300);
      END_STATE();
    case 1294:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      ADVANCE_MAP(
        '"', 1297,
        '#', 524,
        '&', 1291,
        ')', 1307,
        '<', 1298,
        '`', 1301,
        '\t', 1294,
        0x0b, 1294,
        '\f', 1294,
        ' ', 1294,
        0xa0, 1294,
        0x200b, 1294,
        0x2060, 1294,
        0xfeff, 1294,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1295:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      ADVANCE_MAP(
        '"', 1297,
        '#', 524,
        '&', 1291,
        '<', 1298,
        '`', 1301,
        '\t', 1295,
        0x0b, 1295,
        '\f', 1295,
        ' ', 1295,
        0xa0, 1295,
        0x200b, 1295,
        0x2060, 1295,
        0xfeff, 1295,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1296:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      ADVANCE_MAP(
        '"', 1297,
        '#', 524,
        '&', 1290,
        '<', 1298,
        '`', 1301,
        '\t', 1296,
        0x0b, 1296,
        '\f', 1296,
        ' ', 1296,
        0xa0, 1296,
        0x200b, 1296,
        0x2060, 1296,
        0xfeff, 1296,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1297:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead == '"') ADVANCE(1289);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '|') ADVANCE(158);
      if (lookahead != 0) ADVANCE(1297);
      END_STATE();
    case 1298:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead == '#') ADVANCE(1300);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1299:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead == '#') ADVANCE(1299);
      if (lookahead == '>') ADVANCE(531);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '|') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1300);
      END_STATE();
    case 1300:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead == '#') ADVANCE(1299);
      if (lookahead == '`') ADVANCE(1293);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '|') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1300);
      END_STATE();
    case 1301:
      ACCEPT_TOKEN(aux_sym__verbatim_command_argument_chars_token3);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1302:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 1303:
      ACCEPT_TOKEN(aux_sym_param_block_token1);
      END_STATE();
    case 1304:
      ACCEPT_TOKEN(aux_sym_param_block_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1305:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 1306:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 1307:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1308:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 1309:
      ACCEPT_TOKEN(anon_sym_COMMA);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1310:
      ACCEPT_TOKEN(aux_sym_block_name_token1);
      END_STATE();
    case 1311:
      ACCEPT_TOKEN(aux_sym_block_name_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1312:
      ACCEPT_TOKEN(aux_sym_block_name_token2);
      END_STATE();
    case 1313:
      ACCEPT_TOKEN(aux_sym_block_name_token2);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1314:
      ACCEPT_TOKEN(aux_sym_block_name_token3);
      END_STATE();
    case 1315:
      ACCEPT_TOKEN(aux_sym_block_name_token3);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1316:
      ACCEPT_TOKEN(aux_sym_block_name_token4);
      END_STATE();
    case 1317:
      ACCEPT_TOKEN(aux_sym_block_name_token4);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1318:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 1319:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1320:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 1321:
      ACCEPT_TOKEN(aux_sym_if_statement_token1);
      if (lookahead == '-') ADVANCE(1728);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1322:
      ACCEPT_TOKEN(aux_sym_elseif_clause_token1);
      END_STATE();
    case 1323:
      ACCEPT_TOKEN(aux_sym_elseif_clause_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1324:
      ACCEPT_TOKEN(aux_sym_else_clause_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(318);
      END_STATE();
    case 1325:
      ACCEPT_TOKEN(aux_sym_else_clause_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1569);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1326:
      ACCEPT_TOKEN(aux_sym_else_clause_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1327:
      ACCEPT_TOKEN(aux_sym_switch_statement_token1);
      END_STATE();
    case 1328:
      ACCEPT_TOKEN(aux_sym_switch_statement_token1);
      if (lookahead == '-') ADVANCE(1740);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1329:
      ACCEPT_TOKEN(aux_sym_switch_parameter_token1);
      END_STATE();
    case 1330:
      ACCEPT_TOKEN(aux_sym_switch_parameter_token2);
      END_STATE();
    case 1331:
      ACCEPT_TOKEN(aux_sym_switch_parameter_token3);
      END_STATE();
    case 1332:
      ACCEPT_TOKEN(aux_sym_switch_parameter_token4);
      END_STATE();
    case 1333:
      ACCEPT_TOKEN(aux_sym_switch_parameter_token5);
      END_STATE();
    case 1334:
      ACCEPT_TOKEN(aux_sym_switch_condition_token1);
      END_STATE();
    case 1335:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token1);
      END_STATE();
    case 1336:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1337:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token2);
      END_STATE();
    case 1338:
      ACCEPT_TOKEN(aux_sym_for_statement_token1);
      if (lookahead == '-') ADVANCE(1742);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1524);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1339:
      ACCEPT_TOKEN(aux_sym_for_statement_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(234);
      END_STATE();
    case 1340:
      ACCEPT_TOKEN(aux_sym_for_statement_token2);
      END_STATE();
    case 1341:
      ACCEPT_TOKEN(aux_sym_while_statement_token1);
      END_STATE();
    case 1342:
      ACCEPT_TOKEN(aux_sym_while_statement_token1);
      if (lookahead == '-') ADVANCE(1744);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1343:
      ACCEPT_TOKEN(aux_sym_do_statement_token1);
      END_STATE();
    case 1344:
      ACCEPT_TOKEN(aux_sym_do_statement_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1345:
      ACCEPT_TOKEN(aux_sym_do_statement_token2);
      END_STATE();
    case 1346:
      ACCEPT_TOKEN(aux_sym_function_statement_token1);
      if (lookahead == '-') ADVANCE(1730);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1347:
      ACCEPT_TOKEN(aux_sym_function_statement_token2);
      if (lookahead == '-') ADVANCE(1732);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1348:
      ACCEPT_TOKEN(aux_sym_function_statement_token3);
      if (lookahead == '-') ADVANCE(1734);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1349:
      ACCEPT_TOKEN(aux_sym_flow_control_statement_token1);
      if (lookahead == '-') ADVANCE(1714);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1350:
      ACCEPT_TOKEN(aux_sym_flow_control_statement_token2);
      if (lookahead == '-') ADVANCE(1716);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1351:
      ACCEPT_TOKEN(aux_sym_flow_control_statement_token3);
      if (lookahead == '-') ADVANCE(1718);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1352:
      ACCEPT_TOKEN(aux_sym_flow_control_statement_token4);
      if (lookahead == '-') ADVANCE(1720);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1353:
      ACCEPT_TOKEN(aux_sym_flow_control_statement_token5);
      if (lookahead == '-') ADVANCE(1722);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1354:
      ACCEPT_TOKEN(sym_label);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1354);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1355:
      ACCEPT_TOKEN(sym_label);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1355);
      END_STATE();
    case 1356:
      ACCEPT_TOKEN(aux_sym_trap_statement_token1);
      if (lookahead == '-') ADVANCE(1726);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1357:
      ACCEPT_TOKEN(aux_sym_try_statement_token1);
      if (lookahead == '-') ADVANCE(1724);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1358:
      ACCEPT_TOKEN(aux_sym_catch_clause_token1);
      END_STATE();
    case 1359:
      ACCEPT_TOKEN(aux_sym_catch_clause_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1360:
      ACCEPT_TOKEN(aux_sym_finally_clause_token1);
      END_STATE();
    case 1361:
      ACCEPT_TOKEN(aux_sym_finally_clause_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1362:
      ACCEPT_TOKEN(aux_sym_data_statement_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1363:
      ACCEPT_TOKEN(aux_sym_data_commands_allowed_token1);
      END_STATE();
    case 1364:
      ACCEPT_TOKEN(aux_sym_inlinescript_statement_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1365:
      ACCEPT_TOKEN(aux_sym_parallel_statement_token1);
      if (lookahead == '-') ADVANCE(1746);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1366:
      ACCEPT_TOKEN(aux_sym_sequence_statement_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1367:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 1368:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '|') ADVANCE(1301);
      END_STATE();
    case 1369:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 1370:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(1369);
      END_STATE();
    case 1371:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 1372:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(1367);
      END_STATE();
    case 1373:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == '#') ADVANCE(525);
      if (lookahead == '<') ADVANCE(1374);
      if (lookahead == '@') ADVANCE(1378);
      if (lookahead == '`') ADVANCE(51);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1373);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1374:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == '#') ADVANCE(1376);
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          lookahead != '`') ADVANCE(1380);
      END_STATE();
    case 1375:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == '#') ADVANCE(1375);
      if (lookahead == '>') ADVANCE(532);
      if (('"' <= lookahead && lookahead <= '$') ||
          lookahead == '`') ADVANCE(172);
      if (lookahead != 0) ADVANCE(1376);
      END_STATE();
    case 1376:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == '#') ADVANCE(1375);
      if (lookahead == '`') ADVANCE(103);
      if (('"' <= lookahead && lookahead <= '$')) ADVANCE(172);
      if (lookahead != 0) ADVANCE(1376);
      END_STATE();
    case 1377:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == ':') ADVANCE(1379);
      if (lookahead == '?') ADVANCE(1194);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1191);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1378:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      ADVANCE_MAP(
        '?', 1194,
        'G', 1077,
        'g', 1077,
        'L', 1083,
        'l', 1083,
        'P', 1085,
        'p', 1085,
        'S', 1068,
        's', 1068,
        'U', 1088,
        'u', 1088,
        'W', 1082,
        'w', 1082,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1093);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1379:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead == '?') ADVANCE(1194);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1191);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          (lookahead < '_' || 'z' < lookahead)) ADVANCE(1380);
      END_STATE();
    case 1380:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token1);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(1380);
      END_STATE();
    case 1381:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token2);
      END_STATE();
    case 1382:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token2);
      if (lookahead == '\n') ADVANCE(1381);
      if (lookahead == '\r') ADVANCE(52);
      END_STATE();
    case 1383:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token3);
      END_STATE();
    case 1384:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token3);
      if (lookahead == '\n') ADVANCE(1385);
      END_STATE();
    case 1385:
      ACCEPT_TOKEN(aux_sym__expandable_string_literal_immediate_token3);
      if (lookahead == '`') ADVANCE(51);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1373);
      END_STATE();
    case 1386:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE2);
      END_STATE();
    case 1387:
      ACCEPT_TOKEN(anon_sym_DOLLAR2);
      END_STATE();
    case 1388:
      ACCEPT_TOKEN(anon_sym_DOLLAR2);
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '`', 214,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 1389:
      ACCEPT_TOKEN(anon_sym_DOLLAR2);
      ADVANCE_MAP(
        '$', 952,
        '(', 1861,
        '?', 958,
        '^', 955,
        '_', 961,
        '{', 498,
        'G', 978,
        'g', 978,
        'L', 984,
        'l', 984,
        'P', 986,
        'p', 986,
        'S', 969,
        's', 969,
        'U', 989,
        'u', 989,
        'W', 983,
        'w', 983,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(994);
      END_STATE();
    case 1390:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 1391:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(1386);
      END_STATE();
    case 1392:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1604,
        'c', 1604,
        'D', 1514,
        'd', 1514,
        'E', 1627,
        'e', 1627,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1522,
        'p', 1522,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1392,
        0x200b, 1392,
        0x2060, 1392,
        0xfeff, 1392,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1393:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1604,
        'c', 1604,
        'D', 1514,
        'd', 1514,
        'E', 1627,
        'e', 1627,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1533,
        'p', 1533,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1393,
        0x200b, 1393,
        0x2060, 1393,
        0xfeff, 1393,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1394:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1604,
        'c', 1604,
        'D', 1514,
        'd', 1514,
        'E', 1627,
        'e', 1627,
        'F', 1582,
        'f', 1582,
        'I', 1567,
        'i', 1567,
        'P', 1533,
        'p', 1533,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1394,
        0x200b, 1394,
        0x2060, 1394,
        0xfeff, 1394,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1395:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1604,
        'c', 1604,
        'D', 1514,
        'd', 1514,
        'E', 1609,
        'e', 1609,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1533,
        'p', 1533,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1395,
        0x200b, 1395,
        0x2060, 1395,
        0xfeff, 1395,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1396:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1604,
        'c', 1604,
        'D', 1514,
        'd', 1514,
        'E', 1620,
        'e', 1620,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1533,
        'p', 1533,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1396,
        0x200b, 1396,
        0x2060, 1396,
        0xfeff, 1396,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1397:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1660,
        'b', 1660,
        'C', 1529,
        'c', 1529,
        'D', 1514,
        'd', 1514,
        'E', 1627,
        'e', 1627,
        'F', 1582,
        'f', 1582,
        'I', 1567,
        'i', 1567,
        'P', 1533,
        'p', 1533,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1397,
        0x200b, 1397,
        0x2060, 1397,
        0xfeff, 1397,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1398:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1551,
        'b', 1551,
        'C', 1604,
        'c', 1604,
        'D', 1513,
        'd', 1513,
        'E', 1631,
        'e', 1631,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1521,
        'p', 1521,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1398,
        0x200b, 1398,
        0x2060, 1398,
        0xfeff, 1398,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1399:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        ':', 1712,
        'B', 1551,
        'b', 1551,
        'C', 1604,
        'c', 1604,
        'D', 1513,
        'd', 1513,
        'E', 1631,
        'e', 1631,
        'F', 1581,
        'f', 1581,
        'I', 1567,
        'i', 1567,
        'P', 1532,
        'p', 1532,
        'R', 1554,
        'r', 1554,
        'S', 1547,
        's', 1547,
        'T', 1577,
        't', 1577,
        'W', 1576,
        'w', 1576,
        0xa0, 1399,
        0x200b, 1399,
        0x2060, 1399,
        0xfeff, 1399,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1400:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 825,
        '0', 547,
        'B', 1669,
        'b', 1669,
        'C', 1616,
        'c', 1616,
        'E', 1639,
        'e', 1639,
        'F', 1596,
        'f', 1596,
        'I', 1568,
        'i', 1568,
        'P', 1535,
        'p', 1535,
        'R', 1566,
        'r', 1566,
        'S', 1708,
        's', 1708,
        'T', 1579,
        't', 1579,
        'W', 1580,
        'w', 1580,
        0xa0, 1400,
        0x200b, 1400,
        0x2060, 1400,
        0xfeff, 1400,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1401:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 1489,
        '0', 535,
        'B', 1470,
        'b', 1470,
        'C', 1450,
        'c', 1450,
        'E', 1458,
        'e', 1458,
        'F', 1442,
        'f', 1442,
        'I', 1437,
        'i', 1437,
        'P', 1427,
        'p', 1427,
        'R', 1431,
        'r', 1431,
        'S', 1488,
        's', 1488,
        'T', 1440,
        't', 1440,
        'W', 1439,
        'w', 1439,
        0xa0, 1401,
        0x200b, 1401,
        0x2060, 1401,
        0xfeff, 1401,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(536);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          (lookahead < ' ' || '$' < lookahead) &&
          (lookahead < '&' || '9' < lookahead) &&
          lookahead != ';' &&
          lookahead != '<' &&
          (lookahead < '?' || ']' < lookahead) &&
          (lookahead < '_' || '}' < lookahead)) ADVANCE(1713);
      END_STATE();
    case 1402:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      ADVANCE_MAP(
        '#', 528,
        '.', 826,
        'B', 1669,
        'b', 1669,
        'C', 1616,
        'c', 1616,
        'E', 1639,
        'e', 1639,
        'F', 1596,
        'f', 1596,
        'I', 1568,
        'i', 1568,
        'P', 1535,
        'p', 1535,
        'R', 1566,
        'r', 1566,
        'S', 1708,
        's', 1708,
        'T', 1579,
        't', 1579,
        'W', 1580,
        'w', 1580,
        0xa0, 1402,
        0x200b, 1402,
        0x2060, 1402,
        0xfeff, 1402,
      );
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1403:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '+') ADVANCE(496);
      if (lookahead == '-') ADVANCE(1795);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1404:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1728);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1405:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1742);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1406:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1724);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1407:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1722);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1408:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1726);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1409:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1714);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1410:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1718);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1411:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1744);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1412:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1732);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1413:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1720);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1414:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1740);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1415:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1716);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1416:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1730);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1417:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1746);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1418:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1734);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1419:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1738);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1420:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1736);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1421:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'b') ADVANCE(538);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1422:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'b') ADVANCE(586);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1423:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'b') ADVANCE(571);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1424:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1475);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1425:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1467);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1494);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1426:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1449);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1427:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1472);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1428:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1455);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1429:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1441);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1430:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1481);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1431:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1477);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1432:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1426);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1433:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1454);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1434:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1501);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1435:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1505);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1436:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1474);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1437:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1492);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1438:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1456);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1439:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1443);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1469);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1440:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1468);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1425);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1441:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1504);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1442:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1452);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1471);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1459);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1443:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1451);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1444:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1466);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1445:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1461);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1446:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1482);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1447:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1478);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1448:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1438);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1449:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1498);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1450:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1424);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1460);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1451:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1434);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1452:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1479);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1453:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1433);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1454:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1507);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1455:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1453);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1456:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1465);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1457:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1495);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1458:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1483);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1446);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1459:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1430);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1460:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1480);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1461:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1485);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1462:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1503);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1463:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1506);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1464:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1486);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1465:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1487);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1466:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1463);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1467:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1497);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1468:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1464);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1469:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1448);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1470:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1432);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1471:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1493);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1472:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1428);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1473:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1462);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1474:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1502);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1475:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1476);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1476:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1499);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1477:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1484);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1478:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1429);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1479:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1436);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1480:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1445);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1481:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1444);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1482:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1496);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1483:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1457);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1484:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1473);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1485:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1435);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1486:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1500);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1487:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1508);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1488:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1447);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1489:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(584);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1490:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(569);
      if (lookahead == '.' ||
          lookahead == '?' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1491:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1796);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1492:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1729);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1493:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1743);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1494:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1725);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1495:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1739);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1496:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1723);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1497:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1727);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1498:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1715);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1499:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1737);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1500:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1719);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1501:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1745);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1502:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1733);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1503:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1721);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1504:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1741);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1505:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1717);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1506:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1731);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1507:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1747);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1508:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '-') ADVANCE(1735);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1491);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1509:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'b') ADVANCE(565);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1510:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'b') ADVANCE(597);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1511:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'b') ADVANCE(580);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1512:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(496);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(593);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1513:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1687);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1344);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1636);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1514:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1687);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1344);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1515:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1675);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1516:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1651);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1357);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1517:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1362);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1518:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1617);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1519:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1626);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1520:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1599);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1521:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1662);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1648);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1522:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1662);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1523:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1624);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1524:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1537);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1525:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1600);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1526:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1652);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1406);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1527:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1618);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1528:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1611);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1529:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1692);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1515);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1633);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1530:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1665);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1531:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1677);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1532:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1666);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1648);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1533:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1666);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1534:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1622);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1535:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1671);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1536:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1573);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1537:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1574);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1538:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1654);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1539:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1578);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1540:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1575);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1541:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1664);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1542:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1550);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1543:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1565);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1544:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1689);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1545:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1695);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1546:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1317);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1623);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1547:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1655);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1589);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1548:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1342);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1549:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1350);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1550:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1366);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1551:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1572);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1555);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1552:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1325);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1553:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1326);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1554:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1684);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1555:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1520);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1556:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1678);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1557:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1602);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1558:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1634);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1559:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1659);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1560:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1614);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1561:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1525);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1562:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1411);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1563:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1668);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1564:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1415);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1565:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1676);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1566:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1693);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1567:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1321);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1605);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1568:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1404);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1569:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1323);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1570:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1615);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1571:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1621);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1572:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1592);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1573:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1328);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1574:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1336);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1575:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1359);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1576:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1586);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1658);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1577:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1657);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1516);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1578:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1414);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1579:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1670);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1526);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1580:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1595);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1672);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1581:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1608);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1656);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1628);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1582:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1607);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1656);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1628);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1583:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1647);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1584:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1653);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1585:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1635);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1586:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1606);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1587:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1640);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1588:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1682);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1589:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1685);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1590:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1538);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1591:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1690);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1592:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1632);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1593:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1691);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1594:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1650);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1595:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1612);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1596:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1619);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1661);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1642);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1597:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1641);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1598:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1570);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1599:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1349);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1600:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1409);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1601:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1571);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1602:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1365);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1603:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1709);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1604:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1515);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1633);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1605:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1585);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1606:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1548);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1607:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1686);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1528);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1608:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1686);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1609:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1680);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1697);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1610:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1557);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1611:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1603);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1612:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1562);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1613:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1560);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1614:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1417);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1615:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1645);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1616:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1531);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1643);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1617:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1610);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1304);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1618:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1610);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1619:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1694);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1620:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1681);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1697);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1621:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1649);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1622:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1613);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1623:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1813);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1624:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1311);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1625:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1419);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1626:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1590);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1627:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1697);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1628:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1544);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1629:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1352);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1630:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1346);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1631:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1546);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1632:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1313);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1633:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1688);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1634:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1542);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1635:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1556);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1636:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1519);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1637:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1413);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1638:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1416);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1639:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1699);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(1591);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1640:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1701);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1641:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1703);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1642:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1545);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1643:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1696);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1644:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1704);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1645:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1705);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1646:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1706);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1647:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1630);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1648:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1543);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1649:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1707);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1650:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1638);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1651:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1356);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1652:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1408);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1653:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1683);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1654:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1530);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1655:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(1698);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1656:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1338);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1657:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1644);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1658:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1598);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1659:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1347);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1660:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1555);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1661:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1405);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1662:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1518);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1663:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1629);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1664:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1584);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1665:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1523);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1666:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1527);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1667:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1637);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1668:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1412);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1669:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1561);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1670:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1646);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1671:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1534);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1672:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1601);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1673:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1812);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1674:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1315);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1675:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1673);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1676:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1674);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1677:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1679);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1678:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1541);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1679:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1420);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1680:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1552);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1681:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1553);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1682:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1353);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1683:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1364);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1684:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1700);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1685:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1536);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1686:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1559);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1687:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1517);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1688:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1587);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1689:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1583);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1690:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1407);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1691:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1539);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1692:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1540);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1693:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1702);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1694:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1563);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1695:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1594);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1696:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1597);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1697:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1623);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1698:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1558);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1699:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1625);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1700:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1663);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1701:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1549);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1702:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1667);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1703:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1564);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1704:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1351);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1705:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1348);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1706:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1410);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1707:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1418);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1708:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1593);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1709:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1361);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1710:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1711:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(574);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1712:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1354);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1713:
      ACCEPT_TOKEN(aux_sym_command_name_token1);
      if ((!eof && set_contains(aux_sym_command_name_token1_character_set_1, 12, lookahead))) ADVANCE(1713);
      END_STATE();
    case 1714:
      ACCEPT_TOKEN(aux_sym_command_name_token2);
      END_STATE();
    case 1715:
      ACCEPT_TOKEN(aux_sym_command_name_token2);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1716:
      ACCEPT_TOKEN(aux_sym_command_name_token3);
      END_STATE();
    case 1717:
      ACCEPT_TOKEN(aux_sym_command_name_token3);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1718:
      ACCEPT_TOKEN(aux_sym_command_name_token4);
      END_STATE();
    case 1719:
      ACCEPT_TOKEN(aux_sym_command_name_token4);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1720:
      ACCEPT_TOKEN(aux_sym_command_name_token5);
      END_STATE();
    case 1721:
      ACCEPT_TOKEN(aux_sym_command_name_token5);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1722:
      ACCEPT_TOKEN(aux_sym_command_name_token6);
      END_STATE();
    case 1723:
      ACCEPT_TOKEN(aux_sym_command_name_token6);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1724:
      ACCEPT_TOKEN(aux_sym_command_name_token7);
      END_STATE();
    case 1725:
      ACCEPT_TOKEN(aux_sym_command_name_token7);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1726:
      ACCEPT_TOKEN(aux_sym_command_name_token8);
      END_STATE();
    case 1727:
      ACCEPT_TOKEN(aux_sym_command_name_token8);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1728:
      ACCEPT_TOKEN(aux_sym_command_name_token9);
      END_STATE();
    case 1729:
      ACCEPT_TOKEN(aux_sym_command_name_token9);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1730:
      ACCEPT_TOKEN(aux_sym_command_name_token10);
      END_STATE();
    case 1731:
      ACCEPT_TOKEN(aux_sym_command_name_token10);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1732:
      ACCEPT_TOKEN(aux_sym_command_name_token11);
      END_STATE();
    case 1733:
      ACCEPT_TOKEN(aux_sym_command_name_token11);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1734:
      ACCEPT_TOKEN(aux_sym_command_name_token12);
      END_STATE();
    case 1735:
      ACCEPT_TOKEN(aux_sym_command_name_token12);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1736:
      ACCEPT_TOKEN(aux_sym_command_name_token13);
      END_STATE();
    case 1737:
      ACCEPT_TOKEN(aux_sym_command_name_token13);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1738:
      ACCEPT_TOKEN(aux_sym_command_name_token14);
      END_STATE();
    case 1739:
      ACCEPT_TOKEN(aux_sym_command_name_token14);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1740:
      ACCEPT_TOKEN(aux_sym_command_name_token15);
      END_STATE();
    case 1741:
      ACCEPT_TOKEN(aux_sym_command_name_token15);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1742:
      ACCEPT_TOKEN(aux_sym_command_name_token16);
      END_STATE();
    case 1743:
      ACCEPT_TOKEN(aux_sym_command_name_token16);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1744:
      ACCEPT_TOKEN(aux_sym_command_name_token17);
      END_STATE();
    case 1745:
      ACCEPT_TOKEN(aux_sym_command_name_token17);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1746:
      ACCEPT_TOKEN(aux_sym_command_name_token18);
      END_STATE();
    case 1747:
      ACCEPT_TOKEN(aux_sym_command_name_token18);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1748:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '\r') ADVANCE(1);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(172);
      if (lookahead != 0) ADVANCE(1754);
      END_STATE();
    case 1749:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      ADVANCE_MAP(
        '#', 1768,
        '*', 1761,
        '-', 1758,
        '1', 1762,
        '2', 1763,
        '3', 1764,
        '4', 1765,
        '5', 1766,
        '6', 1767,
        ':', 1804,
        '<', 881,
        '>', 839,
        '`', 1771,
        0xa0, 1749,
        0x200b, 1749,
        0x2060, 1749,
        0xfeff, 1749,
      );
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '@' ||
          lookahead == '[') &&
          lookahead != ',' &&
          (lookahead < ':' || '<' < lookahead)) ADVANCE(1771);
      END_STATE();
    case 1750:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      ADVANCE_MAP(
        '#', 1768,
        '*', 1761,
        '-', 1760,
        '1', 1762,
        '2', 1763,
        '3', 1764,
        '4', 1765,
        '5', 1766,
        '6', 1767,
        ':', 1804,
        '<', 881,
        '>', 839,
        '`', 1771,
        0xa0, 1750,
        0x200b, 1750,
        0x2060, 1750,
        0xfeff, 1750,
      );
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '@' ||
          lookahead == '[') &&
          lookahead != ',' &&
          (lookahead < ':' || '<' < lookahead)) ADVANCE(1771);
      END_STATE();
    case 1751:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '#') ADVANCE(1768);
      if (lookahead == '<') ADVANCE(1752);
      if (lookahead == '`') ADVANCE(1771);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1751);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[') &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '<') ADVANCE(1771);
      END_STATE();
    case 1752:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '#') ADVANCE(1754);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1753:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '#') ADVANCE(1753);
      if (lookahead == '>') ADVANCE(1771);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(172);
      if (lookahead != 0) ADVANCE(1754);
      END_STATE();
    case 1754:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '#') ADVANCE(1753);
      if (lookahead == '`') ADVANCE(1748);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(172);
      if (lookahead != 0) ADVANCE(1754);
      END_STATE();
    case 1755:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '%') ADVANCE(1808);
      if (lookahead == '-') ADVANCE(1759);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1756:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '%') ADVANCE(1769);
      if (lookahead == '-') ADVANCE(1759);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1757:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '&') ADVANCE(195);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1758:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '-') ADVANCE(1755);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1759:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '-') ADVANCE(1759);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1760:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '-') ADVANCE(1756);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1761:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(875);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1762:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(1757);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1763:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(845);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1764:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(851);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1765:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(857);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1766:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(863);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1767:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '>') ADVANCE(869);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1768:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(533);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1768);
      END_STATE();
    case 1769:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1797);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1769);
      END_STATE();
    case 1770:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (lookahead == '-' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1770);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          ('@' <= lookahead && lookahead <= '[')) &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1771:
      ACCEPT_TOKEN(aux_sym_command_name_token19);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1772:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 1773:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      if (lookahead == '"') ADVANCE(651);
      END_STATE();
    case 1774:
      ACCEPT_TOKEN(anon_sym_SQUOTE_SQUOTE);
      END_STATE();
    case 1775:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '%') ADVANCE(1806);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1776:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1854);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1777:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1271);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1778:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1859);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1779:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1775);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1780:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1856);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1781:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-') ADVANCE(1273);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1782:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(178);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1783:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(843);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1784:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(849);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1785:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(855);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1786:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(861);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1787:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '>') ADVANCE(867);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1788:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'b') ADVANCE(596);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1789:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1790);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1790:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1792);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1791:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1789);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1792:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1873);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1793:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1794);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1794:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1791);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1795:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(591);
      if (lookahead == '-' ||
          lookahead == '.' ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1796:
      ACCEPT_TOKEN(sym_path_command_name_token);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1797:
      ACCEPT_TOKEN(sym_stop_parsing);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1797);
      END_STATE();
    case 1798:
      ACCEPT_TOKEN(anon_sym_SPACE);
      if (lookahead == ' ') ADVANCE(1798);
      END_STATE();
    case 1799:
      ACCEPT_TOKEN(anon_sym_SPACE);
      if (lookahead == ' ') ADVANCE(1799);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1199);
      END_STATE();
    case 1800:
      ACCEPT_TOKEN(anon_sym_SPACE);
      if (lookahead == ' ') ADVANCE(1800);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1200);
      END_STATE();
    case 1801:
      ACCEPT_TOKEN(anon_sym_SPACE);
      if (lookahead == ' ') ADVANCE(1801);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(1201);
      END_STATE();
    case 1802:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 1803:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(1871);
      END_STATE();
    case 1804:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '$' ||
          lookahead == '-' ||
          lookahead == '@' ||
          lookahead == '[' ||
          lookahead == '`') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1771);
      END_STATE();
    case 1805:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1806:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_PERCENT);
      END_STATE();
    case 1807:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_PERCENT);
      ADVANCE_MAP(
        '\t', 1797,
        0x0b, 1797,
        '\f', 1797,
        ' ', 1797,
        '&', 1797,
        '(', 1797,
        ')', 1797,
        ',', 1797,
        ';', 1797,
        '|', 1797,
        '}', 1797,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1242);
      END_STATE();
    case 1808:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_PERCENT);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          lookahead == ' ' ||
          lookahead == '"' ||
          ('&' <= lookahead && lookahead <= ')') ||
          lookahead == ',' ||
          lookahead == ';' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1797);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1769);
      END_STATE();
    case 1809:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_PERCENT);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1797);
      END_STATE();
    case 1810:
      ACCEPT_TOKEN(aux_sym_class_attribute_token1);
      END_STATE();
    case 1811:
      ACCEPT_TOKEN(aux_sym_class_attribute_token2);
      END_STATE();
    case 1812:
      ACCEPT_TOKEN(aux_sym_class_statement_token1);
      END_STATE();
    case 1813:
      ACCEPT_TOKEN(aux_sym_enum_statement_token1);
      END_STATE();
    case 1814:
      ACCEPT_TOKEN(aux_sym_logical_expression_token1);
      END_STATE();
    case 1815:
      ACCEPT_TOKEN(aux_sym_logical_expression_token2);
      END_STATE();
    case 1816:
      ACCEPT_TOKEN(aux_sym_logical_expression_token3);
      END_STATE();
    case 1817:
      ACCEPT_TOKEN(aux_sym_bitwise_expression_token1);
      END_STATE();
    case 1818:
      ACCEPT_TOKEN(aux_sym_bitwise_expression_token2);
      END_STATE();
    case 1819:
      ACCEPT_TOKEN(aux_sym_bitwise_expression_token3);
      END_STATE();
    case 1820:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(1849);
      END_STATE();
    case 1821:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(1849);
      if (lookahead == '=') ADVANCE(832);
      END_STATE();
    case 1822:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(1850);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != '+' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1823:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '=') ADVANCE(832);
      END_STATE();
    case 1824:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1852,
        '=', 836,
        'A', 386,
        'a', 386,
        'B', 220,
        'b', 220,
        'C', 223,
        'c', 223,
        'E', 435,
        'e', 435,
        'F', 951,
        'f', 951,
        'G', 283,
        'g', 283,
        'I', 272,
        'i', 272,
        'J', 412,
        'j', 412,
        'L', 284,
        'l', 284,
        'M', 224,
        'm', 224,
        'N', 285,
        'n', 285,
        'O', 438,
        'o', 438,
        'P', 226,
        'p', 226,
        'R', 287,
        'r', 287,
        'S', 323,
        's', 323,
        'W', 338,
        'w', 338,
        'X', 414,
        'x', 414,
      );
      END_STATE();
    case 1825:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1851,
        '=', 836,
        'A', 386,
        'a', 386,
        'B', 220,
        'b', 220,
        'C', 256,
        'c', 256,
        'E', 434,
        'e', 434,
        'F', 950,
        'f', 950,
        'G', 283,
        'g', 283,
        'I', 272,
        'i', 272,
        'J', 412,
        'j', 412,
        'L', 284,
        'l', 284,
        'M', 224,
        'm', 224,
        'N', 285,
        'n', 285,
        'O', 438,
        'o', 438,
        'R', 316,
        'r', 316,
        'S', 322,
        's', 322,
        'X', 414,
        'x', 414,
      );
      END_STATE();
    case 1826:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1851,
        '=', 836,
        'A', 386,
        'a', 386,
        'B', 221,
        'b', 221,
        'C', 256,
        'c', 256,
        'E', 434,
        'e', 434,
        'F', 950,
        'f', 950,
        'G', 283,
        'g', 283,
        'I', 272,
        'i', 272,
        'J', 412,
        'j', 412,
        'L', 284,
        'l', 284,
        'M', 224,
        'm', 224,
        'N', 286,
        'n', 286,
        'O', 438,
        'o', 438,
        'R', 316,
        'r', 316,
        'S', 322,
        's', 322,
        'X', 414,
        'x', 414,
      );
      END_STATE();
    case 1827:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1851,
        'A', 386,
        'a', 386,
        'B', 220,
        'b', 220,
        'C', 256,
        'c', 256,
        'E', 434,
        'e', 434,
        'F', 950,
        'f', 950,
        'G', 283,
        'g', 283,
        'I', 272,
        'i', 272,
        'J', 412,
        'j', 412,
        'L', 284,
        'l', 284,
        'M', 224,
        'm', 224,
        'N', 285,
        'n', 285,
        'O', 438,
        'o', 438,
        'R', 316,
        'r', 316,
        'S', 322,
        's', 322,
        'X', 414,
        'x', 414,
      );
      END_STATE();
    case 1828:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1851,
        'B', 393,
        'b', 393,
        'J', 412,
        'j', 412,
        'N', 419,
        'n', 419,
        'S', 426,
        's', 426,
      );
      END_STATE();
    case 1829:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1853,
        'B', 1281,
        'b', 1281,
        'J', 1282,
        'j', 1282,
        'N', 1283,
        'n', 1283,
        'S', 1284,
        's', 1284,
      );
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1830:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1857,
        'B', 1228,
        'b', 1228,
        'J', 1230,
        'j', 1230,
        'N', 1231,
        'n', 1231,
        'S', 1234,
        's', 1234,
      );
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1831:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1860,
        'B', 1229,
        'b', 1229,
        'J', 1232,
        'j', 1232,
        'N', 1233,
        'n', 1233,
        'S', 1235,
        's', 1235,
      );
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != '-' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1832:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1855,
        'B', 1281,
        'b', 1281,
        'J', 1282,
        'j', 1282,
        'N', 1283,
        'n', 1283,
        'S', 1284,
        's', 1284,
      );
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1833:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '-', 1858,
        'B', 1228,
        'b', 1228,
        'J', 1230,
        'j', 1230,
        'N', 1231,
        'n', 1231,
        'S', 1234,
        's', 1234,
      );
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1834:
      ACCEPT_TOKEN(anon_sym_DASH);
      ADVANCE_MAP(
        '=', 836,
        'A', 386,
        'a', 386,
        'B', 221,
        'b', 221,
        'C', 256,
        'c', 256,
        'E', 434,
        'e', 434,
        'F', 950,
        'f', 950,
        'G', 283,
        'g', 283,
        'I', 272,
        'i', 272,
        'J', 412,
        'j', 412,
        'L', 284,
        'l', 284,
        'M', 224,
        'm', 224,
        'N', 286,
        'n', 286,
        'O', 438,
        'o', 438,
        'R', 316,
        'r', 316,
        'S', 322,
        's', 322,
        'X', 414,
        'x', 414,
      );
      END_STATE();
    case 1835:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 1836:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '=') ADVANCE(834);
      END_STATE();
    case 1837:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 1838:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 1839:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(835);
      END_STATE();
    case 1840:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 1841:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(833);
      if (lookahead == '>') ADVANCE(873);
      END_STATE();
    case 1842:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 1843:
      ACCEPT_TOKEN(aux_sym_expression_with_unary_operator_token1);
      END_STATE();
    case 1844:
      ACCEPT_TOKEN(aux_sym_expression_with_unary_operator_token1);
      ADVANCE_MAP(
        'C', 423,
        'c', 423,
        'I', 388,
        'i', 388,
        'L', 354,
        'l', 354,
        'M', 245,
        'm', 245,
      );
      END_STATE();
    case 1845:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 1846:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(831);
      END_STATE();
    case 1847:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1848:
      ACCEPT_TOKEN(aux_sym_expression_with_unary_operator_token2);
      END_STATE();
    case 1849:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 1850:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1851:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 1852:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1809);
      END_STATE();
    case 1853:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1809);
      if (lookahead == '-') ADVANCE(1274);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1854:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1809);
      if (lookahead == '-') ADVANCE(1275);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1855:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1797);
      if (lookahead == '-') ADVANCE(1274);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1287);
      END_STATE();
    case 1856:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1797);
      if (lookahead == '-') ADVANCE(1275);
      if (lookahead == '`') ADVANCE(1287);
      if (lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '\\') ADVANCE(1796);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1276);
      END_STATE();
    case 1857:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1807);
      if (lookahead == '-') ADVANCE(1209);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1858:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '%') ADVANCE(1242);
      if (lookahead == '-') ADVANCE(1209);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(1245);
      if (((!eof && set_contains(sym_generic_token_character_set_1, 11, lookahead)) ||
          lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\'' ||
          ('@' <= lookahead && lookahead <= '[') ||
          lookahead == '{') &&
          lookahead != ',' &&
          lookahead != ';') ADVANCE(1246);
      END_STATE();
    case 1859:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1860:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1861:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LPAREN);
      END_STATE();
    case 1862:
      ACCEPT_TOKEN(anon_sym_AT_LPAREN);
      END_STATE();
    case 1863:
      ACCEPT_TOKEN(anon_sym_AT_LBRACE);
      END_STATE();
    case 1864:
      ACCEPT_TOKEN(anon_sym_AT_LBRACE);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '&' &&
          lookahead != '(' &&
          lookahead != ')' &&
          lookahead != ',' &&
          lookahead != ';' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(1246);
      END_STATE();
    case 1865:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == '.') ADVANCE(1842);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(418);
      END_STATE();
    case 1866:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == '.') ADVANCE(1842);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(418);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 1867:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1793);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '\\' ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1796);
      END_STATE();
    case 1868:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(418);
      END_STATE();
    case 1869:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(418);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 1870:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 1871:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 1872:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 1873:
      ACCEPT_TOKEN(aux_sym_invokation_foreach_expression_token1);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 139},
  [2] = {.lex_state = 514, .external_lex_state = 1},
  [3] = {.lex_state = 514, .external_lex_state = 1},
  [4] = {.lex_state = 514},
  [5] = {.lex_state = 514, .external_lex_state = 1},
  [6] = {.lex_state = 514},
  [7] = {.lex_state = 140, .external_lex_state = 1},
  [8] = {.lex_state = 140},
  [9] = {.lex_state = 140},
  [10] = {.lex_state = 140},
  [11] = {.lex_state = 140},
  [12] = {.lex_state = 140},
  [13] = {.lex_state = 140},
  [14] = {.lex_state = 140},
  [15] = {.lex_state = 140},
  [16] = {.lex_state = 140},
  [17] = {.lex_state = 140},
  [18] = {.lex_state = 140},
  [19] = {.lex_state = 140},
  [20] = {.lex_state = 140},
  [21] = {.lex_state = 140, .external_lex_state = 1},
  [22] = {.lex_state = 140, .external_lex_state = 1},
  [23] = {.lex_state = 140, .external_lex_state = 1},
  [24] = {.lex_state = 140, .external_lex_state = 1},
  [25] = {.lex_state = 140, .external_lex_state = 1},
  [26] = {.lex_state = 140, .external_lex_state = 1},
  [27] = {.lex_state = 140, .external_lex_state = 1},
  [28] = {.lex_state = 140, .external_lex_state = 1},
  [29] = {.lex_state = 140, .external_lex_state = 1},
  [30] = {.lex_state = 140, .external_lex_state = 1},
  [31] = {.lex_state = 140, .external_lex_state = 1},
  [32] = {.lex_state = 140},
  [33] = {.lex_state = 140},
  [34] = {.lex_state = 140},
  [35] = {.lex_state = 140},
  [36] = {.lex_state = 140},
  [37] = {.lex_state = 140},
  [38] = {.lex_state = 140},
  [39] = {.lex_state = 140},
  [40] = {.lex_state = 140},
  [41] = {.lex_state = 140},
  [42] = {.lex_state = 140},
  [43] = {.lex_state = 140},
  [44] = {.lex_state = 141},
  [45] = {.lex_state = 141},
  [46] = {.lex_state = 516},
  [47] = {.lex_state = 516},
  [48] = {.lex_state = 516},
  [49] = {.lex_state = 516},
  [50] = {.lex_state = 516},
  [51] = {.lex_state = 516},
  [52] = {.lex_state = 516},
  [53] = {.lex_state = 516},
  [54] = {.lex_state = 516},
  [55] = {.lex_state = 516},
  [56] = {.lex_state = 516},
  [57] = {.lex_state = 516},
  [58] = {.lex_state = 516},
  [59] = {.lex_state = 516},
  [60] = {.lex_state = 516},
  [61] = {.lex_state = 516},
  [62] = {.lex_state = 516},
  [63] = {.lex_state = 516},
  [64] = {.lex_state = 516},
  [65] = {.lex_state = 516},
  [66] = {.lex_state = 516},
  [67] = {.lex_state = 516},
  [68] = {.lex_state = 516},
  [69] = {.lex_state = 516},
  [70] = {.lex_state = 516},
  [71] = {.lex_state = 516},
  [72] = {.lex_state = 516},
  [73] = {.lex_state = 516},
  [74] = {.lex_state = 516},
  [75] = {.lex_state = 516},
  [76] = {.lex_state = 516},
  [77] = {.lex_state = 516},
  [78] = {.lex_state = 516},
  [79] = {.lex_state = 516},
  [80] = {.lex_state = 516},
  [81] = {.lex_state = 514},
  [82] = {.lex_state = 514, .external_lex_state = 1},
  [83] = {.lex_state = 131},
  [84] = {.lex_state = 131},
  [85] = {.lex_state = 131, .external_lex_state = 1},
  [86] = {.lex_state = 131, .external_lex_state = 1},
  [87] = {.lex_state = 142},
  [88] = {.lex_state = 142},
  [89] = {.lex_state = 142},
  [90] = {.lex_state = 142},
  [91] = {.lex_state = 142},
  [92] = {.lex_state = 142},
  [93] = {.lex_state = 142},
  [94] = {.lex_state = 142},
  [95] = {.lex_state = 142},
  [96] = {.lex_state = 142},
  [97] = {.lex_state = 142},
  [98] = {.lex_state = 142},
  [99] = {.lex_state = 142},
  [100] = {.lex_state = 142},
  [101] = {.lex_state = 142},
  [102] = {.lex_state = 142},
  [103] = {.lex_state = 142},
  [104] = {.lex_state = 142},
  [105] = {.lex_state = 142},
  [106] = {.lex_state = 142},
  [107] = {.lex_state = 142},
  [108] = {.lex_state = 142},
  [109] = {.lex_state = 142},
  [110] = {.lex_state = 142},
  [111] = {.lex_state = 142, .external_lex_state = 1},
  [112] = {.lex_state = 142},
  [113] = {.lex_state = 142},
  [114] = {.lex_state = 142},
  [115] = {.lex_state = 142, .external_lex_state = 1},
  [116] = {.lex_state = 142},
  [117] = {.lex_state = 142},
  [118] = {.lex_state = 142},
  [119] = {.lex_state = 142},
  [120] = {.lex_state = 142},
  [121] = {.lex_state = 142},
  [122] = {.lex_state = 142},
  [123] = {.lex_state = 142},
  [124] = {.lex_state = 142, .external_lex_state = 1},
  [125] = {.lex_state = 142, .external_lex_state = 1},
  [126] = {.lex_state = 142, .external_lex_state = 1},
  [127] = {.lex_state = 142, .external_lex_state = 1},
  [128] = {.lex_state = 142, .external_lex_state = 1},
  [129] = {.lex_state = 142, .external_lex_state = 1},
  [130] = {.lex_state = 142, .external_lex_state = 1},
  [131] = {.lex_state = 142, .external_lex_state = 1},
  [132] = {.lex_state = 142, .external_lex_state = 1},
  [133] = {.lex_state = 142, .external_lex_state = 1},
  [134] = {.lex_state = 142, .external_lex_state = 1},
  [135] = {.lex_state = 142, .external_lex_state = 1},
  [136] = {.lex_state = 142, .external_lex_state = 1},
  [137] = {.lex_state = 142, .external_lex_state = 1},
  [138] = {.lex_state = 142, .external_lex_state = 1},
  [139] = {.lex_state = 142, .external_lex_state = 1},
  [140] = {.lex_state = 142, .external_lex_state = 1},
  [141] = {.lex_state = 142, .external_lex_state = 1},
  [142] = {.lex_state = 142, .external_lex_state = 1},
  [143] = {.lex_state = 142, .external_lex_state = 1},
  [144] = {.lex_state = 142, .external_lex_state = 1},
  [145] = {.lex_state = 142, .external_lex_state = 1},
  [146] = {.lex_state = 142, .external_lex_state = 1},
  [147] = {.lex_state = 142, .external_lex_state = 1},
  [148] = {.lex_state = 142, .external_lex_state = 1},
  [149] = {.lex_state = 142, .external_lex_state = 1},
  [150] = {.lex_state = 142, .external_lex_state = 1},
  [151] = {.lex_state = 142},
  [152] = {.lex_state = 142, .external_lex_state = 1},
  [153] = {.lex_state = 142, .external_lex_state = 1},
  [154] = {.lex_state = 142, .external_lex_state = 1},
  [155] = {.lex_state = 142, .external_lex_state = 1},
  [156] = {.lex_state = 142, .external_lex_state = 1},
  [157] = {.lex_state = 142, .external_lex_state = 1},
  [158] = {.lex_state = 142, .external_lex_state = 1},
  [159] = {.lex_state = 142},
  [160] = {.lex_state = 142, .external_lex_state = 1},
  [161] = {.lex_state = 144},
  [162] = {.lex_state = 144},
  [163] = {.lex_state = 144},
  [164] = {.lex_state = 144},
  [165] = {.lex_state = 144},
  [166] = {.lex_state = 144},
  [167] = {.lex_state = 144},
  [168] = {.lex_state = 144, .external_lex_state = 1},
  [169] = {.lex_state = 144},
  [170] = {.lex_state = 144},
  [171] = {.lex_state = 144, .external_lex_state = 1},
  [172] = {.lex_state = 144},
  [173] = {.lex_state = 144, .external_lex_state = 1},
  [174] = {.lex_state = 144},
  [175] = {.lex_state = 144},
  [176] = {.lex_state = 144, .external_lex_state = 1},
  [177] = {.lex_state = 144, .external_lex_state = 1},
  [178] = {.lex_state = 144, .external_lex_state = 1},
  [179] = {.lex_state = 144, .external_lex_state = 1},
  [180] = {.lex_state = 144, .external_lex_state = 1},
  [181] = {.lex_state = 144, .external_lex_state = 1},
  [182] = {.lex_state = 144, .external_lex_state = 1},
  [183] = {.lex_state = 144, .external_lex_state = 1},
  [184] = {.lex_state = 144, .external_lex_state = 1},
  [185] = {.lex_state = 144, .external_lex_state = 1},
  [186] = {.lex_state = 144},
  [187] = {.lex_state = 144},
  [188] = {.lex_state = 144},
  [189] = {.lex_state = 144, .external_lex_state = 1},
  [190] = {.lex_state = 144},
  [191] = {.lex_state = 144, .external_lex_state = 1},
  [192] = {.lex_state = 144, .external_lex_state = 1},
  [193] = {.lex_state = 144},
  [194] = {.lex_state = 144},
  [195] = {.lex_state = 144, .external_lex_state = 1},
  [196] = {.lex_state = 144, .external_lex_state = 1},
  [197] = {.lex_state = 144},
  [198] = {.lex_state = 144, .external_lex_state = 1},
  [199] = {.lex_state = 144, .external_lex_state = 1},
  [200] = {.lex_state = 144},
  [201] = {.lex_state = 19},
  [202] = {.lex_state = 514, .external_lex_state = 1},
  [203] = {.lex_state = 19},
  [204] = {.lex_state = 19},
  [205] = {.lex_state = 19},
  [206] = {.lex_state = 19},
  [207] = {.lex_state = 19},
  [208] = {.lex_state = 19},
  [209] = {.lex_state = 19},
  [210] = {.lex_state = 19},
  [211] = {.lex_state = 514, .external_lex_state = 1},
  [212] = {.lex_state = 19},
  [213] = {.lex_state = 19},
  [214] = {.lex_state = 514},
  [215] = {.lex_state = 131},
  [216] = {.lex_state = 514},
  [217] = {.lex_state = 19},
  [218] = {.lex_state = 106, .external_lex_state = 1},
  [219] = {.lex_state = 131, .external_lex_state = 1},
  [220] = {.lex_state = 106},
  [221] = {.lex_state = 106, .external_lex_state = 1},
  [222] = {.lex_state = 106},
  [223] = {.lex_state = 138},
  [224] = {.lex_state = 138},
  [225] = {.lex_state = 138},
  [226] = {.lex_state = 126, .external_lex_state = 1},
  [227] = {.lex_state = 138},
  [228] = {.lex_state = 138},
  [229] = {.lex_state = 138},
  [230] = {.lex_state = 138},
  [231] = {.lex_state = 138},
  [232] = {.lex_state = 138},
  [233] = {.lex_state = 138},
  [234] = {.lex_state = 138},
  [235] = {.lex_state = 108},
  [236] = {.lex_state = 138},
  [237] = {.lex_state = 108, .external_lex_state = 1},
  [238] = {.lex_state = 126},
  [239] = {.lex_state = 138},
  [240] = {.lex_state = 108, .external_lex_state = 1},
  [241] = {.lex_state = 108},
  [242] = {.lex_state = 138},
  [243] = {.lex_state = 138},
  [244] = {.lex_state = 138},
  [245] = {.lex_state = 138},
  [246] = {.lex_state = 128},
  [247] = {.lex_state = 128, .external_lex_state = 1},
  [248] = {.lex_state = 138, .external_lex_state = 1},
  [249] = {.lex_state = 138},
  [250] = {.lex_state = 138},
  [251] = {.lex_state = 138},
  [252] = {.lex_state = 138},
  [253] = {.lex_state = 138},
  [254] = {.lex_state = 138},
  [255] = {.lex_state = 138},
  [256] = {.lex_state = 138},
  [257] = {.lex_state = 138},
  [258] = {.lex_state = 138},
  [259] = {.lex_state = 138},
  [260] = {.lex_state = 138},
  [261] = {.lex_state = 138},
  [262] = {.lex_state = 138},
  [263] = {.lex_state = 138},
  [264] = {.lex_state = 138},
  [265] = {.lex_state = 138},
  [266] = {.lex_state = 138},
  [267] = {.lex_state = 138},
  [268] = {.lex_state = 138},
  [269] = {.lex_state = 138},
  [270] = {.lex_state = 138},
  [271] = {.lex_state = 138},
  [272] = {.lex_state = 138},
  [273] = {.lex_state = 138},
  [274] = {.lex_state = 138},
  [275] = {.lex_state = 138},
  [276] = {.lex_state = 138},
  [277] = {.lex_state = 138},
  [278] = {.lex_state = 138},
  [279] = {.lex_state = 138},
  [280] = {.lex_state = 138},
  [281] = {.lex_state = 138},
  [282] = {.lex_state = 138},
  [283] = {.lex_state = 138},
  [284] = {.lex_state = 138},
  [285] = {.lex_state = 134},
  [286] = {.lex_state = 134, .external_lex_state = 1},
  [287] = {.lex_state = 134},
  [288] = {.lex_state = 134, .external_lex_state = 1},
  [289] = {.lex_state = 145},
  [290] = {.lex_state = 145, .external_lex_state = 1},
  [291] = {.lex_state = 138},
  [292] = {.lex_state = 138},
  [293] = {.lex_state = 517},
  [294] = {.lex_state = 518},
  [295] = {.lex_state = 518},
  [296] = {.lex_state = 517},
  [297] = {.lex_state = 518},
  [298] = {.lex_state = 517},
  [299] = {.lex_state = 140, .external_lex_state = 1},
  [300] = {.lex_state = 140, .external_lex_state = 1},
  [301] = {.lex_state = 517},
  [302] = {.lex_state = 141},
  [303] = {.lex_state = 518},
  [304] = {.lex_state = 140, .external_lex_state = 1},
  [305] = {.lex_state = 141},
  [306] = {.lex_state = 518},
  [307] = {.lex_state = 517},
  [308] = {.lex_state = 517},
  [309] = {.lex_state = 518},
  [310] = {.lex_state = 518},
  [311] = {.lex_state = 520},
  [312] = {.lex_state = 517},
  [313] = {.lex_state = 518},
  [314] = {.lex_state = 519},
  [315] = {.lex_state = 518},
  [316] = {.lex_state = 517},
  [317] = {.lex_state = 518},
  [318] = {.lex_state = 517},
  [319] = {.lex_state = 518},
  [320] = {.lex_state = 520},
  [321] = {.lex_state = 518},
  [322] = {.lex_state = 517},
  [323] = {.lex_state = 517},
  [324] = {.lex_state = 519},
  [325] = {.lex_state = 517},
  [326] = {.lex_state = 518},
  [327] = {.lex_state = 518},
  [328] = {.lex_state = 516},
  [329] = {.lex_state = 516},
  [330] = {.lex_state = 516},
  [331] = {.lex_state = 516},
  [332] = {.lex_state = 516},
  [333] = {.lex_state = 516},
  [334] = {.lex_state = 516},
  [335] = {.lex_state = 516},
  [336] = {.lex_state = 516},
  [337] = {.lex_state = 516},
  [338] = {.lex_state = 516},
  [339] = {.lex_state = 516},
  [340] = {.lex_state = 516},
  [341] = {.lex_state = 516},
  [342] = {.lex_state = 516},
  [343] = {.lex_state = 516},
  [344] = {.lex_state = 516},
  [345] = {.lex_state = 516},
  [346] = {.lex_state = 516},
  [347] = {.lex_state = 516},
  [348] = {.lex_state = 516},
  [349] = {.lex_state = 516},
  [350] = {.lex_state = 516},
  [351] = {.lex_state = 516},
  [352] = {.lex_state = 516},
  [353] = {.lex_state = 516},
  [354] = {.lex_state = 516},
  [355] = {.lex_state = 516},
  [356] = {.lex_state = 516},
  [357] = {.lex_state = 516},
  [358] = {.lex_state = 516},
  [359] = {.lex_state = 516},
  [360] = {.lex_state = 516},
  [361] = {.lex_state = 516},
  [362] = {.lex_state = 516},
  [363] = {.lex_state = 516},
  [364] = {.lex_state = 516},
  [365] = {.lex_state = 516},
  [366] = {.lex_state = 516},
  [367] = {.lex_state = 516},
  [368] = {.lex_state = 516},
  [369] = {.lex_state = 516},
  [370] = {.lex_state = 516},
  [371] = {.lex_state = 516},
  [372] = {.lex_state = 516},
  [373] = {.lex_state = 516},
  [374] = {.lex_state = 516},
  [375] = {.lex_state = 516},
  [376] = {.lex_state = 516},
  [377] = {.lex_state = 516},
  [378] = {.lex_state = 516},
  [379] = {.lex_state = 516},
  [380] = {.lex_state = 516},
  [381] = {.lex_state = 516},
  [382] = {.lex_state = 516},
  [383] = {.lex_state = 516},
  [384] = {.lex_state = 516},
  [385] = {.lex_state = 516},
  [386] = {.lex_state = 516},
  [387] = {.lex_state = 516},
  [388] = {.lex_state = 516},
  [389] = {.lex_state = 516},
  [390] = {.lex_state = 516},
  [391] = {.lex_state = 516},
  [392] = {.lex_state = 516},
  [393] = {.lex_state = 516},
  [394] = {.lex_state = 516},
  [395] = {.lex_state = 516},
  [396] = {.lex_state = 516},
  [397] = {.lex_state = 516},
  [398] = {.lex_state = 516},
  [399] = {.lex_state = 516},
  [400] = {.lex_state = 516},
  [401] = {.lex_state = 516},
  [402] = {.lex_state = 516},
  [403] = {.lex_state = 516},
  [404] = {.lex_state = 516},
  [405] = {.lex_state = 516},
  [406] = {.lex_state = 516},
  [407] = {.lex_state = 516},
  [408] = {.lex_state = 516},
  [409] = {.lex_state = 516},
  [410] = {.lex_state = 516},
  [411] = {.lex_state = 516},
  [412] = {.lex_state = 516},
  [413] = {.lex_state = 516},
  [414] = {.lex_state = 516},
  [415] = {.lex_state = 516},
  [416] = {.lex_state = 516},
  [417] = {.lex_state = 516},
  [418] = {.lex_state = 516},
  [419] = {.lex_state = 516},
  [420] = {.lex_state = 516},
  [421] = {.lex_state = 516},
  [422] = {.lex_state = 516},
  [423] = {.lex_state = 516},
  [424] = {.lex_state = 516},
  [425] = {.lex_state = 516},
  [426] = {.lex_state = 516},
  [427] = {.lex_state = 516},
  [428] = {.lex_state = 516},
  [429] = {.lex_state = 516},
  [430] = {.lex_state = 516},
  [431] = {.lex_state = 142},
  [432] = {.lex_state = 516},
  [433] = {.lex_state = 142},
  [434] = {.lex_state = 142, .external_lex_state = 1},
  [435] = {.lex_state = 142, .external_lex_state = 1},
  [436] = {.lex_state = 106, .external_lex_state = 1},
  [437] = {.lex_state = 106},
  [438] = {.lex_state = 126, .external_lex_state = 1},
  [439] = {.lex_state = 126},
  [440] = {.lex_state = 144},
  [441] = {.lex_state = 126, .external_lex_state = 1},
  [442] = {.lex_state = 131},
  [443] = {.lex_state = 108},
  [444] = {.lex_state = 108, .external_lex_state = 1},
  [445] = {.lex_state = 126},
  [446] = {.lex_state = 131},
  [447] = {.lex_state = 144, .external_lex_state = 1},
  [448] = {.lex_state = 126, .external_lex_state = 1},
  [449] = {.lex_state = 126},
  [450] = {.lex_state = 144, .external_lex_state = 1},
  [451] = {.lex_state = 144},
  [452] = {.lex_state = 144},
  [453] = {.lex_state = 128},
  [454] = {.lex_state = 144},
  [455] = {.lex_state = 144, .external_lex_state = 1},
  [456] = {.lex_state = 144},
  [457] = {.lex_state = 128, .external_lex_state = 1},
  [458] = {.lex_state = 128, .external_lex_state = 1},
  [459] = {.lex_state = 128},
  [460] = {.lex_state = 144},
  [461] = {.lex_state = 128},
  [462] = {.lex_state = 144},
  [463] = {.lex_state = 128, .external_lex_state = 1},
  [464] = {.lex_state = 131},
  [465] = {.lex_state = 144},
  [466] = {.lex_state = 144},
  [467] = {.lex_state = 144},
  [468] = {.lex_state = 131},
  [469] = {.lex_state = 144, .external_lex_state = 1},
  [470] = {.lex_state = 131},
  [471] = {.lex_state = 131},
  [472] = {.lex_state = 131},
  [473] = {.lex_state = 144},
  [474] = {.lex_state = 131},
  [475] = {.lex_state = 131},
  [476] = {.lex_state = 150},
  [477] = {.lex_state = 144, .external_lex_state = 1},
  [478] = {.lex_state = 144},
  [479] = {.lex_state = 131},
  [480] = {.lex_state = 144, .external_lex_state = 1},
  [481] = {.lex_state = 131},
  [482] = {.lex_state = 144, .external_lex_state = 1},
  [483] = {.lex_state = 144},
  [484] = {.lex_state = 131},
  [485] = {.lex_state = 131},
  [486] = {.lex_state = 131},
  [487] = {.lex_state = 131},
  [488] = {.lex_state = 134, .external_lex_state = 1},
  [489] = {.lex_state = 134},
  [490] = {.lex_state = 131},
  [491] = {.lex_state = 127, .external_lex_state = 1},
  [492] = {.lex_state = 131},
  [493] = {.lex_state = 131},
  [494] = {.lex_state = 127, .external_lex_state = 1},
  [495] = {.lex_state = 131},
  [496] = {.lex_state = 144},
  [497] = {.lex_state = 131},
  [498] = {.lex_state = 131},
  [499] = {.lex_state = 144},
  [500] = {.lex_state = 131},
  [501] = {.lex_state = 131},
  [502] = {.lex_state = 131},
  [503] = {.lex_state = 131},
  [504] = {.lex_state = 127},
  [505] = {.lex_state = 127},
  [506] = {.lex_state = 150},
  [507] = {.lex_state = 144},
  [508] = {.lex_state = 144},
  [509] = {.lex_state = 150},
  [510] = {.lex_state = 144},
  [511] = {.lex_state = 144},
  [512] = {.lex_state = 144},
  [513] = {.lex_state = 150},
  [514] = {.lex_state = 131},
  [515] = {.lex_state = 150},
  [516] = {.lex_state = 150},
  [517] = {.lex_state = 144},
  [518] = {.lex_state = 144},
  [519] = {.lex_state = 144},
  [520] = {.lex_state = 144, .external_lex_state = 1},
  [521] = {.lex_state = 144, .external_lex_state = 1},
  [522] = {.lex_state = 131},
  [523] = {.lex_state = 131},
  [524] = {.lex_state = 131},
  [525] = {.lex_state = 145, .external_lex_state = 1},
  [526] = {.lex_state = 131},
  [527] = {.lex_state = 145},
  [528] = {.lex_state = 131},
  [529] = {.lex_state = 131},
  [530] = {.lex_state = 131},
  [531] = {.lex_state = 131},
  [532] = {.lex_state = 131},
  [533] = {.lex_state = 131},
  [534] = {.lex_state = 131},
  [535] = {.lex_state = 131},
  [536] = {.lex_state = 131},
  [537] = {.lex_state = 131},
  [538] = {.lex_state = 144},
  [539] = {.lex_state = 131},
  [540] = {.lex_state = 131},
  [541] = {.lex_state = 144},
  [542] = {.lex_state = 131},
  [543] = {.lex_state = 131},
  [544] = {.lex_state = 131},
  [545] = {.lex_state = 144},
  [546] = {.lex_state = 144},
  [547] = {.lex_state = 131},
  [548] = {.lex_state = 131},
  [549] = {.lex_state = 131},
  [550] = {.lex_state = 131},
  [551] = {.lex_state = 131},
  [552] = {.lex_state = 131},
  [553] = {.lex_state = 131},
  [554] = {.lex_state = 131},
  [555] = {.lex_state = 144},
  [556] = {.lex_state = 136},
  [557] = {.lex_state = 131},
  [558] = {.lex_state = 144},
  [559] = {.lex_state = 136},
  [560] = {.lex_state = 514},
  [561] = {.lex_state = 131},
  [562] = {.lex_state = 131},
  [563] = {.lex_state = 514},
  [564] = {.lex_state = 144, .external_lex_state = 1},
  [565] = {.lex_state = 136, .external_lex_state = 1},
  [566] = {.lex_state = 144, .external_lex_state = 1},
  [567] = {.lex_state = 131},
  [568] = {.lex_state = 136, .external_lex_state = 1},
  [569] = {.lex_state = 131},
  [570] = {.lex_state = 131},
  [571] = {.lex_state = 514, .external_lex_state = 1},
  [572] = {.lex_state = 131, .external_lex_state = 1},
  [573] = {.lex_state = 514, .external_lex_state = 1},
  [574] = {.lex_state = 131},
  [575] = {.lex_state = 131},
  [576] = {.lex_state = 131},
  [577] = {.lex_state = 514},
  [578] = {.lex_state = 131},
  [579] = {.lex_state = 514},
  [580] = {.lex_state = 131},
  [581] = {.lex_state = 144},
  [582] = {.lex_state = 131},
  [583] = {.lex_state = 131},
  [584] = {.lex_state = 131},
  [585] = {.lex_state = 144},
  [586] = {.lex_state = 131},
  [587] = {.lex_state = 131},
  [588] = {.lex_state = 131},
  [589] = {.lex_state = 131},
  [590] = {.lex_state = 131},
  [591] = {.lex_state = 131},
  [592] = {.lex_state = 131},
  [593] = {.lex_state = 131},
  [594] = {.lex_state = 514},
  [595] = {.lex_state = 131},
  [596] = {.lex_state = 514},
  [597] = {.lex_state = 131},
  [598] = {.lex_state = 131},
  [599] = {.lex_state = 131},
  [600] = {.lex_state = 131},
  [601] = {.lex_state = 131},
  [602] = {.lex_state = 131},
  [603] = {.lex_state = 131},
  [604] = {.lex_state = 131},
  [605] = {.lex_state = 131},
  [606] = {.lex_state = 131},
  [607] = {.lex_state = 131},
  [608] = {.lex_state = 131},
  [609] = {.lex_state = 131},
  [610] = {.lex_state = 131},
  [611] = {.lex_state = 131},
  [612] = {.lex_state = 131},
  [613] = {.lex_state = 131},
  [614] = {.lex_state = 131},
  [615] = {.lex_state = 131},
  [616] = {.lex_state = 131},
  [617] = {.lex_state = 131},
  [618] = {.lex_state = 131},
  [619] = {.lex_state = 131},
  [620] = {.lex_state = 131},
  [621] = {.lex_state = 131},
  [622] = {.lex_state = 131},
  [623] = {.lex_state = 131},
  [624] = {.lex_state = 131},
  [625] = {.lex_state = 131},
  [626] = {.lex_state = 131},
  [627] = {.lex_state = 131},
  [628] = {.lex_state = 131},
  [629] = {.lex_state = 131},
  [630] = {.lex_state = 131},
  [631] = {.lex_state = 131},
  [632] = {.lex_state = 131},
  [633] = {.lex_state = 131},
  [634] = {.lex_state = 131},
  [635] = {.lex_state = 131},
  [636] = {.lex_state = 131},
  [637] = {.lex_state = 131},
  [638] = {.lex_state = 131},
  [639] = {.lex_state = 131},
  [640] = {.lex_state = 131},
  [641] = {.lex_state = 131},
  [642] = {.lex_state = 131},
  [643] = {.lex_state = 131},
  [644] = {.lex_state = 131},
  [645] = {.lex_state = 131},
  [646] = {.lex_state = 131},
  [647] = {.lex_state = 131},
  [648] = {.lex_state = 131},
  [649] = {.lex_state = 131},
  [650] = {.lex_state = 131},
  [651] = {.lex_state = 131},
  [652] = {.lex_state = 131},
  [653] = {.lex_state = 131},
  [654] = {.lex_state = 131},
  [655] = {.lex_state = 131},
  [656] = {.lex_state = 131},
  [657] = {.lex_state = 131},
  [658] = {.lex_state = 131},
  [659] = {.lex_state = 131},
  [660] = {.lex_state = 131},
  [661] = {.lex_state = 131},
  [662] = {.lex_state = 131},
  [663] = {.lex_state = 131},
  [664] = {.lex_state = 131},
  [665] = {.lex_state = 131},
  [666] = {.lex_state = 131},
  [667] = {.lex_state = 131},
  [668] = {.lex_state = 131},
  [669] = {.lex_state = 131},
  [670] = {.lex_state = 131},
  [671] = {.lex_state = 131},
  [672] = {.lex_state = 131},
  [673] = {.lex_state = 131},
  [674] = {.lex_state = 131},
  [675] = {.lex_state = 131},
  [676] = {.lex_state = 131},
  [677] = {.lex_state = 131},
  [678] = {.lex_state = 131},
  [679] = {.lex_state = 131},
  [680] = {.lex_state = 112},
  [681] = {.lex_state = 112},
  [682] = {.lex_state = 112, .external_lex_state = 1},
  [683] = {.lex_state = 112, .external_lex_state = 1},
  [684] = {.lex_state = 114},
  [685] = {.lex_state = 114, .external_lex_state = 1},
  [686] = {.lex_state = 114, .external_lex_state = 1},
  [687] = {.lex_state = 114},
  [688] = {.lex_state = 138},
  [689] = {.lex_state = 131},
  [690] = {.lex_state = 131},
  [691] = {.lex_state = 131},
  [692] = {.lex_state = 131},
  [693] = {.lex_state = 131},
  [694] = {.lex_state = 131},
  [695] = {.lex_state = 131},
  [696] = {.lex_state = 131},
  [697] = {.lex_state = 131},
  [698] = {.lex_state = 131},
  [699] = {.lex_state = 131},
  [700] = {.lex_state = 131},
  [701] = {.lex_state = 131},
  [702] = {.lex_state = 122, .external_lex_state = 1},
  [703] = {.lex_state = 131},
  [704] = {.lex_state = 131},
  [705] = {.lex_state = 131},
  [706] = {.lex_state = 131},
  [707] = {.lex_state = 131},
  [708] = {.lex_state = 131},
  [709] = {.lex_state = 122, .external_lex_state = 1},
  [710] = {.lex_state = 131},
  [711] = {.lex_state = 122},
  [712] = {.lex_state = 122},
  [713] = {.lex_state = 116, .external_lex_state = 1},
  [714] = {.lex_state = 123, .external_lex_state = 1},
  [715] = {.lex_state = 116, .external_lex_state = 1},
  [716] = {.lex_state = 116},
  [717] = {.lex_state = 123},
  [718] = {.lex_state = 123, .external_lex_state = 1},
  [719] = {.lex_state = 122, .external_lex_state = 1},
  [720] = {.lex_state = 122},
  [721] = {.lex_state = 122, .external_lex_state = 1},
  [722] = {.lex_state = 123},
  [723] = {.lex_state = 116},
  [724] = {.lex_state = 122},
  [725] = {.lex_state = 117, .external_lex_state = 1},
  [726] = {.lex_state = 117},
  [727] = {.lex_state = 123, .external_lex_state = 1},
  [728] = {.lex_state = 123},
  [729] = {.lex_state = 123, .external_lex_state = 1},
  [730] = {.lex_state = 123},
  [731] = {.lex_state = 117, .external_lex_state = 1},
  [732] = {.lex_state = 117},
  [733] = {.lex_state = 160},
  [734] = {.lex_state = 109},
  [735] = {.lex_state = 109},
  [736] = {.lex_state = 118, .external_lex_state = 1},
  [737] = {.lex_state = 116},
  [738] = {.lex_state = 151},
  [739] = {.lex_state = 118},
  [740] = {.lex_state = 109},
  [741] = {.lex_state = 109},
  [742] = {.lex_state = 109},
  [743] = {.lex_state = 151},
  [744] = {.lex_state = 109},
  [745] = {.lex_state = 151},
  [746] = {.lex_state = 151},
  [747] = {.lex_state = 116, .external_lex_state = 1},
  [748] = {.lex_state = 118},
  [749] = {.lex_state = 118, .external_lex_state = 1},
  [750] = {.lex_state = 118, .external_lex_state = 1},
  [751] = {.lex_state = 117, .external_lex_state = 1},
  [752] = {.lex_state = 118},
  [753] = {.lex_state = 118, .external_lex_state = 1},
  [754] = {.lex_state = 118, .external_lex_state = 1},
  [755] = {.lex_state = 118, .external_lex_state = 1},
  [756] = {.lex_state = 118, .external_lex_state = 1},
  [757] = {.lex_state = 117},
  [758] = {.lex_state = 118, .external_lex_state = 1},
  [759] = {.lex_state = 118, .external_lex_state = 1},
  [760] = {.lex_state = 118, .external_lex_state = 1},
  [761] = {.lex_state = 118, .external_lex_state = 1},
  [762] = {.lex_state = 118, .external_lex_state = 1},
  [763] = {.lex_state = 118},
  [764] = {.lex_state = 118},
  [765] = {.lex_state = 118, .external_lex_state = 1},
  [766] = {.lex_state = 118, .external_lex_state = 1},
  [767] = {.lex_state = 118},
  [768] = {.lex_state = 118},
  [769] = {.lex_state = 118},
  [770] = {.lex_state = 118},
  [771] = {.lex_state = 151},
  [772] = {.lex_state = 118},
  [773] = {.lex_state = 118},
  [774] = {.lex_state = 118, .external_lex_state = 1},
  [775] = {.lex_state = 118},
  [776] = {.lex_state = 118},
  [777] = {.lex_state = 118},
  [778] = {.lex_state = 118},
  [779] = {.lex_state = 118},
  [780] = {.lex_state = 118, .external_lex_state = 1},
  [781] = {.lex_state = 118},
  [782] = {.lex_state = 118},
  [783] = {.lex_state = 118, .external_lex_state = 1},
  [784] = {.lex_state = 118, .external_lex_state = 1},
  [785] = {.lex_state = 118},
  [786] = {.lex_state = 118},
  [787] = {.lex_state = 118, .external_lex_state = 1},
  [788] = {.lex_state = 118, .external_lex_state = 1},
  [789] = {.lex_state = 120, .external_lex_state = 1},
  [790] = {.lex_state = 118},
  [791] = {.lex_state = 118, .external_lex_state = 1},
  [792] = {.lex_state = 118, .external_lex_state = 1},
  [793] = {.lex_state = 118, .external_lex_state = 1},
  [794] = {.lex_state = 118},
  [795] = {.lex_state = 118},
  [796] = {.lex_state = 118},
  [797] = {.lex_state = 118},
  [798] = {.lex_state = 118},
  [799] = {.lex_state = 118, .external_lex_state = 1},
  [800] = {.lex_state = 118},
  [801] = {.lex_state = 118, .external_lex_state = 1},
  [802] = {.lex_state = 118, .external_lex_state = 1},
  [803] = {.lex_state = 118, .external_lex_state = 1},
  [804] = {.lex_state = 118, .external_lex_state = 1},
  [805] = {.lex_state = 118, .external_lex_state = 1},
  [806] = {.lex_state = 118, .external_lex_state = 1},
  [807] = {.lex_state = 118, .external_lex_state = 1},
  [808] = {.lex_state = 118, .external_lex_state = 1},
  [809] = {.lex_state = 120},
  [810] = {.lex_state = 118, .external_lex_state = 1},
  [811] = {.lex_state = 118},
  [812] = {.lex_state = 118},
  [813] = {.lex_state = 118},
  [814] = {.lex_state = 118},
  [815] = {.lex_state = 118},
  [816] = {.lex_state = 118, .external_lex_state = 1},
  [817] = {.lex_state = 118},
  [818] = {.lex_state = 118},
  [819] = {.lex_state = 151},
  [820] = {.lex_state = 118, .external_lex_state = 1},
  [821] = {.lex_state = 118},
  [822] = {.lex_state = 120},
  [823] = {.lex_state = 120, .external_lex_state = 1},
  [824] = {.lex_state = 120, .external_lex_state = 1},
  [825] = {.lex_state = 120, .external_lex_state = 1},
  [826] = {.lex_state = 118, .external_lex_state = 1},
  [827] = {.lex_state = 120, .external_lex_state = 1},
  [828] = {.lex_state = 120, .external_lex_state = 1},
  [829] = {.lex_state = 120},
  [830] = {.lex_state = 120},
  [831] = {.lex_state = 120},
  [832] = {.lex_state = 120, .external_lex_state = 1},
  [833] = {.lex_state = 118},
  [834] = {.lex_state = 118, .external_lex_state = 1},
  [835] = {.lex_state = 120},
  [836] = {.lex_state = 120},
  [837] = {.lex_state = 120},
  [838] = {.lex_state = 120},
  [839] = {.lex_state = 120},
  [840] = {.lex_state = 120},
  [841] = {.lex_state = 120, .external_lex_state = 1},
  [842] = {.lex_state = 120, .external_lex_state = 1},
  [843] = {.lex_state = 120},
  [844] = {.lex_state = 120},
  [845] = {.lex_state = 514, .external_lex_state = 1},
  [846] = {.lex_state = 120},
  [847] = {.lex_state = 120, .external_lex_state = 1},
  [848] = {.lex_state = 120},
  [849] = {.lex_state = 120},
  [850] = {.lex_state = 120},
  [851] = {.lex_state = 120},
  [852] = {.lex_state = 120, .external_lex_state = 1},
  [853] = {.lex_state = 120},
  [854] = {.lex_state = 514, .external_lex_state = 1},
  [855] = {.lex_state = 514},
  [856] = {.lex_state = 514},
  [857] = {.lex_state = 118},
  [858] = {.lex_state = 120},
  [859] = {.lex_state = 120, .external_lex_state = 1},
  [860] = {.lex_state = 120},
  [861] = {.lex_state = 120},
  [862] = {.lex_state = 120},
  [863] = {.lex_state = 120},
  [864] = {.lex_state = 120},
  [865] = {.lex_state = 120, .external_lex_state = 1},
  [866] = {.lex_state = 120, .external_lex_state = 1},
  [867] = {.lex_state = 120},
  [868] = {.lex_state = 120, .external_lex_state = 1},
  [869] = {.lex_state = 120},
  [870] = {.lex_state = 120, .external_lex_state = 1},
  [871] = {.lex_state = 120},
  [872] = {.lex_state = 120, .external_lex_state = 1},
  [873] = {.lex_state = 120, .external_lex_state = 1},
  [874] = {.lex_state = 120, .external_lex_state = 1},
  [875] = {.lex_state = 118},
  [876] = {.lex_state = 120, .external_lex_state = 1},
  [877] = {.lex_state = 120, .external_lex_state = 1},
  [878] = {.lex_state = 120, .external_lex_state = 1},
  [879] = {.lex_state = 120, .external_lex_state = 1},
  [880] = {.lex_state = 120, .external_lex_state = 1},
  [881] = {.lex_state = 120, .external_lex_state = 1},
  [882] = {.lex_state = 120, .external_lex_state = 1},
  [883] = {.lex_state = 118},
  [884] = {.lex_state = 120, .external_lex_state = 1},
  [885] = {.lex_state = 120, .external_lex_state = 1},
  [886] = {.lex_state = 514},
  [887] = {.lex_state = 514, .external_lex_state = 1},
  [888] = {.lex_state = 118, .external_lex_state = 1},
  [889] = {.lex_state = 120, .external_lex_state = 1},
  [890] = {.lex_state = 118, .external_lex_state = 1},
  [891] = {.lex_state = 120},
  [892] = {.lex_state = 120, .external_lex_state = 1},
  [893] = {.lex_state = 120, .external_lex_state = 1},
  [894] = {.lex_state = 120, .external_lex_state = 1},
  [895] = {.lex_state = 120, .external_lex_state = 1},
  [896] = {.lex_state = 120},
  [897] = {.lex_state = 120, .external_lex_state = 1},
  [898] = {.lex_state = 120},
  [899] = {.lex_state = 120, .external_lex_state = 1},
  [900] = {.lex_state = 120},
  [901] = {.lex_state = 120},
  [902] = {.lex_state = 120},
  [903] = {.lex_state = 120},
  [904] = {.lex_state = 110, .external_lex_state = 1},
  [905] = {.lex_state = 120, .external_lex_state = 1},
  [906] = {.lex_state = 120},
  [907] = {.lex_state = 120, .external_lex_state = 1},
  [908] = {.lex_state = 120, .external_lex_state = 1},
  [909] = {.lex_state = 120},
  [910] = {.lex_state = 120},
  [911] = {.lex_state = 110},
  [912] = {.lex_state = 110},
  [913] = {.lex_state = 120},
  [914] = {.lex_state = 120, .external_lex_state = 1},
  [915] = {.lex_state = 110},
  [916] = {.lex_state = 110, .external_lex_state = 1},
  [917] = {.lex_state = 110, .external_lex_state = 1},
  [918] = {.lex_state = 111, .external_lex_state = 1},
  [919] = {.lex_state = 111, .external_lex_state = 1},
  [920] = {.lex_state = 160},
  [921] = {.lex_state = 111},
  [922] = {.lex_state = 111},
  [923] = {.lex_state = 111, .external_lex_state = 1},
  [924] = {.lex_state = 111},
  [925] = {.lex_state = 110, .external_lex_state = 1},
  [926] = {.lex_state = 110, .external_lex_state = 1},
  [927] = {.lex_state = 160},
  [928] = {.lex_state = 110},
  [929] = {.lex_state = 151},
  [930] = {.lex_state = 110, .external_lex_state = 1},
  [931] = {.lex_state = 110},
  [932] = {.lex_state = 110},
  [933] = {.lex_state = 160},
  [934] = {.lex_state = 160},
  [935] = {.lex_state = 160},
  [936] = {.lex_state = 160},
  [937] = {.lex_state = 160},
  [938] = {.lex_state = 111, .external_lex_state = 1},
  [939] = {.lex_state = 160},
  [940] = {.lex_state = 160},
  [941] = {.lex_state = 111, .external_lex_state = 1},
  [942] = {.lex_state = 160},
  [943] = {.lex_state = 160},
  [944] = {.lex_state = 111},
  [945] = {.lex_state = 160},
  [946] = {.lex_state = 160},
  [947] = {.lex_state = 111},
  [948] = {.lex_state = 111},
  [949] = {.lex_state = 160},
  [950] = {.lex_state = 160},
  [951] = {.lex_state = 160},
  [952] = {.lex_state = 160},
  [953] = {.lex_state = 160},
  [954] = {.lex_state = 160},
  [955] = {.lex_state = 160},
  [956] = {.lex_state = 160},
  [957] = {.lex_state = 160},
  [958] = {.lex_state = 160},
  [959] = {.lex_state = 160},
  [960] = {.lex_state = 160},
  [961] = {.lex_state = 160},
  [962] = {.lex_state = 160},
  [963] = {.lex_state = 514},
  [964] = {.lex_state = 160},
  [965] = {.lex_state = 160},
  [966] = {.lex_state = 111, .external_lex_state = 1},
  [967] = {.lex_state = 160},
  [968] = {.lex_state = 160},
  [969] = {.lex_state = 514, .external_lex_state = 1},
  [970] = {.lex_state = 160},
  [971] = {.lex_state = 160},
  [972] = {.lex_state = 160},
  [973] = {.lex_state = 160},
  [974] = {.lex_state = 122},
  [975] = {.lex_state = 122},
  [976] = {.lex_state = 122},
  [977] = {.lex_state = 122, .external_lex_state = 1},
  [978] = {.lex_state = 122, .external_lex_state = 1},
  [979] = {.lex_state = 160, .external_lex_state = 1},
  [980] = {.lex_state = 122, .external_lex_state = 1},
  [981] = {.lex_state = 160, .external_lex_state = 1},
  [982] = {.lex_state = 122},
  [983] = {.lex_state = 160, .external_lex_state = 1},
  [984] = {.lex_state = 160, .external_lex_state = 1},
  [985] = {.lex_state = 160, .external_lex_state = 1},
  [986] = {.lex_state = 160, .external_lex_state = 1},
  [987] = {.lex_state = 160, .external_lex_state = 1},
  [988] = {.lex_state = 160, .external_lex_state = 1},
  [989] = {.lex_state = 160, .external_lex_state = 1},
  [990] = {.lex_state = 160, .external_lex_state = 1},
  [991] = {.lex_state = 123, .external_lex_state = 1},
  [992] = {.lex_state = 122},
  [993] = {.lex_state = 122},
  [994] = {.lex_state = 160, .external_lex_state = 1},
  [995] = {.lex_state = 123, .external_lex_state = 1},
  [996] = {.lex_state = 123},
  [997] = {.lex_state = 123},
  [998] = {.lex_state = 122},
  [999] = {.lex_state = 122},
  [1000] = {.lex_state = 160, .external_lex_state = 1},
  [1001] = {.lex_state = 122},
  [1002] = {.lex_state = 160, .external_lex_state = 1},
  [1003] = {.lex_state = 122},
  [1004] = {.lex_state = 160, .external_lex_state = 1},
  [1005] = {.lex_state = 123, .external_lex_state = 1},
  [1006] = {.lex_state = 150},
  [1007] = {.lex_state = 122},
  [1008] = {.lex_state = 160, .external_lex_state = 1},
  [1009] = {.lex_state = 160, .external_lex_state = 1},
  [1010] = {.lex_state = 122, .external_lex_state = 1},
  [1011] = {.lex_state = 122, .external_lex_state = 1},
  [1012] = {.lex_state = 160, .external_lex_state = 1},
  [1013] = {.lex_state = 160, .external_lex_state = 1},
  [1014] = {.lex_state = 160, .external_lex_state = 1},
  [1015] = {.lex_state = 122, .external_lex_state = 1},
  [1016] = {.lex_state = 122, .external_lex_state = 1},
  [1017] = {.lex_state = 122, .external_lex_state = 1},
  [1018] = {.lex_state = 160, .external_lex_state = 1},
  [1019] = {.lex_state = 160, .external_lex_state = 1},
  [1020] = {.lex_state = 122, .external_lex_state = 1},
  [1021] = {.lex_state = 160, .external_lex_state = 1},
  [1022] = {.lex_state = 160, .external_lex_state = 1},
  [1023] = {.lex_state = 122, .external_lex_state = 1},
  [1024] = {.lex_state = 160, .external_lex_state = 1},
  [1025] = {.lex_state = 160, .external_lex_state = 1},
  [1026] = {.lex_state = 160, .external_lex_state = 1},
  [1027] = {.lex_state = 160, .external_lex_state = 1},
  [1028] = {.lex_state = 160, .external_lex_state = 1},
  [1029] = {.lex_state = 160, .external_lex_state = 1},
  [1030] = {.lex_state = 122, .external_lex_state = 1},
  [1031] = {.lex_state = 160, .external_lex_state = 1},
  [1032] = {.lex_state = 160, .external_lex_state = 1},
  [1033] = {.lex_state = 160, .external_lex_state = 1},
  [1034] = {.lex_state = 160, .external_lex_state = 1},
  [1035] = {.lex_state = 160, .external_lex_state = 1},
  [1036] = {.lex_state = 160, .external_lex_state = 1},
  [1037] = {.lex_state = 160},
  [1038] = {.lex_state = 160, .external_lex_state = 1},
  [1039] = {.lex_state = 123},
  [1040] = {.lex_state = 122, .external_lex_state = 1},
  [1041] = {.lex_state = 122},
  [1042] = {.lex_state = 122, .external_lex_state = 1},
  [1043] = {.lex_state = 122, .external_lex_state = 1},
  [1044] = {.lex_state = 122, .external_lex_state = 1},
  [1045] = {.lex_state = 122, .external_lex_state = 1},
  [1046] = {.lex_state = 122, .external_lex_state = 1},
  [1047] = {.lex_state = 122},
  [1048] = {.lex_state = 122},
  [1049] = {.lex_state = 122},
  [1050] = {.lex_state = 123},
  [1051] = {.lex_state = 122},
  [1052] = {.lex_state = 122, .external_lex_state = 1},
  [1053] = {.lex_state = 122, .external_lex_state = 1},
  [1054] = {.lex_state = 122},
  [1055] = {.lex_state = 122},
  [1056] = {.lex_state = 122, .external_lex_state = 1},
  [1057] = {.lex_state = 122},
  [1058] = {.lex_state = 123, .external_lex_state = 1},
  [1059] = {.lex_state = 123},
  [1060] = {.lex_state = 122, .external_lex_state = 1},
  [1061] = {.lex_state = 122},
  [1062] = {.lex_state = 123, .external_lex_state = 1},
  [1063] = {.lex_state = 122, .external_lex_state = 1},
  [1064] = {.lex_state = 123},
  [1065] = {.lex_state = 123, .external_lex_state = 1},
  [1066] = {.lex_state = 123, .external_lex_state = 1},
  [1067] = {.lex_state = 123, .external_lex_state = 1},
  [1068] = {.lex_state = 122, .external_lex_state = 1},
  [1069] = {.lex_state = 123, .external_lex_state = 1},
  [1070] = {.lex_state = 123, .external_lex_state = 1},
  [1071] = {.lex_state = 160, .external_lex_state = 1},
  [1072] = {.lex_state = 160},
  [1073] = {.lex_state = 123, .external_lex_state = 1},
  [1074] = {.lex_state = 122},
  [1075] = {.lex_state = 122},
  [1076] = {.lex_state = 123},
  [1077] = {.lex_state = 123},
  [1078] = {.lex_state = 123},
  [1079] = {.lex_state = 123},
  [1080] = {.lex_state = 123},
  [1081] = {.lex_state = 122},
  [1082] = {.lex_state = 123},
  [1083] = {.lex_state = 123, .external_lex_state = 1},
  [1084] = {.lex_state = 123, .external_lex_state = 1},
  [1085] = {.lex_state = 123, .external_lex_state = 1},
  [1086] = {.lex_state = 160},
  [1087] = {.lex_state = 123},
  [1088] = {.lex_state = 123, .external_lex_state = 1},
  [1089] = {.lex_state = 123, .external_lex_state = 1},
  [1090] = {.lex_state = 123, .external_lex_state = 1},
  [1091] = {.lex_state = 123, .external_lex_state = 1},
  [1092] = {.lex_state = 123},
  [1093] = {.lex_state = 123, .external_lex_state = 1},
  [1094] = {.lex_state = 160},
  [1095] = {.lex_state = 514},
  [1096] = {.lex_state = 160, .external_lex_state = 1},
  [1097] = {.lex_state = 123},
  [1098] = {.lex_state = 123, .external_lex_state = 1},
  [1099] = {.lex_state = 123},
  [1100] = {.lex_state = 160, .external_lex_state = 1},
  [1101] = {.lex_state = 123, .external_lex_state = 1},
  [1102] = {.lex_state = 123},
  [1103] = {.lex_state = 123},
  [1104] = {.lex_state = 160, .external_lex_state = 1},
  [1105] = {.lex_state = 123},
  [1106] = {.lex_state = 123, .external_lex_state = 1},
  [1107] = {.lex_state = 123},
  [1108] = {.lex_state = 123, .external_lex_state = 1},
  [1109] = {.lex_state = 123},
  [1110] = {.lex_state = 123},
  [1111] = {.lex_state = 136},
  [1112] = {.lex_state = 123},
  [1113] = {.lex_state = 514, .external_lex_state = 1},
  [1114] = {.lex_state = 514, .external_lex_state = 1},
  [1115] = {.lex_state = 514},
  [1116] = {.lex_state = 514, .external_lex_state = 1},
  [1117] = {.lex_state = 514},
  [1118] = {.lex_state = 136, .external_lex_state = 1},
  [1119] = {.lex_state = 131},
  [1120] = {.lex_state = 131},
  [1121] = {.lex_state = 131},
  [1122] = {.lex_state = 514},
  [1123] = {.lex_state = 514, .external_lex_state = 1},
  [1124] = {.lex_state = 514},
  [1125] = {.lex_state = 514},
  [1126] = {.lex_state = 514, .external_lex_state = 1},
  [1127] = {.lex_state = 514, .external_lex_state = 1},
  [1128] = {.lex_state = 514},
  [1129] = {.lex_state = 514, .external_lex_state = 1},
  [1130] = {.lex_state = 514, .external_lex_state = 1},
  [1131] = {.lex_state = 514},
  [1132] = {.lex_state = 514},
  [1133] = {.lex_state = 514},
  [1134] = {.lex_state = 514, .external_lex_state = 1},
  [1135] = {.lex_state = 514},
  [1136] = {.lex_state = 514, .external_lex_state = 1},
  [1137] = {.lex_state = 514, .external_lex_state = 1},
  [1138] = {.lex_state = 131},
  [1139] = {.lex_state = 131},
  [1140] = {.lex_state = 151},
  [1141] = {.lex_state = 151},
  [1142] = {.lex_state = 151},
  [1143] = {.lex_state = 173},
  [1144] = {.lex_state = 173},
  [1145] = {.lex_state = 109},
  [1146] = {.lex_state = 144},
  [1147] = {.lex_state = 144},
  [1148] = {.lex_state = 144},
  [1149] = {.lex_state = 144},
  [1150] = {.lex_state = 144},
  [1151] = {.lex_state = 144},
  [1152] = {.lex_state = 144},
  [1153] = {.lex_state = 144},
  [1154] = {.lex_state = 144},
  [1155] = {.lex_state = 144},
  [1156] = {.lex_state = 144},
  [1157] = {.lex_state = 144},
  [1158] = {.lex_state = 144},
  [1159] = {.lex_state = 144},
  [1160] = {.lex_state = 144},
  [1161] = {.lex_state = 144},
  [1162] = {.lex_state = 144},
  [1163] = {.lex_state = 144},
  [1164] = {.lex_state = 144},
  [1165] = {.lex_state = 144},
  [1166] = {.lex_state = 144},
  [1167] = {.lex_state = 144},
  [1168] = {.lex_state = 144},
  [1169] = {.lex_state = 144},
  [1170] = {.lex_state = 144},
  [1171] = {.lex_state = 153},
  [1172] = {.lex_state = 153},
  [1173] = {.lex_state = 153},
  [1174] = {.lex_state = 153},
  [1175] = {.lex_state = 153},
  [1176] = {.lex_state = 154},
  [1177] = {.lex_state = 154},
  [1178] = {.lex_state = 153},
  [1179] = {.lex_state = 154},
  [1180] = {.lex_state = 153},
  [1181] = {.lex_state = 154},
  [1182] = {.lex_state = 153},
  [1183] = {.lex_state = 154},
  [1184] = {.lex_state = 153},
  [1185] = {.lex_state = 154},
  [1186] = {.lex_state = 154},
  [1187] = {.lex_state = 154},
  [1188] = {.lex_state = 153},
  [1189] = {.lex_state = 154},
  [1190] = {.lex_state = 154},
  [1191] = {.lex_state = 154},
  [1192] = {.lex_state = 154},
  [1193] = {.lex_state = 154},
  [1194] = {.lex_state = 154},
  [1195] = {.lex_state = 154},
  [1196] = {.lex_state = 154},
  [1197] = {.lex_state = 55},
  [1198] = {.lex_state = 55},
  [1199] = {.lex_state = 55},
  [1200] = {.lex_state = 153},
  [1201] = {.lex_state = 55},
  [1202] = {.lex_state = 55},
  [1203] = {.lex_state = 55},
  [1204] = {.lex_state = 55},
  [1205] = {.lex_state = 55},
  [1206] = {.lex_state = 55},
  [1207] = {.lex_state = 55},
  [1208] = {.lex_state = 55},
  [1209] = {.lex_state = 55},
  [1210] = {.lex_state = 154},
  [1211] = {.lex_state = 55},
  [1212] = {.lex_state = 55},
  [1213] = {.lex_state = 55},
  [1214] = {.lex_state = 55},
  [1215] = {.lex_state = 55},
  [1216] = {.lex_state = 514},
  [1217] = {.lex_state = 514},
  [1218] = {.lex_state = 514},
  [1219] = {.lex_state = 514},
  [1220] = {.lex_state = 514},
  [1221] = {.lex_state = 514},
  [1222] = {.lex_state = 514},
  [1223] = {.lex_state = 163},
  [1224] = {.lex_state = 163},
  [1225] = {.lex_state = 144},
  [1226] = {.lex_state = 144},
  [1227] = {.lex_state = 144},
  [1228] = {.lex_state = 144},
  [1229] = {.lex_state = 154},
  [1230] = {.lex_state = 154},
  [1231] = {.lex_state = 153},
  [1232] = {.lex_state = 144},
  [1233] = {.lex_state = 55},
  [1234] = {.lex_state = 514},
  [1235] = {.lex_state = 154},
  [1236] = {.lex_state = 153},
  [1237] = {.lex_state = 144},
  [1238] = {.lex_state = 154},
  [1239] = {.lex_state = 55},
  [1240] = {.lex_state = 153},
  [1241] = {.lex_state = 55},
  [1242] = {.lex_state = 144},
  [1243] = {.lex_state = 514},
  [1244] = {.lex_state = 514},
  [1245] = {.lex_state = 514},
  [1246] = {.lex_state = 514},
  [1247] = {.lex_state = 144},
  [1248] = {.lex_state = 162},
  [1249] = {.lex_state = 144},
  [1250] = {.lex_state = 162},
  [1251] = {.lex_state = 144},
  [1252] = {.lex_state = 514},
  [1253] = {.lex_state = 514},
  [1254] = {.lex_state = 144},
  [1255] = {.lex_state = 144},
  [1256] = {.lex_state = 514},
  [1257] = {.lex_state = 144},
  [1258] = {.lex_state = 144},
  [1259] = {.lex_state = 514},
  [1260] = {.lex_state = 514},
  [1261] = {.lex_state = 144},
  [1262] = {.lex_state = 514},
  [1263] = {.lex_state = 160},
  [1264] = {.lex_state = 144},
  [1265] = {.lex_state = 514},
  [1266] = {.lex_state = 144},
  [1267] = {.lex_state = 514},
  [1268] = {.lex_state = 514},
  [1269] = {.lex_state = 160},
  [1270] = {.lex_state = 160},
  [1271] = {.lex_state = 160},
  [1272] = {.lex_state = 162},
  [1273] = {.lex_state = 160},
  [1274] = {.lex_state = 160},
  [1275] = {.lex_state = 131},
  [1276] = {.lex_state = 131},
  [1277] = {.lex_state = 514},
  [1278] = {.lex_state = 514},
  [1279] = {.lex_state = 160},
  [1280] = {.lex_state = 514},
  [1281] = {.lex_state = 514},
  [1282] = {.lex_state = 514},
  [1283] = {.lex_state = 155},
  [1284] = {.lex_state = 156, .external_lex_state = 1},
  [1285] = {.lex_state = 160},
  [1286] = {.lex_state = 514},
  [1287] = {.lex_state = 514},
  [1288] = {.lex_state = 514},
  [1289] = {.lex_state = 514},
  [1290] = {.lex_state = 155},
  [1291] = {.lex_state = 514},
  [1292] = {.lex_state = 514},
  [1293] = {.lex_state = 514},
  [1294] = {.lex_state = 514},
  [1295] = {.lex_state = 160, .external_lex_state = 1},
  [1296] = {.lex_state = 514},
  [1297] = {.lex_state = 514},
  [1298] = {.lex_state = 156, .external_lex_state = 1},
  [1299] = {.lex_state = 514},
  [1300] = {.lex_state = 514},
  [1301] = {.lex_state = 514},
  [1302] = {.lex_state = 514},
  [1303] = {.lex_state = 152},
  [1304] = {.lex_state = 162, .external_lex_state = 1},
  [1305] = {.lex_state = 160, .external_lex_state = 1},
  [1306] = {.lex_state = 152},
  [1307] = {.lex_state = 514, .external_lex_state = 1},
  [1308] = {.lex_state = 514},
  [1309] = {.lex_state = 160},
  [1310] = {.lex_state = 514},
  [1311] = {.lex_state = 514, .external_lex_state = 1},
  [1312] = {.lex_state = 160},
  [1313] = {.lex_state = 514, .external_lex_state = 1},
  [1314] = {.lex_state = 160},
  [1315] = {.lex_state = 514},
  [1316] = {.lex_state = 514, .external_lex_state = 1},
  [1317] = {.lex_state = 152},
  [1318] = {.lex_state = 160},
  [1319] = {.lex_state = 162},
  [1320] = {.lex_state = 514},
  [1321] = {.lex_state = 160},
  [1322] = {.lex_state = 160},
  [1323] = {.lex_state = 152},
  [1324] = {.lex_state = 165},
  [1325] = {.lex_state = 165},
  [1326] = {.lex_state = 152},
  [1327] = {.lex_state = 152},
  [1328] = {.lex_state = 160},
  [1329] = {.lex_state = 165},
  [1330] = {.lex_state = 160},
  [1331] = {.lex_state = 165},
  [1332] = {.lex_state = 514},
  [1333] = {.lex_state = 514},
  [1334] = {.lex_state = 514},
  [1335] = {.lex_state = 514},
  [1336] = {.lex_state = 514},
  [1337] = {.lex_state = 514},
  [1338] = {.lex_state = 514},
  [1339] = {.lex_state = 514},
  [1340] = {.lex_state = 165},
  [1341] = {.lex_state = 514},
  [1342] = {.lex_state = 514, .external_lex_state = 1},
  [1343] = {.lex_state = 165},
  [1344] = {.lex_state = 165},
  [1345] = {.lex_state = 514, .external_lex_state = 1},
  [1346] = {.lex_state = 165},
  [1347] = {.lex_state = 514, .external_lex_state = 1},
  [1348] = {.lex_state = 514},
  [1349] = {.lex_state = 165},
  [1350] = {.lex_state = 165},
  [1351] = {.lex_state = 162, .external_lex_state = 1},
  [1352] = {.lex_state = 514},
  [1353] = {.lex_state = 162},
  [1354] = {.lex_state = 165},
  [1355] = {.lex_state = 162, .external_lex_state = 1},
  [1356] = {.lex_state = 165},
  [1357] = {.lex_state = 165},
  [1358] = {.lex_state = 165},
  [1359] = {.lex_state = 165},
  [1360] = {.lex_state = 514},
  [1361] = {.lex_state = 165},
  [1362] = {.lex_state = 165},
  [1363] = {.lex_state = 165},
  [1364] = {.lex_state = 165},
  [1365] = {.lex_state = 165},
  [1366] = {.lex_state = 514},
  [1367] = {.lex_state = 514, .external_lex_state = 1},
  [1368] = {.lex_state = 514, .external_lex_state = 1},
  [1369] = {.lex_state = 165},
  [1370] = {.lex_state = 165},
  [1371] = {.lex_state = 514},
  [1372] = {.lex_state = 514, .external_lex_state = 1},
  [1373] = {.lex_state = 514, .external_lex_state = 1},
  [1374] = {.lex_state = 514},
  [1375] = {.lex_state = 514},
  [1376] = {.lex_state = 162},
  [1377] = {.lex_state = 514, .external_lex_state = 1},
  [1378] = {.lex_state = 514, .external_lex_state = 1},
  [1379] = {.lex_state = 514},
  [1380] = {.lex_state = 514},
  [1381] = {.lex_state = 165},
  [1382] = {.lex_state = 165},
  [1383] = {.lex_state = 514},
  [1384] = {.lex_state = 514, .external_lex_state = 1},
  [1385] = {.lex_state = 514, .external_lex_state = 1},
  [1386] = {.lex_state = 514, .external_lex_state = 1},
  [1387] = {.lex_state = 514, .external_lex_state = 1},
  [1388] = {.lex_state = 165},
  [1389] = {.lex_state = 514},
  [1390] = {.lex_state = 165},
  [1391] = {.lex_state = 514},
  [1392] = {.lex_state = 157},
  [1393] = {.lex_state = 514, .external_lex_state = 1},
  [1394] = {.lex_state = 131},
  [1395] = {.lex_state = 131},
  [1396] = {.lex_state = 131},
  [1397] = {.lex_state = 514},
  [1398] = {.lex_state = 157},
  [1399] = {.lex_state = 514, .external_lex_state = 1},
  [1400] = {.lex_state = 514},
  [1401] = {.lex_state = 131},
  [1402] = {.lex_state = 131},
  [1403] = {.lex_state = 514},
  [1404] = {.lex_state = 131},
  [1405] = {.lex_state = 514},
  [1406] = {.lex_state = 514},
  [1407] = {.lex_state = 131},
  [1408] = {.lex_state = 160},
  [1409] = {.lex_state = 160},
  [1410] = {.lex_state = 131},
  [1411] = {.lex_state = 131},
  [1412] = {.lex_state = 514},
  [1413] = {.lex_state = 514},
  [1414] = {.lex_state = 131},
  [1415] = {.lex_state = 514},
  [1416] = {.lex_state = 160},
  [1417] = {.lex_state = 131},
  [1418] = {.lex_state = 160},
  [1419] = {.lex_state = 514},
  [1420] = {.lex_state = 147},
  [1421] = {.lex_state = 20},
  [1422] = {.lex_state = 514},
  [1423] = {.lex_state = 514},
  [1424] = {.lex_state = 514},
  [1425] = {.lex_state = 147},
  [1426] = {.lex_state = 147},
  [1427] = {.lex_state = 147},
  [1428] = {.lex_state = 514},
  [1429] = {.lex_state = 147},
  [1430] = {.lex_state = 514},
  [1431] = {.lex_state = 514},
  [1432] = {.lex_state = 160},
  [1433] = {.lex_state = 514},
  [1434] = {.lex_state = 514},
  [1435] = {.lex_state = 514},
  [1436] = {.lex_state = 514},
  [1437] = {.lex_state = 147},
  [1438] = {.lex_state = 514},
  [1439] = {.lex_state = 20},
  [1440] = {.lex_state = 514},
  [1441] = {.lex_state = 514},
  [1442] = {.lex_state = 20},
  [1443] = {.lex_state = 514},
  [1444] = {.lex_state = 147},
  [1445] = {.lex_state = 20},
  [1446] = {.lex_state = 514},
  [1447] = {.lex_state = 514},
  [1448] = {.lex_state = 514},
  [1449] = {.lex_state = 514, .external_lex_state = 1},
  [1450] = {.lex_state = 142},
  [1451] = {.lex_state = 142},
  [1452] = {.lex_state = 142},
  [1453] = {.lex_state = 142},
  [1454] = {.lex_state = 147},
  [1455] = {.lex_state = 514},
  [1456] = {.lex_state = 142},
  [1457] = {.lex_state = 142},
  [1458] = {.lex_state = 514},
  [1459] = {.lex_state = 160},
  [1460] = {.lex_state = 160},
  [1461] = {.lex_state = 514},
  [1462] = {.lex_state = 142},
  [1463] = {.lex_state = 514},
  [1464] = {.lex_state = 514},
  [1465] = {.lex_state = 20},
  [1466] = {.lex_state = 147},
  [1467] = {.lex_state = 514},
  [1468] = {.lex_state = 20},
  [1469] = {.lex_state = 514},
  [1470] = {.lex_state = 514},
  [1471] = {.lex_state = 514},
  [1472] = {.lex_state = 147},
  [1473] = {.lex_state = 20},
  [1474] = {.lex_state = 147},
  [1475] = {.lex_state = 20},
  [1476] = {.lex_state = 514, .external_lex_state = 1},
  [1477] = {.lex_state = 514},
  [1478] = {.lex_state = 514},
  [1479] = {.lex_state = 160},
  [1480] = {.lex_state = 514},
  [1481] = {.lex_state = 147},
  [1482] = {.lex_state = 514},
  [1483] = {.lex_state = 514},
  [1484] = {.lex_state = 147},
  [1485] = {.lex_state = 514},
  [1486] = {.lex_state = 514},
  [1487] = {.lex_state = 160},
  [1488] = {.lex_state = 514},
  [1489] = {.lex_state = 514},
  [1490] = {.lex_state = 20},
  [1491] = {.lex_state = 514},
  [1492] = {.lex_state = 20},
  [1493] = {.lex_state = 514, .external_lex_state = 1},
  [1494] = {.lex_state = 514},
  [1495] = {.lex_state = 20},
  [1496] = {.lex_state = 142},
  [1497] = {.lex_state = 514},
  [1498] = {.lex_state = 514},
  [1499] = {.lex_state = 142},
  [1500] = {.lex_state = 514},
  [1501] = {.lex_state = 165},
  [1502] = {.lex_state = 20},
  [1503] = {.lex_state = 147},
  [1504] = {.lex_state = 514, .external_lex_state = 1},
  [1505] = {.lex_state = 514},
  [1506] = {.lex_state = 514, .external_lex_state = 1},
  [1507] = {.lex_state = 514},
  [1508] = {.lex_state = 142},
  [1509] = {.lex_state = 514, .external_lex_state = 1},
  [1510] = {.lex_state = 147},
  [1511] = {.lex_state = 147},
  [1512] = {.lex_state = 514},
  [1513] = {.lex_state = 514},
  [1514] = {.lex_state = 147},
  [1515] = {.lex_state = 142},
  [1516] = {.lex_state = 160},
  [1517] = {.lex_state = 160},
  [1518] = {.lex_state = 160},
  [1519] = {.lex_state = 168},
  [1520] = {.lex_state = 514},
  [1521] = {.lex_state = 514},
  [1522] = {.lex_state = 514},
  [1523] = {.lex_state = 514},
  [1524] = {.lex_state = 514},
  [1525] = {.lex_state = 514},
  [1526] = {.lex_state = 514},
  [1527] = {.lex_state = 514},
  [1528] = {.lex_state = 514},
  [1529] = {.lex_state = 514},
  [1530] = {.lex_state = 514},
  [1531] = {.lex_state = 514},
  [1532] = {.lex_state = 514},
  [1533] = {.lex_state = 514},
  [1534] = {.lex_state = 514},
  [1535] = {.lex_state = 514},
  [1536] = {.lex_state = 514},
  [1537] = {.lex_state = 514},
  [1538] = {.lex_state = 514},
  [1539] = {.lex_state = 514},
  [1540] = {.lex_state = 514},
  [1541] = {.lex_state = 514},
  [1542] = {.lex_state = 514},
  [1543] = {.lex_state = 514},
  [1544] = {.lex_state = 514},
  [1545] = {.lex_state = 514},
  [1546] = {.lex_state = 514},
  [1547] = {.lex_state = 514, .external_lex_state = 1},
  [1548] = {.lex_state = 514},
  [1549] = {.lex_state = 514},
  [1550] = {.lex_state = 514},
  [1551] = {.lex_state = 514},
  [1552] = {.lex_state = 514},
  [1553] = {.lex_state = 514},
  [1554] = {.lex_state = 514},
  [1555] = {.lex_state = 514},
  [1556] = {.lex_state = 514},
  [1557] = {.lex_state = 514},
  [1558] = {.lex_state = 514},
  [1559] = {.lex_state = 514},
  [1560] = {.lex_state = 514},
  [1561] = {.lex_state = 514},
  [1562] = {.lex_state = 514},
  [1563] = {.lex_state = 514},
  [1564] = {.lex_state = 514},
  [1565] = {.lex_state = 514},
  [1566] = {.lex_state = 514},
  [1567] = {.lex_state = 514},
  [1568] = {.lex_state = 514},
  [1569] = {.lex_state = 514},
  [1570] = {.lex_state = 514, .external_lex_state = 1},
  [1571] = {.lex_state = 514},
  [1572] = {.lex_state = 514},
  [1573] = {.lex_state = 514},
  [1574] = {.lex_state = 514},
  [1575] = {.lex_state = 514},
  [1576] = {.lex_state = 514},
  [1577] = {.lex_state = 514, .external_lex_state = 1},
  [1578] = {.lex_state = 514},
  [1579] = {.lex_state = 514},
  [1580] = {.lex_state = 514},
  [1581] = {.lex_state = 514},
  [1582] = {.lex_state = 514},
  [1583] = {.lex_state = 514},
  [1584] = {.lex_state = 514},
  [1585] = {.lex_state = 514},
  [1586] = {.lex_state = 514},
  [1587] = {.lex_state = 514},
  [1588] = {.lex_state = 514},
  [1589] = {.lex_state = 514},
  [1590] = {.lex_state = 514},
  [1591] = {.lex_state = 514},
  [1592] = {.lex_state = 514},
  [1593] = {.lex_state = 514},
  [1594] = {.lex_state = 514},
  [1595] = {.lex_state = 514},
  [1596] = {.lex_state = 514},
  [1597] = {.lex_state = 514},
  [1598] = {.lex_state = 514},
  [1599] = {.lex_state = 514},
  [1600] = {.lex_state = 514},
  [1601] = {.lex_state = 514},
  [1602] = {.lex_state = 514},
  [1603] = {.lex_state = 514, .external_lex_state = 1},
  [1604] = {.lex_state = 514},
  [1605] = {.lex_state = 514},
  [1606] = {.lex_state = 514},
  [1607] = {.lex_state = 514},
  [1608] = {.lex_state = 514},
  [1609] = {.lex_state = 514},
  [1610] = {.lex_state = 514},
  [1611] = {.lex_state = 514},
  [1612] = {.lex_state = 514},
  [1613] = {.lex_state = 514},
  [1614] = {.lex_state = 514},
  [1615] = {.lex_state = 514},
  [1616] = {.lex_state = 514},
  [1617] = {.lex_state = 514},
  [1618] = {.lex_state = 514},
  [1619] = {.lex_state = 514},
  [1620] = {.lex_state = 514},
  [1621] = {.lex_state = 514},
  [1622] = {.lex_state = 514},
  [1623] = {.lex_state = 514},
  [1624] = {.lex_state = 514},
  [1625] = {.lex_state = 514},
  [1626] = {.lex_state = 514},
  [1627] = {.lex_state = 514},
  [1628] = {.lex_state = 160},
  [1629] = {.lex_state = 514},
  [1630] = {.lex_state = 514},
  [1631] = {.lex_state = 514, .external_lex_state = 1},
  [1632] = {.lex_state = 514},
  [1633] = {.lex_state = 514},
  [1634] = {.lex_state = 514},
  [1635] = {.lex_state = 514},
  [1636] = {.lex_state = 514},
  [1637] = {.lex_state = 514},
  [1638] = {.lex_state = 174},
  [1639] = {.lex_state = 144},
  [1640] = {.lex_state = 514},
  [1641] = {.lex_state = 514},
  [1642] = {.lex_state = 514},
  [1643] = {.lex_state = 514},
  [1644] = {.lex_state = 514},
  [1645] = {.lex_state = 514},
  [1646] = {.lex_state = 514},
  [1647] = {.lex_state = 514},
  [1648] = {.lex_state = 144},
  [1649] = {.lex_state = 514},
  [1650] = {.lex_state = 514},
  [1651] = {.lex_state = 514},
  [1652] = {.lex_state = 514},
  [1653] = {.lex_state = 514},
  [1654] = {.lex_state = 514},
  [1655] = {.lex_state = 514},
  [1656] = {.lex_state = 514},
  [1657] = {.lex_state = 514},
  [1658] = {.lex_state = 144},
  [1659] = {.lex_state = 514},
  [1660] = {.lex_state = 514},
  [1661] = {.lex_state = 514},
  [1662] = {.lex_state = 514},
  [1663] = {.lex_state = 514},
  [1664] = {.lex_state = 514},
  [1665] = {.lex_state = 514},
  [1666] = {.lex_state = 514},
  [1667] = {.lex_state = 514},
  [1668] = {.lex_state = 514},
  [1669] = {.lex_state = 514},
  [1670] = {.lex_state = 514},
  [1671] = {.lex_state = 514},
  [1672] = {.lex_state = 514},
  [1673] = {.lex_state = 514},
  [1674] = {.lex_state = 514},
  [1675] = {.lex_state = 514},
  [1676] = {.lex_state = 514},
  [1677] = {.lex_state = 514},
  [1678] = {.lex_state = 514},
  [1679] = {.lex_state = 514},
  [1680] = {.lex_state = 144},
  [1681] = {.lex_state = 514},
  [1682] = {.lex_state = 514},
  [1683] = {.lex_state = 514},
  [1684] = {.lex_state = 514},
  [1685] = {.lex_state = 514},
  [1686] = {.lex_state = 514},
  [1687] = {.lex_state = 514},
  [1688] = {.lex_state = 174},
  [1689] = {.lex_state = 174},
  [1690] = {.lex_state = 174},
  [1691] = {.lex_state = 514},
  [1692] = {.lex_state = 514},
  [1693] = {.lex_state = 514},
  [1694] = {.lex_state = 514},
  [1695] = {.lex_state = 514},
  [1696] = {.lex_state = 514},
  [1697] = {.lex_state = 514},
  [1698] = {.lex_state = 514},
  [1699] = {.lex_state = 514},
  [1700] = {.lex_state = 514},
  [1701] = {.lex_state = 514},
  [1702] = {.lex_state = 514, .external_lex_state = 1},
  [1703] = {.lex_state = 514},
  [1704] = {.lex_state = 514},
  [1705] = {.lex_state = 514},
  [1706] = {.lex_state = 514},
  [1707] = {.lex_state = 514},
  [1708] = {.lex_state = 514},
  [1709] = {.lex_state = 514},
  [1710] = {.lex_state = 514},
  [1711] = {.lex_state = 514, .external_lex_state = 1},
  [1712] = {.lex_state = 514},
  [1713] = {.lex_state = 514},
  [1714] = {.lex_state = 514},
  [1715] = {.lex_state = 514},
  [1716] = {.lex_state = 514},
  [1717] = {.lex_state = 514},
  [1718] = {.lex_state = 514},
  [1719] = {.lex_state = 514},
  [1720] = {.lex_state = 514},
  [1721] = {.lex_state = 514},
  [1722] = {.lex_state = 514},
  [1723] = {.lex_state = 514},
  [1724] = {.lex_state = 514},
  [1725] = {.lex_state = 514},
  [1726] = {.lex_state = 514},
  [1727] = {.lex_state = 514},
  [1728] = {.lex_state = 514},
  [1729] = {.lex_state = 514},
  [1730] = {.lex_state = 514},
  [1731] = {.lex_state = 514},
  [1732] = {.lex_state = 165},
  [1733] = {.lex_state = 514, .external_lex_state = 1},
  [1734] = {.lex_state = 514},
  [1735] = {.lex_state = 514},
  [1736] = {.lex_state = 514},
  [1737] = {.lex_state = 514},
  [1738] = {.lex_state = 514},
  [1739] = {.lex_state = 514},
  [1740] = {.lex_state = 514},
  [1741] = {.lex_state = 514},
  [1742] = {.lex_state = 514, .external_lex_state = 1},
  [1743] = {.lex_state = 514},
  [1744] = {.lex_state = 514},
  [1745] = {.lex_state = 514, .external_lex_state = 1},
  [1746] = {.lex_state = 514},
  [1747] = {.lex_state = 514},
  [1748] = {.lex_state = 514, .external_lex_state = 1},
  [1749] = {.lex_state = 514},
  [1750] = {.lex_state = 514},
  [1751] = {.lex_state = 514, .external_lex_state = 1},
  [1752] = {.lex_state = 514},
  [1753] = {.lex_state = 514},
  [1754] = {.lex_state = 514},
  [1755] = {.lex_state = 514},
  [1756] = {.lex_state = 514},
  [1757] = {.lex_state = 514},
  [1758] = {.lex_state = 514},
  [1759] = {.lex_state = 514},
  [1760] = {.lex_state = 514},
  [1761] = {.lex_state = 514},
  [1762] = {.lex_state = 514},
  [1763] = {.lex_state = 514},
  [1764] = {.lex_state = 514},
  [1765] = {.lex_state = 514},
  [1766] = {.lex_state = 514},
  [1767] = {.lex_state = 514},
  [1768] = {.lex_state = 514},
  [1769] = {.lex_state = 514, .external_lex_state = 1},
  [1770] = {.lex_state = 514},
  [1771] = {.lex_state = 514},
  [1772] = {.lex_state = 514, .external_lex_state = 1},
  [1773] = {.lex_state = 514},
  [1774] = {.lex_state = 514, .external_lex_state = 1},
  [1775] = {.lex_state = 514},
  [1776] = {.lex_state = 514},
  [1777] = {.lex_state = 514},
  [1778] = {.lex_state = 514, .external_lex_state = 1},
  [1779] = {.lex_state = 514},
  [1780] = {.lex_state = 514},
  [1781] = {.lex_state = 514, .external_lex_state = 1},
  [1782] = {.lex_state = 514},
  [1783] = {.lex_state = 514},
  [1784] = {.lex_state = 514},
  [1785] = {.lex_state = 514},
  [1786] = {.lex_state = 514},
  [1787] = {.lex_state = 514},
  [1788] = {.lex_state = 514},
  [1789] = {.lex_state = 514, .external_lex_state = 1},
  [1790] = {.lex_state = 514},
  [1791] = {.lex_state = 514},
  [1792] = {.lex_state = 514},
  [1793] = {.lex_state = 514},
  [1794] = {.lex_state = 514, .external_lex_state = 1},
  [1795] = {.lex_state = 514},
  [1796] = {.lex_state = 514},
  [1797] = {.lex_state = 514, .external_lex_state = 1},
  [1798] = {.lex_state = 514},
  [1799] = {.lex_state = 514},
  [1800] = {.lex_state = 514},
  [1801] = {.lex_state = 514},
  [1802] = {.lex_state = 514},
  [1803] = {.lex_state = 514},
  [1804] = {.lex_state = 514},
  [1805] = {.lex_state = 514},
  [1806] = {.lex_state = 514},
  [1807] = {.lex_state = 514, .external_lex_state = 1},
  [1808] = {.lex_state = 131},
  [1809] = {.lex_state = 514},
  [1810] = {.lex_state = 514},
  [1811] = {.lex_state = 514},
  [1812] = {.lex_state = 514},
  [1813] = {.lex_state = 514},
  [1814] = {.lex_state = 514},
  [1815] = {.lex_state = 514, .external_lex_state = 1},
  [1816] = {.lex_state = 514, .external_lex_state = 1},
  [1817] = {.lex_state = 514},
  [1818] = {.lex_state = 514},
  [1819] = {.lex_state = 514},
  [1820] = {.lex_state = 514},
  [1821] = {.lex_state = 514},
  [1822] = {.lex_state = 514, .external_lex_state = 1},
  [1823] = {.lex_state = 514},
  [1824] = {.lex_state = 514},
  [1825] = {.lex_state = 514},
  [1826] = {.lex_state = 514},
  [1827] = {.lex_state = 514},
  [1828] = {.lex_state = 514},
  [1829] = {.lex_state = 514},
  [1830] = {.lex_state = 514},
  [1831] = {.lex_state = 514},
  [1832] = {.lex_state = 514},
  [1833] = {.lex_state = 514},
  [1834] = {.lex_state = 514},
  [1835] = {.lex_state = 514},
  [1836] = {.lex_state = 514},
  [1837] = {.lex_state = 514},
  [1838] = {.lex_state = 514},
  [1839] = {.lex_state = 514},
  [1840] = {.lex_state = 514},
  [1841] = {.lex_state = 514},
  [1842] = {.lex_state = 514},
  [1843] = {.lex_state = 514},
  [1844] = {.lex_state = 514},
  [1845] = {.lex_state = 514},
  [1846] = {.lex_state = 514},
  [1847] = {.lex_state = 514},
  [1848] = {.lex_state = 514},
  [1849] = {.lex_state = 514},
  [1850] = {.lex_state = 514},
  [1851] = {.lex_state = 514},
  [1852] = {.lex_state = 514},
  [1853] = {.lex_state = 514},
  [1854] = {.lex_state = 514},
  [1855] = {.lex_state = 514},
  [1856] = {.lex_state = 514},
  [1857] = {.lex_state = 514},
  [1858] = {.lex_state = 514},
  [1859] = {.lex_state = 514},
  [1860] = {.lex_state = 514},
  [1861] = {.lex_state = 514},
  [1862] = {.lex_state = 514},
  [1863] = {.lex_state = 514},
  [1864] = {.lex_state = 514},
  [1865] = {.lex_state = 514},
  [1866] = {.lex_state = 514},
  [1867] = {.lex_state = 514},
  [1868] = {.lex_state = 514},
  [1869] = {.lex_state = 514},
  [1870] = {.lex_state = 514},
  [1871] = {.lex_state = 514, .external_lex_state = 1},
  [1872] = {.lex_state = 514},
  [1873] = {.lex_state = 514},
  [1874] = {.lex_state = 514},
  [1875] = {.lex_state = 514},
  [1876] = {.lex_state = 514},
  [1877] = {.lex_state = 514},
  [1878] = {.lex_state = 514},
  [1879] = {.lex_state = 131},
  [1880] = {.lex_state = 514},
  [1881] = {.lex_state = 514},
  [1882] = {.lex_state = 514, .external_lex_state = 1},
  [1883] = {.lex_state = 514},
  [1884] = {.lex_state = 514},
  [1885] = {.lex_state = 514},
  [1886] = {.lex_state = 514},
  [1887] = {.lex_state = 514},
  [1888] = {.lex_state = 514},
  [1889] = {.lex_state = 514},
  [1890] = {.lex_state = 514},
  [1891] = {.lex_state = 514},
  [1892] = {.lex_state = 514},
  [1893] = {.lex_state = 514},
  [1894] = {.lex_state = 514},
  [1895] = {.lex_state = 514},
  [1896] = {.lex_state = 514},
  [1897] = {.lex_state = 514},
  [1898] = {.lex_state = 514},
  [1899] = {.lex_state = 514},
  [1900] = {.lex_state = 514},
  [1901] = {.lex_state = 514},
  [1902] = {.lex_state = 514},
  [1903] = {.lex_state = 514},
  [1904] = {.lex_state = 514},
  [1905] = {.lex_state = 514},
  [1906] = {.lex_state = 514},
  [1907] = {.lex_state = 514, .external_lex_state = 1},
  [1908] = {.lex_state = 514},
  [1909] = {.lex_state = 514},
  [1910] = {.lex_state = 514},
  [1911] = {.lex_state = 514},
  [1912] = {.lex_state = 514},
  [1913] = {.lex_state = 514},
  [1914] = {.lex_state = 514},
  [1915] = {.lex_state = 514},
  [1916] = {.lex_state = 131},
  [1917] = {.lex_state = 514, .external_lex_state = 1},
  [1918] = {.lex_state = 514},
  [1919] = {.lex_state = 514, .external_lex_state = 1},
  [1920] = {.lex_state = 514, .external_lex_state = 1},
  [1921] = {.lex_state = 514, .external_lex_state = 1},
  [1922] = {.lex_state = 514},
  [1923] = {.lex_state = 514, .external_lex_state = 1},
  [1924] = {.lex_state = 514, .external_lex_state = 1},
  [1925] = {.lex_state = 514},
  [1926] = {.lex_state = 514},
  [1927] = {.lex_state = 514},
  [1928] = {.lex_state = 514},
  [1929] = {.lex_state = 514, .external_lex_state = 1},
  [1930] = {.lex_state = 514},
  [1931] = {.lex_state = 514},
  [1932] = {.lex_state = 514},
  [1933] = {.lex_state = 514},
  [1934] = {.lex_state = 514},
  [1935] = {.lex_state = 514, .external_lex_state = 1},
  [1936] = {.lex_state = 131},
  [1937] = {.lex_state = 514},
  [1938] = {.lex_state = 514, .external_lex_state = 1},
  [1939] = {.lex_state = 514},
  [1940] = {.lex_state = 514},
  [1941] = {.lex_state = 514},
  [1942] = {.lex_state = 514},
  [1943] = {.lex_state = 514},
  [1944] = {.lex_state = 514, .external_lex_state = 1},
  [1945] = {.lex_state = 514},
  [1946] = {.lex_state = 514},
  [1947] = {.lex_state = 514},
  [1948] = {.lex_state = 514},
  [1949] = {.lex_state = 514},
  [1950] = {.lex_state = 514},
  [1951] = {.lex_state = 514},
  [1952] = {.lex_state = 514},
  [1953] = {.lex_state = 514},
  [1954] = {.lex_state = 514},
  [1955] = {.lex_state = 514},
  [1956] = {.lex_state = 514},
  [1957] = {.lex_state = 514},
  [1958] = {.lex_state = 514},
  [1959] = {.lex_state = 514},
  [1960] = {.lex_state = 514, .external_lex_state = 1},
  [1961] = {.lex_state = 514},
  [1962] = {.lex_state = 514},
  [1963] = {.lex_state = 514},
  [1964] = {.lex_state = 514},
  [1965] = {.lex_state = 514},
  [1966] = {.lex_state = 514},
  [1967] = {.lex_state = 514},
  [1968] = {.lex_state = 514},
  [1969] = {.lex_state = 514},
  [1970] = {.lex_state = 514},
  [1971] = {.lex_state = 514},
  [1972] = {.lex_state = 514, .external_lex_state = 1},
  [1973] = {.lex_state = 514},
  [1974] = {.lex_state = 514},
  [1975] = {.lex_state = 514, .external_lex_state = 1},
  [1976] = {.lex_state = 514, .external_lex_state = 1},
  [1977] = {.lex_state = 514},
  [1978] = {.lex_state = 514, .external_lex_state = 1},
  [1979] = {.lex_state = 514},
  [1980] = {.lex_state = 514},
  [1981] = {.lex_state = 514},
  [1982] = {.lex_state = 514, .external_lex_state = 1},
  [1983] = {.lex_state = 514},
  [1984] = {.lex_state = 514, .external_lex_state = 1},
  [1985] = {.lex_state = 514},
  [1986] = {.lex_state = 514},
  [1987] = {.lex_state = 514},
  [1988] = {.lex_state = 514},
  [1989] = {.lex_state = 514, .external_lex_state = 1},
  [1990] = {.lex_state = 514},
  [1991] = {.lex_state = 514},
  [1992] = {.lex_state = 514},
  [1993] = {.lex_state = 514},
  [1994] = {.lex_state = 514},
  [1995] = {.lex_state = 514},
  [1996] = {.lex_state = 514},
  [1997] = {.lex_state = 514, .external_lex_state = 1},
  [1998] = {.lex_state = 514, .external_lex_state = 1},
  [1999] = {.lex_state = 514},
  [2000] = {.lex_state = 514},
  [2001] = {.lex_state = 514},
  [2002] = {.lex_state = 514},
  [2003] = {.lex_state = 514},
  [2004] = {.lex_state = 514},
  [2005] = {.lex_state = 514, .external_lex_state = 1},
  [2006] = {.lex_state = 514},
  [2007] = {.lex_state = 514},
  [2008] = {.lex_state = 514},
  [2009] = {.lex_state = 514, .external_lex_state = 1},
  [2010] = {.lex_state = 514, .external_lex_state = 1},
  [2011] = {.lex_state = 514},
  [2012] = {.lex_state = 514, .external_lex_state = 1},
  [2013] = {.lex_state = 514},
  [2014] = {.lex_state = 514},
  [2015] = {.lex_state = 514},
  [2016] = {.lex_state = 514, .external_lex_state = 1},
  [2017] = {.lex_state = 514, .external_lex_state = 1},
  [2018] = {.lex_state = 514, .external_lex_state = 1},
  [2019] = {.lex_state = 514, .external_lex_state = 1},
  [2020] = {.lex_state = 514},
  [2021] = {.lex_state = 514},
  [2022] = {.lex_state = 514, .external_lex_state = 1},
  [2023] = {.lex_state = 514, .external_lex_state = 1},
  [2024] = {.lex_state = 514},
  [2025] = {.lex_state = 514, .external_lex_state = 1},
  [2026] = {.lex_state = 514},
  [2027] = {.lex_state = 514, .external_lex_state = 1},
  [2028] = {.lex_state = 514},
  [2029] = {.lex_state = 514, .external_lex_state = 1},
  [2030] = {.lex_state = 514, .external_lex_state = 1},
  [2031] = {.lex_state = 514, .external_lex_state = 1},
  [2032] = {.lex_state = 514, .external_lex_state = 1},
  [2033] = {.lex_state = 514, .external_lex_state = 1},
  [2034] = {.lex_state = 514, .external_lex_state = 1},
  [2035] = {.lex_state = 514, .external_lex_state = 1},
  [2036] = {.lex_state = 514},
  [2037] = {.lex_state = 514},
  [2038] = {.lex_state = 131},
  [2039] = {.lex_state = 131},
  [2040] = {.lex_state = 514},
  [2041] = {.lex_state = 514},
  [2042] = {.lex_state = 514},
  [2043] = {.lex_state = 514, .external_lex_state = 1},
  [2044] = {.lex_state = 514, .external_lex_state = 1},
  [2045] = {.lex_state = 514},
  [2046] = {.lex_state = 131},
  [2047] = {.lex_state = 514, .external_lex_state = 1},
  [2048] = {.lex_state = 514, .external_lex_state = 1},
  [2049] = {.lex_state = 514, .external_lex_state = 1},
  [2050] = {.lex_state = 514, .external_lex_state = 1},
  [2051] = {.lex_state = 514, .external_lex_state = 1},
  [2052] = {.lex_state = 514, .external_lex_state = 1},
  [2053] = {.lex_state = 514, .external_lex_state = 1},
  [2054] = {.lex_state = 514, .external_lex_state = 1},
  [2055] = {.lex_state = 514, .external_lex_state = 1},
  [2056] = {.lex_state = 514},
  [2057] = {.lex_state = 514},
  [2058] = {.lex_state = 131},
  [2059] = {.lex_state = 131},
  [2060] = {.lex_state = 514},
  [2061] = {.lex_state = 514, .external_lex_state = 1},
  [2062] = {.lex_state = 514, .external_lex_state = 1},
  [2063] = {.lex_state = 514, .external_lex_state = 1},
  [2064] = {.lex_state = 514},
  [2065] = {.lex_state = 131},
  [2066] = {.lex_state = 514, .external_lex_state = 1},
  [2067] = {.lex_state = 514, .external_lex_state = 1},
  [2068] = {.lex_state = 514, .external_lex_state = 1},
  [2069] = {.lex_state = 514, .external_lex_state = 1},
  [2070] = {.lex_state = 514, .external_lex_state = 1},
  [2071] = {.lex_state = 514, .external_lex_state = 1},
  [2072] = {.lex_state = 514, .external_lex_state = 1},
  [2073] = {.lex_state = 514, .external_lex_state = 1},
  [2074] = {.lex_state = 514, .external_lex_state = 1},
  [2075] = {.lex_state = 514},
  [2076] = {.lex_state = 514, .external_lex_state = 1},
  [2077] = {.lex_state = 131},
  [2078] = {.lex_state = 131},
  [2079] = {.lex_state = 514, .external_lex_state = 1},
  [2080] = {.lex_state = 514},
  [2081] = {.lex_state = 514, .external_lex_state = 1},
  [2082] = {.lex_state = 514},
  [2083] = {.lex_state = 131},
  [2084] = {.lex_state = 514, .external_lex_state = 1},
  [2085] = {.lex_state = 514, .external_lex_state = 1},
  [2086] = {.lex_state = 514, .external_lex_state = 1},
  [2087] = {.lex_state = 514, .external_lex_state = 1},
  [2088] = {.lex_state = 514, .external_lex_state = 1},
  [2089] = {.lex_state = 514, .external_lex_state = 1},
  [2090] = {.lex_state = 514, .external_lex_state = 1},
  [2091] = {.lex_state = 514, .external_lex_state = 1},
  [2092] = {.lex_state = 514, .external_lex_state = 1},
  [2093] = {.lex_state = 514},
  [2094] = {.lex_state = 514},
  [2095] = {.lex_state = 514, .external_lex_state = 1},
  [2096] = {.lex_state = 514},
  [2097] = {.lex_state = 514},
  [2098] = {.lex_state = 514},
  [2099] = {.lex_state = 514},
  [2100] = {.lex_state = 514},
  [2101] = {.lex_state = 514, .external_lex_state = 1},
  [2102] = {.lex_state = 514},
  [2103] = {.lex_state = 514},
  [2104] = {.lex_state = 514},
  [2105] = {.lex_state = 514},
  [2106] = {.lex_state = 514},
  [2107] = {.lex_state = 514},
  [2108] = {.lex_state = 514},
  [2109] = {.lex_state = 514},
  [2110] = {.lex_state = 514},
  [2111] = {.lex_state = 514},
  [2112] = {.lex_state = 514, .external_lex_state = 1},
  [2113] = {.lex_state = 514},
  [2114] = {.lex_state = 514},
  [2115] = {.lex_state = 514},
  [2116] = {.lex_state = 514},
  [2117] = {.lex_state = 514},
  [2118] = {.lex_state = 514, .external_lex_state = 1},
  [2119] = {.lex_state = 514},
  [2120] = {.lex_state = 514},
  [2121] = {.lex_state = 514},
  [2122] = {.lex_state = 514},
  [2123] = {.lex_state = 514},
  [2124] = {.lex_state = 514},
  [2125] = {.lex_state = 514},
  [2126] = {.lex_state = 514},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1),
    [sym_hexadecimal_integer_literal] = ACTIONS(1),
    [sym_real_literal] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [aux_sym_expandable_string_literal_token5] = ACTIONS(1),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1),
    [sym_verbatim_here_string_characters] = ACTIONS(1),
    [sym_simple_name] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_2_GT] = ACTIONS(1),
    [anon_sym_2_GT_GT] = ACTIONS(1),
    [anon_sym_3_GT] = ACTIONS(1),
    [anon_sym_3_GT_GT] = ACTIONS(1),
    [anon_sym_4_GT] = ACTIONS(1),
    [anon_sym_4_GT_GT] = ACTIONS(1),
    [anon_sym_5_GT] = ACTIONS(1),
    [anon_sym_5_GT_GT] = ACTIONS(1),
    [anon_sym_6_GT] = ACTIONS(1),
    [anon_sym_6_GT_GT] = ACTIONS(1),
    [anon_sym_STAR_GT] = ACTIONS(1),
    [anon_sym_STAR_GT_GT] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1),
    [anon_sym_2_GT_AMP1] = ACTIONS(1),
    [anon_sym_3_GT_AMP1] = ACTIONS(1),
    [anon_sym_4_GT_AMP1] = ACTIONS(1),
    [anon_sym_5_GT_AMP1] = ACTIONS(1),
    [anon_sym_6_GT_AMP1] = ACTIONS(1),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1),
    [anon_sym_1_GT_AMP2] = ACTIONS(1),
    [anon_sym_3_GT_AMP2] = ACTIONS(1),
    [anon_sym_4_GT_AMP2] = ACTIONS(1),
    [anon_sym_5_GT_AMP2] = ACTIONS(1),
    [anon_sym_6_GT_AMP2] = ACTIONS(1),
    [aux_sym_comparison_operator_token1] = ACTIONS(1),
    [aux_sym_comparison_operator_token2] = ACTIONS(1),
    [aux_sym_comparison_operator_token3] = ACTIONS(1),
    [aux_sym_comparison_operator_token4] = ACTIONS(1),
    [aux_sym_comparison_operator_token5] = ACTIONS(1),
    [aux_sym_comparison_operator_token6] = ACTIONS(1),
    [aux_sym_comparison_operator_token7] = ACTIONS(1),
    [aux_sym_comparison_operator_token8] = ACTIONS(1),
    [aux_sym_comparison_operator_token9] = ACTIONS(1),
    [aux_sym_comparison_operator_token10] = ACTIONS(1),
    [aux_sym_comparison_operator_token11] = ACTIONS(1),
    [aux_sym_comparison_operator_token12] = ACTIONS(1),
    [aux_sym_comparison_operator_token13] = ACTIONS(1),
    [aux_sym_comparison_operator_token14] = ACTIONS(1),
    [aux_sym_comparison_operator_token15] = ACTIONS(1),
    [aux_sym_comparison_operator_token16] = ACTIONS(1),
    [aux_sym_comparison_operator_token17] = ACTIONS(1),
    [aux_sym_comparison_operator_token18] = ACTIONS(1),
    [aux_sym_comparison_operator_token19] = ACTIONS(1),
    [aux_sym_comparison_operator_token20] = ACTIONS(1),
    [aux_sym_comparison_operator_token21] = ACTIONS(1),
    [aux_sym_comparison_operator_token22] = ACTIONS(1),
    [aux_sym_comparison_operator_token23] = ACTIONS(1),
    [aux_sym_comparison_operator_token24] = ACTIONS(1),
    [aux_sym_comparison_operator_token25] = ACTIONS(1),
    [aux_sym_comparison_operator_token26] = ACTIONS(1),
    [aux_sym_comparison_operator_token27] = ACTIONS(1),
    [aux_sym_comparison_operator_token28] = ACTIONS(1),
    [aux_sym_comparison_operator_token29] = ACTIONS(1),
    [aux_sym_comparison_operator_token30] = ACTIONS(1),
    [aux_sym_comparison_operator_token31] = ACTIONS(1),
    [aux_sym_comparison_operator_token32] = ACTIONS(1),
    [aux_sym_comparison_operator_token33] = ACTIONS(1),
    [aux_sym_comparison_operator_token34] = ACTIONS(1),
    [aux_sym_comparison_operator_token35] = ACTIONS(1),
    [aux_sym_comparison_operator_token36] = ACTIONS(1),
    [aux_sym_comparison_operator_token37] = ACTIONS(1),
    [aux_sym_comparison_operator_token38] = ACTIONS(1),
    [aux_sym_comparison_operator_token39] = ACTIONS(1),
    [aux_sym_comparison_operator_token40] = ACTIONS(1),
    [aux_sym_comparison_operator_token41] = ACTIONS(1),
    [aux_sym_comparison_operator_token42] = ACTIONS(1),
    [aux_sym_comparison_operator_token43] = ACTIONS(1),
    [aux_sym_comparison_operator_token44] = ACTIONS(1),
    [aux_sym_comparison_operator_token45] = ACTIONS(1),
    [aux_sym_comparison_operator_token46] = ACTIONS(1),
    [aux_sym_comparison_operator_token47] = ACTIONS(1),
    [aux_sym_comparison_operator_token48] = ACTIONS(1),
    [aux_sym_comparison_operator_token49] = ACTIONS(1),
    [aux_sym_comparison_operator_token50] = ACTIONS(1),
    [aux_sym_format_operator_token1] = ACTIONS(1),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1),
    [anon_sym_DOLLAR_] = ACTIONS(1),
    [aux_sym_variable_token1] = ACTIONS(1),
    [aux_sym_variable_token2] = ACTIONS(1),
    [sym_braced_variable] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [aux_sym_param_block_token1] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [aux_sym_block_name_token1] = ACTIONS(1),
    [aux_sym_block_name_token2] = ACTIONS(1),
    [aux_sym_block_name_token3] = ACTIONS(1),
    [aux_sym_block_name_token4] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_if_statement_token1] = ACTIONS(1),
    [aux_sym_elseif_clause_token1] = ACTIONS(1),
    [aux_sym_else_clause_token1] = ACTIONS(1),
    [aux_sym_switch_statement_token1] = ACTIONS(1),
    [aux_sym_switch_parameter_token1] = ACTIONS(1),
    [aux_sym_switch_parameter_token2] = ACTIONS(1),
    [aux_sym_switch_parameter_token3] = ACTIONS(1),
    [aux_sym_switch_parameter_token4] = ACTIONS(1),
    [aux_sym_switch_parameter_token5] = ACTIONS(1),
    [aux_sym_switch_condition_token1] = ACTIONS(1),
    [aux_sym_foreach_statement_token1] = ACTIONS(1),
    [aux_sym_foreach_statement_token2] = ACTIONS(1),
    [aux_sym_for_statement_token1] = ACTIONS(1),
    [aux_sym_while_statement_token1] = ACTIONS(1),
    [aux_sym_do_statement_token1] = ACTIONS(1),
    [aux_sym_do_statement_token2] = ACTIONS(1),
    [aux_sym_function_statement_token1] = ACTIONS(1),
    [aux_sym_function_statement_token2] = ACTIONS(1),
    [aux_sym_function_statement_token3] = ACTIONS(1),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1),
    [aux_sym_trap_statement_token1] = ACTIONS(1),
    [aux_sym_try_statement_token1] = ACTIONS(1),
    [aux_sym_catch_clause_token1] = ACTIONS(1),
    [aux_sym_finally_clause_token1] = ACTIONS(1),
    [aux_sym_data_statement_token1] = ACTIONS(1),
    [aux_sym_data_commands_allowed_token1] = ACTIONS(1),
    [aux_sym_parallel_statement_token1] = ACTIONS(1),
    [aux_sym_sequence_statement_token1] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_DOLLAR2] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [anon_sym_SQUOTE_SQUOTE] = ACTIONS(1),
    [sym_stop_parsing] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1),
    [aux_sym_class_attribute_token1] = ACTIONS(1),
    [aux_sym_class_attribute_token2] = ACTIONS(1),
    [aux_sym_class_statement_token1] = ACTIONS(1),
    [aux_sym_enum_statement_token1] = ACTIONS(1),
    [aux_sym_logical_expression_token1] = ACTIONS(1),
    [aux_sym_logical_expression_token2] = ACTIONS(1),
    [aux_sym_logical_expression_token3] = ACTIONS(1),
    [aux_sym_bitwise_expression_token1] = ACTIONS(1),
    [aux_sym_bitwise_expression_token2] = ACTIONS(1),
    [aux_sym_bitwise_expression_token3] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1),
    [anon_sym_AT_LPAREN] = ACTIONS(1),
    [anon_sym_AT_LBRACE] = ACTIONS(1),
    [anon_sym_DOT2] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [sym__statement_terminator] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_program] = STATE(1750),
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_param_block] = STATE(75),
    [sym_statement_list] = STATE(1747),
    [sym__statement] = STATE(76),
    [sym_empty_statement] = STATE(76),
    [sym_if_statement] = STATE(76),
    [sym__labeled_statement] = STATE(76),
    [sym_switch_statement] = STATE(76),
    [sym_foreach_statement] = STATE(76),
    [sym_for_statement] = STATE(76),
    [sym_while_statement] = STATE(76),
    [sym_do_statement] = STATE(76),
    [sym_function_statement] = STATE(76),
    [sym_flow_control_statement] = STATE(1816),
    [sym_trap_statement] = STATE(76),
    [sym_try_statement] = STATE(76),
    [sym_data_statement] = STATE(76),
    [sym_inlinescript_statement] = STATE(76),
    [sym_parallel_statement] = STATE(76),
    [sym_sequence_statement] = STATE(76),
    [sym_pipeline] = STATE(1816),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(76),
    [sym_enum_statement] = STATE(76),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(1829),
    [sym_attribute] = STATE(1252),
    [aux_sym_statement_list_repeat1] = STATE(76),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(23),
    [aux_sym_param_block_token1] = ACTIONS(25),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(31),
    [aux_sym_switch_statement_token1] = ACTIONS(33),
    [aux_sym_foreach_statement_token1] = ACTIONS(35),
    [aux_sym_for_statement_token1] = ACTIONS(37),
    [aux_sym_while_statement_token1] = ACTIONS(39),
    [aux_sym_do_statement_token1] = ACTIONS(41),
    [aux_sym_function_statement_token1] = ACTIONS(43),
    [aux_sym_function_statement_token2] = ACTIONS(43),
    [aux_sym_function_statement_token3] = ACTIONS(43),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(49),
    [aux_sym_trap_statement_token1] = ACTIONS(51),
    [aux_sym_try_statement_token1] = ACTIONS(53),
    [aux_sym_data_statement_token1] = ACTIONS(55),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(57),
    [aux_sym_parallel_statement_token1] = ACTIONS(59),
    [aux_sym_sequence_statement_token1] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(65),
    [aux_sym_enum_statement_token1] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(2)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_unary_expression] = STATE(182),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(83),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(91),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_BANG_EQ] = ACTIONS(93),
    [anon_sym_PLUS_EQ] = ACTIONS(93),
    [anon_sym_STAR_EQ] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_PERCENT_EQ] = ACTIONS(93),
    [anon_sym_DASH_EQ] = ACTIONS(93),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(97),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(97),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [aux_sym_param_block_token1] = ACTIONS(101),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(95),
    [anon_sym_STAR] = ACTIONS(95),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(97),
    [anon_sym_PLUS_PLUS] = ACTIONS(107),
    [anon_sym_DASH_DASH] = ACTIONS(109),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(3)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_unary_expression] = STATE(182),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(83),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(91),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_BANG_EQ] = ACTIONS(93),
    [anon_sym_PLUS_EQ] = ACTIONS(93),
    [anon_sym_STAR_EQ] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_PERCENT_EQ] = ACTIONS(93),
    [anon_sym_DASH_EQ] = ACTIONS(93),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(97),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(97),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(95),
    [anon_sym_STAR] = ACTIONS(95),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(97),
    [anon_sym_PLUS_PLUS] = ACTIONS(107),
    [anon_sym_DASH_DASH] = ACTIONS(109),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(4)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_unary_expression] = STATE(174),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(121),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_BANG_EQ] = ACTIONS(93),
    [anon_sym_PLUS_EQ] = ACTIONS(93),
    [anon_sym_STAR_EQ] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_PERCENT_EQ] = ACTIONS(93),
    [anon_sym_DASH_EQ] = ACTIONS(93),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(131),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(131),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(95),
    [anon_sym_STAR] = ACTIONS(95),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(131),
    [anon_sym_PLUS_PLUS] = ACTIONS(143),
    [anon_sym_DASH_DASH] = ACTIONS(145),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(5)] = {
    [sym__literal] = STATE(150),
    [sym_integer_literal] = STATE(150),
    [sym_string_literal] = STATE(150),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(150),
    [sym_unary_expression] = STATE(134),
    [sym_expression_with_unary_operator] = STATE(135),
    [sym_pre_increment_expression] = STATE(130),
    [sym_pre_decrement_expression] = STATE(130),
    [sym_cast_expression] = STATE(130),
    [sym__primary_expression] = STATE(150),
    [sym__value] = STATE(150),
    [sym_parenthesized_expression] = STATE(150),
    [sym_sub_expression] = STATE(150),
    [sym_array_expression] = STATE(150),
    [sym_script_block_expression] = STATE(150),
    [sym_hash_literal_expression] = STATE(150),
    [sym_post_increment_expression] = STATE(150),
    [sym_post_decrement_expression] = STATE(150),
    [sym_member_access] = STATE(150),
    [sym_element_access] = STATE(150),
    [sym_invokation_expression] = STATE(150),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(5),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(153),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(91),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_BANG_EQ] = ACTIONS(93),
    [anon_sym_PLUS_EQ] = ACTIONS(93),
    [anon_sym_STAR_EQ] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_PERCENT_EQ] = ACTIONS(93),
    [anon_sym_DASH_EQ] = ACTIONS(93),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(155),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(155),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(157),
    [anon_sym_DASH] = ACTIONS(157),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(95),
    [anon_sym_STAR] = ACTIONS(95),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(157),
    [anon_sym_BANG] = ACTIONS(157),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(155),
    [anon_sym_PLUS_PLUS] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(161),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(6)] = {
    [sym__literal] = STATE(151),
    [sym_integer_literal] = STATE(151),
    [sym_string_literal] = STATE(151),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(151),
    [sym_unary_expression] = STATE(110),
    [sym_expression_with_unary_operator] = STATE(104),
    [sym_pre_increment_expression] = STATE(105),
    [sym_pre_decrement_expression] = STATE(105),
    [sym_cast_expression] = STATE(105),
    [sym__primary_expression] = STATE(151),
    [sym__value] = STATE(151),
    [sym_parenthesized_expression] = STATE(151),
    [sym_sub_expression] = STATE(151),
    [sym_array_expression] = STATE(151),
    [sym_script_block_expression] = STATE(151),
    [sym_hash_literal_expression] = STATE(151),
    [sym_post_increment_expression] = STATE(151),
    [sym_post_decrement_expression] = STATE(151),
    [sym_member_access] = STATE(151),
    [sym_element_access] = STATE(151),
    [sym_invokation_expression] = STATE(151),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(6),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(163),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_BANG_EQ] = ACTIONS(93),
    [anon_sym_PLUS_EQ] = ACTIONS(93),
    [anon_sym_STAR_EQ] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_PERCENT_EQ] = ACTIONS(93),
    [anon_sym_DASH_EQ] = ACTIONS(93),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(165),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(165),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(167),
    [anon_sym_DASH] = ACTIONS(167),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(95),
    [anon_sym_STAR] = ACTIONS(95),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(167),
    [anon_sym_BANG] = ACTIONS(167),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(171),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(7)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1864),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(8)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1765),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(213),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(9)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1851),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(215),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(10)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1885),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(217),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(11)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1990),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(219),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(12)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(2113),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(221),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(13)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1736),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(223),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(14)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(2026),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(225),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(15)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1913),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(227),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(16)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1932),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(229),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(17)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(2117),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(231),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(18)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1719),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(233),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(19)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1700),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(235),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(20)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1710),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(237),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(21)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1698),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(22)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1771),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(23)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1980),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(24)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1761),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(25)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1787),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(26)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1809),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(27)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1831),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(28)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1840),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(29)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1848),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(30)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1855),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(31)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1867),
    [sym_param_block] = STATE(1815),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2060),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(175),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(211),
  },
  [STATE(32)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1803),
    [sym_param_block] = STATE(31),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(33)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1802),
    [sym_param_block] = STATE(26),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(34)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1955),
    [sym_param_block] = STATE(23),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(35)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1824),
    [sym_param_block] = STATE(27),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(36)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1739),
    [sym_param_block] = STATE(22),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(37)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1838),
    [sym_param_block] = STATE(28),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(38)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1753),
    [sym_param_block] = STATE(24),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(39)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1846),
    [sym_param_block] = STATE(29),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(40)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1728),
    [sym_param_block] = STATE(7),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(41)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1853),
    [sym_param_block] = STATE(30),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(42)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1780),
    [sym_param_block] = STATE(25),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(43)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block] = STATE(1859),
    [sym_param_block] = STATE(21),
    [sym_script_block_body] = STATE(1880),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(2),
    [sym_attribute_list] = STATE(2040),
    [sym_attribute] = STATE(1252),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [aux_sym_attribute_list_repeat1] = STATE(1252),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [aux_sym_param_block_token1] = ACTIONS(239),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(44)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block_body] = STATE(2024),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_script_block_repeat1] = STATE(302),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(245),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(45)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_script_block_body] = STATE(2116),
    [sym_named_block_list] = STATE(2003),
    [sym_named_block] = STATE(1297),
    [sym_block_name] = STATE(1582),
    [sym_statement_list] = STATE(2008),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_script_block_repeat1] = STATE(44),
    [aux_sym_named_block_list_repeat1] = STATE(1297),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [aux_sym_block_name_token1] = ACTIONS(177),
    [aux_sym_block_name_token2] = ACTIONS(177),
    [aux_sym_block_name_token3] = ACTIONS(177),
    [aux_sym_block_name_token4] = ACTIONS(177),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(247),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(46)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1810),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(249),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(47)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1826),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(251),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(48)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1827),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(253),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(49)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1832),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(255),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(50)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1839),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(257),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(51)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1847),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(259),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(52)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1854),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(261),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(53)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1872),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(263),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(54)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1749),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(265),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(55)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(61),
    [sym_empty_statement] = STATE(61),
    [sym_if_statement] = STATE(61),
    [sym__labeled_statement] = STATE(61),
    [sym_switch_statement] = STATE(61),
    [sym_foreach_statement] = STATE(61),
    [sym_for_statement] = STATE(61),
    [sym_while_statement] = STATE(61),
    [sym_do_statement] = STATE(61),
    [sym_function_statement] = STATE(61),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(61),
    [sym_try_statement] = STATE(61),
    [sym_data_statement] = STATE(61),
    [sym_inlinescript_statement] = STATE(61),
    [sym_parallel_statement] = STATE(61),
    [sym_sequence_statement] = STATE(61),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(61),
    [sym_enum_statement] = STATE(61),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(61),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(267),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(267),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(56)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(2028),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(57)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1759),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(271),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(58)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1760),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(273),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(59)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1908),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(275),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(60)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1893),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(277),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(61)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(61),
    [sym_empty_statement] = STATE(61),
    [sym_if_statement] = STATE(61),
    [sym__labeled_statement] = STATE(61),
    [sym_switch_statement] = STATE(61),
    [sym_foreach_statement] = STATE(61),
    [sym_for_statement] = STATE(61),
    [sym_while_statement] = STATE(61),
    [sym_do_statement] = STATE(61),
    [sym_function_statement] = STATE(61),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(61),
    [sym_try_statement] = STATE(61),
    [sym_data_statement] = STATE(61),
    [sym_inlinescript_statement] = STATE(61),
    [sym_parallel_statement] = STATE(61),
    [sym_sequence_statement] = STATE(61),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(61),
    [sym_enum_statement] = STATE(61),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(61),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(279),
    [sym_hexadecimal_integer_literal] = ACTIONS(279),
    [sym_real_literal] = ACTIONS(282),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(285),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(288),
    [sym_verbatim_string_characters] = ACTIONS(291),
    [sym_verbatim_here_string_characters] = ACTIONS(291),
    [anon_sym_DOT] = ACTIONS(294),
    [anon_sym_LBRACK] = ACTIONS(297),
    [aux_sym_comparison_operator_token37] = ACTIONS(300),
    [aux_sym_comparison_operator_token50] = ACTIONS(300),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(303),
    [anon_sym_DOLLAR_CARET] = ACTIONS(303),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(303),
    [anon_sym_DOLLAR_] = ACTIONS(303),
    [aux_sym_variable_token1] = ACTIONS(303),
    [aux_sym_variable_token2] = ACTIONS(303),
    [sym_braced_variable] = ACTIONS(303),
    [anon_sym_SEMI] = ACTIONS(306),
    [anon_sym_LPAREN] = ACTIONS(309),
    [anon_sym_RPAREN] = ACTIONS(312),
    [anon_sym_COMMA] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(314),
    [anon_sym_RBRACE] = ACTIONS(312),
    [aux_sym_if_statement_token1] = ACTIONS(317),
    [aux_sym_switch_statement_token1] = ACTIONS(320),
    [aux_sym_foreach_statement_token1] = ACTIONS(323),
    [aux_sym_for_statement_token1] = ACTIONS(326),
    [aux_sym_while_statement_token1] = ACTIONS(329),
    [aux_sym_do_statement_token1] = ACTIONS(332),
    [aux_sym_function_statement_token1] = ACTIONS(335),
    [aux_sym_function_statement_token2] = ACTIONS(335),
    [aux_sym_function_statement_token3] = ACTIONS(335),
    [aux_sym_flow_control_statement_token1] = ACTIONS(338),
    [aux_sym_flow_control_statement_token2] = ACTIONS(338),
    [aux_sym_flow_control_statement_token3] = ACTIONS(341),
    [aux_sym_flow_control_statement_token4] = ACTIONS(341),
    [aux_sym_flow_control_statement_token5] = ACTIONS(341),
    [sym_label] = ACTIONS(344),
    [aux_sym_trap_statement_token1] = ACTIONS(347),
    [aux_sym_try_statement_token1] = ACTIONS(350),
    [aux_sym_data_statement_token1] = ACTIONS(353),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(356),
    [aux_sym_parallel_statement_token1] = ACTIONS(359),
    [aux_sym_sequence_statement_token1] = ACTIONS(362),
    [anon_sym_AMP] = ACTIONS(294),
    [aux_sym_command_name_token1] = ACTIONS(365),
    [aux_sym_command_name_token2] = ACTIONS(365),
    [aux_sym_command_name_token3] = ACTIONS(365),
    [aux_sym_command_name_token4] = ACTIONS(365),
    [aux_sym_command_name_token5] = ACTIONS(365),
    [aux_sym_command_name_token6] = ACTIONS(365),
    [aux_sym_command_name_token7] = ACTIONS(365),
    [aux_sym_command_name_token8] = ACTIONS(365),
    [aux_sym_command_name_token9] = ACTIONS(365),
    [aux_sym_command_name_token10] = ACTIONS(365),
    [aux_sym_command_name_token11] = ACTIONS(365),
    [aux_sym_command_name_token12] = ACTIONS(365),
    [aux_sym_command_name_token13] = ACTIONS(365),
    [aux_sym_command_name_token14] = ACTIONS(365),
    [aux_sym_command_name_token15] = ACTIONS(365),
    [aux_sym_command_name_token16] = ACTIONS(365),
    [aux_sym_command_name_token17] = ACTIONS(365),
    [aux_sym_command_name_token18] = ACTIONS(365),
    [aux_sym_class_statement_token1] = ACTIONS(368),
    [aux_sym_enum_statement_token1] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(300),
    [anon_sym_DASH] = ACTIONS(300),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(300),
    [anon_sym_BANG] = ACTIONS(300),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(300),
    [anon_sym_PLUS_PLUS] = ACTIONS(374),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(380),
    [anon_sym_AT_LPAREN] = ACTIONS(383),
    [anon_sym_AT_LBRACE] = ACTIONS(386),
  },
  [STATE(62)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1966),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(389),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(63)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1968),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(391),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(64)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(2014),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(393),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(65)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1818),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(395),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(66)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1819),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(397),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(67)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1756),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(399),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(68)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1757),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(401),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(69)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1762),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(403),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(70)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1782),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(405),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(71)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1783),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(407),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(72)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1788),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(409),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(73)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1804),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(74)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1805),
    [sym__statement] = STATE(55),
    [sym_empty_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym__labeled_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_foreach_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_function_statement] = STATE(55),
    [sym_flow_control_statement] = STATE(1778),
    [sym_trap_statement] = STATE(55),
    [sym_try_statement] = STATE(55),
    [sym_data_statement] = STATE(55),
    [sym_inlinescript_statement] = STATE(55),
    [sym_parallel_statement] = STATE(55),
    [sym_sequence_statement] = STATE(55),
    [sym_pipeline] = STATE(1778),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(55),
    [sym_enum_statement] = STATE(55),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(55),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(413),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(179),
    [aux_sym_switch_statement_token1] = ACTIONS(181),
    [aux_sym_foreach_statement_token1] = ACTIONS(183),
    [aux_sym_for_statement_token1] = ACTIONS(185),
    [aux_sym_while_statement_token1] = ACTIONS(187),
    [aux_sym_do_statement_token1] = ACTIONS(189),
    [aux_sym_function_statement_token1] = ACTIONS(191),
    [aux_sym_function_statement_token2] = ACTIONS(191),
    [aux_sym_function_statement_token3] = ACTIONS(191),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(193),
    [aux_sym_trap_statement_token1] = ACTIONS(195),
    [aux_sym_try_statement_token1] = ACTIONS(197),
    [aux_sym_data_statement_token1] = ACTIONS(199),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(201),
    [aux_sym_parallel_statement_token1] = ACTIONS(203),
    [aux_sym_sequence_statement_token1] = ACTIONS(205),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(207),
    [aux_sym_enum_statement_token1] = ACTIONS(209),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(75)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_statement_list] = STATE(1930),
    [sym__statement] = STATE(76),
    [sym_empty_statement] = STATE(76),
    [sym_if_statement] = STATE(76),
    [sym__labeled_statement] = STATE(76),
    [sym_switch_statement] = STATE(76),
    [sym_foreach_statement] = STATE(76),
    [sym_for_statement] = STATE(76),
    [sym_while_statement] = STATE(76),
    [sym_do_statement] = STATE(76),
    [sym_function_statement] = STATE(76),
    [sym_flow_control_statement] = STATE(1816),
    [sym_trap_statement] = STATE(76),
    [sym_try_statement] = STATE(76),
    [sym_data_statement] = STATE(76),
    [sym_inlinescript_statement] = STATE(76),
    [sym_parallel_statement] = STATE(76),
    [sym_sequence_statement] = STATE(76),
    [sym_pipeline] = STATE(1816),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(76),
    [sym_enum_statement] = STATE(76),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(76),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(23),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(31),
    [aux_sym_switch_statement_token1] = ACTIONS(33),
    [aux_sym_foreach_statement_token1] = ACTIONS(35),
    [aux_sym_for_statement_token1] = ACTIONS(37),
    [aux_sym_while_statement_token1] = ACTIONS(39),
    [aux_sym_do_statement_token1] = ACTIONS(41),
    [aux_sym_function_statement_token1] = ACTIONS(43),
    [aux_sym_function_statement_token2] = ACTIONS(43),
    [aux_sym_function_statement_token3] = ACTIONS(43),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(49),
    [aux_sym_trap_statement_token1] = ACTIONS(51),
    [aux_sym_try_statement_token1] = ACTIONS(53),
    [aux_sym_data_statement_token1] = ACTIONS(55),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(57),
    [aux_sym_parallel_statement_token1] = ACTIONS(59),
    [aux_sym_sequence_statement_token1] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(65),
    [aux_sym_enum_statement_token1] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(76)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(77),
    [sym_empty_statement] = STATE(77),
    [sym_if_statement] = STATE(77),
    [sym__labeled_statement] = STATE(77),
    [sym_switch_statement] = STATE(77),
    [sym_foreach_statement] = STATE(77),
    [sym_for_statement] = STATE(77),
    [sym_while_statement] = STATE(77),
    [sym_do_statement] = STATE(77),
    [sym_function_statement] = STATE(77),
    [sym_flow_control_statement] = STATE(1816),
    [sym_trap_statement] = STATE(77),
    [sym_try_statement] = STATE(77),
    [sym_data_statement] = STATE(77),
    [sym_inlinescript_statement] = STATE(77),
    [sym_parallel_statement] = STATE(77),
    [sym_sequence_statement] = STATE(77),
    [sym_pipeline] = STATE(1816),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(77),
    [sym_enum_statement] = STATE(77),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(77),
    [ts_builtin_sym_end] = ACTIONS(415),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(23),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(31),
    [aux_sym_switch_statement_token1] = ACTIONS(33),
    [aux_sym_foreach_statement_token1] = ACTIONS(35),
    [aux_sym_for_statement_token1] = ACTIONS(37),
    [aux_sym_while_statement_token1] = ACTIONS(39),
    [aux_sym_do_statement_token1] = ACTIONS(41),
    [aux_sym_function_statement_token1] = ACTIONS(43),
    [aux_sym_function_statement_token2] = ACTIONS(43),
    [aux_sym_function_statement_token3] = ACTIONS(43),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(49),
    [aux_sym_trap_statement_token1] = ACTIONS(51),
    [aux_sym_try_statement_token1] = ACTIONS(53),
    [aux_sym_data_statement_token1] = ACTIONS(55),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(57),
    [aux_sym_parallel_statement_token1] = ACTIONS(59),
    [aux_sym_sequence_statement_token1] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(65),
    [aux_sym_enum_statement_token1] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(77)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(77),
    [sym_empty_statement] = STATE(77),
    [sym_if_statement] = STATE(77),
    [sym__labeled_statement] = STATE(77),
    [sym_switch_statement] = STATE(77),
    [sym_foreach_statement] = STATE(77),
    [sym_for_statement] = STATE(77),
    [sym_while_statement] = STATE(77),
    [sym_do_statement] = STATE(77),
    [sym_function_statement] = STATE(77),
    [sym_flow_control_statement] = STATE(1816),
    [sym_trap_statement] = STATE(77),
    [sym_try_statement] = STATE(77),
    [sym_data_statement] = STATE(77),
    [sym_inlinescript_statement] = STATE(77),
    [sym_parallel_statement] = STATE(77),
    [sym_sequence_statement] = STATE(77),
    [sym_pipeline] = STATE(1816),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(77),
    [sym_enum_statement] = STATE(77),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [aux_sym_statement_list_repeat1] = STATE(77),
    [ts_builtin_sym_end] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(279),
    [sym_hexadecimal_integer_literal] = ACTIONS(279),
    [sym_real_literal] = ACTIONS(282),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(285),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(288),
    [sym_verbatim_string_characters] = ACTIONS(291),
    [sym_verbatim_here_string_characters] = ACTIONS(291),
    [anon_sym_DOT] = ACTIONS(294),
    [anon_sym_LBRACK] = ACTIONS(297),
    [aux_sym_comparison_operator_token37] = ACTIONS(300),
    [aux_sym_comparison_operator_token50] = ACTIONS(300),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(303),
    [anon_sym_DOLLAR_CARET] = ACTIONS(303),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(303),
    [anon_sym_DOLLAR_] = ACTIONS(303),
    [aux_sym_variable_token1] = ACTIONS(303),
    [aux_sym_variable_token2] = ACTIONS(303),
    [sym_braced_variable] = ACTIONS(303),
    [anon_sym_SEMI] = ACTIONS(419),
    [anon_sym_LPAREN] = ACTIONS(309),
    [anon_sym_COMMA] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(314),
    [aux_sym_if_statement_token1] = ACTIONS(422),
    [aux_sym_switch_statement_token1] = ACTIONS(425),
    [aux_sym_foreach_statement_token1] = ACTIONS(428),
    [aux_sym_for_statement_token1] = ACTIONS(431),
    [aux_sym_while_statement_token1] = ACTIONS(434),
    [aux_sym_do_statement_token1] = ACTIONS(437),
    [aux_sym_function_statement_token1] = ACTIONS(440),
    [aux_sym_function_statement_token2] = ACTIONS(440),
    [aux_sym_function_statement_token3] = ACTIONS(440),
    [aux_sym_flow_control_statement_token1] = ACTIONS(338),
    [aux_sym_flow_control_statement_token2] = ACTIONS(338),
    [aux_sym_flow_control_statement_token3] = ACTIONS(341),
    [aux_sym_flow_control_statement_token4] = ACTIONS(341),
    [aux_sym_flow_control_statement_token5] = ACTIONS(341),
    [sym_label] = ACTIONS(443),
    [aux_sym_trap_statement_token1] = ACTIONS(446),
    [aux_sym_try_statement_token1] = ACTIONS(449),
    [aux_sym_data_statement_token1] = ACTIONS(452),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(455),
    [aux_sym_parallel_statement_token1] = ACTIONS(458),
    [aux_sym_sequence_statement_token1] = ACTIONS(461),
    [anon_sym_AMP] = ACTIONS(294),
    [aux_sym_command_name_token1] = ACTIONS(365),
    [aux_sym_command_name_token2] = ACTIONS(365),
    [aux_sym_command_name_token3] = ACTIONS(365),
    [aux_sym_command_name_token4] = ACTIONS(365),
    [aux_sym_command_name_token5] = ACTIONS(365),
    [aux_sym_command_name_token6] = ACTIONS(365),
    [aux_sym_command_name_token7] = ACTIONS(365),
    [aux_sym_command_name_token8] = ACTIONS(365),
    [aux_sym_command_name_token9] = ACTIONS(365),
    [aux_sym_command_name_token10] = ACTIONS(365),
    [aux_sym_command_name_token11] = ACTIONS(365),
    [aux_sym_command_name_token12] = ACTIONS(365),
    [aux_sym_command_name_token13] = ACTIONS(365),
    [aux_sym_command_name_token14] = ACTIONS(365),
    [aux_sym_command_name_token15] = ACTIONS(365),
    [aux_sym_command_name_token16] = ACTIONS(365),
    [aux_sym_command_name_token17] = ACTIONS(365),
    [aux_sym_command_name_token18] = ACTIONS(365),
    [aux_sym_class_statement_token1] = ACTIONS(464),
    [aux_sym_enum_statement_token1] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(300),
    [anon_sym_DASH] = ACTIONS(300),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(300),
    [anon_sym_BANG] = ACTIONS(300),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(300),
    [anon_sym_PLUS_PLUS] = ACTIONS(374),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(380),
    [anon_sym_AT_LPAREN] = ACTIONS(383),
    [anon_sym_AT_LBRACE] = ACTIONS(386),
  },
  [STATE(78)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(1960),
    [sym_empty_statement] = STATE(1960),
    [sym_if_statement] = STATE(1960),
    [sym__labeled_statement] = STATE(1960),
    [sym_switch_statement] = STATE(1960),
    [sym_foreach_statement] = STATE(1960),
    [sym_for_statement] = STATE(1960),
    [sym_while_statement] = STATE(1960),
    [sym_do_statement] = STATE(1960),
    [sym_function_statement] = STATE(1960),
    [sym_flow_control_statement] = STATE(1702),
    [sym_trap_statement] = STATE(1960),
    [sym_try_statement] = STATE(1960),
    [sym_data_statement] = STATE(1960),
    [sym_inlinescript_statement] = STATE(1960),
    [sym_parallel_statement] = STATE(1960),
    [sym_sequence_statement] = STATE(1960),
    [sym_pipeline] = STATE(1702),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(1960),
    [sym_enum_statement] = STATE(1960),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(470),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(472),
    [aux_sym_switch_statement_token1] = ACTIONS(474),
    [aux_sym_foreach_statement_token1] = ACTIONS(476),
    [aux_sym_for_statement_token1] = ACTIONS(478),
    [aux_sym_while_statement_token1] = ACTIONS(480),
    [aux_sym_do_statement_token1] = ACTIONS(482),
    [aux_sym_function_statement_token1] = ACTIONS(484),
    [aux_sym_function_statement_token2] = ACTIONS(484),
    [aux_sym_function_statement_token3] = ACTIONS(484),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(486),
    [aux_sym_trap_statement_token1] = ACTIONS(488),
    [aux_sym_try_statement_token1] = ACTIONS(490),
    [aux_sym_data_statement_token1] = ACTIONS(492),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(494),
    [aux_sym_parallel_statement_token1] = ACTIONS(496),
    [aux_sym_sequence_statement_token1] = ACTIONS(498),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(500),
    [aux_sym_enum_statement_token1] = ACTIONS(502),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(79)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(1711),
    [sym_empty_statement] = STATE(1711),
    [sym_if_statement] = STATE(1711),
    [sym__labeled_statement] = STATE(1711),
    [sym_switch_statement] = STATE(1711),
    [sym_foreach_statement] = STATE(1711),
    [sym_for_statement] = STATE(1711),
    [sym_while_statement] = STATE(1711),
    [sym_do_statement] = STATE(1711),
    [sym_function_statement] = STATE(1711),
    [sym_flow_control_statement] = STATE(1702),
    [sym_trap_statement] = STATE(1711),
    [sym_try_statement] = STATE(1711),
    [sym_data_statement] = STATE(1711),
    [sym_inlinescript_statement] = STATE(1711),
    [sym_parallel_statement] = STATE(1711),
    [sym_sequence_statement] = STATE(1711),
    [sym_pipeline] = STATE(1702),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(1711),
    [sym_enum_statement] = STATE(1711),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(470),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(472),
    [aux_sym_switch_statement_token1] = ACTIONS(474),
    [aux_sym_foreach_statement_token1] = ACTIONS(476),
    [aux_sym_for_statement_token1] = ACTIONS(478),
    [aux_sym_while_statement_token1] = ACTIONS(480),
    [aux_sym_do_statement_token1] = ACTIONS(482),
    [aux_sym_function_statement_token1] = ACTIONS(484),
    [aux_sym_function_statement_token2] = ACTIONS(484),
    [aux_sym_function_statement_token3] = ACTIONS(484),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(486),
    [aux_sym_trap_statement_token1] = ACTIONS(488),
    [aux_sym_try_statement_token1] = ACTIONS(490),
    [aux_sym_data_statement_token1] = ACTIONS(492),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(494),
    [aux_sym_parallel_statement_token1] = ACTIONS(496),
    [aux_sym_sequence_statement_token1] = ACTIONS(498),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(500),
    [aux_sym_enum_statement_token1] = ACTIONS(502),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(80)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym__statement] = STATE(1992),
    [sym_empty_statement] = STATE(1992),
    [sym_if_statement] = STATE(1992),
    [sym__labeled_statement] = STATE(1992),
    [sym_switch_statement] = STATE(1992),
    [sym_foreach_statement] = STATE(1992),
    [sym_for_statement] = STATE(1992),
    [sym_while_statement] = STATE(1992),
    [sym_do_statement] = STATE(1992),
    [sym_function_statement] = STATE(1992),
    [sym_flow_control_statement] = STATE(1944),
    [sym_trap_statement] = STATE(1992),
    [sym_try_statement] = STATE(1992),
    [sym_data_statement] = STATE(1992),
    [sym_inlinescript_statement] = STATE(1992),
    [sym_parallel_statement] = STATE(1992),
    [sym_sequence_statement] = STATE(1992),
    [sym_pipeline] = STATE(1944),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym_class_statement] = STATE(1992),
    [sym_enum_statement] = STATE(1992),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(504),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_if_statement_token1] = ACTIONS(506),
    [aux_sym_switch_statement_token1] = ACTIONS(508),
    [aux_sym_foreach_statement_token1] = ACTIONS(510),
    [aux_sym_for_statement_token1] = ACTIONS(512),
    [aux_sym_while_statement_token1] = ACTIONS(514),
    [aux_sym_do_statement_token1] = ACTIONS(516),
    [aux_sym_function_statement_token1] = ACTIONS(518),
    [aux_sym_function_statement_token2] = ACTIONS(518),
    [aux_sym_function_statement_token3] = ACTIONS(518),
    [aux_sym_flow_control_statement_token1] = ACTIONS(45),
    [aux_sym_flow_control_statement_token2] = ACTIONS(45),
    [aux_sym_flow_control_statement_token3] = ACTIONS(47),
    [aux_sym_flow_control_statement_token4] = ACTIONS(47),
    [aux_sym_flow_control_statement_token5] = ACTIONS(47),
    [sym_label] = ACTIONS(520),
    [aux_sym_trap_statement_token1] = ACTIONS(522),
    [aux_sym_try_statement_token1] = ACTIONS(524),
    [aux_sym_data_statement_token1] = ACTIONS(526),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(528),
    [aux_sym_parallel_statement_token1] = ACTIONS(530),
    [aux_sym_sequence_statement_token1] = ACTIONS(532),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [aux_sym_class_statement_token1] = ACTIONS(534),
    [aux_sym_enum_statement_token1] = ACTIONS(536),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(81)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [anon_sym_EQ] = ACTIONS(540),
    [anon_sym_BANG_EQ] = ACTIONS(540),
    [anon_sym_PLUS_EQ] = ACTIONS(540),
    [anon_sym_STAR_EQ] = ACTIONS(540),
    [anon_sym_SLASH_EQ] = ACTIONS(540),
    [anon_sym_PERCENT_EQ] = ACTIONS(540),
    [anon_sym_DASH_EQ] = ACTIONS(540),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(540),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(540),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(540),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(540),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(540),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(540),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(540),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(540),
    [anon_sym_2_GT_AMP1] = ACTIONS(540),
    [anon_sym_3_GT_AMP1] = ACTIONS(540),
    [anon_sym_4_GT_AMP1] = ACTIONS(540),
    [anon_sym_5_GT_AMP1] = ACTIONS(540),
    [anon_sym_6_GT_AMP1] = ACTIONS(540),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(540),
    [anon_sym_1_GT_AMP2] = ACTIONS(540),
    [anon_sym_3_GT_AMP2] = ACTIONS(540),
    [anon_sym_4_GT_AMP2] = ACTIONS(540),
    [anon_sym_5_GT_AMP2] = ACTIONS(540),
    [anon_sym_6_GT_AMP2] = ACTIONS(540),
    [aux_sym_comparison_operator_token1] = ACTIONS(540),
    [aux_sym_comparison_operator_token2] = ACTIONS(540),
    [aux_sym_comparison_operator_token3] = ACTIONS(540),
    [aux_sym_comparison_operator_token4] = ACTIONS(540),
    [aux_sym_comparison_operator_token5] = ACTIONS(540),
    [aux_sym_comparison_operator_token6] = ACTIONS(540),
    [aux_sym_comparison_operator_token7] = ACTIONS(540),
    [aux_sym_comparison_operator_token8] = ACTIONS(540),
    [aux_sym_comparison_operator_token9] = ACTIONS(540),
    [aux_sym_comparison_operator_token10] = ACTIONS(540),
    [aux_sym_comparison_operator_token11] = ACTIONS(540),
    [aux_sym_comparison_operator_token12] = ACTIONS(540),
    [aux_sym_comparison_operator_token13] = ACTIONS(540),
    [aux_sym_comparison_operator_token14] = ACTIONS(540),
    [aux_sym_comparison_operator_token15] = ACTIONS(540),
    [aux_sym_comparison_operator_token16] = ACTIONS(540),
    [aux_sym_comparison_operator_token17] = ACTIONS(540),
    [aux_sym_comparison_operator_token18] = ACTIONS(540),
    [aux_sym_comparison_operator_token19] = ACTIONS(540),
    [aux_sym_comparison_operator_token20] = ACTIONS(540),
    [aux_sym_comparison_operator_token21] = ACTIONS(540),
    [aux_sym_comparison_operator_token22] = ACTIONS(540),
    [aux_sym_comparison_operator_token23] = ACTIONS(540),
    [aux_sym_comparison_operator_token24] = ACTIONS(540),
    [aux_sym_comparison_operator_token25] = ACTIONS(540),
    [aux_sym_comparison_operator_token26] = ACTIONS(540),
    [aux_sym_comparison_operator_token27] = ACTIONS(540),
    [aux_sym_comparison_operator_token28] = ACTIONS(538),
    [aux_sym_comparison_operator_token29] = ACTIONS(540),
    [aux_sym_comparison_operator_token30] = ACTIONS(540),
    [aux_sym_comparison_operator_token31] = ACTIONS(540),
    [aux_sym_comparison_operator_token32] = ACTIONS(540),
    [aux_sym_comparison_operator_token33] = ACTIONS(540),
    [aux_sym_comparison_operator_token34] = ACTIONS(538),
    [aux_sym_comparison_operator_token35] = ACTIONS(540),
    [aux_sym_comparison_operator_token36] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token38] = ACTIONS(540),
    [aux_sym_comparison_operator_token39] = ACTIONS(540),
    [aux_sym_comparison_operator_token40] = ACTIONS(540),
    [aux_sym_comparison_operator_token41] = ACTIONS(540),
    [aux_sym_comparison_operator_token42] = ACTIONS(540),
    [aux_sym_comparison_operator_token43] = ACTIONS(540),
    [aux_sym_comparison_operator_token44] = ACTIONS(540),
    [aux_sym_comparison_operator_token45] = ACTIONS(540),
    [aux_sym_comparison_operator_token46] = ACTIONS(540),
    [aux_sym_comparison_operator_token47] = ACTIONS(540),
    [aux_sym_comparison_operator_token48] = ACTIONS(540),
    [aux_sym_comparison_operator_token49] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [aux_sym_format_operator_token1] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [aux_sym_param_block_token1] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_RPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [anon_sym_AMP_AMP] = ACTIONS(540),
    [anon_sym_PIPE_PIPE] = ACTIONS(540),
    [anon_sym_PIPE] = ACTIONS(538),
    [aux_sym_logical_expression_token1] = ACTIONS(540),
    [aux_sym_logical_expression_token2] = ACTIONS(540),
    [aux_sym_logical_expression_token3] = ACTIONS(540),
    [aux_sym_bitwise_expression_token1] = ACTIONS(540),
    [aux_sym_bitwise_expression_token2] = ACTIONS(540),
    [aux_sym_bitwise_expression_token3] = ACTIONS(540),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [anon_sym_SLASH] = ACTIONS(538),
    [anon_sym_BSLASH] = ACTIONS(540),
    [anon_sym_PERCENT] = ACTIONS(538),
    [anon_sym_STAR] = ACTIONS(538),
    [anon_sym_DOT_DOT] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
  },
  [STATE(82)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [anon_sym_EQ] = ACTIONS(540),
    [anon_sym_BANG_EQ] = ACTIONS(540),
    [anon_sym_PLUS_EQ] = ACTIONS(540),
    [anon_sym_STAR_EQ] = ACTIONS(540),
    [anon_sym_SLASH_EQ] = ACTIONS(540),
    [anon_sym_PERCENT_EQ] = ACTIONS(540),
    [anon_sym_DASH_EQ] = ACTIONS(540),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(540),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(540),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(540),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(540),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(540),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(540),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(540),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(540),
    [anon_sym_2_GT_AMP1] = ACTIONS(540),
    [anon_sym_3_GT_AMP1] = ACTIONS(540),
    [anon_sym_4_GT_AMP1] = ACTIONS(540),
    [anon_sym_5_GT_AMP1] = ACTIONS(540),
    [anon_sym_6_GT_AMP1] = ACTIONS(540),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(540),
    [anon_sym_1_GT_AMP2] = ACTIONS(540),
    [anon_sym_3_GT_AMP2] = ACTIONS(540),
    [anon_sym_4_GT_AMP2] = ACTIONS(540),
    [anon_sym_5_GT_AMP2] = ACTIONS(540),
    [anon_sym_6_GT_AMP2] = ACTIONS(540),
    [aux_sym_comparison_operator_token1] = ACTIONS(540),
    [aux_sym_comparison_operator_token2] = ACTIONS(540),
    [aux_sym_comparison_operator_token3] = ACTIONS(540),
    [aux_sym_comparison_operator_token4] = ACTIONS(540),
    [aux_sym_comparison_operator_token5] = ACTIONS(540),
    [aux_sym_comparison_operator_token6] = ACTIONS(540),
    [aux_sym_comparison_operator_token7] = ACTIONS(540),
    [aux_sym_comparison_operator_token8] = ACTIONS(540),
    [aux_sym_comparison_operator_token9] = ACTIONS(540),
    [aux_sym_comparison_operator_token10] = ACTIONS(540),
    [aux_sym_comparison_operator_token11] = ACTIONS(540),
    [aux_sym_comparison_operator_token12] = ACTIONS(540),
    [aux_sym_comparison_operator_token13] = ACTIONS(540),
    [aux_sym_comparison_operator_token14] = ACTIONS(540),
    [aux_sym_comparison_operator_token15] = ACTIONS(540),
    [aux_sym_comparison_operator_token16] = ACTIONS(540),
    [aux_sym_comparison_operator_token17] = ACTIONS(540),
    [aux_sym_comparison_operator_token18] = ACTIONS(540),
    [aux_sym_comparison_operator_token19] = ACTIONS(540),
    [aux_sym_comparison_operator_token20] = ACTIONS(540),
    [aux_sym_comparison_operator_token21] = ACTIONS(540),
    [aux_sym_comparison_operator_token22] = ACTIONS(540),
    [aux_sym_comparison_operator_token23] = ACTIONS(540),
    [aux_sym_comparison_operator_token24] = ACTIONS(540),
    [aux_sym_comparison_operator_token25] = ACTIONS(540),
    [aux_sym_comparison_operator_token26] = ACTIONS(540),
    [aux_sym_comparison_operator_token27] = ACTIONS(540),
    [aux_sym_comparison_operator_token28] = ACTIONS(538),
    [aux_sym_comparison_operator_token29] = ACTIONS(540),
    [aux_sym_comparison_operator_token30] = ACTIONS(540),
    [aux_sym_comparison_operator_token31] = ACTIONS(540),
    [aux_sym_comparison_operator_token32] = ACTIONS(540),
    [aux_sym_comparison_operator_token33] = ACTIONS(540),
    [aux_sym_comparison_operator_token34] = ACTIONS(538),
    [aux_sym_comparison_operator_token35] = ACTIONS(540),
    [aux_sym_comparison_operator_token36] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token38] = ACTIONS(540),
    [aux_sym_comparison_operator_token39] = ACTIONS(540),
    [aux_sym_comparison_operator_token40] = ACTIONS(540),
    [aux_sym_comparison_operator_token41] = ACTIONS(540),
    [aux_sym_comparison_operator_token42] = ACTIONS(540),
    [aux_sym_comparison_operator_token43] = ACTIONS(540),
    [aux_sym_comparison_operator_token44] = ACTIONS(540),
    [aux_sym_comparison_operator_token45] = ACTIONS(540),
    [aux_sym_comparison_operator_token46] = ACTIONS(540),
    [aux_sym_comparison_operator_token47] = ACTIONS(540),
    [aux_sym_comparison_operator_token48] = ACTIONS(540),
    [aux_sym_comparison_operator_token49] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [aux_sym_format_operator_token1] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [aux_sym_param_block_token1] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [anon_sym_AMP_AMP] = ACTIONS(540),
    [anon_sym_PIPE_PIPE] = ACTIONS(540),
    [anon_sym_PIPE] = ACTIONS(538),
    [aux_sym_logical_expression_token1] = ACTIONS(540),
    [aux_sym_logical_expression_token2] = ACTIONS(540),
    [aux_sym_logical_expression_token3] = ACTIONS(540),
    [aux_sym_bitwise_expression_token1] = ACTIONS(540),
    [aux_sym_bitwise_expression_token2] = ACTIONS(540),
    [aux_sym_bitwise_expression_token3] = ACTIONS(540),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [anon_sym_SLASH] = ACTIONS(538),
    [anon_sym_BSLASH] = ACTIONS(540),
    [anon_sym_PERCENT] = ACTIONS(538),
    [anon_sym_STAR] = ACTIONS(538),
    [anon_sym_DOT_DOT] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
    [sym__statement_terminator] = ACTIONS(540),
  },
  [STATE(83)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(174),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [anon_sym_SLASH] = ACTIONS(93),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(93),
    [anon_sym_STAR] = ACTIONS(93),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(548),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [anon_sym_RBRACK] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(84)] = {
    [sym__literal] = STATE(431),
    [sym_integer_literal] = STATE(431),
    [sym_string_literal] = STATE(431),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(431),
    [sym_unary_expression] = STATE(110),
    [sym_expression_with_unary_operator] = STATE(104),
    [sym_pre_increment_expression] = STATE(105),
    [sym_pre_decrement_expression] = STATE(105),
    [sym_cast_expression] = STATE(105),
    [sym__primary_expression] = STATE(431),
    [sym__value] = STATE(431),
    [sym_parenthesized_expression] = STATE(431),
    [sym_sub_expression] = STATE(431),
    [sym_array_expression] = STATE(431),
    [sym_script_block_expression] = STATE(431),
    [sym_hash_literal_expression] = STATE(431),
    [sym_post_increment_expression] = STATE(431),
    [sym_post_decrement_expression] = STATE(431),
    [sym_member_access] = STATE(431),
    [sym_element_access] = STATE(431),
    [sym_invokation_expression] = STATE(431),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(84),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(554),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(556),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(556),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(139),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(558),
    [anon_sym_DASH] = ACTIONS(558),
    [anon_sym_SLASH] = ACTIONS(93),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(93),
    [anon_sym_STAR] = ACTIONS(93),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(558),
    [anon_sym_BANG] = ACTIONS(556),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(556),
    [anon_sym_PLUS_PLUS] = ACTIONS(560),
    [anon_sym_DASH_DASH] = ACTIONS(562),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [anon_sym_RBRACK] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(85)] = {
    [sym__literal] = STATE(435),
    [sym_integer_literal] = STATE(435),
    [sym_string_literal] = STATE(435),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(435),
    [sym_unary_expression] = STATE(182),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(435),
    [sym__value] = STATE(435),
    [sym_parenthesized_expression] = STATE(435),
    [sym_sub_expression] = STATE(435),
    [sym_array_expression] = STATE(435),
    [sym_script_block_expression] = STATE(435),
    [sym_hash_literal_expression] = STATE(435),
    [sym_post_increment_expression] = STATE(435),
    [sym_post_decrement_expression] = STATE(435),
    [sym_member_access] = STATE(435),
    [sym_element_access] = STATE(435),
    [sym_invokation_expression] = STATE(435),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(85),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(564),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(568),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(568),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(568),
    [anon_sym_LBRACE] = ACTIONS(105),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(570),
    [anon_sym_DASH] = ACTIONS(570),
    [anon_sym_SLASH] = ACTIONS(93),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(93),
    [anon_sym_STAR] = ACTIONS(93),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(570),
    [anon_sym_BANG] = ACTIONS(568),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(568),
    [anon_sym_PLUS_PLUS] = ACTIONS(572),
    [anon_sym_DASH_DASH] = ACTIONS(574),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(86)] = {
    [sym__literal] = STATE(434),
    [sym_integer_literal] = STATE(434),
    [sym_string_literal] = STATE(434),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(434),
    [sym_unary_expression] = STATE(134),
    [sym_expression_with_unary_operator] = STATE(135),
    [sym_pre_increment_expression] = STATE(130),
    [sym_pre_decrement_expression] = STATE(130),
    [sym_cast_expression] = STATE(130),
    [sym__primary_expression] = STATE(434),
    [sym__value] = STATE(434),
    [sym_parenthesized_expression] = STATE(434),
    [sym_sub_expression] = STATE(434),
    [sym_array_expression] = STATE(434),
    [sym_script_block_expression] = STATE(434),
    [sym_hash_literal_expression] = STATE(434),
    [sym_post_increment_expression] = STATE(434),
    [sym_post_decrement_expression] = STATE(434),
    [sym_member_access] = STATE(434),
    [sym_element_access] = STATE(434),
    [sym_invokation_expression] = STATE(434),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(86),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(576),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token1] = ACTIONS(93),
    [aux_sym_comparison_operator_token2] = ACTIONS(93),
    [aux_sym_comparison_operator_token3] = ACTIONS(93),
    [aux_sym_comparison_operator_token4] = ACTIONS(93),
    [aux_sym_comparison_operator_token5] = ACTIONS(93),
    [aux_sym_comparison_operator_token6] = ACTIONS(93),
    [aux_sym_comparison_operator_token7] = ACTIONS(93),
    [aux_sym_comparison_operator_token8] = ACTIONS(93),
    [aux_sym_comparison_operator_token9] = ACTIONS(93),
    [aux_sym_comparison_operator_token10] = ACTIONS(93),
    [aux_sym_comparison_operator_token11] = ACTIONS(93),
    [aux_sym_comparison_operator_token12] = ACTIONS(93),
    [aux_sym_comparison_operator_token13] = ACTIONS(93),
    [aux_sym_comparison_operator_token14] = ACTIONS(93),
    [aux_sym_comparison_operator_token15] = ACTIONS(93),
    [aux_sym_comparison_operator_token16] = ACTIONS(93),
    [aux_sym_comparison_operator_token17] = ACTIONS(93),
    [aux_sym_comparison_operator_token18] = ACTIONS(93),
    [aux_sym_comparison_operator_token19] = ACTIONS(93),
    [aux_sym_comparison_operator_token20] = ACTIONS(93),
    [aux_sym_comparison_operator_token21] = ACTIONS(93),
    [aux_sym_comparison_operator_token22] = ACTIONS(93),
    [aux_sym_comparison_operator_token23] = ACTIONS(93),
    [aux_sym_comparison_operator_token24] = ACTIONS(93),
    [aux_sym_comparison_operator_token25] = ACTIONS(93),
    [aux_sym_comparison_operator_token26] = ACTIONS(93),
    [aux_sym_comparison_operator_token27] = ACTIONS(93),
    [aux_sym_comparison_operator_token28] = ACTIONS(95),
    [aux_sym_comparison_operator_token29] = ACTIONS(93),
    [aux_sym_comparison_operator_token30] = ACTIONS(93),
    [aux_sym_comparison_operator_token31] = ACTIONS(93),
    [aux_sym_comparison_operator_token32] = ACTIONS(93),
    [aux_sym_comparison_operator_token33] = ACTIONS(93),
    [aux_sym_comparison_operator_token34] = ACTIONS(95),
    [aux_sym_comparison_operator_token35] = ACTIONS(93),
    [aux_sym_comparison_operator_token36] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(578),
    [aux_sym_comparison_operator_token38] = ACTIONS(93),
    [aux_sym_comparison_operator_token39] = ACTIONS(93),
    [aux_sym_comparison_operator_token40] = ACTIONS(93),
    [aux_sym_comparison_operator_token41] = ACTIONS(93),
    [aux_sym_comparison_operator_token42] = ACTIONS(93),
    [aux_sym_comparison_operator_token43] = ACTIONS(93),
    [aux_sym_comparison_operator_token44] = ACTIONS(93),
    [aux_sym_comparison_operator_token45] = ACTIONS(93),
    [aux_sym_comparison_operator_token46] = ACTIONS(93),
    [aux_sym_comparison_operator_token47] = ACTIONS(93),
    [aux_sym_comparison_operator_token48] = ACTIONS(93),
    [aux_sym_comparison_operator_token49] = ACTIONS(93),
    [aux_sym_comparison_operator_token50] = ACTIONS(578),
    [aux_sym_format_operator_token1] = ACTIONS(93),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(578),
    [anon_sym_LBRACE] = ACTIONS(105),
    [aux_sym_logical_expression_token1] = ACTIONS(93),
    [aux_sym_logical_expression_token2] = ACTIONS(93),
    [aux_sym_logical_expression_token3] = ACTIONS(93),
    [aux_sym_bitwise_expression_token1] = ACTIONS(93),
    [aux_sym_bitwise_expression_token2] = ACTIONS(93),
    [aux_sym_bitwise_expression_token3] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(580),
    [anon_sym_DASH] = ACTIONS(580),
    [anon_sym_SLASH] = ACTIONS(93),
    [anon_sym_BSLASH] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(93),
    [anon_sym_STAR] = ACTIONS(93),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(580),
    [anon_sym_BANG] = ACTIONS(578),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(578),
    [anon_sym_PLUS_PLUS] = ACTIONS(582),
    [anon_sym_DASH_DASH] = ACTIONS(584),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(87)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(586),
    [anon_sym_EQ] = ACTIONS(586),
    [anon_sym_BANG_EQ] = ACTIONS(586),
    [anon_sym_PLUS_EQ] = ACTIONS(586),
    [anon_sym_STAR_EQ] = ACTIONS(586),
    [anon_sym_SLASH_EQ] = ACTIONS(586),
    [anon_sym_PERCENT_EQ] = ACTIONS(586),
    [anon_sym_DASH_EQ] = ACTIONS(586),
    [anon_sym_GT] = ACTIONS(588),
    [anon_sym_GT_GT] = ACTIONS(586),
    [anon_sym_2_GT] = ACTIONS(588),
    [anon_sym_2_GT_GT] = ACTIONS(586),
    [anon_sym_3_GT] = ACTIONS(588),
    [anon_sym_3_GT_GT] = ACTIONS(586),
    [anon_sym_4_GT] = ACTIONS(588),
    [anon_sym_4_GT_GT] = ACTIONS(586),
    [anon_sym_5_GT] = ACTIONS(588),
    [anon_sym_5_GT_GT] = ACTIONS(586),
    [anon_sym_6_GT] = ACTIONS(588),
    [anon_sym_6_GT_GT] = ACTIONS(586),
    [anon_sym_STAR_GT] = ACTIONS(588),
    [anon_sym_STAR_GT_GT] = ACTIONS(586),
    [anon_sym_LT] = ACTIONS(588),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(586),
    [anon_sym_2_GT_AMP1] = ACTIONS(586),
    [anon_sym_3_GT_AMP1] = ACTIONS(586),
    [anon_sym_4_GT_AMP1] = ACTIONS(586),
    [anon_sym_5_GT_AMP1] = ACTIONS(586),
    [anon_sym_6_GT_AMP1] = ACTIONS(586),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(586),
    [anon_sym_1_GT_AMP2] = ACTIONS(586),
    [anon_sym_3_GT_AMP2] = ACTIONS(586),
    [anon_sym_4_GT_AMP2] = ACTIONS(586),
    [anon_sym_5_GT_AMP2] = ACTIONS(586),
    [anon_sym_6_GT_AMP2] = ACTIONS(586),
    [aux_sym_comparison_operator_token1] = ACTIONS(586),
    [aux_sym_comparison_operator_token2] = ACTIONS(586),
    [aux_sym_comparison_operator_token3] = ACTIONS(586),
    [aux_sym_comparison_operator_token4] = ACTIONS(586),
    [aux_sym_comparison_operator_token5] = ACTIONS(586),
    [aux_sym_comparison_operator_token6] = ACTIONS(586),
    [aux_sym_comparison_operator_token7] = ACTIONS(586),
    [aux_sym_comparison_operator_token8] = ACTIONS(586),
    [aux_sym_comparison_operator_token9] = ACTIONS(586),
    [aux_sym_comparison_operator_token10] = ACTIONS(586),
    [aux_sym_comparison_operator_token11] = ACTIONS(586),
    [aux_sym_comparison_operator_token12] = ACTIONS(586),
    [aux_sym_comparison_operator_token13] = ACTIONS(586),
    [aux_sym_comparison_operator_token14] = ACTIONS(586),
    [aux_sym_comparison_operator_token15] = ACTIONS(586),
    [aux_sym_comparison_operator_token16] = ACTIONS(586),
    [aux_sym_comparison_operator_token17] = ACTIONS(586),
    [aux_sym_comparison_operator_token18] = ACTIONS(586),
    [aux_sym_comparison_operator_token19] = ACTIONS(586),
    [aux_sym_comparison_operator_token20] = ACTIONS(586),
    [aux_sym_comparison_operator_token21] = ACTIONS(586),
    [aux_sym_comparison_operator_token22] = ACTIONS(586),
    [aux_sym_comparison_operator_token23] = ACTIONS(586),
    [aux_sym_comparison_operator_token24] = ACTIONS(586),
    [aux_sym_comparison_operator_token25] = ACTIONS(586),
    [aux_sym_comparison_operator_token26] = ACTIONS(586),
    [aux_sym_comparison_operator_token27] = ACTIONS(586),
    [aux_sym_comparison_operator_token28] = ACTIONS(588),
    [aux_sym_comparison_operator_token29] = ACTIONS(586),
    [aux_sym_comparison_operator_token30] = ACTIONS(586),
    [aux_sym_comparison_operator_token31] = ACTIONS(586),
    [aux_sym_comparison_operator_token32] = ACTIONS(586),
    [aux_sym_comparison_operator_token33] = ACTIONS(586),
    [aux_sym_comparison_operator_token34] = ACTIONS(588),
    [aux_sym_comparison_operator_token35] = ACTIONS(586),
    [aux_sym_comparison_operator_token36] = ACTIONS(586),
    [aux_sym_comparison_operator_token37] = ACTIONS(586),
    [aux_sym_comparison_operator_token38] = ACTIONS(586),
    [aux_sym_comparison_operator_token39] = ACTIONS(586),
    [aux_sym_comparison_operator_token40] = ACTIONS(586),
    [aux_sym_comparison_operator_token41] = ACTIONS(586),
    [aux_sym_comparison_operator_token42] = ACTIONS(586),
    [aux_sym_comparison_operator_token43] = ACTIONS(586),
    [aux_sym_comparison_operator_token44] = ACTIONS(586),
    [aux_sym_comparison_operator_token45] = ACTIONS(586),
    [aux_sym_comparison_operator_token46] = ACTIONS(586),
    [aux_sym_comparison_operator_token47] = ACTIONS(586),
    [aux_sym_comparison_operator_token48] = ACTIONS(586),
    [aux_sym_comparison_operator_token49] = ACTIONS(586),
    [aux_sym_comparison_operator_token50] = ACTIONS(586),
    [aux_sym_format_operator_token1] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(586),
    [anon_sym_RPAREN] = ACTIONS(586),
    [anon_sym_COMMA] = ACTIONS(586),
    [aux_sym_foreach_statement_token2] = ACTIONS(586),
    [anon_sym_AMP_AMP] = ACTIONS(586),
    [anon_sym_PIPE_PIPE] = ACTIONS(586),
    [anon_sym_PIPE] = ACTIONS(588),
    [aux_sym_logical_expression_token1] = ACTIONS(586),
    [aux_sym_logical_expression_token2] = ACTIONS(586),
    [aux_sym_logical_expression_token3] = ACTIONS(586),
    [aux_sym_bitwise_expression_token1] = ACTIONS(586),
    [aux_sym_bitwise_expression_token2] = ACTIONS(586),
    [aux_sym_bitwise_expression_token3] = ACTIONS(586),
    [anon_sym_PLUS] = ACTIONS(588),
    [anon_sym_DASH] = ACTIONS(588),
    [anon_sym_SLASH] = ACTIONS(588),
    [anon_sym_BSLASH] = ACTIONS(586),
    [anon_sym_PERCENT] = ACTIONS(588),
    [anon_sym_STAR] = ACTIONS(588),
    [anon_sym_DOT_DOT] = ACTIONS(586),
    [anon_sym_PLUS_PLUS] = ACTIONS(586),
    [anon_sym_DASH_DASH] = ACTIONS(586),
    [anon_sym_DOT2] = ACTIONS(588),
    [anon_sym_COLON_COLON] = ACTIONS(586),
    [anon_sym_RBRACK] = ACTIONS(586),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(586),
  },
  [STATE(88)] = {
    [sym_argument_list] = STATE(98),
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(590),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_BANG_EQ] = ACTIONS(590),
    [anon_sym_PLUS_EQ] = ACTIONS(590),
    [anon_sym_STAR_EQ] = ACTIONS(590),
    [anon_sym_SLASH_EQ] = ACTIONS(590),
    [anon_sym_PERCENT_EQ] = ACTIONS(590),
    [anon_sym_DASH_EQ] = ACTIONS(590),
    [anon_sym_GT] = ACTIONS(592),
    [anon_sym_GT_GT] = ACTIONS(590),
    [anon_sym_2_GT] = ACTIONS(592),
    [anon_sym_2_GT_GT] = ACTIONS(590),
    [anon_sym_3_GT] = ACTIONS(592),
    [anon_sym_3_GT_GT] = ACTIONS(590),
    [anon_sym_4_GT] = ACTIONS(592),
    [anon_sym_4_GT_GT] = ACTIONS(590),
    [anon_sym_5_GT] = ACTIONS(592),
    [anon_sym_5_GT_GT] = ACTIONS(590),
    [anon_sym_6_GT] = ACTIONS(592),
    [anon_sym_6_GT_GT] = ACTIONS(590),
    [anon_sym_STAR_GT] = ACTIONS(592),
    [anon_sym_STAR_GT_GT] = ACTIONS(590),
    [anon_sym_LT] = ACTIONS(592),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(590),
    [anon_sym_2_GT_AMP1] = ACTIONS(590),
    [anon_sym_3_GT_AMP1] = ACTIONS(590),
    [anon_sym_4_GT_AMP1] = ACTIONS(590),
    [anon_sym_5_GT_AMP1] = ACTIONS(590),
    [anon_sym_6_GT_AMP1] = ACTIONS(590),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(590),
    [anon_sym_1_GT_AMP2] = ACTIONS(590),
    [anon_sym_3_GT_AMP2] = ACTIONS(590),
    [anon_sym_4_GT_AMP2] = ACTIONS(590),
    [anon_sym_5_GT_AMP2] = ACTIONS(590),
    [anon_sym_6_GT_AMP2] = ACTIONS(590),
    [aux_sym_comparison_operator_token1] = ACTIONS(590),
    [aux_sym_comparison_operator_token2] = ACTIONS(590),
    [aux_sym_comparison_operator_token3] = ACTIONS(590),
    [aux_sym_comparison_operator_token4] = ACTIONS(590),
    [aux_sym_comparison_operator_token5] = ACTIONS(590),
    [aux_sym_comparison_operator_token6] = ACTIONS(590),
    [aux_sym_comparison_operator_token7] = ACTIONS(590),
    [aux_sym_comparison_operator_token8] = ACTIONS(590),
    [aux_sym_comparison_operator_token9] = ACTIONS(590),
    [aux_sym_comparison_operator_token10] = ACTIONS(590),
    [aux_sym_comparison_operator_token11] = ACTIONS(590),
    [aux_sym_comparison_operator_token12] = ACTIONS(590),
    [aux_sym_comparison_operator_token13] = ACTIONS(590),
    [aux_sym_comparison_operator_token14] = ACTIONS(590),
    [aux_sym_comparison_operator_token15] = ACTIONS(590),
    [aux_sym_comparison_operator_token16] = ACTIONS(590),
    [aux_sym_comparison_operator_token17] = ACTIONS(590),
    [aux_sym_comparison_operator_token18] = ACTIONS(590),
    [aux_sym_comparison_operator_token19] = ACTIONS(590),
    [aux_sym_comparison_operator_token20] = ACTIONS(590),
    [aux_sym_comparison_operator_token21] = ACTIONS(590),
    [aux_sym_comparison_operator_token22] = ACTIONS(590),
    [aux_sym_comparison_operator_token23] = ACTIONS(590),
    [aux_sym_comparison_operator_token24] = ACTIONS(590),
    [aux_sym_comparison_operator_token25] = ACTIONS(590),
    [aux_sym_comparison_operator_token26] = ACTIONS(590),
    [aux_sym_comparison_operator_token27] = ACTIONS(590),
    [aux_sym_comparison_operator_token28] = ACTIONS(592),
    [aux_sym_comparison_operator_token29] = ACTIONS(590),
    [aux_sym_comparison_operator_token30] = ACTIONS(590),
    [aux_sym_comparison_operator_token31] = ACTIONS(590),
    [aux_sym_comparison_operator_token32] = ACTIONS(590),
    [aux_sym_comparison_operator_token33] = ACTIONS(590),
    [aux_sym_comparison_operator_token34] = ACTIONS(592),
    [aux_sym_comparison_operator_token35] = ACTIONS(590),
    [aux_sym_comparison_operator_token36] = ACTIONS(590),
    [aux_sym_comparison_operator_token37] = ACTIONS(590),
    [aux_sym_comparison_operator_token38] = ACTIONS(590),
    [aux_sym_comparison_operator_token39] = ACTIONS(590),
    [aux_sym_comparison_operator_token40] = ACTIONS(590),
    [aux_sym_comparison_operator_token41] = ACTIONS(590),
    [aux_sym_comparison_operator_token42] = ACTIONS(590),
    [aux_sym_comparison_operator_token43] = ACTIONS(590),
    [aux_sym_comparison_operator_token44] = ACTIONS(590),
    [aux_sym_comparison_operator_token45] = ACTIONS(590),
    [aux_sym_comparison_operator_token46] = ACTIONS(590),
    [aux_sym_comparison_operator_token47] = ACTIONS(590),
    [aux_sym_comparison_operator_token48] = ACTIONS(590),
    [aux_sym_comparison_operator_token49] = ACTIONS(590),
    [aux_sym_comparison_operator_token50] = ACTIONS(590),
    [aux_sym_format_operator_token1] = ACTIONS(590),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(590),
    [anon_sym_COMMA] = ACTIONS(590),
    [anon_sym_AMP_AMP] = ACTIONS(590),
    [anon_sym_PIPE_PIPE] = ACTIONS(590),
    [anon_sym_PIPE] = ACTIONS(592),
    [aux_sym_logical_expression_token1] = ACTIONS(590),
    [aux_sym_logical_expression_token2] = ACTIONS(590),
    [aux_sym_logical_expression_token3] = ACTIONS(590),
    [aux_sym_bitwise_expression_token1] = ACTIONS(590),
    [aux_sym_bitwise_expression_token2] = ACTIONS(590),
    [aux_sym_bitwise_expression_token3] = ACTIONS(590),
    [anon_sym_PLUS] = ACTIONS(592),
    [anon_sym_DASH] = ACTIONS(592),
    [anon_sym_SLASH] = ACTIONS(592),
    [anon_sym_BSLASH] = ACTIONS(590),
    [anon_sym_PERCENT] = ACTIONS(592),
    [anon_sym_STAR] = ACTIONS(592),
    [anon_sym_DOT_DOT] = ACTIONS(590),
    [anon_sym_PLUS_PLUS] = ACTIONS(590),
    [anon_sym_DASH_DASH] = ACTIONS(590),
    [anon_sym_DOT2] = ACTIONS(592),
    [anon_sym_COLON_COLON] = ACTIONS(590),
    [anon_sym_RBRACK] = ACTIONS(590),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(590),
  },
  [STATE(89)] = {
    [sym_argument_list] = STATE(98),
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(590),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_BANG_EQ] = ACTIONS(590),
    [anon_sym_PLUS_EQ] = ACTIONS(590),
    [anon_sym_STAR_EQ] = ACTIONS(590),
    [anon_sym_SLASH_EQ] = ACTIONS(590),
    [anon_sym_PERCENT_EQ] = ACTIONS(590),
    [anon_sym_DASH_EQ] = ACTIONS(590),
    [anon_sym_GT] = ACTIONS(592),
    [anon_sym_GT_GT] = ACTIONS(590),
    [anon_sym_2_GT] = ACTIONS(592),
    [anon_sym_2_GT_GT] = ACTIONS(590),
    [anon_sym_3_GT] = ACTIONS(592),
    [anon_sym_3_GT_GT] = ACTIONS(590),
    [anon_sym_4_GT] = ACTIONS(592),
    [anon_sym_4_GT_GT] = ACTIONS(590),
    [anon_sym_5_GT] = ACTIONS(592),
    [anon_sym_5_GT_GT] = ACTIONS(590),
    [anon_sym_6_GT] = ACTIONS(592),
    [anon_sym_6_GT_GT] = ACTIONS(590),
    [anon_sym_STAR_GT] = ACTIONS(592),
    [anon_sym_STAR_GT_GT] = ACTIONS(590),
    [anon_sym_LT] = ACTIONS(592),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(590),
    [anon_sym_2_GT_AMP1] = ACTIONS(590),
    [anon_sym_3_GT_AMP1] = ACTIONS(590),
    [anon_sym_4_GT_AMP1] = ACTIONS(590),
    [anon_sym_5_GT_AMP1] = ACTIONS(590),
    [anon_sym_6_GT_AMP1] = ACTIONS(590),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(590),
    [anon_sym_1_GT_AMP2] = ACTIONS(590),
    [anon_sym_3_GT_AMP2] = ACTIONS(590),
    [anon_sym_4_GT_AMP2] = ACTIONS(590),
    [anon_sym_5_GT_AMP2] = ACTIONS(590),
    [anon_sym_6_GT_AMP2] = ACTIONS(590),
    [aux_sym_comparison_operator_token1] = ACTIONS(590),
    [aux_sym_comparison_operator_token2] = ACTIONS(590),
    [aux_sym_comparison_operator_token3] = ACTIONS(590),
    [aux_sym_comparison_operator_token4] = ACTIONS(590),
    [aux_sym_comparison_operator_token5] = ACTIONS(590),
    [aux_sym_comparison_operator_token6] = ACTIONS(590),
    [aux_sym_comparison_operator_token7] = ACTIONS(590),
    [aux_sym_comparison_operator_token8] = ACTIONS(590),
    [aux_sym_comparison_operator_token9] = ACTIONS(590),
    [aux_sym_comparison_operator_token10] = ACTIONS(590),
    [aux_sym_comparison_operator_token11] = ACTIONS(590),
    [aux_sym_comparison_operator_token12] = ACTIONS(590),
    [aux_sym_comparison_operator_token13] = ACTIONS(590),
    [aux_sym_comparison_operator_token14] = ACTIONS(590),
    [aux_sym_comparison_operator_token15] = ACTIONS(590),
    [aux_sym_comparison_operator_token16] = ACTIONS(590),
    [aux_sym_comparison_operator_token17] = ACTIONS(590),
    [aux_sym_comparison_operator_token18] = ACTIONS(590),
    [aux_sym_comparison_operator_token19] = ACTIONS(590),
    [aux_sym_comparison_operator_token20] = ACTIONS(590),
    [aux_sym_comparison_operator_token21] = ACTIONS(590),
    [aux_sym_comparison_operator_token22] = ACTIONS(590),
    [aux_sym_comparison_operator_token23] = ACTIONS(590),
    [aux_sym_comparison_operator_token24] = ACTIONS(590),
    [aux_sym_comparison_operator_token25] = ACTIONS(590),
    [aux_sym_comparison_operator_token26] = ACTIONS(590),
    [aux_sym_comparison_operator_token27] = ACTIONS(590),
    [aux_sym_comparison_operator_token28] = ACTIONS(592),
    [aux_sym_comparison_operator_token29] = ACTIONS(590),
    [aux_sym_comparison_operator_token30] = ACTIONS(590),
    [aux_sym_comparison_operator_token31] = ACTIONS(590),
    [aux_sym_comparison_operator_token32] = ACTIONS(590),
    [aux_sym_comparison_operator_token33] = ACTIONS(590),
    [aux_sym_comparison_operator_token34] = ACTIONS(592),
    [aux_sym_comparison_operator_token35] = ACTIONS(590),
    [aux_sym_comparison_operator_token36] = ACTIONS(590),
    [aux_sym_comparison_operator_token37] = ACTIONS(590),
    [aux_sym_comparison_operator_token38] = ACTIONS(590),
    [aux_sym_comparison_operator_token39] = ACTIONS(590),
    [aux_sym_comparison_operator_token40] = ACTIONS(590),
    [aux_sym_comparison_operator_token41] = ACTIONS(590),
    [aux_sym_comparison_operator_token42] = ACTIONS(590),
    [aux_sym_comparison_operator_token43] = ACTIONS(590),
    [aux_sym_comparison_operator_token44] = ACTIONS(590),
    [aux_sym_comparison_operator_token45] = ACTIONS(590),
    [aux_sym_comparison_operator_token46] = ACTIONS(590),
    [aux_sym_comparison_operator_token47] = ACTIONS(590),
    [aux_sym_comparison_operator_token48] = ACTIONS(590),
    [aux_sym_comparison_operator_token49] = ACTIONS(590),
    [aux_sym_comparison_operator_token50] = ACTIONS(590),
    [aux_sym_format_operator_token1] = ACTIONS(590),
    [anon_sym_LPAREN] = ACTIONS(594),
    [anon_sym_RPAREN] = ACTIONS(590),
    [anon_sym_COMMA] = ACTIONS(590),
    [anon_sym_AMP_AMP] = ACTIONS(590),
    [anon_sym_PIPE_PIPE] = ACTIONS(590),
    [anon_sym_PIPE] = ACTIONS(592),
    [aux_sym_logical_expression_token1] = ACTIONS(590),
    [aux_sym_logical_expression_token2] = ACTIONS(590),
    [aux_sym_logical_expression_token3] = ACTIONS(590),
    [aux_sym_bitwise_expression_token1] = ACTIONS(590),
    [aux_sym_bitwise_expression_token2] = ACTIONS(590),
    [aux_sym_bitwise_expression_token3] = ACTIONS(590),
    [anon_sym_PLUS] = ACTIONS(592),
    [anon_sym_DASH] = ACTIONS(592),
    [anon_sym_SLASH] = ACTIONS(592),
    [anon_sym_BSLASH] = ACTIONS(590),
    [anon_sym_PERCENT] = ACTIONS(592),
    [anon_sym_STAR] = ACTIONS(592),
    [anon_sym_DOT_DOT] = ACTIONS(590),
    [anon_sym_PLUS_PLUS] = ACTIONS(590),
    [anon_sym_DASH_DASH] = ACTIONS(590),
    [anon_sym_DOT2] = ACTIONS(592),
    [anon_sym_COLON_COLON] = ACTIONS(590),
    [anon_sym_RBRACK] = ACTIONS(590),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(590),
  },
  [STATE(90)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_EQ] = ACTIONS(596),
    [anon_sym_BANG_EQ] = ACTIONS(596),
    [anon_sym_PLUS_EQ] = ACTIONS(596),
    [anon_sym_STAR_EQ] = ACTIONS(596),
    [anon_sym_SLASH_EQ] = ACTIONS(596),
    [anon_sym_PERCENT_EQ] = ACTIONS(596),
    [anon_sym_DASH_EQ] = ACTIONS(596),
    [anon_sym_GT] = ACTIONS(598),
    [anon_sym_GT_GT] = ACTIONS(596),
    [anon_sym_2_GT] = ACTIONS(598),
    [anon_sym_2_GT_GT] = ACTIONS(596),
    [anon_sym_3_GT] = ACTIONS(598),
    [anon_sym_3_GT_GT] = ACTIONS(596),
    [anon_sym_4_GT] = ACTIONS(598),
    [anon_sym_4_GT_GT] = ACTIONS(596),
    [anon_sym_5_GT] = ACTIONS(598),
    [anon_sym_5_GT_GT] = ACTIONS(596),
    [anon_sym_6_GT] = ACTIONS(598),
    [anon_sym_6_GT_GT] = ACTIONS(596),
    [anon_sym_STAR_GT] = ACTIONS(598),
    [anon_sym_STAR_GT_GT] = ACTIONS(596),
    [anon_sym_LT] = ACTIONS(598),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(596),
    [anon_sym_2_GT_AMP1] = ACTIONS(596),
    [anon_sym_3_GT_AMP1] = ACTIONS(596),
    [anon_sym_4_GT_AMP1] = ACTIONS(596),
    [anon_sym_5_GT_AMP1] = ACTIONS(596),
    [anon_sym_6_GT_AMP1] = ACTIONS(596),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(596),
    [anon_sym_1_GT_AMP2] = ACTIONS(596),
    [anon_sym_3_GT_AMP2] = ACTIONS(596),
    [anon_sym_4_GT_AMP2] = ACTIONS(596),
    [anon_sym_5_GT_AMP2] = ACTIONS(596),
    [anon_sym_6_GT_AMP2] = ACTIONS(596),
    [aux_sym_comparison_operator_token1] = ACTIONS(596),
    [aux_sym_comparison_operator_token2] = ACTIONS(596),
    [aux_sym_comparison_operator_token3] = ACTIONS(596),
    [aux_sym_comparison_operator_token4] = ACTIONS(596),
    [aux_sym_comparison_operator_token5] = ACTIONS(596),
    [aux_sym_comparison_operator_token6] = ACTIONS(596),
    [aux_sym_comparison_operator_token7] = ACTIONS(596),
    [aux_sym_comparison_operator_token8] = ACTIONS(596),
    [aux_sym_comparison_operator_token9] = ACTIONS(596),
    [aux_sym_comparison_operator_token10] = ACTIONS(596),
    [aux_sym_comparison_operator_token11] = ACTIONS(596),
    [aux_sym_comparison_operator_token12] = ACTIONS(596),
    [aux_sym_comparison_operator_token13] = ACTIONS(596),
    [aux_sym_comparison_operator_token14] = ACTIONS(596),
    [aux_sym_comparison_operator_token15] = ACTIONS(596),
    [aux_sym_comparison_operator_token16] = ACTIONS(596),
    [aux_sym_comparison_operator_token17] = ACTIONS(596),
    [aux_sym_comparison_operator_token18] = ACTIONS(596),
    [aux_sym_comparison_operator_token19] = ACTIONS(596),
    [aux_sym_comparison_operator_token20] = ACTIONS(596),
    [aux_sym_comparison_operator_token21] = ACTIONS(596),
    [aux_sym_comparison_operator_token22] = ACTIONS(596),
    [aux_sym_comparison_operator_token23] = ACTIONS(596),
    [aux_sym_comparison_operator_token24] = ACTIONS(596),
    [aux_sym_comparison_operator_token25] = ACTIONS(596),
    [aux_sym_comparison_operator_token26] = ACTIONS(596),
    [aux_sym_comparison_operator_token27] = ACTIONS(596),
    [aux_sym_comparison_operator_token28] = ACTIONS(598),
    [aux_sym_comparison_operator_token29] = ACTIONS(596),
    [aux_sym_comparison_operator_token30] = ACTIONS(596),
    [aux_sym_comparison_operator_token31] = ACTIONS(596),
    [aux_sym_comparison_operator_token32] = ACTIONS(596),
    [aux_sym_comparison_operator_token33] = ACTIONS(596),
    [aux_sym_comparison_operator_token34] = ACTIONS(598),
    [aux_sym_comparison_operator_token35] = ACTIONS(596),
    [aux_sym_comparison_operator_token36] = ACTIONS(596),
    [aux_sym_comparison_operator_token37] = ACTIONS(596),
    [aux_sym_comparison_operator_token38] = ACTIONS(596),
    [aux_sym_comparison_operator_token39] = ACTIONS(596),
    [aux_sym_comparison_operator_token40] = ACTIONS(596),
    [aux_sym_comparison_operator_token41] = ACTIONS(596),
    [aux_sym_comparison_operator_token42] = ACTIONS(596),
    [aux_sym_comparison_operator_token43] = ACTIONS(596),
    [aux_sym_comparison_operator_token44] = ACTIONS(596),
    [aux_sym_comparison_operator_token45] = ACTIONS(596),
    [aux_sym_comparison_operator_token46] = ACTIONS(596),
    [aux_sym_comparison_operator_token47] = ACTIONS(596),
    [aux_sym_comparison_operator_token48] = ACTIONS(596),
    [aux_sym_comparison_operator_token49] = ACTIONS(596),
    [aux_sym_comparison_operator_token50] = ACTIONS(596),
    [aux_sym_format_operator_token1] = ACTIONS(596),
    [anon_sym_LPAREN] = ACTIONS(596),
    [anon_sym_RPAREN] = ACTIONS(596),
    [anon_sym_COMMA] = ACTIONS(596),
    [anon_sym_AMP_AMP] = ACTIONS(596),
    [anon_sym_PIPE_PIPE] = ACTIONS(596),
    [anon_sym_PIPE] = ACTIONS(598),
    [aux_sym_logical_expression_token1] = ACTIONS(596),
    [aux_sym_logical_expression_token2] = ACTIONS(596),
    [aux_sym_logical_expression_token3] = ACTIONS(596),
    [aux_sym_bitwise_expression_token1] = ACTIONS(596),
    [aux_sym_bitwise_expression_token2] = ACTIONS(596),
    [aux_sym_bitwise_expression_token3] = ACTIONS(596),
    [anon_sym_PLUS] = ACTIONS(598),
    [anon_sym_DASH] = ACTIONS(598),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_BSLASH] = ACTIONS(596),
    [anon_sym_PERCENT] = ACTIONS(598),
    [anon_sym_STAR] = ACTIONS(598),
    [anon_sym_DOT_DOT] = ACTIONS(596),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_DOT2] = ACTIONS(598),
    [anon_sym_COLON_COLON] = ACTIONS(596),
    [anon_sym_RBRACK] = ACTIONS(596),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(596),
  },
  [STATE(91)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(600),
    [anon_sym_EQ] = ACTIONS(600),
    [anon_sym_BANG_EQ] = ACTIONS(600),
    [anon_sym_PLUS_EQ] = ACTIONS(600),
    [anon_sym_STAR_EQ] = ACTIONS(600),
    [anon_sym_SLASH_EQ] = ACTIONS(600),
    [anon_sym_PERCENT_EQ] = ACTIONS(600),
    [anon_sym_DASH_EQ] = ACTIONS(600),
    [anon_sym_GT] = ACTIONS(602),
    [anon_sym_GT_GT] = ACTIONS(600),
    [anon_sym_2_GT] = ACTIONS(602),
    [anon_sym_2_GT_GT] = ACTIONS(600),
    [anon_sym_3_GT] = ACTIONS(602),
    [anon_sym_3_GT_GT] = ACTIONS(600),
    [anon_sym_4_GT] = ACTIONS(602),
    [anon_sym_4_GT_GT] = ACTIONS(600),
    [anon_sym_5_GT] = ACTIONS(602),
    [anon_sym_5_GT_GT] = ACTIONS(600),
    [anon_sym_6_GT] = ACTIONS(602),
    [anon_sym_6_GT_GT] = ACTIONS(600),
    [anon_sym_STAR_GT] = ACTIONS(602),
    [anon_sym_STAR_GT_GT] = ACTIONS(600),
    [anon_sym_LT] = ACTIONS(602),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(600),
    [anon_sym_2_GT_AMP1] = ACTIONS(600),
    [anon_sym_3_GT_AMP1] = ACTIONS(600),
    [anon_sym_4_GT_AMP1] = ACTIONS(600),
    [anon_sym_5_GT_AMP1] = ACTIONS(600),
    [anon_sym_6_GT_AMP1] = ACTIONS(600),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(600),
    [anon_sym_1_GT_AMP2] = ACTIONS(600),
    [anon_sym_3_GT_AMP2] = ACTIONS(600),
    [anon_sym_4_GT_AMP2] = ACTIONS(600),
    [anon_sym_5_GT_AMP2] = ACTIONS(600),
    [anon_sym_6_GT_AMP2] = ACTIONS(600),
    [aux_sym_comparison_operator_token1] = ACTIONS(600),
    [aux_sym_comparison_operator_token2] = ACTIONS(600),
    [aux_sym_comparison_operator_token3] = ACTIONS(600),
    [aux_sym_comparison_operator_token4] = ACTIONS(600),
    [aux_sym_comparison_operator_token5] = ACTIONS(600),
    [aux_sym_comparison_operator_token6] = ACTIONS(600),
    [aux_sym_comparison_operator_token7] = ACTIONS(600),
    [aux_sym_comparison_operator_token8] = ACTIONS(600),
    [aux_sym_comparison_operator_token9] = ACTIONS(600),
    [aux_sym_comparison_operator_token10] = ACTIONS(600),
    [aux_sym_comparison_operator_token11] = ACTIONS(600),
    [aux_sym_comparison_operator_token12] = ACTIONS(600),
    [aux_sym_comparison_operator_token13] = ACTIONS(600),
    [aux_sym_comparison_operator_token14] = ACTIONS(600),
    [aux_sym_comparison_operator_token15] = ACTIONS(600),
    [aux_sym_comparison_operator_token16] = ACTIONS(600),
    [aux_sym_comparison_operator_token17] = ACTIONS(600),
    [aux_sym_comparison_operator_token18] = ACTIONS(600),
    [aux_sym_comparison_operator_token19] = ACTIONS(600),
    [aux_sym_comparison_operator_token20] = ACTIONS(600),
    [aux_sym_comparison_operator_token21] = ACTIONS(600),
    [aux_sym_comparison_operator_token22] = ACTIONS(600),
    [aux_sym_comparison_operator_token23] = ACTIONS(600),
    [aux_sym_comparison_operator_token24] = ACTIONS(600),
    [aux_sym_comparison_operator_token25] = ACTIONS(600),
    [aux_sym_comparison_operator_token26] = ACTIONS(600),
    [aux_sym_comparison_operator_token27] = ACTIONS(600),
    [aux_sym_comparison_operator_token28] = ACTIONS(602),
    [aux_sym_comparison_operator_token29] = ACTIONS(600),
    [aux_sym_comparison_operator_token30] = ACTIONS(600),
    [aux_sym_comparison_operator_token31] = ACTIONS(600),
    [aux_sym_comparison_operator_token32] = ACTIONS(600),
    [aux_sym_comparison_operator_token33] = ACTIONS(600),
    [aux_sym_comparison_operator_token34] = ACTIONS(602),
    [aux_sym_comparison_operator_token35] = ACTIONS(600),
    [aux_sym_comparison_operator_token36] = ACTIONS(600),
    [aux_sym_comparison_operator_token37] = ACTIONS(600),
    [aux_sym_comparison_operator_token38] = ACTIONS(600),
    [aux_sym_comparison_operator_token39] = ACTIONS(600),
    [aux_sym_comparison_operator_token40] = ACTIONS(600),
    [aux_sym_comparison_operator_token41] = ACTIONS(600),
    [aux_sym_comparison_operator_token42] = ACTIONS(600),
    [aux_sym_comparison_operator_token43] = ACTIONS(600),
    [aux_sym_comparison_operator_token44] = ACTIONS(600),
    [aux_sym_comparison_operator_token45] = ACTIONS(600),
    [aux_sym_comparison_operator_token46] = ACTIONS(600),
    [aux_sym_comparison_operator_token47] = ACTIONS(600),
    [aux_sym_comparison_operator_token48] = ACTIONS(600),
    [aux_sym_comparison_operator_token49] = ACTIONS(600),
    [aux_sym_comparison_operator_token50] = ACTIONS(600),
    [aux_sym_format_operator_token1] = ACTIONS(600),
    [anon_sym_LPAREN] = ACTIONS(600),
    [anon_sym_RPAREN] = ACTIONS(600),
    [anon_sym_COMMA] = ACTIONS(600),
    [anon_sym_AMP_AMP] = ACTIONS(600),
    [anon_sym_PIPE_PIPE] = ACTIONS(600),
    [anon_sym_PIPE] = ACTIONS(602),
    [aux_sym_logical_expression_token1] = ACTIONS(600),
    [aux_sym_logical_expression_token2] = ACTIONS(600),
    [aux_sym_logical_expression_token3] = ACTIONS(600),
    [aux_sym_bitwise_expression_token1] = ACTIONS(600),
    [aux_sym_bitwise_expression_token2] = ACTIONS(600),
    [aux_sym_bitwise_expression_token3] = ACTIONS(600),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(602),
    [anon_sym_BSLASH] = ACTIONS(600),
    [anon_sym_PERCENT] = ACTIONS(602),
    [anon_sym_STAR] = ACTIONS(602),
    [anon_sym_DOT_DOT] = ACTIONS(600),
    [anon_sym_PLUS_PLUS] = ACTIONS(600),
    [anon_sym_DASH_DASH] = ACTIONS(600),
    [anon_sym_DOT2] = ACTIONS(602),
    [anon_sym_COLON_COLON] = ACTIONS(600),
    [anon_sym_RBRACK] = ACTIONS(600),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(600),
  },
  [STATE(92)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(604),
    [anon_sym_EQ] = ACTIONS(604),
    [anon_sym_BANG_EQ] = ACTIONS(604),
    [anon_sym_PLUS_EQ] = ACTIONS(604),
    [anon_sym_STAR_EQ] = ACTIONS(604),
    [anon_sym_SLASH_EQ] = ACTIONS(604),
    [anon_sym_PERCENT_EQ] = ACTIONS(604),
    [anon_sym_DASH_EQ] = ACTIONS(604),
    [anon_sym_GT] = ACTIONS(607),
    [anon_sym_GT_GT] = ACTIONS(604),
    [anon_sym_2_GT] = ACTIONS(607),
    [anon_sym_2_GT_GT] = ACTIONS(604),
    [anon_sym_3_GT] = ACTIONS(607),
    [anon_sym_3_GT_GT] = ACTIONS(604),
    [anon_sym_4_GT] = ACTIONS(607),
    [anon_sym_4_GT_GT] = ACTIONS(604),
    [anon_sym_5_GT] = ACTIONS(607),
    [anon_sym_5_GT_GT] = ACTIONS(604),
    [anon_sym_6_GT] = ACTIONS(607),
    [anon_sym_6_GT_GT] = ACTIONS(604),
    [anon_sym_STAR_GT] = ACTIONS(607),
    [anon_sym_STAR_GT_GT] = ACTIONS(604),
    [anon_sym_LT] = ACTIONS(607),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(604),
    [anon_sym_2_GT_AMP1] = ACTIONS(604),
    [anon_sym_3_GT_AMP1] = ACTIONS(604),
    [anon_sym_4_GT_AMP1] = ACTIONS(604),
    [anon_sym_5_GT_AMP1] = ACTIONS(604),
    [anon_sym_6_GT_AMP1] = ACTIONS(604),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(604),
    [anon_sym_1_GT_AMP2] = ACTIONS(604),
    [anon_sym_3_GT_AMP2] = ACTIONS(604),
    [anon_sym_4_GT_AMP2] = ACTIONS(604),
    [anon_sym_5_GT_AMP2] = ACTIONS(604),
    [anon_sym_6_GT_AMP2] = ACTIONS(604),
    [aux_sym_comparison_operator_token1] = ACTIONS(604),
    [aux_sym_comparison_operator_token2] = ACTIONS(604),
    [aux_sym_comparison_operator_token3] = ACTIONS(604),
    [aux_sym_comparison_operator_token4] = ACTIONS(604),
    [aux_sym_comparison_operator_token5] = ACTIONS(604),
    [aux_sym_comparison_operator_token6] = ACTIONS(604),
    [aux_sym_comparison_operator_token7] = ACTIONS(604),
    [aux_sym_comparison_operator_token8] = ACTIONS(604),
    [aux_sym_comparison_operator_token9] = ACTIONS(604),
    [aux_sym_comparison_operator_token10] = ACTIONS(604),
    [aux_sym_comparison_operator_token11] = ACTIONS(604),
    [aux_sym_comparison_operator_token12] = ACTIONS(604),
    [aux_sym_comparison_operator_token13] = ACTIONS(604),
    [aux_sym_comparison_operator_token14] = ACTIONS(604),
    [aux_sym_comparison_operator_token15] = ACTIONS(604),
    [aux_sym_comparison_operator_token16] = ACTIONS(604),
    [aux_sym_comparison_operator_token17] = ACTIONS(604),
    [aux_sym_comparison_operator_token18] = ACTIONS(604),
    [aux_sym_comparison_operator_token19] = ACTIONS(604),
    [aux_sym_comparison_operator_token20] = ACTIONS(604),
    [aux_sym_comparison_operator_token21] = ACTIONS(604),
    [aux_sym_comparison_operator_token22] = ACTIONS(604),
    [aux_sym_comparison_operator_token23] = ACTIONS(604),
    [aux_sym_comparison_operator_token24] = ACTIONS(604),
    [aux_sym_comparison_operator_token25] = ACTIONS(604),
    [aux_sym_comparison_operator_token26] = ACTIONS(604),
    [aux_sym_comparison_operator_token27] = ACTIONS(604),
    [aux_sym_comparison_operator_token28] = ACTIONS(607),
    [aux_sym_comparison_operator_token29] = ACTIONS(604),
    [aux_sym_comparison_operator_token30] = ACTIONS(604),
    [aux_sym_comparison_operator_token31] = ACTIONS(604),
    [aux_sym_comparison_operator_token32] = ACTIONS(604),
    [aux_sym_comparison_operator_token33] = ACTIONS(604),
    [aux_sym_comparison_operator_token34] = ACTIONS(607),
    [aux_sym_comparison_operator_token35] = ACTIONS(604),
    [aux_sym_comparison_operator_token36] = ACTIONS(604),
    [aux_sym_comparison_operator_token37] = ACTIONS(604),
    [aux_sym_comparison_operator_token38] = ACTIONS(604),
    [aux_sym_comparison_operator_token39] = ACTIONS(604),
    [aux_sym_comparison_operator_token40] = ACTIONS(604),
    [aux_sym_comparison_operator_token41] = ACTIONS(604),
    [aux_sym_comparison_operator_token42] = ACTIONS(604),
    [aux_sym_comparison_operator_token43] = ACTIONS(604),
    [aux_sym_comparison_operator_token44] = ACTIONS(604),
    [aux_sym_comparison_operator_token45] = ACTIONS(604),
    [aux_sym_comparison_operator_token46] = ACTIONS(604),
    [aux_sym_comparison_operator_token47] = ACTIONS(604),
    [aux_sym_comparison_operator_token48] = ACTIONS(604),
    [aux_sym_comparison_operator_token49] = ACTIONS(604),
    [aux_sym_comparison_operator_token50] = ACTIONS(604),
    [aux_sym_format_operator_token1] = ACTIONS(604),
    [anon_sym_LPAREN] = ACTIONS(604),
    [anon_sym_RPAREN] = ACTIONS(604),
    [anon_sym_COMMA] = ACTIONS(604),
    [anon_sym_AMP_AMP] = ACTIONS(604),
    [anon_sym_PIPE_PIPE] = ACTIONS(604),
    [anon_sym_PIPE] = ACTIONS(607),
    [aux_sym_logical_expression_token1] = ACTIONS(604),
    [aux_sym_logical_expression_token2] = ACTIONS(604),
    [aux_sym_logical_expression_token3] = ACTIONS(604),
    [aux_sym_bitwise_expression_token1] = ACTIONS(604),
    [aux_sym_bitwise_expression_token2] = ACTIONS(604),
    [aux_sym_bitwise_expression_token3] = ACTIONS(604),
    [anon_sym_PLUS] = ACTIONS(607),
    [anon_sym_DASH] = ACTIONS(607),
    [anon_sym_SLASH] = ACTIONS(607),
    [anon_sym_BSLASH] = ACTIONS(604),
    [anon_sym_PERCENT] = ACTIONS(607),
    [anon_sym_STAR] = ACTIONS(607),
    [anon_sym_DOT_DOT] = ACTIONS(604),
    [anon_sym_PLUS_PLUS] = ACTIONS(604),
    [anon_sym_DASH_DASH] = ACTIONS(604),
    [anon_sym_DOT2] = ACTIONS(607),
    [anon_sym_COLON_COLON] = ACTIONS(604),
    [anon_sym_RBRACK] = ACTIONS(604),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(604),
  },
  [STATE(93)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(610),
    [anon_sym_EQ] = ACTIONS(610),
    [anon_sym_BANG_EQ] = ACTIONS(610),
    [anon_sym_PLUS_EQ] = ACTIONS(610),
    [anon_sym_STAR_EQ] = ACTIONS(610),
    [anon_sym_SLASH_EQ] = ACTIONS(610),
    [anon_sym_PERCENT_EQ] = ACTIONS(610),
    [anon_sym_DASH_EQ] = ACTIONS(610),
    [anon_sym_GT] = ACTIONS(612),
    [anon_sym_GT_GT] = ACTIONS(610),
    [anon_sym_2_GT] = ACTIONS(612),
    [anon_sym_2_GT_GT] = ACTIONS(610),
    [anon_sym_3_GT] = ACTIONS(612),
    [anon_sym_3_GT_GT] = ACTIONS(610),
    [anon_sym_4_GT] = ACTIONS(612),
    [anon_sym_4_GT_GT] = ACTIONS(610),
    [anon_sym_5_GT] = ACTIONS(612),
    [anon_sym_5_GT_GT] = ACTIONS(610),
    [anon_sym_6_GT] = ACTIONS(612),
    [anon_sym_6_GT_GT] = ACTIONS(610),
    [anon_sym_STAR_GT] = ACTIONS(612),
    [anon_sym_STAR_GT_GT] = ACTIONS(610),
    [anon_sym_LT] = ACTIONS(612),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(610),
    [anon_sym_2_GT_AMP1] = ACTIONS(610),
    [anon_sym_3_GT_AMP1] = ACTIONS(610),
    [anon_sym_4_GT_AMP1] = ACTIONS(610),
    [anon_sym_5_GT_AMP1] = ACTIONS(610),
    [anon_sym_6_GT_AMP1] = ACTIONS(610),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(610),
    [anon_sym_1_GT_AMP2] = ACTIONS(610),
    [anon_sym_3_GT_AMP2] = ACTIONS(610),
    [anon_sym_4_GT_AMP2] = ACTIONS(610),
    [anon_sym_5_GT_AMP2] = ACTIONS(610),
    [anon_sym_6_GT_AMP2] = ACTIONS(610),
    [aux_sym_comparison_operator_token1] = ACTIONS(610),
    [aux_sym_comparison_operator_token2] = ACTIONS(610),
    [aux_sym_comparison_operator_token3] = ACTIONS(610),
    [aux_sym_comparison_operator_token4] = ACTIONS(610),
    [aux_sym_comparison_operator_token5] = ACTIONS(610),
    [aux_sym_comparison_operator_token6] = ACTIONS(610),
    [aux_sym_comparison_operator_token7] = ACTIONS(610),
    [aux_sym_comparison_operator_token8] = ACTIONS(610),
    [aux_sym_comparison_operator_token9] = ACTIONS(610),
    [aux_sym_comparison_operator_token10] = ACTIONS(610),
    [aux_sym_comparison_operator_token11] = ACTIONS(610),
    [aux_sym_comparison_operator_token12] = ACTIONS(610),
    [aux_sym_comparison_operator_token13] = ACTIONS(610),
    [aux_sym_comparison_operator_token14] = ACTIONS(610),
    [aux_sym_comparison_operator_token15] = ACTIONS(610),
    [aux_sym_comparison_operator_token16] = ACTIONS(610),
    [aux_sym_comparison_operator_token17] = ACTIONS(610),
    [aux_sym_comparison_operator_token18] = ACTIONS(610),
    [aux_sym_comparison_operator_token19] = ACTIONS(610),
    [aux_sym_comparison_operator_token20] = ACTIONS(610),
    [aux_sym_comparison_operator_token21] = ACTIONS(610),
    [aux_sym_comparison_operator_token22] = ACTIONS(610),
    [aux_sym_comparison_operator_token23] = ACTIONS(610),
    [aux_sym_comparison_operator_token24] = ACTIONS(610),
    [aux_sym_comparison_operator_token25] = ACTIONS(610),
    [aux_sym_comparison_operator_token26] = ACTIONS(610),
    [aux_sym_comparison_operator_token27] = ACTIONS(610),
    [aux_sym_comparison_operator_token28] = ACTIONS(612),
    [aux_sym_comparison_operator_token29] = ACTIONS(610),
    [aux_sym_comparison_operator_token30] = ACTIONS(610),
    [aux_sym_comparison_operator_token31] = ACTIONS(610),
    [aux_sym_comparison_operator_token32] = ACTIONS(610),
    [aux_sym_comparison_operator_token33] = ACTIONS(610),
    [aux_sym_comparison_operator_token34] = ACTIONS(612),
    [aux_sym_comparison_operator_token35] = ACTIONS(610),
    [aux_sym_comparison_operator_token36] = ACTIONS(610),
    [aux_sym_comparison_operator_token37] = ACTIONS(610),
    [aux_sym_comparison_operator_token38] = ACTIONS(610),
    [aux_sym_comparison_operator_token39] = ACTIONS(610),
    [aux_sym_comparison_operator_token40] = ACTIONS(610),
    [aux_sym_comparison_operator_token41] = ACTIONS(610),
    [aux_sym_comparison_operator_token42] = ACTIONS(610),
    [aux_sym_comparison_operator_token43] = ACTIONS(610),
    [aux_sym_comparison_operator_token44] = ACTIONS(610),
    [aux_sym_comparison_operator_token45] = ACTIONS(610),
    [aux_sym_comparison_operator_token46] = ACTIONS(610),
    [aux_sym_comparison_operator_token47] = ACTIONS(610),
    [aux_sym_comparison_operator_token48] = ACTIONS(610),
    [aux_sym_comparison_operator_token49] = ACTIONS(610),
    [aux_sym_comparison_operator_token50] = ACTIONS(610),
    [aux_sym_format_operator_token1] = ACTIONS(610),
    [anon_sym_LPAREN] = ACTIONS(610),
    [anon_sym_RPAREN] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(610),
    [anon_sym_AMP_AMP] = ACTIONS(610),
    [anon_sym_PIPE_PIPE] = ACTIONS(610),
    [anon_sym_PIPE] = ACTIONS(612),
    [aux_sym_logical_expression_token1] = ACTIONS(610),
    [aux_sym_logical_expression_token2] = ACTIONS(610),
    [aux_sym_logical_expression_token3] = ACTIONS(610),
    [aux_sym_bitwise_expression_token1] = ACTIONS(610),
    [aux_sym_bitwise_expression_token2] = ACTIONS(610),
    [aux_sym_bitwise_expression_token3] = ACTIONS(610),
    [anon_sym_PLUS] = ACTIONS(612),
    [anon_sym_DASH] = ACTIONS(612),
    [anon_sym_SLASH] = ACTIONS(612),
    [anon_sym_BSLASH] = ACTIONS(610),
    [anon_sym_PERCENT] = ACTIONS(612),
    [anon_sym_STAR] = ACTIONS(612),
    [anon_sym_DOT_DOT] = ACTIONS(610),
    [anon_sym_PLUS_PLUS] = ACTIONS(610),
    [anon_sym_DASH_DASH] = ACTIONS(610),
    [anon_sym_DOT2] = ACTIONS(612),
    [anon_sym_COLON_COLON] = ACTIONS(610),
    [anon_sym_RBRACK] = ACTIONS(610),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(610),
  },
  [STATE(94)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(614),
    [anon_sym_EQ] = ACTIONS(614),
    [anon_sym_BANG_EQ] = ACTIONS(614),
    [anon_sym_PLUS_EQ] = ACTIONS(614),
    [anon_sym_STAR_EQ] = ACTIONS(614),
    [anon_sym_SLASH_EQ] = ACTIONS(614),
    [anon_sym_PERCENT_EQ] = ACTIONS(614),
    [anon_sym_DASH_EQ] = ACTIONS(614),
    [anon_sym_GT] = ACTIONS(616),
    [anon_sym_GT_GT] = ACTIONS(614),
    [anon_sym_2_GT] = ACTIONS(616),
    [anon_sym_2_GT_GT] = ACTIONS(614),
    [anon_sym_3_GT] = ACTIONS(616),
    [anon_sym_3_GT_GT] = ACTIONS(614),
    [anon_sym_4_GT] = ACTIONS(616),
    [anon_sym_4_GT_GT] = ACTIONS(614),
    [anon_sym_5_GT] = ACTIONS(616),
    [anon_sym_5_GT_GT] = ACTIONS(614),
    [anon_sym_6_GT] = ACTIONS(616),
    [anon_sym_6_GT_GT] = ACTIONS(614),
    [anon_sym_STAR_GT] = ACTIONS(616),
    [anon_sym_STAR_GT_GT] = ACTIONS(614),
    [anon_sym_LT] = ACTIONS(616),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(614),
    [anon_sym_2_GT_AMP1] = ACTIONS(614),
    [anon_sym_3_GT_AMP1] = ACTIONS(614),
    [anon_sym_4_GT_AMP1] = ACTIONS(614),
    [anon_sym_5_GT_AMP1] = ACTIONS(614),
    [anon_sym_6_GT_AMP1] = ACTIONS(614),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(614),
    [anon_sym_1_GT_AMP2] = ACTIONS(614),
    [anon_sym_3_GT_AMP2] = ACTIONS(614),
    [anon_sym_4_GT_AMP2] = ACTIONS(614),
    [anon_sym_5_GT_AMP2] = ACTIONS(614),
    [anon_sym_6_GT_AMP2] = ACTIONS(614),
    [aux_sym_comparison_operator_token1] = ACTIONS(614),
    [aux_sym_comparison_operator_token2] = ACTIONS(614),
    [aux_sym_comparison_operator_token3] = ACTIONS(614),
    [aux_sym_comparison_operator_token4] = ACTIONS(614),
    [aux_sym_comparison_operator_token5] = ACTIONS(614),
    [aux_sym_comparison_operator_token6] = ACTIONS(614),
    [aux_sym_comparison_operator_token7] = ACTIONS(614),
    [aux_sym_comparison_operator_token8] = ACTIONS(614),
    [aux_sym_comparison_operator_token9] = ACTIONS(614),
    [aux_sym_comparison_operator_token10] = ACTIONS(614),
    [aux_sym_comparison_operator_token11] = ACTIONS(614),
    [aux_sym_comparison_operator_token12] = ACTIONS(614),
    [aux_sym_comparison_operator_token13] = ACTIONS(614),
    [aux_sym_comparison_operator_token14] = ACTIONS(614),
    [aux_sym_comparison_operator_token15] = ACTIONS(614),
    [aux_sym_comparison_operator_token16] = ACTIONS(614),
    [aux_sym_comparison_operator_token17] = ACTIONS(614),
    [aux_sym_comparison_operator_token18] = ACTIONS(614),
    [aux_sym_comparison_operator_token19] = ACTIONS(614),
    [aux_sym_comparison_operator_token20] = ACTIONS(614),
    [aux_sym_comparison_operator_token21] = ACTIONS(614),
    [aux_sym_comparison_operator_token22] = ACTIONS(614),
    [aux_sym_comparison_operator_token23] = ACTIONS(614),
    [aux_sym_comparison_operator_token24] = ACTIONS(614),
    [aux_sym_comparison_operator_token25] = ACTIONS(614),
    [aux_sym_comparison_operator_token26] = ACTIONS(614),
    [aux_sym_comparison_operator_token27] = ACTIONS(614),
    [aux_sym_comparison_operator_token28] = ACTIONS(616),
    [aux_sym_comparison_operator_token29] = ACTIONS(614),
    [aux_sym_comparison_operator_token30] = ACTIONS(614),
    [aux_sym_comparison_operator_token31] = ACTIONS(614),
    [aux_sym_comparison_operator_token32] = ACTIONS(614),
    [aux_sym_comparison_operator_token33] = ACTIONS(614),
    [aux_sym_comparison_operator_token34] = ACTIONS(616),
    [aux_sym_comparison_operator_token35] = ACTIONS(614),
    [aux_sym_comparison_operator_token36] = ACTIONS(614),
    [aux_sym_comparison_operator_token37] = ACTIONS(614),
    [aux_sym_comparison_operator_token38] = ACTIONS(614),
    [aux_sym_comparison_operator_token39] = ACTIONS(614),
    [aux_sym_comparison_operator_token40] = ACTIONS(614),
    [aux_sym_comparison_operator_token41] = ACTIONS(614),
    [aux_sym_comparison_operator_token42] = ACTIONS(614),
    [aux_sym_comparison_operator_token43] = ACTIONS(614),
    [aux_sym_comparison_operator_token44] = ACTIONS(614),
    [aux_sym_comparison_operator_token45] = ACTIONS(614),
    [aux_sym_comparison_operator_token46] = ACTIONS(614),
    [aux_sym_comparison_operator_token47] = ACTIONS(614),
    [aux_sym_comparison_operator_token48] = ACTIONS(614),
    [aux_sym_comparison_operator_token49] = ACTIONS(614),
    [aux_sym_comparison_operator_token50] = ACTIONS(614),
    [aux_sym_format_operator_token1] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(614),
    [anon_sym_RPAREN] = ACTIONS(614),
    [anon_sym_COMMA] = ACTIONS(614),
    [anon_sym_AMP_AMP] = ACTIONS(614),
    [anon_sym_PIPE_PIPE] = ACTIONS(614),
    [anon_sym_PIPE] = ACTIONS(616),
    [aux_sym_logical_expression_token1] = ACTIONS(614),
    [aux_sym_logical_expression_token2] = ACTIONS(614),
    [aux_sym_logical_expression_token3] = ACTIONS(614),
    [aux_sym_bitwise_expression_token1] = ACTIONS(614),
    [aux_sym_bitwise_expression_token2] = ACTIONS(614),
    [aux_sym_bitwise_expression_token3] = ACTIONS(614),
    [anon_sym_PLUS] = ACTIONS(616),
    [anon_sym_DASH] = ACTIONS(616),
    [anon_sym_SLASH] = ACTIONS(616),
    [anon_sym_BSLASH] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_STAR] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(614),
    [anon_sym_PLUS_PLUS] = ACTIONS(614),
    [anon_sym_DASH_DASH] = ACTIONS(614),
    [anon_sym_DOT2] = ACTIONS(616),
    [anon_sym_COLON_COLON] = ACTIONS(614),
    [anon_sym_RBRACK] = ACTIONS(614),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(614),
  },
  [STATE(95)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(618),
    [anon_sym_EQ] = ACTIONS(618),
    [anon_sym_BANG_EQ] = ACTIONS(618),
    [anon_sym_PLUS_EQ] = ACTIONS(618),
    [anon_sym_STAR_EQ] = ACTIONS(618),
    [anon_sym_SLASH_EQ] = ACTIONS(618),
    [anon_sym_PERCENT_EQ] = ACTIONS(618),
    [anon_sym_DASH_EQ] = ACTIONS(618),
    [anon_sym_GT] = ACTIONS(620),
    [anon_sym_GT_GT] = ACTIONS(618),
    [anon_sym_2_GT] = ACTIONS(620),
    [anon_sym_2_GT_GT] = ACTIONS(618),
    [anon_sym_3_GT] = ACTIONS(620),
    [anon_sym_3_GT_GT] = ACTIONS(618),
    [anon_sym_4_GT] = ACTIONS(620),
    [anon_sym_4_GT_GT] = ACTIONS(618),
    [anon_sym_5_GT] = ACTIONS(620),
    [anon_sym_5_GT_GT] = ACTIONS(618),
    [anon_sym_6_GT] = ACTIONS(620),
    [anon_sym_6_GT_GT] = ACTIONS(618),
    [anon_sym_STAR_GT] = ACTIONS(620),
    [anon_sym_STAR_GT_GT] = ACTIONS(618),
    [anon_sym_LT] = ACTIONS(620),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(618),
    [anon_sym_2_GT_AMP1] = ACTIONS(618),
    [anon_sym_3_GT_AMP1] = ACTIONS(618),
    [anon_sym_4_GT_AMP1] = ACTIONS(618),
    [anon_sym_5_GT_AMP1] = ACTIONS(618),
    [anon_sym_6_GT_AMP1] = ACTIONS(618),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(618),
    [anon_sym_1_GT_AMP2] = ACTIONS(618),
    [anon_sym_3_GT_AMP2] = ACTIONS(618),
    [anon_sym_4_GT_AMP2] = ACTIONS(618),
    [anon_sym_5_GT_AMP2] = ACTIONS(618),
    [anon_sym_6_GT_AMP2] = ACTIONS(618),
    [aux_sym_comparison_operator_token1] = ACTIONS(618),
    [aux_sym_comparison_operator_token2] = ACTIONS(618),
    [aux_sym_comparison_operator_token3] = ACTIONS(618),
    [aux_sym_comparison_operator_token4] = ACTIONS(618),
    [aux_sym_comparison_operator_token5] = ACTIONS(618),
    [aux_sym_comparison_operator_token6] = ACTIONS(618),
    [aux_sym_comparison_operator_token7] = ACTIONS(618),
    [aux_sym_comparison_operator_token8] = ACTIONS(618),
    [aux_sym_comparison_operator_token9] = ACTIONS(618),
    [aux_sym_comparison_operator_token10] = ACTIONS(618),
    [aux_sym_comparison_operator_token11] = ACTIONS(618),
    [aux_sym_comparison_operator_token12] = ACTIONS(618),
    [aux_sym_comparison_operator_token13] = ACTIONS(618),
    [aux_sym_comparison_operator_token14] = ACTIONS(618),
    [aux_sym_comparison_operator_token15] = ACTIONS(618),
    [aux_sym_comparison_operator_token16] = ACTIONS(618),
    [aux_sym_comparison_operator_token17] = ACTIONS(618),
    [aux_sym_comparison_operator_token18] = ACTIONS(618),
    [aux_sym_comparison_operator_token19] = ACTIONS(618),
    [aux_sym_comparison_operator_token20] = ACTIONS(618),
    [aux_sym_comparison_operator_token21] = ACTIONS(618),
    [aux_sym_comparison_operator_token22] = ACTIONS(618),
    [aux_sym_comparison_operator_token23] = ACTIONS(618),
    [aux_sym_comparison_operator_token24] = ACTIONS(618),
    [aux_sym_comparison_operator_token25] = ACTIONS(618),
    [aux_sym_comparison_operator_token26] = ACTIONS(618),
    [aux_sym_comparison_operator_token27] = ACTIONS(618),
    [aux_sym_comparison_operator_token28] = ACTIONS(620),
    [aux_sym_comparison_operator_token29] = ACTIONS(618),
    [aux_sym_comparison_operator_token30] = ACTIONS(618),
    [aux_sym_comparison_operator_token31] = ACTIONS(618),
    [aux_sym_comparison_operator_token32] = ACTIONS(618),
    [aux_sym_comparison_operator_token33] = ACTIONS(618),
    [aux_sym_comparison_operator_token34] = ACTIONS(620),
    [aux_sym_comparison_operator_token35] = ACTIONS(618),
    [aux_sym_comparison_operator_token36] = ACTIONS(618),
    [aux_sym_comparison_operator_token37] = ACTIONS(618),
    [aux_sym_comparison_operator_token38] = ACTIONS(618),
    [aux_sym_comparison_operator_token39] = ACTIONS(618),
    [aux_sym_comparison_operator_token40] = ACTIONS(618),
    [aux_sym_comparison_operator_token41] = ACTIONS(618),
    [aux_sym_comparison_operator_token42] = ACTIONS(618),
    [aux_sym_comparison_operator_token43] = ACTIONS(618),
    [aux_sym_comparison_operator_token44] = ACTIONS(618),
    [aux_sym_comparison_operator_token45] = ACTIONS(618),
    [aux_sym_comparison_operator_token46] = ACTIONS(618),
    [aux_sym_comparison_operator_token47] = ACTIONS(618),
    [aux_sym_comparison_operator_token48] = ACTIONS(618),
    [aux_sym_comparison_operator_token49] = ACTIONS(618),
    [aux_sym_comparison_operator_token50] = ACTIONS(618),
    [aux_sym_format_operator_token1] = ACTIONS(618),
    [anon_sym_LPAREN] = ACTIONS(618),
    [anon_sym_RPAREN] = ACTIONS(618),
    [anon_sym_COMMA] = ACTIONS(618),
    [anon_sym_AMP_AMP] = ACTIONS(618),
    [anon_sym_PIPE_PIPE] = ACTIONS(618),
    [anon_sym_PIPE] = ACTIONS(620),
    [aux_sym_logical_expression_token1] = ACTIONS(618),
    [aux_sym_logical_expression_token2] = ACTIONS(618),
    [aux_sym_logical_expression_token3] = ACTIONS(618),
    [aux_sym_bitwise_expression_token1] = ACTIONS(618),
    [aux_sym_bitwise_expression_token2] = ACTIONS(618),
    [aux_sym_bitwise_expression_token3] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_SLASH] = ACTIONS(620),
    [anon_sym_BSLASH] = ACTIONS(618),
    [anon_sym_PERCENT] = ACTIONS(620),
    [anon_sym_STAR] = ACTIONS(620),
    [anon_sym_DOT_DOT] = ACTIONS(618),
    [anon_sym_PLUS_PLUS] = ACTIONS(618),
    [anon_sym_DASH_DASH] = ACTIONS(618),
    [anon_sym_DOT2] = ACTIONS(620),
    [anon_sym_COLON_COLON] = ACTIONS(618),
    [anon_sym_RBRACK] = ACTIONS(618),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(618),
  },
  [STATE(96)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(622),
    [anon_sym_EQ] = ACTIONS(622),
    [anon_sym_BANG_EQ] = ACTIONS(622),
    [anon_sym_PLUS_EQ] = ACTIONS(622),
    [anon_sym_STAR_EQ] = ACTIONS(622),
    [anon_sym_SLASH_EQ] = ACTIONS(622),
    [anon_sym_PERCENT_EQ] = ACTIONS(622),
    [anon_sym_DASH_EQ] = ACTIONS(622),
    [anon_sym_GT] = ACTIONS(624),
    [anon_sym_GT_GT] = ACTIONS(622),
    [anon_sym_2_GT] = ACTIONS(624),
    [anon_sym_2_GT_GT] = ACTIONS(622),
    [anon_sym_3_GT] = ACTIONS(624),
    [anon_sym_3_GT_GT] = ACTIONS(622),
    [anon_sym_4_GT] = ACTIONS(624),
    [anon_sym_4_GT_GT] = ACTIONS(622),
    [anon_sym_5_GT] = ACTIONS(624),
    [anon_sym_5_GT_GT] = ACTIONS(622),
    [anon_sym_6_GT] = ACTIONS(624),
    [anon_sym_6_GT_GT] = ACTIONS(622),
    [anon_sym_STAR_GT] = ACTIONS(624),
    [anon_sym_STAR_GT_GT] = ACTIONS(622),
    [anon_sym_LT] = ACTIONS(624),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(622),
    [anon_sym_2_GT_AMP1] = ACTIONS(622),
    [anon_sym_3_GT_AMP1] = ACTIONS(622),
    [anon_sym_4_GT_AMP1] = ACTIONS(622),
    [anon_sym_5_GT_AMP1] = ACTIONS(622),
    [anon_sym_6_GT_AMP1] = ACTIONS(622),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(622),
    [anon_sym_1_GT_AMP2] = ACTIONS(622),
    [anon_sym_3_GT_AMP2] = ACTIONS(622),
    [anon_sym_4_GT_AMP2] = ACTIONS(622),
    [anon_sym_5_GT_AMP2] = ACTIONS(622),
    [anon_sym_6_GT_AMP2] = ACTIONS(622),
    [aux_sym_comparison_operator_token1] = ACTIONS(622),
    [aux_sym_comparison_operator_token2] = ACTIONS(622),
    [aux_sym_comparison_operator_token3] = ACTIONS(622),
    [aux_sym_comparison_operator_token4] = ACTIONS(622),
    [aux_sym_comparison_operator_token5] = ACTIONS(622),
    [aux_sym_comparison_operator_token6] = ACTIONS(622),
    [aux_sym_comparison_operator_token7] = ACTIONS(622),
    [aux_sym_comparison_operator_token8] = ACTIONS(622),
    [aux_sym_comparison_operator_token9] = ACTIONS(622),
    [aux_sym_comparison_operator_token10] = ACTIONS(622),
    [aux_sym_comparison_operator_token11] = ACTIONS(622),
    [aux_sym_comparison_operator_token12] = ACTIONS(622),
    [aux_sym_comparison_operator_token13] = ACTIONS(622),
    [aux_sym_comparison_operator_token14] = ACTIONS(622),
    [aux_sym_comparison_operator_token15] = ACTIONS(622),
    [aux_sym_comparison_operator_token16] = ACTIONS(622),
    [aux_sym_comparison_operator_token17] = ACTIONS(622),
    [aux_sym_comparison_operator_token18] = ACTIONS(622),
    [aux_sym_comparison_operator_token19] = ACTIONS(622),
    [aux_sym_comparison_operator_token20] = ACTIONS(622),
    [aux_sym_comparison_operator_token21] = ACTIONS(622),
    [aux_sym_comparison_operator_token22] = ACTIONS(622),
    [aux_sym_comparison_operator_token23] = ACTIONS(622),
    [aux_sym_comparison_operator_token24] = ACTIONS(622),
    [aux_sym_comparison_operator_token25] = ACTIONS(622),
    [aux_sym_comparison_operator_token26] = ACTIONS(622),
    [aux_sym_comparison_operator_token27] = ACTIONS(622),
    [aux_sym_comparison_operator_token28] = ACTIONS(624),
    [aux_sym_comparison_operator_token29] = ACTIONS(622),
    [aux_sym_comparison_operator_token30] = ACTIONS(622),
    [aux_sym_comparison_operator_token31] = ACTIONS(622),
    [aux_sym_comparison_operator_token32] = ACTIONS(622),
    [aux_sym_comparison_operator_token33] = ACTIONS(622),
    [aux_sym_comparison_operator_token34] = ACTIONS(624),
    [aux_sym_comparison_operator_token35] = ACTIONS(622),
    [aux_sym_comparison_operator_token36] = ACTIONS(622),
    [aux_sym_comparison_operator_token37] = ACTIONS(622),
    [aux_sym_comparison_operator_token38] = ACTIONS(622),
    [aux_sym_comparison_operator_token39] = ACTIONS(622),
    [aux_sym_comparison_operator_token40] = ACTIONS(622),
    [aux_sym_comparison_operator_token41] = ACTIONS(622),
    [aux_sym_comparison_operator_token42] = ACTIONS(622),
    [aux_sym_comparison_operator_token43] = ACTIONS(622),
    [aux_sym_comparison_operator_token44] = ACTIONS(622),
    [aux_sym_comparison_operator_token45] = ACTIONS(622),
    [aux_sym_comparison_operator_token46] = ACTIONS(622),
    [aux_sym_comparison_operator_token47] = ACTIONS(622),
    [aux_sym_comparison_operator_token48] = ACTIONS(622),
    [aux_sym_comparison_operator_token49] = ACTIONS(622),
    [aux_sym_comparison_operator_token50] = ACTIONS(622),
    [aux_sym_format_operator_token1] = ACTIONS(622),
    [anon_sym_LPAREN] = ACTIONS(622),
    [anon_sym_RPAREN] = ACTIONS(622),
    [anon_sym_COMMA] = ACTIONS(622),
    [anon_sym_AMP_AMP] = ACTIONS(622),
    [anon_sym_PIPE_PIPE] = ACTIONS(622),
    [anon_sym_PIPE] = ACTIONS(624),
    [aux_sym_logical_expression_token1] = ACTIONS(622),
    [aux_sym_logical_expression_token2] = ACTIONS(622),
    [aux_sym_logical_expression_token3] = ACTIONS(622),
    [aux_sym_bitwise_expression_token1] = ACTIONS(622),
    [aux_sym_bitwise_expression_token2] = ACTIONS(622),
    [aux_sym_bitwise_expression_token3] = ACTIONS(622),
    [anon_sym_PLUS] = ACTIONS(624),
    [anon_sym_DASH] = ACTIONS(624),
    [anon_sym_SLASH] = ACTIONS(624),
    [anon_sym_BSLASH] = ACTIONS(622),
    [anon_sym_PERCENT] = ACTIONS(624),
    [anon_sym_STAR] = ACTIONS(624),
    [anon_sym_DOT_DOT] = ACTIONS(622),
    [anon_sym_PLUS_PLUS] = ACTIONS(622),
    [anon_sym_DASH_DASH] = ACTIONS(622),
    [anon_sym_DOT2] = ACTIONS(624),
    [anon_sym_COLON_COLON] = ACTIONS(622),
    [anon_sym_RBRACK] = ACTIONS(622),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(622),
  },
  [STATE(97)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(626),
    [anon_sym_EQ] = ACTIONS(626),
    [anon_sym_BANG_EQ] = ACTIONS(626),
    [anon_sym_PLUS_EQ] = ACTIONS(626),
    [anon_sym_STAR_EQ] = ACTIONS(626),
    [anon_sym_SLASH_EQ] = ACTIONS(626),
    [anon_sym_PERCENT_EQ] = ACTIONS(626),
    [anon_sym_DASH_EQ] = ACTIONS(626),
    [anon_sym_GT] = ACTIONS(628),
    [anon_sym_GT_GT] = ACTIONS(626),
    [anon_sym_2_GT] = ACTIONS(628),
    [anon_sym_2_GT_GT] = ACTIONS(626),
    [anon_sym_3_GT] = ACTIONS(628),
    [anon_sym_3_GT_GT] = ACTIONS(626),
    [anon_sym_4_GT] = ACTIONS(628),
    [anon_sym_4_GT_GT] = ACTIONS(626),
    [anon_sym_5_GT] = ACTIONS(628),
    [anon_sym_5_GT_GT] = ACTIONS(626),
    [anon_sym_6_GT] = ACTIONS(628),
    [anon_sym_6_GT_GT] = ACTIONS(626),
    [anon_sym_STAR_GT] = ACTIONS(628),
    [anon_sym_STAR_GT_GT] = ACTIONS(626),
    [anon_sym_LT] = ACTIONS(628),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(626),
    [anon_sym_2_GT_AMP1] = ACTIONS(626),
    [anon_sym_3_GT_AMP1] = ACTIONS(626),
    [anon_sym_4_GT_AMP1] = ACTIONS(626),
    [anon_sym_5_GT_AMP1] = ACTIONS(626),
    [anon_sym_6_GT_AMP1] = ACTIONS(626),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(626),
    [anon_sym_1_GT_AMP2] = ACTIONS(626),
    [anon_sym_3_GT_AMP2] = ACTIONS(626),
    [anon_sym_4_GT_AMP2] = ACTIONS(626),
    [anon_sym_5_GT_AMP2] = ACTIONS(626),
    [anon_sym_6_GT_AMP2] = ACTIONS(626),
    [aux_sym_comparison_operator_token1] = ACTIONS(626),
    [aux_sym_comparison_operator_token2] = ACTIONS(626),
    [aux_sym_comparison_operator_token3] = ACTIONS(626),
    [aux_sym_comparison_operator_token4] = ACTIONS(626),
    [aux_sym_comparison_operator_token5] = ACTIONS(626),
    [aux_sym_comparison_operator_token6] = ACTIONS(626),
    [aux_sym_comparison_operator_token7] = ACTIONS(626),
    [aux_sym_comparison_operator_token8] = ACTIONS(626),
    [aux_sym_comparison_operator_token9] = ACTIONS(626),
    [aux_sym_comparison_operator_token10] = ACTIONS(626),
    [aux_sym_comparison_operator_token11] = ACTIONS(626),
    [aux_sym_comparison_operator_token12] = ACTIONS(626),
    [aux_sym_comparison_operator_token13] = ACTIONS(626),
    [aux_sym_comparison_operator_token14] = ACTIONS(626),
    [aux_sym_comparison_operator_token15] = ACTIONS(626),
    [aux_sym_comparison_operator_token16] = ACTIONS(626),
    [aux_sym_comparison_operator_token17] = ACTIONS(626),
    [aux_sym_comparison_operator_token18] = ACTIONS(626),
    [aux_sym_comparison_operator_token19] = ACTIONS(626),
    [aux_sym_comparison_operator_token20] = ACTIONS(626),
    [aux_sym_comparison_operator_token21] = ACTIONS(626),
    [aux_sym_comparison_operator_token22] = ACTIONS(626),
    [aux_sym_comparison_operator_token23] = ACTIONS(626),
    [aux_sym_comparison_operator_token24] = ACTIONS(626),
    [aux_sym_comparison_operator_token25] = ACTIONS(626),
    [aux_sym_comparison_operator_token26] = ACTIONS(626),
    [aux_sym_comparison_operator_token27] = ACTIONS(626),
    [aux_sym_comparison_operator_token28] = ACTIONS(628),
    [aux_sym_comparison_operator_token29] = ACTIONS(626),
    [aux_sym_comparison_operator_token30] = ACTIONS(626),
    [aux_sym_comparison_operator_token31] = ACTIONS(626),
    [aux_sym_comparison_operator_token32] = ACTIONS(626),
    [aux_sym_comparison_operator_token33] = ACTIONS(626),
    [aux_sym_comparison_operator_token34] = ACTIONS(628),
    [aux_sym_comparison_operator_token35] = ACTIONS(626),
    [aux_sym_comparison_operator_token36] = ACTIONS(626),
    [aux_sym_comparison_operator_token37] = ACTIONS(626),
    [aux_sym_comparison_operator_token38] = ACTIONS(626),
    [aux_sym_comparison_operator_token39] = ACTIONS(626),
    [aux_sym_comparison_operator_token40] = ACTIONS(626),
    [aux_sym_comparison_operator_token41] = ACTIONS(626),
    [aux_sym_comparison_operator_token42] = ACTIONS(626),
    [aux_sym_comparison_operator_token43] = ACTIONS(626),
    [aux_sym_comparison_operator_token44] = ACTIONS(626),
    [aux_sym_comparison_operator_token45] = ACTIONS(626),
    [aux_sym_comparison_operator_token46] = ACTIONS(626),
    [aux_sym_comparison_operator_token47] = ACTIONS(626),
    [aux_sym_comparison_operator_token48] = ACTIONS(626),
    [aux_sym_comparison_operator_token49] = ACTIONS(626),
    [aux_sym_comparison_operator_token50] = ACTIONS(626),
    [aux_sym_format_operator_token1] = ACTIONS(626),
    [anon_sym_LPAREN] = ACTIONS(626),
    [anon_sym_RPAREN] = ACTIONS(626),
    [anon_sym_COMMA] = ACTIONS(626),
    [anon_sym_AMP_AMP] = ACTIONS(626),
    [anon_sym_PIPE_PIPE] = ACTIONS(626),
    [anon_sym_PIPE] = ACTIONS(628),
    [aux_sym_logical_expression_token1] = ACTIONS(626),
    [aux_sym_logical_expression_token2] = ACTIONS(626),
    [aux_sym_logical_expression_token3] = ACTIONS(626),
    [aux_sym_bitwise_expression_token1] = ACTIONS(626),
    [aux_sym_bitwise_expression_token2] = ACTIONS(626),
    [aux_sym_bitwise_expression_token3] = ACTIONS(626),
    [anon_sym_PLUS] = ACTIONS(628),
    [anon_sym_DASH] = ACTIONS(628),
    [anon_sym_SLASH] = ACTIONS(628),
    [anon_sym_BSLASH] = ACTIONS(626),
    [anon_sym_PERCENT] = ACTIONS(628),
    [anon_sym_STAR] = ACTIONS(628),
    [anon_sym_DOT_DOT] = ACTIONS(626),
    [anon_sym_PLUS_PLUS] = ACTIONS(626),
    [anon_sym_DASH_DASH] = ACTIONS(626),
    [anon_sym_DOT2] = ACTIONS(628),
    [anon_sym_COLON_COLON] = ACTIONS(626),
    [anon_sym_RBRACK] = ACTIONS(626),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(626),
  },
  [STATE(98)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(630),
    [anon_sym_EQ] = ACTIONS(630),
    [anon_sym_BANG_EQ] = ACTIONS(630),
    [anon_sym_PLUS_EQ] = ACTIONS(630),
    [anon_sym_STAR_EQ] = ACTIONS(630),
    [anon_sym_SLASH_EQ] = ACTIONS(630),
    [anon_sym_PERCENT_EQ] = ACTIONS(630),
    [anon_sym_DASH_EQ] = ACTIONS(630),
    [anon_sym_GT] = ACTIONS(632),
    [anon_sym_GT_GT] = ACTIONS(630),
    [anon_sym_2_GT] = ACTIONS(632),
    [anon_sym_2_GT_GT] = ACTIONS(630),
    [anon_sym_3_GT] = ACTIONS(632),
    [anon_sym_3_GT_GT] = ACTIONS(630),
    [anon_sym_4_GT] = ACTIONS(632),
    [anon_sym_4_GT_GT] = ACTIONS(630),
    [anon_sym_5_GT] = ACTIONS(632),
    [anon_sym_5_GT_GT] = ACTIONS(630),
    [anon_sym_6_GT] = ACTIONS(632),
    [anon_sym_6_GT_GT] = ACTIONS(630),
    [anon_sym_STAR_GT] = ACTIONS(632),
    [anon_sym_STAR_GT_GT] = ACTIONS(630),
    [anon_sym_LT] = ACTIONS(632),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(630),
    [anon_sym_2_GT_AMP1] = ACTIONS(630),
    [anon_sym_3_GT_AMP1] = ACTIONS(630),
    [anon_sym_4_GT_AMP1] = ACTIONS(630),
    [anon_sym_5_GT_AMP1] = ACTIONS(630),
    [anon_sym_6_GT_AMP1] = ACTIONS(630),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(630),
    [anon_sym_1_GT_AMP2] = ACTIONS(630),
    [anon_sym_3_GT_AMP2] = ACTIONS(630),
    [anon_sym_4_GT_AMP2] = ACTIONS(630),
    [anon_sym_5_GT_AMP2] = ACTIONS(630),
    [anon_sym_6_GT_AMP2] = ACTIONS(630),
    [aux_sym_comparison_operator_token1] = ACTIONS(630),
    [aux_sym_comparison_operator_token2] = ACTIONS(630),
    [aux_sym_comparison_operator_token3] = ACTIONS(630),
    [aux_sym_comparison_operator_token4] = ACTIONS(630),
    [aux_sym_comparison_operator_token5] = ACTIONS(630),
    [aux_sym_comparison_operator_token6] = ACTIONS(630),
    [aux_sym_comparison_operator_token7] = ACTIONS(630),
    [aux_sym_comparison_operator_token8] = ACTIONS(630),
    [aux_sym_comparison_operator_token9] = ACTIONS(630),
    [aux_sym_comparison_operator_token10] = ACTIONS(630),
    [aux_sym_comparison_operator_token11] = ACTIONS(630),
    [aux_sym_comparison_operator_token12] = ACTIONS(630),
    [aux_sym_comparison_operator_token13] = ACTIONS(630),
    [aux_sym_comparison_operator_token14] = ACTIONS(630),
    [aux_sym_comparison_operator_token15] = ACTIONS(630),
    [aux_sym_comparison_operator_token16] = ACTIONS(630),
    [aux_sym_comparison_operator_token17] = ACTIONS(630),
    [aux_sym_comparison_operator_token18] = ACTIONS(630),
    [aux_sym_comparison_operator_token19] = ACTIONS(630),
    [aux_sym_comparison_operator_token20] = ACTIONS(630),
    [aux_sym_comparison_operator_token21] = ACTIONS(630),
    [aux_sym_comparison_operator_token22] = ACTIONS(630),
    [aux_sym_comparison_operator_token23] = ACTIONS(630),
    [aux_sym_comparison_operator_token24] = ACTIONS(630),
    [aux_sym_comparison_operator_token25] = ACTIONS(630),
    [aux_sym_comparison_operator_token26] = ACTIONS(630),
    [aux_sym_comparison_operator_token27] = ACTIONS(630),
    [aux_sym_comparison_operator_token28] = ACTIONS(632),
    [aux_sym_comparison_operator_token29] = ACTIONS(630),
    [aux_sym_comparison_operator_token30] = ACTIONS(630),
    [aux_sym_comparison_operator_token31] = ACTIONS(630),
    [aux_sym_comparison_operator_token32] = ACTIONS(630),
    [aux_sym_comparison_operator_token33] = ACTIONS(630),
    [aux_sym_comparison_operator_token34] = ACTIONS(632),
    [aux_sym_comparison_operator_token35] = ACTIONS(630),
    [aux_sym_comparison_operator_token36] = ACTIONS(630),
    [aux_sym_comparison_operator_token37] = ACTIONS(630),
    [aux_sym_comparison_operator_token38] = ACTIONS(630),
    [aux_sym_comparison_operator_token39] = ACTIONS(630),
    [aux_sym_comparison_operator_token40] = ACTIONS(630),
    [aux_sym_comparison_operator_token41] = ACTIONS(630),
    [aux_sym_comparison_operator_token42] = ACTIONS(630),
    [aux_sym_comparison_operator_token43] = ACTIONS(630),
    [aux_sym_comparison_operator_token44] = ACTIONS(630),
    [aux_sym_comparison_operator_token45] = ACTIONS(630),
    [aux_sym_comparison_operator_token46] = ACTIONS(630),
    [aux_sym_comparison_operator_token47] = ACTIONS(630),
    [aux_sym_comparison_operator_token48] = ACTIONS(630),
    [aux_sym_comparison_operator_token49] = ACTIONS(630),
    [aux_sym_comparison_operator_token50] = ACTIONS(630),
    [aux_sym_format_operator_token1] = ACTIONS(630),
    [anon_sym_LPAREN] = ACTIONS(630),
    [anon_sym_RPAREN] = ACTIONS(630),
    [anon_sym_COMMA] = ACTIONS(630),
    [anon_sym_AMP_AMP] = ACTIONS(630),
    [anon_sym_PIPE_PIPE] = ACTIONS(630),
    [anon_sym_PIPE] = ACTIONS(632),
    [aux_sym_logical_expression_token1] = ACTIONS(630),
    [aux_sym_logical_expression_token2] = ACTIONS(630),
    [aux_sym_logical_expression_token3] = ACTIONS(630),
    [aux_sym_bitwise_expression_token1] = ACTIONS(630),
    [aux_sym_bitwise_expression_token2] = ACTIONS(630),
    [aux_sym_bitwise_expression_token3] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(632),
    [anon_sym_DASH] = ACTIONS(632),
    [anon_sym_SLASH] = ACTIONS(632),
    [anon_sym_BSLASH] = ACTIONS(630),
    [anon_sym_PERCENT] = ACTIONS(632),
    [anon_sym_STAR] = ACTIONS(632),
    [anon_sym_DOT_DOT] = ACTIONS(630),
    [anon_sym_PLUS_PLUS] = ACTIONS(630),
    [anon_sym_DASH_DASH] = ACTIONS(630),
    [anon_sym_DOT2] = ACTIONS(632),
    [anon_sym_COLON_COLON] = ACTIONS(630),
    [anon_sym_RBRACK] = ACTIONS(630),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(630),
  },
  [STATE(99)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(634),
    [anon_sym_EQ] = ACTIONS(634),
    [anon_sym_BANG_EQ] = ACTIONS(634),
    [anon_sym_PLUS_EQ] = ACTIONS(634),
    [anon_sym_STAR_EQ] = ACTIONS(634),
    [anon_sym_SLASH_EQ] = ACTIONS(634),
    [anon_sym_PERCENT_EQ] = ACTIONS(634),
    [anon_sym_DASH_EQ] = ACTIONS(634),
    [anon_sym_GT] = ACTIONS(636),
    [anon_sym_GT_GT] = ACTIONS(634),
    [anon_sym_2_GT] = ACTIONS(636),
    [anon_sym_2_GT_GT] = ACTIONS(634),
    [anon_sym_3_GT] = ACTIONS(636),
    [anon_sym_3_GT_GT] = ACTIONS(634),
    [anon_sym_4_GT] = ACTIONS(636),
    [anon_sym_4_GT_GT] = ACTIONS(634),
    [anon_sym_5_GT] = ACTIONS(636),
    [anon_sym_5_GT_GT] = ACTIONS(634),
    [anon_sym_6_GT] = ACTIONS(636),
    [anon_sym_6_GT_GT] = ACTIONS(634),
    [anon_sym_STAR_GT] = ACTIONS(636),
    [anon_sym_STAR_GT_GT] = ACTIONS(634),
    [anon_sym_LT] = ACTIONS(636),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(634),
    [anon_sym_2_GT_AMP1] = ACTIONS(634),
    [anon_sym_3_GT_AMP1] = ACTIONS(634),
    [anon_sym_4_GT_AMP1] = ACTIONS(634),
    [anon_sym_5_GT_AMP1] = ACTIONS(634),
    [anon_sym_6_GT_AMP1] = ACTIONS(634),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(634),
    [anon_sym_1_GT_AMP2] = ACTIONS(634),
    [anon_sym_3_GT_AMP2] = ACTIONS(634),
    [anon_sym_4_GT_AMP2] = ACTIONS(634),
    [anon_sym_5_GT_AMP2] = ACTIONS(634),
    [anon_sym_6_GT_AMP2] = ACTIONS(634),
    [aux_sym_comparison_operator_token1] = ACTIONS(634),
    [aux_sym_comparison_operator_token2] = ACTIONS(634),
    [aux_sym_comparison_operator_token3] = ACTIONS(634),
    [aux_sym_comparison_operator_token4] = ACTIONS(634),
    [aux_sym_comparison_operator_token5] = ACTIONS(634),
    [aux_sym_comparison_operator_token6] = ACTIONS(634),
    [aux_sym_comparison_operator_token7] = ACTIONS(634),
    [aux_sym_comparison_operator_token8] = ACTIONS(634),
    [aux_sym_comparison_operator_token9] = ACTIONS(634),
    [aux_sym_comparison_operator_token10] = ACTIONS(634),
    [aux_sym_comparison_operator_token11] = ACTIONS(634),
    [aux_sym_comparison_operator_token12] = ACTIONS(634),
    [aux_sym_comparison_operator_token13] = ACTIONS(634),
    [aux_sym_comparison_operator_token14] = ACTIONS(634),
    [aux_sym_comparison_operator_token15] = ACTIONS(634),
    [aux_sym_comparison_operator_token16] = ACTIONS(634),
    [aux_sym_comparison_operator_token17] = ACTIONS(634),
    [aux_sym_comparison_operator_token18] = ACTIONS(634),
    [aux_sym_comparison_operator_token19] = ACTIONS(634),
    [aux_sym_comparison_operator_token20] = ACTIONS(634),
    [aux_sym_comparison_operator_token21] = ACTIONS(634),
    [aux_sym_comparison_operator_token22] = ACTIONS(634),
    [aux_sym_comparison_operator_token23] = ACTIONS(634),
    [aux_sym_comparison_operator_token24] = ACTIONS(634),
    [aux_sym_comparison_operator_token25] = ACTIONS(634),
    [aux_sym_comparison_operator_token26] = ACTIONS(634),
    [aux_sym_comparison_operator_token27] = ACTIONS(634),
    [aux_sym_comparison_operator_token28] = ACTIONS(636),
    [aux_sym_comparison_operator_token29] = ACTIONS(634),
    [aux_sym_comparison_operator_token30] = ACTIONS(634),
    [aux_sym_comparison_operator_token31] = ACTIONS(634),
    [aux_sym_comparison_operator_token32] = ACTIONS(634),
    [aux_sym_comparison_operator_token33] = ACTIONS(634),
    [aux_sym_comparison_operator_token34] = ACTIONS(636),
    [aux_sym_comparison_operator_token35] = ACTIONS(634),
    [aux_sym_comparison_operator_token36] = ACTIONS(634),
    [aux_sym_comparison_operator_token37] = ACTIONS(634),
    [aux_sym_comparison_operator_token38] = ACTIONS(634),
    [aux_sym_comparison_operator_token39] = ACTIONS(634),
    [aux_sym_comparison_operator_token40] = ACTIONS(634),
    [aux_sym_comparison_operator_token41] = ACTIONS(634),
    [aux_sym_comparison_operator_token42] = ACTIONS(634),
    [aux_sym_comparison_operator_token43] = ACTIONS(634),
    [aux_sym_comparison_operator_token44] = ACTIONS(634),
    [aux_sym_comparison_operator_token45] = ACTIONS(634),
    [aux_sym_comparison_operator_token46] = ACTIONS(634),
    [aux_sym_comparison_operator_token47] = ACTIONS(634),
    [aux_sym_comparison_operator_token48] = ACTIONS(634),
    [aux_sym_comparison_operator_token49] = ACTIONS(634),
    [aux_sym_comparison_operator_token50] = ACTIONS(634),
    [aux_sym_format_operator_token1] = ACTIONS(634),
    [anon_sym_LPAREN] = ACTIONS(634),
    [anon_sym_RPAREN] = ACTIONS(634),
    [anon_sym_COMMA] = ACTIONS(634),
    [anon_sym_AMP_AMP] = ACTIONS(634),
    [anon_sym_PIPE_PIPE] = ACTIONS(634),
    [anon_sym_PIPE] = ACTIONS(636),
    [aux_sym_logical_expression_token1] = ACTIONS(634),
    [aux_sym_logical_expression_token2] = ACTIONS(634),
    [aux_sym_logical_expression_token3] = ACTIONS(634),
    [aux_sym_bitwise_expression_token1] = ACTIONS(634),
    [aux_sym_bitwise_expression_token2] = ACTIONS(634),
    [aux_sym_bitwise_expression_token3] = ACTIONS(634),
    [anon_sym_PLUS] = ACTIONS(636),
    [anon_sym_DASH] = ACTIONS(636),
    [anon_sym_SLASH] = ACTIONS(636),
    [anon_sym_BSLASH] = ACTIONS(634),
    [anon_sym_PERCENT] = ACTIONS(636),
    [anon_sym_STAR] = ACTIONS(636),
    [anon_sym_DOT_DOT] = ACTIONS(634),
    [anon_sym_PLUS_PLUS] = ACTIONS(634),
    [anon_sym_DASH_DASH] = ACTIONS(634),
    [anon_sym_DOT2] = ACTIONS(636),
    [anon_sym_COLON_COLON] = ACTIONS(634),
    [anon_sym_RBRACK] = ACTIONS(634),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(634),
  },
  [STATE(100)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(638),
    [anon_sym_EQ] = ACTIONS(638),
    [anon_sym_BANG_EQ] = ACTIONS(638),
    [anon_sym_PLUS_EQ] = ACTIONS(638),
    [anon_sym_STAR_EQ] = ACTIONS(638),
    [anon_sym_SLASH_EQ] = ACTIONS(638),
    [anon_sym_PERCENT_EQ] = ACTIONS(638),
    [anon_sym_DASH_EQ] = ACTIONS(638),
    [anon_sym_GT] = ACTIONS(640),
    [anon_sym_GT_GT] = ACTIONS(638),
    [anon_sym_2_GT] = ACTIONS(640),
    [anon_sym_2_GT_GT] = ACTIONS(638),
    [anon_sym_3_GT] = ACTIONS(640),
    [anon_sym_3_GT_GT] = ACTIONS(638),
    [anon_sym_4_GT] = ACTIONS(640),
    [anon_sym_4_GT_GT] = ACTIONS(638),
    [anon_sym_5_GT] = ACTIONS(640),
    [anon_sym_5_GT_GT] = ACTIONS(638),
    [anon_sym_6_GT] = ACTIONS(640),
    [anon_sym_6_GT_GT] = ACTIONS(638),
    [anon_sym_STAR_GT] = ACTIONS(640),
    [anon_sym_STAR_GT_GT] = ACTIONS(638),
    [anon_sym_LT] = ACTIONS(640),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(638),
    [anon_sym_2_GT_AMP1] = ACTIONS(638),
    [anon_sym_3_GT_AMP1] = ACTIONS(638),
    [anon_sym_4_GT_AMP1] = ACTIONS(638),
    [anon_sym_5_GT_AMP1] = ACTIONS(638),
    [anon_sym_6_GT_AMP1] = ACTIONS(638),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(638),
    [anon_sym_1_GT_AMP2] = ACTIONS(638),
    [anon_sym_3_GT_AMP2] = ACTIONS(638),
    [anon_sym_4_GT_AMP2] = ACTIONS(638),
    [anon_sym_5_GT_AMP2] = ACTIONS(638),
    [anon_sym_6_GT_AMP2] = ACTIONS(638),
    [aux_sym_comparison_operator_token1] = ACTIONS(638),
    [aux_sym_comparison_operator_token2] = ACTIONS(638),
    [aux_sym_comparison_operator_token3] = ACTIONS(638),
    [aux_sym_comparison_operator_token4] = ACTIONS(638),
    [aux_sym_comparison_operator_token5] = ACTIONS(638),
    [aux_sym_comparison_operator_token6] = ACTIONS(638),
    [aux_sym_comparison_operator_token7] = ACTIONS(638),
    [aux_sym_comparison_operator_token8] = ACTIONS(638),
    [aux_sym_comparison_operator_token9] = ACTIONS(638),
    [aux_sym_comparison_operator_token10] = ACTIONS(638),
    [aux_sym_comparison_operator_token11] = ACTIONS(638),
    [aux_sym_comparison_operator_token12] = ACTIONS(638),
    [aux_sym_comparison_operator_token13] = ACTIONS(638),
    [aux_sym_comparison_operator_token14] = ACTIONS(638),
    [aux_sym_comparison_operator_token15] = ACTIONS(638),
    [aux_sym_comparison_operator_token16] = ACTIONS(638),
    [aux_sym_comparison_operator_token17] = ACTIONS(638),
    [aux_sym_comparison_operator_token18] = ACTIONS(638),
    [aux_sym_comparison_operator_token19] = ACTIONS(638),
    [aux_sym_comparison_operator_token20] = ACTIONS(638),
    [aux_sym_comparison_operator_token21] = ACTIONS(638),
    [aux_sym_comparison_operator_token22] = ACTIONS(638),
    [aux_sym_comparison_operator_token23] = ACTIONS(638),
    [aux_sym_comparison_operator_token24] = ACTIONS(638),
    [aux_sym_comparison_operator_token25] = ACTIONS(638),
    [aux_sym_comparison_operator_token26] = ACTIONS(638),
    [aux_sym_comparison_operator_token27] = ACTIONS(638),
    [aux_sym_comparison_operator_token28] = ACTIONS(640),
    [aux_sym_comparison_operator_token29] = ACTIONS(638),
    [aux_sym_comparison_operator_token30] = ACTIONS(638),
    [aux_sym_comparison_operator_token31] = ACTIONS(638),
    [aux_sym_comparison_operator_token32] = ACTIONS(638),
    [aux_sym_comparison_operator_token33] = ACTIONS(638),
    [aux_sym_comparison_operator_token34] = ACTIONS(640),
    [aux_sym_comparison_operator_token35] = ACTIONS(638),
    [aux_sym_comparison_operator_token36] = ACTIONS(638),
    [aux_sym_comparison_operator_token37] = ACTIONS(638),
    [aux_sym_comparison_operator_token38] = ACTIONS(638),
    [aux_sym_comparison_operator_token39] = ACTIONS(638),
    [aux_sym_comparison_operator_token40] = ACTIONS(638),
    [aux_sym_comparison_operator_token41] = ACTIONS(638),
    [aux_sym_comparison_operator_token42] = ACTIONS(638),
    [aux_sym_comparison_operator_token43] = ACTIONS(638),
    [aux_sym_comparison_operator_token44] = ACTIONS(638),
    [aux_sym_comparison_operator_token45] = ACTIONS(638),
    [aux_sym_comparison_operator_token46] = ACTIONS(638),
    [aux_sym_comparison_operator_token47] = ACTIONS(638),
    [aux_sym_comparison_operator_token48] = ACTIONS(638),
    [aux_sym_comparison_operator_token49] = ACTIONS(638),
    [aux_sym_comparison_operator_token50] = ACTIONS(638),
    [aux_sym_format_operator_token1] = ACTIONS(638),
    [anon_sym_LPAREN] = ACTIONS(638),
    [anon_sym_RPAREN] = ACTIONS(638),
    [anon_sym_COMMA] = ACTIONS(638),
    [anon_sym_AMP_AMP] = ACTIONS(638),
    [anon_sym_PIPE_PIPE] = ACTIONS(638),
    [anon_sym_PIPE] = ACTIONS(640),
    [aux_sym_logical_expression_token1] = ACTIONS(638),
    [aux_sym_logical_expression_token2] = ACTIONS(638),
    [aux_sym_logical_expression_token3] = ACTIONS(638),
    [aux_sym_bitwise_expression_token1] = ACTIONS(638),
    [aux_sym_bitwise_expression_token2] = ACTIONS(638),
    [aux_sym_bitwise_expression_token3] = ACTIONS(638),
    [anon_sym_PLUS] = ACTIONS(640),
    [anon_sym_DASH] = ACTIONS(640),
    [anon_sym_SLASH] = ACTIONS(640),
    [anon_sym_BSLASH] = ACTIONS(638),
    [anon_sym_PERCENT] = ACTIONS(640),
    [anon_sym_STAR] = ACTIONS(640),
    [anon_sym_DOT_DOT] = ACTIONS(638),
    [anon_sym_PLUS_PLUS] = ACTIONS(638),
    [anon_sym_DASH_DASH] = ACTIONS(638),
    [anon_sym_DOT2] = ACTIONS(640),
    [anon_sym_COLON_COLON] = ACTIONS(638),
    [anon_sym_RBRACK] = ACTIONS(638),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(638),
  },
  [STATE(101)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(642),
    [anon_sym_EQ] = ACTIONS(642),
    [anon_sym_BANG_EQ] = ACTIONS(642),
    [anon_sym_PLUS_EQ] = ACTIONS(642),
    [anon_sym_STAR_EQ] = ACTIONS(642),
    [anon_sym_SLASH_EQ] = ACTIONS(642),
    [anon_sym_PERCENT_EQ] = ACTIONS(642),
    [anon_sym_DASH_EQ] = ACTIONS(642),
    [anon_sym_GT] = ACTIONS(644),
    [anon_sym_GT_GT] = ACTIONS(642),
    [anon_sym_2_GT] = ACTIONS(644),
    [anon_sym_2_GT_GT] = ACTIONS(642),
    [anon_sym_3_GT] = ACTIONS(644),
    [anon_sym_3_GT_GT] = ACTIONS(642),
    [anon_sym_4_GT] = ACTIONS(644),
    [anon_sym_4_GT_GT] = ACTIONS(642),
    [anon_sym_5_GT] = ACTIONS(644),
    [anon_sym_5_GT_GT] = ACTIONS(642),
    [anon_sym_6_GT] = ACTIONS(644),
    [anon_sym_6_GT_GT] = ACTIONS(642),
    [anon_sym_STAR_GT] = ACTIONS(644),
    [anon_sym_STAR_GT_GT] = ACTIONS(642),
    [anon_sym_LT] = ACTIONS(644),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(642),
    [anon_sym_2_GT_AMP1] = ACTIONS(642),
    [anon_sym_3_GT_AMP1] = ACTIONS(642),
    [anon_sym_4_GT_AMP1] = ACTIONS(642),
    [anon_sym_5_GT_AMP1] = ACTIONS(642),
    [anon_sym_6_GT_AMP1] = ACTIONS(642),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(642),
    [anon_sym_1_GT_AMP2] = ACTIONS(642),
    [anon_sym_3_GT_AMP2] = ACTIONS(642),
    [anon_sym_4_GT_AMP2] = ACTIONS(642),
    [anon_sym_5_GT_AMP2] = ACTIONS(642),
    [anon_sym_6_GT_AMP2] = ACTIONS(642),
    [aux_sym_comparison_operator_token1] = ACTIONS(642),
    [aux_sym_comparison_operator_token2] = ACTIONS(642),
    [aux_sym_comparison_operator_token3] = ACTIONS(642),
    [aux_sym_comparison_operator_token4] = ACTIONS(642),
    [aux_sym_comparison_operator_token5] = ACTIONS(642),
    [aux_sym_comparison_operator_token6] = ACTIONS(642),
    [aux_sym_comparison_operator_token7] = ACTIONS(642),
    [aux_sym_comparison_operator_token8] = ACTIONS(642),
    [aux_sym_comparison_operator_token9] = ACTIONS(642),
    [aux_sym_comparison_operator_token10] = ACTIONS(642),
    [aux_sym_comparison_operator_token11] = ACTIONS(642),
    [aux_sym_comparison_operator_token12] = ACTIONS(642),
    [aux_sym_comparison_operator_token13] = ACTIONS(642),
    [aux_sym_comparison_operator_token14] = ACTIONS(642),
    [aux_sym_comparison_operator_token15] = ACTIONS(642),
    [aux_sym_comparison_operator_token16] = ACTIONS(642),
    [aux_sym_comparison_operator_token17] = ACTIONS(642),
    [aux_sym_comparison_operator_token18] = ACTIONS(642),
    [aux_sym_comparison_operator_token19] = ACTIONS(642),
    [aux_sym_comparison_operator_token20] = ACTIONS(642),
    [aux_sym_comparison_operator_token21] = ACTIONS(642),
    [aux_sym_comparison_operator_token22] = ACTIONS(642),
    [aux_sym_comparison_operator_token23] = ACTIONS(642),
    [aux_sym_comparison_operator_token24] = ACTIONS(642),
    [aux_sym_comparison_operator_token25] = ACTIONS(642),
    [aux_sym_comparison_operator_token26] = ACTIONS(642),
    [aux_sym_comparison_operator_token27] = ACTIONS(642),
    [aux_sym_comparison_operator_token28] = ACTIONS(644),
    [aux_sym_comparison_operator_token29] = ACTIONS(642),
    [aux_sym_comparison_operator_token30] = ACTIONS(642),
    [aux_sym_comparison_operator_token31] = ACTIONS(642),
    [aux_sym_comparison_operator_token32] = ACTIONS(642),
    [aux_sym_comparison_operator_token33] = ACTIONS(642),
    [aux_sym_comparison_operator_token34] = ACTIONS(644),
    [aux_sym_comparison_operator_token35] = ACTIONS(642),
    [aux_sym_comparison_operator_token36] = ACTIONS(642),
    [aux_sym_comparison_operator_token37] = ACTIONS(642),
    [aux_sym_comparison_operator_token38] = ACTIONS(642),
    [aux_sym_comparison_operator_token39] = ACTIONS(642),
    [aux_sym_comparison_operator_token40] = ACTIONS(642),
    [aux_sym_comparison_operator_token41] = ACTIONS(642),
    [aux_sym_comparison_operator_token42] = ACTIONS(642),
    [aux_sym_comparison_operator_token43] = ACTIONS(642),
    [aux_sym_comparison_operator_token44] = ACTIONS(642),
    [aux_sym_comparison_operator_token45] = ACTIONS(642),
    [aux_sym_comparison_operator_token46] = ACTIONS(642),
    [aux_sym_comparison_operator_token47] = ACTIONS(642),
    [aux_sym_comparison_operator_token48] = ACTIONS(642),
    [aux_sym_comparison_operator_token49] = ACTIONS(642),
    [aux_sym_comparison_operator_token50] = ACTIONS(642),
    [aux_sym_format_operator_token1] = ACTIONS(642),
    [anon_sym_LPAREN] = ACTIONS(642),
    [anon_sym_RPAREN] = ACTIONS(642),
    [anon_sym_COMMA] = ACTIONS(642),
    [anon_sym_AMP_AMP] = ACTIONS(642),
    [anon_sym_PIPE_PIPE] = ACTIONS(642),
    [anon_sym_PIPE] = ACTIONS(644),
    [aux_sym_logical_expression_token1] = ACTIONS(642),
    [aux_sym_logical_expression_token2] = ACTIONS(642),
    [aux_sym_logical_expression_token3] = ACTIONS(642),
    [aux_sym_bitwise_expression_token1] = ACTIONS(642),
    [aux_sym_bitwise_expression_token2] = ACTIONS(642),
    [aux_sym_bitwise_expression_token3] = ACTIONS(642),
    [anon_sym_PLUS] = ACTIONS(644),
    [anon_sym_DASH] = ACTIONS(644),
    [anon_sym_SLASH] = ACTIONS(644),
    [anon_sym_BSLASH] = ACTIONS(642),
    [anon_sym_PERCENT] = ACTIONS(644),
    [anon_sym_STAR] = ACTIONS(644),
    [anon_sym_DOT_DOT] = ACTIONS(642),
    [anon_sym_PLUS_PLUS] = ACTIONS(642),
    [anon_sym_DASH_DASH] = ACTIONS(642),
    [anon_sym_DOT2] = ACTIONS(644),
    [anon_sym_COLON_COLON] = ACTIONS(642),
    [anon_sym_RBRACK] = ACTIONS(642),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(642),
  },
  [STATE(102)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(646),
    [anon_sym_EQ] = ACTIONS(646),
    [anon_sym_BANG_EQ] = ACTIONS(646),
    [anon_sym_PLUS_EQ] = ACTIONS(646),
    [anon_sym_STAR_EQ] = ACTIONS(646),
    [anon_sym_SLASH_EQ] = ACTIONS(646),
    [anon_sym_PERCENT_EQ] = ACTIONS(646),
    [anon_sym_DASH_EQ] = ACTIONS(646),
    [anon_sym_GT] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(646),
    [anon_sym_2_GT] = ACTIONS(648),
    [anon_sym_2_GT_GT] = ACTIONS(646),
    [anon_sym_3_GT] = ACTIONS(648),
    [anon_sym_3_GT_GT] = ACTIONS(646),
    [anon_sym_4_GT] = ACTIONS(648),
    [anon_sym_4_GT_GT] = ACTIONS(646),
    [anon_sym_5_GT] = ACTIONS(648),
    [anon_sym_5_GT_GT] = ACTIONS(646),
    [anon_sym_6_GT] = ACTIONS(648),
    [anon_sym_6_GT_GT] = ACTIONS(646),
    [anon_sym_STAR_GT] = ACTIONS(648),
    [anon_sym_STAR_GT_GT] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(646),
    [anon_sym_2_GT_AMP1] = ACTIONS(646),
    [anon_sym_3_GT_AMP1] = ACTIONS(646),
    [anon_sym_4_GT_AMP1] = ACTIONS(646),
    [anon_sym_5_GT_AMP1] = ACTIONS(646),
    [anon_sym_6_GT_AMP1] = ACTIONS(646),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(646),
    [anon_sym_1_GT_AMP2] = ACTIONS(646),
    [anon_sym_3_GT_AMP2] = ACTIONS(646),
    [anon_sym_4_GT_AMP2] = ACTIONS(646),
    [anon_sym_5_GT_AMP2] = ACTIONS(646),
    [anon_sym_6_GT_AMP2] = ACTIONS(646),
    [aux_sym_comparison_operator_token1] = ACTIONS(646),
    [aux_sym_comparison_operator_token2] = ACTIONS(646),
    [aux_sym_comparison_operator_token3] = ACTIONS(646),
    [aux_sym_comparison_operator_token4] = ACTIONS(646),
    [aux_sym_comparison_operator_token5] = ACTIONS(646),
    [aux_sym_comparison_operator_token6] = ACTIONS(646),
    [aux_sym_comparison_operator_token7] = ACTIONS(646),
    [aux_sym_comparison_operator_token8] = ACTIONS(646),
    [aux_sym_comparison_operator_token9] = ACTIONS(646),
    [aux_sym_comparison_operator_token10] = ACTIONS(646),
    [aux_sym_comparison_operator_token11] = ACTIONS(646),
    [aux_sym_comparison_operator_token12] = ACTIONS(646),
    [aux_sym_comparison_operator_token13] = ACTIONS(646),
    [aux_sym_comparison_operator_token14] = ACTIONS(646),
    [aux_sym_comparison_operator_token15] = ACTIONS(646),
    [aux_sym_comparison_operator_token16] = ACTIONS(646),
    [aux_sym_comparison_operator_token17] = ACTIONS(646),
    [aux_sym_comparison_operator_token18] = ACTIONS(646),
    [aux_sym_comparison_operator_token19] = ACTIONS(646),
    [aux_sym_comparison_operator_token20] = ACTIONS(646),
    [aux_sym_comparison_operator_token21] = ACTIONS(646),
    [aux_sym_comparison_operator_token22] = ACTIONS(646),
    [aux_sym_comparison_operator_token23] = ACTIONS(646),
    [aux_sym_comparison_operator_token24] = ACTIONS(646),
    [aux_sym_comparison_operator_token25] = ACTIONS(646),
    [aux_sym_comparison_operator_token26] = ACTIONS(646),
    [aux_sym_comparison_operator_token27] = ACTIONS(646),
    [aux_sym_comparison_operator_token28] = ACTIONS(648),
    [aux_sym_comparison_operator_token29] = ACTIONS(646),
    [aux_sym_comparison_operator_token30] = ACTIONS(646),
    [aux_sym_comparison_operator_token31] = ACTIONS(646),
    [aux_sym_comparison_operator_token32] = ACTIONS(646),
    [aux_sym_comparison_operator_token33] = ACTIONS(646),
    [aux_sym_comparison_operator_token34] = ACTIONS(648),
    [aux_sym_comparison_operator_token35] = ACTIONS(646),
    [aux_sym_comparison_operator_token36] = ACTIONS(646),
    [aux_sym_comparison_operator_token37] = ACTIONS(646),
    [aux_sym_comparison_operator_token38] = ACTIONS(646),
    [aux_sym_comparison_operator_token39] = ACTIONS(646),
    [aux_sym_comparison_operator_token40] = ACTIONS(646),
    [aux_sym_comparison_operator_token41] = ACTIONS(646),
    [aux_sym_comparison_operator_token42] = ACTIONS(646),
    [aux_sym_comparison_operator_token43] = ACTIONS(646),
    [aux_sym_comparison_operator_token44] = ACTIONS(646),
    [aux_sym_comparison_operator_token45] = ACTIONS(646),
    [aux_sym_comparison_operator_token46] = ACTIONS(646),
    [aux_sym_comparison_operator_token47] = ACTIONS(646),
    [aux_sym_comparison_operator_token48] = ACTIONS(646),
    [aux_sym_comparison_operator_token49] = ACTIONS(646),
    [aux_sym_comparison_operator_token50] = ACTIONS(646),
    [aux_sym_format_operator_token1] = ACTIONS(646),
    [anon_sym_LPAREN] = ACTIONS(646),
    [anon_sym_RPAREN] = ACTIONS(646),
    [anon_sym_COMMA] = ACTIONS(646),
    [anon_sym_AMP_AMP] = ACTIONS(646),
    [anon_sym_PIPE_PIPE] = ACTIONS(646),
    [anon_sym_PIPE] = ACTIONS(648),
    [aux_sym_logical_expression_token1] = ACTIONS(646),
    [aux_sym_logical_expression_token2] = ACTIONS(646),
    [aux_sym_logical_expression_token3] = ACTIONS(646),
    [aux_sym_bitwise_expression_token1] = ACTIONS(646),
    [aux_sym_bitwise_expression_token2] = ACTIONS(646),
    [aux_sym_bitwise_expression_token3] = ACTIONS(646),
    [anon_sym_PLUS] = ACTIONS(648),
    [anon_sym_DASH] = ACTIONS(648),
    [anon_sym_SLASH] = ACTIONS(648),
    [anon_sym_BSLASH] = ACTIONS(646),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR] = ACTIONS(648),
    [anon_sym_DOT_DOT] = ACTIONS(646),
    [anon_sym_PLUS_PLUS] = ACTIONS(646),
    [anon_sym_DASH_DASH] = ACTIONS(646),
    [anon_sym_DOT2] = ACTIONS(648),
    [anon_sym_COLON_COLON] = ACTIONS(646),
    [anon_sym_RBRACK] = ACTIONS(646),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(646),
  },
  [STATE(103)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_BANG_EQ] = ACTIONS(650),
    [anon_sym_PLUS_EQ] = ACTIONS(650),
    [anon_sym_STAR_EQ] = ACTIONS(650),
    [anon_sym_SLASH_EQ] = ACTIONS(650),
    [anon_sym_PERCENT_EQ] = ACTIONS(650),
    [anon_sym_DASH_EQ] = ACTIONS(650),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_GT_GT] = ACTIONS(650),
    [anon_sym_2_GT] = ACTIONS(652),
    [anon_sym_2_GT_GT] = ACTIONS(650),
    [anon_sym_3_GT] = ACTIONS(652),
    [anon_sym_3_GT_GT] = ACTIONS(650),
    [anon_sym_4_GT] = ACTIONS(652),
    [anon_sym_4_GT_GT] = ACTIONS(650),
    [anon_sym_5_GT] = ACTIONS(652),
    [anon_sym_5_GT_GT] = ACTIONS(650),
    [anon_sym_6_GT] = ACTIONS(652),
    [anon_sym_6_GT_GT] = ACTIONS(650),
    [anon_sym_STAR_GT] = ACTIONS(652),
    [anon_sym_STAR_GT_GT] = ACTIONS(650),
    [anon_sym_LT] = ACTIONS(652),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(650),
    [anon_sym_2_GT_AMP1] = ACTIONS(650),
    [anon_sym_3_GT_AMP1] = ACTIONS(650),
    [anon_sym_4_GT_AMP1] = ACTIONS(650),
    [anon_sym_5_GT_AMP1] = ACTIONS(650),
    [anon_sym_6_GT_AMP1] = ACTIONS(650),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(650),
    [anon_sym_1_GT_AMP2] = ACTIONS(650),
    [anon_sym_3_GT_AMP2] = ACTIONS(650),
    [anon_sym_4_GT_AMP2] = ACTIONS(650),
    [anon_sym_5_GT_AMP2] = ACTIONS(650),
    [anon_sym_6_GT_AMP2] = ACTIONS(650),
    [aux_sym_comparison_operator_token1] = ACTIONS(650),
    [aux_sym_comparison_operator_token2] = ACTIONS(650),
    [aux_sym_comparison_operator_token3] = ACTIONS(650),
    [aux_sym_comparison_operator_token4] = ACTIONS(650),
    [aux_sym_comparison_operator_token5] = ACTIONS(650),
    [aux_sym_comparison_operator_token6] = ACTIONS(650),
    [aux_sym_comparison_operator_token7] = ACTIONS(650),
    [aux_sym_comparison_operator_token8] = ACTIONS(650),
    [aux_sym_comparison_operator_token9] = ACTIONS(650),
    [aux_sym_comparison_operator_token10] = ACTIONS(650),
    [aux_sym_comparison_operator_token11] = ACTIONS(650),
    [aux_sym_comparison_operator_token12] = ACTIONS(650),
    [aux_sym_comparison_operator_token13] = ACTIONS(650),
    [aux_sym_comparison_operator_token14] = ACTIONS(650),
    [aux_sym_comparison_operator_token15] = ACTIONS(650),
    [aux_sym_comparison_operator_token16] = ACTIONS(650),
    [aux_sym_comparison_operator_token17] = ACTIONS(650),
    [aux_sym_comparison_operator_token18] = ACTIONS(650),
    [aux_sym_comparison_operator_token19] = ACTIONS(650),
    [aux_sym_comparison_operator_token20] = ACTIONS(650),
    [aux_sym_comparison_operator_token21] = ACTIONS(650),
    [aux_sym_comparison_operator_token22] = ACTIONS(650),
    [aux_sym_comparison_operator_token23] = ACTIONS(650),
    [aux_sym_comparison_operator_token24] = ACTIONS(650),
    [aux_sym_comparison_operator_token25] = ACTIONS(650),
    [aux_sym_comparison_operator_token26] = ACTIONS(650),
    [aux_sym_comparison_operator_token27] = ACTIONS(650),
    [aux_sym_comparison_operator_token28] = ACTIONS(652),
    [aux_sym_comparison_operator_token29] = ACTIONS(650),
    [aux_sym_comparison_operator_token30] = ACTIONS(650),
    [aux_sym_comparison_operator_token31] = ACTIONS(650),
    [aux_sym_comparison_operator_token32] = ACTIONS(650),
    [aux_sym_comparison_operator_token33] = ACTIONS(650),
    [aux_sym_comparison_operator_token34] = ACTIONS(652),
    [aux_sym_comparison_operator_token35] = ACTIONS(650),
    [aux_sym_comparison_operator_token36] = ACTIONS(650),
    [aux_sym_comparison_operator_token37] = ACTIONS(650),
    [aux_sym_comparison_operator_token38] = ACTIONS(650),
    [aux_sym_comparison_operator_token39] = ACTIONS(650),
    [aux_sym_comparison_operator_token40] = ACTIONS(650),
    [aux_sym_comparison_operator_token41] = ACTIONS(650),
    [aux_sym_comparison_operator_token42] = ACTIONS(650),
    [aux_sym_comparison_operator_token43] = ACTIONS(650),
    [aux_sym_comparison_operator_token44] = ACTIONS(650),
    [aux_sym_comparison_operator_token45] = ACTIONS(650),
    [aux_sym_comparison_operator_token46] = ACTIONS(650),
    [aux_sym_comparison_operator_token47] = ACTIONS(650),
    [aux_sym_comparison_operator_token48] = ACTIONS(650),
    [aux_sym_comparison_operator_token49] = ACTIONS(650),
    [aux_sym_comparison_operator_token50] = ACTIONS(650),
    [aux_sym_format_operator_token1] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_AMP_AMP] = ACTIONS(650),
    [anon_sym_PIPE_PIPE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(652),
    [aux_sym_logical_expression_token1] = ACTIONS(650),
    [aux_sym_logical_expression_token2] = ACTIONS(650),
    [aux_sym_logical_expression_token3] = ACTIONS(650),
    [aux_sym_bitwise_expression_token1] = ACTIONS(650),
    [aux_sym_bitwise_expression_token2] = ACTIONS(650),
    [aux_sym_bitwise_expression_token3] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_SLASH] = ACTIONS(652),
    [anon_sym_BSLASH] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(652),
    [anon_sym_STAR] = ACTIONS(652),
    [anon_sym_DOT_DOT] = ACTIONS(650),
    [anon_sym_PLUS_PLUS] = ACTIONS(650),
    [anon_sym_DASH_DASH] = ACTIONS(650),
    [anon_sym_DOT2] = ACTIONS(652),
    [anon_sym_COLON_COLON] = ACTIONS(650),
    [anon_sym_RBRACK] = ACTIONS(650),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(650),
  },
  [STATE(104)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(654),
    [anon_sym_DASH_DASH] = ACTIONS(654),
    [anon_sym_DOT2] = ACTIONS(656),
    [anon_sym_COLON_COLON] = ACTIONS(654),
    [anon_sym_RBRACK] = ACTIONS(654),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(654),
  },
  [STATE(105)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(658),
    [anon_sym_EQ] = ACTIONS(658),
    [anon_sym_BANG_EQ] = ACTIONS(658),
    [anon_sym_PLUS_EQ] = ACTIONS(658),
    [anon_sym_STAR_EQ] = ACTIONS(658),
    [anon_sym_SLASH_EQ] = ACTIONS(658),
    [anon_sym_PERCENT_EQ] = ACTIONS(658),
    [anon_sym_DASH_EQ] = ACTIONS(658),
    [anon_sym_GT] = ACTIONS(660),
    [anon_sym_GT_GT] = ACTIONS(658),
    [anon_sym_2_GT] = ACTIONS(660),
    [anon_sym_2_GT_GT] = ACTIONS(658),
    [anon_sym_3_GT] = ACTIONS(660),
    [anon_sym_3_GT_GT] = ACTIONS(658),
    [anon_sym_4_GT] = ACTIONS(660),
    [anon_sym_4_GT_GT] = ACTIONS(658),
    [anon_sym_5_GT] = ACTIONS(660),
    [anon_sym_5_GT_GT] = ACTIONS(658),
    [anon_sym_6_GT] = ACTIONS(660),
    [anon_sym_6_GT_GT] = ACTIONS(658),
    [anon_sym_STAR_GT] = ACTIONS(660),
    [anon_sym_STAR_GT_GT] = ACTIONS(658),
    [anon_sym_LT] = ACTIONS(660),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(658),
    [anon_sym_2_GT_AMP1] = ACTIONS(658),
    [anon_sym_3_GT_AMP1] = ACTIONS(658),
    [anon_sym_4_GT_AMP1] = ACTIONS(658),
    [anon_sym_5_GT_AMP1] = ACTIONS(658),
    [anon_sym_6_GT_AMP1] = ACTIONS(658),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(658),
    [anon_sym_1_GT_AMP2] = ACTIONS(658),
    [anon_sym_3_GT_AMP2] = ACTIONS(658),
    [anon_sym_4_GT_AMP2] = ACTIONS(658),
    [anon_sym_5_GT_AMP2] = ACTIONS(658),
    [anon_sym_6_GT_AMP2] = ACTIONS(658),
    [aux_sym_comparison_operator_token1] = ACTIONS(658),
    [aux_sym_comparison_operator_token2] = ACTIONS(658),
    [aux_sym_comparison_operator_token3] = ACTIONS(658),
    [aux_sym_comparison_operator_token4] = ACTIONS(658),
    [aux_sym_comparison_operator_token5] = ACTIONS(658),
    [aux_sym_comparison_operator_token6] = ACTIONS(658),
    [aux_sym_comparison_operator_token7] = ACTIONS(658),
    [aux_sym_comparison_operator_token8] = ACTIONS(658),
    [aux_sym_comparison_operator_token9] = ACTIONS(658),
    [aux_sym_comparison_operator_token10] = ACTIONS(658),
    [aux_sym_comparison_operator_token11] = ACTIONS(658),
    [aux_sym_comparison_operator_token12] = ACTIONS(658),
    [aux_sym_comparison_operator_token13] = ACTIONS(658),
    [aux_sym_comparison_operator_token14] = ACTIONS(658),
    [aux_sym_comparison_operator_token15] = ACTIONS(658),
    [aux_sym_comparison_operator_token16] = ACTIONS(658),
    [aux_sym_comparison_operator_token17] = ACTIONS(658),
    [aux_sym_comparison_operator_token18] = ACTIONS(658),
    [aux_sym_comparison_operator_token19] = ACTIONS(658),
    [aux_sym_comparison_operator_token20] = ACTIONS(658),
    [aux_sym_comparison_operator_token21] = ACTIONS(658),
    [aux_sym_comparison_operator_token22] = ACTIONS(658),
    [aux_sym_comparison_operator_token23] = ACTIONS(658),
    [aux_sym_comparison_operator_token24] = ACTIONS(658),
    [aux_sym_comparison_operator_token25] = ACTIONS(658),
    [aux_sym_comparison_operator_token26] = ACTIONS(658),
    [aux_sym_comparison_operator_token27] = ACTIONS(658),
    [aux_sym_comparison_operator_token28] = ACTIONS(660),
    [aux_sym_comparison_operator_token29] = ACTIONS(658),
    [aux_sym_comparison_operator_token30] = ACTIONS(658),
    [aux_sym_comparison_operator_token31] = ACTIONS(658),
    [aux_sym_comparison_operator_token32] = ACTIONS(658),
    [aux_sym_comparison_operator_token33] = ACTIONS(658),
    [aux_sym_comparison_operator_token34] = ACTIONS(660),
    [aux_sym_comparison_operator_token35] = ACTIONS(658),
    [aux_sym_comparison_operator_token36] = ACTIONS(658),
    [aux_sym_comparison_operator_token37] = ACTIONS(658),
    [aux_sym_comparison_operator_token38] = ACTIONS(658),
    [aux_sym_comparison_operator_token39] = ACTIONS(658),
    [aux_sym_comparison_operator_token40] = ACTIONS(658),
    [aux_sym_comparison_operator_token41] = ACTIONS(658),
    [aux_sym_comparison_operator_token42] = ACTIONS(658),
    [aux_sym_comparison_operator_token43] = ACTIONS(658),
    [aux_sym_comparison_operator_token44] = ACTIONS(658),
    [aux_sym_comparison_operator_token45] = ACTIONS(658),
    [aux_sym_comparison_operator_token46] = ACTIONS(658),
    [aux_sym_comparison_operator_token47] = ACTIONS(658),
    [aux_sym_comparison_operator_token48] = ACTIONS(658),
    [aux_sym_comparison_operator_token49] = ACTIONS(658),
    [aux_sym_comparison_operator_token50] = ACTIONS(658),
    [aux_sym_format_operator_token1] = ACTIONS(658),
    [anon_sym_LPAREN] = ACTIONS(658),
    [anon_sym_RPAREN] = ACTIONS(658),
    [anon_sym_COMMA] = ACTIONS(658),
    [anon_sym_AMP_AMP] = ACTIONS(658),
    [anon_sym_PIPE_PIPE] = ACTIONS(658),
    [anon_sym_PIPE] = ACTIONS(660),
    [aux_sym_logical_expression_token1] = ACTIONS(658),
    [aux_sym_logical_expression_token2] = ACTIONS(658),
    [aux_sym_logical_expression_token3] = ACTIONS(658),
    [aux_sym_bitwise_expression_token1] = ACTIONS(658),
    [aux_sym_bitwise_expression_token2] = ACTIONS(658),
    [aux_sym_bitwise_expression_token3] = ACTIONS(658),
    [anon_sym_PLUS] = ACTIONS(660),
    [anon_sym_DASH] = ACTIONS(660),
    [anon_sym_SLASH] = ACTIONS(660),
    [anon_sym_BSLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(660),
    [anon_sym_STAR] = ACTIONS(660),
    [anon_sym_DOT_DOT] = ACTIONS(658),
    [anon_sym_PLUS_PLUS] = ACTIONS(658),
    [anon_sym_DASH_DASH] = ACTIONS(658),
    [anon_sym_DOT2] = ACTIONS(660),
    [anon_sym_COLON_COLON] = ACTIONS(658),
    [anon_sym_RBRACK] = ACTIONS(658),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(658),
  },
  [STATE(106)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(662),
    [anon_sym_EQ] = ACTIONS(662),
    [anon_sym_BANG_EQ] = ACTIONS(662),
    [anon_sym_PLUS_EQ] = ACTIONS(662),
    [anon_sym_STAR_EQ] = ACTIONS(662),
    [anon_sym_SLASH_EQ] = ACTIONS(662),
    [anon_sym_PERCENT_EQ] = ACTIONS(662),
    [anon_sym_DASH_EQ] = ACTIONS(662),
    [anon_sym_GT] = ACTIONS(664),
    [anon_sym_GT_GT] = ACTIONS(662),
    [anon_sym_2_GT] = ACTIONS(664),
    [anon_sym_2_GT_GT] = ACTIONS(662),
    [anon_sym_3_GT] = ACTIONS(664),
    [anon_sym_3_GT_GT] = ACTIONS(662),
    [anon_sym_4_GT] = ACTIONS(664),
    [anon_sym_4_GT_GT] = ACTIONS(662),
    [anon_sym_5_GT] = ACTIONS(664),
    [anon_sym_5_GT_GT] = ACTIONS(662),
    [anon_sym_6_GT] = ACTIONS(664),
    [anon_sym_6_GT_GT] = ACTIONS(662),
    [anon_sym_STAR_GT] = ACTIONS(664),
    [anon_sym_STAR_GT_GT] = ACTIONS(662),
    [anon_sym_LT] = ACTIONS(664),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(662),
    [anon_sym_2_GT_AMP1] = ACTIONS(662),
    [anon_sym_3_GT_AMP1] = ACTIONS(662),
    [anon_sym_4_GT_AMP1] = ACTIONS(662),
    [anon_sym_5_GT_AMP1] = ACTIONS(662),
    [anon_sym_6_GT_AMP1] = ACTIONS(662),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(662),
    [anon_sym_1_GT_AMP2] = ACTIONS(662),
    [anon_sym_3_GT_AMP2] = ACTIONS(662),
    [anon_sym_4_GT_AMP2] = ACTIONS(662),
    [anon_sym_5_GT_AMP2] = ACTIONS(662),
    [anon_sym_6_GT_AMP2] = ACTIONS(662),
    [aux_sym_comparison_operator_token1] = ACTIONS(662),
    [aux_sym_comparison_operator_token2] = ACTIONS(662),
    [aux_sym_comparison_operator_token3] = ACTIONS(662),
    [aux_sym_comparison_operator_token4] = ACTIONS(662),
    [aux_sym_comparison_operator_token5] = ACTIONS(662),
    [aux_sym_comparison_operator_token6] = ACTIONS(662),
    [aux_sym_comparison_operator_token7] = ACTIONS(662),
    [aux_sym_comparison_operator_token8] = ACTIONS(662),
    [aux_sym_comparison_operator_token9] = ACTIONS(662),
    [aux_sym_comparison_operator_token10] = ACTIONS(662),
    [aux_sym_comparison_operator_token11] = ACTIONS(662),
    [aux_sym_comparison_operator_token12] = ACTIONS(662),
    [aux_sym_comparison_operator_token13] = ACTIONS(662),
    [aux_sym_comparison_operator_token14] = ACTIONS(662),
    [aux_sym_comparison_operator_token15] = ACTIONS(662),
    [aux_sym_comparison_operator_token16] = ACTIONS(662),
    [aux_sym_comparison_operator_token17] = ACTIONS(662),
    [aux_sym_comparison_operator_token18] = ACTIONS(662),
    [aux_sym_comparison_operator_token19] = ACTIONS(662),
    [aux_sym_comparison_operator_token20] = ACTIONS(662),
    [aux_sym_comparison_operator_token21] = ACTIONS(662),
    [aux_sym_comparison_operator_token22] = ACTIONS(662),
    [aux_sym_comparison_operator_token23] = ACTIONS(662),
    [aux_sym_comparison_operator_token24] = ACTIONS(662),
    [aux_sym_comparison_operator_token25] = ACTIONS(662),
    [aux_sym_comparison_operator_token26] = ACTIONS(662),
    [aux_sym_comparison_operator_token27] = ACTIONS(662),
    [aux_sym_comparison_operator_token28] = ACTIONS(664),
    [aux_sym_comparison_operator_token29] = ACTIONS(662),
    [aux_sym_comparison_operator_token30] = ACTIONS(662),
    [aux_sym_comparison_operator_token31] = ACTIONS(662),
    [aux_sym_comparison_operator_token32] = ACTIONS(662),
    [aux_sym_comparison_operator_token33] = ACTIONS(662),
    [aux_sym_comparison_operator_token34] = ACTIONS(664),
    [aux_sym_comparison_operator_token35] = ACTIONS(662),
    [aux_sym_comparison_operator_token36] = ACTIONS(662),
    [aux_sym_comparison_operator_token37] = ACTIONS(662),
    [aux_sym_comparison_operator_token38] = ACTIONS(662),
    [aux_sym_comparison_operator_token39] = ACTIONS(662),
    [aux_sym_comparison_operator_token40] = ACTIONS(662),
    [aux_sym_comparison_operator_token41] = ACTIONS(662),
    [aux_sym_comparison_operator_token42] = ACTIONS(662),
    [aux_sym_comparison_operator_token43] = ACTIONS(662),
    [aux_sym_comparison_operator_token44] = ACTIONS(662),
    [aux_sym_comparison_operator_token45] = ACTIONS(662),
    [aux_sym_comparison_operator_token46] = ACTIONS(662),
    [aux_sym_comparison_operator_token47] = ACTIONS(662),
    [aux_sym_comparison_operator_token48] = ACTIONS(662),
    [aux_sym_comparison_operator_token49] = ACTIONS(662),
    [aux_sym_comparison_operator_token50] = ACTIONS(662),
    [aux_sym_format_operator_token1] = ACTIONS(662),
    [anon_sym_LPAREN] = ACTIONS(662),
    [anon_sym_RPAREN] = ACTIONS(662),
    [anon_sym_COMMA] = ACTIONS(662),
    [anon_sym_AMP_AMP] = ACTIONS(662),
    [anon_sym_PIPE_PIPE] = ACTIONS(662),
    [anon_sym_PIPE] = ACTIONS(664),
    [aux_sym_logical_expression_token1] = ACTIONS(662),
    [aux_sym_logical_expression_token2] = ACTIONS(662),
    [aux_sym_logical_expression_token3] = ACTIONS(662),
    [aux_sym_bitwise_expression_token1] = ACTIONS(662),
    [aux_sym_bitwise_expression_token2] = ACTIONS(662),
    [aux_sym_bitwise_expression_token3] = ACTIONS(662),
    [anon_sym_PLUS] = ACTIONS(664),
    [anon_sym_DASH] = ACTIONS(664),
    [anon_sym_SLASH] = ACTIONS(664),
    [anon_sym_BSLASH] = ACTIONS(662),
    [anon_sym_PERCENT] = ACTIONS(664),
    [anon_sym_STAR] = ACTIONS(664),
    [anon_sym_DOT_DOT] = ACTIONS(662),
    [anon_sym_PLUS_PLUS] = ACTIONS(662),
    [anon_sym_DASH_DASH] = ACTIONS(662),
    [anon_sym_DOT2] = ACTIONS(664),
    [anon_sym_COLON_COLON] = ACTIONS(662),
    [anon_sym_RBRACK] = ACTIONS(662),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(662),
  },
  [STATE(107)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(666),
    [anon_sym_EQ] = ACTIONS(666),
    [anon_sym_BANG_EQ] = ACTIONS(666),
    [anon_sym_PLUS_EQ] = ACTIONS(666),
    [anon_sym_STAR_EQ] = ACTIONS(666),
    [anon_sym_SLASH_EQ] = ACTIONS(666),
    [anon_sym_PERCENT_EQ] = ACTIONS(666),
    [anon_sym_DASH_EQ] = ACTIONS(666),
    [anon_sym_GT] = ACTIONS(668),
    [anon_sym_GT_GT] = ACTIONS(666),
    [anon_sym_2_GT] = ACTIONS(668),
    [anon_sym_2_GT_GT] = ACTIONS(666),
    [anon_sym_3_GT] = ACTIONS(668),
    [anon_sym_3_GT_GT] = ACTIONS(666),
    [anon_sym_4_GT] = ACTIONS(668),
    [anon_sym_4_GT_GT] = ACTIONS(666),
    [anon_sym_5_GT] = ACTIONS(668),
    [anon_sym_5_GT_GT] = ACTIONS(666),
    [anon_sym_6_GT] = ACTIONS(668),
    [anon_sym_6_GT_GT] = ACTIONS(666),
    [anon_sym_STAR_GT] = ACTIONS(668),
    [anon_sym_STAR_GT_GT] = ACTIONS(666),
    [anon_sym_LT] = ACTIONS(668),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(666),
    [anon_sym_2_GT_AMP1] = ACTIONS(666),
    [anon_sym_3_GT_AMP1] = ACTIONS(666),
    [anon_sym_4_GT_AMP1] = ACTIONS(666),
    [anon_sym_5_GT_AMP1] = ACTIONS(666),
    [anon_sym_6_GT_AMP1] = ACTIONS(666),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(666),
    [anon_sym_1_GT_AMP2] = ACTIONS(666),
    [anon_sym_3_GT_AMP2] = ACTIONS(666),
    [anon_sym_4_GT_AMP2] = ACTIONS(666),
    [anon_sym_5_GT_AMP2] = ACTIONS(666),
    [anon_sym_6_GT_AMP2] = ACTIONS(666),
    [aux_sym_comparison_operator_token1] = ACTIONS(666),
    [aux_sym_comparison_operator_token2] = ACTIONS(666),
    [aux_sym_comparison_operator_token3] = ACTIONS(666),
    [aux_sym_comparison_operator_token4] = ACTIONS(666),
    [aux_sym_comparison_operator_token5] = ACTIONS(666),
    [aux_sym_comparison_operator_token6] = ACTIONS(666),
    [aux_sym_comparison_operator_token7] = ACTIONS(666),
    [aux_sym_comparison_operator_token8] = ACTIONS(666),
    [aux_sym_comparison_operator_token9] = ACTIONS(666),
    [aux_sym_comparison_operator_token10] = ACTIONS(666),
    [aux_sym_comparison_operator_token11] = ACTIONS(666),
    [aux_sym_comparison_operator_token12] = ACTIONS(666),
    [aux_sym_comparison_operator_token13] = ACTIONS(666),
    [aux_sym_comparison_operator_token14] = ACTIONS(666),
    [aux_sym_comparison_operator_token15] = ACTIONS(666),
    [aux_sym_comparison_operator_token16] = ACTIONS(666),
    [aux_sym_comparison_operator_token17] = ACTIONS(666),
    [aux_sym_comparison_operator_token18] = ACTIONS(666),
    [aux_sym_comparison_operator_token19] = ACTIONS(666),
    [aux_sym_comparison_operator_token20] = ACTIONS(666),
    [aux_sym_comparison_operator_token21] = ACTIONS(666),
    [aux_sym_comparison_operator_token22] = ACTIONS(666),
    [aux_sym_comparison_operator_token23] = ACTIONS(666),
    [aux_sym_comparison_operator_token24] = ACTIONS(666),
    [aux_sym_comparison_operator_token25] = ACTIONS(666),
    [aux_sym_comparison_operator_token26] = ACTIONS(666),
    [aux_sym_comparison_operator_token27] = ACTIONS(666),
    [aux_sym_comparison_operator_token28] = ACTIONS(668),
    [aux_sym_comparison_operator_token29] = ACTIONS(666),
    [aux_sym_comparison_operator_token30] = ACTIONS(666),
    [aux_sym_comparison_operator_token31] = ACTIONS(666),
    [aux_sym_comparison_operator_token32] = ACTIONS(666),
    [aux_sym_comparison_operator_token33] = ACTIONS(666),
    [aux_sym_comparison_operator_token34] = ACTIONS(668),
    [aux_sym_comparison_operator_token35] = ACTIONS(666),
    [aux_sym_comparison_operator_token36] = ACTIONS(666),
    [aux_sym_comparison_operator_token37] = ACTIONS(666),
    [aux_sym_comparison_operator_token38] = ACTIONS(666),
    [aux_sym_comparison_operator_token39] = ACTIONS(666),
    [aux_sym_comparison_operator_token40] = ACTIONS(666),
    [aux_sym_comparison_operator_token41] = ACTIONS(666),
    [aux_sym_comparison_operator_token42] = ACTIONS(666),
    [aux_sym_comparison_operator_token43] = ACTIONS(666),
    [aux_sym_comparison_operator_token44] = ACTIONS(666),
    [aux_sym_comparison_operator_token45] = ACTIONS(666),
    [aux_sym_comparison_operator_token46] = ACTIONS(666),
    [aux_sym_comparison_operator_token47] = ACTIONS(666),
    [aux_sym_comparison_operator_token48] = ACTIONS(666),
    [aux_sym_comparison_operator_token49] = ACTIONS(666),
    [aux_sym_comparison_operator_token50] = ACTIONS(666),
    [aux_sym_format_operator_token1] = ACTIONS(666),
    [anon_sym_LPAREN] = ACTIONS(666),
    [anon_sym_RPAREN] = ACTIONS(666),
    [anon_sym_COMMA] = ACTIONS(666),
    [anon_sym_AMP_AMP] = ACTIONS(666),
    [anon_sym_PIPE_PIPE] = ACTIONS(666),
    [anon_sym_PIPE] = ACTIONS(668),
    [aux_sym_logical_expression_token1] = ACTIONS(666),
    [aux_sym_logical_expression_token2] = ACTIONS(666),
    [aux_sym_logical_expression_token3] = ACTIONS(666),
    [aux_sym_bitwise_expression_token1] = ACTIONS(666),
    [aux_sym_bitwise_expression_token2] = ACTIONS(666),
    [aux_sym_bitwise_expression_token3] = ACTIONS(666),
    [anon_sym_PLUS] = ACTIONS(668),
    [anon_sym_DASH] = ACTIONS(668),
    [anon_sym_SLASH] = ACTIONS(668),
    [anon_sym_BSLASH] = ACTIONS(666),
    [anon_sym_PERCENT] = ACTIONS(668),
    [anon_sym_STAR] = ACTIONS(668),
    [anon_sym_DOT_DOT] = ACTIONS(666),
    [anon_sym_PLUS_PLUS] = ACTIONS(666),
    [anon_sym_DASH_DASH] = ACTIONS(666),
    [anon_sym_DOT2] = ACTIONS(668),
    [anon_sym_COLON_COLON] = ACTIONS(666),
    [anon_sym_RBRACK] = ACTIONS(666),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(666),
  },
  [STATE(108)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(670),
    [anon_sym_EQ] = ACTIONS(670),
    [anon_sym_BANG_EQ] = ACTIONS(670),
    [anon_sym_PLUS_EQ] = ACTIONS(670),
    [anon_sym_STAR_EQ] = ACTIONS(670),
    [anon_sym_SLASH_EQ] = ACTIONS(670),
    [anon_sym_PERCENT_EQ] = ACTIONS(670),
    [anon_sym_DASH_EQ] = ACTIONS(670),
    [anon_sym_GT] = ACTIONS(672),
    [anon_sym_GT_GT] = ACTIONS(670),
    [anon_sym_2_GT] = ACTIONS(672),
    [anon_sym_2_GT_GT] = ACTIONS(670),
    [anon_sym_3_GT] = ACTIONS(672),
    [anon_sym_3_GT_GT] = ACTIONS(670),
    [anon_sym_4_GT] = ACTIONS(672),
    [anon_sym_4_GT_GT] = ACTIONS(670),
    [anon_sym_5_GT] = ACTIONS(672),
    [anon_sym_5_GT_GT] = ACTIONS(670),
    [anon_sym_6_GT] = ACTIONS(672),
    [anon_sym_6_GT_GT] = ACTIONS(670),
    [anon_sym_STAR_GT] = ACTIONS(672),
    [anon_sym_STAR_GT_GT] = ACTIONS(670),
    [anon_sym_LT] = ACTIONS(672),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(670),
    [anon_sym_2_GT_AMP1] = ACTIONS(670),
    [anon_sym_3_GT_AMP1] = ACTIONS(670),
    [anon_sym_4_GT_AMP1] = ACTIONS(670),
    [anon_sym_5_GT_AMP1] = ACTIONS(670),
    [anon_sym_6_GT_AMP1] = ACTIONS(670),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(670),
    [anon_sym_1_GT_AMP2] = ACTIONS(670),
    [anon_sym_3_GT_AMP2] = ACTIONS(670),
    [anon_sym_4_GT_AMP2] = ACTIONS(670),
    [anon_sym_5_GT_AMP2] = ACTIONS(670),
    [anon_sym_6_GT_AMP2] = ACTIONS(670),
    [aux_sym_comparison_operator_token1] = ACTIONS(670),
    [aux_sym_comparison_operator_token2] = ACTIONS(670),
    [aux_sym_comparison_operator_token3] = ACTIONS(670),
    [aux_sym_comparison_operator_token4] = ACTIONS(670),
    [aux_sym_comparison_operator_token5] = ACTIONS(670),
    [aux_sym_comparison_operator_token6] = ACTIONS(670),
    [aux_sym_comparison_operator_token7] = ACTIONS(670),
    [aux_sym_comparison_operator_token8] = ACTIONS(670),
    [aux_sym_comparison_operator_token9] = ACTIONS(670),
    [aux_sym_comparison_operator_token10] = ACTIONS(670),
    [aux_sym_comparison_operator_token11] = ACTIONS(670),
    [aux_sym_comparison_operator_token12] = ACTIONS(670),
    [aux_sym_comparison_operator_token13] = ACTIONS(670),
    [aux_sym_comparison_operator_token14] = ACTIONS(670),
    [aux_sym_comparison_operator_token15] = ACTIONS(670),
    [aux_sym_comparison_operator_token16] = ACTIONS(670),
    [aux_sym_comparison_operator_token17] = ACTIONS(670),
    [aux_sym_comparison_operator_token18] = ACTIONS(670),
    [aux_sym_comparison_operator_token19] = ACTIONS(670),
    [aux_sym_comparison_operator_token20] = ACTIONS(670),
    [aux_sym_comparison_operator_token21] = ACTIONS(670),
    [aux_sym_comparison_operator_token22] = ACTIONS(670),
    [aux_sym_comparison_operator_token23] = ACTIONS(670),
    [aux_sym_comparison_operator_token24] = ACTIONS(670),
    [aux_sym_comparison_operator_token25] = ACTIONS(670),
    [aux_sym_comparison_operator_token26] = ACTIONS(670),
    [aux_sym_comparison_operator_token27] = ACTIONS(670),
    [aux_sym_comparison_operator_token28] = ACTIONS(672),
    [aux_sym_comparison_operator_token29] = ACTIONS(670),
    [aux_sym_comparison_operator_token30] = ACTIONS(670),
    [aux_sym_comparison_operator_token31] = ACTIONS(670),
    [aux_sym_comparison_operator_token32] = ACTIONS(670),
    [aux_sym_comparison_operator_token33] = ACTIONS(670),
    [aux_sym_comparison_operator_token34] = ACTIONS(672),
    [aux_sym_comparison_operator_token35] = ACTIONS(670),
    [aux_sym_comparison_operator_token36] = ACTIONS(670),
    [aux_sym_comparison_operator_token37] = ACTIONS(670),
    [aux_sym_comparison_operator_token38] = ACTIONS(670),
    [aux_sym_comparison_operator_token39] = ACTIONS(670),
    [aux_sym_comparison_operator_token40] = ACTIONS(670),
    [aux_sym_comparison_operator_token41] = ACTIONS(670),
    [aux_sym_comparison_operator_token42] = ACTIONS(670),
    [aux_sym_comparison_operator_token43] = ACTIONS(670),
    [aux_sym_comparison_operator_token44] = ACTIONS(670),
    [aux_sym_comparison_operator_token45] = ACTIONS(670),
    [aux_sym_comparison_operator_token46] = ACTIONS(670),
    [aux_sym_comparison_operator_token47] = ACTIONS(670),
    [aux_sym_comparison_operator_token48] = ACTIONS(670),
    [aux_sym_comparison_operator_token49] = ACTIONS(670),
    [aux_sym_comparison_operator_token50] = ACTIONS(670),
    [aux_sym_format_operator_token1] = ACTIONS(670),
    [anon_sym_LPAREN] = ACTIONS(670),
    [anon_sym_RPAREN] = ACTIONS(670),
    [anon_sym_COMMA] = ACTIONS(670),
    [anon_sym_AMP_AMP] = ACTIONS(670),
    [anon_sym_PIPE_PIPE] = ACTIONS(670),
    [anon_sym_PIPE] = ACTIONS(672),
    [aux_sym_logical_expression_token1] = ACTIONS(670),
    [aux_sym_logical_expression_token2] = ACTIONS(670),
    [aux_sym_logical_expression_token3] = ACTIONS(670),
    [aux_sym_bitwise_expression_token1] = ACTIONS(670),
    [aux_sym_bitwise_expression_token2] = ACTIONS(670),
    [aux_sym_bitwise_expression_token3] = ACTIONS(670),
    [anon_sym_PLUS] = ACTIONS(672),
    [anon_sym_DASH] = ACTIONS(672),
    [anon_sym_SLASH] = ACTIONS(672),
    [anon_sym_BSLASH] = ACTIONS(670),
    [anon_sym_PERCENT] = ACTIONS(672),
    [anon_sym_STAR] = ACTIONS(672),
    [anon_sym_DOT_DOT] = ACTIONS(670),
    [anon_sym_PLUS_PLUS] = ACTIONS(670),
    [anon_sym_DASH_DASH] = ACTIONS(670),
    [anon_sym_DOT2] = ACTIONS(672),
    [anon_sym_COLON_COLON] = ACTIONS(670),
    [anon_sym_RBRACK] = ACTIONS(670),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(670),
  },
  [STATE(109)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(674),
    [anon_sym_EQ] = ACTIONS(674),
    [anon_sym_BANG_EQ] = ACTIONS(674),
    [anon_sym_PLUS_EQ] = ACTIONS(674),
    [anon_sym_STAR_EQ] = ACTIONS(674),
    [anon_sym_SLASH_EQ] = ACTIONS(674),
    [anon_sym_PERCENT_EQ] = ACTIONS(674),
    [anon_sym_DASH_EQ] = ACTIONS(674),
    [anon_sym_GT] = ACTIONS(676),
    [anon_sym_GT_GT] = ACTIONS(674),
    [anon_sym_2_GT] = ACTIONS(676),
    [anon_sym_2_GT_GT] = ACTIONS(674),
    [anon_sym_3_GT] = ACTIONS(676),
    [anon_sym_3_GT_GT] = ACTIONS(674),
    [anon_sym_4_GT] = ACTIONS(676),
    [anon_sym_4_GT_GT] = ACTIONS(674),
    [anon_sym_5_GT] = ACTIONS(676),
    [anon_sym_5_GT_GT] = ACTIONS(674),
    [anon_sym_6_GT] = ACTIONS(676),
    [anon_sym_6_GT_GT] = ACTIONS(674),
    [anon_sym_STAR_GT] = ACTIONS(676),
    [anon_sym_STAR_GT_GT] = ACTIONS(674),
    [anon_sym_LT] = ACTIONS(676),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(674),
    [anon_sym_2_GT_AMP1] = ACTIONS(674),
    [anon_sym_3_GT_AMP1] = ACTIONS(674),
    [anon_sym_4_GT_AMP1] = ACTIONS(674),
    [anon_sym_5_GT_AMP1] = ACTIONS(674),
    [anon_sym_6_GT_AMP1] = ACTIONS(674),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(674),
    [anon_sym_1_GT_AMP2] = ACTIONS(674),
    [anon_sym_3_GT_AMP2] = ACTIONS(674),
    [anon_sym_4_GT_AMP2] = ACTIONS(674),
    [anon_sym_5_GT_AMP2] = ACTIONS(674),
    [anon_sym_6_GT_AMP2] = ACTIONS(674),
    [aux_sym_comparison_operator_token1] = ACTIONS(674),
    [aux_sym_comparison_operator_token2] = ACTIONS(674),
    [aux_sym_comparison_operator_token3] = ACTIONS(674),
    [aux_sym_comparison_operator_token4] = ACTIONS(674),
    [aux_sym_comparison_operator_token5] = ACTIONS(674),
    [aux_sym_comparison_operator_token6] = ACTIONS(674),
    [aux_sym_comparison_operator_token7] = ACTIONS(674),
    [aux_sym_comparison_operator_token8] = ACTIONS(674),
    [aux_sym_comparison_operator_token9] = ACTIONS(674),
    [aux_sym_comparison_operator_token10] = ACTIONS(674),
    [aux_sym_comparison_operator_token11] = ACTIONS(674),
    [aux_sym_comparison_operator_token12] = ACTIONS(674),
    [aux_sym_comparison_operator_token13] = ACTIONS(674),
    [aux_sym_comparison_operator_token14] = ACTIONS(674),
    [aux_sym_comparison_operator_token15] = ACTIONS(674),
    [aux_sym_comparison_operator_token16] = ACTIONS(674),
    [aux_sym_comparison_operator_token17] = ACTIONS(674),
    [aux_sym_comparison_operator_token18] = ACTIONS(674),
    [aux_sym_comparison_operator_token19] = ACTIONS(674),
    [aux_sym_comparison_operator_token20] = ACTIONS(674),
    [aux_sym_comparison_operator_token21] = ACTIONS(674),
    [aux_sym_comparison_operator_token22] = ACTIONS(674),
    [aux_sym_comparison_operator_token23] = ACTIONS(674),
    [aux_sym_comparison_operator_token24] = ACTIONS(674),
    [aux_sym_comparison_operator_token25] = ACTIONS(674),
    [aux_sym_comparison_operator_token26] = ACTIONS(674),
    [aux_sym_comparison_operator_token27] = ACTIONS(674),
    [aux_sym_comparison_operator_token28] = ACTIONS(676),
    [aux_sym_comparison_operator_token29] = ACTIONS(674),
    [aux_sym_comparison_operator_token30] = ACTIONS(674),
    [aux_sym_comparison_operator_token31] = ACTIONS(674),
    [aux_sym_comparison_operator_token32] = ACTIONS(674),
    [aux_sym_comparison_operator_token33] = ACTIONS(674),
    [aux_sym_comparison_operator_token34] = ACTIONS(676),
    [aux_sym_comparison_operator_token35] = ACTIONS(674),
    [aux_sym_comparison_operator_token36] = ACTIONS(674),
    [aux_sym_comparison_operator_token37] = ACTIONS(674),
    [aux_sym_comparison_operator_token38] = ACTIONS(674),
    [aux_sym_comparison_operator_token39] = ACTIONS(674),
    [aux_sym_comparison_operator_token40] = ACTIONS(674),
    [aux_sym_comparison_operator_token41] = ACTIONS(674),
    [aux_sym_comparison_operator_token42] = ACTIONS(674),
    [aux_sym_comparison_operator_token43] = ACTIONS(674),
    [aux_sym_comparison_operator_token44] = ACTIONS(674),
    [aux_sym_comparison_operator_token45] = ACTIONS(674),
    [aux_sym_comparison_operator_token46] = ACTIONS(674),
    [aux_sym_comparison_operator_token47] = ACTIONS(674),
    [aux_sym_comparison_operator_token48] = ACTIONS(674),
    [aux_sym_comparison_operator_token49] = ACTIONS(674),
    [aux_sym_comparison_operator_token50] = ACTIONS(674),
    [aux_sym_format_operator_token1] = ACTIONS(674),
    [anon_sym_LPAREN] = ACTIONS(674),
    [anon_sym_RPAREN] = ACTIONS(674),
    [anon_sym_COMMA] = ACTIONS(674),
    [anon_sym_AMP_AMP] = ACTIONS(674),
    [anon_sym_PIPE_PIPE] = ACTIONS(674),
    [anon_sym_PIPE] = ACTIONS(676),
    [aux_sym_logical_expression_token1] = ACTIONS(674),
    [aux_sym_logical_expression_token2] = ACTIONS(674),
    [aux_sym_logical_expression_token3] = ACTIONS(674),
    [aux_sym_bitwise_expression_token1] = ACTIONS(674),
    [aux_sym_bitwise_expression_token2] = ACTIONS(674),
    [aux_sym_bitwise_expression_token3] = ACTIONS(674),
    [anon_sym_PLUS] = ACTIONS(676),
    [anon_sym_DASH] = ACTIONS(676),
    [anon_sym_SLASH] = ACTIONS(676),
    [anon_sym_BSLASH] = ACTIONS(674),
    [anon_sym_PERCENT] = ACTIONS(676),
    [anon_sym_STAR] = ACTIONS(676),
    [anon_sym_DOT_DOT] = ACTIONS(674),
    [anon_sym_PLUS_PLUS] = ACTIONS(674),
    [anon_sym_DASH_DASH] = ACTIONS(674),
    [anon_sym_DOT2] = ACTIONS(676),
    [anon_sym_COLON_COLON] = ACTIONS(674),
    [anon_sym_RBRACK] = ACTIONS(674),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(674),
  },
  [STATE(110)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(678),
    [anon_sym_EQ] = ACTIONS(678),
    [anon_sym_BANG_EQ] = ACTIONS(678),
    [anon_sym_PLUS_EQ] = ACTIONS(678),
    [anon_sym_STAR_EQ] = ACTIONS(678),
    [anon_sym_SLASH_EQ] = ACTIONS(678),
    [anon_sym_PERCENT_EQ] = ACTIONS(678),
    [anon_sym_DASH_EQ] = ACTIONS(678),
    [anon_sym_GT] = ACTIONS(680),
    [anon_sym_GT_GT] = ACTIONS(678),
    [anon_sym_2_GT] = ACTIONS(680),
    [anon_sym_2_GT_GT] = ACTIONS(678),
    [anon_sym_3_GT] = ACTIONS(680),
    [anon_sym_3_GT_GT] = ACTIONS(678),
    [anon_sym_4_GT] = ACTIONS(680),
    [anon_sym_4_GT_GT] = ACTIONS(678),
    [anon_sym_5_GT] = ACTIONS(680),
    [anon_sym_5_GT_GT] = ACTIONS(678),
    [anon_sym_6_GT] = ACTIONS(680),
    [anon_sym_6_GT_GT] = ACTIONS(678),
    [anon_sym_STAR_GT] = ACTIONS(680),
    [anon_sym_STAR_GT_GT] = ACTIONS(678),
    [anon_sym_LT] = ACTIONS(680),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(678),
    [anon_sym_2_GT_AMP1] = ACTIONS(678),
    [anon_sym_3_GT_AMP1] = ACTIONS(678),
    [anon_sym_4_GT_AMP1] = ACTIONS(678),
    [anon_sym_5_GT_AMP1] = ACTIONS(678),
    [anon_sym_6_GT_AMP1] = ACTIONS(678),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(678),
    [anon_sym_1_GT_AMP2] = ACTIONS(678),
    [anon_sym_3_GT_AMP2] = ACTIONS(678),
    [anon_sym_4_GT_AMP2] = ACTIONS(678),
    [anon_sym_5_GT_AMP2] = ACTIONS(678),
    [anon_sym_6_GT_AMP2] = ACTIONS(678),
    [aux_sym_comparison_operator_token1] = ACTIONS(678),
    [aux_sym_comparison_operator_token2] = ACTIONS(678),
    [aux_sym_comparison_operator_token3] = ACTIONS(678),
    [aux_sym_comparison_operator_token4] = ACTIONS(678),
    [aux_sym_comparison_operator_token5] = ACTIONS(678),
    [aux_sym_comparison_operator_token6] = ACTIONS(678),
    [aux_sym_comparison_operator_token7] = ACTIONS(678),
    [aux_sym_comparison_operator_token8] = ACTIONS(678),
    [aux_sym_comparison_operator_token9] = ACTIONS(678),
    [aux_sym_comparison_operator_token10] = ACTIONS(678),
    [aux_sym_comparison_operator_token11] = ACTIONS(678),
    [aux_sym_comparison_operator_token12] = ACTIONS(678),
    [aux_sym_comparison_operator_token13] = ACTIONS(678),
    [aux_sym_comparison_operator_token14] = ACTIONS(678),
    [aux_sym_comparison_operator_token15] = ACTIONS(678),
    [aux_sym_comparison_operator_token16] = ACTIONS(678),
    [aux_sym_comparison_operator_token17] = ACTIONS(678),
    [aux_sym_comparison_operator_token18] = ACTIONS(678),
    [aux_sym_comparison_operator_token19] = ACTIONS(678),
    [aux_sym_comparison_operator_token20] = ACTIONS(678),
    [aux_sym_comparison_operator_token21] = ACTIONS(678),
    [aux_sym_comparison_operator_token22] = ACTIONS(678),
    [aux_sym_comparison_operator_token23] = ACTIONS(678),
    [aux_sym_comparison_operator_token24] = ACTIONS(678),
    [aux_sym_comparison_operator_token25] = ACTIONS(678),
    [aux_sym_comparison_operator_token26] = ACTIONS(678),
    [aux_sym_comparison_operator_token27] = ACTIONS(678),
    [aux_sym_comparison_operator_token28] = ACTIONS(680),
    [aux_sym_comparison_operator_token29] = ACTIONS(678),
    [aux_sym_comparison_operator_token30] = ACTIONS(678),
    [aux_sym_comparison_operator_token31] = ACTIONS(678),
    [aux_sym_comparison_operator_token32] = ACTIONS(678),
    [aux_sym_comparison_operator_token33] = ACTIONS(678),
    [aux_sym_comparison_operator_token34] = ACTIONS(680),
    [aux_sym_comparison_operator_token35] = ACTIONS(678),
    [aux_sym_comparison_operator_token36] = ACTIONS(678),
    [aux_sym_comparison_operator_token37] = ACTIONS(678),
    [aux_sym_comparison_operator_token38] = ACTIONS(678),
    [aux_sym_comparison_operator_token39] = ACTIONS(678),
    [aux_sym_comparison_operator_token40] = ACTIONS(678),
    [aux_sym_comparison_operator_token41] = ACTIONS(678),
    [aux_sym_comparison_operator_token42] = ACTIONS(678),
    [aux_sym_comparison_operator_token43] = ACTIONS(678),
    [aux_sym_comparison_operator_token44] = ACTIONS(678),
    [aux_sym_comparison_operator_token45] = ACTIONS(678),
    [aux_sym_comparison_operator_token46] = ACTIONS(678),
    [aux_sym_comparison_operator_token47] = ACTIONS(678),
    [aux_sym_comparison_operator_token48] = ACTIONS(678),
    [aux_sym_comparison_operator_token49] = ACTIONS(678),
    [aux_sym_comparison_operator_token50] = ACTIONS(678),
    [aux_sym_format_operator_token1] = ACTIONS(678),
    [anon_sym_LPAREN] = ACTIONS(678),
    [anon_sym_RPAREN] = ACTIONS(678),
    [anon_sym_COMMA] = ACTIONS(678),
    [anon_sym_AMP_AMP] = ACTIONS(678),
    [anon_sym_PIPE_PIPE] = ACTIONS(678),
    [anon_sym_PIPE] = ACTIONS(680),
    [aux_sym_logical_expression_token1] = ACTIONS(678),
    [aux_sym_logical_expression_token2] = ACTIONS(678),
    [aux_sym_logical_expression_token3] = ACTIONS(678),
    [aux_sym_bitwise_expression_token1] = ACTIONS(678),
    [aux_sym_bitwise_expression_token2] = ACTIONS(678),
    [aux_sym_bitwise_expression_token3] = ACTIONS(678),
    [anon_sym_PLUS] = ACTIONS(680),
    [anon_sym_DASH] = ACTIONS(680),
    [anon_sym_SLASH] = ACTIONS(680),
    [anon_sym_BSLASH] = ACTIONS(678),
    [anon_sym_PERCENT] = ACTIONS(680),
    [anon_sym_STAR] = ACTIONS(680),
    [anon_sym_DOT_DOT] = ACTIONS(678),
    [anon_sym_PLUS_PLUS] = ACTIONS(678),
    [anon_sym_DASH_DASH] = ACTIONS(678),
    [anon_sym_DOT2] = ACTIONS(680),
    [anon_sym_COLON_COLON] = ACTIONS(678),
    [anon_sym_RBRACK] = ACTIONS(678),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(678),
  },
  [STATE(111)] = {
    [sym_argument_list] = STATE(144),
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(590),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_BANG_EQ] = ACTIONS(590),
    [anon_sym_PLUS_EQ] = ACTIONS(590),
    [anon_sym_STAR_EQ] = ACTIONS(590),
    [anon_sym_SLASH_EQ] = ACTIONS(590),
    [anon_sym_PERCENT_EQ] = ACTIONS(590),
    [anon_sym_DASH_EQ] = ACTIONS(590),
    [anon_sym_GT] = ACTIONS(592),
    [anon_sym_GT_GT] = ACTIONS(590),
    [anon_sym_2_GT] = ACTIONS(592),
    [anon_sym_2_GT_GT] = ACTIONS(590),
    [anon_sym_3_GT] = ACTIONS(592),
    [anon_sym_3_GT_GT] = ACTIONS(590),
    [anon_sym_4_GT] = ACTIONS(592),
    [anon_sym_4_GT_GT] = ACTIONS(590),
    [anon_sym_5_GT] = ACTIONS(592),
    [anon_sym_5_GT_GT] = ACTIONS(590),
    [anon_sym_6_GT] = ACTIONS(592),
    [anon_sym_6_GT_GT] = ACTIONS(590),
    [anon_sym_STAR_GT] = ACTIONS(592),
    [anon_sym_STAR_GT_GT] = ACTIONS(590),
    [anon_sym_LT] = ACTIONS(592),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(590),
    [anon_sym_2_GT_AMP1] = ACTIONS(590),
    [anon_sym_3_GT_AMP1] = ACTIONS(590),
    [anon_sym_4_GT_AMP1] = ACTIONS(590),
    [anon_sym_5_GT_AMP1] = ACTIONS(590),
    [anon_sym_6_GT_AMP1] = ACTIONS(590),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(590),
    [anon_sym_1_GT_AMP2] = ACTIONS(590),
    [anon_sym_3_GT_AMP2] = ACTIONS(590),
    [anon_sym_4_GT_AMP2] = ACTIONS(590),
    [anon_sym_5_GT_AMP2] = ACTIONS(590),
    [anon_sym_6_GT_AMP2] = ACTIONS(590),
    [aux_sym_comparison_operator_token1] = ACTIONS(590),
    [aux_sym_comparison_operator_token2] = ACTIONS(590),
    [aux_sym_comparison_operator_token3] = ACTIONS(590),
    [aux_sym_comparison_operator_token4] = ACTIONS(590),
    [aux_sym_comparison_operator_token5] = ACTIONS(590),
    [aux_sym_comparison_operator_token6] = ACTIONS(590),
    [aux_sym_comparison_operator_token7] = ACTIONS(590),
    [aux_sym_comparison_operator_token8] = ACTIONS(590),
    [aux_sym_comparison_operator_token9] = ACTIONS(590),
    [aux_sym_comparison_operator_token10] = ACTIONS(590),
    [aux_sym_comparison_operator_token11] = ACTIONS(590),
    [aux_sym_comparison_operator_token12] = ACTIONS(590),
    [aux_sym_comparison_operator_token13] = ACTIONS(590),
    [aux_sym_comparison_operator_token14] = ACTIONS(590),
    [aux_sym_comparison_operator_token15] = ACTIONS(590),
    [aux_sym_comparison_operator_token16] = ACTIONS(590),
    [aux_sym_comparison_operator_token17] = ACTIONS(590),
    [aux_sym_comparison_operator_token18] = ACTIONS(590),
    [aux_sym_comparison_operator_token19] = ACTIONS(590),
    [aux_sym_comparison_operator_token20] = ACTIONS(590),
    [aux_sym_comparison_operator_token21] = ACTIONS(590),
    [aux_sym_comparison_operator_token22] = ACTIONS(590),
    [aux_sym_comparison_operator_token23] = ACTIONS(590),
    [aux_sym_comparison_operator_token24] = ACTIONS(590),
    [aux_sym_comparison_operator_token25] = ACTIONS(590),
    [aux_sym_comparison_operator_token26] = ACTIONS(590),
    [aux_sym_comparison_operator_token27] = ACTIONS(590),
    [aux_sym_comparison_operator_token28] = ACTIONS(592),
    [aux_sym_comparison_operator_token29] = ACTIONS(590),
    [aux_sym_comparison_operator_token30] = ACTIONS(590),
    [aux_sym_comparison_operator_token31] = ACTIONS(590),
    [aux_sym_comparison_operator_token32] = ACTIONS(590),
    [aux_sym_comparison_operator_token33] = ACTIONS(590),
    [aux_sym_comparison_operator_token34] = ACTIONS(592),
    [aux_sym_comparison_operator_token35] = ACTIONS(590),
    [aux_sym_comparison_operator_token36] = ACTIONS(590),
    [aux_sym_comparison_operator_token37] = ACTIONS(590),
    [aux_sym_comparison_operator_token38] = ACTIONS(590),
    [aux_sym_comparison_operator_token39] = ACTIONS(590),
    [aux_sym_comparison_operator_token40] = ACTIONS(590),
    [aux_sym_comparison_operator_token41] = ACTIONS(590),
    [aux_sym_comparison_operator_token42] = ACTIONS(590),
    [aux_sym_comparison_operator_token43] = ACTIONS(590),
    [aux_sym_comparison_operator_token44] = ACTIONS(590),
    [aux_sym_comparison_operator_token45] = ACTIONS(590),
    [aux_sym_comparison_operator_token46] = ACTIONS(590),
    [aux_sym_comparison_operator_token47] = ACTIONS(590),
    [aux_sym_comparison_operator_token48] = ACTIONS(590),
    [aux_sym_comparison_operator_token49] = ACTIONS(590),
    [aux_sym_comparison_operator_token50] = ACTIONS(590),
    [aux_sym_format_operator_token1] = ACTIONS(590),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_COMMA] = ACTIONS(590),
    [anon_sym_AMP_AMP] = ACTIONS(590),
    [anon_sym_PIPE_PIPE] = ACTIONS(590),
    [anon_sym_PIPE] = ACTIONS(592),
    [aux_sym_logical_expression_token1] = ACTIONS(590),
    [aux_sym_logical_expression_token2] = ACTIONS(590),
    [aux_sym_logical_expression_token3] = ACTIONS(590),
    [aux_sym_bitwise_expression_token1] = ACTIONS(590),
    [aux_sym_bitwise_expression_token2] = ACTIONS(590),
    [aux_sym_bitwise_expression_token3] = ACTIONS(590),
    [anon_sym_PLUS] = ACTIONS(592),
    [anon_sym_DASH] = ACTIONS(592),
    [anon_sym_SLASH] = ACTIONS(592),
    [anon_sym_BSLASH] = ACTIONS(590),
    [anon_sym_PERCENT] = ACTIONS(592),
    [anon_sym_STAR] = ACTIONS(592),
    [anon_sym_DOT_DOT] = ACTIONS(590),
    [anon_sym_PLUS_PLUS] = ACTIONS(590),
    [anon_sym_DASH_DASH] = ACTIONS(590),
    [anon_sym_DOT2] = ACTIONS(592),
    [anon_sym_COLON_COLON] = ACTIONS(590),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(590),
    [sym__statement_terminator] = ACTIONS(590),
  },
  [STATE(112)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(682),
    [anon_sym_EQ] = ACTIONS(682),
    [anon_sym_BANG_EQ] = ACTIONS(682),
    [anon_sym_PLUS_EQ] = ACTIONS(682),
    [anon_sym_STAR_EQ] = ACTIONS(682),
    [anon_sym_SLASH_EQ] = ACTIONS(682),
    [anon_sym_PERCENT_EQ] = ACTIONS(682),
    [anon_sym_DASH_EQ] = ACTIONS(682),
    [anon_sym_GT] = ACTIONS(684),
    [anon_sym_GT_GT] = ACTIONS(682),
    [anon_sym_2_GT] = ACTIONS(684),
    [anon_sym_2_GT_GT] = ACTIONS(682),
    [anon_sym_3_GT] = ACTIONS(684),
    [anon_sym_3_GT_GT] = ACTIONS(682),
    [anon_sym_4_GT] = ACTIONS(684),
    [anon_sym_4_GT_GT] = ACTIONS(682),
    [anon_sym_5_GT] = ACTIONS(684),
    [anon_sym_5_GT_GT] = ACTIONS(682),
    [anon_sym_6_GT] = ACTIONS(684),
    [anon_sym_6_GT_GT] = ACTIONS(682),
    [anon_sym_STAR_GT] = ACTIONS(684),
    [anon_sym_STAR_GT_GT] = ACTIONS(682),
    [anon_sym_LT] = ACTIONS(684),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(682),
    [anon_sym_2_GT_AMP1] = ACTIONS(682),
    [anon_sym_3_GT_AMP1] = ACTIONS(682),
    [anon_sym_4_GT_AMP1] = ACTIONS(682),
    [anon_sym_5_GT_AMP1] = ACTIONS(682),
    [anon_sym_6_GT_AMP1] = ACTIONS(682),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(682),
    [anon_sym_1_GT_AMP2] = ACTIONS(682),
    [anon_sym_3_GT_AMP2] = ACTIONS(682),
    [anon_sym_4_GT_AMP2] = ACTIONS(682),
    [anon_sym_5_GT_AMP2] = ACTIONS(682),
    [anon_sym_6_GT_AMP2] = ACTIONS(682),
    [aux_sym_comparison_operator_token1] = ACTIONS(682),
    [aux_sym_comparison_operator_token2] = ACTIONS(682),
    [aux_sym_comparison_operator_token3] = ACTIONS(682),
    [aux_sym_comparison_operator_token4] = ACTIONS(682),
    [aux_sym_comparison_operator_token5] = ACTIONS(682),
    [aux_sym_comparison_operator_token6] = ACTIONS(682),
    [aux_sym_comparison_operator_token7] = ACTIONS(682),
    [aux_sym_comparison_operator_token8] = ACTIONS(682),
    [aux_sym_comparison_operator_token9] = ACTIONS(682),
    [aux_sym_comparison_operator_token10] = ACTIONS(682),
    [aux_sym_comparison_operator_token11] = ACTIONS(682),
    [aux_sym_comparison_operator_token12] = ACTIONS(682),
    [aux_sym_comparison_operator_token13] = ACTIONS(682),
    [aux_sym_comparison_operator_token14] = ACTIONS(682),
    [aux_sym_comparison_operator_token15] = ACTIONS(682),
    [aux_sym_comparison_operator_token16] = ACTIONS(682),
    [aux_sym_comparison_operator_token17] = ACTIONS(682),
    [aux_sym_comparison_operator_token18] = ACTIONS(682),
    [aux_sym_comparison_operator_token19] = ACTIONS(682),
    [aux_sym_comparison_operator_token20] = ACTIONS(682),
    [aux_sym_comparison_operator_token21] = ACTIONS(682),
    [aux_sym_comparison_operator_token22] = ACTIONS(682),
    [aux_sym_comparison_operator_token23] = ACTIONS(682),
    [aux_sym_comparison_operator_token24] = ACTIONS(682),
    [aux_sym_comparison_operator_token25] = ACTIONS(682),
    [aux_sym_comparison_operator_token26] = ACTIONS(682),
    [aux_sym_comparison_operator_token27] = ACTIONS(682),
    [aux_sym_comparison_operator_token28] = ACTIONS(684),
    [aux_sym_comparison_operator_token29] = ACTIONS(682),
    [aux_sym_comparison_operator_token30] = ACTIONS(682),
    [aux_sym_comparison_operator_token31] = ACTIONS(682),
    [aux_sym_comparison_operator_token32] = ACTIONS(682),
    [aux_sym_comparison_operator_token33] = ACTIONS(682),
    [aux_sym_comparison_operator_token34] = ACTIONS(684),
    [aux_sym_comparison_operator_token35] = ACTIONS(682),
    [aux_sym_comparison_operator_token36] = ACTIONS(682),
    [aux_sym_comparison_operator_token37] = ACTIONS(682),
    [aux_sym_comparison_operator_token38] = ACTIONS(682),
    [aux_sym_comparison_operator_token39] = ACTIONS(682),
    [aux_sym_comparison_operator_token40] = ACTIONS(682),
    [aux_sym_comparison_operator_token41] = ACTIONS(682),
    [aux_sym_comparison_operator_token42] = ACTIONS(682),
    [aux_sym_comparison_operator_token43] = ACTIONS(682),
    [aux_sym_comparison_operator_token44] = ACTIONS(682),
    [aux_sym_comparison_operator_token45] = ACTIONS(682),
    [aux_sym_comparison_operator_token46] = ACTIONS(682),
    [aux_sym_comparison_operator_token47] = ACTIONS(682),
    [aux_sym_comparison_operator_token48] = ACTIONS(682),
    [aux_sym_comparison_operator_token49] = ACTIONS(682),
    [aux_sym_comparison_operator_token50] = ACTIONS(682),
    [aux_sym_format_operator_token1] = ACTIONS(682),
    [anon_sym_LPAREN] = ACTIONS(682),
    [anon_sym_RPAREN] = ACTIONS(682),
    [anon_sym_COMMA] = ACTIONS(682),
    [anon_sym_AMP_AMP] = ACTIONS(682),
    [anon_sym_PIPE_PIPE] = ACTIONS(682),
    [anon_sym_PIPE] = ACTIONS(684),
    [aux_sym_logical_expression_token1] = ACTIONS(682),
    [aux_sym_logical_expression_token2] = ACTIONS(682),
    [aux_sym_logical_expression_token3] = ACTIONS(682),
    [aux_sym_bitwise_expression_token1] = ACTIONS(682),
    [aux_sym_bitwise_expression_token2] = ACTIONS(682),
    [aux_sym_bitwise_expression_token3] = ACTIONS(682),
    [anon_sym_PLUS] = ACTIONS(684),
    [anon_sym_DASH] = ACTIONS(684),
    [anon_sym_SLASH] = ACTIONS(684),
    [anon_sym_BSLASH] = ACTIONS(682),
    [anon_sym_PERCENT] = ACTIONS(684),
    [anon_sym_STAR] = ACTIONS(684),
    [anon_sym_DOT_DOT] = ACTIONS(682),
    [anon_sym_PLUS_PLUS] = ACTIONS(682),
    [anon_sym_DASH_DASH] = ACTIONS(682),
    [anon_sym_DOT2] = ACTIONS(684),
    [anon_sym_COLON_COLON] = ACTIONS(682),
    [anon_sym_RBRACK] = ACTIONS(682),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(682),
  },
  [STATE(113)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(686),
    [anon_sym_EQ] = ACTIONS(686),
    [anon_sym_BANG_EQ] = ACTIONS(686),
    [anon_sym_PLUS_EQ] = ACTIONS(686),
    [anon_sym_STAR_EQ] = ACTIONS(686),
    [anon_sym_SLASH_EQ] = ACTIONS(686),
    [anon_sym_PERCENT_EQ] = ACTIONS(686),
    [anon_sym_DASH_EQ] = ACTIONS(686),
    [anon_sym_GT] = ACTIONS(688),
    [anon_sym_GT_GT] = ACTIONS(686),
    [anon_sym_2_GT] = ACTIONS(688),
    [anon_sym_2_GT_GT] = ACTIONS(686),
    [anon_sym_3_GT] = ACTIONS(688),
    [anon_sym_3_GT_GT] = ACTIONS(686),
    [anon_sym_4_GT] = ACTIONS(688),
    [anon_sym_4_GT_GT] = ACTIONS(686),
    [anon_sym_5_GT] = ACTIONS(688),
    [anon_sym_5_GT_GT] = ACTIONS(686),
    [anon_sym_6_GT] = ACTIONS(688),
    [anon_sym_6_GT_GT] = ACTIONS(686),
    [anon_sym_STAR_GT] = ACTIONS(688),
    [anon_sym_STAR_GT_GT] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(688),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(686),
    [anon_sym_2_GT_AMP1] = ACTIONS(686),
    [anon_sym_3_GT_AMP1] = ACTIONS(686),
    [anon_sym_4_GT_AMP1] = ACTIONS(686),
    [anon_sym_5_GT_AMP1] = ACTIONS(686),
    [anon_sym_6_GT_AMP1] = ACTIONS(686),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(686),
    [anon_sym_1_GT_AMP2] = ACTIONS(686),
    [anon_sym_3_GT_AMP2] = ACTIONS(686),
    [anon_sym_4_GT_AMP2] = ACTIONS(686),
    [anon_sym_5_GT_AMP2] = ACTIONS(686),
    [anon_sym_6_GT_AMP2] = ACTIONS(686),
    [aux_sym_comparison_operator_token1] = ACTIONS(686),
    [aux_sym_comparison_operator_token2] = ACTIONS(686),
    [aux_sym_comparison_operator_token3] = ACTIONS(686),
    [aux_sym_comparison_operator_token4] = ACTIONS(686),
    [aux_sym_comparison_operator_token5] = ACTIONS(686),
    [aux_sym_comparison_operator_token6] = ACTIONS(686),
    [aux_sym_comparison_operator_token7] = ACTIONS(686),
    [aux_sym_comparison_operator_token8] = ACTIONS(686),
    [aux_sym_comparison_operator_token9] = ACTIONS(686),
    [aux_sym_comparison_operator_token10] = ACTIONS(686),
    [aux_sym_comparison_operator_token11] = ACTIONS(686),
    [aux_sym_comparison_operator_token12] = ACTIONS(686),
    [aux_sym_comparison_operator_token13] = ACTIONS(686),
    [aux_sym_comparison_operator_token14] = ACTIONS(686),
    [aux_sym_comparison_operator_token15] = ACTIONS(686),
    [aux_sym_comparison_operator_token16] = ACTIONS(686),
    [aux_sym_comparison_operator_token17] = ACTIONS(686),
    [aux_sym_comparison_operator_token18] = ACTIONS(686),
    [aux_sym_comparison_operator_token19] = ACTIONS(686),
    [aux_sym_comparison_operator_token20] = ACTIONS(686),
    [aux_sym_comparison_operator_token21] = ACTIONS(686),
    [aux_sym_comparison_operator_token22] = ACTIONS(686),
    [aux_sym_comparison_operator_token23] = ACTIONS(686),
    [aux_sym_comparison_operator_token24] = ACTIONS(686),
    [aux_sym_comparison_operator_token25] = ACTIONS(686),
    [aux_sym_comparison_operator_token26] = ACTIONS(686),
    [aux_sym_comparison_operator_token27] = ACTIONS(686),
    [aux_sym_comparison_operator_token28] = ACTIONS(688),
    [aux_sym_comparison_operator_token29] = ACTIONS(686),
    [aux_sym_comparison_operator_token30] = ACTIONS(686),
    [aux_sym_comparison_operator_token31] = ACTIONS(686),
    [aux_sym_comparison_operator_token32] = ACTIONS(686),
    [aux_sym_comparison_operator_token33] = ACTIONS(686),
    [aux_sym_comparison_operator_token34] = ACTIONS(688),
    [aux_sym_comparison_operator_token35] = ACTIONS(686),
    [aux_sym_comparison_operator_token36] = ACTIONS(686),
    [aux_sym_comparison_operator_token37] = ACTIONS(686),
    [aux_sym_comparison_operator_token38] = ACTIONS(686),
    [aux_sym_comparison_operator_token39] = ACTIONS(686),
    [aux_sym_comparison_operator_token40] = ACTIONS(686),
    [aux_sym_comparison_operator_token41] = ACTIONS(686),
    [aux_sym_comparison_operator_token42] = ACTIONS(686),
    [aux_sym_comparison_operator_token43] = ACTIONS(686),
    [aux_sym_comparison_operator_token44] = ACTIONS(686),
    [aux_sym_comparison_operator_token45] = ACTIONS(686),
    [aux_sym_comparison_operator_token46] = ACTIONS(686),
    [aux_sym_comparison_operator_token47] = ACTIONS(686),
    [aux_sym_comparison_operator_token48] = ACTIONS(686),
    [aux_sym_comparison_operator_token49] = ACTIONS(686),
    [aux_sym_comparison_operator_token50] = ACTIONS(686),
    [aux_sym_format_operator_token1] = ACTIONS(686),
    [anon_sym_LPAREN] = ACTIONS(686),
    [anon_sym_RPAREN] = ACTIONS(686),
    [anon_sym_COMMA] = ACTIONS(686),
    [anon_sym_AMP_AMP] = ACTIONS(686),
    [anon_sym_PIPE_PIPE] = ACTIONS(686),
    [anon_sym_PIPE] = ACTIONS(688),
    [aux_sym_logical_expression_token1] = ACTIONS(686),
    [aux_sym_logical_expression_token2] = ACTIONS(686),
    [aux_sym_logical_expression_token3] = ACTIONS(686),
    [aux_sym_bitwise_expression_token1] = ACTIONS(686),
    [aux_sym_bitwise_expression_token2] = ACTIONS(686),
    [aux_sym_bitwise_expression_token3] = ACTIONS(686),
    [anon_sym_PLUS] = ACTIONS(688),
    [anon_sym_DASH] = ACTIONS(688),
    [anon_sym_SLASH] = ACTIONS(688),
    [anon_sym_BSLASH] = ACTIONS(686),
    [anon_sym_PERCENT] = ACTIONS(688),
    [anon_sym_STAR] = ACTIONS(688),
    [anon_sym_DOT_DOT] = ACTIONS(686),
    [anon_sym_PLUS_PLUS] = ACTIONS(686),
    [anon_sym_DASH_DASH] = ACTIONS(686),
    [anon_sym_DOT2] = ACTIONS(688),
    [anon_sym_COLON_COLON] = ACTIONS(686),
    [anon_sym_RBRACK] = ACTIONS(686),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(686),
  },
  [STATE(114)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(690),
    [anon_sym_EQ] = ACTIONS(690),
    [anon_sym_BANG_EQ] = ACTIONS(690),
    [anon_sym_PLUS_EQ] = ACTIONS(690),
    [anon_sym_STAR_EQ] = ACTIONS(690),
    [anon_sym_SLASH_EQ] = ACTIONS(690),
    [anon_sym_PERCENT_EQ] = ACTIONS(690),
    [anon_sym_DASH_EQ] = ACTIONS(690),
    [anon_sym_GT] = ACTIONS(692),
    [anon_sym_GT_GT] = ACTIONS(690),
    [anon_sym_2_GT] = ACTIONS(692),
    [anon_sym_2_GT_GT] = ACTIONS(690),
    [anon_sym_3_GT] = ACTIONS(692),
    [anon_sym_3_GT_GT] = ACTIONS(690),
    [anon_sym_4_GT] = ACTIONS(692),
    [anon_sym_4_GT_GT] = ACTIONS(690),
    [anon_sym_5_GT] = ACTIONS(692),
    [anon_sym_5_GT_GT] = ACTIONS(690),
    [anon_sym_6_GT] = ACTIONS(692),
    [anon_sym_6_GT_GT] = ACTIONS(690),
    [anon_sym_STAR_GT] = ACTIONS(692),
    [anon_sym_STAR_GT_GT] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(690),
    [anon_sym_2_GT_AMP1] = ACTIONS(690),
    [anon_sym_3_GT_AMP1] = ACTIONS(690),
    [anon_sym_4_GT_AMP1] = ACTIONS(690),
    [anon_sym_5_GT_AMP1] = ACTIONS(690),
    [anon_sym_6_GT_AMP1] = ACTIONS(690),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(690),
    [anon_sym_1_GT_AMP2] = ACTIONS(690),
    [anon_sym_3_GT_AMP2] = ACTIONS(690),
    [anon_sym_4_GT_AMP2] = ACTIONS(690),
    [anon_sym_5_GT_AMP2] = ACTIONS(690),
    [anon_sym_6_GT_AMP2] = ACTIONS(690),
    [aux_sym_comparison_operator_token1] = ACTIONS(690),
    [aux_sym_comparison_operator_token2] = ACTIONS(690),
    [aux_sym_comparison_operator_token3] = ACTIONS(690),
    [aux_sym_comparison_operator_token4] = ACTIONS(690),
    [aux_sym_comparison_operator_token5] = ACTIONS(690),
    [aux_sym_comparison_operator_token6] = ACTIONS(690),
    [aux_sym_comparison_operator_token7] = ACTIONS(690),
    [aux_sym_comparison_operator_token8] = ACTIONS(690),
    [aux_sym_comparison_operator_token9] = ACTIONS(690),
    [aux_sym_comparison_operator_token10] = ACTIONS(690),
    [aux_sym_comparison_operator_token11] = ACTIONS(690),
    [aux_sym_comparison_operator_token12] = ACTIONS(690),
    [aux_sym_comparison_operator_token13] = ACTIONS(690),
    [aux_sym_comparison_operator_token14] = ACTIONS(690),
    [aux_sym_comparison_operator_token15] = ACTIONS(690),
    [aux_sym_comparison_operator_token16] = ACTIONS(690),
    [aux_sym_comparison_operator_token17] = ACTIONS(690),
    [aux_sym_comparison_operator_token18] = ACTIONS(690),
    [aux_sym_comparison_operator_token19] = ACTIONS(690),
    [aux_sym_comparison_operator_token20] = ACTIONS(690),
    [aux_sym_comparison_operator_token21] = ACTIONS(690),
    [aux_sym_comparison_operator_token22] = ACTIONS(690),
    [aux_sym_comparison_operator_token23] = ACTIONS(690),
    [aux_sym_comparison_operator_token24] = ACTIONS(690),
    [aux_sym_comparison_operator_token25] = ACTIONS(690),
    [aux_sym_comparison_operator_token26] = ACTIONS(690),
    [aux_sym_comparison_operator_token27] = ACTIONS(690),
    [aux_sym_comparison_operator_token28] = ACTIONS(692),
    [aux_sym_comparison_operator_token29] = ACTIONS(690),
    [aux_sym_comparison_operator_token30] = ACTIONS(690),
    [aux_sym_comparison_operator_token31] = ACTIONS(690),
    [aux_sym_comparison_operator_token32] = ACTIONS(690),
    [aux_sym_comparison_operator_token33] = ACTIONS(690),
    [aux_sym_comparison_operator_token34] = ACTIONS(692),
    [aux_sym_comparison_operator_token35] = ACTIONS(690),
    [aux_sym_comparison_operator_token36] = ACTIONS(690),
    [aux_sym_comparison_operator_token37] = ACTIONS(690),
    [aux_sym_comparison_operator_token38] = ACTIONS(690),
    [aux_sym_comparison_operator_token39] = ACTIONS(690),
    [aux_sym_comparison_operator_token40] = ACTIONS(690),
    [aux_sym_comparison_operator_token41] = ACTIONS(690),
    [aux_sym_comparison_operator_token42] = ACTIONS(690),
    [aux_sym_comparison_operator_token43] = ACTIONS(690),
    [aux_sym_comparison_operator_token44] = ACTIONS(690),
    [aux_sym_comparison_operator_token45] = ACTIONS(690),
    [aux_sym_comparison_operator_token46] = ACTIONS(690),
    [aux_sym_comparison_operator_token47] = ACTIONS(690),
    [aux_sym_comparison_operator_token48] = ACTIONS(690),
    [aux_sym_comparison_operator_token49] = ACTIONS(690),
    [aux_sym_comparison_operator_token50] = ACTIONS(690),
    [aux_sym_format_operator_token1] = ACTIONS(690),
    [anon_sym_LPAREN] = ACTIONS(690),
    [anon_sym_RPAREN] = ACTIONS(690),
    [anon_sym_COMMA] = ACTIONS(690),
    [anon_sym_AMP_AMP] = ACTIONS(690),
    [anon_sym_PIPE_PIPE] = ACTIONS(690),
    [anon_sym_PIPE] = ACTIONS(692),
    [aux_sym_logical_expression_token1] = ACTIONS(690),
    [aux_sym_logical_expression_token2] = ACTIONS(690),
    [aux_sym_logical_expression_token3] = ACTIONS(690),
    [aux_sym_bitwise_expression_token1] = ACTIONS(690),
    [aux_sym_bitwise_expression_token2] = ACTIONS(690),
    [aux_sym_bitwise_expression_token3] = ACTIONS(690),
    [anon_sym_PLUS] = ACTIONS(692),
    [anon_sym_DASH] = ACTIONS(692),
    [anon_sym_SLASH] = ACTIONS(692),
    [anon_sym_BSLASH] = ACTIONS(690),
    [anon_sym_PERCENT] = ACTIONS(692),
    [anon_sym_STAR] = ACTIONS(692),
    [anon_sym_DOT_DOT] = ACTIONS(690),
    [anon_sym_PLUS_PLUS] = ACTIONS(690),
    [anon_sym_DASH_DASH] = ACTIONS(690),
    [anon_sym_DOT2] = ACTIONS(692),
    [anon_sym_COLON_COLON] = ACTIONS(690),
    [anon_sym_RBRACK] = ACTIONS(690),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(690),
  },
  [STATE(115)] = {
    [sym_argument_list] = STATE(144),
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(590),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_BANG_EQ] = ACTIONS(590),
    [anon_sym_PLUS_EQ] = ACTIONS(590),
    [anon_sym_STAR_EQ] = ACTIONS(590),
    [anon_sym_SLASH_EQ] = ACTIONS(590),
    [anon_sym_PERCENT_EQ] = ACTIONS(590),
    [anon_sym_DASH_EQ] = ACTIONS(590),
    [anon_sym_GT] = ACTIONS(592),
    [anon_sym_GT_GT] = ACTIONS(590),
    [anon_sym_2_GT] = ACTIONS(592),
    [anon_sym_2_GT_GT] = ACTIONS(590),
    [anon_sym_3_GT] = ACTIONS(592),
    [anon_sym_3_GT_GT] = ACTIONS(590),
    [anon_sym_4_GT] = ACTIONS(592),
    [anon_sym_4_GT_GT] = ACTIONS(590),
    [anon_sym_5_GT] = ACTIONS(592),
    [anon_sym_5_GT_GT] = ACTIONS(590),
    [anon_sym_6_GT] = ACTIONS(592),
    [anon_sym_6_GT_GT] = ACTIONS(590),
    [anon_sym_STAR_GT] = ACTIONS(592),
    [anon_sym_STAR_GT_GT] = ACTIONS(590),
    [anon_sym_LT] = ACTIONS(592),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(590),
    [anon_sym_2_GT_AMP1] = ACTIONS(590),
    [anon_sym_3_GT_AMP1] = ACTIONS(590),
    [anon_sym_4_GT_AMP1] = ACTIONS(590),
    [anon_sym_5_GT_AMP1] = ACTIONS(590),
    [anon_sym_6_GT_AMP1] = ACTIONS(590),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(590),
    [anon_sym_1_GT_AMP2] = ACTIONS(590),
    [anon_sym_3_GT_AMP2] = ACTIONS(590),
    [anon_sym_4_GT_AMP2] = ACTIONS(590),
    [anon_sym_5_GT_AMP2] = ACTIONS(590),
    [anon_sym_6_GT_AMP2] = ACTIONS(590),
    [aux_sym_comparison_operator_token1] = ACTIONS(590),
    [aux_sym_comparison_operator_token2] = ACTIONS(590),
    [aux_sym_comparison_operator_token3] = ACTIONS(590),
    [aux_sym_comparison_operator_token4] = ACTIONS(590),
    [aux_sym_comparison_operator_token5] = ACTIONS(590),
    [aux_sym_comparison_operator_token6] = ACTIONS(590),
    [aux_sym_comparison_operator_token7] = ACTIONS(590),
    [aux_sym_comparison_operator_token8] = ACTIONS(590),
    [aux_sym_comparison_operator_token9] = ACTIONS(590),
    [aux_sym_comparison_operator_token10] = ACTIONS(590),
    [aux_sym_comparison_operator_token11] = ACTIONS(590),
    [aux_sym_comparison_operator_token12] = ACTIONS(590),
    [aux_sym_comparison_operator_token13] = ACTIONS(590),
    [aux_sym_comparison_operator_token14] = ACTIONS(590),
    [aux_sym_comparison_operator_token15] = ACTIONS(590),
    [aux_sym_comparison_operator_token16] = ACTIONS(590),
    [aux_sym_comparison_operator_token17] = ACTIONS(590),
    [aux_sym_comparison_operator_token18] = ACTIONS(590),
    [aux_sym_comparison_operator_token19] = ACTIONS(590),
    [aux_sym_comparison_operator_token20] = ACTIONS(590),
    [aux_sym_comparison_operator_token21] = ACTIONS(590),
    [aux_sym_comparison_operator_token22] = ACTIONS(590),
    [aux_sym_comparison_operator_token23] = ACTIONS(590),
    [aux_sym_comparison_operator_token24] = ACTIONS(590),
    [aux_sym_comparison_operator_token25] = ACTIONS(590),
    [aux_sym_comparison_operator_token26] = ACTIONS(590),
    [aux_sym_comparison_operator_token27] = ACTIONS(590),
    [aux_sym_comparison_operator_token28] = ACTIONS(592),
    [aux_sym_comparison_operator_token29] = ACTIONS(590),
    [aux_sym_comparison_operator_token30] = ACTIONS(590),
    [aux_sym_comparison_operator_token31] = ACTIONS(590),
    [aux_sym_comparison_operator_token32] = ACTIONS(590),
    [aux_sym_comparison_operator_token33] = ACTIONS(590),
    [aux_sym_comparison_operator_token34] = ACTIONS(592),
    [aux_sym_comparison_operator_token35] = ACTIONS(590),
    [aux_sym_comparison_operator_token36] = ACTIONS(590),
    [aux_sym_comparison_operator_token37] = ACTIONS(590),
    [aux_sym_comparison_operator_token38] = ACTIONS(590),
    [aux_sym_comparison_operator_token39] = ACTIONS(590),
    [aux_sym_comparison_operator_token40] = ACTIONS(590),
    [aux_sym_comparison_operator_token41] = ACTIONS(590),
    [aux_sym_comparison_operator_token42] = ACTIONS(590),
    [aux_sym_comparison_operator_token43] = ACTIONS(590),
    [aux_sym_comparison_operator_token44] = ACTIONS(590),
    [aux_sym_comparison_operator_token45] = ACTIONS(590),
    [aux_sym_comparison_operator_token46] = ACTIONS(590),
    [aux_sym_comparison_operator_token47] = ACTIONS(590),
    [aux_sym_comparison_operator_token48] = ACTIONS(590),
    [aux_sym_comparison_operator_token49] = ACTIONS(590),
    [aux_sym_comparison_operator_token50] = ACTIONS(590),
    [aux_sym_format_operator_token1] = ACTIONS(590),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(590),
    [anon_sym_AMP_AMP] = ACTIONS(590),
    [anon_sym_PIPE_PIPE] = ACTIONS(590),
    [anon_sym_PIPE] = ACTIONS(592),
    [aux_sym_logical_expression_token1] = ACTIONS(590),
    [aux_sym_logical_expression_token2] = ACTIONS(590),
    [aux_sym_logical_expression_token3] = ACTIONS(590),
    [aux_sym_bitwise_expression_token1] = ACTIONS(590),
    [aux_sym_bitwise_expression_token2] = ACTIONS(590),
    [aux_sym_bitwise_expression_token3] = ACTIONS(590),
    [anon_sym_PLUS] = ACTIONS(592),
    [anon_sym_DASH] = ACTIONS(592),
    [anon_sym_SLASH] = ACTIONS(592),
    [anon_sym_BSLASH] = ACTIONS(590),
    [anon_sym_PERCENT] = ACTIONS(592),
    [anon_sym_STAR] = ACTIONS(592),
    [anon_sym_DOT_DOT] = ACTIONS(590),
    [anon_sym_PLUS_PLUS] = ACTIONS(590),
    [anon_sym_DASH_DASH] = ACTIONS(590),
    [anon_sym_DOT2] = ACTIONS(592),
    [anon_sym_COLON_COLON] = ACTIONS(590),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(590),
    [sym__statement_terminator] = ACTIONS(590),
  },
  [STATE(116)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(696),
    [anon_sym_EQ] = ACTIONS(696),
    [anon_sym_BANG_EQ] = ACTIONS(696),
    [anon_sym_PLUS_EQ] = ACTIONS(696),
    [anon_sym_STAR_EQ] = ACTIONS(696),
    [anon_sym_SLASH_EQ] = ACTIONS(696),
    [anon_sym_PERCENT_EQ] = ACTIONS(696),
    [anon_sym_DASH_EQ] = ACTIONS(696),
    [anon_sym_GT] = ACTIONS(698),
    [anon_sym_GT_GT] = ACTIONS(696),
    [anon_sym_2_GT] = ACTIONS(698),
    [anon_sym_2_GT_GT] = ACTIONS(696),
    [anon_sym_3_GT] = ACTIONS(698),
    [anon_sym_3_GT_GT] = ACTIONS(696),
    [anon_sym_4_GT] = ACTIONS(698),
    [anon_sym_4_GT_GT] = ACTIONS(696),
    [anon_sym_5_GT] = ACTIONS(698),
    [anon_sym_5_GT_GT] = ACTIONS(696),
    [anon_sym_6_GT] = ACTIONS(698),
    [anon_sym_6_GT_GT] = ACTIONS(696),
    [anon_sym_STAR_GT] = ACTIONS(698),
    [anon_sym_STAR_GT_GT] = ACTIONS(696),
    [anon_sym_LT] = ACTIONS(698),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(696),
    [anon_sym_2_GT_AMP1] = ACTIONS(696),
    [anon_sym_3_GT_AMP1] = ACTIONS(696),
    [anon_sym_4_GT_AMP1] = ACTIONS(696),
    [anon_sym_5_GT_AMP1] = ACTIONS(696),
    [anon_sym_6_GT_AMP1] = ACTIONS(696),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(696),
    [anon_sym_1_GT_AMP2] = ACTIONS(696),
    [anon_sym_3_GT_AMP2] = ACTIONS(696),
    [anon_sym_4_GT_AMP2] = ACTIONS(696),
    [anon_sym_5_GT_AMP2] = ACTIONS(696),
    [anon_sym_6_GT_AMP2] = ACTIONS(696),
    [aux_sym_comparison_operator_token1] = ACTIONS(696),
    [aux_sym_comparison_operator_token2] = ACTIONS(696),
    [aux_sym_comparison_operator_token3] = ACTIONS(696),
    [aux_sym_comparison_operator_token4] = ACTIONS(696),
    [aux_sym_comparison_operator_token5] = ACTIONS(696),
    [aux_sym_comparison_operator_token6] = ACTIONS(696),
    [aux_sym_comparison_operator_token7] = ACTIONS(696),
    [aux_sym_comparison_operator_token8] = ACTIONS(696),
    [aux_sym_comparison_operator_token9] = ACTIONS(696),
    [aux_sym_comparison_operator_token10] = ACTIONS(696),
    [aux_sym_comparison_operator_token11] = ACTIONS(696),
    [aux_sym_comparison_operator_token12] = ACTIONS(696),
    [aux_sym_comparison_operator_token13] = ACTIONS(696),
    [aux_sym_comparison_operator_token14] = ACTIONS(696),
    [aux_sym_comparison_operator_token15] = ACTIONS(696),
    [aux_sym_comparison_operator_token16] = ACTIONS(696),
    [aux_sym_comparison_operator_token17] = ACTIONS(696),
    [aux_sym_comparison_operator_token18] = ACTIONS(696),
    [aux_sym_comparison_operator_token19] = ACTIONS(696),
    [aux_sym_comparison_operator_token20] = ACTIONS(696),
    [aux_sym_comparison_operator_token21] = ACTIONS(696),
    [aux_sym_comparison_operator_token22] = ACTIONS(696),
    [aux_sym_comparison_operator_token23] = ACTIONS(696),
    [aux_sym_comparison_operator_token24] = ACTIONS(696),
    [aux_sym_comparison_operator_token25] = ACTIONS(696),
    [aux_sym_comparison_operator_token26] = ACTIONS(696),
    [aux_sym_comparison_operator_token27] = ACTIONS(696),
    [aux_sym_comparison_operator_token28] = ACTIONS(698),
    [aux_sym_comparison_operator_token29] = ACTIONS(696),
    [aux_sym_comparison_operator_token30] = ACTIONS(696),
    [aux_sym_comparison_operator_token31] = ACTIONS(696),
    [aux_sym_comparison_operator_token32] = ACTIONS(696),
    [aux_sym_comparison_operator_token33] = ACTIONS(696),
    [aux_sym_comparison_operator_token34] = ACTIONS(698),
    [aux_sym_comparison_operator_token35] = ACTIONS(696),
    [aux_sym_comparison_operator_token36] = ACTIONS(696),
    [aux_sym_comparison_operator_token37] = ACTIONS(696),
    [aux_sym_comparison_operator_token38] = ACTIONS(696),
    [aux_sym_comparison_operator_token39] = ACTIONS(696),
    [aux_sym_comparison_operator_token40] = ACTIONS(696),
    [aux_sym_comparison_operator_token41] = ACTIONS(696),
    [aux_sym_comparison_operator_token42] = ACTIONS(696),
    [aux_sym_comparison_operator_token43] = ACTIONS(696),
    [aux_sym_comparison_operator_token44] = ACTIONS(696),
    [aux_sym_comparison_operator_token45] = ACTIONS(696),
    [aux_sym_comparison_operator_token46] = ACTIONS(696),
    [aux_sym_comparison_operator_token47] = ACTIONS(696),
    [aux_sym_comparison_operator_token48] = ACTIONS(696),
    [aux_sym_comparison_operator_token49] = ACTIONS(696),
    [aux_sym_comparison_operator_token50] = ACTIONS(696),
    [aux_sym_format_operator_token1] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(696),
    [anon_sym_RPAREN] = ACTIONS(696),
    [anon_sym_COMMA] = ACTIONS(696),
    [anon_sym_AMP_AMP] = ACTIONS(696),
    [anon_sym_PIPE_PIPE] = ACTIONS(696),
    [anon_sym_PIPE] = ACTIONS(698),
    [aux_sym_logical_expression_token1] = ACTIONS(696),
    [aux_sym_logical_expression_token2] = ACTIONS(696),
    [aux_sym_logical_expression_token3] = ACTIONS(696),
    [aux_sym_bitwise_expression_token1] = ACTIONS(696),
    [aux_sym_bitwise_expression_token2] = ACTIONS(696),
    [aux_sym_bitwise_expression_token3] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(698),
    [anon_sym_DASH] = ACTIONS(698),
    [anon_sym_SLASH] = ACTIONS(698),
    [anon_sym_BSLASH] = ACTIONS(696),
    [anon_sym_PERCENT] = ACTIONS(698),
    [anon_sym_STAR] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(696),
    [anon_sym_DASH_DASH] = ACTIONS(696),
    [anon_sym_DOT2] = ACTIONS(698),
    [anon_sym_COLON_COLON] = ACTIONS(696),
    [anon_sym_RBRACK] = ACTIONS(696),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(696),
  },
  [STATE(117)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(700),
    [anon_sym_EQ] = ACTIONS(700),
    [anon_sym_BANG_EQ] = ACTIONS(700),
    [anon_sym_PLUS_EQ] = ACTIONS(700),
    [anon_sym_STAR_EQ] = ACTIONS(700),
    [anon_sym_SLASH_EQ] = ACTIONS(700),
    [anon_sym_PERCENT_EQ] = ACTIONS(700),
    [anon_sym_DASH_EQ] = ACTIONS(700),
    [anon_sym_GT] = ACTIONS(702),
    [anon_sym_GT_GT] = ACTIONS(700),
    [anon_sym_2_GT] = ACTIONS(702),
    [anon_sym_2_GT_GT] = ACTIONS(700),
    [anon_sym_3_GT] = ACTIONS(702),
    [anon_sym_3_GT_GT] = ACTIONS(700),
    [anon_sym_4_GT] = ACTIONS(702),
    [anon_sym_4_GT_GT] = ACTIONS(700),
    [anon_sym_5_GT] = ACTIONS(702),
    [anon_sym_5_GT_GT] = ACTIONS(700),
    [anon_sym_6_GT] = ACTIONS(702),
    [anon_sym_6_GT_GT] = ACTIONS(700),
    [anon_sym_STAR_GT] = ACTIONS(702),
    [anon_sym_STAR_GT_GT] = ACTIONS(700),
    [anon_sym_LT] = ACTIONS(702),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(700),
    [anon_sym_2_GT_AMP1] = ACTIONS(700),
    [anon_sym_3_GT_AMP1] = ACTIONS(700),
    [anon_sym_4_GT_AMP1] = ACTIONS(700),
    [anon_sym_5_GT_AMP1] = ACTIONS(700),
    [anon_sym_6_GT_AMP1] = ACTIONS(700),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(700),
    [anon_sym_1_GT_AMP2] = ACTIONS(700),
    [anon_sym_3_GT_AMP2] = ACTIONS(700),
    [anon_sym_4_GT_AMP2] = ACTIONS(700),
    [anon_sym_5_GT_AMP2] = ACTIONS(700),
    [anon_sym_6_GT_AMP2] = ACTIONS(700),
    [aux_sym_comparison_operator_token1] = ACTIONS(700),
    [aux_sym_comparison_operator_token2] = ACTIONS(700),
    [aux_sym_comparison_operator_token3] = ACTIONS(700),
    [aux_sym_comparison_operator_token4] = ACTIONS(700),
    [aux_sym_comparison_operator_token5] = ACTIONS(700),
    [aux_sym_comparison_operator_token6] = ACTIONS(700),
    [aux_sym_comparison_operator_token7] = ACTIONS(700),
    [aux_sym_comparison_operator_token8] = ACTIONS(700),
    [aux_sym_comparison_operator_token9] = ACTIONS(700),
    [aux_sym_comparison_operator_token10] = ACTIONS(700),
    [aux_sym_comparison_operator_token11] = ACTIONS(700),
    [aux_sym_comparison_operator_token12] = ACTIONS(700),
    [aux_sym_comparison_operator_token13] = ACTIONS(700),
    [aux_sym_comparison_operator_token14] = ACTIONS(700),
    [aux_sym_comparison_operator_token15] = ACTIONS(700),
    [aux_sym_comparison_operator_token16] = ACTIONS(700),
    [aux_sym_comparison_operator_token17] = ACTIONS(700),
    [aux_sym_comparison_operator_token18] = ACTIONS(700),
    [aux_sym_comparison_operator_token19] = ACTIONS(700),
    [aux_sym_comparison_operator_token20] = ACTIONS(700),
    [aux_sym_comparison_operator_token21] = ACTIONS(700),
    [aux_sym_comparison_operator_token22] = ACTIONS(700),
    [aux_sym_comparison_operator_token23] = ACTIONS(700),
    [aux_sym_comparison_operator_token24] = ACTIONS(700),
    [aux_sym_comparison_operator_token25] = ACTIONS(700),
    [aux_sym_comparison_operator_token26] = ACTIONS(700),
    [aux_sym_comparison_operator_token27] = ACTIONS(700),
    [aux_sym_comparison_operator_token28] = ACTIONS(702),
    [aux_sym_comparison_operator_token29] = ACTIONS(700),
    [aux_sym_comparison_operator_token30] = ACTIONS(700),
    [aux_sym_comparison_operator_token31] = ACTIONS(700),
    [aux_sym_comparison_operator_token32] = ACTIONS(700),
    [aux_sym_comparison_operator_token33] = ACTIONS(700),
    [aux_sym_comparison_operator_token34] = ACTIONS(702),
    [aux_sym_comparison_operator_token35] = ACTIONS(700),
    [aux_sym_comparison_operator_token36] = ACTIONS(700),
    [aux_sym_comparison_operator_token37] = ACTIONS(700),
    [aux_sym_comparison_operator_token38] = ACTIONS(700),
    [aux_sym_comparison_operator_token39] = ACTIONS(700),
    [aux_sym_comparison_operator_token40] = ACTIONS(700),
    [aux_sym_comparison_operator_token41] = ACTIONS(700),
    [aux_sym_comparison_operator_token42] = ACTIONS(700),
    [aux_sym_comparison_operator_token43] = ACTIONS(700),
    [aux_sym_comparison_operator_token44] = ACTIONS(700),
    [aux_sym_comparison_operator_token45] = ACTIONS(700),
    [aux_sym_comparison_operator_token46] = ACTIONS(700),
    [aux_sym_comparison_operator_token47] = ACTIONS(700),
    [aux_sym_comparison_operator_token48] = ACTIONS(700),
    [aux_sym_comparison_operator_token49] = ACTIONS(700),
    [aux_sym_comparison_operator_token50] = ACTIONS(700),
    [aux_sym_format_operator_token1] = ACTIONS(700),
    [anon_sym_LPAREN] = ACTIONS(700),
    [anon_sym_RPAREN] = ACTIONS(700),
    [anon_sym_COMMA] = ACTIONS(700),
    [anon_sym_AMP_AMP] = ACTIONS(700),
    [anon_sym_PIPE_PIPE] = ACTIONS(700),
    [anon_sym_PIPE] = ACTIONS(702),
    [aux_sym_logical_expression_token1] = ACTIONS(700),
    [aux_sym_logical_expression_token2] = ACTIONS(700),
    [aux_sym_logical_expression_token3] = ACTIONS(700),
    [aux_sym_bitwise_expression_token1] = ACTIONS(700),
    [aux_sym_bitwise_expression_token2] = ACTIONS(700),
    [aux_sym_bitwise_expression_token3] = ACTIONS(700),
    [anon_sym_PLUS] = ACTIONS(702),
    [anon_sym_DASH] = ACTIONS(702),
    [anon_sym_SLASH] = ACTIONS(702),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_PERCENT] = ACTIONS(702),
    [anon_sym_STAR] = ACTIONS(702),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_PLUS_PLUS] = ACTIONS(700),
    [anon_sym_DASH_DASH] = ACTIONS(700),
    [anon_sym_DOT2] = ACTIONS(702),
    [anon_sym_COLON_COLON] = ACTIONS(700),
    [anon_sym_RBRACK] = ACTIONS(700),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(700),
  },
  [STATE(118)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(704),
    [anon_sym_EQ] = ACTIONS(704),
    [anon_sym_BANG_EQ] = ACTIONS(704),
    [anon_sym_PLUS_EQ] = ACTIONS(704),
    [anon_sym_STAR_EQ] = ACTIONS(704),
    [anon_sym_SLASH_EQ] = ACTIONS(704),
    [anon_sym_PERCENT_EQ] = ACTIONS(704),
    [anon_sym_DASH_EQ] = ACTIONS(704),
    [anon_sym_GT] = ACTIONS(706),
    [anon_sym_GT_GT] = ACTIONS(704),
    [anon_sym_2_GT] = ACTIONS(706),
    [anon_sym_2_GT_GT] = ACTIONS(704),
    [anon_sym_3_GT] = ACTIONS(706),
    [anon_sym_3_GT_GT] = ACTIONS(704),
    [anon_sym_4_GT] = ACTIONS(706),
    [anon_sym_4_GT_GT] = ACTIONS(704),
    [anon_sym_5_GT] = ACTIONS(706),
    [anon_sym_5_GT_GT] = ACTIONS(704),
    [anon_sym_6_GT] = ACTIONS(706),
    [anon_sym_6_GT_GT] = ACTIONS(704),
    [anon_sym_STAR_GT] = ACTIONS(706),
    [anon_sym_STAR_GT_GT] = ACTIONS(704),
    [anon_sym_LT] = ACTIONS(706),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(704),
    [anon_sym_2_GT_AMP1] = ACTIONS(704),
    [anon_sym_3_GT_AMP1] = ACTIONS(704),
    [anon_sym_4_GT_AMP1] = ACTIONS(704),
    [anon_sym_5_GT_AMP1] = ACTIONS(704),
    [anon_sym_6_GT_AMP1] = ACTIONS(704),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(704),
    [anon_sym_1_GT_AMP2] = ACTIONS(704),
    [anon_sym_3_GT_AMP2] = ACTIONS(704),
    [anon_sym_4_GT_AMP2] = ACTIONS(704),
    [anon_sym_5_GT_AMP2] = ACTIONS(704),
    [anon_sym_6_GT_AMP2] = ACTIONS(704),
    [aux_sym_comparison_operator_token1] = ACTIONS(704),
    [aux_sym_comparison_operator_token2] = ACTIONS(704),
    [aux_sym_comparison_operator_token3] = ACTIONS(704),
    [aux_sym_comparison_operator_token4] = ACTIONS(704),
    [aux_sym_comparison_operator_token5] = ACTIONS(704),
    [aux_sym_comparison_operator_token6] = ACTIONS(704),
    [aux_sym_comparison_operator_token7] = ACTIONS(704),
    [aux_sym_comparison_operator_token8] = ACTIONS(704),
    [aux_sym_comparison_operator_token9] = ACTIONS(704),
    [aux_sym_comparison_operator_token10] = ACTIONS(704),
    [aux_sym_comparison_operator_token11] = ACTIONS(704),
    [aux_sym_comparison_operator_token12] = ACTIONS(704),
    [aux_sym_comparison_operator_token13] = ACTIONS(704),
    [aux_sym_comparison_operator_token14] = ACTIONS(704),
    [aux_sym_comparison_operator_token15] = ACTIONS(704),
    [aux_sym_comparison_operator_token16] = ACTIONS(704),
    [aux_sym_comparison_operator_token17] = ACTIONS(704),
    [aux_sym_comparison_operator_token18] = ACTIONS(704),
    [aux_sym_comparison_operator_token19] = ACTIONS(704),
    [aux_sym_comparison_operator_token20] = ACTIONS(704),
    [aux_sym_comparison_operator_token21] = ACTIONS(704),
    [aux_sym_comparison_operator_token22] = ACTIONS(704),
    [aux_sym_comparison_operator_token23] = ACTIONS(704),
    [aux_sym_comparison_operator_token24] = ACTIONS(704),
    [aux_sym_comparison_operator_token25] = ACTIONS(704),
    [aux_sym_comparison_operator_token26] = ACTIONS(704),
    [aux_sym_comparison_operator_token27] = ACTIONS(704),
    [aux_sym_comparison_operator_token28] = ACTIONS(706),
    [aux_sym_comparison_operator_token29] = ACTIONS(704),
    [aux_sym_comparison_operator_token30] = ACTIONS(704),
    [aux_sym_comparison_operator_token31] = ACTIONS(704),
    [aux_sym_comparison_operator_token32] = ACTIONS(704),
    [aux_sym_comparison_operator_token33] = ACTIONS(704),
    [aux_sym_comparison_operator_token34] = ACTIONS(706),
    [aux_sym_comparison_operator_token35] = ACTIONS(704),
    [aux_sym_comparison_operator_token36] = ACTIONS(704),
    [aux_sym_comparison_operator_token37] = ACTIONS(704),
    [aux_sym_comparison_operator_token38] = ACTIONS(704),
    [aux_sym_comparison_operator_token39] = ACTIONS(704),
    [aux_sym_comparison_operator_token40] = ACTIONS(704),
    [aux_sym_comparison_operator_token41] = ACTIONS(704),
    [aux_sym_comparison_operator_token42] = ACTIONS(704),
    [aux_sym_comparison_operator_token43] = ACTIONS(704),
    [aux_sym_comparison_operator_token44] = ACTIONS(704),
    [aux_sym_comparison_operator_token45] = ACTIONS(704),
    [aux_sym_comparison_operator_token46] = ACTIONS(704),
    [aux_sym_comparison_operator_token47] = ACTIONS(704),
    [aux_sym_comparison_operator_token48] = ACTIONS(704),
    [aux_sym_comparison_operator_token49] = ACTIONS(704),
    [aux_sym_comparison_operator_token50] = ACTIONS(704),
    [aux_sym_format_operator_token1] = ACTIONS(704),
    [anon_sym_LPAREN] = ACTIONS(704),
    [anon_sym_RPAREN] = ACTIONS(704),
    [anon_sym_COMMA] = ACTIONS(704),
    [anon_sym_AMP_AMP] = ACTIONS(704),
    [anon_sym_PIPE_PIPE] = ACTIONS(704),
    [anon_sym_PIPE] = ACTIONS(706),
    [aux_sym_logical_expression_token1] = ACTIONS(704),
    [aux_sym_logical_expression_token2] = ACTIONS(704),
    [aux_sym_logical_expression_token3] = ACTIONS(704),
    [aux_sym_bitwise_expression_token1] = ACTIONS(704),
    [aux_sym_bitwise_expression_token2] = ACTIONS(704),
    [aux_sym_bitwise_expression_token3] = ACTIONS(704),
    [anon_sym_PLUS] = ACTIONS(706),
    [anon_sym_DASH] = ACTIONS(706),
    [anon_sym_SLASH] = ACTIONS(706),
    [anon_sym_BSLASH] = ACTIONS(704),
    [anon_sym_PERCENT] = ACTIONS(706),
    [anon_sym_STAR] = ACTIONS(706),
    [anon_sym_DOT_DOT] = ACTIONS(704),
    [anon_sym_PLUS_PLUS] = ACTIONS(704),
    [anon_sym_DASH_DASH] = ACTIONS(704),
    [anon_sym_DOT2] = ACTIONS(706),
    [anon_sym_COLON_COLON] = ACTIONS(704),
    [anon_sym_RBRACK] = ACTIONS(704),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(704),
  },
  [STATE(119)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(708),
    [anon_sym_EQ] = ACTIONS(708),
    [anon_sym_BANG_EQ] = ACTIONS(708),
    [anon_sym_PLUS_EQ] = ACTIONS(708),
    [anon_sym_STAR_EQ] = ACTIONS(708),
    [anon_sym_SLASH_EQ] = ACTIONS(708),
    [anon_sym_PERCENT_EQ] = ACTIONS(708),
    [anon_sym_DASH_EQ] = ACTIONS(708),
    [anon_sym_GT] = ACTIONS(710),
    [anon_sym_GT_GT] = ACTIONS(708),
    [anon_sym_2_GT] = ACTIONS(710),
    [anon_sym_2_GT_GT] = ACTIONS(708),
    [anon_sym_3_GT] = ACTIONS(710),
    [anon_sym_3_GT_GT] = ACTIONS(708),
    [anon_sym_4_GT] = ACTIONS(710),
    [anon_sym_4_GT_GT] = ACTIONS(708),
    [anon_sym_5_GT] = ACTIONS(710),
    [anon_sym_5_GT_GT] = ACTIONS(708),
    [anon_sym_6_GT] = ACTIONS(710),
    [anon_sym_6_GT_GT] = ACTIONS(708),
    [anon_sym_STAR_GT] = ACTIONS(710),
    [anon_sym_STAR_GT_GT] = ACTIONS(708),
    [anon_sym_LT] = ACTIONS(710),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(708),
    [anon_sym_2_GT_AMP1] = ACTIONS(708),
    [anon_sym_3_GT_AMP1] = ACTIONS(708),
    [anon_sym_4_GT_AMP1] = ACTIONS(708),
    [anon_sym_5_GT_AMP1] = ACTIONS(708),
    [anon_sym_6_GT_AMP1] = ACTIONS(708),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(708),
    [anon_sym_1_GT_AMP2] = ACTIONS(708),
    [anon_sym_3_GT_AMP2] = ACTIONS(708),
    [anon_sym_4_GT_AMP2] = ACTIONS(708),
    [anon_sym_5_GT_AMP2] = ACTIONS(708),
    [anon_sym_6_GT_AMP2] = ACTIONS(708),
    [aux_sym_comparison_operator_token1] = ACTIONS(708),
    [aux_sym_comparison_operator_token2] = ACTIONS(708),
    [aux_sym_comparison_operator_token3] = ACTIONS(708),
    [aux_sym_comparison_operator_token4] = ACTIONS(708),
    [aux_sym_comparison_operator_token5] = ACTIONS(708),
    [aux_sym_comparison_operator_token6] = ACTIONS(708),
    [aux_sym_comparison_operator_token7] = ACTIONS(708),
    [aux_sym_comparison_operator_token8] = ACTIONS(708),
    [aux_sym_comparison_operator_token9] = ACTIONS(708),
    [aux_sym_comparison_operator_token10] = ACTIONS(708),
    [aux_sym_comparison_operator_token11] = ACTIONS(708),
    [aux_sym_comparison_operator_token12] = ACTIONS(708),
    [aux_sym_comparison_operator_token13] = ACTIONS(708),
    [aux_sym_comparison_operator_token14] = ACTIONS(708),
    [aux_sym_comparison_operator_token15] = ACTIONS(708),
    [aux_sym_comparison_operator_token16] = ACTIONS(708),
    [aux_sym_comparison_operator_token17] = ACTIONS(708),
    [aux_sym_comparison_operator_token18] = ACTIONS(708),
    [aux_sym_comparison_operator_token19] = ACTIONS(708),
    [aux_sym_comparison_operator_token20] = ACTIONS(708),
    [aux_sym_comparison_operator_token21] = ACTIONS(708),
    [aux_sym_comparison_operator_token22] = ACTIONS(708),
    [aux_sym_comparison_operator_token23] = ACTIONS(708),
    [aux_sym_comparison_operator_token24] = ACTIONS(708),
    [aux_sym_comparison_operator_token25] = ACTIONS(708),
    [aux_sym_comparison_operator_token26] = ACTIONS(708),
    [aux_sym_comparison_operator_token27] = ACTIONS(708),
    [aux_sym_comparison_operator_token28] = ACTIONS(710),
    [aux_sym_comparison_operator_token29] = ACTIONS(708),
    [aux_sym_comparison_operator_token30] = ACTIONS(708),
    [aux_sym_comparison_operator_token31] = ACTIONS(708),
    [aux_sym_comparison_operator_token32] = ACTIONS(708),
    [aux_sym_comparison_operator_token33] = ACTIONS(708),
    [aux_sym_comparison_operator_token34] = ACTIONS(710),
    [aux_sym_comparison_operator_token35] = ACTIONS(708),
    [aux_sym_comparison_operator_token36] = ACTIONS(708),
    [aux_sym_comparison_operator_token37] = ACTIONS(708),
    [aux_sym_comparison_operator_token38] = ACTIONS(708),
    [aux_sym_comparison_operator_token39] = ACTIONS(708),
    [aux_sym_comparison_operator_token40] = ACTIONS(708),
    [aux_sym_comparison_operator_token41] = ACTIONS(708),
    [aux_sym_comparison_operator_token42] = ACTIONS(708),
    [aux_sym_comparison_operator_token43] = ACTIONS(708),
    [aux_sym_comparison_operator_token44] = ACTIONS(708),
    [aux_sym_comparison_operator_token45] = ACTIONS(708),
    [aux_sym_comparison_operator_token46] = ACTIONS(708),
    [aux_sym_comparison_operator_token47] = ACTIONS(708),
    [aux_sym_comparison_operator_token48] = ACTIONS(708),
    [aux_sym_comparison_operator_token49] = ACTIONS(708),
    [aux_sym_comparison_operator_token50] = ACTIONS(708),
    [aux_sym_format_operator_token1] = ACTIONS(708),
    [anon_sym_LPAREN] = ACTIONS(708),
    [anon_sym_RPAREN] = ACTIONS(708),
    [anon_sym_COMMA] = ACTIONS(708),
    [anon_sym_AMP_AMP] = ACTIONS(708),
    [anon_sym_PIPE_PIPE] = ACTIONS(708),
    [anon_sym_PIPE] = ACTIONS(710),
    [aux_sym_logical_expression_token1] = ACTIONS(708),
    [aux_sym_logical_expression_token2] = ACTIONS(708),
    [aux_sym_logical_expression_token3] = ACTIONS(708),
    [aux_sym_bitwise_expression_token1] = ACTIONS(708),
    [aux_sym_bitwise_expression_token2] = ACTIONS(708),
    [aux_sym_bitwise_expression_token3] = ACTIONS(708),
    [anon_sym_PLUS] = ACTIONS(710),
    [anon_sym_DASH] = ACTIONS(710),
    [anon_sym_SLASH] = ACTIONS(710),
    [anon_sym_BSLASH] = ACTIONS(708),
    [anon_sym_PERCENT] = ACTIONS(710),
    [anon_sym_STAR] = ACTIONS(710),
    [anon_sym_DOT_DOT] = ACTIONS(708),
    [anon_sym_PLUS_PLUS] = ACTIONS(708),
    [anon_sym_DASH_DASH] = ACTIONS(708),
    [anon_sym_DOT2] = ACTIONS(710),
    [anon_sym_COLON_COLON] = ACTIONS(708),
    [anon_sym_RBRACK] = ACTIONS(708),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(708),
  },
  [STATE(120)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(712),
    [anon_sym_EQ] = ACTIONS(712),
    [anon_sym_BANG_EQ] = ACTIONS(712),
    [anon_sym_PLUS_EQ] = ACTIONS(712),
    [anon_sym_STAR_EQ] = ACTIONS(712),
    [anon_sym_SLASH_EQ] = ACTIONS(712),
    [anon_sym_PERCENT_EQ] = ACTIONS(712),
    [anon_sym_DASH_EQ] = ACTIONS(712),
    [anon_sym_GT] = ACTIONS(714),
    [anon_sym_GT_GT] = ACTIONS(712),
    [anon_sym_2_GT] = ACTIONS(714),
    [anon_sym_2_GT_GT] = ACTIONS(712),
    [anon_sym_3_GT] = ACTIONS(714),
    [anon_sym_3_GT_GT] = ACTIONS(712),
    [anon_sym_4_GT] = ACTIONS(714),
    [anon_sym_4_GT_GT] = ACTIONS(712),
    [anon_sym_5_GT] = ACTIONS(714),
    [anon_sym_5_GT_GT] = ACTIONS(712),
    [anon_sym_6_GT] = ACTIONS(714),
    [anon_sym_6_GT_GT] = ACTIONS(712),
    [anon_sym_STAR_GT] = ACTIONS(714),
    [anon_sym_STAR_GT_GT] = ACTIONS(712),
    [anon_sym_LT] = ACTIONS(714),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(712),
    [anon_sym_2_GT_AMP1] = ACTIONS(712),
    [anon_sym_3_GT_AMP1] = ACTIONS(712),
    [anon_sym_4_GT_AMP1] = ACTIONS(712),
    [anon_sym_5_GT_AMP1] = ACTIONS(712),
    [anon_sym_6_GT_AMP1] = ACTIONS(712),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(712),
    [anon_sym_1_GT_AMP2] = ACTIONS(712),
    [anon_sym_3_GT_AMP2] = ACTIONS(712),
    [anon_sym_4_GT_AMP2] = ACTIONS(712),
    [anon_sym_5_GT_AMP2] = ACTIONS(712),
    [anon_sym_6_GT_AMP2] = ACTIONS(712),
    [aux_sym_comparison_operator_token1] = ACTIONS(712),
    [aux_sym_comparison_operator_token2] = ACTIONS(712),
    [aux_sym_comparison_operator_token3] = ACTIONS(712),
    [aux_sym_comparison_operator_token4] = ACTIONS(712),
    [aux_sym_comparison_operator_token5] = ACTIONS(712),
    [aux_sym_comparison_operator_token6] = ACTIONS(712),
    [aux_sym_comparison_operator_token7] = ACTIONS(712),
    [aux_sym_comparison_operator_token8] = ACTIONS(712),
    [aux_sym_comparison_operator_token9] = ACTIONS(712),
    [aux_sym_comparison_operator_token10] = ACTIONS(712),
    [aux_sym_comparison_operator_token11] = ACTIONS(712),
    [aux_sym_comparison_operator_token12] = ACTIONS(712),
    [aux_sym_comparison_operator_token13] = ACTIONS(712),
    [aux_sym_comparison_operator_token14] = ACTIONS(712),
    [aux_sym_comparison_operator_token15] = ACTIONS(712),
    [aux_sym_comparison_operator_token16] = ACTIONS(712),
    [aux_sym_comparison_operator_token17] = ACTIONS(712),
    [aux_sym_comparison_operator_token18] = ACTIONS(712),
    [aux_sym_comparison_operator_token19] = ACTIONS(712),
    [aux_sym_comparison_operator_token20] = ACTIONS(712),
    [aux_sym_comparison_operator_token21] = ACTIONS(712),
    [aux_sym_comparison_operator_token22] = ACTIONS(712),
    [aux_sym_comparison_operator_token23] = ACTIONS(712),
    [aux_sym_comparison_operator_token24] = ACTIONS(712),
    [aux_sym_comparison_operator_token25] = ACTIONS(712),
    [aux_sym_comparison_operator_token26] = ACTIONS(712),
    [aux_sym_comparison_operator_token27] = ACTIONS(712),
    [aux_sym_comparison_operator_token28] = ACTIONS(714),
    [aux_sym_comparison_operator_token29] = ACTIONS(712),
    [aux_sym_comparison_operator_token30] = ACTIONS(712),
    [aux_sym_comparison_operator_token31] = ACTIONS(712),
    [aux_sym_comparison_operator_token32] = ACTIONS(712),
    [aux_sym_comparison_operator_token33] = ACTIONS(712),
    [aux_sym_comparison_operator_token34] = ACTIONS(714),
    [aux_sym_comparison_operator_token35] = ACTIONS(712),
    [aux_sym_comparison_operator_token36] = ACTIONS(712),
    [aux_sym_comparison_operator_token37] = ACTIONS(712),
    [aux_sym_comparison_operator_token38] = ACTIONS(712),
    [aux_sym_comparison_operator_token39] = ACTIONS(712),
    [aux_sym_comparison_operator_token40] = ACTIONS(712),
    [aux_sym_comparison_operator_token41] = ACTIONS(712),
    [aux_sym_comparison_operator_token42] = ACTIONS(712),
    [aux_sym_comparison_operator_token43] = ACTIONS(712),
    [aux_sym_comparison_operator_token44] = ACTIONS(712),
    [aux_sym_comparison_operator_token45] = ACTIONS(712),
    [aux_sym_comparison_operator_token46] = ACTIONS(712),
    [aux_sym_comparison_operator_token47] = ACTIONS(712),
    [aux_sym_comparison_operator_token48] = ACTIONS(712),
    [aux_sym_comparison_operator_token49] = ACTIONS(712),
    [aux_sym_comparison_operator_token50] = ACTIONS(712),
    [aux_sym_format_operator_token1] = ACTIONS(712),
    [anon_sym_LPAREN] = ACTIONS(712),
    [anon_sym_RPAREN] = ACTIONS(712),
    [anon_sym_COMMA] = ACTIONS(712),
    [anon_sym_AMP_AMP] = ACTIONS(712),
    [anon_sym_PIPE_PIPE] = ACTIONS(712),
    [anon_sym_PIPE] = ACTIONS(714),
    [aux_sym_logical_expression_token1] = ACTIONS(712),
    [aux_sym_logical_expression_token2] = ACTIONS(712),
    [aux_sym_logical_expression_token3] = ACTIONS(712),
    [aux_sym_bitwise_expression_token1] = ACTIONS(712),
    [aux_sym_bitwise_expression_token2] = ACTIONS(712),
    [aux_sym_bitwise_expression_token3] = ACTIONS(712),
    [anon_sym_PLUS] = ACTIONS(714),
    [anon_sym_DASH] = ACTIONS(714),
    [anon_sym_SLASH] = ACTIONS(714),
    [anon_sym_BSLASH] = ACTIONS(712),
    [anon_sym_PERCENT] = ACTIONS(714),
    [anon_sym_STAR] = ACTIONS(714),
    [anon_sym_DOT_DOT] = ACTIONS(712),
    [anon_sym_PLUS_PLUS] = ACTIONS(712),
    [anon_sym_DASH_DASH] = ACTIONS(712),
    [anon_sym_DOT2] = ACTIONS(714),
    [anon_sym_COLON_COLON] = ACTIONS(712),
    [anon_sym_RBRACK] = ACTIONS(712),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(712),
  },
  [STATE(121)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(716),
    [anon_sym_EQ] = ACTIONS(716),
    [anon_sym_BANG_EQ] = ACTIONS(716),
    [anon_sym_PLUS_EQ] = ACTIONS(716),
    [anon_sym_STAR_EQ] = ACTIONS(716),
    [anon_sym_SLASH_EQ] = ACTIONS(716),
    [anon_sym_PERCENT_EQ] = ACTIONS(716),
    [anon_sym_DASH_EQ] = ACTIONS(716),
    [anon_sym_GT] = ACTIONS(718),
    [anon_sym_GT_GT] = ACTIONS(716),
    [anon_sym_2_GT] = ACTIONS(718),
    [anon_sym_2_GT_GT] = ACTIONS(716),
    [anon_sym_3_GT] = ACTIONS(718),
    [anon_sym_3_GT_GT] = ACTIONS(716),
    [anon_sym_4_GT] = ACTIONS(718),
    [anon_sym_4_GT_GT] = ACTIONS(716),
    [anon_sym_5_GT] = ACTIONS(718),
    [anon_sym_5_GT_GT] = ACTIONS(716),
    [anon_sym_6_GT] = ACTIONS(718),
    [anon_sym_6_GT_GT] = ACTIONS(716),
    [anon_sym_STAR_GT] = ACTIONS(718),
    [anon_sym_STAR_GT_GT] = ACTIONS(716),
    [anon_sym_LT] = ACTIONS(718),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(716),
    [anon_sym_2_GT_AMP1] = ACTIONS(716),
    [anon_sym_3_GT_AMP1] = ACTIONS(716),
    [anon_sym_4_GT_AMP1] = ACTIONS(716),
    [anon_sym_5_GT_AMP1] = ACTIONS(716),
    [anon_sym_6_GT_AMP1] = ACTIONS(716),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(716),
    [anon_sym_1_GT_AMP2] = ACTIONS(716),
    [anon_sym_3_GT_AMP2] = ACTIONS(716),
    [anon_sym_4_GT_AMP2] = ACTIONS(716),
    [anon_sym_5_GT_AMP2] = ACTIONS(716),
    [anon_sym_6_GT_AMP2] = ACTIONS(716),
    [aux_sym_comparison_operator_token1] = ACTIONS(716),
    [aux_sym_comparison_operator_token2] = ACTIONS(716),
    [aux_sym_comparison_operator_token3] = ACTIONS(716),
    [aux_sym_comparison_operator_token4] = ACTIONS(716),
    [aux_sym_comparison_operator_token5] = ACTIONS(716),
    [aux_sym_comparison_operator_token6] = ACTIONS(716),
    [aux_sym_comparison_operator_token7] = ACTIONS(716),
    [aux_sym_comparison_operator_token8] = ACTIONS(716),
    [aux_sym_comparison_operator_token9] = ACTIONS(716),
    [aux_sym_comparison_operator_token10] = ACTIONS(716),
    [aux_sym_comparison_operator_token11] = ACTIONS(716),
    [aux_sym_comparison_operator_token12] = ACTIONS(716),
    [aux_sym_comparison_operator_token13] = ACTIONS(716),
    [aux_sym_comparison_operator_token14] = ACTIONS(716),
    [aux_sym_comparison_operator_token15] = ACTIONS(716),
    [aux_sym_comparison_operator_token16] = ACTIONS(716),
    [aux_sym_comparison_operator_token17] = ACTIONS(716),
    [aux_sym_comparison_operator_token18] = ACTIONS(716),
    [aux_sym_comparison_operator_token19] = ACTIONS(716),
    [aux_sym_comparison_operator_token20] = ACTIONS(716),
    [aux_sym_comparison_operator_token21] = ACTIONS(716),
    [aux_sym_comparison_operator_token22] = ACTIONS(716),
    [aux_sym_comparison_operator_token23] = ACTIONS(716),
    [aux_sym_comparison_operator_token24] = ACTIONS(716),
    [aux_sym_comparison_operator_token25] = ACTIONS(716),
    [aux_sym_comparison_operator_token26] = ACTIONS(716),
    [aux_sym_comparison_operator_token27] = ACTIONS(716),
    [aux_sym_comparison_operator_token28] = ACTIONS(718),
    [aux_sym_comparison_operator_token29] = ACTIONS(716),
    [aux_sym_comparison_operator_token30] = ACTIONS(716),
    [aux_sym_comparison_operator_token31] = ACTIONS(716),
    [aux_sym_comparison_operator_token32] = ACTIONS(716),
    [aux_sym_comparison_operator_token33] = ACTIONS(716),
    [aux_sym_comparison_operator_token34] = ACTIONS(718),
    [aux_sym_comparison_operator_token35] = ACTIONS(716),
    [aux_sym_comparison_operator_token36] = ACTIONS(716),
    [aux_sym_comparison_operator_token37] = ACTIONS(716),
    [aux_sym_comparison_operator_token38] = ACTIONS(716),
    [aux_sym_comparison_operator_token39] = ACTIONS(716),
    [aux_sym_comparison_operator_token40] = ACTIONS(716),
    [aux_sym_comparison_operator_token41] = ACTIONS(716),
    [aux_sym_comparison_operator_token42] = ACTIONS(716),
    [aux_sym_comparison_operator_token43] = ACTIONS(716),
    [aux_sym_comparison_operator_token44] = ACTIONS(716),
    [aux_sym_comparison_operator_token45] = ACTIONS(716),
    [aux_sym_comparison_operator_token46] = ACTIONS(716),
    [aux_sym_comparison_operator_token47] = ACTIONS(716),
    [aux_sym_comparison_operator_token48] = ACTIONS(716),
    [aux_sym_comparison_operator_token49] = ACTIONS(716),
    [aux_sym_comparison_operator_token50] = ACTIONS(716),
    [aux_sym_format_operator_token1] = ACTIONS(716),
    [anon_sym_LPAREN] = ACTIONS(716),
    [anon_sym_RPAREN] = ACTIONS(716),
    [anon_sym_COMMA] = ACTIONS(716),
    [anon_sym_AMP_AMP] = ACTIONS(716),
    [anon_sym_PIPE_PIPE] = ACTIONS(716),
    [anon_sym_PIPE] = ACTIONS(718),
    [aux_sym_logical_expression_token1] = ACTIONS(716),
    [aux_sym_logical_expression_token2] = ACTIONS(716),
    [aux_sym_logical_expression_token3] = ACTIONS(716),
    [aux_sym_bitwise_expression_token1] = ACTIONS(716),
    [aux_sym_bitwise_expression_token2] = ACTIONS(716),
    [aux_sym_bitwise_expression_token3] = ACTIONS(716),
    [anon_sym_PLUS] = ACTIONS(718),
    [anon_sym_DASH] = ACTIONS(718),
    [anon_sym_SLASH] = ACTIONS(718),
    [anon_sym_BSLASH] = ACTIONS(716),
    [anon_sym_PERCENT] = ACTIONS(718),
    [anon_sym_STAR] = ACTIONS(718),
    [anon_sym_DOT_DOT] = ACTIONS(716),
    [anon_sym_PLUS_PLUS] = ACTIONS(716),
    [anon_sym_DASH_DASH] = ACTIONS(716),
    [anon_sym_DOT2] = ACTIONS(718),
    [anon_sym_COLON_COLON] = ACTIONS(716),
    [anon_sym_RBRACK] = ACTIONS(716),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(716),
  },
  [STATE(122)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(720),
    [anon_sym_EQ] = ACTIONS(720),
    [anon_sym_BANG_EQ] = ACTIONS(720),
    [anon_sym_PLUS_EQ] = ACTIONS(720),
    [anon_sym_STAR_EQ] = ACTIONS(720),
    [anon_sym_SLASH_EQ] = ACTIONS(720),
    [anon_sym_PERCENT_EQ] = ACTIONS(720),
    [anon_sym_DASH_EQ] = ACTIONS(720),
    [anon_sym_GT] = ACTIONS(722),
    [anon_sym_GT_GT] = ACTIONS(720),
    [anon_sym_2_GT] = ACTIONS(722),
    [anon_sym_2_GT_GT] = ACTIONS(720),
    [anon_sym_3_GT] = ACTIONS(722),
    [anon_sym_3_GT_GT] = ACTIONS(720),
    [anon_sym_4_GT] = ACTIONS(722),
    [anon_sym_4_GT_GT] = ACTIONS(720),
    [anon_sym_5_GT] = ACTIONS(722),
    [anon_sym_5_GT_GT] = ACTIONS(720),
    [anon_sym_6_GT] = ACTIONS(722),
    [anon_sym_6_GT_GT] = ACTIONS(720),
    [anon_sym_STAR_GT] = ACTIONS(722),
    [anon_sym_STAR_GT_GT] = ACTIONS(720),
    [anon_sym_LT] = ACTIONS(722),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(720),
    [anon_sym_2_GT_AMP1] = ACTIONS(720),
    [anon_sym_3_GT_AMP1] = ACTIONS(720),
    [anon_sym_4_GT_AMP1] = ACTIONS(720),
    [anon_sym_5_GT_AMP1] = ACTIONS(720),
    [anon_sym_6_GT_AMP1] = ACTIONS(720),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(720),
    [anon_sym_1_GT_AMP2] = ACTIONS(720),
    [anon_sym_3_GT_AMP2] = ACTIONS(720),
    [anon_sym_4_GT_AMP2] = ACTIONS(720),
    [anon_sym_5_GT_AMP2] = ACTIONS(720),
    [anon_sym_6_GT_AMP2] = ACTIONS(720),
    [aux_sym_comparison_operator_token1] = ACTIONS(720),
    [aux_sym_comparison_operator_token2] = ACTIONS(720),
    [aux_sym_comparison_operator_token3] = ACTIONS(720),
    [aux_sym_comparison_operator_token4] = ACTIONS(720),
    [aux_sym_comparison_operator_token5] = ACTIONS(720),
    [aux_sym_comparison_operator_token6] = ACTIONS(720),
    [aux_sym_comparison_operator_token7] = ACTIONS(720),
    [aux_sym_comparison_operator_token8] = ACTIONS(720),
    [aux_sym_comparison_operator_token9] = ACTIONS(720),
    [aux_sym_comparison_operator_token10] = ACTIONS(720),
    [aux_sym_comparison_operator_token11] = ACTIONS(720),
    [aux_sym_comparison_operator_token12] = ACTIONS(720),
    [aux_sym_comparison_operator_token13] = ACTIONS(720),
    [aux_sym_comparison_operator_token14] = ACTIONS(720),
    [aux_sym_comparison_operator_token15] = ACTIONS(720),
    [aux_sym_comparison_operator_token16] = ACTIONS(720),
    [aux_sym_comparison_operator_token17] = ACTIONS(720),
    [aux_sym_comparison_operator_token18] = ACTIONS(720),
    [aux_sym_comparison_operator_token19] = ACTIONS(720),
    [aux_sym_comparison_operator_token20] = ACTIONS(720),
    [aux_sym_comparison_operator_token21] = ACTIONS(720),
    [aux_sym_comparison_operator_token22] = ACTIONS(720),
    [aux_sym_comparison_operator_token23] = ACTIONS(720),
    [aux_sym_comparison_operator_token24] = ACTIONS(720),
    [aux_sym_comparison_operator_token25] = ACTIONS(720),
    [aux_sym_comparison_operator_token26] = ACTIONS(720),
    [aux_sym_comparison_operator_token27] = ACTIONS(720),
    [aux_sym_comparison_operator_token28] = ACTIONS(722),
    [aux_sym_comparison_operator_token29] = ACTIONS(720),
    [aux_sym_comparison_operator_token30] = ACTIONS(720),
    [aux_sym_comparison_operator_token31] = ACTIONS(720),
    [aux_sym_comparison_operator_token32] = ACTIONS(720),
    [aux_sym_comparison_operator_token33] = ACTIONS(720),
    [aux_sym_comparison_operator_token34] = ACTIONS(722),
    [aux_sym_comparison_operator_token35] = ACTIONS(720),
    [aux_sym_comparison_operator_token36] = ACTIONS(720),
    [aux_sym_comparison_operator_token37] = ACTIONS(720),
    [aux_sym_comparison_operator_token38] = ACTIONS(720),
    [aux_sym_comparison_operator_token39] = ACTIONS(720),
    [aux_sym_comparison_operator_token40] = ACTIONS(720),
    [aux_sym_comparison_operator_token41] = ACTIONS(720),
    [aux_sym_comparison_operator_token42] = ACTIONS(720),
    [aux_sym_comparison_operator_token43] = ACTIONS(720),
    [aux_sym_comparison_operator_token44] = ACTIONS(720),
    [aux_sym_comparison_operator_token45] = ACTIONS(720),
    [aux_sym_comparison_operator_token46] = ACTIONS(720),
    [aux_sym_comparison_operator_token47] = ACTIONS(720),
    [aux_sym_comparison_operator_token48] = ACTIONS(720),
    [aux_sym_comparison_operator_token49] = ACTIONS(720),
    [aux_sym_comparison_operator_token50] = ACTIONS(720),
    [aux_sym_format_operator_token1] = ACTIONS(720),
    [anon_sym_LPAREN] = ACTIONS(720),
    [anon_sym_RPAREN] = ACTIONS(720),
    [anon_sym_COMMA] = ACTIONS(720),
    [anon_sym_AMP_AMP] = ACTIONS(720),
    [anon_sym_PIPE_PIPE] = ACTIONS(720),
    [anon_sym_PIPE] = ACTIONS(722),
    [aux_sym_logical_expression_token1] = ACTIONS(720),
    [aux_sym_logical_expression_token2] = ACTIONS(720),
    [aux_sym_logical_expression_token3] = ACTIONS(720),
    [aux_sym_bitwise_expression_token1] = ACTIONS(720),
    [aux_sym_bitwise_expression_token2] = ACTIONS(720),
    [aux_sym_bitwise_expression_token3] = ACTIONS(720),
    [anon_sym_PLUS] = ACTIONS(722),
    [anon_sym_DASH] = ACTIONS(722),
    [anon_sym_SLASH] = ACTIONS(722),
    [anon_sym_BSLASH] = ACTIONS(720),
    [anon_sym_PERCENT] = ACTIONS(722),
    [anon_sym_STAR] = ACTIONS(722),
    [anon_sym_DOT_DOT] = ACTIONS(720),
    [anon_sym_PLUS_PLUS] = ACTIONS(720),
    [anon_sym_DASH_DASH] = ACTIONS(720),
    [anon_sym_DOT2] = ACTIONS(722),
    [anon_sym_COLON_COLON] = ACTIONS(720),
    [anon_sym_RBRACK] = ACTIONS(720),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(720),
  },
  [STATE(123)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(724),
    [anon_sym_EQ] = ACTIONS(724),
    [anon_sym_BANG_EQ] = ACTIONS(724),
    [anon_sym_PLUS_EQ] = ACTIONS(724),
    [anon_sym_STAR_EQ] = ACTIONS(724),
    [anon_sym_SLASH_EQ] = ACTIONS(724),
    [anon_sym_PERCENT_EQ] = ACTIONS(724),
    [anon_sym_DASH_EQ] = ACTIONS(724),
    [anon_sym_GT] = ACTIONS(726),
    [anon_sym_GT_GT] = ACTIONS(724),
    [anon_sym_2_GT] = ACTIONS(726),
    [anon_sym_2_GT_GT] = ACTIONS(724),
    [anon_sym_3_GT] = ACTIONS(726),
    [anon_sym_3_GT_GT] = ACTIONS(724),
    [anon_sym_4_GT] = ACTIONS(726),
    [anon_sym_4_GT_GT] = ACTIONS(724),
    [anon_sym_5_GT] = ACTIONS(726),
    [anon_sym_5_GT_GT] = ACTIONS(724),
    [anon_sym_6_GT] = ACTIONS(726),
    [anon_sym_6_GT_GT] = ACTIONS(724),
    [anon_sym_STAR_GT] = ACTIONS(726),
    [anon_sym_STAR_GT_GT] = ACTIONS(724),
    [anon_sym_LT] = ACTIONS(726),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(724),
    [anon_sym_2_GT_AMP1] = ACTIONS(724),
    [anon_sym_3_GT_AMP1] = ACTIONS(724),
    [anon_sym_4_GT_AMP1] = ACTIONS(724),
    [anon_sym_5_GT_AMP1] = ACTIONS(724),
    [anon_sym_6_GT_AMP1] = ACTIONS(724),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(724),
    [anon_sym_1_GT_AMP2] = ACTIONS(724),
    [anon_sym_3_GT_AMP2] = ACTIONS(724),
    [anon_sym_4_GT_AMP2] = ACTIONS(724),
    [anon_sym_5_GT_AMP2] = ACTIONS(724),
    [anon_sym_6_GT_AMP2] = ACTIONS(724),
    [aux_sym_comparison_operator_token1] = ACTIONS(724),
    [aux_sym_comparison_operator_token2] = ACTIONS(724),
    [aux_sym_comparison_operator_token3] = ACTIONS(724),
    [aux_sym_comparison_operator_token4] = ACTIONS(724),
    [aux_sym_comparison_operator_token5] = ACTIONS(724),
    [aux_sym_comparison_operator_token6] = ACTIONS(724),
    [aux_sym_comparison_operator_token7] = ACTIONS(724),
    [aux_sym_comparison_operator_token8] = ACTIONS(724),
    [aux_sym_comparison_operator_token9] = ACTIONS(724),
    [aux_sym_comparison_operator_token10] = ACTIONS(724),
    [aux_sym_comparison_operator_token11] = ACTIONS(724),
    [aux_sym_comparison_operator_token12] = ACTIONS(724),
    [aux_sym_comparison_operator_token13] = ACTIONS(724),
    [aux_sym_comparison_operator_token14] = ACTIONS(724),
    [aux_sym_comparison_operator_token15] = ACTIONS(724),
    [aux_sym_comparison_operator_token16] = ACTIONS(724),
    [aux_sym_comparison_operator_token17] = ACTIONS(724),
    [aux_sym_comparison_operator_token18] = ACTIONS(724),
    [aux_sym_comparison_operator_token19] = ACTIONS(724),
    [aux_sym_comparison_operator_token20] = ACTIONS(724),
    [aux_sym_comparison_operator_token21] = ACTIONS(724),
    [aux_sym_comparison_operator_token22] = ACTIONS(724),
    [aux_sym_comparison_operator_token23] = ACTIONS(724),
    [aux_sym_comparison_operator_token24] = ACTIONS(724),
    [aux_sym_comparison_operator_token25] = ACTIONS(724),
    [aux_sym_comparison_operator_token26] = ACTIONS(724),
    [aux_sym_comparison_operator_token27] = ACTIONS(724),
    [aux_sym_comparison_operator_token28] = ACTIONS(726),
    [aux_sym_comparison_operator_token29] = ACTIONS(724),
    [aux_sym_comparison_operator_token30] = ACTIONS(724),
    [aux_sym_comparison_operator_token31] = ACTIONS(724),
    [aux_sym_comparison_operator_token32] = ACTIONS(724),
    [aux_sym_comparison_operator_token33] = ACTIONS(724),
    [aux_sym_comparison_operator_token34] = ACTIONS(726),
    [aux_sym_comparison_operator_token35] = ACTIONS(724),
    [aux_sym_comparison_operator_token36] = ACTIONS(724),
    [aux_sym_comparison_operator_token37] = ACTIONS(724),
    [aux_sym_comparison_operator_token38] = ACTIONS(724),
    [aux_sym_comparison_operator_token39] = ACTIONS(724),
    [aux_sym_comparison_operator_token40] = ACTIONS(724),
    [aux_sym_comparison_operator_token41] = ACTIONS(724),
    [aux_sym_comparison_operator_token42] = ACTIONS(724),
    [aux_sym_comparison_operator_token43] = ACTIONS(724),
    [aux_sym_comparison_operator_token44] = ACTIONS(724),
    [aux_sym_comparison_operator_token45] = ACTIONS(724),
    [aux_sym_comparison_operator_token46] = ACTIONS(724),
    [aux_sym_comparison_operator_token47] = ACTIONS(724),
    [aux_sym_comparison_operator_token48] = ACTIONS(724),
    [aux_sym_comparison_operator_token49] = ACTIONS(724),
    [aux_sym_comparison_operator_token50] = ACTIONS(724),
    [aux_sym_format_operator_token1] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(724),
    [anon_sym_RPAREN] = ACTIONS(724),
    [anon_sym_COMMA] = ACTIONS(724),
    [anon_sym_AMP_AMP] = ACTIONS(724),
    [anon_sym_PIPE_PIPE] = ACTIONS(724),
    [anon_sym_PIPE] = ACTIONS(726),
    [aux_sym_logical_expression_token1] = ACTIONS(724),
    [aux_sym_logical_expression_token2] = ACTIONS(724),
    [aux_sym_logical_expression_token3] = ACTIONS(724),
    [aux_sym_bitwise_expression_token1] = ACTIONS(724),
    [aux_sym_bitwise_expression_token2] = ACTIONS(724),
    [aux_sym_bitwise_expression_token3] = ACTIONS(724),
    [anon_sym_PLUS] = ACTIONS(726),
    [anon_sym_DASH] = ACTIONS(726),
    [anon_sym_SLASH] = ACTIONS(726),
    [anon_sym_BSLASH] = ACTIONS(724),
    [anon_sym_PERCENT] = ACTIONS(726),
    [anon_sym_STAR] = ACTIONS(726),
    [anon_sym_DOT_DOT] = ACTIONS(724),
    [anon_sym_PLUS_PLUS] = ACTIONS(724),
    [anon_sym_DASH_DASH] = ACTIONS(724),
    [anon_sym_DOT2] = ACTIONS(726),
    [anon_sym_COLON_COLON] = ACTIONS(724),
    [anon_sym_RBRACK] = ACTIONS(724),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(724),
  },
  [STATE(124)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(700),
    [anon_sym_EQ] = ACTIONS(700),
    [anon_sym_BANG_EQ] = ACTIONS(700),
    [anon_sym_PLUS_EQ] = ACTIONS(700),
    [anon_sym_STAR_EQ] = ACTIONS(700),
    [anon_sym_SLASH_EQ] = ACTIONS(700),
    [anon_sym_PERCENT_EQ] = ACTIONS(700),
    [anon_sym_DASH_EQ] = ACTIONS(700),
    [anon_sym_GT] = ACTIONS(702),
    [anon_sym_GT_GT] = ACTIONS(700),
    [anon_sym_2_GT] = ACTIONS(702),
    [anon_sym_2_GT_GT] = ACTIONS(700),
    [anon_sym_3_GT] = ACTIONS(702),
    [anon_sym_3_GT_GT] = ACTIONS(700),
    [anon_sym_4_GT] = ACTIONS(702),
    [anon_sym_4_GT_GT] = ACTIONS(700),
    [anon_sym_5_GT] = ACTIONS(702),
    [anon_sym_5_GT_GT] = ACTIONS(700),
    [anon_sym_6_GT] = ACTIONS(702),
    [anon_sym_6_GT_GT] = ACTIONS(700),
    [anon_sym_STAR_GT] = ACTIONS(702),
    [anon_sym_STAR_GT_GT] = ACTIONS(700),
    [anon_sym_LT] = ACTIONS(702),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(700),
    [anon_sym_2_GT_AMP1] = ACTIONS(700),
    [anon_sym_3_GT_AMP1] = ACTIONS(700),
    [anon_sym_4_GT_AMP1] = ACTIONS(700),
    [anon_sym_5_GT_AMP1] = ACTIONS(700),
    [anon_sym_6_GT_AMP1] = ACTIONS(700),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(700),
    [anon_sym_1_GT_AMP2] = ACTIONS(700),
    [anon_sym_3_GT_AMP2] = ACTIONS(700),
    [anon_sym_4_GT_AMP2] = ACTIONS(700),
    [anon_sym_5_GT_AMP2] = ACTIONS(700),
    [anon_sym_6_GT_AMP2] = ACTIONS(700),
    [aux_sym_comparison_operator_token1] = ACTIONS(700),
    [aux_sym_comparison_operator_token2] = ACTIONS(700),
    [aux_sym_comparison_operator_token3] = ACTIONS(700),
    [aux_sym_comparison_operator_token4] = ACTIONS(700),
    [aux_sym_comparison_operator_token5] = ACTIONS(700),
    [aux_sym_comparison_operator_token6] = ACTIONS(700),
    [aux_sym_comparison_operator_token7] = ACTIONS(700),
    [aux_sym_comparison_operator_token8] = ACTIONS(700),
    [aux_sym_comparison_operator_token9] = ACTIONS(700),
    [aux_sym_comparison_operator_token10] = ACTIONS(700),
    [aux_sym_comparison_operator_token11] = ACTIONS(700),
    [aux_sym_comparison_operator_token12] = ACTIONS(700),
    [aux_sym_comparison_operator_token13] = ACTIONS(700),
    [aux_sym_comparison_operator_token14] = ACTIONS(700),
    [aux_sym_comparison_operator_token15] = ACTIONS(700),
    [aux_sym_comparison_operator_token16] = ACTIONS(700),
    [aux_sym_comparison_operator_token17] = ACTIONS(700),
    [aux_sym_comparison_operator_token18] = ACTIONS(700),
    [aux_sym_comparison_operator_token19] = ACTIONS(700),
    [aux_sym_comparison_operator_token20] = ACTIONS(700),
    [aux_sym_comparison_operator_token21] = ACTIONS(700),
    [aux_sym_comparison_operator_token22] = ACTIONS(700),
    [aux_sym_comparison_operator_token23] = ACTIONS(700),
    [aux_sym_comparison_operator_token24] = ACTIONS(700),
    [aux_sym_comparison_operator_token25] = ACTIONS(700),
    [aux_sym_comparison_operator_token26] = ACTIONS(700),
    [aux_sym_comparison_operator_token27] = ACTIONS(700),
    [aux_sym_comparison_operator_token28] = ACTIONS(702),
    [aux_sym_comparison_operator_token29] = ACTIONS(700),
    [aux_sym_comparison_operator_token30] = ACTIONS(700),
    [aux_sym_comparison_operator_token31] = ACTIONS(700),
    [aux_sym_comparison_operator_token32] = ACTIONS(700),
    [aux_sym_comparison_operator_token33] = ACTIONS(700),
    [aux_sym_comparison_operator_token34] = ACTIONS(702),
    [aux_sym_comparison_operator_token35] = ACTIONS(700),
    [aux_sym_comparison_operator_token36] = ACTIONS(700),
    [aux_sym_comparison_operator_token37] = ACTIONS(700),
    [aux_sym_comparison_operator_token38] = ACTIONS(700),
    [aux_sym_comparison_operator_token39] = ACTIONS(700),
    [aux_sym_comparison_operator_token40] = ACTIONS(700),
    [aux_sym_comparison_operator_token41] = ACTIONS(700),
    [aux_sym_comparison_operator_token42] = ACTIONS(700),
    [aux_sym_comparison_operator_token43] = ACTIONS(700),
    [aux_sym_comparison_operator_token44] = ACTIONS(700),
    [aux_sym_comparison_operator_token45] = ACTIONS(700),
    [aux_sym_comparison_operator_token46] = ACTIONS(700),
    [aux_sym_comparison_operator_token47] = ACTIONS(700),
    [aux_sym_comparison_operator_token48] = ACTIONS(700),
    [aux_sym_comparison_operator_token49] = ACTIONS(700),
    [aux_sym_comparison_operator_token50] = ACTIONS(700),
    [aux_sym_format_operator_token1] = ACTIONS(700),
    [anon_sym_LPAREN] = ACTIONS(700),
    [anon_sym_COMMA] = ACTIONS(700),
    [anon_sym_AMP_AMP] = ACTIONS(700),
    [anon_sym_PIPE_PIPE] = ACTIONS(700),
    [anon_sym_PIPE] = ACTIONS(702),
    [aux_sym_logical_expression_token1] = ACTIONS(700),
    [aux_sym_logical_expression_token2] = ACTIONS(700),
    [aux_sym_logical_expression_token3] = ACTIONS(700),
    [aux_sym_bitwise_expression_token1] = ACTIONS(700),
    [aux_sym_bitwise_expression_token2] = ACTIONS(700),
    [aux_sym_bitwise_expression_token3] = ACTIONS(700),
    [anon_sym_PLUS] = ACTIONS(702),
    [anon_sym_DASH] = ACTIONS(702),
    [anon_sym_SLASH] = ACTIONS(702),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_PERCENT] = ACTIONS(702),
    [anon_sym_STAR] = ACTIONS(702),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_PLUS_PLUS] = ACTIONS(700),
    [anon_sym_DASH_DASH] = ACTIONS(700),
    [anon_sym_DOT2] = ACTIONS(702),
    [anon_sym_COLON_COLON] = ACTIONS(700),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(700),
    [sym__statement_terminator] = ACTIONS(700),
  },
  [STATE(125)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(642),
    [anon_sym_EQ] = ACTIONS(642),
    [anon_sym_BANG_EQ] = ACTIONS(642),
    [anon_sym_PLUS_EQ] = ACTIONS(642),
    [anon_sym_STAR_EQ] = ACTIONS(642),
    [anon_sym_SLASH_EQ] = ACTIONS(642),
    [anon_sym_PERCENT_EQ] = ACTIONS(642),
    [anon_sym_DASH_EQ] = ACTIONS(642),
    [anon_sym_GT] = ACTIONS(644),
    [anon_sym_GT_GT] = ACTIONS(642),
    [anon_sym_2_GT] = ACTIONS(644),
    [anon_sym_2_GT_GT] = ACTIONS(642),
    [anon_sym_3_GT] = ACTIONS(644),
    [anon_sym_3_GT_GT] = ACTIONS(642),
    [anon_sym_4_GT] = ACTIONS(644),
    [anon_sym_4_GT_GT] = ACTIONS(642),
    [anon_sym_5_GT] = ACTIONS(644),
    [anon_sym_5_GT_GT] = ACTIONS(642),
    [anon_sym_6_GT] = ACTIONS(644),
    [anon_sym_6_GT_GT] = ACTIONS(642),
    [anon_sym_STAR_GT] = ACTIONS(644),
    [anon_sym_STAR_GT_GT] = ACTIONS(642),
    [anon_sym_LT] = ACTIONS(644),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(642),
    [anon_sym_2_GT_AMP1] = ACTIONS(642),
    [anon_sym_3_GT_AMP1] = ACTIONS(642),
    [anon_sym_4_GT_AMP1] = ACTIONS(642),
    [anon_sym_5_GT_AMP1] = ACTIONS(642),
    [anon_sym_6_GT_AMP1] = ACTIONS(642),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(642),
    [anon_sym_1_GT_AMP2] = ACTIONS(642),
    [anon_sym_3_GT_AMP2] = ACTIONS(642),
    [anon_sym_4_GT_AMP2] = ACTIONS(642),
    [anon_sym_5_GT_AMP2] = ACTIONS(642),
    [anon_sym_6_GT_AMP2] = ACTIONS(642),
    [aux_sym_comparison_operator_token1] = ACTIONS(642),
    [aux_sym_comparison_operator_token2] = ACTIONS(642),
    [aux_sym_comparison_operator_token3] = ACTIONS(642),
    [aux_sym_comparison_operator_token4] = ACTIONS(642),
    [aux_sym_comparison_operator_token5] = ACTIONS(642),
    [aux_sym_comparison_operator_token6] = ACTIONS(642),
    [aux_sym_comparison_operator_token7] = ACTIONS(642),
    [aux_sym_comparison_operator_token8] = ACTIONS(642),
    [aux_sym_comparison_operator_token9] = ACTIONS(642),
    [aux_sym_comparison_operator_token10] = ACTIONS(642),
    [aux_sym_comparison_operator_token11] = ACTIONS(642),
    [aux_sym_comparison_operator_token12] = ACTIONS(642),
    [aux_sym_comparison_operator_token13] = ACTIONS(642),
    [aux_sym_comparison_operator_token14] = ACTIONS(642),
    [aux_sym_comparison_operator_token15] = ACTIONS(642),
    [aux_sym_comparison_operator_token16] = ACTIONS(642),
    [aux_sym_comparison_operator_token17] = ACTIONS(642),
    [aux_sym_comparison_operator_token18] = ACTIONS(642),
    [aux_sym_comparison_operator_token19] = ACTIONS(642),
    [aux_sym_comparison_operator_token20] = ACTIONS(642),
    [aux_sym_comparison_operator_token21] = ACTIONS(642),
    [aux_sym_comparison_operator_token22] = ACTIONS(642),
    [aux_sym_comparison_operator_token23] = ACTIONS(642),
    [aux_sym_comparison_operator_token24] = ACTIONS(642),
    [aux_sym_comparison_operator_token25] = ACTIONS(642),
    [aux_sym_comparison_operator_token26] = ACTIONS(642),
    [aux_sym_comparison_operator_token27] = ACTIONS(642),
    [aux_sym_comparison_operator_token28] = ACTIONS(644),
    [aux_sym_comparison_operator_token29] = ACTIONS(642),
    [aux_sym_comparison_operator_token30] = ACTIONS(642),
    [aux_sym_comparison_operator_token31] = ACTIONS(642),
    [aux_sym_comparison_operator_token32] = ACTIONS(642),
    [aux_sym_comparison_operator_token33] = ACTIONS(642),
    [aux_sym_comparison_operator_token34] = ACTIONS(644),
    [aux_sym_comparison_operator_token35] = ACTIONS(642),
    [aux_sym_comparison_operator_token36] = ACTIONS(642),
    [aux_sym_comparison_operator_token37] = ACTIONS(642),
    [aux_sym_comparison_operator_token38] = ACTIONS(642),
    [aux_sym_comparison_operator_token39] = ACTIONS(642),
    [aux_sym_comparison_operator_token40] = ACTIONS(642),
    [aux_sym_comparison_operator_token41] = ACTIONS(642),
    [aux_sym_comparison_operator_token42] = ACTIONS(642),
    [aux_sym_comparison_operator_token43] = ACTIONS(642),
    [aux_sym_comparison_operator_token44] = ACTIONS(642),
    [aux_sym_comparison_operator_token45] = ACTIONS(642),
    [aux_sym_comparison_operator_token46] = ACTIONS(642),
    [aux_sym_comparison_operator_token47] = ACTIONS(642),
    [aux_sym_comparison_operator_token48] = ACTIONS(642),
    [aux_sym_comparison_operator_token49] = ACTIONS(642),
    [aux_sym_comparison_operator_token50] = ACTIONS(642),
    [aux_sym_format_operator_token1] = ACTIONS(642),
    [anon_sym_LPAREN] = ACTIONS(642),
    [anon_sym_COMMA] = ACTIONS(642),
    [anon_sym_AMP_AMP] = ACTIONS(642),
    [anon_sym_PIPE_PIPE] = ACTIONS(642),
    [anon_sym_PIPE] = ACTIONS(644),
    [aux_sym_logical_expression_token1] = ACTIONS(642),
    [aux_sym_logical_expression_token2] = ACTIONS(642),
    [aux_sym_logical_expression_token3] = ACTIONS(642),
    [aux_sym_bitwise_expression_token1] = ACTIONS(642),
    [aux_sym_bitwise_expression_token2] = ACTIONS(642),
    [aux_sym_bitwise_expression_token3] = ACTIONS(642),
    [anon_sym_PLUS] = ACTIONS(644),
    [anon_sym_DASH] = ACTIONS(644),
    [anon_sym_SLASH] = ACTIONS(644),
    [anon_sym_BSLASH] = ACTIONS(642),
    [anon_sym_PERCENT] = ACTIONS(644),
    [anon_sym_STAR] = ACTIONS(644),
    [anon_sym_DOT_DOT] = ACTIONS(642),
    [anon_sym_PLUS_PLUS] = ACTIONS(642),
    [anon_sym_DASH_DASH] = ACTIONS(642),
    [anon_sym_DOT2] = ACTIONS(644),
    [anon_sym_COLON_COLON] = ACTIONS(642),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(642),
    [sym__statement_terminator] = ACTIONS(642),
  },
  [STATE(126)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(610),
    [anon_sym_EQ] = ACTIONS(610),
    [anon_sym_BANG_EQ] = ACTIONS(610),
    [anon_sym_PLUS_EQ] = ACTIONS(610),
    [anon_sym_STAR_EQ] = ACTIONS(610),
    [anon_sym_SLASH_EQ] = ACTIONS(610),
    [anon_sym_PERCENT_EQ] = ACTIONS(610),
    [anon_sym_DASH_EQ] = ACTIONS(610),
    [anon_sym_GT] = ACTIONS(612),
    [anon_sym_GT_GT] = ACTIONS(610),
    [anon_sym_2_GT] = ACTIONS(612),
    [anon_sym_2_GT_GT] = ACTIONS(610),
    [anon_sym_3_GT] = ACTIONS(612),
    [anon_sym_3_GT_GT] = ACTIONS(610),
    [anon_sym_4_GT] = ACTIONS(612),
    [anon_sym_4_GT_GT] = ACTIONS(610),
    [anon_sym_5_GT] = ACTIONS(612),
    [anon_sym_5_GT_GT] = ACTIONS(610),
    [anon_sym_6_GT] = ACTIONS(612),
    [anon_sym_6_GT_GT] = ACTIONS(610),
    [anon_sym_STAR_GT] = ACTIONS(612),
    [anon_sym_STAR_GT_GT] = ACTIONS(610),
    [anon_sym_LT] = ACTIONS(612),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(610),
    [anon_sym_2_GT_AMP1] = ACTIONS(610),
    [anon_sym_3_GT_AMP1] = ACTIONS(610),
    [anon_sym_4_GT_AMP1] = ACTIONS(610),
    [anon_sym_5_GT_AMP1] = ACTIONS(610),
    [anon_sym_6_GT_AMP1] = ACTIONS(610),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(610),
    [anon_sym_1_GT_AMP2] = ACTIONS(610),
    [anon_sym_3_GT_AMP2] = ACTIONS(610),
    [anon_sym_4_GT_AMP2] = ACTIONS(610),
    [anon_sym_5_GT_AMP2] = ACTIONS(610),
    [anon_sym_6_GT_AMP2] = ACTIONS(610),
    [aux_sym_comparison_operator_token1] = ACTIONS(610),
    [aux_sym_comparison_operator_token2] = ACTIONS(610),
    [aux_sym_comparison_operator_token3] = ACTIONS(610),
    [aux_sym_comparison_operator_token4] = ACTIONS(610),
    [aux_sym_comparison_operator_token5] = ACTIONS(610),
    [aux_sym_comparison_operator_token6] = ACTIONS(610),
    [aux_sym_comparison_operator_token7] = ACTIONS(610),
    [aux_sym_comparison_operator_token8] = ACTIONS(610),
    [aux_sym_comparison_operator_token9] = ACTIONS(610),
    [aux_sym_comparison_operator_token10] = ACTIONS(610),
    [aux_sym_comparison_operator_token11] = ACTIONS(610),
    [aux_sym_comparison_operator_token12] = ACTIONS(610),
    [aux_sym_comparison_operator_token13] = ACTIONS(610),
    [aux_sym_comparison_operator_token14] = ACTIONS(610),
    [aux_sym_comparison_operator_token15] = ACTIONS(610),
    [aux_sym_comparison_operator_token16] = ACTIONS(610),
    [aux_sym_comparison_operator_token17] = ACTIONS(610),
    [aux_sym_comparison_operator_token18] = ACTIONS(610),
    [aux_sym_comparison_operator_token19] = ACTIONS(610),
    [aux_sym_comparison_operator_token20] = ACTIONS(610),
    [aux_sym_comparison_operator_token21] = ACTIONS(610),
    [aux_sym_comparison_operator_token22] = ACTIONS(610),
    [aux_sym_comparison_operator_token23] = ACTIONS(610),
    [aux_sym_comparison_operator_token24] = ACTIONS(610),
    [aux_sym_comparison_operator_token25] = ACTIONS(610),
    [aux_sym_comparison_operator_token26] = ACTIONS(610),
    [aux_sym_comparison_operator_token27] = ACTIONS(610),
    [aux_sym_comparison_operator_token28] = ACTIONS(612),
    [aux_sym_comparison_operator_token29] = ACTIONS(610),
    [aux_sym_comparison_operator_token30] = ACTIONS(610),
    [aux_sym_comparison_operator_token31] = ACTIONS(610),
    [aux_sym_comparison_operator_token32] = ACTIONS(610),
    [aux_sym_comparison_operator_token33] = ACTIONS(610),
    [aux_sym_comparison_operator_token34] = ACTIONS(612),
    [aux_sym_comparison_operator_token35] = ACTIONS(610),
    [aux_sym_comparison_operator_token36] = ACTIONS(610),
    [aux_sym_comparison_operator_token37] = ACTIONS(610),
    [aux_sym_comparison_operator_token38] = ACTIONS(610),
    [aux_sym_comparison_operator_token39] = ACTIONS(610),
    [aux_sym_comparison_operator_token40] = ACTIONS(610),
    [aux_sym_comparison_operator_token41] = ACTIONS(610),
    [aux_sym_comparison_operator_token42] = ACTIONS(610),
    [aux_sym_comparison_operator_token43] = ACTIONS(610),
    [aux_sym_comparison_operator_token44] = ACTIONS(610),
    [aux_sym_comparison_operator_token45] = ACTIONS(610),
    [aux_sym_comparison_operator_token46] = ACTIONS(610),
    [aux_sym_comparison_operator_token47] = ACTIONS(610),
    [aux_sym_comparison_operator_token48] = ACTIONS(610),
    [aux_sym_comparison_operator_token49] = ACTIONS(610),
    [aux_sym_comparison_operator_token50] = ACTIONS(610),
    [aux_sym_format_operator_token1] = ACTIONS(610),
    [anon_sym_LPAREN] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(610),
    [anon_sym_AMP_AMP] = ACTIONS(610),
    [anon_sym_PIPE_PIPE] = ACTIONS(610),
    [anon_sym_PIPE] = ACTIONS(612),
    [aux_sym_logical_expression_token1] = ACTIONS(610),
    [aux_sym_logical_expression_token2] = ACTIONS(610),
    [aux_sym_logical_expression_token3] = ACTIONS(610),
    [aux_sym_bitwise_expression_token1] = ACTIONS(610),
    [aux_sym_bitwise_expression_token2] = ACTIONS(610),
    [aux_sym_bitwise_expression_token3] = ACTIONS(610),
    [anon_sym_PLUS] = ACTIONS(612),
    [anon_sym_DASH] = ACTIONS(612),
    [anon_sym_SLASH] = ACTIONS(612),
    [anon_sym_BSLASH] = ACTIONS(610),
    [anon_sym_PERCENT] = ACTIONS(612),
    [anon_sym_STAR] = ACTIONS(612),
    [anon_sym_DOT_DOT] = ACTIONS(610),
    [anon_sym_PLUS_PLUS] = ACTIONS(610),
    [anon_sym_DASH_DASH] = ACTIONS(610),
    [anon_sym_DOT2] = ACTIONS(612),
    [anon_sym_COLON_COLON] = ACTIONS(610),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(610),
    [sym__statement_terminator] = ACTIONS(610),
  },
  [STATE(127)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(682),
    [anon_sym_EQ] = ACTIONS(682),
    [anon_sym_BANG_EQ] = ACTIONS(682),
    [anon_sym_PLUS_EQ] = ACTIONS(682),
    [anon_sym_STAR_EQ] = ACTIONS(682),
    [anon_sym_SLASH_EQ] = ACTIONS(682),
    [anon_sym_PERCENT_EQ] = ACTIONS(682),
    [anon_sym_DASH_EQ] = ACTIONS(682),
    [anon_sym_GT] = ACTIONS(684),
    [anon_sym_GT_GT] = ACTIONS(682),
    [anon_sym_2_GT] = ACTIONS(684),
    [anon_sym_2_GT_GT] = ACTIONS(682),
    [anon_sym_3_GT] = ACTIONS(684),
    [anon_sym_3_GT_GT] = ACTIONS(682),
    [anon_sym_4_GT] = ACTIONS(684),
    [anon_sym_4_GT_GT] = ACTIONS(682),
    [anon_sym_5_GT] = ACTIONS(684),
    [anon_sym_5_GT_GT] = ACTIONS(682),
    [anon_sym_6_GT] = ACTIONS(684),
    [anon_sym_6_GT_GT] = ACTIONS(682),
    [anon_sym_STAR_GT] = ACTIONS(684),
    [anon_sym_STAR_GT_GT] = ACTIONS(682),
    [anon_sym_LT] = ACTIONS(684),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(682),
    [anon_sym_2_GT_AMP1] = ACTIONS(682),
    [anon_sym_3_GT_AMP1] = ACTIONS(682),
    [anon_sym_4_GT_AMP1] = ACTIONS(682),
    [anon_sym_5_GT_AMP1] = ACTIONS(682),
    [anon_sym_6_GT_AMP1] = ACTIONS(682),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(682),
    [anon_sym_1_GT_AMP2] = ACTIONS(682),
    [anon_sym_3_GT_AMP2] = ACTIONS(682),
    [anon_sym_4_GT_AMP2] = ACTIONS(682),
    [anon_sym_5_GT_AMP2] = ACTIONS(682),
    [anon_sym_6_GT_AMP2] = ACTIONS(682),
    [aux_sym_comparison_operator_token1] = ACTIONS(682),
    [aux_sym_comparison_operator_token2] = ACTIONS(682),
    [aux_sym_comparison_operator_token3] = ACTIONS(682),
    [aux_sym_comparison_operator_token4] = ACTIONS(682),
    [aux_sym_comparison_operator_token5] = ACTIONS(682),
    [aux_sym_comparison_operator_token6] = ACTIONS(682),
    [aux_sym_comparison_operator_token7] = ACTIONS(682),
    [aux_sym_comparison_operator_token8] = ACTIONS(682),
    [aux_sym_comparison_operator_token9] = ACTIONS(682),
    [aux_sym_comparison_operator_token10] = ACTIONS(682),
    [aux_sym_comparison_operator_token11] = ACTIONS(682),
    [aux_sym_comparison_operator_token12] = ACTIONS(682),
    [aux_sym_comparison_operator_token13] = ACTIONS(682),
    [aux_sym_comparison_operator_token14] = ACTIONS(682),
    [aux_sym_comparison_operator_token15] = ACTIONS(682),
    [aux_sym_comparison_operator_token16] = ACTIONS(682),
    [aux_sym_comparison_operator_token17] = ACTIONS(682),
    [aux_sym_comparison_operator_token18] = ACTIONS(682),
    [aux_sym_comparison_operator_token19] = ACTIONS(682),
    [aux_sym_comparison_operator_token20] = ACTIONS(682),
    [aux_sym_comparison_operator_token21] = ACTIONS(682),
    [aux_sym_comparison_operator_token22] = ACTIONS(682),
    [aux_sym_comparison_operator_token23] = ACTIONS(682),
    [aux_sym_comparison_operator_token24] = ACTIONS(682),
    [aux_sym_comparison_operator_token25] = ACTIONS(682),
    [aux_sym_comparison_operator_token26] = ACTIONS(682),
    [aux_sym_comparison_operator_token27] = ACTIONS(682),
    [aux_sym_comparison_operator_token28] = ACTIONS(684),
    [aux_sym_comparison_operator_token29] = ACTIONS(682),
    [aux_sym_comparison_operator_token30] = ACTIONS(682),
    [aux_sym_comparison_operator_token31] = ACTIONS(682),
    [aux_sym_comparison_operator_token32] = ACTIONS(682),
    [aux_sym_comparison_operator_token33] = ACTIONS(682),
    [aux_sym_comparison_operator_token34] = ACTIONS(684),
    [aux_sym_comparison_operator_token35] = ACTIONS(682),
    [aux_sym_comparison_operator_token36] = ACTIONS(682),
    [aux_sym_comparison_operator_token37] = ACTIONS(682),
    [aux_sym_comparison_operator_token38] = ACTIONS(682),
    [aux_sym_comparison_operator_token39] = ACTIONS(682),
    [aux_sym_comparison_operator_token40] = ACTIONS(682),
    [aux_sym_comparison_operator_token41] = ACTIONS(682),
    [aux_sym_comparison_operator_token42] = ACTIONS(682),
    [aux_sym_comparison_operator_token43] = ACTIONS(682),
    [aux_sym_comparison_operator_token44] = ACTIONS(682),
    [aux_sym_comparison_operator_token45] = ACTIONS(682),
    [aux_sym_comparison_operator_token46] = ACTIONS(682),
    [aux_sym_comparison_operator_token47] = ACTIONS(682),
    [aux_sym_comparison_operator_token48] = ACTIONS(682),
    [aux_sym_comparison_operator_token49] = ACTIONS(682),
    [aux_sym_comparison_operator_token50] = ACTIONS(682),
    [aux_sym_format_operator_token1] = ACTIONS(682),
    [anon_sym_LPAREN] = ACTIONS(682),
    [anon_sym_COMMA] = ACTIONS(682),
    [anon_sym_AMP_AMP] = ACTIONS(682),
    [anon_sym_PIPE_PIPE] = ACTIONS(682),
    [anon_sym_PIPE] = ACTIONS(684),
    [aux_sym_logical_expression_token1] = ACTIONS(682),
    [aux_sym_logical_expression_token2] = ACTIONS(682),
    [aux_sym_logical_expression_token3] = ACTIONS(682),
    [aux_sym_bitwise_expression_token1] = ACTIONS(682),
    [aux_sym_bitwise_expression_token2] = ACTIONS(682),
    [aux_sym_bitwise_expression_token3] = ACTIONS(682),
    [anon_sym_PLUS] = ACTIONS(684),
    [anon_sym_DASH] = ACTIONS(684),
    [anon_sym_SLASH] = ACTIONS(684),
    [anon_sym_BSLASH] = ACTIONS(682),
    [anon_sym_PERCENT] = ACTIONS(684),
    [anon_sym_STAR] = ACTIONS(684),
    [anon_sym_DOT_DOT] = ACTIONS(682),
    [anon_sym_PLUS_PLUS] = ACTIONS(682),
    [anon_sym_DASH_DASH] = ACTIONS(682),
    [anon_sym_DOT2] = ACTIONS(684),
    [anon_sym_COLON_COLON] = ACTIONS(682),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(682),
    [sym__statement_terminator] = ACTIONS(682),
  },
  [STATE(128)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(646),
    [anon_sym_EQ] = ACTIONS(646),
    [anon_sym_BANG_EQ] = ACTIONS(646),
    [anon_sym_PLUS_EQ] = ACTIONS(646),
    [anon_sym_STAR_EQ] = ACTIONS(646),
    [anon_sym_SLASH_EQ] = ACTIONS(646),
    [anon_sym_PERCENT_EQ] = ACTIONS(646),
    [anon_sym_DASH_EQ] = ACTIONS(646),
    [anon_sym_GT] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(646),
    [anon_sym_2_GT] = ACTIONS(648),
    [anon_sym_2_GT_GT] = ACTIONS(646),
    [anon_sym_3_GT] = ACTIONS(648),
    [anon_sym_3_GT_GT] = ACTIONS(646),
    [anon_sym_4_GT] = ACTIONS(648),
    [anon_sym_4_GT_GT] = ACTIONS(646),
    [anon_sym_5_GT] = ACTIONS(648),
    [anon_sym_5_GT_GT] = ACTIONS(646),
    [anon_sym_6_GT] = ACTIONS(648),
    [anon_sym_6_GT_GT] = ACTIONS(646),
    [anon_sym_STAR_GT] = ACTIONS(648),
    [anon_sym_STAR_GT_GT] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(646),
    [anon_sym_2_GT_AMP1] = ACTIONS(646),
    [anon_sym_3_GT_AMP1] = ACTIONS(646),
    [anon_sym_4_GT_AMP1] = ACTIONS(646),
    [anon_sym_5_GT_AMP1] = ACTIONS(646),
    [anon_sym_6_GT_AMP1] = ACTIONS(646),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(646),
    [anon_sym_1_GT_AMP2] = ACTIONS(646),
    [anon_sym_3_GT_AMP2] = ACTIONS(646),
    [anon_sym_4_GT_AMP2] = ACTIONS(646),
    [anon_sym_5_GT_AMP2] = ACTIONS(646),
    [anon_sym_6_GT_AMP2] = ACTIONS(646),
    [aux_sym_comparison_operator_token1] = ACTIONS(646),
    [aux_sym_comparison_operator_token2] = ACTIONS(646),
    [aux_sym_comparison_operator_token3] = ACTIONS(646),
    [aux_sym_comparison_operator_token4] = ACTIONS(646),
    [aux_sym_comparison_operator_token5] = ACTIONS(646),
    [aux_sym_comparison_operator_token6] = ACTIONS(646),
    [aux_sym_comparison_operator_token7] = ACTIONS(646),
    [aux_sym_comparison_operator_token8] = ACTIONS(646),
    [aux_sym_comparison_operator_token9] = ACTIONS(646),
    [aux_sym_comparison_operator_token10] = ACTIONS(646),
    [aux_sym_comparison_operator_token11] = ACTIONS(646),
    [aux_sym_comparison_operator_token12] = ACTIONS(646),
    [aux_sym_comparison_operator_token13] = ACTIONS(646),
    [aux_sym_comparison_operator_token14] = ACTIONS(646),
    [aux_sym_comparison_operator_token15] = ACTIONS(646),
    [aux_sym_comparison_operator_token16] = ACTIONS(646),
    [aux_sym_comparison_operator_token17] = ACTIONS(646),
    [aux_sym_comparison_operator_token18] = ACTIONS(646),
    [aux_sym_comparison_operator_token19] = ACTIONS(646),
    [aux_sym_comparison_operator_token20] = ACTIONS(646),
    [aux_sym_comparison_operator_token21] = ACTIONS(646),
    [aux_sym_comparison_operator_token22] = ACTIONS(646),
    [aux_sym_comparison_operator_token23] = ACTIONS(646),
    [aux_sym_comparison_operator_token24] = ACTIONS(646),
    [aux_sym_comparison_operator_token25] = ACTIONS(646),
    [aux_sym_comparison_operator_token26] = ACTIONS(646),
    [aux_sym_comparison_operator_token27] = ACTIONS(646),
    [aux_sym_comparison_operator_token28] = ACTIONS(648),
    [aux_sym_comparison_operator_token29] = ACTIONS(646),
    [aux_sym_comparison_operator_token30] = ACTIONS(646),
    [aux_sym_comparison_operator_token31] = ACTIONS(646),
    [aux_sym_comparison_operator_token32] = ACTIONS(646),
    [aux_sym_comparison_operator_token33] = ACTIONS(646),
    [aux_sym_comparison_operator_token34] = ACTIONS(648),
    [aux_sym_comparison_operator_token35] = ACTIONS(646),
    [aux_sym_comparison_operator_token36] = ACTIONS(646),
    [aux_sym_comparison_operator_token37] = ACTIONS(646),
    [aux_sym_comparison_operator_token38] = ACTIONS(646),
    [aux_sym_comparison_operator_token39] = ACTIONS(646),
    [aux_sym_comparison_operator_token40] = ACTIONS(646),
    [aux_sym_comparison_operator_token41] = ACTIONS(646),
    [aux_sym_comparison_operator_token42] = ACTIONS(646),
    [aux_sym_comparison_operator_token43] = ACTIONS(646),
    [aux_sym_comparison_operator_token44] = ACTIONS(646),
    [aux_sym_comparison_operator_token45] = ACTIONS(646),
    [aux_sym_comparison_operator_token46] = ACTIONS(646),
    [aux_sym_comparison_operator_token47] = ACTIONS(646),
    [aux_sym_comparison_operator_token48] = ACTIONS(646),
    [aux_sym_comparison_operator_token49] = ACTIONS(646),
    [aux_sym_comparison_operator_token50] = ACTIONS(646),
    [aux_sym_format_operator_token1] = ACTIONS(646),
    [anon_sym_LPAREN] = ACTIONS(646),
    [anon_sym_COMMA] = ACTIONS(646),
    [anon_sym_AMP_AMP] = ACTIONS(646),
    [anon_sym_PIPE_PIPE] = ACTIONS(646),
    [anon_sym_PIPE] = ACTIONS(648),
    [aux_sym_logical_expression_token1] = ACTIONS(646),
    [aux_sym_logical_expression_token2] = ACTIONS(646),
    [aux_sym_logical_expression_token3] = ACTIONS(646),
    [aux_sym_bitwise_expression_token1] = ACTIONS(646),
    [aux_sym_bitwise_expression_token2] = ACTIONS(646),
    [aux_sym_bitwise_expression_token3] = ACTIONS(646),
    [anon_sym_PLUS] = ACTIONS(648),
    [anon_sym_DASH] = ACTIONS(648),
    [anon_sym_SLASH] = ACTIONS(648),
    [anon_sym_BSLASH] = ACTIONS(646),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR] = ACTIONS(648),
    [anon_sym_DOT_DOT] = ACTIONS(646),
    [anon_sym_PLUS_PLUS] = ACTIONS(646),
    [anon_sym_DASH_DASH] = ACTIONS(646),
    [anon_sym_DOT2] = ACTIONS(648),
    [anon_sym_COLON_COLON] = ACTIONS(646),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(646),
    [sym__statement_terminator] = ACTIONS(646),
  },
  [STATE(129)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_EQ] = ACTIONS(596),
    [anon_sym_BANG_EQ] = ACTIONS(596),
    [anon_sym_PLUS_EQ] = ACTIONS(596),
    [anon_sym_STAR_EQ] = ACTIONS(596),
    [anon_sym_SLASH_EQ] = ACTIONS(596),
    [anon_sym_PERCENT_EQ] = ACTIONS(596),
    [anon_sym_DASH_EQ] = ACTIONS(596),
    [anon_sym_GT] = ACTIONS(598),
    [anon_sym_GT_GT] = ACTIONS(596),
    [anon_sym_2_GT] = ACTIONS(598),
    [anon_sym_2_GT_GT] = ACTIONS(596),
    [anon_sym_3_GT] = ACTIONS(598),
    [anon_sym_3_GT_GT] = ACTIONS(596),
    [anon_sym_4_GT] = ACTIONS(598),
    [anon_sym_4_GT_GT] = ACTIONS(596),
    [anon_sym_5_GT] = ACTIONS(598),
    [anon_sym_5_GT_GT] = ACTIONS(596),
    [anon_sym_6_GT] = ACTIONS(598),
    [anon_sym_6_GT_GT] = ACTIONS(596),
    [anon_sym_STAR_GT] = ACTIONS(598),
    [anon_sym_STAR_GT_GT] = ACTIONS(596),
    [anon_sym_LT] = ACTIONS(598),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(596),
    [anon_sym_2_GT_AMP1] = ACTIONS(596),
    [anon_sym_3_GT_AMP1] = ACTIONS(596),
    [anon_sym_4_GT_AMP1] = ACTIONS(596),
    [anon_sym_5_GT_AMP1] = ACTIONS(596),
    [anon_sym_6_GT_AMP1] = ACTIONS(596),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(596),
    [anon_sym_1_GT_AMP2] = ACTIONS(596),
    [anon_sym_3_GT_AMP2] = ACTIONS(596),
    [anon_sym_4_GT_AMP2] = ACTIONS(596),
    [anon_sym_5_GT_AMP2] = ACTIONS(596),
    [anon_sym_6_GT_AMP2] = ACTIONS(596),
    [aux_sym_comparison_operator_token1] = ACTIONS(596),
    [aux_sym_comparison_operator_token2] = ACTIONS(596),
    [aux_sym_comparison_operator_token3] = ACTIONS(596),
    [aux_sym_comparison_operator_token4] = ACTIONS(596),
    [aux_sym_comparison_operator_token5] = ACTIONS(596),
    [aux_sym_comparison_operator_token6] = ACTIONS(596),
    [aux_sym_comparison_operator_token7] = ACTIONS(596),
    [aux_sym_comparison_operator_token8] = ACTIONS(596),
    [aux_sym_comparison_operator_token9] = ACTIONS(596),
    [aux_sym_comparison_operator_token10] = ACTIONS(596),
    [aux_sym_comparison_operator_token11] = ACTIONS(596),
    [aux_sym_comparison_operator_token12] = ACTIONS(596),
    [aux_sym_comparison_operator_token13] = ACTIONS(596),
    [aux_sym_comparison_operator_token14] = ACTIONS(596),
    [aux_sym_comparison_operator_token15] = ACTIONS(596),
    [aux_sym_comparison_operator_token16] = ACTIONS(596),
    [aux_sym_comparison_operator_token17] = ACTIONS(596),
    [aux_sym_comparison_operator_token18] = ACTIONS(596),
    [aux_sym_comparison_operator_token19] = ACTIONS(596),
    [aux_sym_comparison_operator_token20] = ACTIONS(596),
    [aux_sym_comparison_operator_token21] = ACTIONS(596),
    [aux_sym_comparison_operator_token22] = ACTIONS(596),
    [aux_sym_comparison_operator_token23] = ACTIONS(596),
    [aux_sym_comparison_operator_token24] = ACTIONS(596),
    [aux_sym_comparison_operator_token25] = ACTIONS(596),
    [aux_sym_comparison_operator_token26] = ACTIONS(596),
    [aux_sym_comparison_operator_token27] = ACTIONS(596),
    [aux_sym_comparison_operator_token28] = ACTIONS(598),
    [aux_sym_comparison_operator_token29] = ACTIONS(596),
    [aux_sym_comparison_operator_token30] = ACTIONS(596),
    [aux_sym_comparison_operator_token31] = ACTIONS(596),
    [aux_sym_comparison_operator_token32] = ACTIONS(596),
    [aux_sym_comparison_operator_token33] = ACTIONS(596),
    [aux_sym_comparison_operator_token34] = ACTIONS(598),
    [aux_sym_comparison_operator_token35] = ACTIONS(596),
    [aux_sym_comparison_operator_token36] = ACTIONS(596),
    [aux_sym_comparison_operator_token37] = ACTIONS(596),
    [aux_sym_comparison_operator_token38] = ACTIONS(596),
    [aux_sym_comparison_operator_token39] = ACTIONS(596),
    [aux_sym_comparison_operator_token40] = ACTIONS(596),
    [aux_sym_comparison_operator_token41] = ACTIONS(596),
    [aux_sym_comparison_operator_token42] = ACTIONS(596),
    [aux_sym_comparison_operator_token43] = ACTIONS(596),
    [aux_sym_comparison_operator_token44] = ACTIONS(596),
    [aux_sym_comparison_operator_token45] = ACTIONS(596),
    [aux_sym_comparison_operator_token46] = ACTIONS(596),
    [aux_sym_comparison_operator_token47] = ACTIONS(596),
    [aux_sym_comparison_operator_token48] = ACTIONS(596),
    [aux_sym_comparison_operator_token49] = ACTIONS(596),
    [aux_sym_comparison_operator_token50] = ACTIONS(596),
    [aux_sym_format_operator_token1] = ACTIONS(596),
    [anon_sym_LPAREN] = ACTIONS(596),
    [anon_sym_COMMA] = ACTIONS(596),
    [anon_sym_AMP_AMP] = ACTIONS(596),
    [anon_sym_PIPE_PIPE] = ACTIONS(596),
    [anon_sym_PIPE] = ACTIONS(598),
    [aux_sym_logical_expression_token1] = ACTIONS(596),
    [aux_sym_logical_expression_token2] = ACTIONS(596),
    [aux_sym_logical_expression_token3] = ACTIONS(596),
    [aux_sym_bitwise_expression_token1] = ACTIONS(596),
    [aux_sym_bitwise_expression_token2] = ACTIONS(596),
    [aux_sym_bitwise_expression_token3] = ACTIONS(596),
    [anon_sym_PLUS] = ACTIONS(598),
    [anon_sym_DASH] = ACTIONS(598),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_BSLASH] = ACTIONS(596),
    [anon_sym_PERCENT] = ACTIONS(598),
    [anon_sym_STAR] = ACTIONS(598),
    [anon_sym_DOT_DOT] = ACTIONS(596),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_DOT2] = ACTIONS(598),
    [anon_sym_COLON_COLON] = ACTIONS(596),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(596),
    [sym__statement_terminator] = ACTIONS(596),
  },
  [STATE(130)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(658),
    [anon_sym_EQ] = ACTIONS(658),
    [anon_sym_BANG_EQ] = ACTIONS(658),
    [anon_sym_PLUS_EQ] = ACTIONS(658),
    [anon_sym_STAR_EQ] = ACTIONS(658),
    [anon_sym_SLASH_EQ] = ACTIONS(658),
    [anon_sym_PERCENT_EQ] = ACTIONS(658),
    [anon_sym_DASH_EQ] = ACTIONS(658),
    [anon_sym_GT] = ACTIONS(660),
    [anon_sym_GT_GT] = ACTIONS(658),
    [anon_sym_2_GT] = ACTIONS(660),
    [anon_sym_2_GT_GT] = ACTIONS(658),
    [anon_sym_3_GT] = ACTIONS(660),
    [anon_sym_3_GT_GT] = ACTIONS(658),
    [anon_sym_4_GT] = ACTIONS(660),
    [anon_sym_4_GT_GT] = ACTIONS(658),
    [anon_sym_5_GT] = ACTIONS(660),
    [anon_sym_5_GT_GT] = ACTIONS(658),
    [anon_sym_6_GT] = ACTIONS(660),
    [anon_sym_6_GT_GT] = ACTIONS(658),
    [anon_sym_STAR_GT] = ACTIONS(660),
    [anon_sym_STAR_GT_GT] = ACTIONS(658),
    [anon_sym_LT] = ACTIONS(660),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(658),
    [anon_sym_2_GT_AMP1] = ACTIONS(658),
    [anon_sym_3_GT_AMP1] = ACTIONS(658),
    [anon_sym_4_GT_AMP1] = ACTIONS(658),
    [anon_sym_5_GT_AMP1] = ACTIONS(658),
    [anon_sym_6_GT_AMP1] = ACTIONS(658),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(658),
    [anon_sym_1_GT_AMP2] = ACTIONS(658),
    [anon_sym_3_GT_AMP2] = ACTIONS(658),
    [anon_sym_4_GT_AMP2] = ACTIONS(658),
    [anon_sym_5_GT_AMP2] = ACTIONS(658),
    [anon_sym_6_GT_AMP2] = ACTIONS(658),
    [aux_sym_comparison_operator_token1] = ACTIONS(658),
    [aux_sym_comparison_operator_token2] = ACTIONS(658),
    [aux_sym_comparison_operator_token3] = ACTIONS(658),
    [aux_sym_comparison_operator_token4] = ACTIONS(658),
    [aux_sym_comparison_operator_token5] = ACTIONS(658),
    [aux_sym_comparison_operator_token6] = ACTIONS(658),
    [aux_sym_comparison_operator_token7] = ACTIONS(658),
    [aux_sym_comparison_operator_token8] = ACTIONS(658),
    [aux_sym_comparison_operator_token9] = ACTIONS(658),
    [aux_sym_comparison_operator_token10] = ACTIONS(658),
    [aux_sym_comparison_operator_token11] = ACTIONS(658),
    [aux_sym_comparison_operator_token12] = ACTIONS(658),
    [aux_sym_comparison_operator_token13] = ACTIONS(658),
    [aux_sym_comparison_operator_token14] = ACTIONS(658),
    [aux_sym_comparison_operator_token15] = ACTIONS(658),
    [aux_sym_comparison_operator_token16] = ACTIONS(658),
    [aux_sym_comparison_operator_token17] = ACTIONS(658),
    [aux_sym_comparison_operator_token18] = ACTIONS(658),
    [aux_sym_comparison_operator_token19] = ACTIONS(658),
    [aux_sym_comparison_operator_token20] = ACTIONS(658),
    [aux_sym_comparison_operator_token21] = ACTIONS(658),
    [aux_sym_comparison_operator_token22] = ACTIONS(658),
    [aux_sym_comparison_operator_token23] = ACTIONS(658),
    [aux_sym_comparison_operator_token24] = ACTIONS(658),
    [aux_sym_comparison_operator_token25] = ACTIONS(658),
    [aux_sym_comparison_operator_token26] = ACTIONS(658),
    [aux_sym_comparison_operator_token27] = ACTIONS(658),
    [aux_sym_comparison_operator_token28] = ACTIONS(660),
    [aux_sym_comparison_operator_token29] = ACTIONS(658),
    [aux_sym_comparison_operator_token30] = ACTIONS(658),
    [aux_sym_comparison_operator_token31] = ACTIONS(658),
    [aux_sym_comparison_operator_token32] = ACTIONS(658),
    [aux_sym_comparison_operator_token33] = ACTIONS(658),
    [aux_sym_comparison_operator_token34] = ACTIONS(660),
    [aux_sym_comparison_operator_token35] = ACTIONS(658),
    [aux_sym_comparison_operator_token36] = ACTIONS(658),
    [aux_sym_comparison_operator_token37] = ACTIONS(658),
    [aux_sym_comparison_operator_token38] = ACTIONS(658),
    [aux_sym_comparison_operator_token39] = ACTIONS(658),
    [aux_sym_comparison_operator_token40] = ACTIONS(658),
    [aux_sym_comparison_operator_token41] = ACTIONS(658),
    [aux_sym_comparison_operator_token42] = ACTIONS(658),
    [aux_sym_comparison_operator_token43] = ACTIONS(658),
    [aux_sym_comparison_operator_token44] = ACTIONS(658),
    [aux_sym_comparison_operator_token45] = ACTIONS(658),
    [aux_sym_comparison_operator_token46] = ACTIONS(658),
    [aux_sym_comparison_operator_token47] = ACTIONS(658),
    [aux_sym_comparison_operator_token48] = ACTIONS(658),
    [aux_sym_comparison_operator_token49] = ACTIONS(658),
    [aux_sym_comparison_operator_token50] = ACTIONS(658),
    [aux_sym_format_operator_token1] = ACTIONS(658),
    [anon_sym_LPAREN] = ACTIONS(658),
    [anon_sym_COMMA] = ACTIONS(658),
    [anon_sym_AMP_AMP] = ACTIONS(658),
    [anon_sym_PIPE_PIPE] = ACTIONS(658),
    [anon_sym_PIPE] = ACTIONS(660),
    [aux_sym_logical_expression_token1] = ACTIONS(658),
    [aux_sym_logical_expression_token2] = ACTIONS(658),
    [aux_sym_logical_expression_token3] = ACTIONS(658),
    [aux_sym_bitwise_expression_token1] = ACTIONS(658),
    [aux_sym_bitwise_expression_token2] = ACTIONS(658),
    [aux_sym_bitwise_expression_token3] = ACTIONS(658),
    [anon_sym_PLUS] = ACTIONS(660),
    [anon_sym_DASH] = ACTIONS(660),
    [anon_sym_SLASH] = ACTIONS(660),
    [anon_sym_BSLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(660),
    [anon_sym_STAR] = ACTIONS(660),
    [anon_sym_DOT_DOT] = ACTIONS(658),
    [anon_sym_PLUS_PLUS] = ACTIONS(658),
    [anon_sym_DASH_DASH] = ACTIONS(658),
    [anon_sym_DOT2] = ACTIONS(660),
    [anon_sym_COLON_COLON] = ACTIONS(658),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(658),
    [sym__statement_terminator] = ACTIONS(658),
  },
  [STATE(131)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(724),
    [anon_sym_EQ] = ACTIONS(724),
    [anon_sym_BANG_EQ] = ACTIONS(724),
    [anon_sym_PLUS_EQ] = ACTIONS(724),
    [anon_sym_STAR_EQ] = ACTIONS(724),
    [anon_sym_SLASH_EQ] = ACTIONS(724),
    [anon_sym_PERCENT_EQ] = ACTIONS(724),
    [anon_sym_DASH_EQ] = ACTIONS(724),
    [anon_sym_GT] = ACTIONS(726),
    [anon_sym_GT_GT] = ACTIONS(724),
    [anon_sym_2_GT] = ACTIONS(726),
    [anon_sym_2_GT_GT] = ACTIONS(724),
    [anon_sym_3_GT] = ACTIONS(726),
    [anon_sym_3_GT_GT] = ACTIONS(724),
    [anon_sym_4_GT] = ACTIONS(726),
    [anon_sym_4_GT_GT] = ACTIONS(724),
    [anon_sym_5_GT] = ACTIONS(726),
    [anon_sym_5_GT_GT] = ACTIONS(724),
    [anon_sym_6_GT] = ACTIONS(726),
    [anon_sym_6_GT_GT] = ACTIONS(724),
    [anon_sym_STAR_GT] = ACTIONS(726),
    [anon_sym_STAR_GT_GT] = ACTIONS(724),
    [anon_sym_LT] = ACTIONS(726),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(724),
    [anon_sym_2_GT_AMP1] = ACTIONS(724),
    [anon_sym_3_GT_AMP1] = ACTIONS(724),
    [anon_sym_4_GT_AMP1] = ACTIONS(724),
    [anon_sym_5_GT_AMP1] = ACTIONS(724),
    [anon_sym_6_GT_AMP1] = ACTIONS(724),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(724),
    [anon_sym_1_GT_AMP2] = ACTIONS(724),
    [anon_sym_3_GT_AMP2] = ACTIONS(724),
    [anon_sym_4_GT_AMP2] = ACTIONS(724),
    [anon_sym_5_GT_AMP2] = ACTIONS(724),
    [anon_sym_6_GT_AMP2] = ACTIONS(724),
    [aux_sym_comparison_operator_token1] = ACTIONS(724),
    [aux_sym_comparison_operator_token2] = ACTIONS(724),
    [aux_sym_comparison_operator_token3] = ACTIONS(724),
    [aux_sym_comparison_operator_token4] = ACTIONS(724),
    [aux_sym_comparison_operator_token5] = ACTIONS(724),
    [aux_sym_comparison_operator_token6] = ACTIONS(724),
    [aux_sym_comparison_operator_token7] = ACTIONS(724),
    [aux_sym_comparison_operator_token8] = ACTIONS(724),
    [aux_sym_comparison_operator_token9] = ACTIONS(724),
    [aux_sym_comparison_operator_token10] = ACTIONS(724),
    [aux_sym_comparison_operator_token11] = ACTIONS(724),
    [aux_sym_comparison_operator_token12] = ACTIONS(724),
    [aux_sym_comparison_operator_token13] = ACTIONS(724),
    [aux_sym_comparison_operator_token14] = ACTIONS(724),
    [aux_sym_comparison_operator_token15] = ACTIONS(724),
    [aux_sym_comparison_operator_token16] = ACTIONS(724),
    [aux_sym_comparison_operator_token17] = ACTIONS(724),
    [aux_sym_comparison_operator_token18] = ACTIONS(724),
    [aux_sym_comparison_operator_token19] = ACTIONS(724),
    [aux_sym_comparison_operator_token20] = ACTIONS(724),
    [aux_sym_comparison_operator_token21] = ACTIONS(724),
    [aux_sym_comparison_operator_token22] = ACTIONS(724),
    [aux_sym_comparison_operator_token23] = ACTIONS(724),
    [aux_sym_comparison_operator_token24] = ACTIONS(724),
    [aux_sym_comparison_operator_token25] = ACTIONS(724),
    [aux_sym_comparison_operator_token26] = ACTIONS(724),
    [aux_sym_comparison_operator_token27] = ACTIONS(724),
    [aux_sym_comparison_operator_token28] = ACTIONS(726),
    [aux_sym_comparison_operator_token29] = ACTIONS(724),
    [aux_sym_comparison_operator_token30] = ACTIONS(724),
    [aux_sym_comparison_operator_token31] = ACTIONS(724),
    [aux_sym_comparison_operator_token32] = ACTIONS(724),
    [aux_sym_comparison_operator_token33] = ACTIONS(724),
    [aux_sym_comparison_operator_token34] = ACTIONS(726),
    [aux_sym_comparison_operator_token35] = ACTIONS(724),
    [aux_sym_comparison_operator_token36] = ACTIONS(724),
    [aux_sym_comparison_operator_token37] = ACTIONS(724),
    [aux_sym_comparison_operator_token38] = ACTIONS(724),
    [aux_sym_comparison_operator_token39] = ACTIONS(724),
    [aux_sym_comparison_operator_token40] = ACTIONS(724),
    [aux_sym_comparison_operator_token41] = ACTIONS(724),
    [aux_sym_comparison_operator_token42] = ACTIONS(724),
    [aux_sym_comparison_operator_token43] = ACTIONS(724),
    [aux_sym_comparison_operator_token44] = ACTIONS(724),
    [aux_sym_comparison_operator_token45] = ACTIONS(724),
    [aux_sym_comparison_operator_token46] = ACTIONS(724),
    [aux_sym_comparison_operator_token47] = ACTIONS(724),
    [aux_sym_comparison_operator_token48] = ACTIONS(724),
    [aux_sym_comparison_operator_token49] = ACTIONS(724),
    [aux_sym_comparison_operator_token50] = ACTIONS(724),
    [aux_sym_format_operator_token1] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(724),
    [anon_sym_COMMA] = ACTIONS(724),
    [anon_sym_AMP_AMP] = ACTIONS(724),
    [anon_sym_PIPE_PIPE] = ACTIONS(724),
    [anon_sym_PIPE] = ACTIONS(726),
    [aux_sym_logical_expression_token1] = ACTIONS(724),
    [aux_sym_logical_expression_token2] = ACTIONS(724),
    [aux_sym_logical_expression_token3] = ACTIONS(724),
    [aux_sym_bitwise_expression_token1] = ACTIONS(724),
    [aux_sym_bitwise_expression_token2] = ACTIONS(724),
    [aux_sym_bitwise_expression_token3] = ACTIONS(724),
    [anon_sym_PLUS] = ACTIONS(726),
    [anon_sym_DASH] = ACTIONS(726),
    [anon_sym_SLASH] = ACTIONS(726),
    [anon_sym_BSLASH] = ACTIONS(724),
    [anon_sym_PERCENT] = ACTIONS(726),
    [anon_sym_STAR] = ACTIONS(726),
    [anon_sym_DOT_DOT] = ACTIONS(724),
    [anon_sym_PLUS_PLUS] = ACTIONS(724),
    [anon_sym_DASH_DASH] = ACTIONS(724),
    [anon_sym_DOT2] = ACTIONS(726),
    [anon_sym_COLON_COLON] = ACTIONS(724),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(724),
    [sym__statement_terminator] = ACTIONS(724),
  },
  [STATE(132)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(670),
    [anon_sym_EQ] = ACTIONS(670),
    [anon_sym_BANG_EQ] = ACTIONS(670),
    [anon_sym_PLUS_EQ] = ACTIONS(670),
    [anon_sym_STAR_EQ] = ACTIONS(670),
    [anon_sym_SLASH_EQ] = ACTIONS(670),
    [anon_sym_PERCENT_EQ] = ACTIONS(670),
    [anon_sym_DASH_EQ] = ACTIONS(670),
    [anon_sym_GT] = ACTIONS(672),
    [anon_sym_GT_GT] = ACTIONS(670),
    [anon_sym_2_GT] = ACTIONS(672),
    [anon_sym_2_GT_GT] = ACTIONS(670),
    [anon_sym_3_GT] = ACTIONS(672),
    [anon_sym_3_GT_GT] = ACTIONS(670),
    [anon_sym_4_GT] = ACTIONS(672),
    [anon_sym_4_GT_GT] = ACTIONS(670),
    [anon_sym_5_GT] = ACTIONS(672),
    [anon_sym_5_GT_GT] = ACTIONS(670),
    [anon_sym_6_GT] = ACTIONS(672),
    [anon_sym_6_GT_GT] = ACTIONS(670),
    [anon_sym_STAR_GT] = ACTIONS(672),
    [anon_sym_STAR_GT_GT] = ACTIONS(670),
    [anon_sym_LT] = ACTIONS(672),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(670),
    [anon_sym_2_GT_AMP1] = ACTIONS(670),
    [anon_sym_3_GT_AMP1] = ACTIONS(670),
    [anon_sym_4_GT_AMP1] = ACTIONS(670),
    [anon_sym_5_GT_AMP1] = ACTIONS(670),
    [anon_sym_6_GT_AMP1] = ACTIONS(670),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(670),
    [anon_sym_1_GT_AMP2] = ACTIONS(670),
    [anon_sym_3_GT_AMP2] = ACTIONS(670),
    [anon_sym_4_GT_AMP2] = ACTIONS(670),
    [anon_sym_5_GT_AMP2] = ACTIONS(670),
    [anon_sym_6_GT_AMP2] = ACTIONS(670),
    [aux_sym_comparison_operator_token1] = ACTIONS(670),
    [aux_sym_comparison_operator_token2] = ACTIONS(670),
    [aux_sym_comparison_operator_token3] = ACTIONS(670),
    [aux_sym_comparison_operator_token4] = ACTIONS(670),
    [aux_sym_comparison_operator_token5] = ACTIONS(670),
    [aux_sym_comparison_operator_token6] = ACTIONS(670),
    [aux_sym_comparison_operator_token7] = ACTIONS(670),
    [aux_sym_comparison_operator_token8] = ACTIONS(670),
    [aux_sym_comparison_operator_token9] = ACTIONS(670),
    [aux_sym_comparison_operator_token10] = ACTIONS(670),
    [aux_sym_comparison_operator_token11] = ACTIONS(670),
    [aux_sym_comparison_operator_token12] = ACTIONS(670),
    [aux_sym_comparison_operator_token13] = ACTIONS(670),
    [aux_sym_comparison_operator_token14] = ACTIONS(670),
    [aux_sym_comparison_operator_token15] = ACTIONS(670),
    [aux_sym_comparison_operator_token16] = ACTIONS(670),
    [aux_sym_comparison_operator_token17] = ACTIONS(670),
    [aux_sym_comparison_operator_token18] = ACTIONS(670),
    [aux_sym_comparison_operator_token19] = ACTIONS(670),
    [aux_sym_comparison_operator_token20] = ACTIONS(670),
    [aux_sym_comparison_operator_token21] = ACTIONS(670),
    [aux_sym_comparison_operator_token22] = ACTIONS(670),
    [aux_sym_comparison_operator_token23] = ACTIONS(670),
    [aux_sym_comparison_operator_token24] = ACTIONS(670),
    [aux_sym_comparison_operator_token25] = ACTIONS(670),
    [aux_sym_comparison_operator_token26] = ACTIONS(670),
    [aux_sym_comparison_operator_token27] = ACTIONS(670),
    [aux_sym_comparison_operator_token28] = ACTIONS(672),
    [aux_sym_comparison_operator_token29] = ACTIONS(670),
    [aux_sym_comparison_operator_token30] = ACTIONS(670),
    [aux_sym_comparison_operator_token31] = ACTIONS(670),
    [aux_sym_comparison_operator_token32] = ACTIONS(670),
    [aux_sym_comparison_operator_token33] = ACTIONS(670),
    [aux_sym_comparison_operator_token34] = ACTIONS(672),
    [aux_sym_comparison_operator_token35] = ACTIONS(670),
    [aux_sym_comparison_operator_token36] = ACTIONS(670),
    [aux_sym_comparison_operator_token37] = ACTIONS(670),
    [aux_sym_comparison_operator_token38] = ACTIONS(670),
    [aux_sym_comparison_operator_token39] = ACTIONS(670),
    [aux_sym_comparison_operator_token40] = ACTIONS(670),
    [aux_sym_comparison_operator_token41] = ACTIONS(670),
    [aux_sym_comparison_operator_token42] = ACTIONS(670),
    [aux_sym_comparison_operator_token43] = ACTIONS(670),
    [aux_sym_comparison_operator_token44] = ACTIONS(670),
    [aux_sym_comparison_operator_token45] = ACTIONS(670),
    [aux_sym_comparison_operator_token46] = ACTIONS(670),
    [aux_sym_comparison_operator_token47] = ACTIONS(670),
    [aux_sym_comparison_operator_token48] = ACTIONS(670),
    [aux_sym_comparison_operator_token49] = ACTIONS(670),
    [aux_sym_comparison_operator_token50] = ACTIONS(670),
    [aux_sym_format_operator_token1] = ACTIONS(670),
    [anon_sym_LPAREN] = ACTIONS(670),
    [anon_sym_COMMA] = ACTIONS(670),
    [anon_sym_AMP_AMP] = ACTIONS(670),
    [anon_sym_PIPE_PIPE] = ACTIONS(670),
    [anon_sym_PIPE] = ACTIONS(672),
    [aux_sym_logical_expression_token1] = ACTIONS(670),
    [aux_sym_logical_expression_token2] = ACTIONS(670),
    [aux_sym_logical_expression_token3] = ACTIONS(670),
    [aux_sym_bitwise_expression_token1] = ACTIONS(670),
    [aux_sym_bitwise_expression_token2] = ACTIONS(670),
    [aux_sym_bitwise_expression_token3] = ACTIONS(670),
    [anon_sym_PLUS] = ACTIONS(672),
    [anon_sym_DASH] = ACTIONS(672),
    [anon_sym_SLASH] = ACTIONS(672),
    [anon_sym_BSLASH] = ACTIONS(670),
    [anon_sym_PERCENT] = ACTIONS(672),
    [anon_sym_STAR] = ACTIONS(672),
    [anon_sym_DOT_DOT] = ACTIONS(670),
    [anon_sym_PLUS_PLUS] = ACTIONS(670),
    [anon_sym_DASH_DASH] = ACTIONS(670),
    [anon_sym_DOT2] = ACTIONS(672),
    [anon_sym_COLON_COLON] = ACTIONS(670),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(670),
    [sym__statement_terminator] = ACTIONS(670),
  },
  [STATE(133)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(674),
    [anon_sym_EQ] = ACTIONS(674),
    [anon_sym_BANG_EQ] = ACTIONS(674),
    [anon_sym_PLUS_EQ] = ACTIONS(674),
    [anon_sym_STAR_EQ] = ACTIONS(674),
    [anon_sym_SLASH_EQ] = ACTIONS(674),
    [anon_sym_PERCENT_EQ] = ACTIONS(674),
    [anon_sym_DASH_EQ] = ACTIONS(674),
    [anon_sym_GT] = ACTIONS(676),
    [anon_sym_GT_GT] = ACTIONS(674),
    [anon_sym_2_GT] = ACTIONS(676),
    [anon_sym_2_GT_GT] = ACTIONS(674),
    [anon_sym_3_GT] = ACTIONS(676),
    [anon_sym_3_GT_GT] = ACTIONS(674),
    [anon_sym_4_GT] = ACTIONS(676),
    [anon_sym_4_GT_GT] = ACTIONS(674),
    [anon_sym_5_GT] = ACTIONS(676),
    [anon_sym_5_GT_GT] = ACTIONS(674),
    [anon_sym_6_GT] = ACTIONS(676),
    [anon_sym_6_GT_GT] = ACTIONS(674),
    [anon_sym_STAR_GT] = ACTIONS(676),
    [anon_sym_STAR_GT_GT] = ACTIONS(674),
    [anon_sym_LT] = ACTIONS(676),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(674),
    [anon_sym_2_GT_AMP1] = ACTIONS(674),
    [anon_sym_3_GT_AMP1] = ACTIONS(674),
    [anon_sym_4_GT_AMP1] = ACTIONS(674),
    [anon_sym_5_GT_AMP1] = ACTIONS(674),
    [anon_sym_6_GT_AMP1] = ACTIONS(674),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(674),
    [anon_sym_1_GT_AMP2] = ACTIONS(674),
    [anon_sym_3_GT_AMP2] = ACTIONS(674),
    [anon_sym_4_GT_AMP2] = ACTIONS(674),
    [anon_sym_5_GT_AMP2] = ACTIONS(674),
    [anon_sym_6_GT_AMP2] = ACTIONS(674),
    [aux_sym_comparison_operator_token1] = ACTIONS(674),
    [aux_sym_comparison_operator_token2] = ACTIONS(674),
    [aux_sym_comparison_operator_token3] = ACTIONS(674),
    [aux_sym_comparison_operator_token4] = ACTIONS(674),
    [aux_sym_comparison_operator_token5] = ACTIONS(674),
    [aux_sym_comparison_operator_token6] = ACTIONS(674),
    [aux_sym_comparison_operator_token7] = ACTIONS(674),
    [aux_sym_comparison_operator_token8] = ACTIONS(674),
    [aux_sym_comparison_operator_token9] = ACTIONS(674),
    [aux_sym_comparison_operator_token10] = ACTIONS(674),
    [aux_sym_comparison_operator_token11] = ACTIONS(674),
    [aux_sym_comparison_operator_token12] = ACTIONS(674),
    [aux_sym_comparison_operator_token13] = ACTIONS(674),
    [aux_sym_comparison_operator_token14] = ACTIONS(674),
    [aux_sym_comparison_operator_token15] = ACTIONS(674),
    [aux_sym_comparison_operator_token16] = ACTIONS(674),
    [aux_sym_comparison_operator_token17] = ACTIONS(674),
    [aux_sym_comparison_operator_token18] = ACTIONS(674),
    [aux_sym_comparison_operator_token19] = ACTIONS(674),
    [aux_sym_comparison_operator_token20] = ACTIONS(674),
    [aux_sym_comparison_operator_token21] = ACTIONS(674),
    [aux_sym_comparison_operator_token22] = ACTIONS(674),
    [aux_sym_comparison_operator_token23] = ACTIONS(674),
    [aux_sym_comparison_operator_token24] = ACTIONS(674),
    [aux_sym_comparison_operator_token25] = ACTIONS(674),
    [aux_sym_comparison_operator_token26] = ACTIONS(674),
    [aux_sym_comparison_operator_token27] = ACTIONS(674),
    [aux_sym_comparison_operator_token28] = ACTIONS(676),
    [aux_sym_comparison_operator_token29] = ACTIONS(674),
    [aux_sym_comparison_operator_token30] = ACTIONS(674),
    [aux_sym_comparison_operator_token31] = ACTIONS(674),
    [aux_sym_comparison_operator_token32] = ACTIONS(674),
    [aux_sym_comparison_operator_token33] = ACTIONS(674),
    [aux_sym_comparison_operator_token34] = ACTIONS(676),
    [aux_sym_comparison_operator_token35] = ACTIONS(674),
    [aux_sym_comparison_operator_token36] = ACTIONS(674),
    [aux_sym_comparison_operator_token37] = ACTIONS(674),
    [aux_sym_comparison_operator_token38] = ACTIONS(674),
    [aux_sym_comparison_operator_token39] = ACTIONS(674),
    [aux_sym_comparison_operator_token40] = ACTIONS(674),
    [aux_sym_comparison_operator_token41] = ACTIONS(674),
    [aux_sym_comparison_operator_token42] = ACTIONS(674),
    [aux_sym_comparison_operator_token43] = ACTIONS(674),
    [aux_sym_comparison_operator_token44] = ACTIONS(674),
    [aux_sym_comparison_operator_token45] = ACTIONS(674),
    [aux_sym_comparison_operator_token46] = ACTIONS(674),
    [aux_sym_comparison_operator_token47] = ACTIONS(674),
    [aux_sym_comparison_operator_token48] = ACTIONS(674),
    [aux_sym_comparison_operator_token49] = ACTIONS(674),
    [aux_sym_comparison_operator_token50] = ACTIONS(674),
    [aux_sym_format_operator_token1] = ACTIONS(674),
    [anon_sym_LPAREN] = ACTIONS(674),
    [anon_sym_COMMA] = ACTIONS(674),
    [anon_sym_AMP_AMP] = ACTIONS(674),
    [anon_sym_PIPE_PIPE] = ACTIONS(674),
    [anon_sym_PIPE] = ACTIONS(676),
    [aux_sym_logical_expression_token1] = ACTIONS(674),
    [aux_sym_logical_expression_token2] = ACTIONS(674),
    [aux_sym_logical_expression_token3] = ACTIONS(674),
    [aux_sym_bitwise_expression_token1] = ACTIONS(674),
    [aux_sym_bitwise_expression_token2] = ACTIONS(674),
    [aux_sym_bitwise_expression_token3] = ACTIONS(674),
    [anon_sym_PLUS] = ACTIONS(676),
    [anon_sym_DASH] = ACTIONS(676),
    [anon_sym_SLASH] = ACTIONS(676),
    [anon_sym_BSLASH] = ACTIONS(674),
    [anon_sym_PERCENT] = ACTIONS(676),
    [anon_sym_STAR] = ACTIONS(676),
    [anon_sym_DOT_DOT] = ACTIONS(674),
    [anon_sym_PLUS_PLUS] = ACTIONS(674),
    [anon_sym_DASH_DASH] = ACTIONS(674),
    [anon_sym_DOT2] = ACTIONS(676),
    [anon_sym_COLON_COLON] = ACTIONS(674),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(674),
    [sym__statement_terminator] = ACTIONS(674),
  },
  [STATE(134)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(678),
    [anon_sym_EQ] = ACTIONS(678),
    [anon_sym_BANG_EQ] = ACTIONS(678),
    [anon_sym_PLUS_EQ] = ACTIONS(678),
    [anon_sym_STAR_EQ] = ACTIONS(678),
    [anon_sym_SLASH_EQ] = ACTIONS(678),
    [anon_sym_PERCENT_EQ] = ACTIONS(678),
    [anon_sym_DASH_EQ] = ACTIONS(678),
    [anon_sym_GT] = ACTIONS(680),
    [anon_sym_GT_GT] = ACTIONS(678),
    [anon_sym_2_GT] = ACTIONS(680),
    [anon_sym_2_GT_GT] = ACTIONS(678),
    [anon_sym_3_GT] = ACTIONS(680),
    [anon_sym_3_GT_GT] = ACTIONS(678),
    [anon_sym_4_GT] = ACTIONS(680),
    [anon_sym_4_GT_GT] = ACTIONS(678),
    [anon_sym_5_GT] = ACTIONS(680),
    [anon_sym_5_GT_GT] = ACTIONS(678),
    [anon_sym_6_GT] = ACTIONS(680),
    [anon_sym_6_GT_GT] = ACTIONS(678),
    [anon_sym_STAR_GT] = ACTIONS(680),
    [anon_sym_STAR_GT_GT] = ACTIONS(678),
    [anon_sym_LT] = ACTIONS(680),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(678),
    [anon_sym_2_GT_AMP1] = ACTIONS(678),
    [anon_sym_3_GT_AMP1] = ACTIONS(678),
    [anon_sym_4_GT_AMP1] = ACTIONS(678),
    [anon_sym_5_GT_AMP1] = ACTIONS(678),
    [anon_sym_6_GT_AMP1] = ACTIONS(678),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(678),
    [anon_sym_1_GT_AMP2] = ACTIONS(678),
    [anon_sym_3_GT_AMP2] = ACTIONS(678),
    [anon_sym_4_GT_AMP2] = ACTIONS(678),
    [anon_sym_5_GT_AMP2] = ACTIONS(678),
    [anon_sym_6_GT_AMP2] = ACTIONS(678),
    [aux_sym_comparison_operator_token1] = ACTIONS(678),
    [aux_sym_comparison_operator_token2] = ACTIONS(678),
    [aux_sym_comparison_operator_token3] = ACTIONS(678),
    [aux_sym_comparison_operator_token4] = ACTIONS(678),
    [aux_sym_comparison_operator_token5] = ACTIONS(678),
    [aux_sym_comparison_operator_token6] = ACTIONS(678),
    [aux_sym_comparison_operator_token7] = ACTIONS(678),
    [aux_sym_comparison_operator_token8] = ACTIONS(678),
    [aux_sym_comparison_operator_token9] = ACTIONS(678),
    [aux_sym_comparison_operator_token10] = ACTIONS(678),
    [aux_sym_comparison_operator_token11] = ACTIONS(678),
    [aux_sym_comparison_operator_token12] = ACTIONS(678),
    [aux_sym_comparison_operator_token13] = ACTIONS(678),
    [aux_sym_comparison_operator_token14] = ACTIONS(678),
    [aux_sym_comparison_operator_token15] = ACTIONS(678),
    [aux_sym_comparison_operator_token16] = ACTIONS(678),
    [aux_sym_comparison_operator_token17] = ACTIONS(678),
    [aux_sym_comparison_operator_token18] = ACTIONS(678),
    [aux_sym_comparison_operator_token19] = ACTIONS(678),
    [aux_sym_comparison_operator_token20] = ACTIONS(678),
    [aux_sym_comparison_operator_token21] = ACTIONS(678),
    [aux_sym_comparison_operator_token22] = ACTIONS(678),
    [aux_sym_comparison_operator_token23] = ACTIONS(678),
    [aux_sym_comparison_operator_token24] = ACTIONS(678),
    [aux_sym_comparison_operator_token25] = ACTIONS(678),
    [aux_sym_comparison_operator_token26] = ACTIONS(678),
    [aux_sym_comparison_operator_token27] = ACTIONS(678),
    [aux_sym_comparison_operator_token28] = ACTIONS(680),
    [aux_sym_comparison_operator_token29] = ACTIONS(678),
    [aux_sym_comparison_operator_token30] = ACTIONS(678),
    [aux_sym_comparison_operator_token31] = ACTIONS(678),
    [aux_sym_comparison_operator_token32] = ACTIONS(678),
    [aux_sym_comparison_operator_token33] = ACTIONS(678),
    [aux_sym_comparison_operator_token34] = ACTIONS(680),
    [aux_sym_comparison_operator_token35] = ACTIONS(678),
    [aux_sym_comparison_operator_token36] = ACTIONS(678),
    [aux_sym_comparison_operator_token37] = ACTIONS(678),
    [aux_sym_comparison_operator_token38] = ACTIONS(678),
    [aux_sym_comparison_operator_token39] = ACTIONS(678),
    [aux_sym_comparison_operator_token40] = ACTIONS(678),
    [aux_sym_comparison_operator_token41] = ACTIONS(678),
    [aux_sym_comparison_operator_token42] = ACTIONS(678),
    [aux_sym_comparison_operator_token43] = ACTIONS(678),
    [aux_sym_comparison_operator_token44] = ACTIONS(678),
    [aux_sym_comparison_operator_token45] = ACTIONS(678),
    [aux_sym_comparison_operator_token46] = ACTIONS(678),
    [aux_sym_comparison_operator_token47] = ACTIONS(678),
    [aux_sym_comparison_operator_token48] = ACTIONS(678),
    [aux_sym_comparison_operator_token49] = ACTIONS(678),
    [aux_sym_comparison_operator_token50] = ACTIONS(678),
    [aux_sym_format_operator_token1] = ACTIONS(678),
    [anon_sym_LPAREN] = ACTIONS(678),
    [anon_sym_COMMA] = ACTIONS(678),
    [anon_sym_AMP_AMP] = ACTIONS(678),
    [anon_sym_PIPE_PIPE] = ACTIONS(678),
    [anon_sym_PIPE] = ACTIONS(680),
    [aux_sym_logical_expression_token1] = ACTIONS(678),
    [aux_sym_logical_expression_token2] = ACTIONS(678),
    [aux_sym_logical_expression_token3] = ACTIONS(678),
    [aux_sym_bitwise_expression_token1] = ACTIONS(678),
    [aux_sym_bitwise_expression_token2] = ACTIONS(678),
    [aux_sym_bitwise_expression_token3] = ACTIONS(678),
    [anon_sym_PLUS] = ACTIONS(680),
    [anon_sym_DASH] = ACTIONS(680),
    [anon_sym_SLASH] = ACTIONS(680),
    [anon_sym_BSLASH] = ACTIONS(678),
    [anon_sym_PERCENT] = ACTIONS(680),
    [anon_sym_STAR] = ACTIONS(680),
    [anon_sym_DOT_DOT] = ACTIONS(678),
    [anon_sym_PLUS_PLUS] = ACTIONS(678),
    [anon_sym_DASH_DASH] = ACTIONS(678),
    [anon_sym_DOT2] = ACTIONS(680),
    [anon_sym_COLON_COLON] = ACTIONS(678),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(678),
    [sym__statement_terminator] = ACTIONS(678),
  },
  [STATE(135)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(654),
    [anon_sym_DASH_DASH] = ACTIONS(654),
    [anon_sym_DOT2] = ACTIONS(656),
    [anon_sym_COLON_COLON] = ACTIONS(654),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(654),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(136)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(600),
    [anon_sym_EQ] = ACTIONS(600),
    [anon_sym_BANG_EQ] = ACTIONS(600),
    [anon_sym_PLUS_EQ] = ACTIONS(600),
    [anon_sym_STAR_EQ] = ACTIONS(600),
    [anon_sym_SLASH_EQ] = ACTIONS(600),
    [anon_sym_PERCENT_EQ] = ACTIONS(600),
    [anon_sym_DASH_EQ] = ACTIONS(600),
    [anon_sym_GT] = ACTIONS(602),
    [anon_sym_GT_GT] = ACTIONS(600),
    [anon_sym_2_GT] = ACTIONS(602),
    [anon_sym_2_GT_GT] = ACTIONS(600),
    [anon_sym_3_GT] = ACTIONS(602),
    [anon_sym_3_GT_GT] = ACTIONS(600),
    [anon_sym_4_GT] = ACTIONS(602),
    [anon_sym_4_GT_GT] = ACTIONS(600),
    [anon_sym_5_GT] = ACTIONS(602),
    [anon_sym_5_GT_GT] = ACTIONS(600),
    [anon_sym_6_GT] = ACTIONS(602),
    [anon_sym_6_GT_GT] = ACTIONS(600),
    [anon_sym_STAR_GT] = ACTIONS(602),
    [anon_sym_STAR_GT_GT] = ACTIONS(600),
    [anon_sym_LT] = ACTIONS(602),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(600),
    [anon_sym_2_GT_AMP1] = ACTIONS(600),
    [anon_sym_3_GT_AMP1] = ACTIONS(600),
    [anon_sym_4_GT_AMP1] = ACTIONS(600),
    [anon_sym_5_GT_AMP1] = ACTIONS(600),
    [anon_sym_6_GT_AMP1] = ACTIONS(600),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(600),
    [anon_sym_1_GT_AMP2] = ACTIONS(600),
    [anon_sym_3_GT_AMP2] = ACTIONS(600),
    [anon_sym_4_GT_AMP2] = ACTIONS(600),
    [anon_sym_5_GT_AMP2] = ACTIONS(600),
    [anon_sym_6_GT_AMP2] = ACTIONS(600),
    [aux_sym_comparison_operator_token1] = ACTIONS(600),
    [aux_sym_comparison_operator_token2] = ACTIONS(600),
    [aux_sym_comparison_operator_token3] = ACTIONS(600),
    [aux_sym_comparison_operator_token4] = ACTIONS(600),
    [aux_sym_comparison_operator_token5] = ACTIONS(600),
    [aux_sym_comparison_operator_token6] = ACTIONS(600),
    [aux_sym_comparison_operator_token7] = ACTIONS(600),
    [aux_sym_comparison_operator_token8] = ACTIONS(600),
    [aux_sym_comparison_operator_token9] = ACTIONS(600),
    [aux_sym_comparison_operator_token10] = ACTIONS(600),
    [aux_sym_comparison_operator_token11] = ACTIONS(600),
    [aux_sym_comparison_operator_token12] = ACTIONS(600),
    [aux_sym_comparison_operator_token13] = ACTIONS(600),
    [aux_sym_comparison_operator_token14] = ACTIONS(600),
    [aux_sym_comparison_operator_token15] = ACTIONS(600),
    [aux_sym_comparison_operator_token16] = ACTIONS(600),
    [aux_sym_comparison_operator_token17] = ACTIONS(600),
    [aux_sym_comparison_operator_token18] = ACTIONS(600),
    [aux_sym_comparison_operator_token19] = ACTIONS(600),
    [aux_sym_comparison_operator_token20] = ACTIONS(600),
    [aux_sym_comparison_operator_token21] = ACTIONS(600),
    [aux_sym_comparison_operator_token22] = ACTIONS(600),
    [aux_sym_comparison_operator_token23] = ACTIONS(600),
    [aux_sym_comparison_operator_token24] = ACTIONS(600),
    [aux_sym_comparison_operator_token25] = ACTIONS(600),
    [aux_sym_comparison_operator_token26] = ACTIONS(600),
    [aux_sym_comparison_operator_token27] = ACTIONS(600),
    [aux_sym_comparison_operator_token28] = ACTIONS(602),
    [aux_sym_comparison_operator_token29] = ACTIONS(600),
    [aux_sym_comparison_operator_token30] = ACTIONS(600),
    [aux_sym_comparison_operator_token31] = ACTIONS(600),
    [aux_sym_comparison_operator_token32] = ACTIONS(600),
    [aux_sym_comparison_operator_token33] = ACTIONS(600),
    [aux_sym_comparison_operator_token34] = ACTIONS(602),
    [aux_sym_comparison_operator_token35] = ACTIONS(600),
    [aux_sym_comparison_operator_token36] = ACTIONS(600),
    [aux_sym_comparison_operator_token37] = ACTIONS(600),
    [aux_sym_comparison_operator_token38] = ACTIONS(600),
    [aux_sym_comparison_operator_token39] = ACTIONS(600),
    [aux_sym_comparison_operator_token40] = ACTIONS(600),
    [aux_sym_comparison_operator_token41] = ACTIONS(600),
    [aux_sym_comparison_operator_token42] = ACTIONS(600),
    [aux_sym_comparison_operator_token43] = ACTIONS(600),
    [aux_sym_comparison_operator_token44] = ACTIONS(600),
    [aux_sym_comparison_operator_token45] = ACTIONS(600),
    [aux_sym_comparison_operator_token46] = ACTIONS(600),
    [aux_sym_comparison_operator_token47] = ACTIONS(600),
    [aux_sym_comparison_operator_token48] = ACTIONS(600),
    [aux_sym_comparison_operator_token49] = ACTIONS(600),
    [aux_sym_comparison_operator_token50] = ACTIONS(600),
    [aux_sym_format_operator_token1] = ACTIONS(600),
    [anon_sym_LPAREN] = ACTIONS(600),
    [anon_sym_COMMA] = ACTIONS(600),
    [anon_sym_AMP_AMP] = ACTIONS(600),
    [anon_sym_PIPE_PIPE] = ACTIONS(600),
    [anon_sym_PIPE] = ACTIONS(602),
    [aux_sym_logical_expression_token1] = ACTIONS(600),
    [aux_sym_logical_expression_token2] = ACTIONS(600),
    [aux_sym_logical_expression_token3] = ACTIONS(600),
    [aux_sym_bitwise_expression_token1] = ACTIONS(600),
    [aux_sym_bitwise_expression_token2] = ACTIONS(600),
    [aux_sym_bitwise_expression_token3] = ACTIONS(600),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(602),
    [anon_sym_BSLASH] = ACTIONS(600),
    [anon_sym_PERCENT] = ACTIONS(602),
    [anon_sym_STAR] = ACTIONS(602),
    [anon_sym_DOT_DOT] = ACTIONS(600),
    [anon_sym_PLUS_PLUS] = ACTIONS(600),
    [anon_sym_DASH_DASH] = ACTIONS(600),
    [anon_sym_DOT2] = ACTIONS(602),
    [anon_sym_COLON_COLON] = ACTIONS(600),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(600),
    [sym__statement_terminator] = ACTIONS(600),
  },
  [STATE(137)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(604),
    [anon_sym_EQ] = ACTIONS(604),
    [anon_sym_BANG_EQ] = ACTIONS(604),
    [anon_sym_PLUS_EQ] = ACTIONS(604),
    [anon_sym_STAR_EQ] = ACTIONS(604),
    [anon_sym_SLASH_EQ] = ACTIONS(604),
    [anon_sym_PERCENT_EQ] = ACTIONS(604),
    [anon_sym_DASH_EQ] = ACTIONS(604),
    [anon_sym_GT] = ACTIONS(607),
    [anon_sym_GT_GT] = ACTIONS(604),
    [anon_sym_2_GT] = ACTIONS(607),
    [anon_sym_2_GT_GT] = ACTIONS(604),
    [anon_sym_3_GT] = ACTIONS(607),
    [anon_sym_3_GT_GT] = ACTIONS(604),
    [anon_sym_4_GT] = ACTIONS(607),
    [anon_sym_4_GT_GT] = ACTIONS(604),
    [anon_sym_5_GT] = ACTIONS(607),
    [anon_sym_5_GT_GT] = ACTIONS(604),
    [anon_sym_6_GT] = ACTIONS(607),
    [anon_sym_6_GT_GT] = ACTIONS(604),
    [anon_sym_STAR_GT] = ACTIONS(607),
    [anon_sym_STAR_GT_GT] = ACTIONS(604),
    [anon_sym_LT] = ACTIONS(607),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(604),
    [anon_sym_2_GT_AMP1] = ACTIONS(604),
    [anon_sym_3_GT_AMP1] = ACTIONS(604),
    [anon_sym_4_GT_AMP1] = ACTIONS(604),
    [anon_sym_5_GT_AMP1] = ACTIONS(604),
    [anon_sym_6_GT_AMP1] = ACTIONS(604),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(604),
    [anon_sym_1_GT_AMP2] = ACTIONS(604),
    [anon_sym_3_GT_AMP2] = ACTIONS(604),
    [anon_sym_4_GT_AMP2] = ACTIONS(604),
    [anon_sym_5_GT_AMP2] = ACTIONS(604),
    [anon_sym_6_GT_AMP2] = ACTIONS(604),
    [aux_sym_comparison_operator_token1] = ACTIONS(604),
    [aux_sym_comparison_operator_token2] = ACTIONS(604),
    [aux_sym_comparison_operator_token3] = ACTIONS(604),
    [aux_sym_comparison_operator_token4] = ACTIONS(604),
    [aux_sym_comparison_operator_token5] = ACTIONS(604),
    [aux_sym_comparison_operator_token6] = ACTIONS(604),
    [aux_sym_comparison_operator_token7] = ACTIONS(604),
    [aux_sym_comparison_operator_token8] = ACTIONS(604),
    [aux_sym_comparison_operator_token9] = ACTIONS(604),
    [aux_sym_comparison_operator_token10] = ACTIONS(604),
    [aux_sym_comparison_operator_token11] = ACTIONS(604),
    [aux_sym_comparison_operator_token12] = ACTIONS(604),
    [aux_sym_comparison_operator_token13] = ACTIONS(604),
    [aux_sym_comparison_operator_token14] = ACTIONS(604),
    [aux_sym_comparison_operator_token15] = ACTIONS(604),
    [aux_sym_comparison_operator_token16] = ACTIONS(604),
    [aux_sym_comparison_operator_token17] = ACTIONS(604),
    [aux_sym_comparison_operator_token18] = ACTIONS(604),
    [aux_sym_comparison_operator_token19] = ACTIONS(604),
    [aux_sym_comparison_operator_token20] = ACTIONS(604),
    [aux_sym_comparison_operator_token21] = ACTIONS(604),
    [aux_sym_comparison_operator_token22] = ACTIONS(604),
    [aux_sym_comparison_operator_token23] = ACTIONS(604),
    [aux_sym_comparison_operator_token24] = ACTIONS(604),
    [aux_sym_comparison_operator_token25] = ACTIONS(604),
    [aux_sym_comparison_operator_token26] = ACTIONS(604),
    [aux_sym_comparison_operator_token27] = ACTIONS(604),
    [aux_sym_comparison_operator_token28] = ACTIONS(607),
    [aux_sym_comparison_operator_token29] = ACTIONS(604),
    [aux_sym_comparison_operator_token30] = ACTIONS(604),
    [aux_sym_comparison_operator_token31] = ACTIONS(604),
    [aux_sym_comparison_operator_token32] = ACTIONS(604),
    [aux_sym_comparison_operator_token33] = ACTIONS(604),
    [aux_sym_comparison_operator_token34] = ACTIONS(607),
    [aux_sym_comparison_operator_token35] = ACTIONS(604),
    [aux_sym_comparison_operator_token36] = ACTIONS(604),
    [aux_sym_comparison_operator_token37] = ACTIONS(604),
    [aux_sym_comparison_operator_token38] = ACTIONS(604),
    [aux_sym_comparison_operator_token39] = ACTIONS(604),
    [aux_sym_comparison_operator_token40] = ACTIONS(604),
    [aux_sym_comparison_operator_token41] = ACTIONS(604),
    [aux_sym_comparison_operator_token42] = ACTIONS(604),
    [aux_sym_comparison_operator_token43] = ACTIONS(604),
    [aux_sym_comparison_operator_token44] = ACTIONS(604),
    [aux_sym_comparison_operator_token45] = ACTIONS(604),
    [aux_sym_comparison_operator_token46] = ACTIONS(604),
    [aux_sym_comparison_operator_token47] = ACTIONS(604),
    [aux_sym_comparison_operator_token48] = ACTIONS(604),
    [aux_sym_comparison_operator_token49] = ACTIONS(604),
    [aux_sym_comparison_operator_token50] = ACTIONS(604),
    [aux_sym_format_operator_token1] = ACTIONS(604),
    [anon_sym_LPAREN] = ACTIONS(604),
    [anon_sym_COMMA] = ACTIONS(604),
    [anon_sym_AMP_AMP] = ACTIONS(604),
    [anon_sym_PIPE_PIPE] = ACTIONS(604),
    [anon_sym_PIPE] = ACTIONS(607),
    [aux_sym_logical_expression_token1] = ACTIONS(604),
    [aux_sym_logical_expression_token2] = ACTIONS(604),
    [aux_sym_logical_expression_token3] = ACTIONS(604),
    [aux_sym_bitwise_expression_token1] = ACTIONS(604),
    [aux_sym_bitwise_expression_token2] = ACTIONS(604),
    [aux_sym_bitwise_expression_token3] = ACTIONS(604),
    [anon_sym_PLUS] = ACTIONS(607),
    [anon_sym_DASH] = ACTIONS(607),
    [anon_sym_SLASH] = ACTIONS(607),
    [anon_sym_BSLASH] = ACTIONS(604),
    [anon_sym_PERCENT] = ACTIONS(607),
    [anon_sym_STAR] = ACTIONS(607),
    [anon_sym_DOT_DOT] = ACTIONS(604),
    [anon_sym_PLUS_PLUS] = ACTIONS(604),
    [anon_sym_DASH_DASH] = ACTIONS(604),
    [anon_sym_DOT2] = ACTIONS(607),
    [anon_sym_COLON_COLON] = ACTIONS(604),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(604),
    [sym__statement_terminator] = ACTIONS(604),
  },
  [STATE(138)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(712),
    [anon_sym_EQ] = ACTIONS(712),
    [anon_sym_BANG_EQ] = ACTIONS(712),
    [anon_sym_PLUS_EQ] = ACTIONS(712),
    [anon_sym_STAR_EQ] = ACTIONS(712),
    [anon_sym_SLASH_EQ] = ACTIONS(712),
    [anon_sym_PERCENT_EQ] = ACTIONS(712),
    [anon_sym_DASH_EQ] = ACTIONS(712),
    [anon_sym_GT] = ACTIONS(714),
    [anon_sym_GT_GT] = ACTIONS(712),
    [anon_sym_2_GT] = ACTIONS(714),
    [anon_sym_2_GT_GT] = ACTIONS(712),
    [anon_sym_3_GT] = ACTIONS(714),
    [anon_sym_3_GT_GT] = ACTIONS(712),
    [anon_sym_4_GT] = ACTIONS(714),
    [anon_sym_4_GT_GT] = ACTIONS(712),
    [anon_sym_5_GT] = ACTIONS(714),
    [anon_sym_5_GT_GT] = ACTIONS(712),
    [anon_sym_6_GT] = ACTIONS(714),
    [anon_sym_6_GT_GT] = ACTIONS(712),
    [anon_sym_STAR_GT] = ACTIONS(714),
    [anon_sym_STAR_GT_GT] = ACTIONS(712),
    [anon_sym_LT] = ACTIONS(714),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(712),
    [anon_sym_2_GT_AMP1] = ACTIONS(712),
    [anon_sym_3_GT_AMP1] = ACTIONS(712),
    [anon_sym_4_GT_AMP1] = ACTIONS(712),
    [anon_sym_5_GT_AMP1] = ACTIONS(712),
    [anon_sym_6_GT_AMP1] = ACTIONS(712),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(712),
    [anon_sym_1_GT_AMP2] = ACTIONS(712),
    [anon_sym_3_GT_AMP2] = ACTIONS(712),
    [anon_sym_4_GT_AMP2] = ACTIONS(712),
    [anon_sym_5_GT_AMP2] = ACTIONS(712),
    [anon_sym_6_GT_AMP2] = ACTIONS(712),
    [aux_sym_comparison_operator_token1] = ACTIONS(712),
    [aux_sym_comparison_operator_token2] = ACTIONS(712),
    [aux_sym_comparison_operator_token3] = ACTIONS(712),
    [aux_sym_comparison_operator_token4] = ACTIONS(712),
    [aux_sym_comparison_operator_token5] = ACTIONS(712),
    [aux_sym_comparison_operator_token6] = ACTIONS(712),
    [aux_sym_comparison_operator_token7] = ACTIONS(712),
    [aux_sym_comparison_operator_token8] = ACTIONS(712),
    [aux_sym_comparison_operator_token9] = ACTIONS(712),
    [aux_sym_comparison_operator_token10] = ACTIONS(712),
    [aux_sym_comparison_operator_token11] = ACTIONS(712),
    [aux_sym_comparison_operator_token12] = ACTIONS(712),
    [aux_sym_comparison_operator_token13] = ACTIONS(712),
    [aux_sym_comparison_operator_token14] = ACTIONS(712),
    [aux_sym_comparison_operator_token15] = ACTIONS(712),
    [aux_sym_comparison_operator_token16] = ACTIONS(712),
    [aux_sym_comparison_operator_token17] = ACTIONS(712),
    [aux_sym_comparison_operator_token18] = ACTIONS(712),
    [aux_sym_comparison_operator_token19] = ACTIONS(712),
    [aux_sym_comparison_operator_token20] = ACTIONS(712),
    [aux_sym_comparison_operator_token21] = ACTIONS(712),
    [aux_sym_comparison_operator_token22] = ACTIONS(712),
    [aux_sym_comparison_operator_token23] = ACTIONS(712),
    [aux_sym_comparison_operator_token24] = ACTIONS(712),
    [aux_sym_comparison_operator_token25] = ACTIONS(712),
    [aux_sym_comparison_operator_token26] = ACTIONS(712),
    [aux_sym_comparison_operator_token27] = ACTIONS(712),
    [aux_sym_comparison_operator_token28] = ACTIONS(714),
    [aux_sym_comparison_operator_token29] = ACTIONS(712),
    [aux_sym_comparison_operator_token30] = ACTIONS(712),
    [aux_sym_comparison_operator_token31] = ACTIONS(712),
    [aux_sym_comparison_operator_token32] = ACTIONS(712),
    [aux_sym_comparison_operator_token33] = ACTIONS(712),
    [aux_sym_comparison_operator_token34] = ACTIONS(714),
    [aux_sym_comparison_operator_token35] = ACTIONS(712),
    [aux_sym_comparison_operator_token36] = ACTIONS(712),
    [aux_sym_comparison_operator_token37] = ACTIONS(712),
    [aux_sym_comparison_operator_token38] = ACTIONS(712),
    [aux_sym_comparison_operator_token39] = ACTIONS(712),
    [aux_sym_comparison_operator_token40] = ACTIONS(712),
    [aux_sym_comparison_operator_token41] = ACTIONS(712),
    [aux_sym_comparison_operator_token42] = ACTIONS(712),
    [aux_sym_comparison_operator_token43] = ACTIONS(712),
    [aux_sym_comparison_operator_token44] = ACTIONS(712),
    [aux_sym_comparison_operator_token45] = ACTIONS(712),
    [aux_sym_comparison_operator_token46] = ACTIONS(712),
    [aux_sym_comparison_operator_token47] = ACTIONS(712),
    [aux_sym_comparison_operator_token48] = ACTIONS(712),
    [aux_sym_comparison_operator_token49] = ACTIONS(712),
    [aux_sym_comparison_operator_token50] = ACTIONS(712),
    [aux_sym_format_operator_token1] = ACTIONS(712),
    [anon_sym_LPAREN] = ACTIONS(712),
    [anon_sym_COMMA] = ACTIONS(712),
    [anon_sym_AMP_AMP] = ACTIONS(712),
    [anon_sym_PIPE_PIPE] = ACTIONS(712),
    [anon_sym_PIPE] = ACTIONS(714),
    [aux_sym_logical_expression_token1] = ACTIONS(712),
    [aux_sym_logical_expression_token2] = ACTIONS(712),
    [aux_sym_logical_expression_token3] = ACTIONS(712),
    [aux_sym_bitwise_expression_token1] = ACTIONS(712),
    [aux_sym_bitwise_expression_token2] = ACTIONS(712),
    [aux_sym_bitwise_expression_token3] = ACTIONS(712),
    [anon_sym_PLUS] = ACTIONS(714),
    [anon_sym_DASH] = ACTIONS(714),
    [anon_sym_SLASH] = ACTIONS(714),
    [anon_sym_BSLASH] = ACTIONS(712),
    [anon_sym_PERCENT] = ACTIONS(714),
    [anon_sym_STAR] = ACTIONS(714),
    [anon_sym_DOT_DOT] = ACTIONS(712),
    [anon_sym_PLUS_PLUS] = ACTIONS(712),
    [anon_sym_DASH_DASH] = ACTIONS(712),
    [anon_sym_DOT2] = ACTIONS(714),
    [anon_sym_COLON_COLON] = ACTIONS(712),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(712),
    [sym__statement_terminator] = ACTIONS(712),
  },
  [STATE(139)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(662),
    [anon_sym_EQ] = ACTIONS(662),
    [anon_sym_BANG_EQ] = ACTIONS(662),
    [anon_sym_PLUS_EQ] = ACTIONS(662),
    [anon_sym_STAR_EQ] = ACTIONS(662),
    [anon_sym_SLASH_EQ] = ACTIONS(662),
    [anon_sym_PERCENT_EQ] = ACTIONS(662),
    [anon_sym_DASH_EQ] = ACTIONS(662),
    [anon_sym_GT] = ACTIONS(664),
    [anon_sym_GT_GT] = ACTIONS(662),
    [anon_sym_2_GT] = ACTIONS(664),
    [anon_sym_2_GT_GT] = ACTIONS(662),
    [anon_sym_3_GT] = ACTIONS(664),
    [anon_sym_3_GT_GT] = ACTIONS(662),
    [anon_sym_4_GT] = ACTIONS(664),
    [anon_sym_4_GT_GT] = ACTIONS(662),
    [anon_sym_5_GT] = ACTIONS(664),
    [anon_sym_5_GT_GT] = ACTIONS(662),
    [anon_sym_6_GT] = ACTIONS(664),
    [anon_sym_6_GT_GT] = ACTIONS(662),
    [anon_sym_STAR_GT] = ACTIONS(664),
    [anon_sym_STAR_GT_GT] = ACTIONS(662),
    [anon_sym_LT] = ACTIONS(664),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(662),
    [anon_sym_2_GT_AMP1] = ACTIONS(662),
    [anon_sym_3_GT_AMP1] = ACTIONS(662),
    [anon_sym_4_GT_AMP1] = ACTIONS(662),
    [anon_sym_5_GT_AMP1] = ACTIONS(662),
    [anon_sym_6_GT_AMP1] = ACTIONS(662),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(662),
    [anon_sym_1_GT_AMP2] = ACTIONS(662),
    [anon_sym_3_GT_AMP2] = ACTIONS(662),
    [anon_sym_4_GT_AMP2] = ACTIONS(662),
    [anon_sym_5_GT_AMP2] = ACTIONS(662),
    [anon_sym_6_GT_AMP2] = ACTIONS(662),
    [aux_sym_comparison_operator_token1] = ACTIONS(662),
    [aux_sym_comparison_operator_token2] = ACTIONS(662),
    [aux_sym_comparison_operator_token3] = ACTIONS(662),
    [aux_sym_comparison_operator_token4] = ACTIONS(662),
    [aux_sym_comparison_operator_token5] = ACTIONS(662),
    [aux_sym_comparison_operator_token6] = ACTIONS(662),
    [aux_sym_comparison_operator_token7] = ACTIONS(662),
    [aux_sym_comparison_operator_token8] = ACTIONS(662),
    [aux_sym_comparison_operator_token9] = ACTIONS(662),
    [aux_sym_comparison_operator_token10] = ACTIONS(662),
    [aux_sym_comparison_operator_token11] = ACTIONS(662),
    [aux_sym_comparison_operator_token12] = ACTIONS(662),
    [aux_sym_comparison_operator_token13] = ACTIONS(662),
    [aux_sym_comparison_operator_token14] = ACTIONS(662),
    [aux_sym_comparison_operator_token15] = ACTIONS(662),
    [aux_sym_comparison_operator_token16] = ACTIONS(662),
    [aux_sym_comparison_operator_token17] = ACTIONS(662),
    [aux_sym_comparison_operator_token18] = ACTIONS(662),
    [aux_sym_comparison_operator_token19] = ACTIONS(662),
    [aux_sym_comparison_operator_token20] = ACTIONS(662),
    [aux_sym_comparison_operator_token21] = ACTIONS(662),
    [aux_sym_comparison_operator_token22] = ACTIONS(662),
    [aux_sym_comparison_operator_token23] = ACTIONS(662),
    [aux_sym_comparison_operator_token24] = ACTIONS(662),
    [aux_sym_comparison_operator_token25] = ACTIONS(662),
    [aux_sym_comparison_operator_token26] = ACTIONS(662),
    [aux_sym_comparison_operator_token27] = ACTIONS(662),
    [aux_sym_comparison_operator_token28] = ACTIONS(664),
    [aux_sym_comparison_operator_token29] = ACTIONS(662),
    [aux_sym_comparison_operator_token30] = ACTIONS(662),
    [aux_sym_comparison_operator_token31] = ACTIONS(662),
    [aux_sym_comparison_operator_token32] = ACTIONS(662),
    [aux_sym_comparison_operator_token33] = ACTIONS(662),
    [aux_sym_comparison_operator_token34] = ACTIONS(664),
    [aux_sym_comparison_operator_token35] = ACTIONS(662),
    [aux_sym_comparison_operator_token36] = ACTIONS(662),
    [aux_sym_comparison_operator_token37] = ACTIONS(662),
    [aux_sym_comparison_operator_token38] = ACTIONS(662),
    [aux_sym_comparison_operator_token39] = ACTIONS(662),
    [aux_sym_comparison_operator_token40] = ACTIONS(662),
    [aux_sym_comparison_operator_token41] = ACTIONS(662),
    [aux_sym_comparison_operator_token42] = ACTIONS(662),
    [aux_sym_comparison_operator_token43] = ACTIONS(662),
    [aux_sym_comparison_operator_token44] = ACTIONS(662),
    [aux_sym_comparison_operator_token45] = ACTIONS(662),
    [aux_sym_comparison_operator_token46] = ACTIONS(662),
    [aux_sym_comparison_operator_token47] = ACTIONS(662),
    [aux_sym_comparison_operator_token48] = ACTIONS(662),
    [aux_sym_comparison_operator_token49] = ACTIONS(662),
    [aux_sym_comparison_operator_token50] = ACTIONS(662),
    [aux_sym_format_operator_token1] = ACTIONS(662),
    [anon_sym_LPAREN] = ACTIONS(662),
    [anon_sym_COMMA] = ACTIONS(662),
    [anon_sym_AMP_AMP] = ACTIONS(662),
    [anon_sym_PIPE_PIPE] = ACTIONS(662),
    [anon_sym_PIPE] = ACTIONS(664),
    [aux_sym_logical_expression_token1] = ACTIONS(662),
    [aux_sym_logical_expression_token2] = ACTIONS(662),
    [aux_sym_logical_expression_token3] = ACTIONS(662),
    [aux_sym_bitwise_expression_token1] = ACTIONS(662),
    [aux_sym_bitwise_expression_token2] = ACTIONS(662),
    [aux_sym_bitwise_expression_token3] = ACTIONS(662),
    [anon_sym_PLUS] = ACTIONS(664),
    [anon_sym_DASH] = ACTIONS(664),
    [anon_sym_SLASH] = ACTIONS(664),
    [anon_sym_BSLASH] = ACTIONS(662),
    [anon_sym_PERCENT] = ACTIONS(664),
    [anon_sym_STAR] = ACTIONS(664),
    [anon_sym_DOT_DOT] = ACTIONS(662),
    [anon_sym_PLUS_PLUS] = ACTIONS(662),
    [anon_sym_DASH_DASH] = ACTIONS(662),
    [anon_sym_DOT2] = ACTIONS(664),
    [anon_sym_COLON_COLON] = ACTIONS(662),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(662),
    [sym__statement_terminator] = ACTIONS(662),
  },
  [STATE(140)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(618),
    [anon_sym_EQ] = ACTIONS(618),
    [anon_sym_BANG_EQ] = ACTIONS(618),
    [anon_sym_PLUS_EQ] = ACTIONS(618),
    [anon_sym_STAR_EQ] = ACTIONS(618),
    [anon_sym_SLASH_EQ] = ACTIONS(618),
    [anon_sym_PERCENT_EQ] = ACTIONS(618),
    [anon_sym_DASH_EQ] = ACTIONS(618),
    [anon_sym_GT] = ACTIONS(620),
    [anon_sym_GT_GT] = ACTIONS(618),
    [anon_sym_2_GT] = ACTIONS(620),
    [anon_sym_2_GT_GT] = ACTIONS(618),
    [anon_sym_3_GT] = ACTIONS(620),
    [anon_sym_3_GT_GT] = ACTIONS(618),
    [anon_sym_4_GT] = ACTIONS(620),
    [anon_sym_4_GT_GT] = ACTIONS(618),
    [anon_sym_5_GT] = ACTIONS(620),
    [anon_sym_5_GT_GT] = ACTIONS(618),
    [anon_sym_6_GT] = ACTIONS(620),
    [anon_sym_6_GT_GT] = ACTIONS(618),
    [anon_sym_STAR_GT] = ACTIONS(620),
    [anon_sym_STAR_GT_GT] = ACTIONS(618),
    [anon_sym_LT] = ACTIONS(620),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(618),
    [anon_sym_2_GT_AMP1] = ACTIONS(618),
    [anon_sym_3_GT_AMP1] = ACTIONS(618),
    [anon_sym_4_GT_AMP1] = ACTIONS(618),
    [anon_sym_5_GT_AMP1] = ACTIONS(618),
    [anon_sym_6_GT_AMP1] = ACTIONS(618),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(618),
    [anon_sym_1_GT_AMP2] = ACTIONS(618),
    [anon_sym_3_GT_AMP2] = ACTIONS(618),
    [anon_sym_4_GT_AMP2] = ACTIONS(618),
    [anon_sym_5_GT_AMP2] = ACTIONS(618),
    [anon_sym_6_GT_AMP2] = ACTIONS(618),
    [aux_sym_comparison_operator_token1] = ACTIONS(618),
    [aux_sym_comparison_operator_token2] = ACTIONS(618),
    [aux_sym_comparison_operator_token3] = ACTIONS(618),
    [aux_sym_comparison_operator_token4] = ACTIONS(618),
    [aux_sym_comparison_operator_token5] = ACTIONS(618),
    [aux_sym_comparison_operator_token6] = ACTIONS(618),
    [aux_sym_comparison_operator_token7] = ACTIONS(618),
    [aux_sym_comparison_operator_token8] = ACTIONS(618),
    [aux_sym_comparison_operator_token9] = ACTIONS(618),
    [aux_sym_comparison_operator_token10] = ACTIONS(618),
    [aux_sym_comparison_operator_token11] = ACTIONS(618),
    [aux_sym_comparison_operator_token12] = ACTIONS(618),
    [aux_sym_comparison_operator_token13] = ACTIONS(618),
    [aux_sym_comparison_operator_token14] = ACTIONS(618),
    [aux_sym_comparison_operator_token15] = ACTIONS(618),
    [aux_sym_comparison_operator_token16] = ACTIONS(618),
    [aux_sym_comparison_operator_token17] = ACTIONS(618),
    [aux_sym_comparison_operator_token18] = ACTIONS(618),
    [aux_sym_comparison_operator_token19] = ACTIONS(618),
    [aux_sym_comparison_operator_token20] = ACTIONS(618),
    [aux_sym_comparison_operator_token21] = ACTIONS(618),
    [aux_sym_comparison_operator_token22] = ACTIONS(618),
    [aux_sym_comparison_operator_token23] = ACTIONS(618),
    [aux_sym_comparison_operator_token24] = ACTIONS(618),
    [aux_sym_comparison_operator_token25] = ACTIONS(618),
    [aux_sym_comparison_operator_token26] = ACTIONS(618),
    [aux_sym_comparison_operator_token27] = ACTIONS(618),
    [aux_sym_comparison_operator_token28] = ACTIONS(620),
    [aux_sym_comparison_operator_token29] = ACTIONS(618),
    [aux_sym_comparison_operator_token30] = ACTIONS(618),
    [aux_sym_comparison_operator_token31] = ACTIONS(618),
    [aux_sym_comparison_operator_token32] = ACTIONS(618),
    [aux_sym_comparison_operator_token33] = ACTIONS(618),
    [aux_sym_comparison_operator_token34] = ACTIONS(620),
    [aux_sym_comparison_operator_token35] = ACTIONS(618),
    [aux_sym_comparison_operator_token36] = ACTIONS(618),
    [aux_sym_comparison_operator_token37] = ACTIONS(618),
    [aux_sym_comparison_operator_token38] = ACTIONS(618),
    [aux_sym_comparison_operator_token39] = ACTIONS(618),
    [aux_sym_comparison_operator_token40] = ACTIONS(618),
    [aux_sym_comparison_operator_token41] = ACTIONS(618),
    [aux_sym_comparison_operator_token42] = ACTIONS(618),
    [aux_sym_comparison_operator_token43] = ACTIONS(618),
    [aux_sym_comparison_operator_token44] = ACTIONS(618),
    [aux_sym_comparison_operator_token45] = ACTIONS(618),
    [aux_sym_comparison_operator_token46] = ACTIONS(618),
    [aux_sym_comparison_operator_token47] = ACTIONS(618),
    [aux_sym_comparison_operator_token48] = ACTIONS(618),
    [aux_sym_comparison_operator_token49] = ACTIONS(618),
    [aux_sym_comparison_operator_token50] = ACTIONS(618),
    [aux_sym_format_operator_token1] = ACTIONS(618),
    [anon_sym_LPAREN] = ACTIONS(618),
    [anon_sym_COMMA] = ACTIONS(618),
    [anon_sym_AMP_AMP] = ACTIONS(618),
    [anon_sym_PIPE_PIPE] = ACTIONS(618),
    [anon_sym_PIPE] = ACTIONS(620),
    [aux_sym_logical_expression_token1] = ACTIONS(618),
    [aux_sym_logical_expression_token2] = ACTIONS(618),
    [aux_sym_logical_expression_token3] = ACTIONS(618),
    [aux_sym_bitwise_expression_token1] = ACTIONS(618),
    [aux_sym_bitwise_expression_token2] = ACTIONS(618),
    [aux_sym_bitwise_expression_token3] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_SLASH] = ACTIONS(620),
    [anon_sym_BSLASH] = ACTIONS(618),
    [anon_sym_PERCENT] = ACTIONS(620),
    [anon_sym_STAR] = ACTIONS(620),
    [anon_sym_DOT_DOT] = ACTIONS(618),
    [anon_sym_PLUS_PLUS] = ACTIONS(618),
    [anon_sym_DASH_DASH] = ACTIONS(618),
    [anon_sym_DOT2] = ACTIONS(620),
    [anon_sym_COLON_COLON] = ACTIONS(618),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(618),
    [sym__statement_terminator] = ACTIONS(618),
  },
  [STATE(141)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(622),
    [anon_sym_EQ] = ACTIONS(622),
    [anon_sym_BANG_EQ] = ACTIONS(622),
    [anon_sym_PLUS_EQ] = ACTIONS(622),
    [anon_sym_STAR_EQ] = ACTIONS(622),
    [anon_sym_SLASH_EQ] = ACTIONS(622),
    [anon_sym_PERCENT_EQ] = ACTIONS(622),
    [anon_sym_DASH_EQ] = ACTIONS(622),
    [anon_sym_GT] = ACTIONS(624),
    [anon_sym_GT_GT] = ACTIONS(622),
    [anon_sym_2_GT] = ACTIONS(624),
    [anon_sym_2_GT_GT] = ACTIONS(622),
    [anon_sym_3_GT] = ACTIONS(624),
    [anon_sym_3_GT_GT] = ACTIONS(622),
    [anon_sym_4_GT] = ACTIONS(624),
    [anon_sym_4_GT_GT] = ACTIONS(622),
    [anon_sym_5_GT] = ACTIONS(624),
    [anon_sym_5_GT_GT] = ACTIONS(622),
    [anon_sym_6_GT] = ACTIONS(624),
    [anon_sym_6_GT_GT] = ACTIONS(622),
    [anon_sym_STAR_GT] = ACTIONS(624),
    [anon_sym_STAR_GT_GT] = ACTIONS(622),
    [anon_sym_LT] = ACTIONS(624),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(622),
    [anon_sym_2_GT_AMP1] = ACTIONS(622),
    [anon_sym_3_GT_AMP1] = ACTIONS(622),
    [anon_sym_4_GT_AMP1] = ACTIONS(622),
    [anon_sym_5_GT_AMP1] = ACTIONS(622),
    [anon_sym_6_GT_AMP1] = ACTIONS(622),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(622),
    [anon_sym_1_GT_AMP2] = ACTIONS(622),
    [anon_sym_3_GT_AMP2] = ACTIONS(622),
    [anon_sym_4_GT_AMP2] = ACTIONS(622),
    [anon_sym_5_GT_AMP2] = ACTIONS(622),
    [anon_sym_6_GT_AMP2] = ACTIONS(622),
    [aux_sym_comparison_operator_token1] = ACTIONS(622),
    [aux_sym_comparison_operator_token2] = ACTIONS(622),
    [aux_sym_comparison_operator_token3] = ACTIONS(622),
    [aux_sym_comparison_operator_token4] = ACTIONS(622),
    [aux_sym_comparison_operator_token5] = ACTIONS(622),
    [aux_sym_comparison_operator_token6] = ACTIONS(622),
    [aux_sym_comparison_operator_token7] = ACTIONS(622),
    [aux_sym_comparison_operator_token8] = ACTIONS(622),
    [aux_sym_comparison_operator_token9] = ACTIONS(622),
    [aux_sym_comparison_operator_token10] = ACTIONS(622),
    [aux_sym_comparison_operator_token11] = ACTIONS(622),
    [aux_sym_comparison_operator_token12] = ACTIONS(622),
    [aux_sym_comparison_operator_token13] = ACTIONS(622),
    [aux_sym_comparison_operator_token14] = ACTIONS(622),
    [aux_sym_comparison_operator_token15] = ACTIONS(622),
    [aux_sym_comparison_operator_token16] = ACTIONS(622),
    [aux_sym_comparison_operator_token17] = ACTIONS(622),
    [aux_sym_comparison_operator_token18] = ACTIONS(622),
    [aux_sym_comparison_operator_token19] = ACTIONS(622),
    [aux_sym_comparison_operator_token20] = ACTIONS(622),
    [aux_sym_comparison_operator_token21] = ACTIONS(622),
    [aux_sym_comparison_operator_token22] = ACTIONS(622),
    [aux_sym_comparison_operator_token23] = ACTIONS(622),
    [aux_sym_comparison_operator_token24] = ACTIONS(622),
    [aux_sym_comparison_operator_token25] = ACTIONS(622),
    [aux_sym_comparison_operator_token26] = ACTIONS(622),
    [aux_sym_comparison_operator_token27] = ACTIONS(622),
    [aux_sym_comparison_operator_token28] = ACTIONS(624),
    [aux_sym_comparison_operator_token29] = ACTIONS(622),
    [aux_sym_comparison_operator_token30] = ACTIONS(622),
    [aux_sym_comparison_operator_token31] = ACTIONS(622),
    [aux_sym_comparison_operator_token32] = ACTIONS(622),
    [aux_sym_comparison_operator_token33] = ACTIONS(622),
    [aux_sym_comparison_operator_token34] = ACTIONS(624),
    [aux_sym_comparison_operator_token35] = ACTIONS(622),
    [aux_sym_comparison_operator_token36] = ACTIONS(622),
    [aux_sym_comparison_operator_token37] = ACTIONS(622),
    [aux_sym_comparison_operator_token38] = ACTIONS(622),
    [aux_sym_comparison_operator_token39] = ACTIONS(622),
    [aux_sym_comparison_operator_token40] = ACTIONS(622),
    [aux_sym_comparison_operator_token41] = ACTIONS(622),
    [aux_sym_comparison_operator_token42] = ACTIONS(622),
    [aux_sym_comparison_operator_token43] = ACTIONS(622),
    [aux_sym_comparison_operator_token44] = ACTIONS(622),
    [aux_sym_comparison_operator_token45] = ACTIONS(622),
    [aux_sym_comparison_operator_token46] = ACTIONS(622),
    [aux_sym_comparison_operator_token47] = ACTIONS(622),
    [aux_sym_comparison_operator_token48] = ACTIONS(622),
    [aux_sym_comparison_operator_token49] = ACTIONS(622),
    [aux_sym_comparison_operator_token50] = ACTIONS(622),
    [aux_sym_format_operator_token1] = ACTIONS(622),
    [anon_sym_LPAREN] = ACTIONS(622),
    [anon_sym_COMMA] = ACTIONS(622),
    [anon_sym_AMP_AMP] = ACTIONS(622),
    [anon_sym_PIPE_PIPE] = ACTIONS(622),
    [anon_sym_PIPE] = ACTIONS(624),
    [aux_sym_logical_expression_token1] = ACTIONS(622),
    [aux_sym_logical_expression_token2] = ACTIONS(622),
    [aux_sym_logical_expression_token3] = ACTIONS(622),
    [aux_sym_bitwise_expression_token1] = ACTIONS(622),
    [aux_sym_bitwise_expression_token2] = ACTIONS(622),
    [aux_sym_bitwise_expression_token3] = ACTIONS(622),
    [anon_sym_PLUS] = ACTIONS(624),
    [anon_sym_DASH] = ACTIONS(624),
    [anon_sym_SLASH] = ACTIONS(624),
    [anon_sym_BSLASH] = ACTIONS(622),
    [anon_sym_PERCENT] = ACTIONS(624),
    [anon_sym_STAR] = ACTIONS(624),
    [anon_sym_DOT_DOT] = ACTIONS(622),
    [anon_sym_PLUS_PLUS] = ACTIONS(622),
    [anon_sym_DASH_DASH] = ACTIONS(622),
    [anon_sym_DOT2] = ACTIONS(624),
    [anon_sym_COLON_COLON] = ACTIONS(622),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(622),
    [sym__statement_terminator] = ACTIONS(622),
  },
  [STATE(142)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(586),
    [anon_sym_EQ] = ACTIONS(586),
    [anon_sym_BANG_EQ] = ACTIONS(586),
    [anon_sym_PLUS_EQ] = ACTIONS(586),
    [anon_sym_STAR_EQ] = ACTIONS(586),
    [anon_sym_SLASH_EQ] = ACTIONS(586),
    [anon_sym_PERCENT_EQ] = ACTIONS(586),
    [anon_sym_DASH_EQ] = ACTIONS(586),
    [anon_sym_GT] = ACTIONS(588),
    [anon_sym_GT_GT] = ACTIONS(586),
    [anon_sym_2_GT] = ACTIONS(588),
    [anon_sym_2_GT_GT] = ACTIONS(586),
    [anon_sym_3_GT] = ACTIONS(588),
    [anon_sym_3_GT_GT] = ACTIONS(586),
    [anon_sym_4_GT] = ACTIONS(588),
    [anon_sym_4_GT_GT] = ACTIONS(586),
    [anon_sym_5_GT] = ACTIONS(588),
    [anon_sym_5_GT_GT] = ACTIONS(586),
    [anon_sym_6_GT] = ACTIONS(588),
    [anon_sym_6_GT_GT] = ACTIONS(586),
    [anon_sym_STAR_GT] = ACTIONS(588),
    [anon_sym_STAR_GT_GT] = ACTIONS(586),
    [anon_sym_LT] = ACTIONS(588),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(586),
    [anon_sym_2_GT_AMP1] = ACTIONS(586),
    [anon_sym_3_GT_AMP1] = ACTIONS(586),
    [anon_sym_4_GT_AMP1] = ACTIONS(586),
    [anon_sym_5_GT_AMP1] = ACTIONS(586),
    [anon_sym_6_GT_AMP1] = ACTIONS(586),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(586),
    [anon_sym_1_GT_AMP2] = ACTIONS(586),
    [anon_sym_3_GT_AMP2] = ACTIONS(586),
    [anon_sym_4_GT_AMP2] = ACTIONS(586),
    [anon_sym_5_GT_AMP2] = ACTIONS(586),
    [anon_sym_6_GT_AMP2] = ACTIONS(586),
    [aux_sym_comparison_operator_token1] = ACTIONS(586),
    [aux_sym_comparison_operator_token2] = ACTIONS(586),
    [aux_sym_comparison_operator_token3] = ACTIONS(586),
    [aux_sym_comparison_operator_token4] = ACTIONS(586),
    [aux_sym_comparison_operator_token5] = ACTIONS(586),
    [aux_sym_comparison_operator_token6] = ACTIONS(586),
    [aux_sym_comparison_operator_token7] = ACTIONS(586),
    [aux_sym_comparison_operator_token8] = ACTIONS(586),
    [aux_sym_comparison_operator_token9] = ACTIONS(586),
    [aux_sym_comparison_operator_token10] = ACTIONS(586),
    [aux_sym_comparison_operator_token11] = ACTIONS(586),
    [aux_sym_comparison_operator_token12] = ACTIONS(586),
    [aux_sym_comparison_operator_token13] = ACTIONS(586),
    [aux_sym_comparison_operator_token14] = ACTIONS(586),
    [aux_sym_comparison_operator_token15] = ACTIONS(586),
    [aux_sym_comparison_operator_token16] = ACTIONS(586),
    [aux_sym_comparison_operator_token17] = ACTIONS(586),
    [aux_sym_comparison_operator_token18] = ACTIONS(586),
    [aux_sym_comparison_operator_token19] = ACTIONS(586),
    [aux_sym_comparison_operator_token20] = ACTIONS(586),
    [aux_sym_comparison_operator_token21] = ACTIONS(586),
    [aux_sym_comparison_operator_token22] = ACTIONS(586),
    [aux_sym_comparison_operator_token23] = ACTIONS(586),
    [aux_sym_comparison_operator_token24] = ACTIONS(586),
    [aux_sym_comparison_operator_token25] = ACTIONS(586),
    [aux_sym_comparison_operator_token26] = ACTIONS(586),
    [aux_sym_comparison_operator_token27] = ACTIONS(586),
    [aux_sym_comparison_operator_token28] = ACTIONS(588),
    [aux_sym_comparison_operator_token29] = ACTIONS(586),
    [aux_sym_comparison_operator_token30] = ACTIONS(586),
    [aux_sym_comparison_operator_token31] = ACTIONS(586),
    [aux_sym_comparison_operator_token32] = ACTIONS(586),
    [aux_sym_comparison_operator_token33] = ACTIONS(586),
    [aux_sym_comparison_operator_token34] = ACTIONS(588),
    [aux_sym_comparison_operator_token35] = ACTIONS(586),
    [aux_sym_comparison_operator_token36] = ACTIONS(586),
    [aux_sym_comparison_operator_token37] = ACTIONS(586),
    [aux_sym_comparison_operator_token38] = ACTIONS(586),
    [aux_sym_comparison_operator_token39] = ACTIONS(586),
    [aux_sym_comparison_operator_token40] = ACTIONS(586),
    [aux_sym_comparison_operator_token41] = ACTIONS(586),
    [aux_sym_comparison_operator_token42] = ACTIONS(586),
    [aux_sym_comparison_operator_token43] = ACTIONS(586),
    [aux_sym_comparison_operator_token44] = ACTIONS(586),
    [aux_sym_comparison_operator_token45] = ACTIONS(586),
    [aux_sym_comparison_operator_token46] = ACTIONS(586),
    [aux_sym_comparison_operator_token47] = ACTIONS(586),
    [aux_sym_comparison_operator_token48] = ACTIONS(586),
    [aux_sym_comparison_operator_token49] = ACTIONS(586),
    [aux_sym_comparison_operator_token50] = ACTIONS(586),
    [aux_sym_format_operator_token1] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(586),
    [anon_sym_COMMA] = ACTIONS(586),
    [anon_sym_AMP_AMP] = ACTIONS(586),
    [anon_sym_PIPE_PIPE] = ACTIONS(586),
    [anon_sym_PIPE] = ACTIONS(588),
    [aux_sym_logical_expression_token1] = ACTIONS(586),
    [aux_sym_logical_expression_token2] = ACTIONS(586),
    [aux_sym_logical_expression_token3] = ACTIONS(586),
    [aux_sym_bitwise_expression_token1] = ACTIONS(586),
    [aux_sym_bitwise_expression_token2] = ACTIONS(586),
    [aux_sym_bitwise_expression_token3] = ACTIONS(586),
    [anon_sym_PLUS] = ACTIONS(588),
    [anon_sym_DASH] = ACTIONS(588),
    [anon_sym_SLASH] = ACTIONS(588),
    [anon_sym_BSLASH] = ACTIONS(586),
    [anon_sym_PERCENT] = ACTIONS(588),
    [anon_sym_STAR] = ACTIONS(588),
    [anon_sym_DOT_DOT] = ACTIONS(586),
    [anon_sym_PLUS_PLUS] = ACTIONS(586),
    [anon_sym_DASH_DASH] = ACTIONS(586),
    [anon_sym_DOT2] = ACTIONS(588),
    [anon_sym_COLON_COLON] = ACTIONS(586),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(586),
    [sym__statement_terminator] = ACTIONS(586),
  },
  [STATE(143)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(626),
    [anon_sym_EQ] = ACTIONS(626),
    [anon_sym_BANG_EQ] = ACTIONS(626),
    [anon_sym_PLUS_EQ] = ACTIONS(626),
    [anon_sym_STAR_EQ] = ACTIONS(626),
    [anon_sym_SLASH_EQ] = ACTIONS(626),
    [anon_sym_PERCENT_EQ] = ACTIONS(626),
    [anon_sym_DASH_EQ] = ACTIONS(626),
    [anon_sym_GT] = ACTIONS(628),
    [anon_sym_GT_GT] = ACTIONS(626),
    [anon_sym_2_GT] = ACTIONS(628),
    [anon_sym_2_GT_GT] = ACTIONS(626),
    [anon_sym_3_GT] = ACTIONS(628),
    [anon_sym_3_GT_GT] = ACTIONS(626),
    [anon_sym_4_GT] = ACTIONS(628),
    [anon_sym_4_GT_GT] = ACTIONS(626),
    [anon_sym_5_GT] = ACTIONS(628),
    [anon_sym_5_GT_GT] = ACTIONS(626),
    [anon_sym_6_GT] = ACTIONS(628),
    [anon_sym_6_GT_GT] = ACTIONS(626),
    [anon_sym_STAR_GT] = ACTIONS(628),
    [anon_sym_STAR_GT_GT] = ACTIONS(626),
    [anon_sym_LT] = ACTIONS(628),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(626),
    [anon_sym_2_GT_AMP1] = ACTIONS(626),
    [anon_sym_3_GT_AMP1] = ACTIONS(626),
    [anon_sym_4_GT_AMP1] = ACTIONS(626),
    [anon_sym_5_GT_AMP1] = ACTIONS(626),
    [anon_sym_6_GT_AMP1] = ACTIONS(626),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(626),
    [anon_sym_1_GT_AMP2] = ACTIONS(626),
    [anon_sym_3_GT_AMP2] = ACTIONS(626),
    [anon_sym_4_GT_AMP2] = ACTIONS(626),
    [anon_sym_5_GT_AMP2] = ACTIONS(626),
    [anon_sym_6_GT_AMP2] = ACTIONS(626),
    [aux_sym_comparison_operator_token1] = ACTIONS(626),
    [aux_sym_comparison_operator_token2] = ACTIONS(626),
    [aux_sym_comparison_operator_token3] = ACTIONS(626),
    [aux_sym_comparison_operator_token4] = ACTIONS(626),
    [aux_sym_comparison_operator_token5] = ACTIONS(626),
    [aux_sym_comparison_operator_token6] = ACTIONS(626),
    [aux_sym_comparison_operator_token7] = ACTIONS(626),
    [aux_sym_comparison_operator_token8] = ACTIONS(626),
    [aux_sym_comparison_operator_token9] = ACTIONS(626),
    [aux_sym_comparison_operator_token10] = ACTIONS(626),
    [aux_sym_comparison_operator_token11] = ACTIONS(626),
    [aux_sym_comparison_operator_token12] = ACTIONS(626),
    [aux_sym_comparison_operator_token13] = ACTIONS(626),
    [aux_sym_comparison_operator_token14] = ACTIONS(626),
    [aux_sym_comparison_operator_token15] = ACTIONS(626),
    [aux_sym_comparison_operator_token16] = ACTIONS(626),
    [aux_sym_comparison_operator_token17] = ACTIONS(626),
    [aux_sym_comparison_operator_token18] = ACTIONS(626),
    [aux_sym_comparison_operator_token19] = ACTIONS(626),
    [aux_sym_comparison_operator_token20] = ACTIONS(626),
    [aux_sym_comparison_operator_token21] = ACTIONS(626),
    [aux_sym_comparison_operator_token22] = ACTIONS(626),
    [aux_sym_comparison_operator_token23] = ACTIONS(626),
    [aux_sym_comparison_operator_token24] = ACTIONS(626),
    [aux_sym_comparison_operator_token25] = ACTIONS(626),
    [aux_sym_comparison_operator_token26] = ACTIONS(626),
    [aux_sym_comparison_operator_token27] = ACTIONS(626),
    [aux_sym_comparison_operator_token28] = ACTIONS(628),
    [aux_sym_comparison_operator_token29] = ACTIONS(626),
    [aux_sym_comparison_operator_token30] = ACTIONS(626),
    [aux_sym_comparison_operator_token31] = ACTIONS(626),
    [aux_sym_comparison_operator_token32] = ACTIONS(626),
    [aux_sym_comparison_operator_token33] = ACTIONS(626),
    [aux_sym_comparison_operator_token34] = ACTIONS(628),
    [aux_sym_comparison_operator_token35] = ACTIONS(626),
    [aux_sym_comparison_operator_token36] = ACTIONS(626),
    [aux_sym_comparison_operator_token37] = ACTIONS(626),
    [aux_sym_comparison_operator_token38] = ACTIONS(626),
    [aux_sym_comparison_operator_token39] = ACTIONS(626),
    [aux_sym_comparison_operator_token40] = ACTIONS(626),
    [aux_sym_comparison_operator_token41] = ACTIONS(626),
    [aux_sym_comparison_operator_token42] = ACTIONS(626),
    [aux_sym_comparison_operator_token43] = ACTIONS(626),
    [aux_sym_comparison_operator_token44] = ACTIONS(626),
    [aux_sym_comparison_operator_token45] = ACTIONS(626),
    [aux_sym_comparison_operator_token46] = ACTIONS(626),
    [aux_sym_comparison_operator_token47] = ACTIONS(626),
    [aux_sym_comparison_operator_token48] = ACTIONS(626),
    [aux_sym_comparison_operator_token49] = ACTIONS(626),
    [aux_sym_comparison_operator_token50] = ACTIONS(626),
    [aux_sym_format_operator_token1] = ACTIONS(626),
    [anon_sym_LPAREN] = ACTIONS(626),
    [anon_sym_COMMA] = ACTIONS(626),
    [anon_sym_AMP_AMP] = ACTIONS(626),
    [anon_sym_PIPE_PIPE] = ACTIONS(626),
    [anon_sym_PIPE] = ACTIONS(628),
    [aux_sym_logical_expression_token1] = ACTIONS(626),
    [aux_sym_logical_expression_token2] = ACTIONS(626),
    [aux_sym_logical_expression_token3] = ACTIONS(626),
    [aux_sym_bitwise_expression_token1] = ACTIONS(626),
    [aux_sym_bitwise_expression_token2] = ACTIONS(626),
    [aux_sym_bitwise_expression_token3] = ACTIONS(626),
    [anon_sym_PLUS] = ACTIONS(628),
    [anon_sym_DASH] = ACTIONS(628),
    [anon_sym_SLASH] = ACTIONS(628),
    [anon_sym_BSLASH] = ACTIONS(626),
    [anon_sym_PERCENT] = ACTIONS(628),
    [anon_sym_STAR] = ACTIONS(628),
    [anon_sym_DOT_DOT] = ACTIONS(626),
    [anon_sym_PLUS_PLUS] = ACTIONS(626),
    [anon_sym_DASH_DASH] = ACTIONS(626),
    [anon_sym_DOT2] = ACTIONS(628),
    [anon_sym_COLON_COLON] = ACTIONS(626),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(626),
    [sym__statement_terminator] = ACTIONS(626),
  },
  [STATE(144)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(630),
    [anon_sym_EQ] = ACTIONS(630),
    [anon_sym_BANG_EQ] = ACTIONS(630),
    [anon_sym_PLUS_EQ] = ACTIONS(630),
    [anon_sym_STAR_EQ] = ACTIONS(630),
    [anon_sym_SLASH_EQ] = ACTIONS(630),
    [anon_sym_PERCENT_EQ] = ACTIONS(630),
    [anon_sym_DASH_EQ] = ACTIONS(630),
    [anon_sym_GT] = ACTIONS(632),
    [anon_sym_GT_GT] = ACTIONS(630),
    [anon_sym_2_GT] = ACTIONS(632),
    [anon_sym_2_GT_GT] = ACTIONS(630),
    [anon_sym_3_GT] = ACTIONS(632),
    [anon_sym_3_GT_GT] = ACTIONS(630),
    [anon_sym_4_GT] = ACTIONS(632),
    [anon_sym_4_GT_GT] = ACTIONS(630),
    [anon_sym_5_GT] = ACTIONS(632),
    [anon_sym_5_GT_GT] = ACTIONS(630),
    [anon_sym_6_GT] = ACTIONS(632),
    [anon_sym_6_GT_GT] = ACTIONS(630),
    [anon_sym_STAR_GT] = ACTIONS(632),
    [anon_sym_STAR_GT_GT] = ACTIONS(630),
    [anon_sym_LT] = ACTIONS(632),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(630),
    [anon_sym_2_GT_AMP1] = ACTIONS(630),
    [anon_sym_3_GT_AMP1] = ACTIONS(630),
    [anon_sym_4_GT_AMP1] = ACTIONS(630),
    [anon_sym_5_GT_AMP1] = ACTIONS(630),
    [anon_sym_6_GT_AMP1] = ACTIONS(630),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(630),
    [anon_sym_1_GT_AMP2] = ACTIONS(630),
    [anon_sym_3_GT_AMP2] = ACTIONS(630),
    [anon_sym_4_GT_AMP2] = ACTIONS(630),
    [anon_sym_5_GT_AMP2] = ACTIONS(630),
    [anon_sym_6_GT_AMP2] = ACTIONS(630),
    [aux_sym_comparison_operator_token1] = ACTIONS(630),
    [aux_sym_comparison_operator_token2] = ACTIONS(630),
    [aux_sym_comparison_operator_token3] = ACTIONS(630),
    [aux_sym_comparison_operator_token4] = ACTIONS(630),
    [aux_sym_comparison_operator_token5] = ACTIONS(630),
    [aux_sym_comparison_operator_token6] = ACTIONS(630),
    [aux_sym_comparison_operator_token7] = ACTIONS(630),
    [aux_sym_comparison_operator_token8] = ACTIONS(630),
    [aux_sym_comparison_operator_token9] = ACTIONS(630),
    [aux_sym_comparison_operator_token10] = ACTIONS(630),
    [aux_sym_comparison_operator_token11] = ACTIONS(630),
    [aux_sym_comparison_operator_token12] = ACTIONS(630),
    [aux_sym_comparison_operator_token13] = ACTIONS(630),
    [aux_sym_comparison_operator_token14] = ACTIONS(630),
    [aux_sym_comparison_operator_token15] = ACTIONS(630),
    [aux_sym_comparison_operator_token16] = ACTIONS(630),
    [aux_sym_comparison_operator_token17] = ACTIONS(630),
    [aux_sym_comparison_operator_token18] = ACTIONS(630),
    [aux_sym_comparison_operator_token19] = ACTIONS(630),
    [aux_sym_comparison_operator_token20] = ACTIONS(630),
    [aux_sym_comparison_operator_token21] = ACTIONS(630),
    [aux_sym_comparison_operator_token22] = ACTIONS(630),
    [aux_sym_comparison_operator_token23] = ACTIONS(630),
    [aux_sym_comparison_operator_token24] = ACTIONS(630),
    [aux_sym_comparison_operator_token25] = ACTIONS(630),
    [aux_sym_comparison_operator_token26] = ACTIONS(630),
    [aux_sym_comparison_operator_token27] = ACTIONS(630),
    [aux_sym_comparison_operator_token28] = ACTIONS(632),
    [aux_sym_comparison_operator_token29] = ACTIONS(630),
    [aux_sym_comparison_operator_token30] = ACTIONS(630),
    [aux_sym_comparison_operator_token31] = ACTIONS(630),
    [aux_sym_comparison_operator_token32] = ACTIONS(630),
    [aux_sym_comparison_operator_token33] = ACTIONS(630),
    [aux_sym_comparison_operator_token34] = ACTIONS(632),
    [aux_sym_comparison_operator_token35] = ACTIONS(630),
    [aux_sym_comparison_operator_token36] = ACTIONS(630),
    [aux_sym_comparison_operator_token37] = ACTIONS(630),
    [aux_sym_comparison_operator_token38] = ACTIONS(630),
    [aux_sym_comparison_operator_token39] = ACTIONS(630),
    [aux_sym_comparison_operator_token40] = ACTIONS(630),
    [aux_sym_comparison_operator_token41] = ACTIONS(630),
    [aux_sym_comparison_operator_token42] = ACTIONS(630),
    [aux_sym_comparison_operator_token43] = ACTIONS(630),
    [aux_sym_comparison_operator_token44] = ACTIONS(630),
    [aux_sym_comparison_operator_token45] = ACTIONS(630),
    [aux_sym_comparison_operator_token46] = ACTIONS(630),
    [aux_sym_comparison_operator_token47] = ACTIONS(630),
    [aux_sym_comparison_operator_token48] = ACTIONS(630),
    [aux_sym_comparison_operator_token49] = ACTIONS(630),
    [aux_sym_comparison_operator_token50] = ACTIONS(630),
    [aux_sym_format_operator_token1] = ACTIONS(630),
    [anon_sym_LPAREN] = ACTIONS(630),
    [anon_sym_COMMA] = ACTIONS(630),
    [anon_sym_AMP_AMP] = ACTIONS(630),
    [anon_sym_PIPE_PIPE] = ACTIONS(630),
    [anon_sym_PIPE] = ACTIONS(632),
    [aux_sym_logical_expression_token1] = ACTIONS(630),
    [aux_sym_logical_expression_token2] = ACTIONS(630),
    [aux_sym_logical_expression_token3] = ACTIONS(630),
    [aux_sym_bitwise_expression_token1] = ACTIONS(630),
    [aux_sym_bitwise_expression_token2] = ACTIONS(630),
    [aux_sym_bitwise_expression_token3] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(632),
    [anon_sym_DASH] = ACTIONS(632),
    [anon_sym_SLASH] = ACTIONS(632),
    [anon_sym_BSLASH] = ACTIONS(630),
    [anon_sym_PERCENT] = ACTIONS(632),
    [anon_sym_STAR] = ACTIONS(632),
    [anon_sym_DOT_DOT] = ACTIONS(630),
    [anon_sym_PLUS_PLUS] = ACTIONS(630),
    [anon_sym_DASH_DASH] = ACTIONS(630),
    [anon_sym_DOT2] = ACTIONS(632),
    [anon_sym_COLON_COLON] = ACTIONS(630),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(630),
    [sym__statement_terminator] = ACTIONS(630),
  },
  [STATE(145)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_BANG_EQ] = ACTIONS(650),
    [anon_sym_PLUS_EQ] = ACTIONS(650),
    [anon_sym_STAR_EQ] = ACTIONS(650),
    [anon_sym_SLASH_EQ] = ACTIONS(650),
    [anon_sym_PERCENT_EQ] = ACTIONS(650),
    [anon_sym_DASH_EQ] = ACTIONS(650),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_GT_GT] = ACTIONS(650),
    [anon_sym_2_GT] = ACTIONS(652),
    [anon_sym_2_GT_GT] = ACTIONS(650),
    [anon_sym_3_GT] = ACTIONS(652),
    [anon_sym_3_GT_GT] = ACTIONS(650),
    [anon_sym_4_GT] = ACTIONS(652),
    [anon_sym_4_GT_GT] = ACTIONS(650),
    [anon_sym_5_GT] = ACTIONS(652),
    [anon_sym_5_GT_GT] = ACTIONS(650),
    [anon_sym_6_GT] = ACTIONS(652),
    [anon_sym_6_GT_GT] = ACTIONS(650),
    [anon_sym_STAR_GT] = ACTIONS(652),
    [anon_sym_STAR_GT_GT] = ACTIONS(650),
    [anon_sym_LT] = ACTIONS(652),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(650),
    [anon_sym_2_GT_AMP1] = ACTIONS(650),
    [anon_sym_3_GT_AMP1] = ACTIONS(650),
    [anon_sym_4_GT_AMP1] = ACTIONS(650),
    [anon_sym_5_GT_AMP1] = ACTIONS(650),
    [anon_sym_6_GT_AMP1] = ACTIONS(650),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(650),
    [anon_sym_1_GT_AMP2] = ACTIONS(650),
    [anon_sym_3_GT_AMP2] = ACTIONS(650),
    [anon_sym_4_GT_AMP2] = ACTIONS(650),
    [anon_sym_5_GT_AMP2] = ACTIONS(650),
    [anon_sym_6_GT_AMP2] = ACTIONS(650),
    [aux_sym_comparison_operator_token1] = ACTIONS(650),
    [aux_sym_comparison_operator_token2] = ACTIONS(650),
    [aux_sym_comparison_operator_token3] = ACTIONS(650),
    [aux_sym_comparison_operator_token4] = ACTIONS(650),
    [aux_sym_comparison_operator_token5] = ACTIONS(650),
    [aux_sym_comparison_operator_token6] = ACTIONS(650),
    [aux_sym_comparison_operator_token7] = ACTIONS(650),
    [aux_sym_comparison_operator_token8] = ACTIONS(650),
    [aux_sym_comparison_operator_token9] = ACTIONS(650),
    [aux_sym_comparison_operator_token10] = ACTIONS(650),
    [aux_sym_comparison_operator_token11] = ACTIONS(650),
    [aux_sym_comparison_operator_token12] = ACTIONS(650),
    [aux_sym_comparison_operator_token13] = ACTIONS(650),
    [aux_sym_comparison_operator_token14] = ACTIONS(650),
    [aux_sym_comparison_operator_token15] = ACTIONS(650),
    [aux_sym_comparison_operator_token16] = ACTIONS(650),
    [aux_sym_comparison_operator_token17] = ACTIONS(650),
    [aux_sym_comparison_operator_token18] = ACTIONS(650),
    [aux_sym_comparison_operator_token19] = ACTIONS(650),
    [aux_sym_comparison_operator_token20] = ACTIONS(650),
    [aux_sym_comparison_operator_token21] = ACTIONS(650),
    [aux_sym_comparison_operator_token22] = ACTIONS(650),
    [aux_sym_comparison_operator_token23] = ACTIONS(650),
    [aux_sym_comparison_operator_token24] = ACTIONS(650),
    [aux_sym_comparison_operator_token25] = ACTIONS(650),
    [aux_sym_comparison_operator_token26] = ACTIONS(650),
    [aux_sym_comparison_operator_token27] = ACTIONS(650),
    [aux_sym_comparison_operator_token28] = ACTIONS(652),
    [aux_sym_comparison_operator_token29] = ACTIONS(650),
    [aux_sym_comparison_operator_token30] = ACTIONS(650),
    [aux_sym_comparison_operator_token31] = ACTIONS(650),
    [aux_sym_comparison_operator_token32] = ACTIONS(650),
    [aux_sym_comparison_operator_token33] = ACTIONS(650),
    [aux_sym_comparison_operator_token34] = ACTIONS(652),
    [aux_sym_comparison_operator_token35] = ACTIONS(650),
    [aux_sym_comparison_operator_token36] = ACTIONS(650),
    [aux_sym_comparison_operator_token37] = ACTIONS(650),
    [aux_sym_comparison_operator_token38] = ACTIONS(650),
    [aux_sym_comparison_operator_token39] = ACTIONS(650),
    [aux_sym_comparison_operator_token40] = ACTIONS(650),
    [aux_sym_comparison_operator_token41] = ACTIONS(650),
    [aux_sym_comparison_operator_token42] = ACTIONS(650),
    [aux_sym_comparison_operator_token43] = ACTIONS(650),
    [aux_sym_comparison_operator_token44] = ACTIONS(650),
    [aux_sym_comparison_operator_token45] = ACTIONS(650),
    [aux_sym_comparison_operator_token46] = ACTIONS(650),
    [aux_sym_comparison_operator_token47] = ACTIONS(650),
    [aux_sym_comparison_operator_token48] = ACTIONS(650),
    [aux_sym_comparison_operator_token49] = ACTIONS(650),
    [aux_sym_comparison_operator_token50] = ACTIONS(650),
    [aux_sym_format_operator_token1] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_AMP_AMP] = ACTIONS(650),
    [anon_sym_PIPE_PIPE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(652),
    [aux_sym_logical_expression_token1] = ACTIONS(650),
    [aux_sym_logical_expression_token2] = ACTIONS(650),
    [aux_sym_logical_expression_token3] = ACTIONS(650),
    [aux_sym_bitwise_expression_token1] = ACTIONS(650),
    [aux_sym_bitwise_expression_token2] = ACTIONS(650),
    [aux_sym_bitwise_expression_token3] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_SLASH] = ACTIONS(652),
    [anon_sym_BSLASH] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(652),
    [anon_sym_STAR] = ACTIONS(652),
    [anon_sym_DOT_DOT] = ACTIONS(650),
    [anon_sym_PLUS_PLUS] = ACTIONS(650),
    [anon_sym_DASH_DASH] = ACTIONS(650),
    [anon_sym_DOT2] = ACTIONS(652),
    [anon_sym_COLON_COLON] = ACTIONS(650),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(650),
    [sym__statement_terminator] = ACTIONS(650),
  },
  [STATE(146)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(716),
    [anon_sym_EQ] = ACTIONS(716),
    [anon_sym_BANG_EQ] = ACTIONS(716),
    [anon_sym_PLUS_EQ] = ACTIONS(716),
    [anon_sym_STAR_EQ] = ACTIONS(716),
    [anon_sym_SLASH_EQ] = ACTIONS(716),
    [anon_sym_PERCENT_EQ] = ACTIONS(716),
    [anon_sym_DASH_EQ] = ACTIONS(716),
    [anon_sym_GT] = ACTIONS(718),
    [anon_sym_GT_GT] = ACTIONS(716),
    [anon_sym_2_GT] = ACTIONS(718),
    [anon_sym_2_GT_GT] = ACTIONS(716),
    [anon_sym_3_GT] = ACTIONS(718),
    [anon_sym_3_GT_GT] = ACTIONS(716),
    [anon_sym_4_GT] = ACTIONS(718),
    [anon_sym_4_GT_GT] = ACTIONS(716),
    [anon_sym_5_GT] = ACTIONS(718),
    [anon_sym_5_GT_GT] = ACTIONS(716),
    [anon_sym_6_GT] = ACTIONS(718),
    [anon_sym_6_GT_GT] = ACTIONS(716),
    [anon_sym_STAR_GT] = ACTIONS(718),
    [anon_sym_STAR_GT_GT] = ACTIONS(716),
    [anon_sym_LT] = ACTIONS(718),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(716),
    [anon_sym_2_GT_AMP1] = ACTIONS(716),
    [anon_sym_3_GT_AMP1] = ACTIONS(716),
    [anon_sym_4_GT_AMP1] = ACTIONS(716),
    [anon_sym_5_GT_AMP1] = ACTIONS(716),
    [anon_sym_6_GT_AMP1] = ACTIONS(716),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(716),
    [anon_sym_1_GT_AMP2] = ACTIONS(716),
    [anon_sym_3_GT_AMP2] = ACTIONS(716),
    [anon_sym_4_GT_AMP2] = ACTIONS(716),
    [anon_sym_5_GT_AMP2] = ACTIONS(716),
    [anon_sym_6_GT_AMP2] = ACTIONS(716),
    [aux_sym_comparison_operator_token1] = ACTIONS(716),
    [aux_sym_comparison_operator_token2] = ACTIONS(716),
    [aux_sym_comparison_operator_token3] = ACTIONS(716),
    [aux_sym_comparison_operator_token4] = ACTIONS(716),
    [aux_sym_comparison_operator_token5] = ACTIONS(716),
    [aux_sym_comparison_operator_token6] = ACTIONS(716),
    [aux_sym_comparison_operator_token7] = ACTIONS(716),
    [aux_sym_comparison_operator_token8] = ACTIONS(716),
    [aux_sym_comparison_operator_token9] = ACTIONS(716),
    [aux_sym_comparison_operator_token10] = ACTIONS(716),
    [aux_sym_comparison_operator_token11] = ACTIONS(716),
    [aux_sym_comparison_operator_token12] = ACTIONS(716),
    [aux_sym_comparison_operator_token13] = ACTIONS(716),
    [aux_sym_comparison_operator_token14] = ACTIONS(716),
    [aux_sym_comparison_operator_token15] = ACTIONS(716),
    [aux_sym_comparison_operator_token16] = ACTIONS(716),
    [aux_sym_comparison_operator_token17] = ACTIONS(716),
    [aux_sym_comparison_operator_token18] = ACTIONS(716),
    [aux_sym_comparison_operator_token19] = ACTIONS(716),
    [aux_sym_comparison_operator_token20] = ACTIONS(716),
    [aux_sym_comparison_operator_token21] = ACTIONS(716),
    [aux_sym_comparison_operator_token22] = ACTIONS(716),
    [aux_sym_comparison_operator_token23] = ACTIONS(716),
    [aux_sym_comparison_operator_token24] = ACTIONS(716),
    [aux_sym_comparison_operator_token25] = ACTIONS(716),
    [aux_sym_comparison_operator_token26] = ACTIONS(716),
    [aux_sym_comparison_operator_token27] = ACTIONS(716),
    [aux_sym_comparison_operator_token28] = ACTIONS(718),
    [aux_sym_comparison_operator_token29] = ACTIONS(716),
    [aux_sym_comparison_operator_token30] = ACTIONS(716),
    [aux_sym_comparison_operator_token31] = ACTIONS(716),
    [aux_sym_comparison_operator_token32] = ACTIONS(716),
    [aux_sym_comparison_operator_token33] = ACTIONS(716),
    [aux_sym_comparison_operator_token34] = ACTIONS(718),
    [aux_sym_comparison_operator_token35] = ACTIONS(716),
    [aux_sym_comparison_operator_token36] = ACTIONS(716),
    [aux_sym_comparison_operator_token37] = ACTIONS(716),
    [aux_sym_comparison_operator_token38] = ACTIONS(716),
    [aux_sym_comparison_operator_token39] = ACTIONS(716),
    [aux_sym_comparison_operator_token40] = ACTIONS(716),
    [aux_sym_comparison_operator_token41] = ACTIONS(716),
    [aux_sym_comparison_operator_token42] = ACTIONS(716),
    [aux_sym_comparison_operator_token43] = ACTIONS(716),
    [aux_sym_comparison_operator_token44] = ACTIONS(716),
    [aux_sym_comparison_operator_token45] = ACTIONS(716),
    [aux_sym_comparison_operator_token46] = ACTIONS(716),
    [aux_sym_comparison_operator_token47] = ACTIONS(716),
    [aux_sym_comparison_operator_token48] = ACTIONS(716),
    [aux_sym_comparison_operator_token49] = ACTIONS(716),
    [aux_sym_comparison_operator_token50] = ACTIONS(716),
    [aux_sym_format_operator_token1] = ACTIONS(716),
    [anon_sym_LPAREN] = ACTIONS(716),
    [anon_sym_COMMA] = ACTIONS(716),
    [anon_sym_AMP_AMP] = ACTIONS(716),
    [anon_sym_PIPE_PIPE] = ACTIONS(716),
    [anon_sym_PIPE] = ACTIONS(718),
    [aux_sym_logical_expression_token1] = ACTIONS(716),
    [aux_sym_logical_expression_token2] = ACTIONS(716),
    [aux_sym_logical_expression_token3] = ACTIONS(716),
    [aux_sym_bitwise_expression_token1] = ACTIONS(716),
    [aux_sym_bitwise_expression_token2] = ACTIONS(716),
    [aux_sym_bitwise_expression_token3] = ACTIONS(716),
    [anon_sym_PLUS] = ACTIONS(718),
    [anon_sym_DASH] = ACTIONS(718),
    [anon_sym_SLASH] = ACTIONS(718),
    [anon_sym_BSLASH] = ACTIONS(716),
    [anon_sym_PERCENT] = ACTIONS(718),
    [anon_sym_STAR] = ACTIONS(718),
    [anon_sym_DOT_DOT] = ACTIONS(716),
    [anon_sym_PLUS_PLUS] = ACTIONS(716),
    [anon_sym_DASH_DASH] = ACTIONS(716),
    [anon_sym_DOT2] = ACTIONS(718),
    [anon_sym_COLON_COLON] = ACTIONS(716),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(716),
    [sym__statement_terminator] = ACTIONS(716),
  },
  [STATE(147)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(638),
    [anon_sym_EQ] = ACTIONS(638),
    [anon_sym_BANG_EQ] = ACTIONS(638),
    [anon_sym_PLUS_EQ] = ACTIONS(638),
    [anon_sym_STAR_EQ] = ACTIONS(638),
    [anon_sym_SLASH_EQ] = ACTIONS(638),
    [anon_sym_PERCENT_EQ] = ACTIONS(638),
    [anon_sym_DASH_EQ] = ACTIONS(638),
    [anon_sym_GT] = ACTIONS(640),
    [anon_sym_GT_GT] = ACTIONS(638),
    [anon_sym_2_GT] = ACTIONS(640),
    [anon_sym_2_GT_GT] = ACTIONS(638),
    [anon_sym_3_GT] = ACTIONS(640),
    [anon_sym_3_GT_GT] = ACTIONS(638),
    [anon_sym_4_GT] = ACTIONS(640),
    [anon_sym_4_GT_GT] = ACTIONS(638),
    [anon_sym_5_GT] = ACTIONS(640),
    [anon_sym_5_GT_GT] = ACTIONS(638),
    [anon_sym_6_GT] = ACTIONS(640),
    [anon_sym_6_GT_GT] = ACTIONS(638),
    [anon_sym_STAR_GT] = ACTIONS(640),
    [anon_sym_STAR_GT_GT] = ACTIONS(638),
    [anon_sym_LT] = ACTIONS(640),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(638),
    [anon_sym_2_GT_AMP1] = ACTIONS(638),
    [anon_sym_3_GT_AMP1] = ACTIONS(638),
    [anon_sym_4_GT_AMP1] = ACTIONS(638),
    [anon_sym_5_GT_AMP1] = ACTIONS(638),
    [anon_sym_6_GT_AMP1] = ACTIONS(638),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(638),
    [anon_sym_1_GT_AMP2] = ACTIONS(638),
    [anon_sym_3_GT_AMP2] = ACTIONS(638),
    [anon_sym_4_GT_AMP2] = ACTIONS(638),
    [anon_sym_5_GT_AMP2] = ACTIONS(638),
    [anon_sym_6_GT_AMP2] = ACTIONS(638),
    [aux_sym_comparison_operator_token1] = ACTIONS(638),
    [aux_sym_comparison_operator_token2] = ACTIONS(638),
    [aux_sym_comparison_operator_token3] = ACTIONS(638),
    [aux_sym_comparison_operator_token4] = ACTIONS(638),
    [aux_sym_comparison_operator_token5] = ACTIONS(638),
    [aux_sym_comparison_operator_token6] = ACTIONS(638),
    [aux_sym_comparison_operator_token7] = ACTIONS(638),
    [aux_sym_comparison_operator_token8] = ACTIONS(638),
    [aux_sym_comparison_operator_token9] = ACTIONS(638),
    [aux_sym_comparison_operator_token10] = ACTIONS(638),
    [aux_sym_comparison_operator_token11] = ACTIONS(638),
    [aux_sym_comparison_operator_token12] = ACTIONS(638),
    [aux_sym_comparison_operator_token13] = ACTIONS(638),
    [aux_sym_comparison_operator_token14] = ACTIONS(638),
    [aux_sym_comparison_operator_token15] = ACTIONS(638),
    [aux_sym_comparison_operator_token16] = ACTIONS(638),
    [aux_sym_comparison_operator_token17] = ACTIONS(638),
    [aux_sym_comparison_operator_token18] = ACTIONS(638),
    [aux_sym_comparison_operator_token19] = ACTIONS(638),
    [aux_sym_comparison_operator_token20] = ACTIONS(638),
    [aux_sym_comparison_operator_token21] = ACTIONS(638),
    [aux_sym_comparison_operator_token22] = ACTIONS(638),
    [aux_sym_comparison_operator_token23] = ACTIONS(638),
    [aux_sym_comparison_operator_token24] = ACTIONS(638),
    [aux_sym_comparison_operator_token25] = ACTIONS(638),
    [aux_sym_comparison_operator_token26] = ACTIONS(638),
    [aux_sym_comparison_operator_token27] = ACTIONS(638),
    [aux_sym_comparison_operator_token28] = ACTIONS(640),
    [aux_sym_comparison_operator_token29] = ACTIONS(638),
    [aux_sym_comparison_operator_token30] = ACTIONS(638),
    [aux_sym_comparison_operator_token31] = ACTIONS(638),
    [aux_sym_comparison_operator_token32] = ACTIONS(638),
    [aux_sym_comparison_operator_token33] = ACTIONS(638),
    [aux_sym_comparison_operator_token34] = ACTIONS(640),
    [aux_sym_comparison_operator_token35] = ACTIONS(638),
    [aux_sym_comparison_operator_token36] = ACTIONS(638),
    [aux_sym_comparison_operator_token37] = ACTIONS(638),
    [aux_sym_comparison_operator_token38] = ACTIONS(638),
    [aux_sym_comparison_operator_token39] = ACTIONS(638),
    [aux_sym_comparison_operator_token40] = ACTIONS(638),
    [aux_sym_comparison_operator_token41] = ACTIONS(638),
    [aux_sym_comparison_operator_token42] = ACTIONS(638),
    [aux_sym_comparison_operator_token43] = ACTIONS(638),
    [aux_sym_comparison_operator_token44] = ACTIONS(638),
    [aux_sym_comparison_operator_token45] = ACTIONS(638),
    [aux_sym_comparison_operator_token46] = ACTIONS(638),
    [aux_sym_comparison_operator_token47] = ACTIONS(638),
    [aux_sym_comparison_operator_token48] = ACTIONS(638),
    [aux_sym_comparison_operator_token49] = ACTIONS(638),
    [aux_sym_comparison_operator_token50] = ACTIONS(638),
    [aux_sym_format_operator_token1] = ACTIONS(638),
    [anon_sym_LPAREN] = ACTIONS(638),
    [anon_sym_COMMA] = ACTIONS(638),
    [anon_sym_AMP_AMP] = ACTIONS(638),
    [anon_sym_PIPE_PIPE] = ACTIONS(638),
    [anon_sym_PIPE] = ACTIONS(640),
    [aux_sym_logical_expression_token1] = ACTIONS(638),
    [aux_sym_logical_expression_token2] = ACTIONS(638),
    [aux_sym_logical_expression_token3] = ACTIONS(638),
    [aux_sym_bitwise_expression_token1] = ACTIONS(638),
    [aux_sym_bitwise_expression_token2] = ACTIONS(638),
    [aux_sym_bitwise_expression_token3] = ACTIONS(638),
    [anon_sym_PLUS] = ACTIONS(640),
    [anon_sym_DASH] = ACTIONS(640),
    [anon_sym_SLASH] = ACTIONS(640),
    [anon_sym_BSLASH] = ACTIONS(638),
    [anon_sym_PERCENT] = ACTIONS(640),
    [anon_sym_STAR] = ACTIONS(640),
    [anon_sym_DOT_DOT] = ACTIONS(638),
    [anon_sym_PLUS_PLUS] = ACTIONS(638),
    [anon_sym_DASH_DASH] = ACTIONS(638),
    [anon_sym_DOT2] = ACTIONS(640),
    [anon_sym_COLON_COLON] = ACTIONS(638),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(638),
    [sym__statement_terminator] = ACTIONS(638),
  },
  [STATE(148)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(634),
    [anon_sym_EQ] = ACTIONS(634),
    [anon_sym_BANG_EQ] = ACTIONS(634),
    [anon_sym_PLUS_EQ] = ACTIONS(634),
    [anon_sym_STAR_EQ] = ACTIONS(634),
    [anon_sym_SLASH_EQ] = ACTIONS(634),
    [anon_sym_PERCENT_EQ] = ACTIONS(634),
    [anon_sym_DASH_EQ] = ACTIONS(634),
    [anon_sym_GT] = ACTIONS(636),
    [anon_sym_GT_GT] = ACTIONS(634),
    [anon_sym_2_GT] = ACTIONS(636),
    [anon_sym_2_GT_GT] = ACTIONS(634),
    [anon_sym_3_GT] = ACTIONS(636),
    [anon_sym_3_GT_GT] = ACTIONS(634),
    [anon_sym_4_GT] = ACTIONS(636),
    [anon_sym_4_GT_GT] = ACTIONS(634),
    [anon_sym_5_GT] = ACTIONS(636),
    [anon_sym_5_GT_GT] = ACTIONS(634),
    [anon_sym_6_GT] = ACTIONS(636),
    [anon_sym_6_GT_GT] = ACTIONS(634),
    [anon_sym_STAR_GT] = ACTIONS(636),
    [anon_sym_STAR_GT_GT] = ACTIONS(634),
    [anon_sym_LT] = ACTIONS(636),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(634),
    [anon_sym_2_GT_AMP1] = ACTIONS(634),
    [anon_sym_3_GT_AMP1] = ACTIONS(634),
    [anon_sym_4_GT_AMP1] = ACTIONS(634),
    [anon_sym_5_GT_AMP1] = ACTIONS(634),
    [anon_sym_6_GT_AMP1] = ACTIONS(634),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(634),
    [anon_sym_1_GT_AMP2] = ACTIONS(634),
    [anon_sym_3_GT_AMP2] = ACTIONS(634),
    [anon_sym_4_GT_AMP2] = ACTIONS(634),
    [anon_sym_5_GT_AMP2] = ACTIONS(634),
    [anon_sym_6_GT_AMP2] = ACTIONS(634),
    [aux_sym_comparison_operator_token1] = ACTIONS(634),
    [aux_sym_comparison_operator_token2] = ACTIONS(634),
    [aux_sym_comparison_operator_token3] = ACTIONS(634),
    [aux_sym_comparison_operator_token4] = ACTIONS(634),
    [aux_sym_comparison_operator_token5] = ACTIONS(634),
    [aux_sym_comparison_operator_token6] = ACTIONS(634),
    [aux_sym_comparison_operator_token7] = ACTIONS(634),
    [aux_sym_comparison_operator_token8] = ACTIONS(634),
    [aux_sym_comparison_operator_token9] = ACTIONS(634),
    [aux_sym_comparison_operator_token10] = ACTIONS(634),
    [aux_sym_comparison_operator_token11] = ACTIONS(634),
    [aux_sym_comparison_operator_token12] = ACTIONS(634),
    [aux_sym_comparison_operator_token13] = ACTIONS(634),
    [aux_sym_comparison_operator_token14] = ACTIONS(634),
    [aux_sym_comparison_operator_token15] = ACTIONS(634),
    [aux_sym_comparison_operator_token16] = ACTIONS(634),
    [aux_sym_comparison_operator_token17] = ACTIONS(634),
    [aux_sym_comparison_operator_token18] = ACTIONS(634),
    [aux_sym_comparison_operator_token19] = ACTIONS(634),
    [aux_sym_comparison_operator_token20] = ACTIONS(634),
    [aux_sym_comparison_operator_token21] = ACTIONS(634),
    [aux_sym_comparison_operator_token22] = ACTIONS(634),
    [aux_sym_comparison_operator_token23] = ACTIONS(634),
    [aux_sym_comparison_operator_token24] = ACTIONS(634),
    [aux_sym_comparison_operator_token25] = ACTIONS(634),
    [aux_sym_comparison_operator_token26] = ACTIONS(634),
    [aux_sym_comparison_operator_token27] = ACTIONS(634),
    [aux_sym_comparison_operator_token28] = ACTIONS(636),
    [aux_sym_comparison_operator_token29] = ACTIONS(634),
    [aux_sym_comparison_operator_token30] = ACTIONS(634),
    [aux_sym_comparison_operator_token31] = ACTIONS(634),
    [aux_sym_comparison_operator_token32] = ACTIONS(634),
    [aux_sym_comparison_operator_token33] = ACTIONS(634),
    [aux_sym_comparison_operator_token34] = ACTIONS(636),
    [aux_sym_comparison_operator_token35] = ACTIONS(634),
    [aux_sym_comparison_operator_token36] = ACTIONS(634),
    [aux_sym_comparison_operator_token37] = ACTIONS(634),
    [aux_sym_comparison_operator_token38] = ACTIONS(634),
    [aux_sym_comparison_operator_token39] = ACTIONS(634),
    [aux_sym_comparison_operator_token40] = ACTIONS(634),
    [aux_sym_comparison_operator_token41] = ACTIONS(634),
    [aux_sym_comparison_operator_token42] = ACTIONS(634),
    [aux_sym_comparison_operator_token43] = ACTIONS(634),
    [aux_sym_comparison_operator_token44] = ACTIONS(634),
    [aux_sym_comparison_operator_token45] = ACTIONS(634),
    [aux_sym_comparison_operator_token46] = ACTIONS(634),
    [aux_sym_comparison_operator_token47] = ACTIONS(634),
    [aux_sym_comparison_operator_token48] = ACTIONS(634),
    [aux_sym_comparison_operator_token49] = ACTIONS(634),
    [aux_sym_comparison_operator_token50] = ACTIONS(634),
    [aux_sym_format_operator_token1] = ACTIONS(634),
    [anon_sym_LPAREN] = ACTIONS(634),
    [anon_sym_COMMA] = ACTIONS(634),
    [anon_sym_AMP_AMP] = ACTIONS(634),
    [anon_sym_PIPE_PIPE] = ACTIONS(634),
    [anon_sym_PIPE] = ACTIONS(636),
    [aux_sym_logical_expression_token1] = ACTIONS(634),
    [aux_sym_logical_expression_token2] = ACTIONS(634),
    [aux_sym_logical_expression_token3] = ACTIONS(634),
    [aux_sym_bitwise_expression_token1] = ACTIONS(634),
    [aux_sym_bitwise_expression_token2] = ACTIONS(634),
    [aux_sym_bitwise_expression_token3] = ACTIONS(634),
    [anon_sym_PLUS] = ACTIONS(636),
    [anon_sym_DASH] = ACTIONS(636),
    [anon_sym_SLASH] = ACTIONS(636),
    [anon_sym_BSLASH] = ACTIONS(634),
    [anon_sym_PERCENT] = ACTIONS(636),
    [anon_sym_STAR] = ACTIONS(636),
    [anon_sym_DOT_DOT] = ACTIONS(634),
    [anon_sym_PLUS_PLUS] = ACTIONS(634),
    [anon_sym_DASH_DASH] = ACTIONS(634),
    [anon_sym_DOT2] = ACTIONS(636),
    [anon_sym_COLON_COLON] = ACTIONS(634),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(634),
    [sym__statement_terminator] = ACTIONS(634),
  },
  [STATE(149)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(666),
    [anon_sym_EQ] = ACTIONS(666),
    [anon_sym_BANG_EQ] = ACTIONS(666),
    [anon_sym_PLUS_EQ] = ACTIONS(666),
    [anon_sym_STAR_EQ] = ACTIONS(666),
    [anon_sym_SLASH_EQ] = ACTIONS(666),
    [anon_sym_PERCENT_EQ] = ACTIONS(666),
    [anon_sym_DASH_EQ] = ACTIONS(666),
    [anon_sym_GT] = ACTIONS(668),
    [anon_sym_GT_GT] = ACTIONS(666),
    [anon_sym_2_GT] = ACTIONS(668),
    [anon_sym_2_GT_GT] = ACTIONS(666),
    [anon_sym_3_GT] = ACTIONS(668),
    [anon_sym_3_GT_GT] = ACTIONS(666),
    [anon_sym_4_GT] = ACTIONS(668),
    [anon_sym_4_GT_GT] = ACTIONS(666),
    [anon_sym_5_GT] = ACTIONS(668),
    [anon_sym_5_GT_GT] = ACTIONS(666),
    [anon_sym_6_GT] = ACTIONS(668),
    [anon_sym_6_GT_GT] = ACTIONS(666),
    [anon_sym_STAR_GT] = ACTIONS(668),
    [anon_sym_STAR_GT_GT] = ACTIONS(666),
    [anon_sym_LT] = ACTIONS(668),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(666),
    [anon_sym_2_GT_AMP1] = ACTIONS(666),
    [anon_sym_3_GT_AMP1] = ACTIONS(666),
    [anon_sym_4_GT_AMP1] = ACTIONS(666),
    [anon_sym_5_GT_AMP1] = ACTIONS(666),
    [anon_sym_6_GT_AMP1] = ACTIONS(666),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(666),
    [anon_sym_1_GT_AMP2] = ACTIONS(666),
    [anon_sym_3_GT_AMP2] = ACTIONS(666),
    [anon_sym_4_GT_AMP2] = ACTIONS(666),
    [anon_sym_5_GT_AMP2] = ACTIONS(666),
    [anon_sym_6_GT_AMP2] = ACTIONS(666),
    [aux_sym_comparison_operator_token1] = ACTIONS(666),
    [aux_sym_comparison_operator_token2] = ACTIONS(666),
    [aux_sym_comparison_operator_token3] = ACTIONS(666),
    [aux_sym_comparison_operator_token4] = ACTIONS(666),
    [aux_sym_comparison_operator_token5] = ACTIONS(666),
    [aux_sym_comparison_operator_token6] = ACTIONS(666),
    [aux_sym_comparison_operator_token7] = ACTIONS(666),
    [aux_sym_comparison_operator_token8] = ACTIONS(666),
    [aux_sym_comparison_operator_token9] = ACTIONS(666),
    [aux_sym_comparison_operator_token10] = ACTIONS(666),
    [aux_sym_comparison_operator_token11] = ACTIONS(666),
    [aux_sym_comparison_operator_token12] = ACTIONS(666),
    [aux_sym_comparison_operator_token13] = ACTIONS(666),
    [aux_sym_comparison_operator_token14] = ACTIONS(666),
    [aux_sym_comparison_operator_token15] = ACTIONS(666),
    [aux_sym_comparison_operator_token16] = ACTIONS(666),
    [aux_sym_comparison_operator_token17] = ACTIONS(666),
    [aux_sym_comparison_operator_token18] = ACTIONS(666),
    [aux_sym_comparison_operator_token19] = ACTIONS(666),
    [aux_sym_comparison_operator_token20] = ACTIONS(666),
    [aux_sym_comparison_operator_token21] = ACTIONS(666),
    [aux_sym_comparison_operator_token22] = ACTIONS(666),
    [aux_sym_comparison_operator_token23] = ACTIONS(666),
    [aux_sym_comparison_operator_token24] = ACTIONS(666),
    [aux_sym_comparison_operator_token25] = ACTIONS(666),
    [aux_sym_comparison_operator_token26] = ACTIONS(666),
    [aux_sym_comparison_operator_token27] = ACTIONS(666),
    [aux_sym_comparison_operator_token28] = ACTIONS(668),
    [aux_sym_comparison_operator_token29] = ACTIONS(666),
    [aux_sym_comparison_operator_token30] = ACTIONS(666),
    [aux_sym_comparison_operator_token31] = ACTIONS(666),
    [aux_sym_comparison_operator_token32] = ACTIONS(666),
    [aux_sym_comparison_operator_token33] = ACTIONS(666),
    [aux_sym_comparison_operator_token34] = ACTIONS(668),
    [aux_sym_comparison_operator_token35] = ACTIONS(666),
    [aux_sym_comparison_operator_token36] = ACTIONS(666),
    [aux_sym_comparison_operator_token37] = ACTIONS(666),
    [aux_sym_comparison_operator_token38] = ACTIONS(666),
    [aux_sym_comparison_operator_token39] = ACTIONS(666),
    [aux_sym_comparison_operator_token40] = ACTIONS(666),
    [aux_sym_comparison_operator_token41] = ACTIONS(666),
    [aux_sym_comparison_operator_token42] = ACTIONS(666),
    [aux_sym_comparison_operator_token43] = ACTIONS(666),
    [aux_sym_comparison_operator_token44] = ACTIONS(666),
    [aux_sym_comparison_operator_token45] = ACTIONS(666),
    [aux_sym_comparison_operator_token46] = ACTIONS(666),
    [aux_sym_comparison_operator_token47] = ACTIONS(666),
    [aux_sym_comparison_operator_token48] = ACTIONS(666),
    [aux_sym_comparison_operator_token49] = ACTIONS(666),
    [aux_sym_comparison_operator_token50] = ACTIONS(666),
    [aux_sym_format_operator_token1] = ACTIONS(666),
    [anon_sym_LPAREN] = ACTIONS(666),
    [anon_sym_COMMA] = ACTIONS(666),
    [anon_sym_AMP_AMP] = ACTIONS(666),
    [anon_sym_PIPE_PIPE] = ACTIONS(666),
    [anon_sym_PIPE] = ACTIONS(668),
    [aux_sym_logical_expression_token1] = ACTIONS(666),
    [aux_sym_logical_expression_token2] = ACTIONS(666),
    [aux_sym_logical_expression_token3] = ACTIONS(666),
    [aux_sym_bitwise_expression_token1] = ACTIONS(666),
    [aux_sym_bitwise_expression_token2] = ACTIONS(666),
    [aux_sym_bitwise_expression_token3] = ACTIONS(666),
    [anon_sym_PLUS] = ACTIONS(668),
    [anon_sym_DASH] = ACTIONS(668),
    [anon_sym_SLASH] = ACTIONS(668),
    [anon_sym_BSLASH] = ACTIONS(666),
    [anon_sym_PERCENT] = ACTIONS(668),
    [anon_sym_STAR] = ACTIONS(668),
    [anon_sym_DOT_DOT] = ACTIONS(666),
    [anon_sym_PLUS_PLUS] = ACTIONS(666),
    [anon_sym_DASH_DASH] = ACTIONS(666),
    [anon_sym_DOT2] = ACTIONS(668),
    [anon_sym_COLON_COLON] = ACTIONS(666),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(666),
    [sym__statement_terminator] = ACTIONS(666),
  },
  [STATE(150)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(728),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(730),
    [anon_sym_DASH_DASH] = ACTIONS(732),
    [anon_sym_DOT2] = ACTIONS(734),
    [anon_sym_COLON_COLON] = ACTIONS(736),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(738),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(151)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(740),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(742),
    [anon_sym_DASH_DASH] = ACTIONS(744),
    [anon_sym_DOT2] = ACTIONS(746),
    [anon_sym_COLON_COLON] = ACTIONS(748),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(750),
  },
  [STATE(152)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(686),
    [anon_sym_EQ] = ACTIONS(686),
    [anon_sym_BANG_EQ] = ACTIONS(686),
    [anon_sym_PLUS_EQ] = ACTIONS(686),
    [anon_sym_STAR_EQ] = ACTIONS(686),
    [anon_sym_SLASH_EQ] = ACTIONS(686),
    [anon_sym_PERCENT_EQ] = ACTIONS(686),
    [anon_sym_DASH_EQ] = ACTIONS(686),
    [anon_sym_GT] = ACTIONS(688),
    [anon_sym_GT_GT] = ACTIONS(686),
    [anon_sym_2_GT] = ACTIONS(688),
    [anon_sym_2_GT_GT] = ACTIONS(686),
    [anon_sym_3_GT] = ACTIONS(688),
    [anon_sym_3_GT_GT] = ACTIONS(686),
    [anon_sym_4_GT] = ACTIONS(688),
    [anon_sym_4_GT_GT] = ACTIONS(686),
    [anon_sym_5_GT] = ACTIONS(688),
    [anon_sym_5_GT_GT] = ACTIONS(686),
    [anon_sym_6_GT] = ACTIONS(688),
    [anon_sym_6_GT_GT] = ACTIONS(686),
    [anon_sym_STAR_GT] = ACTIONS(688),
    [anon_sym_STAR_GT_GT] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(688),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(686),
    [anon_sym_2_GT_AMP1] = ACTIONS(686),
    [anon_sym_3_GT_AMP1] = ACTIONS(686),
    [anon_sym_4_GT_AMP1] = ACTIONS(686),
    [anon_sym_5_GT_AMP1] = ACTIONS(686),
    [anon_sym_6_GT_AMP1] = ACTIONS(686),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(686),
    [anon_sym_1_GT_AMP2] = ACTIONS(686),
    [anon_sym_3_GT_AMP2] = ACTIONS(686),
    [anon_sym_4_GT_AMP2] = ACTIONS(686),
    [anon_sym_5_GT_AMP2] = ACTIONS(686),
    [anon_sym_6_GT_AMP2] = ACTIONS(686),
    [aux_sym_comparison_operator_token1] = ACTIONS(686),
    [aux_sym_comparison_operator_token2] = ACTIONS(686),
    [aux_sym_comparison_operator_token3] = ACTIONS(686),
    [aux_sym_comparison_operator_token4] = ACTIONS(686),
    [aux_sym_comparison_operator_token5] = ACTIONS(686),
    [aux_sym_comparison_operator_token6] = ACTIONS(686),
    [aux_sym_comparison_operator_token7] = ACTIONS(686),
    [aux_sym_comparison_operator_token8] = ACTIONS(686),
    [aux_sym_comparison_operator_token9] = ACTIONS(686),
    [aux_sym_comparison_operator_token10] = ACTIONS(686),
    [aux_sym_comparison_operator_token11] = ACTIONS(686),
    [aux_sym_comparison_operator_token12] = ACTIONS(686),
    [aux_sym_comparison_operator_token13] = ACTIONS(686),
    [aux_sym_comparison_operator_token14] = ACTIONS(686),
    [aux_sym_comparison_operator_token15] = ACTIONS(686),
    [aux_sym_comparison_operator_token16] = ACTIONS(686),
    [aux_sym_comparison_operator_token17] = ACTIONS(686),
    [aux_sym_comparison_operator_token18] = ACTIONS(686),
    [aux_sym_comparison_operator_token19] = ACTIONS(686),
    [aux_sym_comparison_operator_token20] = ACTIONS(686),
    [aux_sym_comparison_operator_token21] = ACTIONS(686),
    [aux_sym_comparison_operator_token22] = ACTIONS(686),
    [aux_sym_comparison_operator_token23] = ACTIONS(686),
    [aux_sym_comparison_operator_token24] = ACTIONS(686),
    [aux_sym_comparison_operator_token25] = ACTIONS(686),
    [aux_sym_comparison_operator_token26] = ACTIONS(686),
    [aux_sym_comparison_operator_token27] = ACTIONS(686),
    [aux_sym_comparison_operator_token28] = ACTIONS(688),
    [aux_sym_comparison_operator_token29] = ACTIONS(686),
    [aux_sym_comparison_operator_token30] = ACTIONS(686),
    [aux_sym_comparison_operator_token31] = ACTIONS(686),
    [aux_sym_comparison_operator_token32] = ACTIONS(686),
    [aux_sym_comparison_operator_token33] = ACTIONS(686),
    [aux_sym_comparison_operator_token34] = ACTIONS(688),
    [aux_sym_comparison_operator_token35] = ACTIONS(686),
    [aux_sym_comparison_operator_token36] = ACTIONS(686),
    [aux_sym_comparison_operator_token37] = ACTIONS(686),
    [aux_sym_comparison_operator_token38] = ACTIONS(686),
    [aux_sym_comparison_operator_token39] = ACTIONS(686),
    [aux_sym_comparison_operator_token40] = ACTIONS(686),
    [aux_sym_comparison_operator_token41] = ACTIONS(686),
    [aux_sym_comparison_operator_token42] = ACTIONS(686),
    [aux_sym_comparison_operator_token43] = ACTIONS(686),
    [aux_sym_comparison_operator_token44] = ACTIONS(686),
    [aux_sym_comparison_operator_token45] = ACTIONS(686),
    [aux_sym_comparison_operator_token46] = ACTIONS(686),
    [aux_sym_comparison_operator_token47] = ACTIONS(686),
    [aux_sym_comparison_operator_token48] = ACTIONS(686),
    [aux_sym_comparison_operator_token49] = ACTIONS(686),
    [aux_sym_comparison_operator_token50] = ACTIONS(686),
    [aux_sym_format_operator_token1] = ACTIONS(686),
    [anon_sym_LPAREN] = ACTIONS(686),
    [anon_sym_COMMA] = ACTIONS(686),
    [anon_sym_AMP_AMP] = ACTIONS(686),
    [anon_sym_PIPE_PIPE] = ACTIONS(686),
    [anon_sym_PIPE] = ACTIONS(688),
    [aux_sym_logical_expression_token1] = ACTIONS(686),
    [aux_sym_logical_expression_token2] = ACTIONS(686),
    [aux_sym_logical_expression_token3] = ACTIONS(686),
    [aux_sym_bitwise_expression_token1] = ACTIONS(686),
    [aux_sym_bitwise_expression_token2] = ACTIONS(686),
    [aux_sym_bitwise_expression_token3] = ACTIONS(686),
    [anon_sym_PLUS] = ACTIONS(688),
    [anon_sym_DASH] = ACTIONS(688),
    [anon_sym_SLASH] = ACTIONS(688),
    [anon_sym_BSLASH] = ACTIONS(686),
    [anon_sym_PERCENT] = ACTIONS(688),
    [anon_sym_STAR] = ACTIONS(688),
    [anon_sym_DOT_DOT] = ACTIONS(686),
    [anon_sym_PLUS_PLUS] = ACTIONS(686),
    [anon_sym_DASH_DASH] = ACTIONS(686),
    [anon_sym_DOT2] = ACTIONS(688),
    [anon_sym_COLON_COLON] = ACTIONS(686),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(686),
    [sym__statement_terminator] = ACTIONS(686),
  },
  [STATE(153)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(690),
    [anon_sym_EQ] = ACTIONS(690),
    [anon_sym_BANG_EQ] = ACTIONS(690),
    [anon_sym_PLUS_EQ] = ACTIONS(690),
    [anon_sym_STAR_EQ] = ACTIONS(690),
    [anon_sym_SLASH_EQ] = ACTIONS(690),
    [anon_sym_PERCENT_EQ] = ACTIONS(690),
    [anon_sym_DASH_EQ] = ACTIONS(690),
    [anon_sym_GT] = ACTIONS(692),
    [anon_sym_GT_GT] = ACTIONS(690),
    [anon_sym_2_GT] = ACTIONS(692),
    [anon_sym_2_GT_GT] = ACTIONS(690),
    [anon_sym_3_GT] = ACTIONS(692),
    [anon_sym_3_GT_GT] = ACTIONS(690),
    [anon_sym_4_GT] = ACTIONS(692),
    [anon_sym_4_GT_GT] = ACTIONS(690),
    [anon_sym_5_GT] = ACTIONS(692),
    [anon_sym_5_GT_GT] = ACTIONS(690),
    [anon_sym_6_GT] = ACTIONS(692),
    [anon_sym_6_GT_GT] = ACTIONS(690),
    [anon_sym_STAR_GT] = ACTIONS(692),
    [anon_sym_STAR_GT_GT] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(690),
    [anon_sym_2_GT_AMP1] = ACTIONS(690),
    [anon_sym_3_GT_AMP1] = ACTIONS(690),
    [anon_sym_4_GT_AMP1] = ACTIONS(690),
    [anon_sym_5_GT_AMP1] = ACTIONS(690),
    [anon_sym_6_GT_AMP1] = ACTIONS(690),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(690),
    [anon_sym_1_GT_AMP2] = ACTIONS(690),
    [anon_sym_3_GT_AMP2] = ACTIONS(690),
    [anon_sym_4_GT_AMP2] = ACTIONS(690),
    [anon_sym_5_GT_AMP2] = ACTIONS(690),
    [anon_sym_6_GT_AMP2] = ACTIONS(690),
    [aux_sym_comparison_operator_token1] = ACTIONS(690),
    [aux_sym_comparison_operator_token2] = ACTIONS(690),
    [aux_sym_comparison_operator_token3] = ACTIONS(690),
    [aux_sym_comparison_operator_token4] = ACTIONS(690),
    [aux_sym_comparison_operator_token5] = ACTIONS(690),
    [aux_sym_comparison_operator_token6] = ACTIONS(690),
    [aux_sym_comparison_operator_token7] = ACTIONS(690),
    [aux_sym_comparison_operator_token8] = ACTIONS(690),
    [aux_sym_comparison_operator_token9] = ACTIONS(690),
    [aux_sym_comparison_operator_token10] = ACTIONS(690),
    [aux_sym_comparison_operator_token11] = ACTIONS(690),
    [aux_sym_comparison_operator_token12] = ACTIONS(690),
    [aux_sym_comparison_operator_token13] = ACTIONS(690),
    [aux_sym_comparison_operator_token14] = ACTIONS(690),
    [aux_sym_comparison_operator_token15] = ACTIONS(690),
    [aux_sym_comparison_operator_token16] = ACTIONS(690),
    [aux_sym_comparison_operator_token17] = ACTIONS(690),
    [aux_sym_comparison_operator_token18] = ACTIONS(690),
    [aux_sym_comparison_operator_token19] = ACTIONS(690),
    [aux_sym_comparison_operator_token20] = ACTIONS(690),
    [aux_sym_comparison_operator_token21] = ACTIONS(690),
    [aux_sym_comparison_operator_token22] = ACTIONS(690),
    [aux_sym_comparison_operator_token23] = ACTIONS(690),
    [aux_sym_comparison_operator_token24] = ACTIONS(690),
    [aux_sym_comparison_operator_token25] = ACTIONS(690),
    [aux_sym_comparison_operator_token26] = ACTIONS(690),
    [aux_sym_comparison_operator_token27] = ACTIONS(690),
    [aux_sym_comparison_operator_token28] = ACTIONS(692),
    [aux_sym_comparison_operator_token29] = ACTIONS(690),
    [aux_sym_comparison_operator_token30] = ACTIONS(690),
    [aux_sym_comparison_operator_token31] = ACTIONS(690),
    [aux_sym_comparison_operator_token32] = ACTIONS(690),
    [aux_sym_comparison_operator_token33] = ACTIONS(690),
    [aux_sym_comparison_operator_token34] = ACTIONS(692),
    [aux_sym_comparison_operator_token35] = ACTIONS(690),
    [aux_sym_comparison_operator_token36] = ACTIONS(690),
    [aux_sym_comparison_operator_token37] = ACTIONS(690),
    [aux_sym_comparison_operator_token38] = ACTIONS(690),
    [aux_sym_comparison_operator_token39] = ACTIONS(690),
    [aux_sym_comparison_operator_token40] = ACTIONS(690),
    [aux_sym_comparison_operator_token41] = ACTIONS(690),
    [aux_sym_comparison_operator_token42] = ACTIONS(690),
    [aux_sym_comparison_operator_token43] = ACTIONS(690),
    [aux_sym_comparison_operator_token44] = ACTIONS(690),
    [aux_sym_comparison_operator_token45] = ACTIONS(690),
    [aux_sym_comparison_operator_token46] = ACTIONS(690),
    [aux_sym_comparison_operator_token47] = ACTIONS(690),
    [aux_sym_comparison_operator_token48] = ACTIONS(690),
    [aux_sym_comparison_operator_token49] = ACTIONS(690),
    [aux_sym_comparison_operator_token50] = ACTIONS(690),
    [aux_sym_format_operator_token1] = ACTIONS(690),
    [anon_sym_LPAREN] = ACTIONS(690),
    [anon_sym_COMMA] = ACTIONS(690),
    [anon_sym_AMP_AMP] = ACTIONS(690),
    [anon_sym_PIPE_PIPE] = ACTIONS(690),
    [anon_sym_PIPE] = ACTIONS(692),
    [aux_sym_logical_expression_token1] = ACTIONS(690),
    [aux_sym_logical_expression_token2] = ACTIONS(690),
    [aux_sym_logical_expression_token3] = ACTIONS(690),
    [aux_sym_bitwise_expression_token1] = ACTIONS(690),
    [aux_sym_bitwise_expression_token2] = ACTIONS(690),
    [aux_sym_bitwise_expression_token3] = ACTIONS(690),
    [anon_sym_PLUS] = ACTIONS(692),
    [anon_sym_DASH] = ACTIONS(692),
    [anon_sym_SLASH] = ACTIONS(692),
    [anon_sym_BSLASH] = ACTIONS(690),
    [anon_sym_PERCENT] = ACTIONS(692),
    [anon_sym_STAR] = ACTIONS(692),
    [anon_sym_DOT_DOT] = ACTIONS(690),
    [anon_sym_PLUS_PLUS] = ACTIONS(690),
    [anon_sym_DASH_DASH] = ACTIONS(690),
    [anon_sym_DOT2] = ACTIONS(692),
    [anon_sym_COLON_COLON] = ACTIONS(690),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(690),
    [sym__statement_terminator] = ACTIONS(690),
  },
  [STATE(154)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(696),
    [anon_sym_EQ] = ACTIONS(696),
    [anon_sym_BANG_EQ] = ACTIONS(696),
    [anon_sym_PLUS_EQ] = ACTIONS(696),
    [anon_sym_STAR_EQ] = ACTIONS(696),
    [anon_sym_SLASH_EQ] = ACTIONS(696),
    [anon_sym_PERCENT_EQ] = ACTIONS(696),
    [anon_sym_DASH_EQ] = ACTIONS(696),
    [anon_sym_GT] = ACTIONS(698),
    [anon_sym_GT_GT] = ACTIONS(696),
    [anon_sym_2_GT] = ACTIONS(698),
    [anon_sym_2_GT_GT] = ACTIONS(696),
    [anon_sym_3_GT] = ACTIONS(698),
    [anon_sym_3_GT_GT] = ACTIONS(696),
    [anon_sym_4_GT] = ACTIONS(698),
    [anon_sym_4_GT_GT] = ACTIONS(696),
    [anon_sym_5_GT] = ACTIONS(698),
    [anon_sym_5_GT_GT] = ACTIONS(696),
    [anon_sym_6_GT] = ACTIONS(698),
    [anon_sym_6_GT_GT] = ACTIONS(696),
    [anon_sym_STAR_GT] = ACTIONS(698),
    [anon_sym_STAR_GT_GT] = ACTIONS(696),
    [anon_sym_LT] = ACTIONS(698),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(696),
    [anon_sym_2_GT_AMP1] = ACTIONS(696),
    [anon_sym_3_GT_AMP1] = ACTIONS(696),
    [anon_sym_4_GT_AMP1] = ACTIONS(696),
    [anon_sym_5_GT_AMP1] = ACTIONS(696),
    [anon_sym_6_GT_AMP1] = ACTIONS(696),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(696),
    [anon_sym_1_GT_AMP2] = ACTIONS(696),
    [anon_sym_3_GT_AMP2] = ACTIONS(696),
    [anon_sym_4_GT_AMP2] = ACTIONS(696),
    [anon_sym_5_GT_AMP2] = ACTIONS(696),
    [anon_sym_6_GT_AMP2] = ACTIONS(696),
    [aux_sym_comparison_operator_token1] = ACTIONS(696),
    [aux_sym_comparison_operator_token2] = ACTIONS(696),
    [aux_sym_comparison_operator_token3] = ACTIONS(696),
    [aux_sym_comparison_operator_token4] = ACTIONS(696),
    [aux_sym_comparison_operator_token5] = ACTIONS(696),
    [aux_sym_comparison_operator_token6] = ACTIONS(696),
    [aux_sym_comparison_operator_token7] = ACTIONS(696),
    [aux_sym_comparison_operator_token8] = ACTIONS(696),
    [aux_sym_comparison_operator_token9] = ACTIONS(696),
    [aux_sym_comparison_operator_token10] = ACTIONS(696),
    [aux_sym_comparison_operator_token11] = ACTIONS(696),
    [aux_sym_comparison_operator_token12] = ACTIONS(696),
    [aux_sym_comparison_operator_token13] = ACTIONS(696),
    [aux_sym_comparison_operator_token14] = ACTIONS(696),
    [aux_sym_comparison_operator_token15] = ACTIONS(696),
    [aux_sym_comparison_operator_token16] = ACTIONS(696),
    [aux_sym_comparison_operator_token17] = ACTIONS(696),
    [aux_sym_comparison_operator_token18] = ACTIONS(696),
    [aux_sym_comparison_operator_token19] = ACTIONS(696),
    [aux_sym_comparison_operator_token20] = ACTIONS(696),
    [aux_sym_comparison_operator_token21] = ACTIONS(696),
    [aux_sym_comparison_operator_token22] = ACTIONS(696),
    [aux_sym_comparison_operator_token23] = ACTIONS(696),
    [aux_sym_comparison_operator_token24] = ACTIONS(696),
    [aux_sym_comparison_operator_token25] = ACTIONS(696),
    [aux_sym_comparison_operator_token26] = ACTIONS(696),
    [aux_sym_comparison_operator_token27] = ACTIONS(696),
    [aux_sym_comparison_operator_token28] = ACTIONS(698),
    [aux_sym_comparison_operator_token29] = ACTIONS(696),
    [aux_sym_comparison_operator_token30] = ACTIONS(696),
    [aux_sym_comparison_operator_token31] = ACTIONS(696),
    [aux_sym_comparison_operator_token32] = ACTIONS(696),
    [aux_sym_comparison_operator_token33] = ACTIONS(696),
    [aux_sym_comparison_operator_token34] = ACTIONS(698),
    [aux_sym_comparison_operator_token35] = ACTIONS(696),
    [aux_sym_comparison_operator_token36] = ACTIONS(696),
    [aux_sym_comparison_operator_token37] = ACTIONS(696),
    [aux_sym_comparison_operator_token38] = ACTIONS(696),
    [aux_sym_comparison_operator_token39] = ACTIONS(696),
    [aux_sym_comparison_operator_token40] = ACTIONS(696),
    [aux_sym_comparison_operator_token41] = ACTIONS(696),
    [aux_sym_comparison_operator_token42] = ACTIONS(696),
    [aux_sym_comparison_operator_token43] = ACTIONS(696),
    [aux_sym_comparison_operator_token44] = ACTIONS(696),
    [aux_sym_comparison_operator_token45] = ACTIONS(696),
    [aux_sym_comparison_operator_token46] = ACTIONS(696),
    [aux_sym_comparison_operator_token47] = ACTIONS(696),
    [aux_sym_comparison_operator_token48] = ACTIONS(696),
    [aux_sym_comparison_operator_token49] = ACTIONS(696),
    [aux_sym_comparison_operator_token50] = ACTIONS(696),
    [aux_sym_format_operator_token1] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(696),
    [anon_sym_COMMA] = ACTIONS(696),
    [anon_sym_AMP_AMP] = ACTIONS(696),
    [anon_sym_PIPE_PIPE] = ACTIONS(696),
    [anon_sym_PIPE] = ACTIONS(698),
    [aux_sym_logical_expression_token1] = ACTIONS(696),
    [aux_sym_logical_expression_token2] = ACTIONS(696),
    [aux_sym_logical_expression_token3] = ACTIONS(696),
    [aux_sym_bitwise_expression_token1] = ACTIONS(696),
    [aux_sym_bitwise_expression_token2] = ACTIONS(696),
    [aux_sym_bitwise_expression_token3] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(698),
    [anon_sym_DASH] = ACTIONS(698),
    [anon_sym_SLASH] = ACTIONS(698),
    [anon_sym_BSLASH] = ACTIONS(696),
    [anon_sym_PERCENT] = ACTIONS(698),
    [anon_sym_STAR] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(696),
    [anon_sym_DASH_DASH] = ACTIONS(696),
    [anon_sym_DOT2] = ACTIONS(698),
    [anon_sym_COLON_COLON] = ACTIONS(696),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(696),
    [sym__statement_terminator] = ACTIONS(696),
  },
  [STATE(155)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(704),
    [anon_sym_EQ] = ACTIONS(704),
    [anon_sym_BANG_EQ] = ACTIONS(704),
    [anon_sym_PLUS_EQ] = ACTIONS(704),
    [anon_sym_STAR_EQ] = ACTIONS(704),
    [anon_sym_SLASH_EQ] = ACTIONS(704),
    [anon_sym_PERCENT_EQ] = ACTIONS(704),
    [anon_sym_DASH_EQ] = ACTIONS(704),
    [anon_sym_GT] = ACTIONS(706),
    [anon_sym_GT_GT] = ACTIONS(704),
    [anon_sym_2_GT] = ACTIONS(706),
    [anon_sym_2_GT_GT] = ACTIONS(704),
    [anon_sym_3_GT] = ACTIONS(706),
    [anon_sym_3_GT_GT] = ACTIONS(704),
    [anon_sym_4_GT] = ACTIONS(706),
    [anon_sym_4_GT_GT] = ACTIONS(704),
    [anon_sym_5_GT] = ACTIONS(706),
    [anon_sym_5_GT_GT] = ACTIONS(704),
    [anon_sym_6_GT] = ACTIONS(706),
    [anon_sym_6_GT_GT] = ACTIONS(704),
    [anon_sym_STAR_GT] = ACTIONS(706),
    [anon_sym_STAR_GT_GT] = ACTIONS(704),
    [anon_sym_LT] = ACTIONS(706),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(704),
    [anon_sym_2_GT_AMP1] = ACTIONS(704),
    [anon_sym_3_GT_AMP1] = ACTIONS(704),
    [anon_sym_4_GT_AMP1] = ACTIONS(704),
    [anon_sym_5_GT_AMP1] = ACTIONS(704),
    [anon_sym_6_GT_AMP1] = ACTIONS(704),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(704),
    [anon_sym_1_GT_AMP2] = ACTIONS(704),
    [anon_sym_3_GT_AMP2] = ACTIONS(704),
    [anon_sym_4_GT_AMP2] = ACTIONS(704),
    [anon_sym_5_GT_AMP2] = ACTIONS(704),
    [anon_sym_6_GT_AMP2] = ACTIONS(704),
    [aux_sym_comparison_operator_token1] = ACTIONS(704),
    [aux_sym_comparison_operator_token2] = ACTIONS(704),
    [aux_sym_comparison_operator_token3] = ACTIONS(704),
    [aux_sym_comparison_operator_token4] = ACTIONS(704),
    [aux_sym_comparison_operator_token5] = ACTIONS(704),
    [aux_sym_comparison_operator_token6] = ACTIONS(704),
    [aux_sym_comparison_operator_token7] = ACTIONS(704),
    [aux_sym_comparison_operator_token8] = ACTIONS(704),
    [aux_sym_comparison_operator_token9] = ACTIONS(704),
    [aux_sym_comparison_operator_token10] = ACTIONS(704),
    [aux_sym_comparison_operator_token11] = ACTIONS(704),
    [aux_sym_comparison_operator_token12] = ACTIONS(704),
    [aux_sym_comparison_operator_token13] = ACTIONS(704),
    [aux_sym_comparison_operator_token14] = ACTIONS(704),
    [aux_sym_comparison_operator_token15] = ACTIONS(704),
    [aux_sym_comparison_operator_token16] = ACTIONS(704),
    [aux_sym_comparison_operator_token17] = ACTIONS(704),
    [aux_sym_comparison_operator_token18] = ACTIONS(704),
    [aux_sym_comparison_operator_token19] = ACTIONS(704),
    [aux_sym_comparison_operator_token20] = ACTIONS(704),
    [aux_sym_comparison_operator_token21] = ACTIONS(704),
    [aux_sym_comparison_operator_token22] = ACTIONS(704),
    [aux_sym_comparison_operator_token23] = ACTIONS(704),
    [aux_sym_comparison_operator_token24] = ACTIONS(704),
    [aux_sym_comparison_operator_token25] = ACTIONS(704),
    [aux_sym_comparison_operator_token26] = ACTIONS(704),
    [aux_sym_comparison_operator_token27] = ACTIONS(704),
    [aux_sym_comparison_operator_token28] = ACTIONS(706),
    [aux_sym_comparison_operator_token29] = ACTIONS(704),
    [aux_sym_comparison_operator_token30] = ACTIONS(704),
    [aux_sym_comparison_operator_token31] = ACTIONS(704),
    [aux_sym_comparison_operator_token32] = ACTIONS(704),
    [aux_sym_comparison_operator_token33] = ACTIONS(704),
    [aux_sym_comparison_operator_token34] = ACTIONS(706),
    [aux_sym_comparison_operator_token35] = ACTIONS(704),
    [aux_sym_comparison_operator_token36] = ACTIONS(704),
    [aux_sym_comparison_operator_token37] = ACTIONS(704),
    [aux_sym_comparison_operator_token38] = ACTIONS(704),
    [aux_sym_comparison_operator_token39] = ACTIONS(704),
    [aux_sym_comparison_operator_token40] = ACTIONS(704),
    [aux_sym_comparison_operator_token41] = ACTIONS(704),
    [aux_sym_comparison_operator_token42] = ACTIONS(704),
    [aux_sym_comparison_operator_token43] = ACTIONS(704),
    [aux_sym_comparison_operator_token44] = ACTIONS(704),
    [aux_sym_comparison_operator_token45] = ACTIONS(704),
    [aux_sym_comparison_operator_token46] = ACTIONS(704),
    [aux_sym_comparison_operator_token47] = ACTIONS(704),
    [aux_sym_comparison_operator_token48] = ACTIONS(704),
    [aux_sym_comparison_operator_token49] = ACTIONS(704),
    [aux_sym_comparison_operator_token50] = ACTIONS(704),
    [aux_sym_format_operator_token1] = ACTIONS(704),
    [anon_sym_LPAREN] = ACTIONS(704),
    [anon_sym_COMMA] = ACTIONS(704),
    [anon_sym_AMP_AMP] = ACTIONS(704),
    [anon_sym_PIPE_PIPE] = ACTIONS(704),
    [anon_sym_PIPE] = ACTIONS(706),
    [aux_sym_logical_expression_token1] = ACTIONS(704),
    [aux_sym_logical_expression_token2] = ACTIONS(704),
    [aux_sym_logical_expression_token3] = ACTIONS(704),
    [aux_sym_bitwise_expression_token1] = ACTIONS(704),
    [aux_sym_bitwise_expression_token2] = ACTIONS(704),
    [aux_sym_bitwise_expression_token3] = ACTIONS(704),
    [anon_sym_PLUS] = ACTIONS(706),
    [anon_sym_DASH] = ACTIONS(706),
    [anon_sym_SLASH] = ACTIONS(706),
    [anon_sym_BSLASH] = ACTIONS(704),
    [anon_sym_PERCENT] = ACTIONS(706),
    [anon_sym_STAR] = ACTIONS(706),
    [anon_sym_DOT_DOT] = ACTIONS(704),
    [anon_sym_PLUS_PLUS] = ACTIONS(704),
    [anon_sym_DASH_DASH] = ACTIONS(704),
    [anon_sym_DOT2] = ACTIONS(706),
    [anon_sym_COLON_COLON] = ACTIONS(704),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(704),
    [sym__statement_terminator] = ACTIONS(704),
  },
  [STATE(156)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(720),
    [anon_sym_EQ] = ACTIONS(720),
    [anon_sym_BANG_EQ] = ACTIONS(720),
    [anon_sym_PLUS_EQ] = ACTIONS(720),
    [anon_sym_STAR_EQ] = ACTIONS(720),
    [anon_sym_SLASH_EQ] = ACTIONS(720),
    [anon_sym_PERCENT_EQ] = ACTIONS(720),
    [anon_sym_DASH_EQ] = ACTIONS(720),
    [anon_sym_GT] = ACTIONS(722),
    [anon_sym_GT_GT] = ACTIONS(720),
    [anon_sym_2_GT] = ACTIONS(722),
    [anon_sym_2_GT_GT] = ACTIONS(720),
    [anon_sym_3_GT] = ACTIONS(722),
    [anon_sym_3_GT_GT] = ACTIONS(720),
    [anon_sym_4_GT] = ACTIONS(722),
    [anon_sym_4_GT_GT] = ACTIONS(720),
    [anon_sym_5_GT] = ACTIONS(722),
    [anon_sym_5_GT_GT] = ACTIONS(720),
    [anon_sym_6_GT] = ACTIONS(722),
    [anon_sym_6_GT_GT] = ACTIONS(720),
    [anon_sym_STAR_GT] = ACTIONS(722),
    [anon_sym_STAR_GT_GT] = ACTIONS(720),
    [anon_sym_LT] = ACTIONS(722),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(720),
    [anon_sym_2_GT_AMP1] = ACTIONS(720),
    [anon_sym_3_GT_AMP1] = ACTIONS(720),
    [anon_sym_4_GT_AMP1] = ACTIONS(720),
    [anon_sym_5_GT_AMP1] = ACTIONS(720),
    [anon_sym_6_GT_AMP1] = ACTIONS(720),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(720),
    [anon_sym_1_GT_AMP2] = ACTIONS(720),
    [anon_sym_3_GT_AMP2] = ACTIONS(720),
    [anon_sym_4_GT_AMP2] = ACTIONS(720),
    [anon_sym_5_GT_AMP2] = ACTIONS(720),
    [anon_sym_6_GT_AMP2] = ACTIONS(720),
    [aux_sym_comparison_operator_token1] = ACTIONS(720),
    [aux_sym_comparison_operator_token2] = ACTIONS(720),
    [aux_sym_comparison_operator_token3] = ACTIONS(720),
    [aux_sym_comparison_operator_token4] = ACTIONS(720),
    [aux_sym_comparison_operator_token5] = ACTIONS(720),
    [aux_sym_comparison_operator_token6] = ACTIONS(720),
    [aux_sym_comparison_operator_token7] = ACTIONS(720),
    [aux_sym_comparison_operator_token8] = ACTIONS(720),
    [aux_sym_comparison_operator_token9] = ACTIONS(720),
    [aux_sym_comparison_operator_token10] = ACTIONS(720),
    [aux_sym_comparison_operator_token11] = ACTIONS(720),
    [aux_sym_comparison_operator_token12] = ACTIONS(720),
    [aux_sym_comparison_operator_token13] = ACTIONS(720),
    [aux_sym_comparison_operator_token14] = ACTIONS(720),
    [aux_sym_comparison_operator_token15] = ACTIONS(720),
    [aux_sym_comparison_operator_token16] = ACTIONS(720),
    [aux_sym_comparison_operator_token17] = ACTIONS(720),
    [aux_sym_comparison_operator_token18] = ACTIONS(720),
    [aux_sym_comparison_operator_token19] = ACTIONS(720),
    [aux_sym_comparison_operator_token20] = ACTIONS(720),
    [aux_sym_comparison_operator_token21] = ACTIONS(720),
    [aux_sym_comparison_operator_token22] = ACTIONS(720),
    [aux_sym_comparison_operator_token23] = ACTIONS(720),
    [aux_sym_comparison_operator_token24] = ACTIONS(720),
    [aux_sym_comparison_operator_token25] = ACTIONS(720),
    [aux_sym_comparison_operator_token26] = ACTIONS(720),
    [aux_sym_comparison_operator_token27] = ACTIONS(720),
    [aux_sym_comparison_operator_token28] = ACTIONS(722),
    [aux_sym_comparison_operator_token29] = ACTIONS(720),
    [aux_sym_comparison_operator_token30] = ACTIONS(720),
    [aux_sym_comparison_operator_token31] = ACTIONS(720),
    [aux_sym_comparison_operator_token32] = ACTIONS(720),
    [aux_sym_comparison_operator_token33] = ACTIONS(720),
    [aux_sym_comparison_operator_token34] = ACTIONS(722),
    [aux_sym_comparison_operator_token35] = ACTIONS(720),
    [aux_sym_comparison_operator_token36] = ACTIONS(720),
    [aux_sym_comparison_operator_token37] = ACTIONS(720),
    [aux_sym_comparison_operator_token38] = ACTIONS(720),
    [aux_sym_comparison_operator_token39] = ACTIONS(720),
    [aux_sym_comparison_operator_token40] = ACTIONS(720),
    [aux_sym_comparison_operator_token41] = ACTIONS(720),
    [aux_sym_comparison_operator_token42] = ACTIONS(720),
    [aux_sym_comparison_operator_token43] = ACTIONS(720),
    [aux_sym_comparison_operator_token44] = ACTIONS(720),
    [aux_sym_comparison_operator_token45] = ACTIONS(720),
    [aux_sym_comparison_operator_token46] = ACTIONS(720),
    [aux_sym_comparison_operator_token47] = ACTIONS(720),
    [aux_sym_comparison_operator_token48] = ACTIONS(720),
    [aux_sym_comparison_operator_token49] = ACTIONS(720),
    [aux_sym_comparison_operator_token50] = ACTIONS(720),
    [aux_sym_format_operator_token1] = ACTIONS(720),
    [anon_sym_LPAREN] = ACTIONS(720),
    [anon_sym_COMMA] = ACTIONS(720),
    [anon_sym_AMP_AMP] = ACTIONS(720),
    [anon_sym_PIPE_PIPE] = ACTIONS(720),
    [anon_sym_PIPE] = ACTIONS(722),
    [aux_sym_logical_expression_token1] = ACTIONS(720),
    [aux_sym_logical_expression_token2] = ACTIONS(720),
    [aux_sym_logical_expression_token3] = ACTIONS(720),
    [aux_sym_bitwise_expression_token1] = ACTIONS(720),
    [aux_sym_bitwise_expression_token2] = ACTIONS(720),
    [aux_sym_bitwise_expression_token3] = ACTIONS(720),
    [anon_sym_PLUS] = ACTIONS(722),
    [anon_sym_DASH] = ACTIONS(722),
    [anon_sym_SLASH] = ACTIONS(722),
    [anon_sym_BSLASH] = ACTIONS(720),
    [anon_sym_PERCENT] = ACTIONS(722),
    [anon_sym_STAR] = ACTIONS(722),
    [anon_sym_DOT_DOT] = ACTIONS(720),
    [anon_sym_PLUS_PLUS] = ACTIONS(720),
    [anon_sym_DASH_DASH] = ACTIONS(720),
    [anon_sym_DOT2] = ACTIONS(722),
    [anon_sym_COLON_COLON] = ACTIONS(720),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(720),
    [sym__statement_terminator] = ACTIONS(720),
  },
  [STATE(157)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(708),
    [anon_sym_EQ] = ACTIONS(708),
    [anon_sym_BANG_EQ] = ACTIONS(708),
    [anon_sym_PLUS_EQ] = ACTIONS(708),
    [anon_sym_STAR_EQ] = ACTIONS(708),
    [anon_sym_SLASH_EQ] = ACTIONS(708),
    [anon_sym_PERCENT_EQ] = ACTIONS(708),
    [anon_sym_DASH_EQ] = ACTIONS(708),
    [anon_sym_GT] = ACTIONS(710),
    [anon_sym_GT_GT] = ACTIONS(708),
    [anon_sym_2_GT] = ACTIONS(710),
    [anon_sym_2_GT_GT] = ACTIONS(708),
    [anon_sym_3_GT] = ACTIONS(710),
    [anon_sym_3_GT_GT] = ACTIONS(708),
    [anon_sym_4_GT] = ACTIONS(710),
    [anon_sym_4_GT_GT] = ACTIONS(708),
    [anon_sym_5_GT] = ACTIONS(710),
    [anon_sym_5_GT_GT] = ACTIONS(708),
    [anon_sym_6_GT] = ACTIONS(710),
    [anon_sym_6_GT_GT] = ACTIONS(708),
    [anon_sym_STAR_GT] = ACTIONS(710),
    [anon_sym_STAR_GT_GT] = ACTIONS(708),
    [anon_sym_LT] = ACTIONS(710),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(708),
    [anon_sym_2_GT_AMP1] = ACTIONS(708),
    [anon_sym_3_GT_AMP1] = ACTIONS(708),
    [anon_sym_4_GT_AMP1] = ACTIONS(708),
    [anon_sym_5_GT_AMP1] = ACTIONS(708),
    [anon_sym_6_GT_AMP1] = ACTIONS(708),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(708),
    [anon_sym_1_GT_AMP2] = ACTIONS(708),
    [anon_sym_3_GT_AMP2] = ACTIONS(708),
    [anon_sym_4_GT_AMP2] = ACTIONS(708),
    [anon_sym_5_GT_AMP2] = ACTIONS(708),
    [anon_sym_6_GT_AMP2] = ACTIONS(708),
    [aux_sym_comparison_operator_token1] = ACTIONS(708),
    [aux_sym_comparison_operator_token2] = ACTIONS(708),
    [aux_sym_comparison_operator_token3] = ACTIONS(708),
    [aux_sym_comparison_operator_token4] = ACTIONS(708),
    [aux_sym_comparison_operator_token5] = ACTIONS(708),
    [aux_sym_comparison_operator_token6] = ACTIONS(708),
    [aux_sym_comparison_operator_token7] = ACTIONS(708),
    [aux_sym_comparison_operator_token8] = ACTIONS(708),
    [aux_sym_comparison_operator_token9] = ACTIONS(708),
    [aux_sym_comparison_operator_token10] = ACTIONS(708),
    [aux_sym_comparison_operator_token11] = ACTIONS(708),
    [aux_sym_comparison_operator_token12] = ACTIONS(708),
    [aux_sym_comparison_operator_token13] = ACTIONS(708),
    [aux_sym_comparison_operator_token14] = ACTIONS(708),
    [aux_sym_comparison_operator_token15] = ACTIONS(708),
    [aux_sym_comparison_operator_token16] = ACTIONS(708),
    [aux_sym_comparison_operator_token17] = ACTIONS(708),
    [aux_sym_comparison_operator_token18] = ACTIONS(708),
    [aux_sym_comparison_operator_token19] = ACTIONS(708),
    [aux_sym_comparison_operator_token20] = ACTIONS(708),
    [aux_sym_comparison_operator_token21] = ACTIONS(708),
    [aux_sym_comparison_operator_token22] = ACTIONS(708),
    [aux_sym_comparison_operator_token23] = ACTIONS(708),
    [aux_sym_comparison_operator_token24] = ACTIONS(708),
    [aux_sym_comparison_operator_token25] = ACTIONS(708),
    [aux_sym_comparison_operator_token26] = ACTIONS(708),
    [aux_sym_comparison_operator_token27] = ACTIONS(708),
    [aux_sym_comparison_operator_token28] = ACTIONS(710),
    [aux_sym_comparison_operator_token29] = ACTIONS(708),
    [aux_sym_comparison_operator_token30] = ACTIONS(708),
    [aux_sym_comparison_operator_token31] = ACTIONS(708),
    [aux_sym_comparison_operator_token32] = ACTIONS(708),
    [aux_sym_comparison_operator_token33] = ACTIONS(708),
    [aux_sym_comparison_operator_token34] = ACTIONS(710),
    [aux_sym_comparison_operator_token35] = ACTIONS(708),
    [aux_sym_comparison_operator_token36] = ACTIONS(708),
    [aux_sym_comparison_operator_token37] = ACTIONS(708),
    [aux_sym_comparison_operator_token38] = ACTIONS(708),
    [aux_sym_comparison_operator_token39] = ACTIONS(708),
    [aux_sym_comparison_operator_token40] = ACTIONS(708),
    [aux_sym_comparison_operator_token41] = ACTIONS(708),
    [aux_sym_comparison_operator_token42] = ACTIONS(708),
    [aux_sym_comparison_operator_token43] = ACTIONS(708),
    [aux_sym_comparison_operator_token44] = ACTIONS(708),
    [aux_sym_comparison_operator_token45] = ACTIONS(708),
    [aux_sym_comparison_operator_token46] = ACTIONS(708),
    [aux_sym_comparison_operator_token47] = ACTIONS(708),
    [aux_sym_comparison_operator_token48] = ACTIONS(708),
    [aux_sym_comparison_operator_token49] = ACTIONS(708),
    [aux_sym_comparison_operator_token50] = ACTIONS(708),
    [aux_sym_format_operator_token1] = ACTIONS(708),
    [anon_sym_LPAREN] = ACTIONS(708),
    [anon_sym_COMMA] = ACTIONS(708),
    [anon_sym_AMP_AMP] = ACTIONS(708),
    [anon_sym_PIPE_PIPE] = ACTIONS(708),
    [anon_sym_PIPE] = ACTIONS(710),
    [aux_sym_logical_expression_token1] = ACTIONS(708),
    [aux_sym_logical_expression_token2] = ACTIONS(708),
    [aux_sym_logical_expression_token3] = ACTIONS(708),
    [aux_sym_bitwise_expression_token1] = ACTIONS(708),
    [aux_sym_bitwise_expression_token2] = ACTIONS(708),
    [aux_sym_bitwise_expression_token3] = ACTIONS(708),
    [anon_sym_PLUS] = ACTIONS(710),
    [anon_sym_DASH] = ACTIONS(710),
    [anon_sym_SLASH] = ACTIONS(710),
    [anon_sym_BSLASH] = ACTIONS(708),
    [anon_sym_PERCENT] = ACTIONS(710),
    [anon_sym_STAR] = ACTIONS(710),
    [anon_sym_DOT_DOT] = ACTIONS(708),
    [anon_sym_PLUS_PLUS] = ACTIONS(708),
    [anon_sym_DASH_DASH] = ACTIONS(708),
    [anon_sym_DOT2] = ACTIONS(710),
    [anon_sym_COLON_COLON] = ACTIONS(708),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(708),
    [sym__statement_terminator] = ACTIONS(708),
  },
  [STATE(158)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(614),
    [anon_sym_EQ] = ACTIONS(614),
    [anon_sym_BANG_EQ] = ACTIONS(614),
    [anon_sym_PLUS_EQ] = ACTIONS(614),
    [anon_sym_STAR_EQ] = ACTIONS(614),
    [anon_sym_SLASH_EQ] = ACTIONS(614),
    [anon_sym_PERCENT_EQ] = ACTIONS(614),
    [anon_sym_DASH_EQ] = ACTIONS(614),
    [anon_sym_GT] = ACTIONS(616),
    [anon_sym_GT_GT] = ACTIONS(614),
    [anon_sym_2_GT] = ACTIONS(616),
    [anon_sym_2_GT_GT] = ACTIONS(614),
    [anon_sym_3_GT] = ACTIONS(616),
    [anon_sym_3_GT_GT] = ACTIONS(614),
    [anon_sym_4_GT] = ACTIONS(616),
    [anon_sym_4_GT_GT] = ACTIONS(614),
    [anon_sym_5_GT] = ACTIONS(616),
    [anon_sym_5_GT_GT] = ACTIONS(614),
    [anon_sym_6_GT] = ACTIONS(616),
    [anon_sym_6_GT_GT] = ACTIONS(614),
    [anon_sym_STAR_GT] = ACTIONS(616),
    [anon_sym_STAR_GT_GT] = ACTIONS(614),
    [anon_sym_LT] = ACTIONS(616),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(614),
    [anon_sym_2_GT_AMP1] = ACTIONS(614),
    [anon_sym_3_GT_AMP1] = ACTIONS(614),
    [anon_sym_4_GT_AMP1] = ACTIONS(614),
    [anon_sym_5_GT_AMP1] = ACTIONS(614),
    [anon_sym_6_GT_AMP1] = ACTIONS(614),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(614),
    [anon_sym_1_GT_AMP2] = ACTIONS(614),
    [anon_sym_3_GT_AMP2] = ACTIONS(614),
    [anon_sym_4_GT_AMP2] = ACTIONS(614),
    [anon_sym_5_GT_AMP2] = ACTIONS(614),
    [anon_sym_6_GT_AMP2] = ACTIONS(614),
    [aux_sym_comparison_operator_token1] = ACTIONS(614),
    [aux_sym_comparison_operator_token2] = ACTIONS(614),
    [aux_sym_comparison_operator_token3] = ACTIONS(614),
    [aux_sym_comparison_operator_token4] = ACTIONS(614),
    [aux_sym_comparison_operator_token5] = ACTIONS(614),
    [aux_sym_comparison_operator_token6] = ACTIONS(614),
    [aux_sym_comparison_operator_token7] = ACTIONS(614),
    [aux_sym_comparison_operator_token8] = ACTIONS(614),
    [aux_sym_comparison_operator_token9] = ACTIONS(614),
    [aux_sym_comparison_operator_token10] = ACTIONS(614),
    [aux_sym_comparison_operator_token11] = ACTIONS(614),
    [aux_sym_comparison_operator_token12] = ACTIONS(614),
    [aux_sym_comparison_operator_token13] = ACTIONS(614),
    [aux_sym_comparison_operator_token14] = ACTIONS(614),
    [aux_sym_comparison_operator_token15] = ACTIONS(614),
    [aux_sym_comparison_operator_token16] = ACTIONS(614),
    [aux_sym_comparison_operator_token17] = ACTIONS(614),
    [aux_sym_comparison_operator_token18] = ACTIONS(614),
    [aux_sym_comparison_operator_token19] = ACTIONS(614),
    [aux_sym_comparison_operator_token20] = ACTIONS(614),
    [aux_sym_comparison_operator_token21] = ACTIONS(614),
    [aux_sym_comparison_operator_token22] = ACTIONS(614),
    [aux_sym_comparison_operator_token23] = ACTIONS(614),
    [aux_sym_comparison_operator_token24] = ACTIONS(614),
    [aux_sym_comparison_operator_token25] = ACTIONS(614),
    [aux_sym_comparison_operator_token26] = ACTIONS(614),
    [aux_sym_comparison_operator_token27] = ACTIONS(614),
    [aux_sym_comparison_operator_token28] = ACTIONS(616),
    [aux_sym_comparison_operator_token29] = ACTIONS(614),
    [aux_sym_comparison_operator_token30] = ACTIONS(614),
    [aux_sym_comparison_operator_token31] = ACTIONS(614),
    [aux_sym_comparison_operator_token32] = ACTIONS(614),
    [aux_sym_comparison_operator_token33] = ACTIONS(614),
    [aux_sym_comparison_operator_token34] = ACTIONS(616),
    [aux_sym_comparison_operator_token35] = ACTIONS(614),
    [aux_sym_comparison_operator_token36] = ACTIONS(614),
    [aux_sym_comparison_operator_token37] = ACTIONS(614),
    [aux_sym_comparison_operator_token38] = ACTIONS(614),
    [aux_sym_comparison_operator_token39] = ACTIONS(614),
    [aux_sym_comparison_operator_token40] = ACTIONS(614),
    [aux_sym_comparison_operator_token41] = ACTIONS(614),
    [aux_sym_comparison_operator_token42] = ACTIONS(614),
    [aux_sym_comparison_operator_token43] = ACTIONS(614),
    [aux_sym_comparison_operator_token44] = ACTIONS(614),
    [aux_sym_comparison_operator_token45] = ACTIONS(614),
    [aux_sym_comparison_operator_token46] = ACTIONS(614),
    [aux_sym_comparison_operator_token47] = ACTIONS(614),
    [aux_sym_comparison_operator_token48] = ACTIONS(614),
    [aux_sym_comparison_operator_token49] = ACTIONS(614),
    [aux_sym_comparison_operator_token50] = ACTIONS(614),
    [aux_sym_format_operator_token1] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(614),
    [anon_sym_COMMA] = ACTIONS(614),
    [anon_sym_AMP_AMP] = ACTIONS(614),
    [anon_sym_PIPE_PIPE] = ACTIONS(614),
    [anon_sym_PIPE] = ACTIONS(616),
    [aux_sym_logical_expression_token1] = ACTIONS(614),
    [aux_sym_logical_expression_token2] = ACTIONS(614),
    [aux_sym_logical_expression_token3] = ACTIONS(614),
    [aux_sym_bitwise_expression_token1] = ACTIONS(614),
    [aux_sym_bitwise_expression_token2] = ACTIONS(614),
    [aux_sym_bitwise_expression_token3] = ACTIONS(614),
    [anon_sym_PLUS] = ACTIONS(616),
    [anon_sym_DASH] = ACTIONS(616),
    [anon_sym_SLASH] = ACTIONS(616),
    [anon_sym_BSLASH] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_STAR] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(614),
    [anon_sym_PLUS_PLUS] = ACTIONS(614),
    [anon_sym_DASH_DASH] = ACTIONS(614),
    [anon_sym_DOT2] = ACTIONS(616),
    [anon_sym_COLON_COLON] = ACTIONS(614),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(614),
    [sym__statement_terminator] = ACTIONS(614),
  },
  [STATE(159)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(740),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(742),
    [anon_sym_DASH_DASH] = ACTIONS(744),
    [anon_sym_DOT2] = ACTIONS(752),
    [anon_sym_COLON_COLON] = ACTIONS(754),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(750),
  },
  [STATE(160)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(728),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(730),
    [anon_sym_DASH_DASH] = ACTIONS(732),
    [anon_sym_DOT2] = ACTIONS(756),
    [anon_sym_COLON_COLON] = ACTIONS(758),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(738),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(161)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(760),
    [anon_sym_BANG_EQ] = ACTIONS(760),
    [anon_sym_PLUS_EQ] = ACTIONS(760),
    [anon_sym_STAR_EQ] = ACTIONS(760),
    [anon_sym_SLASH_EQ] = ACTIONS(760),
    [anon_sym_PERCENT_EQ] = ACTIONS(760),
    [anon_sym_DASH_EQ] = ACTIONS(760),
    [anon_sym_GT] = ACTIONS(762),
    [anon_sym_GT_GT] = ACTIONS(760),
    [anon_sym_2_GT] = ACTIONS(762),
    [anon_sym_2_GT_GT] = ACTIONS(760),
    [anon_sym_3_GT] = ACTIONS(762),
    [anon_sym_3_GT_GT] = ACTIONS(760),
    [anon_sym_4_GT] = ACTIONS(762),
    [anon_sym_4_GT_GT] = ACTIONS(760),
    [anon_sym_5_GT] = ACTIONS(762),
    [anon_sym_5_GT_GT] = ACTIONS(760),
    [anon_sym_6_GT] = ACTIONS(762),
    [anon_sym_6_GT_GT] = ACTIONS(760),
    [anon_sym_STAR_GT] = ACTIONS(762),
    [anon_sym_STAR_GT_GT] = ACTIONS(760),
    [anon_sym_LT] = ACTIONS(762),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(760),
    [anon_sym_2_GT_AMP1] = ACTIONS(760),
    [anon_sym_3_GT_AMP1] = ACTIONS(760),
    [anon_sym_4_GT_AMP1] = ACTIONS(760),
    [anon_sym_5_GT_AMP1] = ACTIONS(760),
    [anon_sym_6_GT_AMP1] = ACTIONS(760),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(760),
    [anon_sym_1_GT_AMP2] = ACTIONS(760),
    [anon_sym_3_GT_AMP2] = ACTIONS(760),
    [anon_sym_4_GT_AMP2] = ACTIONS(760),
    [anon_sym_5_GT_AMP2] = ACTIONS(760),
    [anon_sym_6_GT_AMP2] = ACTIONS(760),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_RPAREN] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [anon_sym_PIPE] = ACTIONS(762),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(762),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(762),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(762),
    [anon_sym_STAR] = ACTIONS(762),
    [anon_sym_DOT_DOT] = ACTIONS(760),
    [anon_sym_RBRACK] = ACTIONS(760),
  },
  [STATE(162)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(764),
    [anon_sym_BANG_EQ] = ACTIONS(764),
    [anon_sym_PLUS_EQ] = ACTIONS(764),
    [anon_sym_STAR_EQ] = ACTIONS(764),
    [anon_sym_SLASH_EQ] = ACTIONS(764),
    [anon_sym_PERCENT_EQ] = ACTIONS(764),
    [anon_sym_DASH_EQ] = ACTIONS(764),
    [anon_sym_GT] = ACTIONS(766),
    [anon_sym_GT_GT] = ACTIONS(764),
    [anon_sym_2_GT] = ACTIONS(766),
    [anon_sym_2_GT_GT] = ACTIONS(764),
    [anon_sym_3_GT] = ACTIONS(766),
    [anon_sym_3_GT_GT] = ACTIONS(764),
    [anon_sym_4_GT] = ACTIONS(766),
    [anon_sym_4_GT_GT] = ACTIONS(764),
    [anon_sym_5_GT] = ACTIONS(766),
    [anon_sym_5_GT_GT] = ACTIONS(764),
    [anon_sym_6_GT] = ACTIONS(766),
    [anon_sym_6_GT_GT] = ACTIONS(764),
    [anon_sym_STAR_GT] = ACTIONS(766),
    [anon_sym_STAR_GT_GT] = ACTIONS(764),
    [anon_sym_LT] = ACTIONS(766),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(764),
    [anon_sym_2_GT_AMP1] = ACTIONS(764),
    [anon_sym_3_GT_AMP1] = ACTIONS(764),
    [anon_sym_4_GT_AMP1] = ACTIONS(764),
    [anon_sym_5_GT_AMP1] = ACTIONS(764),
    [anon_sym_6_GT_AMP1] = ACTIONS(764),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(764),
    [anon_sym_1_GT_AMP2] = ACTIONS(764),
    [anon_sym_3_GT_AMP2] = ACTIONS(764),
    [anon_sym_4_GT_AMP2] = ACTIONS(764),
    [anon_sym_5_GT_AMP2] = ACTIONS(764),
    [anon_sym_6_GT_AMP2] = ACTIONS(764),
    [aux_sym_comparison_operator_token1] = ACTIONS(764),
    [aux_sym_comparison_operator_token2] = ACTIONS(764),
    [aux_sym_comparison_operator_token3] = ACTIONS(764),
    [aux_sym_comparison_operator_token4] = ACTIONS(764),
    [aux_sym_comparison_operator_token5] = ACTIONS(764),
    [aux_sym_comparison_operator_token6] = ACTIONS(764),
    [aux_sym_comparison_operator_token7] = ACTIONS(764),
    [aux_sym_comparison_operator_token8] = ACTIONS(764),
    [aux_sym_comparison_operator_token9] = ACTIONS(764),
    [aux_sym_comparison_operator_token10] = ACTIONS(764),
    [aux_sym_comparison_operator_token11] = ACTIONS(764),
    [aux_sym_comparison_operator_token12] = ACTIONS(764),
    [aux_sym_comparison_operator_token13] = ACTIONS(764),
    [aux_sym_comparison_operator_token14] = ACTIONS(764),
    [aux_sym_comparison_operator_token15] = ACTIONS(764),
    [aux_sym_comparison_operator_token16] = ACTIONS(764),
    [aux_sym_comparison_operator_token17] = ACTIONS(764),
    [aux_sym_comparison_operator_token18] = ACTIONS(764),
    [aux_sym_comparison_operator_token19] = ACTIONS(764),
    [aux_sym_comparison_operator_token20] = ACTIONS(764),
    [aux_sym_comparison_operator_token21] = ACTIONS(764),
    [aux_sym_comparison_operator_token22] = ACTIONS(764),
    [aux_sym_comparison_operator_token23] = ACTIONS(764),
    [aux_sym_comparison_operator_token24] = ACTIONS(764),
    [aux_sym_comparison_operator_token25] = ACTIONS(764),
    [aux_sym_comparison_operator_token26] = ACTIONS(764),
    [aux_sym_comparison_operator_token27] = ACTIONS(764),
    [aux_sym_comparison_operator_token28] = ACTIONS(766),
    [aux_sym_comparison_operator_token29] = ACTIONS(764),
    [aux_sym_comparison_operator_token30] = ACTIONS(764),
    [aux_sym_comparison_operator_token31] = ACTIONS(764),
    [aux_sym_comparison_operator_token32] = ACTIONS(764),
    [aux_sym_comparison_operator_token33] = ACTIONS(764),
    [aux_sym_comparison_operator_token34] = ACTIONS(766),
    [aux_sym_comparison_operator_token35] = ACTIONS(764),
    [aux_sym_comparison_operator_token36] = ACTIONS(764),
    [aux_sym_comparison_operator_token37] = ACTIONS(764),
    [aux_sym_comparison_operator_token38] = ACTIONS(764),
    [aux_sym_comparison_operator_token39] = ACTIONS(764),
    [aux_sym_comparison_operator_token40] = ACTIONS(764),
    [aux_sym_comparison_operator_token41] = ACTIONS(764),
    [aux_sym_comparison_operator_token42] = ACTIONS(764),
    [aux_sym_comparison_operator_token43] = ACTIONS(764),
    [aux_sym_comparison_operator_token44] = ACTIONS(764),
    [aux_sym_comparison_operator_token45] = ACTIONS(764),
    [aux_sym_comparison_operator_token46] = ACTIONS(764),
    [aux_sym_comparison_operator_token47] = ACTIONS(764),
    [aux_sym_comparison_operator_token48] = ACTIONS(764),
    [aux_sym_comparison_operator_token49] = ACTIONS(764),
    [aux_sym_comparison_operator_token50] = ACTIONS(764),
    [aux_sym_format_operator_token1] = ACTIONS(764),
    [anon_sym_RPAREN] = ACTIONS(764),
    [anon_sym_COMMA] = ACTIONS(764),
    [anon_sym_AMP_AMP] = ACTIONS(764),
    [anon_sym_PIPE_PIPE] = ACTIONS(764),
    [anon_sym_PIPE] = ACTIONS(766),
    [aux_sym_logical_expression_token1] = ACTIONS(764),
    [aux_sym_logical_expression_token2] = ACTIONS(764),
    [aux_sym_logical_expression_token3] = ACTIONS(764),
    [aux_sym_bitwise_expression_token1] = ACTIONS(764),
    [aux_sym_bitwise_expression_token2] = ACTIONS(764),
    [aux_sym_bitwise_expression_token3] = ACTIONS(764),
    [anon_sym_PLUS] = ACTIONS(766),
    [anon_sym_DASH] = ACTIONS(766),
    [anon_sym_SLASH] = ACTIONS(766),
    [anon_sym_BSLASH] = ACTIONS(764),
    [anon_sym_PERCENT] = ACTIONS(766),
    [anon_sym_STAR] = ACTIONS(766),
    [anon_sym_DOT_DOT] = ACTIONS(764),
    [anon_sym_RBRACK] = ACTIONS(764),
  },
  [STATE(163)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(768),
    [anon_sym_BANG_EQ] = ACTIONS(768),
    [anon_sym_PLUS_EQ] = ACTIONS(768),
    [anon_sym_STAR_EQ] = ACTIONS(768),
    [anon_sym_SLASH_EQ] = ACTIONS(768),
    [anon_sym_PERCENT_EQ] = ACTIONS(768),
    [anon_sym_DASH_EQ] = ACTIONS(768),
    [anon_sym_GT] = ACTIONS(770),
    [anon_sym_GT_GT] = ACTIONS(768),
    [anon_sym_2_GT] = ACTIONS(770),
    [anon_sym_2_GT_GT] = ACTIONS(768),
    [anon_sym_3_GT] = ACTIONS(770),
    [anon_sym_3_GT_GT] = ACTIONS(768),
    [anon_sym_4_GT] = ACTIONS(770),
    [anon_sym_4_GT_GT] = ACTIONS(768),
    [anon_sym_5_GT] = ACTIONS(770),
    [anon_sym_5_GT_GT] = ACTIONS(768),
    [anon_sym_6_GT] = ACTIONS(770),
    [anon_sym_6_GT_GT] = ACTIONS(768),
    [anon_sym_STAR_GT] = ACTIONS(770),
    [anon_sym_STAR_GT_GT] = ACTIONS(768),
    [anon_sym_LT] = ACTIONS(770),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(768),
    [anon_sym_2_GT_AMP1] = ACTIONS(768),
    [anon_sym_3_GT_AMP1] = ACTIONS(768),
    [anon_sym_4_GT_AMP1] = ACTIONS(768),
    [anon_sym_5_GT_AMP1] = ACTIONS(768),
    [anon_sym_6_GT_AMP1] = ACTIONS(768),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(768),
    [anon_sym_1_GT_AMP2] = ACTIONS(768),
    [anon_sym_3_GT_AMP2] = ACTIONS(768),
    [anon_sym_4_GT_AMP2] = ACTIONS(768),
    [anon_sym_5_GT_AMP2] = ACTIONS(768),
    [anon_sym_6_GT_AMP2] = ACTIONS(768),
    [aux_sym_comparison_operator_token1] = ACTIONS(768),
    [aux_sym_comparison_operator_token2] = ACTIONS(768),
    [aux_sym_comparison_operator_token3] = ACTIONS(768),
    [aux_sym_comparison_operator_token4] = ACTIONS(768),
    [aux_sym_comparison_operator_token5] = ACTIONS(768),
    [aux_sym_comparison_operator_token6] = ACTIONS(768),
    [aux_sym_comparison_operator_token7] = ACTIONS(768),
    [aux_sym_comparison_operator_token8] = ACTIONS(768),
    [aux_sym_comparison_operator_token9] = ACTIONS(768),
    [aux_sym_comparison_operator_token10] = ACTIONS(768),
    [aux_sym_comparison_operator_token11] = ACTIONS(768),
    [aux_sym_comparison_operator_token12] = ACTIONS(768),
    [aux_sym_comparison_operator_token13] = ACTIONS(768),
    [aux_sym_comparison_operator_token14] = ACTIONS(768),
    [aux_sym_comparison_operator_token15] = ACTIONS(768),
    [aux_sym_comparison_operator_token16] = ACTIONS(768),
    [aux_sym_comparison_operator_token17] = ACTIONS(768),
    [aux_sym_comparison_operator_token18] = ACTIONS(768),
    [aux_sym_comparison_operator_token19] = ACTIONS(768),
    [aux_sym_comparison_operator_token20] = ACTIONS(768),
    [aux_sym_comparison_operator_token21] = ACTIONS(768),
    [aux_sym_comparison_operator_token22] = ACTIONS(768),
    [aux_sym_comparison_operator_token23] = ACTIONS(768),
    [aux_sym_comparison_operator_token24] = ACTIONS(768),
    [aux_sym_comparison_operator_token25] = ACTIONS(768),
    [aux_sym_comparison_operator_token26] = ACTIONS(768),
    [aux_sym_comparison_operator_token27] = ACTIONS(768),
    [aux_sym_comparison_operator_token28] = ACTIONS(770),
    [aux_sym_comparison_operator_token29] = ACTIONS(768),
    [aux_sym_comparison_operator_token30] = ACTIONS(768),
    [aux_sym_comparison_operator_token31] = ACTIONS(768),
    [aux_sym_comparison_operator_token32] = ACTIONS(768),
    [aux_sym_comparison_operator_token33] = ACTIONS(768),
    [aux_sym_comparison_operator_token34] = ACTIONS(770),
    [aux_sym_comparison_operator_token35] = ACTIONS(768),
    [aux_sym_comparison_operator_token36] = ACTIONS(768),
    [aux_sym_comparison_operator_token37] = ACTIONS(768),
    [aux_sym_comparison_operator_token38] = ACTIONS(768),
    [aux_sym_comparison_operator_token39] = ACTIONS(768),
    [aux_sym_comparison_operator_token40] = ACTIONS(768),
    [aux_sym_comparison_operator_token41] = ACTIONS(768),
    [aux_sym_comparison_operator_token42] = ACTIONS(768),
    [aux_sym_comparison_operator_token43] = ACTIONS(768),
    [aux_sym_comparison_operator_token44] = ACTIONS(768),
    [aux_sym_comparison_operator_token45] = ACTIONS(768),
    [aux_sym_comparison_operator_token46] = ACTIONS(768),
    [aux_sym_comparison_operator_token47] = ACTIONS(768),
    [aux_sym_comparison_operator_token48] = ACTIONS(768),
    [aux_sym_comparison_operator_token49] = ACTIONS(768),
    [aux_sym_comparison_operator_token50] = ACTIONS(768),
    [aux_sym_format_operator_token1] = ACTIONS(768),
    [anon_sym_RPAREN] = ACTIONS(768),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_AMP_AMP] = ACTIONS(768),
    [anon_sym_PIPE_PIPE] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(770),
    [aux_sym_logical_expression_token1] = ACTIONS(768),
    [aux_sym_logical_expression_token2] = ACTIONS(768),
    [aux_sym_logical_expression_token3] = ACTIONS(768),
    [aux_sym_bitwise_expression_token1] = ACTIONS(768),
    [aux_sym_bitwise_expression_token2] = ACTIONS(768),
    [aux_sym_bitwise_expression_token3] = ACTIONS(768),
    [anon_sym_PLUS] = ACTIONS(770),
    [anon_sym_DASH] = ACTIONS(770),
    [anon_sym_SLASH] = ACTIONS(770),
    [anon_sym_BSLASH] = ACTIONS(768),
    [anon_sym_PERCENT] = ACTIONS(770),
    [anon_sym_STAR] = ACTIONS(770),
    [anon_sym_DOT_DOT] = ACTIONS(768),
    [anon_sym_RBRACK] = ACTIONS(768),
  },
  [STATE(164)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(172),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(772),
    [anon_sym_BANG_EQ] = ACTIONS(772),
    [anon_sym_PLUS_EQ] = ACTIONS(772),
    [anon_sym_STAR_EQ] = ACTIONS(772),
    [anon_sym_SLASH_EQ] = ACTIONS(772),
    [anon_sym_PERCENT_EQ] = ACTIONS(772),
    [anon_sym_DASH_EQ] = ACTIONS(772),
    [anon_sym_GT] = ACTIONS(774),
    [anon_sym_GT_GT] = ACTIONS(772),
    [anon_sym_2_GT] = ACTIONS(774),
    [anon_sym_2_GT_GT] = ACTIONS(772),
    [anon_sym_3_GT] = ACTIONS(774),
    [anon_sym_3_GT_GT] = ACTIONS(772),
    [anon_sym_4_GT] = ACTIONS(774),
    [anon_sym_4_GT_GT] = ACTIONS(772),
    [anon_sym_5_GT] = ACTIONS(774),
    [anon_sym_5_GT_GT] = ACTIONS(772),
    [anon_sym_6_GT] = ACTIONS(774),
    [anon_sym_6_GT_GT] = ACTIONS(772),
    [anon_sym_STAR_GT] = ACTIONS(774),
    [anon_sym_STAR_GT_GT] = ACTIONS(772),
    [anon_sym_LT] = ACTIONS(774),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(772),
    [anon_sym_2_GT_AMP1] = ACTIONS(772),
    [anon_sym_3_GT_AMP1] = ACTIONS(772),
    [anon_sym_4_GT_AMP1] = ACTIONS(772),
    [anon_sym_5_GT_AMP1] = ACTIONS(772),
    [anon_sym_6_GT_AMP1] = ACTIONS(772),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(772),
    [anon_sym_1_GT_AMP2] = ACTIONS(772),
    [anon_sym_3_GT_AMP2] = ACTIONS(772),
    [anon_sym_4_GT_AMP2] = ACTIONS(772),
    [anon_sym_5_GT_AMP2] = ACTIONS(772),
    [anon_sym_6_GT_AMP2] = ACTIONS(772),
    [aux_sym_comparison_operator_token1] = ACTIONS(772),
    [aux_sym_comparison_operator_token2] = ACTIONS(772),
    [aux_sym_comparison_operator_token3] = ACTIONS(772),
    [aux_sym_comparison_operator_token4] = ACTIONS(772),
    [aux_sym_comparison_operator_token5] = ACTIONS(772),
    [aux_sym_comparison_operator_token6] = ACTIONS(772),
    [aux_sym_comparison_operator_token7] = ACTIONS(772),
    [aux_sym_comparison_operator_token8] = ACTIONS(772),
    [aux_sym_comparison_operator_token9] = ACTIONS(772),
    [aux_sym_comparison_operator_token10] = ACTIONS(772),
    [aux_sym_comparison_operator_token11] = ACTIONS(772),
    [aux_sym_comparison_operator_token12] = ACTIONS(772),
    [aux_sym_comparison_operator_token13] = ACTIONS(772),
    [aux_sym_comparison_operator_token14] = ACTIONS(772),
    [aux_sym_comparison_operator_token15] = ACTIONS(772),
    [aux_sym_comparison_operator_token16] = ACTIONS(772),
    [aux_sym_comparison_operator_token17] = ACTIONS(772),
    [aux_sym_comparison_operator_token18] = ACTIONS(772),
    [aux_sym_comparison_operator_token19] = ACTIONS(772),
    [aux_sym_comparison_operator_token20] = ACTIONS(772),
    [aux_sym_comparison_operator_token21] = ACTIONS(772),
    [aux_sym_comparison_operator_token22] = ACTIONS(772),
    [aux_sym_comparison_operator_token23] = ACTIONS(772),
    [aux_sym_comparison_operator_token24] = ACTIONS(772),
    [aux_sym_comparison_operator_token25] = ACTIONS(772),
    [aux_sym_comparison_operator_token26] = ACTIONS(772),
    [aux_sym_comparison_operator_token27] = ACTIONS(772),
    [aux_sym_comparison_operator_token28] = ACTIONS(774),
    [aux_sym_comparison_operator_token29] = ACTIONS(772),
    [aux_sym_comparison_operator_token30] = ACTIONS(772),
    [aux_sym_comparison_operator_token31] = ACTIONS(772),
    [aux_sym_comparison_operator_token32] = ACTIONS(772),
    [aux_sym_comparison_operator_token33] = ACTIONS(772),
    [aux_sym_comparison_operator_token34] = ACTIONS(774),
    [aux_sym_comparison_operator_token35] = ACTIONS(772),
    [aux_sym_comparison_operator_token36] = ACTIONS(772),
    [aux_sym_comparison_operator_token37] = ACTIONS(772),
    [aux_sym_comparison_operator_token38] = ACTIONS(772),
    [aux_sym_comparison_operator_token39] = ACTIONS(772),
    [aux_sym_comparison_operator_token40] = ACTIONS(772),
    [aux_sym_comparison_operator_token41] = ACTIONS(772),
    [aux_sym_comparison_operator_token42] = ACTIONS(772),
    [aux_sym_comparison_operator_token43] = ACTIONS(772),
    [aux_sym_comparison_operator_token44] = ACTIONS(772),
    [aux_sym_comparison_operator_token45] = ACTIONS(772),
    [aux_sym_comparison_operator_token46] = ACTIONS(772),
    [aux_sym_comparison_operator_token47] = ACTIONS(772),
    [aux_sym_comparison_operator_token48] = ACTIONS(772),
    [aux_sym_comparison_operator_token49] = ACTIONS(772),
    [aux_sym_comparison_operator_token50] = ACTIONS(772),
    [aux_sym_format_operator_token1] = ACTIONS(772),
    [anon_sym_RPAREN] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(776),
    [anon_sym_AMP_AMP] = ACTIONS(772),
    [anon_sym_PIPE_PIPE] = ACTIONS(772),
    [anon_sym_PIPE] = ACTIONS(774),
    [aux_sym_logical_expression_token1] = ACTIONS(772),
    [aux_sym_logical_expression_token2] = ACTIONS(772),
    [aux_sym_logical_expression_token3] = ACTIONS(772),
    [aux_sym_bitwise_expression_token1] = ACTIONS(772),
    [aux_sym_bitwise_expression_token2] = ACTIONS(772),
    [aux_sym_bitwise_expression_token3] = ACTIONS(772),
    [anon_sym_PLUS] = ACTIONS(774),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(774),
    [anon_sym_BSLASH] = ACTIONS(772),
    [anon_sym_PERCENT] = ACTIONS(774),
    [anon_sym_STAR] = ACTIONS(774),
    [anon_sym_DOT_DOT] = ACTIONS(772),
  },
  [STATE(165)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_RBRACK] = ACTIONS(654),
  },
  [STATE(166)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(658),
    [anon_sym_BANG_EQ] = ACTIONS(658),
    [anon_sym_PLUS_EQ] = ACTIONS(658),
    [anon_sym_STAR_EQ] = ACTIONS(658),
    [anon_sym_SLASH_EQ] = ACTIONS(658),
    [anon_sym_PERCENT_EQ] = ACTIONS(658),
    [anon_sym_DASH_EQ] = ACTIONS(658),
    [anon_sym_GT] = ACTIONS(660),
    [anon_sym_GT_GT] = ACTIONS(658),
    [anon_sym_2_GT] = ACTIONS(660),
    [anon_sym_2_GT_GT] = ACTIONS(658),
    [anon_sym_3_GT] = ACTIONS(660),
    [anon_sym_3_GT_GT] = ACTIONS(658),
    [anon_sym_4_GT] = ACTIONS(660),
    [anon_sym_4_GT_GT] = ACTIONS(658),
    [anon_sym_5_GT] = ACTIONS(660),
    [anon_sym_5_GT_GT] = ACTIONS(658),
    [anon_sym_6_GT] = ACTIONS(660),
    [anon_sym_6_GT_GT] = ACTIONS(658),
    [anon_sym_STAR_GT] = ACTIONS(660),
    [anon_sym_STAR_GT_GT] = ACTIONS(658),
    [anon_sym_LT] = ACTIONS(660),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(658),
    [anon_sym_2_GT_AMP1] = ACTIONS(658),
    [anon_sym_3_GT_AMP1] = ACTIONS(658),
    [anon_sym_4_GT_AMP1] = ACTIONS(658),
    [anon_sym_5_GT_AMP1] = ACTIONS(658),
    [anon_sym_6_GT_AMP1] = ACTIONS(658),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(658),
    [anon_sym_1_GT_AMP2] = ACTIONS(658),
    [anon_sym_3_GT_AMP2] = ACTIONS(658),
    [anon_sym_4_GT_AMP2] = ACTIONS(658),
    [anon_sym_5_GT_AMP2] = ACTIONS(658),
    [anon_sym_6_GT_AMP2] = ACTIONS(658),
    [aux_sym_comparison_operator_token1] = ACTIONS(658),
    [aux_sym_comparison_operator_token2] = ACTIONS(658),
    [aux_sym_comparison_operator_token3] = ACTIONS(658),
    [aux_sym_comparison_operator_token4] = ACTIONS(658),
    [aux_sym_comparison_operator_token5] = ACTIONS(658),
    [aux_sym_comparison_operator_token6] = ACTIONS(658),
    [aux_sym_comparison_operator_token7] = ACTIONS(658),
    [aux_sym_comparison_operator_token8] = ACTIONS(658),
    [aux_sym_comparison_operator_token9] = ACTIONS(658),
    [aux_sym_comparison_operator_token10] = ACTIONS(658),
    [aux_sym_comparison_operator_token11] = ACTIONS(658),
    [aux_sym_comparison_operator_token12] = ACTIONS(658),
    [aux_sym_comparison_operator_token13] = ACTIONS(658),
    [aux_sym_comparison_operator_token14] = ACTIONS(658),
    [aux_sym_comparison_operator_token15] = ACTIONS(658),
    [aux_sym_comparison_operator_token16] = ACTIONS(658),
    [aux_sym_comparison_operator_token17] = ACTIONS(658),
    [aux_sym_comparison_operator_token18] = ACTIONS(658),
    [aux_sym_comparison_operator_token19] = ACTIONS(658),
    [aux_sym_comparison_operator_token20] = ACTIONS(658),
    [aux_sym_comparison_operator_token21] = ACTIONS(658),
    [aux_sym_comparison_operator_token22] = ACTIONS(658),
    [aux_sym_comparison_operator_token23] = ACTIONS(658),
    [aux_sym_comparison_operator_token24] = ACTIONS(658),
    [aux_sym_comparison_operator_token25] = ACTIONS(658),
    [aux_sym_comparison_operator_token26] = ACTIONS(658),
    [aux_sym_comparison_operator_token27] = ACTIONS(658),
    [aux_sym_comparison_operator_token28] = ACTIONS(660),
    [aux_sym_comparison_operator_token29] = ACTIONS(658),
    [aux_sym_comparison_operator_token30] = ACTIONS(658),
    [aux_sym_comparison_operator_token31] = ACTIONS(658),
    [aux_sym_comparison_operator_token32] = ACTIONS(658),
    [aux_sym_comparison_operator_token33] = ACTIONS(658),
    [aux_sym_comparison_operator_token34] = ACTIONS(660),
    [aux_sym_comparison_operator_token35] = ACTIONS(658),
    [aux_sym_comparison_operator_token36] = ACTIONS(658),
    [aux_sym_comparison_operator_token37] = ACTIONS(658),
    [aux_sym_comparison_operator_token38] = ACTIONS(658),
    [aux_sym_comparison_operator_token39] = ACTIONS(658),
    [aux_sym_comparison_operator_token40] = ACTIONS(658),
    [aux_sym_comparison_operator_token41] = ACTIONS(658),
    [aux_sym_comparison_operator_token42] = ACTIONS(658),
    [aux_sym_comparison_operator_token43] = ACTIONS(658),
    [aux_sym_comparison_operator_token44] = ACTIONS(658),
    [aux_sym_comparison_operator_token45] = ACTIONS(658),
    [aux_sym_comparison_operator_token46] = ACTIONS(658),
    [aux_sym_comparison_operator_token47] = ACTIONS(658),
    [aux_sym_comparison_operator_token48] = ACTIONS(658),
    [aux_sym_comparison_operator_token49] = ACTIONS(658),
    [aux_sym_comparison_operator_token50] = ACTIONS(658),
    [aux_sym_format_operator_token1] = ACTIONS(658),
    [anon_sym_RPAREN] = ACTIONS(658),
    [anon_sym_COMMA] = ACTIONS(658),
    [anon_sym_AMP_AMP] = ACTIONS(658),
    [anon_sym_PIPE_PIPE] = ACTIONS(658),
    [anon_sym_PIPE] = ACTIONS(660),
    [aux_sym_logical_expression_token1] = ACTIONS(658),
    [aux_sym_logical_expression_token2] = ACTIONS(658),
    [aux_sym_logical_expression_token3] = ACTIONS(658),
    [aux_sym_bitwise_expression_token1] = ACTIONS(658),
    [aux_sym_bitwise_expression_token2] = ACTIONS(658),
    [aux_sym_bitwise_expression_token3] = ACTIONS(658),
    [anon_sym_PLUS] = ACTIONS(660),
    [anon_sym_DASH] = ACTIONS(660),
    [anon_sym_SLASH] = ACTIONS(660),
    [anon_sym_BSLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(660),
    [anon_sym_STAR] = ACTIONS(660),
    [anon_sym_DOT_DOT] = ACTIONS(658),
    [anon_sym_RBRACK] = ACTIONS(658),
  },
  [STATE(167)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(724),
    [anon_sym_BANG_EQ] = ACTIONS(724),
    [anon_sym_PLUS_EQ] = ACTIONS(724),
    [anon_sym_STAR_EQ] = ACTIONS(724),
    [anon_sym_SLASH_EQ] = ACTIONS(724),
    [anon_sym_PERCENT_EQ] = ACTIONS(724),
    [anon_sym_DASH_EQ] = ACTIONS(724),
    [anon_sym_GT] = ACTIONS(726),
    [anon_sym_GT_GT] = ACTIONS(724),
    [anon_sym_2_GT] = ACTIONS(726),
    [anon_sym_2_GT_GT] = ACTIONS(724),
    [anon_sym_3_GT] = ACTIONS(726),
    [anon_sym_3_GT_GT] = ACTIONS(724),
    [anon_sym_4_GT] = ACTIONS(726),
    [anon_sym_4_GT_GT] = ACTIONS(724),
    [anon_sym_5_GT] = ACTIONS(726),
    [anon_sym_5_GT_GT] = ACTIONS(724),
    [anon_sym_6_GT] = ACTIONS(726),
    [anon_sym_6_GT_GT] = ACTIONS(724),
    [anon_sym_STAR_GT] = ACTIONS(726),
    [anon_sym_STAR_GT_GT] = ACTIONS(724),
    [anon_sym_LT] = ACTIONS(726),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(724),
    [anon_sym_2_GT_AMP1] = ACTIONS(724),
    [anon_sym_3_GT_AMP1] = ACTIONS(724),
    [anon_sym_4_GT_AMP1] = ACTIONS(724),
    [anon_sym_5_GT_AMP1] = ACTIONS(724),
    [anon_sym_6_GT_AMP1] = ACTIONS(724),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(724),
    [anon_sym_1_GT_AMP2] = ACTIONS(724),
    [anon_sym_3_GT_AMP2] = ACTIONS(724),
    [anon_sym_4_GT_AMP2] = ACTIONS(724),
    [anon_sym_5_GT_AMP2] = ACTIONS(724),
    [anon_sym_6_GT_AMP2] = ACTIONS(724),
    [aux_sym_comparison_operator_token1] = ACTIONS(724),
    [aux_sym_comparison_operator_token2] = ACTIONS(724),
    [aux_sym_comparison_operator_token3] = ACTIONS(724),
    [aux_sym_comparison_operator_token4] = ACTIONS(724),
    [aux_sym_comparison_operator_token5] = ACTIONS(724),
    [aux_sym_comparison_operator_token6] = ACTIONS(724),
    [aux_sym_comparison_operator_token7] = ACTIONS(724),
    [aux_sym_comparison_operator_token8] = ACTIONS(724),
    [aux_sym_comparison_operator_token9] = ACTIONS(724),
    [aux_sym_comparison_operator_token10] = ACTIONS(724),
    [aux_sym_comparison_operator_token11] = ACTIONS(724),
    [aux_sym_comparison_operator_token12] = ACTIONS(724),
    [aux_sym_comparison_operator_token13] = ACTIONS(724),
    [aux_sym_comparison_operator_token14] = ACTIONS(724),
    [aux_sym_comparison_operator_token15] = ACTIONS(724),
    [aux_sym_comparison_operator_token16] = ACTIONS(724),
    [aux_sym_comparison_operator_token17] = ACTIONS(724),
    [aux_sym_comparison_operator_token18] = ACTIONS(724),
    [aux_sym_comparison_operator_token19] = ACTIONS(724),
    [aux_sym_comparison_operator_token20] = ACTIONS(724),
    [aux_sym_comparison_operator_token21] = ACTIONS(724),
    [aux_sym_comparison_operator_token22] = ACTIONS(724),
    [aux_sym_comparison_operator_token23] = ACTIONS(724),
    [aux_sym_comparison_operator_token24] = ACTIONS(724),
    [aux_sym_comparison_operator_token25] = ACTIONS(724),
    [aux_sym_comparison_operator_token26] = ACTIONS(724),
    [aux_sym_comparison_operator_token27] = ACTIONS(724),
    [aux_sym_comparison_operator_token28] = ACTIONS(726),
    [aux_sym_comparison_operator_token29] = ACTIONS(724),
    [aux_sym_comparison_operator_token30] = ACTIONS(724),
    [aux_sym_comparison_operator_token31] = ACTIONS(724),
    [aux_sym_comparison_operator_token32] = ACTIONS(724),
    [aux_sym_comparison_operator_token33] = ACTIONS(724),
    [aux_sym_comparison_operator_token34] = ACTIONS(726),
    [aux_sym_comparison_operator_token35] = ACTIONS(724),
    [aux_sym_comparison_operator_token36] = ACTIONS(724),
    [aux_sym_comparison_operator_token37] = ACTIONS(724),
    [aux_sym_comparison_operator_token38] = ACTIONS(724),
    [aux_sym_comparison_operator_token39] = ACTIONS(724),
    [aux_sym_comparison_operator_token40] = ACTIONS(724),
    [aux_sym_comparison_operator_token41] = ACTIONS(724),
    [aux_sym_comparison_operator_token42] = ACTIONS(724),
    [aux_sym_comparison_operator_token43] = ACTIONS(724),
    [aux_sym_comparison_operator_token44] = ACTIONS(724),
    [aux_sym_comparison_operator_token45] = ACTIONS(724),
    [aux_sym_comparison_operator_token46] = ACTIONS(724),
    [aux_sym_comparison_operator_token47] = ACTIONS(724),
    [aux_sym_comparison_operator_token48] = ACTIONS(724),
    [aux_sym_comparison_operator_token49] = ACTIONS(724),
    [aux_sym_comparison_operator_token50] = ACTIONS(724),
    [aux_sym_format_operator_token1] = ACTIONS(724),
    [anon_sym_RPAREN] = ACTIONS(724),
    [anon_sym_COMMA] = ACTIONS(724),
    [anon_sym_AMP_AMP] = ACTIONS(724),
    [anon_sym_PIPE_PIPE] = ACTIONS(724),
    [anon_sym_PIPE] = ACTIONS(726),
    [aux_sym_logical_expression_token1] = ACTIONS(724),
    [aux_sym_logical_expression_token2] = ACTIONS(724),
    [aux_sym_logical_expression_token3] = ACTIONS(724),
    [aux_sym_bitwise_expression_token1] = ACTIONS(724),
    [aux_sym_bitwise_expression_token2] = ACTIONS(724),
    [aux_sym_bitwise_expression_token3] = ACTIONS(724),
    [anon_sym_PLUS] = ACTIONS(726),
    [anon_sym_DASH] = ACTIONS(726),
    [anon_sym_SLASH] = ACTIONS(726),
    [anon_sym_BSLASH] = ACTIONS(724),
    [anon_sym_PERCENT] = ACTIONS(726),
    [anon_sym_STAR] = ACTIONS(726),
    [anon_sym_DOT_DOT] = ACTIONS(724),
    [anon_sym_RBRACK] = ACTIONS(724),
  },
  [STATE(168)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(168),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(760),
    [anon_sym_BANG_EQ] = ACTIONS(760),
    [anon_sym_PLUS_EQ] = ACTIONS(760),
    [anon_sym_STAR_EQ] = ACTIONS(760),
    [anon_sym_SLASH_EQ] = ACTIONS(760),
    [anon_sym_PERCENT_EQ] = ACTIONS(760),
    [anon_sym_DASH_EQ] = ACTIONS(760),
    [anon_sym_GT] = ACTIONS(762),
    [anon_sym_GT_GT] = ACTIONS(760),
    [anon_sym_2_GT] = ACTIONS(762),
    [anon_sym_2_GT_GT] = ACTIONS(760),
    [anon_sym_3_GT] = ACTIONS(762),
    [anon_sym_3_GT_GT] = ACTIONS(760),
    [anon_sym_4_GT] = ACTIONS(762),
    [anon_sym_4_GT_GT] = ACTIONS(760),
    [anon_sym_5_GT] = ACTIONS(762),
    [anon_sym_5_GT_GT] = ACTIONS(760),
    [anon_sym_6_GT] = ACTIONS(762),
    [anon_sym_6_GT_GT] = ACTIONS(760),
    [anon_sym_STAR_GT] = ACTIONS(762),
    [anon_sym_STAR_GT_GT] = ACTIONS(760),
    [anon_sym_LT] = ACTIONS(762),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(760),
    [anon_sym_2_GT_AMP1] = ACTIONS(760),
    [anon_sym_3_GT_AMP1] = ACTIONS(760),
    [anon_sym_4_GT_AMP1] = ACTIONS(760),
    [anon_sym_5_GT_AMP1] = ACTIONS(760),
    [anon_sym_6_GT_AMP1] = ACTIONS(760),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(760),
    [anon_sym_1_GT_AMP2] = ACTIONS(760),
    [anon_sym_3_GT_AMP2] = ACTIONS(760),
    [anon_sym_4_GT_AMP2] = ACTIONS(760),
    [anon_sym_5_GT_AMP2] = ACTIONS(760),
    [anon_sym_6_GT_AMP2] = ACTIONS(760),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(778),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [anon_sym_PIPE] = ACTIONS(762),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(762),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(762),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(762),
    [anon_sym_STAR] = ACTIONS(762),
    [anon_sym_DOT_DOT] = ACTIONS(760),
    [sym__statement_terminator] = ACTIONS(760),
  },
  [STATE(169)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(670),
    [anon_sym_BANG_EQ] = ACTIONS(670),
    [anon_sym_PLUS_EQ] = ACTIONS(670),
    [anon_sym_STAR_EQ] = ACTIONS(670),
    [anon_sym_SLASH_EQ] = ACTIONS(670),
    [anon_sym_PERCENT_EQ] = ACTIONS(670),
    [anon_sym_DASH_EQ] = ACTIONS(670),
    [anon_sym_GT] = ACTIONS(672),
    [anon_sym_GT_GT] = ACTIONS(670),
    [anon_sym_2_GT] = ACTIONS(672),
    [anon_sym_2_GT_GT] = ACTIONS(670),
    [anon_sym_3_GT] = ACTIONS(672),
    [anon_sym_3_GT_GT] = ACTIONS(670),
    [anon_sym_4_GT] = ACTIONS(672),
    [anon_sym_4_GT_GT] = ACTIONS(670),
    [anon_sym_5_GT] = ACTIONS(672),
    [anon_sym_5_GT_GT] = ACTIONS(670),
    [anon_sym_6_GT] = ACTIONS(672),
    [anon_sym_6_GT_GT] = ACTIONS(670),
    [anon_sym_STAR_GT] = ACTIONS(672),
    [anon_sym_STAR_GT_GT] = ACTIONS(670),
    [anon_sym_LT] = ACTIONS(672),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(670),
    [anon_sym_2_GT_AMP1] = ACTIONS(670),
    [anon_sym_3_GT_AMP1] = ACTIONS(670),
    [anon_sym_4_GT_AMP1] = ACTIONS(670),
    [anon_sym_5_GT_AMP1] = ACTIONS(670),
    [anon_sym_6_GT_AMP1] = ACTIONS(670),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(670),
    [anon_sym_1_GT_AMP2] = ACTIONS(670),
    [anon_sym_3_GT_AMP2] = ACTIONS(670),
    [anon_sym_4_GT_AMP2] = ACTIONS(670),
    [anon_sym_5_GT_AMP2] = ACTIONS(670),
    [anon_sym_6_GT_AMP2] = ACTIONS(670),
    [aux_sym_comparison_operator_token1] = ACTIONS(670),
    [aux_sym_comparison_operator_token2] = ACTIONS(670),
    [aux_sym_comparison_operator_token3] = ACTIONS(670),
    [aux_sym_comparison_operator_token4] = ACTIONS(670),
    [aux_sym_comparison_operator_token5] = ACTIONS(670),
    [aux_sym_comparison_operator_token6] = ACTIONS(670),
    [aux_sym_comparison_operator_token7] = ACTIONS(670),
    [aux_sym_comparison_operator_token8] = ACTIONS(670),
    [aux_sym_comparison_operator_token9] = ACTIONS(670),
    [aux_sym_comparison_operator_token10] = ACTIONS(670),
    [aux_sym_comparison_operator_token11] = ACTIONS(670),
    [aux_sym_comparison_operator_token12] = ACTIONS(670),
    [aux_sym_comparison_operator_token13] = ACTIONS(670),
    [aux_sym_comparison_operator_token14] = ACTIONS(670),
    [aux_sym_comparison_operator_token15] = ACTIONS(670),
    [aux_sym_comparison_operator_token16] = ACTIONS(670),
    [aux_sym_comparison_operator_token17] = ACTIONS(670),
    [aux_sym_comparison_operator_token18] = ACTIONS(670),
    [aux_sym_comparison_operator_token19] = ACTIONS(670),
    [aux_sym_comparison_operator_token20] = ACTIONS(670),
    [aux_sym_comparison_operator_token21] = ACTIONS(670),
    [aux_sym_comparison_operator_token22] = ACTIONS(670),
    [aux_sym_comparison_operator_token23] = ACTIONS(670),
    [aux_sym_comparison_operator_token24] = ACTIONS(670),
    [aux_sym_comparison_operator_token25] = ACTIONS(670),
    [aux_sym_comparison_operator_token26] = ACTIONS(670),
    [aux_sym_comparison_operator_token27] = ACTIONS(670),
    [aux_sym_comparison_operator_token28] = ACTIONS(672),
    [aux_sym_comparison_operator_token29] = ACTIONS(670),
    [aux_sym_comparison_operator_token30] = ACTIONS(670),
    [aux_sym_comparison_operator_token31] = ACTIONS(670),
    [aux_sym_comparison_operator_token32] = ACTIONS(670),
    [aux_sym_comparison_operator_token33] = ACTIONS(670),
    [aux_sym_comparison_operator_token34] = ACTIONS(672),
    [aux_sym_comparison_operator_token35] = ACTIONS(670),
    [aux_sym_comparison_operator_token36] = ACTIONS(670),
    [aux_sym_comparison_operator_token37] = ACTIONS(670),
    [aux_sym_comparison_operator_token38] = ACTIONS(670),
    [aux_sym_comparison_operator_token39] = ACTIONS(670),
    [aux_sym_comparison_operator_token40] = ACTIONS(670),
    [aux_sym_comparison_operator_token41] = ACTIONS(670),
    [aux_sym_comparison_operator_token42] = ACTIONS(670),
    [aux_sym_comparison_operator_token43] = ACTIONS(670),
    [aux_sym_comparison_operator_token44] = ACTIONS(670),
    [aux_sym_comparison_operator_token45] = ACTIONS(670),
    [aux_sym_comparison_operator_token46] = ACTIONS(670),
    [aux_sym_comparison_operator_token47] = ACTIONS(670),
    [aux_sym_comparison_operator_token48] = ACTIONS(670),
    [aux_sym_comparison_operator_token49] = ACTIONS(670),
    [aux_sym_comparison_operator_token50] = ACTIONS(670),
    [aux_sym_format_operator_token1] = ACTIONS(670),
    [anon_sym_RPAREN] = ACTIONS(670),
    [anon_sym_COMMA] = ACTIONS(670),
    [anon_sym_AMP_AMP] = ACTIONS(670),
    [anon_sym_PIPE_PIPE] = ACTIONS(670),
    [anon_sym_PIPE] = ACTIONS(672),
    [aux_sym_logical_expression_token1] = ACTIONS(670),
    [aux_sym_logical_expression_token2] = ACTIONS(670),
    [aux_sym_logical_expression_token3] = ACTIONS(670),
    [aux_sym_bitwise_expression_token1] = ACTIONS(670),
    [aux_sym_bitwise_expression_token2] = ACTIONS(670),
    [aux_sym_bitwise_expression_token3] = ACTIONS(670),
    [anon_sym_PLUS] = ACTIONS(672),
    [anon_sym_DASH] = ACTIONS(672),
    [anon_sym_SLASH] = ACTIONS(672),
    [anon_sym_BSLASH] = ACTIONS(670),
    [anon_sym_PERCENT] = ACTIONS(672),
    [anon_sym_STAR] = ACTIONS(672),
    [anon_sym_DOT_DOT] = ACTIONS(670),
    [anon_sym_RBRACK] = ACTIONS(670),
  },
  [STATE(170)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(674),
    [anon_sym_BANG_EQ] = ACTIONS(674),
    [anon_sym_PLUS_EQ] = ACTIONS(674),
    [anon_sym_STAR_EQ] = ACTIONS(674),
    [anon_sym_SLASH_EQ] = ACTIONS(674),
    [anon_sym_PERCENT_EQ] = ACTIONS(674),
    [anon_sym_DASH_EQ] = ACTIONS(674),
    [anon_sym_GT] = ACTIONS(676),
    [anon_sym_GT_GT] = ACTIONS(674),
    [anon_sym_2_GT] = ACTIONS(676),
    [anon_sym_2_GT_GT] = ACTIONS(674),
    [anon_sym_3_GT] = ACTIONS(676),
    [anon_sym_3_GT_GT] = ACTIONS(674),
    [anon_sym_4_GT] = ACTIONS(676),
    [anon_sym_4_GT_GT] = ACTIONS(674),
    [anon_sym_5_GT] = ACTIONS(676),
    [anon_sym_5_GT_GT] = ACTIONS(674),
    [anon_sym_6_GT] = ACTIONS(676),
    [anon_sym_6_GT_GT] = ACTIONS(674),
    [anon_sym_STAR_GT] = ACTIONS(676),
    [anon_sym_STAR_GT_GT] = ACTIONS(674),
    [anon_sym_LT] = ACTIONS(676),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(674),
    [anon_sym_2_GT_AMP1] = ACTIONS(674),
    [anon_sym_3_GT_AMP1] = ACTIONS(674),
    [anon_sym_4_GT_AMP1] = ACTIONS(674),
    [anon_sym_5_GT_AMP1] = ACTIONS(674),
    [anon_sym_6_GT_AMP1] = ACTIONS(674),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(674),
    [anon_sym_1_GT_AMP2] = ACTIONS(674),
    [anon_sym_3_GT_AMP2] = ACTIONS(674),
    [anon_sym_4_GT_AMP2] = ACTIONS(674),
    [anon_sym_5_GT_AMP2] = ACTIONS(674),
    [anon_sym_6_GT_AMP2] = ACTIONS(674),
    [aux_sym_comparison_operator_token1] = ACTIONS(674),
    [aux_sym_comparison_operator_token2] = ACTIONS(674),
    [aux_sym_comparison_operator_token3] = ACTIONS(674),
    [aux_sym_comparison_operator_token4] = ACTIONS(674),
    [aux_sym_comparison_operator_token5] = ACTIONS(674),
    [aux_sym_comparison_operator_token6] = ACTIONS(674),
    [aux_sym_comparison_operator_token7] = ACTIONS(674),
    [aux_sym_comparison_operator_token8] = ACTIONS(674),
    [aux_sym_comparison_operator_token9] = ACTIONS(674),
    [aux_sym_comparison_operator_token10] = ACTIONS(674),
    [aux_sym_comparison_operator_token11] = ACTIONS(674),
    [aux_sym_comparison_operator_token12] = ACTIONS(674),
    [aux_sym_comparison_operator_token13] = ACTIONS(674),
    [aux_sym_comparison_operator_token14] = ACTIONS(674),
    [aux_sym_comparison_operator_token15] = ACTIONS(674),
    [aux_sym_comparison_operator_token16] = ACTIONS(674),
    [aux_sym_comparison_operator_token17] = ACTIONS(674),
    [aux_sym_comparison_operator_token18] = ACTIONS(674),
    [aux_sym_comparison_operator_token19] = ACTIONS(674),
    [aux_sym_comparison_operator_token20] = ACTIONS(674),
    [aux_sym_comparison_operator_token21] = ACTIONS(674),
    [aux_sym_comparison_operator_token22] = ACTIONS(674),
    [aux_sym_comparison_operator_token23] = ACTIONS(674),
    [aux_sym_comparison_operator_token24] = ACTIONS(674),
    [aux_sym_comparison_operator_token25] = ACTIONS(674),
    [aux_sym_comparison_operator_token26] = ACTIONS(674),
    [aux_sym_comparison_operator_token27] = ACTIONS(674),
    [aux_sym_comparison_operator_token28] = ACTIONS(676),
    [aux_sym_comparison_operator_token29] = ACTIONS(674),
    [aux_sym_comparison_operator_token30] = ACTIONS(674),
    [aux_sym_comparison_operator_token31] = ACTIONS(674),
    [aux_sym_comparison_operator_token32] = ACTIONS(674),
    [aux_sym_comparison_operator_token33] = ACTIONS(674),
    [aux_sym_comparison_operator_token34] = ACTIONS(676),
    [aux_sym_comparison_operator_token35] = ACTIONS(674),
    [aux_sym_comparison_operator_token36] = ACTIONS(674),
    [aux_sym_comparison_operator_token37] = ACTIONS(674),
    [aux_sym_comparison_operator_token38] = ACTIONS(674),
    [aux_sym_comparison_operator_token39] = ACTIONS(674),
    [aux_sym_comparison_operator_token40] = ACTIONS(674),
    [aux_sym_comparison_operator_token41] = ACTIONS(674),
    [aux_sym_comparison_operator_token42] = ACTIONS(674),
    [aux_sym_comparison_operator_token43] = ACTIONS(674),
    [aux_sym_comparison_operator_token44] = ACTIONS(674),
    [aux_sym_comparison_operator_token45] = ACTIONS(674),
    [aux_sym_comparison_operator_token46] = ACTIONS(674),
    [aux_sym_comparison_operator_token47] = ACTIONS(674),
    [aux_sym_comparison_operator_token48] = ACTIONS(674),
    [aux_sym_comparison_operator_token49] = ACTIONS(674),
    [aux_sym_comparison_operator_token50] = ACTIONS(674),
    [aux_sym_format_operator_token1] = ACTIONS(674),
    [anon_sym_RPAREN] = ACTIONS(674),
    [anon_sym_COMMA] = ACTIONS(674),
    [anon_sym_AMP_AMP] = ACTIONS(674),
    [anon_sym_PIPE_PIPE] = ACTIONS(674),
    [anon_sym_PIPE] = ACTIONS(676),
    [aux_sym_logical_expression_token1] = ACTIONS(674),
    [aux_sym_logical_expression_token2] = ACTIONS(674),
    [aux_sym_logical_expression_token3] = ACTIONS(674),
    [aux_sym_bitwise_expression_token1] = ACTIONS(674),
    [aux_sym_bitwise_expression_token2] = ACTIONS(674),
    [aux_sym_bitwise_expression_token3] = ACTIONS(674),
    [anon_sym_PLUS] = ACTIONS(676),
    [anon_sym_DASH] = ACTIONS(676),
    [anon_sym_SLASH] = ACTIONS(676),
    [anon_sym_BSLASH] = ACTIONS(674),
    [anon_sym_PERCENT] = ACTIONS(676),
    [anon_sym_STAR] = ACTIONS(676),
    [anon_sym_DOT_DOT] = ACTIONS(674),
    [anon_sym_RBRACK] = ACTIONS(674),
  },
  [STATE(171)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(173),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(772),
    [anon_sym_BANG_EQ] = ACTIONS(772),
    [anon_sym_PLUS_EQ] = ACTIONS(772),
    [anon_sym_STAR_EQ] = ACTIONS(772),
    [anon_sym_SLASH_EQ] = ACTIONS(772),
    [anon_sym_PERCENT_EQ] = ACTIONS(772),
    [anon_sym_DASH_EQ] = ACTIONS(772),
    [anon_sym_GT] = ACTIONS(774),
    [anon_sym_GT_GT] = ACTIONS(772),
    [anon_sym_2_GT] = ACTIONS(774),
    [anon_sym_2_GT_GT] = ACTIONS(772),
    [anon_sym_3_GT] = ACTIONS(774),
    [anon_sym_3_GT_GT] = ACTIONS(772),
    [anon_sym_4_GT] = ACTIONS(774),
    [anon_sym_4_GT_GT] = ACTIONS(772),
    [anon_sym_5_GT] = ACTIONS(774),
    [anon_sym_5_GT_GT] = ACTIONS(772),
    [anon_sym_6_GT] = ACTIONS(774),
    [anon_sym_6_GT_GT] = ACTIONS(772),
    [anon_sym_STAR_GT] = ACTIONS(774),
    [anon_sym_STAR_GT_GT] = ACTIONS(772),
    [anon_sym_LT] = ACTIONS(774),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(772),
    [anon_sym_2_GT_AMP1] = ACTIONS(772),
    [anon_sym_3_GT_AMP1] = ACTIONS(772),
    [anon_sym_4_GT_AMP1] = ACTIONS(772),
    [anon_sym_5_GT_AMP1] = ACTIONS(772),
    [anon_sym_6_GT_AMP1] = ACTIONS(772),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(772),
    [anon_sym_1_GT_AMP2] = ACTIONS(772),
    [anon_sym_3_GT_AMP2] = ACTIONS(772),
    [anon_sym_4_GT_AMP2] = ACTIONS(772),
    [anon_sym_5_GT_AMP2] = ACTIONS(772),
    [anon_sym_6_GT_AMP2] = ACTIONS(772),
    [aux_sym_comparison_operator_token1] = ACTIONS(772),
    [aux_sym_comparison_operator_token2] = ACTIONS(772),
    [aux_sym_comparison_operator_token3] = ACTIONS(772),
    [aux_sym_comparison_operator_token4] = ACTIONS(772),
    [aux_sym_comparison_operator_token5] = ACTIONS(772),
    [aux_sym_comparison_operator_token6] = ACTIONS(772),
    [aux_sym_comparison_operator_token7] = ACTIONS(772),
    [aux_sym_comparison_operator_token8] = ACTIONS(772),
    [aux_sym_comparison_operator_token9] = ACTIONS(772),
    [aux_sym_comparison_operator_token10] = ACTIONS(772),
    [aux_sym_comparison_operator_token11] = ACTIONS(772),
    [aux_sym_comparison_operator_token12] = ACTIONS(772),
    [aux_sym_comparison_operator_token13] = ACTIONS(772),
    [aux_sym_comparison_operator_token14] = ACTIONS(772),
    [aux_sym_comparison_operator_token15] = ACTIONS(772),
    [aux_sym_comparison_operator_token16] = ACTIONS(772),
    [aux_sym_comparison_operator_token17] = ACTIONS(772),
    [aux_sym_comparison_operator_token18] = ACTIONS(772),
    [aux_sym_comparison_operator_token19] = ACTIONS(772),
    [aux_sym_comparison_operator_token20] = ACTIONS(772),
    [aux_sym_comparison_operator_token21] = ACTIONS(772),
    [aux_sym_comparison_operator_token22] = ACTIONS(772),
    [aux_sym_comparison_operator_token23] = ACTIONS(772),
    [aux_sym_comparison_operator_token24] = ACTIONS(772),
    [aux_sym_comparison_operator_token25] = ACTIONS(772),
    [aux_sym_comparison_operator_token26] = ACTIONS(772),
    [aux_sym_comparison_operator_token27] = ACTIONS(772),
    [aux_sym_comparison_operator_token28] = ACTIONS(774),
    [aux_sym_comparison_operator_token29] = ACTIONS(772),
    [aux_sym_comparison_operator_token30] = ACTIONS(772),
    [aux_sym_comparison_operator_token31] = ACTIONS(772),
    [aux_sym_comparison_operator_token32] = ACTIONS(772),
    [aux_sym_comparison_operator_token33] = ACTIONS(772),
    [aux_sym_comparison_operator_token34] = ACTIONS(774),
    [aux_sym_comparison_operator_token35] = ACTIONS(772),
    [aux_sym_comparison_operator_token36] = ACTIONS(772),
    [aux_sym_comparison_operator_token37] = ACTIONS(772),
    [aux_sym_comparison_operator_token38] = ACTIONS(772),
    [aux_sym_comparison_operator_token39] = ACTIONS(772),
    [aux_sym_comparison_operator_token40] = ACTIONS(772),
    [aux_sym_comparison_operator_token41] = ACTIONS(772),
    [aux_sym_comparison_operator_token42] = ACTIONS(772),
    [aux_sym_comparison_operator_token43] = ACTIONS(772),
    [aux_sym_comparison_operator_token44] = ACTIONS(772),
    [aux_sym_comparison_operator_token45] = ACTIONS(772),
    [aux_sym_comparison_operator_token46] = ACTIONS(772),
    [aux_sym_comparison_operator_token47] = ACTIONS(772),
    [aux_sym_comparison_operator_token48] = ACTIONS(772),
    [aux_sym_comparison_operator_token49] = ACTIONS(772),
    [aux_sym_comparison_operator_token50] = ACTIONS(772),
    [aux_sym_format_operator_token1] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(781),
    [anon_sym_AMP_AMP] = ACTIONS(772),
    [anon_sym_PIPE_PIPE] = ACTIONS(772),
    [anon_sym_PIPE] = ACTIONS(774),
    [aux_sym_logical_expression_token1] = ACTIONS(772),
    [aux_sym_logical_expression_token2] = ACTIONS(772),
    [aux_sym_logical_expression_token3] = ACTIONS(772),
    [aux_sym_bitwise_expression_token1] = ACTIONS(772),
    [aux_sym_bitwise_expression_token2] = ACTIONS(772),
    [aux_sym_bitwise_expression_token3] = ACTIONS(772),
    [anon_sym_PLUS] = ACTIONS(774),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(774),
    [anon_sym_BSLASH] = ACTIONS(772),
    [anon_sym_PERCENT] = ACTIONS(774),
    [anon_sym_STAR] = ACTIONS(774),
    [anon_sym_DOT_DOT] = ACTIONS(772),
    [sym__statement_terminator] = ACTIONS(772),
  },
  [STATE(172)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(175),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(783),
    [anon_sym_BANG_EQ] = ACTIONS(783),
    [anon_sym_PLUS_EQ] = ACTIONS(783),
    [anon_sym_STAR_EQ] = ACTIONS(783),
    [anon_sym_SLASH_EQ] = ACTIONS(783),
    [anon_sym_PERCENT_EQ] = ACTIONS(783),
    [anon_sym_DASH_EQ] = ACTIONS(783),
    [anon_sym_GT] = ACTIONS(785),
    [anon_sym_GT_GT] = ACTIONS(783),
    [anon_sym_2_GT] = ACTIONS(785),
    [anon_sym_2_GT_GT] = ACTIONS(783),
    [anon_sym_3_GT] = ACTIONS(785),
    [anon_sym_3_GT_GT] = ACTIONS(783),
    [anon_sym_4_GT] = ACTIONS(785),
    [anon_sym_4_GT_GT] = ACTIONS(783),
    [anon_sym_5_GT] = ACTIONS(785),
    [anon_sym_5_GT_GT] = ACTIONS(783),
    [anon_sym_6_GT] = ACTIONS(785),
    [anon_sym_6_GT_GT] = ACTIONS(783),
    [anon_sym_STAR_GT] = ACTIONS(785),
    [anon_sym_STAR_GT_GT] = ACTIONS(783),
    [anon_sym_LT] = ACTIONS(785),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(783),
    [anon_sym_2_GT_AMP1] = ACTIONS(783),
    [anon_sym_3_GT_AMP1] = ACTIONS(783),
    [anon_sym_4_GT_AMP1] = ACTIONS(783),
    [anon_sym_5_GT_AMP1] = ACTIONS(783),
    [anon_sym_6_GT_AMP1] = ACTIONS(783),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(783),
    [anon_sym_1_GT_AMP2] = ACTIONS(783),
    [anon_sym_3_GT_AMP2] = ACTIONS(783),
    [anon_sym_4_GT_AMP2] = ACTIONS(783),
    [anon_sym_5_GT_AMP2] = ACTIONS(783),
    [anon_sym_6_GT_AMP2] = ACTIONS(783),
    [aux_sym_comparison_operator_token1] = ACTIONS(783),
    [aux_sym_comparison_operator_token2] = ACTIONS(783),
    [aux_sym_comparison_operator_token3] = ACTIONS(783),
    [aux_sym_comparison_operator_token4] = ACTIONS(783),
    [aux_sym_comparison_operator_token5] = ACTIONS(783),
    [aux_sym_comparison_operator_token6] = ACTIONS(783),
    [aux_sym_comparison_operator_token7] = ACTIONS(783),
    [aux_sym_comparison_operator_token8] = ACTIONS(783),
    [aux_sym_comparison_operator_token9] = ACTIONS(783),
    [aux_sym_comparison_operator_token10] = ACTIONS(783),
    [aux_sym_comparison_operator_token11] = ACTIONS(783),
    [aux_sym_comparison_operator_token12] = ACTIONS(783),
    [aux_sym_comparison_operator_token13] = ACTIONS(783),
    [aux_sym_comparison_operator_token14] = ACTIONS(783),
    [aux_sym_comparison_operator_token15] = ACTIONS(783),
    [aux_sym_comparison_operator_token16] = ACTIONS(783),
    [aux_sym_comparison_operator_token17] = ACTIONS(783),
    [aux_sym_comparison_operator_token18] = ACTIONS(783),
    [aux_sym_comparison_operator_token19] = ACTIONS(783),
    [aux_sym_comparison_operator_token20] = ACTIONS(783),
    [aux_sym_comparison_operator_token21] = ACTIONS(783),
    [aux_sym_comparison_operator_token22] = ACTIONS(783),
    [aux_sym_comparison_operator_token23] = ACTIONS(783),
    [aux_sym_comparison_operator_token24] = ACTIONS(783),
    [aux_sym_comparison_operator_token25] = ACTIONS(783),
    [aux_sym_comparison_operator_token26] = ACTIONS(783),
    [aux_sym_comparison_operator_token27] = ACTIONS(783),
    [aux_sym_comparison_operator_token28] = ACTIONS(785),
    [aux_sym_comparison_operator_token29] = ACTIONS(783),
    [aux_sym_comparison_operator_token30] = ACTIONS(783),
    [aux_sym_comparison_operator_token31] = ACTIONS(783),
    [aux_sym_comparison_operator_token32] = ACTIONS(783),
    [aux_sym_comparison_operator_token33] = ACTIONS(783),
    [aux_sym_comparison_operator_token34] = ACTIONS(785),
    [aux_sym_comparison_operator_token35] = ACTIONS(783),
    [aux_sym_comparison_operator_token36] = ACTIONS(783),
    [aux_sym_comparison_operator_token37] = ACTIONS(783),
    [aux_sym_comparison_operator_token38] = ACTIONS(783),
    [aux_sym_comparison_operator_token39] = ACTIONS(783),
    [aux_sym_comparison_operator_token40] = ACTIONS(783),
    [aux_sym_comparison_operator_token41] = ACTIONS(783),
    [aux_sym_comparison_operator_token42] = ACTIONS(783),
    [aux_sym_comparison_operator_token43] = ACTIONS(783),
    [aux_sym_comparison_operator_token44] = ACTIONS(783),
    [aux_sym_comparison_operator_token45] = ACTIONS(783),
    [aux_sym_comparison_operator_token46] = ACTIONS(783),
    [aux_sym_comparison_operator_token47] = ACTIONS(783),
    [aux_sym_comparison_operator_token48] = ACTIONS(783),
    [aux_sym_comparison_operator_token49] = ACTIONS(783),
    [aux_sym_comparison_operator_token50] = ACTIONS(783),
    [aux_sym_format_operator_token1] = ACTIONS(783),
    [anon_sym_RPAREN] = ACTIONS(783),
    [anon_sym_COMMA] = ACTIONS(776),
    [anon_sym_AMP_AMP] = ACTIONS(783),
    [anon_sym_PIPE_PIPE] = ACTIONS(783),
    [anon_sym_PIPE] = ACTIONS(785),
    [aux_sym_logical_expression_token1] = ACTIONS(783),
    [aux_sym_logical_expression_token2] = ACTIONS(783),
    [aux_sym_logical_expression_token3] = ACTIONS(783),
    [aux_sym_bitwise_expression_token1] = ACTIONS(783),
    [aux_sym_bitwise_expression_token2] = ACTIONS(783),
    [aux_sym_bitwise_expression_token3] = ACTIONS(783),
    [anon_sym_PLUS] = ACTIONS(785),
    [anon_sym_DASH] = ACTIONS(785),
    [anon_sym_SLASH] = ACTIONS(785),
    [anon_sym_BSLASH] = ACTIONS(783),
    [anon_sym_PERCENT] = ACTIONS(785),
    [anon_sym_STAR] = ACTIONS(785),
    [anon_sym_DOT_DOT] = ACTIONS(783),
  },
  [STATE(173)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(168),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(783),
    [anon_sym_BANG_EQ] = ACTIONS(783),
    [anon_sym_PLUS_EQ] = ACTIONS(783),
    [anon_sym_STAR_EQ] = ACTIONS(783),
    [anon_sym_SLASH_EQ] = ACTIONS(783),
    [anon_sym_PERCENT_EQ] = ACTIONS(783),
    [anon_sym_DASH_EQ] = ACTIONS(783),
    [anon_sym_GT] = ACTIONS(785),
    [anon_sym_GT_GT] = ACTIONS(783),
    [anon_sym_2_GT] = ACTIONS(785),
    [anon_sym_2_GT_GT] = ACTIONS(783),
    [anon_sym_3_GT] = ACTIONS(785),
    [anon_sym_3_GT_GT] = ACTIONS(783),
    [anon_sym_4_GT] = ACTIONS(785),
    [anon_sym_4_GT_GT] = ACTIONS(783),
    [anon_sym_5_GT] = ACTIONS(785),
    [anon_sym_5_GT_GT] = ACTIONS(783),
    [anon_sym_6_GT] = ACTIONS(785),
    [anon_sym_6_GT_GT] = ACTIONS(783),
    [anon_sym_STAR_GT] = ACTIONS(785),
    [anon_sym_STAR_GT_GT] = ACTIONS(783),
    [anon_sym_LT] = ACTIONS(785),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(783),
    [anon_sym_2_GT_AMP1] = ACTIONS(783),
    [anon_sym_3_GT_AMP1] = ACTIONS(783),
    [anon_sym_4_GT_AMP1] = ACTIONS(783),
    [anon_sym_5_GT_AMP1] = ACTIONS(783),
    [anon_sym_6_GT_AMP1] = ACTIONS(783),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(783),
    [anon_sym_1_GT_AMP2] = ACTIONS(783),
    [anon_sym_3_GT_AMP2] = ACTIONS(783),
    [anon_sym_4_GT_AMP2] = ACTIONS(783),
    [anon_sym_5_GT_AMP2] = ACTIONS(783),
    [anon_sym_6_GT_AMP2] = ACTIONS(783),
    [aux_sym_comparison_operator_token1] = ACTIONS(783),
    [aux_sym_comparison_operator_token2] = ACTIONS(783),
    [aux_sym_comparison_operator_token3] = ACTIONS(783),
    [aux_sym_comparison_operator_token4] = ACTIONS(783),
    [aux_sym_comparison_operator_token5] = ACTIONS(783),
    [aux_sym_comparison_operator_token6] = ACTIONS(783),
    [aux_sym_comparison_operator_token7] = ACTIONS(783),
    [aux_sym_comparison_operator_token8] = ACTIONS(783),
    [aux_sym_comparison_operator_token9] = ACTIONS(783),
    [aux_sym_comparison_operator_token10] = ACTIONS(783),
    [aux_sym_comparison_operator_token11] = ACTIONS(783),
    [aux_sym_comparison_operator_token12] = ACTIONS(783),
    [aux_sym_comparison_operator_token13] = ACTIONS(783),
    [aux_sym_comparison_operator_token14] = ACTIONS(783),
    [aux_sym_comparison_operator_token15] = ACTIONS(783),
    [aux_sym_comparison_operator_token16] = ACTIONS(783),
    [aux_sym_comparison_operator_token17] = ACTIONS(783),
    [aux_sym_comparison_operator_token18] = ACTIONS(783),
    [aux_sym_comparison_operator_token19] = ACTIONS(783),
    [aux_sym_comparison_operator_token20] = ACTIONS(783),
    [aux_sym_comparison_operator_token21] = ACTIONS(783),
    [aux_sym_comparison_operator_token22] = ACTIONS(783),
    [aux_sym_comparison_operator_token23] = ACTIONS(783),
    [aux_sym_comparison_operator_token24] = ACTIONS(783),
    [aux_sym_comparison_operator_token25] = ACTIONS(783),
    [aux_sym_comparison_operator_token26] = ACTIONS(783),
    [aux_sym_comparison_operator_token27] = ACTIONS(783),
    [aux_sym_comparison_operator_token28] = ACTIONS(785),
    [aux_sym_comparison_operator_token29] = ACTIONS(783),
    [aux_sym_comparison_operator_token30] = ACTIONS(783),
    [aux_sym_comparison_operator_token31] = ACTIONS(783),
    [aux_sym_comparison_operator_token32] = ACTIONS(783),
    [aux_sym_comparison_operator_token33] = ACTIONS(783),
    [aux_sym_comparison_operator_token34] = ACTIONS(785),
    [aux_sym_comparison_operator_token35] = ACTIONS(783),
    [aux_sym_comparison_operator_token36] = ACTIONS(783),
    [aux_sym_comparison_operator_token37] = ACTIONS(783),
    [aux_sym_comparison_operator_token38] = ACTIONS(783),
    [aux_sym_comparison_operator_token39] = ACTIONS(783),
    [aux_sym_comparison_operator_token40] = ACTIONS(783),
    [aux_sym_comparison_operator_token41] = ACTIONS(783),
    [aux_sym_comparison_operator_token42] = ACTIONS(783),
    [aux_sym_comparison_operator_token43] = ACTIONS(783),
    [aux_sym_comparison_operator_token44] = ACTIONS(783),
    [aux_sym_comparison_operator_token45] = ACTIONS(783),
    [aux_sym_comparison_operator_token46] = ACTIONS(783),
    [aux_sym_comparison_operator_token47] = ACTIONS(783),
    [aux_sym_comparison_operator_token48] = ACTIONS(783),
    [aux_sym_comparison_operator_token49] = ACTIONS(783),
    [aux_sym_comparison_operator_token50] = ACTIONS(783),
    [aux_sym_format_operator_token1] = ACTIONS(783),
    [anon_sym_COMMA] = ACTIONS(781),
    [anon_sym_AMP_AMP] = ACTIONS(783),
    [anon_sym_PIPE_PIPE] = ACTIONS(783),
    [anon_sym_PIPE] = ACTIONS(785),
    [aux_sym_logical_expression_token1] = ACTIONS(783),
    [aux_sym_logical_expression_token2] = ACTIONS(783),
    [aux_sym_logical_expression_token3] = ACTIONS(783),
    [aux_sym_bitwise_expression_token1] = ACTIONS(783),
    [aux_sym_bitwise_expression_token2] = ACTIONS(783),
    [aux_sym_bitwise_expression_token3] = ACTIONS(783),
    [anon_sym_PLUS] = ACTIONS(785),
    [anon_sym_DASH] = ACTIONS(785),
    [anon_sym_SLASH] = ACTIONS(785),
    [anon_sym_BSLASH] = ACTIONS(783),
    [anon_sym_PERCENT] = ACTIONS(785),
    [anon_sym_STAR] = ACTIONS(785),
    [anon_sym_DOT_DOT] = ACTIONS(783),
    [sym__statement_terminator] = ACTIONS(783),
  },
  [STATE(174)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(678),
    [anon_sym_BANG_EQ] = ACTIONS(678),
    [anon_sym_PLUS_EQ] = ACTIONS(678),
    [anon_sym_STAR_EQ] = ACTIONS(678),
    [anon_sym_SLASH_EQ] = ACTIONS(678),
    [anon_sym_PERCENT_EQ] = ACTIONS(678),
    [anon_sym_DASH_EQ] = ACTIONS(678),
    [anon_sym_GT] = ACTIONS(680),
    [anon_sym_GT_GT] = ACTIONS(678),
    [anon_sym_2_GT] = ACTIONS(680),
    [anon_sym_2_GT_GT] = ACTIONS(678),
    [anon_sym_3_GT] = ACTIONS(680),
    [anon_sym_3_GT_GT] = ACTIONS(678),
    [anon_sym_4_GT] = ACTIONS(680),
    [anon_sym_4_GT_GT] = ACTIONS(678),
    [anon_sym_5_GT] = ACTIONS(680),
    [anon_sym_5_GT_GT] = ACTIONS(678),
    [anon_sym_6_GT] = ACTIONS(680),
    [anon_sym_6_GT_GT] = ACTIONS(678),
    [anon_sym_STAR_GT] = ACTIONS(680),
    [anon_sym_STAR_GT_GT] = ACTIONS(678),
    [anon_sym_LT] = ACTIONS(680),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(678),
    [anon_sym_2_GT_AMP1] = ACTIONS(678),
    [anon_sym_3_GT_AMP1] = ACTIONS(678),
    [anon_sym_4_GT_AMP1] = ACTIONS(678),
    [anon_sym_5_GT_AMP1] = ACTIONS(678),
    [anon_sym_6_GT_AMP1] = ACTIONS(678),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(678),
    [anon_sym_1_GT_AMP2] = ACTIONS(678),
    [anon_sym_3_GT_AMP2] = ACTIONS(678),
    [anon_sym_4_GT_AMP2] = ACTIONS(678),
    [anon_sym_5_GT_AMP2] = ACTIONS(678),
    [anon_sym_6_GT_AMP2] = ACTIONS(678),
    [aux_sym_comparison_operator_token1] = ACTIONS(678),
    [aux_sym_comparison_operator_token2] = ACTIONS(678),
    [aux_sym_comparison_operator_token3] = ACTIONS(678),
    [aux_sym_comparison_operator_token4] = ACTIONS(678),
    [aux_sym_comparison_operator_token5] = ACTIONS(678),
    [aux_sym_comparison_operator_token6] = ACTIONS(678),
    [aux_sym_comparison_operator_token7] = ACTIONS(678),
    [aux_sym_comparison_operator_token8] = ACTIONS(678),
    [aux_sym_comparison_operator_token9] = ACTIONS(678),
    [aux_sym_comparison_operator_token10] = ACTIONS(678),
    [aux_sym_comparison_operator_token11] = ACTIONS(678),
    [aux_sym_comparison_operator_token12] = ACTIONS(678),
    [aux_sym_comparison_operator_token13] = ACTIONS(678),
    [aux_sym_comparison_operator_token14] = ACTIONS(678),
    [aux_sym_comparison_operator_token15] = ACTIONS(678),
    [aux_sym_comparison_operator_token16] = ACTIONS(678),
    [aux_sym_comparison_operator_token17] = ACTIONS(678),
    [aux_sym_comparison_operator_token18] = ACTIONS(678),
    [aux_sym_comparison_operator_token19] = ACTIONS(678),
    [aux_sym_comparison_operator_token20] = ACTIONS(678),
    [aux_sym_comparison_operator_token21] = ACTIONS(678),
    [aux_sym_comparison_operator_token22] = ACTIONS(678),
    [aux_sym_comparison_operator_token23] = ACTIONS(678),
    [aux_sym_comparison_operator_token24] = ACTIONS(678),
    [aux_sym_comparison_operator_token25] = ACTIONS(678),
    [aux_sym_comparison_operator_token26] = ACTIONS(678),
    [aux_sym_comparison_operator_token27] = ACTIONS(678),
    [aux_sym_comparison_operator_token28] = ACTIONS(680),
    [aux_sym_comparison_operator_token29] = ACTIONS(678),
    [aux_sym_comparison_operator_token30] = ACTIONS(678),
    [aux_sym_comparison_operator_token31] = ACTIONS(678),
    [aux_sym_comparison_operator_token32] = ACTIONS(678),
    [aux_sym_comparison_operator_token33] = ACTIONS(678),
    [aux_sym_comparison_operator_token34] = ACTIONS(680),
    [aux_sym_comparison_operator_token35] = ACTIONS(678),
    [aux_sym_comparison_operator_token36] = ACTIONS(678),
    [aux_sym_comparison_operator_token37] = ACTIONS(678),
    [aux_sym_comparison_operator_token38] = ACTIONS(678),
    [aux_sym_comparison_operator_token39] = ACTIONS(678),
    [aux_sym_comparison_operator_token40] = ACTIONS(678),
    [aux_sym_comparison_operator_token41] = ACTIONS(678),
    [aux_sym_comparison_operator_token42] = ACTIONS(678),
    [aux_sym_comparison_operator_token43] = ACTIONS(678),
    [aux_sym_comparison_operator_token44] = ACTIONS(678),
    [aux_sym_comparison_operator_token45] = ACTIONS(678),
    [aux_sym_comparison_operator_token46] = ACTIONS(678),
    [aux_sym_comparison_operator_token47] = ACTIONS(678),
    [aux_sym_comparison_operator_token48] = ACTIONS(678),
    [aux_sym_comparison_operator_token49] = ACTIONS(678),
    [aux_sym_comparison_operator_token50] = ACTIONS(678),
    [aux_sym_format_operator_token1] = ACTIONS(678),
    [anon_sym_RPAREN] = ACTIONS(678),
    [anon_sym_COMMA] = ACTIONS(678),
    [anon_sym_AMP_AMP] = ACTIONS(678),
    [anon_sym_PIPE_PIPE] = ACTIONS(678),
    [anon_sym_PIPE] = ACTIONS(680),
    [aux_sym_logical_expression_token1] = ACTIONS(678),
    [aux_sym_logical_expression_token2] = ACTIONS(678),
    [aux_sym_logical_expression_token3] = ACTIONS(678),
    [aux_sym_bitwise_expression_token1] = ACTIONS(678),
    [aux_sym_bitwise_expression_token2] = ACTIONS(678),
    [aux_sym_bitwise_expression_token3] = ACTIONS(678),
    [anon_sym_PLUS] = ACTIONS(680),
    [anon_sym_DASH] = ACTIONS(680),
    [anon_sym_SLASH] = ACTIONS(680),
    [anon_sym_BSLASH] = ACTIONS(678),
    [anon_sym_PERCENT] = ACTIONS(680),
    [anon_sym_STAR] = ACTIONS(680),
    [anon_sym_DOT_DOT] = ACTIONS(678),
    [anon_sym_RBRACK] = ACTIONS(678),
  },
  [STATE(175)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(175),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(760),
    [anon_sym_BANG_EQ] = ACTIONS(760),
    [anon_sym_PLUS_EQ] = ACTIONS(760),
    [anon_sym_STAR_EQ] = ACTIONS(760),
    [anon_sym_SLASH_EQ] = ACTIONS(760),
    [anon_sym_PERCENT_EQ] = ACTIONS(760),
    [anon_sym_DASH_EQ] = ACTIONS(760),
    [anon_sym_GT] = ACTIONS(762),
    [anon_sym_GT_GT] = ACTIONS(760),
    [anon_sym_2_GT] = ACTIONS(762),
    [anon_sym_2_GT_GT] = ACTIONS(760),
    [anon_sym_3_GT] = ACTIONS(762),
    [anon_sym_3_GT_GT] = ACTIONS(760),
    [anon_sym_4_GT] = ACTIONS(762),
    [anon_sym_4_GT_GT] = ACTIONS(760),
    [anon_sym_5_GT] = ACTIONS(762),
    [anon_sym_5_GT_GT] = ACTIONS(760),
    [anon_sym_6_GT] = ACTIONS(762),
    [anon_sym_6_GT_GT] = ACTIONS(760),
    [anon_sym_STAR_GT] = ACTIONS(762),
    [anon_sym_STAR_GT_GT] = ACTIONS(760),
    [anon_sym_LT] = ACTIONS(762),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(760),
    [anon_sym_2_GT_AMP1] = ACTIONS(760),
    [anon_sym_3_GT_AMP1] = ACTIONS(760),
    [anon_sym_4_GT_AMP1] = ACTIONS(760),
    [anon_sym_5_GT_AMP1] = ACTIONS(760),
    [anon_sym_6_GT_AMP1] = ACTIONS(760),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(760),
    [anon_sym_1_GT_AMP2] = ACTIONS(760),
    [anon_sym_3_GT_AMP2] = ACTIONS(760),
    [anon_sym_4_GT_AMP2] = ACTIONS(760),
    [anon_sym_5_GT_AMP2] = ACTIONS(760),
    [anon_sym_6_GT_AMP2] = ACTIONS(760),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_RPAREN] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(787),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [anon_sym_PIPE] = ACTIONS(762),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(762),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(762),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(762),
    [anon_sym_STAR] = ACTIONS(762),
    [anon_sym_DOT_DOT] = ACTIONS(760),
  },
  [STATE(176)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(654),
    [anon_sym_BANG_EQ] = ACTIONS(654),
    [anon_sym_PLUS_EQ] = ACTIONS(654),
    [anon_sym_STAR_EQ] = ACTIONS(654),
    [anon_sym_SLASH_EQ] = ACTIONS(654),
    [anon_sym_PERCENT_EQ] = ACTIONS(654),
    [anon_sym_DASH_EQ] = ACTIONS(654),
    [anon_sym_GT] = ACTIONS(656),
    [anon_sym_GT_GT] = ACTIONS(654),
    [anon_sym_2_GT] = ACTIONS(656),
    [anon_sym_2_GT_GT] = ACTIONS(654),
    [anon_sym_3_GT] = ACTIONS(656),
    [anon_sym_3_GT_GT] = ACTIONS(654),
    [anon_sym_4_GT] = ACTIONS(656),
    [anon_sym_4_GT_GT] = ACTIONS(654),
    [anon_sym_5_GT] = ACTIONS(656),
    [anon_sym_5_GT_GT] = ACTIONS(654),
    [anon_sym_6_GT] = ACTIONS(656),
    [anon_sym_6_GT_GT] = ACTIONS(654),
    [anon_sym_STAR_GT] = ACTIONS(656),
    [anon_sym_STAR_GT_GT] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(654),
    [anon_sym_2_GT_AMP1] = ACTIONS(654),
    [anon_sym_3_GT_AMP1] = ACTIONS(654),
    [anon_sym_4_GT_AMP1] = ACTIONS(654),
    [anon_sym_5_GT_AMP1] = ACTIONS(654),
    [anon_sym_6_GT_AMP1] = ACTIONS(654),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(654),
    [anon_sym_1_GT_AMP2] = ACTIONS(654),
    [anon_sym_3_GT_AMP2] = ACTIONS(654),
    [anon_sym_4_GT_AMP2] = ACTIONS(654),
    [anon_sym_5_GT_AMP2] = ACTIONS(654),
    [anon_sym_6_GT_AMP2] = ACTIONS(654),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [anon_sym_AMP_AMP] = ACTIONS(654),
    [anon_sym_PIPE_PIPE] = ACTIONS(654),
    [anon_sym_PIPE] = ACTIONS(656),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(656),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(656),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(177)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(658),
    [anon_sym_BANG_EQ] = ACTIONS(658),
    [anon_sym_PLUS_EQ] = ACTIONS(658),
    [anon_sym_STAR_EQ] = ACTIONS(658),
    [anon_sym_SLASH_EQ] = ACTIONS(658),
    [anon_sym_PERCENT_EQ] = ACTIONS(658),
    [anon_sym_DASH_EQ] = ACTIONS(658),
    [anon_sym_GT] = ACTIONS(660),
    [anon_sym_GT_GT] = ACTIONS(658),
    [anon_sym_2_GT] = ACTIONS(660),
    [anon_sym_2_GT_GT] = ACTIONS(658),
    [anon_sym_3_GT] = ACTIONS(660),
    [anon_sym_3_GT_GT] = ACTIONS(658),
    [anon_sym_4_GT] = ACTIONS(660),
    [anon_sym_4_GT_GT] = ACTIONS(658),
    [anon_sym_5_GT] = ACTIONS(660),
    [anon_sym_5_GT_GT] = ACTIONS(658),
    [anon_sym_6_GT] = ACTIONS(660),
    [anon_sym_6_GT_GT] = ACTIONS(658),
    [anon_sym_STAR_GT] = ACTIONS(660),
    [anon_sym_STAR_GT_GT] = ACTIONS(658),
    [anon_sym_LT] = ACTIONS(660),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(658),
    [anon_sym_2_GT_AMP1] = ACTIONS(658),
    [anon_sym_3_GT_AMP1] = ACTIONS(658),
    [anon_sym_4_GT_AMP1] = ACTIONS(658),
    [anon_sym_5_GT_AMP1] = ACTIONS(658),
    [anon_sym_6_GT_AMP1] = ACTIONS(658),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(658),
    [anon_sym_1_GT_AMP2] = ACTIONS(658),
    [anon_sym_3_GT_AMP2] = ACTIONS(658),
    [anon_sym_4_GT_AMP2] = ACTIONS(658),
    [anon_sym_5_GT_AMP2] = ACTIONS(658),
    [anon_sym_6_GT_AMP2] = ACTIONS(658),
    [aux_sym_comparison_operator_token1] = ACTIONS(658),
    [aux_sym_comparison_operator_token2] = ACTIONS(658),
    [aux_sym_comparison_operator_token3] = ACTIONS(658),
    [aux_sym_comparison_operator_token4] = ACTIONS(658),
    [aux_sym_comparison_operator_token5] = ACTIONS(658),
    [aux_sym_comparison_operator_token6] = ACTIONS(658),
    [aux_sym_comparison_operator_token7] = ACTIONS(658),
    [aux_sym_comparison_operator_token8] = ACTIONS(658),
    [aux_sym_comparison_operator_token9] = ACTIONS(658),
    [aux_sym_comparison_operator_token10] = ACTIONS(658),
    [aux_sym_comparison_operator_token11] = ACTIONS(658),
    [aux_sym_comparison_operator_token12] = ACTIONS(658),
    [aux_sym_comparison_operator_token13] = ACTIONS(658),
    [aux_sym_comparison_operator_token14] = ACTIONS(658),
    [aux_sym_comparison_operator_token15] = ACTIONS(658),
    [aux_sym_comparison_operator_token16] = ACTIONS(658),
    [aux_sym_comparison_operator_token17] = ACTIONS(658),
    [aux_sym_comparison_operator_token18] = ACTIONS(658),
    [aux_sym_comparison_operator_token19] = ACTIONS(658),
    [aux_sym_comparison_operator_token20] = ACTIONS(658),
    [aux_sym_comparison_operator_token21] = ACTIONS(658),
    [aux_sym_comparison_operator_token22] = ACTIONS(658),
    [aux_sym_comparison_operator_token23] = ACTIONS(658),
    [aux_sym_comparison_operator_token24] = ACTIONS(658),
    [aux_sym_comparison_operator_token25] = ACTIONS(658),
    [aux_sym_comparison_operator_token26] = ACTIONS(658),
    [aux_sym_comparison_operator_token27] = ACTIONS(658),
    [aux_sym_comparison_operator_token28] = ACTIONS(660),
    [aux_sym_comparison_operator_token29] = ACTIONS(658),
    [aux_sym_comparison_operator_token30] = ACTIONS(658),
    [aux_sym_comparison_operator_token31] = ACTIONS(658),
    [aux_sym_comparison_operator_token32] = ACTIONS(658),
    [aux_sym_comparison_operator_token33] = ACTIONS(658),
    [aux_sym_comparison_operator_token34] = ACTIONS(660),
    [aux_sym_comparison_operator_token35] = ACTIONS(658),
    [aux_sym_comparison_operator_token36] = ACTIONS(658),
    [aux_sym_comparison_operator_token37] = ACTIONS(658),
    [aux_sym_comparison_operator_token38] = ACTIONS(658),
    [aux_sym_comparison_operator_token39] = ACTIONS(658),
    [aux_sym_comparison_operator_token40] = ACTIONS(658),
    [aux_sym_comparison_operator_token41] = ACTIONS(658),
    [aux_sym_comparison_operator_token42] = ACTIONS(658),
    [aux_sym_comparison_operator_token43] = ACTIONS(658),
    [aux_sym_comparison_operator_token44] = ACTIONS(658),
    [aux_sym_comparison_operator_token45] = ACTIONS(658),
    [aux_sym_comparison_operator_token46] = ACTIONS(658),
    [aux_sym_comparison_operator_token47] = ACTIONS(658),
    [aux_sym_comparison_operator_token48] = ACTIONS(658),
    [aux_sym_comparison_operator_token49] = ACTIONS(658),
    [aux_sym_comparison_operator_token50] = ACTIONS(658),
    [aux_sym_format_operator_token1] = ACTIONS(658),
    [anon_sym_COMMA] = ACTIONS(658),
    [anon_sym_AMP_AMP] = ACTIONS(658),
    [anon_sym_PIPE_PIPE] = ACTIONS(658),
    [anon_sym_PIPE] = ACTIONS(660),
    [aux_sym_logical_expression_token1] = ACTIONS(658),
    [aux_sym_logical_expression_token2] = ACTIONS(658),
    [aux_sym_logical_expression_token3] = ACTIONS(658),
    [aux_sym_bitwise_expression_token1] = ACTIONS(658),
    [aux_sym_bitwise_expression_token2] = ACTIONS(658),
    [aux_sym_bitwise_expression_token3] = ACTIONS(658),
    [anon_sym_PLUS] = ACTIONS(660),
    [anon_sym_DASH] = ACTIONS(660),
    [anon_sym_SLASH] = ACTIONS(660),
    [anon_sym_BSLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(660),
    [anon_sym_STAR] = ACTIONS(660),
    [anon_sym_DOT_DOT] = ACTIONS(658),
    [sym__statement_terminator] = ACTIONS(658),
  },
  [STATE(178)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(670),
    [anon_sym_BANG_EQ] = ACTIONS(670),
    [anon_sym_PLUS_EQ] = ACTIONS(670),
    [anon_sym_STAR_EQ] = ACTIONS(670),
    [anon_sym_SLASH_EQ] = ACTIONS(670),
    [anon_sym_PERCENT_EQ] = ACTIONS(670),
    [anon_sym_DASH_EQ] = ACTIONS(670),
    [anon_sym_GT] = ACTIONS(672),
    [anon_sym_GT_GT] = ACTIONS(670),
    [anon_sym_2_GT] = ACTIONS(672),
    [anon_sym_2_GT_GT] = ACTIONS(670),
    [anon_sym_3_GT] = ACTIONS(672),
    [anon_sym_3_GT_GT] = ACTIONS(670),
    [anon_sym_4_GT] = ACTIONS(672),
    [anon_sym_4_GT_GT] = ACTIONS(670),
    [anon_sym_5_GT] = ACTIONS(672),
    [anon_sym_5_GT_GT] = ACTIONS(670),
    [anon_sym_6_GT] = ACTIONS(672),
    [anon_sym_6_GT_GT] = ACTIONS(670),
    [anon_sym_STAR_GT] = ACTIONS(672),
    [anon_sym_STAR_GT_GT] = ACTIONS(670),
    [anon_sym_LT] = ACTIONS(672),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(670),
    [anon_sym_2_GT_AMP1] = ACTIONS(670),
    [anon_sym_3_GT_AMP1] = ACTIONS(670),
    [anon_sym_4_GT_AMP1] = ACTIONS(670),
    [anon_sym_5_GT_AMP1] = ACTIONS(670),
    [anon_sym_6_GT_AMP1] = ACTIONS(670),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(670),
    [anon_sym_1_GT_AMP2] = ACTIONS(670),
    [anon_sym_3_GT_AMP2] = ACTIONS(670),
    [anon_sym_4_GT_AMP2] = ACTIONS(670),
    [anon_sym_5_GT_AMP2] = ACTIONS(670),
    [anon_sym_6_GT_AMP2] = ACTIONS(670),
    [aux_sym_comparison_operator_token1] = ACTIONS(670),
    [aux_sym_comparison_operator_token2] = ACTIONS(670),
    [aux_sym_comparison_operator_token3] = ACTIONS(670),
    [aux_sym_comparison_operator_token4] = ACTIONS(670),
    [aux_sym_comparison_operator_token5] = ACTIONS(670),
    [aux_sym_comparison_operator_token6] = ACTIONS(670),
    [aux_sym_comparison_operator_token7] = ACTIONS(670),
    [aux_sym_comparison_operator_token8] = ACTIONS(670),
    [aux_sym_comparison_operator_token9] = ACTIONS(670),
    [aux_sym_comparison_operator_token10] = ACTIONS(670),
    [aux_sym_comparison_operator_token11] = ACTIONS(670),
    [aux_sym_comparison_operator_token12] = ACTIONS(670),
    [aux_sym_comparison_operator_token13] = ACTIONS(670),
    [aux_sym_comparison_operator_token14] = ACTIONS(670),
    [aux_sym_comparison_operator_token15] = ACTIONS(670),
    [aux_sym_comparison_operator_token16] = ACTIONS(670),
    [aux_sym_comparison_operator_token17] = ACTIONS(670),
    [aux_sym_comparison_operator_token18] = ACTIONS(670),
    [aux_sym_comparison_operator_token19] = ACTIONS(670),
    [aux_sym_comparison_operator_token20] = ACTIONS(670),
    [aux_sym_comparison_operator_token21] = ACTIONS(670),
    [aux_sym_comparison_operator_token22] = ACTIONS(670),
    [aux_sym_comparison_operator_token23] = ACTIONS(670),
    [aux_sym_comparison_operator_token24] = ACTIONS(670),
    [aux_sym_comparison_operator_token25] = ACTIONS(670),
    [aux_sym_comparison_operator_token26] = ACTIONS(670),
    [aux_sym_comparison_operator_token27] = ACTIONS(670),
    [aux_sym_comparison_operator_token28] = ACTIONS(672),
    [aux_sym_comparison_operator_token29] = ACTIONS(670),
    [aux_sym_comparison_operator_token30] = ACTIONS(670),
    [aux_sym_comparison_operator_token31] = ACTIONS(670),
    [aux_sym_comparison_operator_token32] = ACTIONS(670),
    [aux_sym_comparison_operator_token33] = ACTIONS(670),
    [aux_sym_comparison_operator_token34] = ACTIONS(672),
    [aux_sym_comparison_operator_token35] = ACTIONS(670),
    [aux_sym_comparison_operator_token36] = ACTIONS(670),
    [aux_sym_comparison_operator_token37] = ACTIONS(670),
    [aux_sym_comparison_operator_token38] = ACTIONS(670),
    [aux_sym_comparison_operator_token39] = ACTIONS(670),
    [aux_sym_comparison_operator_token40] = ACTIONS(670),
    [aux_sym_comparison_operator_token41] = ACTIONS(670),
    [aux_sym_comparison_operator_token42] = ACTIONS(670),
    [aux_sym_comparison_operator_token43] = ACTIONS(670),
    [aux_sym_comparison_operator_token44] = ACTIONS(670),
    [aux_sym_comparison_operator_token45] = ACTIONS(670),
    [aux_sym_comparison_operator_token46] = ACTIONS(670),
    [aux_sym_comparison_operator_token47] = ACTIONS(670),
    [aux_sym_comparison_operator_token48] = ACTIONS(670),
    [aux_sym_comparison_operator_token49] = ACTIONS(670),
    [aux_sym_comparison_operator_token50] = ACTIONS(670),
    [aux_sym_format_operator_token1] = ACTIONS(670),
    [anon_sym_COMMA] = ACTIONS(670),
    [anon_sym_AMP_AMP] = ACTIONS(670),
    [anon_sym_PIPE_PIPE] = ACTIONS(670),
    [anon_sym_PIPE] = ACTIONS(672),
    [aux_sym_logical_expression_token1] = ACTIONS(670),
    [aux_sym_logical_expression_token2] = ACTIONS(670),
    [aux_sym_logical_expression_token3] = ACTIONS(670),
    [aux_sym_bitwise_expression_token1] = ACTIONS(670),
    [aux_sym_bitwise_expression_token2] = ACTIONS(670),
    [aux_sym_bitwise_expression_token3] = ACTIONS(670),
    [anon_sym_PLUS] = ACTIONS(672),
    [anon_sym_DASH] = ACTIONS(672),
    [anon_sym_SLASH] = ACTIONS(672),
    [anon_sym_BSLASH] = ACTIONS(670),
    [anon_sym_PERCENT] = ACTIONS(672),
    [anon_sym_STAR] = ACTIONS(672),
    [anon_sym_DOT_DOT] = ACTIONS(670),
    [sym__statement_terminator] = ACTIONS(670),
  },
  [STATE(179)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(674),
    [anon_sym_BANG_EQ] = ACTIONS(674),
    [anon_sym_PLUS_EQ] = ACTIONS(674),
    [anon_sym_STAR_EQ] = ACTIONS(674),
    [anon_sym_SLASH_EQ] = ACTIONS(674),
    [anon_sym_PERCENT_EQ] = ACTIONS(674),
    [anon_sym_DASH_EQ] = ACTIONS(674),
    [anon_sym_GT] = ACTIONS(676),
    [anon_sym_GT_GT] = ACTIONS(674),
    [anon_sym_2_GT] = ACTIONS(676),
    [anon_sym_2_GT_GT] = ACTIONS(674),
    [anon_sym_3_GT] = ACTIONS(676),
    [anon_sym_3_GT_GT] = ACTIONS(674),
    [anon_sym_4_GT] = ACTIONS(676),
    [anon_sym_4_GT_GT] = ACTIONS(674),
    [anon_sym_5_GT] = ACTIONS(676),
    [anon_sym_5_GT_GT] = ACTIONS(674),
    [anon_sym_6_GT] = ACTIONS(676),
    [anon_sym_6_GT_GT] = ACTIONS(674),
    [anon_sym_STAR_GT] = ACTIONS(676),
    [anon_sym_STAR_GT_GT] = ACTIONS(674),
    [anon_sym_LT] = ACTIONS(676),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(674),
    [anon_sym_2_GT_AMP1] = ACTIONS(674),
    [anon_sym_3_GT_AMP1] = ACTIONS(674),
    [anon_sym_4_GT_AMP1] = ACTIONS(674),
    [anon_sym_5_GT_AMP1] = ACTIONS(674),
    [anon_sym_6_GT_AMP1] = ACTIONS(674),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(674),
    [anon_sym_1_GT_AMP2] = ACTIONS(674),
    [anon_sym_3_GT_AMP2] = ACTIONS(674),
    [anon_sym_4_GT_AMP2] = ACTIONS(674),
    [anon_sym_5_GT_AMP2] = ACTIONS(674),
    [anon_sym_6_GT_AMP2] = ACTIONS(674),
    [aux_sym_comparison_operator_token1] = ACTIONS(674),
    [aux_sym_comparison_operator_token2] = ACTIONS(674),
    [aux_sym_comparison_operator_token3] = ACTIONS(674),
    [aux_sym_comparison_operator_token4] = ACTIONS(674),
    [aux_sym_comparison_operator_token5] = ACTIONS(674),
    [aux_sym_comparison_operator_token6] = ACTIONS(674),
    [aux_sym_comparison_operator_token7] = ACTIONS(674),
    [aux_sym_comparison_operator_token8] = ACTIONS(674),
    [aux_sym_comparison_operator_token9] = ACTIONS(674),
    [aux_sym_comparison_operator_token10] = ACTIONS(674),
    [aux_sym_comparison_operator_token11] = ACTIONS(674),
    [aux_sym_comparison_operator_token12] = ACTIONS(674),
    [aux_sym_comparison_operator_token13] = ACTIONS(674),
    [aux_sym_comparison_operator_token14] = ACTIONS(674),
    [aux_sym_comparison_operator_token15] = ACTIONS(674),
    [aux_sym_comparison_operator_token16] = ACTIONS(674),
    [aux_sym_comparison_operator_token17] = ACTIONS(674),
    [aux_sym_comparison_operator_token18] = ACTIONS(674),
    [aux_sym_comparison_operator_token19] = ACTIONS(674),
    [aux_sym_comparison_operator_token20] = ACTIONS(674),
    [aux_sym_comparison_operator_token21] = ACTIONS(674),
    [aux_sym_comparison_operator_token22] = ACTIONS(674),
    [aux_sym_comparison_operator_token23] = ACTIONS(674),
    [aux_sym_comparison_operator_token24] = ACTIONS(674),
    [aux_sym_comparison_operator_token25] = ACTIONS(674),
    [aux_sym_comparison_operator_token26] = ACTIONS(674),
    [aux_sym_comparison_operator_token27] = ACTIONS(674),
    [aux_sym_comparison_operator_token28] = ACTIONS(676),
    [aux_sym_comparison_operator_token29] = ACTIONS(674),
    [aux_sym_comparison_operator_token30] = ACTIONS(674),
    [aux_sym_comparison_operator_token31] = ACTIONS(674),
    [aux_sym_comparison_operator_token32] = ACTIONS(674),
    [aux_sym_comparison_operator_token33] = ACTIONS(674),
    [aux_sym_comparison_operator_token34] = ACTIONS(676),
    [aux_sym_comparison_operator_token35] = ACTIONS(674),
    [aux_sym_comparison_operator_token36] = ACTIONS(674),
    [aux_sym_comparison_operator_token37] = ACTIONS(674),
    [aux_sym_comparison_operator_token38] = ACTIONS(674),
    [aux_sym_comparison_operator_token39] = ACTIONS(674),
    [aux_sym_comparison_operator_token40] = ACTIONS(674),
    [aux_sym_comparison_operator_token41] = ACTIONS(674),
    [aux_sym_comparison_operator_token42] = ACTIONS(674),
    [aux_sym_comparison_operator_token43] = ACTIONS(674),
    [aux_sym_comparison_operator_token44] = ACTIONS(674),
    [aux_sym_comparison_operator_token45] = ACTIONS(674),
    [aux_sym_comparison_operator_token46] = ACTIONS(674),
    [aux_sym_comparison_operator_token47] = ACTIONS(674),
    [aux_sym_comparison_operator_token48] = ACTIONS(674),
    [aux_sym_comparison_operator_token49] = ACTIONS(674),
    [aux_sym_comparison_operator_token50] = ACTIONS(674),
    [aux_sym_format_operator_token1] = ACTIONS(674),
    [anon_sym_COMMA] = ACTIONS(674),
    [anon_sym_AMP_AMP] = ACTIONS(674),
    [anon_sym_PIPE_PIPE] = ACTIONS(674),
    [anon_sym_PIPE] = ACTIONS(676),
    [aux_sym_logical_expression_token1] = ACTIONS(674),
    [aux_sym_logical_expression_token2] = ACTIONS(674),
    [aux_sym_logical_expression_token3] = ACTIONS(674),
    [aux_sym_bitwise_expression_token1] = ACTIONS(674),
    [aux_sym_bitwise_expression_token2] = ACTIONS(674),
    [aux_sym_bitwise_expression_token3] = ACTIONS(674),
    [anon_sym_PLUS] = ACTIONS(676),
    [anon_sym_DASH] = ACTIONS(676),
    [anon_sym_SLASH] = ACTIONS(676),
    [anon_sym_BSLASH] = ACTIONS(674),
    [anon_sym_PERCENT] = ACTIONS(676),
    [anon_sym_STAR] = ACTIONS(676),
    [anon_sym_DOT_DOT] = ACTIONS(674),
    [sym__statement_terminator] = ACTIONS(674),
  },
  [STATE(180)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(760),
    [anon_sym_BANG_EQ] = ACTIONS(760),
    [anon_sym_PLUS_EQ] = ACTIONS(760),
    [anon_sym_STAR_EQ] = ACTIONS(760),
    [anon_sym_SLASH_EQ] = ACTIONS(760),
    [anon_sym_PERCENT_EQ] = ACTIONS(760),
    [anon_sym_DASH_EQ] = ACTIONS(760),
    [anon_sym_GT] = ACTIONS(762),
    [anon_sym_GT_GT] = ACTIONS(760),
    [anon_sym_2_GT] = ACTIONS(762),
    [anon_sym_2_GT_GT] = ACTIONS(760),
    [anon_sym_3_GT] = ACTIONS(762),
    [anon_sym_3_GT_GT] = ACTIONS(760),
    [anon_sym_4_GT] = ACTIONS(762),
    [anon_sym_4_GT_GT] = ACTIONS(760),
    [anon_sym_5_GT] = ACTIONS(762),
    [anon_sym_5_GT_GT] = ACTIONS(760),
    [anon_sym_6_GT] = ACTIONS(762),
    [anon_sym_6_GT_GT] = ACTIONS(760),
    [anon_sym_STAR_GT] = ACTIONS(762),
    [anon_sym_STAR_GT_GT] = ACTIONS(760),
    [anon_sym_LT] = ACTIONS(762),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(760),
    [anon_sym_2_GT_AMP1] = ACTIONS(760),
    [anon_sym_3_GT_AMP1] = ACTIONS(760),
    [anon_sym_4_GT_AMP1] = ACTIONS(760),
    [anon_sym_5_GT_AMP1] = ACTIONS(760),
    [anon_sym_6_GT_AMP1] = ACTIONS(760),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(760),
    [anon_sym_1_GT_AMP2] = ACTIONS(760),
    [anon_sym_3_GT_AMP2] = ACTIONS(760),
    [anon_sym_4_GT_AMP2] = ACTIONS(760),
    [anon_sym_5_GT_AMP2] = ACTIONS(760),
    [anon_sym_6_GT_AMP2] = ACTIONS(760),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [anon_sym_PIPE] = ACTIONS(762),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(762),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(762),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(762),
    [anon_sym_STAR] = ACTIONS(762),
    [anon_sym_DOT_DOT] = ACTIONS(760),
    [sym__statement_terminator] = ACTIONS(760),
  },
  [STATE(181)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(724),
    [anon_sym_BANG_EQ] = ACTIONS(724),
    [anon_sym_PLUS_EQ] = ACTIONS(724),
    [anon_sym_STAR_EQ] = ACTIONS(724),
    [anon_sym_SLASH_EQ] = ACTIONS(724),
    [anon_sym_PERCENT_EQ] = ACTIONS(724),
    [anon_sym_DASH_EQ] = ACTIONS(724),
    [anon_sym_GT] = ACTIONS(726),
    [anon_sym_GT_GT] = ACTIONS(724),
    [anon_sym_2_GT] = ACTIONS(726),
    [anon_sym_2_GT_GT] = ACTIONS(724),
    [anon_sym_3_GT] = ACTIONS(726),
    [anon_sym_3_GT_GT] = ACTIONS(724),
    [anon_sym_4_GT] = ACTIONS(726),
    [anon_sym_4_GT_GT] = ACTIONS(724),
    [anon_sym_5_GT] = ACTIONS(726),
    [anon_sym_5_GT_GT] = ACTIONS(724),
    [anon_sym_6_GT] = ACTIONS(726),
    [anon_sym_6_GT_GT] = ACTIONS(724),
    [anon_sym_STAR_GT] = ACTIONS(726),
    [anon_sym_STAR_GT_GT] = ACTIONS(724),
    [anon_sym_LT] = ACTIONS(726),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(724),
    [anon_sym_2_GT_AMP1] = ACTIONS(724),
    [anon_sym_3_GT_AMP1] = ACTIONS(724),
    [anon_sym_4_GT_AMP1] = ACTIONS(724),
    [anon_sym_5_GT_AMP1] = ACTIONS(724),
    [anon_sym_6_GT_AMP1] = ACTIONS(724),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(724),
    [anon_sym_1_GT_AMP2] = ACTIONS(724),
    [anon_sym_3_GT_AMP2] = ACTIONS(724),
    [anon_sym_4_GT_AMP2] = ACTIONS(724),
    [anon_sym_5_GT_AMP2] = ACTIONS(724),
    [anon_sym_6_GT_AMP2] = ACTIONS(724),
    [aux_sym_comparison_operator_token1] = ACTIONS(724),
    [aux_sym_comparison_operator_token2] = ACTIONS(724),
    [aux_sym_comparison_operator_token3] = ACTIONS(724),
    [aux_sym_comparison_operator_token4] = ACTIONS(724),
    [aux_sym_comparison_operator_token5] = ACTIONS(724),
    [aux_sym_comparison_operator_token6] = ACTIONS(724),
    [aux_sym_comparison_operator_token7] = ACTIONS(724),
    [aux_sym_comparison_operator_token8] = ACTIONS(724),
    [aux_sym_comparison_operator_token9] = ACTIONS(724),
    [aux_sym_comparison_operator_token10] = ACTIONS(724),
    [aux_sym_comparison_operator_token11] = ACTIONS(724),
    [aux_sym_comparison_operator_token12] = ACTIONS(724),
    [aux_sym_comparison_operator_token13] = ACTIONS(724),
    [aux_sym_comparison_operator_token14] = ACTIONS(724),
    [aux_sym_comparison_operator_token15] = ACTIONS(724),
    [aux_sym_comparison_operator_token16] = ACTIONS(724),
    [aux_sym_comparison_operator_token17] = ACTIONS(724),
    [aux_sym_comparison_operator_token18] = ACTIONS(724),
    [aux_sym_comparison_operator_token19] = ACTIONS(724),
    [aux_sym_comparison_operator_token20] = ACTIONS(724),
    [aux_sym_comparison_operator_token21] = ACTIONS(724),
    [aux_sym_comparison_operator_token22] = ACTIONS(724),
    [aux_sym_comparison_operator_token23] = ACTIONS(724),
    [aux_sym_comparison_operator_token24] = ACTIONS(724),
    [aux_sym_comparison_operator_token25] = ACTIONS(724),
    [aux_sym_comparison_operator_token26] = ACTIONS(724),
    [aux_sym_comparison_operator_token27] = ACTIONS(724),
    [aux_sym_comparison_operator_token28] = ACTIONS(726),
    [aux_sym_comparison_operator_token29] = ACTIONS(724),
    [aux_sym_comparison_operator_token30] = ACTIONS(724),
    [aux_sym_comparison_operator_token31] = ACTIONS(724),
    [aux_sym_comparison_operator_token32] = ACTIONS(724),
    [aux_sym_comparison_operator_token33] = ACTIONS(724),
    [aux_sym_comparison_operator_token34] = ACTIONS(726),
    [aux_sym_comparison_operator_token35] = ACTIONS(724),
    [aux_sym_comparison_operator_token36] = ACTIONS(724),
    [aux_sym_comparison_operator_token37] = ACTIONS(724),
    [aux_sym_comparison_operator_token38] = ACTIONS(724),
    [aux_sym_comparison_operator_token39] = ACTIONS(724),
    [aux_sym_comparison_operator_token40] = ACTIONS(724),
    [aux_sym_comparison_operator_token41] = ACTIONS(724),
    [aux_sym_comparison_operator_token42] = ACTIONS(724),
    [aux_sym_comparison_operator_token43] = ACTIONS(724),
    [aux_sym_comparison_operator_token44] = ACTIONS(724),
    [aux_sym_comparison_operator_token45] = ACTIONS(724),
    [aux_sym_comparison_operator_token46] = ACTIONS(724),
    [aux_sym_comparison_operator_token47] = ACTIONS(724),
    [aux_sym_comparison_operator_token48] = ACTIONS(724),
    [aux_sym_comparison_operator_token49] = ACTIONS(724),
    [aux_sym_comparison_operator_token50] = ACTIONS(724),
    [aux_sym_format_operator_token1] = ACTIONS(724),
    [anon_sym_COMMA] = ACTIONS(724),
    [anon_sym_AMP_AMP] = ACTIONS(724),
    [anon_sym_PIPE_PIPE] = ACTIONS(724),
    [anon_sym_PIPE] = ACTIONS(726),
    [aux_sym_logical_expression_token1] = ACTIONS(724),
    [aux_sym_logical_expression_token2] = ACTIONS(724),
    [aux_sym_logical_expression_token3] = ACTIONS(724),
    [aux_sym_bitwise_expression_token1] = ACTIONS(724),
    [aux_sym_bitwise_expression_token2] = ACTIONS(724),
    [aux_sym_bitwise_expression_token3] = ACTIONS(724),
    [anon_sym_PLUS] = ACTIONS(726),
    [anon_sym_DASH] = ACTIONS(726),
    [anon_sym_SLASH] = ACTIONS(726),
    [anon_sym_BSLASH] = ACTIONS(724),
    [anon_sym_PERCENT] = ACTIONS(726),
    [anon_sym_STAR] = ACTIONS(726),
    [anon_sym_DOT_DOT] = ACTIONS(724),
    [sym__statement_terminator] = ACTIONS(724),
  },
  [STATE(182)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(678),
    [anon_sym_BANG_EQ] = ACTIONS(678),
    [anon_sym_PLUS_EQ] = ACTIONS(678),
    [anon_sym_STAR_EQ] = ACTIONS(678),
    [anon_sym_SLASH_EQ] = ACTIONS(678),
    [anon_sym_PERCENT_EQ] = ACTIONS(678),
    [anon_sym_DASH_EQ] = ACTIONS(678),
    [anon_sym_GT] = ACTIONS(680),
    [anon_sym_GT_GT] = ACTIONS(678),
    [anon_sym_2_GT] = ACTIONS(680),
    [anon_sym_2_GT_GT] = ACTIONS(678),
    [anon_sym_3_GT] = ACTIONS(680),
    [anon_sym_3_GT_GT] = ACTIONS(678),
    [anon_sym_4_GT] = ACTIONS(680),
    [anon_sym_4_GT_GT] = ACTIONS(678),
    [anon_sym_5_GT] = ACTIONS(680),
    [anon_sym_5_GT_GT] = ACTIONS(678),
    [anon_sym_6_GT] = ACTIONS(680),
    [anon_sym_6_GT_GT] = ACTIONS(678),
    [anon_sym_STAR_GT] = ACTIONS(680),
    [anon_sym_STAR_GT_GT] = ACTIONS(678),
    [anon_sym_LT] = ACTIONS(680),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(678),
    [anon_sym_2_GT_AMP1] = ACTIONS(678),
    [anon_sym_3_GT_AMP1] = ACTIONS(678),
    [anon_sym_4_GT_AMP1] = ACTIONS(678),
    [anon_sym_5_GT_AMP1] = ACTIONS(678),
    [anon_sym_6_GT_AMP1] = ACTIONS(678),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(678),
    [anon_sym_1_GT_AMP2] = ACTIONS(678),
    [anon_sym_3_GT_AMP2] = ACTIONS(678),
    [anon_sym_4_GT_AMP2] = ACTIONS(678),
    [anon_sym_5_GT_AMP2] = ACTIONS(678),
    [anon_sym_6_GT_AMP2] = ACTIONS(678),
    [aux_sym_comparison_operator_token1] = ACTIONS(678),
    [aux_sym_comparison_operator_token2] = ACTIONS(678),
    [aux_sym_comparison_operator_token3] = ACTIONS(678),
    [aux_sym_comparison_operator_token4] = ACTIONS(678),
    [aux_sym_comparison_operator_token5] = ACTIONS(678),
    [aux_sym_comparison_operator_token6] = ACTIONS(678),
    [aux_sym_comparison_operator_token7] = ACTIONS(678),
    [aux_sym_comparison_operator_token8] = ACTIONS(678),
    [aux_sym_comparison_operator_token9] = ACTIONS(678),
    [aux_sym_comparison_operator_token10] = ACTIONS(678),
    [aux_sym_comparison_operator_token11] = ACTIONS(678),
    [aux_sym_comparison_operator_token12] = ACTIONS(678),
    [aux_sym_comparison_operator_token13] = ACTIONS(678),
    [aux_sym_comparison_operator_token14] = ACTIONS(678),
    [aux_sym_comparison_operator_token15] = ACTIONS(678),
    [aux_sym_comparison_operator_token16] = ACTIONS(678),
    [aux_sym_comparison_operator_token17] = ACTIONS(678),
    [aux_sym_comparison_operator_token18] = ACTIONS(678),
    [aux_sym_comparison_operator_token19] = ACTIONS(678),
    [aux_sym_comparison_operator_token20] = ACTIONS(678),
    [aux_sym_comparison_operator_token21] = ACTIONS(678),
    [aux_sym_comparison_operator_token22] = ACTIONS(678),
    [aux_sym_comparison_operator_token23] = ACTIONS(678),
    [aux_sym_comparison_operator_token24] = ACTIONS(678),
    [aux_sym_comparison_operator_token25] = ACTIONS(678),
    [aux_sym_comparison_operator_token26] = ACTIONS(678),
    [aux_sym_comparison_operator_token27] = ACTIONS(678),
    [aux_sym_comparison_operator_token28] = ACTIONS(680),
    [aux_sym_comparison_operator_token29] = ACTIONS(678),
    [aux_sym_comparison_operator_token30] = ACTIONS(678),
    [aux_sym_comparison_operator_token31] = ACTIONS(678),
    [aux_sym_comparison_operator_token32] = ACTIONS(678),
    [aux_sym_comparison_operator_token33] = ACTIONS(678),
    [aux_sym_comparison_operator_token34] = ACTIONS(680),
    [aux_sym_comparison_operator_token35] = ACTIONS(678),
    [aux_sym_comparison_operator_token36] = ACTIONS(678),
    [aux_sym_comparison_operator_token37] = ACTIONS(678),
    [aux_sym_comparison_operator_token38] = ACTIONS(678),
    [aux_sym_comparison_operator_token39] = ACTIONS(678),
    [aux_sym_comparison_operator_token40] = ACTIONS(678),
    [aux_sym_comparison_operator_token41] = ACTIONS(678),
    [aux_sym_comparison_operator_token42] = ACTIONS(678),
    [aux_sym_comparison_operator_token43] = ACTIONS(678),
    [aux_sym_comparison_operator_token44] = ACTIONS(678),
    [aux_sym_comparison_operator_token45] = ACTIONS(678),
    [aux_sym_comparison_operator_token46] = ACTIONS(678),
    [aux_sym_comparison_operator_token47] = ACTIONS(678),
    [aux_sym_comparison_operator_token48] = ACTIONS(678),
    [aux_sym_comparison_operator_token49] = ACTIONS(678),
    [aux_sym_comparison_operator_token50] = ACTIONS(678),
    [aux_sym_format_operator_token1] = ACTIONS(678),
    [anon_sym_COMMA] = ACTIONS(678),
    [anon_sym_AMP_AMP] = ACTIONS(678),
    [anon_sym_PIPE_PIPE] = ACTIONS(678),
    [anon_sym_PIPE] = ACTIONS(680),
    [aux_sym_logical_expression_token1] = ACTIONS(678),
    [aux_sym_logical_expression_token2] = ACTIONS(678),
    [aux_sym_logical_expression_token3] = ACTIONS(678),
    [aux_sym_bitwise_expression_token1] = ACTIONS(678),
    [aux_sym_bitwise_expression_token2] = ACTIONS(678),
    [aux_sym_bitwise_expression_token3] = ACTIONS(678),
    [anon_sym_PLUS] = ACTIONS(680),
    [anon_sym_DASH] = ACTIONS(680),
    [anon_sym_SLASH] = ACTIONS(680),
    [anon_sym_BSLASH] = ACTIONS(678),
    [anon_sym_PERCENT] = ACTIONS(680),
    [anon_sym_STAR] = ACTIONS(680),
    [anon_sym_DOT_DOT] = ACTIONS(678),
    [sym__statement_terminator] = ACTIONS(678),
  },
  [STATE(183)] = {
    [sym_format_operator] = STATE(586),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(790),
    [anon_sym_BANG_EQ] = ACTIONS(790),
    [anon_sym_PLUS_EQ] = ACTIONS(790),
    [anon_sym_STAR_EQ] = ACTIONS(790),
    [anon_sym_SLASH_EQ] = ACTIONS(790),
    [anon_sym_PERCENT_EQ] = ACTIONS(790),
    [anon_sym_DASH_EQ] = ACTIONS(790),
    [anon_sym_GT] = ACTIONS(792),
    [anon_sym_GT_GT] = ACTIONS(790),
    [anon_sym_2_GT] = ACTIONS(792),
    [anon_sym_2_GT_GT] = ACTIONS(790),
    [anon_sym_3_GT] = ACTIONS(792),
    [anon_sym_3_GT_GT] = ACTIONS(790),
    [anon_sym_4_GT] = ACTIONS(792),
    [anon_sym_4_GT_GT] = ACTIONS(790),
    [anon_sym_5_GT] = ACTIONS(792),
    [anon_sym_5_GT_GT] = ACTIONS(790),
    [anon_sym_6_GT] = ACTIONS(792),
    [anon_sym_6_GT_GT] = ACTIONS(790),
    [anon_sym_STAR_GT] = ACTIONS(792),
    [anon_sym_STAR_GT_GT] = ACTIONS(790),
    [anon_sym_LT] = ACTIONS(792),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(790),
    [anon_sym_2_GT_AMP1] = ACTIONS(790),
    [anon_sym_3_GT_AMP1] = ACTIONS(790),
    [anon_sym_4_GT_AMP1] = ACTIONS(790),
    [anon_sym_5_GT_AMP1] = ACTIONS(790),
    [anon_sym_6_GT_AMP1] = ACTIONS(790),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(790),
    [anon_sym_1_GT_AMP2] = ACTIONS(790),
    [anon_sym_3_GT_AMP2] = ACTIONS(790),
    [anon_sym_4_GT_AMP2] = ACTIONS(790),
    [anon_sym_5_GT_AMP2] = ACTIONS(790),
    [anon_sym_6_GT_AMP2] = ACTIONS(790),
    [aux_sym_comparison_operator_token1] = ACTIONS(790),
    [aux_sym_comparison_operator_token2] = ACTIONS(790),
    [aux_sym_comparison_operator_token3] = ACTIONS(790),
    [aux_sym_comparison_operator_token4] = ACTIONS(790),
    [aux_sym_comparison_operator_token5] = ACTIONS(790),
    [aux_sym_comparison_operator_token6] = ACTIONS(790),
    [aux_sym_comparison_operator_token7] = ACTIONS(790),
    [aux_sym_comparison_operator_token8] = ACTIONS(790),
    [aux_sym_comparison_operator_token9] = ACTIONS(790),
    [aux_sym_comparison_operator_token10] = ACTIONS(790),
    [aux_sym_comparison_operator_token11] = ACTIONS(790),
    [aux_sym_comparison_operator_token12] = ACTIONS(790),
    [aux_sym_comparison_operator_token13] = ACTIONS(790),
    [aux_sym_comparison_operator_token14] = ACTIONS(790),
    [aux_sym_comparison_operator_token15] = ACTIONS(790),
    [aux_sym_comparison_operator_token16] = ACTIONS(790),
    [aux_sym_comparison_operator_token17] = ACTIONS(790),
    [aux_sym_comparison_operator_token18] = ACTIONS(790),
    [aux_sym_comparison_operator_token19] = ACTIONS(790),
    [aux_sym_comparison_operator_token20] = ACTIONS(790),
    [aux_sym_comparison_operator_token21] = ACTIONS(790),
    [aux_sym_comparison_operator_token22] = ACTIONS(790),
    [aux_sym_comparison_operator_token23] = ACTIONS(790),
    [aux_sym_comparison_operator_token24] = ACTIONS(790),
    [aux_sym_comparison_operator_token25] = ACTIONS(790),
    [aux_sym_comparison_operator_token26] = ACTIONS(790),
    [aux_sym_comparison_operator_token27] = ACTIONS(790),
    [aux_sym_comparison_operator_token28] = ACTIONS(792),
    [aux_sym_comparison_operator_token29] = ACTIONS(790),
    [aux_sym_comparison_operator_token30] = ACTIONS(790),
    [aux_sym_comparison_operator_token31] = ACTIONS(790),
    [aux_sym_comparison_operator_token32] = ACTIONS(790),
    [aux_sym_comparison_operator_token33] = ACTIONS(790),
    [aux_sym_comparison_operator_token34] = ACTIONS(792),
    [aux_sym_comparison_operator_token35] = ACTIONS(790),
    [aux_sym_comparison_operator_token36] = ACTIONS(790),
    [aux_sym_comparison_operator_token37] = ACTIONS(790),
    [aux_sym_comparison_operator_token38] = ACTIONS(790),
    [aux_sym_comparison_operator_token39] = ACTIONS(790),
    [aux_sym_comparison_operator_token40] = ACTIONS(790),
    [aux_sym_comparison_operator_token41] = ACTIONS(790),
    [aux_sym_comparison_operator_token42] = ACTIONS(790),
    [aux_sym_comparison_operator_token43] = ACTIONS(790),
    [aux_sym_comparison_operator_token44] = ACTIONS(790),
    [aux_sym_comparison_operator_token45] = ACTIONS(790),
    [aux_sym_comparison_operator_token46] = ACTIONS(790),
    [aux_sym_comparison_operator_token47] = ACTIONS(790),
    [aux_sym_comparison_operator_token48] = ACTIONS(790),
    [aux_sym_comparison_operator_token49] = ACTIONS(790),
    [aux_sym_comparison_operator_token50] = ACTIONS(790),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_AMP_AMP] = ACTIONS(790),
    [anon_sym_PIPE_PIPE] = ACTIONS(790),
    [anon_sym_PIPE] = ACTIONS(792),
    [aux_sym_logical_expression_token1] = ACTIONS(790),
    [aux_sym_logical_expression_token2] = ACTIONS(790),
    [aux_sym_logical_expression_token3] = ACTIONS(790),
    [aux_sym_bitwise_expression_token1] = ACTIONS(790),
    [aux_sym_bitwise_expression_token2] = ACTIONS(790),
    [aux_sym_bitwise_expression_token3] = ACTIONS(790),
    [anon_sym_PLUS] = ACTIONS(792),
    [anon_sym_DASH] = ACTIONS(792),
    [anon_sym_SLASH] = ACTIONS(792),
    [anon_sym_BSLASH] = ACTIONS(790),
    [anon_sym_PERCENT] = ACTIONS(792),
    [anon_sym_STAR] = ACTIONS(792),
    [sym__statement_terminator] = ACTIONS(790),
  },
  [STATE(184)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(764),
    [anon_sym_BANG_EQ] = ACTIONS(764),
    [anon_sym_PLUS_EQ] = ACTIONS(764),
    [anon_sym_STAR_EQ] = ACTIONS(764),
    [anon_sym_SLASH_EQ] = ACTIONS(764),
    [anon_sym_PERCENT_EQ] = ACTIONS(764),
    [anon_sym_DASH_EQ] = ACTIONS(764),
    [anon_sym_GT] = ACTIONS(766),
    [anon_sym_GT_GT] = ACTIONS(764),
    [anon_sym_2_GT] = ACTIONS(766),
    [anon_sym_2_GT_GT] = ACTIONS(764),
    [anon_sym_3_GT] = ACTIONS(766),
    [anon_sym_3_GT_GT] = ACTIONS(764),
    [anon_sym_4_GT] = ACTIONS(766),
    [anon_sym_4_GT_GT] = ACTIONS(764),
    [anon_sym_5_GT] = ACTIONS(766),
    [anon_sym_5_GT_GT] = ACTIONS(764),
    [anon_sym_6_GT] = ACTIONS(766),
    [anon_sym_6_GT_GT] = ACTIONS(764),
    [anon_sym_STAR_GT] = ACTIONS(766),
    [anon_sym_STAR_GT_GT] = ACTIONS(764),
    [anon_sym_LT] = ACTIONS(766),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(764),
    [anon_sym_2_GT_AMP1] = ACTIONS(764),
    [anon_sym_3_GT_AMP1] = ACTIONS(764),
    [anon_sym_4_GT_AMP1] = ACTIONS(764),
    [anon_sym_5_GT_AMP1] = ACTIONS(764),
    [anon_sym_6_GT_AMP1] = ACTIONS(764),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(764),
    [anon_sym_1_GT_AMP2] = ACTIONS(764),
    [anon_sym_3_GT_AMP2] = ACTIONS(764),
    [anon_sym_4_GT_AMP2] = ACTIONS(764),
    [anon_sym_5_GT_AMP2] = ACTIONS(764),
    [anon_sym_6_GT_AMP2] = ACTIONS(764),
    [aux_sym_comparison_operator_token1] = ACTIONS(764),
    [aux_sym_comparison_operator_token2] = ACTIONS(764),
    [aux_sym_comparison_operator_token3] = ACTIONS(764),
    [aux_sym_comparison_operator_token4] = ACTIONS(764),
    [aux_sym_comparison_operator_token5] = ACTIONS(764),
    [aux_sym_comparison_operator_token6] = ACTIONS(764),
    [aux_sym_comparison_operator_token7] = ACTIONS(764),
    [aux_sym_comparison_operator_token8] = ACTIONS(764),
    [aux_sym_comparison_operator_token9] = ACTIONS(764),
    [aux_sym_comparison_operator_token10] = ACTIONS(764),
    [aux_sym_comparison_operator_token11] = ACTIONS(764),
    [aux_sym_comparison_operator_token12] = ACTIONS(764),
    [aux_sym_comparison_operator_token13] = ACTIONS(764),
    [aux_sym_comparison_operator_token14] = ACTIONS(764),
    [aux_sym_comparison_operator_token15] = ACTIONS(764),
    [aux_sym_comparison_operator_token16] = ACTIONS(764),
    [aux_sym_comparison_operator_token17] = ACTIONS(764),
    [aux_sym_comparison_operator_token18] = ACTIONS(764),
    [aux_sym_comparison_operator_token19] = ACTIONS(764),
    [aux_sym_comparison_operator_token20] = ACTIONS(764),
    [aux_sym_comparison_operator_token21] = ACTIONS(764),
    [aux_sym_comparison_operator_token22] = ACTIONS(764),
    [aux_sym_comparison_operator_token23] = ACTIONS(764),
    [aux_sym_comparison_operator_token24] = ACTIONS(764),
    [aux_sym_comparison_operator_token25] = ACTIONS(764),
    [aux_sym_comparison_operator_token26] = ACTIONS(764),
    [aux_sym_comparison_operator_token27] = ACTIONS(764),
    [aux_sym_comparison_operator_token28] = ACTIONS(766),
    [aux_sym_comparison_operator_token29] = ACTIONS(764),
    [aux_sym_comparison_operator_token30] = ACTIONS(764),
    [aux_sym_comparison_operator_token31] = ACTIONS(764),
    [aux_sym_comparison_operator_token32] = ACTIONS(764),
    [aux_sym_comparison_operator_token33] = ACTIONS(764),
    [aux_sym_comparison_operator_token34] = ACTIONS(766),
    [aux_sym_comparison_operator_token35] = ACTIONS(764),
    [aux_sym_comparison_operator_token36] = ACTIONS(764),
    [aux_sym_comparison_operator_token37] = ACTIONS(764),
    [aux_sym_comparison_operator_token38] = ACTIONS(764),
    [aux_sym_comparison_operator_token39] = ACTIONS(764),
    [aux_sym_comparison_operator_token40] = ACTIONS(764),
    [aux_sym_comparison_operator_token41] = ACTIONS(764),
    [aux_sym_comparison_operator_token42] = ACTIONS(764),
    [aux_sym_comparison_operator_token43] = ACTIONS(764),
    [aux_sym_comparison_operator_token44] = ACTIONS(764),
    [aux_sym_comparison_operator_token45] = ACTIONS(764),
    [aux_sym_comparison_operator_token46] = ACTIONS(764),
    [aux_sym_comparison_operator_token47] = ACTIONS(764),
    [aux_sym_comparison_operator_token48] = ACTIONS(764),
    [aux_sym_comparison_operator_token49] = ACTIONS(764),
    [aux_sym_comparison_operator_token50] = ACTIONS(764),
    [aux_sym_format_operator_token1] = ACTIONS(764),
    [anon_sym_AMP_AMP] = ACTIONS(764),
    [anon_sym_PIPE_PIPE] = ACTIONS(764),
    [anon_sym_PIPE] = ACTIONS(766),
    [aux_sym_logical_expression_token1] = ACTIONS(764),
    [aux_sym_logical_expression_token2] = ACTIONS(764),
    [aux_sym_logical_expression_token3] = ACTIONS(764),
    [aux_sym_bitwise_expression_token1] = ACTIONS(764),
    [aux_sym_bitwise_expression_token2] = ACTIONS(764),
    [aux_sym_bitwise_expression_token3] = ACTIONS(764),
    [anon_sym_PLUS] = ACTIONS(766),
    [anon_sym_DASH] = ACTIONS(766),
    [anon_sym_SLASH] = ACTIONS(766),
    [anon_sym_BSLASH] = ACTIONS(764),
    [anon_sym_PERCENT] = ACTIONS(766),
    [anon_sym_STAR] = ACTIONS(766),
    [anon_sym_DOT_DOT] = ACTIONS(764),
    [sym__statement_terminator] = ACTIONS(764),
  },
  [STATE(185)] = {
    [sym_format_operator] = STATE(586),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(796),
    [anon_sym_BANG_EQ] = ACTIONS(796),
    [anon_sym_PLUS_EQ] = ACTIONS(796),
    [anon_sym_STAR_EQ] = ACTIONS(796),
    [anon_sym_SLASH_EQ] = ACTIONS(796),
    [anon_sym_PERCENT_EQ] = ACTIONS(796),
    [anon_sym_DASH_EQ] = ACTIONS(796),
    [anon_sym_GT] = ACTIONS(798),
    [anon_sym_GT_GT] = ACTIONS(796),
    [anon_sym_2_GT] = ACTIONS(798),
    [anon_sym_2_GT_GT] = ACTIONS(796),
    [anon_sym_3_GT] = ACTIONS(798),
    [anon_sym_3_GT_GT] = ACTIONS(796),
    [anon_sym_4_GT] = ACTIONS(798),
    [anon_sym_4_GT_GT] = ACTIONS(796),
    [anon_sym_5_GT] = ACTIONS(798),
    [anon_sym_5_GT_GT] = ACTIONS(796),
    [anon_sym_6_GT] = ACTIONS(798),
    [anon_sym_6_GT_GT] = ACTIONS(796),
    [anon_sym_STAR_GT] = ACTIONS(798),
    [anon_sym_STAR_GT_GT] = ACTIONS(796),
    [anon_sym_LT] = ACTIONS(798),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(796),
    [anon_sym_2_GT_AMP1] = ACTIONS(796),
    [anon_sym_3_GT_AMP1] = ACTIONS(796),
    [anon_sym_4_GT_AMP1] = ACTIONS(796),
    [anon_sym_5_GT_AMP1] = ACTIONS(796),
    [anon_sym_6_GT_AMP1] = ACTIONS(796),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(796),
    [anon_sym_1_GT_AMP2] = ACTIONS(796),
    [anon_sym_3_GT_AMP2] = ACTIONS(796),
    [anon_sym_4_GT_AMP2] = ACTIONS(796),
    [anon_sym_5_GT_AMP2] = ACTIONS(796),
    [anon_sym_6_GT_AMP2] = ACTIONS(796),
    [aux_sym_comparison_operator_token1] = ACTIONS(796),
    [aux_sym_comparison_operator_token2] = ACTIONS(796),
    [aux_sym_comparison_operator_token3] = ACTIONS(796),
    [aux_sym_comparison_operator_token4] = ACTIONS(796),
    [aux_sym_comparison_operator_token5] = ACTIONS(796),
    [aux_sym_comparison_operator_token6] = ACTIONS(796),
    [aux_sym_comparison_operator_token7] = ACTIONS(796),
    [aux_sym_comparison_operator_token8] = ACTIONS(796),
    [aux_sym_comparison_operator_token9] = ACTIONS(796),
    [aux_sym_comparison_operator_token10] = ACTIONS(796),
    [aux_sym_comparison_operator_token11] = ACTIONS(796),
    [aux_sym_comparison_operator_token12] = ACTIONS(796),
    [aux_sym_comparison_operator_token13] = ACTIONS(796),
    [aux_sym_comparison_operator_token14] = ACTIONS(796),
    [aux_sym_comparison_operator_token15] = ACTIONS(796),
    [aux_sym_comparison_operator_token16] = ACTIONS(796),
    [aux_sym_comparison_operator_token17] = ACTIONS(796),
    [aux_sym_comparison_operator_token18] = ACTIONS(796),
    [aux_sym_comparison_operator_token19] = ACTIONS(796),
    [aux_sym_comparison_operator_token20] = ACTIONS(796),
    [aux_sym_comparison_operator_token21] = ACTIONS(796),
    [aux_sym_comparison_operator_token22] = ACTIONS(796),
    [aux_sym_comparison_operator_token23] = ACTIONS(796),
    [aux_sym_comparison_operator_token24] = ACTIONS(796),
    [aux_sym_comparison_operator_token25] = ACTIONS(796),
    [aux_sym_comparison_operator_token26] = ACTIONS(796),
    [aux_sym_comparison_operator_token27] = ACTIONS(796),
    [aux_sym_comparison_operator_token28] = ACTIONS(798),
    [aux_sym_comparison_operator_token29] = ACTIONS(796),
    [aux_sym_comparison_operator_token30] = ACTIONS(796),
    [aux_sym_comparison_operator_token31] = ACTIONS(796),
    [aux_sym_comparison_operator_token32] = ACTIONS(796),
    [aux_sym_comparison_operator_token33] = ACTIONS(796),
    [aux_sym_comparison_operator_token34] = ACTIONS(798),
    [aux_sym_comparison_operator_token35] = ACTIONS(796),
    [aux_sym_comparison_operator_token36] = ACTIONS(796),
    [aux_sym_comparison_operator_token37] = ACTIONS(796),
    [aux_sym_comparison_operator_token38] = ACTIONS(796),
    [aux_sym_comparison_operator_token39] = ACTIONS(796),
    [aux_sym_comparison_operator_token40] = ACTIONS(796),
    [aux_sym_comparison_operator_token41] = ACTIONS(796),
    [aux_sym_comparison_operator_token42] = ACTIONS(796),
    [aux_sym_comparison_operator_token43] = ACTIONS(796),
    [aux_sym_comparison_operator_token44] = ACTIONS(796),
    [aux_sym_comparison_operator_token45] = ACTIONS(796),
    [aux_sym_comparison_operator_token46] = ACTIONS(796),
    [aux_sym_comparison_operator_token47] = ACTIONS(796),
    [aux_sym_comparison_operator_token48] = ACTIONS(796),
    [aux_sym_comparison_operator_token49] = ACTIONS(796),
    [aux_sym_comparison_operator_token50] = ACTIONS(796),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_AMP_AMP] = ACTIONS(796),
    [anon_sym_PIPE_PIPE] = ACTIONS(796),
    [anon_sym_PIPE] = ACTIONS(798),
    [aux_sym_logical_expression_token1] = ACTIONS(796),
    [aux_sym_logical_expression_token2] = ACTIONS(796),
    [aux_sym_logical_expression_token3] = ACTIONS(796),
    [aux_sym_bitwise_expression_token1] = ACTIONS(796),
    [aux_sym_bitwise_expression_token2] = ACTIONS(796),
    [aux_sym_bitwise_expression_token3] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(798),
    [anon_sym_SLASH] = ACTIONS(798),
    [anon_sym_BSLASH] = ACTIONS(796),
    [anon_sym_PERCENT] = ACTIONS(798),
    [anon_sym_STAR] = ACTIONS(798),
    [sym__statement_terminator] = ACTIONS(796),
  },
  [STATE(186)] = {
    [sym_format_operator] = STATE(588),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(790),
    [anon_sym_BANG_EQ] = ACTIONS(790),
    [anon_sym_PLUS_EQ] = ACTIONS(790),
    [anon_sym_STAR_EQ] = ACTIONS(790),
    [anon_sym_SLASH_EQ] = ACTIONS(790),
    [anon_sym_PERCENT_EQ] = ACTIONS(790),
    [anon_sym_DASH_EQ] = ACTIONS(790),
    [anon_sym_GT] = ACTIONS(792),
    [anon_sym_GT_GT] = ACTIONS(790),
    [anon_sym_2_GT] = ACTIONS(792),
    [anon_sym_2_GT_GT] = ACTIONS(790),
    [anon_sym_3_GT] = ACTIONS(792),
    [anon_sym_3_GT_GT] = ACTIONS(790),
    [anon_sym_4_GT] = ACTIONS(792),
    [anon_sym_4_GT_GT] = ACTIONS(790),
    [anon_sym_5_GT] = ACTIONS(792),
    [anon_sym_5_GT_GT] = ACTIONS(790),
    [anon_sym_6_GT] = ACTIONS(792),
    [anon_sym_6_GT_GT] = ACTIONS(790),
    [anon_sym_STAR_GT] = ACTIONS(792),
    [anon_sym_STAR_GT_GT] = ACTIONS(790),
    [anon_sym_LT] = ACTIONS(792),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(790),
    [anon_sym_2_GT_AMP1] = ACTIONS(790),
    [anon_sym_3_GT_AMP1] = ACTIONS(790),
    [anon_sym_4_GT_AMP1] = ACTIONS(790),
    [anon_sym_5_GT_AMP1] = ACTIONS(790),
    [anon_sym_6_GT_AMP1] = ACTIONS(790),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(790),
    [anon_sym_1_GT_AMP2] = ACTIONS(790),
    [anon_sym_3_GT_AMP2] = ACTIONS(790),
    [anon_sym_4_GT_AMP2] = ACTIONS(790),
    [anon_sym_5_GT_AMP2] = ACTIONS(790),
    [anon_sym_6_GT_AMP2] = ACTIONS(790),
    [aux_sym_comparison_operator_token1] = ACTIONS(790),
    [aux_sym_comparison_operator_token2] = ACTIONS(790),
    [aux_sym_comparison_operator_token3] = ACTIONS(790),
    [aux_sym_comparison_operator_token4] = ACTIONS(790),
    [aux_sym_comparison_operator_token5] = ACTIONS(790),
    [aux_sym_comparison_operator_token6] = ACTIONS(790),
    [aux_sym_comparison_operator_token7] = ACTIONS(790),
    [aux_sym_comparison_operator_token8] = ACTIONS(790),
    [aux_sym_comparison_operator_token9] = ACTIONS(790),
    [aux_sym_comparison_operator_token10] = ACTIONS(790),
    [aux_sym_comparison_operator_token11] = ACTIONS(790),
    [aux_sym_comparison_operator_token12] = ACTIONS(790),
    [aux_sym_comparison_operator_token13] = ACTIONS(790),
    [aux_sym_comparison_operator_token14] = ACTIONS(790),
    [aux_sym_comparison_operator_token15] = ACTIONS(790),
    [aux_sym_comparison_operator_token16] = ACTIONS(790),
    [aux_sym_comparison_operator_token17] = ACTIONS(790),
    [aux_sym_comparison_operator_token18] = ACTIONS(790),
    [aux_sym_comparison_operator_token19] = ACTIONS(790),
    [aux_sym_comparison_operator_token20] = ACTIONS(790),
    [aux_sym_comparison_operator_token21] = ACTIONS(790),
    [aux_sym_comparison_operator_token22] = ACTIONS(790),
    [aux_sym_comparison_operator_token23] = ACTIONS(790),
    [aux_sym_comparison_operator_token24] = ACTIONS(790),
    [aux_sym_comparison_operator_token25] = ACTIONS(790),
    [aux_sym_comparison_operator_token26] = ACTIONS(790),
    [aux_sym_comparison_operator_token27] = ACTIONS(790),
    [aux_sym_comparison_operator_token28] = ACTIONS(792),
    [aux_sym_comparison_operator_token29] = ACTIONS(790),
    [aux_sym_comparison_operator_token30] = ACTIONS(790),
    [aux_sym_comparison_operator_token31] = ACTIONS(790),
    [aux_sym_comparison_operator_token32] = ACTIONS(790),
    [aux_sym_comparison_operator_token33] = ACTIONS(790),
    [aux_sym_comparison_operator_token34] = ACTIONS(792),
    [aux_sym_comparison_operator_token35] = ACTIONS(790),
    [aux_sym_comparison_operator_token36] = ACTIONS(790),
    [aux_sym_comparison_operator_token37] = ACTIONS(790),
    [aux_sym_comparison_operator_token38] = ACTIONS(790),
    [aux_sym_comparison_operator_token39] = ACTIONS(790),
    [aux_sym_comparison_operator_token40] = ACTIONS(790),
    [aux_sym_comparison_operator_token41] = ACTIONS(790),
    [aux_sym_comparison_operator_token42] = ACTIONS(790),
    [aux_sym_comparison_operator_token43] = ACTIONS(790),
    [aux_sym_comparison_operator_token44] = ACTIONS(790),
    [aux_sym_comparison_operator_token45] = ACTIONS(790),
    [aux_sym_comparison_operator_token46] = ACTIONS(790),
    [aux_sym_comparison_operator_token47] = ACTIONS(790),
    [aux_sym_comparison_operator_token48] = ACTIONS(790),
    [aux_sym_comparison_operator_token49] = ACTIONS(790),
    [aux_sym_comparison_operator_token50] = ACTIONS(790),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_RPAREN] = ACTIONS(790),
    [anon_sym_AMP_AMP] = ACTIONS(790),
    [anon_sym_PIPE_PIPE] = ACTIONS(790),
    [anon_sym_PIPE] = ACTIONS(792),
    [aux_sym_logical_expression_token1] = ACTIONS(790),
    [aux_sym_logical_expression_token2] = ACTIONS(790),
    [aux_sym_logical_expression_token3] = ACTIONS(790),
    [aux_sym_bitwise_expression_token1] = ACTIONS(790),
    [aux_sym_bitwise_expression_token2] = ACTIONS(790),
    [aux_sym_bitwise_expression_token3] = ACTIONS(790),
    [anon_sym_PLUS] = ACTIONS(792),
    [anon_sym_DASH] = ACTIONS(792),
    [anon_sym_SLASH] = ACTIONS(792),
    [anon_sym_BSLASH] = ACTIONS(790),
    [anon_sym_PERCENT] = ACTIONS(792),
    [anon_sym_STAR] = ACTIONS(792),
  },
  [STATE(187)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(800),
    [anon_sym_BANG_EQ] = ACTIONS(800),
    [anon_sym_PLUS_EQ] = ACTIONS(800),
    [anon_sym_STAR_EQ] = ACTIONS(800),
    [anon_sym_SLASH_EQ] = ACTIONS(800),
    [anon_sym_PERCENT_EQ] = ACTIONS(800),
    [anon_sym_DASH_EQ] = ACTIONS(800),
    [anon_sym_GT] = ACTIONS(802),
    [anon_sym_GT_GT] = ACTIONS(800),
    [anon_sym_2_GT] = ACTIONS(802),
    [anon_sym_2_GT_GT] = ACTIONS(800),
    [anon_sym_3_GT] = ACTIONS(802),
    [anon_sym_3_GT_GT] = ACTIONS(800),
    [anon_sym_4_GT] = ACTIONS(802),
    [anon_sym_4_GT_GT] = ACTIONS(800),
    [anon_sym_5_GT] = ACTIONS(802),
    [anon_sym_5_GT_GT] = ACTIONS(800),
    [anon_sym_6_GT] = ACTIONS(802),
    [anon_sym_6_GT_GT] = ACTIONS(800),
    [anon_sym_STAR_GT] = ACTIONS(802),
    [anon_sym_STAR_GT_GT] = ACTIONS(800),
    [anon_sym_LT] = ACTIONS(802),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(800),
    [anon_sym_2_GT_AMP1] = ACTIONS(800),
    [anon_sym_3_GT_AMP1] = ACTIONS(800),
    [anon_sym_4_GT_AMP1] = ACTIONS(800),
    [anon_sym_5_GT_AMP1] = ACTIONS(800),
    [anon_sym_6_GT_AMP1] = ACTIONS(800),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(800),
    [anon_sym_1_GT_AMP2] = ACTIONS(800),
    [anon_sym_3_GT_AMP2] = ACTIONS(800),
    [anon_sym_4_GT_AMP2] = ACTIONS(800),
    [anon_sym_5_GT_AMP2] = ACTIONS(800),
    [anon_sym_6_GT_AMP2] = ACTIONS(800),
    [aux_sym_comparison_operator_token1] = ACTIONS(800),
    [aux_sym_comparison_operator_token2] = ACTIONS(800),
    [aux_sym_comparison_operator_token3] = ACTIONS(800),
    [aux_sym_comparison_operator_token4] = ACTIONS(800),
    [aux_sym_comparison_operator_token5] = ACTIONS(800),
    [aux_sym_comparison_operator_token6] = ACTIONS(800),
    [aux_sym_comparison_operator_token7] = ACTIONS(800),
    [aux_sym_comparison_operator_token8] = ACTIONS(800),
    [aux_sym_comparison_operator_token9] = ACTIONS(800),
    [aux_sym_comparison_operator_token10] = ACTIONS(800),
    [aux_sym_comparison_operator_token11] = ACTIONS(800),
    [aux_sym_comparison_operator_token12] = ACTIONS(800),
    [aux_sym_comparison_operator_token13] = ACTIONS(800),
    [aux_sym_comparison_operator_token14] = ACTIONS(800),
    [aux_sym_comparison_operator_token15] = ACTIONS(800),
    [aux_sym_comparison_operator_token16] = ACTIONS(800),
    [aux_sym_comparison_operator_token17] = ACTIONS(800),
    [aux_sym_comparison_operator_token18] = ACTIONS(800),
    [aux_sym_comparison_operator_token19] = ACTIONS(800),
    [aux_sym_comparison_operator_token20] = ACTIONS(800),
    [aux_sym_comparison_operator_token21] = ACTIONS(800),
    [aux_sym_comparison_operator_token22] = ACTIONS(800),
    [aux_sym_comparison_operator_token23] = ACTIONS(800),
    [aux_sym_comparison_operator_token24] = ACTIONS(800),
    [aux_sym_comparison_operator_token25] = ACTIONS(800),
    [aux_sym_comparison_operator_token26] = ACTIONS(800),
    [aux_sym_comparison_operator_token27] = ACTIONS(800),
    [aux_sym_comparison_operator_token28] = ACTIONS(802),
    [aux_sym_comparison_operator_token29] = ACTIONS(800),
    [aux_sym_comparison_operator_token30] = ACTIONS(800),
    [aux_sym_comparison_operator_token31] = ACTIONS(800),
    [aux_sym_comparison_operator_token32] = ACTIONS(800),
    [aux_sym_comparison_operator_token33] = ACTIONS(800),
    [aux_sym_comparison_operator_token34] = ACTIONS(802),
    [aux_sym_comparison_operator_token35] = ACTIONS(800),
    [aux_sym_comparison_operator_token36] = ACTIONS(800),
    [aux_sym_comparison_operator_token37] = ACTIONS(800),
    [aux_sym_comparison_operator_token38] = ACTIONS(800),
    [aux_sym_comparison_operator_token39] = ACTIONS(800),
    [aux_sym_comparison_operator_token40] = ACTIONS(800),
    [aux_sym_comparison_operator_token41] = ACTIONS(800),
    [aux_sym_comparison_operator_token42] = ACTIONS(800),
    [aux_sym_comparison_operator_token43] = ACTIONS(800),
    [aux_sym_comparison_operator_token44] = ACTIONS(800),
    [aux_sym_comparison_operator_token45] = ACTIONS(800),
    [aux_sym_comparison_operator_token46] = ACTIONS(800),
    [aux_sym_comparison_operator_token47] = ACTIONS(800),
    [aux_sym_comparison_operator_token48] = ACTIONS(800),
    [aux_sym_comparison_operator_token49] = ACTIONS(800),
    [aux_sym_comparison_operator_token50] = ACTIONS(800),
    [aux_sym_format_operator_token1] = ACTIONS(800),
    [anon_sym_RPAREN] = ACTIONS(800),
    [anon_sym_AMP_AMP] = ACTIONS(800),
    [anon_sym_PIPE_PIPE] = ACTIONS(800),
    [anon_sym_PIPE] = ACTIONS(802),
    [aux_sym_logical_expression_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token2] = ACTIONS(800),
    [aux_sym_logical_expression_token3] = ACTIONS(800),
    [aux_sym_bitwise_expression_token1] = ACTIONS(800),
    [aux_sym_bitwise_expression_token2] = ACTIONS(800),
    [aux_sym_bitwise_expression_token3] = ACTIONS(800),
    [anon_sym_PLUS] = ACTIONS(802),
    [anon_sym_DASH] = ACTIONS(802),
    [anon_sym_SLASH] = ACTIONS(802),
    [anon_sym_BSLASH] = ACTIONS(800),
    [anon_sym_PERCENT] = ACTIONS(802),
    [anon_sym_STAR] = ACTIONS(802),
    [anon_sym_DOT_DOT] = ACTIONS(804),
  },
  [STATE(188)] = {
    [sym_format_operator] = STATE(588),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(796),
    [anon_sym_BANG_EQ] = ACTIONS(796),
    [anon_sym_PLUS_EQ] = ACTIONS(796),
    [anon_sym_STAR_EQ] = ACTIONS(796),
    [anon_sym_SLASH_EQ] = ACTIONS(796),
    [anon_sym_PERCENT_EQ] = ACTIONS(796),
    [anon_sym_DASH_EQ] = ACTIONS(796),
    [anon_sym_GT] = ACTIONS(798),
    [anon_sym_GT_GT] = ACTIONS(796),
    [anon_sym_2_GT] = ACTIONS(798),
    [anon_sym_2_GT_GT] = ACTIONS(796),
    [anon_sym_3_GT] = ACTIONS(798),
    [anon_sym_3_GT_GT] = ACTIONS(796),
    [anon_sym_4_GT] = ACTIONS(798),
    [anon_sym_4_GT_GT] = ACTIONS(796),
    [anon_sym_5_GT] = ACTIONS(798),
    [anon_sym_5_GT_GT] = ACTIONS(796),
    [anon_sym_6_GT] = ACTIONS(798),
    [anon_sym_6_GT_GT] = ACTIONS(796),
    [anon_sym_STAR_GT] = ACTIONS(798),
    [anon_sym_STAR_GT_GT] = ACTIONS(796),
    [anon_sym_LT] = ACTIONS(798),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(796),
    [anon_sym_2_GT_AMP1] = ACTIONS(796),
    [anon_sym_3_GT_AMP1] = ACTIONS(796),
    [anon_sym_4_GT_AMP1] = ACTIONS(796),
    [anon_sym_5_GT_AMP1] = ACTIONS(796),
    [anon_sym_6_GT_AMP1] = ACTIONS(796),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(796),
    [anon_sym_1_GT_AMP2] = ACTIONS(796),
    [anon_sym_3_GT_AMP2] = ACTIONS(796),
    [anon_sym_4_GT_AMP2] = ACTIONS(796),
    [anon_sym_5_GT_AMP2] = ACTIONS(796),
    [anon_sym_6_GT_AMP2] = ACTIONS(796),
    [aux_sym_comparison_operator_token1] = ACTIONS(796),
    [aux_sym_comparison_operator_token2] = ACTIONS(796),
    [aux_sym_comparison_operator_token3] = ACTIONS(796),
    [aux_sym_comparison_operator_token4] = ACTIONS(796),
    [aux_sym_comparison_operator_token5] = ACTIONS(796),
    [aux_sym_comparison_operator_token6] = ACTIONS(796),
    [aux_sym_comparison_operator_token7] = ACTIONS(796),
    [aux_sym_comparison_operator_token8] = ACTIONS(796),
    [aux_sym_comparison_operator_token9] = ACTIONS(796),
    [aux_sym_comparison_operator_token10] = ACTIONS(796),
    [aux_sym_comparison_operator_token11] = ACTIONS(796),
    [aux_sym_comparison_operator_token12] = ACTIONS(796),
    [aux_sym_comparison_operator_token13] = ACTIONS(796),
    [aux_sym_comparison_operator_token14] = ACTIONS(796),
    [aux_sym_comparison_operator_token15] = ACTIONS(796),
    [aux_sym_comparison_operator_token16] = ACTIONS(796),
    [aux_sym_comparison_operator_token17] = ACTIONS(796),
    [aux_sym_comparison_operator_token18] = ACTIONS(796),
    [aux_sym_comparison_operator_token19] = ACTIONS(796),
    [aux_sym_comparison_operator_token20] = ACTIONS(796),
    [aux_sym_comparison_operator_token21] = ACTIONS(796),
    [aux_sym_comparison_operator_token22] = ACTIONS(796),
    [aux_sym_comparison_operator_token23] = ACTIONS(796),
    [aux_sym_comparison_operator_token24] = ACTIONS(796),
    [aux_sym_comparison_operator_token25] = ACTIONS(796),
    [aux_sym_comparison_operator_token26] = ACTIONS(796),
    [aux_sym_comparison_operator_token27] = ACTIONS(796),
    [aux_sym_comparison_operator_token28] = ACTIONS(798),
    [aux_sym_comparison_operator_token29] = ACTIONS(796),
    [aux_sym_comparison_operator_token30] = ACTIONS(796),
    [aux_sym_comparison_operator_token31] = ACTIONS(796),
    [aux_sym_comparison_operator_token32] = ACTIONS(796),
    [aux_sym_comparison_operator_token33] = ACTIONS(796),
    [aux_sym_comparison_operator_token34] = ACTIONS(798),
    [aux_sym_comparison_operator_token35] = ACTIONS(796),
    [aux_sym_comparison_operator_token36] = ACTIONS(796),
    [aux_sym_comparison_operator_token37] = ACTIONS(796),
    [aux_sym_comparison_operator_token38] = ACTIONS(796),
    [aux_sym_comparison_operator_token39] = ACTIONS(796),
    [aux_sym_comparison_operator_token40] = ACTIONS(796),
    [aux_sym_comparison_operator_token41] = ACTIONS(796),
    [aux_sym_comparison_operator_token42] = ACTIONS(796),
    [aux_sym_comparison_operator_token43] = ACTIONS(796),
    [aux_sym_comparison_operator_token44] = ACTIONS(796),
    [aux_sym_comparison_operator_token45] = ACTIONS(796),
    [aux_sym_comparison_operator_token46] = ACTIONS(796),
    [aux_sym_comparison_operator_token47] = ACTIONS(796),
    [aux_sym_comparison_operator_token48] = ACTIONS(796),
    [aux_sym_comparison_operator_token49] = ACTIONS(796),
    [aux_sym_comparison_operator_token50] = ACTIONS(796),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_RPAREN] = ACTIONS(796),
    [anon_sym_AMP_AMP] = ACTIONS(796),
    [anon_sym_PIPE_PIPE] = ACTIONS(796),
    [anon_sym_PIPE] = ACTIONS(798),
    [aux_sym_logical_expression_token1] = ACTIONS(796),
    [aux_sym_logical_expression_token2] = ACTIONS(796),
    [aux_sym_logical_expression_token3] = ACTIONS(796),
    [aux_sym_bitwise_expression_token1] = ACTIONS(796),
    [aux_sym_bitwise_expression_token2] = ACTIONS(796),
    [aux_sym_bitwise_expression_token3] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(798),
    [anon_sym_SLASH] = ACTIONS(798),
    [anon_sym_BSLASH] = ACTIONS(796),
    [anon_sym_PERCENT] = ACTIONS(798),
    [anon_sym_STAR] = ACTIONS(798),
  },
  [STATE(189)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(806),
    [anon_sym_BANG_EQ] = ACTIONS(806),
    [anon_sym_PLUS_EQ] = ACTIONS(806),
    [anon_sym_STAR_EQ] = ACTIONS(806),
    [anon_sym_SLASH_EQ] = ACTIONS(806),
    [anon_sym_PERCENT_EQ] = ACTIONS(806),
    [anon_sym_DASH_EQ] = ACTIONS(806),
    [anon_sym_GT] = ACTIONS(808),
    [anon_sym_GT_GT] = ACTIONS(806),
    [anon_sym_2_GT] = ACTIONS(808),
    [anon_sym_2_GT_GT] = ACTIONS(806),
    [anon_sym_3_GT] = ACTIONS(808),
    [anon_sym_3_GT_GT] = ACTIONS(806),
    [anon_sym_4_GT] = ACTIONS(808),
    [anon_sym_4_GT_GT] = ACTIONS(806),
    [anon_sym_5_GT] = ACTIONS(808),
    [anon_sym_5_GT_GT] = ACTIONS(806),
    [anon_sym_6_GT] = ACTIONS(808),
    [anon_sym_6_GT_GT] = ACTIONS(806),
    [anon_sym_STAR_GT] = ACTIONS(808),
    [anon_sym_STAR_GT_GT] = ACTIONS(806),
    [anon_sym_LT] = ACTIONS(808),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(806),
    [anon_sym_2_GT_AMP1] = ACTIONS(806),
    [anon_sym_3_GT_AMP1] = ACTIONS(806),
    [anon_sym_4_GT_AMP1] = ACTIONS(806),
    [anon_sym_5_GT_AMP1] = ACTIONS(806),
    [anon_sym_6_GT_AMP1] = ACTIONS(806),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(806),
    [anon_sym_1_GT_AMP2] = ACTIONS(806),
    [anon_sym_3_GT_AMP2] = ACTIONS(806),
    [anon_sym_4_GT_AMP2] = ACTIONS(806),
    [anon_sym_5_GT_AMP2] = ACTIONS(806),
    [anon_sym_6_GT_AMP2] = ACTIONS(806),
    [aux_sym_comparison_operator_token1] = ACTIONS(806),
    [aux_sym_comparison_operator_token2] = ACTIONS(806),
    [aux_sym_comparison_operator_token3] = ACTIONS(806),
    [aux_sym_comparison_operator_token4] = ACTIONS(806),
    [aux_sym_comparison_operator_token5] = ACTIONS(806),
    [aux_sym_comparison_operator_token6] = ACTIONS(806),
    [aux_sym_comparison_operator_token7] = ACTIONS(806),
    [aux_sym_comparison_operator_token8] = ACTIONS(806),
    [aux_sym_comparison_operator_token9] = ACTIONS(806),
    [aux_sym_comparison_operator_token10] = ACTIONS(806),
    [aux_sym_comparison_operator_token11] = ACTIONS(806),
    [aux_sym_comparison_operator_token12] = ACTIONS(806),
    [aux_sym_comparison_operator_token13] = ACTIONS(806),
    [aux_sym_comparison_operator_token14] = ACTIONS(806),
    [aux_sym_comparison_operator_token15] = ACTIONS(806),
    [aux_sym_comparison_operator_token16] = ACTIONS(806),
    [aux_sym_comparison_operator_token17] = ACTIONS(806),
    [aux_sym_comparison_operator_token18] = ACTIONS(806),
    [aux_sym_comparison_operator_token19] = ACTIONS(806),
    [aux_sym_comparison_operator_token20] = ACTIONS(806),
    [aux_sym_comparison_operator_token21] = ACTIONS(806),
    [aux_sym_comparison_operator_token22] = ACTIONS(806),
    [aux_sym_comparison_operator_token23] = ACTIONS(806),
    [aux_sym_comparison_operator_token24] = ACTIONS(806),
    [aux_sym_comparison_operator_token25] = ACTIONS(806),
    [aux_sym_comparison_operator_token26] = ACTIONS(806),
    [aux_sym_comparison_operator_token27] = ACTIONS(806),
    [aux_sym_comparison_operator_token28] = ACTIONS(808),
    [aux_sym_comparison_operator_token29] = ACTIONS(806),
    [aux_sym_comparison_operator_token30] = ACTIONS(806),
    [aux_sym_comparison_operator_token31] = ACTIONS(806),
    [aux_sym_comparison_operator_token32] = ACTIONS(806),
    [aux_sym_comparison_operator_token33] = ACTIONS(806),
    [aux_sym_comparison_operator_token34] = ACTIONS(808),
    [aux_sym_comparison_operator_token35] = ACTIONS(806),
    [aux_sym_comparison_operator_token36] = ACTIONS(806),
    [aux_sym_comparison_operator_token37] = ACTIONS(806),
    [aux_sym_comparison_operator_token38] = ACTIONS(806),
    [aux_sym_comparison_operator_token39] = ACTIONS(806),
    [aux_sym_comparison_operator_token40] = ACTIONS(806),
    [aux_sym_comparison_operator_token41] = ACTIONS(806),
    [aux_sym_comparison_operator_token42] = ACTIONS(806),
    [aux_sym_comparison_operator_token43] = ACTIONS(806),
    [aux_sym_comparison_operator_token44] = ACTIONS(806),
    [aux_sym_comparison_operator_token45] = ACTIONS(806),
    [aux_sym_comparison_operator_token46] = ACTIONS(806),
    [aux_sym_comparison_operator_token47] = ACTIONS(806),
    [aux_sym_comparison_operator_token48] = ACTIONS(806),
    [aux_sym_comparison_operator_token49] = ACTIONS(806),
    [aux_sym_comparison_operator_token50] = ACTIONS(806),
    [aux_sym_format_operator_token1] = ACTIONS(806),
    [anon_sym_AMP_AMP] = ACTIONS(806),
    [anon_sym_PIPE_PIPE] = ACTIONS(806),
    [anon_sym_PIPE] = ACTIONS(808),
    [aux_sym_logical_expression_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token2] = ACTIONS(806),
    [aux_sym_logical_expression_token3] = ACTIONS(806),
    [aux_sym_bitwise_expression_token1] = ACTIONS(806),
    [aux_sym_bitwise_expression_token2] = ACTIONS(806),
    [aux_sym_bitwise_expression_token3] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(808),
    [anon_sym_DASH] = ACTIONS(808),
    [anon_sym_SLASH] = ACTIONS(808),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_PERCENT] = ACTIONS(808),
    [anon_sym_STAR] = ACTIONS(808),
    [anon_sym_DOT_DOT] = ACTIONS(810),
    [sym__statement_terminator] = ACTIONS(806),
  },
  [STATE(190)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(806),
    [anon_sym_BANG_EQ] = ACTIONS(806),
    [anon_sym_PLUS_EQ] = ACTIONS(806),
    [anon_sym_STAR_EQ] = ACTIONS(806),
    [anon_sym_SLASH_EQ] = ACTIONS(806),
    [anon_sym_PERCENT_EQ] = ACTIONS(806),
    [anon_sym_DASH_EQ] = ACTIONS(806),
    [anon_sym_GT] = ACTIONS(808),
    [anon_sym_GT_GT] = ACTIONS(806),
    [anon_sym_2_GT] = ACTIONS(808),
    [anon_sym_2_GT_GT] = ACTIONS(806),
    [anon_sym_3_GT] = ACTIONS(808),
    [anon_sym_3_GT_GT] = ACTIONS(806),
    [anon_sym_4_GT] = ACTIONS(808),
    [anon_sym_4_GT_GT] = ACTIONS(806),
    [anon_sym_5_GT] = ACTIONS(808),
    [anon_sym_5_GT_GT] = ACTIONS(806),
    [anon_sym_6_GT] = ACTIONS(808),
    [anon_sym_6_GT_GT] = ACTIONS(806),
    [anon_sym_STAR_GT] = ACTIONS(808),
    [anon_sym_STAR_GT_GT] = ACTIONS(806),
    [anon_sym_LT] = ACTIONS(808),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(806),
    [anon_sym_2_GT_AMP1] = ACTIONS(806),
    [anon_sym_3_GT_AMP1] = ACTIONS(806),
    [anon_sym_4_GT_AMP1] = ACTIONS(806),
    [anon_sym_5_GT_AMP1] = ACTIONS(806),
    [anon_sym_6_GT_AMP1] = ACTIONS(806),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(806),
    [anon_sym_1_GT_AMP2] = ACTIONS(806),
    [anon_sym_3_GT_AMP2] = ACTIONS(806),
    [anon_sym_4_GT_AMP2] = ACTIONS(806),
    [anon_sym_5_GT_AMP2] = ACTIONS(806),
    [anon_sym_6_GT_AMP2] = ACTIONS(806),
    [aux_sym_comparison_operator_token1] = ACTIONS(806),
    [aux_sym_comparison_operator_token2] = ACTIONS(806),
    [aux_sym_comparison_operator_token3] = ACTIONS(806),
    [aux_sym_comparison_operator_token4] = ACTIONS(806),
    [aux_sym_comparison_operator_token5] = ACTIONS(806),
    [aux_sym_comparison_operator_token6] = ACTIONS(806),
    [aux_sym_comparison_operator_token7] = ACTIONS(806),
    [aux_sym_comparison_operator_token8] = ACTIONS(806),
    [aux_sym_comparison_operator_token9] = ACTIONS(806),
    [aux_sym_comparison_operator_token10] = ACTIONS(806),
    [aux_sym_comparison_operator_token11] = ACTIONS(806),
    [aux_sym_comparison_operator_token12] = ACTIONS(806),
    [aux_sym_comparison_operator_token13] = ACTIONS(806),
    [aux_sym_comparison_operator_token14] = ACTIONS(806),
    [aux_sym_comparison_operator_token15] = ACTIONS(806),
    [aux_sym_comparison_operator_token16] = ACTIONS(806),
    [aux_sym_comparison_operator_token17] = ACTIONS(806),
    [aux_sym_comparison_operator_token18] = ACTIONS(806),
    [aux_sym_comparison_operator_token19] = ACTIONS(806),
    [aux_sym_comparison_operator_token20] = ACTIONS(806),
    [aux_sym_comparison_operator_token21] = ACTIONS(806),
    [aux_sym_comparison_operator_token22] = ACTIONS(806),
    [aux_sym_comparison_operator_token23] = ACTIONS(806),
    [aux_sym_comparison_operator_token24] = ACTIONS(806),
    [aux_sym_comparison_operator_token25] = ACTIONS(806),
    [aux_sym_comparison_operator_token26] = ACTIONS(806),
    [aux_sym_comparison_operator_token27] = ACTIONS(806),
    [aux_sym_comparison_operator_token28] = ACTIONS(808),
    [aux_sym_comparison_operator_token29] = ACTIONS(806),
    [aux_sym_comparison_operator_token30] = ACTIONS(806),
    [aux_sym_comparison_operator_token31] = ACTIONS(806),
    [aux_sym_comparison_operator_token32] = ACTIONS(806),
    [aux_sym_comparison_operator_token33] = ACTIONS(806),
    [aux_sym_comparison_operator_token34] = ACTIONS(808),
    [aux_sym_comparison_operator_token35] = ACTIONS(806),
    [aux_sym_comparison_operator_token36] = ACTIONS(806),
    [aux_sym_comparison_operator_token37] = ACTIONS(806),
    [aux_sym_comparison_operator_token38] = ACTIONS(806),
    [aux_sym_comparison_operator_token39] = ACTIONS(806),
    [aux_sym_comparison_operator_token40] = ACTIONS(806),
    [aux_sym_comparison_operator_token41] = ACTIONS(806),
    [aux_sym_comparison_operator_token42] = ACTIONS(806),
    [aux_sym_comparison_operator_token43] = ACTIONS(806),
    [aux_sym_comparison_operator_token44] = ACTIONS(806),
    [aux_sym_comparison_operator_token45] = ACTIONS(806),
    [aux_sym_comparison_operator_token46] = ACTIONS(806),
    [aux_sym_comparison_operator_token47] = ACTIONS(806),
    [aux_sym_comparison_operator_token48] = ACTIONS(806),
    [aux_sym_comparison_operator_token49] = ACTIONS(806),
    [aux_sym_comparison_operator_token50] = ACTIONS(806),
    [aux_sym_format_operator_token1] = ACTIONS(806),
    [anon_sym_RPAREN] = ACTIONS(806),
    [anon_sym_AMP_AMP] = ACTIONS(806),
    [anon_sym_PIPE_PIPE] = ACTIONS(806),
    [anon_sym_PIPE] = ACTIONS(808),
    [aux_sym_logical_expression_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token2] = ACTIONS(806),
    [aux_sym_logical_expression_token3] = ACTIONS(806),
    [aux_sym_bitwise_expression_token1] = ACTIONS(806),
    [aux_sym_bitwise_expression_token2] = ACTIONS(806),
    [aux_sym_bitwise_expression_token3] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(808),
    [anon_sym_DASH] = ACTIONS(808),
    [anon_sym_SLASH] = ACTIONS(808),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_PERCENT] = ACTIONS(808),
    [anon_sym_STAR] = ACTIONS(808),
    [anon_sym_DOT_DOT] = ACTIONS(804),
  },
  [STATE(191)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(768),
    [anon_sym_BANG_EQ] = ACTIONS(768),
    [anon_sym_PLUS_EQ] = ACTIONS(768),
    [anon_sym_STAR_EQ] = ACTIONS(768),
    [anon_sym_SLASH_EQ] = ACTIONS(768),
    [anon_sym_PERCENT_EQ] = ACTIONS(768),
    [anon_sym_DASH_EQ] = ACTIONS(768),
    [anon_sym_GT] = ACTIONS(770),
    [anon_sym_GT_GT] = ACTIONS(768),
    [anon_sym_2_GT] = ACTIONS(770),
    [anon_sym_2_GT_GT] = ACTIONS(768),
    [anon_sym_3_GT] = ACTIONS(770),
    [anon_sym_3_GT_GT] = ACTIONS(768),
    [anon_sym_4_GT] = ACTIONS(770),
    [anon_sym_4_GT_GT] = ACTIONS(768),
    [anon_sym_5_GT] = ACTIONS(770),
    [anon_sym_5_GT_GT] = ACTIONS(768),
    [anon_sym_6_GT] = ACTIONS(770),
    [anon_sym_6_GT_GT] = ACTIONS(768),
    [anon_sym_STAR_GT] = ACTIONS(770),
    [anon_sym_STAR_GT_GT] = ACTIONS(768),
    [anon_sym_LT] = ACTIONS(770),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(768),
    [anon_sym_2_GT_AMP1] = ACTIONS(768),
    [anon_sym_3_GT_AMP1] = ACTIONS(768),
    [anon_sym_4_GT_AMP1] = ACTIONS(768),
    [anon_sym_5_GT_AMP1] = ACTIONS(768),
    [anon_sym_6_GT_AMP1] = ACTIONS(768),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(768),
    [anon_sym_1_GT_AMP2] = ACTIONS(768),
    [anon_sym_3_GT_AMP2] = ACTIONS(768),
    [anon_sym_4_GT_AMP2] = ACTIONS(768),
    [anon_sym_5_GT_AMP2] = ACTIONS(768),
    [anon_sym_6_GT_AMP2] = ACTIONS(768),
    [aux_sym_comparison_operator_token1] = ACTIONS(768),
    [aux_sym_comparison_operator_token2] = ACTIONS(768),
    [aux_sym_comparison_operator_token3] = ACTIONS(768),
    [aux_sym_comparison_operator_token4] = ACTIONS(768),
    [aux_sym_comparison_operator_token5] = ACTIONS(768),
    [aux_sym_comparison_operator_token6] = ACTIONS(768),
    [aux_sym_comparison_operator_token7] = ACTIONS(768),
    [aux_sym_comparison_operator_token8] = ACTIONS(768),
    [aux_sym_comparison_operator_token9] = ACTIONS(768),
    [aux_sym_comparison_operator_token10] = ACTIONS(768),
    [aux_sym_comparison_operator_token11] = ACTIONS(768),
    [aux_sym_comparison_operator_token12] = ACTIONS(768),
    [aux_sym_comparison_operator_token13] = ACTIONS(768),
    [aux_sym_comparison_operator_token14] = ACTIONS(768),
    [aux_sym_comparison_operator_token15] = ACTIONS(768),
    [aux_sym_comparison_operator_token16] = ACTIONS(768),
    [aux_sym_comparison_operator_token17] = ACTIONS(768),
    [aux_sym_comparison_operator_token18] = ACTIONS(768),
    [aux_sym_comparison_operator_token19] = ACTIONS(768),
    [aux_sym_comparison_operator_token20] = ACTIONS(768),
    [aux_sym_comparison_operator_token21] = ACTIONS(768),
    [aux_sym_comparison_operator_token22] = ACTIONS(768),
    [aux_sym_comparison_operator_token23] = ACTIONS(768),
    [aux_sym_comparison_operator_token24] = ACTIONS(768),
    [aux_sym_comparison_operator_token25] = ACTIONS(768),
    [aux_sym_comparison_operator_token26] = ACTIONS(768),
    [aux_sym_comparison_operator_token27] = ACTIONS(768),
    [aux_sym_comparison_operator_token28] = ACTIONS(770),
    [aux_sym_comparison_operator_token29] = ACTIONS(768),
    [aux_sym_comparison_operator_token30] = ACTIONS(768),
    [aux_sym_comparison_operator_token31] = ACTIONS(768),
    [aux_sym_comparison_operator_token32] = ACTIONS(768),
    [aux_sym_comparison_operator_token33] = ACTIONS(768),
    [aux_sym_comparison_operator_token34] = ACTIONS(770),
    [aux_sym_comparison_operator_token35] = ACTIONS(768),
    [aux_sym_comparison_operator_token36] = ACTIONS(768),
    [aux_sym_comparison_operator_token37] = ACTIONS(768),
    [aux_sym_comparison_operator_token38] = ACTIONS(768),
    [aux_sym_comparison_operator_token39] = ACTIONS(768),
    [aux_sym_comparison_operator_token40] = ACTIONS(768),
    [aux_sym_comparison_operator_token41] = ACTIONS(768),
    [aux_sym_comparison_operator_token42] = ACTIONS(768),
    [aux_sym_comparison_operator_token43] = ACTIONS(768),
    [aux_sym_comparison_operator_token44] = ACTIONS(768),
    [aux_sym_comparison_operator_token45] = ACTIONS(768),
    [aux_sym_comparison_operator_token46] = ACTIONS(768),
    [aux_sym_comparison_operator_token47] = ACTIONS(768),
    [aux_sym_comparison_operator_token48] = ACTIONS(768),
    [aux_sym_comparison_operator_token49] = ACTIONS(768),
    [aux_sym_comparison_operator_token50] = ACTIONS(768),
    [aux_sym_format_operator_token1] = ACTIONS(768),
    [anon_sym_AMP_AMP] = ACTIONS(768),
    [anon_sym_PIPE_PIPE] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(770),
    [aux_sym_logical_expression_token1] = ACTIONS(768),
    [aux_sym_logical_expression_token2] = ACTIONS(768),
    [aux_sym_logical_expression_token3] = ACTIONS(768),
    [aux_sym_bitwise_expression_token1] = ACTIONS(768),
    [aux_sym_bitwise_expression_token2] = ACTIONS(768),
    [aux_sym_bitwise_expression_token3] = ACTIONS(768),
    [anon_sym_PLUS] = ACTIONS(770),
    [anon_sym_DASH] = ACTIONS(770),
    [anon_sym_SLASH] = ACTIONS(770),
    [anon_sym_BSLASH] = ACTIONS(768),
    [anon_sym_PERCENT] = ACTIONS(770),
    [anon_sym_STAR] = ACTIONS(770),
    [anon_sym_DOT_DOT] = ACTIONS(768),
    [sym__statement_terminator] = ACTIONS(768),
  },
  [STATE(192)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(800),
    [anon_sym_BANG_EQ] = ACTIONS(800),
    [anon_sym_PLUS_EQ] = ACTIONS(800),
    [anon_sym_STAR_EQ] = ACTIONS(800),
    [anon_sym_SLASH_EQ] = ACTIONS(800),
    [anon_sym_PERCENT_EQ] = ACTIONS(800),
    [anon_sym_DASH_EQ] = ACTIONS(800),
    [anon_sym_GT] = ACTIONS(802),
    [anon_sym_GT_GT] = ACTIONS(800),
    [anon_sym_2_GT] = ACTIONS(802),
    [anon_sym_2_GT_GT] = ACTIONS(800),
    [anon_sym_3_GT] = ACTIONS(802),
    [anon_sym_3_GT_GT] = ACTIONS(800),
    [anon_sym_4_GT] = ACTIONS(802),
    [anon_sym_4_GT_GT] = ACTIONS(800),
    [anon_sym_5_GT] = ACTIONS(802),
    [anon_sym_5_GT_GT] = ACTIONS(800),
    [anon_sym_6_GT] = ACTIONS(802),
    [anon_sym_6_GT_GT] = ACTIONS(800),
    [anon_sym_STAR_GT] = ACTIONS(802),
    [anon_sym_STAR_GT_GT] = ACTIONS(800),
    [anon_sym_LT] = ACTIONS(802),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(800),
    [anon_sym_2_GT_AMP1] = ACTIONS(800),
    [anon_sym_3_GT_AMP1] = ACTIONS(800),
    [anon_sym_4_GT_AMP1] = ACTIONS(800),
    [anon_sym_5_GT_AMP1] = ACTIONS(800),
    [anon_sym_6_GT_AMP1] = ACTIONS(800),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(800),
    [anon_sym_1_GT_AMP2] = ACTIONS(800),
    [anon_sym_3_GT_AMP2] = ACTIONS(800),
    [anon_sym_4_GT_AMP2] = ACTIONS(800),
    [anon_sym_5_GT_AMP2] = ACTIONS(800),
    [anon_sym_6_GT_AMP2] = ACTIONS(800),
    [aux_sym_comparison_operator_token1] = ACTIONS(800),
    [aux_sym_comparison_operator_token2] = ACTIONS(800),
    [aux_sym_comparison_operator_token3] = ACTIONS(800),
    [aux_sym_comparison_operator_token4] = ACTIONS(800),
    [aux_sym_comparison_operator_token5] = ACTIONS(800),
    [aux_sym_comparison_operator_token6] = ACTIONS(800),
    [aux_sym_comparison_operator_token7] = ACTIONS(800),
    [aux_sym_comparison_operator_token8] = ACTIONS(800),
    [aux_sym_comparison_operator_token9] = ACTIONS(800),
    [aux_sym_comparison_operator_token10] = ACTIONS(800),
    [aux_sym_comparison_operator_token11] = ACTIONS(800),
    [aux_sym_comparison_operator_token12] = ACTIONS(800),
    [aux_sym_comparison_operator_token13] = ACTIONS(800),
    [aux_sym_comparison_operator_token14] = ACTIONS(800),
    [aux_sym_comparison_operator_token15] = ACTIONS(800),
    [aux_sym_comparison_operator_token16] = ACTIONS(800),
    [aux_sym_comparison_operator_token17] = ACTIONS(800),
    [aux_sym_comparison_operator_token18] = ACTIONS(800),
    [aux_sym_comparison_operator_token19] = ACTIONS(800),
    [aux_sym_comparison_operator_token20] = ACTIONS(800),
    [aux_sym_comparison_operator_token21] = ACTIONS(800),
    [aux_sym_comparison_operator_token22] = ACTIONS(800),
    [aux_sym_comparison_operator_token23] = ACTIONS(800),
    [aux_sym_comparison_operator_token24] = ACTIONS(800),
    [aux_sym_comparison_operator_token25] = ACTIONS(800),
    [aux_sym_comparison_operator_token26] = ACTIONS(800),
    [aux_sym_comparison_operator_token27] = ACTIONS(800),
    [aux_sym_comparison_operator_token28] = ACTIONS(802),
    [aux_sym_comparison_operator_token29] = ACTIONS(800),
    [aux_sym_comparison_operator_token30] = ACTIONS(800),
    [aux_sym_comparison_operator_token31] = ACTIONS(800),
    [aux_sym_comparison_operator_token32] = ACTIONS(800),
    [aux_sym_comparison_operator_token33] = ACTIONS(800),
    [aux_sym_comparison_operator_token34] = ACTIONS(802),
    [aux_sym_comparison_operator_token35] = ACTIONS(800),
    [aux_sym_comparison_operator_token36] = ACTIONS(800),
    [aux_sym_comparison_operator_token37] = ACTIONS(800),
    [aux_sym_comparison_operator_token38] = ACTIONS(800),
    [aux_sym_comparison_operator_token39] = ACTIONS(800),
    [aux_sym_comparison_operator_token40] = ACTIONS(800),
    [aux_sym_comparison_operator_token41] = ACTIONS(800),
    [aux_sym_comparison_operator_token42] = ACTIONS(800),
    [aux_sym_comparison_operator_token43] = ACTIONS(800),
    [aux_sym_comparison_operator_token44] = ACTIONS(800),
    [aux_sym_comparison_operator_token45] = ACTIONS(800),
    [aux_sym_comparison_operator_token46] = ACTIONS(800),
    [aux_sym_comparison_operator_token47] = ACTIONS(800),
    [aux_sym_comparison_operator_token48] = ACTIONS(800),
    [aux_sym_comparison_operator_token49] = ACTIONS(800),
    [aux_sym_comparison_operator_token50] = ACTIONS(800),
    [aux_sym_format_operator_token1] = ACTIONS(800),
    [anon_sym_AMP_AMP] = ACTIONS(800),
    [anon_sym_PIPE_PIPE] = ACTIONS(800),
    [anon_sym_PIPE] = ACTIONS(802),
    [aux_sym_logical_expression_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token2] = ACTIONS(800),
    [aux_sym_logical_expression_token3] = ACTIONS(800),
    [aux_sym_bitwise_expression_token1] = ACTIONS(800),
    [aux_sym_bitwise_expression_token2] = ACTIONS(800),
    [aux_sym_bitwise_expression_token3] = ACTIONS(800),
    [anon_sym_PLUS] = ACTIONS(802),
    [anon_sym_DASH] = ACTIONS(802),
    [anon_sym_SLASH] = ACTIONS(802),
    [anon_sym_BSLASH] = ACTIONS(800),
    [anon_sym_PERCENT] = ACTIONS(802),
    [anon_sym_STAR] = ACTIONS(802),
    [anon_sym_DOT_DOT] = ACTIONS(810),
    [sym__statement_terminator] = ACTIONS(800),
  },
  [STATE(193)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(812),
    [anon_sym_BANG_EQ] = ACTIONS(812),
    [anon_sym_PLUS_EQ] = ACTIONS(812),
    [anon_sym_STAR_EQ] = ACTIONS(812),
    [anon_sym_SLASH_EQ] = ACTIONS(812),
    [anon_sym_PERCENT_EQ] = ACTIONS(812),
    [anon_sym_DASH_EQ] = ACTIONS(812),
    [anon_sym_GT] = ACTIONS(814),
    [anon_sym_GT_GT] = ACTIONS(812),
    [anon_sym_2_GT] = ACTIONS(814),
    [anon_sym_2_GT_GT] = ACTIONS(812),
    [anon_sym_3_GT] = ACTIONS(814),
    [anon_sym_3_GT_GT] = ACTIONS(812),
    [anon_sym_4_GT] = ACTIONS(814),
    [anon_sym_4_GT_GT] = ACTIONS(812),
    [anon_sym_5_GT] = ACTIONS(814),
    [anon_sym_5_GT_GT] = ACTIONS(812),
    [anon_sym_6_GT] = ACTIONS(814),
    [anon_sym_6_GT_GT] = ACTIONS(812),
    [anon_sym_STAR_GT] = ACTIONS(814),
    [anon_sym_STAR_GT_GT] = ACTIONS(812),
    [anon_sym_LT] = ACTIONS(814),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(812),
    [anon_sym_2_GT_AMP1] = ACTIONS(812),
    [anon_sym_3_GT_AMP1] = ACTIONS(812),
    [anon_sym_4_GT_AMP1] = ACTIONS(812),
    [anon_sym_5_GT_AMP1] = ACTIONS(812),
    [anon_sym_6_GT_AMP1] = ACTIONS(812),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(812),
    [anon_sym_1_GT_AMP2] = ACTIONS(812),
    [anon_sym_3_GT_AMP2] = ACTIONS(812),
    [anon_sym_4_GT_AMP2] = ACTIONS(812),
    [anon_sym_5_GT_AMP2] = ACTIONS(812),
    [anon_sym_6_GT_AMP2] = ACTIONS(812),
    [aux_sym_comparison_operator_token1] = ACTIONS(812),
    [aux_sym_comparison_operator_token2] = ACTIONS(812),
    [aux_sym_comparison_operator_token3] = ACTIONS(812),
    [aux_sym_comparison_operator_token4] = ACTIONS(812),
    [aux_sym_comparison_operator_token5] = ACTIONS(812),
    [aux_sym_comparison_operator_token6] = ACTIONS(812),
    [aux_sym_comparison_operator_token7] = ACTIONS(812),
    [aux_sym_comparison_operator_token8] = ACTIONS(812),
    [aux_sym_comparison_operator_token9] = ACTIONS(812),
    [aux_sym_comparison_operator_token10] = ACTIONS(812),
    [aux_sym_comparison_operator_token11] = ACTIONS(812),
    [aux_sym_comparison_operator_token12] = ACTIONS(812),
    [aux_sym_comparison_operator_token13] = ACTIONS(812),
    [aux_sym_comparison_operator_token14] = ACTIONS(812),
    [aux_sym_comparison_operator_token15] = ACTIONS(812),
    [aux_sym_comparison_operator_token16] = ACTIONS(812),
    [aux_sym_comparison_operator_token17] = ACTIONS(812),
    [aux_sym_comparison_operator_token18] = ACTIONS(812),
    [aux_sym_comparison_operator_token19] = ACTIONS(812),
    [aux_sym_comparison_operator_token20] = ACTIONS(812),
    [aux_sym_comparison_operator_token21] = ACTIONS(812),
    [aux_sym_comparison_operator_token22] = ACTIONS(812),
    [aux_sym_comparison_operator_token23] = ACTIONS(812),
    [aux_sym_comparison_operator_token24] = ACTIONS(812),
    [aux_sym_comparison_operator_token25] = ACTIONS(812),
    [aux_sym_comparison_operator_token26] = ACTIONS(812),
    [aux_sym_comparison_operator_token27] = ACTIONS(812),
    [aux_sym_comparison_operator_token28] = ACTIONS(814),
    [aux_sym_comparison_operator_token29] = ACTIONS(812),
    [aux_sym_comparison_operator_token30] = ACTIONS(812),
    [aux_sym_comparison_operator_token31] = ACTIONS(812),
    [aux_sym_comparison_operator_token32] = ACTIONS(812),
    [aux_sym_comparison_operator_token33] = ACTIONS(812),
    [aux_sym_comparison_operator_token34] = ACTIONS(814),
    [aux_sym_comparison_operator_token35] = ACTIONS(812),
    [aux_sym_comparison_operator_token36] = ACTIONS(812),
    [aux_sym_comparison_operator_token37] = ACTIONS(812),
    [aux_sym_comparison_operator_token38] = ACTIONS(812),
    [aux_sym_comparison_operator_token39] = ACTIONS(812),
    [aux_sym_comparison_operator_token40] = ACTIONS(812),
    [aux_sym_comparison_operator_token41] = ACTIONS(812),
    [aux_sym_comparison_operator_token42] = ACTIONS(812),
    [aux_sym_comparison_operator_token43] = ACTIONS(812),
    [aux_sym_comparison_operator_token44] = ACTIONS(812),
    [aux_sym_comparison_operator_token45] = ACTIONS(812),
    [aux_sym_comparison_operator_token46] = ACTIONS(812),
    [aux_sym_comparison_operator_token47] = ACTIONS(812),
    [aux_sym_comparison_operator_token48] = ACTIONS(812),
    [aux_sym_comparison_operator_token49] = ACTIONS(812),
    [aux_sym_comparison_operator_token50] = ACTIONS(812),
    [anon_sym_RPAREN] = ACTIONS(812),
    [anon_sym_AMP_AMP] = ACTIONS(812),
    [anon_sym_PIPE_PIPE] = ACTIONS(812),
    [anon_sym_PIPE] = ACTIONS(814),
    [aux_sym_logical_expression_token1] = ACTIONS(812),
    [aux_sym_logical_expression_token2] = ACTIONS(812),
    [aux_sym_logical_expression_token3] = ACTIONS(812),
    [aux_sym_bitwise_expression_token1] = ACTIONS(812),
    [aux_sym_bitwise_expression_token2] = ACTIONS(812),
    [aux_sym_bitwise_expression_token3] = ACTIONS(812),
    [anon_sym_PLUS] = ACTIONS(814),
    [anon_sym_DASH] = ACTIONS(814),
    [anon_sym_SLASH] = ACTIONS(816),
    [anon_sym_BSLASH] = ACTIONS(818),
    [anon_sym_PERCENT] = ACTIONS(816),
    [anon_sym_STAR] = ACTIONS(816),
  },
  [STATE(194)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(820),
    [anon_sym_BANG_EQ] = ACTIONS(820),
    [anon_sym_PLUS_EQ] = ACTIONS(820),
    [anon_sym_STAR_EQ] = ACTIONS(820),
    [anon_sym_SLASH_EQ] = ACTIONS(820),
    [anon_sym_PERCENT_EQ] = ACTIONS(820),
    [anon_sym_DASH_EQ] = ACTIONS(820),
    [anon_sym_GT] = ACTIONS(822),
    [anon_sym_GT_GT] = ACTIONS(820),
    [anon_sym_2_GT] = ACTIONS(822),
    [anon_sym_2_GT_GT] = ACTIONS(820),
    [anon_sym_3_GT] = ACTIONS(822),
    [anon_sym_3_GT_GT] = ACTIONS(820),
    [anon_sym_4_GT] = ACTIONS(822),
    [anon_sym_4_GT_GT] = ACTIONS(820),
    [anon_sym_5_GT] = ACTIONS(822),
    [anon_sym_5_GT_GT] = ACTIONS(820),
    [anon_sym_6_GT] = ACTIONS(822),
    [anon_sym_6_GT_GT] = ACTIONS(820),
    [anon_sym_STAR_GT] = ACTIONS(822),
    [anon_sym_STAR_GT_GT] = ACTIONS(820),
    [anon_sym_LT] = ACTIONS(822),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(820),
    [anon_sym_2_GT_AMP1] = ACTIONS(820),
    [anon_sym_3_GT_AMP1] = ACTIONS(820),
    [anon_sym_4_GT_AMP1] = ACTIONS(820),
    [anon_sym_5_GT_AMP1] = ACTIONS(820),
    [anon_sym_6_GT_AMP1] = ACTIONS(820),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(820),
    [anon_sym_1_GT_AMP2] = ACTIONS(820),
    [anon_sym_3_GT_AMP2] = ACTIONS(820),
    [anon_sym_4_GT_AMP2] = ACTIONS(820),
    [anon_sym_5_GT_AMP2] = ACTIONS(820),
    [anon_sym_6_GT_AMP2] = ACTIONS(820),
    [aux_sym_comparison_operator_token1] = ACTIONS(820),
    [aux_sym_comparison_operator_token2] = ACTIONS(820),
    [aux_sym_comparison_operator_token3] = ACTIONS(820),
    [aux_sym_comparison_operator_token4] = ACTIONS(820),
    [aux_sym_comparison_operator_token5] = ACTIONS(820),
    [aux_sym_comparison_operator_token6] = ACTIONS(820),
    [aux_sym_comparison_operator_token7] = ACTIONS(820),
    [aux_sym_comparison_operator_token8] = ACTIONS(820),
    [aux_sym_comparison_operator_token9] = ACTIONS(820),
    [aux_sym_comparison_operator_token10] = ACTIONS(820),
    [aux_sym_comparison_operator_token11] = ACTIONS(820),
    [aux_sym_comparison_operator_token12] = ACTIONS(820),
    [aux_sym_comparison_operator_token13] = ACTIONS(820),
    [aux_sym_comparison_operator_token14] = ACTIONS(820),
    [aux_sym_comparison_operator_token15] = ACTIONS(820),
    [aux_sym_comparison_operator_token16] = ACTIONS(820),
    [aux_sym_comparison_operator_token17] = ACTIONS(820),
    [aux_sym_comparison_operator_token18] = ACTIONS(820),
    [aux_sym_comparison_operator_token19] = ACTIONS(820),
    [aux_sym_comparison_operator_token20] = ACTIONS(820),
    [aux_sym_comparison_operator_token21] = ACTIONS(820),
    [aux_sym_comparison_operator_token22] = ACTIONS(820),
    [aux_sym_comparison_operator_token23] = ACTIONS(820),
    [aux_sym_comparison_operator_token24] = ACTIONS(820),
    [aux_sym_comparison_operator_token25] = ACTIONS(820),
    [aux_sym_comparison_operator_token26] = ACTIONS(820),
    [aux_sym_comparison_operator_token27] = ACTIONS(820),
    [aux_sym_comparison_operator_token28] = ACTIONS(822),
    [aux_sym_comparison_operator_token29] = ACTIONS(820),
    [aux_sym_comparison_operator_token30] = ACTIONS(820),
    [aux_sym_comparison_operator_token31] = ACTIONS(820),
    [aux_sym_comparison_operator_token32] = ACTIONS(820),
    [aux_sym_comparison_operator_token33] = ACTIONS(820),
    [aux_sym_comparison_operator_token34] = ACTIONS(822),
    [aux_sym_comparison_operator_token35] = ACTIONS(820),
    [aux_sym_comparison_operator_token36] = ACTIONS(820),
    [aux_sym_comparison_operator_token37] = ACTIONS(820),
    [aux_sym_comparison_operator_token38] = ACTIONS(820),
    [aux_sym_comparison_operator_token39] = ACTIONS(820),
    [aux_sym_comparison_operator_token40] = ACTIONS(820),
    [aux_sym_comparison_operator_token41] = ACTIONS(820),
    [aux_sym_comparison_operator_token42] = ACTIONS(820),
    [aux_sym_comparison_operator_token43] = ACTIONS(820),
    [aux_sym_comparison_operator_token44] = ACTIONS(820),
    [aux_sym_comparison_operator_token45] = ACTIONS(820),
    [aux_sym_comparison_operator_token46] = ACTIONS(820),
    [aux_sym_comparison_operator_token47] = ACTIONS(820),
    [aux_sym_comparison_operator_token48] = ACTIONS(820),
    [aux_sym_comparison_operator_token49] = ACTIONS(820),
    [aux_sym_comparison_operator_token50] = ACTIONS(820),
    [anon_sym_RPAREN] = ACTIONS(820),
    [anon_sym_AMP_AMP] = ACTIONS(820),
    [anon_sym_PIPE_PIPE] = ACTIONS(820),
    [anon_sym_PIPE] = ACTIONS(822),
    [aux_sym_logical_expression_token1] = ACTIONS(820),
    [aux_sym_logical_expression_token2] = ACTIONS(820),
    [aux_sym_logical_expression_token3] = ACTIONS(820),
    [aux_sym_bitwise_expression_token1] = ACTIONS(820),
    [aux_sym_bitwise_expression_token2] = ACTIONS(820),
    [aux_sym_bitwise_expression_token3] = ACTIONS(820),
    [anon_sym_PLUS] = ACTIONS(822),
    [anon_sym_DASH] = ACTIONS(822),
    [anon_sym_SLASH] = ACTIONS(816),
    [anon_sym_BSLASH] = ACTIONS(818),
    [anon_sym_PERCENT] = ACTIONS(816),
    [anon_sym_STAR] = ACTIONS(816),
  },
  [STATE(195)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(812),
    [anon_sym_BANG_EQ] = ACTIONS(812),
    [anon_sym_PLUS_EQ] = ACTIONS(812),
    [anon_sym_STAR_EQ] = ACTIONS(812),
    [anon_sym_SLASH_EQ] = ACTIONS(812),
    [anon_sym_PERCENT_EQ] = ACTIONS(812),
    [anon_sym_DASH_EQ] = ACTIONS(812),
    [anon_sym_GT] = ACTIONS(814),
    [anon_sym_GT_GT] = ACTIONS(812),
    [anon_sym_2_GT] = ACTIONS(814),
    [anon_sym_2_GT_GT] = ACTIONS(812),
    [anon_sym_3_GT] = ACTIONS(814),
    [anon_sym_3_GT_GT] = ACTIONS(812),
    [anon_sym_4_GT] = ACTIONS(814),
    [anon_sym_4_GT_GT] = ACTIONS(812),
    [anon_sym_5_GT] = ACTIONS(814),
    [anon_sym_5_GT_GT] = ACTIONS(812),
    [anon_sym_6_GT] = ACTIONS(814),
    [anon_sym_6_GT_GT] = ACTIONS(812),
    [anon_sym_STAR_GT] = ACTIONS(814),
    [anon_sym_STAR_GT_GT] = ACTIONS(812),
    [anon_sym_LT] = ACTIONS(814),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(812),
    [anon_sym_2_GT_AMP1] = ACTIONS(812),
    [anon_sym_3_GT_AMP1] = ACTIONS(812),
    [anon_sym_4_GT_AMP1] = ACTIONS(812),
    [anon_sym_5_GT_AMP1] = ACTIONS(812),
    [anon_sym_6_GT_AMP1] = ACTIONS(812),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(812),
    [anon_sym_1_GT_AMP2] = ACTIONS(812),
    [anon_sym_3_GT_AMP2] = ACTIONS(812),
    [anon_sym_4_GT_AMP2] = ACTIONS(812),
    [anon_sym_5_GT_AMP2] = ACTIONS(812),
    [anon_sym_6_GT_AMP2] = ACTIONS(812),
    [aux_sym_comparison_operator_token1] = ACTIONS(812),
    [aux_sym_comparison_operator_token2] = ACTIONS(812),
    [aux_sym_comparison_operator_token3] = ACTIONS(812),
    [aux_sym_comparison_operator_token4] = ACTIONS(812),
    [aux_sym_comparison_operator_token5] = ACTIONS(812),
    [aux_sym_comparison_operator_token6] = ACTIONS(812),
    [aux_sym_comparison_operator_token7] = ACTIONS(812),
    [aux_sym_comparison_operator_token8] = ACTIONS(812),
    [aux_sym_comparison_operator_token9] = ACTIONS(812),
    [aux_sym_comparison_operator_token10] = ACTIONS(812),
    [aux_sym_comparison_operator_token11] = ACTIONS(812),
    [aux_sym_comparison_operator_token12] = ACTIONS(812),
    [aux_sym_comparison_operator_token13] = ACTIONS(812),
    [aux_sym_comparison_operator_token14] = ACTIONS(812),
    [aux_sym_comparison_operator_token15] = ACTIONS(812),
    [aux_sym_comparison_operator_token16] = ACTIONS(812),
    [aux_sym_comparison_operator_token17] = ACTIONS(812),
    [aux_sym_comparison_operator_token18] = ACTIONS(812),
    [aux_sym_comparison_operator_token19] = ACTIONS(812),
    [aux_sym_comparison_operator_token20] = ACTIONS(812),
    [aux_sym_comparison_operator_token21] = ACTIONS(812),
    [aux_sym_comparison_operator_token22] = ACTIONS(812),
    [aux_sym_comparison_operator_token23] = ACTIONS(812),
    [aux_sym_comparison_operator_token24] = ACTIONS(812),
    [aux_sym_comparison_operator_token25] = ACTIONS(812),
    [aux_sym_comparison_operator_token26] = ACTIONS(812),
    [aux_sym_comparison_operator_token27] = ACTIONS(812),
    [aux_sym_comparison_operator_token28] = ACTIONS(814),
    [aux_sym_comparison_operator_token29] = ACTIONS(812),
    [aux_sym_comparison_operator_token30] = ACTIONS(812),
    [aux_sym_comparison_operator_token31] = ACTIONS(812),
    [aux_sym_comparison_operator_token32] = ACTIONS(812),
    [aux_sym_comparison_operator_token33] = ACTIONS(812),
    [aux_sym_comparison_operator_token34] = ACTIONS(814),
    [aux_sym_comparison_operator_token35] = ACTIONS(812),
    [aux_sym_comparison_operator_token36] = ACTIONS(812),
    [aux_sym_comparison_operator_token37] = ACTIONS(812),
    [aux_sym_comparison_operator_token38] = ACTIONS(812),
    [aux_sym_comparison_operator_token39] = ACTIONS(812),
    [aux_sym_comparison_operator_token40] = ACTIONS(812),
    [aux_sym_comparison_operator_token41] = ACTIONS(812),
    [aux_sym_comparison_operator_token42] = ACTIONS(812),
    [aux_sym_comparison_operator_token43] = ACTIONS(812),
    [aux_sym_comparison_operator_token44] = ACTIONS(812),
    [aux_sym_comparison_operator_token45] = ACTIONS(812),
    [aux_sym_comparison_operator_token46] = ACTIONS(812),
    [aux_sym_comparison_operator_token47] = ACTIONS(812),
    [aux_sym_comparison_operator_token48] = ACTIONS(812),
    [aux_sym_comparison_operator_token49] = ACTIONS(812),
    [aux_sym_comparison_operator_token50] = ACTIONS(812),
    [anon_sym_AMP_AMP] = ACTIONS(812),
    [anon_sym_PIPE_PIPE] = ACTIONS(812),
    [anon_sym_PIPE] = ACTIONS(814),
    [aux_sym_logical_expression_token1] = ACTIONS(812),
    [aux_sym_logical_expression_token2] = ACTIONS(812),
    [aux_sym_logical_expression_token3] = ACTIONS(812),
    [aux_sym_bitwise_expression_token1] = ACTIONS(812),
    [aux_sym_bitwise_expression_token2] = ACTIONS(812),
    [aux_sym_bitwise_expression_token3] = ACTIONS(812),
    [anon_sym_PLUS] = ACTIONS(814),
    [anon_sym_DASH] = ACTIONS(814),
    [anon_sym_SLASH] = ACTIONS(824),
    [anon_sym_BSLASH] = ACTIONS(826),
    [anon_sym_PERCENT] = ACTIONS(824),
    [anon_sym_STAR] = ACTIONS(824),
    [sym__statement_terminator] = ACTIONS(812),
  },
  [STATE(196)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(820),
    [anon_sym_BANG_EQ] = ACTIONS(820),
    [anon_sym_PLUS_EQ] = ACTIONS(820),
    [anon_sym_STAR_EQ] = ACTIONS(820),
    [anon_sym_SLASH_EQ] = ACTIONS(820),
    [anon_sym_PERCENT_EQ] = ACTIONS(820),
    [anon_sym_DASH_EQ] = ACTIONS(820),
    [anon_sym_GT] = ACTIONS(822),
    [anon_sym_GT_GT] = ACTIONS(820),
    [anon_sym_2_GT] = ACTIONS(822),
    [anon_sym_2_GT_GT] = ACTIONS(820),
    [anon_sym_3_GT] = ACTIONS(822),
    [anon_sym_3_GT_GT] = ACTIONS(820),
    [anon_sym_4_GT] = ACTIONS(822),
    [anon_sym_4_GT_GT] = ACTIONS(820),
    [anon_sym_5_GT] = ACTIONS(822),
    [anon_sym_5_GT_GT] = ACTIONS(820),
    [anon_sym_6_GT] = ACTIONS(822),
    [anon_sym_6_GT_GT] = ACTIONS(820),
    [anon_sym_STAR_GT] = ACTIONS(822),
    [anon_sym_STAR_GT_GT] = ACTIONS(820),
    [anon_sym_LT] = ACTIONS(822),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(820),
    [anon_sym_2_GT_AMP1] = ACTIONS(820),
    [anon_sym_3_GT_AMP1] = ACTIONS(820),
    [anon_sym_4_GT_AMP1] = ACTIONS(820),
    [anon_sym_5_GT_AMP1] = ACTIONS(820),
    [anon_sym_6_GT_AMP1] = ACTIONS(820),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(820),
    [anon_sym_1_GT_AMP2] = ACTIONS(820),
    [anon_sym_3_GT_AMP2] = ACTIONS(820),
    [anon_sym_4_GT_AMP2] = ACTIONS(820),
    [anon_sym_5_GT_AMP2] = ACTIONS(820),
    [anon_sym_6_GT_AMP2] = ACTIONS(820),
    [aux_sym_comparison_operator_token1] = ACTIONS(820),
    [aux_sym_comparison_operator_token2] = ACTIONS(820),
    [aux_sym_comparison_operator_token3] = ACTIONS(820),
    [aux_sym_comparison_operator_token4] = ACTIONS(820),
    [aux_sym_comparison_operator_token5] = ACTIONS(820),
    [aux_sym_comparison_operator_token6] = ACTIONS(820),
    [aux_sym_comparison_operator_token7] = ACTIONS(820),
    [aux_sym_comparison_operator_token8] = ACTIONS(820),
    [aux_sym_comparison_operator_token9] = ACTIONS(820),
    [aux_sym_comparison_operator_token10] = ACTIONS(820),
    [aux_sym_comparison_operator_token11] = ACTIONS(820),
    [aux_sym_comparison_operator_token12] = ACTIONS(820),
    [aux_sym_comparison_operator_token13] = ACTIONS(820),
    [aux_sym_comparison_operator_token14] = ACTIONS(820),
    [aux_sym_comparison_operator_token15] = ACTIONS(820),
    [aux_sym_comparison_operator_token16] = ACTIONS(820),
    [aux_sym_comparison_operator_token17] = ACTIONS(820),
    [aux_sym_comparison_operator_token18] = ACTIONS(820),
    [aux_sym_comparison_operator_token19] = ACTIONS(820),
    [aux_sym_comparison_operator_token20] = ACTIONS(820),
    [aux_sym_comparison_operator_token21] = ACTIONS(820),
    [aux_sym_comparison_operator_token22] = ACTIONS(820),
    [aux_sym_comparison_operator_token23] = ACTIONS(820),
    [aux_sym_comparison_operator_token24] = ACTIONS(820),
    [aux_sym_comparison_operator_token25] = ACTIONS(820),
    [aux_sym_comparison_operator_token26] = ACTIONS(820),
    [aux_sym_comparison_operator_token27] = ACTIONS(820),
    [aux_sym_comparison_operator_token28] = ACTIONS(822),
    [aux_sym_comparison_operator_token29] = ACTIONS(820),
    [aux_sym_comparison_operator_token30] = ACTIONS(820),
    [aux_sym_comparison_operator_token31] = ACTIONS(820),
    [aux_sym_comparison_operator_token32] = ACTIONS(820),
    [aux_sym_comparison_operator_token33] = ACTIONS(820),
    [aux_sym_comparison_operator_token34] = ACTIONS(822),
    [aux_sym_comparison_operator_token35] = ACTIONS(820),
    [aux_sym_comparison_operator_token36] = ACTIONS(820),
    [aux_sym_comparison_operator_token37] = ACTIONS(820),
    [aux_sym_comparison_operator_token38] = ACTIONS(820),
    [aux_sym_comparison_operator_token39] = ACTIONS(820),
    [aux_sym_comparison_operator_token40] = ACTIONS(820),
    [aux_sym_comparison_operator_token41] = ACTIONS(820),
    [aux_sym_comparison_operator_token42] = ACTIONS(820),
    [aux_sym_comparison_operator_token43] = ACTIONS(820),
    [aux_sym_comparison_operator_token44] = ACTIONS(820),
    [aux_sym_comparison_operator_token45] = ACTIONS(820),
    [aux_sym_comparison_operator_token46] = ACTIONS(820),
    [aux_sym_comparison_operator_token47] = ACTIONS(820),
    [aux_sym_comparison_operator_token48] = ACTIONS(820),
    [aux_sym_comparison_operator_token49] = ACTIONS(820),
    [aux_sym_comparison_operator_token50] = ACTIONS(820),
    [anon_sym_AMP_AMP] = ACTIONS(820),
    [anon_sym_PIPE_PIPE] = ACTIONS(820),
    [anon_sym_PIPE] = ACTIONS(822),
    [aux_sym_logical_expression_token1] = ACTIONS(820),
    [aux_sym_logical_expression_token2] = ACTIONS(820),
    [aux_sym_logical_expression_token3] = ACTIONS(820),
    [aux_sym_bitwise_expression_token1] = ACTIONS(820),
    [aux_sym_bitwise_expression_token2] = ACTIONS(820),
    [aux_sym_bitwise_expression_token3] = ACTIONS(820),
    [anon_sym_PLUS] = ACTIONS(822),
    [anon_sym_DASH] = ACTIONS(822),
    [anon_sym_SLASH] = ACTIONS(824),
    [anon_sym_BSLASH] = ACTIONS(826),
    [anon_sym_PERCENT] = ACTIONS(824),
    [anon_sym_STAR] = ACTIONS(824),
    [sym__statement_terminator] = ACTIONS(820),
  },
  [STATE(197)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(828),
    [anon_sym_BANG_EQ] = ACTIONS(828),
    [anon_sym_PLUS_EQ] = ACTIONS(828),
    [anon_sym_STAR_EQ] = ACTIONS(828),
    [anon_sym_SLASH_EQ] = ACTIONS(828),
    [anon_sym_PERCENT_EQ] = ACTIONS(828),
    [anon_sym_DASH_EQ] = ACTIONS(828),
    [anon_sym_GT] = ACTIONS(830),
    [anon_sym_GT_GT] = ACTIONS(828),
    [anon_sym_2_GT] = ACTIONS(830),
    [anon_sym_2_GT_GT] = ACTIONS(828),
    [anon_sym_3_GT] = ACTIONS(830),
    [anon_sym_3_GT_GT] = ACTIONS(828),
    [anon_sym_4_GT] = ACTIONS(830),
    [anon_sym_4_GT_GT] = ACTIONS(828),
    [anon_sym_5_GT] = ACTIONS(830),
    [anon_sym_5_GT_GT] = ACTIONS(828),
    [anon_sym_6_GT] = ACTIONS(830),
    [anon_sym_6_GT_GT] = ACTIONS(828),
    [anon_sym_STAR_GT] = ACTIONS(830),
    [anon_sym_STAR_GT_GT] = ACTIONS(828),
    [anon_sym_LT] = ACTIONS(830),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(828),
    [anon_sym_2_GT_AMP1] = ACTIONS(828),
    [anon_sym_3_GT_AMP1] = ACTIONS(828),
    [anon_sym_4_GT_AMP1] = ACTIONS(828),
    [anon_sym_5_GT_AMP1] = ACTIONS(828),
    [anon_sym_6_GT_AMP1] = ACTIONS(828),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(828),
    [anon_sym_1_GT_AMP2] = ACTIONS(828),
    [anon_sym_3_GT_AMP2] = ACTIONS(828),
    [anon_sym_4_GT_AMP2] = ACTIONS(828),
    [anon_sym_5_GT_AMP2] = ACTIONS(828),
    [anon_sym_6_GT_AMP2] = ACTIONS(828),
    [aux_sym_comparison_operator_token1] = ACTIONS(828),
    [aux_sym_comparison_operator_token2] = ACTIONS(828),
    [aux_sym_comparison_operator_token3] = ACTIONS(828),
    [aux_sym_comparison_operator_token4] = ACTIONS(828),
    [aux_sym_comparison_operator_token5] = ACTIONS(828),
    [aux_sym_comparison_operator_token6] = ACTIONS(828),
    [aux_sym_comparison_operator_token7] = ACTIONS(828),
    [aux_sym_comparison_operator_token8] = ACTIONS(828),
    [aux_sym_comparison_operator_token9] = ACTIONS(828),
    [aux_sym_comparison_operator_token10] = ACTIONS(828),
    [aux_sym_comparison_operator_token11] = ACTIONS(828),
    [aux_sym_comparison_operator_token12] = ACTIONS(828),
    [aux_sym_comparison_operator_token13] = ACTIONS(828),
    [aux_sym_comparison_operator_token14] = ACTIONS(828),
    [aux_sym_comparison_operator_token15] = ACTIONS(828),
    [aux_sym_comparison_operator_token16] = ACTIONS(828),
    [aux_sym_comparison_operator_token17] = ACTIONS(828),
    [aux_sym_comparison_operator_token18] = ACTIONS(828),
    [aux_sym_comparison_operator_token19] = ACTIONS(828),
    [aux_sym_comparison_operator_token20] = ACTIONS(828),
    [aux_sym_comparison_operator_token21] = ACTIONS(828),
    [aux_sym_comparison_operator_token22] = ACTIONS(828),
    [aux_sym_comparison_operator_token23] = ACTIONS(828),
    [aux_sym_comparison_operator_token24] = ACTIONS(828),
    [aux_sym_comparison_operator_token25] = ACTIONS(828),
    [aux_sym_comparison_operator_token26] = ACTIONS(828),
    [aux_sym_comparison_operator_token27] = ACTIONS(828),
    [aux_sym_comparison_operator_token28] = ACTIONS(830),
    [aux_sym_comparison_operator_token29] = ACTIONS(828),
    [aux_sym_comparison_operator_token30] = ACTIONS(828),
    [aux_sym_comparison_operator_token31] = ACTIONS(828),
    [aux_sym_comparison_operator_token32] = ACTIONS(828),
    [aux_sym_comparison_operator_token33] = ACTIONS(828),
    [aux_sym_comparison_operator_token34] = ACTIONS(830),
    [aux_sym_comparison_operator_token35] = ACTIONS(828),
    [aux_sym_comparison_operator_token36] = ACTIONS(828),
    [aux_sym_comparison_operator_token37] = ACTIONS(828),
    [aux_sym_comparison_operator_token38] = ACTIONS(828),
    [aux_sym_comparison_operator_token39] = ACTIONS(828),
    [aux_sym_comparison_operator_token40] = ACTIONS(828),
    [aux_sym_comparison_operator_token41] = ACTIONS(828),
    [aux_sym_comparison_operator_token42] = ACTIONS(828),
    [aux_sym_comparison_operator_token43] = ACTIONS(828),
    [aux_sym_comparison_operator_token44] = ACTIONS(828),
    [aux_sym_comparison_operator_token45] = ACTIONS(828),
    [aux_sym_comparison_operator_token46] = ACTIONS(828),
    [aux_sym_comparison_operator_token47] = ACTIONS(828),
    [aux_sym_comparison_operator_token48] = ACTIONS(828),
    [aux_sym_comparison_operator_token49] = ACTIONS(828),
    [aux_sym_comparison_operator_token50] = ACTIONS(828),
    [anon_sym_RPAREN] = ACTIONS(828),
    [anon_sym_AMP_AMP] = ACTIONS(828),
    [anon_sym_PIPE_PIPE] = ACTIONS(828),
    [anon_sym_PIPE] = ACTIONS(830),
    [aux_sym_logical_expression_token1] = ACTIONS(828),
    [aux_sym_logical_expression_token2] = ACTIONS(828),
    [aux_sym_logical_expression_token3] = ACTIONS(828),
    [aux_sym_bitwise_expression_token1] = ACTIONS(828),
    [aux_sym_bitwise_expression_token2] = ACTIONS(828),
    [aux_sym_bitwise_expression_token3] = ACTIONS(828),
    [anon_sym_PLUS] = ACTIONS(832),
    [anon_sym_DASH] = ACTIONS(832),
  },
  [STATE(198)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(828),
    [anon_sym_BANG_EQ] = ACTIONS(828),
    [anon_sym_PLUS_EQ] = ACTIONS(828),
    [anon_sym_STAR_EQ] = ACTIONS(828),
    [anon_sym_SLASH_EQ] = ACTIONS(828),
    [anon_sym_PERCENT_EQ] = ACTIONS(828),
    [anon_sym_DASH_EQ] = ACTIONS(828),
    [anon_sym_GT] = ACTIONS(830),
    [anon_sym_GT_GT] = ACTIONS(828),
    [anon_sym_2_GT] = ACTIONS(830),
    [anon_sym_2_GT_GT] = ACTIONS(828),
    [anon_sym_3_GT] = ACTIONS(830),
    [anon_sym_3_GT_GT] = ACTIONS(828),
    [anon_sym_4_GT] = ACTIONS(830),
    [anon_sym_4_GT_GT] = ACTIONS(828),
    [anon_sym_5_GT] = ACTIONS(830),
    [anon_sym_5_GT_GT] = ACTIONS(828),
    [anon_sym_6_GT] = ACTIONS(830),
    [anon_sym_6_GT_GT] = ACTIONS(828),
    [anon_sym_STAR_GT] = ACTIONS(830),
    [anon_sym_STAR_GT_GT] = ACTIONS(828),
    [anon_sym_LT] = ACTIONS(830),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(828),
    [anon_sym_2_GT_AMP1] = ACTIONS(828),
    [anon_sym_3_GT_AMP1] = ACTIONS(828),
    [anon_sym_4_GT_AMP1] = ACTIONS(828),
    [anon_sym_5_GT_AMP1] = ACTIONS(828),
    [anon_sym_6_GT_AMP1] = ACTIONS(828),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(828),
    [anon_sym_1_GT_AMP2] = ACTIONS(828),
    [anon_sym_3_GT_AMP2] = ACTIONS(828),
    [anon_sym_4_GT_AMP2] = ACTIONS(828),
    [anon_sym_5_GT_AMP2] = ACTIONS(828),
    [anon_sym_6_GT_AMP2] = ACTIONS(828),
    [aux_sym_comparison_operator_token1] = ACTIONS(828),
    [aux_sym_comparison_operator_token2] = ACTIONS(828),
    [aux_sym_comparison_operator_token3] = ACTIONS(828),
    [aux_sym_comparison_operator_token4] = ACTIONS(828),
    [aux_sym_comparison_operator_token5] = ACTIONS(828),
    [aux_sym_comparison_operator_token6] = ACTIONS(828),
    [aux_sym_comparison_operator_token7] = ACTIONS(828),
    [aux_sym_comparison_operator_token8] = ACTIONS(828),
    [aux_sym_comparison_operator_token9] = ACTIONS(828),
    [aux_sym_comparison_operator_token10] = ACTIONS(828),
    [aux_sym_comparison_operator_token11] = ACTIONS(828),
    [aux_sym_comparison_operator_token12] = ACTIONS(828),
    [aux_sym_comparison_operator_token13] = ACTIONS(828),
    [aux_sym_comparison_operator_token14] = ACTIONS(828),
    [aux_sym_comparison_operator_token15] = ACTIONS(828),
    [aux_sym_comparison_operator_token16] = ACTIONS(828),
    [aux_sym_comparison_operator_token17] = ACTIONS(828),
    [aux_sym_comparison_operator_token18] = ACTIONS(828),
    [aux_sym_comparison_operator_token19] = ACTIONS(828),
    [aux_sym_comparison_operator_token20] = ACTIONS(828),
    [aux_sym_comparison_operator_token21] = ACTIONS(828),
    [aux_sym_comparison_operator_token22] = ACTIONS(828),
    [aux_sym_comparison_operator_token23] = ACTIONS(828),
    [aux_sym_comparison_operator_token24] = ACTIONS(828),
    [aux_sym_comparison_operator_token25] = ACTIONS(828),
    [aux_sym_comparison_operator_token26] = ACTIONS(828),
    [aux_sym_comparison_operator_token27] = ACTIONS(828),
    [aux_sym_comparison_operator_token28] = ACTIONS(830),
    [aux_sym_comparison_operator_token29] = ACTIONS(828),
    [aux_sym_comparison_operator_token30] = ACTIONS(828),
    [aux_sym_comparison_operator_token31] = ACTIONS(828),
    [aux_sym_comparison_operator_token32] = ACTIONS(828),
    [aux_sym_comparison_operator_token33] = ACTIONS(828),
    [aux_sym_comparison_operator_token34] = ACTIONS(830),
    [aux_sym_comparison_operator_token35] = ACTIONS(828),
    [aux_sym_comparison_operator_token36] = ACTIONS(828),
    [aux_sym_comparison_operator_token37] = ACTIONS(828),
    [aux_sym_comparison_operator_token38] = ACTIONS(828),
    [aux_sym_comparison_operator_token39] = ACTIONS(828),
    [aux_sym_comparison_operator_token40] = ACTIONS(828),
    [aux_sym_comparison_operator_token41] = ACTIONS(828),
    [aux_sym_comparison_operator_token42] = ACTIONS(828),
    [aux_sym_comparison_operator_token43] = ACTIONS(828),
    [aux_sym_comparison_operator_token44] = ACTIONS(828),
    [aux_sym_comparison_operator_token45] = ACTIONS(828),
    [aux_sym_comparison_operator_token46] = ACTIONS(828),
    [aux_sym_comparison_operator_token47] = ACTIONS(828),
    [aux_sym_comparison_operator_token48] = ACTIONS(828),
    [aux_sym_comparison_operator_token49] = ACTIONS(828),
    [aux_sym_comparison_operator_token50] = ACTIONS(828),
    [anon_sym_AMP_AMP] = ACTIONS(828),
    [anon_sym_PIPE_PIPE] = ACTIONS(828),
    [anon_sym_PIPE] = ACTIONS(830),
    [aux_sym_logical_expression_token1] = ACTIONS(828),
    [aux_sym_logical_expression_token2] = ACTIONS(828),
    [aux_sym_logical_expression_token3] = ACTIONS(828),
    [aux_sym_bitwise_expression_token1] = ACTIONS(828),
    [aux_sym_bitwise_expression_token2] = ACTIONS(828),
    [aux_sym_bitwise_expression_token3] = ACTIONS(828),
    [anon_sym_PLUS] = ACTIONS(834),
    [anon_sym_DASH] = ACTIONS(834),
    [sym__statement_terminator] = ACTIONS(828),
  },
  [STATE(199)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(836),
    [anon_sym_BANG_EQ] = ACTIONS(836),
    [anon_sym_PLUS_EQ] = ACTIONS(836),
    [anon_sym_STAR_EQ] = ACTIONS(836),
    [anon_sym_SLASH_EQ] = ACTIONS(836),
    [anon_sym_PERCENT_EQ] = ACTIONS(836),
    [anon_sym_DASH_EQ] = ACTIONS(836),
    [anon_sym_GT] = ACTIONS(838),
    [anon_sym_GT_GT] = ACTIONS(836),
    [anon_sym_2_GT] = ACTIONS(838),
    [anon_sym_2_GT_GT] = ACTIONS(836),
    [anon_sym_3_GT] = ACTIONS(838),
    [anon_sym_3_GT_GT] = ACTIONS(836),
    [anon_sym_4_GT] = ACTIONS(838),
    [anon_sym_4_GT_GT] = ACTIONS(836),
    [anon_sym_5_GT] = ACTIONS(838),
    [anon_sym_5_GT_GT] = ACTIONS(836),
    [anon_sym_6_GT] = ACTIONS(838),
    [anon_sym_6_GT_GT] = ACTIONS(836),
    [anon_sym_STAR_GT] = ACTIONS(838),
    [anon_sym_STAR_GT_GT] = ACTIONS(836),
    [anon_sym_LT] = ACTIONS(838),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(836),
    [anon_sym_2_GT_AMP1] = ACTIONS(836),
    [anon_sym_3_GT_AMP1] = ACTIONS(836),
    [anon_sym_4_GT_AMP1] = ACTIONS(836),
    [anon_sym_5_GT_AMP1] = ACTIONS(836),
    [anon_sym_6_GT_AMP1] = ACTIONS(836),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(836),
    [anon_sym_1_GT_AMP2] = ACTIONS(836),
    [anon_sym_3_GT_AMP2] = ACTIONS(836),
    [anon_sym_4_GT_AMP2] = ACTIONS(836),
    [anon_sym_5_GT_AMP2] = ACTIONS(836),
    [anon_sym_6_GT_AMP2] = ACTIONS(836),
    [aux_sym_comparison_operator_token1] = ACTIONS(836),
    [aux_sym_comparison_operator_token2] = ACTIONS(836),
    [aux_sym_comparison_operator_token3] = ACTIONS(836),
    [aux_sym_comparison_operator_token4] = ACTIONS(836),
    [aux_sym_comparison_operator_token5] = ACTIONS(836),
    [aux_sym_comparison_operator_token6] = ACTIONS(836),
    [aux_sym_comparison_operator_token7] = ACTIONS(836),
    [aux_sym_comparison_operator_token8] = ACTIONS(836),
    [aux_sym_comparison_operator_token9] = ACTIONS(836),
    [aux_sym_comparison_operator_token10] = ACTIONS(836),
    [aux_sym_comparison_operator_token11] = ACTIONS(836),
    [aux_sym_comparison_operator_token12] = ACTIONS(836),
    [aux_sym_comparison_operator_token13] = ACTIONS(836),
    [aux_sym_comparison_operator_token14] = ACTIONS(836),
    [aux_sym_comparison_operator_token15] = ACTIONS(836),
    [aux_sym_comparison_operator_token16] = ACTIONS(836),
    [aux_sym_comparison_operator_token17] = ACTIONS(836),
    [aux_sym_comparison_operator_token18] = ACTIONS(836),
    [aux_sym_comparison_operator_token19] = ACTIONS(836),
    [aux_sym_comparison_operator_token20] = ACTIONS(836),
    [aux_sym_comparison_operator_token21] = ACTIONS(836),
    [aux_sym_comparison_operator_token22] = ACTIONS(836),
    [aux_sym_comparison_operator_token23] = ACTIONS(836),
    [aux_sym_comparison_operator_token24] = ACTIONS(836),
    [aux_sym_comparison_operator_token25] = ACTIONS(836),
    [aux_sym_comparison_operator_token26] = ACTIONS(836),
    [aux_sym_comparison_operator_token27] = ACTIONS(836),
    [aux_sym_comparison_operator_token28] = ACTIONS(838),
    [aux_sym_comparison_operator_token29] = ACTIONS(836),
    [aux_sym_comparison_operator_token30] = ACTIONS(836),
    [aux_sym_comparison_operator_token31] = ACTIONS(836),
    [aux_sym_comparison_operator_token32] = ACTIONS(836),
    [aux_sym_comparison_operator_token33] = ACTIONS(836),
    [aux_sym_comparison_operator_token34] = ACTIONS(838),
    [aux_sym_comparison_operator_token35] = ACTIONS(836),
    [aux_sym_comparison_operator_token36] = ACTIONS(836),
    [aux_sym_comparison_operator_token37] = ACTIONS(836),
    [aux_sym_comparison_operator_token38] = ACTIONS(836),
    [aux_sym_comparison_operator_token39] = ACTIONS(836),
    [aux_sym_comparison_operator_token40] = ACTIONS(836),
    [aux_sym_comparison_operator_token41] = ACTIONS(836),
    [aux_sym_comparison_operator_token42] = ACTIONS(836),
    [aux_sym_comparison_operator_token43] = ACTIONS(836),
    [aux_sym_comparison_operator_token44] = ACTIONS(836),
    [aux_sym_comparison_operator_token45] = ACTIONS(836),
    [aux_sym_comparison_operator_token46] = ACTIONS(836),
    [aux_sym_comparison_operator_token47] = ACTIONS(836),
    [aux_sym_comparison_operator_token48] = ACTIONS(836),
    [aux_sym_comparison_operator_token49] = ACTIONS(836),
    [aux_sym_comparison_operator_token50] = ACTIONS(836),
    [anon_sym_AMP_AMP] = ACTIONS(836),
    [anon_sym_PIPE_PIPE] = ACTIONS(836),
    [anon_sym_PIPE] = ACTIONS(838),
    [aux_sym_logical_expression_token1] = ACTIONS(836),
    [aux_sym_logical_expression_token2] = ACTIONS(836),
    [aux_sym_logical_expression_token3] = ACTIONS(836),
    [aux_sym_bitwise_expression_token1] = ACTIONS(836),
    [aux_sym_bitwise_expression_token2] = ACTIONS(836),
    [aux_sym_bitwise_expression_token3] = ACTIONS(836),
    [anon_sym_PLUS] = ACTIONS(834),
    [anon_sym_DASH] = ACTIONS(834),
    [sym__statement_terminator] = ACTIONS(836),
  },
  [STATE(200)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(836),
    [anon_sym_BANG_EQ] = ACTIONS(836),
    [anon_sym_PLUS_EQ] = ACTIONS(836),
    [anon_sym_STAR_EQ] = ACTIONS(836),
    [anon_sym_SLASH_EQ] = ACTIONS(836),
    [anon_sym_PERCENT_EQ] = ACTIONS(836),
    [anon_sym_DASH_EQ] = ACTIONS(836),
    [anon_sym_GT] = ACTIONS(838),
    [anon_sym_GT_GT] = ACTIONS(836),
    [anon_sym_2_GT] = ACTIONS(838),
    [anon_sym_2_GT_GT] = ACTIONS(836),
    [anon_sym_3_GT] = ACTIONS(838),
    [anon_sym_3_GT_GT] = ACTIONS(836),
    [anon_sym_4_GT] = ACTIONS(838),
    [anon_sym_4_GT_GT] = ACTIONS(836),
    [anon_sym_5_GT] = ACTIONS(838),
    [anon_sym_5_GT_GT] = ACTIONS(836),
    [anon_sym_6_GT] = ACTIONS(838),
    [anon_sym_6_GT_GT] = ACTIONS(836),
    [anon_sym_STAR_GT] = ACTIONS(838),
    [anon_sym_STAR_GT_GT] = ACTIONS(836),
    [anon_sym_LT] = ACTIONS(838),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(836),
    [anon_sym_2_GT_AMP1] = ACTIONS(836),
    [anon_sym_3_GT_AMP1] = ACTIONS(836),
    [anon_sym_4_GT_AMP1] = ACTIONS(836),
    [anon_sym_5_GT_AMP1] = ACTIONS(836),
    [anon_sym_6_GT_AMP1] = ACTIONS(836),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(836),
    [anon_sym_1_GT_AMP2] = ACTIONS(836),
    [anon_sym_3_GT_AMP2] = ACTIONS(836),
    [anon_sym_4_GT_AMP2] = ACTIONS(836),
    [anon_sym_5_GT_AMP2] = ACTIONS(836),
    [anon_sym_6_GT_AMP2] = ACTIONS(836),
    [aux_sym_comparison_operator_token1] = ACTIONS(836),
    [aux_sym_comparison_operator_token2] = ACTIONS(836),
    [aux_sym_comparison_operator_token3] = ACTIONS(836),
    [aux_sym_comparison_operator_token4] = ACTIONS(836),
    [aux_sym_comparison_operator_token5] = ACTIONS(836),
    [aux_sym_comparison_operator_token6] = ACTIONS(836),
    [aux_sym_comparison_operator_token7] = ACTIONS(836),
    [aux_sym_comparison_operator_token8] = ACTIONS(836),
    [aux_sym_comparison_operator_token9] = ACTIONS(836),
    [aux_sym_comparison_operator_token10] = ACTIONS(836),
    [aux_sym_comparison_operator_token11] = ACTIONS(836),
    [aux_sym_comparison_operator_token12] = ACTIONS(836),
    [aux_sym_comparison_operator_token13] = ACTIONS(836),
    [aux_sym_comparison_operator_token14] = ACTIONS(836),
    [aux_sym_comparison_operator_token15] = ACTIONS(836),
    [aux_sym_comparison_operator_token16] = ACTIONS(836),
    [aux_sym_comparison_operator_token17] = ACTIONS(836),
    [aux_sym_comparison_operator_token18] = ACTIONS(836),
    [aux_sym_comparison_operator_token19] = ACTIONS(836),
    [aux_sym_comparison_operator_token20] = ACTIONS(836),
    [aux_sym_comparison_operator_token21] = ACTIONS(836),
    [aux_sym_comparison_operator_token22] = ACTIONS(836),
    [aux_sym_comparison_operator_token23] = ACTIONS(836),
    [aux_sym_comparison_operator_token24] = ACTIONS(836),
    [aux_sym_comparison_operator_token25] = ACTIONS(836),
    [aux_sym_comparison_operator_token26] = ACTIONS(836),
    [aux_sym_comparison_operator_token27] = ACTIONS(836),
    [aux_sym_comparison_operator_token28] = ACTIONS(838),
    [aux_sym_comparison_operator_token29] = ACTIONS(836),
    [aux_sym_comparison_operator_token30] = ACTIONS(836),
    [aux_sym_comparison_operator_token31] = ACTIONS(836),
    [aux_sym_comparison_operator_token32] = ACTIONS(836),
    [aux_sym_comparison_operator_token33] = ACTIONS(836),
    [aux_sym_comparison_operator_token34] = ACTIONS(838),
    [aux_sym_comparison_operator_token35] = ACTIONS(836),
    [aux_sym_comparison_operator_token36] = ACTIONS(836),
    [aux_sym_comparison_operator_token37] = ACTIONS(836),
    [aux_sym_comparison_operator_token38] = ACTIONS(836),
    [aux_sym_comparison_operator_token39] = ACTIONS(836),
    [aux_sym_comparison_operator_token40] = ACTIONS(836),
    [aux_sym_comparison_operator_token41] = ACTIONS(836),
    [aux_sym_comparison_operator_token42] = ACTIONS(836),
    [aux_sym_comparison_operator_token43] = ACTIONS(836),
    [aux_sym_comparison_operator_token44] = ACTIONS(836),
    [aux_sym_comparison_operator_token45] = ACTIONS(836),
    [aux_sym_comparison_operator_token46] = ACTIONS(836),
    [aux_sym_comparison_operator_token47] = ACTIONS(836),
    [aux_sym_comparison_operator_token48] = ACTIONS(836),
    [aux_sym_comparison_operator_token49] = ACTIONS(836),
    [aux_sym_comparison_operator_token50] = ACTIONS(836),
    [anon_sym_RPAREN] = ACTIONS(836),
    [anon_sym_AMP_AMP] = ACTIONS(836),
    [anon_sym_PIPE_PIPE] = ACTIONS(836),
    [anon_sym_PIPE] = ACTIONS(838),
    [aux_sym_logical_expression_token1] = ACTIONS(836),
    [aux_sym_logical_expression_token2] = ACTIONS(836),
    [aux_sym_logical_expression_token3] = ACTIONS(836),
    [aux_sym_bitwise_expression_token1] = ACTIONS(836),
    [aux_sym_bitwise_expression_token2] = ACTIONS(836),
    [aux_sym_bitwise_expression_token3] = ACTIONS(836),
    [anon_sym_PLUS] = ACTIONS(832),
    [anon_sym_DASH] = ACTIONS(832),
  },
  [STATE(201)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2048),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(840),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(842),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(844),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(202)] = {
    [sym_comparison_operator] = STATE(553),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(846),
    [anon_sym_BANG_EQ] = ACTIONS(846),
    [anon_sym_PLUS_EQ] = ACTIONS(846),
    [anon_sym_STAR_EQ] = ACTIONS(846),
    [anon_sym_SLASH_EQ] = ACTIONS(846),
    [anon_sym_PERCENT_EQ] = ACTIONS(846),
    [anon_sym_DASH_EQ] = ACTIONS(846),
    [anon_sym_GT] = ACTIONS(848),
    [anon_sym_GT_GT] = ACTIONS(846),
    [anon_sym_2_GT] = ACTIONS(848),
    [anon_sym_2_GT_GT] = ACTIONS(846),
    [anon_sym_3_GT] = ACTIONS(848),
    [anon_sym_3_GT_GT] = ACTIONS(846),
    [anon_sym_4_GT] = ACTIONS(848),
    [anon_sym_4_GT_GT] = ACTIONS(846),
    [anon_sym_5_GT] = ACTIONS(848),
    [anon_sym_5_GT_GT] = ACTIONS(846),
    [anon_sym_6_GT] = ACTIONS(848),
    [anon_sym_6_GT_GT] = ACTIONS(846),
    [anon_sym_STAR_GT] = ACTIONS(848),
    [anon_sym_STAR_GT_GT] = ACTIONS(846),
    [anon_sym_LT] = ACTIONS(848),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(846),
    [anon_sym_2_GT_AMP1] = ACTIONS(846),
    [anon_sym_3_GT_AMP1] = ACTIONS(846),
    [anon_sym_4_GT_AMP1] = ACTIONS(846),
    [anon_sym_5_GT_AMP1] = ACTIONS(846),
    [anon_sym_6_GT_AMP1] = ACTIONS(846),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(846),
    [anon_sym_1_GT_AMP2] = ACTIONS(846),
    [anon_sym_3_GT_AMP2] = ACTIONS(846),
    [anon_sym_4_GT_AMP2] = ACTIONS(846),
    [anon_sym_5_GT_AMP2] = ACTIONS(846),
    [anon_sym_6_GT_AMP2] = ACTIONS(846),
    [aux_sym_comparison_operator_token1] = ACTIONS(850),
    [aux_sym_comparison_operator_token2] = ACTIONS(850),
    [aux_sym_comparison_operator_token3] = ACTIONS(850),
    [aux_sym_comparison_operator_token4] = ACTIONS(850),
    [aux_sym_comparison_operator_token5] = ACTIONS(850),
    [aux_sym_comparison_operator_token6] = ACTIONS(850),
    [aux_sym_comparison_operator_token7] = ACTIONS(850),
    [aux_sym_comparison_operator_token8] = ACTIONS(850),
    [aux_sym_comparison_operator_token9] = ACTIONS(850),
    [aux_sym_comparison_operator_token10] = ACTIONS(850),
    [aux_sym_comparison_operator_token11] = ACTIONS(850),
    [aux_sym_comparison_operator_token12] = ACTIONS(850),
    [aux_sym_comparison_operator_token13] = ACTIONS(850),
    [aux_sym_comparison_operator_token14] = ACTIONS(850),
    [aux_sym_comparison_operator_token15] = ACTIONS(850),
    [aux_sym_comparison_operator_token16] = ACTIONS(850),
    [aux_sym_comparison_operator_token17] = ACTIONS(850),
    [aux_sym_comparison_operator_token18] = ACTIONS(850),
    [aux_sym_comparison_operator_token19] = ACTIONS(850),
    [aux_sym_comparison_operator_token20] = ACTIONS(850),
    [aux_sym_comparison_operator_token21] = ACTIONS(850),
    [aux_sym_comparison_operator_token22] = ACTIONS(850),
    [aux_sym_comparison_operator_token23] = ACTIONS(850),
    [aux_sym_comparison_operator_token24] = ACTIONS(850),
    [aux_sym_comparison_operator_token25] = ACTIONS(850),
    [aux_sym_comparison_operator_token26] = ACTIONS(850),
    [aux_sym_comparison_operator_token27] = ACTIONS(850),
    [aux_sym_comparison_operator_token28] = ACTIONS(852),
    [aux_sym_comparison_operator_token29] = ACTIONS(850),
    [aux_sym_comparison_operator_token30] = ACTIONS(850),
    [aux_sym_comparison_operator_token31] = ACTIONS(850),
    [aux_sym_comparison_operator_token32] = ACTIONS(850),
    [aux_sym_comparison_operator_token33] = ACTIONS(850),
    [aux_sym_comparison_operator_token34] = ACTIONS(852),
    [aux_sym_comparison_operator_token35] = ACTIONS(850),
    [aux_sym_comparison_operator_token36] = ACTIONS(850),
    [aux_sym_comparison_operator_token37] = ACTIONS(850),
    [aux_sym_comparison_operator_token38] = ACTIONS(850),
    [aux_sym_comparison_operator_token39] = ACTIONS(850),
    [aux_sym_comparison_operator_token40] = ACTIONS(850),
    [aux_sym_comparison_operator_token41] = ACTIONS(850),
    [aux_sym_comparison_operator_token42] = ACTIONS(850),
    [aux_sym_comparison_operator_token43] = ACTIONS(850),
    [aux_sym_comparison_operator_token44] = ACTIONS(850),
    [aux_sym_comparison_operator_token45] = ACTIONS(850),
    [aux_sym_comparison_operator_token46] = ACTIONS(850),
    [aux_sym_comparison_operator_token47] = ACTIONS(850),
    [aux_sym_comparison_operator_token48] = ACTIONS(850),
    [aux_sym_comparison_operator_token49] = ACTIONS(850),
    [aux_sym_comparison_operator_token50] = ACTIONS(850),
    [anon_sym_AMP_AMP] = ACTIONS(846),
    [anon_sym_PIPE_PIPE] = ACTIONS(846),
    [anon_sym_PIPE] = ACTIONS(848),
    [aux_sym_logical_expression_token1] = ACTIONS(846),
    [aux_sym_logical_expression_token2] = ACTIONS(846),
    [aux_sym_logical_expression_token3] = ACTIONS(846),
    [aux_sym_bitwise_expression_token1] = ACTIONS(846),
    [aux_sym_bitwise_expression_token2] = ACTIONS(846),
    [aux_sym_bitwise_expression_token3] = ACTIONS(846),
    [sym__statement_terminator] = ACTIONS(846),
  },
  [STATE(203)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_initializer] = STATE(2063),
    [sym_pipeline] = STATE(1748),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(856),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(858),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(204)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2051),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(862),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(864),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(205)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2067),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(866),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(206)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2070),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(872),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(874),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(876),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(207)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_initializer] = STATE(2081),
    [sym_pipeline] = STATE(1748),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(878),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(880),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(882),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(208)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2085),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(884),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(886),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(888),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(209)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2088),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(890),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(892),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(894),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(210)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_initializer] = STATE(2118),
    [sym_pipeline] = STATE(1748),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(896),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(898),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(900),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(211)] = {
    [sym_comparison_operator] = STATE(553),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(902),
    [anon_sym_BANG_EQ] = ACTIONS(902),
    [anon_sym_PLUS_EQ] = ACTIONS(902),
    [anon_sym_STAR_EQ] = ACTIONS(902),
    [anon_sym_SLASH_EQ] = ACTIONS(902),
    [anon_sym_PERCENT_EQ] = ACTIONS(902),
    [anon_sym_DASH_EQ] = ACTIONS(902),
    [anon_sym_GT] = ACTIONS(904),
    [anon_sym_GT_GT] = ACTIONS(902),
    [anon_sym_2_GT] = ACTIONS(904),
    [anon_sym_2_GT_GT] = ACTIONS(902),
    [anon_sym_3_GT] = ACTIONS(904),
    [anon_sym_3_GT_GT] = ACTIONS(902),
    [anon_sym_4_GT] = ACTIONS(904),
    [anon_sym_4_GT_GT] = ACTIONS(902),
    [anon_sym_5_GT] = ACTIONS(904),
    [anon_sym_5_GT_GT] = ACTIONS(902),
    [anon_sym_6_GT] = ACTIONS(904),
    [anon_sym_6_GT_GT] = ACTIONS(902),
    [anon_sym_STAR_GT] = ACTIONS(904),
    [anon_sym_STAR_GT_GT] = ACTIONS(902),
    [anon_sym_LT] = ACTIONS(904),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(902),
    [anon_sym_2_GT_AMP1] = ACTIONS(902),
    [anon_sym_3_GT_AMP1] = ACTIONS(902),
    [anon_sym_4_GT_AMP1] = ACTIONS(902),
    [anon_sym_5_GT_AMP1] = ACTIONS(902),
    [anon_sym_6_GT_AMP1] = ACTIONS(902),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(902),
    [anon_sym_1_GT_AMP2] = ACTIONS(902),
    [anon_sym_3_GT_AMP2] = ACTIONS(902),
    [anon_sym_4_GT_AMP2] = ACTIONS(902),
    [anon_sym_5_GT_AMP2] = ACTIONS(902),
    [anon_sym_6_GT_AMP2] = ACTIONS(902),
    [aux_sym_comparison_operator_token1] = ACTIONS(850),
    [aux_sym_comparison_operator_token2] = ACTIONS(850),
    [aux_sym_comparison_operator_token3] = ACTIONS(850),
    [aux_sym_comparison_operator_token4] = ACTIONS(850),
    [aux_sym_comparison_operator_token5] = ACTIONS(850),
    [aux_sym_comparison_operator_token6] = ACTIONS(850),
    [aux_sym_comparison_operator_token7] = ACTIONS(850),
    [aux_sym_comparison_operator_token8] = ACTIONS(850),
    [aux_sym_comparison_operator_token9] = ACTIONS(850),
    [aux_sym_comparison_operator_token10] = ACTIONS(850),
    [aux_sym_comparison_operator_token11] = ACTIONS(850),
    [aux_sym_comparison_operator_token12] = ACTIONS(850),
    [aux_sym_comparison_operator_token13] = ACTIONS(850),
    [aux_sym_comparison_operator_token14] = ACTIONS(850),
    [aux_sym_comparison_operator_token15] = ACTIONS(850),
    [aux_sym_comparison_operator_token16] = ACTIONS(850),
    [aux_sym_comparison_operator_token17] = ACTIONS(850),
    [aux_sym_comparison_operator_token18] = ACTIONS(850),
    [aux_sym_comparison_operator_token19] = ACTIONS(850),
    [aux_sym_comparison_operator_token20] = ACTIONS(850),
    [aux_sym_comparison_operator_token21] = ACTIONS(850),
    [aux_sym_comparison_operator_token22] = ACTIONS(850),
    [aux_sym_comparison_operator_token23] = ACTIONS(850),
    [aux_sym_comparison_operator_token24] = ACTIONS(850),
    [aux_sym_comparison_operator_token25] = ACTIONS(850),
    [aux_sym_comparison_operator_token26] = ACTIONS(850),
    [aux_sym_comparison_operator_token27] = ACTIONS(850),
    [aux_sym_comparison_operator_token28] = ACTIONS(852),
    [aux_sym_comparison_operator_token29] = ACTIONS(850),
    [aux_sym_comparison_operator_token30] = ACTIONS(850),
    [aux_sym_comparison_operator_token31] = ACTIONS(850),
    [aux_sym_comparison_operator_token32] = ACTIONS(850),
    [aux_sym_comparison_operator_token33] = ACTIONS(850),
    [aux_sym_comparison_operator_token34] = ACTIONS(852),
    [aux_sym_comparison_operator_token35] = ACTIONS(850),
    [aux_sym_comparison_operator_token36] = ACTIONS(850),
    [aux_sym_comparison_operator_token37] = ACTIONS(850),
    [aux_sym_comparison_operator_token38] = ACTIONS(850),
    [aux_sym_comparison_operator_token39] = ACTIONS(850),
    [aux_sym_comparison_operator_token40] = ACTIONS(850),
    [aux_sym_comparison_operator_token41] = ACTIONS(850),
    [aux_sym_comparison_operator_token42] = ACTIONS(850),
    [aux_sym_comparison_operator_token43] = ACTIONS(850),
    [aux_sym_comparison_operator_token44] = ACTIONS(850),
    [aux_sym_comparison_operator_token45] = ACTIONS(850),
    [aux_sym_comparison_operator_token46] = ACTIONS(850),
    [aux_sym_comparison_operator_token47] = ACTIONS(850),
    [aux_sym_comparison_operator_token48] = ACTIONS(850),
    [aux_sym_comparison_operator_token49] = ACTIONS(850),
    [aux_sym_comparison_operator_token50] = ACTIONS(850),
    [anon_sym_AMP_AMP] = ACTIONS(902),
    [anon_sym_PIPE_PIPE] = ACTIONS(902),
    [anon_sym_PIPE] = ACTIONS(904),
    [aux_sym_logical_expression_token1] = ACTIONS(902),
    [aux_sym_logical_expression_token2] = ACTIONS(902),
    [aux_sym_logical_expression_token3] = ACTIONS(902),
    [aux_sym_bitwise_expression_token1] = ACTIONS(902),
    [aux_sym_bitwise_expression_token2] = ACTIONS(902),
    [aux_sym_bitwise_expression_token3] = ACTIONS(902),
    [sym__statement_terminator] = ACTIONS(902),
  },
  [STATE(212)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_initializer] = STATE(2044),
    [sym_pipeline] = STATE(1748),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(906),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(908),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(910),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(213)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(2112),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(912),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(914),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(916),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(214)] = {
    [sym_comparison_operator] = STATE(551),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(846),
    [anon_sym_BANG_EQ] = ACTIONS(846),
    [anon_sym_PLUS_EQ] = ACTIONS(846),
    [anon_sym_STAR_EQ] = ACTIONS(846),
    [anon_sym_SLASH_EQ] = ACTIONS(846),
    [anon_sym_PERCENT_EQ] = ACTIONS(846),
    [anon_sym_DASH_EQ] = ACTIONS(846),
    [anon_sym_GT] = ACTIONS(848),
    [anon_sym_GT_GT] = ACTIONS(846),
    [anon_sym_2_GT] = ACTIONS(848),
    [anon_sym_2_GT_GT] = ACTIONS(846),
    [anon_sym_3_GT] = ACTIONS(848),
    [anon_sym_3_GT_GT] = ACTIONS(846),
    [anon_sym_4_GT] = ACTIONS(848),
    [anon_sym_4_GT_GT] = ACTIONS(846),
    [anon_sym_5_GT] = ACTIONS(848),
    [anon_sym_5_GT_GT] = ACTIONS(846),
    [anon_sym_6_GT] = ACTIONS(848),
    [anon_sym_6_GT_GT] = ACTIONS(846),
    [anon_sym_STAR_GT] = ACTIONS(848),
    [anon_sym_STAR_GT_GT] = ACTIONS(846),
    [anon_sym_LT] = ACTIONS(848),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(846),
    [anon_sym_2_GT_AMP1] = ACTIONS(846),
    [anon_sym_3_GT_AMP1] = ACTIONS(846),
    [anon_sym_4_GT_AMP1] = ACTIONS(846),
    [anon_sym_5_GT_AMP1] = ACTIONS(846),
    [anon_sym_6_GT_AMP1] = ACTIONS(846),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(846),
    [anon_sym_1_GT_AMP2] = ACTIONS(846),
    [anon_sym_3_GT_AMP2] = ACTIONS(846),
    [anon_sym_4_GT_AMP2] = ACTIONS(846),
    [anon_sym_5_GT_AMP2] = ACTIONS(846),
    [anon_sym_6_GT_AMP2] = ACTIONS(846),
    [aux_sym_comparison_operator_token1] = ACTIONS(850),
    [aux_sym_comparison_operator_token2] = ACTIONS(850),
    [aux_sym_comparison_operator_token3] = ACTIONS(850),
    [aux_sym_comparison_operator_token4] = ACTIONS(850),
    [aux_sym_comparison_operator_token5] = ACTIONS(850),
    [aux_sym_comparison_operator_token6] = ACTIONS(850),
    [aux_sym_comparison_operator_token7] = ACTIONS(850),
    [aux_sym_comparison_operator_token8] = ACTIONS(850),
    [aux_sym_comparison_operator_token9] = ACTIONS(850),
    [aux_sym_comparison_operator_token10] = ACTIONS(850),
    [aux_sym_comparison_operator_token11] = ACTIONS(850),
    [aux_sym_comparison_operator_token12] = ACTIONS(850),
    [aux_sym_comparison_operator_token13] = ACTIONS(850),
    [aux_sym_comparison_operator_token14] = ACTIONS(850),
    [aux_sym_comparison_operator_token15] = ACTIONS(850),
    [aux_sym_comparison_operator_token16] = ACTIONS(850),
    [aux_sym_comparison_operator_token17] = ACTIONS(850),
    [aux_sym_comparison_operator_token18] = ACTIONS(850),
    [aux_sym_comparison_operator_token19] = ACTIONS(850),
    [aux_sym_comparison_operator_token20] = ACTIONS(850),
    [aux_sym_comparison_operator_token21] = ACTIONS(850),
    [aux_sym_comparison_operator_token22] = ACTIONS(850),
    [aux_sym_comparison_operator_token23] = ACTIONS(850),
    [aux_sym_comparison_operator_token24] = ACTIONS(850),
    [aux_sym_comparison_operator_token25] = ACTIONS(850),
    [aux_sym_comparison_operator_token26] = ACTIONS(850),
    [aux_sym_comparison_operator_token27] = ACTIONS(850),
    [aux_sym_comparison_operator_token28] = ACTIONS(852),
    [aux_sym_comparison_operator_token29] = ACTIONS(850),
    [aux_sym_comparison_operator_token30] = ACTIONS(850),
    [aux_sym_comparison_operator_token31] = ACTIONS(850),
    [aux_sym_comparison_operator_token32] = ACTIONS(850),
    [aux_sym_comparison_operator_token33] = ACTIONS(850),
    [aux_sym_comparison_operator_token34] = ACTIONS(852),
    [aux_sym_comparison_operator_token35] = ACTIONS(850),
    [aux_sym_comparison_operator_token36] = ACTIONS(850),
    [aux_sym_comparison_operator_token37] = ACTIONS(850),
    [aux_sym_comparison_operator_token38] = ACTIONS(850),
    [aux_sym_comparison_operator_token39] = ACTIONS(850),
    [aux_sym_comparison_operator_token40] = ACTIONS(850),
    [aux_sym_comparison_operator_token41] = ACTIONS(850),
    [aux_sym_comparison_operator_token42] = ACTIONS(850),
    [aux_sym_comparison_operator_token43] = ACTIONS(850),
    [aux_sym_comparison_operator_token44] = ACTIONS(850),
    [aux_sym_comparison_operator_token45] = ACTIONS(850),
    [aux_sym_comparison_operator_token46] = ACTIONS(850),
    [aux_sym_comparison_operator_token47] = ACTIONS(850),
    [aux_sym_comparison_operator_token48] = ACTIONS(850),
    [aux_sym_comparison_operator_token49] = ACTIONS(850),
    [aux_sym_comparison_operator_token50] = ACTIONS(850),
    [anon_sym_RPAREN] = ACTIONS(846),
    [anon_sym_AMP_AMP] = ACTIONS(846),
    [anon_sym_PIPE_PIPE] = ACTIONS(846),
    [anon_sym_PIPE] = ACTIONS(848),
    [aux_sym_logical_expression_token1] = ACTIONS(846),
    [aux_sym_logical_expression_token2] = ACTIONS(846),
    [aux_sym_logical_expression_token3] = ACTIONS(846),
    [aux_sym_bitwise_expression_token1] = ACTIONS(846),
    [aux_sym_bitwise_expression_token2] = ACTIONS(846),
    [aux_sym_bitwise_expression_token3] = ACTIONS(846),
  },
  [STATE(215)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [aux_sym_comparison_operator_token1] = ACTIONS(540),
    [aux_sym_comparison_operator_token2] = ACTIONS(540),
    [aux_sym_comparison_operator_token3] = ACTIONS(540),
    [aux_sym_comparison_operator_token4] = ACTIONS(540),
    [aux_sym_comparison_operator_token5] = ACTIONS(540),
    [aux_sym_comparison_operator_token6] = ACTIONS(540),
    [aux_sym_comparison_operator_token7] = ACTIONS(540),
    [aux_sym_comparison_operator_token8] = ACTIONS(540),
    [aux_sym_comparison_operator_token9] = ACTIONS(540),
    [aux_sym_comparison_operator_token10] = ACTIONS(540),
    [aux_sym_comparison_operator_token11] = ACTIONS(540),
    [aux_sym_comparison_operator_token12] = ACTIONS(540),
    [aux_sym_comparison_operator_token13] = ACTIONS(540),
    [aux_sym_comparison_operator_token14] = ACTIONS(540),
    [aux_sym_comparison_operator_token15] = ACTIONS(540),
    [aux_sym_comparison_operator_token16] = ACTIONS(540),
    [aux_sym_comparison_operator_token17] = ACTIONS(540),
    [aux_sym_comparison_operator_token18] = ACTIONS(540),
    [aux_sym_comparison_operator_token19] = ACTIONS(540),
    [aux_sym_comparison_operator_token20] = ACTIONS(540),
    [aux_sym_comparison_operator_token21] = ACTIONS(540),
    [aux_sym_comparison_operator_token22] = ACTIONS(540),
    [aux_sym_comparison_operator_token23] = ACTIONS(540),
    [aux_sym_comparison_operator_token24] = ACTIONS(540),
    [aux_sym_comparison_operator_token25] = ACTIONS(540),
    [aux_sym_comparison_operator_token26] = ACTIONS(540),
    [aux_sym_comparison_operator_token27] = ACTIONS(540),
    [aux_sym_comparison_operator_token28] = ACTIONS(538),
    [aux_sym_comparison_operator_token29] = ACTIONS(540),
    [aux_sym_comparison_operator_token30] = ACTIONS(540),
    [aux_sym_comparison_operator_token31] = ACTIONS(540),
    [aux_sym_comparison_operator_token32] = ACTIONS(540),
    [aux_sym_comparison_operator_token33] = ACTIONS(540),
    [aux_sym_comparison_operator_token34] = ACTIONS(538),
    [aux_sym_comparison_operator_token35] = ACTIONS(540),
    [aux_sym_comparison_operator_token36] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token38] = ACTIONS(540),
    [aux_sym_comparison_operator_token39] = ACTIONS(540),
    [aux_sym_comparison_operator_token40] = ACTIONS(540),
    [aux_sym_comparison_operator_token41] = ACTIONS(540),
    [aux_sym_comparison_operator_token42] = ACTIONS(540),
    [aux_sym_comparison_operator_token43] = ACTIONS(540),
    [aux_sym_comparison_operator_token44] = ACTIONS(540),
    [aux_sym_comparison_operator_token45] = ACTIONS(540),
    [aux_sym_comparison_operator_token46] = ACTIONS(540),
    [aux_sym_comparison_operator_token47] = ACTIONS(540),
    [aux_sym_comparison_operator_token48] = ACTIONS(540),
    [aux_sym_comparison_operator_token49] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [aux_sym_format_operator_token1] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_RPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [aux_sym_logical_expression_token1] = ACTIONS(540),
    [aux_sym_logical_expression_token2] = ACTIONS(540),
    [aux_sym_logical_expression_token3] = ACTIONS(540),
    [aux_sym_bitwise_expression_token1] = ACTIONS(540),
    [aux_sym_bitwise_expression_token2] = ACTIONS(540),
    [aux_sym_bitwise_expression_token3] = ACTIONS(540),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [anon_sym_SLASH] = ACTIONS(540),
    [anon_sym_BSLASH] = ACTIONS(540),
    [anon_sym_PERCENT] = ACTIONS(540),
    [anon_sym_STAR] = ACTIONS(540),
    [anon_sym_DOT_DOT] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [anon_sym_RBRACK] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
  },
  [STATE(216)] = {
    [sym_comparison_operator] = STATE(551),
    [sym_comment] = ACTIONS(79),
    [anon_sym_EQ] = ACTIONS(902),
    [anon_sym_BANG_EQ] = ACTIONS(902),
    [anon_sym_PLUS_EQ] = ACTIONS(902),
    [anon_sym_STAR_EQ] = ACTIONS(902),
    [anon_sym_SLASH_EQ] = ACTIONS(902),
    [anon_sym_PERCENT_EQ] = ACTIONS(902),
    [anon_sym_DASH_EQ] = ACTIONS(902),
    [anon_sym_GT] = ACTIONS(904),
    [anon_sym_GT_GT] = ACTIONS(902),
    [anon_sym_2_GT] = ACTIONS(904),
    [anon_sym_2_GT_GT] = ACTIONS(902),
    [anon_sym_3_GT] = ACTIONS(904),
    [anon_sym_3_GT_GT] = ACTIONS(902),
    [anon_sym_4_GT] = ACTIONS(904),
    [anon_sym_4_GT_GT] = ACTIONS(902),
    [anon_sym_5_GT] = ACTIONS(904),
    [anon_sym_5_GT_GT] = ACTIONS(902),
    [anon_sym_6_GT] = ACTIONS(904),
    [anon_sym_6_GT_GT] = ACTIONS(902),
    [anon_sym_STAR_GT] = ACTIONS(904),
    [anon_sym_STAR_GT_GT] = ACTIONS(902),
    [anon_sym_LT] = ACTIONS(904),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(902),
    [anon_sym_2_GT_AMP1] = ACTIONS(902),
    [anon_sym_3_GT_AMP1] = ACTIONS(902),
    [anon_sym_4_GT_AMP1] = ACTIONS(902),
    [anon_sym_5_GT_AMP1] = ACTIONS(902),
    [anon_sym_6_GT_AMP1] = ACTIONS(902),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(902),
    [anon_sym_1_GT_AMP2] = ACTIONS(902),
    [anon_sym_3_GT_AMP2] = ACTIONS(902),
    [anon_sym_4_GT_AMP2] = ACTIONS(902),
    [anon_sym_5_GT_AMP2] = ACTIONS(902),
    [anon_sym_6_GT_AMP2] = ACTIONS(902),
    [aux_sym_comparison_operator_token1] = ACTIONS(850),
    [aux_sym_comparison_operator_token2] = ACTIONS(850),
    [aux_sym_comparison_operator_token3] = ACTIONS(850),
    [aux_sym_comparison_operator_token4] = ACTIONS(850),
    [aux_sym_comparison_operator_token5] = ACTIONS(850),
    [aux_sym_comparison_operator_token6] = ACTIONS(850),
    [aux_sym_comparison_operator_token7] = ACTIONS(850),
    [aux_sym_comparison_operator_token8] = ACTIONS(850),
    [aux_sym_comparison_operator_token9] = ACTIONS(850),
    [aux_sym_comparison_operator_token10] = ACTIONS(850),
    [aux_sym_comparison_operator_token11] = ACTIONS(850),
    [aux_sym_comparison_operator_token12] = ACTIONS(850),
    [aux_sym_comparison_operator_token13] = ACTIONS(850),
    [aux_sym_comparison_operator_token14] = ACTIONS(850),
    [aux_sym_comparison_operator_token15] = ACTIONS(850),
    [aux_sym_comparison_operator_token16] = ACTIONS(850),
    [aux_sym_comparison_operator_token17] = ACTIONS(850),
    [aux_sym_comparison_operator_token18] = ACTIONS(850),
    [aux_sym_comparison_operator_token19] = ACTIONS(850),
    [aux_sym_comparison_operator_token20] = ACTIONS(850),
    [aux_sym_comparison_operator_token21] = ACTIONS(850),
    [aux_sym_comparison_operator_token22] = ACTIONS(850),
    [aux_sym_comparison_operator_token23] = ACTIONS(850),
    [aux_sym_comparison_operator_token24] = ACTIONS(850),
    [aux_sym_comparison_operator_token25] = ACTIONS(850),
    [aux_sym_comparison_operator_token26] = ACTIONS(850),
    [aux_sym_comparison_operator_token27] = ACTIONS(850),
    [aux_sym_comparison_operator_token28] = ACTIONS(852),
    [aux_sym_comparison_operator_token29] = ACTIONS(850),
    [aux_sym_comparison_operator_token30] = ACTIONS(850),
    [aux_sym_comparison_operator_token31] = ACTIONS(850),
    [aux_sym_comparison_operator_token32] = ACTIONS(850),
    [aux_sym_comparison_operator_token33] = ACTIONS(850),
    [aux_sym_comparison_operator_token34] = ACTIONS(852),
    [aux_sym_comparison_operator_token35] = ACTIONS(850),
    [aux_sym_comparison_operator_token36] = ACTIONS(850),
    [aux_sym_comparison_operator_token37] = ACTIONS(850),
    [aux_sym_comparison_operator_token38] = ACTIONS(850),
    [aux_sym_comparison_operator_token39] = ACTIONS(850),
    [aux_sym_comparison_operator_token40] = ACTIONS(850),
    [aux_sym_comparison_operator_token41] = ACTIONS(850),
    [aux_sym_comparison_operator_token42] = ACTIONS(850),
    [aux_sym_comparison_operator_token43] = ACTIONS(850),
    [aux_sym_comparison_operator_token44] = ACTIONS(850),
    [aux_sym_comparison_operator_token45] = ACTIONS(850),
    [aux_sym_comparison_operator_token46] = ACTIONS(850),
    [aux_sym_comparison_operator_token47] = ACTIONS(850),
    [aux_sym_comparison_operator_token48] = ACTIONS(850),
    [aux_sym_comparison_operator_token49] = ACTIONS(850),
    [aux_sym_comparison_operator_token50] = ACTIONS(850),
    [anon_sym_RPAREN] = ACTIONS(902),
    [anon_sym_AMP_AMP] = ACTIONS(902),
    [anon_sym_PIPE_PIPE] = ACTIONS(902),
    [anon_sym_PIPE] = ACTIONS(904),
    [aux_sym_logical_expression_token1] = ACTIONS(902),
    [aux_sym_logical_expression_token2] = ACTIONS(902),
    [aux_sym_logical_expression_token3] = ACTIONS(902),
    [aux_sym_bitwise_expression_token1] = ACTIONS(902),
    [aux_sym_bitwise_expression_token2] = ACTIONS(902),
    [aux_sym_bitwise_expression_token3] = ACTIONS(902),
  },
  [STATE(217)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_condition] = STATE(1935),
    [sym_pipeline] = STATE(1972),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_SEMI] = ACTIONS(918),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(920),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_for_statement_token2] = ACTIONS(922),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(218)] = {
    [sym__literal] = STATE(753),
    [sym_integer_literal] = STATE(753),
    [sym_string_literal] = STATE(753),
    [sym_expandable_string_literal] = STATE(820),
    [sym_expandable_here_string_literal] = STATE(820),
    [sym_variable] = STATE(753),
    [sym_unary_expression] = STATE(1023),
    [sym_expression_with_unary_operator] = STATE(1010),
    [sym_pre_increment_expression] = STATE(1011),
    [sym_pre_decrement_expression] = STATE(1011),
    [sym_cast_expression] = STATE(1011),
    [sym__primary_expression] = STATE(753),
    [sym__value] = STATE(753),
    [sym_parenthesized_expression] = STATE(753),
    [sym_sub_expression] = STATE(753),
    [sym_array_expression] = STATE(753),
    [sym_script_block_expression] = STATE(753),
    [sym_hash_literal_expression] = STATE(753),
    [sym_post_increment_expression] = STATE(753),
    [sym_post_decrement_expression] = STATE(753),
    [sym_member_access] = STATE(753),
    [sym_element_access] = STATE(753),
    [sym_invokation_expression] = STATE(753),
    [sym_invokation_foreach_expression] = STATE(754),
    [sym_type_literal] = STATE(218),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(924),
    [sym_hexadecimal_integer_literal] = ACTIONS(924),
    [sym_real_literal] = ACTIONS(926),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(928),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(930),
    [sym_verbatim_string_characters] = ACTIONS(932),
    [sym_verbatim_here_string_characters] = ACTIONS(932),
    [anon_sym_LBRACK] = ACTIONS(934),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(936),
    [aux_sym_comparison_operator_token50] = ACTIONS(936),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(938),
    [anon_sym_DOLLAR_CARET] = ACTIONS(938),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(938),
    [anon_sym_DOLLAR_] = ACTIONS(938),
    [aux_sym_variable_token1] = ACTIONS(938),
    [aux_sym_variable_token2] = ACTIONS(938),
    [sym_braced_variable] = ACTIONS(938),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(940),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_LBRACE] = ACTIONS(942),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(936),
    [anon_sym_DASH] = ACTIONS(936),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(936),
    [anon_sym_BANG] = ACTIONS(936),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(936),
    [anon_sym_PLUS_PLUS] = ACTIONS(944),
    [anon_sym_DASH_DASH] = ACTIONS(946),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(948),
    [anon_sym_AT_LPAREN] = ACTIONS(950),
    [anon_sym_AT_LBRACE] = ACTIONS(952),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(219)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [aux_sym_comparison_operator_token1] = ACTIONS(540),
    [aux_sym_comparison_operator_token2] = ACTIONS(540),
    [aux_sym_comparison_operator_token3] = ACTIONS(540),
    [aux_sym_comparison_operator_token4] = ACTIONS(540),
    [aux_sym_comparison_operator_token5] = ACTIONS(540),
    [aux_sym_comparison_operator_token6] = ACTIONS(540),
    [aux_sym_comparison_operator_token7] = ACTIONS(540),
    [aux_sym_comparison_operator_token8] = ACTIONS(540),
    [aux_sym_comparison_operator_token9] = ACTIONS(540),
    [aux_sym_comparison_operator_token10] = ACTIONS(540),
    [aux_sym_comparison_operator_token11] = ACTIONS(540),
    [aux_sym_comparison_operator_token12] = ACTIONS(540),
    [aux_sym_comparison_operator_token13] = ACTIONS(540),
    [aux_sym_comparison_operator_token14] = ACTIONS(540),
    [aux_sym_comparison_operator_token15] = ACTIONS(540),
    [aux_sym_comparison_operator_token16] = ACTIONS(540),
    [aux_sym_comparison_operator_token17] = ACTIONS(540),
    [aux_sym_comparison_operator_token18] = ACTIONS(540),
    [aux_sym_comparison_operator_token19] = ACTIONS(540),
    [aux_sym_comparison_operator_token20] = ACTIONS(540),
    [aux_sym_comparison_operator_token21] = ACTIONS(540),
    [aux_sym_comparison_operator_token22] = ACTIONS(540),
    [aux_sym_comparison_operator_token23] = ACTIONS(540),
    [aux_sym_comparison_operator_token24] = ACTIONS(540),
    [aux_sym_comparison_operator_token25] = ACTIONS(540),
    [aux_sym_comparison_operator_token26] = ACTIONS(540),
    [aux_sym_comparison_operator_token27] = ACTIONS(540),
    [aux_sym_comparison_operator_token28] = ACTIONS(538),
    [aux_sym_comparison_operator_token29] = ACTIONS(540),
    [aux_sym_comparison_operator_token30] = ACTIONS(540),
    [aux_sym_comparison_operator_token31] = ACTIONS(540),
    [aux_sym_comparison_operator_token32] = ACTIONS(540),
    [aux_sym_comparison_operator_token33] = ACTIONS(540),
    [aux_sym_comparison_operator_token34] = ACTIONS(538),
    [aux_sym_comparison_operator_token35] = ACTIONS(540),
    [aux_sym_comparison_operator_token36] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token38] = ACTIONS(540),
    [aux_sym_comparison_operator_token39] = ACTIONS(540),
    [aux_sym_comparison_operator_token40] = ACTIONS(540),
    [aux_sym_comparison_operator_token41] = ACTIONS(540),
    [aux_sym_comparison_operator_token42] = ACTIONS(540),
    [aux_sym_comparison_operator_token43] = ACTIONS(540),
    [aux_sym_comparison_operator_token44] = ACTIONS(540),
    [aux_sym_comparison_operator_token45] = ACTIONS(540),
    [aux_sym_comparison_operator_token46] = ACTIONS(540),
    [aux_sym_comparison_operator_token47] = ACTIONS(540),
    [aux_sym_comparison_operator_token48] = ACTIONS(540),
    [aux_sym_comparison_operator_token49] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [aux_sym_format_operator_token1] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [aux_sym_logical_expression_token1] = ACTIONS(540),
    [aux_sym_logical_expression_token2] = ACTIONS(540),
    [aux_sym_logical_expression_token3] = ACTIONS(540),
    [aux_sym_bitwise_expression_token1] = ACTIONS(540),
    [aux_sym_bitwise_expression_token2] = ACTIONS(540),
    [aux_sym_bitwise_expression_token3] = ACTIONS(540),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [anon_sym_SLASH] = ACTIONS(540),
    [anon_sym_BSLASH] = ACTIONS(540),
    [anon_sym_PERCENT] = ACTIONS(540),
    [anon_sym_STAR] = ACTIONS(540),
    [anon_sym_DOT_DOT] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
    [sym__statement_terminator] = ACTIONS(540),
  },
  [STATE(220)] = {
    [sym__literal] = STATE(794),
    [sym_integer_literal] = STATE(794),
    [sym_string_literal] = STATE(794),
    [sym_expandable_string_literal] = STATE(782),
    [sym_expandable_here_string_literal] = STATE(782),
    [sym_variable] = STATE(794),
    [sym_unary_expression] = STATE(815),
    [sym_expression_with_unary_operator] = STATE(779),
    [sym_pre_increment_expression] = STATE(811),
    [sym_pre_decrement_expression] = STATE(811),
    [sym_cast_expression] = STATE(811),
    [sym__primary_expression] = STATE(794),
    [sym__value] = STATE(794),
    [sym_parenthesized_expression] = STATE(794),
    [sym_sub_expression] = STATE(794),
    [sym_array_expression] = STATE(794),
    [sym_script_block_expression] = STATE(794),
    [sym_hash_literal_expression] = STATE(794),
    [sym_post_increment_expression] = STATE(794),
    [sym_post_decrement_expression] = STATE(794),
    [sym_member_access] = STATE(794),
    [sym_element_access] = STATE(794),
    [sym_invokation_expression] = STATE(794),
    [sym_invokation_foreach_expression] = STATE(796),
    [sym_type_literal] = STATE(220),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(954),
    [sym_hexadecimal_integer_literal] = ACTIONS(954),
    [sym_real_literal] = ACTIONS(956),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(958),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(960),
    [sym_verbatim_string_characters] = ACTIONS(962),
    [sym_verbatim_here_string_characters] = ACTIONS(962),
    [anon_sym_LBRACK] = ACTIONS(964),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(966),
    [aux_sym_comparison_operator_token50] = ACTIONS(966),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(968),
    [anon_sym_DOLLAR_CARET] = ACTIONS(968),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(968),
    [anon_sym_DOLLAR_] = ACTIONS(968),
    [aux_sym_variable_token1] = ACTIONS(968),
    [aux_sym_variable_token2] = ACTIONS(968),
    [sym_braced_variable] = ACTIONS(968),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(970),
    [anon_sym_RPAREN] = ACTIONS(95),
    [anon_sym_COMMA] = ACTIONS(966),
    [anon_sym_LBRACE] = ACTIONS(972),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(966),
    [anon_sym_DASH] = ACTIONS(966),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(966),
    [anon_sym_BANG] = ACTIONS(966),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(966),
    [anon_sym_PLUS_PLUS] = ACTIONS(974),
    [anon_sym_DASH_DASH] = ACTIONS(976),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(978),
    [anon_sym_AT_LPAREN] = ACTIONS(980),
    [anon_sym_AT_LBRACE] = ACTIONS(982),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(221)] = {
    [sym__literal] = STATE(753),
    [sym_integer_literal] = STATE(753),
    [sym_string_literal] = STATE(753),
    [sym_expandable_string_literal] = STATE(820),
    [sym_expandable_here_string_literal] = STATE(820),
    [sym_variable] = STATE(753),
    [sym_unary_expression] = STATE(808),
    [sym_expression_with_unary_operator] = STATE(810),
    [sym_pre_increment_expression] = STATE(803),
    [sym_pre_decrement_expression] = STATE(803),
    [sym_cast_expression] = STATE(803),
    [sym__primary_expression] = STATE(753),
    [sym__value] = STATE(753),
    [sym_parenthesized_expression] = STATE(753),
    [sym_sub_expression] = STATE(753),
    [sym_array_expression] = STATE(753),
    [sym_script_block_expression] = STATE(753),
    [sym_hash_literal_expression] = STATE(753),
    [sym_post_increment_expression] = STATE(753),
    [sym_post_decrement_expression] = STATE(753),
    [sym_member_access] = STATE(753),
    [sym_element_access] = STATE(753),
    [sym_invokation_expression] = STATE(753),
    [sym_invokation_foreach_expression] = STATE(754),
    [sym_type_literal] = STATE(221),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(924),
    [sym_hexadecimal_integer_literal] = ACTIONS(924),
    [sym_real_literal] = ACTIONS(926),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(928),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(930),
    [sym_verbatim_string_characters] = ACTIONS(932),
    [sym_verbatim_here_string_characters] = ACTIONS(932),
    [anon_sym_LBRACK] = ACTIONS(934),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(984),
    [aux_sym_comparison_operator_token50] = ACTIONS(984),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(938),
    [anon_sym_DOLLAR_CARET] = ACTIONS(938),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(938),
    [anon_sym_DOLLAR_] = ACTIONS(938),
    [aux_sym_variable_token1] = ACTIONS(938),
    [aux_sym_variable_token2] = ACTIONS(938),
    [sym_braced_variable] = ACTIONS(938),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(940),
    [anon_sym_COMMA] = ACTIONS(984),
    [anon_sym_LBRACE] = ACTIONS(942),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(984),
    [anon_sym_DASH] = ACTIONS(984),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(984),
    [anon_sym_BANG] = ACTIONS(984),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(984),
    [anon_sym_PLUS_PLUS] = ACTIONS(986),
    [anon_sym_DASH_DASH] = ACTIONS(988),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(948),
    [anon_sym_AT_LPAREN] = ACTIONS(950),
    [anon_sym_AT_LBRACE] = ACTIONS(952),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(222)] = {
    [sym__literal] = STATE(794),
    [sym_integer_literal] = STATE(794),
    [sym_string_literal] = STATE(794),
    [sym_expandable_string_literal] = STATE(782),
    [sym_expandable_here_string_literal] = STATE(782),
    [sym_variable] = STATE(794),
    [sym_unary_expression] = STATE(992),
    [sym_expression_with_unary_operator] = STATE(998),
    [sym_pre_increment_expression] = STATE(982),
    [sym_pre_decrement_expression] = STATE(982),
    [sym_cast_expression] = STATE(982),
    [sym__primary_expression] = STATE(794),
    [sym__value] = STATE(794),
    [sym_parenthesized_expression] = STATE(794),
    [sym_sub_expression] = STATE(794),
    [sym_array_expression] = STATE(794),
    [sym_script_block_expression] = STATE(794),
    [sym_hash_literal_expression] = STATE(794),
    [sym_post_increment_expression] = STATE(794),
    [sym_post_decrement_expression] = STATE(794),
    [sym_member_access] = STATE(794),
    [sym_element_access] = STATE(794),
    [sym_invokation_expression] = STATE(794),
    [sym_invokation_foreach_expression] = STATE(796),
    [sym_type_literal] = STATE(222),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(954),
    [sym_hexadecimal_integer_literal] = ACTIONS(954),
    [sym_real_literal] = ACTIONS(956),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(958),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(960),
    [sym_verbatim_string_characters] = ACTIONS(962),
    [sym_verbatim_here_string_characters] = ACTIONS(962),
    [anon_sym_LBRACK] = ACTIONS(964),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(990),
    [aux_sym_comparison_operator_token50] = ACTIONS(990),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(968),
    [anon_sym_DOLLAR_CARET] = ACTIONS(968),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(968),
    [anon_sym_DOLLAR_] = ACTIONS(968),
    [aux_sym_variable_token1] = ACTIONS(968),
    [aux_sym_variable_token2] = ACTIONS(968),
    [sym_braced_variable] = ACTIONS(968),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(970),
    [anon_sym_RPAREN] = ACTIONS(95),
    [anon_sym_COMMA] = ACTIONS(990),
    [anon_sym_LBRACE] = ACTIONS(972),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(990),
    [anon_sym_DASH] = ACTIONS(990),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(990),
    [anon_sym_BANG] = ACTIONS(990),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(990),
    [anon_sym_PLUS_PLUS] = ACTIONS(992),
    [anon_sym_DASH_DASH] = ACTIONS(994),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(978),
    [anon_sym_AT_LPAREN] = ACTIONS(980),
    [anon_sym_AT_LBRACE] = ACTIONS(982),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(223)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2054),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(996),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(224)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2050),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(998),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(225)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2091),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1000),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(226)] = {
    [sym__literal] = STATE(753),
    [sym_integer_literal] = STATE(753),
    [sym_string_literal] = STATE(753),
    [sym_expandable_string_literal] = STATE(820),
    [sym_expandable_here_string_literal] = STATE(820),
    [sym_variable] = STATE(753),
    [sym_array_literal_expression] = STATE(1044),
    [sym_unary_expression] = STATE(977),
    [sym_expression_with_unary_operator] = STATE(1010),
    [sym_pre_increment_expression] = STATE(1011),
    [sym_pre_decrement_expression] = STATE(1011),
    [sym_cast_expression] = STATE(1011),
    [sym__primary_expression] = STATE(753),
    [sym__value] = STATE(753),
    [sym_parenthesized_expression] = STATE(753),
    [sym_sub_expression] = STATE(753),
    [sym_array_expression] = STATE(753),
    [sym_script_block_expression] = STATE(753),
    [sym_hash_literal_expression] = STATE(753),
    [sym_post_increment_expression] = STATE(753),
    [sym_post_decrement_expression] = STATE(753),
    [sym_member_access] = STATE(753),
    [sym_element_access] = STATE(753),
    [sym_invokation_expression] = STATE(753),
    [sym_invokation_foreach_expression] = STATE(754),
    [sym_type_literal] = STATE(218),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(924),
    [sym_hexadecimal_integer_literal] = ACTIONS(924),
    [sym_real_literal] = ACTIONS(926),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(928),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(930),
    [sym_verbatim_string_characters] = ACTIONS(932),
    [sym_verbatim_here_string_characters] = ACTIONS(932),
    [anon_sym_LBRACK] = ACTIONS(934),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(936),
    [aux_sym_comparison_operator_token50] = ACTIONS(936),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(938),
    [anon_sym_DOLLAR_CARET] = ACTIONS(938),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(938),
    [anon_sym_DOLLAR_] = ACTIONS(938),
    [aux_sym_variable_token1] = ACTIONS(938),
    [aux_sym_variable_token2] = ACTIONS(938),
    [sym_braced_variable] = ACTIONS(938),
    [sym_generic_token] = ACTIONS(1004),
    [sym_command_parameter] = ACTIONS(1002),
    [anon_sym_LPAREN] = ACTIONS(940),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_LBRACE] = ACTIONS(942),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [sym_stop_parsing] = ACTIONS(1002),
    [anon_sym_SPACE] = ACTIONS(1002),
    [anon_sym_COLON] = ACTIONS(1002),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(936),
    [anon_sym_DASH] = ACTIONS(936),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(936),
    [anon_sym_BANG] = ACTIONS(936),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(936),
    [anon_sym_PLUS_PLUS] = ACTIONS(944),
    [anon_sym_DASH_DASH] = ACTIONS(946),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(948),
    [anon_sym_AT_LPAREN] = ACTIONS(950),
    [anon_sym_AT_LBRACE] = ACTIONS(952),
    [sym__statement_terminator] = ACTIONS(1006),
  },
  [STATE(227)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2087),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1008),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(228)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(1976),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1010),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(229)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(1745),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1012),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(230)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2092),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1014),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(231)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2090),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1016),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(232)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(1807),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1018),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(233)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(1751),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1020),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(234)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2053),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1022),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(235)] = {
    [sym__literal] = STATE(835),
    [sym_integer_literal] = STATE(835),
    [sym_string_literal] = STATE(835),
    [sym_expandable_string_literal] = STATE(830),
    [sym_expandable_here_string_literal] = STATE(830),
    [sym_variable] = STATE(835),
    [sym_unary_expression] = STATE(1080),
    [sym_expression_with_unary_operator] = STATE(1064),
    [sym_pre_increment_expression] = STATE(1076),
    [sym_pre_decrement_expression] = STATE(1076),
    [sym_cast_expression] = STATE(1076),
    [sym__primary_expression] = STATE(835),
    [sym__value] = STATE(835),
    [sym_parenthesized_expression] = STATE(835),
    [sym_sub_expression] = STATE(835),
    [sym_array_expression] = STATE(835),
    [sym_script_block_expression] = STATE(835),
    [sym_hash_literal_expression] = STATE(835),
    [sym_post_increment_expression] = STATE(835),
    [sym_post_decrement_expression] = STATE(835),
    [sym_member_access] = STATE(835),
    [sym_element_access] = STATE(835),
    [sym_invokation_expression] = STATE(835),
    [sym_invokation_foreach_expression] = STATE(837),
    [sym_type_literal] = STATE(235),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1024),
    [sym_hexadecimal_integer_literal] = ACTIONS(1024),
    [sym_real_literal] = ACTIONS(1026),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1028),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1030),
    [sym_verbatim_string_characters] = ACTIONS(1032),
    [sym_verbatim_here_string_characters] = ACTIONS(1032),
    [anon_sym_LBRACK] = ACTIONS(1034),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(1036),
    [aux_sym_comparison_operator_token50] = ACTIONS(1036),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1038),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1038),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1038),
    [anon_sym_DOLLAR_] = ACTIONS(1038),
    [aux_sym_variable_token1] = ACTIONS(1038),
    [aux_sym_variable_token2] = ACTIONS(1038),
    [sym_braced_variable] = ACTIONS(1038),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(1040),
    [anon_sym_RPAREN] = ACTIONS(95),
    [anon_sym_COMMA] = ACTIONS(1036),
    [anon_sym_LBRACE] = ACTIONS(1042),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1036),
    [anon_sym_DASH] = ACTIONS(1036),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1036),
    [anon_sym_BANG] = ACTIONS(1036),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1036),
    [anon_sym_PLUS_PLUS] = ACTIONS(1044),
    [anon_sym_DASH_DASH] = ACTIONS(1046),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1048),
    [anon_sym_AT_LPAREN] = ACTIONS(1050),
    [anon_sym_AT_LBRACE] = ACTIONS(1052),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(236)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2055),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(237)] = {
    [sym__literal] = STATE(859),
    [sym_integer_literal] = STATE(859),
    [sym_string_literal] = STATE(859),
    [sym_expandable_string_literal] = STATE(842),
    [sym_expandable_here_string_literal] = STATE(842),
    [sym_variable] = STATE(859),
    [sym_unary_expression] = STATE(1070),
    [sym_expression_with_unary_operator] = STATE(1058),
    [sym_pre_increment_expression] = STATE(1065),
    [sym_pre_decrement_expression] = STATE(1065),
    [sym_cast_expression] = STATE(1065),
    [sym__primary_expression] = STATE(859),
    [sym__value] = STATE(859),
    [sym_parenthesized_expression] = STATE(859),
    [sym_sub_expression] = STATE(859),
    [sym_array_expression] = STATE(859),
    [sym_script_block_expression] = STATE(859),
    [sym_hash_literal_expression] = STATE(859),
    [sym_post_increment_expression] = STATE(859),
    [sym_post_decrement_expression] = STATE(859),
    [sym_member_access] = STATE(859),
    [sym_element_access] = STATE(859),
    [sym_invokation_expression] = STATE(859),
    [sym_invokation_foreach_expression] = STATE(868),
    [sym_type_literal] = STATE(237),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1056),
    [sym_hexadecimal_integer_literal] = ACTIONS(1056),
    [sym_real_literal] = ACTIONS(1058),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1060),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1062),
    [sym_verbatim_string_characters] = ACTIONS(1064),
    [sym_verbatim_here_string_characters] = ACTIONS(1064),
    [anon_sym_LBRACK] = ACTIONS(1066),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(1068),
    [aux_sym_comparison_operator_token50] = ACTIONS(1068),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1070),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1070),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1070),
    [anon_sym_DOLLAR_] = ACTIONS(1070),
    [aux_sym_variable_token1] = ACTIONS(1070),
    [aux_sym_variable_token2] = ACTIONS(1070),
    [sym_braced_variable] = ACTIONS(1070),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1068),
    [anon_sym_LBRACE] = ACTIONS(1074),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1068),
    [anon_sym_DASH] = ACTIONS(1068),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1068),
    [anon_sym_BANG] = ACTIONS(1068),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1068),
    [anon_sym_PLUS_PLUS] = ACTIONS(1076),
    [anon_sym_DASH_DASH] = ACTIONS(1078),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1080),
    [anon_sym_AT_LPAREN] = ACTIONS(1082),
    [anon_sym_AT_LBRACE] = ACTIONS(1084),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(238)] = {
    [sym__literal] = STATE(794),
    [sym_integer_literal] = STATE(794),
    [sym_string_literal] = STATE(794),
    [sym_expandable_string_literal] = STATE(782),
    [sym_expandable_here_string_literal] = STATE(782),
    [sym_variable] = STATE(794),
    [sym_array_literal_expression] = STATE(1047),
    [sym_unary_expression] = STATE(974),
    [sym_expression_with_unary_operator] = STATE(998),
    [sym_pre_increment_expression] = STATE(982),
    [sym_pre_decrement_expression] = STATE(982),
    [sym_cast_expression] = STATE(982),
    [sym__primary_expression] = STATE(794),
    [sym__value] = STATE(794),
    [sym_parenthesized_expression] = STATE(794),
    [sym_sub_expression] = STATE(794),
    [sym_array_expression] = STATE(794),
    [sym_script_block_expression] = STATE(794),
    [sym_hash_literal_expression] = STATE(794),
    [sym_post_increment_expression] = STATE(794),
    [sym_post_decrement_expression] = STATE(794),
    [sym_member_access] = STATE(794),
    [sym_element_access] = STATE(794),
    [sym_invokation_expression] = STATE(794),
    [sym_invokation_foreach_expression] = STATE(796),
    [sym_type_literal] = STATE(222),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(954),
    [sym_hexadecimal_integer_literal] = ACTIONS(954),
    [sym_real_literal] = ACTIONS(956),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(958),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(960),
    [sym_verbatim_string_characters] = ACTIONS(962),
    [sym_verbatim_here_string_characters] = ACTIONS(962),
    [anon_sym_LBRACK] = ACTIONS(964),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(990),
    [aux_sym_comparison_operator_token50] = ACTIONS(990),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(968),
    [anon_sym_DOLLAR_CARET] = ACTIONS(968),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(968),
    [anon_sym_DOLLAR_] = ACTIONS(968),
    [aux_sym_variable_token1] = ACTIONS(968),
    [aux_sym_variable_token2] = ACTIONS(968),
    [sym_braced_variable] = ACTIONS(968),
    [sym_generic_token] = ACTIONS(1086),
    [sym_command_parameter] = ACTIONS(1002),
    [anon_sym_LPAREN] = ACTIONS(970),
    [anon_sym_RPAREN] = ACTIONS(1002),
    [anon_sym_COMMA] = ACTIONS(990),
    [anon_sym_LBRACE] = ACTIONS(972),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [sym_stop_parsing] = ACTIONS(1002),
    [anon_sym_SPACE] = ACTIONS(1002),
    [anon_sym_COLON] = ACTIONS(1002),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(990),
    [anon_sym_DASH] = ACTIONS(990),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(990),
    [anon_sym_BANG] = ACTIONS(990),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(990),
    [anon_sym_PLUS_PLUS] = ACTIONS(992),
    [anon_sym_DASH_DASH] = ACTIONS(994),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(978),
    [anon_sym_AT_LPAREN] = ACTIONS(980),
    [anon_sym_AT_LBRACE] = ACTIONS(982),
  },
  [STATE(239)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2073),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1088),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(240)] = {
    [sym__literal] = STATE(859),
    [sym_integer_literal] = STATE(859),
    [sym_string_literal] = STATE(859),
    [sym_expandable_string_literal] = STATE(842),
    [sym_expandable_here_string_literal] = STATE(842),
    [sym_variable] = STATE(859),
    [sym_unary_expression] = STATE(825),
    [sym_expression_with_unary_operator] = STATE(874),
    [sym_pre_increment_expression] = STATE(847),
    [sym_pre_decrement_expression] = STATE(847),
    [sym_cast_expression] = STATE(847),
    [sym__primary_expression] = STATE(859),
    [sym__value] = STATE(859),
    [sym_parenthesized_expression] = STATE(859),
    [sym_sub_expression] = STATE(859),
    [sym_array_expression] = STATE(859),
    [sym_script_block_expression] = STATE(859),
    [sym_hash_literal_expression] = STATE(859),
    [sym_post_increment_expression] = STATE(859),
    [sym_post_decrement_expression] = STATE(859),
    [sym_member_access] = STATE(859),
    [sym_element_access] = STATE(859),
    [sym_invokation_expression] = STATE(859),
    [sym_invokation_foreach_expression] = STATE(868),
    [sym_type_literal] = STATE(240),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1056),
    [sym_hexadecimal_integer_literal] = ACTIONS(1056),
    [sym_real_literal] = ACTIONS(1058),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1060),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1062),
    [sym_verbatim_string_characters] = ACTIONS(1064),
    [sym_verbatim_here_string_characters] = ACTIONS(1064),
    [anon_sym_LBRACK] = ACTIONS(1066),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(1090),
    [aux_sym_comparison_operator_token50] = ACTIONS(1090),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1070),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1070),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1070),
    [anon_sym_DOLLAR_] = ACTIONS(1070),
    [aux_sym_variable_token1] = ACTIONS(1070),
    [aux_sym_variable_token2] = ACTIONS(1070),
    [sym_braced_variable] = ACTIONS(1070),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1090),
    [anon_sym_LBRACE] = ACTIONS(1074),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1090),
    [anon_sym_DASH] = ACTIONS(1090),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1090),
    [anon_sym_BANG] = ACTIONS(1090),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1090),
    [anon_sym_PLUS_PLUS] = ACTIONS(1092),
    [anon_sym_DASH_DASH] = ACTIONS(1094),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1080),
    [anon_sym_AT_LPAREN] = ACTIONS(1082),
    [anon_sym_AT_LBRACE] = ACTIONS(1084),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(241)] = {
    [sym__literal] = STATE(835),
    [sym_integer_literal] = STATE(835),
    [sym_string_literal] = STATE(835),
    [sym_expandable_string_literal] = STATE(830),
    [sym_expandable_here_string_literal] = STATE(830),
    [sym_variable] = STATE(835),
    [sym_unary_expression] = STATE(867),
    [sym_expression_with_unary_operator] = STATE(831),
    [sym_pre_increment_expression] = STATE(839),
    [sym_pre_decrement_expression] = STATE(839),
    [sym_cast_expression] = STATE(839),
    [sym__primary_expression] = STATE(835),
    [sym__value] = STATE(835),
    [sym_parenthesized_expression] = STATE(835),
    [sym_sub_expression] = STATE(835),
    [sym_array_expression] = STATE(835),
    [sym_script_block_expression] = STATE(835),
    [sym_hash_literal_expression] = STATE(835),
    [sym_post_increment_expression] = STATE(835),
    [sym_post_decrement_expression] = STATE(835),
    [sym_member_access] = STATE(835),
    [sym_element_access] = STATE(835),
    [sym_invokation_expression] = STATE(835),
    [sym_invokation_foreach_expression] = STATE(837),
    [sym_type_literal] = STATE(241),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1024),
    [sym_hexadecimal_integer_literal] = ACTIONS(1024),
    [sym_real_literal] = ACTIONS(1026),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1028),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1030),
    [sym_verbatim_string_characters] = ACTIONS(1032),
    [sym_verbatim_here_string_characters] = ACTIONS(1032),
    [anon_sym_LBRACK] = ACTIONS(1034),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(95),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(95),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(95),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(95),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(95),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(95),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(95),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(95),
    [anon_sym_2_GT_AMP1] = ACTIONS(95),
    [anon_sym_3_GT_AMP1] = ACTIONS(95),
    [anon_sym_4_GT_AMP1] = ACTIONS(95),
    [anon_sym_5_GT_AMP1] = ACTIONS(95),
    [anon_sym_6_GT_AMP1] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(95),
    [anon_sym_1_GT_AMP2] = ACTIONS(95),
    [anon_sym_3_GT_AMP2] = ACTIONS(95),
    [anon_sym_4_GT_AMP2] = ACTIONS(95),
    [anon_sym_5_GT_AMP2] = ACTIONS(95),
    [anon_sym_6_GT_AMP2] = ACTIONS(95),
    [aux_sym_comparison_operator_token37] = ACTIONS(1096),
    [aux_sym_comparison_operator_token50] = ACTIONS(1096),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1038),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1038),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1038),
    [anon_sym_DOLLAR_] = ACTIONS(1038),
    [aux_sym_variable_token1] = ACTIONS(1038),
    [aux_sym_variable_token2] = ACTIONS(1038),
    [sym_braced_variable] = ACTIONS(1038),
    [sym_command_parameter] = ACTIONS(95),
    [anon_sym_LPAREN] = ACTIONS(1040),
    [anon_sym_RPAREN] = ACTIONS(95),
    [anon_sym_COMMA] = ACTIONS(1096),
    [anon_sym_LBRACE] = ACTIONS(1042),
    [anon_sym_AMP_AMP] = ACTIONS(95),
    [anon_sym_PIPE_PIPE] = ACTIONS(95),
    [anon_sym_PIPE] = ACTIONS(95),
    [sym_stop_parsing] = ACTIONS(95),
    [anon_sym_SPACE] = ACTIONS(93),
    [anon_sym_COLON] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1096),
    [anon_sym_DASH] = ACTIONS(1096),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1096),
    [anon_sym_BANG] = ACTIONS(1096),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1096),
    [anon_sym_PLUS_PLUS] = ACTIONS(1098),
    [anon_sym_DASH_DASH] = ACTIONS(1100),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1048),
    [anon_sym_AT_LPAREN] = ACTIONS(1050),
    [anon_sym_AT_LBRACE] = ACTIONS(1052),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(95),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(242)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2069),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(243)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2072),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1104),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(244)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_for_iterator] = STATE(2074),
    [sym_pipeline] = STATE(1920),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_RPAREN] = ACTIONS(1106),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(245)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1709),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(246)] = {
    [sym__literal] = STATE(835),
    [sym_integer_literal] = STATE(835),
    [sym_string_literal] = STATE(835),
    [sym_expandable_string_literal] = STATE(830),
    [sym_expandable_here_string_literal] = STATE(830),
    [sym_variable] = STATE(835),
    [sym_array_literal_expression] = STATE(1102),
    [sym_unary_expression] = STATE(1039),
    [sym_expression_with_unary_operator] = STATE(1064),
    [sym_pre_increment_expression] = STATE(1076),
    [sym_pre_decrement_expression] = STATE(1076),
    [sym_cast_expression] = STATE(1076),
    [sym__primary_expression] = STATE(835),
    [sym__value] = STATE(835),
    [sym_parenthesized_expression] = STATE(835),
    [sym_sub_expression] = STATE(835),
    [sym_array_expression] = STATE(835),
    [sym_script_block_expression] = STATE(835),
    [sym_hash_literal_expression] = STATE(835),
    [sym_post_increment_expression] = STATE(835),
    [sym_post_decrement_expression] = STATE(835),
    [sym_member_access] = STATE(835),
    [sym_element_access] = STATE(835),
    [sym_invokation_expression] = STATE(835),
    [sym_invokation_foreach_expression] = STATE(837),
    [sym_type_literal] = STATE(235),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1024),
    [sym_hexadecimal_integer_literal] = ACTIONS(1024),
    [sym_real_literal] = ACTIONS(1026),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1028),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1030),
    [sym_verbatim_string_characters] = ACTIONS(1032),
    [sym_verbatim_here_string_characters] = ACTIONS(1032),
    [anon_sym_LBRACK] = ACTIONS(1034),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(1036),
    [aux_sym_comparison_operator_token50] = ACTIONS(1036),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1038),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1038),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1038),
    [anon_sym_DOLLAR_] = ACTIONS(1038),
    [aux_sym_variable_token1] = ACTIONS(1038),
    [aux_sym_variable_token2] = ACTIONS(1038),
    [sym_braced_variable] = ACTIONS(1038),
    [sym_generic_token] = ACTIONS(1134),
    [sym_command_parameter] = ACTIONS(1002),
    [anon_sym_LPAREN] = ACTIONS(1040),
    [anon_sym_RPAREN] = ACTIONS(1002),
    [anon_sym_COMMA] = ACTIONS(1036),
    [anon_sym_LBRACE] = ACTIONS(1042),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [sym_stop_parsing] = ACTIONS(1002),
    [anon_sym_SPACE] = ACTIONS(1002),
    [anon_sym_COLON] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(1036),
    [anon_sym_DASH] = ACTIONS(1036),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1036),
    [anon_sym_BANG] = ACTIONS(1036),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1036),
    [anon_sym_PLUS_PLUS] = ACTIONS(1044),
    [anon_sym_DASH_DASH] = ACTIONS(1046),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1048),
    [anon_sym_AT_LPAREN] = ACTIONS(1050),
    [anon_sym_AT_LBRACE] = ACTIONS(1052),
  },
  [STATE(247)] = {
    [sym__literal] = STATE(859),
    [sym_integer_literal] = STATE(859),
    [sym_string_literal] = STATE(859),
    [sym_expandable_string_literal] = STATE(842),
    [sym_expandable_here_string_literal] = STATE(842),
    [sym_variable] = STATE(859),
    [sym_array_literal_expression] = STATE(1098),
    [sym_unary_expression] = STATE(1005),
    [sym_expression_with_unary_operator] = STATE(1058),
    [sym_pre_increment_expression] = STATE(1065),
    [sym_pre_decrement_expression] = STATE(1065),
    [sym_cast_expression] = STATE(1065),
    [sym__primary_expression] = STATE(859),
    [sym__value] = STATE(859),
    [sym_parenthesized_expression] = STATE(859),
    [sym_sub_expression] = STATE(859),
    [sym_array_expression] = STATE(859),
    [sym_script_block_expression] = STATE(859),
    [sym_hash_literal_expression] = STATE(859),
    [sym_post_increment_expression] = STATE(859),
    [sym_post_decrement_expression] = STATE(859),
    [sym_member_access] = STATE(859),
    [sym_element_access] = STATE(859),
    [sym_invokation_expression] = STATE(859),
    [sym_invokation_foreach_expression] = STATE(868),
    [sym_type_literal] = STATE(237),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1056),
    [sym_hexadecimal_integer_literal] = ACTIONS(1056),
    [sym_real_literal] = ACTIONS(1058),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1060),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1062),
    [sym_verbatim_string_characters] = ACTIONS(1064),
    [sym_verbatim_here_string_characters] = ACTIONS(1064),
    [anon_sym_LBRACK] = ACTIONS(1066),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(1068),
    [aux_sym_comparison_operator_token50] = ACTIONS(1068),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1070),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1070),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1070),
    [anon_sym_DOLLAR_] = ACTIONS(1070),
    [aux_sym_variable_token1] = ACTIONS(1070),
    [aux_sym_variable_token2] = ACTIONS(1070),
    [sym_braced_variable] = ACTIONS(1070),
    [sym_generic_token] = ACTIONS(1136),
    [sym_command_parameter] = ACTIONS(1002),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1068),
    [anon_sym_LBRACE] = ACTIONS(1074),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [sym_stop_parsing] = ACTIONS(1002),
    [anon_sym_SPACE] = ACTIONS(1002),
    [anon_sym_COLON] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(1068),
    [anon_sym_DASH] = ACTIONS(1068),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1068),
    [anon_sym_BANG] = ACTIONS(1068),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1068),
    [anon_sym_PLUS_PLUS] = ACTIONS(1076),
    [anon_sym_DASH_DASH] = ACTIONS(1078),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1080),
    [anon_sym_AT_LPAREN] = ACTIONS(1082),
    [anon_sym_AT_LBRACE] = ACTIONS(1084),
    [sym__statement_terminator] = ACTIONS(1006),
  },
  [STATE(248)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_pipeline] = STATE(2095),
    [sym_pipeline_chain] = STATE(1342),
    [sym_left_assignment_expression] = STATE(1292),
    [sym_assignment_expression] = STATE(1742),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(845),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
    [sym__statement_terminator] = ACTIONS(1138),
  },
  [STATE(249)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1931),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(250)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1868),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(251)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1718),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(252)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1730),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(253)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1894),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(254)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1942),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(255)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_while_condition] = STATE(1985),
    [sym_pipeline] = STATE(1724),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(256)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1910),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(257)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1995),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(258)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1729),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(259)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(2015),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(260)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1967),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(261)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1844),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(262)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1800),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(263)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1857),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(264)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1727),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(265)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1862),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(266)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1954),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(267)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1752),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(268)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1779),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(269)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1801),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(270)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1823),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(271)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1837),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(272)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1845),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(273)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1852),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(274)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1858),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(275)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1892),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(276)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1901),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(277)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1905),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(278)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1940),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(279)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1948),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(280)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1952),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(281)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1957),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(282)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1983),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(283)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(1991),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(284)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline] = STATE(2000),
    [sym_pipeline_chain] = STATE(1348),
    [sym_left_assignment_expression] = STATE(1287),
    [sym_assignment_expression] = STATE(1934),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(855),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(285)] = {
    [sym__literal] = STATE(1072),
    [sym_integer_literal] = STATE(1072),
    [sym_string_literal] = STATE(1072),
    [sym_expandable_string_literal] = STATE(933),
    [sym_expandable_here_string_literal] = STATE(933),
    [sym_variable] = STATE(1072),
    [sym_unary_expression] = STATE(174),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(1072),
    [sym__value] = STATE(1072),
    [sym_parenthesized_expression] = STATE(1072),
    [sym_sub_expression] = STATE(1072),
    [sym_array_expression] = STATE(1072),
    [sym_script_block_expression] = STATE(1072),
    [sym_hash_literal_expression] = STATE(1072),
    [sym_post_increment_expression] = STATE(1072),
    [sym_post_decrement_expression] = STATE(1072),
    [sym_member_access] = STATE(1072),
    [sym_element_access] = STATE(1072),
    [sym_invokation_expression] = STATE(1072),
    [sym_invokation_foreach_expression] = STATE(945),
    [sym_type_literal] = STATE(285),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(1140),
    [sym_hexadecimal_integer_literal] = ACTIONS(1142),
    [sym_real_literal] = ACTIONS(1144),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1146),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1148),
    [sym_verbatim_string_characters] = ACTIONS(1150),
    [sym_verbatim_here_string_characters] = ACTIONS(1150),
    [anon_sym_LBRACK] = ACTIONS(1152),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(1154),
    [aux_sym_comparison_operator_token50] = ACTIONS(1154),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1156),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1156),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1156),
    [anon_sym_DOLLAR_] = ACTIONS(1158),
    [aux_sym_variable_token1] = ACTIONS(1158),
    [aux_sym_variable_token2] = ACTIONS(1156),
    [sym_braced_variable] = ACTIONS(1156),
    [anon_sym_LPAREN] = ACTIONS(1160),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1162),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1164),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1154),
    [anon_sym_BANG] = ACTIONS(1154),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1154),
    [anon_sym_PLUS_PLUS] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1170),
    [anon_sym_AT_LPAREN] = ACTIONS(1172),
    [anon_sym_AT_LBRACE] = ACTIONS(1174),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(286)] = {
    [sym__literal] = STATE(1071),
    [sym_integer_literal] = STATE(1071),
    [sym_string_literal] = STATE(1071),
    [sym_expandable_string_literal] = STATE(1009),
    [sym_expandable_here_string_literal] = STATE(1009),
    [sym_variable] = STATE(1071),
    [sym_unary_expression] = STATE(182),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(1071),
    [sym__value] = STATE(1071),
    [sym_parenthesized_expression] = STATE(1071),
    [sym_sub_expression] = STATE(1071),
    [sym_array_expression] = STATE(1071),
    [sym_script_block_expression] = STATE(1071),
    [sym_hash_literal_expression] = STATE(1071),
    [sym_post_increment_expression] = STATE(1071),
    [sym_post_decrement_expression] = STATE(1071),
    [sym_member_access] = STATE(1071),
    [sym_element_access] = STATE(1071),
    [sym_invokation_expression] = STATE(1071),
    [sym_invokation_foreach_expression] = STATE(1012),
    [sym_type_literal] = STATE(286),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(1176),
    [sym_hexadecimal_integer_literal] = ACTIONS(1178),
    [sym_real_literal] = ACTIONS(1180),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1182),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1184),
    [sym_verbatim_string_characters] = ACTIONS(1186),
    [sym_verbatim_here_string_characters] = ACTIONS(1186),
    [anon_sym_LBRACK] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(1190),
    [aux_sym_comparison_operator_token50] = ACTIONS(1190),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1192),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1192),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1192),
    [anon_sym_DOLLAR_] = ACTIONS(1194),
    [aux_sym_variable_token1] = ACTIONS(1194),
    [aux_sym_variable_token2] = ACTIONS(1192),
    [sym_braced_variable] = ACTIONS(1192),
    [anon_sym_LPAREN] = ACTIONS(1196),
    [anon_sym_COMMA] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1198),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1190),
    [anon_sym_BANG] = ACTIONS(1190),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1190),
    [anon_sym_PLUS_PLUS] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1206),
    [anon_sym_AT_LPAREN] = ACTIONS(1208),
    [anon_sym_AT_LBRACE] = ACTIONS(1210),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(287)] = {
    [sym__literal] = STATE(1037),
    [sym_integer_literal] = STATE(1037),
    [sym_string_literal] = STATE(1037),
    [sym_expandable_string_literal] = STATE(933),
    [sym_expandable_here_string_literal] = STATE(933),
    [sym_variable] = STATE(1037),
    [sym_unary_expression] = STATE(962),
    [sym_expression_with_unary_operator] = STATE(956),
    [sym_pre_increment_expression] = STATE(957),
    [sym_pre_decrement_expression] = STATE(957),
    [sym_cast_expression] = STATE(957),
    [sym__primary_expression] = STATE(1037),
    [sym__value] = STATE(1037),
    [sym_parenthesized_expression] = STATE(1037),
    [sym_sub_expression] = STATE(1037),
    [sym_array_expression] = STATE(1037),
    [sym_script_block_expression] = STATE(1037),
    [sym_hash_literal_expression] = STATE(1037),
    [sym_post_increment_expression] = STATE(1037),
    [sym_post_decrement_expression] = STATE(1037),
    [sym_member_access] = STATE(1037),
    [sym_element_access] = STATE(1037),
    [sym_invokation_expression] = STATE(1037),
    [sym_invokation_foreach_expression] = STATE(945),
    [sym_type_literal] = STATE(287),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(1140),
    [sym_hexadecimal_integer_literal] = ACTIONS(1142),
    [sym_real_literal] = ACTIONS(1212),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1146),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1148),
    [sym_verbatim_string_characters] = ACTIONS(1150),
    [sym_verbatim_here_string_characters] = ACTIONS(1150),
    [anon_sym_LBRACK] = ACTIONS(1152),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(1214),
    [aux_sym_comparison_operator_token50] = ACTIONS(1214),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1156),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1156),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1156),
    [anon_sym_DOLLAR_] = ACTIONS(1158),
    [aux_sym_variable_token1] = ACTIONS(1158),
    [aux_sym_variable_token2] = ACTIONS(1156),
    [sym_braced_variable] = ACTIONS(1156),
    [anon_sym_LPAREN] = ACTIONS(1160),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1162),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1216),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1214),
    [anon_sym_BANG] = ACTIONS(1214),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1214),
    [anon_sym_PLUS_PLUS] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1170),
    [anon_sym_AT_LPAREN] = ACTIONS(1172),
    [anon_sym_AT_LBRACE] = ACTIONS(1174),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
  },
  [STATE(288)] = {
    [sym__literal] = STATE(1036),
    [sym_integer_literal] = STATE(1036),
    [sym_string_literal] = STATE(1036),
    [sym_expandable_string_literal] = STATE(1009),
    [sym_expandable_here_string_literal] = STATE(1009),
    [sym_variable] = STATE(1036),
    [sym_unary_expression] = STATE(990),
    [sym_expression_with_unary_operator] = STATE(985),
    [sym_pre_increment_expression] = STATE(986),
    [sym_pre_decrement_expression] = STATE(986),
    [sym_cast_expression] = STATE(986),
    [sym__primary_expression] = STATE(1036),
    [sym__value] = STATE(1036),
    [sym_parenthesized_expression] = STATE(1036),
    [sym_sub_expression] = STATE(1036),
    [sym_array_expression] = STATE(1036),
    [sym_script_block_expression] = STATE(1036),
    [sym_hash_literal_expression] = STATE(1036),
    [sym_post_increment_expression] = STATE(1036),
    [sym_post_decrement_expression] = STATE(1036),
    [sym_member_access] = STATE(1036),
    [sym_element_access] = STATE(1036),
    [sym_invokation_expression] = STATE(1036),
    [sym_invokation_foreach_expression] = STATE(1012),
    [sym_type_literal] = STATE(288),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(1176),
    [sym_hexadecimal_integer_literal] = ACTIONS(1178),
    [sym_real_literal] = ACTIONS(1222),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1182),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1184),
    [sym_verbatim_string_characters] = ACTIONS(1186),
    [sym_verbatim_here_string_characters] = ACTIONS(1186),
    [anon_sym_LBRACK] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(95),
    [anon_sym_GT_GT] = ACTIONS(93),
    [anon_sym_2_GT] = ACTIONS(95),
    [anon_sym_2_GT_GT] = ACTIONS(93),
    [anon_sym_3_GT] = ACTIONS(95),
    [anon_sym_3_GT_GT] = ACTIONS(93),
    [anon_sym_4_GT] = ACTIONS(95),
    [anon_sym_4_GT_GT] = ACTIONS(93),
    [anon_sym_5_GT] = ACTIONS(95),
    [anon_sym_5_GT_GT] = ACTIONS(93),
    [anon_sym_6_GT] = ACTIONS(95),
    [anon_sym_6_GT_GT] = ACTIONS(93),
    [anon_sym_STAR_GT] = ACTIONS(95),
    [anon_sym_STAR_GT_GT] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(93),
    [anon_sym_2_GT_AMP1] = ACTIONS(93),
    [anon_sym_3_GT_AMP1] = ACTIONS(93),
    [anon_sym_4_GT_AMP1] = ACTIONS(93),
    [anon_sym_5_GT_AMP1] = ACTIONS(93),
    [anon_sym_6_GT_AMP1] = ACTIONS(93),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(93),
    [anon_sym_1_GT_AMP2] = ACTIONS(93),
    [anon_sym_3_GT_AMP2] = ACTIONS(93),
    [anon_sym_4_GT_AMP2] = ACTIONS(93),
    [anon_sym_5_GT_AMP2] = ACTIONS(93),
    [anon_sym_6_GT_AMP2] = ACTIONS(93),
    [aux_sym_comparison_operator_token37] = ACTIONS(1224),
    [aux_sym_comparison_operator_token50] = ACTIONS(1224),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1192),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1192),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1192),
    [anon_sym_DOLLAR_] = ACTIONS(1194),
    [aux_sym_variable_token1] = ACTIONS(1194),
    [aux_sym_variable_token2] = ACTIONS(1192),
    [sym_braced_variable] = ACTIONS(1192),
    [anon_sym_LPAREN] = ACTIONS(1196),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1198),
    [anon_sym_AMP_AMP] = ACTIONS(93),
    [anon_sym_PIPE_PIPE] = ACTIONS(93),
    [anon_sym_PIPE] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_DASH] = ACTIONS(1226),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [anon_sym_DASH_DASH] = ACTIONS(1230),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1206),
    [anon_sym_AT_LPAREN] = ACTIONS(1208),
    [anon_sym_AT_LBRACE] = ACTIONS(1210),
    [anon_sym_DOT2] = ACTIONS(95),
    [anon_sym_COLON_COLON] = ACTIONS(93),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(93),
    [sym__statement_terminator] = ACTIONS(93),
  },
  [STATE(289)] = {
    [sym__literal] = STATE(1072),
    [sym_integer_literal] = STATE(1072),
    [sym_string_literal] = STATE(1072),
    [sym_expandable_string_literal] = STATE(933),
    [sym_expandable_here_string_literal] = STATE(933),
    [sym_variable] = STATE(1072),
    [sym_array_literal_expression] = STATE(1128),
    [sym_unary_expression] = STATE(1124),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(1072),
    [sym__value] = STATE(1072),
    [sym_parenthesized_expression] = STATE(1072),
    [sym_sub_expression] = STATE(1072),
    [sym_array_expression] = STATE(1072),
    [sym_script_block_expression] = STATE(1072),
    [sym_hash_literal_expression] = STATE(1072),
    [sym_post_increment_expression] = STATE(1072),
    [sym_post_decrement_expression] = STATE(1072),
    [sym_member_access] = STATE(1072),
    [sym_element_access] = STATE(1072),
    [sym_invokation_expression] = STATE(1072),
    [sym_invokation_foreach_expression] = STATE(945),
    [sym_type_literal] = STATE(285),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1140),
    [sym_hexadecimal_integer_literal] = ACTIONS(1140),
    [sym_real_literal] = ACTIONS(1232),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1234),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1236),
    [sym_verbatim_string_characters] = ACTIONS(1238),
    [sym_verbatim_here_string_characters] = ACTIONS(1238),
    [anon_sym_LBRACK] = ACTIONS(1240),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(1164),
    [aux_sym_comparison_operator_token50] = ACTIONS(1164),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1158),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1158),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1158),
    [anon_sym_DOLLAR_] = ACTIONS(1158),
    [aux_sym_variable_token1] = ACTIONS(1158),
    [aux_sym_variable_token2] = ACTIONS(1158),
    [sym_braced_variable] = ACTIONS(1158),
    [sym_generic_token] = ACTIONS(1242),
    [anon_sym_LPAREN] = ACTIONS(1244),
    [anon_sym_RPAREN] = ACTIONS(1002),
    [anon_sym_COMMA] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1246),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1164),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [anon_sym_DASH_DASH] = ACTIONS(1250),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1252),
    [anon_sym_AT_LPAREN] = ACTIONS(1254),
    [anon_sym_AT_LBRACE] = ACTIONS(1256),
  },
  [STATE(290)] = {
    [sym__literal] = STATE(1071),
    [sym_integer_literal] = STATE(1071),
    [sym_string_literal] = STATE(1071),
    [sym_expandable_string_literal] = STATE(1009),
    [sym_expandable_here_string_literal] = STATE(1009),
    [sym_variable] = STATE(1071),
    [sym_array_literal_expression] = STATE(1137),
    [sym_unary_expression] = STATE(1127),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(1071),
    [sym__value] = STATE(1071),
    [sym_parenthesized_expression] = STATE(1071),
    [sym_sub_expression] = STATE(1071),
    [sym_array_expression] = STATE(1071),
    [sym_script_block_expression] = STATE(1071),
    [sym_hash_literal_expression] = STATE(1071),
    [sym_post_increment_expression] = STATE(1071),
    [sym_post_decrement_expression] = STATE(1071),
    [sym_member_access] = STATE(1071),
    [sym_element_access] = STATE(1071),
    [sym_invokation_expression] = STATE(1071),
    [sym_invokation_foreach_expression] = STATE(1012),
    [sym_type_literal] = STATE(286),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1176),
    [sym_hexadecimal_integer_literal] = ACTIONS(1176),
    [sym_real_literal] = ACTIONS(1258),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1260),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1262),
    [sym_verbatim_string_characters] = ACTIONS(1264),
    [sym_verbatim_here_string_characters] = ACTIONS(1264),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_GT] = ACTIONS(1002),
    [anon_sym_GT_GT] = ACTIONS(1002),
    [anon_sym_2_GT] = ACTIONS(1002),
    [anon_sym_2_GT_GT] = ACTIONS(1002),
    [anon_sym_3_GT] = ACTIONS(1002),
    [anon_sym_3_GT_GT] = ACTIONS(1002),
    [anon_sym_4_GT] = ACTIONS(1002),
    [anon_sym_4_GT_GT] = ACTIONS(1002),
    [anon_sym_5_GT] = ACTIONS(1002),
    [anon_sym_5_GT_GT] = ACTIONS(1002),
    [anon_sym_6_GT] = ACTIONS(1002),
    [anon_sym_6_GT_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT] = ACTIONS(1002),
    [anon_sym_STAR_GT_GT] = ACTIONS(1002),
    [anon_sym_LT] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1002),
    [anon_sym_2_GT_AMP1] = ACTIONS(1002),
    [anon_sym_3_GT_AMP1] = ACTIONS(1002),
    [anon_sym_4_GT_AMP1] = ACTIONS(1002),
    [anon_sym_5_GT_AMP1] = ACTIONS(1002),
    [anon_sym_6_GT_AMP1] = ACTIONS(1002),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1002),
    [anon_sym_1_GT_AMP2] = ACTIONS(1002),
    [anon_sym_3_GT_AMP2] = ACTIONS(1002),
    [anon_sym_4_GT_AMP2] = ACTIONS(1002),
    [anon_sym_5_GT_AMP2] = ACTIONS(1002),
    [anon_sym_6_GT_AMP2] = ACTIONS(1002),
    [aux_sym_comparison_operator_token37] = ACTIONS(1200),
    [aux_sym_comparison_operator_token50] = ACTIONS(1200),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1194),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1194),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1194),
    [anon_sym_DOLLAR_] = ACTIONS(1194),
    [aux_sym_variable_token1] = ACTIONS(1194),
    [aux_sym_variable_token2] = ACTIONS(1194),
    [sym_braced_variable] = ACTIONS(1194),
    [sym_generic_token] = ACTIONS(1268),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1002),
    [anon_sym_PIPE_PIPE] = ACTIONS(1002),
    [anon_sym_PIPE] = ACTIONS(1002),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1278),
    [anon_sym_AT_LPAREN] = ACTIONS(1280),
    [anon_sym_AT_LBRACE] = ACTIONS(1282),
    [sym__statement_terminator] = ACTIONS(1006),
  },
  [STATE(291)] = {
    [sym__literal] = STATE(159),
    [sym_integer_literal] = STATE(159),
    [sym_string_literal] = STATE(159),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(159),
    [sym_pipeline_chain] = STATE(1455),
    [sym_command] = STATE(1319),
    [sym_command_invokation_operator] = STATE(516),
    [sym_command_name] = STATE(711),
    [sym__expression] = STATE(1095),
    [sym_logical_expression] = STATE(963),
    [sym_bitwise_expression] = STATE(856),
    [sym_comparison_expression] = STATE(214),
    [sym_additive_expression] = STATE(197),
    [sym_multiplicative_expression] = STATE(193),
    [sym_format_expression] = STATE(186),
    [sym_range_expression] = STATE(187),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(164),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(159),
    [sym__value] = STATE(159),
    [sym_parenthesized_expression] = STATE(159),
    [sym_sub_expression] = STATE(159),
    [sym_array_expression] = STATE(159),
    [sym_script_block_expression] = STATE(159),
    [sym_hash_literal_expression] = STATE(159),
    [sym_post_increment_expression] = STATE(159),
    [sym_post_decrement_expression] = STATE(159),
    [sym_member_access] = STATE(159),
    [sym_element_access] = STATE(159),
    [sym_invokation_expression] = STATE(159),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(4),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(117),
    [sym_real_literal] = ACTIONS(1108),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1110),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1112),
    [sym_verbatim_string_characters] = ACTIONS(1114),
    [sym_verbatim_here_string_characters] = ACTIONS(1114),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [aux_sym_comparison_operator_token37] = ACTIONS(141),
    [aux_sym_comparison_operator_token50] = ACTIONS(141),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(135),
    [anon_sym_DOLLAR_CARET] = ACTIONS(135),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(135),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(135),
    [sym_braced_variable] = ACTIONS(135),
    [anon_sym_LPAREN] = ACTIONS(1118),
    [anon_sym_COMMA] = ACTIONS(141),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(1122),
    [aux_sym_command_name_token2] = ACTIONS(1122),
    [aux_sym_command_name_token3] = ACTIONS(1122),
    [aux_sym_command_name_token4] = ACTIONS(1122),
    [aux_sym_command_name_token5] = ACTIONS(1122),
    [aux_sym_command_name_token6] = ACTIONS(1122),
    [aux_sym_command_name_token7] = ACTIONS(1122),
    [aux_sym_command_name_token8] = ACTIONS(1122),
    [aux_sym_command_name_token9] = ACTIONS(1122),
    [aux_sym_command_name_token10] = ACTIONS(1122),
    [aux_sym_command_name_token11] = ACTIONS(1122),
    [aux_sym_command_name_token12] = ACTIONS(1122),
    [aux_sym_command_name_token13] = ACTIONS(1122),
    [aux_sym_command_name_token14] = ACTIONS(1122),
    [aux_sym_command_name_token15] = ACTIONS(1122),
    [aux_sym_command_name_token16] = ACTIONS(1122),
    [aux_sym_command_name_token17] = ACTIONS(1122),
    [aux_sym_command_name_token18] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(141),
    [anon_sym_BANG] = ACTIONS(141),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1128),
    [anon_sym_AT_LPAREN] = ACTIONS(1130),
    [anon_sym_AT_LBRACE] = ACTIONS(1132),
  },
  [STATE(292)] = {
    [sym__literal] = STATE(160),
    [sym_integer_literal] = STATE(160),
    [sym_string_literal] = STATE(160),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(160),
    [sym_pipeline_chain] = STATE(1476),
    [sym_command] = STATE(1304),
    [sym_command_invokation_operator] = STATE(515),
    [sym_command_name] = STATE(709),
    [sym__expression] = STATE(1113),
    [sym_logical_expression] = STATE(969),
    [sym_bitwise_expression] = STATE(854),
    [sym_comparison_expression] = STATE(202),
    [sym_additive_expression] = STATE(198),
    [sym_multiplicative_expression] = STATE(195),
    [sym_format_expression] = STATE(183),
    [sym_range_expression] = STATE(192),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(171),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(160),
    [sym__value] = STATE(160),
    [sym_parenthesized_expression] = STATE(160),
    [sym_sub_expression] = STATE(160),
    [sym_array_expression] = STATE(160),
    [sym_script_block_expression] = STATE(160),
    [sym_hash_literal_expression] = STATE(160),
    [sym_post_increment_expression] = STATE(160),
    [sym_post_decrement_expression] = STATE(160),
    [sym_member_access] = STATE(160),
    [sym_element_access] = STATE(160),
    [sym_invokation_expression] = STATE(160),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(3),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(5),
    [sym_real_literal] = ACTIONS(7),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(9),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(11),
    [sym_verbatim_string_characters] = ACTIONS(13),
    [sym_verbatim_here_string_characters] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(241),
    [aux_sym_comparison_operator_token37] = ACTIONS(19),
    [aux_sym_comparison_operator_token50] = ACTIONS(19),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(21),
    [anon_sym_DOLLAR_CARET] = ACTIONS(21),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(21),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(21),
    [sym_braced_variable] = ACTIONS(21),
    [anon_sym_LPAREN] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(19),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_AMP] = ACTIONS(15),
    [aux_sym_command_name_token1] = ACTIONS(63),
    [aux_sym_command_name_token2] = ACTIONS(63),
    [aux_sym_command_name_token3] = ACTIONS(63),
    [aux_sym_command_name_token4] = ACTIONS(63),
    [aux_sym_command_name_token5] = ACTIONS(63),
    [aux_sym_command_name_token6] = ACTIONS(63),
    [aux_sym_command_name_token7] = ACTIONS(63),
    [aux_sym_command_name_token8] = ACTIONS(63),
    [aux_sym_command_name_token9] = ACTIONS(63),
    [aux_sym_command_name_token10] = ACTIONS(63),
    [aux_sym_command_name_token11] = ACTIONS(63),
    [aux_sym_command_name_token12] = ACTIONS(63),
    [aux_sym_command_name_token13] = ACTIONS(63),
    [aux_sym_command_name_token14] = ACTIONS(63),
    [aux_sym_command_name_token15] = ACTIONS(63),
    [aux_sym_command_name_token16] = ACTIONS(63),
    [aux_sym_command_name_token17] = ACTIONS(63),
    [aux_sym_command_name_token18] = ACTIONS(63),
    [anon_sym_PLUS] = ACTIONS(19),
    [anon_sym_DASH] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(19),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(19),
    [anon_sym_PLUS_PLUS] = ACTIONS(69),
    [anon_sym_DASH_DASH] = ACTIONS(71),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(73),
    [anon_sym_AT_LPAREN] = ACTIONS(75),
    [anon_sym_AT_LBRACE] = ACTIONS(77),
  },
  [STATE(293)] = {
    [sym_elseif_clauses] = STATE(311),
    [sym_elseif_clause] = STATE(298),
    [sym_else_clause] = STATE(337),
    [aux_sym_elseif_clauses_repeat1] = STATE(298),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1284),
    [sym_hexadecimal_integer_literal] = ACTIONS(1284),
    [sym_real_literal] = ACTIONS(1284),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1284),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1284),
    [sym_verbatim_string_characters] = ACTIONS(1284),
    [sym_verbatim_here_string_characters] = ACTIONS(1284),
    [anon_sym_DOT] = ACTIONS(1284),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [aux_sym_comparison_operator_token37] = ACTIONS(1284),
    [aux_sym_comparison_operator_token50] = ACTIONS(1284),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1284),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1284),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1284),
    [anon_sym_DOLLAR_] = ACTIONS(1284),
    [aux_sym_variable_token1] = ACTIONS(1284),
    [aux_sym_variable_token2] = ACTIONS(1284),
    [sym_braced_variable] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym_LPAREN] = ACTIONS(1284),
    [anon_sym_RPAREN] = ACTIONS(1284),
    [anon_sym_COMMA] = ACTIONS(1284),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_RBRACE] = ACTIONS(1284),
    [aux_sym_if_statement_token1] = ACTIONS(1284),
    [aux_sym_elseif_clause_token1] = ACTIONS(1286),
    [aux_sym_else_clause_token1] = ACTIONS(1288),
    [aux_sym_switch_statement_token1] = ACTIONS(1284),
    [aux_sym_foreach_statement_token1] = ACTIONS(1284),
    [aux_sym_for_statement_token1] = ACTIONS(1284),
    [aux_sym_while_statement_token1] = ACTIONS(1284),
    [aux_sym_do_statement_token1] = ACTIONS(1284),
    [aux_sym_function_statement_token1] = ACTIONS(1284),
    [aux_sym_function_statement_token2] = ACTIONS(1284),
    [aux_sym_function_statement_token3] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1284),
    [sym_label] = ACTIONS(1284),
    [aux_sym_trap_statement_token1] = ACTIONS(1284),
    [aux_sym_try_statement_token1] = ACTIONS(1284),
    [aux_sym_data_statement_token1] = ACTIONS(1284),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1284),
    [aux_sym_parallel_statement_token1] = ACTIONS(1284),
    [aux_sym_sequence_statement_token1] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [aux_sym_command_name_token1] = ACTIONS(1284),
    [aux_sym_command_name_token2] = ACTIONS(1284),
    [aux_sym_command_name_token3] = ACTIONS(1284),
    [aux_sym_command_name_token4] = ACTIONS(1284),
    [aux_sym_command_name_token5] = ACTIONS(1284),
    [aux_sym_command_name_token6] = ACTIONS(1284),
    [aux_sym_command_name_token7] = ACTIONS(1284),
    [aux_sym_command_name_token8] = ACTIONS(1284),
    [aux_sym_command_name_token9] = ACTIONS(1284),
    [aux_sym_command_name_token10] = ACTIONS(1284),
    [aux_sym_command_name_token11] = ACTIONS(1284),
    [aux_sym_command_name_token12] = ACTIONS(1284),
    [aux_sym_command_name_token13] = ACTIONS(1284),
    [aux_sym_command_name_token14] = ACTIONS(1284),
    [aux_sym_command_name_token15] = ACTIONS(1284),
    [aux_sym_command_name_token16] = ACTIONS(1284),
    [aux_sym_command_name_token17] = ACTIONS(1284),
    [aux_sym_command_name_token18] = ACTIONS(1284),
    [aux_sym_class_statement_token1] = ACTIONS(1284),
    [aux_sym_enum_statement_token1] = ACTIONS(1284),
    [anon_sym_PLUS] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1284),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1284),
    [anon_sym_AT_LPAREN] = ACTIONS(1284),
    [anon_sym_AT_LBRACE] = ACTIONS(1284),
  },
  [STATE(294)] = {
    [sym_catch_clauses] = STATE(314),
    [sym_catch_clause] = STATE(303),
    [sym_finally_clause] = STATE(345),
    [aux_sym_catch_clauses_repeat1] = STATE(303),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1290),
    [sym_hexadecimal_integer_literal] = ACTIONS(1290),
    [sym_real_literal] = ACTIONS(1290),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1290),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1290),
    [sym_verbatim_string_characters] = ACTIONS(1290),
    [sym_verbatim_here_string_characters] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_LBRACK] = ACTIONS(1290),
    [aux_sym_comparison_operator_token37] = ACTIONS(1290),
    [aux_sym_comparison_operator_token50] = ACTIONS(1290),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1290),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1290),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1290),
    [anon_sym_DOLLAR_] = ACTIONS(1290),
    [aux_sym_variable_token1] = ACTIONS(1290),
    [aux_sym_variable_token2] = ACTIONS(1290),
    [sym_braced_variable] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_RPAREN] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [aux_sym_if_statement_token1] = ACTIONS(1290),
    [aux_sym_switch_statement_token1] = ACTIONS(1290),
    [aux_sym_foreach_statement_token1] = ACTIONS(1290),
    [aux_sym_for_statement_token1] = ACTIONS(1290),
    [aux_sym_while_statement_token1] = ACTIONS(1290),
    [aux_sym_do_statement_token1] = ACTIONS(1290),
    [aux_sym_function_statement_token1] = ACTIONS(1290),
    [aux_sym_function_statement_token2] = ACTIONS(1290),
    [aux_sym_function_statement_token3] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1290),
    [sym_label] = ACTIONS(1290),
    [aux_sym_trap_statement_token1] = ACTIONS(1290),
    [aux_sym_try_statement_token1] = ACTIONS(1290),
    [aux_sym_catch_clause_token1] = ACTIONS(1292),
    [aux_sym_finally_clause_token1] = ACTIONS(1294),
    [aux_sym_data_statement_token1] = ACTIONS(1290),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1290),
    [aux_sym_parallel_statement_token1] = ACTIONS(1290),
    [aux_sym_sequence_statement_token1] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [aux_sym_command_name_token1] = ACTIONS(1290),
    [aux_sym_command_name_token2] = ACTIONS(1290),
    [aux_sym_command_name_token3] = ACTIONS(1290),
    [aux_sym_command_name_token4] = ACTIONS(1290),
    [aux_sym_command_name_token5] = ACTIONS(1290),
    [aux_sym_command_name_token6] = ACTIONS(1290),
    [aux_sym_command_name_token7] = ACTIONS(1290),
    [aux_sym_command_name_token8] = ACTIONS(1290),
    [aux_sym_command_name_token9] = ACTIONS(1290),
    [aux_sym_command_name_token10] = ACTIONS(1290),
    [aux_sym_command_name_token11] = ACTIONS(1290),
    [aux_sym_command_name_token12] = ACTIONS(1290),
    [aux_sym_command_name_token13] = ACTIONS(1290),
    [aux_sym_command_name_token14] = ACTIONS(1290),
    [aux_sym_command_name_token15] = ACTIONS(1290),
    [aux_sym_command_name_token16] = ACTIONS(1290),
    [aux_sym_command_name_token17] = ACTIONS(1290),
    [aux_sym_command_name_token18] = ACTIONS(1290),
    [aux_sym_class_statement_token1] = ACTIONS(1290),
    [aux_sym_enum_statement_token1] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1290),
    [anon_sym_PLUS_PLUS] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1290),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1290),
    [anon_sym_AT_LPAREN] = ACTIONS(1290),
    [anon_sym_AT_LBRACE] = ACTIONS(1290),
  },
  [STATE(295)] = {
    [sym_catch_clauses] = STATE(324),
    [sym_catch_clause] = STATE(309),
    [sym_finally_clause] = STATE(420),
    [aux_sym_catch_clauses_repeat1] = STATE(309),
    [ts_builtin_sym_end] = ACTIONS(1296),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1290),
    [sym_hexadecimal_integer_literal] = ACTIONS(1290),
    [sym_real_literal] = ACTIONS(1290),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1290),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1290),
    [sym_verbatim_string_characters] = ACTIONS(1290),
    [sym_verbatim_here_string_characters] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_LBRACK] = ACTIONS(1290),
    [aux_sym_comparison_operator_token37] = ACTIONS(1290),
    [aux_sym_comparison_operator_token50] = ACTIONS(1290),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1290),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1290),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1290),
    [anon_sym_DOLLAR_] = ACTIONS(1290),
    [aux_sym_variable_token1] = ACTIONS(1290),
    [aux_sym_variable_token2] = ACTIONS(1290),
    [sym_braced_variable] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [aux_sym_if_statement_token1] = ACTIONS(1290),
    [aux_sym_switch_statement_token1] = ACTIONS(1290),
    [aux_sym_foreach_statement_token1] = ACTIONS(1290),
    [aux_sym_for_statement_token1] = ACTIONS(1290),
    [aux_sym_while_statement_token1] = ACTIONS(1290),
    [aux_sym_do_statement_token1] = ACTIONS(1290),
    [aux_sym_function_statement_token1] = ACTIONS(1290),
    [aux_sym_function_statement_token2] = ACTIONS(1290),
    [aux_sym_function_statement_token3] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1290),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1290),
    [sym_label] = ACTIONS(1290),
    [aux_sym_trap_statement_token1] = ACTIONS(1290),
    [aux_sym_try_statement_token1] = ACTIONS(1290),
    [aux_sym_catch_clause_token1] = ACTIONS(1298),
    [aux_sym_finally_clause_token1] = ACTIONS(1300),
    [aux_sym_data_statement_token1] = ACTIONS(1290),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1290),
    [aux_sym_parallel_statement_token1] = ACTIONS(1290),
    [aux_sym_sequence_statement_token1] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [aux_sym_command_name_token1] = ACTIONS(1290),
    [aux_sym_command_name_token2] = ACTIONS(1290),
    [aux_sym_command_name_token3] = ACTIONS(1290),
    [aux_sym_command_name_token4] = ACTIONS(1290),
    [aux_sym_command_name_token5] = ACTIONS(1290),
    [aux_sym_command_name_token6] = ACTIONS(1290),
    [aux_sym_command_name_token7] = ACTIONS(1290),
    [aux_sym_command_name_token8] = ACTIONS(1290),
    [aux_sym_command_name_token9] = ACTIONS(1290),
    [aux_sym_command_name_token10] = ACTIONS(1290),
    [aux_sym_command_name_token11] = ACTIONS(1290),
    [aux_sym_command_name_token12] = ACTIONS(1290),
    [aux_sym_command_name_token13] = ACTIONS(1290),
    [aux_sym_command_name_token14] = ACTIONS(1290),
    [aux_sym_command_name_token15] = ACTIONS(1290),
    [aux_sym_command_name_token16] = ACTIONS(1290),
    [aux_sym_command_name_token17] = ACTIONS(1290),
    [aux_sym_command_name_token18] = ACTIONS(1290),
    [aux_sym_class_statement_token1] = ACTIONS(1290),
    [aux_sym_enum_statement_token1] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1290),
    [anon_sym_PLUS_PLUS] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1290),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1290),
    [anon_sym_AT_LPAREN] = ACTIONS(1290),
    [anon_sym_AT_LBRACE] = ACTIONS(1290),
  },
  [STATE(296)] = {
    [sym_elseif_clauses] = STATE(320),
    [sym_elseif_clause] = STATE(307),
    [sym_else_clause] = STATE(388),
    [aux_sym_elseif_clauses_repeat1] = STATE(307),
    [ts_builtin_sym_end] = ACTIONS(1302),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1284),
    [sym_hexadecimal_integer_literal] = ACTIONS(1284),
    [sym_real_literal] = ACTIONS(1284),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1284),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1284),
    [sym_verbatim_string_characters] = ACTIONS(1284),
    [sym_verbatim_here_string_characters] = ACTIONS(1284),
    [anon_sym_DOT] = ACTIONS(1284),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [aux_sym_comparison_operator_token37] = ACTIONS(1284),
    [aux_sym_comparison_operator_token50] = ACTIONS(1284),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1284),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1284),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1284),
    [anon_sym_DOLLAR_] = ACTIONS(1284),
    [aux_sym_variable_token1] = ACTIONS(1284),
    [aux_sym_variable_token2] = ACTIONS(1284),
    [sym_braced_variable] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym_LPAREN] = ACTIONS(1284),
    [anon_sym_COMMA] = ACTIONS(1284),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [aux_sym_if_statement_token1] = ACTIONS(1284),
    [aux_sym_elseif_clause_token1] = ACTIONS(1304),
    [aux_sym_else_clause_token1] = ACTIONS(1306),
    [aux_sym_switch_statement_token1] = ACTIONS(1284),
    [aux_sym_foreach_statement_token1] = ACTIONS(1284),
    [aux_sym_for_statement_token1] = ACTIONS(1284),
    [aux_sym_while_statement_token1] = ACTIONS(1284),
    [aux_sym_do_statement_token1] = ACTIONS(1284),
    [aux_sym_function_statement_token1] = ACTIONS(1284),
    [aux_sym_function_statement_token2] = ACTIONS(1284),
    [aux_sym_function_statement_token3] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1284),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1284),
    [sym_label] = ACTIONS(1284),
    [aux_sym_trap_statement_token1] = ACTIONS(1284),
    [aux_sym_try_statement_token1] = ACTIONS(1284),
    [aux_sym_data_statement_token1] = ACTIONS(1284),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1284),
    [aux_sym_parallel_statement_token1] = ACTIONS(1284),
    [aux_sym_sequence_statement_token1] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [aux_sym_command_name_token1] = ACTIONS(1284),
    [aux_sym_command_name_token2] = ACTIONS(1284),
    [aux_sym_command_name_token3] = ACTIONS(1284),
    [aux_sym_command_name_token4] = ACTIONS(1284),
    [aux_sym_command_name_token5] = ACTIONS(1284),
    [aux_sym_command_name_token6] = ACTIONS(1284),
    [aux_sym_command_name_token7] = ACTIONS(1284),
    [aux_sym_command_name_token8] = ACTIONS(1284),
    [aux_sym_command_name_token9] = ACTIONS(1284),
    [aux_sym_command_name_token10] = ACTIONS(1284),
    [aux_sym_command_name_token11] = ACTIONS(1284),
    [aux_sym_command_name_token12] = ACTIONS(1284),
    [aux_sym_command_name_token13] = ACTIONS(1284),
    [aux_sym_command_name_token14] = ACTIONS(1284),
    [aux_sym_command_name_token15] = ACTIONS(1284),
    [aux_sym_command_name_token16] = ACTIONS(1284),
    [aux_sym_command_name_token17] = ACTIONS(1284),
    [aux_sym_command_name_token18] = ACTIONS(1284),
    [aux_sym_class_statement_token1] = ACTIONS(1284),
    [aux_sym_enum_statement_token1] = ACTIONS(1284),
    [anon_sym_PLUS] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1284),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1284),
    [anon_sym_AT_LPAREN] = ACTIONS(1284),
    [anon_sym_AT_LBRACE] = ACTIONS(1284),
  },
  [STATE(297)] = {
    [sym_catch_clause] = STATE(297),
    [aux_sym_catch_clauses_repeat1] = STATE(297),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1308),
    [sym_hexadecimal_integer_literal] = ACTIONS(1308),
    [sym_real_literal] = ACTIONS(1308),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1308),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1308),
    [sym_verbatim_string_characters] = ACTIONS(1308),
    [sym_verbatim_here_string_characters] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1308),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_comparison_operator_token37] = ACTIONS(1308),
    [aux_sym_comparison_operator_token50] = ACTIONS(1308),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1308),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1308),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_] = ACTIONS(1308),
    [aux_sym_variable_token1] = ACTIONS(1308),
    [aux_sym_variable_token2] = ACTIONS(1308),
    [sym_braced_variable] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_RPAREN] = ACTIONS(1308),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_if_statement_token1] = ACTIONS(1308),
    [aux_sym_switch_statement_token1] = ACTIONS(1308),
    [aux_sym_foreach_statement_token1] = ACTIONS(1308),
    [aux_sym_for_statement_token1] = ACTIONS(1308),
    [aux_sym_while_statement_token1] = ACTIONS(1308),
    [aux_sym_do_statement_token1] = ACTIONS(1308),
    [aux_sym_function_statement_token1] = ACTIONS(1308),
    [aux_sym_function_statement_token2] = ACTIONS(1308),
    [aux_sym_function_statement_token3] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1308),
    [sym_label] = ACTIONS(1308),
    [aux_sym_trap_statement_token1] = ACTIONS(1308),
    [aux_sym_try_statement_token1] = ACTIONS(1308),
    [aux_sym_catch_clause_token1] = ACTIONS(1310),
    [aux_sym_finally_clause_token1] = ACTIONS(1308),
    [aux_sym_data_statement_token1] = ACTIONS(1308),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1308),
    [aux_sym_parallel_statement_token1] = ACTIONS(1308),
    [aux_sym_sequence_statement_token1] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [aux_sym_command_name_token1] = ACTIONS(1308),
    [aux_sym_command_name_token2] = ACTIONS(1308),
    [aux_sym_command_name_token3] = ACTIONS(1308),
    [aux_sym_command_name_token4] = ACTIONS(1308),
    [aux_sym_command_name_token5] = ACTIONS(1308),
    [aux_sym_command_name_token6] = ACTIONS(1308),
    [aux_sym_command_name_token7] = ACTIONS(1308),
    [aux_sym_command_name_token8] = ACTIONS(1308),
    [aux_sym_command_name_token9] = ACTIONS(1308),
    [aux_sym_command_name_token10] = ACTIONS(1308),
    [aux_sym_command_name_token11] = ACTIONS(1308),
    [aux_sym_command_name_token12] = ACTIONS(1308),
    [aux_sym_command_name_token13] = ACTIONS(1308),
    [aux_sym_command_name_token14] = ACTIONS(1308),
    [aux_sym_command_name_token15] = ACTIONS(1308),
    [aux_sym_command_name_token16] = ACTIONS(1308),
    [aux_sym_command_name_token17] = ACTIONS(1308),
    [aux_sym_command_name_token18] = ACTIONS(1308),
    [aux_sym_class_statement_token1] = ACTIONS(1308),
    [aux_sym_enum_statement_token1] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_AT_LPAREN] = ACTIONS(1308),
    [anon_sym_AT_LBRACE] = ACTIONS(1308),
  },
  [STATE(298)] = {
    [sym_elseif_clause] = STATE(301),
    [aux_sym_elseif_clauses_repeat1] = STATE(301),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1313),
    [sym_hexadecimal_integer_literal] = ACTIONS(1313),
    [sym_real_literal] = ACTIONS(1313),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1313),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1313),
    [sym_verbatim_string_characters] = ACTIONS(1313),
    [sym_verbatim_here_string_characters] = ACTIONS(1313),
    [anon_sym_DOT] = ACTIONS(1313),
    [anon_sym_LBRACK] = ACTIONS(1313),
    [aux_sym_comparison_operator_token37] = ACTIONS(1313),
    [aux_sym_comparison_operator_token50] = ACTIONS(1313),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1313),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1313),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1313),
    [anon_sym_DOLLAR_] = ACTIONS(1313),
    [aux_sym_variable_token1] = ACTIONS(1313),
    [aux_sym_variable_token2] = ACTIONS(1313),
    [sym_braced_variable] = ACTIONS(1313),
    [anon_sym_SEMI] = ACTIONS(1313),
    [anon_sym_LPAREN] = ACTIONS(1313),
    [anon_sym_RPAREN] = ACTIONS(1313),
    [anon_sym_COMMA] = ACTIONS(1313),
    [anon_sym_LBRACE] = ACTIONS(1313),
    [anon_sym_RBRACE] = ACTIONS(1313),
    [aux_sym_if_statement_token1] = ACTIONS(1313),
    [aux_sym_elseif_clause_token1] = ACTIONS(1286),
    [aux_sym_else_clause_token1] = ACTIONS(1313),
    [aux_sym_switch_statement_token1] = ACTIONS(1313),
    [aux_sym_foreach_statement_token1] = ACTIONS(1313),
    [aux_sym_for_statement_token1] = ACTIONS(1313),
    [aux_sym_while_statement_token1] = ACTIONS(1313),
    [aux_sym_do_statement_token1] = ACTIONS(1313),
    [aux_sym_function_statement_token1] = ACTIONS(1313),
    [aux_sym_function_statement_token2] = ACTIONS(1313),
    [aux_sym_function_statement_token3] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1313),
    [sym_label] = ACTIONS(1313),
    [aux_sym_trap_statement_token1] = ACTIONS(1313),
    [aux_sym_try_statement_token1] = ACTIONS(1313),
    [aux_sym_data_statement_token1] = ACTIONS(1313),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1313),
    [aux_sym_parallel_statement_token1] = ACTIONS(1313),
    [aux_sym_sequence_statement_token1] = ACTIONS(1313),
    [anon_sym_AMP] = ACTIONS(1313),
    [aux_sym_command_name_token1] = ACTIONS(1313),
    [aux_sym_command_name_token2] = ACTIONS(1313),
    [aux_sym_command_name_token3] = ACTIONS(1313),
    [aux_sym_command_name_token4] = ACTIONS(1313),
    [aux_sym_command_name_token5] = ACTIONS(1313),
    [aux_sym_command_name_token6] = ACTIONS(1313),
    [aux_sym_command_name_token7] = ACTIONS(1313),
    [aux_sym_command_name_token8] = ACTIONS(1313),
    [aux_sym_command_name_token9] = ACTIONS(1313),
    [aux_sym_command_name_token10] = ACTIONS(1313),
    [aux_sym_command_name_token11] = ACTIONS(1313),
    [aux_sym_command_name_token12] = ACTIONS(1313),
    [aux_sym_command_name_token13] = ACTIONS(1313),
    [aux_sym_command_name_token14] = ACTIONS(1313),
    [aux_sym_command_name_token15] = ACTIONS(1313),
    [aux_sym_command_name_token16] = ACTIONS(1313),
    [aux_sym_command_name_token17] = ACTIONS(1313),
    [aux_sym_command_name_token18] = ACTIONS(1313),
    [aux_sym_class_statement_token1] = ACTIONS(1313),
    [aux_sym_enum_statement_token1] = ACTIONS(1313),
    [anon_sym_PLUS] = ACTIONS(1313),
    [anon_sym_DASH] = ACTIONS(1313),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1313),
    [anon_sym_BANG] = ACTIONS(1313),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1313),
    [anon_sym_PLUS_PLUS] = ACTIONS(1313),
    [anon_sym_DASH_DASH] = ACTIONS(1313),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1313),
    [anon_sym_AT_LPAREN] = ACTIONS(1313),
    [anon_sym_AT_LBRACE] = ACTIONS(1313),
  },
  [STATE(299)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1315),
    [sym_hexadecimal_integer_literal] = ACTIONS(1315),
    [sym_real_literal] = ACTIONS(1315),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1315),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1315),
    [sym_verbatim_string_characters] = ACTIONS(1315),
    [sym_verbatim_here_string_characters] = ACTIONS(1315),
    [anon_sym_DOT] = ACTIONS(1315),
    [anon_sym_LBRACK] = ACTIONS(1315),
    [aux_sym_comparison_operator_token37] = ACTIONS(1315),
    [aux_sym_comparison_operator_token50] = ACTIONS(1315),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1315),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1315),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1315),
    [anon_sym_DOLLAR_] = ACTIONS(1315),
    [aux_sym_variable_token1] = ACTIONS(1315),
    [aux_sym_variable_token2] = ACTIONS(1315),
    [sym_braced_variable] = ACTIONS(1315),
    [anon_sym_SEMI] = ACTIONS(1315),
    [aux_sym_param_block_token1] = ACTIONS(1315),
    [anon_sym_LPAREN] = ACTIONS(1315),
    [anon_sym_COMMA] = ACTIONS(1315),
    [aux_sym_block_name_token1] = ACTIONS(1315),
    [aux_sym_block_name_token2] = ACTIONS(1315),
    [aux_sym_block_name_token3] = ACTIONS(1315),
    [aux_sym_block_name_token4] = ACTIONS(1315),
    [anon_sym_LBRACE] = ACTIONS(1315),
    [aux_sym_if_statement_token1] = ACTIONS(1315),
    [aux_sym_switch_statement_token1] = ACTIONS(1315),
    [aux_sym_foreach_statement_token1] = ACTIONS(1315),
    [aux_sym_for_statement_token1] = ACTIONS(1315),
    [aux_sym_while_statement_token1] = ACTIONS(1315),
    [aux_sym_do_statement_token1] = ACTIONS(1315),
    [aux_sym_function_statement_token1] = ACTIONS(1315),
    [aux_sym_function_statement_token2] = ACTIONS(1315),
    [aux_sym_function_statement_token3] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1315),
    [sym_label] = ACTIONS(1315),
    [aux_sym_trap_statement_token1] = ACTIONS(1315),
    [aux_sym_try_statement_token1] = ACTIONS(1315),
    [aux_sym_data_statement_token1] = ACTIONS(1315),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1315),
    [aux_sym_parallel_statement_token1] = ACTIONS(1315),
    [aux_sym_sequence_statement_token1] = ACTIONS(1315),
    [anon_sym_AMP] = ACTIONS(1315),
    [aux_sym_command_name_token1] = ACTIONS(1315),
    [aux_sym_command_name_token2] = ACTIONS(1315),
    [aux_sym_command_name_token3] = ACTIONS(1315),
    [aux_sym_command_name_token4] = ACTIONS(1315),
    [aux_sym_command_name_token5] = ACTIONS(1315),
    [aux_sym_command_name_token6] = ACTIONS(1315),
    [aux_sym_command_name_token7] = ACTIONS(1315),
    [aux_sym_command_name_token8] = ACTIONS(1315),
    [aux_sym_command_name_token9] = ACTIONS(1315),
    [aux_sym_command_name_token10] = ACTIONS(1315),
    [aux_sym_command_name_token11] = ACTIONS(1315),
    [aux_sym_command_name_token12] = ACTIONS(1315),
    [aux_sym_command_name_token13] = ACTIONS(1315),
    [aux_sym_command_name_token14] = ACTIONS(1315),
    [aux_sym_command_name_token15] = ACTIONS(1315),
    [aux_sym_command_name_token16] = ACTIONS(1315),
    [aux_sym_command_name_token17] = ACTIONS(1315),
    [aux_sym_command_name_token18] = ACTIONS(1315),
    [aux_sym_class_statement_token1] = ACTIONS(1315),
    [aux_sym_enum_statement_token1] = ACTIONS(1315),
    [anon_sym_PLUS] = ACTIONS(1315),
    [anon_sym_DASH] = ACTIONS(1315),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1315),
    [anon_sym_BANG] = ACTIONS(1315),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1315),
    [anon_sym_PLUS_PLUS] = ACTIONS(1315),
    [anon_sym_DASH_DASH] = ACTIONS(1315),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1315),
    [anon_sym_AT_LPAREN] = ACTIONS(1315),
    [anon_sym_AT_LBRACE] = ACTIONS(1315),
    [sym__statement_terminator] = ACTIONS(1317),
  },
  [STATE(300)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1319),
    [sym_hexadecimal_integer_literal] = ACTIONS(1319),
    [sym_real_literal] = ACTIONS(1319),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1319),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1319),
    [sym_verbatim_string_characters] = ACTIONS(1319),
    [sym_verbatim_here_string_characters] = ACTIONS(1319),
    [anon_sym_DOT] = ACTIONS(1319),
    [anon_sym_LBRACK] = ACTIONS(1319),
    [aux_sym_comparison_operator_token37] = ACTIONS(1319),
    [aux_sym_comparison_operator_token50] = ACTIONS(1319),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1319),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1319),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1319),
    [anon_sym_DOLLAR_] = ACTIONS(1319),
    [aux_sym_variable_token1] = ACTIONS(1319),
    [aux_sym_variable_token2] = ACTIONS(1319),
    [sym_braced_variable] = ACTIONS(1319),
    [anon_sym_SEMI] = ACTIONS(1319),
    [aux_sym_param_block_token1] = ACTIONS(1319),
    [anon_sym_LPAREN] = ACTIONS(1319),
    [anon_sym_COMMA] = ACTIONS(1319),
    [aux_sym_block_name_token1] = ACTIONS(1319),
    [aux_sym_block_name_token2] = ACTIONS(1319),
    [aux_sym_block_name_token3] = ACTIONS(1319),
    [aux_sym_block_name_token4] = ACTIONS(1319),
    [anon_sym_LBRACE] = ACTIONS(1319),
    [aux_sym_if_statement_token1] = ACTIONS(1319),
    [aux_sym_switch_statement_token1] = ACTIONS(1319),
    [aux_sym_foreach_statement_token1] = ACTIONS(1319),
    [aux_sym_for_statement_token1] = ACTIONS(1319),
    [aux_sym_while_statement_token1] = ACTIONS(1319),
    [aux_sym_do_statement_token1] = ACTIONS(1319),
    [aux_sym_function_statement_token1] = ACTIONS(1319),
    [aux_sym_function_statement_token2] = ACTIONS(1319),
    [aux_sym_function_statement_token3] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1319),
    [sym_label] = ACTIONS(1319),
    [aux_sym_trap_statement_token1] = ACTIONS(1319),
    [aux_sym_try_statement_token1] = ACTIONS(1319),
    [aux_sym_data_statement_token1] = ACTIONS(1319),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1319),
    [aux_sym_parallel_statement_token1] = ACTIONS(1319),
    [aux_sym_sequence_statement_token1] = ACTIONS(1319),
    [anon_sym_AMP] = ACTIONS(1319),
    [aux_sym_command_name_token1] = ACTIONS(1319),
    [aux_sym_command_name_token2] = ACTIONS(1319),
    [aux_sym_command_name_token3] = ACTIONS(1319),
    [aux_sym_command_name_token4] = ACTIONS(1319),
    [aux_sym_command_name_token5] = ACTIONS(1319),
    [aux_sym_command_name_token6] = ACTIONS(1319),
    [aux_sym_command_name_token7] = ACTIONS(1319),
    [aux_sym_command_name_token8] = ACTIONS(1319),
    [aux_sym_command_name_token9] = ACTIONS(1319),
    [aux_sym_command_name_token10] = ACTIONS(1319),
    [aux_sym_command_name_token11] = ACTIONS(1319),
    [aux_sym_command_name_token12] = ACTIONS(1319),
    [aux_sym_command_name_token13] = ACTIONS(1319),
    [aux_sym_command_name_token14] = ACTIONS(1319),
    [aux_sym_command_name_token15] = ACTIONS(1319),
    [aux_sym_command_name_token16] = ACTIONS(1319),
    [aux_sym_command_name_token17] = ACTIONS(1319),
    [aux_sym_command_name_token18] = ACTIONS(1319),
    [aux_sym_class_statement_token1] = ACTIONS(1319),
    [aux_sym_enum_statement_token1] = ACTIONS(1319),
    [anon_sym_PLUS] = ACTIONS(1319),
    [anon_sym_DASH] = ACTIONS(1319),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1319),
    [anon_sym_BANG] = ACTIONS(1319),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1319),
    [anon_sym_PLUS_PLUS] = ACTIONS(1319),
    [anon_sym_DASH_DASH] = ACTIONS(1319),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1319),
    [anon_sym_AT_LPAREN] = ACTIONS(1319),
    [anon_sym_AT_LBRACE] = ACTIONS(1319),
    [sym__statement_terminator] = ACTIONS(1321),
  },
  [STATE(301)] = {
    [sym_elseif_clause] = STATE(301),
    [aux_sym_elseif_clauses_repeat1] = STATE(301),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1323),
    [sym_hexadecimal_integer_literal] = ACTIONS(1323),
    [sym_real_literal] = ACTIONS(1323),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1323),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1323),
    [sym_verbatim_string_characters] = ACTIONS(1323),
    [sym_verbatim_here_string_characters] = ACTIONS(1323),
    [anon_sym_DOT] = ACTIONS(1323),
    [anon_sym_LBRACK] = ACTIONS(1323),
    [aux_sym_comparison_operator_token37] = ACTIONS(1323),
    [aux_sym_comparison_operator_token50] = ACTIONS(1323),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1323),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1323),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1323),
    [anon_sym_DOLLAR_] = ACTIONS(1323),
    [aux_sym_variable_token1] = ACTIONS(1323),
    [aux_sym_variable_token2] = ACTIONS(1323),
    [sym_braced_variable] = ACTIONS(1323),
    [anon_sym_SEMI] = ACTIONS(1323),
    [anon_sym_LPAREN] = ACTIONS(1323),
    [anon_sym_RPAREN] = ACTIONS(1323),
    [anon_sym_COMMA] = ACTIONS(1323),
    [anon_sym_LBRACE] = ACTIONS(1323),
    [anon_sym_RBRACE] = ACTIONS(1323),
    [aux_sym_if_statement_token1] = ACTIONS(1323),
    [aux_sym_elseif_clause_token1] = ACTIONS(1325),
    [aux_sym_else_clause_token1] = ACTIONS(1323),
    [aux_sym_switch_statement_token1] = ACTIONS(1323),
    [aux_sym_foreach_statement_token1] = ACTIONS(1323),
    [aux_sym_for_statement_token1] = ACTIONS(1323),
    [aux_sym_while_statement_token1] = ACTIONS(1323),
    [aux_sym_do_statement_token1] = ACTIONS(1323),
    [aux_sym_function_statement_token1] = ACTIONS(1323),
    [aux_sym_function_statement_token2] = ACTIONS(1323),
    [aux_sym_function_statement_token3] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1323),
    [sym_label] = ACTIONS(1323),
    [aux_sym_trap_statement_token1] = ACTIONS(1323),
    [aux_sym_try_statement_token1] = ACTIONS(1323),
    [aux_sym_data_statement_token1] = ACTIONS(1323),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1323),
    [aux_sym_parallel_statement_token1] = ACTIONS(1323),
    [aux_sym_sequence_statement_token1] = ACTIONS(1323),
    [anon_sym_AMP] = ACTIONS(1323),
    [aux_sym_command_name_token1] = ACTIONS(1323),
    [aux_sym_command_name_token2] = ACTIONS(1323),
    [aux_sym_command_name_token3] = ACTIONS(1323),
    [aux_sym_command_name_token4] = ACTIONS(1323),
    [aux_sym_command_name_token5] = ACTIONS(1323),
    [aux_sym_command_name_token6] = ACTIONS(1323),
    [aux_sym_command_name_token7] = ACTIONS(1323),
    [aux_sym_command_name_token8] = ACTIONS(1323),
    [aux_sym_command_name_token9] = ACTIONS(1323),
    [aux_sym_command_name_token10] = ACTIONS(1323),
    [aux_sym_command_name_token11] = ACTIONS(1323),
    [aux_sym_command_name_token12] = ACTIONS(1323),
    [aux_sym_command_name_token13] = ACTIONS(1323),
    [aux_sym_command_name_token14] = ACTIONS(1323),
    [aux_sym_command_name_token15] = ACTIONS(1323),
    [aux_sym_command_name_token16] = ACTIONS(1323),
    [aux_sym_command_name_token17] = ACTIONS(1323),
    [aux_sym_command_name_token18] = ACTIONS(1323),
    [aux_sym_class_statement_token1] = ACTIONS(1323),
    [aux_sym_enum_statement_token1] = ACTIONS(1323),
    [anon_sym_PLUS] = ACTIONS(1323),
    [anon_sym_DASH] = ACTIONS(1323),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1323),
    [anon_sym_BANG] = ACTIONS(1323),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1323),
    [anon_sym_PLUS_PLUS] = ACTIONS(1323),
    [anon_sym_DASH_DASH] = ACTIONS(1323),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1323),
    [anon_sym_AT_LPAREN] = ACTIONS(1323),
    [anon_sym_AT_LBRACE] = ACTIONS(1323),
  },
  [STATE(302)] = {
    [aux_sym_script_block_repeat1] = STATE(302),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1328),
    [sym_hexadecimal_integer_literal] = ACTIONS(1328),
    [sym_real_literal] = ACTIONS(1328),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1328),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1328),
    [sym_verbatim_string_characters] = ACTIONS(1328),
    [sym_verbatim_here_string_characters] = ACTIONS(1328),
    [anon_sym_DOT] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1328),
    [aux_sym_comparison_operator_token37] = ACTIONS(1328),
    [aux_sym_comparison_operator_token50] = ACTIONS(1328),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1328),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1328),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1328),
    [anon_sym_DOLLAR_] = ACTIONS(1328),
    [aux_sym_variable_token1] = ACTIONS(1328),
    [aux_sym_variable_token2] = ACTIONS(1328),
    [sym_braced_variable] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1330),
    [anon_sym_LPAREN] = ACTIONS(1328),
    [anon_sym_COMMA] = ACTIONS(1328),
    [aux_sym_block_name_token1] = ACTIONS(1328),
    [aux_sym_block_name_token2] = ACTIONS(1328),
    [aux_sym_block_name_token3] = ACTIONS(1328),
    [aux_sym_block_name_token4] = ACTIONS(1328),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_RBRACE] = ACTIONS(1328),
    [aux_sym_if_statement_token1] = ACTIONS(1328),
    [aux_sym_switch_statement_token1] = ACTIONS(1328),
    [aux_sym_foreach_statement_token1] = ACTIONS(1328),
    [aux_sym_for_statement_token1] = ACTIONS(1328),
    [aux_sym_while_statement_token1] = ACTIONS(1328),
    [aux_sym_do_statement_token1] = ACTIONS(1328),
    [aux_sym_function_statement_token1] = ACTIONS(1328),
    [aux_sym_function_statement_token2] = ACTIONS(1328),
    [aux_sym_function_statement_token3] = ACTIONS(1328),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1328),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1328),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1328),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1328),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1328),
    [sym_label] = ACTIONS(1328),
    [aux_sym_trap_statement_token1] = ACTIONS(1328),
    [aux_sym_try_statement_token1] = ACTIONS(1328),
    [aux_sym_data_statement_token1] = ACTIONS(1328),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1328),
    [aux_sym_parallel_statement_token1] = ACTIONS(1328),
    [aux_sym_sequence_statement_token1] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [aux_sym_command_name_token1] = ACTIONS(1328),
    [aux_sym_command_name_token2] = ACTIONS(1328),
    [aux_sym_command_name_token3] = ACTIONS(1328),
    [aux_sym_command_name_token4] = ACTIONS(1328),
    [aux_sym_command_name_token5] = ACTIONS(1328),
    [aux_sym_command_name_token6] = ACTIONS(1328),
    [aux_sym_command_name_token7] = ACTIONS(1328),
    [aux_sym_command_name_token8] = ACTIONS(1328),
    [aux_sym_command_name_token9] = ACTIONS(1328),
    [aux_sym_command_name_token10] = ACTIONS(1328),
    [aux_sym_command_name_token11] = ACTIONS(1328),
    [aux_sym_command_name_token12] = ACTIONS(1328),
    [aux_sym_command_name_token13] = ACTIONS(1328),
    [aux_sym_command_name_token14] = ACTIONS(1328),
    [aux_sym_command_name_token15] = ACTIONS(1328),
    [aux_sym_command_name_token16] = ACTIONS(1328),
    [aux_sym_command_name_token17] = ACTIONS(1328),
    [aux_sym_command_name_token18] = ACTIONS(1328),
    [aux_sym_class_statement_token1] = ACTIONS(1328),
    [aux_sym_enum_statement_token1] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1328),
    [anon_sym_AT_LPAREN] = ACTIONS(1328),
    [anon_sym_AT_LBRACE] = ACTIONS(1328),
  },
  [STATE(303)] = {
    [sym_catch_clause] = STATE(297),
    [aux_sym_catch_clauses_repeat1] = STATE(297),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1333),
    [sym_hexadecimal_integer_literal] = ACTIONS(1333),
    [sym_real_literal] = ACTIONS(1333),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1333),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1333),
    [sym_verbatim_string_characters] = ACTIONS(1333),
    [sym_verbatim_here_string_characters] = ACTIONS(1333),
    [anon_sym_DOT] = ACTIONS(1333),
    [anon_sym_LBRACK] = ACTIONS(1333),
    [aux_sym_comparison_operator_token37] = ACTIONS(1333),
    [aux_sym_comparison_operator_token50] = ACTIONS(1333),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1333),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1333),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1333),
    [anon_sym_DOLLAR_] = ACTIONS(1333),
    [aux_sym_variable_token1] = ACTIONS(1333),
    [aux_sym_variable_token2] = ACTIONS(1333),
    [sym_braced_variable] = ACTIONS(1333),
    [anon_sym_SEMI] = ACTIONS(1333),
    [anon_sym_LPAREN] = ACTIONS(1333),
    [anon_sym_RPAREN] = ACTIONS(1333),
    [anon_sym_COMMA] = ACTIONS(1333),
    [anon_sym_LBRACE] = ACTIONS(1333),
    [anon_sym_RBRACE] = ACTIONS(1333),
    [aux_sym_if_statement_token1] = ACTIONS(1333),
    [aux_sym_switch_statement_token1] = ACTIONS(1333),
    [aux_sym_foreach_statement_token1] = ACTIONS(1333),
    [aux_sym_for_statement_token1] = ACTIONS(1333),
    [aux_sym_while_statement_token1] = ACTIONS(1333),
    [aux_sym_do_statement_token1] = ACTIONS(1333),
    [aux_sym_function_statement_token1] = ACTIONS(1333),
    [aux_sym_function_statement_token2] = ACTIONS(1333),
    [aux_sym_function_statement_token3] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1333),
    [sym_label] = ACTIONS(1333),
    [aux_sym_trap_statement_token1] = ACTIONS(1333),
    [aux_sym_try_statement_token1] = ACTIONS(1333),
    [aux_sym_catch_clause_token1] = ACTIONS(1292),
    [aux_sym_finally_clause_token1] = ACTIONS(1333),
    [aux_sym_data_statement_token1] = ACTIONS(1333),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1333),
    [aux_sym_parallel_statement_token1] = ACTIONS(1333),
    [aux_sym_sequence_statement_token1] = ACTIONS(1333),
    [anon_sym_AMP] = ACTIONS(1333),
    [aux_sym_command_name_token1] = ACTIONS(1333),
    [aux_sym_command_name_token2] = ACTIONS(1333),
    [aux_sym_command_name_token3] = ACTIONS(1333),
    [aux_sym_command_name_token4] = ACTIONS(1333),
    [aux_sym_command_name_token5] = ACTIONS(1333),
    [aux_sym_command_name_token6] = ACTIONS(1333),
    [aux_sym_command_name_token7] = ACTIONS(1333),
    [aux_sym_command_name_token8] = ACTIONS(1333),
    [aux_sym_command_name_token9] = ACTIONS(1333),
    [aux_sym_command_name_token10] = ACTIONS(1333),
    [aux_sym_command_name_token11] = ACTIONS(1333),
    [aux_sym_command_name_token12] = ACTIONS(1333),
    [aux_sym_command_name_token13] = ACTIONS(1333),
    [aux_sym_command_name_token14] = ACTIONS(1333),
    [aux_sym_command_name_token15] = ACTIONS(1333),
    [aux_sym_command_name_token16] = ACTIONS(1333),
    [aux_sym_command_name_token17] = ACTIONS(1333),
    [aux_sym_command_name_token18] = ACTIONS(1333),
    [aux_sym_class_statement_token1] = ACTIONS(1333),
    [aux_sym_enum_statement_token1] = ACTIONS(1333),
    [anon_sym_PLUS] = ACTIONS(1333),
    [anon_sym_DASH] = ACTIONS(1333),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1333),
    [anon_sym_BANG] = ACTIONS(1333),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1333),
    [anon_sym_PLUS_PLUS] = ACTIONS(1333),
    [anon_sym_DASH_DASH] = ACTIONS(1333),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1333),
    [anon_sym_AT_LPAREN] = ACTIONS(1333),
    [anon_sym_AT_LBRACE] = ACTIONS(1333),
  },
  [STATE(304)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1335),
    [sym_hexadecimal_integer_literal] = ACTIONS(1335),
    [sym_real_literal] = ACTIONS(1335),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1335),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1335),
    [sym_verbatim_string_characters] = ACTIONS(1335),
    [sym_verbatim_here_string_characters] = ACTIONS(1335),
    [anon_sym_DOT] = ACTIONS(1335),
    [anon_sym_LBRACK] = ACTIONS(1335),
    [aux_sym_comparison_operator_token37] = ACTIONS(1335),
    [aux_sym_comparison_operator_token50] = ACTIONS(1335),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1335),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1335),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1335),
    [anon_sym_DOLLAR_] = ACTIONS(1335),
    [aux_sym_variable_token1] = ACTIONS(1335),
    [aux_sym_variable_token2] = ACTIONS(1335),
    [sym_braced_variable] = ACTIONS(1335),
    [anon_sym_SEMI] = ACTIONS(1335),
    [aux_sym_param_block_token1] = ACTIONS(1335),
    [anon_sym_LPAREN] = ACTIONS(1335),
    [anon_sym_COMMA] = ACTIONS(1335),
    [aux_sym_block_name_token1] = ACTIONS(1335),
    [aux_sym_block_name_token2] = ACTIONS(1335),
    [aux_sym_block_name_token3] = ACTIONS(1335),
    [aux_sym_block_name_token4] = ACTIONS(1335),
    [anon_sym_LBRACE] = ACTIONS(1335),
    [aux_sym_if_statement_token1] = ACTIONS(1335),
    [aux_sym_switch_statement_token1] = ACTIONS(1335),
    [aux_sym_foreach_statement_token1] = ACTIONS(1335),
    [aux_sym_for_statement_token1] = ACTIONS(1335),
    [aux_sym_while_statement_token1] = ACTIONS(1335),
    [aux_sym_do_statement_token1] = ACTIONS(1335),
    [aux_sym_function_statement_token1] = ACTIONS(1335),
    [aux_sym_function_statement_token2] = ACTIONS(1335),
    [aux_sym_function_statement_token3] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1335),
    [sym_label] = ACTIONS(1335),
    [aux_sym_trap_statement_token1] = ACTIONS(1335),
    [aux_sym_try_statement_token1] = ACTIONS(1335),
    [aux_sym_data_statement_token1] = ACTIONS(1335),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1335),
    [aux_sym_parallel_statement_token1] = ACTIONS(1335),
    [aux_sym_sequence_statement_token1] = ACTIONS(1335),
    [anon_sym_AMP] = ACTIONS(1335),
    [aux_sym_command_name_token1] = ACTIONS(1335),
    [aux_sym_command_name_token2] = ACTIONS(1335),
    [aux_sym_command_name_token3] = ACTIONS(1335),
    [aux_sym_command_name_token4] = ACTIONS(1335),
    [aux_sym_command_name_token5] = ACTIONS(1335),
    [aux_sym_command_name_token6] = ACTIONS(1335),
    [aux_sym_command_name_token7] = ACTIONS(1335),
    [aux_sym_command_name_token8] = ACTIONS(1335),
    [aux_sym_command_name_token9] = ACTIONS(1335),
    [aux_sym_command_name_token10] = ACTIONS(1335),
    [aux_sym_command_name_token11] = ACTIONS(1335),
    [aux_sym_command_name_token12] = ACTIONS(1335),
    [aux_sym_command_name_token13] = ACTIONS(1335),
    [aux_sym_command_name_token14] = ACTIONS(1335),
    [aux_sym_command_name_token15] = ACTIONS(1335),
    [aux_sym_command_name_token16] = ACTIONS(1335),
    [aux_sym_command_name_token17] = ACTIONS(1335),
    [aux_sym_command_name_token18] = ACTIONS(1335),
    [aux_sym_class_statement_token1] = ACTIONS(1335),
    [aux_sym_enum_statement_token1] = ACTIONS(1335),
    [anon_sym_PLUS] = ACTIONS(1335),
    [anon_sym_DASH] = ACTIONS(1335),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1335),
    [anon_sym_BANG] = ACTIONS(1335),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1335),
    [anon_sym_PLUS_PLUS] = ACTIONS(1335),
    [anon_sym_DASH_DASH] = ACTIONS(1335),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1335),
    [anon_sym_AT_LPAREN] = ACTIONS(1335),
    [anon_sym_AT_LBRACE] = ACTIONS(1335),
    [sym__statement_terminator] = ACTIONS(1337),
  },
  [STATE(305)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1339),
    [sym_hexadecimal_integer_literal] = ACTIONS(1339),
    [sym_real_literal] = ACTIONS(1339),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1339),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1339),
    [sym_verbatim_string_characters] = ACTIONS(1339),
    [sym_verbatim_here_string_characters] = ACTIONS(1339),
    [anon_sym_DOT] = ACTIONS(1339),
    [anon_sym_LBRACK] = ACTIONS(1339),
    [aux_sym_comparison_operator_token37] = ACTIONS(1339),
    [aux_sym_comparison_operator_token50] = ACTIONS(1339),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1339),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1339),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1339),
    [anon_sym_DOLLAR_] = ACTIONS(1339),
    [aux_sym_variable_token1] = ACTIONS(1339),
    [aux_sym_variable_token2] = ACTIONS(1339),
    [sym_braced_variable] = ACTIONS(1339),
    [anon_sym_SEMI] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(1339),
    [anon_sym_COMMA] = ACTIONS(1339),
    [aux_sym_block_name_token1] = ACTIONS(1341),
    [aux_sym_block_name_token2] = ACTIONS(1341),
    [aux_sym_block_name_token3] = ACTIONS(1341),
    [aux_sym_block_name_token4] = ACTIONS(1341),
    [anon_sym_LBRACE] = ACTIONS(1339),
    [anon_sym_RBRACE] = ACTIONS(1339),
    [aux_sym_if_statement_token1] = ACTIONS(1339),
    [aux_sym_switch_statement_token1] = ACTIONS(1339),
    [aux_sym_foreach_statement_token1] = ACTIONS(1339),
    [aux_sym_for_statement_token1] = ACTIONS(1339),
    [aux_sym_while_statement_token1] = ACTIONS(1339),
    [aux_sym_do_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token2] = ACTIONS(1339),
    [aux_sym_function_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1339),
    [sym_label] = ACTIONS(1339),
    [aux_sym_trap_statement_token1] = ACTIONS(1339),
    [aux_sym_try_statement_token1] = ACTIONS(1339),
    [aux_sym_data_statement_token1] = ACTIONS(1339),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1339),
    [aux_sym_parallel_statement_token1] = ACTIONS(1339),
    [aux_sym_sequence_statement_token1] = ACTIONS(1339),
    [anon_sym_AMP] = ACTIONS(1339),
    [aux_sym_command_name_token1] = ACTIONS(1339),
    [aux_sym_command_name_token2] = ACTIONS(1339),
    [aux_sym_command_name_token3] = ACTIONS(1339),
    [aux_sym_command_name_token4] = ACTIONS(1339),
    [aux_sym_command_name_token5] = ACTIONS(1339),
    [aux_sym_command_name_token6] = ACTIONS(1339),
    [aux_sym_command_name_token7] = ACTIONS(1339),
    [aux_sym_command_name_token8] = ACTIONS(1339),
    [aux_sym_command_name_token9] = ACTIONS(1339),
    [aux_sym_command_name_token10] = ACTIONS(1339),
    [aux_sym_command_name_token11] = ACTIONS(1339),
    [aux_sym_command_name_token12] = ACTIONS(1339),
    [aux_sym_command_name_token13] = ACTIONS(1339),
    [aux_sym_command_name_token14] = ACTIONS(1339),
    [aux_sym_command_name_token15] = ACTIONS(1339),
    [aux_sym_command_name_token16] = ACTIONS(1339),
    [aux_sym_command_name_token17] = ACTIONS(1339),
    [aux_sym_command_name_token18] = ACTIONS(1339),
    [aux_sym_class_statement_token1] = ACTIONS(1339),
    [aux_sym_enum_statement_token1] = ACTIONS(1339),
    [anon_sym_PLUS] = ACTIONS(1339),
    [anon_sym_DASH] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1339),
    [anon_sym_BANG] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1339),
    [anon_sym_PLUS_PLUS] = ACTIONS(1339),
    [anon_sym_DASH_DASH] = ACTIONS(1339),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LBRACE] = ACTIONS(1339),
  },
  [STATE(306)] = {
    [sym_catch_clause] = STATE(306),
    [aux_sym_catch_clauses_repeat1] = STATE(306),
    [ts_builtin_sym_end] = ACTIONS(1343),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1308),
    [sym_hexadecimal_integer_literal] = ACTIONS(1308),
    [sym_real_literal] = ACTIONS(1308),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1308),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1308),
    [sym_verbatim_string_characters] = ACTIONS(1308),
    [sym_verbatim_here_string_characters] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1308),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_comparison_operator_token37] = ACTIONS(1308),
    [aux_sym_comparison_operator_token50] = ACTIONS(1308),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1308),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1308),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_] = ACTIONS(1308),
    [aux_sym_variable_token1] = ACTIONS(1308),
    [aux_sym_variable_token2] = ACTIONS(1308),
    [sym_braced_variable] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [aux_sym_if_statement_token1] = ACTIONS(1308),
    [aux_sym_switch_statement_token1] = ACTIONS(1308),
    [aux_sym_foreach_statement_token1] = ACTIONS(1308),
    [aux_sym_for_statement_token1] = ACTIONS(1308),
    [aux_sym_while_statement_token1] = ACTIONS(1308),
    [aux_sym_do_statement_token1] = ACTIONS(1308),
    [aux_sym_function_statement_token1] = ACTIONS(1308),
    [aux_sym_function_statement_token2] = ACTIONS(1308),
    [aux_sym_function_statement_token3] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1308),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1308),
    [sym_label] = ACTIONS(1308),
    [aux_sym_trap_statement_token1] = ACTIONS(1308),
    [aux_sym_try_statement_token1] = ACTIONS(1308),
    [aux_sym_catch_clause_token1] = ACTIONS(1345),
    [aux_sym_finally_clause_token1] = ACTIONS(1308),
    [aux_sym_data_statement_token1] = ACTIONS(1308),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1308),
    [aux_sym_parallel_statement_token1] = ACTIONS(1308),
    [aux_sym_sequence_statement_token1] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [aux_sym_command_name_token1] = ACTIONS(1308),
    [aux_sym_command_name_token2] = ACTIONS(1308),
    [aux_sym_command_name_token3] = ACTIONS(1308),
    [aux_sym_command_name_token4] = ACTIONS(1308),
    [aux_sym_command_name_token5] = ACTIONS(1308),
    [aux_sym_command_name_token6] = ACTIONS(1308),
    [aux_sym_command_name_token7] = ACTIONS(1308),
    [aux_sym_command_name_token8] = ACTIONS(1308),
    [aux_sym_command_name_token9] = ACTIONS(1308),
    [aux_sym_command_name_token10] = ACTIONS(1308),
    [aux_sym_command_name_token11] = ACTIONS(1308),
    [aux_sym_command_name_token12] = ACTIONS(1308),
    [aux_sym_command_name_token13] = ACTIONS(1308),
    [aux_sym_command_name_token14] = ACTIONS(1308),
    [aux_sym_command_name_token15] = ACTIONS(1308),
    [aux_sym_command_name_token16] = ACTIONS(1308),
    [aux_sym_command_name_token17] = ACTIONS(1308),
    [aux_sym_command_name_token18] = ACTIONS(1308),
    [aux_sym_class_statement_token1] = ACTIONS(1308),
    [aux_sym_enum_statement_token1] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_AT_LPAREN] = ACTIONS(1308),
    [anon_sym_AT_LBRACE] = ACTIONS(1308),
  },
  [STATE(307)] = {
    [sym_elseif_clause] = STATE(308),
    [aux_sym_elseif_clauses_repeat1] = STATE(308),
    [ts_builtin_sym_end] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1313),
    [sym_hexadecimal_integer_literal] = ACTIONS(1313),
    [sym_real_literal] = ACTIONS(1313),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1313),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1313),
    [sym_verbatim_string_characters] = ACTIONS(1313),
    [sym_verbatim_here_string_characters] = ACTIONS(1313),
    [anon_sym_DOT] = ACTIONS(1313),
    [anon_sym_LBRACK] = ACTIONS(1313),
    [aux_sym_comparison_operator_token37] = ACTIONS(1313),
    [aux_sym_comparison_operator_token50] = ACTIONS(1313),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1313),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1313),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1313),
    [anon_sym_DOLLAR_] = ACTIONS(1313),
    [aux_sym_variable_token1] = ACTIONS(1313),
    [aux_sym_variable_token2] = ACTIONS(1313),
    [sym_braced_variable] = ACTIONS(1313),
    [anon_sym_SEMI] = ACTIONS(1313),
    [anon_sym_LPAREN] = ACTIONS(1313),
    [anon_sym_COMMA] = ACTIONS(1313),
    [anon_sym_LBRACE] = ACTIONS(1313),
    [aux_sym_if_statement_token1] = ACTIONS(1313),
    [aux_sym_elseif_clause_token1] = ACTIONS(1304),
    [aux_sym_else_clause_token1] = ACTIONS(1313),
    [aux_sym_switch_statement_token1] = ACTIONS(1313),
    [aux_sym_foreach_statement_token1] = ACTIONS(1313),
    [aux_sym_for_statement_token1] = ACTIONS(1313),
    [aux_sym_while_statement_token1] = ACTIONS(1313),
    [aux_sym_do_statement_token1] = ACTIONS(1313),
    [aux_sym_function_statement_token1] = ACTIONS(1313),
    [aux_sym_function_statement_token2] = ACTIONS(1313),
    [aux_sym_function_statement_token3] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1313),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1313),
    [sym_label] = ACTIONS(1313),
    [aux_sym_trap_statement_token1] = ACTIONS(1313),
    [aux_sym_try_statement_token1] = ACTIONS(1313),
    [aux_sym_data_statement_token1] = ACTIONS(1313),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1313),
    [aux_sym_parallel_statement_token1] = ACTIONS(1313),
    [aux_sym_sequence_statement_token1] = ACTIONS(1313),
    [anon_sym_AMP] = ACTIONS(1313),
    [aux_sym_command_name_token1] = ACTIONS(1313),
    [aux_sym_command_name_token2] = ACTIONS(1313),
    [aux_sym_command_name_token3] = ACTIONS(1313),
    [aux_sym_command_name_token4] = ACTIONS(1313),
    [aux_sym_command_name_token5] = ACTIONS(1313),
    [aux_sym_command_name_token6] = ACTIONS(1313),
    [aux_sym_command_name_token7] = ACTIONS(1313),
    [aux_sym_command_name_token8] = ACTIONS(1313),
    [aux_sym_command_name_token9] = ACTIONS(1313),
    [aux_sym_command_name_token10] = ACTIONS(1313),
    [aux_sym_command_name_token11] = ACTIONS(1313),
    [aux_sym_command_name_token12] = ACTIONS(1313),
    [aux_sym_command_name_token13] = ACTIONS(1313),
    [aux_sym_command_name_token14] = ACTIONS(1313),
    [aux_sym_command_name_token15] = ACTIONS(1313),
    [aux_sym_command_name_token16] = ACTIONS(1313),
    [aux_sym_command_name_token17] = ACTIONS(1313),
    [aux_sym_command_name_token18] = ACTIONS(1313),
    [aux_sym_class_statement_token1] = ACTIONS(1313),
    [aux_sym_enum_statement_token1] = ACTIONS(1313),
    [anon_sym_PLUS] = ACTIONS(1313),
    [anon_sym_DASH] = ACTIONS(1313),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1313),
    [anon_sym_BANG] = ACTIONS(1313),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1313),
    [anon_sym_PLUS_PLUS] = ACTIONS(1313),
    [anon_sym_DASH_DASH] = ACTIONS(1313),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1313),
    [anon_sym_AT_LPAREN] = ACTIONS(1313),
    [anon_sym_AT_LBRACE] = ACTIONS(1313),
  },
  [STATE(308)] = {
    [sym_elseif_clause] = STATE(308),
    [aux_sym_elseif_clauses_repeat1] = STATE(308),
    [ts_builtin_sym_end] = ACTIONS(1350),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1323),
    [sym_hexadecimal_integer_literal] = ACTIONS(1323),
    [sym_real_literal] = ACTIONS(1323),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1323),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1323),
    [sym_verbatim_string_characters] = ACTIONS(1323),
    [sym_verbatim_here_string_characters] = ACTIONS(1323),
    [anon_sym_DOT] = ACTIONS(1323),
    [anon_sym_LBRACK] = ACTIONS(1323),
    [aux_sym_comparison_operator_token37] = ACTIONS(1323),
    [aux_sym_comparison_operator_token50] = ACTIONS(1323),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1323),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1323),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1323),
    [anon_sym_DOLLAR_] = ACTIONS(1323),
    [aux_sym_variable_token1] = ACTIONS(1323),
    [aux_sym_variable_token2] = ACTIONS(1323),
    [sym_braced_variable] = ACTIONS(1323),
    [anon_sym_SEMI] = ACTIONS(1323),
    [anon_sym_LPAREN] = ACTIONS(1323),
    [anon_sym_COMMA] = ACTIONS(1323),
    [anon_sym_LBRACE] = ACTIONS(1323),
    [aux_sym_if_statement_token1] = ACTIONS(1323),
    [aux_sym_elseif_clause_token1] = ACTIONS(1352),
    [aux_sym_else_clause_token1] = ACTIONS(1323),
    [aux_sym_switch_statement_token1] = ACTIONS(1323),
    [aux_sym_foreach_statement_token1] = ACTIONS(1323),
    [aux_sym_for_statement_token1] = ACTIONS(1323),
    [aux_sym_while_statement_token1] = ACTIONS(1323),
    [aux_sym_do_statement_token1] = ACTIONS(1323),
    [aux_sym_function_statement_token1] = ACTIONS(1323),
    [aux_sym_function_statement_token2] = ACTIONS(1323),
    [aux_sym_function_statement_token3] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1323),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1323),
    [sym_label] = ACTIONS(1323),
    [aux_sym_trap_statement_token1] = ACTIONS(1323),
    [aux_sym_try_statement_token1] = ACTIONS(1323),
    [aux_sym_data_statement_token1] = ACTIONS(1323),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1323),
    [aux_sym_parallel_statement_token1] = ACTIONS(1323),
    [aux_sym_sequence_statement_token1] = ACTIONS(1323),
    [anon_sym_AMP] = ACTIONS(1323),
    [aux_sym_command_name_token1] = ACTIONS(1323),
    [aux_sym_command_name_token2] = ACTIONS(1323),
    [aux_sym_command_name_token3] = ACTIONS(1323),
    [aux_sym_command_name_token4] = ACTIONS(1323),
    [aux_sym_command_name_token5] = ACTIONS(1323),
    [aux_sym_command_name_token6] = ACTIONS(1323),
    [aux_sym_command_name_token7] = ACTIONS(1323),
    [aux_sym_command_name_token8] = ACTIONS(1323),
    [aux_sym_command_name_token9] = ACTIONS(1323),
    [aux_sym_command_name_token10] = ACTIONS(1323),
    [aux_sym_command_name_token11] = ACTIONS(1323),
    [aux_sym_command_name_token12] = ACTIONS(1323),
    [aux_sym_command_name_token13] = ACTIONS(1323),
    [aux_sym_command_name_token14] = ACTIONS(1323),
    [aux_sym_command_name_token15] = ACTIONS(1323),
    [aux_sym_command_name_token16] = ACTIONS(1323),
    [aux_sym_command_name_token17] = ACTIONS(1323),
    [aux_sym_command_name_token18] = ACTIONS(1323),
    [aux_sym_class_statement_token1] = ACTIONS(1323),
    [aux_sym_enum_statement_token1] = ACTIONS(1323),
    [anon_sym_PLUS] = ACTIONS(1323),
    [anon_sym_DASH] = ACTIONS(1323),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1323),
    [anon_sym_BANG] = ACTIONS(1323),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1323),
    [anon_sym_PLUS_PLUS] = ACTIONS(1323),
    [anon_sym_DASH_DASH] = ACTIONS(1323),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1323),
    [anon_sym_AT_LPAREN] = ACTIONS(1323),
    [anon_sym_AT_LBRACE] = ACTIONS(1323),
  },
  [STATE(309)] = {
    [sym_catch_clause] = STATE(306),
    [aux_sym_catch_clauses_repeat1] = STATE(306),
    [ts_builtin_sym_end] = ACTIONS(1355),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1333),
    [sym_hexadecimal_integer_literal] = ACTIONS(1333),
    [sym_real_literal] = ACTIONS(1333),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1333),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1333),
    [sym_verbatim_string_characters] = ACTIONS(1333),
    [sym_verbatim_here_string_characters] = ACTIONS(1333),
    [anon_sym_DOT] = ACTIONS(1333),
    [anon_sym_LBRACK] = ACTIONS(1333),
    [aux_sym_comparison_operator_token37] = ACTIONS(1333),
    [aux_sym_comparison_operator_token50] = ACTIONS(1333),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1333),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1333),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1333),
    [anon_sym_DOLLAR_] = ACTIONS(1333),
    [aux_sym_variable_token1] = ACTIONS(1333),
    [aux_sym_variable_token2] = ACTIONS(1333),
    [sym_braced_variable] = ACTIONS(1333),
    [anon_sym_SEMI] = ACTIONS(1333),
    [anon_sym_LPAREN] = ACTIONS(1333),
    [anon_sym_COMMA] = ACTIONS(1333),
    [anon_sym_LBRACE] = ACTIONS(1333),
    [aux_sym_if_statement_token1] = ACTIONS(1333),
    [aux_sym_switch_statement_token1] = ACTIONS(1333),
    [aux_sym_foreach_statement_token1] = ACTIONS(1333),
    [aux_sym_for_statement_token1] = ACTIONS(1333),
    [aux_sym_while_statement_token1] = ACTIONS(1333),
    [aux_sym_do_statement_token1] = ACTIONS(1333),
    [aux_sym_function_statement_token1] = ACTIONS(1333),
    [aux_sym_function_statement_token2] = ACTIONS(1333),
    [aux_sym_function_statement_token3] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1333),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1333),
    [sym_label] = ACTIONS(1333),
    [aux_sym_trap_statement_token1] = ACTIONS(1333),
    [aux_sym_try_statement_token1] = ACTIONS(1333),
    [aux_sym_catch_clause_token1] = ACTIONS(1298),
    [aux_sym_finally_clause_token1] = ACTIONS(1333),
    [aux_sym_data_statement_token1] = ACTIONS(1333),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1333),
    [aux_sym_parallel_statement_token1] = ACTIONS(1333),
    [aux_sym_sequence_statement_token1] = ACTIONS(1333),
    [anon_sym_AMP] = ACTIONS(1333),
    [aux_sym_command_name_token1] = ACTIONS(1333),
    [aux_sym_command_name_token2] = ACTIONS(1333),
    [aux_sym_command_name_token3] = ACTIONS(1333),
    [aux_sym_command_name_token4] = ACTIONS(1333),
    [aux_sym_command_name_token5] = ACTIONS(1333),
    [aux_sym_command_name_token6] = ACTIONS(1333),
    [aux_sym_command_name_token7] = ACTIONS(1333),
    [aux_sym_command_name_token8] = ACTIONS(1333),
    [aux_sym_command_name_token9] = ACTIONS(1333),
    [aux_sym_command_name_token10] = ACTIONS(1333),
    [aux_sym_command_name_token11] = ACTIONS(1333),
    [aux_sym_command_name_token12] = ACTIONS(1333),
    [aux_sym_command_name_token13] = ACTIONS(1333),
    [aux_sym_command_name_token14] = ACTIONS(1333),
    [aux_sym_command_name_token15] = ACTIONS(1333),
    [aux_sym_command_name_token16] = ACTIONS(1333),
    [aux_sym_command_name_token17] = ACTIONS(1333),
    [aux_sym_command_name_token18] = ACTIONS(1333),
    [aux_sym_class_statement_token1] = ACTIONS(1333),
    [aux_sym_enum_statement_token1] = ACTIONS(1333),
    [anon_sym_PLUS] = ACTIONS(1333),
    [anon_sym_DASH] = ACTIONS(1333),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1333),
    [anon_sym_BANG] = ACTIONS(1333),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1333),
    [anon_sym_PLUS_PLUS] = ACTIONS(1333),
    [anon_sym_DASH_DASH] = ACTIONS(1333),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1333),
    [anon_sym_AT_LPAREN] = ACTIONS(1333),
    [anon_sym_AT_LBRACE] = ACTIONS(1333),
  },
  [STATE(310)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1357),
    [sym_hexadecimal_integer_literal] = ACTIONS(1357),
    [sym_real_literal] = ACTIONS(1357),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1357),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1357),
    [sym_verbatim_string_characters] = ACTIONS(1357),
    [sym_verbatim_here_string_characters] = ACTIONS(1357),
    [anon_sym_DOT] = ACTIONS(1357),
    [anon_sym_LBRACK] = ACTIONS(1357),
    [aux_sym_comparison_operator_token37] = ACTIONS(1357),
    [aux_sym_comparison_operator_token50] = ACTIONS(1357),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1357),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1357),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1357),
    [anon_sym_DOLLAR_] = ACTIONS(1357),
    [aux_sym_variable_token1] = ACTIONS(1357),
    [aux_sym_variable_token2] = ACTIONS(1357),
    [sym_braced_variable] = ACTIONS(1357),
    [anon_sym_SEMI] = ACTIONS(1357),
    [anon_sym_LPAREN] = ACTIONS(1357),
    [anon_sym_RPAREN] = ACTIONS(1357),
    [anon_sym_COMMA] = ACTIONS(1357),
    [anon_sym_LBRACE] = ACTIONS(1357),
    [anon_sym_RBRACE] = ACTIONS(1357),
    [aux_sym_if_statement_token1] = ACTIONS(1357),
    [aux_sym_switch_statement_token1] = ACTIONS(1357),
    [aux_sym_foreach_statement_token1] = ACTIONS(1357),
    [aux_sym_for_statement_token1] = ACTIONS(1357),
    [aux_sym_while_statement_token1] = ACTIONS(1357),
    [aux_sym_do_statement_token1] = ACTIONS(1357),
    [aux_sym_function_statement_token1] = ACTIONS(1357),
    [aux_sym_function_statement_token2] = ACTIONS(1357),
    [aux_sym_function_statement_token3] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1357),
    [sym_label] = ACTIONS(1357),
    [aux_sym_trap_statement_token1] = ACTIONS(1357),
    [aux_sym_try_statement_token1] = ACTIONS(1357),
    [aux_sym_catch_clause_token1] = ACTIONS(1357),
    [aux_sym_finally_clause_token1] = ACTIONS(1357),
    [aux_sym_data_statement_token1] = ACTIONS(1357),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1357),
    [aux_sym_parallel_statement_token1] = ACTIONS(1357),
    [aux_sym_sequence_statement_token1] = ACTIONS(1357),
    [anon_sym_AMP] = ACTIONS(1357),
    [aux_sym_command_name_token1] = ACTIONS(1357),
    [aux_sym_command_name_token2] = ACTIONS(1357),
    [aux_sym_command_name_token3] = ACTIONS(1357),
    [aux_sym_command_name_token4] = ACTIONS(1357),
    [aux_sym_command_name_token5] = ACTIONS(1357),
    [aux_sym_command_name_token6] = ACTIONS(1357),
    [aux_sym_command_name_token7] = ACTIONS(1357),
    [aux_sym_command_name_token8] = ACTIONS(1357),
    [aux_sym_command_name_token9] = ACTIONS(1357),
    [aux_sym_command_name_token10] = ACTIONS(1357),
    [aux_sym_command_name_token11] = ACTIONS(1357),
    [aux_sym_command_name_token12] = ACTIONS(1357),
    [aux_sym_command_name_token13] = ACTIONS(1357),
    [aux_sym_command_name_token14] = ACTIONS(1357),
    [aux_sym_command_name_token15] = ACTIONS(1357),
    [aux_sym_command_name_token16] = ACTIONS(1357),
    [aux_sym_command_name_token17] = ACTIONS(1357),
    [aux_sym_command_name_token18] = ACTIONS(1357),
    [aux_sym_class_statement_token1] = ACTIONS(1357),
    [aux_sym_enum_statement_token1] = ACTIONS(1357),
    [anon_sym_PLUS] = ACTIONS(1357),
    [anon_sym_DASH] = ACTIONS(1357),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1357),
    [anon_sym_BANG] = ACTIONS(1357),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1357),
    [anon_sym_PLUS_PLUS] = ACTIONS(1357),
    [anon_sym_DASH_DASH] = ACTIONS(1357),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1357),
    [anon_sym_AT_LPAREN] = ACTIONS(1357),
    [anon_sym_AT_LBRACE] = ACTIONS(1357),
  },
  [STATE(311)] = {
    [sym_else_clause] = STATE(360),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1359),
    [sym_hexadecimal_integer_literal] = ACTIONS(1359),
    [sym_real_literal] = ACTIONS(1359),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1359),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1359),
    [sym_verbatim_string_characters] = ACTIONS(1359),
    [sym_verbatim_here_string_characters] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [aux_sym_comparison_operator_token37] = ACTIONS(1359),
    [aux_sym_comparison_operator_token50] = ACTIONS(1359),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1359),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1359),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1359),
    [anon_sym_DOLLAR_] = ACTIONS(1359),
    [aux_sym_variable_token1] = ACTIONS(1359),
    [aux_sym_variable_token2] = ACTIONS(1359),
    [sym_braced_variable] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_RPAREN] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [anon_sym_RBRACE] = ACTIONS(1359),
    [aux_sym_if_statement_token1] = ACTIONS(1359),
    [aux_sym_else_clause_token1] = ACTIONS(1288),
    [aux_sym_switch_statement_token1] = ACTIONS(1359),
    [aux_sym_foreach_statement_token1] = ACTIONS(1359),
    [aux_sym_for_statement_token1] = ACTIONS(1359),
    [aux_sym_while_statement_token1] = ACTIONS(1359),
    [aux_sym_do_statement_token1] = ACTIONS(1359),
    [aux_sym_function_statement_token1] = ACTIONS(1359),
    [aux_sym_function_statement_token2] = ACTIONS(1359),
    [aux_sym_function_statement_token3] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1359),
    [sym_label] = ACTIONS(1359),
    [aux_sym_trap_statement_token1] = ACTIONS(1359),
    [aux_sym_try_statement_token1] = ACTIONS(1359),
    [aux_sym_data_statement_token1] = ACTIONS(1359),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1359),
    [aux_sym_parallel_statement_token1] = ACTIONS(1359),
    [aux_sym_sequence_statement_token1] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [aux_sym_command_name_token1] = ACTIONS(1359),
    [aux_sym_command_name_token2] = ACTIONS(1359),
    [aux_sym_command_name_token3] = ACTIONS(1359),
    [aux_sym_command_name_token4] = ACTIONS(1359),
    [aux_sym_command_name_token5] = ACTIONS(1359),
    [aux_sym_command_name_token6] = ACTIONS(1359),
    [aux_sym_command_name_token7] = ACTIONS(1359),
    [aux_sym_command_name_token8] = ACTIONS(1359),
    [aux_sym_command_name_token9] = ACTIONS(1359),
    [aux_sym_command_name_token10] = ACTIONS(1359),
    [aux_sym_command_name_token11] = ACTIONS(1359),
    [aux_sym_command_name_token12] = ACTIONS(1359),
    [aux_sym_command_name_token13] = ACTIONS(1359),
    [aux_sym_command_name_token14] = ACTIONS(1359),
    [aux_sym_command_name_token15] = ACTIONS(1359),
    [aux_sym_command_name_token16] = ACTIONS(1359),
    [aux_sym_command_name_token17] = ACTIONS(1359),
    [aux_sym_command_name_token18] = ACTIONS(1359),
    [aux_sym_class_statement_token1] = ACTIONS(1359),
    [aux_sym_enum_statement_token1] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1359),
    [anon_sym_BANG] = ACTIONS(1359),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1359),
    [anon_sym_PLUS_PLUS] = ACTIONS(1359),
    [anon_sym_DASH_DASH] = ACTIONS(1359),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1359),
    [anon_sym_AT_LPAREN] = ACTIONS(1359),
    [anon_sym_AT_LBRACE] = ACTIONS(1359),
  },
  [STATE(312)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1361),
    [sym_hexadecimal_integer_literal] = ACTIONS(1361),
    [sym_real_literal] = ACTIONS(1361),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1361),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1361),
    [sym_verbatim_string_characters] = ACTIONS(1361),
    [sym_verbatim_here_string_characters] = ACTIONS(1361),
    [anon_sym_DOT] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1361),
    [aux_sym_comparison_operator_token37] = ACTIONS(1361),
    [aux_sym_comparison_operator_token50] = ACTIONS(1361),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1361),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1361),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1361),
    [anon_sym_DOLLAR_] = ACTIONS(1361),
    [aux_sym_variable_token1] = ACTIONS(1361),
    [aux_sym_variable_token2] = ACTIONS(1361),
    [sym_braced_variable] = ACTIONS(1361),
    [anon_sym_SEMI] = ACTIONS(1361),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1361),
    [anon_sym_COMMA] = ACTIONS(1361),
    [anon_sym_LBRACE] = ACTIONS(1361),
    [anon_sym_RBRACE] = ACTIONS(1361),
    [aux_sym_if_statement_token1] = ACTIONS(1361),
    [aux_sym_elseif_clause_token1] = ACTIONS(1361),
    [aux_sym_else_clause_token1] = ACTIONS(1361),
    [aux_sym_switch_statement_token1] = ACTIONS(1361),
    [aux_sym_foreach_statement_token1] = ACTIONS(1361),
    [aux_sym_for_statement_token1] = ACTIONS(1361),
    [aux_sym_while_statement_token1] = ACTIONS(1361),
    [aux_sym_do_statement_token1] = ACTIONS(1361),
    [aux_sym_function_statement_token1] = ACTIONS(1361),
    [aux_sym_function_statement_token2] = ACTIONS(1361),
    [aux_sym_function_statement_token3] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1361),
    [sym_label] = ACTIONS(1361),
    [aux_sym_trap_statement_token1] = ACTIONS(1361),
    [aux_sym_try_statement_token1] = ACTIONS(1361),
    [aux_sym_data_statement_token1] = ACTIONS(1361),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1361),
    [aux_sym_parallel_statement_token1] = ACTIONS(1361),
    [aux_sym_sequence_statement_token1] = ACTIONS(1361),
    [anon_sym_AMP] = ACTIONS(1361),
    [aux_sym_command_name_token1] = ACTIONS(1361),
    [aux_sym_command_name_token2] = ACTIONS(1361),
    [aux_sym_command_name_token3] = ACTIONS(1361),
    [aux_sym_command_name_token4] = ACTIONS(1361),
    [aux_sym_command_name_token5] = ACTIONS(1361),
    [aux_sym_command_name_token6] = ACTIONS(1361),
    [aux_sym_command_name_token7] = ACTIONS(1361),
    [aux_sym_command_name_token8] = ACTIONS(1361),
    [aux_sym_command_name_token9] = ACTIONS(1361),
    [aux_sym_command_name_token10] = ACTIONS(1361),
    [aux_sym_command_name_token11] = ACTIONS(1361),
    [aux_sym_command_name_token12] = ACTIONS(1361),
    [aux_sym_command_name_token13] = ACTIONS(1361),
    [aux_sym_command_name_token14] = ACTIONS(1361),
    [aux_sym_command_name_token15] = ACTIONS(1361),
    [aux_sym_command_name_token16] = ACTIONS(1361),
    [aux_sym_command_name_token17] = ACTIONS(1361),
    [aux_sym_command_name_token18] = ACTIONS(1361),
    [aux_sym_class_statement_token1] = ACTIONS(1361),
    [aux_sym_enum_statement_token1] = ACTIONS(1361),
    [anon_sym_PLUS] = ACTIONS(1361),
    [anon_sym_DASH] = ACTIONS(1361),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1361),
    [anon_sym_BANG] = ACTIONS(1361),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1361),
    [anon_sym_PLUS_PLUS] = ACTIONS(1361),
    [anon_sym_DASH_DASH] = ACTIONS(1361),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1361),
    [anon_sym_AT_LPAREN] = ACTIONS(1361),
    [anon_sym_AT_LBRACE] = ACTIONS(1361),
  },
  [STATE(313)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [anon_sym_RBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_catch_clause_token1] = ACTIONS(1363),
    [aux_sym_finally_clause_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(314)] = {
    [sym_finally_clause] = STATE(328),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1365),
    [sym_hexadecimal_integer_literal] = ACTIONS(1365),
    [sym_real_literal] = ACTIONS(1365),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1365),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1365),
    [sym_verbatim_string_characters] = ACTIONS(1365),
    [sym_verbatim_here_string_characters] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1365),
    [anon_sym_LBRACK] = ACTIONS(1365),
    [aux_sym_comparison_operator_token37] = ACTIONS(1365),
    [aux_sym_comparison_operator_token50] = ACTIONS(1365),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1365),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1365),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1365),
    [anon_sym_DOLLAR_] = ACTIONS(1365),
    [aux_sym_variable_token1] = ACTIONS(1365),
    [aux_sym_variable_token2] = ACTIONS(1365),
    [sym_braced_variable] = ACTIONS(1365),
    [anon_sym_SEMI] = ACTIONS(1365),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [anon_sym_RPAREN] = ACTIONS(1365),
    [anon_sym_COMMA] = ACTIONS(1365),
    [anon_sym_LBRACE] = ACTIONS(1365),
    [anon_sym_RBRACE] = ACTIONS(1365),
    [aux_sym_if_statement_token1] = ACTIONS(1365),
    [aux_sym_switch_statement_token1] = ACTIONS(1365),
    [aux_sym_foreach_statement_token1] = ACTIONS(1365),
    [aux_sym_for_statement_token1] = ACTIONS(1365),
    [aux_sym_while_statement_token1] = ACTIONS(1365),
    [aux_sym_do_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token2] = ACTIONS(1365),
    [aux_sym_function_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1365),
    [sym_label] = ACTIONS(1365),
    [aux_sym_trap_statement_token1] = ACTIONS(1365),
    [aux_sym_try_statement_token1] = ACTIONS(1365),
    [aux_sym_finally_clause_token1] = ACTIONS(1294),
    [aux_sym_data_statement_token1] = ACTIONS(1365),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1365),
    [aux_sym_parallel_statement_token1] = ACTIONS(1365),
    [aux_sym_sequence_statement_token1] = ACTIONS(1365),
    [anon_sym_AMP] = ACTIONS(1365),
    [aux_sym_command_name_token1] = ACTIONS(1365),
    [aux_sym_command_name_token2] = ACTIONS(1365),
    [aux_sym_command_name_token3] = ACTIONS(1365),
    [aux_sym_command_name_token4] = ACTIONS(1365),
    [aux_sym_command_name_token5] = ACTIONS(1365),
    [aux_sym_command_name_token6] = ACTIONS(1365),
    [aux_sym_command_name_token7] = ACTIONS(1365),
    [aux_sym_command_name_token8] = ACTIONS(1365),
    [aux_sym_command_name_token9] = ACTIONS(1365),
    [aux_sym_command_name_token10] = ACTIONS(1365),
    [aux_sym_command_name_token11] = ACTIONS(1365),
    [aux_sym_command_name_token12] = ACTIONS(1365),
    [aux_sym_command_name_token13] = ACTIONS(1365),
    [aux_sym_command_name_token14] = ACTIONS(1365),
    [aux_sym_command_name_token15] = ACTIONS(1365),
    [aux_sym_command_name_token16] = ACTIONS(1365),
    [aux_sym_command_name_token17] = ACTIONS(1365),
    [aux_sym_command_name_token18] = ACTIONS(1365),
    [aux_sym_class_statement_token1] = ACTIONS(1365),
    [aux_sym_enum_statement_token1] = ACTIONS(1365),
    [anon_sym_PLUS] = ACTIONS(1365),
    [anon_sym_DASH] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1365),
    [anon_sym_BANG] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1365),
    [anon_sym_PLUS_PLUS] = ACTIONS(1365),
    [anon_sym_DASH_DASH] = ACTIONS(1365),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LBRACE] = ACTIONS(1365),
  },
  [STATE(315)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_RPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [anon_sym_RBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_catch_clause_token1] = ACTIONS(1367),
    [aux_sym_finally_clause_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(316)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [anon_sym_RBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_elseif_clause_token1] = ACTIONS(1363),
    [aux_sym_else_clause_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(317)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1369),
    [sym_hexadecimal_integer_literal] = ACTIONS(1369),
    [sym_real_literal] = ACTIONS(1369),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1369),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1369),
    [sym_verbatim_string_characters] = ACTIONS(1369),
    [sym_verbatim_here_string_characters] = ACTIONS(1369),
    [anon_sym_DOT] = ACTIONS(1369),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [aux_sym_comparison_operator_token37] = ACTIONS(1369),
    [aux_sym_comparison_operator_token50] = ACTIONS(1369),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1369),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1369),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1369),
    [anon_sym_DOLLAR_] = ACTIONS(1369),
    [aux_sym_variable_token1] = ACTIONS(1369),
    [aux_sym_variable_token2] = ACTIONS(1369),
    [sym_braced_variable] = ACTIONS(1369),
    [anon_sym_SEMI] = ACTIONS(1369),
    [anon_sym_LPAREN] = ACTIONS(1369),
    [anon_sym_RPAREN] = ACTIONS(1369),
    [anon_sym_COMMA] = ACTIONS(1369),
    [anon_sym_LBRACE] = ACTIONS(1369),
    [anon_sym_RBRACE] = ACTIONS(1369),
    [aux_sym_if_statement_token1] = ACTIONS(1369),
    [aux_sym_switch_statement_token1] = ACTIONS(1369),
    [aux_sym_foreach_statement_token1] = ACTIONS(1369),
    [aux_sym_for_statement_token1] = ACTIONS(1369),
    [aux_sym_while_statement_token1] = ACTIONS(1369),
    [aux_sym_do_statement_token1] = ACTIONS(1369),
    [aux_sym_function_statement_token1] = ACTIONS(1369),
    [aux_sym_function_statement_token2] = ACTIONS(1369),
    [aux_sym_function_statement_token3] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1369),
    [sym_label] = ACTIONS(1369),
    [aux_sym_trap_statement_token1] = ACTIONS(1369),
    [aux_sym_try_statement_token1] = ACTIONS(1369),
    [aux_sym_catch_clause_token1] = ACTIONS(1369),
    [aux_sym_finally_clause_token1] = ACTIONS(1369),
    [aux_sym_data_statement_token1] = ACTIONS(1369),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1369),
    [aux_sym_parallel_statement_token1] = ACTIONS(1369),
    [aux_sym_sequence_statement_token1] = ACTIONS(1369),
    [anon_sym_AMP] = ACTIONS(1369),
    [aux_sym_command_name_token1] = ACTIONS(1369),
    [aux_sym_command_name_token2] = ACTIONS(1369),
    [aux_sym_command_name_token3] = ACTIONS(1369),
    [aux_sym_command_name_token4] = ACTIONS(1369),
    [aux_sym_command_name_token5] = ACTIONS(1369),
    [aux_sym_command_name_token6] = ACTIONS(1369),
    [aux_sym_command_name_token7] = ACTIONS(1369),
    [aux_sym_command_name_token8] = ACTIONS(1369),
    [aux_sym_command_name_token9] = ACTIONS(1369),
    [aux_sym_command_name_token10] = ACTIONS(1369),
    [aux_sym_command_name_token11] = ACTIONS(1369),
    [aux_sym_command_name_token12] = ACTIONS(1369),
    [aux_sym_command_name_token13] = ACTIONS(1369),
    [aux_sym_command_name_token14] = ACTIONS(1369),
    [aux_sym_command_name_token15] = ACTIONS(1369),
    [aux_sym_command_name_token16] = ACTIONS(1369),
    [aux_sym_command_name_token17] = ACTIONS(1369),
    [aux_sym_command_name_token18] = ACTIONS(1369),
    [aux_sym_class_statement_token1] = ACTIONS(1369),
    [aux_sym_enum_statement_token1] = ACTIONS(1369),
    [anon_sym_PLUS] = ACTIONS(1369),
    [anon_sym_DASH] = ACTIONS(1369),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1369),
    [anon_sym_BANG] = ACTIONS(1369),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1369),
    [anon_sym_PLUS_PLUS] = ACTIONS(1369),
    [anon_sym_DASH_DASH] = ACTIONS(1369),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1369),
    [anon_sym_AT_LPAREN] = ACTIONS(1369),
    [anon_sym_AT_LBRACE] = ACTIONS(1369),
  },
  [STATE(318)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_RPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [anon_sym_RBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_elseif_clause_token1] = ACTIONS(1367),
    [aux_sym_else_clause_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(319)] = {
    [ts_builtin_sym_end] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1369),
    [sym_hexadecimal_integer_literal] = ACTIONS(1369),
    [sym_real_literal] = ACTIONS(1369),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1369),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1369),
    [sym_verbatim_string_characters] = ACTIONS(1369),
    [sym_verbatim_here_string_characters] = ACTIONS(1369),
    [anon_sym_DOT] = ACTIONS(1369),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [aux_sym_comparison_operator_token37] = ACTIONS(1369),
    [aux_sym_comparison_operator_token50] = ACTIONS(1369),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1369),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1369),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1369),
    [anon_sym_DOLLAR_] = ACTIONS(1369),
    [aux_sym_variable_token1] = ACTIONS(1369),
    [aux_sym_variable_token2] = ACTIONS(1369),
    [sym_braced_variable] = ACTIONS(1369),
    [anon_sym_SEMI] = ACTIONS(1369),
    [anon_sym_LPAREN] = ACTIONS(1369),
    [anon_sym_COMMA] = ACTIONS(1369),
    [anon_sym_LBRACE] = ACTIONS(1369),
    [aux_sym_if_statement_token1] = ACTIONS(1369),
    [aux_sym_switch_statement_token1] = ACTIONS(1369),
    [aux_sym_foreach_statement_token1] = ACTIONS(1369),
    [aux_sym_for_statement_token1] = ACTIONS(1369),
    [aux_sym_while_statement_token1] = ACTIONS(1369),
    [aux_sym_do_statement_token1] = ACTIONS(1369),
    [aux_sym_function_statement_token1] = ACTIONS(1369),
    [aux_sym_function_statement_token2] = ACTIONS(1369),
    [aux_sym_function_statement_token3] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1369),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1369),
    [sym_label] = ACTIONS(1369),
    [aux_sym_trap_statement_token1] = ACTIONS(1369),
    [aux_sym_try_statement_token1] = ACTIONS(1369),
    [aux_sym_catch_clause_token1] = ACTIONS(1369),
    [aux_sym_finally_clause_token1] = ACTIONS(1369),
    [aux_sym_data_statement_token1] = ACTIONS(1369),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1369),
    [aux_sym_parallel_statement_token1] = ACTIONS(1369),
    [aux_sym_sequence_statement_token1] = ACTIONS(1369),
    [anon_sym_AMP] = ACTIONS(1369),
    [aux_sym_command_name_token1] = ACTIONS(1369),
    [aux_sym_command_name_token2] = ACTIONS(1369),
    [aux_sym_command_name_token3] = ACTIONS(1369),
    [aux_sym_command_name_token4] = ACTIONS(1369),
    [aux_sym_command_name_token5] = ACTIONS(1369),
    [aux_sym_command_name_token6] = ACTIONS(1369),
    [aux_sym_command_name_token7] = ACTIONS(1369),
    [aux_sym_command_name_token8] = ACTIONS(1369),
    [aux_sym_command_name_token9] = ACTIONS(1369),
    [aux_sym_command_name_token10] = ACTIONS(1369),
    [aux_sym_command_name_token11] = ACTIONS(1369),
    [aux_sym_command_name_token12] = ACTIONS(1369),
    [aux_sym_command_name_token13] = ACTIONS(1369),
    [aux_sym_command_name_token14] = ACTIONS(1369),
    [aux_sym_command_name_token15] = ACTIONS(1369),
    [aux_sym_command_name_token16] = ACTIONS(1369),
    [aux_sym_command_name_token17] = ACTIONS(1369),
    [aux_sym_command_name_token18] = ACTIONS(1369),
    [aux_sym_class_statement_token1] = ACTIONS(1369),
    [aux_sym_enum_statement_token1] = ACTIONS(1369),
    [anon_sym_PLUS] = ACTIONS(1369),
    [anon_sym_DASH] = ACTIONS(1369),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1369),
    [anon_sym_BANG] = ACTIONS(1369),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1369),
    [anon_sym_PLUS_PLUS] = ACTIONS(1369),
    [anon_sym_DASH_DASH] = ACTIONS(1369),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1369),
    [anon_sym_AT_LPAREN] = ACTIONS(1369),
    [anon_sym_AT_LBRACE] = ACTIONS(1369),
  },
  [STATE(320)] = {
    [sym_else_clause] = STATE(395),
    [ts_builtin_sym_end] = ACTIONS(1373),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1359),
    [sym_hexadecimal_integer_literal] = ACTIONS(1359),
    [sym_real_literal] = ACTIONS(1359),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1359),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1359),
    [sym_verbatim_string_characters] = ACTIONS(1359),
    [sym_verbatim_here_string_characters] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [aux_sym_comparison_operator_token37] = ACTIONS(1359),
    [aux_sym_comparison_operator_token50] = ACTIONS(1359),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1359),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1359),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1359),
    [anon_sym_DOLLAR_] = ACTIONS(1359),
    [aux_sym_variable_token1] = ACTIONS(1359),
    [aux_sym_variable_token2] = ACTIONS(1359),
    [sym_braced_variable] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [aux_sym_if_statement_token1] = ACTIONS(1359),
    [aux_sym_else_clause_token1] = ACTIONS(1306),
    [aux_sym_switch_statement_token1] = ACTIONS(1359),
    [aux_sym_foreach_statement_token1] = ACTIONS(1359),
    [aux_sym_for_statement_token1] = ACTIONS(1359),
    [aux_sym_while_statement_token1] = ACTIONS(1359),
    [aux_sym_do_statement_token1] = ACTIONS(1359),
    [aux_sym_function_statement_token1] = ACTIONS(1359),
    [aux_sym_function_statement_token2] = ACTIONS(1359),
    [aux_sym_function_statement_token3] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1359),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1359),
    [sym_label] = ACTIONS(1359),
    [aux_sym_trap_statement_token1] = ACTIONS(1359),
    [aux_sym_try_statement_token1] = ACTIONS(1359),
    [aux_sym_data_statement_token1] = ACTIONS(1359),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1359),
    [aux_sym_parallel_statement_token1] = ACTIONS(1359),
    [aux_sym_sequence_statement_token1] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [aux_sym_command_name_token1] = ACTIONS(1359),
    [aux_sym_command_name_token2] = ACTIONS(1359),
    [aux_sym_command_name_token3] = ACTIONS(1359),
    [aux_sym_command_name_token4] = ACTIONS(1359),
    [aux_sym_command_name_token5] = ACTIONS(1359),
    [aux_sym_command_name_token6] = ACTIONS(1359),
    [aux_sym_command_name_token7] = ACTIONS(1359),
    [aux_sym_command_name_token8] = ACTIONS(1359),
    [aux_sym_command_name_token9] = ACTIONS(1359),
    [aux_sym_command_name_token10] = ACTIONS(1359),
    [aux_sym_command_name_token11] = ACTIONS(1359),
    [aux_sym_command_name_token12] = ACTIONS(1359),
    [aux_sym_command_name_token13] = ACTIONS(1359),
    [aux_sym_command_name_token14] = ACTIONS(1359),
    [aux_sym_command_name_token15] = ACTIONS(1359),
    [aux_sym_command_name_token16] = ACTIONS(1359),
    [aux_sym_command_name_token17] = ACTIONS(1359),
    [aux_sym_command_name_token18] = ACTIONS(1359),
    [aux_sym_class_statement_token1] = ACTIONS(1359),
    [aux_sym_enum_statement_token1] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1359),
    [anon_sym_BANG] = ACTIONS(1359),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1359),
    [anon_sym_PLUS_PLUS] = ACTIONS(1359),
    [anon_sym_DASH_DASH] = ACTIONS(1359),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1359),
    [anon_sym_AT_LPAREN] = ACTIONS(1359),
    [anon_sym_AT_LBRACE] = ACTIONS(1359),
  },
  [STATE(321)] = {
    [ts_builtin_sym_end] = ACTIONS(1375),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1357),
    [sym_hexadecimal_integer_literal] = ACTIONS(1357),
    [sym_real_literal] = ACTIONS(1357),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1357),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1357),
    [sym_verbatim_string_characters] = ACTIONS(1357),
    [sym_verbatim_here_string_characters] = ACTIONS(1357),
    [anon_sym_DOT] = ACTIONS(1357),
    [anon_sym_LBRACK] = ACTIONS(1357),
    [aux_sym_comparison_operator_token37] = ACTIONS(1357),
    [aux_sym_comparison_operator_token50] = ACTIONS(1357),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1357),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1357),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1357),
    [anon_sym_DOLLAR_] = ACTIONS(1357),
    [aux_sym_variable_token1] = ACTIONS(1357),
    [aux_sym_variable_token2] = ACTIONS(1357),
    [sym_braced_variable] = ACTIONS(1357),
    [anon_sym_SEMI] = ACTIONS(1357),
    [anon_sym_LPAREN] = ACTIONS(1357),
    [anon_sym_COMMA] = ACTIONS(1357),
    [anon_sym_LBRACE] = ACTIONS(1357),
    [aux_sym_if_statement_token1] = ACTIONS(1357),
    [aux_sym_switch_statement_token1] = ACTIONS(1357),
    [aux_sym_foreach_statement_token1] = ACTIONS(1357),
    [aux_sym_for_statement_token1] = ACTIONS(1357),
    [aux_sym_while_statement_token1] = ACTIONS(1357),
    [aux_sym_do_statement_token1] = ACTIONS(1357),
    [aux_sym_function_statement_token1] = ACTIONS(1357),
    [aux_sym_function_statement_token2] = ACTIONS(1357),
    [aux_sym_function_statement_token3] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1357),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1357),
    [sym_label] = ACTIONS(1357),
    [aux_sym_trap_statement_token1] = ACTIONS(1357),
    [aux_sym_try_statement_token1] = ACTIONS(1357),
    [aux_sym_catch_clause_token1] = ACTIONS(1357),
    [aux_sym_finally_clause_token1] = ACTIONS(1357),
    [aux_sym_data_statement_token1] = ACTIONS(1357),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1357),
    [aux_sym_parallel_statement_token1] = ACTIONS(1357),
    [aux_sym_sequence_statement_token1] = ACTIONS(1357),
    [anon_sym_AMP] = ACTIONS(1357),
    [aux_sym_command_name_token1] = ACTIONS(1357),
    [aux_sym_command_name_token2] = ACTIONS(1357),
    [aux_sym_command_name_token3] = ACTIONS(1357),
    [aux_sym_command_name_token4] = ACTIONS(1357),
    [aux_sym_command_name_token5] = ACTIONS(1357),
    [aux_sym_command_name_token6] = ACTIONS(1357),
    [aux_sym_command_name_token7] = ACTIONS(1357),
    [aux_sym_command_name_token8] = ACTIONS(1357),
    [aux_sym_command_name_token9] = ACTIONS(1357),
    [aux_sym_command_name_token10] = ACTIONS(1357),
    [aux_sym_command_name_token11] = ACTIONS(1357),
    [aux_sym_command_name_token12] = ACTIONS(1357),
    [aux_sym_command_name_token13] = ACTIONS(1357),
    [aux_sym_command_name_token14] = ACTIONS(1357),
    [aux_sym_command_name_token15] = ACTIONS(1357),
    [aux_sym_command_name_token16] = ACTIONS(1357),
    [aux_sym_command_name_token17] = ACTIONS(1357),
    [aux_sym_command_name_token18] = ACTIONS(1357),
    [aux_sym_class_statement_token1] = ACTIONS(1357),
    [aux_sym_enum_statement_token1] = ACTIONS(1357),
    [anon_sym_PLUS] = ACTIONS(1357),
    [anon_sym_DASH] = ACTIONS(1357),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1357),
    [anon_sym_BANG] = ACTIONS(1357),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1357),
    [anon_sym_PLUS_PLUS] = ACTIONS(1357),
    [anon_sym_DASH_DASH] = ACTIONS(1357),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1357),
    [anon_sym_AT_LPAREN] = ACTIONS(1357),
    [anon_sym_AT_LBRACE] = ACTIONS(1357),
  },
  [STATE(322)] = {
    [ts_builtin_sym_end] = ACTIONS(1377),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_elseif_clause_token1] = ACTIONS(1367),
    [aux_sym_else_clause_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(323)] = {
    [ts_builtin_sym_end] = ACTIONS(1379),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_elseif_clause_token1] = ACTIONS(1363),
    [aux_sym_else_clause_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(324)] = {
    [sym_finally_clause] = STATE(424),
    [ts_builtin_sym_end] = ACTIONS(1381),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1365),
    [sym_hexadecimal_integer_literal] = ACTIONS(1365),
    [sym_real_literal] = ACTIONS(1365),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1365),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1365),
    [sym_verbatim_string_characters] = ACTIONS(1365),
    [sym_verbatim_here_string_characters] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1365),
    [anon_sym_LBRACK] = ACTIONS(1365),
    [aux_sym_comparison_operator_token37] = ACTIONS(1365),
    [aux_sym_comparison_operator_token50] = ACTIONS(1365),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1365),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1365),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1365),
    [anon_sym_DOLLAR_] = ACTIONS(1365),
    [aux_sym_variable_token1] = ACTIONS(1365),
    [aux_sym_variable_token2] = ACTIONS(1365),
    [sym_braced_variable] = ACTIONS(1365),
    [anon_sym_SEMI] = ACTIONS(1365),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [anon_sym_COMMA] = ACTIONS(1365),
    [anon_sym_LBRACE] = ACTIONS(1365),
    [aux_sym_if_statement_token1] = ACTIONS(1365),
    [aux_sym_switch_statement_token1] = ACTIONS(1365),
    [aux_sym_foreach_statement_token1] = ACTIONS(1365),
    [aux_sym_for_statement_token1] = ACTIONS(1365),
    [aux_sym_while_statement_token1] = ACTIONS(1365),
    [aux_sym_do_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token2] = ACTIONS(1365),
    [aux_sym_function_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1365),
    [sym_label] = ACTIONS(1365),
    [aux_sym_trap_statement_token1] = ACTIONS(1365),
    [aux_sym_try_statement_token1] = ACTIONS(1365),
    [aux_sym_finally_clause_token1] = ACTIONS(1300),
    [aux_sym_data_statement_token1] = ACTIONS(1365),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1365),
    [aux_sym_parallel_statement_token1] = ACTIONS(1365),
    [aux_sym_sequence_statement_token1] = ACTIONS(1365),
    [anon_sym_AMP] = ACTIONS(1365),
    [aux_sym_command_name_token1] = ACTIONS(1365),
    [aux_sym_command_name_token2] = ACTIONS(1365),
    [aux_sym_command_name_token3] = ACTIONS(1365),
    [aux_sym_command_name_token4] = ACTIONS(1365),
    [aux_sym_command_name_token5] = ACTIONS(1365),
    [aux_sym_command_name_token6] = ACTIONS(1365),
    [aux_sym_command_name_token7] = ACTIONS(1365),
    [aux_sym_command_name_token8] = ACTIONS(1365),
    [aux_sym_command_name_token9] = ACTIONS(1365),
    [aux_sym_command_name_token10] = ACTIONS(1365),
    [aux_sym_command_name_token11] = ACTIONS(1365),
    [aux_sym_command_name_token12] = ACTIONS(1365),
    [aux_sym_command_name_token13] = ACTIONS(1365),
    [aux_sym_command_name_token14] = ACTIONS(1365),
    [aux_sym_command_name_token15] = ACTIONS(1365),
    [aux_sym_command_name_token16] = ACTIONS(1365),
    [aux_sym_command_name_token17] = ACTIONS(1365),
    [aux_sym_command_name_token18] = ACTIONS(1365),
    [aux_sym_class_statement_token1] = ACTIONS(1365),
    [aux_sym_enum_statement_token1] = ACTIONS(1365),
    [anon_sym_PLUS] = ACTIONS(1365),
    [anon_sym_DASH] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1365),
    [anon_sym_BANG] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1365),
    [anon_sym_PLUS_PLUS] = ACTIONS(1365),
    [anon_sym_DASH_DASH] = ACTIONS(1365),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LBRACE] = ACTIONS(1365),
  },
  [STATE(325)] = {
    [ts_builtin_sym_end] = ACTIONS(1383),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1361),
    [sym_hexadecimal_integer_literal] = ACTIONS(1361),
    [sym_real_literal] = ACTIONS(1361),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1361),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1361),
    [sym_verbatim_string_characters] = ACTIONS(1361),
    [sym_verbatim_here_string_characters] = ACTIONS(1361),
    [anon_sym_DOT] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1361),
    [aux_sym_comparison_operator_token37] = ACTIONS(1361),
    [aux_sym_comparison_operator_token50] = ACTIONS(1361),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1361),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1361),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1361),
    [anon_sym_DOLLAR_] = ACTIONS(1361),
    [aux_sym_variable_token1] = ACTIONS(1361),
    [aux_sym_variable_token2] = ACTIONS(1361),
    [sym_braced_variable] = ACTIONS(1361),
    [anon_sym_SEMI] = ACTIONS(1361),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_COMMA] = ACTIONS(1361),
    [anon_sym_LBRACE] = ACTIONS(1361),
    [aux_sym_if_statement_token1] = ACTIONS(1361),
    [aux_sym_elseif_clause_token1] = ACTIONS(1361),
    [aux_sym_else_clause_token1] = ACTIONS(1361),
    [aux_sym_switch_statement_token1] = ACTIONS(1361),
    [aux_sym_foreach_statement_token1] = ACTIONS(1361),
    [aux_sym_for_statement_token1] = ACTIONS(1361),
    [aux_sym_while_statement_token1] = ACTIONS(1361),
    [aux_sym_do_statement_token1] = ACTIONS(1361),
    [aux_sym_function_statement_token1] = ACTIONS(1361),
    [aux_sym_function_statement_token2] = ACTIONS(1361),
    [aux_sym_function_statement_token3] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1361),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1361),
    [sym_label] = ACTIONS(1361),
    [aux_sym_trap_statement_token1] = ACTIONS(1361),
    [aux_sym_try_statement_token1] = ACTIONS(1361),
    [aux_sym_data_statement_token1] = ACTIONS(1361),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1361),
    [aux_sym_parallel_statement_token1] = ACTIONS(1361),
    [aux_sym_sequence_statement_token1] = ACTIONS(1361),
    [anon_sym_AMP] = ACTIONS(1361),
    [aux_sym_command_name_token1] = ACTIONS(1361),
    [aux_sym_command_name_token2] = ACTIONS(1361),
    [aux_sym_command_name_token3] = ACTIONS(1361),
    [aux_sym_command_name_token4] = ACTIONS(1361),
    [aux_sym_command_name_token5] = ACTIONS(1361),
    [aux_sym_command_name_token6] = ACTIONS(1361),
    [aux_sym_command_name_token7] = ACTIONS(1361),
    [aux_sym_command_name_token8] = ACTIONS(1361),
    [aux_sym_command_name_token9] = ACTIONS(1361),
    [aux_sym_command_name_token10] = ACTIONS(1361),
    [aux_sym_command_name_token11] = ACTIONS(1361),
    [aux_sym_command_name_token12] = ACTIONS(1361),
    [aux_sym_command_name_token13] = ACTIONS(1361),
    [aux_sym_command_name_token14] = ACTIONS(1361),
    [aux_sym_command_name_token15] = ACTIONS(1361),
    [aux_sym_command_name_token16] = ACTIONS(1361),
    [aux_sym_command_name_token17] = ACTIONS(1361),
    [aux_sym_command_name_token18] = ACTIONS(1361),
    [aux_sym_class_statement_token1] = ACTIONS(1361),
    [aux_sym_enum_statement_token1] = ACTIONS(1361),
    [anon_sym_PLUS] = ACTIONS(1361),
    [anon_sym_DASH] = ACTIONS(1361),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1361),
    [anon_sym_BANG] = ACTIONS(1361),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1361),
    [anon_sym_PLUS_PLUS] = ACTIONS(1361),
    [anon_sym_DASH_DASH] = ACTIONS(1361),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1361),
    [anon_sym_AT_LPAREN] = ACTIONS(1361),
    [anon_sym_AT_LBRACE] = ACTIONS(1361),
  },
  [STATE(326)] = {
    [ts_builtin_sym_end] = ACTIONS(1379),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_catch_clause_token1] = ACTIONS(1363),
    [aux_sym_finally_clause_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(327)] = {
    [ts_builtin_sym_end] = ACTIONS(1377),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_catch_clause_token1] = ACTIONS(1367),
    [aux_sym_finally_clause_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(328)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1385),
    [sym_hexadecimal_integer_literal] = ACTIONS(1385),
    [sym_real_literal] = ACTIONS(1385),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1385),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1385),
    [sym_verbatim_string_characters] = ACTIONS(1385),
    [sym_verbatim_here_string_characters] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [aux_sym_comparison_operator_token37] = ACTIONS(1385),
    [aux_sym_comparison_operator_token50] = ACTIONS(1385),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1385),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1385),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1385),
    [anon_sym_DOLLAR_] = ACTIONS(1385),
    [aux_sym_variable_token1] = ACTIONS(1385),
    [aux_sym_variable_token2] = ACTIONS(1385),
    [sym_braced_variable] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_RPAREN] = ACTIONS(1385),
    [anon_sym_COMMA] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [anon_sym_RBRACE] = ACTIONS(1385),
    [aux_sym_if_statement_token1] = ACTIONS(1385),
    [aux_sym_switch_statement_token1] = ACTIONS(1385),
    [aux_sym_foreach_statement_token1] = ACTIONS(1385),
    [aux_sym_for_statement_token1] = ACTIONS(1385),
    [aux_sym_while_statement_token1] = ACTIONS(1385),
    [aux_sym_do_statement_token1] = ACTIONS(1385),
    [aux_sym_function_statement_token1] = ACTIONS(1385),
    [aux_sym_function_statement_token2] = ACTIONS(1385),
    [aux_sym_function_statement_token3] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1385),
    [sym_label] = ACTIONS(1385),
    [aux_sym_trap_statement_token1] = ACTIONS(1385),
    [aux_sym_try_statement_token1] = ACTIONS(1385),
    [aux_sym_data_statement_token1] = ACTIONS(1385),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1385),
    [aux_sym_parallel_statement_token1] = ACTIONS(1385),
    [aux_sym_sequence_statement_token1] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [aux_sym_command_name_token1] = ACTIONS(1385),
    [aux_sym_command_name_token2] = ACTIONS(1385),
    [aux_sym_command_name_token3] = ACTIONS(1385),
    [aux_sym_command_name_token4] = ACTIONS(1385),
    [aux_sym_command_name_token5] = ACTIONS(1385),
    [aux_sym_command_name_token6] = ACTIONS(1385),
    [aux_sym_command_name_token7] = ACTIONS(1385),
    [aux_sym_command_name_token8] = ACTIONS(1385),
    [aux_sym_command_name_token9] = ACTIONS(1385),
    [aux_sym_command_name_token10] = ACTIONS(1385),
    [aux_sym_command_name_token11] = ACTIONS(1385),
    [aux_sym_command_name_token12] = ACTIONS(1385),
    [aux_sym_command_name_token13] = ACTIONS(1385),
    [aux_sym_command_name_token14] = ACTIONS(1385),
    [aux_sym_command_name_token15] = ACTIONS(1385),
    [aux_sym_command_name_token16] = ACTIONS(1385),
    [aux_sym_command_name_token17] = ACTIONS(1385),
    [aux_sym_command_name_token18] = ACTIONS(1385),
    [aux_sym_class_statement_token1] = ACTIONS(1385),
    [aux_sym_enum_statement_token1] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1385),
    [anon_sym_BANG] = ACTIONS(1385),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1385),
    [anon_sym_PLUS_PLUS] = ACTIONS(1385),
    [anon_sym_DASH_DASH] = ACTIONS(1385),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1385),
    [anon_sym_AT_LPAREN] = ACTIONS(1385),
    [anon_sym_AT_LBRACE] = ACTIONS(1385),
  },
  [STATE(329)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1387),
    [sym_hexadecimal_integer_literal] = ACTIONS(1387),
    [sym_real_literal] = ACTIONS(1387),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1387),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1387),
    [sym_verbatim_string_characters] = ACTIONS(1387),
    [sym_verbatim_here_string_characters] = ACTIONS(1387),
    [anon_sym_DOT] = ACTIONS(1387),
    [anon_sym_LBRACK] = ACTIONS(1387),
    [aux_sym_comparison_operator_token37] = ACTIONS(1387),
    [aux_sym_comparison_operator_token50] = ACTIONS(1387),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1387),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1387),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1387),
    [anon_sym_DOLLAR_] = ACTIONS(1387),
    [aux_sym_variable_token1] = ACTIONS(1387),
    [aux_sym_variable_token2] = ACTIONS(1387),
    [sym_braced_variable] = ACTIONS(1387),
    [anon_sym_SEMI] = ACTIONS(1387),
    [anon_sym_LPAREN] = ACTIONS(1387),
    [anon_sym_RPAREN] = ACTIONS(1387),
    [anon_sym_COMMA] = ACTIONS(1387),
    [anon_sym_LBRACE] = ACTIONS(1387),
    [anon_sym_RBRACE] = ACTIONS(1387),
    [aux_sym_if_statement_token1] = ACTIONS(1387),
    [aux_sym_switch_statement_token1] = ACTIONS(1387),
    [aux_sym_foreach_statement_token1] = ACTIONS(1387),
    [aux_sym_for_statement_token1] = ACTIONS(1387),
    [aux_sym_while_statement_token1] = ACTIONS(1387),
    [aux_sym_do_statement_token1] = ACTIONS(1387),
    [aux_sym_function_statement_token1] = ACTIONS(1387),
    [aux_sym_function_statement_token2] = ACTIONS(1387),
    [aux_sym_function_statement_token3] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1387),
    [sym_label] = ACTIONS(1387),
    [aux_sym_trap_statement_token1] = ACTIONS(1387),
    [aux_sym_try_statement_token1] = ACTIONS(1387),
    [aux_sym_data_statement_token1] = ACTIONS(1387),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1387),
    [aux_sym_parallel_statement_token1] = ACTIONS(1387),
    [aux_sym_sequence_statement_token1] = ACTIONS(1387),
    [anon_sym_AMP] = ACTIONS(1387),
    [aux_sym_command_name_token1] = ACTIONS(1387),
    [aux_sym_command_name_token2] = ACTIONS(1387),
    [aux_sym_command_name_token3] = ACTIONS(1387),
    [aux_sym_command_name_token4] = ACTIONS(1387),
    [aux_sym_command_name_token5] = ACTIONS(1387),
    [aux_sym_command_name_token6] = ACTIONS(1387),
    [aux_sym_command_name_token7] = ACTIONS(1387),
    [aux_sym_command_name_token8] = ACTIONS(1387),
    [aux_sym_command_name_token9] = ACTIONS(1387),
    [aux_sym_command_name_token10] = ACTIONS(1387),
    [aux_sym_command_name_token11] = ACTIONS(1387),
    [aux_sym_command_name_token12] = ACTIONS(1387),
    [aux_sym_command_name_token13] = ACTIONS(1387),
    [aux_sym_command_name_token14] = ACTIONS(1387),
    [aux_sym_command_name_token15] = ACTIONS(1387),
    [aux_sym_command_name_token16] = ACTIONS(1387),
    [aux_sym_command_name_token17] = ACTIONS(1387),
    [aux_sym_command_name_token18] = ACTIONS(1387),
    [aux_sym_class_statement_token1] = ACTIONS(1387),
    [aux_sym_enum_statement_token1] = ACTIONS(1387),
    [anon_sym_PLUS] = ACTIONS(1387),
    [anon_sym_DASH] = ACTIONS(1387),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1387),
    [anon_sym_BANG] = ACTIONS(1387),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1387),
    [anon_sym_PLUS_PLUS] = ACTIONS(1387),
    [anon_sym_DASH_DASH] = ACTIONS(1387),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1387),
    [anon_sym_AT_LPAREN] = ACTIONS(1387),
    [anon_sym_AT_LBRACE] = ACTIONS(1387),
  },
  [STATE(330)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1389),
    [sym_hexadecimal_integer_literal] = ACTIONS(1389),
    [sym_real_literal] = ACTIONS(1389),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1389),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1389),
    [sym_verbatim_string_characters] = ACTIONS(1389),
    [sym_verbatim_here_string_characters] = ACTIONS(1389),
    [anon_sym_DOT] = ACTIONS(1389),
    [anon_sym_LBRACK] = ACTIONS(1389),
    [aux_sym_comparison_operator_token37] = ACTIONS(1389),
    [aux_sym_comparison_operator_token50] = ACTIONS(1389),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1389),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1389),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1389),
    [anon_sym_DOLLAR_] = ACTIONS(1389),
    [aux_sym_variable_token1] = ACTIONS(1389),
    [aux_sym_variable_token2] = ACTIONS(1389),
    [sym_braced_variable] = ACTIONS(1389),
    [anon_sym_SEMI] = ACTIONS(1389),
    [anon_sym_LPAREN] = ACTIONS(1389),
    [anon_sym_RPAREN] = ACTIONS(1389),
    [anon_sym_COMMA] = ACTIONS(1389),
    [anon_sym_LBRACE] = ACTIONS(1389),
    [anon_sym_RBRACE] = ACTIONS(1389),
    [aux_sym_if_statement_token1] = ACTIONS(1389),
    [aux_sym_switch_statement_token1] = ACTIONS(1389),
    [aux_sym_foreach_statement_token1] = ACTIONS(1389),
    [aux_sym_for_statement_token1] = ACTIONS(1389),
    [aux_sym_while_statement_token1] = ACTIONS(1389),
    [aux_sym_do_statement_token1] = ACTIONS(1389),
    [aux_sym_function_statement_token1] = ACTIONS(1389),
    [aux_sym_function_statement_token2] = ACTIONS(1389),
    [aux_sym_function_statement_token3] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1389),
    [sym_label] = ACTIONS(1389),
    [aux_sym_trap_statement_token1] = ACTIONS(1389),
    [aux_sym_try_statement_token1] = ACTIONS(1389),
    [aux_sym_data_statement_token1] = ACTIONS(1389),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1389),
    [aux_sym_parallel_statement_token1] = ACTIONS(1389),
    [aux_sym_sequence_statement_token1] = ACTIONS(1389),
    [anon_sym_AMP] = ACTIONS(1389),
    [aux_sym_command_name_token1] = ACTIONS(1389),
    [aux_sym_command_name_token2] = ACTIONS(1389),
    [aux_sym_command_name_token3] = ACTIONS(1389),
    [aux_sym_command_name_token4] = ACTIONS(1389),
    [aux_sym_command_name_token5] = ACTIONS(1389),
    [aux_sym_command_name_token6] = ACTIONS(1389),
    [aux_sym_command_name_token7] = ACTIONS(1389),
    [aux_sym_command_name_token8] = ACTIONS(1389),
    [aux_sym_command_name_token9] = ACTIONS(1389),
    [aux_sym_command_name_token10] = ACTIONS(1389),
    [aux_sym_command_name_token11] = ACTIONS(1389),
    [aux_sym_command_name_token12] = ACTIONS(1389),
    [aux_sym_command_name_token13] = ACTIONS(1389),
    [aux_sym_command_name_token14] = ACTIONS(1389),
    [aux_sym_command_name_token15] = ACTIONS(1389),
    [aux_sym_command_name_token16] = ACTIONS(1389),
    [aux_sym_command_name_token17] = ACTIONS(1389),
    [aux_sym_command_name_token18] = ACTIONS(1389),
    [aux_sym_class_statement_token1] = ACTIONS(1389),
    [aux_sym_enum_statement_token1] = ACTIONS(1389),
    [anon_sym_PLUS] = ACTIONS(1389),
    [anon_sym_DASH] = ACTIONS(1389),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1389),
    [anon_sym_BANG] = ACTIONS(1389),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1389),
    [anon_sym_PLUS_PLUS] = ACTIONS(1389),
    [anon_sym_DASH_DASH] = ACTIONS(1389),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1389),
    [anon_sym_AT_LPAREN] = ACTIONS(1389),
    [anon_sym_AT_LBRACE] = ACTIONS(1389),
  },
  [STATE(331)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1391),
    [sym_hexadecimal_integer_literal] = ACTIONS(1391),
    [sym_real_literal] = ACTIONS(1391),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1391),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1391),
    [sym_verbatim_string_characters] = ACTIONS(1391),
    [sym_verbatim_here_string_characters] = ACTIONS(1391),
    [anon_sym_DOT] = ACTIONS(1391),
    [anon_sym_LBRACK] = ACTIONS(1391),
    [aux_sym_comparison_operator_token37] = ACTIONS(1391),
    [aux_sym_comparison_operator_token50] = ACTIONS(1391),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1391),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1391),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1391),
    [anon_sym_DOLLAR_] = ACTIONS(1391),
    [aux_sym_variable_token1] = ACTIONS(1391),
    [aux_sym_variable_token2] = ACTIONS(1391),
    [sym_braced_variable] = ACTIONS(1391),
    [anon_sym_SEMI] = ACTIONS(1391),
    [anon_sym_LPAREN] = ACTIONS(1391),
    [anon_sym_RPAREN] = ACTIONS(1391),
    [anon_sym_COMMA] = ACTIONS(1391),
    [anon_sym_LBRACE] = ACTIONS(1391),
    [anon_sym_RBRACE] = ACTIONS(1391),
    [aux_sym_if_statement_token1] = ACTIONS(1391),
    [aux_sym_switch_statement_token1] = ACTIONS(1391),
    [aux_sym_foreach_statement_token1] = ACTIONS(1391),
    [aux_sym_for_statement_token1] = ACTIONS(1391),
    [aux_sym_while_statement_token1] = ACTIONS(1391),
    [aux_sym_do_statement_token1] = ACTIONS(1391),
    [aux_sym_function_statement_token1] = ACTIONS(1391),
    [aux_sym_function_statement_token2] = ACTIONS(1391),
    [aux_sym_function_statement_token3] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1391),
    [sym_label] = ACTIONS(1391),
    [aux_sym_trap_statement_token1] = ACTIONS(1391),
    [aux_sym_try_statement_token1] = ACTIONS(1391),
    [aux_sym_data_statement_token1] = ACTIONS(1391),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1391),
    [aux_sym_parallel_statement_token1] = ACTIONS(1391),
    [aux_sym_sequence_statement_token1] = ACTIONS(1391),
    [anon_sym_AMP] = ACTIONS(1391),
    [aux_sym_command_name_token1] = ACTIONS(1391),
    [aux_sym_command_name_token2] = ACTIONS(1391),
    [aux_sym_command_name_token3] = ACTIONS(1391),
    [aux_sym_command_name_token4] = ACTIONS(1391),
    [aux_sym_command_name_token5] = ACTIONS(1391),
    [aux_sym_command_name_token6] = ACTIONS(1391),
    [aux_sym_command_name_token7] = ACTIONS(1391),
    [aux_sym_command_name_token8] = ACTIONS(1391),
    [aux_sym_command_name_token9] = ACTIONS(1391),
    [aux_sym_command_name_token10] = ACTIONS(1391),
    [aux_sym_command_name_token11] = ACTIONS(1391),
    [aux_sym_command_name_token12] = ACTIONS(1391),
    [aux_sym_command_name_token13] = ACTIONS(1391),
    [aux_sym_command_name_token14] = ACTIONS(1391),
    [aux_sym_command_name_token15] = ACTIONS(1391),
    [aux_sym_command_name_token16] = ACTIONS(1391),
    [aux_sym_command_name_token17] = ACTIONS(1391),
    [aux_sym_command_name_token18] = ACTIONS(1391),
    [aux_sym_class_statement_token1] = ACTIONS(1391),
    [aux_sym_enum_statement_token1] = ACTIONS(1391),
    [anon_sym_PLUS] = ACTIONS(1391),
    [anon_sym_DASH] = ACTIONS(1391),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1391),
    [anon_sym_BANG] = ACTIONS(1391),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1391),
    [anon_sym_PLUS_PLUS] = ACTIONS(1391),
    [anon_sym_DASH_DASH] = ACTIONS(1391),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1391),
    [anon_sym_AT_LPAREN] = ACTIONS(1391),
    [anon_sym_AT_LBRACE] = ACTIONS(1391),
  },
  [STATE(332)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1393),
    [sym_hexadecimal_integer_literal] = ACTIONS(1393),
    [sym_real_literal] = ACTIONS(1393),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1393),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1393),
    [sym_verbatim_string_characters] = ACTIONS(1393),
    [sym_verbatim_here_string_characters] = ACTIONS(1393),
    [anon_sym_DOT] = ACTIONS(1393),
    [anon_sym_LBRACK] = ACTIONS(1393),
    [aux_sym_comparison_operator_token37] = ACTIONS(1393),
    [aux_sym_comparison_operator_token50] = ACTIONS(1393),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1393),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1393),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1393),
    [anon_sym_DOLLAR_] = ACTIONS(1393),
    [aux_sym_variable_token1] = ACTIONS(1393),
    [aux_sym_variable_token2] = ACTIONS(1393),
    [sym_braced_variable] = ACTIONS(1393),
    [anon_sym_SEMI] = ACTIONS(1393),
    [anon_sym_LPAREN] = ACTIONS(1393),
    [anon_sym_RPAREN] = ACTIONS(1393),
    [anon_sym_COMMA] = ACTIONS(1393),
    [anon_sym_LBRACE] = ACTIONS(1393),
    [anon_sym_RBRACE] = ACTIONS(1393),
    [aux_sym_if_statement_token1] = ACTIONS(1393),
    [aux_sym_switch_statement_token1] = ACTIONS(1393),
    [aux_sym_foreach_statement_token1] = ACTIONS(1393),
    [aux_sym_for_statement_token1] = ACTIONS(1393),
    [aux_sym_while_statement_token1] = ACTIONS(1393),
    [aux_sym_do_statement_token1] = ACTIONS(1393),
    [aux_sym_function_statement_token1] = ACTIONS(1393),
    [aux_sym_function_statement_token2] = ACTIONS(1393),
    [aux_sym_function_statement_token3] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1393),
    [sym_label] = ACTIONS(1393),
    [aux_sym_trap_statement_token1] = ACTIONS(1393),
    [aux_sym_try_statement_token1] = ACTIONS(1393),
    [aux_sym_data_statement_token1] = ACTIONS(1393),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1393),
    [aux_sym_parallel_statement_token1] = ACTIONS(1393),
    [aux_sym_sequence_statement_token1] = ACTIONS(1393),
    [anon_sym_AMP] = ACTIONS(1393),
    [aux_sym_command_name_token1] = ACTIONS(1393),
    [aux_sym_command_name_token2] = ACTIONS(1393),
    [aux_sym_command_name_token3] = ACTIONS(1393),
    [aux_sym_command_name_token4] = ACTIONS(1393),
    [aux_sym_command_name_token5] = ACTIONS(1393),
    [aux_sym_command_name_token6] = ACTIONS(1393),
    [aux_sym_command_name_token7] = ACTIONS(1393),
    [aux_sym_command_name_token8] = ACTIONS(1393),
    [aux_sym_command_name_token9] = ACTIONS(1393),
    [aux_sym_command_name_token10] = ACTIONS(1393),
    [aux_sym_command_name_token11] = ACTIONS(1393),
    [aux_sym_command_name_token12] = ACTIONS(1393),
    [aux_sym_command_name_token13] = ACTIONS(1393),
    [aux_sym_command_name_token14] = ACTIONS(1393),
    [aux_sym_command_name_token15] = ACTIONS(1393),
    [aux_sym_command_name_token16] = ACTIONS(1393),
    [aux_sym_command_name_token17] = ACTIONS(1393),
    [aux_sym_command_name_token18] = ACTIONS(1393),
    [aux_sym_class_statement_token1] = ACTIONS(1393),
    [aux_sym_enum_statement_token1] = ACTIONS(1393),
    [anon_sym_PLUS] = ACTIONS(1393),
    [anon_sym_DASH] = ACTIONS(1393),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1393),
    [anon_sym_BANG] = ACTIONS(1393),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1393),
    [anon_sym_PLUS_PLUS] = ACTIONS(1393),
    [anon_sym_DASH_DASH] = ACTIONS(1393),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1393),
    [anon_sym_AT_LPAREN] = ACTIONS(1393),
    [anon_sym_AT_LBRACE] = ACTIONS(1393),
  },
  [STATE(333)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1395),
    [sym_hexadecimal_integer_literal] = ACTIONS(1395),
    [sym_real_literal] = ACTIONS(1395),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1395),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1395),
    [sym_verbatim_string_characters] = ACTIONS(1395),
    [sym_verbatim_here_string_characters] = ACTIONS(1395),
    [anon_sym_DOT] = ACTIONS(1395),
    [anon_sym_LBRACK] = ACTIONS(1395),
    [aux_sym_comparison_operator_token37] = ACTIONS(1395),
    [aux_sym_comparison_operator_token50] = ACTIONS(1395),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1395),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1395),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1395),
    [anon_sym_DOLLAR_] = ACTIONS(1395),
    [aux_sym_variable_token1] = ACTIONS(1395),
    [aux_sym_variable_token2] = ACTIONS(1395),
    [sym_braced_variable] = ACTIONS(1395),
    [anon_sym_SEMI] = ACTIONS(1395),
    [anon_sym_LPAREN] = ACTIONS(1395),
    [anon_sym_RPAREN] = ACTIONS(1395),
    [anon_sym_COMMA] = ACTIONS(1395),
    [anon_sym_LBRACE] = ACTIONS(1395),
    [anon_sym_RBRACE] = ACTIONS(1395),
    [aux_sym_if_statement_token1] = ACTIONS(1395),
    [aux_sym_switch_statement_token1] = ACTIONS(1395),
    [aux_sym_foreach_statement_token1] = ACTIONS(1395),
    [aux_sym_for_statement_token1] = ACTIONS(1395),
    [aux_sym_while_statement_token1] = ACTIONS(1395),
    [aux_sym_do_statement_token1] = ACTIONS(1395),
    [aux_sym_function_statement_token1] = ACTIONS(1395),
    [aux_sym_function_statement_token2] = ACTIONS(1395),
    [aux_sym_function_statement_token3] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1395),
    [sym_label] = ACTIONS(1395),
    [aux_sym_trap_statement_token1] = ACTIONS(1395),
    [aux_sym_try_statement_token1] = ACTIONS(1395),
    [aux_sym_data_statement_token1] = ACTIONS(1395),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1395),
    [aux_sym_parallel_statement_token1] = ACTIONS(1395),
    [aux_sym_sequence_statement_token1] = ACTIONS(1395),
    [anon_sym_AMP] = ACTIONS(1395),
    [aux_sym_command_name_token1] = ACTIONS(1395),
    [aux_sym_command_name_token2] = ACTIONS(1395),
    [aux_sym_command_name_token3] = ACTIONS(1395),
    [aux_sym_command_name_token4] = ACTIONS(1395),
    [aux_sym_command_name_token5] = ACTIONS(1395),
    [aux_sym_command_name_token6] = ACTIONS(1395),
    [aux_sym_command_name_token7] = ACTIONS(1395),
    [aux_sym_command_name_token8] = ACTIONS(1395),
    [aux_sym_command_name_token9] = ACTIONS(1395),
    [aux_sym_command_name_token10] = ACTIONS(1395),
    [aux_sym_command_name_token11] = ACTIONS(1395),
    [aux_sym_command_name_token12] = ACTIONS(1395),
    [aux_sym_command_name_token13] = ACTIONS(1395),
    [aux_sym_command_name_token14] = ACTIONS(1395),
    [aux_sym_command_name_token15] = ACTIONS(1395),
    [aux_sym_command_name_token16] = ACTIONS(1395),
    [aux_sym_command_name_token17] = ACTIONS(1395),
    [aux_sym_command_name_token18] = ACTIONS(1395),
    [aux_sym_class_statement_token1] = ACTIONS(1395),
    [aux_sym_enum_statement_token1] = ACTIONS(1395),
    [anon_sym_PLUS] = ACTIONS(1395),
    [anon_sym_DASH] = ACTIONS(1395),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1395),
    [anon_sym_BANG] = ACTIONS(1395),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1395),
    [anon_sym_PLUS_PLUS] = ACTIONS(1395),
    [anon_sym_DASH_DASH] = ACTIONS(1395),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1395),
    [anon_sym_AT_LPAREN] = ACTIONS(1395),
    [anon_sym_AT_LBRACE] = ACTIONS(1395),
  },
  [STATE(334)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1397),
    [sym_hexadecimal_integer_literal] = ACTIONS(1397),
    [sym_real_literal] = ACTIONS(1397),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1397),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1397),
    [sym_verbatim_string_characters] = ACTIONS(1397),
    [sym_verbatim_here_string_characters] = ACTIONS(1397),
    [anon_sym_DOT] = ACTIONS(1397),
    [anon_sym_LBRACK] = ACTIONS(1397),
    [aux_sym_comparison_operator_token37] = ACTIONS(1397),
    [aux_sym_comparison_operator_token50] = ACTIONS(1397),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1397),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1397),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1397),
    [anon_sym_DOLLAR_] = ACTIONS(1397),
    [aux_sym_variable_token1] = ACTIONS(1397),
    [aux_sym_variable_token2] = ACTIONS(1397),
    [sym_braced_variable] = ACTIONS(1397),
    [anon_sym_SEMI] = ACTIONS(1397),
    [anon_sym_LPAREN] = ACTIONS(1397),
    [anon_sym_RPAREN] = ACTIONS(1397),
    [anon_sym_COMMA] = ACTIONS(1397),
    [anon_sym_LBRACE] = ACTIONS(1397),
    [anon_sym_RBRACE] = ACTIONS(1397),
    [aux_sym_if_statement_token1] = ACTIONS(1397),
    [aux_sym_switch_statement_token1] = ACTIONS(1397),
    [aux_sym_foreach_statement_token1] = ACTIONS(1397),
    [aux_sym_for_statement_token1] = ACTIONS(1397),
    [aux_sym_while_statement_token1] = ACTIONS(1397),
    [aux_sym_do_statement_token1] = ACTIONS(1397),
    [aux_sym_function_statement_token1] = ACTIONS(1397),
    [aux_sym_function_statement_token2] = ACTIONS(1397),
    [aux_sym_function_statement_token3] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1397),
    [sym_label] = ACTIONS(1397),
    [aux_sym_trap_statement_token1] = ACTIONS(1397),
    [aux_sym_try_statement_token1] = ACTIONS(1397),
    [aux_sym_data_statement_token1] = ACTIONS(1397),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1397),
    [aux_sym_parallel_statement_token1] = ACTIONS(1397),
    [aux_sym_sequence_statement_token1] = ACTIONS(1397),
    [anon_sym_AMP] = ACTIONS(1397),
    [aux_sym_command_name_token1] = ACTIONS(1397),
    [aux_sym_command_name_token2] = ACTIONS(1397),
    [aux_sym_command_name_token3] = ACTIONS(1397),
    [aux_sym_command_name_token4] = ACTIONS(1397),
    [aux_sym_command_name_token5] = ACTIONS(1397),
    [aux_sym_command_name_token6] = ACTIONS(1397),
    [aux_sym_command_name_token7] = ACTIONS(1397),
    [aux_sym_command_name_token8] = ACTIONS(1397),
    [aux_sym_command_name_token9] = ACTIONS(1397),
    [aux_sym_command_name_token10] = ACTIONS(1397),
    [aux_sym_command_name_token11] = ACTIONS(1397),
    [aux_sym_command_name_token12] = ACTIONS(1397),
    [aux_sym_command_name_token13] = ACTIONS(1397),
    [aux_sym_command_name_token14] = ACTIONS(1397),
    [aux_sym_command_name_token15] = ACTIONS(1397),
    [aux_sym_command_name_token16] = ACTIONS(1397),
    [aux_sym_command_name_token17] = ACTIONS(1397),
    [aux_sym_command_name_token18] = ACTIONS(1397),
    [aux_sym_class_statement_token1] = ACTIONS(1397),
    [aux_sym_enum_statement_token1] = ACTIONS(1397),
    [anon_sym_PLUS] = ACTIONS(1397),
    [anon_sym_DASH] = ACTIONS(1397),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1397),
    [anon_sym_BANG] = ACTIONS(1397),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1397),
    [anon_sym_PLUS_PLUS] = ACTIONS(1397),
    [anon_sym_DASH_DASH] = ACTIONS(1397),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1397),
    [anon_sym_AT_LPAREN] = ACTIONS(1397),
    [anon_sym_AT_LBRACE] = ACTIONS(1397),
  },
  [STATE(335)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1399),
    [sym_hexadecimal_integer_literal] = ACTIONS(1399),
    [sym_real_literal] = ACTIONS(1399),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1399),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1399),
    [sym_verbatim_string_characters] = ACTIONS(1399),
    [sym_verbatim_here_string_characters] = ACTIONS(1399),
    [anon_sym_DOT] = ACTIONS(1399),
    [anon_sym_LBRACK] = ACTIONS(1399),
    [aux_sym_comparison_operator_token37] = ACTIONS(1399),
    [aux_sym_comparison_operator_token50] = ACTIONS(1399),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1399),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1399),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1399),
    [anon_sym_DOLLAR_] = ACTIONS(1399),
    [aux_sym_variable_token1] = ACTIONS(1399),
    [aux_sym_variable_token2] = ACTIONS(1399),
    [sym_braced_variable] = ACTIONS(1399),
    [anon_sym_SEMI] = ACTIONS(1399),
    [anon_sym_LPAREN] = ACTIONS(1399),
    [anon_sym_RPAREN] = ACTIONS(1399),
    [anon_sym_COMMA] = ACTIONS(1399),
    [anon_sym_LBRACE] = ACTIONS(1399),
    [anon_sym_RBRACE] = ACTIONS(1399),
    [aux_sym_if_statement_token1] = ACTIONS(1399),
    [aux_sym_switch_statement_token1] = ACTIONS(1399),
    [aux_sym_foreach_statement_token1] = ACTIONS(1399),
    [aux_sym_for_statement_token1] = ACTIONS(1399),
    [aux_sym_while_statement_token1] = ACTIONS(1399),
    [aux_sym_do_statement_token1] = ACTIONS(1399),
    [aux_sym_function_statement_token1] = ACTIONS(1399),
    [aux_sym_function_statement_token2] = ACTIONS(1399),
    [aux_sym_function_statement_token3] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1399),
    [sym_label] = ACTIONS(1399),
    [aux_sym_trap_statement_token1] = ACTIONS(1399),
    [aux_sym_try_statement_token1] = ACTIONS(1399),
    [aux_sym_data_statement_token1] = ACTIONS(1399),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1399),
    [aux_sym_parallel_statement_token1] = ACTIONS(1399),
    [aux_sym_sequence_statement_token1] = ACTIONS(1399),
    [anon_sym_AMP] = ACTIONS(1399),
    [aux_sym_command_name_token1] = ACTIONS(1399),
    [aux_sym_command_name_token2] = ACTIONS(1399),
    [aux_sym_command_name_token3] = ACTIONS(1399),
    [aux_sym_command_name_token4] = ACTIONS(1399),
    [aux_sym_command_name_token5] = ACTIONS(1399),
    [aux_sym_command_name_token6] = ACTIONS(1399),
    [aux_sym_command_name_token7] = ACTIONS(1399),
    [aux_sym_command_name_token8] = ACTIONS(1399),
    [aux_sym_command_name_token9] = ACTIONS(1399),
    [aux_sym_command_name_token10] = ACTIONS(1399),
    [aux_sym_command_name_token11] = ACTIONS(1399),
    [aux_sym_command_name_token12] = ACTIONS(1399),
    [aux_sym_command_name_token13] = ACTIONS(1399),
    [aux_sym_command_name_token14] = ACTIONS(1399),
    [aux_sym_command_name_token15] = ACTIONS(1399),
    [aux_sym_command_name_token16] = ACTIONS(1399),
    [aux_sym_command_name_token17] = ACTIONS(1399),
    [aux_sym_command_name_token18] = ACTIONS(1399),
    [aux_sym_class_statement_token1] = ACTIONS(1399),
    [aux_sym_enum_statement_token1] = ACTIONS(1399),
    [anon_sym_PLUS] = ACTIONS(1399),
    [anon_sym_DASH] = ACTIONS(1399),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1399),
    [anon_sym_BANG] = ACTIONS(1399),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1399),
    [anon_sym_PLUS_PLUS] = ACTIONS(1399),
    [anon_sym_DASH_DASH] = ACTIONS(1399),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1399),
    [anon_sym_AT_LPAREN] = ACTIONS(1399),
    [anon_sym_AT_LBRACE] = ACTIONS(1399),
  },
  [STATE(336)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1401),
    [sym_hexadecimal_integer_literal] = ACTIONS(1401),
    [sym_real_literal] = ACTIONS(1401),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1401),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1401),
    [sym_verbatim_string_characters] = ACTIONS(1401),
    [sym_verbatim_here_string_characters] = ACTIONS(1401),
    [anon_sym_DOT] = ACTIONS(1401),
    [anon_sym_LBRACK] = ACTIONS(1401),
    [aux_sym_comparison_operator_token37] = ACTIONS(1401),
    [aux_sym_comparison_operator_token50] = ACTIONS(1401),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1401),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1401),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1401),
    [anon_sym_DOLLAR_] = ACTIONS(1401),
    [aux_sym_variable_token1] = ACTIONS(1401),
    [aux_sym_variable_token2] = ACTIONS(1401),
    [sym_braced_variable] = ACTIONS(1401),
    [anon_sym_SEMI] = ACTIONS(1401),
    [anon_sym_LPAREN] = ACTIONS(1401),
    [anon_sym_RPAREN] = ACTIONS(1401),
    [anon_sym_COMMA] = ACTIONS(1401),
    [anon_sym_LBRACE] = ACTIONS(1401),
    [anon_sym_RBRACE] = ACTIONS(1401),
    [aux_sym_if_statement_token1] = ACTIONS(1401),
    [aux_sym_switch_statement_token1] = ACTIONS(1401),
    [aux_sym_foreach_statement_token1] = ACTIONS(1401),
    [aux_sym_for_statement_token1] = ACTIONS(1401),
    [aux_sym_while_statement_token1] = ACTIONS(1401),
    [aux_sym_do_statement_token1] = ACTIONS(1401),
    [aux_sym_function_statement_token1] = ACTIONS(1401),
    [aux_sym_function_statement_token2] = ACTIONS(1401),
    [aux_sym_function_statement_token3] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1401),
    [sym_label] = ACTIONS(1401),
    [aux_sym_trap_statement_token1] = ACTIONS(1401),
    [aux_sym_try_statement_token1] = ACTIONS(1401),
    [aux_sym_data_statement_token1] = ACTIONS(1401),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1401),
    [aux_sym_parallel_statement_token1] = ACTIONS(1401),
    [aux_sym_sequence_statement_token1] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(1401),
    [aux_sym_command_name_token1] = ACTIONS(1401),
    [aux_sym_command_name_token2] = ACTIONS(1401),
    [aux_sym_command_name_token3] = ACTIONS(1401),
    [aux_sym_command_name_token4] = ACTIONS(1401),
    [aux_sym_command_name_token5] = ACTIONS(1401),
    [aux_sym_command_name_token6] = ACTIONS(1401),
    [aux_sym_command_name_token7] = ACTIONS(1401),
    [aux_sym_command_name_token8] = ACTIONS(1401),
    [aux_sym_command_name_token9] = ACTIONS(1401),
    [aux_sym_command_name_token10] = ACTIONS(1401),
    [aux_sym_command_name_token11] = ACTIONS(1401),
    [aux_sym_command_name_token12] = ACTIONS(1401),
    [aux_sym_command_name_token13] = ACTIONS(1401),
    [aux_sym_command_name_token14] = ACTIONS(1401),
    [aux_sym_command_name_token15] = ACTIONS(1401),
    [aux_sym_command_name_token16] = ACTIONS(1401),
    [aux_sym_command_name_token17] = ACTIONS(1401),
    [aux_sym_command_name_token18] = ACTIONS(1401),
    [aux_sym_class_statement_token1] = ACTIONS(1401),
    [aux_sym_enum_statement_token1] = ACTIONS(1401),
    [anon_sym_PLUS] = ACTIONS(1401),
    [anon_sym_DASH] = ACTIONS(1401),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1401),
    [anon_sym_BANG] = ACTIONS(1401),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1401),
    [anon_sym_PLUS_PLUS] = ACTIONS(1401),
    [anon_sym_DASH_DASH] = ACTIONS(1401),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1401),
    [anon_sym_AT_LPAREN] = ACTIONS(1401),
    [anon_sym_AT_LBRACE] = ACTIONS(1401),
  },
  [STATE(337)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1403),
    [sym_hexadecimal_integer_literal] = ACTIONS(1403),
    [sym_real_literal] = ACTIONS(1403),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1403),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1403),
    [sym_verbatim_string_characters] = ACTIONS(1403),
    [sym_verbatim_here_string_characters] = ACTIONS(1403),
    [anon_sym_DOT] = ACTIONS(1403),
    [anon_sym_LBRACK] = ACTIONS(1403),
    [aux_sym_comparison_operator_token37] = ACTIONS(1403),
    [aux_sym_comparison_operator_token50] = ACTIONS(1403),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1403),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1403),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1403),
    [anon_sym_DOLLAR_] = ACTIONS(1403),
    [aux_sym_variable_token1] = ACTIONS(1403),
    [aux_sym_variable_token2] = ACTIONS(1403),
    [sym_braced_variable] = ACTIONS(1403),
    [anon_sym_SEMI] = ACTIONS(1403),
    [anon_sym_LPAREN] = ACTIONS(1403),
    [anon_sym_RPAREN] = ACTIONS(1403),
    [anon_sym_COMMA] = ACTIONS(1403),
    [anon_sym_LBRACE] = ACTIONS(1403),
    [anon_sym_RBRACE] = ACTIONS(1403),
    [aux_sym_if_statement_token1] = ACTIONS(1403),
    [aux_sym_switch_statement_token1] = ACTIONS(1403),
    [aux_sym_foreach_statement_token1] = ACTIONS(1403),
    [aux_sym_for_statement_token1] = ACTIONS(1403),
    [aux_sym_while_statement_token1] = ACTIONS(1403),
    [aux_sym_do_statement_token1] = ACTIONS(1403),
    [aux_sym_function_statement_token1] = ACTIONS(1403),
    [aux_sym_function_statement_token2] = ACTIONS(1403),
    [aux_sym_function_statement_token3] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1403),
    [sym_label] = ACTIONS(1403),
    [aux_sym_trap_statement_token1] = ACTIONS(1403),
    [aux_sym_try_statement_token1] = ACTIONS(1403),
    [aux_sym_data_statement_token1] = ACTIONS(1403),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1403),
    [aux_sym_parallel_statement_token1] = ACTIONS(1403),
    [aux_sym_sequence_statement_token1] = ACTIONS(1403),
    [anon_sym_AMP] = ACTIONS(1403),
    [aux_sym_command_name_token1] = ACTIONS(1403),
    [aux_sym_command_name_token2] = ACTIONS(1403),
    [aux_sym_command_name_token3] = ACTIONS(1403),
    [aux_sym_command_name_token4] = ACTIONS(1403),
    [aux_sym_command_name_token5] = ACTIONS(1403),
    [aux_sym_command_name_token6] = ACTIONS(1403),
    [aux_sym_command_name_token7] = ACTIONS(1403),
    [aux_sym_command_name_token8] = ACTIONS(1403),
    [aux_sym_command_name_token9] = ACTIONS(1403),
    [aux_sym_command_name_token10] = ACTIONS(1403),
    [aux_sym_command_name_token11] = ACTIONS(1403),
    [aux_sym_command_name_token12] = ACTIONS(1403),
    [aux_sym_command_name_token13] = ACTIONS(1403),
    [aux_sym_command_name_token14] = ACTIONS(1403),
    [aux_sym_command_name_token15] = ACTIONS(1403),
    [aux_sym_command_name_token16] = ACTIONS(1403),
    [aux_sym_command_name_token17] = ACTIONS(1403),
    [aux_sym_command_name_token18] = ACTIONS(1403),
    [aux_sym_class_statement_token1] = ACTIONS(1403),
    [aux_sym_enum_statement_token1] = ACTIONS(1403),
    [anon_sym_PLUS] = ACTIONS(1403),
    [anon_sym_DASH] = ACTIONS(1403),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1403),
    [anon_sym_BANG] = ACTIONS(1403),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1403),
    [anon_sym_PLUS_PLUS] = ACTIONS(1403),
    [anon_sym_DASH_DASH] = ACTIONS(1403),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1403),
    [anon_sym_AT_LPAREN] = ACTIONS(1403),
    [anon_sym_AT_LBRACE] = ACTIONS(1403),
  },
  [STATE(338)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1405),
    [sym_hexadecimal_integer_literal] = ACTIONS(1405),
    [sym_real_literal] = ACTIONS(1405),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1405),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1405),
    [sym_verbatim_string_characters] = ACTIONS(1405),
    [sym_verbatim_here_string_characters] = ACTIONS(1405),
    [anon_sym_DOT] = ACTIONS(1405),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_comparison_operator_token37] = ACTIONS(1405),
    [aux_sym_comparison_operator_token50] = ACTIONS(1405),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1405),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1405),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1405),
    [anon_sym_DOLLAR_] = ACTIONS(1405),
    [aux_sym_variable_token1] = ACTIONS(1405),
    [aux_sym_variable_token2] = ACTIONS(1405),
    [sym_braced_variable] = ACTIONS(1405),
    [anon_sym_SEMI] = ACTIONS(1405),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_RPAREN] = ACTIONS(1405),
    [anon_sym_COMMA] = ACTIONS(1405),
    [anon_sym_LBRACE] = ACTIONS(1405),
    [anon_sym_RBRACE] = ACTIONS(1405),
    [aux_sym_if_statement_token1] = ACTIONS(1405),
    [aux_sym_switch_statement_token1] = ACTIONS(1405),
    [aux_sym_foreach_statement_token1] = ACTIONS(1405),
    [aux_sym_for_statement_token1] = ACTIONS(1405),
    [aux_sym_while_statement_token1] = ACTIONS(1405),
    [aux_sym_do_statement_token1] = ACTIONS(1405),
    [aux_sym_function_statement_token1] = ACTIONS(1405),
    [aux_sym_function_statement_token2] = ACTIONS(1405),
    [aux_sym_function_statement_token3] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1405),
    [sym_label] = ACTIONS(1405),
    [aux_sym_trap_statement_token1] = ACTIONS(1405),
    [aux_sym_try_statement_token1] = ACTIONS(1405),
    [aux_sym_data_statement_token1] = ACTIONS(1405),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1405),
    [aux_sym_parallel_statement_token1] = ACTIONS(1405),
    [aux_sym_sequence_statement_token1] = ACTIONS(1405),
    [anon_sym_AMP] = ACTIONS(1405),
    [aux_sym_command_name_token1] = ACTIONS(1405),
    [aux_sym_command_name_token2] = ACTIONS(1405),
    [aux_sym_command_name_token3] = ACTIONS(1405),
    [aux_sym_command_name_token4] = ACTIONS(1405),
    [aux_sym_command_name_token5] = ACTIONS(1405),
    [aux_sym_command_name_token6] = ACTIONS(1405),
    [aux_sym_command_name_token7] = ACTIONS(1405),
    [aux_sym_command_name_token8] = ACTIONS(1405),
    [aux_sym_command_name_token9] = ACTIONS(1405),
    [aux_sym_command_name_token10] = ACTIONS(1405),
    [aux_sym_command_name_token11] = ACTIONS(1405),
    [aux_sym_command_name_token12] = ACTIONS(1405),
    [aux_sym_command_name_token13] = ACTIONS(1405),
    [aux_sym_command_name_token14] = ACTIONS(1405),
    [aux_sym_command_name_token15] = ACTIONS(1405),
    [aux_sym_command_name_token16] = ACTIONS(1405),
    [aux_sym_command_name_token17] = ACTIONS(1405),
    [aux_sym_command_name_token18] = ACTIONS(1405),
    [aux_sym_class_statement_token1] = ACTIONS(1405),
    [aux_sym_enum_statement_token1] = ACTIONS(1405),
    [anon_sym_PLUS] = ACTIONS(1405),
    [anon_sym_DASH] = ACTIONS(1405),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1405),
    [anon_sym_BANG] = ACTIONS(1405),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1405),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1405),
    [anon_sym_AT_LPAREN] = ACTIONS(1405),
    [anon_sym_AT_LBRACE] = ACTIONS(1405),
  },
  [STATE(339)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1407),
    [sym_hexadecimal_integer_literal] = ACTIONS(1407),
    [sym_real_literal] = ACTIONS(1407),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1407),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1407),
    [sym_verbatim_string_characters] = ACTIONS(1407),
    [sym_verbatim_here_string_characters] = ACTIONS(1407),
    [anon_sym_DOT] = ACTIONS(1407),
    [anon_sym_LBRACK] = ACTIONS(1407),
    [aux_sym_comparison_operator_token37] = ACTIONS(1407),
    [aux_sym_comparison_operator_token50] = ACTIONS(1407),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1407),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1407),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1407),
    [anon_sym_DOLLAR_] = ACTIONS(1407),
    [aux_sym_variable_token1] = ACTIONS(1407),
    [aux_sym_variable_token2] = ACTIONS(1407),
    [sym_braced_variable] = ACTIONS(1407),
    [anon_sym_SEMI] = ACTIONS(1407),
    [anon_sym_LPAREN] = ACTIONS(1407),
    [anon_sym_RPAREN] = ACTIONS(1407),
    [anon_sym_COMMA] = ACTIONS(1407),
    [anon_sym_LBRACE] = ACTIONS(1407),
    [anon_sym_RBRACE] = ACTIONS(1407),
    [aux_sym_if_statement_token1] = ACTIONS(1407),
    [aux_sym_switch_statement_token1] = ACTIONS(1407),
    [aux_sym_foreach_statement_token1] = ACTIONS(1407),
    [aux_sym_for_statement_token1] = ACTIONS(1407),
    [aux_sym_while_statement_token1] = ACTIONS(1407),
    [aux_sym_do_statement_token1] = ACTIONS(1407),
    [aux_sym_function_statement_token1] = ACTIONS(1407),
    [aux_sym_function_statement_token2] = ACTIONS(1407),
    [aux_sym_function_statement_token3] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1407),
    [sym_label] = ACTIONS(1407),
    [aux_sym_trap_statement_token1] = ACTIONS(1407),
    [aux_sym_try_statement_token1] = ACTIONS(1407),
    [aux_sym_data_statement_token1] = ACTIONS(1407),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1407),
    [aux_sym_parallel_statement_token1] = ACTIONS(1407),
    [aux_sym_sequence_statement_token1] = ACTIONS(1407),
    [anon_sym_AMP] = ACTIONS(1407),
    [aux_sym_command_name_token1] = ACTIONS(1407),
    [aux_sym_command_name_token2] = ACTIONS(1407),
    [aux_sym_command_name_token3] = ACTIONS(1407),
    [aux_sym_command_name_token4] = ACTIONS(1407),
    [aux_sym_command_name_token5] = ACTIONS(1407),
    [aux_sym_command_name_token6] = ACTIONS(1407),
    [aux_sym_command_name_token7] = ACTIONS(1407),
    [aux_sym_command_name_token8] = ACTIONS(1407),
    [aux_sym_command_name_token9] = ACTIONS(1407),
    [aux_sym_command_name_token10] = ACTIONS(1407),
    [aux_sym_command_name_token11] = ACTIONS(1407),
    [aux_sym_command_name_token12] = ACTIONS(1407),
    [aux_sym_command_name_token13] = ACTIONS(1407),
    [aux_sym_command_name_token14] = ACTIONS(1407),
    [aux_sym_command_name_token15] = ACTIONS(1407),
    [aux_sym_command_name_token16] = ACTIONS(1407),
    [aux_sym_command_name_token17] = ACTIONS(1407),
    [aux_sym_command_name_token18] = ACTIONS(1407),
    [aux_sym_class_statement_token1] = ACTIONS(1407),
    [aux_sym_enum_statement_token1] = ACTIONS(1407),
    [anon_sym_PLUS] = ACTIONS(1407),
    [anon_sym_DASH] = ACTIONS(1407),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1407),
    [anon_sym_BANG] = ACTIONS(1407),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1407),
    [anon_sym_PLUS_PLUS] = ACTIONS(1407),
    [anon_sym_DASH_DASH] = ACTIONS(1407),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1407),
    [anon_sym_AT_LPAREN] = ACTIONS(1407),
    [anon_sym_AT_LBRACE] = ACTIONS(1407),
  },
  [STATE(340)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1409),
    [sym_hexadecimal_integer_literal] = ACTIONS(1409),
    [sym_real_literal] = ACTIONS(1409),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1409),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1409),
    [sym_verbatim_string_characters] = ACTIONS(1409),
    [sym_verbatim_here_string_characters] = ACTIONS(1409),
    [anon_sym_DOT] = ACTIONS(1409),
    [anon_sym_LBRACK] = ACTIONS(1409),
    [aux_sym_comparison_operator_token37] = ACTIONS(1409),
    [aux_sym_comparison_operator_token50] = ACTIONS(1409),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1409),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1409),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1409),
    [anon_sym_DOLLAR_] = ACTIONS(1409),
    [aux_sym_variable_token1] = ACTIONS(1409),
    [aux_sym_variable_token2] = ACTIONS(1409),
    [sym_braced_variable] = ACTIONS(1409),
    [anon_sym_SEMI] = ACTIONS(1409),
    [anon_sym_LPAREN] = ACTIONS(1409),
    [anon_sym_RPAREN] = ACTIONS(1409),
    [anon_sym_COMMA] = ACTIONS(1409),
    [anon_sym_LBRACE] = ACTIONS(1409),
    [anon_sym_RBRACE] = ACTIONS(1409),
    [aux_sym_if_statement_token1] = ACTIONS(1409),
    [aux_sym_switch_statement_token1] = ACTIONS(1409),
    [aux_sym_foreach_statement_token1] = ACTIONS(1409),
    [aux_sym_for_statement_token1] = ACTIONS(1409),
    [aux_sym_while_statement_token1] = ACTIONS(1409),
    [aux_sym_do_statement_token1] = ACTIONS(1409),
    [aux_sym_function_statement_token1] = ACTIONS(1409),
    [aux_sym_function_statement_token2] = ACTIONS(1409),
    [aux_sym_function_statement_token3] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1409),
    [sym_label] = ACTIONS(1409),
    [aux_sym_trap_statement_token1] = ACTIONS(1409),
    [aux_sym_try_statement_token1] = ACTIONS(1409),
    [aux_sym_data_statement_token1] = ACTIONS(1409),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1409),
    [aux_sym_parallel_statement_token1] = ACTIONS(1409),
    [aux_sym_sequence_statement_token1] = ACTIONS(1409),
    [anon_sym_AMP] = ACTIONS(1409),
    [aux_sym_command_name_token1] = ACTIONS(1409),
    [aux_sym_command_name_token2] = ACTIONS(1409),
    [aux_sym_command_name_token3] = ACTIONS(1409),
    [aux_sym_command_name_token4] = ACTIONS(1409),
    [aux_sym_command_name_token5] = ACTIONS(1409),
    [aux_sym_command_name_token6] = ACTIONS(1409),
    [aux_sym_command_name_token7] = ACTIONS(1409),
    [aux_sym_command_name_token8] = ACTIONS(1409),
    [aux_sym_command_name_token9] = ACTIONS(1409),
    [aux_sym_command_name_token10] = ACTIONS(1409),
    [aux_sym_command_name_token11] = ACTIONS(1409),
    [aux_sym_command_name_token12] = ACTIONS(1409),
    [aux_sym_command_name_token13] = ACTIONS(1409),
    [aux_sym_command_name_token14] = ACTIONS(1409),
    [aux_sym_command_name_token15] = ACTIONS(1409),
    [aux_sym_command_name_token16] = ACTIONS(1409),
    [aux_sym_command_name_token17] = ACTIONS(1409),
    [aux_sym_command_name_token18] = ACTIONS(1409),
    [aux_sym_class_statement_token1] = ACTIONS(1409),
    [aux_sym_enum_statement_token1] = ACTIONS(1409),
    [anon_sym_PLUS] = ACTIONS(1409),
    [anon_sym_DASH] = ACTIONS(1409),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1409),
    [anon_sym_BANG] = ACTIONS(1409),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1409),
    [anon_sym_PLUS_PLUS] = ACTIONS(1409),
    [anon_sym_DASH_DASH] = ACTIONS(1409),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1409),
    [anon_sym_AT_LPAREN] = ACTIONS(1409),
    [anon_sym_AT_LBRACE] = ACTIONS(1409),
  },
  [STATE(341)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1411),
    [sym_hexadecimal_integer_literal] = ACTIONS(1411),
    [sym_real_literal] = ACTIONS(1411),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1411),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1411),
    [sym_verbatim_string_characters] = ACTIONS(1411),
    [sym_verbatim_here_string_characters] = ACTIONS(1411),
    [anon_sym_DOT] = ACTIONS(1411),
    [anon_sym_LBRACK] = ACTIONS(1411),
    [aux_sym_comparison_operator_token37] = ACTIONS(1411),
    [aux_sym_comparison_operator_token50] = ACTIONS(1411),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1411),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1411),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1411),
    [anon_sym_DOLLAR_] = ACTIONS(1411),
    [aux_sym_variable_token1] = ACTIONS(1411),
    [aux_sym_variable_token2] = ACTIONS(1411),
    [sym_braced_variable] = ACTIONS(1411),
    [anon_sym_SEMI] = ACTIONS(1411),
    [anon_sym_LPAREN] = ACTIONS(1411),
    [anon_sym_RPAREN] = ACTIONS(1411),
    [anon_sym_COMMA] = ACTIONS(1411),
    [anon_sym_LBRACE] = ACTIONS(1411),
    [anon_sym_RBRACE] = ACTIONS(1411),
    [aux_sym_if_statement_token1] = ACTIONS(1411),
    [aux_sym_switch_statement_token1] = ACTIONS(1411),
    [aux_sym_foreach_statement_token1] = ACTIONS(1411),
    [aux_sym_for_statement_token1] = ACTIONS(1411),
    [aux_sym_while_statement_token1] = ACTIONS(1411),
    [aux_sym_do_statement_token1] = ACTIONS(1411),
    [aux_sym_function_statement_token1] = ACTIONS(1411),
    [aux_sym_function_statement_token2] = ACTIONS(1411),
    [aux_sym_function_statement_token3] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1411),
    [sym_label] = ACTIONS(1411),
    [aux_sym_trap_statement_token1] = ACTIONS(1411),
    [aux_sym_try_statement_token1] = ACTIONS(1411),
    [aux_sym_data_statement_token1] = ACTIONS(1411),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1411),
    [aux_sym_parallel_statement_token1] = ACTIONS(1411),
    [aux_sym_sequence_statement_token1] = ACTIONS(1411),
    [anon_sym_AMP] = ACTIONS(1411),
    [aux_sym_command_name_token1] = ACTIONS(1411),
    [aux_sym_command_name_token2] = ACTIONS(1411),
    [aux_sym_command_name_token3] = ACTIONS(1411),
    [aux_sym_command_name_token4] = ACTIONS(1411),
    [aux_sym_command_name_token5] = ACTIONS(1411),
    [aux_sym_command_name_token6] = ACTIONS(1411),
    [aux_sym_command_name_token7] = ACTIONS(1411),
    [aux_sym_command_name_token8] = ACTIONS(1411),
    [aux_sym_command_name_token9] = ACTIONS(1411),
    [aux_sym_command_name_token10] = ACTIONS(1411),
    [aux_sym_command_name_token11] = ACTIONS(1411),
    [aux_sym_command_name_token12] = ACTIONS(1411),
    [aux_sym_command_name_token13] = ACTIONS(1411),
    [aux_sym_command_name_token14] = ACTIONS(1411),
    [aux_sym_command_name_token15] = ACTIONS(1411),
    [aux_sym_command_name_token16] = ACTIONS(1411),
    [aux_sym_command_name_token17] = ACTIONS(1411),
    [aux_sym_command_name_token18] = ACTIONS(1411),
    [aux_sym_class_statement_token1] = ACTIONS(1411),
    [aux_sym_enum_statement_token1] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1411),
    [anon_sym_DASH] = ACTIONS(1411),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1411),
    [anon_sym_BANG] = ACTIONS(1411),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1411),
    [anon_sym_PLUS_PLUS] = ACTIONS(1411),
    [anon_sym_DASH_DASH] = ACTIONS(1411),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1411),
    [anon_sym_AT_LPAREN] = ACTIONS(1411),
    [anon_sym_AT_LBRACE] = ACTIONS(1411),
  },
  [STATE(342)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1413),
    [sym_hexadecimal_integer_literal] = ACTIONS(1413),
    [sym_real_literal] = ACTIONS(1413),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1413),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1413),
    [sym_verbatim_string_characters] = ACTIONS(1413),
    [sym_verbatim_here_string_characters] = ACTIONS(1413),
    [anon_sym_DOT] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1413),
    [aux_sym_comparison_operator_token37] = ACTIONS(1413),
    [aux_sym_comparison_operator_token50] = ACTIONS(1413),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1413),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1413),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1413),
    [anon_sym_DOLLAR_] = ACTIONS(1413),
    [aux_sym_variable_token1] = ACTIONS(1413),
    [aux_sym_variable_token2] = ACTIONS(1413),
    [sym_braced_variable] = ACTIONS(1413),
    [anon_sym_SEMI] = ACTIONS(1413),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_RPAREN] = ACTIONS(1413),
    [anon_sym_COMMA] = ACTIONS(1413),
    [anon_sym_LBRACE] = ACTIONS(1413),
    [anon_sym_RBRACE] = ACTIONS(1413),
    [aux_sym_if_statement_token1] = ACTIONS(1413),
    [aux_sym_switch_statement_token1] = ACTIONS(1413),
    [aux_sym_foreach_statement_token1] = ACTIONS(1413),
    [aux_sym_for_statement_token1] = ACTIONS(1413),
    [aux_sym_while_statement_token1] = ACTIONS(1413),
    [aux_sym_do_statement_token1] = ACTIONS(1413),
    [aux_sym_function_statement_token1] = ACTIONS(1413),
    [aux_sym_function_statement_token2] = ACTIONS(1413),
    [aux_sym_function_statement_token3] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1413),
    [sym_label] = ACTIONS(1413),
    [aux_sym_trap_statement_token1] = ACTIONS(1413),
    [aux_sym_try_statement_token1] = ACTIONS(1413),
    [aux_sym_data_statement_token1] = ACTIONS(1413),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1413),
    [aux_sym_parallel_statement_token1] = ACTIONS(1413),
    [aux_sym_sequence_statement_token1] = ACTIONS(1413),
    [anon_sym_AMP] = ACTIONS(1413),
    [aux_sym_command_name_token1] = ACTIONS(1413),
    [aux_sym_command_name_token2] = ACTIONS(1413),
    [aux_sym_command_name_token3] = ACTIONS(1413),
    [aux_sym_command_name_token4] = ACTIONS(1413),
    [aux_sym_command_name_token5] = ACTIONS(1413),
    [aux_sym_command_name_token6] = ACTIONS(1413),
    [aux_sym_command_name_token7] = ACTIONS(1413),
    [aux_sym_command_name_token8] = ACTIONS(1413),
    [aux_sym_command_name_token9] = ACTIONS(1413),
    [aux_sym_command_name_token10] = ACTIONS(1413),
    [aux_sym_command_name_token11] = ACTIONS(1413),
    [aux_sym_command_name_token12] = ACTIONS(1413),
    [aux_sym_command_name_token13] = ACTIONS(1413),
    [aux_sym_command_name_token14] = ACTIONS(1413),
    [aux_sym_command_name_token15] = ACTIONS(1413),
    [aux_sym_command_name_token16] = ACTIONS(1413),
    [aux_sym_command_name_token17] = ACTIONS(1413),
    [aux_sym_command_name_token18] = ACTIONS(1413),
    [aux_sym_class_statement_token1] = ACTIONS(1413),
    [aux_sym_enum_statement_token1] = ACTIONS(1413),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1413),
    [anon_sym_PLUS_PLUS] = ACTIONS(1413),
    [anon_sym_DASH_DASH] = ACTIONS(1413),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1413),
    [anon_sym_AT_LPAREN] = ACTIONS(1413),
    [anon_sym_AT_LBRACE] = ACTIONS(1413),
  },
  [STATE(343)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1415),
    [sym_hexadecimal_integer_literal] = ACTIONS(1415),
    [sym_real_literal] = ACTIONS(1415),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1415),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1415),
    [sym_verbatim_string_characters] = ACTIONS(1415),
    [sym_verbatim_here_string_characters] = ACTIONS(1415),
    [anon_sym_DOT] = ACTIONS(1415),
    [anon_sym_LBRACK] = ACTIONS(1415),
    [aux_sym_comparison_operator_token37] = ACTIONS(1415),
    [aux_sym_comparison_operator_token50] = ACTIONS(1415),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1415),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1415),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1415),
    [anon_sym_DOLLAR_] = ACTIONS(1415),
    [aux_sym_variable_token1] = ACTIONS(1415),
    [aux_sym_variable_token2] = ACTIONS(1415),
    [sym_braced_variable] = ACTIONS(1415),
    [anon_sym_SEMI] = ACTIONS(1415),
    [anon_sym_LPAREN] = ACTIONS(1415),
    [anon_sym_RPAREN] = ACTIONS(1415),
    [anon_sym_COMMA] = ACTIONS(1415),
    [anon_sym_LBRACE] = ACTIONS(1415),
    [anon_sym_RBRACE] = ACTIONS(1415),
    [aux_sym_if_statement_token1] = ACTIONS(1415),
    [aux_sym_switch_statement_token1] = ACTIONS(1415),
    [aux_sym_foreach_statement_token1] = ACTIONS(1415),
    [aux_sym_for_statement_token1] = ACTIONS(1415),
    [aux_sym_while_statement_token1] = ACTIONS(1415),
    [aux_sym_do_statement_token1] = ACTIONS(1415),
    [aux_sym_function_statement_token1] = ACTIONS(1415),
    [aux_sym_function_statement_token2] = ACTIONS(1415),
    [aux_sym_function_statement_token3] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1415),
    [sym_label] = ACTIONS(1415),
    [aux_sym_trap_statement_token1] = ACTIONS(1415),
    [aux_sym_try_statement_token1] = ACTIONS(1415),
    [aux_sym_data_statement_token1] = ACTIONS(1415),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1415),
    [aux_sym_parallel_statement_token1] = ACTIONS(1415),
    [aux_sym_sequence_statement_token1] = ACTIONS(1415),
    [anon_sym_AMP] = ACTIONS(1415),
    [aux_sym_command_name_token1] = ACTIONS(1415),
    [aux_sym_command_name_token2] = ACTIONS(1415),
    [aux_sym_command_name_token3] = ACTIONS(1415),
    [aux_sym_command_name_token4] = ACTIONS(1415),
    [aux_sym_command_name_token5] = ACTIONS(1415),
    [aux_sym_command_name_token6] = ACTIONS(1415),
    [aux_sym_command_name_token7] = ACTIONS(1415),
    [aux_sym_command_name_token8] = ACTIONS(1415),
    [aux_sym_command_name_token9] = ACTIONS(1415),
    [aux_sym_command_name_token10] = ACTIONS(1415),
    [aux_sym_command_name_token11] = ACTIONS(1415),
    [aux_sym_command_name_token12] = ACTIONS(1415),
    [aux_sym_command_name_token13] = ACTIONS(1415),
    [aux_sym_command_name_token14] = ACTIONS(1415),
    [aux_sym_command_name_token15] = ACTIONS(1415),
    [aux_sym_command_name_token16] = ACTIONS(1415),
    [aux_sym_command_name_token17] = ACTIONS(1415),
    [aux_sym_command_name_token18] = ACTIONS(1415),
    [aux_sym_class_statement_token1] = ACTIONS(1415),
    [aux_sym_enum_statement_token1] = ACTIONS(1415),
    [anon_sym_PLUS] = ACTIONS(1415),
    [anon_sym_DASH] = ACTIONS(1415),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1415),
    [anon_sym_BANG] = ACTIONS(1415),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1415),
    [anon_sym_PLUS_PLUS] = ACTIONS(1415),
    [anon_sym_DASH_DASH] = ACTIONS(1415),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1415),
    [anon_sym_AT_LPAREN] = ACTIONS(1415),
    [anon_sym_AT_LBRACE] = ACTIONS(1415),
  },
  [STATE(344)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1417),
    [sym_hexadecimal_integer_literal] = ACTIONS(1417),
    [sym_real_literal] = ACTIONS(1417),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1417),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1417),
    [sym_verbatim_string_characters] = ACTIONS(1417),
    [sym_verbatim_here_string_characters] = ACTIONS(1417),
    [anon_sym_DOT] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1417),
    [aux_sym_comparison_operator_token37] = ACTIONS(1417),
    [aux_sym_comparison_operator_token50] = ACTIONS(1417),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1417),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1417),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1417),
    [anon_sym_DOLLAR_] = ACTIONS(1417),
    [aux_sym_variable_token1] = ACTIONS(1417),
    [aux_sym_variable_token2] = ACTIONS(1417),
    [sym_braced_variable] = ACTIONS(1417),
    [anon_sym_SEMI] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1417),
    [anon_sym_RPAREN] = ACTIONS(1417),
    [anon_sym_COMMA] = ACTIONS(1417),
    [anon_sym_LBRACE] = ACTIONS(1417),
    [anon_sym_RBRACE] = ACTIONS(1417),
    [aux_sym_if_statement_token1] = ACTIONS(1417),
    [aux_sym_switch_statement_token1] = ACTIONS(1417),
    [aux_sym_foreach_statement_token1] = ACTIONS(1417),
    [aux_sym_for_statement_token1] = ACTIONS(1417),
    [aux_sym_while_statement_token1] = ACTIONS(1417),
    [aux_sym_do_statement_token1] = ACTIONS(1417),
    [aux_sym_function_statement_token1] = ACTIONS(1417),
    [aux_sym_function_statement_token2] = ACTIONS(1417),
    [aux_sym_function_statement_token3] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1417),
    [sym_label] = ACTIONS(1417),
    [aux_sym_trap_statement_token1] = ACTIONS(1417),
    [aux_sym_try_statement_token1] = ACTIONS(1417),
    [aux_sym_data_statement_token1] = ACTIONS(1417),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1417),
    [aux_sym_parallel_statement_token1] = ACTIONS(1417),
    [aux_sym_sequence_statement_token1] = ACTIONS(1417),
    [anon_sym_AMP] = ACTIONS(1417),
    [aux_sym_command_name_token1] = ACTIONS(1417),
    [aux_sym_command_name_token2] = ACTIONS(1417),
    [aux_sym_command_name_token3] = ACTIONS(1417),
    [aux_sym_command_name_token4] = ACTIONS(1417),
    [aux_sym_command_name_token5] = ACTIONS(1417),
    [aux_sym_command_name_token6] = ACTIONS(1417),
    [aux_sym_command_name_token7] = ACTIONS(1417),
    [aux_sym_command_name_token8] = ACTIONS(1417),
    [aux_sym_command_name_token9] = ACTIONS(1417),
    [aux_sym_command_name_token10] = ACTIONS(1417),
    [aux_sym_command_name_token11] = ACTIONS(1417),
    [aux_sym_command_name_token12] = ACTIONS(1417),
    [aux_sym_command_name_token13] = ACTIONS(1417),
    [aux_sym_command_name_token14] = ACTIONS(1417),
    [aux_sym_command_name_token15] = ACTIONS(1417),
    [aux_sym_command_name_token16] = ACTIONS(1417),
    [aux_sym_command_name_token17] = ACTIONS(1417),
    [aux_sym_command_name_token18] = ACTIONS(1417),
    [aux_sym_class_statement_token1] = ACTIONS(1417),
    [aux_sym_enum_statement_token1] = ACTIONS(1417),
    [anon_sym_PLUS] = ACTIONS(1417),
    [anon_sym_DASH] = ACTIONS(1417),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1417),
    [anon_sym_BANG] = ACTIONS(1417),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1417),
    [anon_sym_PLUS_PLUS] = ACTIONS(1417),
    [anon_sym_DASH_DASH] = ACTIONS(1417),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1417),
    [anon_sym_AT_LPAREN] = ACTIONS(1417),
    [anon_sym_AT_LBRACE] = ACTIONS(1417),
  },
  [STATE(345)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1365),
    [sym_hexadecimal_integer_literal] = ACTIONS(1365),
    [sym_real_literal] = ACTIONS(1365),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1365),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1365),
    [sym_verbatim_string_characters] = ACTIONS(1365),
    [sym_verbatim_here_string_characters] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1365),
    [anon_sym_LBRACK] = ACTIONS(1365),
    [aux_sym_comparison_operator_token37] = ACTIONS(1365),
    [aux_sym_comparison_operator_token50] = ACTIONS(1365),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1365),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1365),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1365),
    [anon_sym_DOLLAR_] = ACTIONS(1365),
    [aux_sym_variable_token1] = ACTIONS(1365),
    [aux_sym_variable_token2] = ACTIONS(1365),
    [sym_braced_variable] = ACTIONS(1365),
    [anon_sym_SEMI] = ACTIONS(1365),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [anon_sym_RPAREN] = ACTIONS(1365),
    [anon_sym_COMMA] = ACTIONS(1365),
    [anon_sym_LBRACE] = ACTIONS(1365),
    [anon_sym_RBRACE] = ACTIONS(1365),
    [aux_sym_if_statement_token1] = ACTIONS(1365),
    [aux_sym_switch_statement_token1] = ACTIONS(1365),
    [aux_sym_foreach_statement_token1] = ACTIONS(1365),
    [aux_sym_for_statement_token1] = ACTIONS(1365),
    [aux_sym_while_statement_token1] = ACTIONS(1365),
    [aux_sym_do_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token2] = ACTIONS(1365),
    [aux_sym_function_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1365),
    [sym_label] = ACTIONS(1365),
    [aux_sym_trap_statement_token1] = ACTIONS(1365),
    [aux_sym_try_statement_token1] = ACTIONS(1365),
    [aux_sym_data_statement_token1] = ACTIONS(1365),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1365),
    [aux_sym_parallel_statement_token1] = ACTIONS(1365),
    [aux_sym_sequence_statement_token1] = ACTIONS(1365),
    [anon_sym_AMP] = ACTIONS(1365),
    [aux_sym_command_name_token1] = ACTIONS(1365),
    [aux_sym_command_name_token2] = ACTIONS(1365),
    [aux_sym_command_name_token3] = ACTIONS(1365),
    [aux_sym_command_name_token4] = ACTIONS(1365),
    [aux_sym_command_name_token5] = ACTIONS(1365),
    [aux_sym_command_name_token6] = ACTIONS(1365),
    [aux_sym_command_name_token7] = ACTIONS(1365),
    [aux_sym_command_name_token8] = ACTIONS(1365),
    [aux_sym_command_name_token9] = ACTIONS(1365),
    [aux_sym_command_name_token10] = ACTIONS(1365),
    [aux_sym_command_name_token11] = ACTIONS(1365),
    [aux_sym_command_name_token12] = ACTIONS(1365),
    [aux_sym_command_name_token13] = ACTIONS(1365),
    [aux_sym_command_name_token14] = ACTIONS(1365),
    [aux_sym_command_name_token15] = ACTIONS(1365),
    [aux_sym_command_name_token16] = ACTIONS(1365),
    [aux_sym_command_name_token17] = ACTIONS(1365),
    [aux_sym_command_name_token18] = ACTIONS(1365),
    [aux_sym_class_statement_token1] = ACTIONS(1365),
    [aux_sym_enum_statement_token1] = ACTIONS(1365),
    [anon_sym_PLUS] = ACTIONS(1365),
    [anon_sym_DASH] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1365),
    [anon_sym_BANG] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1365),
    [anon_sym_PLUS_PLUS] = ACTIONS(1365),
    [anon_sym_DASH_DASH] = ACTIONS(1365),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LBRACE] = ACTIONS(1365),
  },
  [STATE(346)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1419),
    [sym_hexadecimal_integer_literal] = ACTIONS(1419),
    [sym_real_literal] = ACTIONS(1419),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1419),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1419),
    [sym_verbatim_string_characters] = ACTIONS(1419),
    [sym_verbatim_here_string_characters] = ACTIONS(1419),
    [anon_sym_DOT] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [aux_sym_comparison_operator_token37] = ACTIONS(1419),
    [aux_sym_comparison_operator_token50] = ACTIONS(1419),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1419),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1419),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1419),
    [anon_sym_DOLLAR_] = ACTIONS(1419),
    [aux_sym_variable_token1] = ACTIONS(1419),
    [aux_sym_variable_token2] = ACTIONS(1419),
    [sym_braced_variable] = ACTIONS(1419),
    [anon_sym_SEMI] = ACTIONS(1419),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1419),
    [anon_sym_COMMA] = ACTIONS(1419),
    [anon_sym_LBRACE] = ACTIONS(1419),
    [anon_sym_RBRACE] = ACTIONS(1419),
    [aux_sym_if_statement_token1] = ACTIONS(1419),
    [aux_sym_switch_statement_token1] = ACTIONS(1419),
    [aux_sym_foreach_statement_token1] = ACTIONS(1419),
    [aux_sym_for_statement_token1] = ACTIONS(1419),
    [aux_sym_while_statement_token1] = ACTIONS(1419),
    [aux_sym_do_statement_token1] = ACTIONS(1419),
    [aux_sym_function_statement_token1] = ACTIONS(1419),
    [aux_sym_function_statement_token2] = ACTIONS(1419),
    [aux_sym_function_statement_token3] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1419),
    [sym_label] = ACTIONS(1419),
    [aux_sym_trap_statement_token1] = ACTIONS(1419),
    [aux_sym_try_statement_token1] = ACTIONS(1419),
    [aux_sym_data_statement_token1] = ACTIONS(1419),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1419),
    [aux_sym_parallel_statement_token1] = ACTIONS(1419),
    [aux_sym_sequence_statement_token1] = ACTIONS(1419),
    [anon_sym_AMP] = ACTIONS(1419),
    [aux_sym_command_name_token1] = ACTIONS(1419),
    [aux_sym_command_name_token2] = ACTIONS(1419),
    [aux_sym_command_name_token3] = ACTIONS(1419),
    [aux_sym_command_name_token4] = ACTIONS(1419),
    [aux_sym_command_name_token5] = ACTIONS(1419),
    [aux_sym_command_name_token6] = ACTIONS(1419),
    [aux_sym_command_name_token7] = ACTIONS(1419),
    [aux_sym_command_name_token8] = ACTIONS(1419),
    [aux_sym_command_name_token9] = ACTIONS(1419),
    [aux_sym_command_name_token10] = ACTIONS(1419),
    [aux_sym_command_name_token11] = ACTIONS(1419),
    [aux_sym_command_name_token12] = ACTIONS(1419),
    [aux_sym_command_name_token13] = ACTIONS(1419),
    [aux_sym_command_name_token14] = ACTIONS(1419),
    [aux_sym_command_name_token15] = ACTIONS(1419),
    [aux_sym_command_name_token16] = ACTIONS(1419),
    [aux_sym_command_name_token17] = ACTIONS(1419),
    [aux_sym_command_name_token18] = ACTIONS(1419),
    [aux_sym_class_statement_token1] = ACTIONS(1419),
    [aux_sym_enum_statement_token1] = ACTIONS(1419),
    [anon_sym_PLUS] = ACTIONS(1419),
    [anon_sym_DASH] = ACTIONS(1419),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1419),
    [anon_sym_BANG] = ACTIONS(1419),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1419),
    [anon_sym_PLUS_PLUS] = ACTIONS(1419),
    [anon_sym_DASH_DASH] = ACTIONS(1419),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1419),
    [anon_sym_AT_LPAREN] = ACTIONS(1419),
    [anon_sym_AT_LBRACE] = ACTIONS(1419),
  },
  [STATE(347)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1421),
    [sym_hexadecimal_integer_literal] = ACTIONS(1421),
    [sym_real_literal] = ACTIONS(1421),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1421),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1421),
    [sym_verbatim_string_characters] = ACTIONS(1421),
    [sym_verbatim_here_string_characters] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1421),
    [aux_sym_comparison_operator_token37] = ACTIONS(1421),
    [aux_sym_comparison_operator_token50] = ACTIONS(1421),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1421),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1421),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1421),
    [anon_sym_DOLLAR_] = ACTIONS(1421),
    [aux_sym_variable_token1] = ACTIONS(1421),
    [aux_sym_variable_token2] = ACTIONS(1421),
    [sym_braced_variable] = ACTIONS(1421),
    [anon_sym_SEMI] = ACTIONS(1421),
    [anon_sym_LPAREN] = ACTIONS(1421),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_COMMA] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1421),
    [anon_sym_RBRACE] = ACTIONS(1421),
    [aux_sym_if_statement_token1] = ACTIONS(1421),
    [aux_sym_switch_statement_token1] = ACTIONS(1421),
    [aux_sym_foreach_statement_token1] = ACTIONS(1421),
    [aux_sym_for_statement_token1] = ACTIONS(1421),
    [aux_sym_while_statement_token1] = ACTIONS(1421),
    [aux_sym_do_statement_token1] = ACTIONS(1421),
    [aux_sym_function_statement_token1] = ACTIONS(1421),
    [aux_sym_function_statement_token2] = ACTIONS(1421),
    [aux_sym_function_statement_token3] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1421),
    [sym_label] = ACTIONS(1421),
    [aux_sym_trap_statement_token1] = ACTIONS(1421),
    [aux_sym_try_statement_token1] = ACTIONS(1421),
    [aux_sym_data_statement_token1] = ACTIONS(1421),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1421),
    [aux_sym_parallel_statement_token1] = ACTIONS(1421),
    [aux_sym_sequence_statement_token1] = ACTIONS(1421),
    [anon_sym_AMP] = ACTIONS(1421),
    [aux_sym_command_name_token1] = ACTIONS(1421),
    [aux_sym_command_name_token2] = ACTIONS(1421),
    [aux_sym_command_name_token3] = ACTIONS(1421),
    [aux_sym_command_name_token4] = ACTIONS(1421),
    [aux_sym_command_name_token5] = ACTIONS(1421),
    [aux_sym_command_name_token6] = ACTIONS(1421),
    [aux_sym_command_name_token7] = ACTIONS(1421),
    [aux_sym_command_name_token8] = ACTIONS(1421),
    [aux_sym_command_name_token9] = ACTIONS(1421),
    [aux_sym_command_name_token10] = ACTIONS(1421),
    [aux_sym_command_name_token11] = ACTIONS(1421),
    [aux_sym_command_name_token12] = ACTIONS(1421),
    [aux_sym_command_name_token13] = ACTIONS(1421),
    [aux_sym_command_name_token14] = ACTIONS(1421),
    [aux_sym_command_name_token15] = ACTIONS(1421),
    [aux_sym_command_name_token16] = ACTIONS(1421),
    [aux_sym_command_name_token17] = ACTIONS(1421),
    [aux_sym_command_name_token18] = ACTIONS(1421),
    [aux_sym_class_statement_token1] = ACTIONS(1421),
    [aux_sym_enum_statement_token1] = ACTIONS(1421),
    [anon_sym_PLUS] = ACTIONS(1421),
    [anon_sym_DASH] = ACTIONS(1421),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1421),
    [anon_sym_BANG] = ACTIONS(1421),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1421),
    [anon_sym_PLUS_PLUS] = ACTIONS(1421),
    [anon_sym_DASH_DASH] = ACTIONS(1421),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1421),
    [anon_sym_AT_LPAREN] = ACTIONS(1421),
    [anon_sym_AT_LBRACE] = ACTIONS(1421),
  },
  [STATE(348)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1423),
    [sym_hexadecimal_integer_literal] = ACTIONS(1423),
    [sym_real_literal] = ACTIONS(1423),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1423),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1423),
    [sym_verbatim_string_characters] = ACTIONS(1423),
    [sym_verbatim_here_string_characters] = ACTIONS(1423),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1423),
    [aux_sym_comparison_operator_token37] = ACTIONS(1423),
    [aux_sym_comparison_operator_token50] = ACTIONS(1423),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1423),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1423),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1423),
    [anon_sym_DOLLAR_] = ACTIONS(1423),
    [aux_sym_variable_token1] = ACTIONS(1423),
    [aux_sym_variable_token2] = ACTIONS(1423),
    [sym_braced_variable] = ACTIONS(1423),
    [anon_sym_SEMI] = ACTIONS(1423),
    [anon_sym_LPAREN] = ACTIONS(1423),
    [anon_sym_RPAREN] = ACTIONS(1423),
    [anon_sym_COMMA] = ACTIONS(1423),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_RBRACE] = ACTIONS(1423),
    [aux_sym_if_statement_token1] = ACTIONS(1423),
    [aux_sym_switch_statement_token1] = ACTIONS(1423),
    [aux_sym_foreach_statement_token1] = ACTIONS(1423),
    [aux_sym_for_statement_token1] = ACTIONS(1423),
    [aux_sym_while_statement_token1] = ACTIONS(1423),
    [aux_sym_do_statement_token1] = ACTIONS(1423),
    [aux_sym_function_statement_token1] = ACTIONS(1423),
    [aux_sym_function_statement_token2] = ACTIONS(1423),
    [aux_sym_function_statement_token3] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1423),
    [sym_label] = ACTIONS(1423),
    [aux_sym_trap_statement_token1] = ACTIONS(1423),
    [aux_sym_try_statement_token1] = ACTIONS(1423),
    [aux_sym_data_statement_token1] = ACTIONS(1423),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1423),
    [aux_sym_parallel_statement_token1] = ACTIONS(1423),
    [aux_sym_sequence_statement_token1] = ACTIONS(1423),
    [anon_sym_AMP] = ACTIONS(1423),
    [aux_sym_command_name_token1] = ACTIONS(1423),
    [aux_sym_command_name_token2] = ACTIONS(1423),
    [aux_sym_command_name_token3] = ACTIONS(1423),
    [aux_sym_command_name_token4] = ACTIONS(1423),
    [aux_sym_command_name_token5] = ACTIONS(1423),
    [aux_sym_command_name_token6] = ACTIONS(1423),
    [aux_sym_command_name_token7] = ACTIONS(1423),
    [aux_sym_command_name_token8] = ACTIONS(1423),
    [aux_sym_command_name_token9] = ACTIONS(1423),
    [aux_sym_command_name_token10] = ACTIONS(1423),
    [aux_sym_command_name_token11] = ACTIONS(1423),
    [aux_sym_command_name_token12] = ACTIONS(1423),
    [aux_sym_command_name_token13] = ACTIONS(1423),
    [aux_sym_command_name_token14] = ACTIONS(1423),
    [aux_sym_command_name_token15] = ACTIONS(1423),
    [aux_sym_command_name_token16] = ACTIONS(1423),
    [aux_sym_command_name_token17] = ACTIONS(1423),
    [aux_sym_command_name_token18] = ACTIONS(1423),
    [aux_sym_class_statement_token1] = ACTIONS(1423),
    [aux_sym_enum_statement_token1] = ACTIONS(1423),
    [anon_sym_PLUS] = ACTIONS(1423),
    [anon_sym_DASH] = ACTIONS(1423),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1423),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1423),
    [anon_sym_PLUS_PLUS] = ACTIONS(1423),
    [anon_sym_DASH_DASH] = ACTIONS(1423),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1423),
    [anon_sym_AT_LPAREN] = ACTIONS(1423),
    [anon_sym_AT_LBRACE] = ACTIONS(1423),
  },
  [STATE(349)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1339),
    [sym_hexadecimal_integer_literal] = ACTIONS(1339),
    [sym_real_literal] = ACTIONS(1339),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1339),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1339),
    [sym_verbatim_string_characters] = ACTIONS(1339),
    [sym_verbatim_here_string_characters] = ACTIONS(1339),
    [anon_sym_DOT] = ACTIONS(1339),
    [anon_sym_LBRACK] = ACTIONS(1339),
    [aux_sym_comparison_operator_token37] = ACTIONS(1339),
    [aux_sym_comparison_operator_token50] = ACTIONS(1339),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1339),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1339),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1339),
    [anon_sym_DOLLAR_] = ACTIONS(1339),
    [aux_sym_variable_token1] = ACTIONS(1339),
    [aux_sym_variable_token2] = ACTIONS(1339),
    [sym_braced_variable] = ACTIONS(1339),
    [anon_sym_SEMI] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(1339),
    [anon_sym_RPAREN] = ACTIONS(1339),
    [anon_sym_COMMA] = ACTIONS(1339),
    [anon_sym_LBRACE] = ACTIONS(1339),
    [anon_sym_RBRACE] = ACTIONS(1339),
    [aux_sym_if_statement_token1] = ACTIONS(1339),
    [aux_sym_switch_statement_token1] = ACTIONS(1339),
    [aux_sym_foreach_statement_token1] = ACTIONS(1339),
    [aux_sym_for_statement_token1] = ACTIONS(1339),
    [aux_sym_while_statement_token1] = ACTIONS(1339),
    [aux_sym_do_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token2] = ACTIONS(1339),
    [aux_sym_function_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1339),
    [sym_label] = ACTIONS(1339),
    [aux_sym_trap_statement_token1] = ACTIONS(1339),
    [aux_sym_try_statement_token1] = ACTIONS(1339),
    [aux_sym_data_statement_token1] = ACTIONS(1339),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1339),
    [aux_sym_parallel_statement_token1] = ACTIONS(1339),
    [aux_sym_sequence_statement_token1] = ACTIONS(1339),
    [anon_sym_AMP] = ACTIONS(1339),
    [aux_sym_command_name_token1] = ACTIONS(1339),
    [aux_sym_command_name_token2] = ACTIONS(1339),
    [aux_sym_command_name_token3] = ACTIONS(1339),
    [aux_sym_command_name_token4] = ACTIONS(1339),
    [aux_sym_command_name_token5] = ACTIONS(1339),
    [aux_sym_command_name_token6] = ACTIONS(1339),
    [aux_sym_command_name_token7] = ACTIONS(1339),
    [aux_sym_command_name_token8] = ACTIONS(1339),
    [aux_sym_command_name_token9] = ACTIONS(1339),
    [aux_sym_command_name_token10] = ACTIONS(1339),
    [aux_sym_command_name_token11] = ACTIONS(1339),
    [aux_sym_command_name_token12] = ACTIONS(1339),
    [aux_sym_command_name_token13] = ACTIONS(1339),
    [aux_sym_command_name_token14] = ACTIONS(1339),
    [aux_sym_command_name_token15] = ACTIONS(1339),
    [aux_sym_command_name_token16] = ACTIONS(1339),
    [aux_sym_command_name_token17] = ACTIONS(1339),
    [aux_sym_command_name_token18] = ACTIONS(1339),
    [aux_sym_class_statement_token1] = ACTIONS(1339),
    [aux_sym_enum_statement_token1] = ACTIONS(1339),
    [anon_sym_PLUS] = ACTIONS(1339),
    [anon_sym_DASH] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1339),
    [anon_sym_BANG] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1339),
    [anon_sym_PLUS_PLUS] = ACTIONS(1339),
    [anon_sym_DASH_DASH] = ACTIONS(1339),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LBRACE] = ACTIONS(1339),
  },
  [STATE(350)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1425),
    [sym_hexadecimal_integer_literal] = ACTIONS(1425),
    [sym_real_literal] = ACTIONS(1425),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1425),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1425),
    [sym_verbatim_string_characters] = ACTIONS(1425),
    [sym_verbatim_here_string_characters] = ACTIONS(1425),
    [anon_sym_DOT] = ACTIONS(1425),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [aux_sym_comparison_operator_token37] = ACTIONS(1425),
    [aux_sym_comparison_operator_token50] = ACTIONS(1425),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1425),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1425),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1425),
    [anon_sym_DOLLAR_] = ACTIONS(1425),
    [aux_sym_variable_token1] = ACTIONS(1425),
    [aux_sym_variable_token2] = ACTIONS(1425),
    [sym_braced_variable] = ACTIONS(1425),
    [anon_sym_SEMI] = ACTIONS(1425),
    [anon_sym_LPAREN] = ACTIONS(1425),
    [anon_sym_RPAREN] = ACTIONS(1425),
    [anon_sym_COMMA] = ACTIONS(1425),
    [anon_sym_LBRACE] = ACTIONS(1425),
    [anon_sym_RBRACE] = ACTIONS(1425),
    [aux_sym_if_statement_token1] = ACTIONS(1425),
    [aux_sym_switch_statement_token1] = ACTIONS(1425),
    [aux_sym_foreach_statement_token1] = ACTIONS(1425),
    [aux_sym_for_statement_token1] = ACTIONS(1425),
    [aux_sym_while_statement_token1] = ACTIONS(1425),
    [aux_sym_do_statement_token1] = ACTIONS(1425),
    [aux_sym_function_statement_token1] = ACTIONS(1425),
    [aux_sym_function_statement_token2] = ACTIONS(1425),
    [aux_sym_function_statement_token3] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1425),
    [sym_label] = ACTIONS(1425),
    [aux_sym_trap_statement_token1] = ACTIONS(1425),
    [aux_sym_try_statement_token1] = ACTIONS(1425),
    [aux_sym_data_statement_token1] = ACTIONS(1425),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1425),
    [aux_sym_parallel_statement_token1] = ACTIONS(1425),
    [aux_sym_sequence_statement_token1] = ACTIONS(1425),
    [anon_sym_AMP] = ACTIONS(1425),
    [aux_sym_command_name_token1] = ACTIONS(1425),
    [aux_sym_command_name_token2] = ACTIONS(1425),
    [aux_sym_command_name_token3] = ACTIONS(1425),
    [aux_sym_command_name_token4] = ACTIONS(1425),
    [aux_sym_command_name_token5] = ACTIONS(1425),
    [aux_sym_command_name_token6] = ACTIONS(1425),
    [aux_sym_command_name_token7] = ACTIONS(1425),
    [aux_sym_command_name_token8] = ACTIONS(1425),
    [aux_sym_command_name_token9] = ACTIONS(1425),
    [aux_sym_command_name_token10] = ACTIONS(1425),
    [aux_sym_command_name_token11] = ACTIONS(1425),
    [aux_sym_command_name_token12] = ACTIONS(1425),
    [aux_sym_command_name_token13] = ACTIONS(1425),
    [aux_sym_command_name_token14] = ACTIONS(1425),
    [aux_sym_command_name_token15] = ACTIONS(1425),
    [aux_sym_command_name_token16] = ACTIONS(1425),
    [aux_sym_command_name_token17] = ACTIONS(1425),
    [aux_sym_command_name_token18] = ACTIONS(1425),
    [aux_sym_class_statement_token1] = ACTIONS(1425),
    [aux_sym_enum_statement_token1] = ACTIONS(1425),
    [anon_sym_PLUS] = ACTIONS(1425),
    [anon_sym_DASH] = ACTIONS(1425),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1425),
    [anon_sym_BANG] = ACTIONS(1425),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1425),
    [anon_sym_PLUS_PLUS] = ACTIONS(1425),
    [anon_sym_DASH_DASH] = ACTIONS(1425),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1425),
    [anon_sym_AT_LPAREN] = ACTIONS(1425),
    [anon_sym_AT_LBRACE] = ACTIONS(1425),
  },
  [STATE(351)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1427),
    [sym_hexadecimal_integer_literal] = ACTIONS(1427),
    [sym_real_literal] = ACTIONS(1427),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1427),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1427),
    [sym_verbatim_string_characters] = ACTIONS(1427),
    [sym_verbatim_here_string_characters] = ACTIONS(1427),
    [anon_sym_DOT] = ACTIONS(1427),
    [anon_sym_LBRACK] = ACTIONS(1427),
    [aux_sym_comparison_operator_token37] = ACTIONS(1427),
    [aux_sym_comparison_operator_token50] = ACTIONS(1427),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1427),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1427),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1427),
    [anon_sym_DOLLAR_] = ACTIONS(1427),
    [aux_sym_variable_token1] = ACTIONS(1427),
    [aux_sym_variable_token2] = ACTIONS(1427),
    [sym_braced_variable] = ACTIONS(1427),
    [anon_sym_SEMI] = ACTIONS(1427),
    [anon_sym_LPAREN] = ACTIONS(1427),
    [anon_sym_RPAREN] = ACTIONS(1427),
    [anon_sym_COMMA] = ACTIONS(1427),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_RBRACE] = ACTIONS(1427),
    [aux_sym_if_statement_token1] = ACTIONS(1427),
    [aux_sym_switch_statement_token1] = ACTIONS(1427),
    [aux_sym_foreach_statement_token1] = ACTIONS(1427),
    [aux_sym_for_statement_token1] = ACTIONS(1427),
    [aux_sym_while_statement_token1] = ACTIONS(1427),
    [aux_sym_do_statement_token1] = ACTIONS(1427),
    [aux_sym_function_statement_token1] = ACTIONS(1427),
    [aux_sym_function_statement_token2] = ACTIONS(1427),
    [aux_sym_function_statement_token3] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1427),
    [sym_label] = ACTIONS(1427),
    [aux_sym_trap_statement_token1] = ACTIONS(1427),
    [aux_sym_try_statement_token1] = ACTIONS(1427),
    [aux_sym_data_statement_token1] = ACTIONS(1427),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1427),
    [aux_sym_parallel_statement_token1] = ACTIONS(1427),
    [aux_sym_sequence_statement_token1] = ACTIONS(1427),
    [anon_sym_AMP] = ACTIONS(1427),
    [aux_sym_command_name_token1] = ACTIONS(1427),
    [aux_sym_command_name_token2] = ACTIONS(1427),
    [aux_sym_command_name_token3] = ACTIONS(1427),
    [aux_sym_command_name_token4] = ACTIONS(1427),
    [aux_sym_command_name_token5] = ACTIONS(1427),
    [aux_sym_command_name_token6] = ACTIONS(1427),
    [aux_sym_command_name_token7] = ACTIONS(1427),
    [aux_sym_command_name_token8] = ACTIONS(1427),
    [aux_sym_command_name_token9] = ACTIONS(1427),
    [aux_sym_command_name_token10] = ACTIONS(1427),
    [aux_sym_command_name_token11] = ACTIONS(1427),
    [aux_sym_command_name_token12] = ACTIONS(1427),
    [aux_sym_command_name_token13] = ACTIONS(1427),
    [aux_sym_command_name_token14] = ACTIONS(1427),
    [aux_sym_command_name_token15] = ACTIONS(1427),
    [aux_sym_command_name_token16] = ACTIONS(1427),
    [aux_sym_command_name_token17] = ACTIONS(1427),
    [aux_sym_command_name_token18] = ACTIONS(1427),
    [aux_sym_class_statement_token1] = ACTIONS(1427),
    [aux_sym_enum_statement_token1] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1427),
    [anon_sym_DASH] = ACTIONS(1427),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1427),
    [anon_sym_BANG] = ACTIONS(1427),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1427),
    [anon_sym_PLUS_PLUS] = ACTIONS(1427),
    [anon_sym_DASH_DASH] = ACTIONS(1427),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1427),
    [anon_sym_AT_LPAREN] = ACTIONS(1427),
    [anon_sym_AT_LBRACE] = ACTIONS(1427),
  },
  [STATE(352)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1429),
    [sym_hexadecimal_integer_literal] = ACTIONS(1429),
    [sym_real_literal] = ACTIONS(1429),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1429),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1429),
    [sym_verbatim_string_characters] = ACTIONS(1429),
    [sym_verbatim_here_string_characters] = ACTIONS(1429),
    [anon_sym_DOT] = ACTIONS(1429),
    [anon_sym_LBRACK] = ACTIONS(1429),
    [aux_sym_comparison_operator_token37] = ACTIONS(1429),
    [aux_sym_comparison_operator_token50] = ACTIONS(1429),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1429),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1429),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1429),
    [anon_sym_DOLLAR_] = ACTIONS(1429),
    [aux_sym_variable_token1] = ACTIONS(1429),
    [aux_sym_variable_token2] = ACTIONS(1429),
    [sym_braced_variable] = ACTIONS(1429),
    [anon_sym_SEMI] = ACTIONS(1429),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_RPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1429),
    [anon_sym_LBRACE] = ACTIONS(1429),
    [anon_sym_RBRACE] = ACTIONS(1429),
    [aux_sym_if_statement_token1] = ACTIONS(1429),
    [aux_sym_switch_statement_token1] = ACTIONS(1429),
    [aux_sym_foreach_statement_token1] = ACTIONS(1429),
    [aux_sym_for_statement_token1] = ACTIONS(1429),
    [aux_sym_while_statement_token1] = ACTIONS(1429),
    [aux_sym_do_statement_token1] = ACTIONS(1429),
    [aux_sym_function_statement_token1] = ACTIONS(1429),
    [aux_sym_function_statement_token2] = ACTIONS(1429),
    [aux_sym_function_statement_token3] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1429),
    [sym_label] = ACTIONS(1429),
    [aux_sym_trap_statement_token1] = ACTIONS(1429),
    [aux_sym_try_statement_token1] = ACTIONS(1429),
    [aux_sym_data_statement_token1] = ACTIONS(1429),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1429),
    [aux_sym_parallel_statement_token1] = ACTIONS(1429),
    [aux_sym_sequence_statement_token1] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(1429),
    [aux_sym_command_name_token1] = ACTIONS(1429),
    [aux_sym_command_name_token2] = ACTIONS(1429),
    [aux_sym_command_name_token3] = ACTIONS(1429),
    [aux_sym_command_name_token4] = ACTIONS(1429),
    [aux_sym_command_name_token5] = ACTIONS(1429),
    [aux_sym_command_name_token6] = ACTIONS(1429),
    [aux_sym_command_name_token7] = ACTIONS(1429),
    [aux_sym_command_name_token8] = ACTIONS(1429),
    [aux_sym_command_name_token9] = ACTIONS(1429),
    [aux_sym_command_name_token10] = ACTIONS(1429),
    [aux_sym_command_name_token11] = ACTIONS(1429),
    [aux_sym_command_name_token12] = ACTIONS(1429),
    [aux_sym_command_name_token13] = ACTIONS(1429),
    [aux_sym_command_name_token14] = ACTIONS(1429),
    [aux_sym_command_name_token15] = ACTIONS(1429),
    [aux_sym_command_name_token16] = ACTIONS(1429),
    [aux_sym_command_name_token17] = ACTIONS(1429),
    [aux_sym_command_name_token18] = ACTIONS(1429),
    [aux_sym_class_statement_token1] = ACTIONS(1429),
    [aux_sym_enum_statement_token1] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1429),
    [anon_sym_BANG] = ACTIONS(1429),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1429),
    [anon_sym_PLUS_PLUS] = ACTIONS(1429),
    [anon_sym_DASH_DASH] = ACTIONS(1429),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1429),
    [anon_sym_AT_LPAREN] = ACTIONS(1429),
    [anon_sym_AT_LBRACE] = ACTIONS(1429),
  },
  [STATE(353)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [anon_sym_RBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(354)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1431),
    [sym_hexadecimal_integer_literal] = ACTIONS(1431),
    [sym_real_literal] = ACTIONS(1431),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1431),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1431),
    [sym_verbatim_string_characters] = ACTIONS(1431),
    [sym_verbatim_here_string_characters] = ACTIONS(1431),
    [anon_sym_DOT] = ACTIONS(1431),
    [anon_sym_LBRACK] = ACTIONS(1431),
    [aux_sym_comparison_operator_token37] = ACTIONS(1431),
    [aux_sym_comparison_operator_token50] = ACTIONS(1431),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1431),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1431),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1431),
    [anon_sym_DOLLAR_] = ACTIONS(1431),
    [aux_sym_variable_token1] = ACTIONS(1431),
    [aux_sym_variable_token2] = ACTIONS(1431),
    [sym_braced_variable] = ACTIONS(1431),
    [anon_sym_SEMI] = ACTIONS(1431),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [anon_sym_RPAREN] = ACTIONS(1431),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_LBRACE] = ACTIONS(1431),
    [anon_sym_RBRACE] = ACTIONS(1431),
    [aux_sym_if_statement_token1] = ACTIONS(1431),
    [aux_sym_switch_statement_token1] = ACTIONS(1431),
    [aux_sym_foreach_statement_token1] = ACTIONS(1431),
    [aux_sym_for_statement_token1] = ACTIONS(1431),
    [aux_sym_while_statement_token1] = ACTIONS(1431),
    [aux_sym_do_statement_token1] = ACTIONS(1431),
    [aux_sym_function_statement_token1] = ACTIONS(1431),
    [aux_sym_function_statement_token2] = ACTIONS(1431),
    [aux_sym_function_statement_token3] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1431),
    [sym_label] = ACTIONS(1431),
    [aux_sym_trap_statement_token1] = ACTIONS(1431),
    [aux_sym_try_statement_token1] = ACTIONS(1431),
    [aux_sym_data_statement_token1] = ACTIONS(1431),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1431),
    [aux_sym_parallel_statement_token1] = ACTIONS(1431),
    [aux_sym_sequence_statement_token1] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(1431),
    [aux_sym_command_name_token1] = ACTIONS(1431),
    [aux_sym_command_name_token2] = ACTIONS(1431),
    [aux_sym_command_name_token3] = ACTIONS(1431),
    [aux_sym_command_name_token4] = ACTIONS(1431),
    [aux_sym_command_name_token5] = ACTIONS(1431),
    [aux_sym_command_name_token6] = ACTIONS(1431),
    [aux_sym_command_name_token7] = ACTIONS(1431),
    [aux_sym_command_name_token8] = ACTIONS(1431),
    [aux_sym_command_name_token9] = ACTIONS(1431),
    [aux_sym_command_name_token10] = ACTIONS(1431),
    [aux_sym_command_name_token11] = ACTIONS(1431),
    [aux_sym_command_name_token12] = ACTIONS(1431),
    [aux_sym_command_name_token13] = ACTIONS(1431),
    [aux_sym_command_name_token14] = ACTIONS(1431),
    [aux_sym_command_name_token15] = ACTIONS(1431),
    [aux_sym_command_name_token16] = ACTIONS(1431),
    [aux_sym_command_name_token17] = ACTIONS(1431),
    [aux_sym_command_name_token18] = ACTIONS(1431),
    [aux_sym_class_statement_token1] = ACTIONS(1431),
    [aux_sym_enum_statement_token1] = ACTIONS(1431),
    [anon_sym_PLUS] = ACTIONS(1431),
    [anon_sym_DASH] = ACTIONS(1431),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1431),
    [anon_sym_BANG] = ACTIONS(1431),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1431),
    [anon_sym_PLUS_PLUS] = ACTIONS(1431),
    [anon_sym_DASH_DASH] = ACTIONS(1431),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1431),
    [anon_sym_AT_LPAREN] = ACTIONS(1431),
    [anon_sym_AT_LBRACE] = ACTIONS(1431),
  },
  [STATE(355)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1433),
    [sym_hexadecimal_integer_literal] = ACTIONS(1433),
    [sym_real_literal] = ACTIONS(1433),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1433),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1433),
    [sym_verbatim_string_characters] = ACTIONS(1433),
    [sym_verbatim_here_string_characters] = ACTIONS(1433),
    [anon_sym_DOT] = ACTIONS(1433),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [aux_sym_comparison_operator_token37] = ACTIONS(1433),
    [aux_sym_comparison_operator_token50] = ACTIONS(1433),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1433),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1433),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1433),
    [anon_sym_DOLLAR_] = ACTIONS(1433),
    [aux_sym_variable_token1] = ACTIONS(1433),
    [aux_sym_variable_token2] = ACTIONS(1433),
    [sym_braced_variable] = ACTIONS(1433),
    [anon_sym_SEMI] = ACTIONS(1433),
    [anon_sym_LPAREN] = ACTIONS(1433),
    [anon_sym_RPAREN] = ACTIONS(1433),
    [anon_sym_COMMA] = ACTIONS(1433),
    [anon_sym_LBRACE] = ACTIONS(1433),
    [anon_sym_RBRACE] = ACTIONS(1433),
    [aux_sym_if_statement_token1] = ACTIONS(1433),
    [aux_sym_switch_statement_token1] = ACTIONS(1433),
    [aux_sym_foreach_statement_token1] = ACTIONS(1433),
    [aux_sym_for_statement_token1] = ACTIONS(1433),
    [aux_sym_while_statement_token1] = ACTIONS(1433),
    [aux_sym_do_statement_token1] = ACTIONS(1433),
    [aux_sym_function_statement_token1] = ACTIONS(1433),
    [aux_sym_function_statement_token2] = ACTIONS(1433),
    [aux_sym_function_statement_token3] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1433),
    [sym_label] = ACTIONS(1433),
    [aux_sym_trap_statement_token1] = ACTIONS(1433),
    [aux_sym_try_statement_token1] = ACTIONS(1433),
    [aux_sym_data_statement_token1] = ACTIONS(1433),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1433),
    [aux_sym_parallel_statement_token1] = ACTIONS(1433),
    [aux_sym_sequence_statement_token1] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1433),
    [aux_sym_command_name_token1] = ACTIONS(1433),
    [aux_sym_command_name_token2] = ACTIONS(1433),
    [aux_sym_command_name_token3] = ACTIONS(1433),
    [aux_sym_command_name_token4] = ACTIONS(1433),
    [aux_sym_command_name_token5] = ACTIONS(1433),
    [aux_sym_command_name_token6] = ACTIONS(1433),
    [aux_sym_command_name_token7] = ACTIONS(1433),
    [aux_sym_command_name_token8] = ACTIONS(1433),
    [aux_sym_command_name_token9] = ACTIONS(1433),
    [aux_sym_command_name_token10] = ACTIONS(1433),
    [aux_sym_command_name_token11] = ACTIONS(1433),
    [aux_sym_command_name_token12] = ACTIONS(1433),
    [aux_sym_command_name_token13] = ACTIONS(1433),
    [aux_sym_command_name_token14] = ACTIONS(1433),
    [aux_sym_command_name_token15] = ACTIONS(1433),
    [aux_sym_command_name_token16] = ACTIONS(1433),
    [aux_sym_command_name_token17] = ACTIONS(1433),
    [aux_sym_command_name_token18] = ACTIONS(1433),
    [aux_sym_class_statement_token1] = ACTIONS(1433),
    [aux_sym_enum_statement_token1] = ACTIONS(1433),
    [anon_sym_PLUS] = ACTIONS(1433),
    [anon_sym_DASH] = ACTIONS(1433),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1433),
    [anon_sym_BANG] = ACTIONS(1433),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1433),
    [anon_sym_PLUS_PLUS] = ACTIONS(1433),
    [anon_sym_DASH_DASH] = ACTIONS(1433),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1433),
    [anon_sym_AT_LPAREN] = ACTIONS(1433),
    [anon_sym_AT_LBRACE] = ACTIONS(1433),
  },
  [STATE(356)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1435),
    [sym_hexadecimal_integer_literal] = ACTIONS(1435),
    [sym_real_literal] = ACTIONS(1435),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1435),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1435),
    [sym_verbatim_string_characters] = ACTIONS(1435),
    [sym_verbatim_here_string_characters] = ACTIONS(1435),
    [anon_sym_DOT] = ACTIONS(1435),
    [anon_sym_LBRACK] = ACTIONS(1435),
    [aux_sym_comparison_operator_token37] = ACTIONS(1435),
    [aux_sym_comparison_operator_token50] = ACTIONS(1435),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1435),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1435),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1435),
    [anon_sym_DOLLAR_] = ACTIONS(1435),
    [aux_sym_variable_token1] = ACTIONS(1435),
    [aux_sym_variable_token2] = ACTIONS(1435),
    [sym_braced_variable] = ACTIONS(1435),
    [anon_sym_SEMI] = ACTIONS(1435),
    [anon_sym_LPAREN] = ACTIONS(1435),
    [anon_sym_RPAREN] = ACTIONS(1435),
    [anon_sym_COMMA] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1435),
    [anon_sym_RBRACE] = ACTIONS(1435),
    [aux_sym_if_statement_token1] = ACTIONS(1435),
    [aux_sym_switch_statement_token1] = ACTIONS(1435),
    [aux_sym_foreach_statement_token1] = ACTIONS(1435),
    [aux_sym_for_statement_token1] = ACTIONS(1435),
    [aux_sym_while_statement_token1] = ACTIONS(1435),
    [aux_sym_do_statement_token1] = ACTIONS(1435),
    [aux_sym_function_statement_token1] = ACTIONS(1435),
    [aux_sym_function_statement_token2] = ACTIONS(1435),
    [aux_sym_function_statement_token3] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1435),
    [sym_label] = ACTIONS(1435),
    [aux_sym_trap_statement_token1] = ACTIONS(1435),
    [aux_sym_try_statement_token1] = ACTIONS(1435),
    [aux_sym_data_statement_token1] = ACTIONS(1435),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1435),
    [aux_sym_parallel_statement_token1] = ACTIONS(1435),
    [aux_sym_sequence_statement_token1] = ACTIONS(1435),
    [anon_sym_AMP] = ACTIONS(1435),
    [aux_sym_command_name_token1] = ACTIONS(1435),
    [aux_sym_command_name_token2] = ACTIONS(1435),
    [aux_sym_command_name_token3] = ACTIONS(1435),
    [aux_sym_command_name_token4] = ACTIONS(1435),
    [aux_sym_command_name_token5] = ACTIONS(1435),
    [aux_sym_command_name_token6] = ACTIONS(1435),
    [aux_sym_command_name_token7] = ACTIONS(1435),
    [aux_sym_command_name_token8] = ACTIONS(1435),
    [aux_sym_command_name_token9] = ACTIONS(1435),
    [aux_sym_command_name_token10] = ACTIONS(1435),
    [aux_sym_command_name_token11] = ACTIONS(1435),
    [aux_sym_command_name_token12] = ACTIONS(1435),
    [aux_sym_command_name_token13] = ACTIONS(1435),
    [aux_sym_command_name_token14] = ACTIONS(1435),
    [aux_sym_command_name_token15] = ACTIONS(1435),
    [aux_sym_command_name_token16] = ACTIONS(1435),
    [aux_sym_command_name_token17] = ACTIONS(1435),
    [aux_sym_command_name_token18] = ACTIONS(1435),
    [aux_sym_class_statement_token1] = ACTIONS(1435),
    [aux_sym_enum_statement_token1] = ACTIONS(1435),
    [anon_sym_PLUS] = ACTIONS(1435),
    [anon_sym_DASH] = ACTIONS(1435),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1435),
    [anon_sym_BANG] = ACTIONS(1435),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1435),
    [anon_sym_PLUS_PLUS] = ACTIONS(1435),
    [anon_sym_DASH_DASH] = ACTIONS(1435),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1435),
    [anon_sym_AT_LPAREN] = ACTIONS(1435),
    [anon_sym_AT_LBRACE] = ACTIONS(1435),
  },
  [STATE(357)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_RPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [anon_sym_RBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(358)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1437),
    [sym_hexadecimal_integer_literal] = ACTIONS(1437),
    [sym_real_literal] = ACTIONS(1437),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1437),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1437),
    [sym_verbatim_string_characters] = ACTIONS(1437),
    [sym_verbatim_here_string_characters] = ACTIONS(1437),
    [anon_sym_DOT] = ACTIONS(1437),
    [anon_sym_LBRACK] = ACTIONS(1437),
    [aux_sym_comparison_operator_token37] = ACTIONS(1437),
    [aux_sym_comparison_operator_token50] = ACTIONS(1437),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1437),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1437),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1437),
    [anon_sym_DOLLAR_] = ACTIONS(1437),
    [aux_sym_variable_token1] = ACTIONS(1437),
    [aux_sym_variable_token2] = ACTIONS(1437),
    [sym_braced_variable] = ACTIONS(1437),
    [anon_sym_SEMI] = ACTIONS(1437),
    [anon_sym_LPAREN] = ACTIONS(1437),
    [anon_sym_RPAREN] = ACTIONS(1437),
    [anon_sym_COMMA] = ACTIONS(1437),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(1437),
    [aux_sym_if_statement_token1] = ACTIONS(1437),
    [aux_sym_switch_statement_token1] = ACTIONS(1437),
    [aux_sym_foreach_statement_token1] = ACTIONS(1437),
    [aux_sym_for_statement_token1] = ACTIONS(1437),
    [aux_sym_while_statement_token1] = ACTIONS(1437),
    [aux_sym_do_statement_token1] = ACTIONS(1437),
    [aux_sym_function_statement_token1] = ACTIONS(1437),
    [aux_sym_function_statement_token2] = ACTIONS(1437),
    [aux_sym_function_statement_token3] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1437),
    [sym_label] = ACTIONS(1437),
    [aux_sym_trap_statement_token1] = ACTIONS(1437),
    [aux_sym_try_statement_token1] = ACTIONS(1437),
    [aux_sym_data_statement_token1] = ACTIONS(1437),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1437),
    [aux_sym_parallel_statement_token1] = ACTIONS(1437),
    [aux_sym_sequence_statement_token1] = ACTIONS(1437),
    [anon_sym_AMP] = ACTIONS(1437),
    [aux_sym_command_name_token1] = ACTIONS(1437),
    [aux_sym_command_name_token2] = ACTIONS(1437),
    [aux_sym_command_name_token3] = ACTIONS(1437),
    [aux_sym_command_name_token4] = ACTIONS(1437),
    [aux_sym_command_name_token5] = ACTIONS(1437),
    [aux_sym_command_name_token6] = ACTIONS(1437),
    [aux_sym_command_name_token7] = ACTIONS(1437),
    [aux_sym_command_name_token8] = ACTIONS(1437),
    [aux_sym_command_name_token9] = ACTIONS(1437),
    [aux_sym_command_name_token10] = ACTIONS(1437),
    [aux_sym_command_name_token11] = ACTIONS(1437),
    [aux_sym_command_name_token12] = ACTIONS(1437),
    [aux_sym_command_name_token13] = ACTIONS(1437),
    [aux_sym_command_name_token14] = ACTIONS(1437),
    [aux_sym_command_name_token15] = ACTIONS(1437),
    [aux_sym_command_name_token16] = ACTIONS(1437),
    [aux_sym_command_name_token17] = ACTIONS(1437),
    [aux_sym_command_name_token18] = ACTIONS(1437),
    [aux_sym_class_statement_token1] = ACTIONS(1437),
    [aux_sym_enum_statement_token1] = ACTIONS(1437),
    [anon_sym_PLUS] = ACTIONS(1437),
    [anon_sym_DASH] = ACTIONS(1437),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1437),
    [anon_sym_BANG] = ACTIONS(1437),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1437),
    [anon_sym_PLUS_PLUS] = ACTIONS(1437),
    [anon_sym_DASH_DASH] = ACTIONS(1437),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1437),
    [anon_sym_AT_LPAREN] = ACTIONS(1437),
    [anon_sym_AT_LBRACE] = ACTIONS(1437),
  },
  [STATE(359)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1439),
    [sym_hexadecimal_integer_literal] = ACTIONS(1439),
    [sym_real_literal] = ACTIONS(1439),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1439),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1439),
    [sym_verbatim_string_characters] = ACTIONS(1439),
    [sym_verbatim_here_string_characters] = ACTIONS(1439),
    [anon_sym_DOT] = ACTIONS(1439),
    [anon_sym_LBRACK] = ACTIONS(1439),
    [aux_sym_comparison_operator_token37] = ACTIONS(1439),
    [aux_sym_comparison_operator_token50] = ACTIONS(1439),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1439),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1439),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1439),
    [anon_sym_DOLLAR_] = ACTIONS(1439),
    [aux_sym_variable_token1] = ACTIONS(1439),
    [aux_sym_variable_token2] = ACTIONS(1439),
    [sym_braced_variable] = ACTIONS(1439),
    [anon_sym_SEMI] = ACTIONS(1439),
    [anon_sym_LPAREN] = ACTIONS(1439),
    [anon_sym_RPAREN] = ACTIONS(1439),
    [anon_sym_COMMA] = ACTIONS(1439),
    [anon_sym_LBRACE] = ACTIONS(1439),
    [anon_sym_RBRACE] = ACTIONS(1439),
    [aux_sym_if_statement_token1] = ACTIONS(1439),
    [aux_sym_switch_statement_token1] = ACTIONS(1439),
    [aux_sym_foreach_statement_token1] = ACTIONS(1439),
    [aux_sym_for_statement_token1] = ACTIONS(1439),
    [aux_sym_while_statement_token1] = ACTIONS(1439),
    [aux_sym_do_statement_token1] = ACTIONS(1439),
    [aux_sym_function_statement_token1] = ACTIONS(1439),
    [aux_sym_function_statement_token2] = ACTIONS(1439),
    [aux_sym_function_statement_token3] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1439),
    [sym_label] = ACTIONS(1439),
    [aux_sym_trap_statement_token1] = ACTIONS(1439),
    [aux_sym_try_statement_token1] = ACTIONS(1439),
    [aux_sym_data_statement_token1] = ACTIONS(1439),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1439),
    [aux_sym_parallel_statement_token1] = ACTIONS(1439),
    [aux_sym_sequence_statement_token1] = ACTIONS(1439),
    [anon_sym_AMP] = ACTIONS(1439),
    [aux_sym_command_name_token1] = ACTIONS(1439),
    [aux_sym_command_name_token2] = ACTIONS(1439),
    [aux_sym_command_name_token3] = ACTIONS(1439),
    [aux_sym_command_name_token4] = ACTIONS(1439),
    [aux_sym_command_name_token5] = ACTIONS(1439),
    [aux_sym_command_name_token6] = ACTIONS(1439),
    [aux_sym_command_name_token7] = ACTIONS(1439),
    [aux_sym_command_name_token8] = ACTIONS(1439),
    [aux_sym_command_name_token9] = ACTIONS(1439),
    [aux_sym_command_name_token10] = ACTIONS(1439),
    [aux_sym_command_name_token11] = ACTIONS(1439),
    [aux_sym_command_name_token12] = ACTIONS(1439),
    [aux_sym_command_name_token13] = ACTIONS(1439),
    [aux_sym_command_name_token14] = ACTIONS(1439),
    [aux_sym_command_name_token15] = ACTIONS(1439),
    [aux_sym_command_name_token16] = ACTIONS(1439),
    [aux_sym_command_name_token17] = ACTIONS(1439),
    [aux_sym_command_name_token18] = ACTIONS(1439),
    [aux_sym_class_statement_token1] = ACTIONS(1439),
    [aux_sym_enum_statement_token1] = ACTIONS(1439),
    [anon_sym_PLUS] = ACTIONS(1439),
    [anon_sym_DASH] = ACTIONS(1439),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1439),
    [anon_sym_BANG] = ACTIONS(1439),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1439),
    [anon_sym_PLUS_PLUS] = ACTIONS(1439),
    [anon_sym_DASH_DASH] = ACTIONS(1439),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1439),
    [anon_sym_AT_LPAREN] = ACTIONS(1439),
    [anon_sym_AT_LBRACE] = ACTIONS(1439),
  },
  [STATE(360)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1441),
    [sym_hexadecimal_integer_literal] = ACTIONS(1441),
    [sym_real_literal] = ACTIONS(1441),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1441),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1441),
    [sym_verbatim_string_characters] = ACTIONS(1441),
    [sym_verbatim_here_string_characters] = ACTIONS(1441),
    [anon_sym_DOT] = ACTIONS(1441),
    [anon_sym_LBRACK] = ACTIONS(1441),
    [aux_sym_comparison_operator_token37] = ACTIONS(1441),
    [aux_sym_comparison_operator_token50] = ACTIONS(1441),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1441),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1441),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1441),
    [anon_sym_DOLLAR_] = ACTIONS(1441),
    [aux_sym_variable_token1] = ACTIONS(1441),
    [aux_sym_variable_token2] = ACTIONS(1441),
    [sym_braced_variable] = ACTIONS(1441),
    [anon_sym_SEMI] = ACTIONS(1441),
    [anon_sym_LPAREN] = ACTIONS(1441),
    [anon_sym_RPAREN] = ACTIONS(1441),
    [anon_sym_COMMA] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1441),
    [anon_sym_RBRACE] = ACTIONS(1441),
    [aux_sym_if_statement_token1] = ACTIONS(1441),
    [aux_sym_switch_statement_token1] = ACTIONS(1441),
    [aux_sym_foreach_statement_token1] = ACTIONS(1441),
    [aux_sym_for_statement_token1] = ACTIONS(1441),
    [aux_sym_while_statement_token1] = ACTIONS(1441),
    [aux_sym_do_statement_token1] = ACTIONS(1441),
    [aux_sym_function_statement_token1] = ACTIONS(1441),
    [aux_sym_function_statement_token2] = ACTIONS(1441),
    [aux_sym_function_statement_token3] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1441),
    [sym_label] = ACTIONS(1441),
    [aux_sym_trap_statement_token1] = ACTIONS(1441),
    [aux_sym_try_statement_token1] = ACTIONS(1441),
    [aux_sym_data_statement_token1] = ACTIONS(1441),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1441),
    [aux_sym_parallel_statement_token1] = ACTIONS(1441),
    [aux_sym_sequence_statement_token1] = ACTIONS(1441),
    [anon_sym_AMP] = ACTIONS(1441),
    [aux_sym_command_name_token1] = ACTIONS(1441),
    [aux_sym_command_name_token2] = ACTIONS(1441),
    [aux_sym_command_name_token3] = ACTIONS(1441),
    [aux_sym_command_name_token4] = ACTIONS(1441),
    [aux_sym_command_name_token5] = ACTIONS(1441),
    [aux_sym_command_name_token6] = ACTIONS(1441),
    [aux_sym_command_name_token7] = ACTIONS(1441),
    [aux_sym_command_name_token8] = ACTIONS(1441),
    [aux_sym_command_name_token9] = ACTIONS(1441),
    [aux_sym_command_name_token10] = ACTIONS(1441),
    [aux_sym_command_name_token11] = ACTIONS(1441),
    [aux_sym_command_name_token12] = ACTIONS(1441),
    [aux_sym_command_name_token13] = ACTIONS(1441),
    [aux_sym_command_name_token14] = ACTIONS(1441),
    [aux_sym_command_name_token15] = ACTIONS(1441),
    [aux_sym_command_name_token16] = ACTIONS(1441),
    [aux_sym_command_name_token17] = ACTIONS(1441),
    [aux_sym_command_name_token18] = ACTIONS(1441),
    [aux_sym_class_statement_token1] = ACTIONS(1441),
    [aux_sym_enum_statement_token1] = ACTIONS(1441),
    [anon_sym_PLUS] = ACTIONS(1441),
    [anon_sym_DASH] = ACTIONS(1441),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1441),
    [anon_sym_BANG] = ACTIONS(1441),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1441),
    [anon_sym_PLUS_PLUS] = ACTIONS(1441),
    [anon_sym_DASH_DASH] = ACTIONS(1441),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1441),
    [anon_sym_AT_LPAREN] = ACTIONS(1441),
    [anon_sym_AT_LBRACE] = ACTIONS(1441),
  },
  [STATE(361)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1443),
    [sym_hexadecimal_integer_literal] = ACTIONS(1443),
    [sym_real_literal] = ACTIONS(1443),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1443),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1443),
    [sym_verbatim_string_characters] = ACTIONS(1443),
    [sym_verbatim_here_string_characters] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1443),
    [anon_sym_LBRACK] = ACTIONS(1443),
    [aux_sym_comparison_operator_token37] = ACTIONS(1443),
    [aux_sym_comparison_operator_token50] = ACTIONS(1443),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1443),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1443),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1443),
    [anon_sym_DOLLAR_] = ACTIONS(1443),
    [aux_sym_variable_token1] = ACTIONS(1443),
    [aux_sym_variable_token2] = ACTIONS(1443),
    [sym_braced_variable] = ACTIONS(1443),
    [anon_sym_SEMI] = ACTIONS(1443),
    [anon_sym_LPAREN] = ACTIONS(1443),
    [anon_sym_RPAREN] = ACTIONS(1443),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1443),
    [anon_sym_RBRACE] = ACTIONS(1443),
    [aux_sym_if_statement_token1] = ACTIONS(1443),
    [aux_sym_switch_statement_token1] = ACTIONS(1443),
    [aux_sym_foreach_statement_token1] = ACTIONS(1443),
    [aux_sym_for_statement_token1] = ACTIONS(1443),
    [aux_sym_while_statement_token1] = ACTIONS(1443),
    [aux_sym_do_statement_token1] = ACTIONS(1443),
    [aux_sym_function_statement_token1] = ACTIONS(1443),
    [aux_sym_function_statement_token2] = ACTIONS(1443),
    [aux_sym_function_statement_token3] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1443),
    [sym_label] = ACTIONS(1443),
    [aux_sym_trap_statement_token1] = ACTIONS(1443),
    [aux_sym_try_statement_token1] = ACTIONS(1443),
    [aux_sym_data_statement_token1] = ACTIONS(1443),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1443),
    [aux_sym_parallel_statement_token1] = ACTIONS(1443),
    [aux_sym_sequence_statement_token1] = ACTIONS(1443),
    [anon_sym_AMP] = ACTIONS(1443),
    [aux_sym_command_name_token1] = ACTIONS(1443),
    [aux_sym_command_name_token2] = ACTIONS(1443),
    [aux_sym_command_name_token3] = ACTIONS(1443),
    [aux_sym_command_name_token4] = ACTIONS(1443),
    [aux_sym_command_name_token5] = ACTIONS(1443),
    [aux_sym_command_name_token6] = ACTIONS(1443),
    [aux_sym_command_name_token7] = ACTIONS(1443),
    [aux_sym_command_name_token8] = ACTIONS(1443),
    [aux_sym_command_name_token9] = ACTIONS(1443),
    [aux_sym_command_name_token10] = ACTIONS(1443),
    [aux_sym_command_name_token11] = ACTIONS(1443),
    [aux_sym_command_name_token12] = ACTIONS(1443),
    [aux_sym_command_name_token13] = ACTIONS(1443),
    [aux_sym_command_name_token14] = ACTIONS(1443),
    [aux_sym_command_name_token15] = ACTIONS(1443),
    [aux_sym_command_name_token16] = ACTIONS(1443),
    [aux_sym_command_name_token17] = ACTIONS(1443),
    [aux_sym_command_name_token18] = ACTIONS(1443),
    [aux_sym_class_statement_token1] = ACTIONS(1443),
    [aux_sym_enum_statement_token1] = ACTIONS(1443),
    [anon_sym_PLUS] = ACTIONS(1443),
    [anon_sym_DASH] = ACTIONS(1443),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1443),
    [anon_sym_BANG] = ACTIONS(1443),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1443),
    [anon_sym_PLUS_PLUS] = ACTIONS(1443),
    [anon_sym_DASH_DASH] = ACTIONS(1443),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1443),
    [anon_sym_AT_LPAREN] = ACTIONS(1443),
    [anon_sym_AT_LBRACE] = ACTIONS(1443),
  },
  [STATE(362)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1445),
    [sym_hexadecimal_integer_literal] = ACTIONS(1445),
    [sym_real_literal] = ACTIONS(1445),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1445),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1445),
    [sym_verbatim_string_characters] = ACTIONS(1445),
    [sym_verbatim_here_string_characters] = ACTIONS(1445),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACK] = ACTIONS(1445),
    [aux_sym_comparison_operator_token37] = ACTIONS(1445),
    [aux_sym_comparison_operator_token50] = ACTIONS(1445),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1445),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1445),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1445),
    [anon_sym_DOLLAR_] = ACTIONS(1445),
    [aux_sym_variable_token1] = ACTIONS(1445),
    [aux_sym_variable_token2] = ACTIONS(1445),
    [sym_braced_variable] = ACTIONS(1445),
    [anon_sym_SEMI] = ACTIONS(1445),
    [anon_sym_LPAREN] = ACTIONS(1445),
    [anon_sym_RPAREN] = ACTIONS(1445),
    [anon_sym_COMMA] = ACTIONS(1445),
    [anon_sym_LBRACE] = ACTIONS(1445),
    [anon_sym_RBRACE] = ACTIONS(1445),
    [aux_sym_if_statement_token1] = ACTIONS(1445),
    [aux_sym_switch_statement_token1] = ACTIONS(1445),
    [aux_sym_foreach_statement_token1] = ACTIONS(1445),
    [aux_sym_for_statement_token1] = ACTIONS(1445),
    [aux_sym_while_statement_token1] = ACTIONS(1445),
    [aux_sym_do_statement_token1] = ACTIONS(1445),
    [aux_sym_function_statement_token1] = ACTIONS(1445),
    [aux_sym_function_statement_token2] = ACTIONS(1445),
    [aux_sym_function_statement_token3] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1445),
    [sym_label] = ACTIONS(1445),
    [aux_sym_trap_statement_token1] = ACTIONS(1445),
    [aux_sym_try_statement_token1] = ACTIONS(1445),
    [aux_sym_data_statement_token1] = ACTIONS(1445),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1445),
    [aux_sym_parallel_statement_token1] = ACTIONS(1445),
    [aux_sym_sequence_statement_token1] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(1445),
    [aux_sym_command_name_token1] = ACTIONS(1445),
    [aux_sym_command_name_token2] = ACTIONS(1445),
    [aux_sym_command_name_token3] = ACTIONS(1445),
    [aux_sym_command_name_token4] = ACTIONS(1445),
    [aux_sym_command_name_token5] = ACTIONS(1445),
    [aux_sym_command_name_token6] = ACTIONS(1445),
    [aux_sym_command_name_token7] = ACTIONS(1445),
    [aux_sym_command_name_token8] = ACTIONS(1445),
    [aux_sym_command_name_token9] = ACTIONS(1445),
    [aux_sym_command_name_token10] = ACTIONS(1445),
    [aux_sym_command_name_token11] = ACTIONS(1445),
    [aux_sym_command_name_token12] = ACTIONS(1445),
    [aux_sym_command_name_token13] = ACTIONS(1445),
    [aux_sym_command_name_token14] = ACTIONS(1445),
    [aux_sym_command_name_token15] = ACTIONS(1445),
    [aux_sym_command_name_token16] = ACTIONS(1445),
    [aux_sym_command_name_token17] = ACTIONS(1445),
    [aux_sym_command_name_token18] = ACTIONS(1445),
    [aux_sym_class_statement_token1] = ACTIONS(1445),
    [aux_sym_enum_statement_token1] = ACTIONS(1445),
    [anon_sym_PLUS] = ACTIONS(1445),
    [anon_sym_DASH] = ACTIONS(1445),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1445),
    [anon_sym_BANG] = ACTIONS(1445),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1445),
    [anon_sym_PLUS_PLUS] = ACTIONS(1445),
    [anon_sym_DASH_DASH] = ACTIONS(1445),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1445),
    [anon_sym_AT_LPAREN] = ACTIONS(1445),
    [anon_sym_AT_LBRACE] = ACTIONS(1445),
  },
  [STATE(363)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1447),
    [sym_hexadecimal_integer_literal] = ACTIONS(1447),
    [sym_real_literal] = ACTIONS(1447),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1447),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1447),
    [sym_verbatim_string_characters] = ACTIONS(1447),
    [sym_verbatim_here_string_characters] = ACTIONS(1447),
    [anon_sym_DOT] = ACTIONS(1447),
    [anon_sym_LBRACK] = ACTIONS(1447),
    [aux_sym_comparison_operator_token37] = ACTIONS(1447),
    [aux_sym_comparison_operator_token50] = ACTIONS(1447),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1447),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1447),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1447),
    [anon_sym_DOLLAR_] = ACTIONS(1447),
    [aux_sym_variable_token1] = ACTIONS(1447),
    [aux_sym_variable_token2] = ACTIONS(1447),
    [sym_braced_variable] = ACTIONS(1447),
    [anon_sym_SEMI] = ACTIONS(1447),
    [anon_sym_LPAREN] = ACTIONS(1447),
    [anon_sym_RPAREN] = ACTIONS(1447),
    [anon_sym_COMMA] = ACTIONS(1447),
    [anon_sym_LBRACE] = ACTIONS(1447),
    [anon_sym_RBRACE] = ACTIONS(1447),
    [aux_sym_if_statement_token1] = ACTIONS(1447),
    [aux_sym_switch_statement_token1] = ACTIONS(1447),
    [aux_sym_foreach_statement_token1] = ACTIONS(1447),
    [aux_sym_for_statement_token1] = ACTIONS(1447),
    [aux_sym_while_statement_token1] = ACTIONS(1447),
    [aux_sym_do_statement_token1] = ACTIONS(1447),
    [aux_sym_function_statement_token1] = ACTIONS(1447),
    [aux_sym_function_statement_token2] = ACTIONS(1447),
    [aux_sym_function_statement_token3] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1447),
    [sym_label] = ACTIONS(1447),
    [aux_sym_trap_statement_token1] = ACTIONS(1447),
    [aux_sym_try_statement_token1] = ACTIONS(1447),
    [aux_sym_data_statement_token1] = ACTIONS(1447),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1447),
    [aux_sym_parallel_statement_token1] = ACTIONS(1447),
    [aux_sym_sequence_statement_token1] = ACTIONS(1447),
    [anon_sym_AMP] = ACTIONS(1447),
    [aux_sym_command_name_token1] = ACTIONS(1447),
    [aux_sym_command_name_token2] = ACTIONS(1447),
    [aux_sym_command_name_token3] = ACTIONS(1447),
    [aux_sym_command_name_token4] = ACTIONS(1447),
    [aux_sym_command_name_token5] = ACTIONS(1447),
    [aux_sym_command_name_token6] = ACTIONS(1447),
    [aux_sym_command_name_token7] = ACTIONS(1447),
    [aux_sym_command_name_token8] = ACTIONS(1447),
    [aux_sym_command_name_token9] = ACTIONS(1447),
    [aux_sym_command_name_token10] = ACTIONS(1447),
    [aux_sym_command_name_token11] = ACTIONS(1447),
    [aux_sym_command_name_token12] = ACTIONS(1447),
    [aux_sym_command_name_token13] = ACTIONS(1447),
    [aux_sym_command_name_token14] = ACTIONS(1447),
    [aux_sym_command_name_token15] = ACTIONS(1447),
    [aux_sym_command_name_token16] = ACTIONS(1447),
    [aux_sym_command_name_token17] = ACTIONS(1447),
    [aux_sym_command_name_token18] = ACTIONS(1447),
    [aux_sym_class_statement_token1] = ACTIONS(1447),
    [aux_sym_enum_statement_token1] = ACTIONS(1447),
    [anon_sym_PLUS] = ACTIONS(1447),
    [anon_sym_DASH] = ACTIONS(1447),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1447),
    [anon_sym_BANG] = ACTIONS(1447),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1447),
    [anon_sym_PLUS_PLUS] = ACTIONS(1447),
    [anon_sym_DASH_DASH] = ACTIONS(1447),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1447),
    [anon_sym_AT_LPAREN] = ACTIONS(1447),
    [anon_sym_AT_LBRACE] = ACTIONS(1447),
  },
  [STATE(364)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1449),
    [sym_hexadecimal_integer_literal] = ACTIONS(1449),
    [sym_real_literal] = ACTIONS(1449),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1449),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1449),
    [sym_verbatim_string_characters] = ACTIONS(1449),
    [sym_verbatim_here_string_characters] = ACTIONS(1449),
    [anon_sym_DOT] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1449),
    [aux_sym_comparison_operator_token37] = ACTIONS(1449),
    [aux_sym_comparison_operator_token50] = ACTIONS(1449),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1449),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1449),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1449),
    [anon_sym_DOLLAR_] = ACTIONS(1449),
    [aux_sym_variable_token1] = ACTIONS(1449),
    [aux_sym_variable_token2] = ACTIONS(1449),
    [sym_braced_variable] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1449),
    [anon_sym_RPAREN] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1449),
    [anon_sym_LBRACE] = ACTIONS(1449),
    [anon_sym_RBRACE] = ACTIONS(1449),
    [aux_sym_if_statement_token1] = ACTIONS(1449),
    [aux_sym_switch_statement_token1] = ACTIONS(1449),
    [aux_sym_foreach_statement_token1] = ACTIONS(1449),
    [aux_sym_for_statement_token1] = ACTIONS(1449),
    [aux_sym_while_statement_token1] = ACTIONS(1449),
    [aux_sym_do_statement_token1] = ACTIONS(1449),
    [aux_sym_function_statement_token1] = ACTIONS(1449),
    [aux_sym_function_statement_token2] = ACTIONS(1449),
    [aux_sym_function_statement_token3] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1449),
    [sym_label] = ACTIONS(1449),
    [aux_sym_trap_statement_token1] = ACTIONS(1449),
    [aux_sym_try_statement_token1] = ACTIONS(1449),
    [aux_sym_data_statement_token1] = ACTIONS(1449),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1449),
    [aux_sym_parallel_statement_token1] = ACTIONS(1449),
    [aux_sym_sequence_statement_token1] = ACTIONS(1449),
    [anon_sym_AMP] = ACTIONS(1449),
    [aux_sym_command_name_token1] = ACTIONS(1449),
    [aux_sym_command_name_token2] = ACTIONS(1449),
    [aux_sym_command_name_token3] = ACTIONS(1449),
    [aux_sym_command_name_token4] = ACTIONS(1449),
    [aux_sym_command_name_token5] = ACTIONS(1449),
    [aux_sym_command_name_token6] = ACTIONS(1449),
    [aux_sym_command_name_token7] = ACTIONS(1449),
    [aux_sym_command_name_token8] = ACTIONS(1449),
    [aux_sym_command_name_token9] = ACTIONS(1449),
    [aux_sym_command_name_token10] = ACTIONS(1449),
    [aux_sym_command_name_token11] = ACTIONS(1449),
    [aux_sym_command_name_token12] = ACTIONS(1449),
    [aux_sym_command_name_token13] = ACTIONS(1449),
    [aux_sym_command_name_token14] = ACTIONS(1449),
    [aux_sym_command_name_token15] = ACTIONS(1449),
    [aux_sym_command_name_token16] = ACTIONS(1449),
    [aux_sym_command_name_token17] = ACTIONS(1449),
    [aux_sym_command_name_token18] = ACTIONS(1449),
    [aux_sym_class_statement_token1] = ACTIONS(1449),
    [aux_sym_enum_statement_token1] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1449),
    [anon_sym_BANG] = ACTIONS(1449),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1449),
    [anon_sym_DASH_DASH] = ACTIONS(1449),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1449),
    [anon_sym_AT_LPAREN] = ACTIONS(1449),
    [anon_sym_AT_LBRACE] = ACTIONS(1449),
  },
  [STATE(365)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1451),
    [sym_hexadecimal_integer_literal] = ACTIONS(1451),
    [sym_real_literal] = ACTIONS(1451),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1451),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1451),
    [sym_verbatim_string_characters] = ACTIONS(1451),
    [sym_verbatim_here_string_characters] = ACTIONS(1451),
    [anon_sym_DOT] = ACTIONS(1451),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_comparison_operator_token37] = ACTIONS(1451),
    [aux_sym_comparison_operator_token50] = ACTIONS(1451),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1451),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1451),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1451),
    [anon_sym_DOLLAR_] = ACTIONS(1451),
    [aux_sym_variable_token1] = ACTIONS(1451),
    [aux_sym_variable_token2] = ACTIONS(1451),
    [sym_braced_variable] = ACTIONS(1451),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_RPAREN] = ACTIONS(1451),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [anon_sym_RBRACE] = ACTIONS(1451),
    [aux_sym_if_statement_token1] = ACTIONS(1451),
    [aux_sym_switch_statement_token1] = ACTIONS(1451),
    [aux_sym_foreach_statement_token1] = ACTIONS(1451),
    [aux_sym_for_statement_token1] = ACTIONS(1451),
    [aux_sym_while_statement_token1] = ACTIONS(1451),
    [aux_sym_do_statement_token1] = ACTIONS(1451),
    [aux_sym_function_statement_token1] = ACTIONS(1451),
    [aux_sym_function_statement_token2] = ACTIONS(1451),
    [aux_sym_function_statement_token3] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1451),
    [sym_label] = ACTIONS(1451),
    [aux_sym_trap_statement_token1] = ACTIONS(1451),
    [aux_sym_try_statement_token1] = ACTIONS(1451),
    [aux_sym_data_statement_token1] = ACTIONS(1451),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1451),
    [aux_sym_parallel_statement_token1] = ACTIONS(1451),
    [aux_sym_sequence_statement_token1] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1451),
    [aux_sym_command_name_token1] = ACTIONS(1451),
    [aux_sym_command_name_token2] = ACTIONS(1451),
    [aux_sym_command_name_token3] = ACTIONS(1451),
    [aux_sym_command_name_token4] = ACTIONS(1451),
    [aux_sym_command_name_token5] = ACTIONS(1451),
    [aux_sym_command_name_token6] = ACTIONS(1451),
    [aux_sym_command_name_token7] = ACTIONS(1451),
    [aux_sym_command_name_token8] = ACTIONS(1451),
    [aux_sym_command_name_token9] = ACTIONS(1451),
    [aux_sym_command_name_token10] = ACTIONS(1451),
    [aux_sym_command_name_token11] = ACTIONS(1451),
    [aux_sym_command_name_token12] = ACTIONS(1451),
    [aux_sym_command_name_token13] = ACTIONS(1451),
    [aux_sym_command_name_token14] = ACTIONS(1451),
    [aux_sym_command_name_token15] = ACTIONS(1451),
    [aux_sym_command_name_token16] = ACTIONS(1451),
    [aux_sym_command_name_token17] = ACTIONS(1451),
    [aux_sym_command_name_token18] = ACTIONS(1451),
    [aux_sym_class_statement_token1] = ACTIONS(1451),
    [aux_sym_enum_statement_token1] = ACTIONS(1451),
    [anon_sym_PLUS] = ACTIONS(1451),
    [anon_sym_DASH] = ACTIONS(1451),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1451),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1451),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1451),
    [anon_sym_AT_LPAREN] = ACTIONS(1451),
    [anon_sym_AT_LBRACE] = ACTIONS(1451),
  },
  [STATE(366)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1453),
    [sym_hexadecimal_integer_literal] = ACTIONS(1453),
    [sym_real_literal] = ACTIONS(1453),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1453),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1453),
    [sym_verbatim_string_characters] = ACTIONS(1453),
    [sym_verbatim_here_string_characters] = ACTIONS(1453),
    [anon_sym_DOT] = ACTIONS(1453),
    [anon_sym_LBRACK] = ACTIONS(1453),
    [aux_sym_comparison_operator_token37] = ACTIONS(1453),
    [aux_sym_comparison_operator_token50] = ACTIONS(1453),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1453),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1453),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1453),
    [anon_sym_DOLLAR_] = ACTIONS(1453),
    [aux_sym_variable_token1] = ACTIONS(1453),
    [aux_sym_variable_token2] = ACTIONS(1453),
    [sym_braced_variable] = ACTIONS(1453),
    [anon_sym_SEMI] = ACTIONS(1453),
    [anon_sym_LPAREN] = ACTIONS(1453),
    [anon_sym_RPAREN] = ACTIONS(1453),
    [anon_sym_COMMA] = ACTIONS(1453),
    [anon_sym_LBRACE] = ACTIONS(1453),
    [anon_sym_RBRACE] = ACTIONS(1453),
    [aux_sym_if_statement_token1] = ACTIONS(1453),
    [aux_sym_switch_statement_token1] = ACTIONS(1453),
    [aux_sym_foreach_statement_token1] = ACTIONS(1453),
    [aux_sym_for_statement_token1] = ACTIONS(1453),
    [aux_sym_while_statement_token1] = ACTIONS(1453),
    [aux_sym_do_statement_token1] = ACTIONS(1453),
    [aux_sym_function_statement_token1] = ACTIONS(1453),
    [aux_sym_function_statement_token2] = ACTIONS(1453),
    [aux_sym_function_statement_token3] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1453),
    [sym_label] = ACTIONS(1453),
    [aux_sym_trap_statement_token1] = ACTIONS(1453),
    [aux_sym_try_statement_token1] = ACTIONS(1453),
    [aux_sym_data_statement_token1] = ACTIONS(1453),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1453),
    [aux_sym_parallel_statement_token1] = ACTIONS(1453),
    [aux_sym_sequence_statement_token1] = ACTIONS(1453),
    [anon_sym_AMP] = ACTIONS(1453),
    [aux_sym_command_name_token1] = ACTIONS(1453),
    [aux_sym_command_name_token2] = ACTIONS(1453),
    [aux_sym_command_name_token3] = ACTIONS(1453),
    [aux_sym_command_name_token4] = ACTIONS(1453),
    [aux_sym_command_name_token5] = ACTIONS(1453),
    [aux_sym_command_name_token6] = ACTIONS(1453),
    [aux_sym_command_name_token7] = ACTIONS(1453),
    [aux_sym_command_name_token8] = ACTIONS(1453),
    [aux_sym_command_name_token9] = ACTIONS(1453),
    [aux_sym_command_name_token10] = ACTIONS(1453),
    [aux_sym_command_name_token11] = ACTIONS(1453),
    [aux_sym_command_name_token12] = ACTIONS(1453),
    [aux_sym_command_name_token13] = ACTIONS(1453),
    [aux_sym_command_name_token14] = ACTIONS(1453),
    [aux_sym_command_name_token15] = ACTIONS(1453),
    [aux_sym_command_name_token16] = ACTIONS(1453),
    [aux_sym_command_name_token17] = ACTIONS(1453),
    [aux_sym_command_name_token18] = ACTIONS(1453),
    [aux_sym_class_statement_token1] = ACTIONS(1453),
    [aux_sym_enum_statement_token1] = ACTIONS(1453),
    [anon_sym_PLUS] = ACTIONS(1453),
    [anon_sym_DASH] = ACTIONS(1453),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1453),
    [anon_sym_BANG] = ACTIONS(1453),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1453),
    [anon_sym_PLUS_PLUS] = ACTIONS(1453),
    [anon_sym_DASH_DASH] = ACTIONS(1453),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1453),
    [anon_sym_AT_LPAREN] = ACTIONS(1453),
    [anon_sym_AT_LBRACE] = ACTIONS(1453),
  },
  [STATE(367)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1455),
    [sym_hexadecimal_integer_literal] = ACTIONS(1455),
    [sym_real_literal] = ACTIONS(1455),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1455),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1455),
    [sym_verbatim_string_characters] = ACTIONS(1455),
    [sym_verbatim_here_string_characters] = ACTIONS(1455),
    [anon_sym_DOT] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1455),
    [aux_sym_comparison_operator_token37] = ACTIONS(1455),
    [aux_sym_comparison_operator_token50] = ACTIONS(1455),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1455),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1455),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1455),
    [anon_sym_DOLLAR_] = ACTIONS(1455),
    [aux_sym_variable_token1] = ACTIONS(1455),
    [aux_sym_variable_token2] = ACTIONS(1455),
    [sym_braced_variable] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1455),
    [anon_sym_RPAREN] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1455),
    [anon_sym_LBRACE] = ACTIONS(1455),
    [anon_sym_RBRACE] = ACTIONS(1455),
    [aux_sym_if_statement_token1] = ACTIONS(1455),
    [aux_sym_switch_statement_token1] = ACTIONS(1455),
    [aux_sym_foreach_statement_token1] = ACTIONS(1455),
    [aux_sym_for_statement_token1] = ACTIONS(1455),
    [aux_sym_while_statement_token1] = ACTIONS(1455),
    [aux_sym_do_statement_token1] = ACTIONS(1455),
    [aux_sym_function_statement_token1] = ACTIONS(1455),
    [aux_sym_function_statement_token2] = ACTIONS(1455),
    [aux_sym_function_statement_token3] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1455),
    [sym_label] = ACTIONS(1455),
    [aux_sym_trap_statement_token1] = ACTIONS(1455),
    [aux_sym_try_statement_token1] = ACTIONS(1455),
    [aux_sym_data_statement_token1] = ACTIONS(1455),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1455),
    [aux_sym_parallel_statement_token1] = ACTIONS(1455),
    [aux_sym_sequence_statement_token1] = ACTIONS(1455),
    [anon_sym_AMP] = ACTIONS(1455),
    [aux_sym_command_name_token1] = ACTIONS(1455),
    [aux_sym_command_name_token2] = ACTIONS(1455),
    [aux_sym_command_name_token3] = ACTIONS(1455),
    [aux_sym_command_name_token4] = ACTIONS(1455),
    [aux_sym_command_name_token5] = ACTIONS(1455),
    [aux_sym_command_name_token6] = ACTIONS(1455),
    [aux_sym_command_name_token7] = ACTIONS(1455),
    [aux_sym_command_name_token8] = ACTIONS(1455),
    [aux_sym_command_name_token9] = ACTIONS(1455),
    [aux_sym_command_name_token10] = ACTIONS(1455),
    [aux_sym_command_name_token11] = ACTIONS(1455),
    [aux_sym_command_name_token12] = ACTIONS(1455),
    [aux_sym_command_name_token13] = ACTIONS(1455),
    [aux_sym_command_name_token14] = ACTIONS(1455),
    [aux_sym_command_name_token15] = ACTIONS(1455),
    [aux_sym_command_name_token16] = ACTIONS(1455),
    [aux_sym_command_name_token17] = ACTIONS(1455),
    [aux_sym_command_name_token18] = ACTIONS(1455),
    [aux_sym_class_statement_token1] = ACTIONS(1455),
    [aux_sym_enum_statement_token1] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1455),
    [anon_sym_BANG] = ACTIONS(1455),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1455),
    [anon_sym_DASH_DASH] = ACTIONS(1455),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1455),
    [anon_sym_AT_LPAREN] = ACTIONS(1455),
    [anon_sym_AT_LBRACE] = ACTIONS(1455),
  },
  [STATE(368)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1457),
    [sym_hexadecimal_integer_literal] = ACTIONS(1457),
    [sym_real_literal] = ACTIONS(1457),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1457),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1457),
    [sym_verbatim_string_characters] = ACTIONS(1457),
    [sym_verbatim_here_string_characters] = ACTIONS(1457),
    [anon_sym_DOT] = ACTIONS(1457),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_comparison_operator_token37] = ACTIONS(1457),
    [aux_sym_comparison_operator_token50] = ACTIONS(1457),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1457),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1457),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1457),
    [anon_sym_DOLLAR_] = ACTIONS(1457),
    [aux_sym_variable_token1] = ACTIONS(1457),
    [aux_sym_variable_token2] = ACTIONS(1457),
    [sym_braced_variable] = ACTIONS(1457),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_RPAREN] = ACTIONS(1457),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_if_statement_token1] = ACTIONS(1457),
    [aux_sym_switch_statement_token1] = ACTIONS(1457),
    [aux_sym_foreach_statement_token1] = ACTIONS(1457),
    [aux_sym_for_statement_token1] = ACTIONS(1457),
    [aux_sym_while_statement_token1] = ACTIONS(1457),
    [aux_sym_do_statement_token1] = ACTIONS(1457),
    [aux_sym_function_statement_token1] = ACTIONS(1457),
    [aux_sym_function_statement_token2] = ACTIONS(1457),
    [aux_sym_function_statement_token3] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1457),
    [sym_label] = ACTIONS(1457),
    [aux_sym_trap_statement_token1] = ACTIONS(1457),
    [aux_sym_try_statement_token1] = ACTIONS(1457),
    [aux_sym_data_statement_token1] = ACTIONS(1457),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1457),
    [aux_sym_parallel_statement_token1] = ACTIONS(1457),
    [aux_sym_sequence_statement_token1] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1457),
    [aux_sym_command_name_token1] = ACTIONS(1457),
    [aux_sym_command_name_token2] = ACTIONS(1457),
    [aux_sym_command_name_token3] = ACTIONS(1457),
    [aux_sym_command_name_token4] = ACTIONS(1457),
    [aux_sym_command_name_token5] = ACTIONS(1457),
    [aux_sym_command_name_token6] = ACTIONS(1457),
    [aux_sym_command_name_token7] = ACTIONS(1457),
    [aux_sym_command_name_token8] = ACTIONS(1457),
    [aux_sym_command_name_token9] = ACTIONS(1457),
    [aux_sym_command_name_token10] = ACTIONS(1457),
    [aux_sym_command_name_token11] = ACTIONS(1457),
    [aux_sym_command_name_token12] = ACTIONS(1457),
    [aux_sym_command_name_token13] = ACTIONS(1457),
    [aux_sym_command_name_token14] = ACTIONS(1457),
    [aux_sym_command_name_token15] = ACTIONS(1457),
    [aux_sym_command_name_token16] = ACTIONS(1457),
    [aux_sym_command_name_token17] = ACTIONS(1457),
    [aux_sym_command_name_token18] = ACTIONS(1457),
    [aux_sym_class_statement_token1] = ACTIONS(1457),
    [aux_sym_enum_statement_token1] = ACTIONS(1457),
    [anon_sym_PLUS] = ACTIONS(1457),
    [anon_sym_DASH] = ACTIONS(1457),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1457),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1457),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1457),
    [anon_sym_AT_LPAREN] = ACTIONS(1457),
    [anon_sym_AT_LBRACE] = ACTIONS(1457),
  },
  [STATE(369)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1459),
    [sym_hexadecimal_integer_literal] = ACTIONS(1459),
    [sym_real_literal] = ACTIONS(1459),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1459),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1459),
    [sym_verbatim_string_characters] = ACTIONS(1459),
    [sym_verbatim_here_string_characters] = ACTIONS(1459),
    [anon_sym_DOT] = ACTIONS(1459),
    [anon_sym_LBRACK] = ACTIONS(1459),
    [aux_sym_comparison_operator_token37] = ACTIONS(1459),
    [aux_sym_comparison_operator_token50] = ACTIONS(1459),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1459),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1459),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1459),
    [anon_sym_DOLLAR_] = ACTIONS(1459),
    [aux_sym_variable_token1] = ACTIONS(1459),
    [aux_sym_variable_token2] = ACTIONS(1459),
    [sym_braced_variable] = ACTIONS(1459),
    [anon_sym_SEMI] = ACTIONS(1459),
    [anon_sym_LPAREN] = ACTIONS(1459),
    [anon_sym_RPAREN] = ACTIONS(1459),
    [anon_sym_COMMA] = ACTIONS(1459),
    [anon_sym_LBRACE] = ACTIONS(1459),
    [anon_sym_RBRACE] = ACTIONS(1459),
    [aux_sym_if_statement_token1] = ACTIONS(1459),
    [aux_sym_switch_statement_token1] = ACTIONS(1459),
    [aux_sym_foreach_statement_token1] = ACTIONS(1459),
    [aux_sym_for_statement_token1] = ACTIONS(1459),
    [aux_sym_while_statement_token1] = ACTIONS(1459),
    [aux_sym_do_statement_token1] = ACTIONS(1459),
    [aux_sym_function_statement_token1] = ACTIONS(1459),
    [aux_sym_function_statement_token2] = ACTIONS(1459),
    [aux_sym_function_statement_token3] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1459),
    [sym_label] = ACTIONS(1459),
    [aux_sym_trap_statement_token1] = ACTIONS(1459),
    [aux_sym_try_statement_token1] = ACTIONS(1459),
    [aux_sym_data_statement_token1] = ACTIONS(1459),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1459),
    [aux_sym_parallel_statement_token1] = ACTIONS(1459),
    [aux_sym_sequence_statement_token1] = ACTIONS(1459),
    [anon_sym_AMP] = ACTIONS(1459),
    [aux_sym_command_name_token1] = ACTIONS(1459),
    [aux_sym_command_name_token2] = ACTIONS(1459),
    [aux_sym_command_name_token3] = ACTIONS(1459),
    [aux_sym_command_name_token4] = ACTIONS(1459),
    [aux_sym_command_name_token5] = ACTIONS(1459),
    [aux_sym_command_name_token6] = ACTIONS(1459),
    [aux_sym_command_name_token7] = ACTIONS(1459),
    [aux_sym_command_name_token8] = ACTIONS(1459),
    [aux_sym_command_name_token9] = ACTIONS(1459),
    [aux_sym_command_name_token10] = ACTIONS(1459),
    [aux_sym_command_name_token11] = ACTIONS(1459),
    [aux_sym_command_name_token12] = ACTIONS(1459),
    [aux_sym_command_name_token13] = ACTIONS(1459),
    [aux_sym_command_name_token14] = ACTIONS(1459),
    [aux_sym_command_name_token15] = ACTIONS(1459),
    [aux_sym_command_name_token16] = ACTIONS(1459),
    [aux_sym_command_name_token17] = ACTIONS(1459),
    [aux_sym_command_name_token18] = ACTIONS(1459),
    [aux_sym_class_statement_token1] = ACTIONS(1459),
    [aux_sym_enum_statement_token1] = ACTIONS(1459),
    [anon_sym_PLUS] = ACTIONS(1459),
    [anon_sym_DASH] = ACTIONS(1459),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1459),
    [anon_sym_BANG] = ACTIONS(1459),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1459),
    [anon_sym_PLUS_PLUS] = ACTIONS(1459),
    [anon_sym_DASH_DASH] = ACTIONS(1459),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1459),
    [anon_sym_AT_LPAREN] = ACTIONS(1459),
    [anon_sym_AT_LBRACE] = ACTIONS(1459),
  },
  [STATE(370)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1461),
    [sym_hexadecimal_integer_literal] = ACTIONS(1461),
    [sym_real_literal] = ACTIONS(1461),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1461),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1461),
    [sym_verbatim_string_characters] = ACTIONS(1461),
    [sym_verbatim_here_string_characters] = ACTIONS(1461),
    [anon_sym_DOT] = ACTIONS(1461),
    [anon_sym_LBRACK] = ACTIONS(1461),
    [aux_sym_comparison_operator_token37] = ACTIONS(1461),
    [aux_sym_comparison_operator_token50] = ACTIONS(1461),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1461),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1461),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1461),
    [anon_sym_DOLLAR_] = ACTIONS(1461),
    [aux_sym_variable_token1] = ACTIONS(1461),
    [aux_sym_variable_token2] = ACTIONS(1461),
    [sym_braced_variable] = ACTIONS(1461),
    [anon_sym_SEMI] = ACTIONS(1461),
    [anon_sym_LPAREN] = ACTIONS(1461),
    [anon_sym_RPAREN] = ACTIONS(1461),
    [anon_sym_COMMA] = ACTIONS(1461),
    [anon_sym_LBRACE] = ACTIONS(1461),
    [anon_sym_RBRACE] = ACTIONS(1461),
    [aux_sym_if_statement_token1] = ACTIONS(1461),
    [aux_sym_switch_statement_token1] = ACTIONS(1461),
    [aux_sym_foreach_statement_token1] = ACTIONS(1461),
    [aux_sym_for_statement_token1] = ACTIONS(1461),
    [aux_sym_while_statement_token1] = ACTIONS(1461),
    [aux_sym_do_statement_token1] = ACTIONS(1461),
    [aux_sym_function_statement_token1] = ACTIONS(1461),
    [aux_sym_function_statement_token2] = ACTIONS(1461),
    [aux_sym_function_statement_token3] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1461),
    [sym_label] = ACTIONS(1461),
    [aux_sym_trap_statement_token1] = ACTIONS(1461),
    [aux_sym_try_statement_token1] = ACTIONS(1461),
    [aux_sym_data_statement_token1] = ACTIONS(1461),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1461),
    [aux_sym_parallel_statement_token1] = ACTIONS(1461),
    [aux_sym_sequence_statement_token1] = ACTIONS(1461),
    [anon_sym_AMP] = ACTIONS(1461),
    [aux_sym_command_name_token1] = ACTIONS(1461),
    [aux_sym_command_name_token2] = ACTIONS(1461),
    [aux_sym_command_name_token3] = ACTIONS(1461),
    [aux_sym_command_name_token4] = ACTIONS(1461),
    [aux_sym_command_name_token5] = ACTIONS(1461),
    [aux_sym_command_name_token6] = ACTIONS(1461),
    [aux_sym_command_name_token7] = ACTIONS(1461),
    [aux_sym_command_name_token8] = ACTIONS(1461),
    [aux_sym_command_name_token9] = ACTIONS(1461),
    [aux_sym_command_name_token10] = ACTIONS(1461),
    [aux_sym_command_name_token11] = ACTIONS(1461),
    [aux_sym_command_name_token12] = ACTIONS(1461),
    [aux_sym_command_name_token13] = ACTIONS(1461),
    [aux_sym_command_name_token14] = ACTIONS(1461),
    [aux_sym_command_name_token15] = ACTIONS(1461),
    [aux_sym_command_name_token16] = ACTIONS(1461),
    [aux_sym_command_name_token17] = ACTIONS(1461),
    [aux_sym_command_name_token18] = ACTIONS(1461),
    [aux_sym_class_statement_token1] = ACTIONS(1461),
    [aux_sym_enum_statement_token1] = ACTIONS(1461),
    [anon_sym_PLUS] = ACTIONS(1461),
    [anon_sym_DASH] = ACTIONS(1461),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1461),
    [anon_sym_BANG] = ACTIONS(1461),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1461),
    [anon_sym_PLUS_PLUS] = ACTIONS(1461),
    [anon_sym_DASH_DASH] = ACTIONS(1461),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1461),
    [anon_sym_AT_LPAREN] = ACTIONS(1461),
    [anon_sym_AT_LBRACE] = ACTIONS(1461),
  },
  [STATE(371)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1463),
    [sym_hexadecimal_integer_literal] = ACTIONS(1463),
    [sym_real_literal] = ACTIONS(1463),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1463),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1463),
    [sym_verbatim_string_characters] = ACTIONS(1463),
    [sym_verbatim_here_string_characters] = ACTIONS(1463),
    [anon_sym_DOT] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1463),
    [aux_sym_comparison_operator_token37] = ACTIONS(1463),
    [aux_sym_comparison_operator_token50] = ACTIONS(1463),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1463),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1463),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1463),
    [anon_sym_DOLLAR_] = ACTIONS(1463),
    [aux_sym_variable_token1] = ACTIONS(1463),
    [aux_sym_variable_token2] = ACTIONS(1463),
    [sym_braced_variable] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1463),
    [anon_sym_RPAREN] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1463),
    [anon_sym_LBRACE] = ACTIONS(1463),
    [anon_sym_RBRACE] = ACTIONS(1463),
    [aux_sym_if_statement_token1] = ACTIONS(1463),
    [aux_sym_switch_statement_token1] = ACTIONS(1463),
    [aux_sym_foreach_statement_token1] = ACTIONS(1463),
    [aux_sym_for_statement_token1] = ACTIONS(1463),
    [aux_sym_while_statement_token1] = ACTIONS(1463),
    [aux_sym_do_statement_token1] = ACTIONS(1463),
    [aux_sym_function_statement_token1] = ACTIONS(1463),
    [aux_sym_function_statement_token2] = ACTIONS(1463),
    [aux_sym_function_statement_token3] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1463),
    [sym_label] = ACTIONS(1463),
    [aux_sym_trap_statement_token1] = ACTIONS(1463),
    [aux_sym_try_statement_token1] = ACTIONS(1463),
    [aux_sym_data_statement_token1] = ACTIONS(1463),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1463),
    [aux_sym_parallel_statement_token1] = ACTIONS(1463),
    [aux_sym_sequence_statement_token1] = ACTIONS(1463),
    [anon_sym_AMP] = ACTIONS(1463),
    [aux_sym_command_name_token1] = ACTIONS(1463),
    [aux_sym_command_name_token2] = ACTIONS(1463),
    [aux_sym_command_name_token3] = ACTIONS(1463),
    [aux_sym_command_name_token4] = ACTIONS(1463),
    [aux_sym_command_name_token5] = ACTIONS(1463),
    [aux_sym_command_name_token6] = ACTIONS(1463),
    [aux_sym_command_name_token7] = ACTIONS(1463),
    [aux_sym_command_name_token8] = ACTIONS(1463),
    [aux_sym_command_name_token9] = ACTIONS(1463),
    [aux_sym_command_name_token10] = ACTIONS(1463),
    [aux_sym_command_name_token11] = ACTIONS(1463),
    [aux_sym_command_name_token12] = ACTIONS(1463),
    [aux_sym_command_name_token13] = ACTIONS(1463),
    [aux_sym_command_name_token14] = ACTIONS(1463),
    [aux_sym_command_name_token15] = ACTIONS(1463),
    [aux_sym_command_name_token16] = ACTIONS(1463),
    [aux_sym_command_name_token17] = ACTIONS(1463),
    [aux_sym_command_name_token18] = ACTIONS(1463),
    [aux_sym_class_statement_token1] = ACTIONS(1463),
    [aux_sym_enum_statement_token1] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1463),
    [anon_sym_BANG] = ACTIONS(1463),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1463),
    [anon_sym_DASH_DASH] = ACTIONS(1463),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1463),
    [anon_sym_AT_LPAREN] = ACTIONS(1463),
    [anon_sym_AT_LBRACE] = ACTIONS(1463),
  },
  [STATE(372)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1465),
    [sym_hexadecimal_integer_literal] = ACTIONS(1465),
    [sym_real_literal] = ACTIONS(1465),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1465),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1465),
    [sym_verbatim_string_characters] = ACTIONS(1465),
    [sym_verbatim_here_string_characters] = ACTIONS(1465),
    [anon_sym_DOT] = ACTIONS(1465),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_comparison_operator_token37] = ACTIONS(1465),
    [aux_sym_comparison_operator_token50] = ACTIONS(1465),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1465),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1465),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1465),
    [anon_sym_DOLLAR_] = ACTIONS(1465),
    [aux_sym_variable_token1] = ACTIONS(1465),
    [aux_sym_variable_token2] = ACTIONS(1465),
    [sym_braced_variable] = ACTIONS(1465),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_RPAREN] = ACTIONS(1465),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [anon_sym_RBRACE] = ACTIONS(1465),
    [aux_sym_if_statement_token1] = ACTIONS(1465),
    [aux_sym_switch_statement_token1] = ACTIONS(1465),
    [aux_sym_foreach_statement_token1] = ACTIONS(1465),
    [aux_sym_for_statement_token1] = ACTIONS(1465),
    [aux_sym_while_statement_token1] = ACTIONS(1465),
    [aux_sym_do_statement_token1] = ACTIONS(1465),
    [aux_sym_function_statement_token1] = ACTIONS(1465),
    [aux_sym_function_statement_token2] = ACTIONS(1465),
    [aux_sym_function_statement_token3] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1465),
    [sym_label] = ACTIONS(1465),
    [aux_sym_trap_statement_token1] = ACTIONS(1465),
    [aux_sym_try_statement_token1] = ACTIONS(1465),
    [aux_sym_data_statement_token1] = ACTIONS(1465),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1465),
    [aux_sym_parallel_statement_token1] = ACTIONS(1465),
    [aux_sym_sequence_statement_token1] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1465),
    [aux_sym_command_name_token1] = ACTIONS(1465),
    [aux_sym_command_name_token2] = ACTIONS(1465),
    [aux_sym_command_name_token3] = ACTIONS(1465),
    [aux_sym_command_name_token4] = ACTIONS(1465),
    [aux_sym_command_name_token5] = ACTIONS(1465),
    [aux_sym_command_name_token6] = ACTIONS(1465),
    [aux_sym_command_name_token7] = ACTIONS(1465),
    [aux_sym_command_name_token8] = ACTIONS(1465),
    [aux_sym_command_name_token9] = ACTIONS(1465),
    [aux_sym_command_name_token10] = ACTIONS(1465),
    [aux_sym_command_name_token11] = ACTIONS(1465),
    [aux_sym_command_name_token12] = ACTIONS(1465),
    [aux_sym_command_name_token13] = ACTIONS(1465),
    [aux_sym_command_name_token14] = ACTIONS(1465),
    [aux_sym_command_name_token15] = ACTIONS(1465),
    [aux_sym_command_name_token16] = ACTIONS(1465),
    [aux_sym_command_name_token17] = ACTIONS(1465),
    [aux_sym_command_name_token18] = ACTIONS(1465),
    [aux_sym_class_statement_token1] = ACTIONS(1465),
    [aux_sym_enum_statement_token1] = ACTIONS(1465),
    [anon_sym_PLUS] = ACTIONS(1465),
    [anon_sym_DASH] = ACTIONS(1465),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1465),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1465),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1465),
    [anon_sym_AT_LPAREN] = ACTIONS(1465),
    [anon_sym_AT_LBRACE] = ACTIONS(1465),
  },
  [STATE(373)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1467),
    [sym_hexadecimal_integer_literal] = ACTIONS(1467),
    [sym_real_literal] = ACTIONS(1467),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1467),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1467),
    [sym_verbatim_string_characters] = ACTIONS(1467),
    [sym_verbatim_here_string_characters] = ACTIONS(1467),
    [anon_sym_DOT] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1467),
    [aux_sym_comparison_operator_token37] = ACTIONS(1467),
    [aux_sym_comparison_operator_token50] = ACTIONS(1467),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1467),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1467),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1467),
    [anon_sym_DOLLAR_] = ACTIONS(1467),
    [aux_sym_variable_token1] = ACTIONS(1467),
    [aux_sym_variable_token2] = ACTIONS(1467),
    [sym_braced_variable] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1467),
    [anon_sym_RPAREN] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1467),
    [anon_sym_LBRACE] = ACTIONS(1467),
    [anon_sym_RBRACE] = ACTIONS(1467),
    [aux_sym_if_statement_token1] = ACTIONS(1467),
    [aux_sym_switch_statement_token1] = ACTIONS(1467),
    [aux_sym_foreach_statement_token1] = ACTIONS(1467),
    [aux_sym_for_statement_token1] = ACTIONS(1467),
    [aux_sym_while_statement_token1] = ACTIONS(1467),
    [aux_sym_do_statement_token1] = ACTIONS(1467),
    [aux_sym_function_statement_token1] = ACTIONS(1467),
    [aux_sym_function_statement_token2] = ACTIONS(1467),
    [aux_sym_function_statement_token3] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1467),
    [sym_label] = ACTIONS(1467),
    [aux_sym_trap_statement_token1] = ACTIONS(1467),
    [aux_sym_try_statement_token1] = ACTIONS(1467),
    [aux_sym_data_statement_token1] = ACTIONS(1467),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1467),
    [aux_sym_parallel_statement_token1] = ACTIONS(1467),
    [aux_sym_sequence_statement_token1] = ACTIONS(1467),
    [anon_sym_AMP] = ACTIONS(1467),
    [aux_sym_command_name_token1] = ACTIONS(1467),
    [aux_sym_command_name_token2] = ACTIONS(1467),
    [aux_sym_command_name_token3] = ACTIONS(1467),
    [aux_sym_command_name_token4] = ACTIONS(1467),
    [aux_sym_command_name_token5] = ACTIONS(1467),
    [aux_sym_command_name_token6] = ACTIONS(1467),
    [aux_sym_command_name_token7] = ACTIONS(1467),
    [aux_sym_command_name_token8] = ACTIONS(1467),
    [aux_sym_command_name_token9] = ACTIONS(1467),
    [aux_sym_command_name_token10] = ACTIONS(1467),
    [aux_sym_command_name_token11] = ACTIONS(1467),
    [aux_sym_command_name_token12] = ACTIONS(1467),
    [aux_sym_command_name_token13] = ACTIONS(1467),
    [aux_sym_command_name_token14] = ACTIONS(1467),
    [aux_sym_command_name_token15] = ACTIONS(1467),
    [aux_sym_command_name_token16] = ACTIONS(1467),
    [aux_sym_command_name_token17] = ACTIONS(1467),
    [aux_sym_command_name_token18] = ACTIONS(1467),
    [aux_sym_class_statement_token1] = ACTIONS(1467),
    [aux_sym_enum_statement_token1] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1467),
    [anon_sym_BANG] = ACTIONS(1467),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1467),
    [anon_sym_DASH_DASH] = ACTIONS(1467),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1467),
    [anon_sym_AT_LPAREN] = ACTIONS(1467),
    [anon_sym_AT_LBRACE] = ACTIONS(1467),
  },
  [STATE(374)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1469),
    [sym_hexadecimal_integer_literal] = ACTIONS(1469),
    [sym_real_literal] = ACTIONS(1469),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1469),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1469),
    [sym_verbatim_string_characters] = ACTIONS(1469),
    [sym_verbatim_here_string_characters] = ACTIONS(1469),
    [anon_sym_DOT] = ACTIONS(1469),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_comparison_operator_token37] = ACTIONS(1469),
    [aux_sym_comparison_operator_token50] = ACTIONS(1469),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1469),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1469),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1469),
    [anon_sym_DOLLAR_] = ACTIONS(1469),
    [aux_sym_variable_token1] = ACTIONS(1469),
    [aux_sym_variable_token2] = ACTIONS(1469),
    [sym_braced_variable] = ACTIONS(1469),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_RPAREN] = ACTIONS(1469),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_if_statement_token1] = ACTIONS(1469),
    [aux_sym_switch_statement_token1] = ACTIONS(1469),
    [aux_sym_foreach_statement_token1] = ACTIONS(1469),
    [aux_sym_for_statement_token1] = ACTIONS(1469),
    [aux_sym_while_statement_token1] = ACTIONS(1469),
    [aux_sym_do_statement_token1] = ACTIONS(1469),
    [aux_sym_function_statement_token1] = ACTIONS(1469),
    [aux_sym_function_statement_token2] = ACTIONS(1469),
    [aux_sym_function_statement_token3] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1469),
    [sym_label] = ACTIONS(1469),
    [aux_sym_trap_statement_token1] = ACTIONS(1469),
    [aux_sym_try_statement_token1] = ACTIONS(1469),
    [aux_sym_data_statement_token1] = ACTIONS(1469),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1469),
    [aux_sym_parallel_statement_token1] = ACTIONS(1469),
    [aux_sym_sequence_statement_token1] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1469),
    [aux_sym_command_name_token1] = ACTIONS(1469),
    [aux_sym_command_name_token2] = ACTIONS(1469),
    [aux_sym_command_name_token3] = ACTIONS(1469),
    [aux_sym_command_name_token4] = ACTIONS(1469),
    [aux_sym_command_name_token5] = ACTIONS(1469),
    [aux_sym_command_name_token6] = ACTIONS(1469),
    [aux_sym_command_name_token7] = ACTIONS(1469),
    [aux_sym_command_name_token8] = ACTIONS(1469),
    [aux_sym_command_name_token9] = ACTIONS(1469),
    [aux_sym_command_name_token10] = ACTIONS(1469),
    [aux_sym_command_name_token11] = ACTIONS(1469),
    [aux_sym_command_name_token12] = ACTIONS(1469),
    [aux_sym_command_name_token13] = ACTIONS(1469),
    [aux_sym_command_name_token14] = ACTIONS(1469),
    [aux_sym_command_name_token15] = ACTIONS(1469),
    [aux_sym_command_name_token16] = ACTIONS(1469),
    [aux_sym_command_name_token17] = ACTIONS(1469),
    [aux_sym_command_name_token18] = ACTIONS(1469),
    [aux_sym_class_statement_token1] = ACTIONS(1469),
    [aux_sym_enum_statement_token1] = ACTIONS(1469),
    [anon_sym_PLUS] = ACTIONS(1469),
    [anon_sym_DASH] = ACTIONS(1469),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1469),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1469),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1469),
    [anon_sym_AT_LPAREN] = ACTIONS(1469),
    [anon_sym_AT_LBRACE] = ACTIONS(1469),
  },
  [STATE(375)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1471),
    [sym_hexadecimal_integer_literal] = ACTIONS(1471),
    [sym_real_literal] = ACTIONS(1471),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1471),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1471),
    [sym_verbatim_string_characters] = ACTIONS(1471),
    [sym_verbatim_here_string_characters] = ACTIONS(1471),
    [anon_sym_DOT] = ACTIONS(1471),
    [anon_sym_LBRACK] = ACTIONS(1471),
    [aux_sym_comparison_operator_token37] = ACTIONS(1471),
    [aux_sym_comparison_operator_token50] = ACTIONS(1471),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1471),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1471),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1471),
    [anon_sym_DOLLAR_] = ACTIONS(1471),
    [aux_sym_variable_token1] = ACTIONS(1471),
    [aux_sym_variable_token2] = ACTIONS(1471),
    [sym_braced_variable] = ACTIONS(1471),
    [anon_sym_SEMI] = ACTIONS(1471),
    [anon_sym_LPAREN] = ACTIONS(1471),
    [anon_sym_RPAREN] = ACTIONS(1471),
    [anon_sym_COMMA] = ACTIONS(1471),
    [anon_sym_LBRACE] = ACTIONS(1471),
    [anon_sym_RBRACE] = ACTIONS(1471),
    [aux_sym_if_statement_token1] = ACTIONS(1471),
    [aux_sym_switch_statement_token1] = ACTIONS(1471),
    [aux_sym_foreach_statement_token1] = ACTIONS(1471),
    [aux_sym_for_statement_token1] = ACTIONS(1471),
    [aux_sym_while_statement_token1] = ACTIONS(1471),
    [aux_sym_do_statement_token1] = ACTIONS(1471),
    [aux_sym_function_statement_token1] = ACTIONS(1471),
    [aux_sym_function_statement_token2] = ACTIONS(1471),
    [aux_sym_function_statement_token3] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1471),
    [sym_label] = ACTIONS(1471),
    [aux_sym_trap_statement_token1] = ACTIONS(1471),
    [aux_sym_try_statement_token1] = ACTIONS(1471),
    [aux_sym_data_statement_token1] = ACTIONS(1471),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1471),
    [aux_sym_parallel_statement_token1] = ACTIONS(1471),
    [aux_sym_sequence_statement_token1] = ACTIONS(1471),
    [anon_sym_AMP] = ACTIONS(1471),
    [aux_sym_command_name_token1] = ACTIONS(1471),
    [aux_sym_command_name_token2] = ACTIONS(1471),
    [aux_sym_command_name_token3] = ACTIONS(1471),
    [aux_sym_command_name_token4] = ACTIONS(1471),
    [aux_sym_command_name_token5] = ACTIONS(1471),
    [aux_sym_command_name_token6] = ACTIONS(1471),
    [aux_sym_command_name_token7] = ACTIONS(1471),
    [aux_sym_command_name_token8] = ACTIONS(1471),
    [aux_sym_command_name_token9] = ACTIONS(1471),
    [aux_sym_command_name_token10] = ACTIONS(1471),
    [aux_sym_command_name_token11] = ACTIONS(1471),
    [aux_sym_command_name_token12] = ACTIONS(1471),
    [aux_sym_command_name_token13] = ACTIONS(1471),
    [aux_sym_command_name_token14] = ACTIONS(1471),
    [aux_sym_command_name_token15] = ACTIONS(1471),
    [aux_sym_command_name_token16] = ACTIONS(1471),
    [aux_sym_command_name_token17] = ACTIONS(1471),
    [aux_sym_command_name_token18] = ACTIONS(1471),
    [aux_sym_class_statement_token1] = ACTIONS(1471),
    [aux_sym_enum_statement_token1] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1471),
    [anon_sym_DASH] = ACTIONS(1471),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1471),
    [anon_sym_BANG] = ACTIONS(1471),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1471),
    [anon_sym_PLUS_PLUS] = ACTIONS(1471),
    [anon_sym_DASH_DASH] = ACTIONS(1471),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1471),
    [anon_sym_AT_LPAREN] = ACTIONS(1471),
    [anon_sym_AT_LBRACE] = ACTIONS(1471),
  },
  [STATE(376)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1473),
    [sym_hexadecimal_integer_literal] = ACTIONS(1473),
    [sym_real_literal] = ACTIONS(1473),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1473),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1473),
    [sym_verbatim_string_characters] = ACTIONS(1473),
    [sym_verbatim_here_string_characters] = ACTIONS(1473),
    [anon_sym_DOT] = ACTIONS(1473),
    [anon_sym_LBRACK] = ACTIONS(1473),
    [aux_sym_comparison_operator_token37] = ACTIONS(1473),
    [aux_sym_comparison_operator_token50] = ACTIONS(1473),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1473),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1473),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1473),
    [anon_sym_DOLLAR_] = ACTIONS(1473),
    [aux_sym_variable_token1] = ACTIONS(1473),
    [aux_sym_variable_token2] = ACTIONS(1473),
    [sym_braced_variable] = ACTIONS(1473),
    [anon_sym_SEMI] = ACTIONS(1473),
    [anon_sym_LPAREN] = ACTIONS(1473),
    [anon_sym_RPAREN] = ACTIONS(1473),
    [anon_sym_COMMA] = ACTIONS(1473),
    [anon_sym_LBRACE] = ACTIONS(1473),
    [anon_sym_RBRACE] = ACTIONS(1473),
    [aux_sym_if_statement_token1] = ACTIONS(1473),
    [aux_sym_switch_statement_token1] = ACTIONS(1473),
    [aux_sym_foreach_statement_token1] = ACTIONS(1473),
    [aux_sym_for_statement_token1] = ACTIONS(1473),
    [aux_sym_while_statement_token1] = ACTIONS(1473),
    [aux_sym_do_statement_token1] = ACTIONS(1473),
    [aux_sym_function_statement_token1] = ACTIONS(1473),
    [aux_sym_function_statement_token2] = ACTIONS(1473),
    [aux_sym_function_statement_token3] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1473),
    [sym_label] = ACTIONS(1473),
    [aux_sym_trap_statement_token1] = ACTIONS(1473),
    [aux_sym_try_statement_token1] = ACTIONS(1473),
    [aux_sym_data_statement_token1] = ACTIONS(1473),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1473),
    [aux_sym_parallel_statement_token1] = ACTIONS(1473),
    [aux_sym_sequence_statement_token1] = ACTIONS(1473),
    [anon_sym_AMP] = ACTIONS(1473),
    [aux_sym_command_name_token1] = ACTIONS(1473),
    [aux_sym_command_name_token2] = ACTIONS(1473),
    [aux_sym_command_name_token3] = ACTIONS(1473),
    [aux_sym_command_name_token4] = ACTIONS(1473),
    [aux_sym_command_name_token5] = ACTIONS(1473),
    [aux_sym_command_name_token6] = ACTIONS(1473),
    [aux_sym_command_name_token7] = ACTIONS(1473),
    [aux_sym_command_name_token8] = ACTIONS(1473),
    [aux_sym_command_name_token9] = ACTIONS(1473),
    [aux_sym_command_name_token10] = ACTIONS(1473),
    [aux_sym_command_name_token11] = ACTIONS(1473),
    [aux_sym_command_name_token12] = ACTIONS(1473),
    [aux_sym_command_name_token13] = ACTIONS(1473),
    [aux_sym_command_name_token14] = ACTIONS(1473),
    [aux_sym_command_name_token15] = ACTIONS(1473),
    [aux_sym_command_name_token16] = ACTIONS(1473),
    [aux_sym_command_name_token17] = ACTIONS(1473),
    [aux_sym_command_name_token18] = ACTIONS(1473),
    [aux_sym_class_statement_token1] = ACTIONS(1473),
    [aux_sym_enum_statement_token1] = ACTIONS(1473),
    [anon_sym_PLUS] = ACTIONS(1473),
    [anon_sym_DASH] = ACTIONS(1473),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1473),
    [anon_sym_BANG] = ACTIONS(1473),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1473),
    [anon_sym_PLUS_PLUS] = ACTIONS(1473),
    [anon_sym_DASH_DASH] = ACTIONS(1473),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1473),
    [anon_sym_AT_LPAREN] = ACTIONS(1473),
    [anon_sym_AT_LBRACE] = ACTIONS(1473),
  },
  [STATE(377)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1475),
    [sym_hexadecimal_integer_literal] = ACTIONS(1475),
    [sym_real_literal] = ACTIONS(1475),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1475),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1475),
    [sym_verbatim_string_characters] = ACTIONS(1475),
    [sym_verbatim_here_string_characters] = ACTIONS(1475),
    [anon_sym_DOT] = ACTIONS(1475),
    [anon_sym_LBRACK] = ACTIONS(1475),
    [aux_sym_comparison_operator_token37] = ACTIONS(1475),
    [aux_sym_comparison_operator_token50] = ACTIONS(1475),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1475),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1475),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1475),
    [anon_sym_DOLLAR_] = ACTIONS(1475),
    [aux_sym_variable_token1] = ACTIONS(1475),
    [aux_sym_variable_token2] = ACTIONS(1475),
    [sym_braced_variable] = ACTIONS(1475),
    [anon_sym_SEMI] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1475),
    [anon_sym_RPAREN] = ACTIONS(1475),
    [anon_sym_COMMA] = ACTIONS(1475),
    [anon_sym_LBRACE] = ACTIONS(1475),
    [anon_sym_RBRACE] = ACTIONS(1475),
    [aux_sym_if_statement_token1] = ACTIONS(1475),
    [aux_sym_switch_statement_token1] = ACTIONS(1475),
    [aux_sym_foreach_statement_token1] = ACTIONS(1475),
    [aux_sym_for_statement_token1] = ACTIONS(1475),
    [aux_sym_while_statement_token1] = ACTIONS(1475),
    [aux_sym_do_statement_token1] = ACTIONS(1475),
    [aux_sym_function_statement_token1] = ACTIONS(1475),
    [aux_sym_function_statement_token2] = ACTIONS(1475),
    [aux_sym_function_statement_token3] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1475),
    [sym_label] = ACTIONS(1475),
    [aux_sym_trap_statement_token1] = ACTIONS(1475),
    [aux_sym_try_statement_token1] = ACTIONS(1475),
    [aux_sym_data_statement_token1] = ACTIONS(1475),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1475),
    [aux_sym_parallel_statement_token1] = ACTIONS(1475),
    [aux_sym_sequence_statement_token1] = ACTIONS(1475),
    [anon_sym_AMP] = ACTIONS(1475),
    [aux_sym_command_name_token1] = ACTIONS(1475),
    [aux_sym_command_name_token2] = ACTIONS(1475),
    [aux_sym_command_name_token3] = ACTIONS(1475),
    [aux_sym_command_name_token4] = ACTIONS(1475),
    [aux_sym_command_name_token5] = ACTIONS(1475),
    [aux_sym_command_name_token6] = ACTIONS(1475),
    [aux_sym_command_name_token7] = ACTIONS(1475),
    [aux_sym_command_name_token8] = ACTIONS(1475),
    [aux_sym_command_name_token9] = ACTIONS(1475),
    [aux_sym_command_name_token10] = ACTIONS(1475),
    [aux_sym_command_name_token11] = ACTIONS(1475),
    [aux_sym_command_name_token12] = ACTIONS(1475),
    [aux_sym_command_name_token13] = ACTIONS(1475),
    [aux_sym_command_name_token14] = ACTIONS(1475),
    [aux_sym_command_name_token15] = ACTIONS(1475),
    [aux_sym_command_name_token16] = ACTIONS(1475),
    [aux_sym_command_name_token17] = ACTIONS(1475),
    [aux_sym_command_name_token18] = ACTIONS(1475),
    [aux_sym_class_statement_token1] = ACTIONS(1475),
    [aux_sym_enum_statement_token1] = ACTIONS(1475),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1475),
    [anon_sym_BANG] = ACTIONS(1475),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1475),
    [anon_sym_PLUS_PLUS] = ACTIONS(1475),
    [anon_sym_DASH_DASH] = ACTIONS(1475),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1475),
    [anon_sym_AT_LPAREN] = ACTIONS(1475),
    [anon_sym_AT_LBRACE] = ACTIONS(1475),
  },
  [STATE(378)] = {
    [ts_builtin_sym_end] = ACTIONS(1477),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1451),
    [sym_hexadecimal_integer_literal] = ACTIONS(1451),
    [sym_real_literal] = ACTIONS(1451),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1451),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1451),
    [sym_verbatim_string_characters] = ACTIONS(1451),
    [sym_verbatim_here_string_characters] = ACTIONS(1451),
    [anon_sym_DOT] = ACTIONS(1451),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_comparison_operator_token37] = ACTIONS(1451),
    [aux_sym_comparison_operator_token50] = ACTIONS(1451),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1451),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1451),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1451),
    [anon_sym_DOLLAR_] = ACTIONS(1451),
    [aux_sym_variable_token1] = ACTIONS(1451),
    [aux_sym_variable_token2] = ACTIONS(1451),
    [sym_braced_variable] = ACTIONS(1451),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [aux_sym_if_statement_token1] = ACTIONS(1451),
    [aux_sym_switch_statement_token1] = ACTIONS(1451),
    [aux_sym_foreach_statement_token1] = ACTIONS(1451),
    [aux_sym_for_statement_token1] = ACTIONS(1451),
    [aux_sym_while_statement_token1] = ACTIONS(1451),
    [aux_sym_do_statement_token1] = ACTIONS(1451),
    [aux_sym_function_statement_token1] = ACTIONS(1451),
    [aux_sym_function_statement_token2] = ACTIONS(1451),
    [aux_sym_function_statement_token3] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1451),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1451),
    [sym_label] = ACTIONS(1451),
    [aux_sym_trap_statement_token1] = ACTIONS(1451),
    [aux_sym_try_statement_token1] = ACTIONS(1451),
    [aux_sym_data_statement_token1] = ACTIONS(1451),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1451),
    [aux_sym_parallel_statement_token1] = ACTIONS(1451),
    [aux_sym_sequence_statement_token1] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1451),
    [aux_sym_command_name_token1] = ACTIONS(1451),
    [aux_sym_command_name_token2] = ACTIONS(1451),
    [aux_sym_command_name_token3] = ACTIONS(1451),
    [aux_sym_command_name_token4] = ACTIONS(1451),
    [aux_sym_command_name_token5] = ACTIONS(1451),
    [aux_sym_command_name_token6] = ACTIONS(1451),
    [aux_sym_command_name_token7] = ACTIONS(1451),
    [aux_sym_command_name_token8] = ACTIONS(1451),
    [aux_sym_command_name_token9] = ACTIONS(1451),
    [aux_sym_command_name_token10] = ACTIONS(1451),
    [aux_sym_command_name_token11] = ACTIONS(1451),
    [aux_sym_command_name_token12] = ACTIONS(1451),
    [aux_sym_command_name_token13] = ACTIONS(1451),
    [aux_sym_command_name_token14] = ACTIONS(1451),
    [aux_sym_command_name_token15] = ACTIONS(1451),
    [aux_sym_command_name_token16] = ACTIONS(1451),
    [aux_sym_command_name_token17] = ACTIONS(1451),
    [aux_sym_command_name_token18] = ACTIONS(1451),
    [aux_sym_class_statement_token1] = ACTIONS(1451),
    [aux_sym_enum_statement_token1] = ACTIONS(1451),
    [anon_sym_PLUS] = ACTIONS(1451),
    [anon_sym_DASH] = ACTIONS(1451),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1451),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1451),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1451),
    [anon_sym_AT_LPAREN] = ACTIONS(1451),
    [anon_sym_AT_LBRACE] = ACTIONS(1451),
  },
  [STATE(379)] = {
    [ts_builtin_sym_end] = ACTIONS(1479),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1463),
    [sym_hexadecimal_integer_literal] = ACTIONS(1463),
    [sym_real_literal] = ACTIONS(1463),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1463),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1463),
    [sym_verbatim_string_characters] = ACTIONS(1463),
    [sym_verbatim_here_string_characters] = ACTIONS(1463),
    [anon_sym_DOT] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1463),
    [aux_sym_comparison_operator_token37] = ACTIONS(1463),
    [aux_sym_comparison_operator_token50] = ACTIONS(1463),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1463),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1463),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1463),
    [anon_sym_DOLLAR_] = ACTIONS(1463),
    [aux_sym_variable_token1] = ACTIONS(1463),
    [aux_sym_variable_token2] = ACTIONS(1463),
    [sym_braced_variable] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1463),
    [anon_sym_LBRACE] = ACTIONS(1463),
    [aux_sym_if_statement_token1] = ACTIONS(1463),
    [aux_sym_switch_statement_token1] = ACTIONS(1463),
    [aux_sym_foreach_statement_token1] = ACTIONS(1463),
    [aux_sym_for_statement_token1] = ACTIONS(1463),
    [aux_sym_while_statement_token1] = ACTIONS(1463),
    [aux_sym_do_statement_token1] = ACTIONS(1463),
    [aux_sym_function_statement_token1] = ACTIONS(1463),
    [aux_sym_function_statement_token2] = ACTIONS(1463),
    [aux_sym_function_statement_token3] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1463),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1463),
    [sym_label] = ACTIONS(1463),
    [aux_sym_trap_statement_token1] = ACTIONS(1463),
    [aux_sym_try_statement_token1] = ACTIONS(1463),
    [aux_sym_data_statement_token1] = ACTIONS(1463),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1463),
    [aux_sym_parallel_statement_token1] = ACTIONS(1463),
    [aux_sym_sequence_statement_token1] = ACTIONS(1463),
    [anon_sym_AMP] = ACTIONS(1463),
    [aux_sym_command_name_token1] = ACTIONS(1463),
    [aux_sym_command_name_token2] = ACTIONS(1463),
    [aux_sym_command_name_token3] = ACTIONS(1463),
    [aux_sym_command_name_token4] = ACTIONS(1463),
    [aux_sym_command_name_token5] = ACTIONS(1463),
    [aux_sym_command_name_token6] = ACTIONS(1463),
    [aux_sym_command_name_token7] = ACTIONS(1463),
    [aux_sym_command_name_token8] = ACTIONS(1463),
    [aux_sym_command_name_token9] = ACTIONS(1463),
    [aux_sym_command_name_token10] = ACTIONS(1463),
    [aux_sym_command_name_token11] = ACTIONS(1463),
    [aux_sym_command_name_token12] = ACTIONS(1463),
    [aux_sym_command_name_token13] = ACTIONS(1463),
    [aux_sym_command_name_token14] = ACTIONS(1463),
    [aux_sym_command_name_token15] = ACTIONS(1463),
    [aux_sym_command_name_token16] = ACTIONS(1463),
    [aux_sym_command_name_token17] = ACTIONS(1463),
    [aux_sym_command_name_token18] = ACTIONS(1463),
    [aux_sym_class_statement_token1] = ACTIONS(1463),
    [aux_sym_enum_statement_token1] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1463),
    [anon_sym_BANG] = ACTIONS(1463),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1463),
    [anon_sym_DASH_DASH] = ACTIONS(1463),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1463),
    [anon_sym_AT_LPAREN] = ACTIONS(1463),
    [anon_sym_AT_LBRACE] = ACTIONS(1463),
  },
  [STATE(380)] = {
    [ts_builtin_sym_end] = ACTIONS(1481),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1431),
    [sym_hexadecimal_integer_literal] = ACTIONS(1431),
    [sym_real_literal] = ACTIONS(1431),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1431),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1431),
    [sym_verbatim_string_characters] = ACTIONS(1431),
    [sym_verbatim_here_string_characters] = ACTIONS(1431),
    [anon_sym_DOT] = ACTIONS(1431),
    [anon_sym_LBRACK] = ACTIONS(1431),
    [aux_sym_comparison_operator_token37] = ACTIONS(1431),
    [aux_sym_comparison_operator_token50] = ACTIONS(1431),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1431),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1431),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1431),
    [anon_sym_DOLLAR_] = ACTIONS(1431),
    [aux_sym_variable_token1] = ACTIONS(1431),
    [aux_sym_variable_token2] = ACTIONS(1431),
    [sym_braced_variable] = ACTIONS(1431),
    [anon_sym_SEMI] = ACTIONS(1431),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_LBRACE] = ACTIONS(1431),
    [aux_sym_if_statement_token1] = ACTIONS(1431),
    [aux_sym_switch_statement_token1] = ACTIONS(1431),
    [aux_sym_foreach_statement_token1] = ACTIONS(1431),
    [aux_sym_for_statement_token1] = ACTIONS(1431),
    [aux_sym_while_statement_token1] = ACTIONS(1431),
    [aux_sym_do_statement_token1] = ACTIONS(1431),
    [aux_sym_function_statement_token1] = ACTIONS(1431),
    [aux_sym_function_statement_token2] = ACTIONS(1431),
    [aux_sym_function_statement_token3] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1431),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1431),
    [sym_label] = ACTIONS(1431),
    [aux_sym_trap_statement_token1] = ACTIONS(1431),
    [aux_sym_try_statement_token1] = ACTIONS(1431),
    [aux_sym_data_statement_token1] = ACTIONS(1431),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1431),
    [aux_sym_parallel_statement_token1] = ACTIONS(1431),
    [aux_sym_sequence_statement_token1] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(1431),
    [aux_sym_command_name_token1] = ACTIONS(1431),
    [aux_sym_command_name_token2] = ACTIONS(1431),
    [aux_sym_command_name_token3] = ACTIONS(1431),
    [aux_sym_command_name_token4] = ACTIONS(1431),
    [aux_sym_command_name_token5] = ACTIONS(1431),
    [aux_sym_command_name_token6] = ACTIONS(1431),
    [aux_sym_command_name_token7] = ACTIONS(1431),
    [aux_sym_command_name_token8] = ACTIONS(1431),
    [aux_sym_command_name_token9] = ACTIONS(1431),
    [aux_sym_command_name_token10] = ACTIONS(1431),
    [aux_sym_command_name_token11] = ACTIONS(1431),
    [aux_sym_command_name_token12] = ACTIONS(1431),
    [aux_sym_command_name_token13] = ACTIONS(1431),
    [aux_sym_command_name_token14] = ACTIONS(1431),
    [aux_sym_command_name_token15] = ACTIONS(1431),
    [aux_sym_command_name_token16] = ACTIONS(1431),
    [aux_sym_command_name_token17] = ACTIONS(1431),
    [aux_sym_command_name_token18] = ACTIONS(1431),
    [aux_sym_class_statement_token1] = ACTIONS(1431),
    [aux_sym_enum_statement_token1] = ACTIONS(1431),
    [anon_sym_PLUS] = ACTIONS(1431),
    [anon_sym_DASH] = ACTIONS(1431),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1431),
    [anon_sym_BANG] = ACTIONS(1431),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1431),
    [anon_sym_PLUS_PLUS] = ACTIONS(1431),
    [anon_sym_DASH_DASH] = ACTIONS(1431),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1431),
    [anon_sym_AT_LPAREN] = ACTIONS(1431),
    [anon_sym_AT_LBRACE] = ACTIONS(1431),
  },
  [STATE(381)] = {
    [ts_builtin_sym_end] = ACTIONS(1483),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1339),
    [sym_hexadecimal_integer_literal] = ACTIONS(1339),
    [sym_real_literal] = ACTIONS(1339),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1339),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1339),
    [sym_verbatim_string_characters] = ACTIONS(1339),
    [sym_verbatim_here_string_characters] = ACTIONS(1339),
    [anon_sym_DOT] = ACTIONS(1339),
    [anon_sym_LBRACK] = ACTIONS(1339),
    [aux_sym_comparison_operator_token37] = ACTIONS(1339),
    [aux_sym_comparison_operator_token50] = ACTIONS(1339),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1339),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1339),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1339),
    [anon_sym_DOLLAR_] = ACTIONS(1339),
    [aux_sym_variable_token1] = ACTIONS(1339),
    [aux_sym_variable_token2] = ACTIONS(1339),
    [sym_braced_variable] = ACTIONS(1339),
    [anon_sym_SEMI] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(1339),
    [anon_sym_COMMA] = ACTIONS(1339),
    [anon_sym_LBRACE] = ACTIONS(1339),
    [aux_sym_if_statement_token1] = ACTIONS(1339),
    [aux_sym_switch_statement_token1] = ACTIONS(1339),
    [aux_sym_foreach_statement_token1] = ACTIONS(1339),
    [aux_sym_for_statement_token1] = ACTIONS(1339),
    [aux_sym_while_statement_token1] = ACTIONS(1339),
    [aux_sym_do_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token1] = ACTIONS(1339),
    [aux_sym_function_statement_token2] = ACTIONS(1339),
    [aux_sym_function_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1339),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1339),
    [sym_label] = ACTIONS(1339),
    [aux_sym_trap_statement_token1] = ACTIONS(1339),
    [aux_sym_try_statement_token1] = ACTIONS(1339),
    [aux_sym_data_statement_token1] = ACTIONS(1339),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1339),
    [aux_sym_parallel_statement_token1] = ACTIONS(1339),
    [aux_sym_sequence_statement_token1] = ACTIONS(1339),
    [anon_sym_AMP] = ACTIONS(1339),
    [aux_sym_command_name_token1] = ACTIONS(1339),
    [aux_sym_command_name_token2] = ACTIONS(1339),
    [aux_sym_command_name_token3] = ACTIONS(1339),
    [aux_sym_command_name_token4] = ACTIONS(1339),
    [aux_sym_command_name_token5] = ACTIONS(1339),
    [aux_sym_command_name_token6] = ACTIONS(1339),
    [aux_sym_command_name_token7] = ACTIONS(1339),
    [aux_sym_command_name_token8] = ACTIONS(1339),
    [aux_sym_command_name_token9] = ACTIONS(1339),
    [aux_sym_command_name_token10] = ACTIONS(1339),
    [aux_sym_command_name_token11] = ACTIONS(1339),
    [aux_sym_command_name_token12] = ACTIONS(1339),
    [aux_sym_command_name_token13] = ACTIONS(1339),
    [aux_sym_command_name_token14] = ACTIONS(1339),
    [aux_sym_command_name_token15] = ACTIONS(1339),
    [aux_sym_command_name_token16] = ACTIONS(1339),
    [aux_sym_command_name_token17] = ACTIONS(1339),
    [aux_sym_command_name_token18] = ACTIONS(1339),
    [aux_sym_class_statement_token1] = ACTIONS(1339),
    [aux_sym_enum_statement_token1] = ACTIONS(1339),
    [anon_sym_PLUS] = ACTIONS(1339),
    [anon_sym_DASH] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1339),
    [anon_sym_BANG] = ACTIONS(1339),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1339),
    [anon_sym_PLUS_PLUS] = ACTIONS(1339),
    [anon_sym_DASH_DASH] = ACTIONS(1339),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LPAREN] = ACTIONS(1339),
    [anon_sym_AT_LBRACE] = ACTIONS(1339),
  },
  [STATE(382)] = {
    [ts_builtin_sym_end] = ACTIONS(1485),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1467),
    [sym_hexadecimal_integer_literal] = ACTIONS(1467),
    [sym_real_literal] = ACTIONS(1467),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1467),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1467),
    [sym_verbatim_string_characters] = ACTIONS(1467),
    [sym_verbatim_here_string_characters] = ACTIONS(1467),
    [anon_sym_DOT] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1467),
    [aux_sym_comparison_operator_token37] = ACTIONS(1467),
    [aux_sym_comparison_operator_token50] = ACTIONS(1467),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1467),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1467),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1467),
    [anon_sym_DOLLAR_] = ACTIONS(1467),
    [aux_sym_variable_token1] = ACTIONS(1467),
    [aux_sym_variable_token2] = ACTIONS(1467),
    [sym_braced_variable] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1467),
    [anon_sym_LBRACE] = ACTIONS(1467),
    [aux_sym_if_statement_token1] = ACTIONS(1467),
    [aux_sym_switch_statement_token1] = ACTIONS(1467),
    [aux_sym_foreach_statement_token1] = ACTIONS(1467),
    [aux_sym_for_statement_token1] = ACTIONS(1467),
    [aux_sym_while_statement_token1] = ACTIONS(1467),
    [aux_sym_do_statement_token1] = ACTIONS(1467),
    [aux_sym_function_statement_token1] = ACTIONS(1467),
    [aux_sym_function_statement_token2] = ACTIONS(1467),
    [aux_sym_function_statement_token3] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1467),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1467),
    [sym_label] = ACTIONS(1467),
    [aux_sym_trap_statement_token1] = ACTIONS(1467),
    [aux_sym_try_statement_token1] = ACTIONS(1467),
    [aux_sym_data_statement_token1] = ACTIONS(1467),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1467),
    [aux_sym_parallel_statement_token1] = ACTIONS(1467),
    [aux_sym_sequence_statement_token1] = ACTIONS(1467),
    [anon_sym_AMP] = ACTIONS(1467),
    [aux_sym_command_name_token1] = ACTIONS(1467),
    [aux_sym_command_name_token2] = ACTIONS(1467),
    [aux_sym_command_name_token3] = ACTIONS(1467),
    [aux_sym_command_name_token4] = ACTIONS(1467),
    [aux_sym_command_name_token5] = ACTIONS(1467),
    [aux_sym_command_name_token6] = ACTIONS(1467),
    [aux_sym_command_name_token7] = ACTIONS(1467),
    [aux_sym_command_name_token8] = ACTIONS(1467),
    [aux_sym_command_name_token9] = ACTIONS(1467),
    [aux_sym_command_name_token10] = ACTIONS(1467),
    [aux_sym_command_name_token11] = ACTIONS(1467),
    [aux_sym_command_name_token12] = ACTIONS(1467),
    [aux_sym_command_name_token13] = ACTIONS(1467),
    [aux_sym_command_name_token14] = ACTIONS(1467),
    [aux_sym_command_name_token15] = ACTIONS(1467),
    [aux_sym_command_name_token16] = ACTIONS(1467),
    [aux_sym_command_name_token17] = ACTIONS(1467),
    [aux_sym_command_name_token18] = ACTIONS(1467),
    [aux_sym_class_statement_token1] = ACTIONS(1467),
    [aux_sym_enum_statement_token1] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1467),
    [anon_sym_BANG] = ACTIONS(1467),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1467),
    [anon_sym_DASH_DASH] = ACTIONS(1467),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1467),
    [anon_sym_AT_LPAREN] = ACTIONS(1467),
    [anon_sym_AT_LBRACE] = ACTIONS(1467),
  },
  [STATE(383)] = {
    [ts_builtin_sym_end] = ACTIONS(1487),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1471),
    [sym_hexadecimal_integer_literal] = ACTIONS(1471),
    [sym_real_literal] = ACTIONS(1471),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1471),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1471),
    [sym_verbatim_string_characters] = ACTIONS(1471),
    [sym_verbatim_here_string_characters] = ACTIONS(1471),
    [anon_sym_DOT] = ACTIONS(1471),
    [anon_sym_LBRACK] = ACTIONS(1471),
    [aux_sym_comparison_operator_token37] = ACTIONS(1471),
    [aux_sym_comparison_operator_token50] = ACTIONS(1471),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1471),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1471),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1471),
    [anon_sym_DOLLAR_] = ACTIONS(1471),
    [aux_sym_variable_token1] = ACTIONS(1471),
    [aux_sym_variable_token2] = ACTIONS(1471),
    [sym_braced_variable] = ACTIONS(1471),
    [anon_sym_SEMI] = ACTIONS(1471),
    [anon_sym_LPAREN] = ACTIONS(1471),
    [anon_sym_COMMA] = ACTIONS(1471),
    [anon_sym_LBRACE] = ACTIONS(1471),
    [aux_sym_if_statement_token1] = ACTIONS(1471),
    [aux_sym_switch_statement_token1] = ACTIONS(1471),
    [aux_sym_foreach_statement_token1] = ACTIONS(1471),
    [aux_sym_for_statement_token1] = ACTIONS(1471),
    [aux_sym_while_statement_token1] = ACTIONS(1471),
    [aux_sym_do_statement_token1] = ACTIONS(1471),
    [aux_sym_function_statement_token1] = ACTIONS(1471),
    [aux_sym_function_statement_token2] = ACTIONS(1471),
    [aux_sym_function_statement_token3] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1471),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1471),
    [sym_label] = ACTIONS(1471),
    [aux_sym_trap_statement_token1] = ACTIONS(1471),
    [aux_sym_try_statement_token1] = ACTIONS(1471),
    [aux_sym_data_statement_token1] = ACTIONS(1471),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1471),
    [aux_sym_parallel_statement_token1] = ACTIONS(1471),
    [aux_sym_sequence_statement_token1] = ACTIONS(1471),
    [anon_sym_AMP] = ACTIONS(1471),
    [aux_sym_command_name_token1] = ACTIONS(1471),
    [aux_sym_command_name_token2] = ACTIONS(1471),
    [aux_sym_command_name_token3] = ACTIONS(1471),
    [aux_sym_command_name_token4] = ACTIONS(1471),
    [aux_sym_command_name_token5] = ACTIONS(1471),
    [aux_sym_command_name_token6] = ACTIONS(1471),
    [aux_sym_command_name_token7] = ACTIONS(1471),
    [aux_sym_command_name_token8] = ACTIONS(1471),
    [aux_sym_command_name_token9] = ACTIONS(1471),
    [aux_sym_command_name_token10] = ACTIONS(1471),
    [aux_sym_command_name_token11] = ACTIONS(1471),
    [aux_sym_command_name_token12] = ACTIONS(1471),
    [aux_sym_command_name_token13] = ACTIONS(1471),
    [aux_sym_command_name_token14] = ACTIONS(1471),
    [aux_sym_command_name_token15] = ACTIONS(1471),
    [aux_sym_command_name_token16] = ACTIONS(1471),
    [aux_sym_command_name_token17] = ACTIONS(1471),
    [aux_sym_command_name_token18] = ACTIONS(1471),
    [aux_sym_class_statement_token1] = ACTIONS(1471),
    [aux_sym_enum_statement_token1] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1471),
    [anon_sym_DASH] = ACTIONS(1471),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1471),
    [anon_sym_BANG] = ACTIONS(1471),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1471),
    [anon_sym_PLUS_PLUS] = ACTIONS(1471),
    [anon_sym_DASH_DASH] = ACTIONS(1471),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1471),
    [anon_sym_AT_LPAREN] = ACTIONS(1471),
    [anon_sym_AT_LBRACE] = ACTIONS(1471),
  },
  [STATE(384)] = {
    [ts_builtin_sym_end] = ACTIONS(1489),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1473),
    [sym_hexadecimal_integer_literal] = ACTIONS(1473),
    [sym_real_literal] = ACTIONS(1473),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1473),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1473),
    [sym_verbatim_string_characters] = ACTIONS(1473),
    [sym_verbatim_here_string_characters] = ACTIONS(1473),
    [anon_sym_DOT] = ACTIONS(1473),
    [anon_sym_LBRACK] = ACTIONS(1473),
    [aux_sym_comparison_operator_token37] = ACTIONS(1473),
    [aux_sym_comparison_operator_token50] = ACTIONS(1473),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1473),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1473),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1473),
    [anon_sym_DOLLAR_] = ACTIONS(1473),
    [aux_sym_variable_token1] = ACTIONS(1473),
    [aux_sym_variable_token2] = ACTIONS(1473),
    [sym_braced_variable] = ACTIONS(1473),
    [anon_sym_SEMI] = ACTIONS(1473),
    [anon_sym_LPAREN] = ACTIONS(1473),
    [anon_sym_COMMA] = ACTIONS(1473),
    [anon_sym_LBRACE] = ACTIONS(1473),
    [aux_sym_if_statement_token1] = ACTIONS(1473),
    [aux_sym_switch_statement_token1] = ACTIONS(1473),
    [aux_sym_foreach_statement_token1] = ACTIONS(1473),
    [aux_sym_for_statement_token1] = ACTIONS(1473),
    [aux_sym_while_statement_token1] = ACTIONS(1473),
    [aux_sym_do_statement_token1] = ACTIONS(1473),
    [aux_sym_function_statement_token1] = ACTIONS(1473),
    [aux_sym_function_statement_token2] = ACTIONS(1473),
    [aux_sym_function_statement_token3] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1473),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1473),
    [sym_label] = ACTIONS(1473),
    [aux_sym_trap_statement_token1] = ACTIONS(1473),
    [aux_sym_try_statement_token1] = ACTIONS(1473),
    [aux_sym_data_statement_token1] = ACTIONS(1473),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1473),
    [aux_sym_parallel_statement_token1] = ACTIONS(1473),
    [aux_sym_sequence_statement_token1] = ACTIONS(1473),
    [anon_sym_AMP] = ACTIONS(1473),
    [aux_sym_command_name_token1] = ACTIONS(1473),
    [aux_sym_command_name_token2] = ACTIONS(1473),
    [aux_sym_command_name_token3] = ACTIONS(1473),
    [aux_sym_command_name_token4] = ACTIONS(1473),
    [aux_sym_command_name_token5] = ACTIONS(1473),
    [aux_sym_command_name_token6] = ACTIONS(1473),
    [aux_sym_command_name_token7] = ACTIONS(1473),
    [aux_sym_command_name_token8] = ACTIONS(1473),
    [aux_sym_command_name_token9] = ACTIONS(1473),
    [aux_sym_command_name_token10] = ACTIONS(1473),
    [aux_sym_command_name_token11] = ACTIONS(1473),
    [aux_sym_command_name_token12] = ACTIONS(1473),
    [aux_sym_command_name_token13] = ACTIONS(1473),
    [aux_sym_command_name_token14] = ACTIONS(1473),
    [aux_sym_command_name_token15] = ACTIONS(1473),
    [aux_sym_command_name_token16] = ACTIONS(1473),
    [aux_sym_command_name_token17] = ACTIONS(1473),
    [aux_sym_command_name_token18] = ACTIONS(1473),
    [aux_sym_class_statement_token1] = ACTIONS(1473),
    [aux_sym_enum_statement_token1] = ACTIONS(1473),
    [anon_sym_PLUS] = ACTIONS(1473),
    [anon_sym_DASH] = ACTIONS(1473),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1473),
    [anon_sym_BANG] = ACTIONS(1473),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1473),
    [anon_sym_PLUS_PLUS] = ACTIONS(1473),
    [anon_sym_DASH_DASH] = ACTIONS(1473),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1473),
    [anon_sym_AT_LPAREN] = ACTIONS(1473),
    [anon_sym_AT_LBRACE] = ACTIONS(1473),
  },
  [STATE(385)] = {
    [ts_builtin_sym_end] = ACTIONS(1491),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1429),
    [sym_hexadecimal_integer_literal] = ACTIONS(1429),
    [sym_real_literal] = ACTIONS(1429),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1429),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1429),
    [sym_verbatim_string_characters] = ACTIONS(1429),
    [sym_verbatim_here_string_characters] = ACTIONS(1429),
    [anon_sym_DOT] = ACTIONS(1429),
    [anon_sym_LBRACK] = ACTIONS(1429),
    [aux_sym_comparison_operator_token37] = ACTIONS(1429),
    [aux_sym_comparison_operator_token50] = ACTIONS(1429),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1429),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1429),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1429),
    [anon_sym_DOLLAR_] = ACTIONS(1429),
    [aux_sym_variable_token1] = ACTIONS(1429),
    [aux_sym_variable_token2] = ACTIONS(1429),
    [sym_braced_variable] = ACTIONS(1429),
    [anon_sym_SEMI] = ACTIONS(1429),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1429),
    [anon_sym_LBRACE] = ACTIONS(1429),
    [aux_sym_if_statement_token1] = ACTIONS(1429),
    [aux_sym_switch_statement_token1] = ACTIONS(1429),
    [aux_sym_foreach_statement_token1] = ACTIONS(1429),
    [aux_sym_for_statement_token1] = ACTIONS(1429),
    [aux_sym_while_statement_token1] = ACTIONS(1429),
    [aux_sym_do_statement_token1] = ACTIONS(1429),
    [aux_sym_function_statement_token1] = ACTIONS(1429),
    [aux_sym_function_statement_token2] = ACTIONS(1429),
    [aux_sym_function_statement_token3] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1429),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1429),
    [sym_label] = ACTIONS(1429),
    [aux_sym_trap_statement_token1] = ACTIONS(1429),
    [aux_sym_try_statement_token1] = ACTIONS(1429),
    [aux_sym_data_statement_token1] = ACTIONS(1429),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1429),
    [aux_sym_parallel_statement_token1] = ACTIONS(1429),
    [aux_sym_sequence_statement_token1] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(1429),
    [aux_sym_command_name_token1] = ACTIONS(1429),
    [aux_sym_command_name_token2] = ACTIONS(1429),
    [aux_sym_command_name_token3] = ACTIONS(1429),
    [aux_sym_command_name_token4] = ACTIONS(1429),
    [aux_sym_command_name_token5] = ACTIONS(1429),
    [aux_sym_command_name_token6] = ACTIONS(1429),
    [aux_sym_command_name_token7] = ACTIONS(1429),
    [aux_sym_command_name_token8] = ACTIONS(1429),
    [aux_sym_command_name_token9] = ACTIONS(1429),
    [aux_sym_command_name_token10] = ACTIONS(1429),
    [aux_sym_command_name_token11] = ACTIONS(1429),
    [aux_sym_command_name_token12] = ACTIONS(1429),
    [aux_sym_command_name_token13] = ACTIONS(1429),
    [aux_sym_command_name_token14] = ACTIONS(1429),
    [aux_sym_command_name_token15] = ACTIONS(1429),
    [aux_sym_command_name_token16] = ACTIONS(1429),
    [aux_sym_command_name_token17] = ACTIONS(1429),
    [aux_sym_command_name_token18] = ACTIONS(1429),
    [aux_sym_class_statement_token1] = ACTIONS(1429),
    [aux_sym_enum_statement_token1] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1429),
    [anon_sym_BANG] = ACTIONS(1429),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1429),
    [anon_sym_PLUS_PLUS] = ACTIONS(1429),
    [anon_sym_DASH_DASH] = ACTIONS(1429),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1429),
    [anon_sym_AT_LPAREN] = ACTIONS(1429),
    [anon_sym_AT_LBRACE] = ACTIONS(1429),
  },
  [STATE(386)] = {
    [ts_builtin_sym_end] = ACTIONS(1493),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1417),
    [sym_hexadecimal_integer_literal] = ACTIONS(1417),
    [sym_real_literal] = ACTIONS(1417),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1417),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1417),
    [sym_verbatim_string_characters] = ACTIONS(1417),
    [sym_verbatim_here_string_characters] = ACTIONS(1417),
    [anon_sym_DOT] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1417),
    [aux_sym_comparison_operator_token37] = ACTIONS(1417),
    [aux_sym_comparison_operator_token50] = ACTIONS(1417),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1417),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1417),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1417),
    [anon_sym_DOLLAR_] = ACTIONS(1417),
    [aux_sym_variable_token1] = ACTIONS(1417),
    [aux_sym_variable_token2] = ACTIONS(1417),
    [sym_braced_variable] = ACTIONS(1417),
    [anon_sym_SEMI] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1417),
    [anon_sym_COMMA] = ACTIONS(1417),
    [anon_sym_LBRACE] = ACTIONS(1417),
    [aux_sym_if_statement_token1] = ACTIONS(1417),
    [aux_sym_switch_statement_token1] = ACTIONS(1417),
    [aux_sym_foreach_statement_token1] = ACTIONS(1417),
    [aux_sym_for_statement_token1] = ACTIONS(1417),
    [aux_sym_while_statement_token1] = ACTIONS(1417),
    [aux_sym_do_statement_token1] = ACTIONS(1417),
    [aux_sym_function_statement_token1] = ACTIONS(1417),
    [aux_sym_function_statement_token2] = ACTIONS(1417),
    [aux_sym_function_statement_token3] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1417),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1417),
    [sym_label] = ACTIONS(1417),
    [aux_sym_trap_statement_token1] = ACTIONS(1417),
    [aux_sym_try_statement_token1] = ACTIONS(1417),
    [aux_sym_data_statement_token1] = ACTIONS(1417),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1417),
    [aux_sym_parallel_statement_token1] = ACTIONS(1417),
    [aux_sym_sequence_statement_token1] = ACTIONS(1417),
    [anon_sym_AMP] = ACTIONS(1417),
    [aux_sym_command_name_token1] = ACTIONS(1417),
    [aux_sym_command_name_token2] = ACTIONS(1417),
    [aux_sym_command_name_token3] = ACTIONS(1417),
    [aux_sym_command_name_token4] = ACTIONS(1417),
    [aux_sym_command_name_token5] = ACTIONS(1417),
    [aux_sym_command_name_token6] = ACTIONS(1417),
    [aux_sym_command_name_token7] = ACTIONS(1417),
    [aux_sym_command_name_token8] = ACTIONS(1417),
    [aux_sym_command_name_token9] = ACTIONS(1417),
    [aux_sym_command_name_token10] = ACTIONS(1417),
    [aux_sym_command_name_token11] = ACTIONS(1417),
    [aux_sym_command_name_token12] = ACTIONS(1417),
    [aux_sym_command_name_token13] = ACTIONS(1417),
    [aux_sym_command_name_token14] = ACTIONS(1417),
    [aux_sym_command_name_token15] = ACTIONS(1417),
    [aux_sym_command_name_token16] = ACTIONS(1417),
    [aux_sym_command_name_token17] = ACTIONS(1417),
    [aux_sym_command_name_token18] = ACTIONS(1417),
    [aux_sym_class_statement_token1] = ACTIONS(1417),
    [aux_sym_enum_statement_token1] = ACTIONS(1417),
    [anon_sym_PLUS] = ACTIONS(1417),
    [anon_sym_DASH] = ACTIONS(1417),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1417),
    [anon_sym_BANG] = ACTIONS(1417),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1417),
    [anon_sym_PLUS_PLUS] = ACTIONS(1417),
    [anon_sym_DASH_DASH] = ACTIONS(1417),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1417),
    [anon_sym_AT_LPAREN] = ACTIONS(1417),
    [anon_sym_AT_LBRACE] = ACTIONS(1417),
  },
  [STATE(387)] = {
    [ts_builtin_sym_end] = ACTIONS(1495),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1449),
    [sym_hexadecimal_integer_literal] = ACTIONS(1449),
    [sym_real_literal] = ACTIONS(1449),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1449),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1449),
    [sym_verbatim_string_characters] = ACTIONS(1449),
    [sym_verbatim_here_string_characters] = ACTIONS(1449),
    [anon_sym_DOT] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1449),
    [aux_sym_comparison_operator_token37] = ACTIONS(1449),
    [aux_sym_comparison_operator_token50] = ACTIONS(1449),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1449),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1449),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1449),
    [anon_sym_DOLLAR_] = ACTIONS(1449),
    [aux_sym_variable_token1] = ACTIONS(1449),
    [aux_sym_variable_token2] = ACTIONS(1449),
    [sym_braced_variable] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1449),
    [anon_sym_LBRACE] = ACTIONS(1449),
    [aux_sym_if_statement_token1] = ACTIONS(1449),
    [aux_sym_switch_statement_token1] = ACTIONS(1449),
    [aux_sym_foreach_statement_token1] = ACTIONS(1449),
    [aux_sym_for_statement_token1] = ACTIONS(1449),
    [aux_sym_while_statement_token1] = ACTIONS(1449),
    [aux_sym_do_statement_token1] = ACTIONS(1449),
    [aux_sym_function_statement_token1] = ACTIONS(1449),
    [aux_sym_function_statement_token2] = ACTIONS(1449),
    [aux_sym_function_statement_token3] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1449),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1449),
    [sym_label] = ACTIONS(1449),
    [aux_sym_trap_statement_token1] = ACTIONS(1449),
    [aux_sym_try_statement_token1] = ACTIONS(1449),
    [aux_sym_data_statement_token1] = ACTIONS(1449),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1449),
    [aux_sym_parallel_statement_token1] = ACTIONS(1449),
    [aux_sym_sequence_statement_token1] = ACTIONS(1449),
    [anon_sym_AMP] = ACTIONS(1449),
    [aux_sym_command_name_token1] = ACTIONS(1449),
    [aux_sym_command_name_token2] = ACTIONS(1449),
    [aux_sym_command_name_token3] = ACTIONS(1449),
    [aux_sym_command_name_token4] = ACTIONS(1449),
    [aux_sym_command_name_token5] = ACTIONS(1449),
    [aux_sym_command_name_token6] = ACTIONS(1449),
    [aux_sym_command_name_token7] = ACTIONS(1449),
    [aux_sym_command_name_token8] = ACTIONS(1449),
    [aux_sym_command_name_token9] = ACTIONS(1449),
    [aux_sym_command_name_token10] = ACTIONS(1449),
    [aux_sym_command_name_token11] = ACTIONS(1449),
    [aux_sym_command_name_token12] = ACTIONS(1449),
    [aux_sym_command_name_token13] = ACTIONS(1449),
    [aux_sym_command_name_token14] = ACTIONS(1449),
    [aux_sym_command_name_token15] = ACTIONS(1449),
    [aux_sym_command_name_token16] = ACTIONS(1449),
    [aux_sym_command_name_token17] = ACTIONS(1449),
    [aux_sym_command_name_token18] = ACTIONS(1449),
    [aux_sym_class_statement_token1] = ACTIONS(1449),
    [aux_sym_enum_statement_token1] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1449),
    [anon_sym_BANG] = ACTIONS(1449),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1449),
    [anon_sym_DASH_DASH] = ACTIONS(1449),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1449),
    [anon_sym_AT_LPAREN] = ACTIONS(1449),
    [anon_sym_AT_LBRACE] = ACTIONS(1449),
  },
  [STATE(388)] = {
    [ts_builtin_sym_end] = ACTIONS(1497),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1403),
    [sym_hexadecimal_integer_literal] = ACTIONS(1403),
    [sym_real_literal] = ACTIONS(1403),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1403),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1403),
    [sym_verbatim_string_characters] = ACTIONS(1403),
    [sym_verbatim_here_string_characters] = ACTIONS(1403),
    [anon_sym_DOT] = ACTIONS(1403),
    [anon_sym_LBRACK] = ACTIONS(1403),
    [aux_sym_comparison_operator_token37] = ACTIONS(1403),
    [aux_sym_comparison_operator_token50] = ACTIONS(1403),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1403),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1403),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1403),
    [anon_sym_DOLLAR_] = ACTIONS(1403),
    [aux_sym_variable_token1] = ACTIONS(1403),
    [aux_sym_variable_token2] = ACTIONS(1403),
    [sym_braced_variable] = ACTIONS(1403),
    [anon_sym_SEMI] = ACTIONS(1403),
    [anon_sym_LPAREN] = ACTIONS(1403),
    [anon_sym_COMMA] = ACTIONS(1403),
    [anon_sym_LBRACE] = ACTIONS(1403),
    [aux_sym_if_statement_token1] = ACTIONS(1403),
    [aux_sym_switch_statement_token1] = ACTIONS(1403),
    [aux_sym_foreach_statement_token1] = ACTIONS(1403),
    [aux_sym_for_statement_token1] = ACTIONS(1403),
    [aux_sym_while_statement_token1] = ACTIONS(1403),
    [aux_sym_do_statement_token1] = ACTIONS(1403),
    [aux_sym_function_statement_token1] = ACTIONS(1403),
    [aux_sym_function_statement_token2] = ACTIONS(1403),
    [aux_sym_function_statement_token3] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1403),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1403),
    [sym_label] = ACTIONS(1403),
    [aux_sym_trap_statement_token1] = ACTIONS(1403),
    [aux_sym_try_statement_token1] = ACTIONS(1403),
    [aux_sym_data_statement_token1] = ACTIONS(1403),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1403),
    [aux_sym_parallel_statement_token1] = ACTIONS(1403),
    [aux_sym_sequence_statement_token1] = ACTIONS(1403),
    [anon_sym_AMP] = ACTIONS(1403),
    [aux_sym_command_name_token1] = ACTIONS(1403),
    [aux_sym_command_name_token2] = ACTIONS(1403),
    [aux_sym_command_name_token3] = ACTIONS(1403),
    [aux_sym_command_name_token4] = ACTIONS(1403),
    [aux_sym_command_name_token5] = ACTIONS(1403),
    [aux_sym_command_name_token6] = ACTIONS(1403),
    [aux_sym_command_name_token7] = ACTIONS(1403),
    [aux_sym_command_name_token8] = ACTIONS(1403),
    [aux_sym_command_name_token9] = ACTIONS(1403),
    [aux_sym_command_name_token10] = ACTIONS(1403),
    [aux_sym_command_name_token11] = ACTIONS(1403),
    [aux_sym_command_name_token12] = ACTIONS(1403),
    [aux_sym_command_name_token13] = ACTIONS(1403),
    [aux_sym_command_name_token14] = ACTIONS(1403),
    [aux_sym_command_name_token15] = ACTIONS(1403),
    [aux_sym_command_name_token16] = ACTIONS(1403),
    [aux_sym_command_name_token17] = ACTIONS(1403),
    [aux_sym_command_name_token18] = ACTIONS(1403),
    [aux_sym_class_statement_token1] = ACTIONS(1403),
    [aux_sym_enum_statement_token1] = ACTIONS(1403),
    [anon_sym_PLUS] = ACTIONS(1403),
    [anon_sym_DASH] = ACTIONS(1403),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1403),
    [anon_sym_BANG] = ACTIONS(1403),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1403),
    [anon_sym_PLUS_PLUS] = ACTIONS(1403),
    [anon_sym_DASH_DASH] = ACTIONS(1403),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1403),
    [anon_sym_AT_LPAREN] = ACTIONS(1403),
    [anon_sym_AT_LBRACE] = ACTIONS(1403),
  },
  [STATE(389)] = {
    [ts_builtin_sym_end] = ACTIONS(1499),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1419),
    [sym_hexadecimal_integer_literal] = ACTIONS(1419),
    [sym_real_literal] = ACTIONS(1419),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1419),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1419),
    [sym_verbatim_string_characters] = ACTIONS(1419),
    [sym_verbatim_here_string_characters] = ACTIONS(1419),
    [anon_sym_DOT] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [aux_sym_comparison_operator_token37] = ACTIONS(1419),
    [aux_sym_comparison_operator_token50] = ACTIONS(1419),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1419),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1419),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1419),
    [anon_sym_DOLLAR_] = ACTIONS(1419),
    [aux_sym_variable_token1] = ACTIONS(1419),
    [aux_sym_variable_token2] = ACTIONS(1419),
    [sym_braced_variable] = ACTIONS(1419),
    [anon_sym_SEMI] = ACTIONS(1419),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_COMMA] = ACTIONS(1419),
    [anon_sym_LBRACE] = ACTIONS(1419),
    [aux_sym_if_statement_token1] = ACTIONS(1419),
    [aux_sym_switch_statement_token1] = ACTIONS(1419),
    [aux_sym_foreach_statement_token1] = ACTIONS(1419),
    [aux_sym_for_statement_token1] = ACTIONS(1419),
    [aux_sym_while_statement_token1] = ACTIONS(1419),
    [aux_sym_do_statement_token1] = ACTIONS(1419),
    [aux_sym_function_statement_token1] = ACTIONS(1419),
    [aux_sym_function_statement_token2] = ACTIONS(1419),
    [aux_sym_function_statement_token3] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1419),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1419),
    [sym_label] = ACTIONS(1419),
    [aux_sym_trap_statement_token1] = ACTIONS(1419),
    [aux_sym_try_statement_token1] = ACTIONS(1419),
    [aux_sym_data_statement_token1] = ACTIONS(1419),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1419),
    [aux_sym_parallel_statement_token1] = ACTIONS(1419),
    [aux_sym_sequence_statement_token1] = ACTIONS(1419),
    [anon_sym_AMP] = ACTIONS(1419),
    [aux_sym_command_name_token1] = ACTIONS(1419),
    [aux_sym_command_name_token2] = ACTIONS(1419),
    [aux_sym_command_name_token3] = ACTIONS(1419),
    [aux_sym_command_name_token4] = ACTIONS(1419),
    [aux_sym_command_name_token5] = ACTIONS(1419),
    [aux_sym_command_name_token6] = ACTIONS(1419),
    [aux_sym_command_name_token7] = ACTIONS(1419),
    [aux_sym_command_name_token8] = ACTIONS(1419),
    [aux_sym_command_name_token9] = ACTIONS(1419),
    [aux_sym_command_name_token10] = ACTIONS(1419),
    [aux_sym_command_name_token11] = ACTIONS(1419),
    [aux_sym_command_name_token12] = ACTIONS(1419),
    [aux_sym_command_name_token13] = ACTIONS(1419),
    [aux_sym_command_name_token14] = ACTIONS(1419),
    [aux_sym_command_name_token15] = ACTIONS(1419),
    [aux_sym_command_name_token16] = ACTIONS(1419),
    [aux_sym_command_name_token17] = ACTIONS(1419),
    [aux_sym_command_name_token18] = ACTIONS(1419),
    [aux_sym_class_statement_token1] = ACTIONS(1419),
    [aux_sym_enum_statement_token1] = ACTIONS(1419),
    [anon_sym_PLUS] = ACTIONS(1419),
    [anon_sym_DASH] = ACTIONS(1419),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1419),
    [anon_sym_BANG] = ACTIONS(1419),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1419),
    [anon_sym_PLUS_PLUS] = ACTIONS(1419),
    [anon_sym_DASH_DASH] = ACTIONS(1419),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1419),
    [anon_sym_AT_LPAREN] = ACTIONS(1419),
    [anon_sym_AT_LBRACE] = ACTIONS(1419),
  },
  [STATE(390)] = {
    [ts_builtin_sym_end] = ACTIONS(1501),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1421),
    [sym_hexadecimal_integer_literal] = ACTIONS(1421),
    [sym_real_literal] = ACTIONS(1421),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1421),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1421),
    [sym_verbatim_string_characters] = ACTIONS(1421),
    [sym_verbatim_here_string_characters] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1421),
    [aux_sym_comparison_operator_token37] = ACTIONS(1421),
    [aux_sym_comparison_operator_token50] = ACTIONS(1421),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1421),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1421),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1421),
    [anon_sym_DOLLAR_] = ACTIONS(1421),
    [aux_sym_variable_token1] = ACTIONS(1421),
    [aux_sym_variable_token2] = ACTIONS(1421),
    [sym_braced_variable] = ACTIONS(1421),
    [anon_sym_SEMI] = ACTIONS(1421),
    [anon_sym_LPAREN] = ACTIONS(1421),
    [anon_sym_COMMA] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1421),
    [aux_sym_if_statement_token1] = ACTIONS(1421),
    [aux_sym_switch_statement_token1] = ACTIONS(1421),
    [aux_sym_foreach_statement_token1] = ACTIONS(1421),
    [aux_sym_for_statement_token1] = ACTIONS(1421),
    [aux_sym_while_statement_token1] = ACTIONS(1421),
    [aux_sym_do_statement_token1] = ACTIONS(1421),
    [aux_sym_function_statement_token1] = ACTIONS(1421),
    [aux_sym_function_statement_token2] = ACTIONS(1421),
    [aux_sym_function_statement_token3] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1421),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1421),
    [sym_label] = ACTIONS(1421),
    [aux_sym_trap_statement_token1] = ACTIONS(1421),
    [aux_sym_try_statement_token1] = ACTIONS(1421),
    [aux_sym_data_statement_token1] = ACTIONS(1421),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1421),
    [aux_sym_parallel_statement_token1] = ACTIONS(1421),
    [aux_sym_sequence_statement_token1] = ACTIONS(1421),
    [anon_sym_AMP] = ACTIONS(1421),
    [aux_sym_command_name_token1] = ACTIONS(1421),
    [aux_sym_command_name_token2] = ACTIONS(1421),
    [aux_sym_command_name_token3] = ACTIONS(1421),
    [aux_sym_command_name_token4] = ACTIONS(1421),
    [aux_sym_command_name_token5] = ACTIONS(1421),
    [aux_sym_command_name_token6] = ACTIONS(1421),
    [aux_sym_command_name_token7] = ACTIONS(1421),
    [aux_sym_command_name_token8] = ACTIONS(1421),
    [aux_sym_command_name_token9] = ACTIONS(1421),
    [aux_sym_command_name_token10] = ACTIONS(1421),
    [aux_sym_command_name_token11] = ACTIONS(1421),
    [aux_sym_command_name_token12] = ACTIONS(1421),
    [aux_sym_command_name_token13] = ACTIONS(1421),
    [aux_sym_command_name_token14] = ACTIONS(1421),
    [aux_sym_command_name_token15] = ACTIONS(1421),
    [aux_sym_command_name_token16] = ACTIONS(1421),
    [aux_sym_command_name_token17] = ACTIONS(1421),
    [aux_sym_command_name_token18] = ACTIONS(1421),
    [aux_sym_class_statement_token1] = ACTIONS(1421),
    [aux_sym_enum_statement_token1] = ACTIONS(1421),
    [anon_sym_PLUS] = ACTIONS(1421),
    [anon_sym_DASH] = ACTIONS(1421),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1421),
    [anon_sym_BANG] = ACTIONS(1421),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1421),
    [anon_sym_PLUS_PLUS] = ACTIONS(1421),
    [anon_sym_DASH_DASH] = ACTIONS(1421),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1421),
    [anon_sym_AT_LPAREN] = ACTIONS(1421),
    [anon_sym_AT_LBRACE] = ACTIONS(1421),
  },
  [STATE(391)] = {
    [ts_builtin_sym_end] = ACTIONS(1503),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1425),
    [sym_hexadecimal_integer_literal] = ACTIONS(1425),
    [sym_real_literal] = ACTIONS(1425),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1425),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1425),
    [sym_verbatim_string_characters] = ACTIONS(1425),
    [sym_verbatim_here_string_characters] = ACTIONS(1425),
    [anon_sym_DOT] = ACTIONS(1425),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [aux_sym_comparison_operator_token37] = ACTIONS(1425),
    [aux_sym_comparison_operator_token50] = ACTIONS(1425),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1425),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1425),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1425),
    [anon_sym_DOLLAR_] = ACTIONS(1425),
    [aux_sym_variable_token1] = ACTIONS(1425),
    [aux_sym_variable_token2] = ACTIONS(1425),
    [sym_braced_variable] = ACTIONS(1425),
    [anon_sym_SEMI] = ACTIONS(1425),
    [anon_sym_LPAREN] = ACTIONS(1425),
    [anon_sym_COMMA] = ACTIONS(1425),
    [anon_sym_LBRACE] = ACTIONS(1425),
    [aux_sym_if_statement_token1] = ACTIONS(1425),
    [aux_sym_switch_statement_token1] = ACTIONS(1425),
    [aux_sym_foreach_statement_token1] = ACTIONS(1425),
    [aux_sym_for_statement_token1] = ACTIONS(1425),
    [aux_sym_while_statement_token1] = ACTIONS(1425),
    [aux_sym_do_statement_token1] = ACTIONS(1425),
    [aux_sym_function_statement_token1] = ACTIONS(1425),
    [aux_sym_function_statement_token2] = ACTIONS(1425),
    [aux_sym_function_statement_token3] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1425),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1425),
    [sym_label] = ACTIONS(1425),
    [aux_sym_trap_statement_token1] = ACTIONS(1425),
    [aux_sym_try_statement_token1] = ACTIONS(1425),
    [aux_sym_data_statement_token1] = ACTIONS(1425),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1425),
    [aux_sym_parallel_statement_token1] = ACTIONS(1425),
    [aux_sym_sequence_statement_token1] = ACTIONS(1425),
    [anon_sym_AMP] = ACTIONS(1425),
    [aux_sym_command_name_token1] = ACTIONS(1425),
    [aux_sym_command_name_token2] = ACTIONS(1425),
    [aux_sym_command_name_token3] = ACTIONS(1425),
    [aux_sym_command_name_token4] = ACTIONS(1425),
    [aux_sym_command_name_token5] = ACTIONS(1425),
    [aux_sym_command_name_token6] = ACTIONS(1425),
    [aux_sym_command_name_token7] = ACTIONS(1425),
    [aux_sym_command_name_token8] = ACTIONS(1425),
    [aux_sym_command_name_token9] = ACTIONS(1425),
    [aux_sym_command_name_token10] = ACTIONS(1425),
    [aux_sym_command_name_token11] = ACTIONS(1425),
    [aux_sym_command_name_token12] = ACTIONS(1425),
    [aux_sym_command_name_token13] = ACTIONS(1425),
    [aux_sym_command_name_token14] = ACTIONS(1425),
    [aux_sym_command_name_token15] = ACTIONS(1425),
    [aux_sym_command_name_token16] = ACTIONS(1425),
    [aux_sym_command_name_token17] = ACTIONS(1425),
    [aux_sym_command_name_token18] = ACTIONS(1425),
    [aux_sym_class_statement_token1] = ACTIONS(1425),
    [aux_sym_enum_statement_token1] = ACTIONS(1425),
    [anon_sym_PLUS] = ACTIONS(1425),
    [anon_sym_DASH] = ACTIONS(1425),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1425),
    [anon_sym_BANG] = ACTIONS(1425),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1425),
    [anon_sym_PLUS_PLUS] = ACTIONS(1425),
    [anon_sym_DASH_DASH] = ACTIONS(1425),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1425),
    [anon_sym_AT_LPAREN] = ACTIONS(1425),
    [anon_sym_AT_LBRACE] = ACTIONS(1425),
  },
  [STATE(392)] = {
    [ts_builtin_sym_end] = ACTIONS(1505),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1427),
    [sym_hexadecimal_integer_literal] = ACTIONS(1427),
    [sym_real_literal] = ACTIONS(1427),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1427),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1427),
    [sym_verbatim_string_characters] = ACTIONS(1427),
    [sym_verbatim_here_string_characters] = ACTIONS(1427),
    [anon_sym_DOT] = ACTIONS(1427),
    [anon_sym_LBRACK] = ACTIONS(1427),
    [aux_sym_comparison_operator_token37] = ACTIONS(1427),
    [aux_sym_comparison_operator_token50] = ACTIONS(1427),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1427),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1427),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1427),
    [anon_sym_DOLLAR_] = ACTIONS(1427),
    [aux_sym_variable_token1] = ACTIONS(1427),
    [aux_sym_variable_token2] = ACTIONS(1427),
    [sym_braced_variable] = ACTIONS(1427),
    [anon_sym_SEMI] = ACTIONS(1427),
    [anon_sym_LPAREN] = ACTIONS(1427),
    [anon_sym_COMMA] = ACTIONS(1427),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [aux_sym_if_statement_token1] = ACTIONS(1427),
    [aux_sym_switch_statement_token1] = ACTIONS(1427),
    [aux_sym_foreach_statement_token1] = ACTIONS(1427),
    [aux_sym_for_statement_token1] = ACTIONS(1427),
    [aux_sym_while_statement_token1] = ACTIONS(1427),
    [aux_sym_do_statement_token1] = ACTIONS(1427),
    [aux_sym_function_statement_token1] = ACTIONS(1427),
    [aux_sym_function_statement_token2] = ACTIONS(1427),
    [aux_sym_function_statement_token3] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1427),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1427),
    [sym_label] = ACTIONS(1427),
    [aux_sym_trap_statement_token1] = ACTIONS(1427),
    [aux_sym_try_statement_token1] = ACTIONS(1427),
    [aux_sym_data_statement_token1] = ACTIONS(1427),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1427),
    [aux_sym_parallel_statement_token1] = ACTIONS(1427),
    [aux_sym_sequence_statement_token1] = ACTIONS(1427),
    [anon_sym_AMP] = ACTIONS(1427),
    [aux_sym_command_name_token1] = ACTIONS(1427),
    [aux_sym_command_name_token2] = ACTIONS(1427),
    [aux_sym_command_name_token3] = ACTIONS(1427),
    [aux_sym_command_name_token4] = ACTIONS(1427),
    [aux_sym_command_name_token5] = ACTIONS(1427),
    [aux_sym_command_name_token6] = ACTIONS(1427),
    [aux_sym_command_name_token7] = ACTIONS(1427),
    [aux_sym_command_name_token8] = ACTIONS(1427),
    [aux_sym_command_name_token9] = ACTIONS(1427),
    [aux_sym_command_name_token10] = ACTIONS(1427),
    [aux_sym_command_name_token11] = ACTIONS(1427),
    [aux_sym_command_name_token12] = ACTIONS(1427),
    [aux_sym_command_name_token13] = ACTIONS(1427),
    [aux_sym_command_name_token14] = ACTIONS(1427),
    [aux_sym_command_name_token15] = ACTIONS(1427),
    [aux_sym_command_name_token16] = ACTIONS(1427),
    [aux_sym_command_name_token17] = ACTIONS(1427),
    [aux_sym_command_name_token18] = ACTIONS(1427),
    [aux_sym_class_statement_token1] = ACTIONS(1427),
    [aux_sym_enum_statement_token1] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1427),
    [anon_sym_DASH] = ACTIONS(1427),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1427),
    [anon_sym_BANG] = ACTIONS(1427),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1427),
    [anon_sym_PLUS_PLUS] = ACTIONS(1427),
    [anon_sym_DASH_DASH] = ACTIONS(1427),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1427),
    [anon_sym_AT_LPAREN] = ACTIONS(1427),
    [anon_sym_AT_LBRACE] = ACTIONS(1427),
  },
  [STATE(393)] = {
    [ts_builtin_sym_end] = ACTIONS(1507),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1475),
    [sym_hexadecimal_integer_literal] = ACTIONS(1475),
    [sym_real_literal] = ACTIONS(1475),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1475),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1475),
    [sym_verbatim_string_characters] = ACTIONS(1475),
    [sym_verbatim_here_string_characters] = ACTIONS(1475),
    [anon_sym_DOT] = ACTIONS(1475),
    [anon_sym_LBRACK] = ACTIONS(1475),
    [aux_sym_comparison_operator_token37] = ACTIONS(1475),
    [aux_sym_comparison_operator_token50] = ACTIONS(1475),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1475),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1475),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1475),
    [anon_sym_DOLLAR_] = ACTIONS(1475),
    [aux_sym_variable_token1] = ACTIONS(1475),
    [aux_sym_variable_token2] = ACTIONS(1475),
    [sym_braced_variable] = ACTIONS(1475),
    [anon_sym_SEMI] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1475),
    [anon_sym_COMMA] = ACTIONS(1475),
    [anon_sym_LBRACE] = ACTIONS(1475),
    [aux_sym_if_statement_token1] = ACTIONS(1475),
    [aux_sym_switch_statement_token1] = ACTIONS(1475),
    [aux_sym_foreach_statement_token1] = ACTIONS(1475),
    [aux_sym_for_statement_token1] = ACTIONS(1475),
    [aux_sym_while_statement_token1] = ACTIONS(1475),
    [aux_sym_do_statement_token1] = ACTIONS(1475),
    [aux_sym_function_statement_token1] = ACTIONS(1475),
    [aux_sym_function_statement_token2] = ACTIONS(1475),
    [aux_sym_function_statement_token3] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1475),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1475),
    [sym_label] = ACTIONS(1475),
    [aux_sym_trap_statement_token1] = ACTIONS(1475),
    [aux_sym_try_statement_token1] = ACTIONS(1475),
    [aux_sym_data_statement_token1] = ACTIONS(1475),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1475),
    [aux_sym_parallel_statement_token1] = ACTIONS(1475),
    [aux_sym_sequence_statement_token1] = ACTIONS(1475),
    [anon_sym_AMP] = ACTIONS(1475),
    [aux_sym_command_name_token1] = ACTIONS(1475),
    [aux_sym_command_name_token2] = ACTIONS(1475),
    [aux_sym_command_name_token3] = ACTIONS(1475),
    [aux_sym_command_name_token4] = ACTIONS(1475),
    [aux_sym_command_name_token5] = ACTIONS(1475),
    [aux_sym_command_name_token6] = ACTIONS(1475),
    [aux_sym_command_name_token7] = ACTIONS(1475),
    [aux_sym_command_name_token8] = ACTIONS(1475),
    [aux_sym_command_name_token9] = ACTIONS(1475),
    [aux_sym_command_name_token10] = ACTIONS(1475),
    [aux_sym_command_name_token11] = ACTIONS(1475),
    [aux_sym_command_name_token12] = ACTIONS(1475),
    [aux_sym_command_name_token13] = ACTIONS(1475),
    [aux_sym_command_name_token14] = ACTIONS(1475),
    [aux_sym_command_name_token15] = ACTIONS(1475),
    [aux_sym_command_name_token16] = ACTIONS(1475),
    [aux_sym_command_name_token17] = ACTIONS(1475),
    [aux_sym_command_name_token18] = ACTIONS(1475),
    [aux_sym_class_statement_token1] = ACTIONS(1475),
    [aux_sym_enum_statement_token1] = ACTIONS(1475),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1475),
    [anon_sym_BANG] = ACTIONS(1475),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1475),
    [anon_sym_PLUS_PLUS] = ACTIONS(1475),
    [anon_sym_DASH_DASH] = ACTIONS(1475),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1475),
    [anon_sym_AT_LPAREN] = ACTIONS(1475),
    [anon_sym_AT_LBRACE] = ACTIONS(1475),
  },
  [STATE(394)] = {
    [ts_builtin_sym_end] = ACTIONS(1509),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1439),
    [sym_hexadecimal_integer_literal] = ACTIONS(1439),
    [sym_real_literal] = ACTIONS(1439),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1439),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1439),
    [sym_verbatim_string_characters] = ACTIONS(1439),
    [sym_verbatim_here_string_characters] = ACTIONS(1439),
    [anon_sym_DOT] = ACTIONS(1439),
    [anon_sym_LBRACK] = ACTIONS(1439),
    [aux_sym_comparison_operator_token37] = ACTIONS(1439),
    [aux_sym_comparison_operator_token50] = ACTIONS(1439),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1439),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1439),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1439),
    [anon_sym_DOLLAR_] = ACTIONS(1439),
    [aux_sym_variable_token1] = ACTIONS(1439),
    [aux_sym_variable_token2] = ACTIONS(1439),
    [sym_braced_variable] = ACTIONS(1439),
    [anon_sym_SEMI] = ACTIONS(1439),
    [anon_sym_LPAREN] = ACTIONS(1439),
    [anon_sym_COMMA] = ACTIONS(1439),
    [anon_sym_LBRACE] = ACTIONS(1439),
    [aux_sym_if_statement_token1] = ACTIONS(1439),
    [aux_sym_switch_statement_token1] = ACTIONS(1439),
    [aux_sym_foreach_statement_token1] = ACTIONS(1439),
    [aux_sym_for_statement_token1] = ACTIONS(1439),
    [aux_sym_while_statement_token1] = ACTIONS(1439),
    [aux_sym_do_statement_token1] = ACTIONS(1439),
    [aux_sym_function_statement_token1] = ACTIONS(1439),
    [aux_sym_function_statement_token2] = ACTIONS(1439),
    [aux_sym_function_statement_token3] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1439),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1439),
    [sym_label] = ACTIONS(1439),
    [aux_sym_trap_statement_token1] = ACTIONS(1439),
    [aux_sym_try_statement_token1] = ACTIONS(1439),
    [aux_sym_data_statement_token1] = ACTIONS(1439),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1439),
    [aux_sym_parallel_statement_token1] = ACTIONS(1439),
    [aux_sym_sequence_statement_token1] = ACTIONS(1439),
    [anon_sym_AMP] = ACTIONS(1439),
    [aux_sym_command_name_token1] = ACTIONS(1439),
    [aux_sym_command_name_token2] = ACTIONS(1439),
    [aux_sym_command_name_token3] = ACTIONS(1439),
    [aux_sym_command_name_token4] = ACTIONS(1439),
    [aux_sym_command_name_token5] = ACTIONS(1439),
    [aux_sym_command_name_token6] = ACTIONS(1439),
    [aux_sym_command_name_token7] = ACTIONS(1439),
    [aux_sym_command_name_token8] = ACTIONS(1439),
    [aux_sym_command_name_token9] = ACTIONS(1439),
    [aux_sym_command_name_token10] = ACTIONS(1439),
    [aux_sym_command_name_token11] = ACTIONS(1439),
    [aux_sym_command_name_token12] = ACTIONS(1439),
    [aux_sym_command_name_token13] = ACTIONS(1439),
    [aux_sym_command_name_token14] = ACTIONS(1439),
    [aux_sym_command_name_token15] = ACTIONS(1439),
    [aux_sym_command_name_token16] = ACTIONS(1439),
    [aux_sym_command_name_token17] = ACTIONS(1439),
    [aux_sym_command_name_token18] = ACTIONS(1439),
    [aux_sym_class_statement_token1] = ACTIONS(1439),
    [aux_sym_enum_statement_token1] = ACTIONS(1439),
    [anon_sym_PLUS] = ACTIONS(1439),
    [anon_sym_DASH] = ACTIONS(1439),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1439),
    [anon_sym_BANG] = ACTIONS(1439),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1439),
    [anon_sym_PLUS_PLUS] = ACTIONS(1439),
    [anon_sym_DASH_DASH] = ACTIONS(1439),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1439),
    [anon_sym_AT_LPAREN] = ACTIONS(1439),
    [anon_sym_AT_LBRACE] = ACTIONS(1439),
  },
  [STATE(395)] = {
    [ts_builtin_sym_end] = ACTIONS(1511),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1441),
    [sym_hexadecimal_integer_literal] = ACTIONS(1441),
    [sym_real_literal] = ACTIONS(1441),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1441),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1441),
    [sym_verbatim_string_characters] = ACTIONS(1441),
    [sym_verbatim_here_string_characters] = ACTIONS(1441),
    [anon_sym_DOT] = ACTIONS(1441),
    [anon_sym_LBRACK] = ACTIONS(1441),
    [aux_sym_comparison_operator_token37] = ACTIONS(1441),
    [aux_sym_comparison_operator_token50] = ACTIONS(1441),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1441),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1441),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1441),
    [anon_sym_DOLLAR_] = ACTIONS(1441),
    [aux_sym_variable_token1] = ACTIONS(1441),
    [aux_sym_variable_token2] = ACTIONS(1441),
    [sym_braced_variable] = ACTIONS(1441),
    [anon_sym_SEMI] = ACTIONS(1441),
    [anon_sym_LPAREN] = ACTIONS(1441),
    [anon_sym_COMMA] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1441),
    [aux_sym_if_statement_token1] = ACTIONS(1441),
    [aux_sym_switch_statement_token1] = ACTIONS(1441),
    [aux_sym_foreach_statement_token1] = ACTIONS(1441),
    [aux_sym_for_statement_token1] = ACTIONS(1441),
    [aux_sym_while_statement_token1] = ACTIONS(1441),
    [aux_sym_do_statement_token1] = ACTIONS(1441),
    [aux_sym_function_statement_token1] = ACTIONS(1441),
    [aux_sym_function_statement_token2] = ACTIONS(1441),
    [aux_sym_function_statement_token3] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1441),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1441),
    [sym_label] = ACTIONS(1441),
    [aux_sym_trap_statement_token1] = ACTIONS(1441),
    [aux_sym_try_statement_token1] = ACTIONS(1441),
    [aux_sym_data_statement_token1] = ACTIONS(1441),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1441),
    [aux_sym_parallel_statement_token1] = ACTIONS(1441),
    [aux_sym_sequence_statement_token1] = ACTIONS(1441),
    [anon_sym_AMP] = ACTIONS(1441),
    [aux_sym_command_name_token1] = ACTIONS(1441),
    [aux_sym_command_name_token2] = ACTIONS(1441),
    [aux_sym_command_name_token3] = ACTIONS(1441),
    [aux_sym_command_name_token4] = ACTIONS(1441),
    [aux_sym_command_name_token5] = ACTIONS(1441),
    [aux_sym_command_name_token6] = ACTIONS(1441),
    [aux_sym_command_name_token7] = ACTIONS(1441),
    [aux_sym_command_name_token8] = ACTIONS(1441),
    [aux_sym_command_name_token9] = ACTIONS(1441),
    [aux_sym_command_name_token10] = ACTIONS(1441),
    [aux_sym_command_name_token11] = ACTIONS(1441),
    [aux_sym_command_name_token12] = ACTIONS(1441),
    [aux_sym_command_name_token13] = ACTIONS(1441),
    [aux_sym_command_name_token14] = ACTIONS(1441),
    [aux_sym_command_name_token15] = ACTIONS(1441),
    [aux_sym_command_name_token16] = ACTIONS(1441),
    [aux_sym_command_name_token17] = ACTIONS(1441),
    [aux_sym_command_name_token18] = ACTIONS(1441),
    [aux_sym_class_statement_token1] = ACTIONS(1441),
    [aux_sym_enum_statement_token1] = ACTIONS(1441),
    [anon_sym_PLUS] = ACTIONS(1441),
    [anon_sym_DASH] = ACTIONS(1441),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1441),
    [anon_sym_BANG] = ACTIONS(1441),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1441),
    [anon_sym_PLUS_PLUS] = ACTIONS(1441),
    [anon_sym_DASH_DASH] = ACTIONS(1441),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1441),
    [anon_sym_AT_LPAREN] = ACTIONS(1441),
    [anon_sym_AT_LBRACE] = ACTIONS(1441),
  },
  [STATE(396)] = {
    [ts_builtin_sym_end] = ACTIONS(1513),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1453),
    [sym_hexadecimal_integer_literal] = ACTIONS(1453),
    [sym_real_literal] = ACTIONS(1453),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1453),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1453),
    [sym_verbatim_string_characters] = ACTIONS(1453),
    [sym_verbatim_here_string_characters] = ACTIONS(1453),
    [anon_sym_DOT] = ACTIONS(1453),
    [anon_sym_LBRACK] = ACTIONS(1453),
    [aux_sym_comparison_operator_token37] = ACTIONS(1453),
    [aux_sym_comparison_operator_token50] = ACTIONS(1453),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1453),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1453),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1453),
    [anon_sym_DOLLAR_] = ACTIONS(1453),
    [aux_sym_variable_token1] = ACTIONS(1453),
    [aux_sym_variable_token2] = ACTIONS(1453),
    [sym_braced_variable] = ACTIONS(1453),
    [anon_sym_SEMI] = ACTIONS(1453),
    [anon_sym_LPAREN] = ACTIONS(1453),
    [anon_sym_COMMA] = ACTIONS(1453),
    [anon_sym_LBRACE] = ACTIONS(1453),
    [aux_sym_if_statement_token1] = ACTIONS(1453),
    [aux_sym_switch_statement_token1] = ACTIONS(1453),
    [aux_sym_foreach_statement_token1] = ACTIONS(1453),
    [aux_sym_for_statement_token1] = ACTIONS(1453),
    [aux_sym_while_statement_token1] = ACTIONS(1453),
    [aux_sym_do_statement_token1] = ACTIONS(1453),
    [aux_sym_function_statement_token1] = ACTIONS(1453),
    [aux_sym_function_statement_token2] = ACTIONS(1453),
    [aux_sym_function_statement_token3] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1453),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1453),
    [sym_label] = ACTIONS(1453),
    [aux_sym_trap_statement_token1] = ACTIONS(1453),
    [aux_sym_try_statement_token1] = ACTIONS(1453),
    [aux_sym_data_statement_token1] = ACTIONS(1453),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1453),
    [aux_sym_parallel_statement_token1] = ACTIONS(1453),
    [aux_sym_sequence_statement_token1] = ACTIONS(1453),
    [anon_sym_AMP] = ACTIONS(1453),
    [aux_sym_command_name_token1] = ACTIONS(1453),
    [aux_sym_command_name_token2] = ACTIONS(1453),
    [aux_sym_command_name_token3] = ACTIONS(1453),
    [aux_sym_command_name_token4] = ACTIONS(1453),
    [aux_sym_command_name_token5] = ACTIONS(1453),
    [aux_sym_command_name_token6] = ACTIONS(1453),
    [aux_sym_command_name_token7] = ACTIONS(1453),
    [aux_sym_command_name_token8] = ACTIONS(1453),
    [aux_sym_command_name_token9] = ACTIONS(1453),
    [aux_sym_command_name_token10] = ACTIONS(1453),
    [aux_sym_command_name_token11] = ACTIONS(1453),
    [aux_sym_command_name_token12] = ACTIONS(1453),
    [aux_sym_command_name_token13] = ACTIONS(1453),
    [aux_sym_command_name_token14] = ACTIONS(1453),
    [aux_sym_command_name_token15] = ACTIONS(1453),
    [aux_sym_command_name_token16] = ACTIONS(1453),
    [aux_sym_command_name_token17] = ACTIONS(1453),
    [aux_sym_command_name_token18] = ACTIONS(1453),
    [aux_sym_class_statement_token1] = ACTIONS(1453),
    [aux_sym_enum_statement_token1] = ACTIONS(1453),
    [anon_sym_PLUS] = ACTIONS(1453),
    [anon_sym_DASH] = ACTIONS(1453),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1453),
    [anon_sym_BANG] = ACTIONS(1453),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1453),
    [anon_sym_PLUS_PLUS] = ACTIONS(1453),
    [anon_sym_DASH_DASH] = ACTIONS(1453),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1453),
    [anon_sym_AT_LPAREN] = ACTIONS(1453),
    [anon_sym_AT_LBRACE] = ACTIONS(1453),
  },
  [STATE(397)] = {
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1465),
    [sym_hexadecimal_integer_literal] = ACTIONS(1465),
    [sym_real_literal] = ACTIONS(1465),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1465),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1465),
    [sym_verbatim_string_characters] = ACTIONS(1465),
    [sym_verbatim_here_string_characters] = ACTIONS(1465),
    [anon_sym_DOT] = ACTIONS(1465),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_comparison_operator_token37] = ACTIONS(1465),
    [aux_sym_comparison_operator_token50] = ACTIONS(1465),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1465),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1465),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1465),
    [anon_sym_DOLLAR_] = ACTIONS(1465),
    [aux_sym_variable_token1] = ACTIONS(1465),
    [aux_sym_variable_token2] = ACTIONS(1465),
    [sym_braced_variable] = ACTIONS(1465),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [aux_sym_if_statement_token1] = ACTIONS(1465),
    [aux_sym_switch_statement_token1] = ACTIONS(1465),
    [aux_sym_foreach_statement_token1] = ACTIONS(1465),
    [aux_sym_for_statement_token1] = ACTIONS(1465),
    [aux_sym_while_statement_token1] = ACTIONS(1465),
    [aux_sym_do_statement_token1] = ACTIONS(1465),
    [aux_sym_function_statement_token1] = ACTIONS(1465),
    [aux_sym_function_statement_token2] = ACTIONS(1465),
    [aux_sym_function_statement_token3] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1465),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1465),
    [sym_label] = ACTIONS(1465),
    [aux_sym_trap_statement_token1] = ACTIONS(1465),
    [aux_sym_try_statement_token1] = ACTIONS(1465),
    [aux_sym_data_statement_token1] = ACTIONS(1465),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1465),
    [aux_sym_parallel_statement_token1] = ACTIONS(1465),
    [aux_sym_sequence_statement_token1] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1465),
    [aux_sym_command_name_token1] = ACTIONS(1465),
    [aux_sym_command_name_token2] = ACTIONS(1465),
    [aux_sym_command_name_token3] = ACTIONS(1465),
    [aux_sym_command_name_token4] = ACTIONS(1465),
    [aux_sym_command_name_token5] = ACTIONS(1465),
    [aux_sym_command_name_token6] = ACTIONS(1465),
    [aux_sym_command_name_token7] = ACTIONS(1465),
    [aux_sym_command_name_token8] = ACTIONS(1465),
    [aux_sym_command_name_token9] = ACTIONS(1465),
    [aux_sym_command_name_token10] = ACTIONS(1465),
    [aux_sym_command_name_token11] = ACTIONS(1465),
    [aux_sym_command_name_token12] = ACTIONS(1465),
    [aux_sym_command_name_token13] = ACTIONS(1465),
    [aux_sym_command_name_token14] = ACTIONS(1465),
    [aux_sym_command_name_token15] = ACTIONS(1465),
    [aux_sym_command_name_token16] = ACTIONS(1465),
    [aux_sym_command_name_token17] = ACTIONS(1465),
    [aux_sym_command_name_token18] = ACTIONS(1465),
    [aux_sym_class_statement_token1] = ACTIONS(1465),
    [aux_sym_enum_statement_token1] = ACTIONS(1465),
    [anon_sym_PLUS] = ACTIONS(1465),
    [anon_sym_DASH] = ACTIONS(1465),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1465),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1465),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1465),
    [anon_sym_AT_LPAREN] = ACTIONS(1465),
    [anon_sym_AT_LBRACE] = ACTIONS(1465),
  },
  [STATE(398)] = {
    [ts_builtin_sym_end] = ACTIONS(1517),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1387),
    [sym_hexadecimal_integer_literal] = ACTIONS(1387),
    [sym_real_literal] = ACTIONS(1387),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1387),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1387),
    [sym_verbatim_string_characters] = ACTIONS(1387),
    [sym_verbatim_here_string_characters] = ACTIONS(1387),
    [anon_sym_DOT] = ACTIONS(1387),
    [anon_sym_LBRACK] = ACTIONS(1387),
    [aux_sym_comparison_operator_token37] = ACTIONS(1387),
    [aux_sym_comparison_operator_token50] = ACTIONS(1387),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1387),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1387),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1387),
    [anon_sym_DOLLAR_] = ACTIONS(1387),
    [aux_sym_variable_token1] = ACTIONS(1387),
    [aux_sym_variable_token2] = ACTIONS(1387),
    [sym_braced_variable] = ACTIONS(1387),
    [anon_sym_SEMI] = ACTIONS(1387),
    [anon_sym_LPAREN] = ACTIONS(1387),
    [anon_sym_COMMA] = ACTIONS(1387),
    [anon_sym_LBRACE] = ACTIONS(1387),
    [aux_sym_if_statement_token1] = ACTIONS(1387),
    [aux_sym_switch_statement_token1] = ACTIONS(1387),
    [aux_sym_foreach_statement_token1] = ACTIONS(1387),
    [aux_sym_for_statement_token1] = ACTIONS(1387),
    [aux_sym_while_statement_token1] = ACTIONS(1387),
    [aux_sym_do_statement_token1] = ACTIONS(1387),
    [aux_sym_function_statement_token1] = ACTIONS(1387),
    [aux_sym_function_statement_token2] = ACTIONS(1387),
    [aux_sym_function_statement_token3] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1387),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1387),
    [sym_label] = ACTIONS(1387),
    [aux_sym_trap_statement_token1] = ACTIONS(1387),
    [aux_sym_try_statement_token1] = ACTIONS(1387),
    [aux_sym_data_statement_token1] = ACTIONS(1387),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1387),
    [aux_sym_parallel_statement_token1] = ACTIONS(1387),
    [aux_sym_sequence_statement_token1] = ACTIONS(1387),
    [anon_sym_AMP] = ACTIONS(1387),
    [aux_sym_command_name_token1] = ACTIONS(1387),
    [aux_sym_command_name_token2] = ACTIONS(1387),
    [aux_sym_command_name_token3] = ACTIONS(1387),
    [aux_sym_command_name_token4] = ACTIONS(1387),
    [aux_sym_command_name_token5] = ACTIONS(1387),
    [aux_sym_command_name_token6] = ACTIONS(1387),
    [aux_sym_command_name_token7] = ACTIONS(1387),
    [aux_sym_command_name_token8] = ACTIONS(1387),
    [aux_sym_command_name_token9] = ACTIONS(1387),
    [aux_sym_command_name_token10] = ACTIONS(1387),
    [aux_sym_command_name_token11] = ACTIONS(1387),
    [aux_sym_command_name_token12] = ACTIONS(1387),
    [aux_sym_command_name_token13] = ACTIONS(1387),
    [aux_sym_command_name_token14] = ACTIONS(1387),
    [aux_sym_command_name_token15] = ACTIONS(1387),
    [aux_sym_command_name_token16] = ACTIONS(1387),
    [aux_sym_command_name_token17] = ACTIONS(1387),
    [aux_sym_command_name_token18] = ACTIONS(1387),
    [aux_sym_class_statement_token1] = ACTIONS(1387),
    [aux_sym_enum_statement_token1] = ACTIONS(1387),
    [anon_sym_PLUS] = ACTIONS(1387),
    [anon_sym_DASH] = ACTIONS(1387),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1387),
    [anon_sym_BANG] = ACTIONS(1387),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1387),
    [anon_sym_PLUS_PLUS] = ACTIONS(1387),
    [anon_sym_DASH_DASH] = ACTIONS(1387),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1387),
    [anon_sym_AT_LPAREN] = ACTIONS(1387),
    [anon_sym_AT_LBRACE] = ACTIONS(1387),
  },
  [STATE(399)] = {
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1389),
    [sym_hexadecimal_integer_literal] = ACTIONS(1389),
    [sym_real_literal] = ACTIONS(1389),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1389),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1389),
    [sym_verbatim_string_characters] = ACTIONS(1389),
    [sym_verbatim_here_string_characters] = ACTIONS(1389),
    [anon_sym_DOT] = ACTIONS(1389),
    [anon_sym_LBRACK] = ACTIONS(1389),
    [aux_sym_comparison_operator_token37] = ACTIONS(1389),
    [aux_sym_comparison_operator_token50] = ACTIONS(1389),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1389),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1389),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1389),
    [anon_sym_DOLLAR_] = ACTIONS(1389),
    [aux_sym_variable_token1] = ACTIONS(1389),
    [aux_sym_variable_token2] = ACTIONS(1389),
    [sym_braced_variable] = ACTIONS(1389),
    [anon_sym_SEMI] = ACTIONS(1389),
    [anon_sym_LPAREN] = ACTIONS(1389),
    [anon_sym_COMMA] = ACTIONS(1389),
    [anon_sym_LBRACE] = ACTIONS(1389),
    [aux_sym_if_statement_token1] = ACTIONS(1389),
    [aux_sym_switch_statement_token1] = ACTIONS(1389),
    [aux_sym_foreach_statement_token1] = ACTIONS(1389),
    [aux_sym_for_statement_token1] = ACTIONS(1389),
    [aux_sym_while_statement_token1] = ACTIONS(1389),
    [aux_sym_do_statement_token1] = ACTIONS(1389),
    [aux_sym_function_statement_token1] = ACTIONS(1389),
    [aux_sym_function_statement_token2] = ACTIONS(1389),
    [aux_sym_function_statement_token3] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1389),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1389),
    [sym_label] = ACTIONS(1389),
    [aux_sym_trap_statement_token1] = ACTIONS(1389),
    [aux_sym_try_statement_token1] = ACTIONS(1389),
    [aux_sym_data_statement_token1] = ACTIONS(1389),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1389),
    [aux_sym_parallel_statement_token1] = ACTIONS(1389),
    [aux_sym_sequence_statement_token1] = ACTIONS(1389),
    [anon_sym_AMP] = ACTIONS(1389),
    [aux_sym_command_name_token1] = ACTIONS(1389),
    [aux_sym_command_name_token2] = ACTIONS(1389),
    [aux_sym_command_name_token3] = ACTIONS(1389),
    [aux_sym_command_name_token4] = ACTIONS(1389),
    [aux_sym_command_name_token5] = ACTIONS(1389),
    [aux_sym_command_name_token6] = ACTIONS(1389),
    [aux_sym_command_name_token7] = ACTIONS(1389),
    [aux_sym_command_name_token8] = ACTIONS(1389),
    [aux_sym_command_name_token9] = ACTIONS(1389),
    [aux_sym_command_name_token10] = ACTIONS(1389),
    [aux_sym_command_name_token11] = ACTIONS(1389),
    [aux_sym_command_name_token12] = ACTIONS(1389),
    [aux_sym_command_name_token13] = ACTIONS(1389),
    [aux_sym_command_name_token14] = ACTIONS(1389),
    [aux_sym_command_name_token15] = ACTIONS(1389),
    [aux_sym_command_name_token16] = ACTIONS(1389),
    [aux_sym_command_name_token17] = ACTIONS(1389),
    [aux_sym_command_name_token18] = ACTIONS(1389),
    [aux_sym_class_statement_token1] = ACTIONS(1389),
    [aux_sym_enum_statement_token1] = ACTIONS(1389),
    [anon_sym_PLUS] = ACTIONS(1389),
    [anon_sym_DASH] = ACTIONS(1389),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1389),
    [anon_sym_BANG] = ACTIONS(1389),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1389),
    [anon_sym_PLUS_PLUS] = ACTIONS(1389),
    [anon_sym_DASH_DASH] = ACTIONS(1389),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1389),
    [anon_sym_AT_LPAREN] = ACTIONS(1389),
    [anon_sym_AT_LBRACE] = ACTIONS(1389),
  },
  [STATE(400)] = {
    [ts_builtin_sym_end] = ACTIONS(1521),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1391),
    [sym_hexadecimal_integer_literal] = ACTIONS(1391),
    [sym_real_literal] = ACTIONS(1391),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1391),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1391),
    [sym_verbatim_string_characters] = ACTIONS(1391),
    [sym_verbatim_here_string_characters] = ACTIONS(1391),
    [anon_sym_DOT] = ACTIONS(1391),
    [anon_sym_LBRACK] = ACTIONS(1391),
    [aux_sym_comparison_operator_token37] = ACTIONS(1391),
    [aux_sym_comparison_operator_token50] = ACTIONS(1391),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1391),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1391),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1391),
    [anon_sym_DOLLAR_] = ACTIONS(1391),
    [aux_sym_variable_token1] = ACTIONS(1391),
    [aux_sym_variable_token2] = ACTIONS(1391),
    [sym_braced_variable] = ACTIONS(1391),
    [anon_sym_SEMI] = ACTIONS(1391),
    [anon_sym_LPAREN] = ACTIONS(1391),
    [anon_sym_COMMA] = ACTIONS(1391),
    [anon_sym_LBRACE] = ACTIONS(1391),
    [aux_sym_if_statement_token1] = ACTIONS(1391),
    [aux_sym_switch_statement_token1] = ACTIONS(1391),
    [aux_sym_foreach_statement_token1] = ACTIONS(1391),
    [aux_sym_for_statement_token1] = ACTIONS(1391),
    [aux_sym_while_statement_token1] = ACTIONS(1391),
    [aux_sym_do_statement_token1] = ACTIONS(1391),
    [aux_sym_function_statement_token1] = ACTIONS(1391),
    [aux_sym_function_statement_token2] = ACTIONS(1391),
    [aux_sym_function_statement_token3] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1391),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1391),
    [sym_label] = ACTIONS(1391),
    [aux_sym_trap_statement_token1] = ACTIONS(1391),
    [aux_sym_try_statement_token1] = ACTIONS(1391),
    [aux_sym_data_statement_token1] = ACTIONS(1391),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1391),
    [aux_sym_parallel_statement_token1] = ACTIONS(1391),
    [aux_sym_sequence_statement_token1] = ACTIONS(1391),
    [anon_sym_AMP] = ACTIONS(1391),
    [aux_sym_command_name_token1] = ACTIONS(1391),
    [aux_sym_command_name_token2] = ACTIONS(1391),
    [aux_sym_command_name_token3] = ACTIONS(1391),
    [aux_sym_command_name_token4] = ACTIONS(1391),
    [aux_sym_command_name_token5] = ACTIONS(1391),
    [aux_sym_command_name_token6] = ACTIONS(1391),
    [aux_sym_command_name_token7] = ACTIONS(1391),
    [aux_sym_command_name_token8] = ACTIONS(1391),
    [aux_sym_command_name_token9] = ACTIONS(1391),
    [aux_sym_command_name_token10] = ACTIONS(1391),
    [aux_sym_command_name_token11] = ACTIONS(1391),
    [aux_sym_command_name_token12] = ACTIONS(1391),
    [aux_sym_command_name_token13] = ACTIONS(1391),
    [aux_sym_command_name_token14] = ACTIONS(1391),
    [aux_sym_command_name_token15] = ACTIONS(1391),
    [aux_sym_command_name_token16] = ACTIONS(1391),
    [aux_sym_command_name_token17] = ACTIONS(1391),
    [aux_sym_command_name_token18] = ACTIONS(1391),
    [aux_sym_class_statement_token1] = ACTIONS(1391),
    [aux_sym_enum_statement_token1] = ACTIONS(1391),
    [anon_sym_PLUS] = ACTIONS(1391),
    [anon_sym_DASH] = ACTIONS(1391),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1391),
    [anon_sym_BANG] = ACTIONS(1391),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1391),
    [anon_sym_PLUS_PLUS] = ACTIONS(1391),
    [anon_sym_DASH_DASH] = ACTIONS(1391),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1391),
    [anon_sym_AT_LPAREN] = ACTIONS(1391),
    [anon_sym_AT_LBRACE] = ACTIONS(1391),
  },
  [STATE(401)] = {
    [ts_builtin_sym_end] = ACTIONS(1523),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1393),
    [sym_hexadecimal_integer_literal] = ACTIONS(1393),
    [sym_real_literal] = ACTIONS(1393),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1393),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1393),
    [sym_verbatim_string_characters] = ACTIONS(1393),
    [sym_verbatim_here_string_characters] = ACTIONS(1393),
    [anon_sym_DOT] = ACTIONS(1393),
    [anon_sym_LBRACK] = ACTIONS(1393),
    [aux_sym_comparison_operator_token37] = ACTIONS(1393),
    [aux_sym_comparison_operator_token50] = ACTIONS(1393),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1393),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1393),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1393),
    [anon_sym_DOLLAR_] = ACTIONS(1393),
    [aux_sym_variable_token1] = ACTIONS(1393),
    [aux_sym_variable_token2] = ACTIONS(1393),
    [sym_braced_variable] = ACTIONS(1393),
    [anon_sym_SEMI] = ACTIONS(1393),
    [anon_sym_LPAREN] = ACTIONS(1393),
    [anon_sym_COMMA] = ACTIONS(1393),
    [anon_sym_LBRACE] = ACTIONS(1393),
    [aux_sym_if_statement_token1] = ACTIONS(1393),
    [aux_sym_switch_statement_token1] = ACTIONS(1393),
    [aux_sym_foreach_statement_token1] = ACTIONS(1393),
    [aux_sym_for_statement_token1] = ACTIONS(1393),
    [aux_sym_while_statement_token1] = ACTIONS(1393),
    [aux_sym_do_statement_token1] = ACTIONS(1393),
    [aux_sym_function_statement_token1] = ACTIONS(1393),
    [aux_sym_function_statement_token2] = ACTIONS(1393),
    [aux_sym_function_statement_token3] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1393),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1393),
    [sym_label] = ACTIONS(1393),
    [aux_sym_trap_statement_token1] = ACTIONS(1393),
    [aux_sym_try_statement_token1] = ACTIONS(1393),
    [aux_sym_data_statement_token1] = ACTIONS(1393),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1393),
    [aux_sym_parallel_statement_token1] = ACTIONS(1393),
    [aux_sym_sequence_statement_token1] = ACTIONS(1393),
    [anon_sym_AMP] = ACTIONS(1393),
    [aux_sym_command_name_token1] = ACTIONS(1393),
    [aux_sym_command_name_token2] = ACTIONS(1393),
    [aux_sym_command_name_token3] = ACTIONS(1393),
    [aux_sym_command_name_token4] = ACTIONS(1393),
    [aux_sym_command_name_token5] = ACTIONS(1393),
    [aux_sym_command_name_token6] = ACTIONS(1393),
    [aux_sym_command_name_token7] = ACTIONS(1393),
    [aux_sym_command_name_token8] = ACTIONS(1393),
    [aux_sym_command_name_token9] = ACTIONS(1393),
    [aux_sym_command_name_token10] = ACTIONS(1393),
    [aux_sym_command_name_token11] = ACTIONS(1393),
    [aux_sym_command_name_token12] = ACTIONS(1393),
    [aux_sym_command_name_token13] = ACTIONS(1393),
    [aux_sym_command_name_token14] = ACTIONS(1393),
    [aux_sym_command_name_token15] = ACTIONS(1393),
    [aux_sym_command_name_token16] = ACTIONS(1393),
    [aux_sym_command_name_token17] = ACTIONS(1393),
    [aux_sym_command_name_token18] = ACTIONS(1393),
    [aux_sym_class_statement_token1] = ACTIONS(1393),
    [aux_sym_enum_statement_token1] = ACTIONS(1393),
    [anon_sym_PLUS] = ACTIONS(1393),
    [anon_sym_DASH] = ACTIONS(1393),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1393),
    [anon_sym_BANG] = ACTIONS(1393),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1393),
    [anon_sym_PLUS_PLUS] = ACTIONS(1393),
    [anon_sym_DASH_DASH] = ACTIONS(1393),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1393),
    [anon_sym_AT_LPAREN] = ACTIONS(1393),
    [anon_sym_AT_LBRACE] = ACTIONS(1393),
  },
  [STATE(402)] = {
    [ts_builtin_sym_end] = ACTIONS(1525),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1395),
    [sym_hexadecimal_integer_literal] = ACTIONS(1395),
    [sym_real_literal] = ACTIONS(1395),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1395),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1395),
    [sym_verbatim_string_characters] = ACTIONS(1395),
    [sym_verbatim_here_string_characters] = ACTIONS(1395),
    [anon_sym_DOT] = ACTIONS(1395),
    [anon_sym_LBRACK] = ACTIONS(1395),
    [aux_sym_comparison_operator_token37] = ACTIONS(1395),
    [aux_sym_comparison_operator_token50] = ACTIONS(1395),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1395),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1395),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1395),
    [anon_sym_DOLLAR_] = ACTIONS(1395),
    [aux_sym_variable_token1] = ACTIONS(1395),
    [aux_sym_variable_token2] = ACTIONS(1395),
    [sym_braced_variable] = ACTIONS(1395),
    [anon_sym_SEMI] = ACTIONS(1395),
    [anon_sym_LPAREN] = ACTIONS(1395),
    [anon_sym_COMMA] = ACTIONS(1395),
    [anon_sym_LBRACE] = ACTIONS(1395),
    [aux_sym_if_statement_token1] = ACTIONS(1395),
    [aux_sym_switch_statement_token1] = ACTIONS(1395),
    [aux_sym_foreach_statement_token1] = ACTIONS(1395),
    [aux_sym_for_statement_token1] = ACTIONS(1395),
    [aux_sym_while_statement_token1] = ACTIONS(1395),
    [aux_sym_do_statement_token1] = ACTIONS(1395),
    [aux_sym_function_statement_token1] = ACTIONS(1395),
    [aux_sym_function_statement_token2] = ACTIONS(1395),
    [aux_sym_function_statement_token3] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1395),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1395),
    [sym_label] = ACTIONS(1395),
    [aux_sym_trap_statement_token1] = ACTIONS(1395),
    [aux_sym_try_statement_token1] = ACTIONS(1395),
    [aux_sym_data_statement_token1] = ACTIONS(1395),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1395),
    [aux_sym_parallel_statement_token1] = ACTIONS(1395),
    [aux_sym_sequence_statement_token1] = ACTIONS(1395),
    [anon_sym_AMP] = ACTIONS(1395),
    [aux_sym_command_name_token1] = ACTIONS(1395),
    [aux_sym_command_name_token2] = ACTIONS(1395),
    [aux_sym_command_name_token3] = ACTIONS(1395),
    [aux_sym_command_name_token4] = ACTIONS(1395),
    [aux_sym_command_name_token5] = ACTIONS(1395),
    [aux_sym_command_name_token6] = ACTIONS(1395),
    [aux_sym_command_name_token7] = ACTIONS(1395),
    [aux_sym_command_name_token8] = ACTIONS(1395),
    [aux_sym_command_name_token9] = ACTIONS(1395),
    [aux_sym_command_name_token10] = ACTIONS(1395),
    [aux_sym_command_name_token11] = ACTIONS(1395),
    [aux_sym_command_name_token12] = ACTIONS(1395),
    [aux_sym_command_name_token13] = ACTIONS(1395),
    [aux_sym_command_name_token14] = ACTIONS(1395),
    [aux_sym_command_name_token15] = ACTIONS(1395),
    [aux_sym_command_name_token16] = ACTIONS(1395),
    [aux_sym_command_name_token17] = ACTIONS(1395),
    [aux_sym_command_name_token18] = ACTIONS(1395),
    [aux_sym_class_statement_token1] = ACTIONS(1395),
    [aux_sym_enum_statement_token1] = ACTIONS(1395),
    [anon_sym_PLUS] = ACTIONS(1395),
    [anon_sym_DASH] = ACTIONS(1395),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1395),
    [anon_sym_BANG] = ACTIONS(1395),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1395),
    [anon_sym_PLUS_PLUS] = ACTIONS(1395),
    [anon_sym_DASH_DASH] = ACTIONS(1395),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1395),
    [anon_sym_AT_LPAREN] = ACTIONS(1395),
    [anon_sym_AT_LBRACE] = ACTIONS(1395),
  },
  [STATE(403)] = {
    [ts_builtin_sym_end] = ACTIONS(1527),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1397),
    [sym_hexadecimal_integer_literal] = ACTIONS(1397),
    [sym_real_literal] = ACTIONS(1397),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1397),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1397),
    [sym_verbatim_string_characters] = ACTIONS(1397),
    [sym_verbatim_here_string_characters] = ACTIONS(1397),
    [anon_sym_DOT] = ACTIONS(1397),
    [anon_sym_LBRACK] = ACTIONS(1397),
    [aux_sym_comparison_operator_token37] = ACTIONS(1397),
    [aux_sym_comparison_operator_token50] = ACTIONS(1397),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1397),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1397),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1397),
    [anon_sym_DOLLAR_] = ACTIONS(1397),
    [aux_sym_variable_token1] = ACTIONS(1397),
    [aux_sym_variable_token2] = ACTIONS(1397),
    [sym_braced_variable] = ACTIONS(1397),
    [anon_sym_SEMI] = ACTIONS(1397),
    [anon_sym_LPAREN] = ACTIONS(1397),
    [anon_sym_COMMA] = ACTIONS(1397),
    [anon_sym_LBRACE] = ACTIONS(1397),
    [aux_sym_if_statement_token1] = ACTIONS(1397),
    [aux_sym_switch_statement_token1] = ACTIONS(1397),
    [aux_sym_foreach_statement_token1] = ACTIONS(1397),
    [aux_sym_for_statement_token1] = ACTIONS(1397),
    [aux_sym_while_statement_token1] = ACTIONS(1397),
    [aux_sym_do_statement_token1] = ACTIONS(1397),
    [aux_sym_function_statement_token1] = ACTIONS(1397),
    [aux_sym_function_statement_token2] = ACTIONS(1397),
    [aux_sym_function_statement_token3] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1397),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1397),
    [sym_label] = ACTIONS(1397),
    [aux_sym_trap_statement_token1] = ACTIONS(1397),
    [aux_sym_try_statement_token1] = ACTIONS(1397),
    [aux_sym_data_statement_token1] = ACTIONS(1397),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1397),
    [aux_sym_parallel_statement_token1] = ACTIONS(1397),
    [aux_sym_sequence_statement_token1] = ACTIONS(1397),
    [anon_sym_AMP] = ACTIONS(1397),
    [aux_sym_command_name_token1] = ACTIONS(1397),
    [aux_sym_command_name_token2] = ACTIONS(1397),
    [aux_sym_command_name_token3] = ACTIONS(1397),
    [aux_sym_command_name_token4] = ACTIONS(1397),
    [aux_sym_command_name_token5] = ACTIONS(1397),
    [aux_sym_command_name_token6] = ACTIONS(1397),
    [aux_sym_command_name_token7] = ACTIONS(1397),
    [aux_sym_command_name_token8] = ACTIONS(1397),
    [aux_sym_command_name_token9] = ACTIONS(1397),
    [aux_sym_command_name_token10] = ACTIONS(1397),
    [aux_sym_command_name_token11] = ACTIONS(1397),
    [aux_sym_command_name_token12] = ACTIONS(1397),
    [aux_sym_command_name_token13] = ACTIONS(1397),
    [aux_sym_command_name_token14] = ACTIONS(1397),
    [aux_sym_command_name_token15] = ACTIONS(1397),
    [aux_sym_command_name_token16] = ACTIONS(1397),
    [aux_sym_command_name_token17] = ACTIONS(1397),
    [aux_sym_command_name_token18] = ACTIONS(1397),
    [aux_sym_class_statement_token1] = ACTIONS(1397),
    [aux_sym_enum_statement_token1] = ACTIONS(1397),
    [anon_sym_PLUS] = ACTIONS(1397),
    [anon_sym_DASH] = ACTIONS(1397),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1397),
    [anon_sym_BANG] = ACTIONS(1397),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1397),
    [anon_sym_PLUS_PLUS] = ACTIONS(1397),
    [anon_sym_DASH_DASH] = ACTIONS(1397),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1397),
    [anon_sym_AT_LPAREN] = ACTIONS(1397),
    [anon_sym_AT_LBRACE] = ACTIONS(1397),
  },
  [STATE(404)] = {
    [ts_builtin_sym_end] = ACTIONS(1529),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1399),
    [sym_hexadecimal_integer_literal] = ACTIONS(1399),
    [sym_real_literal] = ACTIONS(1399),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1399),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1399),
    [sym_verbatim_string_characters] = ACTIONS(1399),
    [sym_verbatim_here_string_characters] = ACTIONS(1399),
    [anon_sym_DOT] = ACTIONS(1399),
    [anon_sym_LBRACK] = ACTIONS(1399),
    [aux_sym_comparison_operator_token37] = ACTIONS(1399),
    [aux_sym_comparison_operator_token50] = ACTIONS(1399),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1399),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1399),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1399),
    [anon_sym_DOLLAR_] = ACTIONS(1399),
    [aux_sym_variable_token1] = ACTIONS(1399),
    [aux_sym_variable_token2] = ACTIONS(1399),
    [sym_braced_variable] = ACTIONS(1399),
    [anon_sym_SEMI] = ACTIONS(1399),
    [anon_sym_LPAREN] = ACTIONS(1399),
    [anon_sym_COMMA] = ACTIONS(1399),
    [anon_sym_LBRACE] = ACTIONS(1399),
    [aux_sym_if_statement_token1] = ACTIONS(1399),
    [aux_sym_switch_statement_token1] = ACTIONS(1399),
    [aux_sym_foreach_statement_token1] = ACTIONS(1399),
    [aux_sym_for_statement_token1] = ACTIONS(1399),
    [aux_sym_while_statement_token1] = ACTIONS(1399),
    [aux_sym_do_statement_token1] = ACTIONS(1399),
    [aux_sym_function_statement_token1] = ACTIONS(1399),
    [aux_sym_function_statement_token2] = ACTIONS(1399),
    [aux_sym_function_statement_token3] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1399),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1399),
    [sym_label] = ACTIONS(1399),
    [aux_sym_trap_statement_token1] = ACTIONS(1399),
    [aux_sym_try_statement_token1] = ACTIONS(1399),
    [aux_sym_data_statement_token1] = ACTIONS(1399),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1399),
    [aux_sym_parallel_statement_token1] = ACTIONS(1399),
    [aux_sym_sequence_statement_token1] = ACTIONS(1399),
    [anon_sym_AMP] = ACTIONS(1399),
    [aux_sym_command_name_token1] = ACTIONS(1399),
    [aux_sym_command_name_token2] = ACTIONS(1399),
    [aux_sym_command_name_token3] = ACTIONS(1399),
    [aux_sym_command_name_token4] = ACTIONS(1399),
    [aux_sym_command_name_token5] = ACTIONS(1399),
    [aux_sym_command_name_token6] = ACTIONS(1399),
    [aux_sym_command_name_token7] = ACTIONS(1399),
    [aux_sym_command_name_token8] = ACTIONS(1399),
    [aux_sym_command_name_token9] = ACTIONS(1399),
    [aux_sym_command_name_token10] = ACTIONS(1399),
    [aux_sym_command_name_token11] = ACTIONS(1399),
    [aux_sym_command_name_token12] = ACTIONS(1399),
    [aux_sym_command_name_token13] = ACTIONS(1399),
    [aux_sym_command_name_token14] = ACTIONS(1399),
    [aux_sym_command_name_token15] = ACTIONS(1399),
    [aux_sym_command_name_token16] = ACTIONS(1399),
    [aux_sym_command_name_token17] = ACTIONS(1399),
    [aux_sym_command_name_token18] = ACTIONS(1399),
    [aux_sym_class_statement_token1] = ACTIONS(1399),
    [aux_sym_enum_statement_token1] = ACTIONS(1399),
    [anon_sym_PLUS] = ACTIONS(1399),
    [anon_sym_DASH] = ACTIONS(1399),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1399),
    [anon_sym_BANG] = ACTIONS(1399),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1399),
    [anon_sym_PLUS_PLUS] = ACTIONS(1399),
    [anon_sym_DASH_DASH] = ACTIONS(1399),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1399),
    [anon_sym_AT_LPAREN] = ACTIONS(1399),
    [anon_sym_AT_LBRACE] = ACTIONS(1399),
  },
  [STATE(405)] = {
    [ts_builtin_sym_end] = ACTIONS(1531),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1401),
    [sym_hexadecimal_integer_literal] = ACTIONS(1401),
    [sym_real_literal] = ACTIONS(1401),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1401),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1401),
    [sym_verbatim_string_characters] = ACTIONS(1401),
    [sym_verbatim_here_string_characters] = ACTIONS(1401),
    [anon_sym_DOT] = ACTIONS(1401),
    [anon_sym_LBRACK] = ACTIONS(1401),
    [aux_sym_comparison_operator_token37] = ACTIONS(1401),
    [aux_sym_comparison_operator_token50] = ACTIONS(1401),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1401),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1401),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1401),
    [anon_sym_DOLLAR_] = ACTIONS(1401),
    [aux_sym_variable_token1] = ACTIONS(1401),
    [aux_sym_variable_token2] = ACTIONS(1401),
    [sym_braced_variable] = ACTIONS(1401),
    [anon_sym_SEMI] = ACTIONS(1401),
    [anon_sym_LPAREN] = ACTIONS(1401),
    [anon_sym_COMMA] = ACTIONS(1401),
    [anon_sym_LBRACE] = ACTIONS(1401),
    [aux_sym_if_statement_token1] = ACTIONS(1401),
    [aux_sym_switch_statement_token1] = ACTIONS(1401),
    [aux_sym_foreach_statement_token1] = ACTIONS(1401),
    [aux_sym_for_statement_token1] = ACTIONS(1401),
    [aux_sym_while_statement_token1] = ACTIONS(1401),
    [aux_sym_do_statement_token1] = ACTIONS(1401),
    [aux_sym_function_statement_token1] = ACTIONS(1401),
    [aux_sym_function_statement_token2] = ACTIONS(1401),
    [aux_sym_function_statement_token3] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1401),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1401),
    [sym_label] = ACTIONS(1401),
    [aux_sym_trap_statement_token1] = ACTIONS(1401),
    [aux_sym_try_statement_token1] = ACTIONS(1401),
    [aux_sym_data_statement_token1] = ACTIONS(1401),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1401),
    [aux_sym_parallel_statement_token1] = ACTIONS(1401),
    [aux_sym_sequence_statement_token1] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(1401),
    [aux_sym_command_name_token1] = ACTIONS(1401),
    [aux_sym_command_name_token2] = ACTIONS(1401),
    [aux_sym_command_name_token3] = ACTIONS(1401),
    [aux_sym_command_name_token4] = ACTIONS(1401),
    [aux_sym_command_name_token5] = ACTIONS(1401),
    [aux_sym_command_name_token6] = ACTIONS(1401),
    [aux_sym_command_name_token7] = ACTIONS(1401),
    [aux_sym_command_name_token8] = ACTIONS(1401),
    [aux_sym_command_name_token9] = ACTIONS(1401),
    [aux_sym_command_name_token10] = ACTIONS(1401),
    [aux_sym_command_name_token11] = ACTIONS(1401),
    [aux_sym_command_name_token12] = ACTIONS(1401),
    [aux_sym_command_name_token13] = ACTIONS(1401),
    [aux_sym_command_name_token14] = ACTIONS(1401),
    [aux_sym_command_name_token15] = ACTIONS(1401),
    [aux_sym_command_name_token16] = ACTIONS(1401),
    [aux_sym_command_name_token17] = ACTIONS(1401),
    [aux_sym_command_name_token18] = ACTIONS(1401),
    [aux_sym_class_statement_token1] = ACTIONS(1401),
    [aux_sym_enum_statement_token1] = ACTIONS(1401),
    [anon_sym_PLUS] = ACTIONS(1401),
    [anon_sym_DASH] = ACTIONS(1401),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1401),
    [anon_sym_BANG] = ACTIONS(1401),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1401),
    [anon_sym_PLUS_PLUS] = ACTIONS(1401),
    [anon_sym_DASH_DASH] = ACTIONS(1401),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1401),
    [anon_sym_AT_LPAREN] = ACTIONS(1401),
    [anon_sym_AT_LBRACE] = ACTIONS(1401),
  },
  [STATE(406)] = {
    [ts_builtin_sym_end] = ACTIONS(1377),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1367),
    [sym_hexadecimal_integer_literal] = ACTIONS(1367),
    [sym_real_literal] = ACTIONS(1367),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1367),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1367),
    [sym_verbatim_string_characters] = ACTIONS(1367),
    [sym_verbatim_here_string_characters] = ACTIONS(1367),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1367),
    [aux_sym_comparison_operator_token37] = ACTIONS(1367),
    [aux_sym_comparison_operator_token50] = ACTIONS(1367),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1367),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1367),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1367),
    [anon_sym_DOLLAR_] = ACTIONS(1367),
    [aux_sym_variable_token1] = ACTIONS(1367),
    [aux_sym_variable_token2] = ACTIONS(1367),
    [sym_braced_variable] = ACTIONS(1367),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1367),
    [anon_sym_COMMA] = ACTIONS(1367),
    [anon_sym_LBRACE] = ACTIONS(1367),
    [aux_sym_if_statement_token1] = ACTIONS(1367),
    [aux_sym_switch_statement_token1] = ACTIONS(1367),
    [aux_sym_foreach_statement_token1] = ACTIONS(1367),
    [aux_sym_for_statement_token1] = ACTIONS(1367),
    [aux_sym_while_statement_token1] = ACTIONS(1367),
    [aux_sym_do_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token1] = ACTIONS(1367),
    [aux_sym_function_statement_token2] = ACTIONS(1367),
    [aux_sym_function_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1367),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1367),
    [sym_label] = ACTIONS(1367),
    [aux_sym_trap_statement_token1] = ACTIONS(1367),
    [aux_sym_try_statement_token1] = ACTIONS(1367),
    [aux_sym_data_statement_token1] = ACTIONS(1367),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1367),
    [aux_sym_parallel_statement_token1] = ACTIONS(1367),
    [aux_sym_sequence_statement_token1] = ACTIONS(1367),
    [anon_sym_AMP] = ACTIONS(1367),
    [aux_sym_command_name_token1] = ACTIONS(1367),
    [aux_sym_command_name_token2] = ACTIONS(1367),
    [aux_sym_command_name_token3] = ACTIONS(1367),
    [aux_sym_command_name_token4] = ACTIONS(1367),
    [aux_sym_command_name_token5] = ACTIONS(1367),
    [aux_sym_command_name_token6] = ACTIONS(1367),
    [aux_sym_command_name_token7] = ACTIONS(1367),
    [aux_sym_command_name_token8] = ACTIONS(1367),
    [aux_sym_command_name_token9] = ACTIONS(1367),
    [aux_sym_command_name_token10] = ACTIONS(1367),
    [aux_sym_command_name_token11] = ACTIONS(1367),
    [aux_sym_command_name_token12] = ACTIONS(1367),
    [aux_sym_command_name_token13] = ACTIONS(1367),
    [aux_sym_command_name_token14] = ACTIONS(1367),
    [aux_sym_command_name_token15] = ACTIONS(1367),
    [aux_sym_command_name_token16] = ACTIONS(1367),
    [aux_sym_command_name_token17] = ACTIONS(1367),
    [aux_sym_command_name_token18] = ACTIONS(1367),
    [aux_sym_class_statement_token1] = ACTIONS(1367),
    [aux_sym_enum_statement_token1] = ACTIONS(1367),
    [anon_sym_PLUS] = ACTIONS(1367),
    [anon_sym_DASH] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1367),
    [anon_sym_BANG] = ACTIONS(1367),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1367),
    [anon_sym_PLUS_PLUS] = ACTIONS(1367),
    [anon_sym_DASH_DASH] = ACTIONS(1367),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LPAREN] = ACTIONS(1367),
    [anon_sym_AT_LBRACE] = ACTIONS(1367),
  },
  [STATE(407)] = {
    [ts_builtin_sym_end] = ACTIONS(1533),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1405),
    [sym_hexadecimal_integer_literal] = ACTIONS(1405),
    [sym_real_literal] = ACTIONS(1405),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1405),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1405),
    [sym_verbatim_string_characters] = ACTIONS(1405),
    [sym_verbatim_here_string_characters] = ACTIONS(1405),
    [anon_sym_DOT] = ACTIONS(1405),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_comparison_operator_token37] = ACTIONS(1405),
    [aux_sym_comparison_operator_token50] = ACTIONS(1405),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1405),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1405),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1405),
    [anon_sym_DOLLAR_] = ACTIONS(1405),
    [aux_sym_variable_token1] = ACTIONS(1405),
    [aux_sym_variable_token2] = ACTIONS(1405),
    [sym_braced_variable] = ACTIONS(1405),
    [anon_sym_SEMI] = ACTIONS(1405),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_COMMA] = ACTIONS(1405),
    [anon_sym_LBRACE] = ACTIONS(1405),
    [aux_sym_if_statement_token1] = ACTIONS(1405),
    [aux_sym_switch_statement_token1] = ACTIONS(1405),
    [aux_sym_foreach_statement_token1] = ACTIONS(1405),
    [aux_sym_for_statement_token1] = ACTIONS(1405),
    [aux_sym_while_statement_token1] = ACTIONS(1405),
    [aux_sym_do_statement_token1] = ACTIONS(1405),
    [aux_sym_function_statement_token1] = ACTIONS(1405),
    [aux_sym_function_statement_token2] = ACTIONS(1405),
    [aux_sym_function_statement_token3] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1405),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1405),
    [sym_label] = ACTIONS(1405),
    [aux_sym_trap_statement_token1] = ACTIONS(1405),
    [aux_sym_try_statement_token1] = ACTIONS(1405),
    [aux_sym_data_statement_token1] = ACTIONS(1405),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1405),
    [aux_sym_parallel_statement_token1] = ACTIONS(1405),
    [aux_sym_sequence_statement_token1] = ACTIONS(1405),
    [anon_sym_AMP] = ACTIONS(1405),
    [aux_sym_command_name_token1] = ACTIONS(1405),
    [aux_sym_command_name_token2] = ACTIONS(1405),
    [aux_sym_command_name_token3] = ACTIONS(1405),
    [aux_sym_command_name_token4] = ACTIONS(1405),
    [aux_sym_command_name_token5] = ACTIONS(1405),
    [aux_sym_command_name_token6] = ACTIONS(1405),
    [aux_sym_command_name_token7] = ACTIONS(1405),
    [aux_sym_command_name_token8] = ACTIONS(1405),
    [aux_sym_command_name_token9] = ACTIONS(1405),
    [aux_sym_command_name_token10] = ACTIONS(1405),
    [aux_sym_command_name_token11] = ACTIONS(1405),
    [aux_sym_command_name_token12] = ACTIONS(1405),
    [aux_sym_command_name_token13] = ACTIONS(1405),
    [aux_sym_command_name_token14] = ACTIONS(1405),
    [aux_sym_command_name_token15] = ACTIONS(1405),
    [aux_sym_command_name_token16] = ACTIONS(1405),
    [aux_sym_command_name_token17] = ACTIONS(1405),
    [aux_sym_command_name_token18] = ACTIONS(1405),
    [aux_sym_class_statement_token1] = ACTIONS(1405),
    [aux_sym_enum_statement_token1] = ACTIONS(1405),
    [anon_sym_PLUS] = ACTIONS(1405),
    [anon_sym_DASH] = ACTIONS(1405),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1405),
    [anon_sym_BANG] = ACTIONS(1405),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1405),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1405),
    [anon_sym_AT_LPAREN] = ACTIONS(1405),
    [anon_sym_AT_LBRACE] = ACTIONS(1405),
  },
  [STATE(408)] = {
    [ts_builtin_sym_end] = ACTIONS(1535),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1407),
    [sym_hexadecimal_integer_literal] = ACTIONS(1407),
    [sym_real_literal] = ACTIONS(1407),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1407),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1407),
    [sym_verbatim_string_characters] = ACTIONS(1407),
    [sym_verbatim_here_string_characters] = ACTIONS(1407),
    [anon_sym_DOT] = ACTIONS(1407),
    [anon_sym_LBRACK] = ACTIONS(1407),
    [aux_sym_comparison_operator_token37] = ACTIONS(1407),
    [aux_sym_comparison_operator_token50] = ACTIONS(1407),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1407),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1407),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1407),
    [anon_sym_DOLLAR_] = ACTIONS(1407),
    [aux_sym_variable_token1] = ACTIONS(1407),
    [aux_sym_variable_token2] = ACTIONS(1407),
    [sym_braced_variable] = ACTIONS(1407),
    [anon_sym_SEMI] = ACTIONS(1407),
    [anon_sym_LPAREN] = ACTIONS(1407),
    [anon_sym_COMMA] = ACTIONS(1407),
    [anon_sym_LBRACE] = ACTIONS(1407),
    [aux_sym_if_statement_token1] = ACTIONS(1407),
    [aux_sym_switch_statement_token1] = ACTIONS(1407),
    [aux_sym_foreach_statement_token1] = ACTIONS(1407),
    [aux_sym_for_statement_token1] = ACTIONS(1407),
    [aux_sym_while_statement_token1] = ACTIONS(1407),
    [aux_sym_do_statement_token1] = ACTIONS(1407),
    [aux_sym_function_statement_token1] = ACTIONS(1407),
    [aux_sym_function_statement_token2] = ACTIONS(1407),
    [aux_sym_function_statement_token3] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1407),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1407),
    [sym_label] = ACTIONS(1407),
    [aux_sym_trap_statement_token1] = ACTIONS(1407),
    [aux_sym_try_statement_token1] = ACTIONS(1407),
    [aux_sym_data_statement_token1] = ACTIONS(1407),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1407),
    [aux_sym_parallel_statement_token1] = ACTIONS(1407),
    [aux_sym_sequence_statement_token1] = ACTIONS(1407),
    [anon_sym_AMP] = ACTIONS(1407),
    [aux_sym_command_name_token1] = ACTIONS(1407),
    [aux_sym_command_name_token2] = ACTIONS(1407),
    [aux_sym_command_name_token3] = ACTIONS(1407),
    [aux_sym_command_name_token4] = ACTIONS(1407),
    [aux_sym_command_name_token5] = ACTIONS(1407),
    [aux_sym_command_name_token6] = ACTIONS(1407),
    [aux_sym_command_name_token7] = ACTIONS(1407),
    [aux_sym_command_name_token8] = ACTIONS(1407),
    [aux_sym_command_name_token9] = ACTIONS(1407),
    [aux_sym_command_name_token10] = ACTIONS(1407),
    [aux_sym_command_name_token11] = ACTIONS(1407),
    [aux_sym_command_name_token12] = ACTIONS(1407),
    [aux_sym_command_name_token13] = ACTIONS(1407),
    [aux_sym_command_name_token14] = ACTIONS(1407),
    [aux_sym_command_name_token15] = ACTIONS(1407),
    [aux_sym_command_name_token16] = ACTIONS(1407),
    [aux_sym_command_name_token17] = ACTIONS(1407),
    [aux_sym_command_name_token18] = ACTIONS(1407),
    [aux_sym_class_statement_token1] = ACTIONS(1407),
    [aux_sym_enum_statement_token1] = ACTIONS(1407),
    [anon_sym_PLUS] = ACTIONS(1407),
    [anon_sym_DASH] = ACTIONS(1407),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1407),
    [anon_sym_BANG] = ACTIONS(1407),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1407),
    [anon_sym_PLUS_PLUS] = ACTIONS(1407),
    [anon_sym_DASH_DASH] = ACTIONS(1407),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1407),
    [anon_sym_AT_LPAREN] = ACTIONS(1407),
    [anon_sym_AT_LBRACE] = ACTIONS(1407),
  },
  [STATE(409)] = {
    [ts_builtin_sym_end] = ACTIONS(1537),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1409),
    [sym_hexadecimal_integer_literal] = ACTIONS(1409),
    [sym_real_literal] = ACTIONS(1409),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1409),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1409),
    [sym_verbatim_string_characters] = ACTIONS(1409),
    [sym_verbatim_here_string_characters] = ACTIONS(1409),
    [anon_sym_DOT] = ACTIONS(1409),
    [anon_sym_LBRACK] = ACTIONS(1409),
    [aux_sym_comparison_operator_token37] = ACTIONS(1409),
    [aux_sym_comparison_operator_token50] = ACTIONS(1409),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1409),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1409),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1409),
    [anon_sym_DOLLAR_] = ACTIONS(1409),
    [aux_sym_variable_token1] = ACTIONS(1409),
    [aux_sym_variable_token2] = ACTIONS(1409),
    [sym_braced_variable] = ACTIONS(1409),
    [anon_sym_SEMI] = ACTIONS(1409),
    [anon_sym_LPAREN] = ACTIONS(1409),
    [anon_sym_COMMA] = ACTIONS(1409),
    [anon_sym_LBRACE] = ACTIONS(1409),
    [aux_sym_if_statement_token1] = ACTIONS(1409),
    [aux_sym_switch_statement_token1] = ACTIONS(1409),
    [aux_sym_foreach_statement_token1] = ACTIONS(1409),
    [aux_sym_for_statement_token1] = ACTIONS(1409),
    [aux_sym_while_statement_token1] = ACTIONS(1409),
    [aux_sym_do_statement_token1] = ACTIONS(1409),
    [aux_sym_function_statement_token1] = ACTIONS(1409),
    [aux_sym_function_statement_token2] = ACTIONS(1409),
    [aux_sym_function_statement_token3] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1409),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1409),
    [sym_label] = ACTIONS(1409),
    [aux_sym_trap_statement_token1] = ACTIONS(1409),
    [aux_sym_try_statement_token1] = ACTIONS(1409),
    [aux_sym_data_statement_token1] = ACTIONS(1409),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1409),
    [aux_sym_parallel_statement_token1] = ACTIONS(1409),
    [aux_sym_sequence_statement_token1] = ACTIONS(1409),
    [anon_sym_AMP] = ACTIONS(1409),
    [aux_sym_command_name_token1] = ACTIONS(1409),
    [aux_sym_command_name_token2] = ACTIONS(1409),
    [aux_sym_command_name_token3] = ACTIONS(1409),
    [aux_sym_command_name_token4] = ACTIONS(1409),
    [aux_sym_command_name_token5] = ACTIONS(1409),
    [aux_sym_command_name_token6] = ACTIONS(1409),
    [aux_sym_command_name_token7] = ACTIONS(1409),
    [aux_sym_command_name_token8] = ACTIONS(1409),
    [aux_sym_command_name_token9] = ACTIONS(1409),
    [aux_sym_command_name_token10] = ACTIONS(1409),
    [aux_sym_command_name_token11] = ACTIONS(1409),
    [aux_sym_command_name_token12] = ACTIONS(1409),
    [aux_sym_command_name_token13] = ACTIONS(1409),
    [aux_sym_command_name_token14] = ACTIONS(1409),
    [aux_sym_command_name_token15] = ACTIONS(1409),
    [aux_sym_command_name_token16] = ACTIONS(1409),
    [aux_sym_command_name_token17] = ACTIONS(1409),
    [aux_sym_command_name_token18] = ACTIONS(1409),
    [aux_sym_class_statement_token1] = ACTIONS(1409),
    [aux_sym_enum_statement_token1] = ACTIONS(1409),
    [anon_sym_PLUS] = ACTIONS(1409),
    [anon_sym_DASH] = ACTIONS(1409),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1409),
    [anon_sym_BANG] = ACTIONS(1409),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1409),
    [anon_sym_PLUS_PLUS] = ACTIONS(1409),
    [anon_sym_DASH_DASH] = ACTIONS(1409),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1409),
    [anon_sym_AT_LPAREN] = ACTIONS(1409),
    [anon_sym_AT_LBRACE] = ACTIONS(1409),
  },
  [STATE(410)] = {
    [ts_builtin_sym_end] = ACTIONS(1539),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1413),
    [sym_hexadecimal_integer_literal] = ACTIONS(1413),
    [sym_real_literal] = ACTIONS(1413),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1413),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1413),
    [sym_verbatim_string_characters] = ACTIONS(1413),
    [sym_verbatim_here_string_characters] = ACTIONS(1413),
    [anon_sym_DOT] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1413),
    [aux_sym_comparison_operator_token37] = ACTIONS(1413),
    [aux_sym_comparison_operator_token50] = ACTIONS(1413),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1413),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1413),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1413),
    [anon_sym_DOLLAR_] = ACTIONS(1413),
    [aux_sym_variable_token1] = ACTIONS(1413),
    [aux_sym_variable_token2] = ACTIONS(1413),
    [sym_braced_variable] = ACTIONS(1413),
    [anon_sym_SEMI] = ACTIONS(1413),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_COMMA] = ACTIONS(1413),
    [anon_sym_LBRACE] = ACTIONS(1413),
    [aux_sym_if_statement_token1] = ACTIONS(1413),
    [aux_sym_switch_statement_token1] = ACTIONS(1413),
    [aux_sym_foreach_statement_token1] = ACTIONS(1413),
    [aux_sym_for_statement_token1] = ACTIONS(1413),
    [aux_sym_while_statement_token1] = ACTIONS(1413),
    [aux_sym_do_statement_token1] = ACTIONS(1413),
    [aux_sym_function_statement_token1] = ACTIONS(1413),
    [aux_sym_function_statement_token2] = ACTIONS(1413),
    [aux_sym_function_statement_token3] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1413),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1413),
    [sym_label] = ACTIONS(1413),
    [aux_sym_trap_statement_token1] = ACTIONS(1413),
    [aux_sym_try_statement_token1] = ACTIONS(1413),
    [aux_sym_data_statement_token1] = ACTIONS(1413),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1413),
    [aux_sym_parallel_statement_token1] = ACTIONS(1413),
    [aux_sym_sequence_statement_token1] = ACTIONS(1413),
    [anon_sym_AMP] = ACTIONS(1413),
    [aux_sym_command_name_token1] = ACTIONS(1413),
    [aux_sym_command_name_token2] = ACTIONS(1413),
    [aux_sym_command_name_token3] = ACTIONS(1413),
    [aux_sym_command_name_token4] = ACTIONS(1413),
    [aux_sym_command_name_token5] = ACTIONS(1413),
    [aux_sym_command_name_token6] = ACTIONS(1413),
    [aux_sym_command_name_token7] = ACTIONS(1413),
    [aux_sym_command_name_token8] = ACTIONS(1413),
    [aux_sym_command_name_token9] = ACTIONS(1413),
    [aux_sym_command_name_token10] = ACTIONS(1413),
    [aux_sym_command_name_token11] = ACTIONS(1413),
    [aux_sym_command_name_token12] = ACTIONS(1413),
    [aux_sym_command_name_token13] = ACTIONS(1413),
    [aux_sym_command_name_token14] = ACTIONS(1413),
    [aux_sym_command_name_token15] = ACTIONS(1413),
    [aux_sym_command_name_token16] = ACTIONS(1413),
    [aux_sym_command_name_token17] = ACTIONS(1413),
    [aux_sym_command_name_token18] = ACTIONS(1413),
    [aux_sym_class_statement_token1] = ACTIONS(1413),
    [aux_sym_enum_statement_token1] = ACTIONS(1413),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1413),
    [anon_sym_PLUS_PLUS] = ACTIONS(1413),
    [anon_sym_DASH_DASH] = ACTIONS(1413),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1413),
    [anon_sym_AT_LPAREN] = ACTIONS(1413),
    [anon_sym_AT_LBRACE] = ACTIONS(1413),
  },
  [STATE(411)] = {
    [ts_builtin_sym_end] = ACTIONS(1541),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1443),
    [sym_hexadecimal_integer_literal] = ACTIONS(1443),
    [sym_real_literal] = ACTIONS(1443),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1443),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1443),
    [sym_verbatim_string_characters] = ACTIONS(1443),
    [sym_verbatim_here_string_characters] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1443),
    [anon_sym_LBRACK] = ACTIONS(1443),
    [aux_sym_comparison_operator_token37] = ACTIONS(1443),
    [aux_sym_comparison_operator_token50] = ACTIONS(1443),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1443),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1443),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1443),
    [anon_sym_DOLLAR_] = ACTIONS(1443),
    [aux_sym_variable_token1] = ACTIONS(1443),
    [aux_sym_variable_token2] = ACTIONS(1443),
    [sym_braced_variable] = ACTIONS(1443),
    [anon_sym_SEMI] = ACTIONS(1443),
    [anon_sym_LPAREN] = ACTIONS(1443),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1443),
    [aux_sym_if_statement_token1] = ACTIONS(1443),
    [aux_sym_switch_statement_token1] = ACTIONS(1443),
    [aux_sym_foreach_statement_token1] = ACTIONS(1443),
    [aux_sym_for_statement_token1] = ACTIONS(1443),
    [aux_sym_while_statement_token1] = ACTIONS(1443),
    [aux_sym_do_statement_token1] = ACTIONS(1443),
    [aux_sym_function_statement_token1] = ACTIONS(1443),
    [aux_sym_function_statement_token2] = ACTIONS(1443),
    [aux_sym_function_statement_token3] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1443),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1443),
    [sym_label] = ACTIONS(1443),
    [aux_sym_trap_statement_token1] = ACTIONS(1443),
    [aux_sym_try_statement_token1] = ACTIONS(1443),
    [aux_sym_data_statement_token1] = ACTIONS(1443),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1443),
    [aux_sym_parallel_statement_token1] = ACTIONS(1443),
    [aux_sym_sequence_statement_token1] = ACTIONS(1443),
    [anon_sym_AMP] = ACTIONS(1443),
    [aux_sym_command_name_token1] = ACTIONS(1443),
    [aux_sym_command_name_token2] = ACTIONS(1443),
    [aux_sym_command_name_token3] = ACTIONS(1443),
    [aux_sym_command_name_token4] = ACTIONS(1443),
    [aux_sym_command_name_token5] = ACTIONS(1443),
    [aux_sym_command_name_token6] = ACTIONS(1443),
    [aux_sym_command_name_token7] = ACTIONS(1443),
    [aux_sym_command_name_token8] = ACTIONS(1443),
    [aux_sym_command_name_token9] = ACTIONS(1443),
    [aux_sym_command_name_token10] = ACTIONS(1443),
    [aux_sym_command_name_token11] = ACTIONS(1443),
    [aux_sym_command_name_token12] = ACTIONS(1443),
    [aux_sym_command_name_token13] = ACTIONS(1443),
    [aux_sym_command_name_token14] = ACTIONS(1443),
    [aux_sym_command_name_token15] = ACTIONS(1443),
    [aux_sym_command_name_token16] = ACTIONS(1443),
    [aux_sym_command_name_token17] = ACTIONS(1443),
    [aux_sym_command_name_token18] = ACTIONS(1443),
    [aux_sym_class_statement_token1] = ACTIONS(1443),
    [aux_sym_enum_statement_token1] = ACTIONS(1443),
    [anon_sym_PLUS] = ACTIONS(1443),
    [anon_sym_DASH] = ACTIONS(1443),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1443),
    [anon_sym_BANG] = ACTIONS(1443),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1443),
    [anon_sym_PLUS_PLUS] = ACTIONS(1443),
    [anon_sym_DASH_DASH] = ACTIONS(1443),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1443),
    [anon_sym_AT_LPAREN] = ACTIONS(1443),
    [anon_sym_AT_LBRACE] = ACTIONS(1443),
  },
  [STATE(412)] = {
    [ts_builtin_sym_end] = ACTIONS(1543),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1445),
    [sym_hexadecimal_integer_literal] = ACTIONS(1445),
    [sym_real_literal] = ACTIONS(1445),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1445),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1445),
    [sym_verbatim_string_characters] = ACTIONS(1445),
    [sym_verbatim_here_string_characters] = ACTIONS(1445),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACK] = ACTIONS(1445),
    [aux_sym_comparison_operator_token37] = ACTIONS(1445),
    [aux_sym_comparison_operator_token50] = ACTIONS(1445),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1445),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1445),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1445),
    [anon_sym_DOLLAR_] = ACTIONS(1445),
    [aux_sym_variable_token1] = ACTIONS(1445),
    [aux_sym_variable_token2] = ACTIONS(1445),
    [sym_braced_variable] = ACTIONS(1445),
    [anon_sym_SEMI] = ACTIONS(1445),
    [anon_sym_LPAREN] = ACTIONS(1445),
    [anon_sym_COMMA] = ACTIONS(1445),
    [anon_sym_LBRACE] = ACTIONS(1445),
    [aux_sym_if_statement_token1] = ACTIONS(1445),
    [aux_sym_switch_statement_token1] = ACTIONS(1445),
    [aux_sym_foreach_statement_token1] = ACTIONS(1445),
    [aux_sym_for_statement_token1] = ACTIONS(1445),
    [aux_sym_while_statement_token1] = ACTIONS(1445),
    [aux_sym_do_statement_token1] = ACTIONS(1445),
    [aux_sym_function_statement_token1] = ACTIONS(1445),
    [aux_sym_function_statement_token2] = ACTIONS(1445),
    [aux_sym_function_statement_token3] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1445),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1445),
    [sym_label] = ACTIONS(1445),
    [aux_sym_trap_statement_token1] = ACTIONS(1445),
    [aux_sym_try_statement_token1] = ACTIONS(1445),
    [aux_sym_data_statement_token1] = ACTIONS(1445),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1445),
    [aux_sym_parallel_statement_token1] = ACTIONS(1445),
    [aux_sym_sequence_statement_token1] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(1445),
    [aux_sym_command_name_token1] = ACTIONS(1445),
    [aux_sym_command_name_token2] = ACTIONS(1445),
    [aux_sym_command_name_token3] = ACTIONS(1445),
    [aux_sym_command_name_token4] = ACTIONS(1445),
    [aux_sym_command_name_token5] = ACTIONS(1445),
    [aux_sym_command_name_token6] = ACTIONS(1445),
    [aux_sym_command_name_token7] = ACTIONS(1445),
    [aux_sym_command_name_token8] = ACTIONS(1445),
    [aux_sym_command_name_token9] = ACTIONS(1445),
    [aux_sym_command_name_token10] = ACTIONS(1445),
    [aux_sym_command_name_token11] = ACTIONS(1445),
    [aux_sym_command_name_token12] = ACTIONS(1445),
    [aux_sym_command_name_token13] = ACTIONS(1445),
    [aux_sym_command_name_token14] = ACTIONS(1445),
    [aux_sym_command_name_token15] = ACTIONS(1445),
    [aux_sym_command_name_token16] = ACTIONS(1445),
    [aux_sym_command_name_token17] = ACTIONS(1445),
    [aux_sym_command_name_token18] = ACTIONS(1445),
    [aux_sym_class_statement_token1] = ACTIONS(1445),
    [aux_sym_enum_statement_token1] = ACTIONS(1445),
    [anon_sym_PLUS] = ACTIONS(1445),
    [anon_sym_DASH] = ACTIONS(1445),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1445),
    [anon_sym_BANG] = ACTIONS(1445),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1445),
    [anon_sym_PLUS_PLUS] = ACTIONS(1445),
    [anon_sym_DASH_DASH] = ACTIONS(1445),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1445),
    [anon_sym_AT_LPAREN] = ACTIONS(1445),
    [anon_sym_AT_LBRACE] = ACTIONS(1445),
  },
  [STATE(413)] = {
    [ts_builtin_sym_end] = ACTIONS(1545),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1435),
    [sym_hexadecimal_integer_literal] = ACTIONS(1435),
    [sym_real_literal] = ACTIONS(1435),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1435),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1435),
    [sym_verbatim_string_characters] = ACTIONS(1435),
    [sym_verbatim_here_string_characters] = ACTIONS(1435),
    [anon_sym_DOT] = ACTIONS(1435),
    [anon_sym_LBRACK] = ACTIONS(1435),
    [aux_sym_comparison_operator_token37] = ACTIONS(1435),
    [aux_sym_comparison_operator_token50] = ACTIONS(1435),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1435),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1435),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1435),
    [anon_sym_DOLLAR_] = ACTIONS(1435),
    [aux_sym_variable_token1] = ACTIONS(1435),
    [aux_sym_variable_token2] = ACTIONS(1435),
    [sym_braced_variable] = ACTIONS(1435),
    [anon_sym_SEMI] = ACTIONS(1435),
    [anon_sym_LPAREN] = ACTIONS(1435),
    [anon_sym_COMMA] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1435),
    [aux_sym_if_statement_token1] = ACTIONS(1435),
    [aux_sym_switch_statement_token1] = ACTIONS(1435),
    [aux_sym_foreach_statement_token1] = ACTIONS(1435),
    [aux_sym_for_statement_token1] = ACTIONS(1435),
    [aux_sym_while_statement_token1] = ACTIONS(1435),
    [aux_sym_do_statement_token1] = ACTIONS(1435),
    [aux_sym_function_statement_token1] = ACTIONS(1435),
    [aux_sym_function_statement_token2] = ACTIONS(1435),
    [aux_sym_function_statement_token3] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1435),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1435),
    [sym_label] = ACTIONS(1435),
    [aux_sym_trap_statement_token1] = ACTIONS(1435),
    [aux_sym_try_statement_token1] = ACTIONS(1435),
    [aux_sym_data_statement_token1] = ACTIONS(1435),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1435),
    [aux_sym_parallel_statement_token1] = ACTIONS(1435),
    [aux_sym_sequence_statement_token1] = ACTIONS(1435),
    [anon_sym_AMP] = ACTIONS(1435),
    [aux_sym_command_name_token1] = ACTIONS(1435),
    [aux_sym_command_name_token2] = ACTIONS(1435),
    [aux_sym_command_name_token3] = ACTIONS(1435),
    [aux_sym_command_name_token4] = ACTIONS(1435),
    [aux_sym_command_name_token5] = ACTIONS(1435),
    [aux_sym_command_name_token6] = ACTIONS(1435),
    [aux_sym_command_name_token7] = ACTIONS(1435),
    [aux_sym_command_name_token8] = ACTIONS(1435),
    [aux_sym_command_name_token9] = ACTIONS(1435),
    [aux_sym_command_name_token10] = ACTIONS(1435),
    [aux_sym_command_name_token11] = ACTIONS(1435),
    [aux_sym_command_name_token12] = ACTIONS(1435),
    [aux_sym_command_name_token13] = ACTIONS(1435),
    [aux_sym_command_name_token14] = ACTIONS(1435),
    [aux_sym_command_name_token15] = ACTIONS(1435),
    [aux_sym_command_name_token16] = ACTIONS(1435),
    [aux_sym_command_name_token17] = ACTIONS(1435),
    [aux_sym_command_name_token18] = ACTIONS(1435),
    [aux_sym_class_statement_token1] = ACTIONS(1435),
    [aux_sym_enum_statement_token1] = ACTIONS(1435),
    [anon_sym_PLUS] = ACTIONS(1435),
    [anon_sym_DASH] = ACTIONS(1435),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1435),
    [anon_sym_BANG] = ACTIONS(1435),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1435),
    [anon_sym_PLUS_PLUS] = ACTIONS(1435),
    [anon_sym_DASH_DASH] = ACTIONS(1435),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1435),
    [anon_sym_AT_LPAREN] = ACTIONS(1435),
    [anon_sym_AT_LBRACE] = ACTIONS(1435),
  },
  [STATE(414)] = {
    [ts_builtin_sym_end] = ACTIONS(1547),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1455),
    [sym_hexadecimal_integer_literal] = ACTIONS(1455),
    [sym_real_literal] = ACTIONS(1455),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1455),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1455),
    [sym_verbatim_string_characters] = ACTIONS(1455),
    [sym_verbatim_here_string_characters] = ACTIONS(1455),
    [anon_sym_DOT] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1455),
    [aux_sym_comparison_operator_token37] = ACTIONS(1455),
    [aux_sym_comparison_operator_token50] = ACTIONS(1455),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1455),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1455),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1455),
    [anon_sym_DOLLAR_] = ACTIONS(1455),
    [aux_sym_variable_token1] = ACTIONS(1455),
    [aux_sym_variable_token2] = ACTIONS(1455),
    [sym_braced_variable] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1455),
    [anon_sym_LBRACE] = ACTIONS(1455),
    [aux_sym_if_statement_token1] = ACTIONS(1455),
    [aux_sym_switch_statement_token1] = ACTIONS(1455),
    [aux_sym_foreach_statement_token1] = ACTIONS(1455),
    [aux_sym_for_statement_token1] = ACTIONS(1455),
    [aux_sym_while_statement_token1] = ACTIONS(1455),
    [aux_sym_do_statement_token1] = ACTIONS(1455),
    [aux_sym_function_statement_token1] = ACTIONS(1455),
    [aux_sym_function_statement_token2] = ACTIONS(1455),
    [aux_sym_function_statement_token3] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1455),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1455),
    [sym_label] = ACTIONS(1455),
    [aux_sym_trap_statement_token1] = ACTIONS(1455),
    [aux_sym_try_statement_token1] = ACTIONS(1455),
    [aux_sym_data_statement_token1] = ACTIONS(1455),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1455),
    [aux_sym_parallel_statement_token1] = ACTIONS(1455),
    [aux_sym_sequence_statement_token1] = ACTIONS(1455),
    [anon_sym_AMP] = ACTIONS(1455),
    [aux_sym_command_name_token1] = ACTIONS(1455),
    [aux_sym_command_name_token2] = ACTIONS(1455),
    [aux_sym_command_name_token3] = ACTIONS(1455),
    [aux_sym_command_name_token4] = ACTIONS(1455),
    [aux_sym_command_name_token5] = ACTIONS(1455),
    [aux_sym_command_name_token6] = ACTIONS(1455),
    [aux_sym_command_name_token7] = ACTIONS(1455),
    [aux_sym_command_name_token8] = ACTIONS(1455),
    [aux_sym_command_name_token9] = ACTIONS(1455),
    [aux_sym_command_name_token10] = ACTIONS(1455),
    [aux_sym_command_name_token11] = ACTIONS(1455),
    [aux_sym_command_name_token12] = ACTIONS(1455),
    [aux_sym_command_name_token13] = ACTIONS(1455),
    [aux_sym_command_name_token14] = ACTIONS(1455),
    [aux_sym_command_name_token15] = ACTIONS(1455),
    [aux_sym_command_name_token16] = ACTIONS(1455),
    [aux_sym_command_name_token17] = ACTIONS(1455),
    [aux_sym_command_name_token18] = ACTIONS(1455),
    [aux_sym_class_statement_token1] = ACTIONS(1455),
    [aux_sym_enum_statement_token1] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1455),
    [anon_sym_BANG] = ACTIONS(1455),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1455),
    [anon_sym_DASH_DASH] = ACTIONS(1455),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1455),
    [anon_sym_AT_LPAREN] = ACTIONS(1455),
    [anon_sym_AT_LBRACE] = ACTIONS(1455),
  },
  [STATE(415)] = {
    [ts_builtin_sym_end] = ACTIONS(1549),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1437),
    [sym_hexadecimal_integer_literal] = ACTIONS(1437),
    [sym_real_literal] = ACTIONS(1437),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1437),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1437),
    [sym_verbatim_string_characters] = ACTIONS(1437),
    [sym_verbatim_here_string_characters] = ACTIONS(1437),
    [anon_sym_DOT] = ACTIONS(1437),
    [anon_sym_LBRACK] = ACTIONS(1437),
    [aux_sym_comparison_operator_token37] = ACTIONS(1437),
    [aux_sym_comparison_operator_token50] = ACTIONS(1437),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1437),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1437),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1437),
    [anon_sym_DOLLAR_] = ACTIONS(1437),
    [aux_sym_variable_token1] = ACTIONS(1437),
    [aux_sym_variable_token2] = ACTIONS(1437),
    [sym_braced_variable] = ACTIONS(1437),
    [anon_sym_SEMI] = ACTIONS(1437),
    [anon_sym_LPAREN] = ACTIONS(1437),
    [anon_sym_COMMA] = ACTIONS(1437),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [aux_sym_if_statement_token1] = ACTIONS(1437),
    [aux_sym_switch_statement_token1] = ACTIONS(1437),
    [aux_sym_foreach_statement_token1] = ACTIONS(1437),
    [aux_sym_for_statement_token1] = ACTIONS(1437),
    [aux_sym_while_statement_token1] = ACTIONS(1437),
    [aux_sym_do_statement_token1] = ACTIONS(1437),
    [aux_sym_function_statement_token1] = ACTIONS(1437),
    [aux_sym_function_statement_token2] = ACTIONS(1437),
    [aux_sym_function_statement_token3] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1437),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1437),
    [sym_label] = ACTIONS(1437),
    [aux_sym_trap_statement_token1] = ACTIONS(1437),
    [aux_sym_try_statement_token1] = ACTIONS(1437),
    [aux_sym_data_statement_token1] = ACTIONS(1437),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1437),
    [aux_sym_parallel_statement_token1] = ACTIONS(1437),
    [aux_sym_sequence_statement_token1] = ACTIONS(1437),
    [anon_sym_AMP] = ACTIONS(1437),
    [aux_sym_command_name_token1] = ACTIONS(1437),
    [aux_sym_command_name_token2] = ACTIONS(1437),
    [aux_sym_command_name_token3] = ACTIONS(1437),
    [aux_sym_command_name_token4] = ACTIONS(1437),
    [aux_sym_command_name_token5] = ACTIONS(1437),
    [aux_sym_command_name_token6] = ACTIONS(1437),
    [aux_sym_command_name_token7] = ACTIONS(1437),
    [aux_sym_command_name_token8] = ACTIONS(1437),
    [aux_sym_command_name_token9] = ACTIONS(1437),
    [aux_sym_command_name_token10] = ACTIONS(1437),
    [aux_sym_command_name_token11] = ACTIONS(1437),
    [aux_sym_command_name_token12] = ACTIONS(1437),
    [aux_sym_command_name_token13] = ACTIONS(1437),
    [aux_sym_command_name_token14] = ACTIONS(1437),
    [aux_sym_command_name_token15] = ACTIONS(1437),
    [aux_sym_command_name_token16] = ACTIONS(1437),
    [aux_sym_command_name_token17] = ACTIONS(1437),
    [aux_sym_command_name_token18] = ACTIONS(1437),
    [aux_sym_class_statement_token1] = ACTIONS(1437),
    [aux_sym_enum_statement_token1] = ACTIONS(1437),
    [anon_sym_PLUS] = ACTIONS(1437),
    [anon_sym_DASH] = ACTIONS(1437),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1437),
    [anon_sym_BANG] = ACTIONS(1437),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1437),
    [anon_sym_PLUS_PLUS] = ACTIONS(1437),
    [anon_sym_DASH_DASH] = ACTIONS(1437),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1437),
    [anon_sym_AT_LPAREN] = ACTIONS(1437),
    [anon_sym_AT_LBRACE] = ACTIONS(1437),
  },
  [STATE(416)] = {
    [ts_builtin_sym_end] = ACTIONS(1551),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1411),
    [sym_hexadecimal_integer_literal] = ACTIONS(1411),
    [sym_real_literal] = ACTIONS(1411),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1411),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1411),
    [sym_verbatim_string_characters] = ACTIONS(1411),
    [sym_verbatim_here_string_characters] = ACTIONS(1411),
    [anon_sym_DOT] = ACTIONS(1411),
    [anon_sym_LBRACK] = ACTIONS(1411),
    [aux_sym_comparison_operator_token37] = ACTIONS(1411),
    [aux_sym_comparison_operator_token50] = ACTIONS(1411),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1411),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1411),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1411),
    [anon_sym_DOLLAR_] = ACTIONS(1411),
    [aux_sym_variable_token1] = ACTIONS(1411),
    [aux_sym_variable_token2] = ACTIONS(1411),
    [sym_braced_variable] = ACTIONS(1411),
    [anon_sym_SEMI] = ACTIONS(1411),
    [anon_sym_LPAREN] = ACTIONS(1411),
    [anon_sym_COMMA] = ACTIONS(1411),
    [anon_sym_LBRACE] = ACTIONS(1411),
    [aux_sym_if_statement_token1] = ACTIONS(1411),
    [aux_sym_switch_statement_token1] = ACTIONS(1411),
    [aux_sym_foreach_statement_token1] = ACTIONS(1411),
    [aux_sym_for_statement_token1] = ACTIONS(1411),
    [aux_sym_while_statement_token1] = ACTIONS(1411),
    [aux_sym_do_statement_token1] = ACTIONS(1411),
    [aux_sym_function_statement_token1] = ACTIONS(1411),
    [aux_sym_function_statement_token2] = ACTIONS(1411),
    [aux_sym_function_statement_token3] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1411),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1411),
    [sym_label] = ACTIONS(1411),
    [aux_sym_trap_statement_token1] = ACTIONS(1411),
    [aux_sym_try_statement_token1] = ACTIONS(1411),
    [aux_sym_data_statement_token1] = ACTIONS(1411),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1411),
    [aux_sym_parallel_statement_token1] = ACTIONS(1411),
    [aux_sym_sequence_statement_token1] = ACTIONS(1411),
    [anon_sym_AMP] = ACTIONS(1411),
    [aux_sym_command_name_token1] = ACTIONS(1411),
    [aux_sym_command_name_token2] = ACTIONS(1411),
    [aux_sym_command_name_token3] = ACTIONS(1411),
    [aux_sym_command_name_token4] = ACTIONS(1411),
    [aux_sym_command_name_token5] = ACTIONS(1411),
    [aux_sym_command_name_token6] = ACTIONS(1411),
    [aux_sym_command_name_token7] = ACTIONS(1411),
    [aux_sym_command_name_token8] = ACTIONS(1411),
    [aux_sym_command_name_token9] = ACTIONS(1411),
    [aux_sym_command_name_token10] = ACTIONS(1411),
    [aux_sym_command_name_token11] = ACTIONS(1411),
    [aux_sym_command_name_token12] = ACTIONS(1411),
    [aux_sym_command_name_token13] = ACTIONS(1411),
    [aux_sym_command_name_token14] = ACTIONS(1411),
    [aux_sym_command_name_token15] = ACTIONS(1411),
    [aux_sym_command_name_token16] = ACTIONS(1411),
    [aux_sym_command_name_token17] = ACTIONS(1411),
    [aux_sym_command_name_token18] = ACTIONS(1411),
    [aux_sym_class_statement_token1] = ACTIONS(1411),
    [aux_sym_enum_statement_token1] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1411),
    [anon_sym_DASH] = ACTIONS(1411),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1411),
    [anon_sym_BANG] = ACTIONS(1411),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1411),
    [anon_sym_PLUS_PLUS] = ACTIONS(1411),
    [anon_sym_DASH_DASH] = ACTIONS(1411),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1411),
    [anon_sym_AT_LPAREN] = ACTIONS(1411),
    [anon_sym_AT_LBRACE] = ACTIONS(1411),
  },
  [STATE(417)] = {
    [ts_builtin_sym_end] = ACTIONS(1553),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1459),
    [sym_hexadecimal_integer_literal] = ACTIONS(1459),
    [sym_real_literal] = ACTIONS(1459),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1459),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1459),
    [sym_verbatim_string_characters] = ACTIONS(1459),
    [sym_verbatim_here_string_characters] = ACTIONS(1459),
    [anon_sym_DOT] = ACTIONS(1459),
    [anon_sym_LBRACK] = ACTIONS(1459),
    [aux_sym_comparison_operator_token37] = ACTIONS(1459),
    [aux_sym_comparison_operator_token50] = ACTIONS(1459),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1459),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1459),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1459),
    [anon_sym_DOLLAR_] = ACTIONS(1459),
    [aux_sym_variable_token1] = ACTIONS(1459),
    [aux_sym_variable_token2] = ACTIONS(1459),
    [sym_braced_variable] = ACTIONS(1459),
    [anon_sym_SEMI] = ACTIONS(1459),
    [anon_sym_LPAREN] = ACTIONS(1459),
    [anon_sym_COMMA] = ACTIONS(1459),
    [anon_sym_LBRACE] = ACTIONS(1459),
    [aux_sym_if_statement_token1] = ACTIONS(1459),
    [aux_sym_switch_statement_token1] = ACTIONS(1459),
    [aux_sym_foreach_statement_token1] = ACTIONS(1459),
    [aux_sym_for_statement_token1] = ACTIONS(1459),
    [aux_sym_while_statement_token1] = ACTIONS(1459),
    [aux_sym_do_statement_token1] = ACTIONS(1459),
    [aux_sym_function_statement_token1] = ACTIONS(1459),
    [aux_sym_function_statement_token2] = ACTIONS(1459),
    [aux_sym_function_statement_token3] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1459),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1459),
    [sym_label] = ACTIONS(1459),
    [aux_sym_trap_statement_token1] = ACTIONS(1459),
    [aux_sym_try_statement_token1] = ACTIONS(1459),
    [aux_sym_data_statement_token1] = ACTIONS(1459),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1459),
    [aux_sym_parallel_statement_token1] = ACTIONS(1459),
    [aux_sym_sequence_statement_token1] = ACTIONS(1459),
    [anon_sym_AMP] = ACTIONS(1459),
    [aux_sym_command_name_token1] = ACTIONS(1459),
    [aux_sym_command_name_token2] = ACTIONS(1459),
    [aux_sym_command_name_token3] = ACTIONS(1459),
    [aux_sym_command_name_token4] = ACTIONS(1459),
    [aux_sym_command_name_token5] = ACTIONS(1459),
    [aux_sym_command_name_token6] = ACTIONS(1459),
    [aux_sym_command_name_token7] = ACTIONS(1459),
    [aux_sym_command_name_token8] = ACTIONS(1459),
    [aux_sym_command_name_token9] = ACTIONS(1459),
    [aux_sym_command_name_token10] = ACTIONS(1459),
    [aux_sym_command_name_token11] = ACTIONS(1459),
    [aux_sym_command_name_token12] = ACTIONS(1459),
    [aux_sym_command_name_token13] = ACTIONS(1459),
    [aux_sym_command_name_token14] = ACTIONS(1459),
    [aux_sym_command_name_token15] = ACTIONS(1459),
    [aux_sym_command_name_token16] = ACTIONS(1459),
    [aux_sym_command_name_token17] = ACTIONS(1459),
    [aux_sym_command_name_token18] = ACTIONS(1459),
    [aux_sym_class_statement_token1] = ACTIONS(1459),
    [aux_sym_enum_statement_token1] = ACTIONS(1459),
    [anon_sym_PLUS] = ACTIONS(1459),
    [anon_sym_DASH] = ACTIONS(1459),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1459),
    [anon_sym_BANG] = ACTIONS(1459),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1459),
    [anon_sym_PLUS_PLUS] = ACTIONS(1459),
    [anon_sym_DASH_DASH] = ACTIONS(1459),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1459),
    [anon_sym_AT_LPAREN] = ACTIONS(1459),
    [anon_sym_AT_LBRACE] = ACTIONS(1459),
  },
  [STATE(418)] = {
    [ts_builtin_sym_end] = ACTIONS(1555),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1469),
    [sym_hexadecimal_integer_literal] = ACTIONS(1469),
    [sym_real_literal] = ACTIONS(1469),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1469),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1469),
    [sym_verbatim_string_characters] = ACTIONS(1469),
    [sym_verbatim_here_string_characters] = ACTIONS(1469),
    [anon_sym_DOT] = ACTIONS(1469),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_comparison_operator_token37] = ACTIONS(1469),
    [aux_sym_comparison_operator_token50] = ACTIONS(1469),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1469),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1469),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1469),
    [anon_sym_DOLLAR_] = ACTIONS(1469),
    [aux_sym_variable_token1] = ACTIONS(1469),
    [aux_sym_variable_token2] = ACTIONS(1469),
    [sym_braced_variable] = ACTIONS(1469),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [aux_sym_if_statement_token1] = ACTIONS(1469),
    [aux_sym_switch_statement_token1] = ACTIONS(1469),
    [aux_sym_foreach_statement_token1] = ACTIONS(1469),
    [aux_sym_for_statement_token1] = ACTIONS(1469),
    [aux_sym_while_statement_token1] = ACTIONS(1469),
    [aux_sym_do_statement_token1] = ACTIONS(1469),
    [aux_sym_function_statement_token1] = ACTIONS(1469),
    [aux_sym_function_statement_token2] = ACTIONS(1469),
    [aux_sym_function_statement_token3] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1469),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1469),
    [sym_label] = ACTIONS(1469),
    [aux_sym_trap_statement_token1] = ACTIONS(1469),
    [aux_sym_try_statement_token1] = ACTIONS(1469),
    [aux_sym_data_statement_token1] = ACTIONS(1469),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1469),
    [aux_sym_parallel_statement_token1] = ACTIONS(1469),
    [aux_sym_sequence_statement_token1] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1469),
    [aux_sym_command_name_token1] = ACTIONS(1469),
    [aux_sym_command_name_token2] = ACTIONS(1469),
    [aux_sym_command_name_token3] = ACTIONS(1469),
    [aux_sym_command_name_token4] = ACTIONS(1469),
    [aux_sym_command_name_token5] = ACTIONS(1469),
    [aux_sym_command_name_token6] = ACTIONS(1469),
    [aux_sym_command_name_token7] = ACTIONS(1469),
    [aux_sym_command_name_token8] = ACTIONS(1469),
    [aux_sym_command_name_token9] = ACTIONS(1469),
    [aux_sym_command_name_token10] = ACTIONS(1469),
    [aux_sym_command_name_token11] = ACTIONS(1469),
    [aux_sym_command_name_token12] = ACTIONS(1469),
    [aux_sym_command_name_token13] = ACTIONS(1469),
    [aux_sym_command_name_token14] = ACTIONS(1469),
    [aux_sym_command_name_token15] = ACTIONS(1469),
    [aux_sym_command_name_token16] = ACTIONS(1469),
    [aux_sym_command_name_token17] = ACTIONS(1469),
    [aux_sym_command_name_token18] = ACTIONS(1469),
    [aux_sym_class_statement_token1] = ACTIONS(1469),
    [aux_sym_enum_statement_token1] = ACTIONS(1469),
    [anon_sym_PLUS] = ACTIONS(1469),
    [anon_sym_DASH] = ACTIONS(1469),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1469),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1469),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1469),
    [anon_sym_AT_LPAREN] = ACTIONS(1469),
    [anon_sym_AT_LBRACE] = ACTIONS(1469),
  },
  [STATE(419)] = {
    [ts_builtin_sym_end] = ACTIONS(1557),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1415),
    [sym_hexadecimal_integer_literal] = ACTIONS(1415),
    [sym_real_literal] = ACTIONS(1415),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1415),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1415),
    [sym_verbatim_string_characters] = ACTIONS(1415),
    [sym_verbatim_here_string_characters] = ACTIONS(1415),
    [anon_sym_DOT] = ACTIONS(1415),
    [anon_sym_LBRACK] = ACTIONS(1415),
    [aux_sym_comparison_operator_token37] = ACTIONS(1415),
    [aux_sym_comparison_operator_token50] = ACTIONS(1415),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1415),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1415),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1415),
    [anon_sym_DOLLAR_] = ACTIONS(1415),
    [aux_sym_variable_token1] = ACTIONS(1415),
    [aux_sym_variable_token2] = ACTIONS(1415),
    [sym_braced_variable] = ACTIONS(1415),
    [anon_sym_SEMI] = ACTIONS(1415),
    [anon_sym_LPAREN] = ACTIONS(1415),
    [anon_sym_COMMA] = ACTIONS(1415),
    [anon_sym_LBRACE] = ACTIONS(1415),
    [aux_sym_if_statement_token1] = ACTIONS(1415),
    [aux_sym_switch_statement_token1] = ACTIONS(1415),
    [aux_sym_foreach_statement_token1] = ACTIONS(1415),
    [aux_sym_for_statement_token1] = ACTIONS(1415),
    [aux_sym_while_statement_token1] = ACTIONS(1415),
    [aux_sym_do_statement_token1] = ACTIONS(1415),
    [aux_sym_function_statement_token1] = ACTIONS(1415),
    [aux_sym_function_statement_token2] = ACTIONS(1415),
    [aux_sym_function_statement_token3] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1415),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1415),
    [sym_label] = ACTIONS(1415),
    [aux_sym_trap_statement_token1] = ACTIONS(1415),
    [aux_sym_try_statement_token1] = ACTIONS(1415),
    [aux_sym_data_statement_token1] = ACTIONS(1415),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1415),
    [aux_sym_parallel_statement_token1] = ACTIONS(1415),
    [aux_sym_sequence_statement_token1] = ACTIONS(1415),
    [anon_sym_AMP] = ACTIONS(1415),
    [aux_sym_command_name_token1] = ACTIONS(1415),
    [aux_sym_command_name_token2] = ACTIONS(1415),
    [aux_sym_command_name_token3] = ACTIONS(1415),
    [aux_sym_command_name_token4] = ACTIONS(1415),
    [aux_sym_command_name_token5] = ACTIONS(1415),
    [aux_sym_command_name_token6] = ACTIONS(1415),
    [aux_sym_command_name_token7] = ACTIONS(1415),
    [aux_sym_command_name_token8] = ACTIONS(1415),
    [aux_sym_command_name_token9] = ACTIONS(1415),
    [aux_sym_command_name_token10] = ACTIONS(1415),
    [aux_sym_command_name_token11] = ACTIONS(1415),
    [aux_sym_command_name_token12] = ACTIONS(1415),
    [aux_sym_command_name_token13] = ACTIONS(1415),
    [aux_sym_command_name_token14] = ACTIONS(1415),
    [aux_sym_command_name_token15] = ACTIONS(1415),
    [aux_sym_command_name_token16] = ACTIONS(1415),
    [aux_sym_command_name_token17] = ACTIONS(1415),
    [aux_sym_command_name_token18] = ACTIONS(1415),
    [aux_sym_class_statement_token1] = ACTIONS(1415),
    [aux_sym_enum_statement_token1] = ACTIONS(1415),
    [anon_sym_PLUS] = ACTIONS(1415),
    [anon_sym_DASH] = ACTIONS(1415),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1415),
    [anon_sym_BANG] = ACTIONS(1415),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1415),
    [anon_sym_PLUS_PLUS] = ACTIONS(1415),
    [anon_sym_DASH_DASH] = ACTIONS(1415),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1415),
    [anon_sym_AT_LPAREN] = ACTIONS(1415),
    [anon_sym_AT_LBRACE] = ACTIONS(1415),
  },
  [STATE(420)] = {
    [ts_builtin_sym_end] = ACTIONS(1381),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1365),
    [sym_hexadecimal_integer_literal] = ACTIONS(1365),
    [sym_real_literal] = ACTIONS(1365),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1365),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1365),
    [sym_verbatim_string_characters] = ACTIONS(1365),
    [sym_verbatim_here_string_characters] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1365),
    [anon_sym_LBRACK] = ACTIONS(1365),
    [aux_sym_comparison_operator_token37] = ACTIONS(1365),
    [aux_sym_comparison_operator_token50] = ACTIONS(1365),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1365),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1365),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1365),
    [anon_sym_DOLLAR_] = ACTIONS(1365),
    [aux_sym_variable_token1] = ACTIONS(1365),
    [aux_sym_variable_token2] = ACTIONS(1365),
    [sym_braced_variable] = ACTIONS(1365),
    [anon_sym_SEMI] = ACTIONS(1365),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [anon_sym_COMMA] = ACTIONS(1365),
    [anon_sym_LBRACE] = ACTIONS(1365),
    [aux_sym_if_statement_token1] = ACTIONS(1365),
    [aux_sym_switch_statement_token1] = ACTIONS(1365),
    [aux_sym_foreach_statement_token1] = ACTIONS(1365),
    [aux_sym_for_statement_token1] = ACTIONS(1365),
    [aux_sym_while_statement_token1] = ACTIONS(1365),
    [aux_sym_do_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token1] = ACTIONS(1365),
    [aux_sym_function_statement_token2] = ACTIONS(1365),
    [aux_sym_function_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1365),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1365),
    [sym_label] = ACTIONS(1365),
    [aux_sym_trap_statement_token1] = ACTIONS(1365),
    [aux_sym_try_statement_token1] = ACTIONS(1365),
    [aux_sym_data_statement_token1] = ACTIONS(1365),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1365),
    [aux_sym_parallel_statement_token1] = ACTIONS(1365),
    [aux_sym_sequence_statement_token1] = ACTIONS(1365),
    [anon_sym_AMP] = ACTIONS(1365),
    [aux_sym_command_name_token1] = ACTIONS(1365),
    [aux_sym_command_name_token2] = ACTIONS(1365),
    [aux_sym_command_name_token3] = ACTIONS(1365),
    [aux_sym_command_name_token4] = ACTIONS(1365),
    [aux_sym_command_name_token5] = ACTIONS(1365),
    [aux_sym_command_name_token6] = ACTIONS(1365),
    [aux_sym_command_name_token7] = ACTIONS(1365),
    [aux_sym_command_name_token8] = ACTIONS(1365),
    [aux_sym_command_name_token9] = ACTIONS(1365),
    [aux_sym_command_name_token10] = ACTIONS(1365),
    [aux_sym_command_name_token11] = ACTIONS(1365),
    [aux_sym_command_name_token12] = ACTIONS(1365),
    [aux_sym_command_name_token13] = ACTIONS(1365),
    [aux_sym_command_name_token14] = ACTIONS(1365),
    [aux_sym_command_name_token15] = ACTIONS(1365),
    [aux_sym_command_name_token16] = ACTIONS(1365),
    [aux_sym_command_name_token17] = ACTIONS(1365),
    [aux_sym_command_name_token18] = ACTIONS(1365),
    [aux_sym_class_statement_token1] = ACTIONS(1365),
    [aux_sym_enum_statement_token1] = ACTIONS(1365),
    [anon_sym_PLUS] = ACTIONS(1365),
    [anon_sym_DASH] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1365),
    [anon_sym_BANG] = ACTIONS(1365),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1365),
    [anon_sym_PLUS_PLUS] = ACTIONS(1365),
    [anon_sym_DASH_DASH] = ACTIONS(1365),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LPAREN] = ACTIONS(1365),
    [anon_sym_AT_LBRACE] = ACTIONS(1365),
  },
  [STATE(421)] = {
    [ts_builtin_sym_end] = ACTIONS(1379),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1363),
    [sym_hexadecimal_integer_literal] = ACTIONS(1363),
    [sym_real_literal] = ACTIONS(1363),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1363),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1363),
    [sym_verbatim_string_characters] = ACTIONS(1363),
    [sym_verbatim_here_string_characters] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [aux_sym_comparison_operator_token37] = ACTIONS(1363),
    [aux_sym_comparison_operator_token50] = ACTIONS(1363),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1363),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1363),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1363),
    [anon_sym_DOLLAR_] = ACTIONS(1363),
    [aux_sym_variable_token1] = ACTIONS(1363),
    [aux_sym_variable_token2] = ACTIONS(1363),
    [sym_braced_variable] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [aux_sym_if_statement_token1] = ACTIONS(1363),
    [aux_sym_switch_statement_token1] = ACTIONS(1363),
    [aux_sym_foreach_statement_token1] = ACTIONS(1363),
    [aux_sym_for_statement_token1] = ACTIONS(1363),
    [aux_sym_while_statement_token1] = ACTIONS(1363),
    [aux_sym_do_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token1] = ACTIONS(1363),
    [aux_sym_function_statement_token2] = ACTIONS(1363),
    [aux_sym_function_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1363),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1363),
    [sym_label] = ACTIONS(1363),
    [aux_sym_trap_statement_token1] = ACTIONS(1363),
    [aux_sym_try_statement_token1] = ACTIONS(1363),
    [aux_sym_data_statement_token1] = ACTIONS(1363),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1363),
    [aux_sym_parallel_statement_token1] = ACTIONS(1363),
    [aux_sym_sequence_statement_token1] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [aux_sym_command_name_token1] = ACTIONS(1363),
    [aux_sym_command_name_token2] = ACTIONS(1363),
    [aux_sym_command_name_token3] = ACTIONS(1363),
    [aux_sym_command_name_token4] = ACTIONS(1363),
    [aux_sym_command_name_token5] = ACTIONS(1363),
    [aux_sym_command_name_token6] = ACTIONS(1363),
    [aux_sym_command_name_token7] = ACTIONS(1363),
    [aux_sym_command_name_token8] = ACTIONS(1363),
    [aux_sym_command_name_token9] = ACTIONS(1363),
    [aux_sym_command_name_token10] = ACTIONS(1363),
    [aux_sym_command_name_token11] = ACTIONS(1363),
    [aux_sym_command_name_token12] = ACTIONS(1363),
    [aux_sym_command_name_token13] = ACTIONS(1363),
    [aux_sym_command_name_token14] = ACTIONS(1363),
    [aux_sym_command_name_token15] = ACTIONS(1363),
    [aux_sym_command_name_token16] = ACTIONS(1363),
    [aux_sym_command_name_token17] = ACTIONS(1363),
    [aux_sym_command_name_token18] = ACTIONS(1363),
    [aux_sym_class_statement_token1] = ACTIONS(1363),
    [aux_sym_enum_statement_token1] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1363),
    [anon_sym_BANG] = ACTIONS(1363),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1363),
    [anon_sym_PLUS_PLUS] = ACTIONS(1363),
    [anon_sym_DASH_DASH] = ACTIONS(1363),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LPAREN] = ACTIONS(1363),
    [anon_sym_AT_LBRACE] = ACTIONS(1363),
  },
  [STATE(422)] = {
    [ts_builtin_sym_end] = ACTIONS(1559),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1423),
    [sym_hexadecimal_integer_literal] = ACTIONS(1423),
    [sym_real_literal] = ACTIONS(1423),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1423),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1423),
    [sym_verbatim_string_characters] = ACTIONS(1423),
    [sym_verbatim_here_string_characters] = ACTIONS(1423),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1423),
    [aux_sym_comparison_operator_token37] = ACTIONS(1423),
    [aux_sym_comparison_operator_token50] = ACTIONS(1423),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1423),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1423),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1423),
    [anon_sym_DOLLAR_] = ACTIONS(1423),
    [aux_sym_variable_token1] = ACTIONS(1423),
    [aux_sym_variable_token2] = ACTIONS(1423),
    [sym_braced_variable] = ACTIONS(1423),
    [anon_sym_SEMI] = ACTIONS(1423),
    [anon_sym_LPAREN] = ACTIONS(1423),
    [anon_sym_COMMA] = ACTIONS(1423),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [aux_sym_if_statement_token1] = ACTIONS(1423),
    [aux_sym_switch_statement_token1] = ACTIONS(1423),
    [aux_sym_foreach_statement_token1] = ACTIONS(1423),
    [aux_sym_for_statement_token1] = ACTIONS(1423),
    [aux_sym_while_statement_token1] = ACTIONS(1423),
    [aux_sym_do_statement_token1] = ACTIONS(1423),
    [aux_sym_function_statement_token1] = ACTIONS(1423),
    [aux_sym_function_statement_token2] = ACTIONS(1423),
    [aux_sym_function_statement_token3] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1423),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1423),
    [sym_label] = ACTIONS(1423),
    [aux_sym_trap_statement_token1] = ACTIONS(1423),
    [aux_sym_try_statement_token1] = ACTIONS(1423),
    [aux_sym_data_statement_token1] = ACTIONS(1423),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1423),
    [aux_sym_parallel_statement_token1] = ACTIONS(1423),
    [aux_sym_sequence_statement_token1] = ACTIONS(1423),
    [anon_sym_AMP] = ACTIONS(1423),
    [aux_sym_command_name_token1] = ACTIONS(1423),
    [aux_sym_command_name_token2] = ACTIONS(1423),
    [aux_sym_command_name_token3] = ACTIONS(1423),
    [aux_sym_command_name_token4] = ACTIONS(1423),
    [aux_sym_command_name_token5] = ACTIONS(1423),
    [aux_sym_command_name_token6] = ACTIONS(1423),
    [aux_sym_command_name_token7] = ACTIONS(1423),
    [aux_sym_command_name_token8] = ACTIONS(1423),
    [aux_sym_command_name_token9] = ACTIONS(1423),
    [aux_sym_command_name_token10] = ACTIONS(1423),
    [aux_sym_command_name_token11] = ACTIONS(1423),
    [aux_sym_command_name_token12] = ACTIONS(1423),
    [aux_sym_command_name_token13] = ACTIONS(1423),
    [aux_sym_command_name_token14] = ACTIONS(1423),
    [aux_sym_command_name_token15] = ACTIONS(1423),
    [aux_sym_command_name_token16] = ACTIONS(1423),
    [aux_sym_command_name_token17] = ACTIONS(1423),
    [aux_sym_command_name_token18] = ACTIONS(1423),
    [aux_sym_class_statement_token1] = ACTIONS(1423),
    [aux_sym_enum_statement_token1] = ACTIONS(1423),
    [anon_sym_PLUS] = ACTIONS(1423),
    [anon_sym_DASH] = ACTIONS(1423),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1423),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1423),
    [anon_sym_PLUS_PLUS] = ACTIONS(1423),
    [anon_sym_DASH_DASH] = ACTIONS(1423),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1423),
    [anon_sym_AT_LPAREN] = ACTIONS(1423),
    [anon_sym_AT_LBRACE] = ACTIONS(1423),
  },
  [STATE(423)] = {
    [ts_builtin_sym_end] = ACTIONS(1561),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1447),
    [sym_hexadecimal_integer_literal] = ACTIONS(1447),
    [sym_real_literal] = ACTIONS(1447),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1447),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1447),
    [sym_verbatim_string_characters] = ACTIONS(1447),
    [sym_verbatim_here_string_characters] = ACTIONS(1447),
    [anon_sym_DOT] = ACTIONS(1447),
    [anon_sym_LBRACK] = ACTIONS(1447),
    [aux_sym_comparison_operator_token37] = ACTIONS(1447),
    [aux_sym_comparison_operator_token50] = ACTIONS(1447),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1447),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1447),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1447),
    [anon_sym_DOLLAR_] = ACTIONS(1447),
    [aux_sym_variable_token1] = ACTIONS(1447),
    [aux_sym_variable_token2] = ACTIONS(1447),
    [sym_braced_variable] = ACTIONS(1447),
    [anon_sym_SEMI] = ACTIONS(1447),
    [anon_sym_LPAREN] = ACTIONS(1447),
    [anon_sym_COMMA] = ACTIONS(1447),
    [anon_sym_LBRACE] = ACTIONS(1447),
    [aux_sym_if_statement_token1] = ACTIONS(1447),
    [aux_sym_switch_statement_token1] = ACTIONS(1447),
    [aux_sym_foreach_statement_token1] = ACTIONS(1447),
    [aux_sym_for_statement_token1] = ACTIONS(1447),
    [aux_sym_while_statement_token1] = ACTIONS(1447),
    [aux_sym_do_statement_token1] = ACTIONS(1447),
    [aux_sym_function_statement_token1] = ACTIONS(1447),
    [aux_sym_function_statement_token2] = ACTIONS(1447),
    [aux_sym_function_statement_token3] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1447),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1447),
    [sym_label] = ACTIONS(1447),
    [aux_sym_trap_statement_token1] = ACTIONS(1447),
    [aux_sym_try_statement_token1] = ACTIONS(1447),
    [aux_sym_data_statement_token1] = ACTIONS(1447),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1447),
    [aux_sym_parallel_statement_token1] = ACTIONS(1447),
    [aux_sym_sequence_statement_token1] = ACTIONS(1447),
    [anon_sym_AMP] = ACTIONS(1447),
    [aux_sym_command_name_token1] = ACTIONS(1447),
    [aux_sym_command_name_token2] = ACTIONS(1447),
    [aux_sym_command_name_token3] = ACTIONS(1447),
    [aux_sym_command_name_token4] = ACTIONS(1447),
    [aux_sym_command_name_token5] = ACTIONS(1447),
    [aux_sym_command_name_token6] = ACTIONS(1447),
    [aux_sym_command_name_token7] = ACTIONS(1447),
    [aux_sym_command_name_token8] = ACTIONS(1447),
    [aux_sym_command_name_token9] = ACTIONS(1447),
    [aux_sym_command_name_token10] = ACTIONS(1447),
    [aux_sym_command_name_token11] = ACTIONS(1447),
    [aux_sym_command_name_token12] = ACTIONS(1447),
    [aux_sym_command_name_token13] = ACTIONS(1447),
    [aux_sym_command_name_token14] = ACTIONS(1447),
    [aux_sym_command_name_token15] = ACTIONS(1447),
    [aux_sym_command_name_token16] = ACTIONS(1447),
    [aux_sym_command_name_token17] = ACTIONS(1447),
    [aux_sym_command_name_token18] = ACTIONS(1447),
    [aux_sym_class_statement_token1] = ACTIONS(1447),
    [aux_sym_enum_statement_token1] = ACTIONS(1447),
    [anon_sym_PLUS] = ACTIONS(1447),
    [anon_sym_DASH] = ACTIONS(1447),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1447),
    [anon_sym_BANG] = ACTIONS(1447),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1447),
    [anon_sym_PLUS_PLUS] = ACTIONS(1447),
    [anon_sym_DASH_DASH] = ACTIONS(1447),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1447),
    [anon_sym_AT_LPAREN] = ACTIONS(1447),
    [anon_sym_AT_LBRACE] = ACTIONS(1447),
  },
  [STATE(424)] = {
    [ts_builtin_sym_end] = ACTIONS(1563),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1385),
    [sym_hexadecimal_integer_literal] = ACTIONS(1385),
    [sym_real_literal] = ACTIONS(1385),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1385),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1385),
    [sym_verbatim_string_characters] = ACTIONS(1385),
    [sym_verbatim_here_string_characters] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [aux_sym_comparison_operator_token37] = ACTIONS(1385),
    [aux_sym_comparison_operator_token50] = ACTIONS(1385),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1385),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1385),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1385),
    [anon_sym_DOLLAR_] = ACTIONS(1385),
    [aux_sym_variable_token1] = ACTIONS(1385),
    [aux_sym_variable_token2] = ACTIONS(1385),
    [sym_braced_variable] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_COMMA] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [aux_sym_if_statement_token1] = ACTIONS(1385),
    [aux_sym_switch_statement_token1] = ACTIONS(1385),
    [aux_sym_foreach_statement_token1] = ACTIONS(1385),
    [aux_sym_for_statement_token1] = ACTIONS(1385),
    [aux_sym_while_statement_token1] = ACTIONS(1385),
    [aux_sym_do_statement_token1] = ACTIONS(1385),
    [aux_sym_function_statement_token1] = ACTIONS(1385),
    [aux_sym_function_statement_token2] = ACTIONS(1385),
    [aux_sym_function_statement_token3] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1385),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1385),
    [sym_label] = ACTIONS(1385),
    [aux_sym_trap_statement_token1] = ACTIONS(1385),
    [aux_sym_try_statement_token1] = ACTIONS(1385),
    [aux_sym_data_statement_token1] = ACTIONS(1385),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1385),
    [aux_sym_parallel_statement_token1] = ACTIONS(1385),
    [aux_sym_sequence_statement_token1] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [aux_sym_command_name_token1] = ACTIONS(1385),
    [aux_sym_command_name_token2] = ACTIONS(1385),
    [aux_sym_command_name_token3] = ACTIONS(1385),
    [aux_sym_command_name_token4] = ACTIONS(1385),
    [aux_sym_command_name_token5] = ACTIONS(1385),
    [aux_sym_command_name_token6] = ACTIONS(1385),
    [aux_sym_command_name_token7] = ACTIONS(1385),
    [aux_sym_command_name_token8] = ACTIONS(1385),
    [aux_sym_command_name_token9] = ACTIONS(1385),
    [aux_sym_command_name_token10] = ACTIONS(1385),
    [aux_sym_command_name_token11] = ACTIONS(1385),
    [aux_sym_command_name_token12] = ACTIONS(1385),
    [aux_sym_command_name_token13] = ACTIONS(1385),
    [aux_sym_command_name_token14] = ACTIONS(1385),
    [aux_sym_command_name_token15] = ACTIONS(1385),
    [aux_sym_command_name_token16] = ACTIONS(1385),
    [aux_sym_command_name_token17] = ACTIONS(1385),
    [aux_sym_command_name_token18] = ACTIONS(1385),
    [aux_sym_class_statement_token1] = ACTIONS(1385),
    [aux_sym_enum_statement_token1] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1385),
    [anon_sym_BANG] = ACTIONS(1385),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1385),
    [anon_sym_PLUS_PLUS] = ACTIONS(1385),
    [anon_sym_DASH_DASH] = ACTIONS(1385),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1385),
    [anon_sym_AT_LPAREN] = ACTIONS(1385),
    [anon_sym_AT_LBRACE] = ACTIONS(1385),
  },
  [STATE(425)] = {
    [ts_builtin_sym_end] = ACTIONS(1565),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1457),
    [sym_hexadecimal_integer_literal] = ACTIONS(1457),
    [sym_real_literal] = ACTIONS(1457),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1457),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1457),
    [sym_verbatim_string_characters] = ACTIONS(1457),
    [sym_verbatim_here_string_characters] = ACTIONS(1457),
    [anon_sym_DOT] = ACTIONS(1457),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_comparison_operator_token37] = ACTIONS(1457),
    [aux_sym_comparison_operator_token50] = ACTIONS(1457),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1457),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1457),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1457),
    [anon_sym_DOLLAR_] = ACTIONS(1457),
    [aux_sym_variable_token1] = ACTIONS(1457),
    [aux_sym_variable_token2] = ACTIONS(1457),
    [sym_braced_variable] = ACTIONS(1457),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [aux_sym_if_statement_token1] = ACTIONS(1457),
    [aux_sym_switch_statement_token1] = ACTIONS(1457),
    [aux_sym_foreach_statement_token1] = ACTIONS(1457),
    [aux_sym_for_statement_token1] = ACTIONS(1457),
    [aux_sym_while_statement_token1] = ACTIONS(1457),
    [aux_sym_do_statement_token1] = ACTIONS(1457),
    [aux_sym_function_statement_token1] = ACTIONS(1457),
    [aux_sym_function_statement_token2] = ACTIONS(1457),
    [aux_sym_function_statement_token3] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1457),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1457),
    [sym_label] = ACTIONS(1457),
    [aux_sym_trap_statement_token1] = ACTIONS(1457),
    [aux_sym_try_statement_token1] = ACTIONS(1457),
    [aux_sym_data_statement_token1] = ACTIONS(1457),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1457),
    [aux_sym_parallel_statement_token1] = ACTIONS(1457),
    [aux_sym_sequence_statement_token1] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1457),
    [aux_sym_command_name_token1] = ACTIONS(1457),
    [aux_sym_command_name_token2] = ACTIONS(1457),
    [aux_sym_command_name_token3] = ACTIONS(1457),
    [aux_sym_command_name_token4] = ACTIONS(1457),
    [aux_sym_command_name_token5] = ACTIONS(1457),
    [aux_sym_command_name_token6] = ACTIONS(1457),
    [aux_sym_command_name_token7] = ACTIONS(1457),
    [aux_sym_command_name_token8] = ACTIONS(1457),
    [aux_sym_command_name_token9] = ACTIONS(1457),
    [aux_sym_command_name_token10] = ACTIONS(1457),
    [aux_sym_command_name_token11] = ACTIONS(1457),
    [aux_sym_command_name_token12] = ACTIONS(1457),
    [aux_sym_command_name_token13] = ACTIONS(1457),
    [aux_sym_command_name_token14] = ACTIONS(1457),
    [aux_sym_command_name_token15] = ACTIONS(1457),
    [aux_sym_command_name_token16] = ACTIONS(1457),
    [aux_sym_command_name_token17] = ACTIONS(1457),
    [aux_sym_command_name_token18] = ACTIONS(1457),
    [aux_sym_class_statement_token1] = ACTIONS(1457),
    [aux_sym_enum_statement_token1] = ACTIONS(1457),
    [anon_sym_PLUS] = ACTIONS(1457),
    [anon_sym_DASH] = ACTIONS(1457),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1457),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1457),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1457),
    [anon_sym_AT_LPAREN] = ACTIONS(1457),
    [anon_sym_AT_LBRACE] = ACTIONS(1457),
  },
  [STATE(426)] = {
    [ts_builtin_sym_end] = ACTIONS(1567),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1461),
    [sym_hexadecimal_integer_literal] = ACTIONS(1461),
    [sym_real_literal] = ACTIONS(1461),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1461),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1461),
    [sym_verbatim_string_characters] = ACTIONS(1461),
    [sym_verbatim_here_string_characters] = ACTIONS(1461),
    [anon_sym_DOT] = ACTIONS(1461),
    [anon_sym_LBRACK] = ACTIONS(1461),
    [aux_sym_comparison_operator_token37] = ACTIONS(1461),
    [aux_sym_comparison_operator_token50] = ACTIONS(1461),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1461),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1461),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1461),
    [anon_sym_DOLLAR_] = ACTIONS(1461),
    [aux_sym_variable_token1] = ACTIONS(1461),
    [aux_sym_variable_token2] = ACTIONS(1461),
    [sym_braced_variable] = ACTIONS(1461),
    [anon_sym_SEMI] = ACTIONS(1461),
    [anon_sym_LPAREN] = ACTIONS(1461),
    [anon_sym_COMMA] = ACTIONS(1461),
    [anon_sym_LBRACE] = ACTIONS(1461),
    [aux_sym_if_statement_token1] = ACTIONS(1461),
    [aux_sym_switch_statement_token1] = ACTIONS(1461),
    [aux_sym_foreach_statement_token1] = ACTIONS(1461),
    [aux_sym_for_statement_token1] = ACTIONS(1461),
    [aux_sym_while_statement_token1] = ACTIONS(1461),
    [aux_sym_do_statement_token1] = ACTIONS(1461),
    [aux_sym_function_statement_token1] = ACTIONS(1461),
    [aux_sym_function_statement_token2] = ACTIONS(1461),
    [aux_sym_function_statement_token3] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1461),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1461),
    [sym_label] = ACTIONS(1461),
    [aux_sym_trap_statement_token1] = ACTIONS(1461),
    [aux_sym_try_statement_token1] = ACTIONS(1461),
    [aux_sym_data_statement_token1] = ACTIONS(1461),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1461),
    [aux_sym_parallel_statement_token1] = ACTIONS(1461),
    [aux_sym_sequence_statement_token1] = ACTIONS(1461),
    [anon_sym_AMP] = ACTIONS(1461),
    [aux_sym_command_name_token1] = ACTIONS(1461),
    [aux_sym_command_name_token2] = ACTIONS(1461),
    [aux_sym_command_name_token3] = ACTIONS(1461),
    [aux_sym_command_name_token4] = ACTIONS(1461),
    [aux_sym_command_name_token5] = ACTIONS(1461),
    [aux_sym_command_name_token6] = ACTIONS(1461),
    [aux_sym_command_name_token7] = ACTIONS(1461),
    [aux_sym_command_name_token8] = ACTIONS(1461),
    [aux_sym_command_name_token9] = ACTIONS(1461),
    [aux_sym_command_name_token10] = ACTIONS(1461),
    [aux_sym_command_name_token11] = ACTIONS(1461),
    [aux_sym_command_name_token12] = ACTIONS(1461),
    [aux_sym_command_name_token13] = ACTIONS(1461),
    [aux_sym_command_name_token14] = ACTIONS(1461),
    [aux_sym_command_name_token15] = ACTIONS(1461),
    [aux_sym_command_name_token16] = ACTIONS(1461),
    [aux_sym_command_name_token17] = ACTIONS(1461),
    [aux_sym_command_name_token18] = ACTIONS(1461),
    [aux_sym_class_statement_token1] = ACTIONS(1461),
    [aux_sym_enum_statement_token1] = ACTIONS(1461),
    [anon_sym_PLUS] = ACTIONS(1461),
    [anon_sym_DASH] = ACTIONS(1461),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1461),
    [anon_sym_BANG] = ACTIONS(1461),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1461),
    [anon_sym_PLUS_PLUS] = ACTIONS(1461),
    [anon_sym_DASH_DASH] = ACTIONS(1461),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1461),
    [anon_sym_AT_LPAREN] = ACTIONS(1461),
    [anon_sym_AT_LBRACE] = ACTIONS(1461),
  },
  [STATE(427)] = {
    [ts_builtin_sym_end] = ACTIONS(1569),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1433),
    [sym_hexadecimal_integer_literal] = ACTIONS(1433),
    [sym_real_literal] = ACTIONS(1433),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1433),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1433),
    [sym_verbatim_string_characters] = ACTIONS(1433),
    [sym_verbatim_here_string_characters] = ACTIONS(1433),
    [anon_sym_DOT] = ACTIONS(1433),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [aux_sym_comparison_operator_token37] = ACTIONS(1433),
    [aux_sym_comparison_operator_token50] = ACTIONS(1433),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1433),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1433),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1433),
    [anon_sym_DOLLAR_] = ACTIONS(1433),
    [aux_sym_variable_token1] = ACTIONS(1433),
    [aux_sym_variable_token2] = ACTIONS(1433),
    [sym_braced_variable] = ACTIONS(1433),
    [anon_sym_SEMI] = ACTIONS(1433),
    [anon_sym_LPAREN] = ACTIONS(1433),
    [anon_sym_COMMA] = ACTIONS(1433),
    [anon_sym_LBRACE] = ACTIONS(1433),
    [aux_sym_if_statement_token1] = ACTIONS(1433),
    [aux_sym_switch_statement_token1] = ACTIONS(1433),
    [aux_sym_foreach_statement_token1] = ACTIONS(1433),
    [aux_sym_for_statement_token1] = ACTIONS(1433),
    [aux_sym_while_statement_token1] = ACTIONS(1433),
    [aux_sym_do_statement_token1] = ACTIONS(1433),
    [aux_sym_function_statement_token1] = ACTIONS(1433),
    [aux_sym_function_statement_token2] = ACTIONS(1433),
    [aux_sym_function_statement_token3] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1433),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1433),
    [sym_label] = ACTIONS(1433),
    [aux_sym_trap_statement_token1] = ACTIONS(1433),
    [aux_sym_try_statement_token1] = ACTIONS(1433),
    [aux_sym_data_statement_token1] = ACTIONS(1433),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1433),
    [aux_sym_parallel_statement_token1] = ACTIONS(1433),
    [aux_sym_sequence_statement_token1] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1433),
    [aux_sym_command_name_token1] = ACTIONS(1433),
    [aux_sym_command_name_token2] = ACTIONS(1433),
    [aux_sym_command_name_token3] = ACTIONS(1433),
    [aux_sym_command_name_token4] = ACTIONS(1433),
    [aux_sym_command_name_token5] = ACTIONS(1433),
    [aux_sym_command_name_token6] = ACTIONS(1433),
    [aux_sym_command_name_token7] = ACTIONS(1433),
    [aux_sym_command_name_token8] = ACTIONS(1433),
    [aux_sym_command_name_token9] = ACTIONS(1433),
    [aux_sym_command_name_token10] = ACTIONS(1433),
    [aux_sym_command_name_token11] = ACTIONS(1433),
    [aux_sym_command_name_token12] = ACTIONS(1433),
    [aux_sym_command_name_token13] = ACTIONS(1433),
    [aux_sym_command_name_token14] = ACTIONS(1433),
    [aux_sym_command_name_token15] = ACTIONS(1433),
    [aux_sym_command_name_token16] = ACTIONS(1433),
    [aux_sym_command_name_token17] = ACTIONS(1433),
    [aux_sym_command_name_token18] = ACTIONS(1433),
    [aux_sym_class_statement_token1] = ACTIONS(1433),
    [aux_sym_enum_statement_token1] = ACTIONS(1433),
    [anon_sym_PLUS] = ACTIONS(1433),
    [anon_sym_DASH] = ACTIONS(1433),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1433),
    [anon_sym_BANG] = ACTIONS(1433),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1433),
    [anon_sym_PLUS_PLUS] = ACTIONS(1433),
    [anon_sym_DASH_DASH] = ACTIONS(1433),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1433),
    [anon_sym_AT_LPAREN] = ACTIONS(1433),
    [anon_sym_AT_LBRACE] = ACTIONS(1433),
  },
  [STATE(428)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1315),
    [sym_hexadecimal_integer_literal] = ACTIONS(1315),
    [sym_real_literal] = ACTIONS(1315),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1315),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1315),
    [sym_verbatim_string_characters] = ACTIONS(1315),
    [sym_verbatim_here_string_characters] = ACTIONS(1315),
    [anon_sym_DOT] = ACTIONS(1315),
    [anon_sym_LBRACK] = ACTIONS(1315),
    [aux_sym_comparison_operator_token37] = ACTIONS(1315),
    [aux_sym_comparison_operator_token50] = ACTIONS(1315),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1315),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1315),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1315),
    [anon_sym_DOLLAR_] = ACTIONS(1315),
    [aux_sym_variable_token1] = ACTIONS(1315),
    [aux_sym_variable_token2] = ACTIONS(1315),
    [sym_braced_variable] = ACTIONS(1315),
    [anon_sym_SEMI] = ACTIONS(1315),
    [anon_sym_LPAREN] = ACTIONS(1315),
    [anon_sym_COMMA] = ACTIONS(1315),
    [anon_sym_LBRACE] = ACTIONS(1315),
    [aux_sym_if_statement_token1] = ACTIONS(1315),
    [aux_sym_switch_statement_token1] = ACTIONS(1315),
    [aux_sym_foreach_statement_token1] = ACTIONS(1315),
    [aux_sym_for_statement_token1] = ACTIONS(1315),
    [aux_sym_while_statement_token1] = ACTIONS(1315),
    [aux_sym_do_statement_token1] = ACTIONS(1315),
    [aux_sym_function_statement_token1] = ACTIONS(1315),
    [aux_sym_function_statement_token2] = ACTIONS(1315),
    [aux_sym_function_statement_token3] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1315),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1315),
    [sym_label] = ACTIONS(1315),
    [aux_sym_trap_statement_token1] = ACTIONS(1315),
    [aux_sym_try_statement_token1] = ACTIONS(1315),
    [aux_sym_data_statement_token1] = ACTIONS(1315),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1315),
    [aux_sym_parallel_statement_token1] = ACTIONS(1315),
    [aux_sym_sequence_statement_token1] = ACTIONS(1315),
    [anon_sym_AMP] = ACTIONS(1315),
    [aux_sym_command_name_token1] = ACTIONS(1315),
    [aux_sym_command_name_token2] = ACTIONS(1315),
    [aux_sym_command_name_token3] = ACTIONS(1315),
    [aux_sym_command_name_token4] = ACTIONS(1315),
    [aux_sym_command_name_token5] = ACTIONS(1315),
    [aux_sym_command_name_token6] = ACTIONS(1315),
    [aux_sym_command_name_token7] = ACTIONS(1315),
    [aux_sym_command_name_token8] = ACTIONS(1315),
    [aux_sym_command_name_token9] = ACTIONS(1315),
    [aux_sym_command_name_token10] = ACTIONS(1315),
    [aux_sym_command_name_token11] = ACTIONS(1315),
    [aux_sym_command_name_token12] = ACTIONS(1315),
    [aux_sym_command_name_token13] = ACTIONS(1315),
    [aux_sym_command_name_token14] = ACTIONS(1315),
    [aux_sym_command_name_token15] = ACTIONS(1315),
    [aux_sym_command_name_token16] = ACTIONS(1315),
    [aux_sym_command_name_token17] = ACTIONS(1315),
    [aux_sym_command_name_token18] = ACTIONS(1315),
    [aux_sym_class_statement_token1] = ACTIONS(1315),
    [aux_sym_enum_statement_token1] = ACTIONS(1315),
    [anon_sym_PLUS] = ACTIONS(1315),
    [anon_sym_DASH] = ACTIONS(1315),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1315),
    [anon_sym_BANG] = ACTIONS(1315),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1315),
    [anon_sym_PLUS_PLUS] = ACTIONS(1315),
    [anon_sym_DASH_DASH] = ACTIONS(1315),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1315),
    [anon_sym_AT_LPAREN] = ACTIONS(1315),
    [anon_sym_AT_LBRACE] = ACTIONS(1315),
  },
  [STATE(429)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1571),
    [sym_hexadecimal_integer_literal] = ACTIONS(1571),
    [sym_real_literal] = ACTIONS(1571),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1571),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1571),
    [sym_verbatim_string_characters] = ACTIONS(1571),
    [sym_verbatim_here_string_characters] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1571),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [aux_sym_comparison_operator_token37] = ACTIONS(1571),
    [aux_sym_comparison_operator_token50] = ACTIONS(1571),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1571),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1571),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1571),
    [anon_sym_DOLLAR_] = ACTIONS(1571),
    [aux_sym_variable_token1] = ACTIONS(1571),
    [aux_sym_variable_token2] = ACTIONS(1571),
    [sym_braced_variable] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1571),
    [anon_sym_COMMA] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [aux_sym_if_statement_token1] = ACTIONS(1571),
    [aux_sym_switch_statement_token1] = ACTIONS(1571),
    [aux_sym_foreach_statement_token1] = ACTIONS(1571),
    [aux_sym_for_statement_token1] = ACTIONS(1571),
    [aux_sym_while_statement_token1] = ACTIONS(1571),
    [aux_sym_do_statement_token1] = ACTIONS(1571),
    [aux_sym_function_statement_token1] = ACTIONS(1571),
    [aux_sym_function_statement_token2] = ACTIONS(1571),
    [aux_sym_function_statement_token3] = ACTIONS(1571),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1571),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1571),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1571),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1571),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1571),
    [sym_label] = ACTIONS(1571),
    [aux_sym_trap_statement_token1] = ACTIONS(1571),
    [aux_sym_try_statement_token1] = ACTIONS(1571),
    [aux_sym_data_statement_token1] = ACTIONS(1571),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1571),
    [aux_sym_parallel_statement_token1] = ACTIONS(1571),
    [aux_sym_sequence_statement_token1] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [aux_sym_command_name_token1] = ACTIONS(1571),
    [aux_sym_command_name_token2] = ACTIONS(1571),
    [aux_sym_command_name_token3] = ACTIONS(1571),
    [aux_sym_command_name_token4] = ACTIONS(1571),
    [aux_sym_command_name_token5] = ACTIONS(1571),
    [aux_sym_command_name_token6] = ACTIONS(1571),
    [aux_sym_command_name_token7] = ACTIONS(1571),
    [aux_sym_command_name_token8] = ACTIONS(1571),
    [aux_sym_command_name_token9] = ACTIONS(1571),
    [aux_sym_command_name_token10] = ACTIONS(1571),
    [aux_sym_command_name_token11] = ACTIONS(1571),
    [aux_sym_command_name_token12] = ACTIONS(1571),
    [aux_sym_command_name_token13] = ACTIONS(1571),
    [aux_sym_command_name_token14] = ACTIONS(1571),
    [aux_sym_command_name_token15] = ACTIONS(1571),
    [aux_sym_command_name_token16] = ACTIONS(1571),
    [aux_sym_command_name_token17] = ACTIONS(1571),
    [aux_sym_command_name_token18] = ACTIONS(1571),
    [aux_sym_class_statement_token1] = ACTIONS(1571),
    [aux_sym_enum_statement_token1] = ACTIONS(1571),
    [anon_sym_PLUS] = ACTIONS(1571),
    [anon_sym_DASH] = ACTIONS(1571),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1571),
    [anon_sym_BANG] = ACTIONS(1571),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1571),
    [anon_sym_PLUS_PLUS] = ACTIONS(1571),
    [anon_sym_DASH_DASH] = ACTIONS(1571),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1571),
    [anon_sym_AT_LPAREN] = ACTIONS(1571),
    [anon_sym_AT_LBRACE] = ACTIONS(1571),
  },
  [STATE(430)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1319),
    [sym_hexadecimal_integer_literal] = ACTIONS(1319),
    [sym_real_literal] = ACTIONS(1319),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1319),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1319),
    [sym_verbatim_string_characters] = ACTIONS(1319),
    [sym_verbatim_here_string_characters] = ACTIONS(1319),
    [anon_sym_DOT] = ACTIONS(1319),
    [anon_sym_LBRACK] = ACTIONS(1319),
    [aux_sym_comparison_operator_token37] = ACTIONS(1319),
    [aux_sym_comparison_operator_token50] = ACTIONS(1319),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1319),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1319),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1319),
    [anon_sym_DOLLAR_] = ACTIONS(1319),
    [aux_sym_variable_token1] = ACTIONS(1319),
    [aux_sym_variable_token2] = ACTIONS(1319),
    [sym_braced_variable] = ACTIONS(1319),
    [anon_sym_SEMI] = ACTIONS(1319),
    [anon_sym_LPAREN] = ACTIONS(1319),
    [anon_sym_COMMA] = ACTIONS(1319),
    [anon_sym_LBRACE] = ACTIONS(1319),
    [aux_sym_if_statement_token1] = ACTIONS(1319),
    [aux_sym_switch_statement_token1] = ACTIONS(1319),
    [aux_sym_foreach_statement_token1] = ACTIONS(1319),
    [aux_sym_for_statement_token1] = ACTIONS(1319),
    [aux_sym_while_statement_token1] = ACTIONS(1319),
    [aux_sym_do_statement_token1] = ACTIONS(1319),
    [aux_sym_function_statement_token1] = ACTIONS(1319),
    [aux_sym_function_statement_token2] = ACTIONS(1319),
    [aux_sym_function_statement_token3] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1319),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1319),
    [sym_label] = ACTIONS(1319),
    [aux_sym_trap_statement_token1] = ACTIONS(1319),
    [aux_sym_try_statement_token1] = ACTIONS(1319),
    [aux_sym_data_statement_token1] = ACTIONS(1319),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1319),
    [aux_sym_parallel_statement_token1] = ACTIONS(1319),
    [aux_sym_sequence_statement_token1] = ACTIONS(1319),
    [anon_sym_AMP] = ACTIONS(1319),
    [aux_sym_command_name_token1] = ACTIONS(1319),
    [aux_sym_command_name_token2] = ACTIONS(1319),
    [aux_sym_command_name_token3] = ACTIONS(1319),
    [aux_sym_command_name_token4] = ACTIONS(1319),
    [aux_sym_command_name_token5] = ACTIONS(1319),
    [aux_sym_command_name_token6] = ACTIONS(1319),
    [aux_sym_command_name_token7] = ACTIONS(1319),
    [aux_sym_command_name_token8] = ACTIONS(1319),
    [aux_sym_command_name_token9] = ACTIONS(1319),
    [aux_sym_command_name_token10] = ACTIONS(1319),
    [aux_sym_command_name_token11] = ACTIONS(1319),
    [aux_sym_command_name_token12] = ACTIONS(1319),
    [aux_sym_command_name_token13] = ACTIONS(1319),
    [aux_sym_command_name_token14] = ACTIONS(1319),
    [aux_sym_command_name_token15] = ACTIONS(1319),
    [aux_sym_command_name_token16] = ACTIONS(1319),
    [aux_sym_command_name_token17] = ACTIONS(1319),
    [aux_sym_command_name_token18] = ACTIONS(1319),
    [aux_sym_class_statement_token1] = ACTIONS(1319),
    [aux_sym_enum_statement_token1] = ACTIONS(1319),
    [anon_sym_PLUS] = ACTIONS(1319),
    [anon_sym_DASH] = ACTIONS(1319),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1319),
    [anon_sym_BANG] = ACTIONS(1319),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1319),
    [anon_sym_PLUS_PLUS] = ACTIONS(1319),
    [anon_sym_DASH_DASH] = ACTIONS(1319),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1319),
    [anon_sym_AT_LPAREN] = ACTIONS(1319),
    [anon_sym_AT_LBRACE] = ACTIONS(1319),
  },
  [STATE(431)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(740),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(654),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(654),
    [anon_sym_STAR] = ACTIONS(654),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(742),
    [anon_sym_DASH_DASH] = ACTIONS(744),
    [anon_sym_DOT2] = ACTIONS(1573),
    [anon_sym_COLON_COLON] = ACTIONS(1575),
    [anon_sym_RBRACK] = ACTIONS(654),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(750),
  },
  [STATE(432)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1335),
    [sym_hexadecimal_integer_literal] = ACTIONS(1335),
    [sym_real_literal] = ACTIONS(1335),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1335),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1335),
    [sym_verbatim_string_characters] = ACTIONS(1335),
    [sym_verbatim_here_string_characters] = ACTIONS(1335),
    [anon_sym_DOT] = ACTIONS(1335),
    [anon_sym_LBRACK] = ACTIONS(1335),
    [aux_sym_comparison_operator_token37] = ACTIONS(1335),
    [aux_sym_comparison_operator_token50] = ACTIONS(1335),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1335),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1335),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1335),
    [anon_sym_DOLLAR_] = ACTIONS(1335),
    [aux_sym_variable_token1] = ACTIONS(1335),
    [aux_sym_variable_token2] = ACTIONS(1335),
    [sym_braced_variable] = ACTIONS(1335),
    [anon_sym_SEMI] = ACTIONS(1335),
    [anon_sym_LPAREN] = ACTIONS(1335),
    [anon_sym_COMMA] = ACTIONS(1335),
    [anon_sym_LBRACE] = ACTIONS(1335),
    [aux_sym_if_statement_token1] = ACTIONS(1335),
    [aux_sym_switch_statement_token1] = ACTIONS(1335),
    [aux_sym_foreach_statement_token1] = ACTIONS(1335),
    [aux_sym_for_statement_token1] = ACTIONS(1335),
    [aux_sym_while_statement_token1] = ACTIONS(1335),
    [aux_sym_do_statement_token1] = ACTIONS(1335),
    [aux_sym_function_statement_token1] = ACTIONS(1335),
    [aux_sym_function_statement_token2] = ACTIONS(1335),
    [aux_sym_function_statement_token3] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token1] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token2] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token3] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token4] = ACTIONS(1335),
    [aux_sym_flow_control_statement_token5] = ACTIONS(1335),
    [sym_label] = ACTIONS(1335),
    [aux_sym_trap_statement_token1] = ACTIONS(1335),
    [aux_sym_try_statement_token1] = ACTIONS(1335),
    [aux_sym_data_statement_token1] = ACTIONS(1335),
    [aux_sym_inlinescript_statement_token1] = ACTIONS(1335),
    [aux_sym_parallel_statement_token1] = ACTIONS(1335),
    [aux_sym_sequence_statement_token1] = ACTIONS(1335),
    [anon_sym_AMP] = ACTIONS(1335),
    [aux_sym_command_name_token1] = ACTIONS(1335),
    [aux_sym_command_name_token2] = ACTIONS(1335),
    [aux_sym_command_name_token3] = ACTIONS(1335),
    [aux_sym_command_name_token4] = ACTIONS(1335),
    [aux_sym_command_name_token5] = ACTIONS(1335),
    [aux_sym_command_name_token6] = ACTIONS(1335),
    [aux_sym_command_name_token7] = ACTIONS(1335),
    [aux_sym_command_name_token8] = ACTIONS(1335),
    [aux_sym_command_name_token9] = ACTIONS(1335),
    [aux_sym_command_name_token10] = ACTIONS(1335),
    [aux_sym_command_name_token11] = ACTIONS(1335),
    [aux_sym_command_name_token12] = ACTIONS(1335),
    [aux_sym_command_name_token13] = ACTIONS(1335),
    [aux_sym_command_name_token14] = ACTIONS(1335),
    [aux_sym_command_name_token15] = ACTIONS(1335),
    [aux_sym_command_name_token16] = ACTIONS(1335),
    [aux_sym_command_name_token17] = ACTIONS(1335),
    [aux_sym_command_name_token18] = ACTIONS(1335),
    [aux_sym_class_statement_token1] = ACTIONS(1335),
    [aux_sym_enum_statement_token1] = ACTIONS(1335),
    [anon_sym_PLUS] = ACTIONS(1335),
    [anon_sym_DASH] = ACTIONS(1335),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1335),
    [anon_sym_BANG] = ACTIONS(1335),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1335),
    [anon_sym_PLUS_PLUS] = ACTIONS(1335),
    [anon_sym_DASH_DASH] = ACTIONS(1335),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1335),
    [anon_sym_AT_LPAREN] = ACTIONS(1335),
    [anon_sym_AT_LBRACE] = ACTIONS(1335),
  },
  [STATE(433)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(740),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(654),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(654),
    [anon_sym_STAR] = ACTIONS(654),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(742),
    [anon_sym_DASH_DASH] = ACTIONS(744),
    [anon_sym_DOT2] = ACTIONS(1577),
    [anon_sym_COLON_COLON] = ACTIONS(1579),
    [anon_sym_RBRACK] = ACTIONS(654),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(750),
  },
  [STATE(434)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(728),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_LPAREN] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(654),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(654),
    [anon_sym_STAR] = ACTIONS(654),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(730),
    [anon_sym_DASH_DASH] = ACTIONS(732),
    [anon_sym_DOT2] = ACTIONS(1581),
    [anon_sym_COLON_COLON] = ACTIONS(1583),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(738),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(435)] = {
    [sym_comment] = ACTIONS(79),
    [anon_sym_LBRACK] = ACTIONS(728),
    [aux_sym_comparison_operator_token1] = ACTIONS(654),
    [aux_sym_comparison_operator_token2] = ACTIONS(654),
    [aux_sym_comparison_operator_token3] = ACTIONS(654),
    [aux_sym_comparison_operator_token4] = ACTIONS(654),
    [aux_sym_comparison_operator_token5] = ACTIONS(654),
    [aux_sym_comparison_operator_token6] = ACTIONS(654),
    [aux_sym_comparison_operator_token7] = ACTIONS(654),
    [aux_sym_comparison_operator_token8] = ACTIONS(654),
    [aux_sym_comparison_operator_token9] = ACTIONS(654),
    [aux_sym_comparison_operator_token10] = ACTIONS(654),
    [aux_sym_comparison_operator_token11] = ACTIONS(654),
    [aux_sym_comparison_operator_token12] = ACTIONS(654),
    [aux_sym_comparison_operator_token13] = ACTIONS(654),
    [aux_sym_comparison_operator_token14] = ACTIONS(654),
    [aux_sym_comparison_operator_token15] = ACTIONS(654),
    [aux_sym_comparison_operator_token16] = ACTIONS(654),
    [aux_sym_comparison_operator_token17] = ACTIONS(654),
    [aux_sym_comparison_operator_token18] = ACTIONS(654),
    [aux_sym_comparison_operator_token19] = ACTIONS(654),
    [aux_sym_comparison_operator_token20] = ACTIONS(654),
    [aux_sym_comparison_operator_token21] = ACTIONS(654),
    [aux_sym_comparison_operator_token22] = ACTIONS(654),
    [aux_sym_comparison_operator_token23] = ACTIONS(654),
    [aux_sym_comparison_operator_token24] = ACTIONS(654),
    [aux_sym_comparison_operator_token25] = ACTIONS(654),
    [aux_sym_comparison_operator_token26] = ACTIONS(654),
    [aux_sym_comparison_operator_token27] = ACTIONS(654),
    [aux_sym_comparison_operator_token28] = ACTIONS(656),
    [aux_sym_comparison_operator_token29] = ACTIONS(654),
    [aux_sym_comparison_operator_token30] = ACTIONS(654),
    [aux_sym_comparison_operator_token31] = ACTIONS(654),
    [aux_sym_comparison_operator_token32] = ACTIONS(654),
    [aux_sym_comparison_operator_token33] = ACTIONS(654),
    [aux_sym_comparison_operator_token34] = ACTIONS(656),
    [aux_sym_comparison_operator_token35] = ACTIONS(654),
    [aux_sym_comparison_operator_token36] = ACTIONS(654),
    [aux_sym_comparison_operator_token37] = ACTIONS(654),
    [aux_sym_comparison_operator_token38] = ACTIONS(654),
    [aux_sym_comparison_operator_token39] = ACTIONS(654),
    [aux_sym_comparison_operator_token40] = ACTIONS(654),
    [aux_sym_comparison_operator_token41] = ACTIONS(654),
    [aux_sym_comparison_operator_token42] = ACTIONS(654),
    [aux_sym_comparison_operator_token43] = ACTIONS(654),
    [aux_sym_comparison_operator_token44] = ACTIONS(654),
    [aux_sym_comparison_operator_token45] = ACTIONS(654),
    [aux_sym_comparison_operator_token46] = ACTIONS(654),
    [aux_sym_comparison_operator_token47] = ACTIONS(654),
    [aux_sym_comparison_operator_token48] = ACTIONS(654),
    [aux_sym_comparison_operator_token49] = ACTIONS(654),
    [aux_sym_comparison_operator_token50] = ACTIONS(654),
    [aux_sym_format_operator_token1] = ACTIONS(654),
    [anon_sym_COMMA] = ACTIONS(654),
    [aux_sym_logical_expression_token1] = ACTIONS(654),
    [aux_sym_logical_expression_token2] = ACTIONS(654),
    [aux_sym_logical_expression_token3] = ACTIONS(654),
    [aux_sym_bitwise_expression_token1] = ACTIONS(654),
    [aux_sym_bitwise_expression_token2] = ACTIONS(654),
    [aux_sym_bitwise_expression_token3] = ACTIONS(654),
    [anon_sym_PLUS] = ACTIONS(656),
    [anon_sym_DASH] = ACTIONS(656),
    [anon_sym_SLASH] = ACTIONS(654),
    [anon_sym_BSLASH] = ACTIONS(654),
    [anon_sym_PERCENT] = ACTIONS(654),
    [anon_sym_STAR] = ACTIONS(654),
    [anon_sym_DOT_DOT] = ACTIONS(654),
    [anon_sym_PLUS_PLUS] = ACTIONS(730),
    [anon_sym_DASH_DASH] = ACTIONS(732),
    [anon_sym_DOT2] = ACTIONS(1585),
    [anon_sym_COLON_COLON] = ACTIONS(1587),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(738),
    [sym__statement_terminator] = ACTIONS(654),
  },
  [STATE(436)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(538),
    [sym_real_literal] = ACTIONS(538),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(538),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(538),
    [sym_verbatim_string_characters] = ACTIONS(538),
    [sym_verbatim_here_string_characters] = ACTIONS(538),
    [anon_sym_LBRACK] = ACTIONS(538),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(538),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(538),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(538),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(538),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(538),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(538),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(538),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(538),
    [anon_sym_2_GT_AMP1] = ACTIONS(538),
    [anon_sym_3_GT_AMP1] = ACTIONS(538),
    [anon_sym_4_GT_AMP1] = ACTIONS(538),
    [anon_sym_5_GT_AMP1] = ACTIONS(538),
    [anon_sym_6_GT_AMP1] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(538),
    [anon_sym_1_GT_AMP2] = ACTIONS(538),
    [anon_sym_3_GT_AMP2] = ACTIONS(538),
    [anon_sym_4_GT_AMP2] = ACTIONS(538),
    [anon_sym_5_GT_AMP2] = ACTIONS(538),
    [anon_sym_6_GT_AMP2] = ACTIONS(538),
    [aux_sym_comparison_operator_token37] = ACTIONS(538),
    [aux_sym_comparison_operator_token50] = ACTIONS(538),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(538),
    [anon_sym_DOLLAR_CARET] = ACTIONS(538),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(538),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(538),
    [sym_braced_variable] = ACTIONS(538),
    [sym_command_parameter] = ACTIONS(538),
    [anon_sym_LPAREN] = ACTIONS(538),
    [anon_sym_COMMA] = ACTIONS(538),
    [anon_sym_LBRACE] = ACTIONS(538),
    [anon_sym_AMP_AMP] = ACTIONS(538),
    [anon_sym_PIPE_PIPE] = ACTIONS(538),
    [anon_sym_PIPE] = ACTIONS(538),
    [sym_stop_parsing] = ACTIONS(538),
    [anon_sym_SPACE] = ACTIONS(540),
    [anon_sym_COLON] = ACTIONS(538),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(538),
    [anon_sym_PLUS_PLUS] = ACTIONS(538),
    [anon_sym_DASH_DASH] = ACTIONS(538),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LBRACE] = ACTIONS(538),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(538),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
    [sym__statement_terminator] = ACTIONS(540),
  },
  [STATE(437)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(538),
    [sym_real_literal] = ACTIONS(538),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(538),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(538),
    [sym_verbatim_string_characters] = ACTIONS(538),
    [sym_verbatim_here_string_characters] = ACTIONS(538),
    [anon_sym_LBRACK] = ACTIONS(538),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(538),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(538),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(538),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(538),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(538),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(538),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(538),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(538),
    [anon_sym_2_GT_AMP1] = ACTIONS(538),
    [anon_sym_3_GT_AMP1] = ACTIONS(538),
    [anon_sym_4_GT_AMP1] = ACTIONS(538),
    [anon_sym_5_GT_AMP1] = ACTIONS(538),
    [anon_sym_6_GT_AMP1] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(538),
    [anon_sym_1_GT_AMP2] = ACTIONS(538),
    [anon_sym_3_GT_AMP2] = ACTIONS(538),
    [anon_sym_4_GT_AMP2] = ACTIONS(538),
    [anon_sym_5_GT_AMP2] = ACTIONS(538),
    [anon_sym_6_GT_AMP2] = ACTIONS(538),
    [aux_sym_comparison_operator_token37] = ACTIONS(538),
    [aux_sym_comparison_operator_token50] = ACTIONS(538),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(538),
    [anon_sym_DOLLAR_CARET] = ACTIONS(538),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(538),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(538),
    [sym_braced_variable] = ACTIONS(538),
    [sym_command_parameter] = ACTIONS(538),
    [anon_sym_LPAREN] = ACTIONS(538),
    [anon_sym_RPAREN] = ACTIONS(538),
    [anon_sym_COMMA] = ACTIONS(538),
    [anon_sym_LBRACE] = ACTIONS(538),
    [anon_sym_AMP_AMP] = ACTIONS(538),
    [anon_sym_PIPE_PIPE] = ACTIONS(538),
    [anon_sym_PIPE] = ACTIONS(538),
    [sym_stop_parsing] = ACTIONS(538),
    [anon_sym_SPACE] = ACTIONS(540),
    [anon_sym_COLON] = ACTIONS(538),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(538),
    [anon_sym_PLUS_PLUS] = ACTIONS(538),
    [anon_sym_DASH_DASH] = ACTIONS(538),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LBRACE] = ACTIONS(538),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(538),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
  },
  [STATE(438)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(441),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1591),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
    [sym__statement_terminator] = ACTIONS(1593),
  },
  [STATE(439)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(439),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [sym_command_parameter] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_RPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [sym_stop_parsing] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1597),
    [anon_sym_COLON] = ACTIONS(1595),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1595),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
  },
  [STATE(440)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(440),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_RPAREN] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(1600),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(760),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(760),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(760),
    [anon_sym_STAR] = ACTIONS(760),
    [anon_sym_DOT_DOT] = ACTIONS(760),
    [anon_sym_RBRACK] = ACTIONS(760),
  },
  [STATE(441)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(441),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [sym_command_parameter] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [sym_stop_parsing] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1603),
    [anon_sym_COLON] = ACTIONS(1595),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1595),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
    [sym__statement_terminator] = ACTIONS(1606),
  },
  [STATE(442)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1678),
    [sym_logical_expression] = STATE(1379),
    [sym_bitwise_expression] = STATE(1293),
    [sym_comparison_expression] = STATE(560),
    [sym_additive_expression] = STATE(546),
    [sym_multiplicative_expression] = STATE(496),
    [sym_format_expression] = STATE(460),
    [sym_range_expression] = STATE(462),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(454),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_attribute_arguments] = STATE(1896),
    [sym_attribute_argument] = STATE(1436),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [sym_simple_name] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1610),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(443)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(538),
    [sym_real_literal] = ACTIONS(538),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(538),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(538),
    [sym_verbatim_string_characters] = ACTIONS(538),
    [sym_verbatim_here_string_characters] = ACTIONS(538),
    [anon_sym_LBRACK] = ACTIONS(538),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(538),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(538),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(538),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(538),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(538),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(538),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(538),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(538),
    [anon_sym_2_GT_AMP1] = ACTIONS(538),
    [anon_sym_3_GT_AMP1] = ACTIONS(538),
    [anon_sym_4_GT_AMP1] = ACTIONS(538),
    [anon_sym_5_GT_AMP1] = ACTIONS(538),
    [anon_sym_6_GT_AMP1] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(538),
    [anon_sym_1_GT_AMP2] = ACTIONS(538),
    [anon_sym_3_GT_AMP2] = ACTIONS(538),
    [anon_sym_4_GT_AMP2] = ACTIONS(538),
    [anon_sym_5_GT_AMP2] = ACTIONS(538),
    [anon_sym_6_GT_AMP2] = ACTIONS(538),
    [aux_sym_comparison_operator_token37] = ACTIONS(538),
    [aux_sym_comparison_operator_token50] = ACTIONS(538),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(538),
    [anon_sym_DOLLAR_CARET] = ACTIONS(538),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(538),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(538),
    [sym_braced_variable] = ACTIONS(538),
    [sym_command_parameter] = ACTIONS(538),
    [anon_sym_LPAREN] = ACTIONS(538),
    [anon_sym_RPAREN] = ACTIONS(538),
    [anon_sym_COMMA] = ACTIONS(538),
    [anon_sym_LBRACE] = ACTIONS(538),
    [anon_sym_AMP_AMP] = ACTIONS(538),
    [anon_sym_PIPE_PIPE] = ACTIONS(538),
    [anon_sym_PIPE] = ACTIONS(538),
    [sym_stop_parsing] = ACTIONS(538),
    [anon_sym_SPACE] = ACTIONS(540),
    [anon_sym_COLON] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(538),
    [anon_sym_PLUS_PLUS] = ACTIONS(538),
    [anon_sym_DASH_DASH] = ACTIONS(538),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LBRACE] = ACTIONS(538),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(538),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
  },
  [STATE(444)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(538),
    [sym_real_literal] = ACTIONS(538),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(538),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(538),
    [sym_verbatim_string_characters] = ACTIONS(538),
    [sym_verbatim_here_string_characters] = ACTIONS(538),
    [anon_sym_LBRACK] = ACTIONS(538),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(538),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(538),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(538),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(538),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(538),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(538),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(538),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(538),
    [anon_sym_2_GT_AMP1] = ACTIONS(538),
    [anon_sym_3_GT_AMP1] = ACTIONS(538),
    [anon_sym_4_GT_AMP1] = ACTIONS(538),
    [anon_sym_5_GT_AMP1] = ACTIONS(538),
    [anon_sym_6_GT_AMP1] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(538),
    [anon_sym_1_GT_AMP2] = ACTIONS(538),
    [anon_sym_3_GT_AMP2] = ACTIONS(538),
    [anon_sym_4_GT_AMP2] = ACTIONS(538),
    [anon_sym_5_GT_AMP2] = ACTIONS(538),
    [anon_sym_6_GT_AMP2] = ACTIONS(538),
    [aux_sym_comparison_operator_token37] = ACTIONS(538),
    [aux_sym_comparison_operator_token50] = ACTIONS(538),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(538),
    [anon_sym_DOLLAR_CARET] = ACTIONS(538),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(538),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(538),
    [sym_braced_variable] = ACTIONS(538),
    [sym_command_parameter] = ACTIONS(538),
    [anon_sym_LPAREN] = ACTIONS(538),
    [anon_sym_COMMA] = ACTIONS(538),
    [anon_sym_LBRACE] = ACTIONS(538),
    [anon_sym_AMP_AMP] = ACTIONS(538),
    [anon_sym_PIPE_PIPE] = ACTIONS(538),
    [anon_sym_PIPE] = ACTIONS(538),
    [sym_stop_parsing] = ACTIONS(538),
    [anon_sym_SPACE] = ACTIONS(540),
    [anon_sym_COLON] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(538),
    [anon_sym_BANG] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(538),
    [anon_sym_PLUS_PLUS] = ACTIONS(538),
    [anon_sym_DASH_DASH] = ACTIONS(538),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LPAREN] = ACTIONS(538),
    [anon_sym_AT_LBRACE] = ACTIONS(538),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(538),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
    [sym__statement_terminator] = ACTIONS(540),
  },
  [STATE(445)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(439),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1612),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
  },
  [STATE(446)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1678),
    [sym_logical_expression] = STATE(1379),
    [sym_bitwise_expression] = STATE(1293),
    [sym_comparison_expression] = STATE(560),
    [sym_additive_expression] = STATE(546),
    [sym_multiplicative_expression] = STATE(496),
    [sym_format_expression] = STATE(460),
    [sym_range_expression] = STATE(462),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(454),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_attribute_arguments] = STATE(1977),
    [sym_attribute_argument] = STATE(1436),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [sym_simple_name] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(447)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(447),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(760),
    [aux_sym_comparison_operator_token2] = ACTIONS(760),
    [aux_sym_comparison_operator_token3] = ACTIONS(760),
    [aux_sym_comparison_operator_token4] = ACTIONS(760),
    [aux_sym_comparison_operator_token5] = ACTIONS(760),
    [aux_sym_comparison_operator_token6] = ACTIONS(760),
    [aux_sym_comparison_operator_token7] = ACTIONS(760),
    [aux_sym_comparison_operator_token8] = ACTIONS(760),
    [aux_sym_comparison_operator_token9] = ACTIONS(760),
    [aux_sym_comparison_operator_token10] = ACTIONS(760),
    [aux_sym_comparison_operator_token11] = ACTIONS(760),
    [aux_sym_comparison_operator_token12] = ACTIONS(760),
    [aux_sym_comparison_operator_token13] = ACTIONS(760),
    [aux_sym_comparison_operator_token14] = ACTIONS(760),
    [aux_sym_comparison_operator_token15] = ACTIONS(760),
    [aux_sym_comparison_operator_token16] = ACTIONS(760),
    [aux_sym_comparison_operator_token17] = ACTIONS(760),
    [aux_sym_comparison_operator_token18] = ACTIONS(760),
    [aux_sym_comparison_operator_token19] = ACTIONS(760),
    [aux_sym_comparison_operator_token20] = ACTIONS(760),
    [aux_sym_comparison_operator_token21] = ACTIONS(760),
    [aux_sym_comparison_operator_token22] = ACTIONS(760),
    [aux_sym_comparison_operator_token23] = ACTIONS(760),
    [aux_sym_comparison_operator_token24] = ACTIONS(760),
    [aux_sym_comparison_operator_token25] = ACTIONS(760),
    [aux_sym_comparison_operator_token26] = ACTIONS(760),
    [aux_sym_comparison_operator_token27] = ACTIONS(760),
    [aux_sym_comparison_operator_token28] = ACTIONS(762),
    [aux_sym_comparison_operator_token29] = ACTIONS(760),
    [aux_sym_comparison_operator_token30] = ACTIONS(760),
    [aux_sym_comparison_operator_token31] = ACTIONS(760),
    [aux_sym_comparison_operator_token32] = ACTIONS(760),
    [aux_sym_comparison_operator_token33] = ACTIONS(760),
    [aux_sym_comparison_operator_token34] = ACTIONS(762),
    [aux_sym_comparison_operator_token35] = ACTIONS(760),
    [aux_sym_comparison_operator_token36] = ACTIONS(760),
    [aux_sym_comparison_operator_token37] = ACTIONS(760),
    [aux_sym_comparison_operator_token38] = ACTIONS(760),
    [aux_sym_comparison_operator_token39] = ACTIONS(760),
    [aux_sym_comparison_operator_token40] = ACTIONS(760),
    [aux_sym_comparison_operator_token41] = ACTIONS(760),
    [aux_sym_comparison_operator_token42] = ACTIONS(760),
    [aux_sym_comparison_operator_token43] = ACTIONS(760),
    [aux_sym_comparison_operator_token44] = ACTIONS(760),
    [aux_sym_comparison_operator_token45] = ACTIONS(760),
    [aux_sym_comparison_operator_token46] = ACTIONS(760),
    [aux_sym_comparison_operator_token47] = ACTIONS(760),
    [aux_sym_comparison_operator_token48] = ACTIONS(760),
    [aux_sym_comparison_operator_token49] = ACTIONS(760),
    [aux_sym_comparison_operator_token50] = ACTIONS(760),
    [aux_sym_format_operator_token1] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(1616),
    [aux_sym_logical_expression_token1] = ACTIONS(760),
    [aux_sym_logical_expression_token2] = ACTIONS(760),
    [aux_sym_logical_expression_token3] = ACTIONS(760),
    [aux_sym_bitwise_expression_token1] = ACTIONS(760),
    [aux_sym_bitwise_expression_token2] = ACTIONS(760),
    [aux_sym_bitwise_expression_token3] = ACTIONS(760),
    [anon_sym_PLUS] = ACTIONS(760),
    [anon_sym_DASH] = ACTIONS(762),
    [anon_sym_SLASH] = ACTIONS(760),
    [anon_sym_BSLASH] = ACTIONS(760),
    [anon_sym_PERCENT] = ACTIONS(760),
    [anon_sym_STAR] = ACTIONS(760),
    [anon_sym_DOT_DOT] = ACTIONS(760),
    [sym__statement_terminator] = ACTIONS(760),
  },
  [STATE(448)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1589),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
    [sym__statement_terminator] = ACTIONS(1593),
  },
  [STATE(449)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1589),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_DASH_DASH_PERCENT] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
  },
  [STATE(450)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(455),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(772),
    [aux_sym_comparison_operator_token2] = ACTIONS(772),
    [aux_sym_comparison_operator_token3] = ACTIONS(772),
    [aux_sym_comparison_operator_token4] = ACTIONS(772),
    [aux_sym_comparison_operator_token5] = ACTIONS(772),
    [aux_sym_comparison_operator_token6] = ACTIONS(772),
    [aux_sym_comparison_operator_token7] = ACTIONS(772),
    [aux_sym_comparison_operator_token8] = ACTIONS(772),
    [aux_sym_comparison_operator_token9] = ACTIONS(772),
    [aux_sym_comparison_operator_token10] = ACTIONS(772),
    [aux_sym_comparison_operator_token11] = ACTIONS(772),
    [aux_sym_comparison_operator_token12] = ACTIONS(772),
    [aux_sym_comparison_operator_token13] = ACTIONS(772),
    [aux_sym_comparison_operator_token14] = ACTIONS(772),
    [aux_sym_comparison_operator_token15] = ACTIONS(772),
    [aux_sym_comparison_operator_token16] = ACTIONS(772),
    [aux_sym_comparison_operator_token17] = ACTIONS(772),
    [aux_sym_comparison_operator_token18] = ACTIONS(772),
    [aux_sym_comparison_operator_token19] = ACTIONS(772),
    [aux_sym_comparison_operator_token20] = ACTIONS(772),
    [aux_sym_comparison_operator_token21] = ACTIONS(772),
    [aux_sym_comparison_operator_token22] = ACTIONS(772),
    [aux_sym_comparison_operator_token23] = ACTIONS(772),
    [aux_sym_comparison_operator_token24] = ACTIONS(772),
    [aux_sym_comparison_operator_token25] = ACTIONS(772),
    [aux_sym_comparison_operator_token26] = ACTIONS(772),
    [aux_sym_comparison_operator_token27] = ACTIONS(772),
    [aux_sym_comparison_operator_token28] = ACTIONS(774),
    [aux_sym_comparison_operator_token29] = ACTIONS(772),
    [aux_sym_comparison_operator_token30] = ACTIONS(772),
    [aux_sym_comparison_operator_token31] = ACTIONS(772),
    [aux_sym_comparison_operator_token32] = ACTIONS(772),
    [aux_sym_comparison_operator_token33] = ACTIONS(772),
    [aux_sym_comparison_operator_token34] = ACTIONS(774),
    [aux_sym_comparison_operator_token35] = ACTIONS(772),
    [aux_sym_comparison_operator_token36] = ACTIONS(772),
    [aux_sym_comparison_operator_token37] = ACTIONS(772),
    [aux_sym_comparison_operator_token38] = ACTIONS(772),
    [aux_sym_comparison_operator_token39] = ACTIONS(772),
    [aux_sym_comparison_operator_token40] = ACTIONS(772),
    [aux_sym_comparison_operator_token41] = ACTIONS(772),
    [aux_sym_comparison_operator_token42] = ACTIONS(772),
    [aux_sym_comparison_operator_token43] = ACTIONS(772),
    [aux_sym_comparison_operator_token44] = ACTIONS(772),
    [aux_sym_comparison_operator_token45] = ACTIONS(772),
    [aux_sym_comparison_operator_token46] = ACTIONS(772),
    [aux_sym_comparison_operator_token47] = ACTIONS(772),
    [aux_sym_comparison_operator_token48] = ACTIONS(772),
    [aux_sym_comparison_operator_token49] = ACTIONS(772),
    [aux_sym_comparison_operator_token50] = ACTIONS(772),
    [aux_sym_format_operator_token1] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(1619),
    [aux_sym_logical_expression_token1] = ACTIONS(772),
    [aux_sym_logical_expression_token2] = ACTIONS(772),
    [aux_sym_logical_expression_token3] = ACTIONS(772),
    [aux_sym_bitwise_expression_token1] = ACTIONS(772),
    [aux_sym_bitwise_expression_token2] = ACTIONS(772),
    [aux_sym_bitwise_expression_token3] = ACTIONS(772),
    [anon_sym_PLUS] = ACTIONS(772),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(772),
    [anon_sym_BSLASH] = ACTIONS(772),
    [anon_sym_PERCENT] = ACTIONS(772),
    [anon_sym_STAR] = ACTIONS(772),
    [anon_sym_DOT_DOT] = ACTIONS(772),
    [sym__statement_terminator] = ACTIONS(772),
  },
  [STATE(451)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(440),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(783),
    [aux_sym_comparison_operator_token2] = ACTIONS(783),
    [aux_sym_comparison_operator_token3] = ACTIONS(783),
    [aux_sym_comparison_operator_token4] = ACTIONS(783),
    [aux_sym_comparison_operator_token5] = ACTIONS(783),
    [aux_sym_comparison_operator_token6] = ACTIONS(783),
    [aux_sym_comparison_operator_token7] = ACTIONS(783),
    [aux_sym_comparison_operator_token8] = ACTIONS(783),
    [aux_sym_comparison_operator_token9] = ACTIONS(783),
    [aux_sym_comparison_operator_token10] = ACTIONS(783),
    [aux_sym_comparison_operator_token11] = ACTIONS(783),
    [aux_sym_comparison_operator_token12] = ACTIONS(783),
    [aux_sym_comparison_operator_token13] = ACTIONS(783),
    [aux_sym_comparison_operator_token14] = ACTIONS(783),
    [aux_sym_comparison_operator_token15] = ACTIONS(783),
    [aux_sym_comparison_operator_token16] = ACTIONS(783),
    [aux_sym_comparison_operator_token17] = ACTIONS(783),
    [aux_sym_comparison_operator_token18] = ACTIONS(783),
    [aux_sym_comparison_operator_token19] = ACTIONS(783),
    [aux_sym_comparison_operator_token20] = ACTIONS(783),
    [aux_sym_comparison_operator_token21] = ACTIONS(783),
    [aux_sym_comparison_operator_token22] = ACTIONS(783),
    [aux_sym_comparison_operator_token23] = ACTIONS(783),
    [aux_sym_comparison_operator_token24] = ACTIONS(783),
    [aux_sym_comparison_operator_token25] = ACTIONS(783),
    [aux_sym_comparison_operator_token26] = ACTIONS(783),
    [aux_sym_comparison_operator_token27] = ACTIONS(783),
    [aux_sym_comparison_operator_token28] = ACTIONS(785),
    [aux_sym_comparison_operator_token29] = ACTIONS(783),
    [aux_sym_comparison_operator_token30] = ACTIONS(783),
    [aux_sym_comparison_operator_token31] = ACTIONS(783),
    [aux_sym_comparison_operator_token32] = ACTIONS(783),
    [aux_sym_comparison_operator_token33] = ACTIONS(783),
    [aux_sym_comparison_operator_token34] = ACTIONS(785),
    [aux_sym_comparison_operator_token35] = ACTIONS(783),
    [aux_sym_comparison_operator_token36] = ACTIONS(783),
    [aux_sym_comparison_operator_token37] = ACTIONS(783),
    [aux_sym_comparison_operator_token38] = ACTIONS(783),
    [aux_sym_comparison_operator_token39] = ACTIONS(783),
    [aux_sym_comparison_operator_token40] = ACTIONS(783),
    [aux_sym_comparison_operator_token41] = ACTIONS(783),
    [aux_sym_comparison_operator_token42] = ACTIONS(783),
    [aux_sym_comparison_operator_token43] = ACTIONS(783),
    [aux_sym_comparison_operator_token44] = ACTIONS(783),
    [aux_sym_comparison_operator_token45] = ACTIONS(783),
    [aux_sym_comparison_operator_token46] = ACTIONS(783),
    [aux_sym_comparison_operator_token47] = ACTIONS(783),
    [aux_sym_comparison_operator_token48] = ACTIONS(783),
    [aux_sym_comparison_operator_token49] = ACTIONS(783),
    [aux_sym_comparison_operator_token50] = ACTIONS(783),
    [aux_sym_format_operator_token1] = ACTIONS(783),
    [anon_sym_COMMA] = ACTIONS(1621),
    [aux_sym_logical_expression_token1] = ACTIONS(783),
    [aux_sym_logical_expression_token2] = ACTIONS(783),
    [aux_sym_logical_expression_token3] = ACTIONS(783),
    [aux_sym_bitwise_expression_token1] = ACTIONS(783),
    [aux_sym_bitwise_expression_token2] = ACTIONS(783),
    [aux_sym_bitwise_expression_token3] = ACTIONS(783),
    [anon_sym_PLUS] = ACTIONS(783),
    [anon_sym_DASH] = ACTIONS(785),
    [anon_sym_SLASH] = ACTIONS(783),
    [anon_sym_BSLASH] = ACTIONS(783),
    [anon_sym_PERCENT] = ACTIONS(783),
    [anon_sym_STAR] = ACTIONS(783),
    [anon_sym_DOT_DOT] = ACTIONS(783),
    [anon_sym_RBRACK] = ACTIONS(783),
  },
  [STATE(452)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(451),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(772),
    [aux_sym_comparison_operator_token2] = ACTIONS(772),
    [aux_sym_comparison_operator_token3] = ACTIONS(772),
    [aux_sym_comparison_operator_token4] = ACTIONS(772),
    [aux_sym_comparison_operator_token5] = ACTIONS(772),
    [aux_sym_comparison_operator_token6] = ACTIONS(772),
    [aux_sym_comparison_operator_token7] = ACTIONS(772),
    [aux_sym_comparison_operator_token8] = ACTIONS(772),
    [aux_sym_comparison_operator_token9] = ACTIONS(772),
    [aux_sym_comparison_operator_token10] = ACTIONS(772),
    [aux_sym_comparison_operator_token11] = ACTIONS(772),
    [aux_sym_comparison_operator_token12] = ACTIONS(772),
    [aux_sym_comparison_operator_token13] = ACTIONS(772),
    [aux_sym_comparison_operator_token14] = ACTIONS(772),
    [aux_sym_comparison_operator_token15] = ACTIONS(772),
    [aux_sym_comparison_operator_token16] = ACTIONS(772),
    [aux_sym_comparison_operator_token17] = ACTIONS(772),
    [aux_sym_comparison_operator_token18] = ACTIONS(772),
    [aux_sym_comparison_operator_token19] = ACTIONS(772),
    [aux_sym_comparison_operator_token20] = ACTIONS(772),
    [aux_sym_comparison_operator_token21] = ACTIONS(772),
    [aux_sym_comparison_operator_token22] = ACTIONS(772),
    [aux_sym_comparison_operator_token23] = ACTIONS(772),
    [aux_sym_comparison_operator_token24] = ACTIONS(772),
    [aux_sym_comparison_operator_token25] = ACTIONS(772),
    [aux_sym_comparison_operator_token26] = ACTIONS(772),
    [aux_sym_comparison_operator_token27] = ACTIONS(772),
    [aux_sym_comparison_operator_token28] = ACTIONS(774),
    [aux_sym_comparison_operator_token29] = ACTIONS(772),
    [aux_sym_comparison_operator_token30] = ACTIONS(772),
    [aux_sym_comparison_operator_token31] = ACTIONS(772),
    [aux_sym_comparison_operator_token32] = ACTIONS(772),
    [aux_sym_comparison_operator_token33] = ACTIONS(772),
    [aux_sym_comparison_operator_token34] = ACTIONS(774),
    [aux_sym_comparison_operator_token35] = ACTIONS(772),
    [aux_sym_comparison_operator_token36] = ACTIONS(772),
    [aux_sym_comparison_operator_token37] = ACTIONS(772),
    [aux_sym_comparison_operator_token38] = ACTIONS(772),
    [aux_sym_comparison_operator_token39] = ACTIONS(772),
    [aux_sym_comparison_operator_token40] = ACTIONS(772),
    [aux_sym_comparison_operator_token41] = ACTIONS(772),
    [aux_sym_comparison_operator_token42] = ACTIONS(772),
    [aux_sym_comparison_operator_token43] = ACTIONS(772),
    [aux_sym_comparison_operator_token44] = ACTIONS(772),
    [aux_sym_comparison_operator_token45] = ACTIONS(772),
    [aux_sym_comparison_operator_token46] = ACTIONS(772),
    [aux_sym_comparison_operator_token47] = ACTIONS(772),
    [aux_sym_comparison_operator_token48] = ACTIONS(772),
    [aux_sym_comparison_operator_token49] = ACTIONS(772),
    [aux_sym_comparison_operator_token50] = ACTIONS(772),
    [aux_sym_format_operator_token1] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(1621),
    [aux_sym_logical_expression_token1] = ACTIONS(772),
    [aux_sym_logical_expression_token2] = ACTIONS(772),
    [aux_sym_logical_expression_token3] = ACTIONS(772),
    [aux_sym_bitwise_expression_token1] = ACTIONS(772),
    [aux_sym_bitwise_expression_token2] = ACTIONS(772),
    [aux_sym_bitwise_expression_token3] = ACTIONS(772),
    [anon_sym_PLUS] = ACTIONS(772),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(772),
    [anon_sym_BSLASH] = ACTIONS(772),
    [anon_sym_PERCENT] = ACTIONS(772),
    [anon_sym_STAR] = ACTIONS(772),
    [anon_sym_DOT_DOT] = ACTIONS(772),
    [anon_sym_RBRACK] = ACTIONS(772),
  },
  [STATE(453)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(459),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1623),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
  },
  [STATE(454)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(456),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(772),
    [aux_sym_comparison_operator_token2] = ACTIONS(772),
    [aux_sym_comparison_operator_token3] = ACTIONS(772),
    [aux_sym_comparison_operator_token4] = ACTIONS(772),
    [aux_sym_comparison_operator_token5] = ACTIONS(772),
    [aux_sym_comparison_operator_token6] = ACTIONS(772),
    [aux_sym_comparison_operator_token7] = ACTIONS(772),
    [aux_sym_comparison_operator_token8] = ACTIONS(772),
    [aux_sym_comparison_operator_token9] = ACTIONS(772),
    [aux_sym_comparison_operator_token10] = ACTIONS(772),
    [aux_sym_comparison_operator_token11] = ACTIONS(772),
    [aux_sym_comparison_operator_token12] = ACTIONS(772),
    [aux_sym_comparison_operator_token13] = ACTIONS(772),
    [aux_sym_comparison_operator_token14] = ACTIONS(772),
    [aux_sym_comparison_operator_token15] = ACTIONS(772),
    [aux_sym_comparison_operator_token16] = ACTIONS(772),
    [aux_sym_comparison_operator_token17] = ACTIONS(772),
    [aux_sym_comparison_operator_token18] = ACTIONS(772),
    [aux_sym_comparison_operator_token19] = ACTIONS(772),
    [aux_sym_comparison_operator_token20] = ACTIONS(772),
    [aux_sym_comparison_operator_token21] = ACTIONS(772),
    [aux_sym_comparison_operator_token22] = ACTIONS(772),
    [aux_sym_comparison_operator_token23] = ACTIONS(772),
    [aux_sym_comparison_operator_token24] = ACTIONS(772),
    [aux_sym_comparison_operator_token25] = ACTIONS(772),
    [aux_sym_comparison_operator_token26] = ACTIONS(772),
    [aux_sym_comparison_operator_token27] = ACTIONS(772),
    [aux_sym_comparison_operator_token28] = ACTIONS(774),
    [aux_sym_comparison_operator_token29] = ACTIONS(772),
    [aux_sym_comparison_operator_token30] = ACTIONS(772),
    [aux_sym_comparison_operator_token31] = ACTIONS(772),
    [aux_sym_comparison_operator_token32] = ACTIONS(772),
    [aux_sym_comparison_operator_token33] = ACTIONS(772),
    [aux_sym_comparison_operator_token34] = ACTIONS(774),
    [aux_sym_comparison_operator_token35] = ACTIONS(772),
    [aux_sym_comparison_operator_token36] = ACTIONS(772),
    [aux_sym_comparison_operator_token37] = ACTIONS(772),
    [aux_sym_comparison_operator_token38] = ACTIONS(772),
    [aux_sym_comparison_operator_token39] = ACTIONS(772),
    [aux_sym_comparison_operator_token40] = ACTIONS(772),
    [aux_sym_comparison_operator_token41] = ACTIONS(772),
    [aux_sym_comparison_operator_token42] = ACTIONS(772),
    [aux_sym_comparison_operator_token43] = ACTIONS(772),
    [aux_sym_comparison_operator_token44] = ACTIONS(772),
    [aux_sym_comparison_operator_token45] = ACTIONS(772),
    [aux_sym_comparison_operator_token46] = ACTIONS(772),
    [aux_sym_comparison_operator_token47] = ACTIONS(772),
    [aux_sym_comparison_operator_token48] = ACTIONS(772),
    [aux_sym_comparison_operator_token49] = ACTIONS(772),
    [aux_sym_comparison_operator_token50] = ACTIONS(772),
    [aux_sym_format_operator_token1] = ACTIONS(772),
    [anon_sym_RPAREN] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(772),
    [aux_sym_logical_expression_token1] = ACTIONS(772),
    [aux_sym_logical_expression_token2] = ACTIONS(772),
    [aux_sym_logical_expression_token3] = ACTIONS(772),
    [aux_sym_bitwise_expression_token1] = ACTIONS(772),
    [aux_sym_bitwise_expression_token2] = ACTIONS(772),
    [aux_sym_bitwise_expression_token3] = ACTIONS(772),
    [anon_sym_PLUS] = ACTIONS(772),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(772),
    [anon_sym_BSLASH] = ACTIONS(772),
    [anon_sym_PERCENT] = ACTIONS(772),
    [anon_sym_STAR] = ACTIONS(772),
    [anon_sym_DOT_DOT] = ACTIONS(772),
  },
  [STATE(455)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(447),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(783),
    [aux_sym_comparison_operator_token2] = ACTIONS(783),
    [aux_sym_comparison_operator_token3] = ACTIONS(783),
    [aux_sym_comparison_operator_token4] = ACTIONS(783),
    [aux_sym_comparison_operator_token5] = ACTIONS(783),
    [aux_sym_comparison_operator_token6] = ACTIONS(783),
    [aux_sym_comparison_operator_token7] = ACTIONS(783),
    [aux_sym_comparison_operator_token8] = ACTIONS(783),
    [aux_sym_comparison_operator_token9] = ACTIONS(783),
    [aux_sym_comparison_operator_token10] = ACTIONS(783),
    [aux_sym_comparison_operator_token11] = ACTIONS(783),
    [aux_sym_comparison_operator_token12] = ACTIONS(783),
    [aux_sym_comparison_operator_token13] = ACTIONS(783),
    [aux_sym_comparison_operator_token14] = ACTIONS(783),
    [aux_sym_comparison_operator_token15] = ACTIONS(783),
    [aux_sym_comparison_operator_token16] = ACTIONS(783),
    [aux_sym_comparison_operator_token17] = ACTIONS(783),
    [aux_sym_comparison_operator_token18] = ACTIONS(783),
    [aux_sym_comparison_operator_token19] = ACTIONS(783),
    [aux_sym_comparison_operator_token20] = ACTIONS(783),
    [aux_sym_comparison_operator_token21] = ACTIONS(783),
    [aux_sym_comparison_operator_token22] = ACTIONS(783),
    [aux_sym_comparison_operator_token23] = ACTIONS(783),
    [aux_sym_comparison_operator_token24] = ACTIONS(783),
    [aux_sym_comparison_operator_token25] = ACTIONS(783),
    [aux_sym_comparison_operator_token26] = ACTIONS(783),
    [aux_sym_comparison_operator_token27] = ACTIONS(783),
    [aux_sym_comparison_operator_token28] = ACTIONS(785),
    [aux_sym_comparison_operator_token29] = ACTIONS(783),
    [aux_sym_comparison_operator_token30] = ACTIONS(783),
    [aux_sym_comparison_operator_token31] = ACTIONS(783),
    [aux_sym_comparison_operator_token32] = ACTIONS(783),
    [aux_sym_comparison_operator_token33] = ACTIONS(783),
    [aux_sym_comparison_operator_token34] = ACTIONS(785),
    [aux_sym_comparison_operator_token35] = ACTIONS(783),
    [aux_sym_comparison_operator_token36] = ACTIONS(783),
    [aux_sym_comparison_operator_token37] = ACTIONS(783),
    [aux_sym_comparison_operator_token38] = ACTIONS(783),
    [aux_sym_comparison_operator_token39] = ACTIONS(783),
    [aux_sym_comparison_operator_token40] = ACTIONS(783),
    [aux_sym_comparison_operator_token41] = ACTIONS(783),
    [aux_sym_comparison_operator_token42] = ACTIONS(783),
    [aux_sym_comparison_operator_token43] = ACTIONS(783),
    [aux_sym_comparison_operator_token44] = ACTIONS(783),
    [aux_sym_comparison_operator_token45] = ACTIONS(783),
    [aux_sym_comparison_operator_token46] = ACTIONS(783),
    [aux_sym_comparison_operator_token47] = ACTIONS(783),
    [aux_sym_comparison_operator_token48] = ACTIONS(783),
    [aux_sym_comparison_operator_token49] = ACTIONS(783),
    [aux_sym_comparison_operator_token50] = ACTIONS(783),
    [aux_sym_format_operator_token1] = ACTIONS(783),
    [anon_sym_COMMA] = ACTIONS(1619),
    [aux_sym_logical_expression_token1] = ACTIONS(783),
    [aux_sym_logical_expression_token2] = ACTIONS(783),
    [aux_sym_logical_expression_token3] = ACTIONS(783),
    [aux_sym_bitwise_expression_token1] = ACTIONS(783),
    [aux_sym_bitwise_expression_token2] = ACTIONS(783),
    [aux_sym_bitwise_expression_token3] = ACTIONS(783),
    [anon_sym_PLUS] = ACTIONS(783),
    [anon_sym_DASH] = ACTIONS(785),
    [anon_sym_SLASH] = ACTIONS(783),
    [anon_sym_BSLASH] = ACTIONS(783),
    [anon_sym_PERCENT] = ACTIONS(783),
    [anon_sym_STAR] = ACTIONS(783),
    [anon_sym_DOT_DOT] = ACTIONS(783),
    [sym__statement_terminator] = ACTIONS(783),
  },
  [STATE(456)] = {
    [aux_sym_array_literal_expression_repeat1] = STATE(440),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(783),
    [aux_sym_comparison_operator_token2] = ACTIONS(783),
    [aux_sym_comparison_operator_token3] = ACTIONS(783),
    [aux_sym_comparison_operator_token4] = ACTIONS(783),
    [aux_sym_comparison_operator_token5] = ACTIONS(783),
    [aux_sym_comparison_operator_token6] = ACTIONS(783),
    [aux_sym_comparison_operator_token7] = ACTIONS(783),
    [aux_sym_comparison_operator_token8] = ACTIONS(783),
    [aux_sym_comparison_operator_token9] = ACTIONS(783),
    [aux_sym_comparison_operator_token10] = ACTIONS(783),
    [aux_sym_comparison_operator_token11] = ACTIONS(783),
    [aux_sym_comparison_operator_token12] = ACTIONS(783),
    [aux_sym_comparison_operator_token13] = ACTIONS(783),
    [aux_sym_comparison_operator_token14] = ACTIONS(783),
    [aux_sym_comparison_operator_token15] = ACTIONS(783),
    [aux_sym_comparison_operator_token16] = ACTIONS(783),
    [aux_sym_comparison_operator_token17] = ACTIONS(783),
    [aux_sym_comparison_operator_token18] = ACTIONS(783),
    [aux_sym_comparison_operator_token19] = ACTIONS(783),
    [aux_sym_comparison_operator_token20] = ACTIONS(783),
    [aux_sym_comparison_operator_token21] = ACTIONS(783),
    [aux_sym_comparison_operator_token22] = ACTIONS(783),
    [aux_sym_comparison_operator_token23] = ACTIONS(783),
    [aux_sym_comparison_operator_token24] = ACTIONS(783),
    [aux_sym_comparison_operator_token25] = ACTIONS(783),
    [aux_sym_comparison_operator_token26] = ACTIONS(783),
    [aux_sym_comparison_operator_token27] = ACTIONS(783),
    [aux_sym_comparison_operator_token28] = ACTIONS(785),
    [aux_sym_comparison_operator_token29] = ACTIONS(783),
    [aux_sym_comparison_operator_token30] = ACTIONS(783),
    [aux_sym_comparison_operator_token31] = ACTIONS(783),
    [aux_sym_comparison_operator_token32] = ACTIONS(783),
    [aux_sym_comparison_operator_token33] = ACTIONS(783),
    [aux_sym_comparison_operator_token34] = ACTIONS(785),
    [aux_sym_comparison_operator_token35] = ACTIONS(783),
    [aux_sym_comparison_operator_token36] = ACTIONS(783),
    [aux_sym_comparison_operator_token37] = ACTIONS(783),
    [aux_sym_comparison_operator_token38] = ACTIONS(783),
    [aux_sym_comparison_operator_token39] = ACTIONS(783),
    [aux_sym_comparison_operator_token40] = ACTIONS(783),
    [aux_sym_comparison_operator_token41] = ACTIONS(783),
    [aux_sym_comparison_operator_token42] = ACTIONS(783),
    [aux_sym_comparison_operator_token43] = ACTIONS(783),
    [aux_sym_comparison_operator_token44] = ACTIONS(783),
    [aux_sym_comparison_operator_token45] = ACTIONS(783),
    [aux_sym_comparison_operator_token46] = ACTIONS(783),
    [aux_sym_comparison_operator_token47] = ACTIONS(783),
    [aux_sym_comparison_operator_token48] = ACTIONS(783),
    [aux_sym_comparison_operator_token49] = ACTIONS(783),
    [aux_sym_comparison_operator_token50] = ACTIONS(783),
    [aux_sym_format_operator_token1] = ACTIONS(783),
    [anon_sym_RPAREN] = ACTIONS(783),
    [anon_sym_COMMA] = ACTIONS(783),
    [aux_sym_logical_expression_token1] = ACTIONS(783),
    [aux_sym_logical_expression_token2] = ACTIONS(783),
    [aux_sym_logical_expression_token3] = ACTIONS(783),
    [aux_sym_bitwise_expression_token1] = ACTIONS(783),
    [aux_sym_bitwise_expression_token2] = ACTIONS(783),
    [aux_sym_bitwise_expression_token3] = ACTIONS(783),
    [anon_sym_PLUS] = ACTIONS(783),
    [anon_sym_DASH] = ACTIONS(785),
    [anon_sym_SLASH] = ACTIONS(783),
    [anon_sym_BSLASH] = ACTIONS(783),
    [anon_sym_PERCENT] = ACTIONS(783),
    [anon_sym_STAR] = ACTIONS(783),
    [anon_sym_DOT_DOT] = ACTIONS(783),
  },
  [STATE(457)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(458),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1625),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
    [sym__statement_terminator] = ACTIONS(1593),
  },
  [STATE(458)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(458),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [sym_command_parameter] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [sym_stop_parsing] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1627),
    [anon_sym_COLON] = ACTIONS(1595),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
    [sym__statement_terminator] = ACTIONS(1606),
  },
  [STATE(459)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(459),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [sym_command_parameter] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_RPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [sym_stop_parsing] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1630),
    [anon_sym_COLON] = ACTIONS(1595),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
  },
  [STATE(460)] = {
    [sym_format_operator] = STATE(584),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(790),
    [aux_sym_comparison_operator_token2] = ACTIONS(790),
    [aux_sym_comparison_operator_token3] = ACTIONS(790),
    [aux_sym_comparison_operator_token4] = ACTIONS(790),
    [aux_sym_comparison_operator_token5] = ACTIONS(790),
    [aux_sym_comparison_operator_token6] = ACTIONS(790),
    [aux_sym_comparison_operator_token7] = ACTIONS(790),
    [aux_sym_comparison_operator_token8] = ACTIONS(790),
    [aux_sym_comparison_operator_token9] = ACTIONS(790),
    [aux_sym_comparison_operator_token10] = ACTIONS(790),
    [aux_sym_comparison_operator_token11] = ACTIONS(790),
    [aux_sym_comparison_operator_token12] = ACTIONS(790),
    [aux_sym_comparison_operator_token13] = ACTIONS(790),
    [aux_sym_comparison_operator_token14] = ACTIONS(790),
    [aux_sym_comparison_operator_token15] = ACTIONS(790),
    [aux_sym_comparison_operator_token16] = ACTIONS(790),
    [aux_sym_comparison_operator_token17] = ACTIONS(790),
    [aux_sym_comparison_operator_token18] = ACTIONS(790),
    [aux_sym_comparison_operator_token19] = ACTIONS(790),
    [aux_sym_comparison_operator_token20] = ACTIONS(790),
    [aux_sym_comparison_operator_token21] = ACTIONS(790),
    [aux_sym_comparison_operator_token22] = ACTIONS(790),
    [aux_sym_comparison_operator_token23] = ACTIONS(790),
    [aux_sym_comparison_operator_token24] = ACTIONS(790),
    [aux_sym_comparison_operator_token25] = ACTIONS(790),
    [aux_sym_comparison_operator_token26] = ACTIONS(790),
    [aux_sym_comparison_operator_token27] = ACTIONS(790),
    [aux_sym_comparison_operator_token28] = ACTIONS(792),
    [aux_sym_comparison_operator_token29] = ACTIONS(790),
    [aux_sym_comparison_operator_token30] = ACTIONS(790),
    [aux_sym_comparison_operator_token31] = ACTIONS(790),
    [aux_sym_comparison_operator_token32] = ACTIONS(790),
    [aux_sym_comparison_operator_token33] = ACTIONS(790),
    [aux_sym_comparison_operator_token34] = ACTIONS(792),
    [aux_sym_comparison_operator_token35] = ACTIONS(790),
    [aux_sym_comparison_operator_token36] = ACTIONS(790),
    [aux_sym_comparison_operator_token37] = ACTIONS(790),
    [aux_sym_comparison_operator_token38] = ACTIONS(790),
    [aux_sym_comparison_operator_token39] = ACTIONS(790),
    [aux_sym_comparison_operator_token40] = ACTIONS(790),
    [aux_sym_comparison_operator_token41] = ACTIONS(790),
    [aux_sym_comparison_operator_token42] = ACTIONS(790),
    [aux_sym_comparison_operator_token43] = ACTIONS(790),
    [aux_sym_comparison_operator_token44] = ACTIONS(790),
    [aux_sym_comparison_operator_token45] = ACTIONS(790),
    [aux_sym_comparison_operator_token46] = ACTIONS(790),
    [aux_sym_comparison_operator_token47] = ACTIONS(790),
    [aux_sym_comparison_operator_token48] = ACTIONS(790),
    [aux_sym_comparison_operator_token49] = ACTIONS(790),
    [aux_sym_comparison_operator_token50] = ACTIONS(790),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_RPAREN] = ACTIONS(790),
    [anon_sym_COMMA] = ACTIONS(790),
    [aux_sym_logical_expression_token1] = ACTIONS(790),
    [aux_sym_logical_expression_token2] = ACTIONS(790),
    [aux_sym_logical_expression_token3] = ACTIONS(790),
    [aux_sym_bitwise_expression_token1] = ACTIONS(790),
    [aux_sym_bitwise_expression_token2] = ACTIONS(790),
    [aux_sym_bitwise_expression_token3] = ACTIONS(790),
    [anon_sym_PLUS] = ACTIONS(790),
    [anon_sym_DASH] = ACTIONS(792),
    [anon_sym_SLASH] = ACTIONS(790),
    [anon_sym_BSLASH] = ACTIONS(790),
    [anon_sym_PERCENT] = ACTIONS(790),
    [anon_sym_STAR] = ACTIONS(790),
  },
  [STATE(461)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1589),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
  },
  [STATE(462)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(800),
    [aux_sym_comparison_operator_token2] = ACTIONS(800),
    [aux_sym_comparison_operator_token3] = ACTIONS(800),
    [aux_sym_comparison_operator_token4] = ACTIONS(800),
    [aux_sym_comparison_operator_token5] = ACTIONS(800),
    [aux_sym_comparison_operator_token6] = ACTIONS(800),
    [aux_sym_comparison_operator_token7] = ACTIONS(800),
    [aux_sym_comparison_operator_token8] = ACTIONS(800),
    [aux_sym_comparison_operator_token9] = ACTIONS(800),
    [aux_sym_comparison_operator_token10] = ACTIONS(800),
    [aux_sym_comparison_operator_token11] = ACTIONS(800),
    [aux_sym_comparison_operator_token12] = ACTIONS(800),
    [aux_sym_comparison_operator_token13] = ACTIONS(800),
    [aux_sym_comparison_operator_token14] = ACTIONS(800),
    [aux_sym_comparison_operator_token15] = ACTIONS(800),
    [aux_sym_comparison_operator_token16] = ACTIONS(800),
    [aux_sym_comparison_operator_token17] = ACTIONS(800),
    [aux_sym_comparison_operator_token18] = ACTIONS(800),
    [aux_sym_comparison_operator_token19] = ACTIONS(800),
    [aux_sym_comparison_operator_token20] = ACTIONS(800),
    [aux_sym_comparison_operator_token21] = ACTIONS(800),
    [aux_sym_comparison_operator_token22] = ACTIONS(800),
    [aux_sym_comparison_operator_token23] = ACTIONS(800),
    [aux_sym_comparison_operator_token24] = ACTIONS(800),
    [aux_sym_comparison_operator_token25] = ACTIONS(800),
    [aux_sym_comparison_operator_token26] = ACTIONS(800),
    [aux_sym_comparison_operator_token27] = ACTIONS(800),
    [aux_sym_comparison_operator_token28] = ACTIONS(802),
    [aux_sym_comparison_operator_token29] = ACTIONS(800),
    [aux_sym_comparison_operator_token30] = ACTIONS(800),
    [aux_sym_comparison_operator_token31] = ACTIONS(800),
    [aux_sym_comparison_operator_token32] = ACTIONS(800),
    [aux_sym_comparison_operator_token33] = ACTIONS(800),
    [aux_sym_comparison_operator_token34] = ACTIONS(802),
    [aux_sym_comparison_operator_token35] = ACTIONS(800),
    [aux_sym_comparison_operator_token36] = ACTIONS(800),
    [aux_sym_comparison_operator_token37] = ACTIONS(800),
    [aux_sym_comparison_operator_token38] = ACTIONS(800),
    [aux_sym_comparison_operator_token39] = ACTIONS(800),
    [aux_sym_comparison_operator_token40] = ACTIONS(800),
    [aux_sym_comparison_operator_token41] = ACTIONS(800),
    [aux_sym_comparison_operator_token42] = ACTIONS(800),
    [aux_sym_comparison_operator_token43] = ACTIONS(800),
    [aux_sym_comparison_operator_token44] = ACTIONS(800),
    [aux_sym_comparison_operator_token45] = ACTIONS(800),
    [aux_sym_comparison_operator_token46] = ACTIONS(800),
    [aux_sym_comparison_operator_token47] = ACTIONS(800),
    [aux_sym_comparison_operator_token48] = ACTIONS(800),
    [aux_sym_comparison_operator_token49] = ACTIONS(800),
    [aux_sym_comparison_operator_token50] = ACTIONS(800),
    [aux_sym_format_operator_token1] = ACTIONS(800),
    [anon_sym_RPAREN] = ACTIONS(800),
    [anon_sym_COMMA] = ACTIONS(800),
    [aux_sym_logical_expression_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token2] = ACTIONS(800),
    [aux_sym_logical_expression_token3] = ACTIONS(800),
    [aux_sym_bitwise_expression_token1] = ACTIONS(800),
    [aux_sym_bitwise_expression_token2] = ACTIONS(800),
    [aux_sym_bitwise_expression_token3] = ACTIONS(800),
    [anon_sym_PLUS] = ACTIONS(800),
    [anon_sym_DASH] = ACTIONS(802),
    [anon_sym_SLASH] = ACTIONS(800),
    [anon_sym_BSLASH] = ACTIONS(800),
    [anon_sym_PERCENT] = ACTIONS(800),
    [anon_sym_STAR] = ACTIONS(800),
    [anon_sym_DOT_DOT] = ACTIONS(1633),
  },
  [STATE(463)] = {
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [sym_command_parameter] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [sym_stop_parsing] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1589),
    [anon_sym_COLON] = ACTIONS(1589),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
    [sym__statement_terminator] = ACTIONS(1593),
  },
  [STATE(464)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1678),
    [sym_logical_expression] = STATE(1379),
    [sym_bitwise_expression] = STATE(1293),
    [sym_comparison_expression] = STATE(560),
    [sym_additive_expression] = STATE(546),
    [sym_multiplicative_expression] = STATE(496),
    [sym_format_expression] = STATE(460),
    [sym_range_expression] = STATE(462),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(454),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_attribute_argument] = STATE(1622),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [sym_simple_name] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(465)] = {
    [sym_format_operator] = STATE(584),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(796),
    [aux_sym_comparison_operator_token2] = ACTIONS(796),
    [aux_sym_comparison_operator_token3] = ACTIONS(796),
    [aux_sym_comparison_operator_token4] = ACTIONS(796),
    [aux_sym_comparison_operator_token5] = ACTIONS(796),
    [aux_sym_comparison_operator_token6] = ACTIONS(796),
    [aux_sym_comparison_operator_token7] = ACTIONS(796),
    [aux_sym_comparison_operator_token8] = ACTIONS(796),
    [aux_sym_comparison_operator_token9] = ACTIONS(796),
    [aux_sym_comparison_operator_token10] = ACTIONS(796),
    [aux_sym_comparison_operator_token11] = ACTIONS(796),
    [aux_sym_comparison_operator_token12] = ACTIONS(796),
    [aux_sym_comparison_operator_token13] = ACTIONS(796),
    [aux_sym_comparison_operator_token14] = ACTIONS(796),
    [aux_sym_comparison_operator_token15] = ACTIONS(796),
    [aux_sym_comparison_operator_token16] = ACTIONS(796),
    [aux_sym_comparison_operator_token17] = ACTIONS(796),
    [aux_sym_comparison_operator_token18] = ACTIONS(796),
    [aux_sym_comparison_operator_token19] = ACTIONS(796),
    [aux_sym_comparison_operator_token20] = ACTIONS(796),
    [aux_sym_comparison_operator_token21] = ACTIONS(796),
    [aux_sym_comparison_operator_token22] = ACTIONS(796),
    [aux_sym_comparison_operator_token23] = ACTIONS(796),
    [aux_sym_comparison_operator_token24] = ACTIONS(796),
    [aux_sym_comparison_operator_token25] = ACTIONS(796),
    [aux_sym_comparison_operator_token26] = ACTIONS(796),
    [aux_sym_comparison_operator_token27] = ACTIONS(796),
    [aux_sym_comparison_operator_token28] = ACTIONS(798),
    [aux_sym_comparison_operator_token29] = ACTIONS(796),
    [aux_sym_comparison_operator_token30] = ACTIONS(796),
    [aux_sym_comparison_operator_token31] = ACTIONS(796),
    [aux_sym_comparison_operator_token32] = ACTIONS(796),
    [aux_sym_comparison_operator_token33] = ACTIONS(796),
    [aux_sym_comparison_operator_token34] = ACTIONS(798),
    [aux_sym_comparison_operator_token35] = ACTIONS(796),
    [aux_sym_comparison_operator_token36] = ACTIONS(796),
    [aux_sym_comparison_operator_token37] = ACTIONS(796),
    [aux_sym_comparison_operator_token38] = ACTIONS(796),
    [aux_sym_comparison_operator_token39] = ACTIONS(796),
    [aux_sym_comparison_operator_token40] = ACTIONS(796),
    [aux_sym_comparison_operator_token41] = ACTIONS(796),
    [aux_sym_comparison_operator_token42] = ACTIONS(796),
    [aux_sym_comparison_operator_token43] = ACTIONS(796),
    [aux_sym_comparison_operator_token44] = ACTIONS(796),
    [aux_sym_comparison_operator_token45] = ACTIONS(796),
    [aux_sym_comparison_operator_token46] = ACTIONS(796),
    [aux_sym_comparison_operator_token47] = ACTIONS(796),
    [aux_sym_comparison_operator_token48] = ACTIONS(796),
    [aux_sym_comparison_operator_token49] = ACTIONS(796),
    [aux_sym_comparison_operator_token50] = ACTIONS(796),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [anon_sym_RPAREN] = ACTIONS(796),
    [anon_sym_COMMA] = ACTIONS(796),
    [aux_sym_logical_expression_token1] = ACTIONS(796),
    [aux_sym_logical_expression_token2] = ACTIONS(796),
    [aux_sym_logical_expression_token3] = ACTIONS(796),
    [aux_sym_bitwise_expression_token1] = ACTIONS(796),
    [aux_sym_bitwise_expression_token2] = ACTIONS(796),
    [aux_sym_bitwise_expression_token3] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(796),
    [anon_sym_DASH] = ACTIONS(798),
    [anon_sym_SLASH] = ACTIONS(796),
    [anon_sym_BSLASH] = ACTIONS(796),
    [anon_sym_PERCENT] = ACTIONS(796),
    [anon_sym_STAR] = ACTIONS(796),
  },
  [STATE(466)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(806),
    [aux_sym_comparison_operator_token2] = ACTIONS(806),
    [aux_sym_comparison_operator_token3] = ACTIONS(806),
    [aux_sym_comparison_operator_token4] = ACTIONS(806),
    [aux_sym_comparison_operator_token5] = ACTIONS(806),
    [aux_sym_comparison_operator_token6] = ACTIONS(806),
    [aux_sym_comparison_operator_token7] = ACTIONS(806),
    [aux_sym_comparison_operator_token8] = ACTIONS(806),
    [aux_sym_comparison_operator_token9] = ACTIONS(806),
    [aux_sym_comparison_operator_token10] = ACTIONS(806),
    [aux_sym_comparison_operator_token11] = ACTIONS(806),
    [aux_sym_comparison_operator_token12] = ACTIONS(806),
    [aux_sym_comparison_operator_token13] = ACTIONS(806),
    [aux_sym_comparison_operator_token14] = ACTIONS(806),
    [aux_sym_comparison_operator_token15] = ACTIONS(806),
    [aux_sym_comparison_operator_token16] = ACTIONS(806),
    [aux_sym_comparison_operator_token17] = ACTIONS(806),
    [aux_sym_comparison_operator_token18] = ACTIONS(806),
    [aux_sym_comparison_operator_token19] = ACTIONS(806),
    [aux_sym_comparison_operator_token20] = ACTIONS(806),
    [aux_sym_comparison_operator_token21] = ACTIONS(806),
    [aux_sym_comparison_operator_token22] = ACTIONS(806),
    [aux_sym_comparison_operator_token23] = ACTIONS(806),
    [aux_sym_comparison_operator_token24] = ACTIONS(806),
    [aux_sym_comparison_operator_token25] = ACTIONS(806),
    [aux_sym_comparison_operator_token26] = ACTIONS(806),
    [aux_sym_comparison_operator_token27] = ACTIONS(806),
    [aux_sym_comparison_operator_token28] = ACTIONS(808),
    [aux_sym_comparison_operator_token29] = ACTIONS(806),
    [aux_sym_comparison_operator_token30] = ACTIONS(806),
    [aux_sym_comparison_operator_token31] = ACTIONS(806),
    [aux_sym_comparison_operator_token32] = ACTIONS(806),
    [aux_sym_comparison_operator_token33] = ACTIONS(806),
    [aux_sym_comparison_operator_token34] = ACTIONS(808),
    [aux_sym_comparison_operator_token35] = ACTIONS(806),
    [aux_sym_comparison_operator_token36] = ACTIONS(806),
    [aux_sym_comparison_operator_token37] = ACTIONS(806),
    [aux_sym_comparison_operator_token38] = ACTIONS(806),
    [aux_sym_comparison_operator_token39] = ACTIONS(806),
    [aux_sym_comparison_operator_token40] = ACTIONS(806),
    [aux_sym_comparison_operator_token41] = ACTIONS(806),
    [aux_sym_comparison_operator_token42] = ACTIONS(806),
    [aux_sym_comparison_operator_token43] = ACTIONS(806),
    [aux_sym_comparison_operator_token44] = ACTIONS(806),
    [aux_sym_comparison_operator_token45] = ACTIONS(806),
    [aux_sym_comparison_operator_token46] = ACTIONS(806),
    [aux_sym_comparison_operator_token47] = ACTIONS(806),
    [aux_sym_comparison_operator_token48] = ACTIONS(806),
    [aux_sym_comparison_operator_token49] = ACTIONS(806),
    [aux_sym_comparison_operator_token50] = ACTIONS(806),
    [aux_sym_format_operator_token1] = ACTIONS(806),
    [anon_sym_RPAREN] = ACTIONS(806),
    [anon_sym_COMMA] = ACTIONS(806),
    [aux_sym_logical_expression_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token2] = ACTIONS(806),
    [aux_sym_logical_expression_token3] = ACTIONS(806),
    [aux_sym_bitwise_expression_token1] = ACTIONS(806),
    [aux_sym_bitwise_expression_token2] = ACTIONS(806),
    [aux_sym_bitwise_expression_token3] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_DASH] = ACTIONS(808),
    [anon_sym_SLASH] = ACTIONS(806),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_PERCENT] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(806),
    [anon_sym_DOT_DOT] = ACTIONS(1633),
  },
  [STATE(467)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(800),
    [aux_sym_comparison_operator_token2] = ACTIONS(800),
    [aux_sym_comparison_operator_token3] = ACTIONS(800),
    [aux_sym_comparison_operator_token4] = ACTIONS(800),
    [aux_sym_comparison_operator_token5] = ACTIONS(800),
    [aux_sym_comparison_operator_token6] = ACTIONS(800),
    [aux_sym_comparison_operator_token7] = ACTIONS(800),
    [aux_sym_comparison_operator_token8] = ACTIONS(800),
    [aux_sym_comparison_operator_token9] = ACTIONS(800),
    [aux_sym_comparison_operator_token10] = ACTIONS(800),
    [aux_sym_comparison_operator_token11] = ACTIONS(800),
    [aux_sym_comparison_operator_token12] = ACTIONS(800),
    [aux_sym_comparison_operator_token13] = ACTIONS(800),
    [aux_sym_comparison_operator_token14] = ACTIONS(800),
    [aux_sym_comparison_operator_token15] = ACTIONS(800),
    [aux_sym_comparison_operator_token16] = ACTIONS(800),
    [aux_sym_comparison_operator_token17] = ACTIONS(800),
    [aux_sym_comparison_operator_token18] = ACTIONS(800),
    [aux_sym_comparison_operator_token19] = ACTIONS(800),
    [aux_sym_comparison_operator_token20] = ACTIONS(800),
    [aux_sym_comparison_operator_token21] = ACTIONS(800),
    [aux_sym_comparison_operator_token22] = ACTIONS(800),
    [aux_sym_comparison_operator_token23] = ACTIONS(800),
    [aux_sym_comparison_operator_token24] = ACTIONS(800),
    [aux_sym_comparison_operator_token25] = ACTIONS(800),
    [aux_sym_comparison_operator_token26] = ACTIONS(800),
    [aux_sym_comparison_operator_token27] = ACTIONS(800),
    [aux_sym_comparison_operator_token28] = ACTIONS(802),
    [aux_sym_comparison_operator_token29] = ACTIONS(800),
    [aux_sym_comparison_operator_token30] = ACTIONS(800),
    [aux_sym_comparison_operator_token31] = ACTIONS(800),
    [aux_sym_comparison_operator_token32] = ACTIONS(800),
    [aux_sym_comparison_operator_token33] = ACTIONS(800),
    [aux_sym_comparison_operator_token34] = ACTIONS(802),
    [aux_sym_comparison_operator_token35] = ACTIONS(800),
    [aux_sym_comparison_operator_token36] = ACTIONS(800),
    [aux_sym_comparison_operator_token37] = ACTIONS(800),
    [aux_sym_comparison_operator_token38] = ACTIONS(800),
    [aux_sym_comparison_operator_token39] = ACTIONS(800),
    [aux_sym_comparison_operator_token40] = ACTIONS(800),
    [aux_sym_comparison_operator_token41] = ACTIONS(800),
    [aux_sym_comparison_operator_token42] = ACTIONS(800),
    [aux_sym_comparison_operator_token43] = ACTIONS(800),
    [aux_sym_comparison_operator_token44] = ACTIONS(800),
    [aux_sym_comparison_operator_token45] = ACTIONS(800),
    [aux_sym_comparison_operator_token46] = ACTIONS(800),
    [aux_sym_comparison_operator_token47] = ACTIONS(800),
    [aux_sym_comparison_operator_token48] = ACTIONS(800),
    [aux_sym_comparison_operator_token49] = ACTIONS(800),
    [aux_sym_comparison_operator_token50] = ACTIONS(800),
    [aux_sym_format_operator_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token2] = ACTIONS(800),
    [aux_sym_logical_expression_token3] = ACTIONS(800),
    [aux_sym_bitwise_expression_token1] = ACTIONS(800),
    [aux_sym_bitwise_expression_token2] = ACTIONS(800),
    [aux_sym_bitwise_expression_token3] = ACTIONS(800),
    [anon_sym_PLUS] = ACTIONS(800),
    [anon_sym_DASH] = ACTIONS(802),
    [anon_sym_SLASH] = ACTIONS(800),
    [anon_sym_BSLASH] = ACTIONS(800),
    [anon_sym_PERCENT] = ACTIONS(800),
    [anon_sym_STAR] = ACTIONS(800),
    [anon_sym_DOT_DOT] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(800),
  },
  [STATE(468)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(2111),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1637),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(469)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(800),
    [aux_sym_comparison_operator_token2] = ACTIONS(800),
    [aux_sym_comparison_operator_token3] = ACTIONS(800),
    [aux_sym_comparison_operator_token4] = ACTIONS(800),
    [aux_sym_comparison_operator_token5] = ACTIONS(800),
    [aux_sym_comparison_operator_token6] = ACTIONS(800),
    [aux_sym_comparison_operator_token7] = ACTIONS(800),
    [aux_sym_comparison_operator_token8] = ACTIONS(800),
    [aux_sym_comparison_operator_token9] = ACTIONS(800),
    [aux_sym_comparison_operator_token10] = ACTIONS(800),
    [aux_sym_comparison_operator_token11] = ACTIONS(800),
    [aux_sym_comparison_operator_token12] = ACTIONS(800),
    [aux_sym_comparison_operator_token13] = ACTIONS(800),
    [aux_sym_comparison_operator_token14] = ACTIONS(800),
    [aux_sym_comparison_operator_token15] = ACTIONS(800),
    [aux_sym_comparison_operator_token16] = ACTIONS(800),
    [aux_sym_comparison_operator_token17] = ACTIONS(800),
    [aux_sym_comparison_operator_token18] = ACTIONS(800),
    [aux_sym_comparison_operator_token19] = ACTIONS(800),
    [aux_sym_comparison_operator_token20] = ACTIONS(800),
    [aux_sym_comparison_operator_token21] = ACTIONS(800),
    [aux_sym_comparison_operator_token22] = ACTIONS(800),
    [aux_sym_comparison_operator_token23] = ACTIONS(800),
    [aux_sym_comparison_operator_token24] = ACTIONS(800),
    [aux_sym_comparison_operator_token25] = ACTIONS(800),
    [aux_sym_comparison_operator_token26] = ACTIONS(800),
    [aux_sym_comparison_operator_token27] = ACTIONS(800),
    [aux_sym_comparison_operator_token28] = ACTIONS(802),
    [aux_sym_comparison_operator_token29] = ACTIONS(800),
    [aux_sym_comparison_operator_token30] = ACTIONS(800),
    [aux_sym_comparison_operator_token31] = ACTIONS(800),
    [aux_sym_comparison_operator_token32] = ACTIONS(800),
    [aux_sym_comparison_operator_token33] = ACTIONS(800),
    [aux_sym_comparison_operator_token34] = ACTIONS(802),
    [aux_sym_comparison_operator_token35] = ACTIONS(800),
    [aux_sym_comparison_operator_token36] = ACTIONS(800),
    [aux_sym_comparison_operator_token37] = ACTIONS(800),
    [aux_sym_comparison_operator_token38] = ACTIONS(800),
    [aux_sym_comparison_operator_token39] = ACTIONS(800),
    [aux_sym_comparison_operator_token40] = ACTIONS(800),
    [aux_sym_comparison_operator_token41] = ACTIONS(800),
    [aux_sym_comparison_operator_token42] = ACTIONS(800),
    [aux_sym_comparison_operator_token43] = ACTIONS(800),
    [aux_sym_comparison_operator_token44] = ACTIONS(800),
    [aux_sym_comparison_operator_token45] = ACTIONS(800),
    [aux_sym_comparison_operator_token46] = ACTIONS(800),
    [aux_sym_comparison_operator_token47] = ACTIONS(800),
    [aux_sym_comparison_operator_token48] = ACTIONS(800),
    [aux_sym_comparison_operator_token49] = ACTIONS(800),
    [aux_sym_comparison_operator_token50] = ACTIONS(800),
    [aux_sym_format_operator_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token1] = ACTIONS(800),
    [aux_sym_logical_expression_token2] = ACTIONS(800),
    [aux_sym_logical_expression_token3] = ACTIONS(800),
    [aux_sym_bitwise_expression_token1] = ACTIONS(800),
    [aux_sym_bitwise_expression_token2] = ACTIONS(800),
    [aux_sym_bitwise_expression_token3] = ACTIONS(800),
    [anon_sym_PLUS] = ACTIONS(800),
    [anon_sym_DASH] = ACTIONS(802),
    [anon_sym_SLASH] = ACTIONS(800),
    [anon_sym_BSLASH] = ACTIONS(800),
    [anon_sym_PERCENT] = ACTIONS(800),
    [anon_sym_STAR] = ACTIONS(800),
    [anon_sym_DOT_DOT] = ACTIONS(1639),
    [sym__statement_terminator] = ACTIONS(800),
  },
  [STATE(470)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1706),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(471)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1768),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1643),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(472)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1842),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1645),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(473)] = {
    [sym_format_operator] = STATE(587),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(796),
    [aux_sym_comparison_operator_token2] = ACTIONS(796),
    [aux_sym_comparison_operator_token3] = ACTIONS(796),
    [aux_sym_comparison_operator_token4] = ACTIONS(796),
    [aux_sym_comparison_operator_token5] = ACTIONS(796),
    [aux_sym_comparison_operator_token6] = ACTIONS(796),
    [aux_sym_comparison_operator_token7] = ACTIONS(796),
    [aux_sym_comparison_operator_token8] = ACTIONS(796),
    [aux_sym_comparison_operator_token9] = ACTIONS(796),
    [aux_sym_comparison_operator_token10] = ACTIONS(796),
    [aux_sym_comparison_operator_token11] = ACTIONS(796),
    [aux_sym_comparison_operator_token12] = ACTIONS(796),
    [aux_sym_comparison_operator_token13] = ACTIONS(796),
    [aux_sym_comparison_operator_token14] = ACTIONS(796),
    [aux_sym_comparison_operator_token15] = ACTIONS(796),
    [aux_sym_comparison_operator_token16] = ACTIONS(796),
    [aux_sym_comparison_operator_token17] = ACTIONS(796),
    [aux_sym_comparison_operator_token18] = ACTIONS(796),
    [aux_sym_comparison_operator_token19] = ACTIONS(796),
    [aux_sym_comparison_operator_token20] = ACTIONS(796),
    [aux_sym_comparison_operator_token21] = ACTIONS(796),
    [aux_sym_comparison_operator_token22] = ACTIONS(796),
    [aux_sym_comparison_operator_token23] = ACTIONS(796),
    [aux_sym_comparison_operator_token24] = ACTIONS(796),
    [aux_sym_comparison_operator_token25] = ACTIONS(796),
    [aux_sym_comparison_operator_token26] = ACTIONS(796),
    [aux_sym_comparison_operator_token27] = ACTIONS(796),
    [aux_sym_comparison_operator_token28] = ACTIONS(798),
    [aux_sym_comparison_operator_token29] = ACTIONS(796),
    [aux_sym_comparison_operator_token30] = ACTIONS(796),
    [aux_sym_comparison_operator_token31] = ACTIONS(796),
    [aux_sym_comparison_operator_token32] = ACTIONS(796),
    [aux_sym_comparison_operator_token33] = ACTIONS(796),
    [aux_sym_comparison_operator_token34] = ACTIONS(798),
    [aux_sym_comparison_operator_token35] = ACTIONS(796),
    [aux_sym_comparison_operator_token36] = ACTIONS(796),
    [aux_sym_comparison_operator_token37] = ACTIONS(796),
    [aux_sym_comparison_operator_token38] = ACTIONS(796),
    [aux_sym_comparison_operator_token39] = ACTIONS(796),
    [aux_sym_comparison_operator_token40] = ACTIONS(796),
    [aux_sym_comparison_operator_token41] = ACTIONS(796),
    [aux_sym_comparison_operator_token42] = ACTIONS(796),
    [aux_sym_comparison_operator_token43] = ACTIONS(796),
    [aux_sym_comparison_operator_token44] = ACTIONS(796),
    [aux_sym_comparison_operator_token45] = ACTIONS(796),
    [aux_sym_comparison_operator_token46] = ACTIONS(796),
    [aux_sym_comparison_operator_token47] = ACTIONS(796),
    [aux_sym_comparison_operator_token48] = ACTIONS(796),
    [aux_sym_comparison_operator_token49] = ACTIONS(796),
    [aux_sym_comparison_operator_token50] = ACTIONS(796),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [aux_sym_logical_expression_token1] = ACTIONS(796),
    [aux_sym_logical_expression_token2] = ACTIONS(796),
    [aux_sym_logical_expression_token3] = ACTIONS(796),
    [aux_sym_bitwise_expression_token1] = ACTIONS(796),
    [aux_sym_bitwise_expression_token2] = ACTIONS(796),
    [aux_sym_bitwise_expression_token3] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(796),
    [anon_sym_DASH] = ACTIONS(798),
    [anon_sym_SLASH] = ACTIONS(796),
    [anon_sym_BSLASH] = ACTIONS(796),
    [anon_sym_PERCENT] = ACTIONS(796),
    [anon_sym_STAR] = ACTIONS(796),
    [anon_sym_RBRACK] = ACTIONS(796),
  },
  [STATE(474)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1791),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1647),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(475)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1813),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1649),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(476)] = {
    [sym__literal] = STATE(1285),
    [sym_integer_literal] = STATE(1285),
    [sym_string_literal] = STATE(1285),
    [sym_expandable_string_literal] = STATE(933),
    [sym_expandable_here_string_literal] = STATE(933),
    [sym_variable] = STATE(1223),
    [sym_data_commands_list] = STATE(1835),
    [sym_data_command] = STATE(1482),
    [sym_command_name] = STATE(1632),
    [sym_path_command_name] = STATE(1632),
    [sym_command_name_expr] = STATE(1569),
    [sym__primary_expression] = STATE(1285),
    [sym__value] = STATE(1285),
    [sym_parenthesized_expression] = STATE(1285),
    [sym_sub_expression] = STATE(1285),
    [sym_array_expression] = STATE(1285),
    [sym_script_block_expression] = STATE(1285),
    [sym_hash_literal_expression] = STATE(1285),
    [sym_post_increment_expression] = STATE(1285),
    [sym_post_decrement_expression] = STATE(1285),
    [sym_member_access] = STATE(1285),
    [sym_element_access] = STATE(1285),
    [sym_invokation_expression] = STATE(1285),
    [sym_invokation_foreach_expression] = STATE(945),
    [sym_type_literal] = STATE(1285),
    [aux_sym_path_command_name_repeat1] = STATE(1248),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1140),
    [sym_hexadecimal_integer_literal] = ACTIONS(1140),
    [sym_real_literal] = ACTIONS(1651),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1234),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1236),
    [sym_verbatim_string_characters] = ACTIONS(1238),
    [sym_verbatim_here_string_characters] = ACTIONS(1238),
    [anon_sym_LBRACK] = ACTIONS(1653),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1655),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1655),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1655),
    [anon_sym_DOLLAR_] = ACTIONS(1655),
    [aux_sym_variable_token1] = ACTIONS(1655),
    [aux_sym_variable_token2] = ACTIONS(1655),
    [sym_braced_variable] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1244),
    [anon_sym_LBRACE] = ACTIONS(1246),
    [aux_sym_command_name_token1] = ACTIONS(1657),
    [aux_sym_command_name_token2] = ACTIONS(1657),
    [aux_sym_command_name_token3] = ACTIONS(1657),
    [aux_sym_command_name_token4] = ACTIONS(1657),
    [aux_sym_command_name_token5] = ACTIONS(1657),
    [aux_sym_command_name_token6] = ACTIONS(1657),
    [aux_sym_command_name_token7] = ACTIONS(1657),
    [aux_sym_command_name_token8] = ACTIONS(1657),
    [aux_sym_command_name_token9] = ACTIONS(1657),
    [aux_sym_command_name_token10] = ACTIONS(1657),
    [aux_sym_command_name_token11] = ACTIONS(1657),
    [aux_sym_command_name_token12] = ACTIONS(1657),
    [aux_sym_command_name_token13] = ACTIONS(1657),
    [aux_sym_command_name_token14] = ACTIONS(1657),
    [aux_sym_command_name_token15] = ACTIONS(1657),
    [aux_sym_command_name_token16] = ACTIONS(1657),
    [aux_sym_command_name_token17] = ACTIONS(1657),
    [aux_sym_command_name_token18] = ACTIONS(1657),
    [sym_path_command_name_token] = ACTIONS(1659),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1252),
    [anon_sym_AT_LPAREN] = ACTIONS(1254),
    [anon_sym_AT_LBRACE] = ACTIONS(1256),
  },
  [STATE(477)] = {
    [sym_format_operator] = STATE(583),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(790),
    [aux_sym_comparison_operator_token2] = ACTIONS(790),
    [aux_sym_comparison_operator_token3] = ACTIONS(790),
    [aux_sym_comparison_operator_token4] = ACTIONS(790),
    [aux_sym_comparison_operator_token5] = ACTIONS(790),
    [aux_sym_comparison_operator_token6] = ACTIONS(790),
    [aux_sym_comparison_operator_token7] = ACTIONS(790),
    [aux_sym_comparison_operator_token8] = ACTIONS(790),
    [aux_sym_comparison_operator_token9] = ACTIONS(790),
    [aux_sym_comparison_operator_token10] = ACTIONS(790),
    [aux_sym_comparison_operator_token11] = ACTIONS(790),
    [aux_sym_comparison_operator_token12] = ACTIONS(790),
    [aux_sym_comparison_operator_token13] = ACTIONS(790),
    [aux_sym_comparison_operator_token14] = ACTIONS(790),
    [aux_sym_comparison_operator_token15] = ACTIONS(790),
    [aux_sym_comparison_operator_token16] = ACTIONS(790),
    [aux_sym_comparison_operator_token17] = ACTIONS(790),
    [aux_sym_comparison_operator_token18] = ACTIONS(790),
    [aux_sym_comparison_operator_token19] = ACTIONS(790),
    [aux_sym_comparison_operator_token20] = ACTIONS(790),
    [aux_sym_comparison_operator_token21] = ACTIONS(790),
    [aux_sym_comparison_operator_token22] = ACTIONS(790),
    [aux_sym_comparison_operator_token23] = ACTIONS(790),
    [aux_sym_comparison_operator_token24] = ACTIONS(790),
    [aux_sym_comparison_operator_token25] = ACTIONS(790),
    [aux_sym_comparison_operator_token26] = ACTIONS(790),
    [aux_sym_comparison_operator_token27] = ACTIONS(790),
    [aux_sym_comparison_operator_token28] = ACTIONS(792),
    [aux_sym_comparison_operator_token29] = ACTIONS(790),
    [aux_sym_comparison_operator_token30] = ACTIONS(790),
    [aux_sym_comparison_operator_token31] = ACTIONS(790),
    [aux_sym_comparison_operator_token32] = ACTIONS(790),
    [aux_sym_comparison_operator_token33] = ACTIONS(790),
    [aux_sym_comparison_operator_token34] = ACTIONS(792),
    [aux_sym_comparison_operator_token35] = ACTIONS(790),
    [aux_sym_comparison_operator_token36] = ACTIONS(790),
    [aux_sym_comparison_operator_token37] = ACTIONS(790),
    [aux_sym_comparison_operator_token38] = ACTIONS(790),
    [aux_sym_comparison_operator_token39] = ACTIONS(790),
    [aux_sym_comparison_operator_token40] = ACTIONS(790),
    [aux_sym_comparison_operator_token41] = ACTIONS(790),
    [aux_sym_comparison_operator_token42] = ACTIONS(790),
    [aux_sym_comparison_operator_token43] = ACTIONS(790),
    [aux_sym_comparison_operator_token44] = ACTIONS(790),
    [aux_sym_comparison_operator_token45] = ACTIONS(790),
    [aux_sym_comparison_operator_token46] = ACTIONS(790),
    [aux_sym_comparison_operator_token47] = ACTIONS(790),
    [aux_sym_comparison_operator_token48] = ACTIONS(790),
    [aux_sym_comparison_operator_token49] = ACTIONS(790),
    [aux_sym_comparison_operator_token50] = ACTIONS(790),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [aux_sym_logical_expression_token1] = ACTIONS(790),
    [aux_sym_logical_expression_token2] = ACTIONS(790),
    [aux_sym_logical_expression_token3] = ACTIONS(790),
    [aux_sym_bitwise_expression_token1] = ACTIONS(790),
    [aux_sym_bitwise_expression_token2] = ACTIONS(790),
    [aux_sym_bitwise_expression_token3] = ACTIONS(790),
    [anon_sym_PLUS] = ACTIONS(790),
    [anon_sym_DASH] = ACTIONS(792),
    [anon_sym_SLASH] = ACTIONS(790),
    [anon_sym_BSLASH] = ACTIONS(790),
    [anon_sym_PERCENT] = ACTIONS(790),
    [anon_sym_STAR] = ACTIONS(790),
    [sym__statement_terminator] = ACTIONS(790),
  },
  [STATE(478)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(806),
    [aux_sym_comparison_operator_token2] = ACTIONS(806),
    [aux_sym_comparison_operator_token3] = ACTIONS(806),
    [aux_sym_comparison_operator_token4] = ACTIONS(806),
    [aux_sym_comparison_operator_token5] = ACTIONS(806),
    [aux_sym_comparison_operator_token6] = ACTIONS(806),
    [aux_sym_comparison_operator_token7] = ACTIONS(806),
    [aux_sym_comparison_operator_token8] = ACTIONS(806),
    [aux_sym_comparison_operator_token9] = ACTIONS(806),
    [aux_sym_comparison_operator_token10] = ACTIONS(806),
    [aux_sym_comparison_operator_token11] = ACTIONS(806),
    [aux_sym_comparison_operator_token12] = ACTIONS(806),
    [aux_sym_comparison_operator_token13] = ACTIONS(806),
    [aux_sym_comparison_operator_token14] = ACTIONS(806),
    [aux_sym_comparison_operator_token15] = ACTIONS(806),
    [aux_sym_comparison_operator_token16] = ACTIONS(806),
    [aux_sym_comparison_operator_token17] = ACTIONS(806),
    [aux_sym_comparison_operator_token18] = ACTIONS(806),
    [aux_sym_comparison_operator_token19] = ACTIONS(806),
    [aux_sym_comparison_operator_token20] = ACTIONS(806),
    [aux_sym_comparison_operator_token21] = ACTIONS(806),
    [aux_sym_comparison_operator_token22] = ACTIONS(806),
    [aux_sym_comparison_operator_token23] = ACTIONS(806),
    [aux_sym_comparison_operator_token24] = ACTIONS(806),
    [aux_sym_comparison_operator_token25] = ACTIONS(806),
    [aux_sym_comparison_operator_token26] = ACTIONS(806),
    [aux_sym_comparison_operator_token27] = ACTIONS(806),
    [aux_sym_comparison_operator_token28] = ACTIONS(808),
    [aux_sym_comparison_operator_token29] = ACTIONS(806),
    [aux_sym_comparison_operator_token30] = ACTIONS(806),
    [aux_sym_comparison_operator_token31] = ACTIONS(806),
    [aux_sym_comparison_operator_token32] = ACTIONS(806),
    [aux_sym_comparison_operator_token33] = ACTIONS(806),
    [aux_sym_comparison_operator_token34] = ACTIONS(808),
    [aux_sym_comparison_operator_token35] = ACTIONS(806),
    [aux_sym_comparison_operator_token36] = ACTIONS(806),
    [aux_sym_comparison_operator_token37] = ACTIONS(806),
    [aux_sym_comparison_operator_token38] = ACTIONS(806),
    [aux_sym_comparison_operator_token39] = ACTIONS(806),
    [aux_sym_comparison_operator_token40] = ACTIONS(806),
    [aux_sym_comparison_operator_token41] = ACTIONS(806),
    [aux_sym_comparison_operator_token42] = ACTIONS(806),
    [aux_sym_comparison_operator_token43] = ACTIONS(806),
    [aux_sym_comparison_operator_token44] = ACTIONS(806),
    [aux_sym_comparison_operator_token45] = ACTIONS(806),
    [aux_sym_comparison_operator_token46] = ACTIONS(806),
    [aux_sym_comparison_operator_token47] = ACTIONS(806),
    [aux_sym_comparison_operator_token48] = ACTIONS(806),
    [aux_sym_comparison_operator_token49] = ACTIONS(806),
    [aux_sym_comparison_operator_token50] = ACTIONS(806),
    [aux_sym_format_operator_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token2] = ACTIONS(806),
    [aux_sym_logical_expression_token3] = ACTIONS(806),
    [aux_sym_bitwise_expression_token1] = ACTIONS(806),
    [aux_sym_bitwise_expression_token2] = ACTIONS(806),
    [aux_sym_bitwise_expression_token3] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_DASH] = ACTIONS(808),
    [anon_sym_SLASH] = ACTIONS(806),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_PERCENT] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(806),
    [anon_sym_DOT_DOT] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(806),
  },
  [STATE(479)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1925),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1661),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(480)] = {
    [sym_format_operator] = STATE(583),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(796),
    [aux_sym_comparison_operator_token2] = ACTIONS(796),
    [aux_sym_comparison_operator_token3] = ACTIONS(796),
    [aux_sym_comparison_operator_token4] = ACTIONS(796),
    [aux_sym_comparison_operator_token5] = ACTIONS(796),
    [aux_sym_comparison_operator_token6] = ACTIONS(796),
    [aux_sym_comparison_operator_token7] = ACTIONS(796),
    [aux_sym_comparison_operator_token8] = ACTIONS(796),
    [aux_sym_comparison_operator_token9] = ACTIONS(796),
    [aux_sym_comparison_operator_token10] = ACTIONS(796),
    [aux_sym_comparison_operator_token11] = ACTIONS(796),
    [aux_sym_comparison_operator_token12] = ACTIONS(796),
    [aux_sym_comparison_operator_token13] = ACTIONS(796),
    [aux_sym_comparison_operator_token14] = ACTIONS(796),
    [aux_sym_comparison_operator_token15] = ACTIONS(796),
    [aux_sym_comparison_operator_token16] = ACTIONS(796),
    [aux_sym_comparison_operator_token17] = ACTIONS(796),
    [aux_sym_comparison_operator_token18] = ACTIONS(796),
    [aux_sym_comparison_operator_token19] = ACTIONS(796),
    [aux_sym_comparison_operator_token20] = ACTIONS(796),
    [aux_sym_comparison_operator_token21] = ACTIONS(796),
    [aux_sym_comparison_operator_token22] = ACTIONS(796),
    [aux_sym_comparison_operator_token23] = ACTIONS(796),
    [aux_sym_comparison_operator_token24] = ACTIONS(796),
    [aux_sym_comparison_operator_token25] = ACTIONS(796),
    [aux_sym_comparison_operator_token26] = ACTIONS(796),
    [aux_sym_comparison_operator_token27] = ACTIONS(796),
    [aux_sym_comparison_operator_token28] = ACTIONS(798),
    [aux_sym_comparison_operator_token29] = ACTIONS(796),
    [aux_sym_comparison_operator_token30] = ACTIONS(796),
    [aux_sym_comparison_operator_token31] = ACTIONS(796),
    [aux_sym_comparison_operator_token32] = ACTIONS(796),
    [aux_sym_comparison_operator_token33] = ACTIONS(796),
    [aux_sym_comparison_operator_token34] = ACTIONS(798),
    [aux_sym_comparison_operator_token35] = ACTIONS(796),
    [aux_sym_comparison_operator_token36] = ACTIONS(796),
    [aux_sym_comparison_operator_token37] = ACTIONS(796),
    [aux_sym_comparison_operator_token38] = ACTIONS(796),
    [aux_sym_comparison_operator_token39] = ACTIONS(796),
    [aux_sym_comparison_operator_token40] = ACTIONS(796),
    [aux_sym_comparison_operator_token41] = ACTIONS(796),
    [aux_sym_comparison_operator_token42] = ACTIONS(796),
    [aux_sym_comparison_operator_token43] = ACTIONS(796),
    [aux_sym_comparison_operator_token44] = ACTIONS(796),
    [aux_sym_comparison_operator_token45] = ACTIONS(796),
    [aux_sym_comparison_operator_token46] = ACTIONS(796),
    [aux_sym_comparison_operator_token47] = ACTIONS(796),
    [aux_sym_comparison_operator_token48] = ACTIONS(796),
    [aux_sym_comparison_operator_token49] = ACTIONS(796),
    [aux_sym_comparison_operator_token50] = ACTIONS(796),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [aux_sym_logical_expression_token1] = ACTIONS(796),
    [aux_sym_logical_expression_token2] = ACTIONS(796),
    [aux_sym_logical_expression_token3] = ACTIONS(796),
    [aux_sym_bitwise_expression_token1] = ACTIONS(796),
    [aux_sym_bitwise_expression_token2] = ACTIONS(796),
    [aux_sym_bitwise_expression_token3] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(796),
    [anon_sym_DASH] = ACTIONS(798),
    [anon_sym_SLASH] = ACTIONS(796),
    [anon_sym_BSLASH] = ACTIONS(796),
    [anon_sym_PERCENT] = ACTIONS(796),
    [anon_sym_STAR] = ACTIONS(796),
    [sym__statement_terminator] = ACTIONS(796),
  },
  [STATE(481)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym_unary_expression] = STATE(517),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_argument_expression_list] = STATE(1712),
    [sym_argument_expression] = STATE(1471),
    [sym_logical_argument_expression] = STATE(1337),
    [sym_bitwise_argument_expression] = STATE(1339),
    [sym_comparison_argument_expression] = STATE(596),
    [sym_additive_argument_expression] = STATE(581),
    [sym_multiplicative_argument_expression] = STATE(541),
    [sym_format_argument_expression] = STATE(518),
    [sym_range_argument_expression] = STATE(508),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_RPAREN] = ACTIONS(1663),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(482)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(806),
    [aux_sym_comparison_operator_token2] = ACTIONS(806),
    [aux_sym_comparison_operator_token3] = ACTIONS(806),
    [aux_sym_comparison_operator_token4] = ACTIONS(806),
    [aux_sym_comparison_operator_token5] = ACTIONS(806),
    [aux_sym_comparison_operator_token6] = ACTIONS(806),
    [aux_sym_comparison_operator_token7] = ACTIONS(806),
    [aux_sym_comparison_operator_token8] = ACTIONS(806),
    [aux_sym_comparison_operator_token9] = ACTIONS(806),
    [aux_sym_comparison_operator_token10] = ACTIONS(806),
    [aux_sym_comparison_operator_token11] = ACTIONS(806),
    [aux_sym_comparison_operator_token12] = ACTIONS(806),
    [aux_sym_comparison_operator_token13] = ACTIONS(806),
    [aux_sym_comparison_operator_token14] = ACTIONS(806),
    [aux_sym_comparison_operator_token15] = ACTIONS(806),
    [aux_sym_comparison_operator_token16] = ACTIONS(806),
    [aux_sym_comparison_operator_token17] = ACTIONS(806),
    [aux_sym_comparison_operator_token18] = ACTIONS(806),
    [aux_sym_comparison_operator_token19] = ACTIONS(806),
    [aux_sym_comparison_operator_token20] = ACTIONS(806),
    [aux_sym_comparison_operator_token21] = ACTIONS(806),
    [aux_sym_comparison_operator_token22] = ACTIONS(806),
    [aux_sym_comparison_operator_token23] = ACTIONS(806),
    [aux_sym_comparison_operator_token24] = ACTIONS(806),
    [aux_sym_comparison_operator_token25] = ACTIONS(806),
    [aux_sym_comparison_operator_token26] = ACTIONS(806),
    [aux_sym_comparison_operator_token27] = ACTIONS(806),
    [aux_sym_comparison_operator_token28] = ACTIONS(808),
    [aux_sym_comparison_operator_token29] = ACTIONS(806),
    [aux_sym_comparison_operator_token30] = ACTIONS(806),
    [aux_sym_comparison_operator_token31] = ACTIONS(806),
    [aux_sym_comparison_operator_token32] = ACTIONS(806),
    [aux_sym_comparison_operator_token33] = ACTIONS(806),
    [aux_sym_comparison_operator_token34] = ACTIONS(808),
    [aux_sym_comparison_operator_token35] = ACTIONS(806),
    [aux_sym_comparison_operator_token36] = ACTIONS(806),
    [aux_sym_comparison_operator_token37] = ACTIONS(806),
    [aux_sym_comparison_operator_token38] = ACTIONS(806),
    [aux_sym_comparison_operator_token39] = ACTIONS(806),
    [aux_sym_comparison_operator_token40] = ACTIONS(806),
    [aux_sym_comparison_operator_token41] = ACTIONS(806),
    [aux_sym_comparison_operator_token42] = ACTIONS(806),
    [aux_sym_comparison_operator_token43] = ACTIONS(806),
    [aux_sym_comparison_operator_token44] = ACTIONS(806),
    [aux_sym_comparison_operator_token45] = ACTIONS(806),
    [aux_sym_comparison_operator_token46] = ACTIONS(806),
    [aux_sym_comparison_operator_token47] = ACTIONS(806),
    [aux_sym_comparison_operator_token48] = ACTIONS(806),
    [aux_sym_comparison_operator_token49] = ACTIONS(806),
    [aux_sym_comparison_operator_token50] = ACTIONS(806),
    [aux_sym_format_operator_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token1] = ACTIONS(806),
    [aux_sym_logical_expression_token2] = ACTIONS(806),
    [aux_sym_logical_expression_token3] = ACTIONS(806),
    [aux_sym_bitwise_expression_token1] = ACTIONS(806),
    [aux_sym_bitwise_expression_token2] = ACTIONS(806),
    [aux_sym_bitwise_expression_token3] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_DASH] = ACTIONS(808),
    [anon_sym_SLASH] = ACTIONS(806),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_PERCENT] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(806),
    [anon_sym_DOT_DOT] = ACTIONS(1639),
    [sym__statement_terminator] = ACTIONS(806),
  },
  [STATE(483)] = {
    [sym_format_operator] = STATE(587),
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(790),
    [aux_sym_comparison_operator_token2] = ACTIONS(790),
    [aux_sym_comparison_operator_token3] = ACTIONS(790),
    [aux_sym_comparison_operator_token4] = ACTIONS(790),
    [aux_sym_comparison_operator_token5] = ACTIONS(790),
    [aux_sym_comparison_operator_token6] = ACTIONS(790),
    [aux_sym_comparison_operator_token7] = ACTIONS(790),
    [aux_sym_comparison_operator_token8] = ACTIONS(790),
    [aux_sym_comparison_operator_token9] = ACTIONS(790),
    [aux_sym_comparison_operator_token10] = ACTIONS(790),
    [aux_sym_comparison_operator_token11] = ACTIONS(790),
    [aux_sym_comparison_operator_token12] = ACTIONS(790),
    [aux_sym_comparison_operator_token13] = ACTIONS(790),
    [aux_sym_comparison_operator_token14] = ACTIONS(790),
    [aux_sym_comparison_operator_token15] = ACTIONS(790),
    [aux_sym_comparison_operator_token16] = ACTIONS(790),
    [aux_sym_comparison_operator_token17] = ACTIONS(790),
    [aux_sym_comparison_operator_token18] = ACTIONS(790),
    [aux_sym_comparison_operator_token19] = ACTIONS(790),
    [aux_sym_comparison_operator_token20] = ACTIONS(790),
    [aux_sym_comparison_operator_token21] = ACTIONS(790),
    [aux_sym_comparison_operator_token22] = ACTIONS(790),
    [aux_sym_comparison_operator_token23] = ACTIONS(790),
    [aux_sym_comparison_operator_token24] = ACTIONS(790),
    [aux_sym_comparison_operator_token25] = ACTIONS(790),
    [aux_sym_comparison_operator_token26] = ACTIONS(790),
    [aux_sym_comparison_operator_token27] = ACTIONS(790),
    [aux_sym_comparison_operator_token28] = ACTIONS(792),
    [aux_sym_comparison_operator_token29] = ACTIONS(790),
    [aux_sym_comparison_operator_token30] = ACTIONS(790),
    [aux_sym_comparison_operator_token31] = ACTIONS(790),
    [aux_sym_comparison_operator_token32] = ACTIONS(790),
    [aux_sym_comparison_operator_token33] = ACTIONS(790),
    [aux_sym_comparison_operator_token34] = ACTIONS(792),
    [aux_sym_comparison_operator_token35] = ACTIONS(790),
    [aux_sym_comparison_operator_token36] = ACTIONS(790),
    [aux_sym_comparison_operator_token37] = ACTIONS(790),
    [aux_sym_comparison_operator_token38] = ACTIONS(790),
    [aux_sym_comparison_operator_token39] = ACTIONS(790),
    [aux_sym_comparison_operator_token40] = ACTIONS(790),
    [aux_sym_comparison_operator_token41] = ACTIONS(790),
    [aux_sym_comparison_operator_token42] = ACTIONS(790),
    [aux_sym_comparison_operator_token43] = ACTIONS(790),
    [aux_sym_comparison_operator_token44] = ACTIONS(790),
    [aux_sym_comparison_operator_token45] = ACTIONS(790),
    [aux_sym_comparison_operator_token46] = ACTIONS(790),
    [aux_sym_comparison_operator_token47] = ACTIONS(790),
    [aux_sym_comparison_operator_token48] = ACTIONS(790),
    [aux_sym_comparison_operator_token49] = ACTIONS(790),
    [aux_sym_comparison_operator_token50] = ACTIONS(790),
    [aux_sym_format_operator_token1] = ACTIONS(794),
    [aux_sym_logical_expression_token1] = ACTIONS(790),
    [aux_sym_logical_expression_token2] = ACTIONS(790),
    [aux_sym_logical_expression_token3] = ACTIONS(790),
    [aux_sym_bitwise_expression_token1] = ACTIONS(790),
    [aux_sym_bitwise_expression_token2] = ACTIONS(790),
    [aux_sym_bitwise_expression_token3] = ACTIONS(790),
    [anon_sym_PLUS] = ACTIONS(790),
    [anon_sym_DASH] = ACTIONS(792),
    [anon_sym_SLASH] = ACTIONS(790),
    [anon_sym_BSLASH] = ACTIONS(790),
    [anon_sym_PERCENT] = ACTIONS(790),
    [anon_sym_STAR] = ACTIONS(790),
    [anon_sym_RBRACK] = ACTIONS(790),
  },
  [STATE(484)] = {
    [sym__literal] = STATE(435),
    [sym_integer_literal] = STATE(435),
    [sym_string_literal] = STATE(435),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(435),
    [sym__expression] = STATE(1924),
    [sym_logical_expression] = STATE(1399),
    [sym_bitwise_expression] = STATE(1311),
    [sym_comparison_expression] = STATE(571),
    [sym_additive_expression] = STATE(566),
    [sym_multiplicative_expression] = STATE(521),
    [sym_format_expression] = STATE(477),
    [sym_range_expression] = STATE(469),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(450),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(435),
    [sym__value] = STATE(435),
    [sym_parenthesized_expression] = STATE(435),
    [sym_sub_expression] = STATE(435),
    [sym_array_expression] = STATE(435),
    [sym_script_block_expression] = STATE(435),
    [sym_hash_literal_expression] = STATE(435),
    [sym_post_increment_expression] = STATE(435),
    [sym_post_decrement_expression] = STATE(435),
    [sym_member_access] = STATE(435),
    [sym_element_access] = STATE(435),
    [sym_invokation_expression] = STATE(435),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(85),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(564),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token37] = ACTIONS(568),
    [aux_sym_comparison_operator_token50] = ACTIONS(568),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(568),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_PLUS] = ACTIONS(570),
    [anon_sym_DASH] = ACTIONS(570),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(568),
    [anon_sym_BANG] = ACTIONS(568),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(568),
    [anon_sym_PLUS_PLUS] = ACTIONS(572),
    [anon_sym_DASH_DASH] = ACTIONS(574),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
  },
  [STATE(485)] = {
    [sym__literal] = STATE(435),
    [sym_integer_literal] = STATE(435),
    [sym_string_literal] = STATE(435),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(435),
    [sym__expression] = STATE(1797),
    [sym_logical_expression] = STATE(1399),
    [sym_bitwise_expression] = STATE(1311),
    [sym_comparison_expression] = STATE(571),
    [sym_additive_expression] = STATE(566),
    [sym_multiplicative_expression] = STATE(521),
    [sym_format_expression] = STATE(477),
    [sym_range_expression] = STATE(469),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(450),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(435),
    [sym__value] = STATE(435),
    [sym_parenthesized_expression] = STATE(435),
    [sym_sub_expression] = STATE(435),
    [sym_array_expression] = STATE(435),
    [sym_script_block_expression] = STATE(435),
    [sym_hash_literal_expression] = STATE(435),
    [sym_post_increment_expression] = STATE(435),
    [sym_post_decrement_expression] = STATE(435),
    [sym_member_access] = STATE(435),
    [sym_element_access] = STATE(435),
    [sym_invokation_expression] = STATE(435),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(85),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(564),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token37] = ACTIONS(568),
    [aux_sym_comparison_operator_token50] = ACTIONS(568),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(568),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_PLUS] = ACTIONS(570),
    [anon_sym_DASH] = ACTIONS(570),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(568),
    [anon_sym_BANG] = ACTIONS(568),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(568),
    [anon_sym_PLUS_PLUS] = ACTIONS(572),
    [anon_sym_DASH_DASH] = ACTIONS(574),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
  },
  [STATE(486)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1669),
    [sym_logical_expression] = STATE(1379),
    [sym_bitwise_expression] = STATE(1293),
    [sym_comparison_expression] = STATE(560),
    [sym_additive_expression] = STATE(546),
    [sym_multiplicative_expression] = STATE(496),
    [sym_format_expression] = STATE(460),
    [sym_range_expression] = STATE(462),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(454),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(487)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1898),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(488)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(540),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(540),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(540),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(540),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(540),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(540),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(540),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(540),
    [anon_sym_2_GT_AMP1] = ACTIONS(540),
    [anon_sym_3_GT_AMP1] = ACTIONS(540),
    [anon_sym_4_GT_AMP1] = ACTIONS(540),
    [anon_sym_5_GT_AMP1] = ACTIONS(540),
    [anon_sym_6_GT_AMP1] = ACTIONS(540),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(540),
    [anon_sym_1_GT_AMP2] = ACTIONS(540),
    [anon_sym_3_GT_AMP2] = ACTIONS(540),
    [anon_sym_4_GT_AMP2] = ACTIONS(540),
    [anon_sym_5_GT_AMP2] = ACTIONS(540),
    [anon_sym_6_GT_AMP2] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [anon_sym_AMP_AMP] = ACTIONS(540),
    [anon_sym_PIPE_PIPE] = ACTIONS(540),
    [anon_sym_PIPE] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(540),
    [anon_sym_BANG] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
    [sym__statement_terminator] = ACTIONS(540),
  },
  [STATE(489)] = {
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(538),
    [sym_hexadecimal_integer_literal] = ACTIONS(540),
    [sym_real_literal] = ACTIONS(540),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(540),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(540),
    [sym_verbatim_string_characters] = ACTIONS(540),
    [sym_verbatim_here_string_characters] = ACTIONS(540),
    [anon_sym_LBRACK] = ACTIONS(540),
    [anon_sym_GT] = ACTIONS(538),
    [anon_sym_GT_GT] = ACTIONS(540),
    [anon_sym_2_GT] = ACTIONS(538),
    [anon_sym_2_GT_GT] = ACTIONS(540),
    [anon_sym_3_GT] = ACTIONS(538),
    [anon_sym_3_GT_GT] = ACTIONS(540),
    [anon_sym_4_GT] = ACTIONS(538),
    [anon_sym_4_GT_GT] = ACTIONS(540),
    [anon_sym_5_GT] = ACTIONS(538),
    [anon_sym_5_GT_GT] = ACTIONS(540),
    [anon_sym_6_GT] = ACTIONS(538),
    [anon_sym_6_GT_GT] = ACTIONS(540),
    [anon_sym_STAR_GT] = ACTIONS(538),
    [anon_sym_STAR_GT_GT] = ACTIONS(540),
    [anon_sym_LT] = ACTIONS(538),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(540),
    [anon_sym_2_GT_AMP1] = ACTIONS(540),
    [anon_sym_3_GT_AMP1] = ACTIONS(540),
    [anon_sym_4_GT_AMP1] = ACTIONS(540),
    [anon_sym_5_GT_AMP1] = ACTIONS(540),
    [anon_sym_6_GT_AMP1] = ACTIONS(540),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(540),
    [anon_sym_1_GT_AMP2] = ACTIONS(540),
    [anon_sym_3_GT_AMP2] = ACTIONS(540),
    [anon_sym_4_GT_AMP2] = ACTIONS(540),
    [anon_sym_5_GT_AMP2] = ACTIONS(540),
    [anon_sym_6_GT_AMP2] = ACTIONS(540),
    [aux_sym_comparison_operator_token37] = ACTIONS(540),
    [aux_sym_comparison_operator_token50] = ACTIONS(540),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(540),
    [anon_sym_DOLLAR_CARET] = ACTIONS(540),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(540),
    [anon_sym_DOLLAR_] = ACTIONS(538),
    [aux_sym_variable_token1] = ACTIONS(538),
    [aux_sym_variable_token2] = ACTIONS(540),
    [sym_braced_variable] = ACTIONS(540),
    [anon_sym_LPAREN] = ACTIONS(540),
    [anon_sym_RPAREN] = ACTIONS(540),
    [anon_sym_COMMA] = ACTIONS(540),
    [anon_sym_LBRACE] = ACTIONS(540),
    [anon_sym_AMP_AMP] = ACTIONS(540),
    [anon_sym_PIPE_PIPE] = ACTIONS(540),
    [anon_sym_PIPE] = ACTIONS(538),
    [anon_sym_PLUS] = ACTIONS(538),
    [anon_sym_DASH] = ACTIONS(538),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(540),
    [anon_sym_BANG] = ACTIONS(540),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(540),
    [anon_sym_DASH_DASH] = ACTIONS(540),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LPAREN] = ACTIONS(540),
    [anon_sym_AT_LBRACE] = ACTIONS(540),
    [anon_sym_DOT2] = ACTIONS(538),
    [anon_sym_COLON_COLON] = ACTIONS(540),
    [aux_sym_invokation_foreach_expression_token1] = ACTIONS(540),
  },
  [STATE(490)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1870),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(491)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(494),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1665),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
    [sym__statement_terminator] = ACTIONS(1593),
  },
  [STATE(492)] = {
    [sym__literal] = STATE(435),
    [sym_integer_literal] = STATE(435),
    [sym_string_literal] = STATE(435),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(435),
    [sym__expression] = STATE(2012),
    [sym_logical_expression] = STATE(1399),
    [sym_bitwise_expression] = STATE(1311),
    [sym_comparison_expression] = STATE(571),
    [sym_additive_expression] = STATE(566),
    [sym_multiplicative_expression] = STATE(521),
    [sym_format_expression] = STATE(477),
    [sym_range_expression] = STATE(469),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(450),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(435),
    [sym__value] = STATE(435),
    [sym_parenthesized_expression] = STATE(435),
    [sym_sub_expression] = STATE(435),
    [sym_array_expression] = STATE(435),
    [sym_script_block_expression] = STATE(435),
    [sym_hash_literal_expression] = STATE(435),
    [sym_post_increment_expression] = STATE(435),
    [sym_post_decrement_expression] = STATE(435),
    [sym_member_access] = STATE(435),
    [sym_element_access] = STATE(435),
    [sym_invokation_expression] = STATE(435),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(85),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(564),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token37] = ACTIONS(568),
    [aux_sym_comparison_operator_token50] = ACTIONS(568),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(568),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_PLUS] = ACTIONS(570),
    [anon_sym_DASH] = ACTIONS(570),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(568),
    [anon_sym_BANG] = ACTIONS(568),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(568),
    [anon_sym_PLUS_PLUS] = ACTIONS(572),
    [anon_sym_DASH_DASH] = ACTIONS(574),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
  },
  [STATE(493)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1617),
    [sym_logical_expression] = STATE(1379),
    [sym_bitwise_expression] = STATE(1293),
    [sym_comparison_expression] = STATE(560),
    [sym_additive_expression] = STATE(546),
    [sym_multiplicative_expression] = STATE(496),
    [sym_format_expression] = STATE(460),
    [sym_range_expression] = STATE(462),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(454),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(494)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(494),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1667),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
    [sym__statement_terminator] = ACTIONS(1606),
  },
  [STATE(495)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1988),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(496)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(812),
    [aux_sym_comparison_operator_token2] = ACTIONS(812),
    [aux_sym_comparison_operator_token3] = ACTIONS(812),
    [aux_sym_comparison_operator_token4] = ACTIONS(812),
    [aux_sym_comparison_operator_token5] = ACTIONS(812),
    [aux_sym_comparison_operator_token6] = ACTIONS(812),
    [aux_sym_comparison_operator_token7] = ACTIONS(812),
    [aux_sym_comparison_operator_token8] = ACTIONS(812),
    [aux_sym_comparison_operator_token9] = ACTIONS(812),
    [aux_sym_comparison_operator_token10] = ACTIONS(812),
    [aux_sym_comparison_operator_token11] = ACTIONS(812),
    [aux_sym_comparison_operator_token12] = ACTIONS(812),
    [aux_sym_comparison_operator_token13] = ACTIONS(812),
    [aux_sym_comparison_operator_token14] = ACTIONS(812),
    [aux_sym_comparison_operator_token15] = ACTIONS(812),
    [aux_sym_comparison_operator_token16] = ACTIONS(812),
    [aux_sym_comparison_operator_token17] = ACTIONS(812),
    [aux_sym_comparison_operator_token18] = ACTIONS(812),
    [aux_sym_comparison_operator_token19] = ACTIONS(812),
    [aux_sym_comparison_operator_token20] = ACTIONS(812),
    [aux_sym_comparison_operator_token21] = ACTIONS(812),
    [aux_sym_comparison_operator_token22] = ACTIONS(812),
    [aux_sym_comparison_operator_token23] = ACTIONS(812),
    [aux_sym_comparison_operator_token24] = ACTIONS(812),
    [aux_sym_comparison_operator_token25] = ACTIONS(812),
    [aux_sym_comparison_operator_token26] = ACTIONS(812),
    [aux_sym_comparison_operator_token27] = ACTIONS(812),
    [aux_sym_comparison_operator_token28] = ACTIONS(814),
    [aux_sym_comparison_operator_token29] = ACTIONS(812),
    [aux_sym_comparison_operator_token30] = ACTIONS(812),
    [aux_sym_comparison_operator_token31] = ACTIONS(812),
    [aux_sym_comparison_operator_token32] = ACTIONS(812),
    [aux_sym_comparison_operator_token33] = ACTIONS(812),
    [aux_sym_comparison_operator_token34] = ACTIONS(814),
    [aux_sym_comparison_operator_token35] = ACTIONS(812),
    [aux_sym_comparison_operator_token36] = ACTIONS(812),
    [aux_sym_comparison_operator_token37] = ACTIONS(812),
    [aux_sym_comparison_operator_token38] = ACTIONS(812),
    [aux_sym_comparison_operator_token39] = ACTIONS(812),
    [aux_sym_comparison_operator_token40] = ACTIONS(812),
    [aux_sym_comparison_operator_token41] = ACTIONS(812),
    [aux_sym_comparison_operator_token42] = ACTIONS(812),
    [aux_sym_comparison_operator_token43] = ACTIONS(812),
    [aux_sym_comparison_operator_token44] = ACTIONS(812),
    [aux_sym_comparison_operator_token45] = ACTIONS(812),
    [aux_sym_comparison_operator_token46] = ACTIONS(812),
    [aux_sym_comparison_operator_token47] = ACTIONS(812),
    [aux_sym_comparison_operator_token48] = ACTIONS(812),
    [aux_sym_comparison_operator_token49] = ACTIONS(812),
    [aux_sym_comparison_operator_token50] = ACTIONS(812),
    [anon_sym_RPAREN] = ACTIONS(812),
    [anon_sym_COMMA] = ACTIONS(812),
    [aux_sym_logical_expression_token1] = ACTIONS(812),
    [aux_sym_logical_expression_token2] = ACTIONS(812),
    [aux_sym_logical_expression_token3] = ACTIONS(812),
    [aux_sym_bitwise_expression_token1] = ACTIONS(812),
    [aux_sym_bitwise_expression_token2] = ACTIONS(812),
    [aux_sym_bitwise_expression_token3] = ACTIONS(812),
    [anon_sym_PLUS] = ACTIONS(812),
    [anon_sym_DASH] = ACTIONS(814),
    [anon_sym_SLASH] = ACTIONS(1670),
    [anon_sym_BSLASH] = ACTIONS(1670),
    [anon_sym_PERCENT] = ACTIONS(1670),
    [anon_sym_STAR] = ACTIONS(1670),
  },
  [STATE(497)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(2114),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(498)] = {
    [sym__literal] = STATE(435),
    [sym_integer_literal] = STATE(435),
    [sym_string_literal] = STATE(435),
    [sym_expandable_string_literal] = STATE(157),
    [sym_expandable_here_string_literal] = STATE(157),
    [sym_variable] = STATE(435),
    [sym__expression] = STATE(1929),
    [sym_logical_expression] = STATE(1399),
    [sym_bitwise_expression] = STATE(1311),
    [sym_comparison_expression] = STATE(571),
    [sym_additive_expression] = STATE(566),
    [sym_multiplicative_expression] = STATE(521),
    [sym_format_expression] = STATE(477),
    [sym_range_expression] = STATE(469),
    [sym_array_literal_expression] = STATE(191),
    [sym_unary_expression] = STATE(450),
    [sym_expression_with_unary_operator] = STATE(176),
    [sym_pre_increment_expression] = STATE(177),
    [sym_pre_decrement_expression] = STATE(177),
    [sym_cast_expression] = STATE(177),
    [sym__primary_expression] = STATE(435),
    [sym__value] = STATE(435),
    [sym_parenthesized_expression] = STATE(435),
    [sym_sub_expression] = STATE(435),
    [sym_array_expression] = STATE(435),
    [sym_script_block_expression] = STATE(435),
    [sym_hash_literal_expression] = STATE(435),
    [sym_post_increment_expression] = STATE(435),
    [sym_post_decrement_expression] = STATE(435),
    [sym_member_access] = STATE(435),
    [sym_element_access] = STATE(435),
    [sym_invokation_expression] = STATE(435),
    [sym_invokation_foreach_expression] = STATE(138),
    [sym_type_literal] = STATE(85),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(5),
    [sym_hexadecimal_integer_literal] = ACTIONS(81),
    [sym_real_literal] = ACTIONS(564),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(85),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(87),
    [sym_verbatim_string_characters] = ACTIONS(89),
    [sym_verbatim_here_string_characters] = ACTIONS(89),
    [anon_sym_LBRACK] = ACTIONS(566),
    [aux_sym_comparison_operator_token37] = ACTIONS(568),
    [aux_sym_comparison_operator_token50] = ACTIONS(568),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(99),
    [anon_sym_DOLLAR_CARET] = ACTIONS(99),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(99),
    [anon_sym_DOLLAR_] = ACTIONS(21),
    [aux_sym_variable_token1] = ACTIONS(21),
    [aux_sym_variable_token2] = ACTIONS(99),
    [sym_braced_variable] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(103),
    [anon_sym_COMMA] = ACTIONS(568),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_PLUS] = ACTIONS(570),
    [anon_sym_DASH] = ACTIONS(570),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(568),
    [anon_sym_BANG] = ACTIONS(568),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(568),
    [anon_sym_PLUS_PLUS] = ACTIONS(572),
    [anon_sym_DASH_DASH] = ACTIONS(574),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(111),
    [anon_sym_AT_LPAREN] = ACTIONS(113),
    [anon_sym_AT_LBRACE] = ACTIONS(115),
  },
  [STATE(499)] = {
    [sym_comment] = ACTIONS(79),
    [aux_sym_comparison_operator_token1] = ACTIONS(820),
    [aux_sym_comparison_operator_token2] = ACTIONS(820),
    [aux_sym_comparison_operator_token3] = ACTIONS(820),
    [aux_sym_comparison_operator_token4] = ACTIONS(820),
    [aux_sym_comparison_operator_token5] = ACTIONS(820),
    [aux_sym_comparison_operator_token6] = ACTIONS(820),
    [aux_sym_comparison_operator_token7] = ACTIONS(820),
    [aux_sym_comparison_operator_token8] = ACTIONS(820),
    [aux_sym_comparison_operator_token9] = ACTIONS(820),
    [aux_sym_comparison_operator_token10] = ACTIONS(820),
    [aux_sym_comparison_operator_token11] = ACTIONS(820),
    [aux_sym_comparison_operator_token12] = ACTIONS(820),
    [aux_sym_comparison_operator_token13] = ACTIONS(820),
    [aux_sym_comparison_operator_token14] = ACTIONS(820),
    [aux_sym_comparison_operator_token15] = ACTIONS(820),
    [aux_sym_comparison_operator_token16] = ACTIONS(820),
    [aux_sym_comparison_operator_token17] = ACTIONS(820),
    [aux_sym_comparison_operator_token18] = ACTIONS(820),
    [aux_sym_comparison_operator_token19] = ACTIONS(820),
    [aux_sym_comparison_operator_token20] = ACTIONS(820),
    [aux_sym_comparison_operator_token21] = ACTIONS(820),
    [aux_sym_comparison_operator_token22] = ACTIONS(820),
    [aux_sym_comparison_operator_token23] = ACTIONS(820),
    [aux_sym_comparison_operator_token24] = ACTIONS(820),
    [aux_sym_comparison_operator_token25] = ACTIONS(820),
    [aux_sym_comparison_operator_token26] = ACTIONS(820),
    [aux_sym_comparison_operator_token27] = ACTIONS(820),
    [aux_sym_comparison_operator_token28] = ACTIONS(822),
    [aux_sym_comparison_operator_token29] = ACTIONS(820),
    [aux_sym_comparison_operator_token30] = ACTIONS(820),
    [aux_sym_comparison_operator_token31] = ACTIONS(820),
    [aux_sym_comparison_operator_token32] = ACTIONS(820),
    [aux_sym_comparison_operator_token33] = ACTIONS(820),
    [aux_sym_comparison_operator_token34] = ACTIONS(822),
    [aux_sym_comparison_operator_token35] = ACTIONS(820),
    [aux_sym_comparison_operator_token36] = ACTIONS(820),
    [aux_sym_comparison_operator_token37] = ACTIONS(820),
    [aux_sym_comparison_operator_token38] = ACTIONS(820),
    [aux_sym_comparison_operator_token39] = ACTIONS(820),
    [aux_sym_comparison_operator_token40] = ACTIONS(820),
    [aux_sym_comparison_operator_token41] = ACTIONS(820),
    [aux_sym_comparison_operator_token42] = ACTIONS(820),
    [aux_sym_comparison_operator_token43] = ACTIONS(820),
    [aux_sym_comparison_operator_token44] = ACTIONS(820),
    [aux_sym_comparison_operator_token45] = ACTIONS(820),
    [aux_sym_comparison_operator_token46] = ACTIONS(820),
    [aux_sym_comparison_operator_token47] = ACTIONS(820),
    [aux_sym_comparison_operator_token48] = ACTIONS(820),
    [aux_sym_comparison_operator_token49] = ACTIONS(820),
    [aux_sym_comparison_operator_token50] = ACTIONS(820),
    [anon_sym_RPAREN] = ACTIONS(820),
    [anon_sym_COMMA] = ACTIONS(820),
    [aux_sym_logical_expression_token1] = ACTIONS(820),
    [aux_sym_logical_expression_token2] = ACTIONS(820),
    [aux_sym_logical_expression_token3] = ACTIONS(820),
    [aux_sym_bitwise_expression_token1] = ACTIONS(820),
    [aux_sym_bitwise_expression_token2] = ACTIONS(820),
    [aux_sym_bitwise_expression_token3] = ACTIONS(820),
    [anon_sym_PLUS] = ACTIONS(820),
    [anon_sym_DASH] = ACTIONS(822),
    [anon_sym_SLASH] = ACTIONS(1670),
    [anon_sym_BSLASH] = ACTIONS(1670),
    [anon_sym_PERCENT] = ACTIONS(1670),
    [anon_sym_STAR] = ACTIONS(1670),
  },
  [STATE(500)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1766),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(501)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1790),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(502)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1812),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(503)] = {
    [sym__literal] = STATE(433),
    [sym_integer_literal] = STATE(433),
    [sym_string_literal] = STATE(433),
    [sym_expandable_string_literal] = STATE(119),
    [sym_expandable_here_string_literal] = STATE(119),
    [sym_variable] = STATE(433),
    [sym__expression] = STATE(1833),
    [sym_logical_expression] = STATE(1405),
    [sym_bitwise_expression] = STATE(1320),
    [sym_comparison_expression] = STATE(577),
    [sym_additive_expression] = STATE(558),
    [sym_multiplicative_expression] = STATE(512),
    [sym_format_expression] = STATE(483),
    [sym_range_expression] = STATE(467),
    [sym_array_literal_expression] = STATE(163),
    [sym_unary_expression] = STATE(452),
    [sym_expression_with_unary_operator] = STATE(165),
    [sym_pre_increment_expression] = STATE(166),
    [sym_pre_decrement_expression] = STATE(166),
    [sym_cast_expression] = STATE(166),
    [sym__primary_expression] = STATE(433),
    [sym__value] = STATE(433),
    [sym_parenthesized_expression] = STATE(433),
    [sym_sub_expression] = STATE(433),
    [sym_array_expression] = STATE(433),
    [sym_script_block_expression] = STATE(433),
    [sym_hash_literal_expression] = STATE(433),
    [sym_post_increment_expression] = STATE(433),
    [sym_post_decrement_expression] = STATE(433),
    [sym_member_access] = STATE(433),
    [sym_element_access] = STATE(433),
    [sym_invokation_expression] = STATE(433),
    [sym_invokation_foreach_expression] = STATE(120),
    [sym_type_literal] = STATE(83),
    [sym_comment] = ACTIONS(79),
    [sym_decimal_integer_literal] = ACTIONS(117),
    [sym_hexadecimal_integer_literal] = ACTIONS(119),
    [sym_real_literal] = ACTIONS(542),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(123),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(125),
    [sym_verbatim_string_characters] = ACTIONS(127),
    [sym_verbatim_here_string_characters] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(544),
    [aux_sym_comparison_operator_token37] = ACTIONS(546),
    [aux_sym_comparison_operator_token50] = ACTIONS(546),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(133),
    [anon_sym_DOLLAR_CARET] = ACTIONS(133),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(133),
    [anon_sym_DOLLAR_] = ACTIONS(135),
    [aux_sym_variable_token1] = ACTIONS(135),
    [aux_sym_variable_token2] = ACTIONS(133),
    [sym_braced_variable] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(137),
    [anon_sym_COMMA] = ACTIONS(546),
    [anon_sym_LBRACE] = ACTIONS(139),
    [anon_sym_PLUS] = ACTIONS(548),
    [anon_sym_DASH] = ACTIONS(548),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(546),
    [anon_sym_BANG] = ACTIONS(546),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(546),
    [anon_sym_PLUS_PLUS] = ACTIONS(550),
    [anon_sym_DASH_DASH] = ACTIONS(552),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(147),
    [anon_sym_AT_LPAREN] = ACTIONS(149),
    [anon_sym_AT_LBRACE] = ACTIONS(151),
  },
  [STATE(504)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(505),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1589),
    [sym_hexadecimal_integer_literal] = ACTIONS(1589),
    [sym_real_literal] = ACTIONS(1589),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1589),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1589),
    [sym_verbatim_string_characters] = ACTIONS(1589),
    [sym_verbatim_here_string_characters] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1589),
    [anon_sym_GT] = ACTIONS(1589),
    [anon_sym_GT_GT] = ACTIONS(1589),
    [anon_sym_2_GT] = ACTIONS(1589),
    [anon_sym_2_GT_GT] = ACTIONS(1589),
    [anon_sym_3_GT] = ACTIONS(1589),
    [anon_sym_3_GT_GT] = ACTIONS(1589),
    [anon_sym_4_GT] = ACTIONS(1589),
    [anon_sym_4_GT_GT] = ACTIONS(1589),
    [anon_sym_5_GT] = ACTIONS(1589),
    [anon_sym_5_GT_GT] = ACTIONS(1589),
    [anon_sym_6_GT] = ACTIONS(1589),
    [anon_sym_6_GT_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT] = ACTIONS(1589),
    [anon_sym_STAR_GT_GT] = ACTIONS(1589),
    [anon_sym_LT] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1589),
    [anon_sym_2_GT_AMP1] = ACTIONS(1589),
    [anon_sym_3_GT_AMP1] = ACTIONS(1589),
    [anon_sym_4_GT_AMP1] = ACTIONS(1589),
    [anon_sym_5_GT_AMP1] = ACTIONS(1589),
    [anon_sym_6_GT_AMP1] = ACTIONS(1589),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1589),
    [anon_sym_1_GT_AMP2] = ACTIONS(1589),
    [anon_sym_3_GT_AMP2] = ACTIONS(1589),
    [anon_sym_4_GT_AMP2] = ACTIONS(1589),
    [anon_sym_5_GT_AMP2] = ACTIONS(1589),
    [anon_sym_6_GT_AMP2] = ACTIONS(1589),
    [aux_sym_comparison_operator_token37] = ACTIONS(1589),
    [aux_sym_comparison_operator_token50] = ACTIONS(1589),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1589),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1589),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1589),
    [anon_sym_DOLLAR_] = ACTIONS(1589),
    [aux_sym_variable_token1] = ACTIONS(1589),
    [aux_sym_variable_token2] = ACTIONS(1589),
    [sym_braced_variable] = ACTIONS(1589),
    [sym_generic_token] = ACTIONS(1589),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_LBRACE] = ACTIONS(1589),
    [anon_sym_AMP_AMP] = ACTIONS(1589),
    [anon_sym_PIPE_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1589),
    [anon_sym_SPACE] = ACTIONS(1672),
    [anon_sym_PLUS] = ACTIONS(1589),
    [anon_sym_DASH] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1589),
    [anon_sym_BANG] = ACTIONS(1589),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1589),
    [anon_sym_PLUS_PLUS] = ACTIONS(1589),
    [anon_sym_DASH_DASH] = ACTIONS(1589),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LPAREN] = ACTIONS(1589),
    [anon_sym_AT_LBRACE] = ACTIONS(1589),
  },
  [STATE(505)] = {
    [aux_sym_command_argument_sep_repeat1] = STATE(505),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1595),
    [sym_hexadecimal_integer_literal] = ACTIONS(1595),
    [sym_real_literal] = ACTIONS(1595),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1595),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1595),
    [sym_verbatim_string_characters] = ACTIONS(1595),
    [sym_verbatim_here_string_characters] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_GT] = ACTIONS(1595),
    [anon_sym_GT_GT] = ACTIONS(1595),
    [anon_sym_2_GT] = ACTIONS(1595),
    [anon_sym_2_GT_GT] = ACTIONS(1595),
    [anon_sym_3_GT] = ACTIONS(1595),
    [anon_sym_3_GT_GT] = ACTIONS(1595),
    [anon_sym_4_GT] = ACTIONS(1595),
    [anon_sym_4_GT_GT] = ACTIONS(1595),
    [anon_sym_5_GT] = ACTIONS(1595),
    [anon_sym_5_GT_GT] = ACTIONS(1595),
    [anon_sym_6_GT] = ACTIONS(1595),
    [anon_sym_6_GT_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT] = ACTIONS(1595),
    [anon_sym_STAR_GT_GT] = ACTIONS(1595),
    [anon_sym_LT] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP1] = ACTIONS(1595),
    [anon_sym_2_GT_AMP1] = ACTIONS(1595),
    [anon_sym_3_GT_AMP1] = ACTIONS(1595),
    [anon_sym_4_GT_AMP1] = ACTIONS(1595),
    [anon_sym_5_GT_AMP1] = ACTIONS(1595),
    [anon_sym_6_GT_AMP1] = ACTIONS(1595),
    [anon_sym_STAR_GT_AMP2] = ACTIONS(1595),
    [anon_sym_1_GT_AMP2] = ACTIONS(1595),
    [anon_sym_3_GT_AMP2] = ACTIONS(1595),
    [anon_sym_4_GT_AMP2] = ACTIONS(1595),
    [anon_sym_5_GT_AMP2] = ACTIONS(1595),
    [anon_sym_6_GT_AMP2] = ACTIONS(1595),
    [aux_sym_comparison_operator_token37] = ACTIONS(1595),
    [aux_sym_comparison_operator_token50] = ACTIONS(1595),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1595),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1595),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1595),
    [anon_sym_DOLLAR_] = ACTIONS(1595),
    [aux_sym_variable_token1] = ACTIONS(1595),
    [aux_sym_variable_token2] = ACTIONS(1595),
    [sym_braced_variable] = ACTIONS(1595),
    [sym_generic_token] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1595),
    [anon_sym_RPAREN] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_SPACE] = ACTIONS(1674),
    [anon_sym_PLUS] = ACTIONS(1595),
    [anon_sym_DASH] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token1] = ACTIONS(1595),
    [anon_sym_BANG] = ACTIONS(1595),
    [aux_sym_expression_with_unary_operator_token2] = ACTIONS(1595),
    [anon_sym_PLUS_PLUS] = ACTIONS(1595),
    [anon_sym_DASH_DASH] = ACTIONS(1595),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LPAREN] = ACTIONS(1595),
    [anon_sym_AT_LBRACE] = ACTIONS(1595),
  },
  [STATE(506)] = {
    [sym__literal] = STATE(1285),
    [sym_integer_literal] = STATE(1285),
    [sym_string_literal] = STATE(1285),
    [sym_expandable_string_literal] = STATE(933),
    [sym_expandable_here_string_literal] = STATE(933),
    [sym_variable] = STATE(1223),
    [sym_data_command] = STATE(1565),
    [sym_command_name] = STATE(1632),
    [sym_path_command_name] = STATE(1632),
    [sym_command_name_expr] = STATE(1569),
    [sym__primary_expression] = STATE(1285),
    [sym__value] = STATE(1285),
    [sym_parenthesized_expression] = STATE(1285),
    [sym_sub_expression] = STATE(1285),
    [sym_array_expression] = STATE(1285),
    [sym_script_block_expression] = STATE(1285),
    [sym_hash_literal_expression] = STATE(1285),
    [sym_post_increment_expression] = STATE(1285),
    [sym_post_decrement_expression] = STATE(1285),
    [sym_member_access] = STATE(1285),
    [sym_element_access] = STATE(1285),
    [sym_invokation_expression] = STATE(1285),
    [sym_invokation_foreach_expression] = STATE(945),
    [sym_type_literal] = STATE(1285),
    [aux_sym_path_command_name_repeat1] = STATE(1248),
    [sym_comment] = ACTIONS(3),
    [sym_decimal_integer_literal] = ACTIONS(1140),
    [sym_hexadecimal_integer_literal] = ACTIONS(1140),
    [sym_real_literal] = ACTIONS(1651),
    [aux_sym_expandable_string_literal_token1] = ACTIONS(1234),
    [aux_sym_expandable_here_string_literal_token1] = ACTIONS(1236),
    [sym_verbatim_string_characters] = ACTIONS(1238),
    [sym_verbatim_here_string_characters] = ACTIONS(1238),
    [anon_sym_LBRACK] = ACTIONS(1653),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1655),
    [anon_sym_DOLLAR_CARET] = ACTIONS(1655),
    [anon_sym_DOLLAR_QMARK] = ACTIONS(1655),
    [anon_sym_DOLLAR_] = ACTIONS(1655),
    [aux_sym_variable_token1] = ACTIONS(1655),
    [aux_sym_variable_token2] = ACTIONS(1655),
    [sym_braced_variable] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1244),
    [anon_sym_LBRACE] = ACTIONS(1246),
    [aux_sym_command_name_token1] = ACTIONS(1657),
    [aux_sym_command_name_token2] = ACTIONS(1657),
    [aux_sym_command_name_token3] = ACTIONS(1657),
    [aux_sym_command_name_token4] = ACTIONS(1657),
    [aux_sym_command_name_token5] = ACTIONS(1657),
    [aux_sym_command_name_token6] = ACTIONS(1657),
    [aux_sym_command_name_token7] = ACTIONS(1657),
    [aux_sym_command_name_token8] = ACTIONS(1657),
    [aux_sym_command_name_token9] = ACTIONS(1657),
    [aux_sym_command_name_token10] = ACTIONS(1657),
    [aux_sym_command_name_token11] = ACTIONS(1657),
    [aux_sym_command_name_token12] = ACTIONS(1657),
    [aux_sym_command_name_token13] = ACTIONS(1657),
    [aux_sym_command_name_token14] = ACTIONS(1657),
    [aux_sym_command_name_token15] = ACTIONS(1657),
    [aux_sym_command_name_token16] = ACTIONS(1657),
    [aux_sym_command_name_token17] = ACTIONS(1657),
    [aux_sym_command_name_token18] = ACTIONS(1657),
    [sym_path_command_name_token] = ACTIONS(1659),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1252),
    [anon_sym_AT_LPAREN] = ACTIONS(1254),
    [anon_sym_AT_LBRACE] = ACTIONS(1256),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(794), 1,
      aux_sym_format_operator_token1,
    STATE(591), 1,
      sym_format_operator,
    ACTIONS(1679), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1677), 58,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
  [75] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1685), 1,
      anon_sym_DOT_DOT,
    ACTIONS(1683), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1681), 59,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_format_operator_token1,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
  [148] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1028), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1030), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1040), 1,
      anon_sym_LPAREN,
    ACTIONS(1042), 1,
      anon_sym_LBRACE,
    ACTIONS(1048), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1052), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1687), 1,
      sym_real_literal,
    ACTIONS(1689), 1,
      anon_sym_LBRACK,
    ACTIONS(1695), 1,
      sym_path_command_name_token,
    STATE(687), 1,
      sym_variable,
    STATE(717), 1,
      sym_command_name_expr,
    STATE(726), 1,
      aux_sym_path_command_name_repeat1,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1024), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1032), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1103), 2,
      sym_command_name,
      sym_path_command_name,
    ACTIONS(1691), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(906), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
    ACTIONS(1693), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [257] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1685), 1,
      anon_sym_DOT_DOT,
    ACTIONS(1699), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1697), 59,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_format_operator_token1,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
  [330] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1703), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1701), 60,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_format_operator_token1,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_DOT_DOT,
  [401] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(814), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1705), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(812), 56,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_PLUS,
      anon_sym_RBRACK,
  [474] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1060), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1062), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1072), 1,
      anon_sym_LPAREN,
    ACTIONS(1074), 1,
      anon_sym_LBRACE,
    ACTIONS(1080), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1082), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1084), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1707), 1,
      sym_real_literal,
    ACTIONS(1709), 1,
      anon_sym_LBRACK,
    ACTIONS(1715), 1,
      sym_path_command_name_token,
    STATE(686), 1,
      sym_variable,
    STATE(714), 1,
      sym_command_name_expr,
    STATE(725), 1,
      aux_sym_path_command_name_repeat1,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1056), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1064), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1106), 2,
      sym_command_name,
      sym_path_command_name,
    ACTIONS(1711), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(908), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
    ACTIONS(1713), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [583] = 34,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(508), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    STATE(518), 1,
      sym_format_argument_expression,
    STATE(541), 1,
      sym_multiplicative_argument_expression,
    STATE(581), 1,
      sym_additive_argument_expression,
    STATE(596), 1,
      sym_comparison_argument_expression,
    STATE(1337), 1,
      sym_logical_argument_expression,
    STATE(1339), 1,
      sym_bitwise_argument_expression,
    STATE(1568), 1,
      sym_argument_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [716] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(928), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(930), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(942), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(950), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(952), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1717), 1,
      sym_real_literal,
    ACTIONS(1719), 1,
      anon_sym_LBRACK,
    ACTIONS(1723), 1,
      sym_path_command_name_token,
    STATE(683), 1,
      sym_variable,
    STATE(702), 1,
      sym_command_name_expr,
    STATE(713), 1,
      aux_sym_path_command_name_repeat1,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    ACTIONS(924), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(932), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1052), 2,
      sym_command_name,
      sym_path_command_name,
    ACTIONS(1721), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(888), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
    ACTIONS(63), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [825] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(958), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(960), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(970), 1,
      anon_sym_LPAREN,
    ACTIONS(972), 1,
      anon_sym_LBRACE,
    ACTIONS(978), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(980), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(982), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1725), 1,
      sym_real_literal,
    ACTIONS(1727), 1,
      anon_sym_LBRACK,
    ACTIONS(1731), 1,
      sym_path_command_name_token,
    STATE(680), 1,
      sym_variable,
    STATE(712), 1,
      sym_command_name_expr,
    STATE(716), 1,
      aux_sym_path_command_name_repeat1,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    ACTIONS(954), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(962), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1041), 2,
      sym_command_name,
      sym_path_command_name,
    ACTIONS(1729), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(857), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
    ACTIONS(1122), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [934] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1735), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1733), 60,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_format_operator_token1,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_DOT_DOT,
  [1005] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(794), 1,
      aux_sym_format_operator_token1,
    STATE(591), 1,
      sym_format_operator,
    ACTIONS(1739), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1737), 58,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
  [1080] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(822), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1705), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(820), 56,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_PLUS,
      anon_sym_RBRACK,
  [1153] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(822), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1741), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(820), 56,
      sym__statement_terminator,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_PLUS,
  [1226] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(814), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1741), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(812), 56,
      sym__statement_terminator,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_PLUS,
  [1299] = 33,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(186), 1,
      sym_format_expression,
    STATE(187), 1,
      sym_range_expression,
    STATE(193), 1,
      sym_multiplicative_expression,
    STATE(197), 1,
      sym_additive_expression,
    STATE(214), 1,
      sym_comparison_expression,
    STATE(886), 1,
      sym_bitwise_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [1429] = 33,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(469), 1,
      sym_range_expression,
    STATE(477), 1,
      sym_format_expression,
    STATE(521), 1,
      sym_multiplicative_expression,
    STATE(566), 1,
      sym_additive_expression,
    STATE(571), 1,
      sym_comparison_expression,
    STATE(1307), 1,
      sym_bitwise_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [1559] = 33,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(183), 1,
      sym_format_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(192), 1,
      sym_range_expression,
    STATE(195), 1,
      sym_multiplicative_expression,
    STATE(198), 1,
      sym_additive_expression,
    STATE(202), 1,
      sym_comparison_expression,
    STATE(887), 1,
      sym_bitwise_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [1689] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1593), 1,
      sym__statement_terminator,
    ACTIONS(1589), 61,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_generic_token,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [1759] = 33,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(467), 1,
      sym_range_expression,
    STATE(483), 1,
      sym_format_expression,
    STATE(512), 1,
      sym_multiplicative_expression,
    STATE(558), 1,
      sym_additive_expression,
    STATE(577), 1,
      sym_comparison_expression,
    STATE(1315), 1,
      sym_bitwise_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [1889] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1589), 62,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_generic_token,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [1957] = 33,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(460), 1,
      sym_format_expression,
    STATE(462), 1,
      sym_range_expression,
    STATE(496), 1,
      sym_multiplicative_expression,
    STATE(546), 1,
      sym_additive_expression,
    STATE(560), 1,
      sym_comparison_expression,
    STATE(1286), 1,
      sym_bitwise_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2087] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1751), 1,
      anon_sym_RBRACE,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1806), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2212] = 32,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(508), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    STATE(518), 1,
      sym_format_argument_expression,
    STATE(541), 1,
      sym_multiplicative_argument_expression,
    STATE(581), 1,
      sym_additive_argument_expression,
    STATE(596), 1,
      sym_comparison_argument_expression,
    STATE(1334), 1,
      sym_bitwise_argument_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2339] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1759), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1820), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2464] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1761), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1828), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2589] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1763), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1784), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2714] = 32,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(183), 1,
      sym_format_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(192), 1,
      sym_range_expression,
    STATE(195), 1,
      sym_multiplicative_expression,
    STATE(198), 1,
      sym_additive_expression,
    STATE(211), 1,
      sym_comparison_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2841] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1765), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1969), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [2966] = 32,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(469), 1,
      sym_range_expression,
    STATE(477), 1,
      sym_format_expression,
    STATE(521), 1,
      sym_multiplicative_expression,
    STATE(566), 1,
      sym_additive_expression,
    STATE(573), 1,
      sym_comparison_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3093] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1767), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1763), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3218] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1771), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1773), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(1769), 54,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
  [3289] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1775), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2119), 1,
      sym_hash_literal_body,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3414] = 32,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(186), 1,
      sym_format_expression,
    STATE(187), 1,
      sym_range_expression,
    STATE(193), 1,
      sym_multiplicative_expression,
    STATE(197), 1,
      sym_additive_expression,
    STATE(216), 1,
      sym_comparison_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3541] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1779), 3,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
      anon_sym_DASH,
    ACTIONS(1773), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_PERCENT,
      anon_sym_STAR,
    ACTIONS(1777), 54,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_PLUS,
  [3612] = 32,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(460), 1,
      sym_format_expression,
    STATE(462), 1,
      sym_range_expression,
    STATE(496), 1,
      sym_multiplicative_expression,
    STATE(546), 1,
      sym_additive_expression,
    STATE(563), 1,
      sym_comparison_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3739] = 32,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(467), 1,
      sym_range_expression,
    STATE(483), 1,
      sym_format_expression,
    STATE(512), 1,
      sym_multiplicative_expression,
    STATE(558), 1,
      sym_additive_expression,
    STATE(579), 1,
      sym_comparison_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3866] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1781), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1758), 1,
      sym_hash_literal_body,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(548), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [3991] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1783), 1,
      anon_sym_PLUS,
    ACTIONS(1785), 1,
      anon_sym_DASH,
    ACTIONS(838), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(836), 56,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
  [4063] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1783), 1,
      anon_sym_PLUS,
    ACTIONS(1785), 1,
      anon_sym_DASH,
    ACTIONS(830), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(828), 56,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
  [4135] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1787), 1,
      sym_decimal_integer_literal,
    ACTIONS(1790), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1793), 1,
      sym_real_literal,
    ACTIONS(1796), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1799), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1805), 1,
      sym_simple_name,
    ACTIONS(1808), 1,
      anon_sym_LBRACK,
    ACTIONS(1820), 1,
      anon_sym_LPAREN,
    ACTIONS(1823), 1,
      anon_sym_LBRACE,
    ACTIONS(1826), 1,
      anon_sym_RBRACE,
    ACTIONS(1831), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1834), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1837), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1840), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1843), 1,
      anon_sym_AT_LBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1802), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1817), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1828), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(547), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1814), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1811), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4257] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1745), 1,
      sym_simple_name,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1846), 1,
      anon_sym_RBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(2115), 1,
      sym_unary_expression,
    STATE(2124), 1,
      sym_key_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(547), 2,
      sym_hash_entry,
      aux_sym_hash_literal_body_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4379] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(460), 1,
      sym_format_expression,
    STATE(462), 1,
      sym_range_expression,
    STATE(496), 1,
      sym_multiplicative_expression,
    STATE(545), 1,
      sym_additive_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4503] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(467), 1,
      sym_range_expression,
    STATE(483), 1,
      sym_format_expression,
    STATE(512), 1,
      sym_multiplicative_expression,
    STATE(555), 1,
      sym_additive_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4627] = 31,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(186), 1,
      sym_format_expression,
    STATE(187), 1,
      sym_range_expression,
    STATE(193), 1,
      sym_multiplicative_expression,
    STATE(200), 1,
      sym_additive_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4751] = 31,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(469), 1,
      sym_range_expression,
    STATE(477), 1,
      sym_format_expression,
    STATE(521), 1,
      sym_multiplicative_expression,
    STATE(564), 1,
      sym_additive_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4875] = 31,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(183), 1,
      sym_format_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(192), 1,
      sym_range_expression,
    STATE(195), 1,
      sym_multiplicative_expression,
    STATE(199), 1,
      sym_additive_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [4999] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(186), 1,
      sym_format_expression,
    STATE(187), 1,
      sym_range_expression,
    STATE(194), 1,
      sym_multiplicative_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5120] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_PLUS,
    ACTIONS(1850), 1,
      anon_sym_DASH,
    ACTIONS(838), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(836), 55,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_RBRACK,
  [5191] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1852), 1,
      sym_real_literal,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(962), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(93), 3,
      anon_sym_EQ,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1273), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5308] = 30,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(183), 1,
      sym_format_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(192), 1,
      sym_range_expression,
    STATE(196), 1,
      sym_multiplicative_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5429] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_PLUS,
    ACTIONS(1850), 1,
      anon_sym_DASH,
    ACTIONS(830), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(828), 55,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_RBRACK,
  [5500] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(174), 1,
      sym_unary_expression,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(93), 3,
      anon_sym_EQ,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5617] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(549), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(846), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [5688] = 30,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(469), 1,
      sym_range_expression,
    STATE(477), 1,
      sym_format_expression,
    STATE(520), 1,
      sym_multiplicative_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5809] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(467), 1,
      sym_range_expression,
    STATE(483), 1,
      sym_format_expression,
    STATE(519), 1,
      sym_multiplicative_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [5930] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(549), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(902), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [6001] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1862), 1,
      anon_sym_PLUS,
    ACTIONS(1864), 1,
      anon_sym_DASH,
    ACTIONS(838), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(836), 55,
      sym__statement_terminator,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
  [6072] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1866), 1,
      sym_real_literal,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1876), 1,
      anon_sym_DASH_DASH,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(182), 1,
      sym_unary_expression,
    STATE(565), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1872), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(93), 3,
      sym__statement_terminator,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1870), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1305), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6189] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1862), 1,
      anon_sym_PLUS,
    ACTIONS(1864), 1,
      anon_sym_DASH,
    ACTIONS(830), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(828), 55,
      sym__statement_terminator,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
  [6260] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(460), 1,
      sym_format_expression,
    STATE(462), 1,
      sym_range_expression,
    STATE(499), 1,
      sym_multiplicative_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6381] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1878), 1,
      sym_real_literal,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(990), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(93), 3,
      sym__statement_terminator,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1295), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6498] = 30,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(508), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    STATE(518), 1,
      sym_format_argument_expression,
    STATE(541), 1,
      sym_multiplicative_argument_expression,
    STATE(585), 1,
      sym_additive_argument_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6619] = 29,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(462), 1,
      sym_range_expression,
    STATE(465), 1,
      sym_format_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6737] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(552), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(846), 7,
      sym__statement_terminator,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [6807] = 29,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1138), 1,
      sym__statement_terminator,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1866), 1,
      sym_real_literal,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1876), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1888), 1,
      sym_label,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(565), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    STATE(2076), 1,
      sym_unary_expression,
    STATE(2095), 1,
      sym_label_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1872), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1870), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1305), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [6925] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(552), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(902), 7,
      sym__statement_terminator,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [6995] = 29,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(185), 1,
      sym_format_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(192), 1,
      sym_range_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7113] = 29,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(508), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    STATE(518), 1,
      sym_format_argument_expression,
    STATE(538), 1,
      sym_multiplicative_argument_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7231] = 29,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(467), 1,
      sym_range_expression,
    STATE(473), 1,
      sym_format_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7349] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(550), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(846), 7,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_RBRACK,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [7419] = 29,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(469), 1,
      sym_range_expression,
    STATE(480), 1,
      sym_format_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7537] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(550), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(902), 7,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
      anon_sym_RBRACK,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [7607] = 29,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(187), 1,
      sym_range_expression,
    STATE(188), 1,
      sym_format_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7725] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1894), 1,
      anon_sym_PLUS,
    ACTIONS(1896), 1,
      anon_sym_DASH,
    ACTIONS(1892), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(1890), 53,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [7794] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(507), 1,
      sym_format_argument_expression,
    STATE(508), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [7909] = 28,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(191), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    STATE(482), 1,
      sym_range_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8024] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    STATE(466), 1,
      sym_range_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8139] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1894), 1,
      anon_sym_PLUS,
    ACTIONS(1896), 1,
      anon_sym_DASH,
    ACTIONS(1900), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(1898), 53,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [8208] = 28,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(189), 1,
      sym_range_expression,
    STATE(191), 1,
      sym_array_literal_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8323] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    STATE(478), 1,
      sym_range_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8438] = 28,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(163), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(190), 1,
      sym_range_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8553] = 27,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(162), 1,
      sym_array_literal_expression,
    STATE(164), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8665] = 27,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(162), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(454), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8777] = 27,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(510), 1,
      sym_range_argument_expression,
    STATE(517), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [8889] = 27,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(162), 1,
      sym_array_literal_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(452), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9001] = 27,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(184), 1,
      sym_array_literal_expression,
    STATE(450), 1,
      sym_unary_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9113] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(569), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(1902), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [9181] = 27,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(171), 1,
      sym_unary_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(184), 1,
      sym_array_literal_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9293] = 5,
    ACTIONS(79), 1,
      sym_comment,
    STATE(569), 1,
      sym_comparison_operator,
    ACTIONS(852), 2,
      aux_sym_comparison_operator_token28,
      aux_sym_comparison_operator_token34,
    ACTIONS(1904), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
    ACTIONS(850), 48,
      aux_sym_comparison_operator_token1,
      aux_sym_comparison_operator_token2,
      aux_sym_comparison_operator_token3,
      aux_sym_comparison_operator_token4,
      aux_sym_comparison_operator_token5,
      aux_sym_comparison_operator_token6,
      aux_sym_comparison_operator_token7,
      aux_sym_comparison_operator_token8,
      aux_sym_comparison_operator_token9,
      aux_sym_comparison_operator_token10,
      aux_sym_comparison_operator_token11,
      aux_sym_comparison_operator_token12,
      aux_sym_comparison_operator_token13,
      aux_sym_comparison_operator_token14,
      aux_sym_comparison_operator_token15,
      aux_sym_comparison_operator_token16,
      aux_sym_comparison_operator_token17,
      aux_sym_comparison_operator_token18,
      aux_sym_comparison_operator_token19,
      aux_sym_comparison_operator_token20,
      aux_sym_comparison_operator_token21,
      aux_sym_comparison_operator_token22,
      aux_sym_comparison_operator_token23,
      aux_sym_comparison_operator_token24,
      aux_sym_comparison_operator_token25,
      aux_sym_comparison_operator_token26,
      aux_sym_comparison_operator_token27,
      aux_sym_comparison_operator_token29,
      aux_sym_comparison_operator_token30,
      aux_sym_comparison_operator_token31,
      aux_sym_comparison_operator_token32,
      aux_sym_comparison_operator_token33,
      aux_sym_comparison_operator_token35,
      aux_sym_comparison_operator_token36,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token38,
      aux_sym_comparison_operator_token39,
      aux_sym_comparison_operator_token40,
      aux_sym_comparison_operator_token41,
      aux_sym_comparison_operator_token42,
      aux_sym_comparison_operator_token43,
      aux_sym_comparison_operator_token44,
      aux_sym_comparison_operator_token45,
      aux_sym_comparison_operator_token46,
      aux_sym_comparison_operator_token47,
      aux_sym_comparison_operator_token48,
      aux_sym_comparison_operator_token49,
      aux_sym_comparison_operator_token50,
  [9361] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(178), 1,
      sym_unary_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9470] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(181), 1,
      sym_unary_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9579] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(153), 1,
      sym_real_literal,
    ACTIONS(159), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(161), 1,
      anon_sym_DASH_DASH,
    STATE(5), 1,
      sym_type_literal,
    STATE(131), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(157), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(155), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(150), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9688] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(153), 1,
      sym_real_literal,
    ACTIONS(159), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(161), 1,
      anon_sym_DASH_DASH,
    STATE(5), 1,
      sym_type_literal,
    STATE(133), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(157), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(155), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(150), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9797] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1180), 1,
      sym_real_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1202), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1204), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(180), 1,
      sym_unary_expression,
    STATE(286), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1200), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1190), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1071), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [9906] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1866), 1,
      sym_real_literal,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1876), 1,
      anon_sym_DASH_DASH,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(181), 1,
      sym_unary_expression,
    STATE(565), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1872), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1870), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1305), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10015] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1866), 1,
      sym_real_literal,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1876), 1,
      anon_sym_DASH_DASH,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(178), 1,
      sym_unary_expression,
    STATE(565), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1872), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1870), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1305), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10124] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(180), 1,
      sym_unary_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10233] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1926), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1928), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    STATE(218), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(1010), 1,
      sym_expression_with_unary_operator,
    STATE(1015), 1,
      sym_unary_expression,
    ACTIONS(936), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1011), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1918), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10342] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1926), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1928), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    STATE(218), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(1010), 1,
      sym_expression_with_unary_operator,
    STATE(1016), 1,
      sym_unary_expression,
    ACTIONS(936), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1011), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1918), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10451] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1926), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1928), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    STATE(218), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(1010), 1,
      sym_expression_with_unary_operator,
    STATE(1017), 1,
      sym_unary_expression,
    ACTIONS(936), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1011), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1918), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10560] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1866), 1,
      sym_real_literal,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1876), 1,
      anon_sym_DASH_DASH,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(179), 1,
      sym_unary_expression,
    STATE(565), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1872), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1870), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1305), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10669] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(167), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10778] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(169), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10887] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(170), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [10996] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(179), 1,
      sym_unary_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11105] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(161), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11214] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1878), 1,
      sym_real_literal,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(987), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1295), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11323] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1878), 1,
      sym_real_literal,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(988), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1295), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11432] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1878), 1,
      sym_real_literal,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(989), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1295), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11541] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1956), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1958), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    STATE(222), 1,
      sym_type_literal,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(998), 1,
      sym_expression_with_unary_operator,
    STATE(1007), 1,
      sym_unary_expression,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(990), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(982), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1948), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11650] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1852), 1,
      sym_real_literal,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(958), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1273), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11759] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1852), 1,
      sym_real_literal,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(960), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1273), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11868] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1852), 1,
      sym_real_literal,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(961), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1273), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [11977] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1968), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1970), 1,
      anon_sym_DASH_DASH,
    STATE(221), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(804), 1,
      sym_unary_expression,
    STATE(810), 1,
      sym_expression_with_unary_operator,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(984), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(803), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1966), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12086] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1968), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1970), 1,
      anon_sym_DASH_DASH,
    STATE(221), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(806), 1,
      sym_unary_expression,
    STATE(810), 1,
      sym_expression_with_unary_operator,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(984), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(803), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1966), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12195] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1968), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1970), 1,
      anon_sym_DASH_DASH,
    STATE(221), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(807), 1,
      sym_unary_expression,
    STATE(810), 1,
      sym_expression_with_unary_operator,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(984), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(803), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1966), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12304] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1974), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1976), 1,
      anon_sym_DASH_DASH,
    STATE(220), 1,
      sym_type_literal,
    STATE(779), 1,
      sym_expression_with_unary_operator,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(812), 1,
      sym_unary_expression,
    ACTIONS(966), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(811), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1972), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12413] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1974), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1976), 1,
      anon_sym_DASH_DASH,
    STATE(220), 1,
      sym_type_literal,
    STATE(779), 1,
      sym_expression_with_unary_operator,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(813), 1,
      sym_unary_expression,
    ACTIONS(966), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(811), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1972), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12522] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1974), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1976), 1,
      anon_sym_DASH_DASH,
    STATE(220), 1,
      sym_type_literal,
    STATE(779), 1,
      sym_expression_with_unary_operator,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(814), 1,
      sym_unary_expression,
    ACTIONS(966), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(811), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1972), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12631] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(1998), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2000), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    STATE(240), 1,
      sym_type_literal,
    STATE(852), 1,
      sym_unary_expression,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(874), 1,
      sym_expression_with_unary_operator,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(847), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1990), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12740] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(1998), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2000), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    STATE(240), 1,
      sym_type_literal,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(874), 1,
      sym_expression_with_unary_operator,
    STATE(894), 1,
      sym_unary_expression,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(847), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1990), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12849] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(1998), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2000), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    STATE(240), 1,
      sym_type_literal,
    STATE(866), 1,
      sym_unary_expression,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(874), 1,
      sym_expression_with_unary_operator,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(847), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1990), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [12958] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2028), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2030), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    STATE(241), 1,
      sym_type_literal,
    STATE(831), 1,
      sym_expression_with_unary_operator,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(849), 1,
      sym_unary_expression,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1096), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(839), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2020), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13067] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2028), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2030), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    STATE(241), 1,
      sym_type_literal,
    STATE(831), 1,
      sym_expression_with_unary_operator,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(858), 1,
      sym_unary_expression,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1096), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(839), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2020), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13176] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2028), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2030), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    STATE(241), 1,
      sym_type_literal,
    STATE(831), 1,
      sym_expression_with_unary_operator,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(864), 1,
      sym_unary_expression,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1096), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(839), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2020), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13285] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(542), 1,
      sym_real_literal,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(550), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(552), 1,
      anon_sym_DASH_DASH,
    STATE(83), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(511), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(548), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(546), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(433), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13394] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(181), 1,
      sym_unary_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13503] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(178), 1,
      sym_unary_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13612] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(564), 1,
      sym_real_literal,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(572), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(574), 1,
      anon_sym_DASH_DASH,
    STATE(85), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(179), 1,
      sym_unary_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(570), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(568), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(435), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13721] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(161), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13830] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(163), 1,
      sym_real_literal,
    ACTIONS(169), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(171), 1,
      anon_sym_DASH_DASH,
    STATE(6), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(123), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(167), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(165), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(151), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [13939] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(163), 1,
      sym_real_literal,
    ACTIONS(169), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(171), 1,
      anon_sym_DASH_DASH,
    STATE(6), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(108), 1,
      sym_unary_expression,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(167), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(165), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(151), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14048] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(163), 1,
      sym_real_literal,
    ACTIONS(169), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(171), 1,
      anon_sym_DASH_DASH,
    STATE(6), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(109), 1,
      sym_unary_expression,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(167), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(165), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(151), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14157] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1956), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1958), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    STATE(222), 1,
      sym_type_literal,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(993), 1,
      sym_unary_expression,
    STATE(998), 1,
      sym_expression_with_unary_operator,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(990), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(982), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1948), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14266] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1956), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1958), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    STATE(222), 1,
      sym_type_literal,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(998), 1,
      sym_expression_with_unary_operator,
    STATE(999), 1,
      sym_unary_expression,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(990), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(982), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1948), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14375] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1938), 1,
      sym_real_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1956), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1958), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    STATE(222), 1,
      sym_type_literal,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(998), 1,
      sym_expression_with_unary_operator,
    STATE(1001), 1,
      sym_unary_expression,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(990), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(982), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1948), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(794), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14484] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1908), 1,
      sym_real_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1926), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1928), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    STATE(218), 1,
      sym_type_literal,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(1010), 1,
      sym_expression_with_unary_operator,
    STATE(1030), 1,
      sym_unary_expression,
    ACTIONS(936), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1011), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1918), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(753), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14593] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1180), 1,
      sym_real_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1202), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1204), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(181), 1,
      sym_unary_expression,
    STATE(286), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1200), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1190), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1071), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14702] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1180), 1,
      sym_real_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1202), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1204), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(178), 1,
      sym_unary_expression,
    STATE(286), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1200), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1190), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1071), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14811] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1180), 1,
      sym_real_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1202), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1204), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(179), 1,
      sym_unary_expression,
    STATE(286), 1,
      sym_type_literal,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1200), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1190), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1071), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [14920] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1144), 1,
      sym_real_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1166), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1168), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    STATE(161), 1,
      sym_unary_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(285), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1164), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1154), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1072), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15029] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1144), 1,
      sym_real_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1166), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1168), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(167), 1,
      sym_unary_expression,
    STATE(285), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1164), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1154), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1072), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15138] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1144), 1,
      sym_real_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1166), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1168), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(169), 1,
      sym_unary_expression,
    STATE(285), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1164), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1154), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1072), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15247] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1144), 1,
      sym_real_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1166), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1168), 1,
      anon_sym_DASH_DASH,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(170), 1,
      sym_unary_expression,
    STATE(285), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1164), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1154), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1072), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15356] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(576), 1,
      sym_real_literal,
    ACTIONS(582), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(584), 1,
      anon_sym_DASH_DASH,
    STATE(86), 1,
      sym_type_literal,
    STATE(131), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(580), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(578), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(434), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15465] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(576), 1,
      sym_real_literal,
    ACTIONS(582), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(584), 1,
      anon_sym_DASH_DASH,
    STATE(86), 1,
      sym_type_literal,
    STATE(132), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(580), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(578), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(434), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15574] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(576), 1,
      sym_real_literal,
    ACTIONS(582), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(584), 1,
      anon_sym_DASH_DASH,
    STATE(86), 1,
      sym_type_literal,
    STATE(133), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(580), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(578), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(434), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15683] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(83), 1,
      sym_real_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(109), 1,
      anon_sym_DASH_DASH,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    STATE(3), 1,
      sym_type_literal,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    STATE(176), 1,
      sym_expression_with_unary_operator,
    STATE(180), 1,
      sym_unary_expression,
    ACTIONS(19), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(177), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(97), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(160), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15792] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(554), 1,
      sym_real_literal,
    ACTIONS(560), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(562), 1,
      anon_sym_DASH_DASH,
    STATE(84), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(123), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(558), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(556), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(431), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [15901] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(554), 1,
      sym_real_literal,
    ACTIONS(560), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(562), 1,
      anon_sym_DASH_DASH,
    STATE(84), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(108), 1,
      sym_unary_expression,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(558), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(556), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(431), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16010] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(554), 1,
      sym_real_literal,
    ACTIONS(560), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(562), 1,
      anon_sym_DASH_DASH,
    STATE(84), 1,
      sym_type_literal,
    STATE(104), 1,
      sym_expression_with_unary_operator,
    STATE(109), 1,
      sym_unary_expression,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(558), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(556), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(431), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16119] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1222), 1,
      sym_real_literal,
    ACTIONS(1228), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1230), 1,
      anon_sym_DASH_DASH,
    STATE(288), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(987), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1226), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1224), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1036), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16228] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1222), 1,
      sym_real_literal,
    ACTIONS(1228), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1230), 1,
      anon_sym_DASH_DASH,
    STATE(288), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(988), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1226), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1224), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1036), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16337] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1222), 1,
      sym_real_literal,
    ACTIONS(1228), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1230), 1,
      anon_sym_DASH_DASH,
    STATE(288), 1,
      sym_type_literal,
    STATE(985), 1,
      sym_expression_with_unary_operator,
    STATE(989), 1,
      sym_unary_expression,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1226), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1224), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1036), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16446] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2040), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2042), 1,
      anon_sym_DASH_DASH,
    STATE(237), 1,
      sym_type_literal,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(1058), 1,
      sym_expression_with_unary_operator,
    STATE(1073), 1,
      sym_unary_expression,
    ACTIONS(1068), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1065), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2038), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16555] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1212), 1,
      sym_real_literal,
    ACTIONS(1218), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1220), 1,
      anon_sym_DASH_DASH,
    STATE(287), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(958), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1214), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1037), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16664] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1212), 1,
      sym_real_literal,
    ACTIONS(1218), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1220), 1,
      anon_sym_DASH_DASH,
    STATE(287), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(960), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1214), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1037), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16773] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1212), 1,
      sym_real_literal,
    ACTIONS(1218), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1220), 1,
      anon_sym_DASH_DASH,
    STATE(287), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(956), 1,
      sym_expression_with_unary_operator,
    STATE(961), 1,
      sym_unary_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1214), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1037), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16882] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2046), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2048), 1,
      anon_sym_DASH_DASH,
    STATE(235), 1,
      sym_type_literal,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(1059), 1,
      sym_unary_expression,
    STATE(1064), 1,
      sym_expression_with_unary_operator,
    ACTIONS(1036), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1076), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2044), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [16991] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2040), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2042), 1,
      anon_sym_DASH_DASH,
    STATE(237), 1,
      sym_type_literal,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(1058), 1,
      sym_expression_with_unary_operator,
    STATE(1066), 1,
      sym_unary_expression,
    ACTIONS(1068), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1065), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2038), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17100] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2040), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2042), 1,
      anon_sym_DASH_DASH,
    STATE(237), 1,
      sym_type_literal,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(1058), 1,
      sym_expression_with_unary_operator,
    STATE(1067), 1,
      sym_unary_expression,
    ACTIONS(1068), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1065), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2038), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17209] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1980), 1,
      sym_real_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2040), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2042), 1,
      anon_sym_DASH_DASH,
    STATE(237), 1,
      sym_type_literal,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(1058), 1,
      sym_expression_with_unary_operator,
    STATE(1069), 1,
      sym_unary_expression,
    ACTIONS(1068), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1065), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2038), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(859), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17318] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(167), 1,
      sym_unary_expression,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17427] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2046), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2048), 1,
      anon_sym_DASH_DASH,
    STATE(235), 1,
      sym_type_literal,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(1064), 1,
      sym_expression_with_unary_operator,
    STATE(1077), 1,
      sym_unary_expression,
    ACTIONS(1036), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1076), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2044), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17536] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2046), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2048), 1,
      anon_sym_DASH_DASH,
    STATE(235), 1,
      sym_type_literal,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(1064), 1,
      sym_expression_with_unary_operator,
    STATE(1078), 1,
      sym_unary_expression,
    ACTIONS(1036), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1076), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2044), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17645] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2010), 1,
      sym_real_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2046), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2048), 1,
      anon_sym_DASH_DASH,
    STATE(235), 1,
      sym_type_literal,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(1064), 1,
      sym_expression_with_unary_operator,
    STATE(1079), 1,
      sym_unary_expression,
    ACTIONS(1036), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1076), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2044), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(835), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17754] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(169), 1,
      sym_unary_expression,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17863] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(167), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [17972] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(169), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [18081] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(121), 1,
      sym_real_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(143), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(145), 1,
      anon_sym_DASH_DASH,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    STATE(4), 1,
      sym_type_literal,
    STATE(120), 1,
      sym_invokation_foreach_expression,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(170), 1,
      sym_unary_expression,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(141), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(131), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(159), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [18190] = 26,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1743), 1,
      sym_real_literal,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1755), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1757), 1,
      anon_sym_DASH_DASH,
    STATE(165), 1,
      sym_expression_with_unary_operator,
    STATE(170), 1,
      sym_unary_expression,
    STATE(559), 1,
      sym_type_literal,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1753), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(166), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1749), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1314), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [18299] = 26,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(153), 1,
      sym_real_literal,
    ACTIONS(159), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(161), 1,
      anon_sym_DASH_DASH,
    STATE(5), 1,
      sym_type_literal,
    STATE(132), 1,
      sym_unary_expression,
    STATE(135), 1,
      sym_expression_with_unary_operator,
    STATE(138), 1,
      sym_invokation_foreach_expression,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(157), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(155), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(150), 16,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
  [18408] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2055), 1,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2053), 8,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_path_command_name_token,
    ACTIONS(2050), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [18474] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 50,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [18534] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 49,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [18594] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2055), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2053), 8,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_path_command_name_token,
    ACTIONS(2050), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [18660] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 49,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [18719] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 48,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [18778] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2055), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2053), 8,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_path_command_name_token,
    ACTIONS(2050), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [18843] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2055), 1,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2053), 8,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_path_command_name_token,
    ACTIONS(2050), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [18908] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2058), 50,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_DOT,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [18964] = 25,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(582), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(584), 1,
      anon_sym_DASH_DASH,
    STATE(86), 1,
      sym_type_literal,
    STATE(115), 1,
      sym_member_name,
    STATE(137), 1,
      sym_string_literal,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(580), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2060), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(578), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(136), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19065] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(979), 1,
      sym_member_name,
    STATE(1032), 1,
      sym_string_literal,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2062), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1031), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19166] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(927), 1,
      sym_member_name,
    STATE(934), 1,
      sym_string_literal,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2064), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(973), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19267] = 25,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(566), 1,
      anon_sym_LBRACK,
    ACTIONS(582), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(584), 1,
      anon_sym_DASH_DASH,
    STATE(86), 1,
      sym_type_literal,
    STATE(111), 1,
      sym_member_name,
    STATE(137), 1,
      sym_string_literal,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(580), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2060), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(578), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(136), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19368] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(560), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(562), 1,
      anon_sym_DASH_DASH,
    STATE(84), 1,
      sym_type_literal,
    STATE(89), 1,
      sym_member_name,
    STATE(92), 1,
      sym_string_literal,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(558), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2066), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(556), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(91), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19469] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(544), 1,
      anon_sym_LBRACK,
    ACTIONS(560), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(562), 1,
      anon_sym_DASH_DASH,
    STATE(84), 1,
      sym_type_literal,
    STATE(88), 1,
      sym_member_name,
    STATE(92), 1,
      sym_string_literal,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(558), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2066), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(556), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(91), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19570] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(924), 1,
      sym_decimal_integer_literal,
    ACTIONS(1906), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1910), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1912), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1916), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_LPAREN,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    ACTIONS(1930), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1934), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1968), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1970), 1,
      anon_sym_DASH_DASH,
    STATE(221), 1,
      sym_type_literal,
    STATE(736), 1,
      sym_member_name,
    STATE(759), 1,
      sym_string_literal,
    ACTIONS(938), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(984), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1914), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(2068), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(803), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1920), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1966), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(758), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19671] = 25,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(159), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(161), 1,
      anon_sym_DASH_DASH,
    STATE(5), 1,
      sym_type_literal,
    STATE(111), 1,
      sym_member_name,
    STATE(137), 1,
      sym_string_literal,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(157), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2060), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(155), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(136), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19772] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1228), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1230), 1,
      anon_sym_DASH_DASH,
    STATE(288), 1,
      sym_type_literal,
    STATE(979), 1,
      sym_member_name,
    STATE(1032), 1,
      sym_string_literal,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1226), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2062), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1224), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1031), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19873] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1218), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1220), 1,
      anon_sym_DASH_DASH,
    STATE(287), 1,
      sym_type_literal,
    STATE(927), 1,
      sym_member_name,
    STATE(934), 1,
      sym_string_literal,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2064), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1214), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(973), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [19974] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(954), 1,
      sym_decimal_integer_literal,
    ACTIONS(1936), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1940), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1942), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1952), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    ACTIONS(1960), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1962), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1964), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1974), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1976), 1,
      anon_sym_DASH_DASH,
    STATE(220), 1,
      sym_type_literal,
    STATE(739), 1,
      sym_member_name,
    STATE(773), 1,
      sym_string_literal,
    ACTIONS(966), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(968), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1944), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(2070), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(811), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1950), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1972), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(772), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20075] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(169), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(171), 1,
      anon_sym_DASH_DASH,
    STATE(6), 1,
      sym_type_literal,
    STATE(89), 1,
      sym_member_name,
    STATE(92), 1,
      sym_string_literal,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(167), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2066), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(165), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(91), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20176] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1188), 1,
      anon_sym_LBRACK,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1228), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1230), 1,
      anon_sym_DASH_DASH,
    STATE(288), 1,
      sym_type_literal,
    STATE(981), 1,
      sym_member_name,
    STATE(1032), 1,
      sym_string_literal,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1226), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2062), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1224), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1031), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20277] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 1,
      anon_sym_LPAREN,
    ACTIONS(2080), 1,
      anon_sym_LBRACE,
    ACTIONS(2084), 1,
      anon_sym_SPACE,
    ACTIONS(2086), 1,
      anon_sym_COLON,
    ACTIONS(2088), 1,
      sym__statement_terminator,
    STATE(226), 1,
      sym_command_argument_sep,
    STATE(438), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(740), 1,
      sym_file_redirection_operator,
    STATE(1040), 1,
      sym_merging_redirection_operator,
    STATE(1351), 1,
      sym_command_elements,
    ACTIONS(2076), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(719), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1020), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2082), 4,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2074), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [20362] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1152), 1,
      anon_sym_LBRACK,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1218), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1220), 1,
      anon_sym_DASH_DASH,
    STATE(287), 1,
      sym_type_literal,
    STATE(920), 1,
      sym_member_name,
    STATE(934), 1,
      sym_string_literal,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2064), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1214), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(973), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20463] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1056), 1,
      sym_decimal_integer_literal,
    ACTIONS(1978), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1982), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1984), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1988), 1,
      anon_sym_LBRACK,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    ACTIONS(1998), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2000), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2004), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2006), 1,
      anon_sym_AT_LBRACE,
    STATE(240), 1,
      sym_type_literal,
    STATE(789), 1,
      sym_member_name,
    STATE(865), 1,
      sym_string_literal,
    ACTIONS(1070), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1986), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(2090), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(847), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1992), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1990), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(899), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20564] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1140), 1,
      sym_decimal_integer_literal,
    ACTIONS(1142), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1146), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1148), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1160), 1,
      anon_sym_LPAREN,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1172), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1174), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1747), 1,
      anon_sym_LBRACK,
    ACTIONS(1858), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1860), 1,
      anon_sym_DASH_DASH,
    STATE(556), 1,
      sym_type_literal,
    STATE(920), 1,
      sym_member_name,
    STATE(934), 1,
      sym_string_literal,
    ACTIONS(1150), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1158), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1856), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2064), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(957), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1156), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1854), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(973), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20665] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(117), 1,
      sym_decimal_integer_literal,
    ACTIONS(119), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(123), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(125), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(137), 1,
      anon_sym_LPAREN,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    ACTIONS(147), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(149), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(151), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(169), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(171), 1,
      anon_sym_DASH_DASH,
    STATE(6), 1,
      sym_type_literal,
    STATE(88), 1,
      sym_member_name,
    STATE(92), 1,
      sym_string_literal,
    ACTIONS(127), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(167), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2066), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(119), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(105), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(165), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(91), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20766] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1176), 1,
      sym_decimal_integer_literal,
    ACTIONS(1178), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(1182), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1184), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1196), 1,
      anon_sym_LPAREN,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    ACTIONS(1206), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1208), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1210), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1868), 1,
      anon_sym_LBRACK,
    ACTIONS(1884), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(1886), 1,
      anon_sym_DASH_DASH,
    STATE(568), 1,
      sym_type_literal,
    STATE(981), 1,
      sym_member_name,
    STATE(1032), 1,
      sym_string_literal,
    ACTIONS(1186), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(1194), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1882), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2062), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(986), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(1192), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(1880), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(1031), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20867] = 25,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(85), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(87), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(91), 1,
      anon_sym_LBRACK,
    ACTIONS(103), 1,
      anon_sym_LPAREN,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(113), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(115), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(159), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(161), 1,
      anon_sym_DASH_DASH,
    STATE(5), 1,
      sym_type_literal,
    STATE(115), 1,
      sym_member_name,
    STATE(137), 1,
      sym_string_literal,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(89), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(157), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2060), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(157), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(130), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(155), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(136), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [20968] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 1,
      anon_sym_LPAREN,
    ACTIONS(2080), 1,
      anon_sym_LBRACE,
    ACTIONS(2084), 1,
      anon_sym_SPACE,
    ACTIONS(2086), 1,
      anon_sym_COLON,
    ACTIONS(2094), 1,
      sym__statement_terminator,
    STATE(226), 1,
      sym_command_argument_sep,
    STATE(438), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(740), 1,
      sym_file_redirection_operator,
    STATE(1040), 1,
      sym_merging_redirection_operator,
    STATE(1355), 1,
      sym_command_elements,
    ACTIONS(2076), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(719), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1020), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2092), 4,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2074), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21053] = 25,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1024), 1,
      sym_decimal_integer_literal,
    ACTIONS(2008), 1,
      sym_hexadecimal_integer_literal,
    ACTIONS(2012), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2014), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2018), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_LPAREN,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    ACTIONS(2028), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2030), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2032), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2034), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2036), 1,
      anon_sym_AT_LBRACE,
    STATE(241), 1,
      sym_type_literal,
    STATE(809), 1,
      sym_member_name,
    STATE(863), 1,
      sym_string_literal,
    ACTIONS(1038), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(1096), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2016), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    ACTIONS(2096), 2,
      sym_real_literal,
      sym_simple_name,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(839), 3,
      sym_pre_increment_expression,
      sym_pre_decrement_expression,
      sym_cast_expression,
    ACTIONS(2022), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2020), 6,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_COMMA,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
    STATE(862), 10,
      sym__literal,
      sym_integer_literal,
      sym_variable,
      sym_expression_with_unary_operator,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
  [21154] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2102), 1,
      anon_sym_LPAREN,
    ACTIONS(2104), 1,
      anon_sym_LBRACE,
    ACTIONS(2106), 1,
      anon_sym_SPACE,
    ACTIONS(2108), 1,
      anon_sym_COLON,
    STATE(238), 1,
      sym_command_argument_sep,
    STATE(445), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(744), 1,
      sym_file_redirection_operator,
    STATE(1061), 1,
      sym_merging_redirection_operator,
    STATE(1353), 1,
      sym_command_elements,
    ACTIONS(2100), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(720), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1003), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2092), 5,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2098), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21237] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2102), 1,
      anon_sym_LPAREN,
    ACTIONS(2104), 1,
      anon_sym_LBRACE,
    ACTIONS(2106), 1,
      anon_sym_SPACE,
    ACTIONS(2108), 1,
      anon_sym_COLON,
    STATE(238), 1,
      sym_command_argument_sep,
    STATE(445), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(744), 1,
      sym_file_redirection_operator,
    STATE(1061), 1,
      sym_merging_redirection_operator,
    STATE(1376), 1,
      sym_command_elements,
    ACTIONS(2100), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(720), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1003), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2082), 5,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2098), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21320] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2114), 1,
      sym_path_command_name_token,
    ACTIONS(2116), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    STATE(715), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2112), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2110), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [21382] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2088), 1,
      sym__statement_terminator,
    ACTIONS(2122), 1,
      anon_sym_LPAREN,
    ACTIONS(2124), 1,
      anon_sym_LBRACE,
    ACTIONS(2126), 1,
      anon_sym_SPACE,
    ACTIONS(2128), 1,
      anon_sym_COLON,
    STATE(247), 1,
      sym_command_argument_sep,
    STATE(457), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(742), 1,
      sym_file_redirection_operator,
    STATE(1093), 1,
      sym_merging_redirection_operator,
    STATE(1351), 1,
      sym_command_elements,
    ACTIONS(2120), 2,
      sym_command_parameter,
      sym_stop_parsing,
    ACTIONS(2082), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    STATE(729), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1062), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2118), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21466] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2135), 1,
      sym_path_command_name_token,
    ACTIONS(2138), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    STATE(715), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2132), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2130), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [21528] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2116), 1,
      anon_sym_SPACE,
    ACTIONS(2142), 1,
      sym_path_command_name_token,
    STATE(723), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2140), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2110), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [21590] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2148), 1,
      anon_sym_LPAREN,
    ACTIONS(2150), 1,
      anon_sym_LBRACE,
    ACTIONS(2152), 1,
      anon_sym_SPACE,
    ACTIONS(2154), 1,
      anon_sym_COLON,
    STATE(246), 1,
      sym_command_argument_sep,
    STATE(453), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(741), 1,
      sym_file_redirection_operator,
    STATE(1097), 1,
      sym_merging_redirection_operator,
    STATE(1376), 1,
      sym_command_elements,
    ACTIONS(2146), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(730), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1050), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2082), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    ACTIONS(2144), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21672] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym__statement_terminator,
    ACTIONS(2122), 1,
      anon_sym_LPAREN,
    ACTIONS(2124), 1,
      anon_sym_LBRACE,
    ACTIONS(2126), 1,
      anon_sym_SPACE,
    ACTIONS(2128), 1,
      anon_sym_COLON,
    STATE(247), 1,
      sym_command_argument_sep,
    STATE(457), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(742), 1,
      sym_file_redirection_operator,
    STATE(1093), 1,
      sym_merging_redirection_operator,
    STATE(1355), 1,
      sym_command_elements,
    ACTIONS(2120), 2,
      sym_command_parameter,
      sym_stop_parsing,
    ACTIONS(2092), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    STATE(729), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1062), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2118), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21756] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 1,
      anon_sym_LPAREN,
    ACTIONS(2080), 1,
      anon_sym_LBRACE,
    ACTIONS(2084), 1,
      anon_sym_SPACE,
    ACTIONS(2086), 1,
      anon_sym_COLON,
    ACTIONS(2160), 1,
      sym__statement_terminator,
    STATE(226), 1,
      sym_command_argument_sep,
    STATE(438), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(740), 1,
      sym_file_redirection_operator,
    STATE(1040), 1,
      sym_merging_redirection_operator,
    ACTIONS(2156), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(721), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1020), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2158), 4,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2074), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21838] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2102), 1,
      anon_sym_LPAREN,
    ACTIONS(2104), 1,
      anon_sym_LBRACE,
    ACTIONS(2106), 1,
      anon_sym_SPACE,
    ACTIONS(2108), 1,
      anon_sym_COLON,
    STATE(238), 1,
      sym_command_argument_sep,
    STATE(445), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(744), 1,
      sym_file_redirection_operator,
    STATE(1061), 1,
      sym_merging_redirection_operator,
    ACTIONS(2162), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(724), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1003), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2158), 5,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2098), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [21918] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2173), 1,
      anon_sym_LPAREN,
    ACTIONS(2176), 1,
      anon_sym_LBRACE,
    ACTIONS(2181), 1,
      anon_sym_SPACE,
    ACTIONS(2184), 1,
      anon_sym_COLON,
    ACTIONS(2187), 1,
      sym__statement_terminator,
    STATE(226), 1,
      sym_command_argument_sep,
    STATE(438), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(740), 1,
      sym_file_redirection_operator,
    STATE(1040), 1,
      sym_merging_redirection_operator,
    ACTIONS(2170), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(721), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1020), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2179), 4,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2167), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2164), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22000] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2148), 1,
      anon_sym_LPAREN,
    ACTIONS(2150), 1,
      anon_sym_LBRACE,
    ACTIONS(2152), 1,
      anon_sym_SPACE,
    ACTIONS(2154), 1,
      anon_sym_COLON,
    STATE(246), 1,
      sym_command_argument_sep,
    STATE(453), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(741), 1,
      sym_file_redirection_operator,
    STATE(1097), 1,
      sym_merging_redirection_operator,
    STATE(1353), 1,
      sym_command_elements,
    ACTIONS(2146), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(730), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1050), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2092), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    ACTIONS(2144), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22082] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2138), 1,
      anon_sym_SPACE,
    ACTIONS(2192), 1,
      sym_path_command_name_token,
    STATE(723), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2189), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2130), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [22144] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2201), 1,
      anon_sym_LPAREN,
    ACTIONS(2204), 1,
      anon_sym_LBRACE,
    ACTIONS(2207), 1,
      anon_sym_SPACE,
    ACTIONS(2210), 1,
      anon_sym_COLON,
    STATE(238), 1,
      sym_command_argument_sep,
    STATE(445), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(744), 1,
      sym_file_redirection_operator,
    STATE(1061), 1,
      sym_merging_redirection_operator,
    ACTIONS(2198), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(724), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1003), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2179), 5,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      anon_sym_DASH_DASH_PERCENT,
    ACTIONS(2195), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2164), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22224] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2215), 1,
      sym_path_command_name_token,
    ACTIONS(2116), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    STATE(731), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2213), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2110), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [22285] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2116), 1,
      anon_sym_SPACE,
    ACTIONS(2219), 1,
      sym_path_command_name_token,
    STATE(732), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2217), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2110), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [22346] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2187), 1,
      sym__statement_terminator,
    ACTIONS(2227), 1,
      anon_sym_LPAREN,
    ACTIONS(2230), 1,
      anon_sym_LBRACE,
    ACTIONS(2233), 1,
      anon_sym_SPACE,
    ACTIONS(2236), 1,
      anon_sym_COLON,
    STATE(247), 1,
      sym_command_argument_sep,
    STATE(457), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(742), 1,
      sym_file_redirection_operator,
    STATE(1093), 1,
      sym_merging_redirection_operator,
    ACTIONS(2224), 2,
      sym_command_parameter,
      sym_stop_parsing,
    ACTIONS(2179), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    STATE(727), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1062), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2221), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2164), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22427] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2245), 1,
      anon_sym_LPAREN,
    ACTIONS(2248), 1,
      anon_sym_LBRACE,
    ACTIONS(2251), 1,
      anon_sym_SPACE,
    ACTIONS(2254), 1,
      anon_sym_COLON,
    STATE(246), 1,
      sym_command_argument_sep,
    STATE(453), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(741), 1,
      sym_file_redirection_operator,
    STATE(1097), 1,
      sym_merging_redirection_operator,
    ACTIONS(2242), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(728), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1050), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2179), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    ACTIONS(2239), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2164), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22506] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2122), 1,
      anon_sym_LPAREN,
    ACTIONS(2124), 1,
      anon_sym_LBRACE,
    ACTIONS(2126), 1,
      anon_sym_SPACE,
    ACTIONS(2128), 1,
      anon_sym_COLON,
    ACTIONS(2160), 1,
      sym__statement_terminator,
    STATE(247), 1,
      sym_command_argument_sep,
    STATE(457), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(742), 1,
      sym_file_redirection_operator,
    STATE(1093), 1,
      sym_merging_redirection_operator,
    ACTIONS(2257), 2,
      sym_command_parameter,
      sym_stop_parsing,
    ACTIONS(2158), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    STATE(727), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1062), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2118), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22587] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2148), 1,
      anon_sym_LPAREN,
    ACTIONS(2150), 1,
      anon_sym_LBRACE,
    ACTIONS(2152), 1,
      anon_sym_SPACE,
    ACTIONS(2154), 1,
      anon_sym_COLON,
    STATE(246), 1,
      sym_command_argument_sep,
    STATE(453), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(741), 1,
      sym_file_redirection_operator,
    STATE(1097), 1,
      sym_merging_redirection_operator,
    ACTIONS(2259), 2,
      sym_command_parameter,
      sym_stop_parsing,
    STATE(728), 3,
      sym__command_element,
      sym_redirection,
      aux_sym_command_elements_repeat1,
    STATE(1050), 3,
      sym__command_argument,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(2158), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    ACTIONS(2144), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
    ACTIONS(2072), 15,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
  [22666] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2264), 1,
      sym_path_command_name_token,
    ACTIONS(2138), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    STATE(731), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2261), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2130), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [22727] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2138), 1,
      anon_sym_SPACE,
    ACTIONS(2270), 1,
      sym_path_command_name_token,
    STATE(732), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(2267), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    ACTIONS(2130), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [22788] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(538), 12,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(540), 35,
      sym_simple_name,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [22843] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2273), 1,
      sym_real_literal,
    ACTIONS(2275), 1,
      anon_sym_SPACE,
    ACTIONS(2277), 1,
      anon_sym_COLON,
    STATE(289), 1,
      sym_command_argument_sep,
    STATE(504), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1131), 1,
      sym__command_argument,
    STATE(1133), 1,
      sym_redirected_file_name,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1086), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1094), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [22937] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1260), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1262), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1270), 1,
      anon_sym_LPAREN,
    ACTIONS(1272), 1,
      anon_sym_LBRACE,
    ACTIONS(1278), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1280), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1282), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(2279), 1,
      sym_real_literal,
    ACTIONS(2281), 1,
      anon_sym_LBRACK,
    ACTIONS(2283), 1,
      anon_sym_SPACE,
    ACTIONS(2285), 1,
      anon_sym_COLON,
    STATE(290), 1,
      sym_command_argument_sep,
    STATE(491), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(1012), 1,
      sym_invokation_foreach_expression,
    STATE(1130), 1,
      sym_redirected_file_name,
    STATE(1134), 1,
      sym__command_argument,
    ACTIONS(1176), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1264), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(1009), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(1100), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(1194), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1096), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23031] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(780), 1,
      sym_argument_list,
    ACTIONS(590), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(592), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [23087] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 1,
      anon_sym_SPACE,
    ACTIONS(588), 45,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [23141] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2287), 1,
      sym_real_literal,
    ACTIONS(2289), 1,
      sym__command_token,
    ACTIONS(2291), 1,
      anon_sym_RBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    STATE(2120), 1,
      sym_switch_clauses,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(771), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23231] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(778), 1,
      sym_argument_list,
    ACTIONS(590), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(592), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [23287] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(928), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(930), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(942), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(950), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(952), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1719), 1,
      anon_sym_LBRACK,
    ACTIONS(2084), 1,
      anon_sym_SPACE,
    ACTIONS(2086), 1,
      anon_sym_COLON,
    ACTIONS(2293), 1,
      sym_real_literal,
    STATE(226), 1,
      sym_command_argument_sep,
    STATE(438), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(754), 1,
      sym_invokation_foreach_expression,
    STATE(1042), 1,
      sym__command_argument,
    STATE(1043), 1,
      sym_redirected_file_name,
    ACTIONS(924), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(932), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(820), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(826), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(938), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(890), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23381] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1028), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1030), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1040), 1,
      anon_sym_LPAREN,
    ACTIONS(1042), 1,
      anon_sym_LBRACE,
    ACTIONS(1048), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1052), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1689), 1,
      anon_sym_LBRACK,
    ACTIONS(2152), 1,
      anon_sym_SPACE,
    ACTIONS(2154), 1,
      anon_sym_COLON,
    ACTIONS(2295), 1,
      sym_real_literal,
    STATE(246), 1,
      sym_command_argument_sep,
    STATE(453), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(837), 1,
      sym_invokation_foreach_expression,
    STATE(1082), 1,
      sym__command_argument,
    STATE(1099), 1,
      sym_redirected_file_name,
    ACTIONS(1024), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1032), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(830), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(910), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(1038), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(909), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23475] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1060), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1062), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1072), 1,
      anon_sym_LPAREN,
    ACTIONS(1074), 1,
      anon_sym_LBRACE,
    ACTIONS(1080), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1082), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1084), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1709), 1,
      anon_sym_LBRACK,
    ACTIONS(2126), 1,
      anon_sym_SPACE,
    ACTIONS(2128), 1,
      anon_sym_COLON,
    ACTIONS(2297), 1,
      sym_real_literal,
    STATE(247), 1,
      sym_command_argument_sep,
    STATE(457), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(868), 1,
      sym_invokation_foreach_expression,
    STATE(1089), 1,
      sym__command_argument,
    STATE(1090), 1,
      sym_redirected_file_name,
    ACTIONS(1056), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1064), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(842), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(905), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(1070), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(907), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23569] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2287), 1,
      sym_real_literal,
    ACTIONS(2289), 1,
      sym__command_token,
    ACTIONS(2299), 1,
      anon_sym_RBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    STATE(2037), 1,
      sym_switch_clauses,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(771), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23659] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(958), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(960), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(970), 1,
      anon_sym_LPAREN,
    ACTIONS(972), 1,
      anon_sym_LBRACE,
    ACTIONS(978), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(980), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(982), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1727), 1,
      anon_sym_LBRACK,
    ACTIONS(2106), 1,
      anon_sym_SPACE,
    ACTIONS(2108), 1,
      anon_sym_COLON,
    ACTIONS(2301), 1,
      sym_real_literal,
    STATE(238), 1,
      sym_command_argument_sep,
    STATE(445), 1,
      aux_sym_command_argument_sep_repeat1,
    STATE(796), 1,
      sym_invokation_foreach_expression,
    STATE(1074), 1,
      sym__command_argument,
    STATE(1075), 1,
      sym_redirected_file_name,
    ACTIONS(954), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(962), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(782), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    STATE(883), 2,
      sym_parenthesized_expression,
      sym_script_block_expression,
    ACTIONS(968), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(833), 15,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_sub_expression,
      sym_array_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23753] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2287), 1,
      sym_real_literal,
    ACTIONS(2289), 1,
      sym__command_token,
    ACTIONS(2303), 1,
      anon_sym_RBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    STATE(2021), 1,
      sym_switch_clauses,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(771), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23843] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2287), 1,
      sym_real_literal,
    ACTIONS(2289), 1,
      sym__command_token,
    ACTIONS(2305), 1,
      anon_sym_RBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    STATE(1904), 1,
      sym_switch_clauses,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(771), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [23933] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(588), 44,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [23987] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24040] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24093] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(682), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(684), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24146] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(588), 43,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
  [24199] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(634), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(636), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24252] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACK,
    ACTIONS(2309), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2311), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2315), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(654), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2313), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(656), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [24315] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(714), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24368] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(650), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(652), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24421] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(718), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24474] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 1,
      anon_sym_SPACE,
    ACTIONS(588), 44,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_path_command_name_token,
      sym_stop_parsing,
      anon_sym_COLON,
  [24527] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(602), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24580] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(604), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(607), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24633] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(634), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(636), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24686] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(614), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(616), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(666), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(668), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24792] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(686), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(688), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24845] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(690), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(692), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24898] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(618), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(620), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [24951] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(624), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25004] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(698), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25057] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(706), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25110] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(610), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(612), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25163] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(682), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(684), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25216] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2287), 1,
      sym_real_literal,
    ACTIONS(2289), 1,
      sym__command_token,
    ACTIONS(2317), 1,
      anon_sym_RBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(819), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [25303] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(602), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25356] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(604), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(607), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(626), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(628), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25462] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(614), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(616), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25515] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(618), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(620), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25568] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(626), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(628), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25621] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(630), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(632), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25674] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(630), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(632), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25780] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(702), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25833] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(710), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25886] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(640), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25939] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(686), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(688), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [25992] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(722), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26045] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(664), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26098] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(690), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(692), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26151] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(644), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26204] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(828), 1,
      sym_argument_list,
    ACTIONS(590), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(592), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26259] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(624), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26312] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(698), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26365] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(706), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26418] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(722), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26471] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 1,
      anon_sym_SPACE,
    ACTIONS(2319), 1,
      anon_sym_LBRACK,
    ACTIONS(2321), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2323), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2327), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2325), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(656), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [26534] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(640), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(714), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26640] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(650), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(652), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26693] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(718), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26746] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(664), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26799] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(644), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(648), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26905] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(598), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [26958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(660), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27011] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(726), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27064] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(702), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(672), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27170] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(676), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(680), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27276] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(871), 1,
      sym_argument_list,
    ACTIONS(590), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(592), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27331] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27384] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(660), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27437] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(726), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27490] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(672), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27543] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(676), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27596] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(680), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27649] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(610), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(612), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27702] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(648), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27755] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(598), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27808] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2332), 1,
      sym_real_literal,
    ACTIONS(2335), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(2338), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(2344), 1,
      anon_sym_LBRACK,
    ACTIONS(2350), 1,
      sym__command_token,
    ACTIONS(2353), 1,
      anon_sym_LPAREN,
    ACTIONS(2356), 1,
      anon_sym_LBRACE,
    ACTIONS(2359), 1,
      anon_sym_RBRACE,
    ACTIONS(2361), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2364), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(2367), 1,
      anon_sym_AT_LBRACE,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1521), 1,
      sym_switch_clause_condition,
    ACTIONS(2329), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(2341), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(819), 2,
      sym_switch_clause,
      aux_sym_switch_clauses_repeat1,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(2347), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1312), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [27895] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(710), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [27948] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(666), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(668), 43,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28001] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(614), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(616), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28053] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(614), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(616), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28105] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(618), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(620), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28157] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(680), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28209] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1006), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(1002), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [28265] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(626), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(628), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28317] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(630), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(632), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28369] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(702), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28421] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(710), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28473] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28525] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28577] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2319), 1,
      anon_sym_LBRACK,
    ACTIONS(2321), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2323), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2327), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2372), 1,
      anon_sym_SPACE,
    ACTIONS(2325), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2370), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [28639] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(540), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(538), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28691] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 1,
      anon_sym_SPACE,
    ACTIONS(2374), 1,
      anon_sym_LBRACK,
    ACTIONS(2376), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2378), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2382), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2380), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(656), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [28753] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(588), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28805] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(714), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(650), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(652), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(660), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [28961] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(718), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29013] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(702), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29065] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(710), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(634), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(636), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29169] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(666), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(668), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29221] = 12,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    STATE(735), 1,
      sym_file_redirection_operator,
    STATE(1129), 1,
      sym_merging_redirection_operator,
    STATE(1367), 1,
      aux_sym__pipeline_tail,
    STATE(1368), 1,
      sym_redirections,
    STATE(1114), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2390), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2384), 7,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2388), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [29291] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(686), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(688), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29343] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(660), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29395] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(690), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(692), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29447] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(726), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29499] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(624), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(698), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29603] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(726), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29655] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(706), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29707] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2398), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2396), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2394), 32,
      sym__statement_terminator,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [29761] = 12,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    STATE(734), 1,
      sym_file_redirection_operator,
    STATE(1132), 1,
      sym_merging_redirection_operator,
    STATE(1338), 1,
      sym_redirections,
    STATE(1341), 1,
      aux_sym__pipeline_tail,
    STATE(1115), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2390), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2384), 7,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2400), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [29831] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2404), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2396), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2394), 32,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [29885] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2319), 1,
      anon_sym_LBRACK,
    ACTIONS(2321), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2323), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2327), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2408), 1,
      anon_sym_SPACE,
    ACTIONS(2325), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2406), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [29947] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(672), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [29999] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2410), 1,
      anon_sym_LBRACK,
    ACTIONS(2412), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2414), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2418), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(654), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2416), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(656), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [30061] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(610), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(612), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30113] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(682), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(684), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30165] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(602), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30217] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(604), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(607), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30269] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(676), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30321] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(604), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(607), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30373] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(676), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30425] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(680), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30477] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(714), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30529] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(626), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(628), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30581] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(598), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30633] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(630), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(632), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30685] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(650), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(652), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30737] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(718), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30789] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30841] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(540), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(538), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30893] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(634), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(636), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(666), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(668), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [30997] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(686), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(688), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31049] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(722), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31101] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(664), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31153] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(690), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(692), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31205] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(624), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31257] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1006), 1,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(1002), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [31313] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(698), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31365] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(706), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31417] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2404), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2422), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2420), 32,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [31471] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2398), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2422), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2420), 32,
      sym__statement_terminator,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [31525] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACK,
    ACTIONS(2309), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2311), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2315), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2313), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2408), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2406), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [31587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(640), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31639] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACK,
    ACTIONS(2309), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2311), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2315), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2313), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2372), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2370), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [31701] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(648), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31753] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(644), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31805] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(648), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(672), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(610), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(612), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [31961] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(598), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32013] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(682), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(684), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32065] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(640), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(602), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32169] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(722), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32221] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(664), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32273] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(644), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32325] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(618), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(620), 42,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32377] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2429), 1,
      aux_sym_command_name_token19,
    ACTIONS(2432), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2435), 1,
      sym__statement_terminator,
    STATE(904), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2424), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [32436] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1006), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(1002), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32491] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2374), 1,
      anon_sym_LBRACK,
    ACTIONS(2376), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2378), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2382), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2408), 1,
      anon_sym_SPACE,
    ACTIONS(2380), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2406), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32552] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2410), 1,
      anon_sym_LBRACK,
    ACTIONS(2412), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2414), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2418), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2372), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2416), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2370), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32613] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2410), 1,
      anon_sym_LBRACK,
    ACTIONS(2412), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2414), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2418), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2408), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2416), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2406), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32674] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2372), 1,
      anon_sym_SPACE,
    ACTIONS(2374), 1,
      anon_sym_LBRACK,
    ACTIONS(2376), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2378), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2382), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2380), 2,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(2370), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32735] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1006), 1,
      anon_sym_SPACE,
    ACTIONS(95), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
    ACTIONS(1002), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [32790] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2440), 1,
      aux_sym_command_name_token19,
    ACTIONS(2443), 1,
      anon_sym_DQUOTE2,
    STATE(911), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2437), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [32847] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2450), 1,
      aux_sym_command_name_token19,
    ACTIONS(2452), 1,
      anon_sym_DQUOTE2,
    STATE(915), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2446), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2448), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [32904] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(540), 2,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(538), 41,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [32955] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(540), 3,
      sym__statement_terminator,
      anon_sym_SPACE,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(538), 40,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
      anon_sym_COLON_COLON,
  [33006] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2452), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2458), 1,
      aux_sym_command_name_token19,
    STATE(911), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2454), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2456), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33063] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2462), 1,
      aux_sym_command_name_token19,
    ACTIONS(2464), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2466), 1,
      sym__statement_terminator,
    STATE(917), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2460), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2448), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33122] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2464), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2470), 1,
      aux_sym_command_name_token19,
    ACTIONS(2472), 1,
      sym__statement_terminator,
    STATE(904), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2468), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2456), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33181] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 1,
      sym__statement_terminator,
    ACTIONS(2477), 1,
      aux_sym_command_name_token19,
    ACTIONS(2480), 1,
      anon_sym_DQUOTE2,
    STATE(918), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2474), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33239] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2466), 1,
      sym__statement_terminator,
    ACTIONS(2485), 1,
      aux_sym_command_name_token19,
    ACTIONS(2487), 1,
      anon_sym_DQUOTE2,
    STATE(923), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2483), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2448), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33297] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2489), 1,
      anon_sym_LPAREN,
    STATE(940), 1,
      sym_argument_list,
    ACTIONS(592), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(590), 30,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [33351] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      aux_sym_command_name_token19,
    ACTIONS(2495), 1,
      anon_sym_DQUOTE2,
    STATE(922), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2491), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2448), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33407] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2495), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2499), 1,
      aux_sym_command_name_token19,
    STATE(924), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2497), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2456), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33463] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2472), 1,
      sym__statement_terminator,
    ACTIONS(2487), 1,
      anon_sym_DQUOTE2,
    ACTIONS(2503), 1,
      aux_sym_command_name_token19,
    STATE(918), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2501), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2456), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33521] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2508), 1,
      aux_sym_command_name_token19,
    ACTIONS(2511), 1,
      anon_sym_DQUOTE2,
    STATE(924), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2505), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [33577] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2514), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2516), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33627] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2520), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33677] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(940), 1,
      sym_argument_list,
    ACTIONS(592), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(590), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [33729] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 40,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33779] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 1,
      aux_sym_expandable_string_literal_token1,
    ACTIONS(1236), 1,
      aux_sym_expandable_here_string_literal_token1,
    ACTIONS(1244), 1,
      anon_sym_LPAREN,
    ACTIONS(1246), 1,
      anon_sym_LBRACE,
    ACTIONS(1252), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1254), 1,
      anon_sym_AT_LPAREN,
    ACTIONS(1256), 1,
      anon_sym_AT_LBRACE,
    ACTIONS(1653), 1,
      anon_sym_LBRACK,
    ACTIONS(2522), 1,
      sym_real_literal,
    ACTIONS(2524), 1,
      sym__command_token,
    STATE(945), 1,
      sym_invokation_foreach_expression,
    STATE(1881), 1,
      sym_switch_filename,
    ACTIONS(1140), 2,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
    ACTIONS(1238), 2,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
    STATE(933), 2,
      sym_expandable_string_literal,
      sym_expandable_here_string_literal,
    ACTIONS(1158), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
    STATE(1318), 17,
      sym__literal,
      sym_integer_literal,
      sym_string_literal,
      sym_variable,
      sym__primary_expression,
      sym__value,
      sym_parenthesized_expression,
      sym_sub_expression,
      sym_array_expression,
      sym_script_block_expression,
      sym_hash_literal_expression,
      sym_post_increment_expression,
      sym_post_decrement_expression,
      sym_member_access,
      sym_element_access,
      sym_invokation_expression,
      sym_type_literal,
  [33859] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2520), 40,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [33959] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2514), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2516), 40,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [34009] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(710), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(708), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34058] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(607), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(604), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34107] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(616), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(614), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34156] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(620), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(618), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34205] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(624), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(622), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34254] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2520), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [34303] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(628), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(626), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34352] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(632), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(630), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34401] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2514), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2516), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [34450] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(636), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(634), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34499] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(668), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(666), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34548] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [34597] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(714), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(712), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34646] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(688), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(686), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34695] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2520), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [34744] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2514), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2516), 39,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [34793] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(588), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(586), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34842] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(640), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(638), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34891] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(644), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(642), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34940] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(692), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(690), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [34989] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(698), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(696), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35038] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(648), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(646), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35087] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(598), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(596), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35136] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(656), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(654), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35185] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(660), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(658), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35234] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(726), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(724), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35283] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(706), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(704), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35332] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(672), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(670), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35381] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(676), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(674), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35430] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(680), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(678), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35479] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2530), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
    ACTIONS(2528), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2526), 29,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35530] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(722), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(720), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35579] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(664), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(662), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35628] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 3,
      sym__statement_terminator,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
      sym_stop_parsing,
      anon_sym_SPACE,
      anon_sym_COLON,
  [35677] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(652), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(650), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35726] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(718), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(716), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35775] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2532), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
    ACTIONS(2528), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2526), 29,
      sym__statement_terminator,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35826] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(612), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(610), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35875] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(684), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(682), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35924] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(702), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(700), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [35973] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(602), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(600), 31,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36022] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(772), 1,
      anon_sym_SPACE,
    ACTIONS(2534), 1,
      anon_sym_COMMA,
    STATE(975), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(774), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36074] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(783), 1,
      anon_sym_SPACE,
    ACTIONS(2534), 1,
      anon_sym_COMMA,
    STATE(976), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(785), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36126] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 1,
      anon_sym_SPACE,
    ACTIONS(2536), 1,
      anon_sym_COMMA,
    STATE(976), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(762), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36178] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2539), 1,
      anon_sym_COMMA,
    STATE(978), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(772), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(774), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36230] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2539), 1,
      anon_sym_COMMA,
    STATE(980), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(783), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(785), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36282] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(983), 1,
      sym_argument_list,
    ACTIONS(592), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(590), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36332] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2541), 1,
      anon_sym_COMMA,
    STATE(980), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(760), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(762), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36384] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2544), 1,
      anon_sym_LPAREN,
    STATE(983), 1,
      sym_argument_list,
    ACTIONS(592), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(590), 28,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36436] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_SPACE,
    ACTIONS(660), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36483] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(632), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(630), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36530] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(588), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(586), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36577] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(656), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(654), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36624] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(660), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(658), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36671] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(726), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(724), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36718] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(672), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(670), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36765] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(676), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(674), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36812] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(680), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(678), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [36859] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2546), 1,
      anon_sym_COMMA,
    STATE(991), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(760), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(762), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [36910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 1,
      anon_sym_SPACE,
    ACTIONS(680), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [36957] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 1,
      anon_sym_SPACE,
    ACTIONS(726), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37004] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(644), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(642), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37051] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2549), 1,
      anon_sym_COMMA,
    STATE(991), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(783), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(785), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [37102] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 1,
      anon_sym_SPACE,
    ACTIONS(2551), 1,
      anon_sym_COMMA,
    STATE(996), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(762), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [37153] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(783), 1,
      anon_sym_SPACE,
    ACTIONS(2554), 1,
      anon_sym_COMMA,
    STATE(996), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(785), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [37204] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 1,
      anon_sym_SPACE,
    ACTIONS(656), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37251] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 1,
      anon_sym_SPACE,
    ACTIONS(672), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37298] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(648), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(646), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37345] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 1,
      anon_sym_SPACE,
    ACTIONS(676), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37392] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(598), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(596), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37439] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2558), 1,
      anon_sym_LPAREN,
    ACTIONS(2560), 1,
      anon_sym_SPACE,
    STATE(1048), 1,
      sym_argument_list,
    ACTIONS(2556), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37490] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(640), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(638), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37537] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2549), 1,
      anon_sym_COMMA,
    STATE(995), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(772), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(774), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [37588] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2562), 39,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
      sym_path_command_name_token,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [37633] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 1,
      anon_sym_SPACE,
    ACTIONS(762), 38,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37680] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(702), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(700), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37727] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(710), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(708), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37774] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(656), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37821] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(660), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [37868] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(714), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(712), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37915] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(652), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(650), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [37962] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(718), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(716), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38009] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(726), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38056] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(672), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38103] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(676), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38150] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(636), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(634), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38197] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(668), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(666), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38244] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2564), 1,
      anon_sym_LPAREN,
    STATE(1045), 1,
      sym_argument_list,
    ACTIONS(2560), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2556), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38295] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(688), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(686), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38342] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(692), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(690), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38389] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(680), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38436] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(698), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(696), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38483] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(706), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(704), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38530] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(722), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(720), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38577] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(664), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(662), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38624] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(612), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(610), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38671] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(684), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(682), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38718] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(762), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [38765] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(602), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(600), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38812] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(607), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(604), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38859] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(616), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(614), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38906] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(620), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(618), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [38953] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(624), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(622), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [39000] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2566), 1,
      anon_sym_LBRACK,
    ACTIONS(2568), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2570), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2572), 1,
      anon_sym_DOT2,
    ACTIONS(2574), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2576), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(654), 24,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39059] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2584), 1,
      anon_sym_DOT2,
    ACTIONS(2586), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(656), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(654), 24,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39118] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(628), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(626), 29,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [39165] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(772), 1,
      anon_sym_SPACE,
    ACTIONS(2554), 1,
      anon_sym_COMMA,
    STATE(997), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(774), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [39216] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2592), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2590), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39262] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2408), 1,
      anon_sym_SPACE,
    ACTIONS(2406), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39308] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2372), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2370), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39354] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2596), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2594), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39400] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2600), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2598), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39446] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2604), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2602), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39492] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(598), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39538] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2600), 1,
      anon_sym_SPACE,
    ACTIONS(2598), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39584] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2604), 1,
      anon_sym_SPACE,
    ACTIONS(2602), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39630] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 1,
      anon_sym_SPACE,
    ACTIONS(648), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39676] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2560), 1,
      anon_sym_SPACE,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    STATE(1087), 1,
      sym_argument_list,
    ACTIONS(2556), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [39726] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 1,
      anon_sym_SPACE,
    ACTIONS(598), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39772] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2408), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2406), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39818] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(648), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39864] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 1,
      anon_sym_SPACE,
    ACTIONS(722), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 1,
      anon_sym_SPACE,
    ACTIONS(664), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [39956] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2608), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40002] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 1,
      anon_sym_SPACE,
    ACTIONS(624), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40048] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(656), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40094] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 1,
      anon_sym_SPACE,
    ACTIONS(762), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40140] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(624), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40186] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2592), 1,
      anon_sym_SPACE,
    ACTIONS(2590), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40232] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2612), 1,
      anon_sym_LPAREN,
    STATE(1101), 1,
      sym_argument_list,
    ACTIONS(2560), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2556), 34,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40282] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(722), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40328] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(654), 1,
      anon_sym_SPACE,
    ACTIONS(656), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40374] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(660), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40420] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(726), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40466] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(672), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40512] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(664), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40558] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(676), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40604] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(680), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40650] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2566), 1,
      anon_sym_LBRACK,
    ACTIONS(2568), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2570), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2576), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2614), 1,
      anon_sym_DOT2,
    ACTIONS(2616), 1,
      anon_sym_COLON_COLON,
    ACTIONS(656), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(654), 23,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40708] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2618), 1,
      anon_sym_DOT2,
    ACTIONS(2620), 1,
      anon_sym_COLON_COLON,
    ACTIONS(656), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(654), 23,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40766] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(760), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(762), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40812] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2372), 1,
      anon_sym_SPACE,
    ACTIONS(2370), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40858] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2596), 1,
      anon_sym_SPACE,
    ACTIONS(2594), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [40904] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_SPACE,
    ACTIONS(660), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40950] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 1,
      anon_sym_SPACE,
    ACTIONS(726), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [40996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(670), 1,
      anon_sym_SPACE,
    ACTIONS(672), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41042] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(674), 1,
      anon_sym_SPACE,
    ACTIONS(676), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41088] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(678), 1,
      anon_sym_SPACE,
    ACTIONS(680), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 1,
      anon_sym_SPACE,
    ACTIONS(2608), 37,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
      anon_sym_DASH_DASH_PERCENT,
  [41180] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2372), 1,
      anon_sym_SPACE,
    ACTIONS(2370), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41225] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(722), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41270] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(664), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41315] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(648), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41360] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(93), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1002), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(1006), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2604), 1,
      anon_sym_SPACE,
    ACTIONS(2602), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41454] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(624), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41499] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2372), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2370), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41544] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2596), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2594), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41589] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2608), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41634] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 1,
      anon_sym_SPACE,
    ACTIONS(2608), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41679] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2592), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2590), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41724] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2618), 1,
      anon_sym_DOT2,
    ACTIONS(2620), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2370), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2372), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41781] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    STATE(734), 1,
      sym_file_redirection_operator,
    STATE(1132), 1,
      sym_merging_redirection_operator,
    STATE(1338), 1,
      sym_redirections,
    STATE(1341), 1,
      aux_sym__pipeline_tail,
    STATE(1115), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2390), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2400), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [41842] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2566), 1,
      anon_sym_LBRACK,
    ACTIONS(2568), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2570), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2576), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(2614), 1,
      anon_sym_DOT2,
    ACTIONS(2616), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2370), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2372), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41899] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2592), 1,
      anon_sym_SPACE,
    ACTIONS(2590), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41944] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2600), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2598), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [41989] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2596), 1,
      anon_sym_SPACE,
    ACTIONS(2594), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42034] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 1,
      anon_sym_DOT2,
    ACTIONS(93), 5,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(1002), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(1006), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [42083] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2604), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2602), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2600), 1,
      anon_sym_SPACE,
    ACTIONS(2598), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42173] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2408), 1,
      anon_sym_SPACE,
    ACTIONS(2406), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42218] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(538), 10,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
      anon_sym_DOT2,
    ACTIONS(540), 27,
      sym__statement_terminator,
      anon_sym_LBRACK,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [42263] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(646), 1,
      anon_sym_SPACE,
    ACTIONS(648), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42308] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2408), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(2406), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42353] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 1,
      anon_sym_SPACE,
    ACTIONS(598), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42398] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(596), 2,
      sym__statement_terminator,
      anon_sym_SPACE,
    ACTIONS(598), 35,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42443] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 1,
      anon_sym_SPACE,
    ACTIONS(722), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42488] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(662), 1,
      anon_sym_SPACE,
    ACTIONS(664), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42533] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(538), 7,
      sym_decimal_integer_literal,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT2,
    ACTIONS(540), 30,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_EQ,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [42578] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(622), 1,
      anon_sym_SPACE,
    ACTIONS(624), 36,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_2_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_LT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      sym_command_parameter,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
      sym_stop_parsing,
      anon_sym_COLON,
  [42623] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    STATE(735), 1,
      sym_file_redirection_operator,
    STATE(1129), 1,
      sym_merging_redirection_operator,
    STATE(1367), 1,
      aux_sym__pipeline_tail,
    STATE(1368), 1,
      sym_redirections,
    STATE(1114), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2390), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2388), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [42684] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2624), 1,
      anon_sym_PIPE,
    STATE(735), 1,
      sym_file_redirection_operator,
    STATE(1129), 1,
      sym_merging_redirection_operator,
    STATE(1116), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2622), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2388), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [42739] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2624), 1,
      anon_sym_PIPE,
    STATE(734), 1,
      sym_file_redirection_operator,
    STATE(1132), 1,
      sym_merging_redirection_operator,
    STATE(1117), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2622), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2386), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2072), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2400), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [42794] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    STATE(735), 1,
      sym_file_redirection_operator,
    STATE(1129), 1,
      sym_merging_redirection_operator,
    STATE(1116), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2635), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2629), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2626), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2632), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [42849] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    STATE(734), 1,
      sym_file_redirection_operator,
    STATE(1132), 1,
      sym_merging_redirection_operator,
    STATE(1117), 2,
      sym_redirection,
      aux_sym_redirections_repeat1,
    ACTIONS(2635), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2629), 7,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
    ACTIONS(2626), 8,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
    ACTIONS(2639), 12,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
  [42904] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(538), 6,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT2,
    ACTIONS(540), 28,
      sym__statement_terminator,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [42946] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2646), 1,
      anon_sym_SEMI,
    STATE(1120), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(2642), 5,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2644), 27,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      sym_simple_name,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [42992] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2650), 1,
      anon_sym_SEMI,
    STATE(1120), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(1328), 5,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2648), 27,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      sym_simple_name,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43038] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2657), 1,
      anon_sym_SEMI,
    STATE(1119), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(2653), 5,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2655), 27,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      sym_simple_name,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43084] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2659), 1,
      anon_sym_COMMA,
    STATE(1125), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(785), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(783), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43129] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2661), 1,
      anon_sym_COMMA,
    STATE(1126), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(785), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(783), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43174] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2659), 1,
      anon_sym_COMMA,
    STATE(1122), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(774), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(772), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43219] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2663), 1,
      anon_sym_COMMA,
    STATE(1125), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(762), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(760), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43264] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2666), 1,
      anon_sym_COMMA,
    STATE(1126), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(762), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(760), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43309] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2661), 1,
      anon_sym_COMMA,
    STATE(1123), 1,
      aux_sym_array_literal_expression_repeat1,
    ACTIONS(774), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(772), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43354] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2598), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2600), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43393] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2590), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2592), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43432] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2594), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2596), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43471] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2370), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2372), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43510] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2590), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2592), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43549] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2594), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2596), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43588] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2370), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2372), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43627] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2608), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2610), 22,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43666] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2608), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2610), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43705] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2598), 9,
      anon_sym_GT,
      anon_sym_2_GT,
      anon_sym_3_GT,
      anon_sym_4_GT,
      anon_sym_5_GT,
      anon_sym_6_GT,
      anon_sym_STAR_GT,
      anon_sym_LT,
      anon_sym_PIPE,
    ACTIONS(2600), 22,
      sym__statement_terminator,
      anon_sym_GT_GT,
      anon_sym_2_GT_GT,
      anon_sym_3_GT_GT,
      anon_sym_4_GT_GT,
      anon_sym_5_GT_GT,
      anon_sym_6_GT_GT,
      anon_sym_STAR_GT_GT,
      anon_sym_STAR_GT_AMP1,
      anon_sym_2_GT_AMP1,
      anon_sym_3_GT_AMP1,
      anon_sym_4_GT_AMP1,
      anon_sym_5_GT_AMP1,
      anon_sym_6_GT_AMP1,
      anon_sym_STAR_GT_AMP2,
      anon_sym_1_GT_AMP2,
      anon_sym_3_GT_AMP2,
      anon_sym_4_GT_AMP2,
      anon_sym_5_GT_AMP2,
      anon_sym_6_GT_AMP2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [43744] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2669), 5,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2671), 25,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43782] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2673), 5,
      sym_decimal_integer_literal,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2675), 25,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      aux_sym_comparison_operator_token37,
      aux_sym_comparison_operator_token50,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_expression_with_unary_operator_token1,
      anon_sym_BANG,
      aux_sym_expression_with_unary_operator_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43820] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2679), 1,
      anon_sym_SEMI,
    STATE(1142), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(2677), 22,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym__command_token,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43854] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2683), 1,
      anon_sym_SEMI,
    STATE(1140), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(2681), 22,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym__command_token,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43888] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2685), 1,
      anon_sym_SEMI,
    STATE(1142), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(1328), 22,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      sym__command_token,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [43922] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(513), 1,
      sym_command_invokation_operator,
    STATE(718), 1,
      sym_command_name,
    STATE(1393), 1,
      sym_command,
    ACTIONS(15), 2,
      anon_sym_DOT,
      anon_sym_AMP,
    ACTIONS(1713), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [43959] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(509), 1,
      sym_command_invokation_operator,
    STATE(722), 1,
      sym_command_name,
    STATE(1403), 1,
      sym_command,
    ACTIONS(15), 2,
      anon_sym_DOT,
      anon_sym_AMP,
    ACTIONS(1693), 18,
      aux_sym_command_name_token1,
      aux_sym_command_name_token2,
      aux_sym_command_name_token3,
      aux_sym_command_name_token4,
      aux_sym_command_name_token5,
      aux_sym_command_name_token6,
      aux_sym_command_name_token7,
      aux_sym_command_name_token8,
      aux_sym_command_name_token9,
      aux_sym_command_name_token10,
      aux_sym_command_name_token11,
      aux_sym_command_name_token12,
      aux_sym_command_name_token13,
      aux_sym_command_name_token14,
      aux_sym_command_name_token15,
      aux_sym_command_name_token16,
      aux_sym_command_name_token17,
      aux_sym_command_name_token18,
  [43996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2690), 1,
      anon_sym_SPACE,
    ACTIONS(2688), 21,
      sym_decimal_integer_literal,
      sym_hexadecimal_integer_literal,
      sym_real_literal,
      aux_sym_expandable_string_literal_token1,
      aux_sym_expandable_here_string_literal_token1,
      sym_verbatim_string_characters,
      sym_verbatim_here_string_characters,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_AT_LPAREN,
      anon_sym_AT_LBRACE,
  [44026] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2696), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1163), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44074] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2700), 1,
      sym_simple_name,
    ACTIONS(2703), 1,
      anon_sym_LBRACK,
    ACTIONS(2712), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(2709), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2714), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(2706), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44122] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2717), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44170] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2719), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1158), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44218] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2721), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1151), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44266] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2723), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44314] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2725), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1153), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44362] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2727), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44410] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2727), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1155), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44458] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44506] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1148), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44554] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2733), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44602] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2735), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44650] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2737), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1162), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44698] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2739), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1161), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44746] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2741), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44794] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2743), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44842] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2745), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44890] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2745), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1165), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44938] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2747), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [44986] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2717), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1169), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45034] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2743), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1168), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45082] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2749), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45130] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2751), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1147), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45178] = 13,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2692), 1,
      sym_simple_name,
    ACTIONS(2694), 1,
      anon_sym_LBRACK,
    ACTIONS(2753), 1,
      anon_sym_RBRACE,
    STATE(1227), 1,
      sym_attribute,
    STATE(1247), 1,
      sym_type_literal,
    STATE(1570), 1,
      sym_variable,
    STATE(1871), 1,
      sym_class_property_definition,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1157), 2,
      sym_class_method_definition,
      aux_sym_class_statement_repeat2,
    STATE(1226), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45226] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2759), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2761), 1,
      anon_sym_DQUOTE,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1327), 1,
      sym__expandable_string_literal_immediate,
    STATE(1453), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1178), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2757), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45265] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2767), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2769), 1,
      anon_sym_DQUOTE,
    STATE(966), 1,
      sym__expandable_string_literal_immediate,
    STATE(1496), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1182), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2765), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45304] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2773), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2775), 1,
      anon_sym_DQUOTE,
    STATE(944), 1,
      sym__expandable_string_literal_immediate,
    STATE(1450), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1188), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2771), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45343] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2779), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2781), 1,
      anon_sym_DQUOTE,
    STATE(928), 1,
      sym__expandable_string_literal_immediate,
    STATE(1515), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1180), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2777), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45382] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2785), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2787), 1,
      anon_sym_DQUOTE,
    STATE(930), 1,
      sym__expandable_string_literal_immediate,
    STATE(1456), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1184), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2783), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45421] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2789), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2795), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1503), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2791), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1181), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45459] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2805), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1484), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45497] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2809), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2811), 1,
      anon_sym_DQUOTE,
    STATE(1499), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2807), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45533] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2813), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2817), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1429), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2815), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1195), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45571] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2819), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2821), 1,
      anon_sym_DQUOTE,
    STATE(1462), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2807), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45607] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2823), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1514), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45645] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2825), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2827), 1,
      anon_sym_DQUOTE,
    STATE(1508), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2807), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45681] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2829), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2833), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1472), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2831), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1187), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45719] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2835), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2837), 1,
      anon_sym_DQUOTE,
    STATE(1451), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2807), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45755] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2839), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2843), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1511), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2841), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1186), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45793] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2845), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1437), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45831] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2847), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1426), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45869] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2763), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2849), 1,
      anon_sym_DOLLAR2,
    ACTIONS(2851), 1,
      anon_sym_DQUOTE,
    STATE(1457), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2807), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2755), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45905] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2853), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2857), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1474), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2855), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1190), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45943] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2859), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1481), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [45981] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2861), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2865), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1510), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2863), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1192), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46019] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2867), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1466), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46057] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2869), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2873), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1427), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2871), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1194), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46095] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2875), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1444), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46133] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2801), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2877), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1425), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2803), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46171] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      anon_sym_DOLLAR,
    ACTIONS(2799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2879), 1,
      aux_sym_expandable_string_literal_token2,
    ACTIONS(2883), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1454), 1,
      aux_sym_expandable_string_literal_repeat2,
    ACTIONS(2881), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1177), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2797), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46209] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2889), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2887), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1203), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46242] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2899), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46275] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2905), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2901), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2903), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1198), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46308] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2915), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2913), 2,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
    STATE(1200), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym__expandable_string_literal_immediate_repeat1,
    ACTIONS(2910), 4,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
    ACTIONS(2907), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46339] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2918), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46372] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2920), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46405] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2922), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46438] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2928), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2924), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2926), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1202), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46471] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2930), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46504] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2936), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2932), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2934), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1208), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46537] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2942), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2938), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2940), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1209), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46570] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2944), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46603] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2946), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46636] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(2959), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2948), 2,
      aux_sym_expandable_string_literal_token2,
      anon_sym_DOLLAR,
    ACTIONS(2951), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    STATE(1210), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_string_literal_repeat1,
    ACTIONS(2956), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46669] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2968), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2973), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2962), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2965), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2970), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46702] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2980), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2976), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2978), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1201), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46735] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2986), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2982), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2984), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1214), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46768] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2988), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2895), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2897), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1211), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46801] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2994), 1,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(2990), 2,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
    ACTIONS(2992), 3,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
    STATE(1205), 3,
      sym_variable,
      sym_sub_expression,
      aux_sym_expandable_here_string_literal_repeat1,
    ACTIONS(2891), 7,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46834] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(2998), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1713), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46874] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3000), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1841), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46914] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1877), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46954] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3004), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1926), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [46994] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3006), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1866), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47034] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3008), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1717), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47074] = 11,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    ACTIONS(3010), 1,
      anon_sym_RPAREN,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1431), 1,
      sym_script_parameter,
    STATE(1876), 1,
      sym_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47114] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(95), 2,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
    ACTIONS(2055), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(2053), 3,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      sym_path_command_name_token,
    ACTIONS(93), 4,
      anon_sym_LBRACK,
      anon_sym_PLUS_PLUS,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3012), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47144] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(588), 5,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      sym_path_command_name_token,
      anon_sym_DASH_DASH,
      anon_sym_DOT2,
    ACTIONS(586), 11,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_COLON_COLON,
      aux_sym_invokation_foreach_expression_token1,
  [47168] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3014), 1,
      sym_simple_name,
    ACTIONS(3016), 1,
      anon_sym_LBRACK,
    STATE(1275), 1,
      sym_type_literal,
    STATE(1547), 1,
      sym_variable,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1242), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47203] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3016), 1,
      anon_sym_LBRACK,
    ACTIONS(3018), 1,
      sym_simple_name,
    STATE(1276), 1,
      sym_type_literal,
    STATE(1603), 1,
      sym_variable,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1242), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47238] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3016), 1,
      anon_sym_LBRACK,
    ACTIONS(3018), 1,
      sym_simple_name,
    STATE(1276), 1,
      sym_type_literal,
    STATE(1603), 1,
      sym_variable,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2698), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1225), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47273] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3020), 1,
      anon_sym_SEMI,
    STATE(1228), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(1328), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2648), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47299] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    ACTIONS(644), 11,
      aux_sym_expandable_string_literal_token2,
      anon_sym_DOLLAR,
      aux_sym_expandable_string_literal_token5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47321] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3027), 2,
      anon_sym_DOLLAR,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3025), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    ACTIONS(3023), 9,
      aux_sym_expandable_string_literal_token2,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47345] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(588), 14,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR_LPAREN,
  [47365] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3032), 1,
      anon_sym_SEMI,
    STATE(1237), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(3030), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(2712), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47391] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 3,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(640), 11,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47413] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    STATE(1280), 1,
      sym_type_literal,
    STATE(1289), 1,
      sym_attribute_list,
    STATE(1400), 1,
      sym_variable,
    STATE(1672), 1,
      sym_script_parameter,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1252), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47447] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    ACTIONS(588), 11,
      aux_sym_expandable_string_literal_token2,
      anon_sym_DOLLAR,
      aux_sym_expandable_string_literal_token5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47469] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(644), 14,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR_LPAREN,
  [47489] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3038), 1,
      anon_sym_SEMI,
    STATE(1228), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(3034), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3036), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47515] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(638), 3,
      aux_sym_expandable_string_literal_token3,
      aux_sym_expandable_string_literal_token4,
      anon_sym_DQUOTE_DQUOTE,
    ACTIONS(640), 11,
      aux_sym_expandable_string_literal_token2,
      anon_sym_DOLLAR,
      aux_sym_expandable_string_literal_token5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47537] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(642), 3,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(644), 11,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47559] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(640), 14,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym__expandable_string_literal_immediate_token1,
      aux_sym__expandable_string_literal_immediate_token2,
      aux_sym__expandable_string_literal_immediate_token3,
      anon_sym_DQUOTE_DQUOTE2,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR_LPAREN,
  [47579] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(586), 3,
      aux_sym_expandable_string_literal_token4,
      aux_sym_expandable_here_string_literal_token4,
      aux_sym_expandable_here_string_literal_token5,
    ACTIONS(588), 11,
      anon_sym_DOLLAR,
      aux_sym_expandable_here_string_literal_token2,
      aux_sym_expandable_here_string_literal_token3,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_DOLLAR_LPAREN,
  [47601] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    STATE(1242), 2,
      sym_class_attribute,
      aux_sym_class_property_definition_repeat1,
    ACTIONS(3040), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3042), 6,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47626] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3047), 1,
      anon_sym_RPAREN,
    STATE(1296), 1,
      sym_type_literal,
    STATE(1505), 1,
      sym_class_method_parameter,
    STATE(1534), 1,
      sym_variable,
    STATE(1795), 1,
      sym_class_method_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47659] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3049), 1,
      anon_sym_RPAREN,
    STATE(1296), 1,
      sym_type_literal,
    STATE(1505), 1,
      sym_class_method_parameter,
    STATE(1534), 1,
      sym_variable,
    STATE(1922), 1,
      sym_class_method_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47692] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3051), 1,
      anon_sym_RPAREN,
    STATE(1296), 1,
      sym_type_literal,
    STATE(1505), 1,
      sym_class_method_parameter,
    STATE(1534), 1,
      sym_variable,
    STATE(2007), 1,
      sym_class_method_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47725] = 9,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3053), 1,
      anon_sym_RPAREN,
    STATE(1296), 1,
      sym_type_literal,
    STATE(1505), 1,
      sym_class_method_parameter,
    STATE(1534), 1,
      sym_variable,
    STATE(1770), 1,
      sym_class_method_parameter_list,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47758] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3055), 1,
      sym_simple_name,
    STATE(1603), 1,
      sym_variable,
    ACTIONS(3061), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(101), 3,
      anon_sym_LBRACK,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
    ACTIONS(3058), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47784] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3068), 1,
      sym_path_command_name_token,
    ACTIONS(2116), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(3066), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1250), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(3064), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47810] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3070), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3072), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47830] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3080), 1,
      sym_path_command_name_token,
    ACTIONS(2138), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(3077), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1250), 2,
      sym_variable,
      aux_sym_path_command_name_repeat1,
    ACTIONS(3074), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [47856] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3083), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3085), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47876] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2996), 1,
      anon_sym_LBRACK,
    STATE(1280), 1,
      sym_type_literal,
    ACTIONS(3089), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1253), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(3087), 6,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_param_block_token1,
  [47902] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3091), 1,
      anon_sym_LBRACK,
    STATE(1280), 1,
      sym_type_literal,
    ACTIONS(3096), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    STATE(1253), 2,
      sym_attribute,
      aux_sym_attribute_list_repeat1,
    ACTIONS(3094), 6,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_param_block_token1,
  [47928] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3098), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3100), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47948] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3102), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3104), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [47968] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1363), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1379), 11,
      anon_sym_RPAREN,
      aux_sym_block_name_token1,
      aux_sym_block_name_token2,
      aux_sym_block_name_token3,
      aux_sym_block_name_token4,
      anon_sym_RBRACE,
      aux_sym_elseif_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [47988] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3106), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3108), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [48008] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3110), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3112), 9,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_RBRACE,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [48028] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1367), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1377), 11,
      anon_sym_RPAREN,
      aux_sym_block_name_token1,
      aux_sym_block_name_token2,
      aux_sym_block_name_token3,
      aux_sym_block_name_token4,
      anon_sym_RBRACE,
      aux_sym_elseif_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [48048] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(512), 1,
      aux_sym_for_statement_token1,
    ACTIONS(3114), 1,
      aux_sym_switch_statement_token1,
    ACTIONS(3116), 1,
      aux_sym_foreach_statement_token1,
    ACTIONS(3118), 1,
      aux_sym_while_statement_token1,
    ACTIONS(3120), 1,
      aux_sym_do_statement_token1,
    STATE(1814), 6,
      sym__labeled_statement,
      sym_switch_statement,
      sym_foreach_statement,
      sym_for_statement,
      sym_while_statement,
      sym_do_statement,
  [48075] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3122), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3124), 8,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [48094] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(478), 1,
      aux_sym_for_statement_token1,
    ACTIONS(3126), 1,
      aux_sym_switch_statement_token1,
    ACTIONS(3128), 1,
      aux_sym_foreach_statement_token1,
    ACTIONS(3130), 1,
      aux_sym_while_statement_token1,
    ACTIONS(3132), 1,
      aux_sym_do_statement_token1,
    STATE(1769), 6,
      sym__labeled_statement,
      sym_switch_statement,
      sym_foreach_statement,
      sym_for_statement,
      sym_while_statement,
      sym_do_statement,
  [48121] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1460), 1,
      sym_switch_parameters,
    STATE(1685), 1,
      sym_switch_condition,
    STATE(1274), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3136), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48148] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3140), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3142), 8,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [48167] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(185), 1,
      aux_sym_for_statement_token1,
    ACTIONS(3144), 1,
      aux_sym_switch_statement_token1,
    ACTIONS(3146), 1,
      aux_sym_foreach_statement_token1,
    ACTIONS(3148), 1,
      aux_sym_while_statement_token1,
    ACTIONS(3150), 1,
      aux_sym_do_statement_token1,
    STATE(361), 6,
      sym__labeled_statement,
      sym_switch_statement,
      sym_foreach_statement,
      sym_for_statement,
      sym_while_statement,
      sym_do_statement,
  [48194] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3152), 3,
      sym_simple_name,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3154), 8,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_class_attribute_token1,
      aux_sym_class_attribute_token2,
  [48213] = 7,
    ACTIONS(37), 1,
      aux_sym_for_statement_token1,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3156), 1,
      aux_sym_switch_statement_token1,
    ACTIONS(3158), 1,
      aux_sym_foreach_statement_token1,
    ACTIONS(3160), 1,
      aux_sym_while_statement_token1,
    ACTIONS(3162), 1,
      aux_sym_do_statement_token1,
    STATE(411), 6,
      sym__labeled_statement,
      sym_switch_statement,
      sym_foreach_statement,
      sym_for_statement,
      sym_while_statement,
      sym_do_statement,
  [48240] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    STATE(1296), 1,
      sym_type_literal,
    STATE(1534), 1,
      sym_variable,
    STATE(1647), 1,
      sym_class_method_parameter,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48267] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1459), 1,
      sym_switch_parameters,
    STATE(1533), 1,
      sym_switch_condition,
    STATE(1274), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3136), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48294] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1432), 1,
      sym_switch_parameters,
    STATE(1627), 1,
      sym_switch_condition,
    STATE(1274), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3136), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48321] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1487), 1,
      sym_switch_parameters,
    STATE(1578), 1,
      sym_switch_condition,
    STATE(1274), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3136), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48348] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(588), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(586), 8,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      sym_path_command_name_token,
  [48366] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3164), 1,
      anon_sym_DOT2,
    ACTIONS(3166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(654), 4,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [48394] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3168), 2,
      anon_sym_LPAREN,
      aux_sym_switch_condition_token1,
    STATE(1279), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3136), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48413] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3170), 1,
      sym_simple_name,
    STATE(1631), 1,
      sym_variable,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48434] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3172), 1,
      sym_simple_name,
    STATE(1547), 1,
      sym_variable,
    ACTIONS(21), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(99), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48455] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3122), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3124), 7,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_param_block_token1,
  [48472] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3152), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(3154), 7,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_param_block_token1,
  [48489] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3174), 2,
      anon_sym_LPAREN,
      aux_sym_switch_condition_token1,
    STATE(1279), 2,
      sym_switch_parameter,
      aux_sym_switch_parameters_repeat1,
    ACTIONS(3176), 5,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
  [48508] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3179), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(101), 7,
      anon_sym_LBRACK,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
      aux_sym_param_block_token1,
  [48525] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2109), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48543] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2103), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48561] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1283), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3181), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
    ACTIONS(3184), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
  [48579] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3190), 1,
      sym__statement_terminator,
    STATE(1298), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3186), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
    ACTIONS(3188), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
  [48599] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3192), 1,
      anon_sym_DOT2,
    ACTIONS(3194), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2408), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [48625] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3196), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2420), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [48641] = 3,
    ACTIONS(79), 1,
      sym_comment,
    STATE(80), 1,
      sym_assignement_operator,
    ACTIONS(3198), 7,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
  [48657] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3203), 1,
      anon_sym_RBRACE,
    STATE(1582), 1,
      sym_block_name,
    STATE(1288), 2,
      sym_named_block,
      aux_sym_named_block_list_repeat1,
    ACTIONS(3200), 4,
      aux_sym_block_name_token1,
      aux_sym_block_name_token2,
      aux_sym_block_name_token3,
      aux_sym_block_name_token4,
  [48677] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(1391), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48695] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1283), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3205), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
    ACTIONS(3188), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
  [48713] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2013), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48731] = 3,
    ACTIONS(79), 1,
      sym_comment,
    STATE(79), 1,
      sym_assignement_operator,
    ACTIONS(3198), 7,
      anon_sym_EQ,
      anon_sym_BANG_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_DASH_EQ,
  [48747] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3196), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2394), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [48763] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(1775), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48781] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2566), 1,
      anon_sym_LBRACK,
    ACTIONS(2568), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2570), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2576), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3207), 1,
      anon_sym_DOT2,
    ACTIONS(3209), 1,
      anon_sym_COLON_COLON,
    ACTIONS(654), 2,
      sym__statement_terminator,
      anon_sym_LPAREN,
  [48807] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(1624), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48825] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_RBRACE,
    STATE(1582), 1,
      sym_block_name,
    STATE(1288), 2,
      sym_named_block,
      aux_sym_named_block_list_repeat1,
    ACTIONS(3211), 4,
      aux_sym_block_name_token1,
      aux_sym_block_name_token2,
      aux_sym_block_name_token3,
      aux_sym_block_name_token4,
  [48845] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3218), 1,
      sym__statement_terminator,
    STATE(1298), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3184), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE,
    ACTIONS(3215), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
  [48865] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2096), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48883] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2097), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48901] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2102), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48919] = 4,
    ACTIONS(79), 1,
      sym_comment,
    STATE(2108), 1,
      sym_variable,
    ACTIONS(135), 2,
      anon_sym_DOLLAR_,
      aux_sym_variable_token1,
    ACTIONS(133), 5,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_DOLLAR_CARET,
      anon_sym_DOLLAR_QMARK,
      aux_sym_variable_token2,
      sym_braced_variable,
  [48937] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3222), 1,
      aux_sym_command_name_token19,
    ACTIONS(3224), 1,
      anon_sym_DQUOTE2,
    STATE(1317), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2448), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(3220), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
  [48958] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    ACTIONS(3226), 1,
      anon_sym_DASH_DASH_PERCENT,
    STATE(1367), 1,
      aux_sym__pipeline_tail,
    STATE(1368), 1,
      sym_verbatim_command_argument,
    ACTIONS(2390), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [48979] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(654), 1,
      sym__statement_terminator,
    ACTIONS(2566), 1,
      anon_sym_LBRACK,
    ACTIONS(2568), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2570), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2576), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3228), 1,
      anon_sym_DOT2,
    ACTIONS(3230), 1,
      anon_sym_COLON_COLON,
  [49004] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3235), 1,
      aux_sym_command_name_token19,
    ACTIONS(3238), 1,
      anon_sym_DQUOTE2,
    STATE(1306), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2427), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(3232), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
  [49025] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3241), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2420), 4,
      sym__statement_terminator,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [49040] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1302), 1,
      anon_sym_RPAREN,
    ACTIONS(3243), 1,
      aux_sym_elseif_clause_token1,
    ACTIONS(3245), 1,
      aux_sym_else_clause_token1,
    STATE(1498), 1,
      sym_elseif_clauses,
    STATE(1915), 1,
      sym_else_clause,
    STATE(1360), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [49063] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3247), 7,
      anon_sym_LPAREN,
      aux_sym_switch_parameter_token1,
      aux_sym_switch_parameter_token2,
      aux_sym_switch_parameter_token3,
      aux_sym_switch_parameter_token4,
      aux_sym_switch_parameter_token5,
      aux_sym_switch_condition_token1,
  [49076] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1296), 1,
      anon_sym_RPAREN,
    ACTIONS(3249), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3251), 1,
      aux_sym_finally_clause_token1,
    STATE(1463), 1,
      sym_catch_clauses,
    STATE(1963), 1,
      sym_finally_clause,
    STATE(1380), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [49099] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3241), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2394), 4,
      sym__statement_terminator,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [49114] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3192), 1,
      anon_sym_DOT2,
    ACTIONS(3194), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3253), 1,
      anon_sym_LBRACE,
  [49139] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1296), 1,
      sym__statement_terminator,
    ACTIONS(3255), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3257), 1,
      aux_sym_finally_clause_token1,
    STATE(1449), 1,
      sym_catch_clauses,
    STATE(1921), 1,
      sym_finally_clause,
    STATE(1372), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [49162] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(654), 1,
      anon_sym_EQ,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3192), 1,
      anon_sym_DOT2,
    ACTIONS(3194), 1,
      anon_sym_COLON_COLON,
  [49187] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3259), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2420), 4,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_RBRACK,
  [49202] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1302), 1,
      sym__statement_terminator,
    ACTIONS(3261), 1,
      aux_sym_elseif_clause_token1,
    ACTIONS(3263), 1,
      aux_sym_else_clause_token1,
    STATE(1506), 1,
      sym_elseif_clauses,
    STATE(2027), 1,
      sym_else_clause,
    STATE(1386), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [49225] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3224), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3267), 1,
      aux_sym_command_name_token19,
    STATE(1306), 1,
      aux_sym_command_name_repeat1,
    ACTIONS(2456), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(3265), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
  [49246] = 8,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(2582), 1,
      anon_sym_DASH_DASH,
    ACTIONS(2588), 1,
      aux_sym_invokation_foreach_expression_token1,
    ACTIONS(3192), 1,
      anon_sym_DOT2,
    ACTIONS(3194), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3269), 1,
      anon_sym_LBRACE,
  [49271] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    ACTIONS(3271), 1,
      anon_sym_DASH_DASH_PERCENT,
    STATE(1338), 1,
      sym_verbatim_command_argument,
    STATE(1341), 1,
      aux_sym__pipeline_tail,
    ACTIONS(2390), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49292] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3259), 3,
      aux_sym_bitwise_expression_token1,
      aux_sym_bitwise_expression_token2,
      aux_sym_bitwise_expression_token3,
    ACTIONS(2394), 4,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
      anon_sym_RBRACK,
  [49307] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3273), 1,
      sym_simple_name,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    STATE(414), 1,
      sym_statement_block,
    STATE(1409), 1,
      sym_data_name,
    STATE(1637), 1,
      sym_data_commands_allowed,
  [49329] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3273), 1,
      sym_simple_name,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1408), 1,
      sym_data_name,
    STATE(1536), 1,
      sym_data_commands_allowed,
    STATE(1774), 1,
      sym_statement_block,
  [49351] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2520), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
  [49365] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1486), 1,
      sym_type_spec,
    STATE(1767), 1,
      sym_generic_type_arguments,
  [49387] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1532), 1,
      sym_type_spec,
    STATE(2041), 1,
      sym_attribute_name,
  [49409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2514), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2516), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
  [49423] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2435), 2,
      anon_sym_DQUOTE_DQUOTE,
      anon_sym_SQUOTE_SQUOTE,
    ACTIONS(2427), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_command_name_token19,
      anon_sym_DQUOTE2,
  [49437] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3273), 1,
      sym_simple_name,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1418), 1,
      sym_data_name,
    STATE(1580), 1,
      sym_data_commands_allowed,
    STATE(1897), 1,
      sym_statement_block,
  [49459] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1625), 1,
      sym_type_spec,
    STATE(1891), 1,
      sym_attribute_name,
  [49481] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3273), 1,
      sym_simple_name,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(367), 1,
      sym_statement_block,
    STATE(1416), 1,
      sym_data_name,
    STATE(1630), 1,
      sym_data_commands_allowed,
  [49503] = 7,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1634), 1,
      sym_type_spec,
    STATE(1891), 1,
      sym_attribute_name,
  [49525] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3287), 1,
      anon_sym_RPAREN,
    STATE(291), 1,
      sym_pipeline_chain_tail,
    STATE(1366), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3289), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49542] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1422), 1,
      sym_type_literal,
    STATE(1477), 1,
      sym_statement_block,
    STATE(1592), 1,
      sym_catch_type_list,
  [49561] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3291), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [49572] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3293), 1,
      anon_sym_LBRACE,
    STATE(319), 1,
      sym_statement_block,
    STATE(1422), 1,
      sym_type_literal,
    STATE(1566), 1,
      sym_catch_type_list,
  [49591] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3295), 1,
      anon_sym_LBRACE,
    STATE(317), 1,
      sym_statement_block,
    STATE(1422), 1,
      sym_type_literal,
    STATE(1653), 1,
      sym_catch_type_list,
  [49610] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3297), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3299), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [49623] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    STATE(1375), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3301), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49638] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3303), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [49649] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(2057), 1,
      sym_type_spec,
  [49668] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    STATE(1374), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3301), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49683] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3305), 1,
      sym__statement_terminator,
    STATE(292), 1,
      sym_pipeline_chain_tail,
    STATE(1385), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3289), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49700] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1773), 1,
      sym_type_spec,
  [49719] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3307), 5,
      anon_sym_DOT,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [49730] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1367), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1377), 4,
      sym__statement_terminator,
      aux_sym_elseif_clause_token1,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [49743] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1875), 1,
      sym_type_spec,
  [49762] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3312), 1,
      sym__statement_terminator,
    STATE(292), 1,
      sym_pipeline_chain_tail,
    STATE(1347), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3309), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49779] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3305), 1,
      anon_sym_RPAREN,
    STATE(291), 1,
      sym_pipeline_chain_tail,
    STATE(1332), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3289), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [49796] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1798), 1,
      sym_type_spec,
  [49815] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1645), 1,
      sym_type_spec,
  [49834] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3316), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 4,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_DASH_DASH_PERCENT,
  [49847] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3318), 5,
      aux_sym_block_name_token1,
      aux_sym_block_name_token2,
      aux_sym_block_name_token3,
      aux_sym_block_name_token4,
      anon_sym_RBRACE,
  [49858] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3322), 1,
      anon_sym_PIPE,
    ACTIONS(3320), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_DASH_DASH_PERCENT,
  [49871] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1994), 1,
      sym_type_spec,
  [49890] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3322), 1,
      anon_sym_PIPE,
    ACTIONS(3320), 4,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_DASH_DASH_PERCENT,
  [49903] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1723), 1,
      sym_type_spec,
  [49922] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1720), 1,
      sym_type_spec,
  [49941] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1843), 1,
      sym_type_spec,
  [49960] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1950), 1,
      sym_type_spec,
  [49979] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1313), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1348), 1,
      anon_sym_RPAREN,
    ACTIONS(3243), 1,
      aux_sym_elseif_clause_token1,
    STATE(1371), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [49996] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1734), 1,
      sym_type_spec,
  [50015] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1737), 1,
      sym_type_spec,
  [50034] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1740), 1,
      sym_type_spec,
  [50053] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1743), 1,
      sym_type_spec,
  [50072] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1746), 1,
      sym_type_spec,
  [50091] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3312), 1,
      anon_sym_RPAREN,
    STATE(291), 1,
      sym_pipeline_chain_tail,
    STATE(1366), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3309), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50108] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    STATE(1377), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3301), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50123] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    STATE(1378), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3301), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50138] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1903), 1,
      sym_type_spec,
  [50157] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3324), 5,
      anon_sym_DOT,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [50168] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1323), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1350), 1,
      anon_sym_RPAREN,
    ACTIONS(3326), 1,
      aux_sym_elseif_clause_token1,
    STATE(1371), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [50185] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3255), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(1355), 2,
      sym__statement_terminator,
      aux_sym_finally_clause_token1,
    STATE(1384), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [50200] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1363), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1379), 4,
      sym__statement_terminator,
      aux_sym_elseif_clause_token1,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [50213] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3331), 1,
      anon_sym_PIPE,
    STATE(1374), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3329), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50228] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2402), 1,
      anon_sym_PIPE,
    STATE(1374), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3334), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50243] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3316), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 4,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_DASH_DASH_PERCENT,
  [50256] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3336), 1,
      anon_sym_PIPE,
    STATE(1377), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3329), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50271] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2392), 1,
      anon_sym_PIPE,
    STATE(1377), 1,
      aux_sym__pipeline_tail,
    ACTIONS(3334), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50286] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2526), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3339), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [50299] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3249), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(1355), 2,
      anon_sym_RPAREN,
      aux_sym_finally_clause_token1,
    STATE(1383), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [50314] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1889), 1,
      sym_type_spec,
  [50333] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3341), 1,
      anon_sym_DOT,
    ACTIONS(3343), 1,
      anon_sym_LBRACK,
    ACTIONS(3345), 3,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [50348] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3347), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(1343), 2,
      anon_sym_RPAREN,
      aux_sym_finally_clause_token1,
    STATE(1383), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [50363] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3350), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(1343), 2,
      sym__statement_terminator,
      aux_sym_finally_clause_token1,
    STATE(1384), 2,
      sym_catch_clause,
      aux_sym_catch_clauses_repeat1,
  [50378] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3287), 1,
      sym__statement_terminator,
    STATE(292), 1,
      sym_pipeline_chain_tail,
    STATE(1347), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(3289), 2,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50395] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1313), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1348), 1,
      sym__statement_terminator,
    ACTIONS(3261), 1,
      aux_sym_elseif_clause_token1,
    STATE(1387), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [50412] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1323), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1350), 1,
      sym__statement_terminator,
    ACTIONS(3353), 1,
      aux_sym_elseif_clause_token1,
    STATE(1387), 2,
      sym_elseif_clause,
      aux_sym_elseif_clauses_repeat1,
  [50429] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1836), 1,
      sym_type_spec,
  [50448] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1422), 1,
      sym_type_literal,
    STATE(1493), 1,
      sym_statement_block,
    STATE(1546), 1,
      sym_catch_type_list,
  [50467] = 6,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3281), 1,
      sym_type_identifier,
    STATE(1324), 1,
      sym_generic_type_name,
    STATE(1382), 1,
      sym_type_name,
    STATE(1413), 1,
      sym_array_type_name,
    STATE(1705), 1,
      sym_type_spec,
  [50486] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3356), 1,
      anon_sym_EQ,
    STATE(1683), 1,
      sym_script_parameter_default,
    ACTIONS(3358), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [50500] = 3,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1290), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3360), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
  [50512] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3362), 1,
      anon_sym_PIPE,
    ACTIONS(3329), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50524] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3366), 1,
      anon_sym_RBRACE,
    STATE(1407), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50540] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3368), 1,
      anon_sym_RBRACE,
    STATE(1410), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50556] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3370), 1,
      anon_sym_RBRACE,
    STATE(1401), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50572] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(412), 1,
      sym_statement_block,
    STATE(1615), 1,
      sym_type_literal,
  [50588] = 3,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1284), 1,
      aux_sym__verbatim_command_argument_chars,
    ACTIONS(3372), 3,
      aux_sym__verbatim_command_argument_chars_token1,
      aux_sym__verbatim_command_argument_chars_token2,
      aux_sym__verbatim_command_argument_chars_token3,
  [50600] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2526), 1,
      sym__statement_terminator,
    ACTIONS(3374), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [50612] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3356), 1,
      anon_sym_EQ,
    STATE(1687), 1,
      sym_script_parameter_default,
    ACTIONS(3376), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [50626] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3378), 1,
      sym_simple_name,
    ACTIONS(3381), 1,
      anon_sym_RBRACE,
    STATE(1401), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50642] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3383), 1,
      anon_sym_RBRACE,
    STATE(1414), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50658] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3362), 1,
      anon_sym_PIPE,
    ACTIONS(3329), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [50670] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2646), 1,
      anon_sym_SEMI,
    STATE(1120), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(3385), 2,
      sym_simple_name,
      anon_sym_RBRACE,
  [50684] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2526), 1,
      anon_sym_RBRACK,
    ACTIONS(3387), 3,
      aux_sym_logical_expression_token1,
      aux_sym_logical_expression_token2,
      aux_sym_logical_expression_token3,
  [50696] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1535), 1,
      sym_type_literal,
    STATE(1772), 1,
      sym_statement_block,
  [50712] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3389), 1,
      anon_sym_RBRACE,
    STATE(1401), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50728] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1541), 1,
      sym_data_commands_allowed,
    STATE(1923), 1,
      sym_statement_block,
  [50744] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    STATE(422), 1,
      sym_statement_block,
    STATE(1571), 1,
      sym_data_commands_allowed,
  [50760] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3391), 1,
      anon_sym_RBRACE,
    STATE(1401), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50776] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3393), 1,
      anon_sym_SEMI,
    STATE(1404), 1,
      aux_sym_script_block_repeat1,
    ACTIONS(3381), 2,
      sym_simple_name,
      anon_sym_RBRACE,
  [50790] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1579), 1,
      sym_type_literal,
    STATE(1825), 1,
      sym_statement_block,
  [50806] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3395), 1,
      anon_sym_COMMA,
    ACTIONS(3397), 1,
      anon_sym_RBRACK,
    STATE(1478), 1,
      aux_sym_dimension_repeat1,
    STATE(1767), 1,
      sym_dimension,
  [50822] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3399), 1,
      anon_sym_RBRACE,
    STATE(1401), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50838] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(362), 1,
      sym_statement_block,
    STATE(1629), 1,
      sym_type_literal,
  [50854] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(348), 1,
      sym_statement_block,
    STATE(1644), 1,
      sym_data_commands_allowed,
  [50870] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3364), 1,
      sym_simple_name,
    ACTIONS(3401), 1,
      anon_sym_RBRACE,
    STATE(1396), 1,
      aux_sym_enum_statement_repeat1,
    STATE(1938), 1,
      sym_enum_member,
  [50886] = 5,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3277), 1,
      aux_sym_data_commands_allowed_token1,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1588), 1,
      sym_data_commands_allowed,
    STATE(1703), 1,
      sym_statement_block,
  [50902] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3403), 1,
      anon_sym_RPAREN,
    ACTIONS(3405), 1,
      anon_sym_COMMA,
    STATE(1458), 1,
      aux_sym_attribute_arguments_repeat1,
  [50915] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR,
    ACTIONS(3410), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [50928] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3412), 1,
      anon_sym_SEMI,
    ACTIONS(3414), 1,
      anon_sym_RPAREN,
    ACTIONS(3416), 1,
      aux_sym_for_statement_token2,
  [50941] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3418), 1,
      anon_sym_COMMA,
    ACTIONS(3420), 1,
      anon_sym_LBRACE,
    STATE(1447), 1,
      aux_sym_catch_type_list_repeat1,
  [50954] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3422), 3,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [50963] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3426), 1,
      anon_sym_LBRACE,
    STATE(1435), 1,
      aux_sym_class_statement_repeat1,
  [50976] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3430), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [50989] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3432), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51002] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2875), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51015] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3434), 1,
      anon_sym_COMMA,
    ACTIONS(3437), 1,
      anon_sym_RBRACK,
    STATE(1428), 1,
      aux_sym_generic_type_arguments_repeat1,
  [51028] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2877), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51041] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_RPAREN,
    ACTIONS(3441), 1,
      anon_sym_COMMA,
    STATE(1469), 1,
      aux_sym_argument_expression_list_repeat1,
  [51054] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3443), 1,
      anon_sym_RPAREN,
    ACTIONS(3445), 1,
      anon_sym_COMMA,
    STATE(1500), 1,
      aux_sym_parameter_list_repeat1,
  [51067] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1641), 1,
      sym_switch_condition,
  [51080] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3447), 1,
      anon_sym_RPAREN,
    ACTIONS(3449), 1,
      anon_sym_COMMA,
    STATE(1433), 1,
      aux_sym_class_method_parameter_list_repeat1,
  [51093] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3452), 1,
      anon_sym_LPAREN,
    ACTIONS(3454), 1,
      anon_sym_LBRACE,
    STATE(1895), 1,
      sym_function_parameter_declaration,
  [51106] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3456), 1,
      anon_sym_COMMA,
    ACTIONS(3459), 1,
      anon_sym_LBRACE,
    STATE(1435), 1,
      aux_sym_class_statement_repeat1,
  [51119] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3405), 1,
      anon_sym_COMMA,
    ACTIONS(3461), 1,
      anon_sym_RPAREN,
    STATE(1419), 1,
      aux_sym_attribute_arguments_repeat1,
  [51132] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3463), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51145] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3465), 1,
      anon_sym_COMMA,
    ACTIONS(3467), 1,
      anon_sym_LBRACE,
    STATE(1446), 1,
      aux_sym_data_commands_list_repeat1,
  [51158] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3469), 1,
      anon_sym_SEMI,
    ACTIONS(3471), 1,
      anon_sym_RPAREN,
    ACTIONS(3473), 1,
      aux_sym_for_statement_token2,
  [51171] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_COMMA,
    ACTIONS(3477), 1,
      anon_sym_RBRACK,
    STATE(1428), 1,
      aux_sym_generic_type_arguments_repeat1,
  [51184] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3479), 1,
      anon_sym_LBRACE,
    STATE(1443), 1,
      aux_sym_class_statement_repeat1,
  [51197] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3481), 1,
      anon_sym_SEMI,
    ACTIONS(3483), 1,
      anon_sym_RPAREN,
    ACTIONS(3485), 1,
      aux_sym_for_statement_token2,
  [51210] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3487), 1,
      anon_sym_LBRACE,
    STATE(1435), 1,
      aux_sym_class_statement_repeat1,
  [51223] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3489), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51236] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3491), 1,
      anon_sym_SEMI,
    ACTIONS(3493), 1,
      anon_sym_RPAREN,
    ACTIONS(3495), 1,
      aux_sym_for_statement_token2,
  [51249] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3497), 1,
      anon_sym_COMMA,
    ACTIONS(3500), 1,
      anon_sym_LBRACE,
    STATE(1446), 1,
      aux_sym_data_commands_list_repeat1,
  [51262] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3418), 1,
      anon_sym_COMMA,
    ACTIONS(3502), 1,
      anon_sym_LBRACE,
    STATE(1497), 1,
      aux_sym_catch_type_list_repeat1,
  [51275] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3504), 3,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [51284] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1381), 1,
      sym__statement_terminator,
    ACTIONS(3257), 1,
      aux_sym_finally_clause_token1,
    STATE(1998), 1,
      sym_finally_clause,
  [51297] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3508), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51310] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3510), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51323] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3512), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3515), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51336] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3517), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51349] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2805), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51362] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3312), 3,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [51371] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3519), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51384] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3521), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51397] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3523), 1,
      anon_sym_RPAREN,
    ACTIONS(3525), 1,
      anon_sym_COMMA,
    STATE(1458), 1,
      aux_sym_attribute_arguments_repeat1,
  [51410] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1538), 1,
      sym_switch_condition,
  [51423] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1661), 1,
      sym_switch_condition,
  [51436] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3452), 1,
      anon_sym_LPAREN,
    ACTIONS(3528), 1,
      anon_sym_LBRACE,
    STATE(1943), 1,
      sym_function_parameter_declaration,
  [51449] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3530), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51462] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1381), 1,
      anon_sym_RPAREN,
    ACTIONS(3251), 1,
      aux_sym_finally_clause_token1,
    STATE(1970), 1,
      sym_finally_clause,
  [51475] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3452), 1,
      anon_sym_LPAREN,
    ACTIONS(3532), 1,
      anon_sym_LBRACE,
    STATE(1707), 1,
      sym_function_parameter_declaration,
  [51488] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3534), 1,
      anon_sym_SEMI,
    ACTIONS(3536), 1,
      anon_sym_RPAREN,
    ACTIONS(3538), 1,
      aux_sym_for_statement_token2,
  [51501] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3540), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51514] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3542), 1,
      anon_sym_LBRACE,
    STATE(1470), 1,
      aux_sym_class_statement_repeat1,
  [51527] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3544), 1,
      anon_sym_SEMI,
    ACTIONS(3546), 1,
      anon_sym_RPAREN,
    ACTIONS(3548), 1,
      aux_sym_for_statement_token2,
  [51540] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3550), 1,
      anon_sym_RPAREN,
    ACTIONS(3552), 1,
      anon_sym_COMMA,
    STATE(1469), 1,
      aux_sym_argument_expression_list_repeat1,
  [51553] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3555), 1,
      anon_sym_LBRACE,
    STATE(1435), 1,
      aux_sym_class_statement_repeat1,
  [51566] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3441), 1,
      anon_sym_COMMA,
    ACTIONS(3557), 1,
      anon_sym_RPAREN,
    STATE(1430), 1,
      aux_sym_argument_expression_list_repeat1,
  [51579] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2847), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51592] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3559), 1,
      anon_sym_SEMI,
    ACTIONS(3561), 1,
      anon_sym_RPAREN,
    ACTIONS(3563), 1,
      aux_sym_for_statement_token2,
  [51605] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2859), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51618] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3565), 1,
      anon_sym_SEMI,
    ACTIONS(3567), 1,
      anon_sym_RPAREN,
    ACTIONS(3569), 1,
      aux_sym_for_statement_token2,
  [51631] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3312), 3,
      sym__statement_terminator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [51640] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1371), 3,
      anon_sym_RPAREN,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [51649] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3571), 1,
      anon_sym_COMMA,
    ACTIONS(3573), 1,
      anon_sym_RBRACK,
    STATE(1512), 1,
      aux_sym_dimension_repeat1,
  [51662] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3575), 1,
      anon_sym_LPAREN,
    ACTIONS(3577), 1,
      aux_sym_switch_parameter_token5,
    STATE(1796), 1,
      sym_foreach_parameter,
  [51675] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3579), 1,
      anon_sym_EQ,
    ACTIONS(3581), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [51686] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3583), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51699] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3465), 1,
      anon_sym_COMMA,
    ACTIONS(3585), 1,
      anon_sym_LBRACE,
    STATE(1438), 1,
      aux_sym_data_commands_list_repeat1,
  [51712] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1375), 3,
      anon_sym_RPAREN,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [51721] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3587), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51734] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3589), 1,
      anon_sym_LBRACE,
    STATE(1424), 1,
      aux_sym_class_statement_repeat1,
  [51747] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_COMMA,
    ACTIONS(3591), 1,
      anon_sym_RBRACK,
    STATE(1440), 1,
      aux_sym_generic_type_arguments_repeat1,
  [51760] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3134), 1,
      anon_sym_LPAREN,
    ACTIONS(3138), 1,
      aux_sym_switch_condition_token1,
    STATE(1583), 1,
      sym_switch_condition,
  [51773] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3452), 1,
      anon_sym_LPAREN,
    ACTIONS(3593), 1,
      anon_sym_LBRACE,
    STATE(1986), 1,
      sym_function_parameter_declaration,
  [51786] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3595), 1,
      anon_sym_RPAREN,
    ACTIONS(3597), 1,
      anon_sym_COMMA,
    STATE(1489), 1,
      aux_sym_parameter_list_repeat1,
  [51799] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3600), 1,
      anon_sym_SEMI,
    ACTIONS(3602), 1,
      anon_sym_RPAREN,
    ACTIONS(3604), 1,
      aux_sym_for_statement_token2,
  [51812] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3606), 1,
      anon_sym_LBRACE,
    STATE(1494), 1,
      aux_sym_class_statement_repeat1,
  [51825] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3608), 1,
      anon_sym_SEMI,
    ACTIONS(3610), 1,
      anon_sym_RPAREN,
    ACTIONS(3612), 1,
      aux_sym_for_statement_token2,
  [51838] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1371), 3,
      sym__statement_terminator,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [51847] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3424), 1,
      anon_sym_COMMA,
    ACTIONS(3614), 1,
      anon_sym_LBRACE,
    STATE(1435), 1,
      aux_sym_class_statement_repeat1,
  [51860] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_SEMI,
    ACTIONS(3618), 1,
      anon_sym_RPAREN,
    ACTIONS(3620), 1,
      aux_sym_for_statement_token2,
  [51873] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3622), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51886] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3624), 1,
      anon_sym_COMMA,
    ACTIONS(3627), 1,
      anon_sym_LBRACE,
    STATE(1497), 1,
      aux_sym_catch_type_list_repeat1,
  [51899] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1373), 1,
      anon_sym_RPAREN,
    ACTIONS(3629), 1,
      aux_sym_else_clause_token1,
    STATE(1999), 1,
      sym_else_clause,
  [51912] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3631), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [51925] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3445), 1,
      anon_sym_COMMA,
    ACTIONS(3633), 1,
      anon_sym_RPAREN,
    STATE(1489), 1,
      aux_sym_parameter_list_repeat1,
  [51938] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3635), 1,
      sym_type_identifier,
    ACTIONS(3637), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [51949] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3639), 1,
      anon_sym_SEMI,
    ACTIONS(3641), 1,
      anon_sym_RPAREN,
    ACTIONS(3643), 1,
      aux_sym_for_statement_token2,
  [51962] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2823), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [51975] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1375), 3,
      sym__statement_terminator,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
  [51984] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3645), 1,
      anon_sym_RPAREN,
    ACTIONS(3647), 1,
      anon_sym_COMMA,
    STATE(1507), 1,
      aux_sym_class_method_parameter_list_repeat1,
  [51997] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1373), 1,
      sym__statement_terminator,
    ACTIONS(3649), 1,
      aux_sym_else_clause_token1,
    STATE(2035), 1,
      sym_else_clause,
  [52010] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3647), 1,
      anon_sym_COMMA,
    ACTIONS(3651), 1,
      anon_sym_RPAREN,
    STATE(1433), 1,
      aux_sym_class_method_parameter_list_repeat1,
  [52023] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3653), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [52036] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1361), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1383), 2,
      sym__statement_terminator,
      aux_sym_elseif_clause_token1,
  [52047] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [52060] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2845), 1,
      aux_sym_expandable_string_literal_token5,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [52073] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3655), 1,
      anon_sym_COMMA,
    ACTIONS(3658), 1,
      anon_sym_RBRACK,
    STATE(1512), 1,
      aux_sym_dimension_repeat1,
  [52086] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1361), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1383), 2,
      anon_sym_RPAREN,
      aux_sym_elseif_clause_token1,
  [52097] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3428), 1,
      anon_sym_DOLLAR,
    ACTIONS(3660), 1,
      aux_sym_expandable_string_literal_token5,
    STATE(1420), 1,
      aux_sym_expandable_string_literal_repeat2,
  [52110] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3506), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3662), 1,
      anon_sym_DQUOTE,
    STATE(1452), 1,
      aux_sym__expandable_string_literal_immediate_repeat2,
  [52123] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3577), 1,
      aux_sym_switch_parameter_token5,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    STATE(2121), 1,
      sym_foreach_parameter,
  [52136] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3577), 1,
      aux_sym_switch_parameter_token5,
    ACTIONS(3666), 1,
      anon_sym_LPAREN,
    STATE(2123), 1,
      sym_foreach_parameter,
  [52149] = 4,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3577), 1,
      aux_sym_switch_parameter_token5,
    ACTIONS(3668), 1,
      anon_sym_LPAREN,
    STATE(2125), 1,
      sym_foreach_parameter,
  [52162] = 4,
    ACTIONS(5), 1,
      sym_decimal_integer_literal,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(81), 1,
      sym_hexadecimal_integer_literal,
    STATE(1822), 1,
      sym_integer_literal,
  [52175] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3670), 2,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
  [52183] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1882), 1,
      sym_statement_block,
  [52193] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(418), 1,
      sym_statement_block,
  [52203] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(378), 1,
      sym_statement_block,
  [52213] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(383), 1,
      sym_statement_block,
  [52223] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(410), 1,
      sym_statement_block,
  [52233] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1313), 1,
      sym_statement_block,
  [52243] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1781), 1,
      sym_statement_block,
  [52253] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1789), 1,
      sym_statement_block,
  [52263] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1794), 1,
      sym_statement_block,
  [52273] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(384), 1,
      sym_statement_block,
  [52283] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3627), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [52291] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3672), 1,
      anon_sym_LPAREN,
    ACTIONS(3674), 1,
      anon_sym_RBRACK,
  [52301] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3676), 1,
      anon_sym_LBRACE,
    STATE(1917), 1,
      sym_switch_body,
  [52311] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3678), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [52319] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1919), 1,
      sym_statement_block,
  [52329] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1923), 1,
      sym_statement_block,
  [52339] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1198), 1,
      anon_sym_LBRACE,
    STATE(1033), 1,
      sym_script_block_expression,
  [52349] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3676), 1,
      anon_sym_LBRACE,
    STATE(1978), 1,
      sym_switch_body,
  [52359] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1984), 1,
      sym_statement_block,
  [52369] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1997), 1,
      sym_statement_block,
  [52379] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2005), 1,
      sym_statement_block,
  [52389] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(389), 1,
      sym_statement_block,
  [52399] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1316), 1,
      sym_statement_block,
  [52409] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2017), 1,
      sym_statement_block,
  [52419] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2018), 1,
      sym_statement_block,
  [52429] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1504), 1,
      sym_statement_block,
  [52439] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3680), 1,
      anon_sym_EQ,
    ACTIONS(3682), 1,
      sym__statement_terminator,
  [52449] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2029), 1,
      sym_statement_block,
  [52459] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2030), 1,
      sym_statement_block,
  [52469] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2034), 1,
      sym_statement_block,
  [52479] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2043), 1,
      sym_statement_block,
  [52489] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2047), 1,
      sym_statement_block,
  [52499] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2049), 1,
      sym_statement_block,
  [52509] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2061), 1,
      sym_statement_block,
  [52519] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2062), 1,
      sym_statement_block,
  [52529] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2066), 1,
      sym_statement_block,
  [52539] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2068), 1,
      sym_statement_block,
  [52549] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3459), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [52557] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2079), 1,
      sym_statement_block,
  [52567] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(1509), 1,
      sym_statement_block,
  [52577] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2084), 1,
      sym_statement_block,
  [52587] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2086), 1,
      sym_statement_block,
  [52597] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2089), 1,
      sym_statement_block,
  [52607] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_LBRACE,
    STATE(2101), 1,
      sym_statement_block,
  [52617] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3500), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [52625] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3293), 1,
      anon_sym_LBRACE,
    STATE(321), 1,
      sym_statement_block,
  [52635] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(423), 1,
      sym_statement_block,
  [52645] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3550), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [52653] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3684), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [52661] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3686), 1,
      anon_sym_EQ,
    ACTIONS(3688), 1,
      sym__statement_terminator,
  [52671] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(425), 1,
      sym_statement_block,
  [52681] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(390), 1,
      sym_statement_block,
  [52691] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1310), 1,
      sym_statement_block,
  [52701] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1909), 1,
      sym_statement_block,
  [52711] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(2107), 1,
      sym_statement_block,
  [52721] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1738), 1,
      sym_statement_block,
  [52731] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3690), 1,
      anon_sym_EQ,
    ACTIONS(3692), 1,
      sym__statement_terminator,
  [52741] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3694), 1,
      anon_sym_LBRACE,
    STATE(1887), 1,
      sym_switch_body,
  [52751] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1902), 1,
      sym_statement_block,
  [52761] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1703), 1,
      sym_statement_block,
  [52771] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1162), 1,
      anon_sym_LBRACE,
    STATE(935), 1,
      sym_script_block_expression,
  [52781] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1352), 1,
      sym_statement_block,
  [52791] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3694), 1,
      anon_sym_LBRACE,
    STATE(2126), 1,
      sym_switch_body,
  [52801] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1946), 1,
      sym_statement_block,
  [52811] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3696), 2,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
  [52819] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3698), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [52827] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1965), 1,
      sym_statement_block,
  [52837] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1701), 1,
      sym_statement_block,
  [52847] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1308), 1,
      sym_statement_block,
  [52857] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1777), 1,
      sym_statement_block,
  [52867] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1799), 1,
      sym_statement_block,
  [52877] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1483), 1,
      sym_statement_block,
  [52887] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1933), 1,
      sym_statement_block,
  [52897] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1945), 1,
      sym_statement_block,
  [52907] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1993), 1,
      sym_statement_block,
  [52917] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(2080), 1,
      sym_statement_block,
  [52927] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1722), 1,
      sym_statement_block,
  [52937] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1735), 1,
      sym_statement_block,
  [52947] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1714), 1,
      sym_statement_block,
  [52957] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1715), 1,
      sym_statement_block,
  [52967] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1744), 1,
      sym_statement_block,
  [52977] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1785), 1,
      sym_statement_block,
  [52987] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3700), 1,
      anon_sym_EQ,
    ACTIONS(3702), 1,
      sym__statement_terminator,
  [52997] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1964), 1,
      sym_statement_block,
  [53007] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1513), 1,
      sym_statement_block,
  [53017] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1974), 1,
      sym_statement_block,
  [53027] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1708), 1,
      sym_statement_block,
  [53037] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1725), 1,
      sym_statement_block,
  [53047] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1726), 1,
      sym_statement_block,
  [53057] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(105), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym_script_block_expression,
  [53067] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(400), 1,
      sym_statement_block,
  [53077] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(401), 1,
      sym_statement_block,
  [53087] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(402), 1,
      sym_statement_block,
  [53097] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(403), 1,
      sym_statement_block,
  [53107] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(419), 1,
      sym_statement_block,
  [53117] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(417), 1,
      sym_statement_block,
  [53127] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3704), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53135] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3295), 1,
      anon_sym_LBRACE,
    STATE(294), 1,
      sym_statement_block,
  [53145] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(369), 1,
      sym_statement_block,
  [53155] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(374), 1,
      sym_statement_block,
  [53165] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(365), 1,
      sym_statement_block,
  [53175] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3523), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53183] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1954), 1,
      anon_sym_LBRACE,
    STATE(775), 1,
      sym_script_block_expression,
  [53193] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3706), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53201] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3672), 1,
      anon_sym_LPAREN,
    ACTIONS(3708), 1,
      anon_sym_RBRACK,
  [53211] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1996), 1,
      anon_sym_LBRACE,
    STATE(823), 1,
      sym_script_block_expression,
  [53221] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3710), 1,
      anon_sym_LBRACE,
    STATE(341), 1,
      sym_switch_body,
  [53231] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3712), 2,
      anon_sym_LBRACE,
      aux_sym_data_commands_allowed_token1,
  [53239] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(343), 1,
      sym_statement_block,
  [53249] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(348), 1,
      sym_statement_block,
  [53259] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3714), 1,
      anon_sym_EQ,
    ACTIONS(3716), 1,
      sym__statement_terminator,
  [53269] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2408), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [53277] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2026), 1,
      anon_sym_LBRACE,
    STATE(822), 1,
      sym_script_block_expression,
  [53287] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3672), 1,
      anon_sym_LPAREN,
    ACTIONS(3718), 1,
      anon_sym_RBRACK,
  [53297] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1585), 1,
      sym_statement_block,
  [53307] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(394), 1,
      sym_statement_block,
  [53317] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(422), 1,
      sym_statement_block,
  [53327] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3720), 1,
      sym__command_token,
    STATE(1464), 1,
      sym_function_name,
  [53337] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3722), 1,
      anon_sym_LBRACE,
    ACTIONS(3724), 1,
      anon_sym_COLON,
  [53347] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(139), 1,
      anon_sym_LBRACE,
    STATE(94), 1,
      sym_script_block_expression,
  [53357] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3710), 1,
      anon_sym_LBRACE,
    STATE(354), 1,
      sym_switch_body,
  [53367] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(356), 1,
      sym_statement_block,
  [53377] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(363), 1,
      sym_statement_block,
  [53387] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(368), 1,
      sym_statement_block,
  [53397] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3437), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [53405] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(405), 1,
      sym_statement_block,
  [53415] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3447), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53423] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3726), 1,
      anon_sym_LBRACE,
    ACTIONS(3728), 1,
      anon_sym_COLON,
  [53433] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    STATE(1531), 1,
      sym_type_literal,
  [53443] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3730), 1,
      anon_sym_LBRACE,
    STATE(293), 1,
      sym_statement_block,
  [53453] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(375), 1,
      sym_statement_block,
  [53463] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(376), 1,
      sym_statement_block,
  [53473] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3295), 1,
      anon_sym_LBRACE,
    STATE(310), 1,
      sym_statement_block,
  [53483] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(346), 1,
      sym_statement_block,
  [53493] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(347), 1,
      sym_statement_block,
  [53503] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(359), 1,
      sym_statement_block,
  [53513] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(366), 1,
      sym_statement_block,
  [53523] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3732), 1,
      anon_sym_LBRACE,
    ACTIONS(3734), 1,
      anon_sym_COLON,
  [53533] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(372), 1,
      sym_statement_block,
  [53543] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(329), 1,
      sym_statement_block,
  [53553] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3736), 1,
      anon_sym_LBRACE,
    STATE(380), 1,
      sym_switch_body,
  [53563] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(331), 1,
      sym_statement_block,
  [53573] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(332), 1,
      sym_statement_block,
  [53583] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(333), 1,
      sym_statement_block,
  [53593] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(334), 1,
      sym_statement_block,
  [53603] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(336), 1,
      sym_statement_block,
  [53613] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3730), 1,
      anon_sym_LBRACE,
    STATE(312), 1,
      sym_statement_block,
  [53623] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(338), 1,
      sym_statement_block,
  [53633] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3738), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53641] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(339), 1,
      sym_statement_block,
  [53651] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(340), 1,
      sym_statement_block,
  [53661] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3595), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53669] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_LBRACE,
    STATE(342), 1,
      sym_statement_block,
  [53679] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(396), 1,
      sym_statement_block,
  [53689] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3740), 1,
      anon_sym_LBRACE,
    STATE(325), 1,
      sym_statement_block,
  [53699] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(407), 1,
      sym_statement_block,
  [53709] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(408), 1,
      sym_statement_block,
  [53719] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3581), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53727] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(409), 1,
      sym_statement_block,
  [53737] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3742), 1,
      anon_sym_LBRACE,
    ACTIONS(3744), 1,
      anon_sym_COLON,
  [53747] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(397), 1,
      sym_statement_block,
  [53757] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3293), 1,
      anon_sym_LBRACE,
    STATE(295), 1,
      sym_statement_block,
  [53767] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3746), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53775] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(398), 1,
      sym_statement_block,
  [53785] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3736), 1,
      anon_sym_LBRACE,
    STATE(416), 1,
      sym_switch_body,
  [53795] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_LBRACE,
    STATE(413), 1,
      sym_statement_block,
  [53805] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3358), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [53813] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3720), 1,
      sym__command_token,
    STATE(1434), 1,
      sym_function_name,
  [53823] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3720), 1,
      sym__command_token,
    STATE(1461), 1,
      sym_function_name,
  [53833] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3720), 1,
      sym__command_token,
    STATE(1488), 1,
      sym_function_name,
  [53843] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3748), 2,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
  [53851] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3750), 2,
      aux_sym_while_statement_token1,
      aux_sym_do_statement_token2,
  [53859] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1691), 1,
      sym_statement_block,
  [53869] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1520), 1,
      sym_statement_block,
  [53879] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3283), 1,
      anon_sym_LBRACE,
    STATE(1692), 1,
      sym_statement_block,
  [53889] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3740), 1,
      anon_sym_LBRACE,
    STATE(296), 1,
      sym_statement_block,
  [53899] = 3,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1924), 1,
      anon_sym_LBRACE,
    STATE(761), 1,
      sym_script_block_expression,
  [53909] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3752), 1,
      anon_sym_RBRACE,
  [53916] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3754), 1,
      anon_sym_LBRACE,
  [53923] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3756), 1,
      anon_sym_RBRACE,
  [53930] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1565), 1,
      anon_sym_RPAREN,
  [53937] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3758), 1,
      sym__statement_terminator,
  [53944] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1559), 1,
      anon_sym_RPAREN,
  [53951] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1567), 1,
      anon_sym_RPAREN,
  [53958] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3760), 1,
      anon_sym_RBRACK,
  [53965] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3762), 1,
      anon_sym_RPAREN,
  [53972] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3764), 1,
      anon_sym_LBRACE,
  [53979] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1535), 1,
      anon_sym_RPAREN,
  [53986] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3766), 1,
      anon_sym_RPAREN,
  [53993] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3768), 1,
      anon_sym_RBRACE,
  [54000] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3770), 1,
      sym__statement_terminator,
  [54007] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3772), 1,
      anon_sym_RPAREN,
  [54014] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3774), 1,
      anon_sym_RPAREN,
  [54021] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1521), 1,
      anon_sym_RPAREN,
  [54028] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1523), 1,
      anon_sym_RPAREN,
  [54035] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1479), 1,
      anon_sym_RPAREN,
  [54042] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3002), 1,
      anon_sym_RPAREN,
  [54049] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3776), 1,
      anon_sym_RPAREN,
  [54056] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3778), 1,
      anon_sym_RBRACE,
  [54063] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3780), 1,
      anon_sym_RBRACK,
  [54070] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3782), 1,
      anon_sym_LBRACE,
  [54077] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1515), 1,
      anon_sym_RPAREN,
  [54084] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3784), 1,
      anon_sym_RBRACK,
  [54091] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3786), 1,
      anon_sym_RPAREN,
  [54098] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1537), 1,
      anon_sym_RPAREN,
  [54105] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1539), 1,
      anon_sym_RPAREN,
  [54112] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3788), 1,
      anon_sym_RPAREN,
  [54119] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3790), 1,
      anon_sym_RBRACE,
  [54126] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3792), 1,
      anon_sym_RPAREN,
  [54133] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3794), 1,
      anon_sym_RPAREN,
  [54140] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3796), 1,
      anon_sym_LBRACE,
  [54147] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3798), 1,
      sym_type_identifier,
  [54154] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1483), 1,
      sym__statement_terminator,
  [54161] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3800), 1,
      anon_sym_RBRACK,
  [54168] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1517), 1,
      anon_sym_RPAREN,
  [54175] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3802), 1,
      anon_sym_RBRACE,
  [54182] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3804), 1,
      anon_sym_RBRACK,
  [54189] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1477), 1,
      anon_sym_RPAREN,
  [54196] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3806), 1,
      anon_sym_RBRACE,
  [54203] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3808), 1,
      anon_sym_RBRACK,
  [54210] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3810), 1,
      anon_sym_LPAREN,
  [54217] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3305), 1,
      sym__statement_terminator,
  [54224] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3812), 1,
      anon_sym_RBRACK,
  [54231] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1525), 1,
      anon_sym_RPAREN,
  [54238] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3814), 1,
      sym__statement_terminator,
  [54245] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3816), 1,
      anon_sym_RBRACK,
  [54252] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3818), 1,
      ts_builtin_sym_end,
  [54259] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3820), 1,
      sym__statement_terminator,
  [54266] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_RBRACE,
  [54273] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3824), 1,
      ts_builtin_sym_end,
  [54280] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3826), 1,
      sym__statement_terminator,
  [54287] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3828), 1,
      anon_sym_RPAREN,
  [54294] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3830), 1,
      anon_sym_RBRACE,
  [54301] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3832), 1,
      anon_sym_LBRACE,
  [54308] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1483), 1,
      anon_sym_RPAREN,
  [54315] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3834), 1,
      anon_sym_RPAREN,
  [54322] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3836), 1,
      anon_sym_RPAREN,
  [54329] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3838), 1,
      anon_sym_RBRACE,
  [54336] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3840), 1,
      anon_sym_RPAREN,
  [54343] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3842), 1,
      anon_sym_RPAREN,
  [54350] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3844), 1,
      anon_sym_RBRACE,
  [54357] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3846), 1,
      anon_sym_RBRACE,
  [54364] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3848), 1,
      anon_sym_RBRACE,
  [54371] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1485), 1,
      anon_sym_RPAREN,
  [54378] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3850), 1,
      anon_sym_RBRACE,
  [54385] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3852), 1,
      anon_sym_RBRACK,
  [54392] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3854), 1,
      anon_sym_RBRACK,
  [54399] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3856), 1,
      anon_sym_RPAREN,
  [54406] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1541), 1,
      sym__statement_terminator,
  [54413] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_RPAREN,
  [54420] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3858), 1,
      anon_sym_RBRACE,
  [54427] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1543), 1,
      sym__statement_terminator,
  [54434] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3860), 1,
      anon_sym_RBRACK,
  [54441] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1547), 1,
      sym__statement_terminator,
  [54448] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3862), 1,
      aux_sym_foreach_statement_token2,
  [54455] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3864), 1,
      anon_sym_LPAREN,
  [54462] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1487), 1,
      anon_sym_RPAREN,
  [54469] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3866), 1,
      sym__statement_terminator,
  [54476] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3868), 1,
      anon_sym_RPAREN,
  [54483] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3870), 1,
      anon_sym_RBRACE,
  [54490] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1553), 1,
      sym__statement_terminator,
  [54497] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3872), 1,
      anon_sym_RPAREN,
  [54504] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3874), 1,
      anon_sym_RPAREN,
  [54511] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3876), 1,
      anon_sym_RBRACE,
  [54518] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1527), 1,
      anon_sym_RPAREN,
  [54525] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3878), 1,
      anon_sym_LBRACE,
  [54532] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3880), 1,
      anon_sym_RBRACE,
  [54539] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3882), 1,
      anon_sym_RBRACE,
  [54546] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1555), 1,
      sym__statement_terminator,
  [54553] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3884), 1,
      anon_sym_RBRACK,
  [54560] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3886), 1,
      anon_sym_RPAREN,
  [54567] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3888), 1,
      anon_sym_LPAREN,
  [54574] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3890), 1,
      anon_sym_LPAREN,
  [54581] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1477), 1,
      sym__statement_terminator,
  [54588] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3049), 1,
      anon_sym_RPAREN,
  [54595] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3892), 1,
      anon_sym_LPAREN,
  [54602] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym__statement_terminator,
  [54609] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3894), 1,
      anon_sym_RBRACK,
  [54616] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1489), 1,
      anon_sym_RPAREN,
  [54623] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3896), 1,
      anon_sym_RPAREN,
  [54630] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3898), 1,
      anon_sym_RPAREN,
  [54637] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3900), 1,
      anon_sym_RBRACE,
  [54644] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3902), 1,
      anon_sym_RBRACE,
  [54651] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3904), 1,
      anon_sym_RPAREN,
  [54658] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3906), 1,
      anon_sym_RPAREN,
  [54665] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3908), 1,
      anon_sym_RBRACE,
  [54672] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3910), 1,
      sym__statement_terminator,
  [54679] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3912), 1,
      sym_simple_name,
  [54686] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3914), 1,
      anon_sym_RBRACE,
  [54693] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3916), 1,
      anon_sym_RBRACE,
  [54700] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3918), 1,
      anon_sym_LPAREN,
  [54707] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3920), 1,
      anon_sym_RBRACK,
  [54714] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3922), 1,
      anon_sym_RPAREN,
  [54721] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1541), 1,
      anon_sym_RPAREN,
  [54728] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(211), 1,
      sym__statement_terminator,
  [54735] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3924), 1,
      sym__statement_terminator,
  [54742] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3926), 1,
      anon_sym_LPAREN,
  [54749] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3928), 1,
      anon_sym_RPAREN,
  [54756] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3930), 1,
      anon_sym_RPAREN,
  [54763] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3932), 1,
      anon_sym_RBRACE,
  [54770] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3934), 1,
      anon_sym_LPAREN,
  [54777] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3936), 1,
      sym__statement_terminator,
  [54784] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3938), 1,
      anon_sym_RPAREN,
  [54791] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3940), 1,
      anon_sym_RBRACE,
  [54798] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1543), 1,
      anon_sym_RPAREN,
  [54805] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3942), 1,
      anon_sym_RPAREN,
  [54812] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3944), 1,
      anon_sym_RPAREN,
  [54819] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3946), 1,
      anon_sym_RBRACE,
  [54826] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3948), 1,
      aux_sym_param_block_token1,
  [54833] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1491), 1,
      anon_sym_RPAREN,
  [54840] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3950), 1,
      anon_sym_RBRACE,
  [54847] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3952), 1,
      anon_sym_RBRACE,
  [54854] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3954), 1,
      anon_sym_RBRACK,
  [54861] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3956), 1,
      anon_sym_RPAREN,
  [54868] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3958), 1,
      anon_sym_LBRACE,
  [54875] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3674), 1,
      anon_sym_RBRACK,
  [54882] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3960), 1,
      anon_sym_RPAREN,
  [54889] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3962), 1,
      anon_sym_RBRACE,
  [54896] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3964), 1,
      anon_sym_RPAREN,
  [54903] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3966), 1,
      anon_sym_RBRACE,
  [54910] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(2998), 1,
      anon_sym_RPAREN,
  [54917] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3968), 1,
      anon_sym_RPAREN,
  [54924] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3970), 1,
      anon_sym_RBRACK,
  [54931] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3972), 1,
      anon_sym_RPAREN,
  [54938] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3974), 1,
      anon_sym_RPAREN,
  [54945] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3976), 1,
      anon_sym_RBRACE,
  [54952] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3978), 1,
      anon_sym_RPAREN,
  [54959] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_RBRACE,
  [54966] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3982), 1,
      anon_sym_RPAREN,
  [54973] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3984), 1,
      anon_sym_LPAREN,
  [54980] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3986), 1,
      anon_sym_RBRACE,
  [54987] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3988), 1,
      anon_sym_RPAREN,
  [54994] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3990), 1,
      anon_sym_RBRACE,
  [55001] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3992), 1,
      anon_sym_RPAREN,
  [55008] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3994), 1,
      anon_sym_RBRACE,
  [55015] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3996), 1,
      anon_sym_RPAREN,
  [55022] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3998), 1,
      anon_sym_RPAREN,
  [55029] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4000), 1,
      anon_sym_RPAREN,
  [55036] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4002), 1,
      anon_sym_RBRACE,
  [55043] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3253), 1,
      anon_sym_LBRACE,
  [55050] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4004), 1,
      anon_sym_RPAREN,
  [55057] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4006), 1,
      anon_sym_RPAREN,
  [55064] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
  [55071] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4010), 1,
      anon_sym_RBRACE,
  [55078] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4012), 1,
      anon_sym_RPAREN,
  [55085] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3004), 1,
      anon_sym_RPAREN,
  [55092] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4014), 1,
      anon_sym_RBRACE,
  [55099] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4016), 1,
      anon_sym_RPAREN,
  [55106] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4018), 1,
      anon_sym_LPAREN,
  [55113] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4020), 1,
      anon_sym_RBRACK,
  [55120] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4022), 1,
      sym__statement_terminator,
  [55127] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4024), 1,
      anon_sym_RBRACE,
  [55134] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4026), 1,
      anon_sym_LBRACE,
  [55141] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3269), 1,
      anon_sym_LBRACE,
  [55148] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4028), 1,
      anon_sym_RBRACK,
  [55155] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4030), 1,
      anon_sym_RPAREN,
  [55162] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4032), 1,
      anon_sym_RPAREN,
  [55169] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4034), 1,
      anon_sym_RPAREN,
  [55176] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4036), 1,
      sym_simple_name,
  [55183] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4038), 1,
      anon_sym_RBRACE,
  [55190] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4040), 1,
      anon_sym_LBRACE,
  [55197] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4042), 1,
      sym__statement_terminator,
  [55204] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1493), 1,
      anon_sym_RPAREN,
  [55211] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1495), 1,
      anon_sym_RPAREN,
  [55218] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4044), 1,
      anon_sym_RBRACE,
  [55225] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4046), 1,
      anon_sym_LBRACE,
  [55232] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_RPAREN,
  [55239] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4048), 1,
      anon_sym_LBRACE,
  [55246] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3708), 1,
      anon_sym_RBRACK,
  [55253] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4050), 1,
      anon_sym_LPAREN,
  [55260] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4052), 1,
      anon_sym_LPAREN,
  [55267] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4054), 1,
      anon_sym_RPAREN,
  [55274] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4056), 1,
      anon_sym_RPAREN,
  [55281] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4058), 1,
      anon_sym_RPAREN,
  [55288] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4060), 1,
      anon_sym_LBRACE,
  [55295] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4062), 1,
      anon_sym_RPAREN,
  [55302] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1547), 1,
      anon_sym_RPAREN,
  [55309] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4064), 1,
      anon_sym_RBRACK,
  [55316] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4066), 1,
      anon_sym_LBRACE,
  [55323] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4068), 1,
      anon_sym_RBRACK,
  [55330] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4070), 1,
      anon_sym_RPAREN,
  [55337] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1557), 1,
      anon_sym_RPAREN,
  [55344] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4072), 1,
      anon_sym_RBRACK,
  [55351] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4074), 1,
      anon_sym_RBRACE,
  [55358] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4076), 1,
      anon_sym_RPAREN,
  [55365] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_RPAREN,
  [55372] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1317), 1,
      sym__statement_terminator,
  [55379] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_RBRACE,
  [55386] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1553), 1,
      anon_sym_RPAREN,
  [55393] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4082), 1,
      anon_sym_RPAREN,
  [55400] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_RPAREN,
  [55407] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4086), 1,
      anon_sym_RPAREN,
  [55414] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4088), 1,
      anon_sym_RBRACE,
  [55421] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4090), 1,
      anon_sym_RPAREN,
  [55428] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1497), 1,
      anon_sym_RPAREN,
  [55435] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4092), 1,
      sym_simple_name,
  [55442] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1551), 1,
      sym__statement_terminator,
  [55449] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4094), 1,
      anon_sym_RBRACK,
  [55456] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1557), 1,
      sym__statement_terminator,
  [55463] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4096), 1,
      sym__statement_terminator,
  [55470] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1381), 1,
      sym__statement_terminator,
  [55477] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3051), 1,
      anon_sym_RPAREN,
  [55484] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1559), 1,
      sym__statement_terminator,
  [55491] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4098), 1,
      sym__statement_terminator,
  [55498] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4100), 1,
      anon_sym_RPAREN,
  [55505] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4102), 1,
      anon_sym_RPAREN,
  [55512] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4104), 1,
      anon_sym_LPAREN,
  [55519] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4106), 1,
      anon_sym_RBRACK,
  [55526] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3682), 1,
      sym__statement_terminator,
  [55533] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4108), 1,
      ts_builtin_sym_end,
  [55540] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4110), 1,
      anon_sym_RPAREN,
  [55547] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_RBRACE,
  [55554] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1499), 1,
      anon_sym_RPAREN,
  [55561] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3305), 1,
      anon_sym_RPAREN,
  [55568] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4114), 1,
      sym__statement_terminator,
  [55575] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4116), 1,
      sym_simple_name,
  [55582] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LBRACE,
  [55589] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4120), 1,
      sym__statement_terminator,
  [55596] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4122), 1,
      anon_sym_LPAREN,
  [55603] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4124), 1,
      anon_sym_RPAREN,
  [55610] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1569), 1,
      anon_sym_RPAREN,
  [55617] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4126), 1,
      anon_sym_RPAREN,
  [55624] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4128), 1,
      anon_sym_LBRACE,
  [55631] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4130), 1,
      sym__statement_terminator,
  [55638] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1501), 1,
      anon_sym_RPAREN,
  [55645] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1545), 1,
      anon_sym_RPAREN,
  [55652] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4132), 1,
      anon_sym_LPAREN,
  [55659] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4134), 1,
      anon_sym_RPAREN,
  [55666] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_RPAREN,
  [55673] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4136), 1,
      anon_sym_RBRACK,
  [55680] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1503), 1,
      anon_sym_RPAREN,
  [55687] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4138), 1,
      anon_sym_RPAREN,
  [55694] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4140), 1,
      anon_sym_RPAREN,
  [55701] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4142), 1,
      anon_sym_RPAREN,
  [55708] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4144), 1,
      anon_sym_RBRACE,
  [55715] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1529), 1,
      anon_sym_RPAREN,
  [55722] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4146), 1,
      anon_sym_RPAREN,
  [55729] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4148), 1,
      anon_sym_RPAREN,
  [55736] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4150), 1,
      anon_sym_RPAREN,
  [55743] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4152), 1,
      sym__statement_terminator,
  [55750] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4154), 1,
      anon_sym_RPAREN,
  [55757] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4156), 1,
      anon_sym_LPAREN,
  [55764] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1381), 1,
      anon_sym_RPAREN,
  [55771] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1531), 1,
      anon_sym_RPAREN,
  [55778] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1561), 1,
      anon_sym_RPAREN,
  [55785] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4158), 1,
      anon_sym_RPAREN,
  [55792] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4160), 1,
      anon_sym_RPAREN,
  [55799] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_RPAREN,
  [55806] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4164), 1,
      anon_sym_RBRACE,
  [55813] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1563), 1,
      anon_sym_RPAREN,
  [55820] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4166), 1,
      anon_sym_RBRACK,
  [55827] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4168), 1,
      sym__statement_terminator,
  [55834] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1505), 1,
      anon_sym_RPAREN,
  [55841] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1533), 1,
      anon_sym_RPAREN,
  [55848] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1337), 1,
      sym__statement_terminator,
  [55855] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4170), 1,
      sym__statement_terminator,
  [55862] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4172), 1,
      anon_sym_RPAREN,
  [55869] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1481), 1,
      sym__statement_terminator,
  [55876] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4174), 1,
      anon_sym_RPAREN,
  [55883] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4176), 1,
      anon_sym_RBRACE,
  [55890] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4178), 1,
      anon_sym_LBRACE,
  [55897] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1569), 1,
      sym__statement_terminator,
  [55904] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4180), 1,
      anon_sym_RPAREN,
  [55911] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1545), 1,
      sym__statement_terminator,
  [55918] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4182), 1,
      anon_sym_RPAREN,
  [55925] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4184), 1,
      anon_sym_LBRACE,
  [55932] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1507), 1,
      anon_sym_RPAREN,
  [55939] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4186), 1,
      anon_sym_RBRACK,
  [55946] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1549), 1,
      sym__statement_terminator,
  [55953] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4188), 1,
      anon_sym_RBRACE,
  [55960] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4190), 1,
      anon_sym_RPAREN,
  [55967] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3770), 1,
      anon_sym_RPAREN,
  [55974] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1509), 1,
      anon_sym_RPAREN,
  [55981] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4192), 1,
      anon_sym_RBRACK,
  [55988] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4194), 1,
      anon_sym_RPAREN,
  [55995] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4196), 1,
      anon_sym_RPAREN,
  [56002] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1561), 1,
      sym__statement_terminator,
  [56009] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1563), 1,
      sym__statement_terminator,
  [56016] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1511), 1,
      anon_sym_RPAREN,
  [56023] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4198), 1,
      anon_sym_RPAREN,
  [56030] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4200), 1,
      anon_sym_RPAREN,
  [56037] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4202), 1,
      anon_sym_RPAREN,
  [56044] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4204), 1,
      anon_sym_RBRACE,
  [56051] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4206), 1,
      anon_sym_RPAREN,
  [56058] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1565), 1,
      sym__statement_terminator,
  [56065] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4208), 1,
      anon_sym_LBRACE,
  [56072] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4210), 1,
      anon_sym_RPAREN,
  [56079] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4212), 1,
      anon_sym_RBRACE,
  [56086] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1567), 1,
      sym__statement_terminator,
  [56093] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1479), 1,
      sym__statement_terminator,
  [56100] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4214), 1,
      anon_sym_LPAREN,
  [56107] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4216), 1,
      sym__statement_terminator,
  [56114] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4218), 1,
      aux_sym_foreach_statement_token2,
  [56121] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4220), 1,
      anon_sym_RBRACE,
  [56128] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4222), 1,
      anon_sym_RPAREN,
  [56135] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1485), 1,
      sym__statement_terminator,
  [56142] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1487), 1,
      sym__statement_terminator,
  [56149] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1489), 1,
      sym__statement_terminator,
  [56156] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1491), 1,
      sym__statement_terminator,
  [56163] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1519), 1,
      anon_sym_RPAREN,
  [56170] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4224), 1,
      anon_sym_RBRACE,
  [56177] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1493), 1,
      sym__statement_terminator,
  [56184] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1495), 1,
      sym__statement_terminator,
  [56191] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4226), 1,
      anon_sym_RBRACE,
  [56198] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1321), 1,
      sym__statement_terminator,
  [56205] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4228), 1,
      anon_sym_RBRACE,
  [56212] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1497), 1,
      sym__statement_terminator,
  [56219] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4230), 1,
      anon_sym_RPAREN,
  [56226] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1499), 1,
      sym__statement_terminator,
  [56233] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1501), 1,
      sym__statement_terminator,
  [56240] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1503), 1,
      sym__statement_terminator,
  [56247] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1505), 1,
      sym__statement_terminator,
  [56254] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1507), 1,
      sym__statement_terminator,
  [56261] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1509), 1,
      sym__statement_terminator,
  [56268] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1511), 1,
      sym__statement_terminator,
  [56275] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4232), 1,
      anon_sym_LPAREN,
  [56282] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4234), 1,
      anon_sym_RBRACE,
  [56289] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4236), 1,
      sym_simple_name,
  [56296] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4238), 1,
      sym_simple_name,
  [56303] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4240), 1,
      aux_sym_param_block_token1,
  [56310] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4242), 1,
      anon_sym_LPAREN,
  [56317] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4244), 1,
      anon_sym_RPAREN,
  [56324] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1513), 1,
      sym__statement_terminator,
  [56331] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4246), 1,
      sym__statement_terminator,
  [56338] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4248), 1,
      anon_sym_LPAREN,
  [56345] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4250), 1,
      sym_simple_name,
  [56352] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1515), 1,
      sym__statement_terminator,
  [56359] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4252), 1,
      sym__statement_terminator,
  [56366] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1517), 1,
      sym__statement_terminator,
  [56373] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4254), 1,
      sym__statement_terminator,
  [56380] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4256), 1,
      sym__statement_terminator,
  [56387] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1519), 1,
      sym__statement_terminator,
  [56394] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4258), 1,
      sym__statement_terminator,
  [56401] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4260), 1,
      sym__statement_terminator,
  [56408] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4262), 1,
      sym__statement_terminator,
  [56415] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4264), 1,
      anon_sym_LPAREN,
  [56422] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(3718), 1,
      anon_sym_RBRACK,
  [56429] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4266), 1,
      sym_simple_name,
  [56436] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4268), 1,
      sym_simple_name,
  [56443] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4270), 1,
      aux_sym_param_block_token1,
  [56450] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1521), 1,
      sym__statement_terminator,
  [56457] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1523), 1,
      sym__statement_terminator,
  [56464] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4272), 1,
      sym__statement_terminator,
  [56471] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_LPAREN,
  [56478] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4276), 1,
      sym_simple_name,
  [56485] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1525), 1,
      sym__statement_terminator,
  [56492] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4278), 1,
      sym__statement_terminator,
  [56499] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1527), 1,
      sym__statement_terminator,
  [56506] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4280), 1,
      sym__statement_terminator,
  [56513] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4282), 1,
      sym__statement_terminator,
  [56520] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1529), 1,
      sym__statement_terminator,
  [56527] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4284), 1,
      sym__statement_terminator,
  [56534] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4286), 1,
      sym__statement_terminator,
  [56541] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4288), 1,
      sym__statement_terminator,
  [56548] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4290), 1,
      anon_sym_LPAREN,
  [56555] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4292), 1,
      sym__statement_terminator,
  [56562] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4294), 1,
      sym_simple_name,
  [56569] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4296), 1,
      sym_simple_name,
  [56576] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1531), 1,
      sym__statement_terminator,
  [56583] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1513), 1,
      anon_sym_RPAREN,
  [56590] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4298), 1,
      sym__statement_terminator,
  [56597] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4300), 1,
      anon_sym_LPAREN,
  [56604] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4302), 1,
      sym_simple_name,
  [56611] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1533), 1,
      sym__statement_terminator,
  [56618] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4304), 1,
      sym__statement_terminator,
  [56625] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1535), 1,
      sym__statement_terminator,
  [56632] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4306), 1,
      sym__statement_terminator,
  [56639] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4308), 1,
      sym__statement_terminator,
  [56646] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1537), 1,
      sym__statement_terminator,
  [56653] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4310), 1,
      sym__statement_terminator,
  [56660] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4312), 1,
      sym__statement_terminator,
  [56667] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4314), 1,
      sym__statement_terminator,
  [56674] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4316), 1,
      anon_sym_LPAREN,
  [56681] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4318), 1,
      anon_sym_LPAREN,
  [56688] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4320), 1,
      sym__statement_terminator,
  [56695] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4322), 1,
      aux_sym_foreach_statement_token2,
  [56702] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4324), 1,
      aux_sym_foreach_statement_token2,
  [56709] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4326), 1,
      anon_sym_LPAREN,
  [56716] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4328), 1,
      anon_sym_LPAREN,
  [56723] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4330), 1,
      anon_sym_LPAREN,
  [56730] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1539), 1,
      sym__statement_terminator,
  [56737] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4332), 1,
      aux_sym_foreach_statement_token2,
  [56744] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4334), 1,
      aux_sym_foreach_statement_token2,
  [56751] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4336), 1,
      anon_sym_LPAREN,
  [56758] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4338), 1,
      anon_sym_LPAREN,
  [56765] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4340), 1,
      anon_sym_LPAREN,
  [56772] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_RPAREN,
  [56779] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4342), 1,
      aux_sym_foreach_statement_token2,
  [56786] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4344), 1,
      aux_sym_foreach_statement_token2,
  [56793] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4346), 1,
      anon_sym_LPAREN,
  [56800] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4348), 1,
      anon_sym_RPAREN,
  [56807] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4350), 1,
      sym__statement_terminator,
  [56814] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4352), 1,
      anon_sym_RBRACE,
  [56821] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4354), 1,
      anon_sym_RBRACK,
  [56828] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4356), 1,
      anon_sym_EQ,
  [56835] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_RBRACE,
  [56842] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4360), 1,
      anon_sym_RBRACE,
  [56849] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4362), 1,
      sym__statement_terminator,
  [56856] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4364), 1,
      anon_sym_RBRACE,
  [56863] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4366), 1,
      anon_sym_RBRACE,
  [56870] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4368), 1,
      anon_sym_LPAREN,
  [56877] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4370), 1,
      anon_sym_LBRACE,
  [56884] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4372), 1,
      anon_sym_LPAREN,
  [56891] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4374), 1,
      anon_sym_EQ,
  [56898] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(4376), 1,
      anon_sym_LPAREN,
  [56905] = 2,
    ACTIONS(79), 1,
      sym_comment,
    ACTIONS(1481), 1,
      anon_sym_RPAREN,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(507)] = 0,
  [SMALL_STATE(508)] = 75,
  [SMALL_STATE(509)] = 148,
  [SMALL_STATE(510)] = 257,
  [SMALL_STATE(511)] = 330,
  [SMALL_STATE(512)] = 401,
  [SMALL_STATE(513)] = 474,
  [SMALL_STATE(514)] = 583,
  [SMALL_STATE(515)] = 716,
  [SMALL_STATE(516)] = 825,
  [SMALL_STATE(517)] = 934,
  [SMALL_STATE(518)] = 1005,
  [SMALL_STATE(519)] = 1080,
  [SMALL_STATE(520)] = 1153,
  [SMALL_STATE(521)] = 1226,
  [SMALL_STATE(522)] = 1299,
  [SMALL_STATE(523)] = 1429,
  [SMALL_STATE(524)] = 1559,
  [SMALL_STATE(525)] = 1689,
  [SMALL_STATE(526)] = 1759,
  [SMALL_STATE(527)] = 1889,
  [SMALL_STATE(528)] = 1957,
  [SMALL_STATE(529)] = 2087,
  [SMALL_STATE(530)] = 2212,
  [SMALL_STATE(531)] = 2339,
  [SMALL_STATE(532)] = 2464,
  [SMALL_STATE(533)] = 2589,
  [SMALL_STATE(534)] = 2714,
  [SMALL_STATE(535)] = 2841,
  [SMALL_STATE(536)] = 2966,
  [SMALL_STATE(537)] = 3093,
  [SMALL_STATE(538)] = 3218,
  [SMALL_STATE(539)] = 3289,
  [SMALL_STATE(540)] = 3414,
  [SMALL_STATE(541)] = 3541,
  [SMALL_STATE(542)] = 3612,
  [SMALL_STATE(543)] = 3739,
  [SMALL_STATE(544)] = 3866,
  [SMALL_STATE(545)] = 3991,
  [SMALL_STATE(546)] = 4063,
  [SMALL_STATE(547)] = 4135,
  [SMALL_STATE(548)] = 4257,
  [SMALL_STATE(549)] = 4379,
  [SMALL_STATE(550)] = 4503,
  [SMALL_STATE(551)] = 4627,
  [SMALL_STATE(552)] = 4751,
  [SMALL_STATE(553)] = 4875,
  [SMALL_STATE(554)] = 4999,
  [SMALL_STATE(555)] = 5120,
  [SMALL_STATE(556)] = 5191,
  [SMALL_STATE(557)] = 5308,
  [SMALL_STATE(558)] = 5429,
  [SMALL_STATE(559)] = 5500,
  [SMALL_STATE(560)] = 5617,
  [SMALL_STATE(561)] = 5688,
  [SMALL_STATE(562)] = 5809,
  [SMALL_STATE(563)] = 5930,
  [SMALL_STATE(564)] = 6001,
  [SMALL_STATE(565)] = 6072,
  [SMALL_STATE(566)] = 6189,
  [SMALL_STATE(567)] = 6260,
  [SMALL_STATE(568)] = 6381,
  [SMALL_STATE(569)] = 6498,
  [SMALL_STATE(570)] = 6619,
  [SMALL_STATE(571)] = 6737,
  [SMALL_STATE(572)] = 6807,
  [SMALL_STATE(573)] = 6925,
  [SMALL_STATE(574)] = 6995,
  [SMALL_STATE(575)] = 7113,
  [SMALL_STATE(576)] = 7231,
  [SMALL_STATE(577)] = 7349,
  [SMALL_STATE(578)] = 7419,
  [SMALL_STATE(579)] = 7537,
  [SMALL_STATE(580)] = 7607,
  [SMALL_STATE(581)] = 7725,
  [SMALL_STATE(582)] = 7794,
  [SMALL_STATE(583)] = 7909,
  [SMALL_STATE(584)] = 8024,
  [SMALL_STATE(585)] = 8139,
  [SMALL_STATE(586)] = 8208,
  [SMALL_STATE(587)] = 8323,
  [SMALL_STATE(588)] = 8438,
  [SMALL_STATE(589)] = 8553,
  [SMALL_STATE(590)] = 8665,
  [SMALL_STATE(591)] = 8777,
  [SMALL_STATE(592)] = 8889,
  [SMALL_STATE(593)] = 9001,
  [SMALL_STATE(594)] = 9113,
  [SMALL_STATE(595)] = 9181,
  [SMALL_STATE(596)] = 9293,
  [SMALL_STATE(597)] = 9361,
  [SMALL_STATE(598)] = 9470,
  [SMALL_STATE(599)] = 9579,
  [SMALL_STATE(600)] = 9688,
  [SMALL_STATE(601)] = 9797,
  [SMALL_STATE(602)] = 9906,
  [SMALL_STATE(603)] = 10015,
  [SMALL_STATE(604)] = 10124,
  [SMALL_STATE(605)] = 10233,
  [SMALL_STATE(606)] = 10342,
  [SMALL_STATE(607)] = 10451,
  [SMALL_STATE(608)] = 10560,
  [SMALL_STATE(609)] = 10669,
  [SMALL_STATE(610)] = 10778,
  [SMALL_STATE(611)] = 10887,
  [SMALL_STATE(612)] = 10996,
  [SMALL_STATE(613)] = 11105,
  [SMALL_STATE(614)] = 11214,
  [SMALL_STATE(615)] = 11323,
  [SMALL_STATE(616)] = 11432,
  [SMALL_STATE(617)] = 11541,
  [SMALL_STATE(618)] = 11650,
  [SMALL_STATE(619)] = 11759,
  [SMALL_STATE(620)] = 11868,
  [SMALL_STATE(621)] = 11977,
  [SMALL_STATE(622)] = 12086,
  [SMALL_STATE(623)] = 12195,
  [SMALL_STATE(624)] = 12304,
  [SMALL_STATE(625)] = 12413,
  [SMALL_STATE(626)] = 12522,
  [SMALL_STATE(627)] = 12631,
  [SMALL_STATE(628)] = 12740,
  [SMALL_STATE(629)] = 12849,
  [SMALL_STATE(630)] = 12958,
  [SMALL_STATE(631)] = 13067,
  [SMALL_STATE(632)] = 13176,
  [SMALL_STATE(633)] = 13285,
  [SMALL_STATE(634)] = 13394,
  [SMALL_STATE(635)] = 13503,
  [SMALL_STATE(636)] = 13612,
  [SMALL_STATE(637)] = 13721,
  [SMALL_STATE(638)] = 13830,
  [SMALL_STATE(639)] = 13939,
  [SMALL_STATE(640)] = 14048,
  [SMALL_STATE(641)] = 14157,
  [SMALL_STATE(642)] = 14266,
  [SMALL_STATE(643)] = 14375,
  [SMALL_STATE(644)] = 14484,
  [SMALL_STATE(645)] = 14593,
  [SMALL_STATE(646)] = 14702,
  [SMALL_STATE(647)] = 14811,
  [SMALL_STATE(648)] = 14920,
  [SMALL_STATE(649)] = 15029,
  [SMALL_STATE(650)] = 15138,
  [SMALL_STATE(651)] = 15247,
  [SMALL_STATE(652)] = 15356,
  [SMALL_STATE(653)] = 15465,
  [SMALL_STATE(654)] = 15574,
  [SMALL_STATE(655)] = 15683,
  [SMALL_STATE(656)] = 15792,
  [SMALL_STATE(657)] = 15901,
  [SMALL_STATE(658)] = 16010,
  [SMALL_STATE(659)] = 16119,
  [SMALL_STATE(660)] = 16228,
  [SMALL_STATE(661)] = 16337,
  [SMALL_STATE(662)] = 16446,
  [SMALL_STATE(663)] = 16555,
  [SMALL_STATE(664)] = 16664,
  [SMALL_STATE(665)] = 16773,
  [SMALL_STATE(666)] = 16882,
  [SMALL_STATE(667)] = 16991,
  [SMALL_STATE(668)] = 17100,
  [SMALL_STATE(669)] = 17209,
  [SMALL_STATE(670)] = 17318,
  [SMALL_STATE(671)] = 17427,
  [SMALL_STATE(672)] = 17536,
  [SMALL_STATE(673)] = 17645,
  [SMALL_STATE(674)] = 17754,
  [SMALL_STATE(675)] = 17863,
  [SMALL_STATE(676)] = 17972,
  [SMALL_STATE(677)] = 18081,
  [SMALL_STATE(678)] = 18190,
  [SMALL_STATE(679)] = 18299,
  [SMALL_STATE(680)] = 18408,
  [SMALL_STATE(681)] = 18474,
  [SMALL_STATE(682)] = 18534,
  [SMALL_STATE(683)] = 18594,
  [SMALL_STATE(684)] = 18660,
  [SMALL_STATE(685)] = 18719,
  [SMALL_STATE(686)] = 18778,
  [SMALL_STATE(687)] = 18843,
  [SMALL_STATE(688)] = 18908,
  [SMALL_STATE(689)] = 18964,
  [SMALL_STATE(690)] = 19065,
  [SMALL_STATE(691)] = 19166,
  [SMALL_STATE(692)] = 19267,
  [SMALL_STATE(693)] = 19368,
  [SMALL_STATE(694)] = 19469,
  [SMALL_STATE(695)] = 19570,
  [SMALL_STATE(696)] = 19671,
  [SMALL_STATE(697)] = 19772,
  [SMALL_STATE(698)] = 19873,
  [SMALL_STATE(699)] = 19974,
  [SMALL_STATE(700)] = 20075,
  [SMALL_STATE(701)] = 20176,
  [SMALL_STATE(702)] = 20277,
  [SMALL_STATE(703)] = 20362,
  [SMALL_STATE(704)] = 20463,
  [SMALL_STATE(705)] = 20564,
  [SMALL_STATE(706)] = 20665,
  [SMALL_STATE(707)] = 20766,
  [SMALL_STATE(708)] = 20867,
  [SMALL_STATE(709)] = 20968,
  [SMALL_STATE(710)] = 21053,
  [SMALL_STATE(711)] = 21154,
  [SMALL_STATE(712)] = 21237,
  [SMALL_STATE(713)] = 21320,
  [SMALL_STATE(714)] = 21382,
  [SMALL_STATE(715)] = 21466,
  [SMALL_STATE(716)] = 21528,
  [SMALL_STATE(717)] = 21590,
  [SMALL_STATE(718)] = 21672,
  [SMALL_STATE(719)] = 21756,
  [SMALL_STATE(720)] = 21838,
  [SMALL_STATE(721)] = 21918,
  [SMALL_STATE(722)] = 22000,
  [SMALL_STATE(723)] = 22082,
  [SMALL_STATE(724)] = 22144,
  [SMALL_STATE(725)] = 22224,
  [SMALL_STATE(726)] = 22285,
  [SMALL_STATE(727)] = 22346,
  [SMALL_STATE(728)] = 22427,
  [SMALL_STATE(729)] = 22506,
  [SMALL_STATE(730)] = 22587,
  [SMALL_STATE(731)] = 22666,
  [SMALL_STATE(732)] = 22727,
  [SMALL_STATE(733)] = 22788,
  [SMALL_STATE(734)] = 22843,
  [SMALL_STATE(735)] = 22937,
  [SMALL_STATE(736)] = 23031,
  [SMALL_STATE(737)] = 23087,
  [SMALL_STATE(738)] = 23141,
  [SMALL_STATE(739)] = 23231,
  [SMALL_STATE(740)] = 23287,
  [SMALL_STATE(741)] = 23381,
  [SMALL_STATE(742)] = 23475,
  [SMALL_STATE(743)] = 23569,
  [SMALL_STATE(744)] = 23659,
  [SMALL_STATE(745)] = 23753,
  [SMALL_STATE(746)] = 23843,
  [SMALL_STATE(747)] = 23933,
  [SMALL_STATE(748)] = 23987,
  [SMALL_STATE(749)] = 24040,
  [SMALL_STATE(750)] = 24093,
  [SMALL_STATE(751)] = 24146,
  [SMALL_STATE(752)] = 24199,
  [SMALL_STATE(753)] = 24252,
  [SMALL_STATE(754)] = 24315,
  [SMALL_STATE(755)] = 24368,
  [SMALL_STATE(756)] = 24421,
  [SMALL_STATE(757)] = 24474,
  [SMALL_STATE(758)] = 24527,
  [SMALL_STATE(759)] = 24580,
  [SMALL_STATE(760)] = 24633,
  [SMALL_STATE(761)] = 24686,
  [SMALL_STATE(762)] = 24739,
  [SMALL_STATE(763)] = 24792,
  [SMALL_STATE(764)] = 24845,
  [SMALL_STATE(765)] = 24898,
  [SMALL_STATE(766)] = 24951,
  [SMALL_STATE(767)] = 25004,
  [SMALL_STATE(768)] = 25057,
  [SMALL_STATE(769)] = 25110,
  [SMALL_STATE(770)] = 25163,
  [SMALL_STATE(771)] = 25216,
  [SMALL_STATE(772)] = 25303,
  [SMALL_STATE(773)] = 25356,
  [SMALL_STATE(774)] = 25409,
  [SMALL_STATE(775)] = 25462,
  [SMALL_STATE(776)] = 25515,
  [SMALL_STATE(777)] = 25568,
  [SMALL_STATE(778)] = 25621,
  [SMALL_STATE(779)] = 25674,
  [SMALL_STATE(780)] = 25727,
  [SMALL_STATE(781)] = 25780,
  [SMALL_STATE(782)] = 25833,
  [SMALL_STATE(783)] = 25886,
  [SMALL_STATE(784)] = 25939,
  [SMALL_STATE(785)] = 25992,
  [SMALL_STATE(786)] = 26045,
  [SMALL_STATE(787)] = 26098,
  [SMALL_STATE(788)] = 26151,
  [SMALL_STATE(789)] = 26204,
  [SMALL_STATE(790)] = 26259,
  [SMALL_STATE(791)] = 26312,
  [SMALL_STATE(792)] = 26365,
  [SMALL_STATE(793)] = 26418,
  [SMALL_STATE(794)] = 26471,
  [SMALL_STATE(795)] = 26534,
  [SMALL_STATE(796)] = 26587,
  [SMALL_STATE(797)] = 26640,
  [SMALL_STATE(798)] = 26693,
  [SMALL_STATE(799)] = 26746,
  [SMALL_STATE(800)] = 26799,
  [SMALL_STATE(801)] = 26852,
  [SMALL_STATE(802)] = 26905,
  [SMALL_STATE(803)] = 26958,
  [SMALL_STATE(804)] = 27011,
  [SMALL_STATE(805)] = 27064,
  [SMALL_STATE(806)] = 27117,
  [SMALL_STATE(807)] = 27170,
  [SMALL_STATE(808)] = 27223,
  [SMALL_STATE(809)] = 27276,
  [SMALL_STATE(810)] = 27331,
  [SMALL_STATE(811)] = 27384,
  [SMALL_STATE(812)] = 27437,
  [SMALL_STATE(813)] = 27490,
  [SMALL_STATE(814)] = 27543,
  [SMALL_STATE(815)] = 27596,
  [SMALL_STATE(816)] = 27649,
  [SMALL_STATE(817)] = 27702,
  [SMALL_STATE(818)] = 27755,
  [SMALL_STATE(819)] = 27808,
  [SMALL_STATE(820)] = 27895,
  [SMALL_STATE(821)] = 27948,
  [SMALL_STATE(822)] = 28001,
  [SMALL_STATE(823)] = 28053,
  [SMALL_STATE(824)] = 28105,
  [SMALL_STATE(825)] = 28157,
  [SMALL_STATE(826)] = 28209,
  [SMALL_STATE(827)] = 28265,
  [SMALL_STATE(828)] = 28317,
  [SMALL_STATE(829)] = 28369,
  [SMALL_STATE(830)] = 28421,
  [SMALL_STATE(831)] = 28473,
  [SMALL_STATE(832)] = 28525,
  [SMALL_STATE(833)] = 28577,
  [SMALL_STATE(834)] = 28639,
  [SMALL_STATE(835)] = 28691,
  [SMALL_STATE(836)] = 28753,
  [SMALL_STATE(837)] = 28805,
  [SMALL_STATE(838)] = 28857,
  [SMALL_STATE(839)] = 28909,
  [SMALL_STATE(840)] = 28961,
  [SMALL_STATE(841)] = 29013,
  [SMALL_STATE(842)] = 29065,
  [SMALL_STATE(843)] = 29117,
  [SMALL_STATE(844)] = 29169,
  [SMALL_STATE(845)] = 29221,
  [SMALL_STATE(846)] = 29291,
  [SMALL_STATE(847)] = 29343,
  [SMALL_STATE(848)] = 29395,
  [SMALL_STATE(849)] = 29447,
  [SMALL_STATE(850)] = 29499,
  [SMALL_STATE(851)] = 29551,
  [SMALL_STATE(852)] = 29603,
  [SMALL_STATE(853)] = 29655,
  [SMALL_STATE(854)] = 29707,
  [SMALL_STATE(855)] = 29761,
  [SMALL_STATE(856)] = 29831,
  [SMALL_STATE(857)] = 29885,
  [SMALL_STATE(858)] = 29947,
  [SMALL_STATE(859)] = 29999,
  [SMALL_STATE(860)] = 30061,
  [SMALL_STATE(861)] = 30113,
  [SMALL_STATE(862)] = 30165,
  [SMALL_STATE(863)] = 30217,
  [SMALL_STATE(864)] = 30269,
  [SMALL_STATE(865)] = 30321,
  [SMALL_STATE(866)] = 30373,
  [SMALL_STATE(867)] = 30425,
  [SMALL_STATE(868)] = 30477,
  [SMALL_STATE(869)] = 30529,
  [SMALL_STATE(870)] = 30581,
  [SMALL_STATE(871)] = 30633,
  [SMALL_STATE(872)] = 30685,
  [SMALL_STATE(873)] = 30737,
  [SMALL_STATE(874)] = 30789,
  [SMALL_STATE(875)] = 30841,
  [SMALL_STATE(876)] = 30893,
  [SMALL_STATE(877)] = 30945,
  [SMALL_STATE(878)] = 30997,
  [SMALL_STATE(879)] = 31049,
  [SMALL_STATE(880)] = 31101,
  [SMALL_STATE(881)] = 31153,
  [SMALL_STATE(882)] = 31205,
  [SMALL_STATE(883)] = 31257,
  [SMALL_STATE(884)] = 31313,
  [SMALL_STATE(885)] = 31365,
  [SMALL_STATE(886)] = 31417,
  [SMALL_STATE(887)] = 31471,
  [SMALL_STATE(888)] = 31525,
  [SMALL_STATE(889)] = 31587,
  [SMALL_STATE(890)] = 31639,
  [SMALL_STATE(891)] = 31701,
  [SMALL_STATE(892)] = 31753,
  [SMALL_STATE(893)] = 31805,
  [SMALL_STATE(894)] = 31857,
  [SMALL_STATE(895)] = 31909,
  [SMALL_STATE(896)] = 31961,
  [SMALL_STATE(897)] = 32013,
  [SMALL_STATE(898)] = 32065,
  [SMALL_STATE(899)] = 32117,
  [SMALL_STATE(900)] = 32169,
  [SMALL_STATE(901)] = 32221,
  [SMALL_STATE(902)] = 32273,
  [SMALL_STATE(903)] = 32325,
  [SMALL_STATE(904)] = 32377,
  [SMALL_STATE(905)] = 32436,
  [SMALL_STATE(906)] = 32491,
  [SMALL_STATE(907)] = 32552,
  [SMALL_STATE(908)] = 32613,
  [SMALL_STATE(909)] = 32674,
  [SMALL_STATE(910)] = 32735,
  [SMALL_STATE(911)] = 32790,
  [SMALL_STATE(912)] = 32847,
  [SMALL_STATE(913)] = 32904,
  [SMALL_STATE(914)] = 32955,
  [SMALL_STATE(915)] = 33006,
  [SMALL_STATE(916)] = 33063,
  [SMALL_STATE(917)] = 33122,
  [SMALL_STATE(918)] = 33181,
  [SMALL_STATE(919)] = 33239,
  [SMALL_STATE(920)] = 33297,
  [SMALL_STATE(921)] = 33351,
  [SMALL_STATE(922)] = 33407,
  [SMALL_STATE(923)] = 33463,
  [SMALL_STATE(924)] = 33521,
  [SMALL_STATE(925)] = 33577,
  [SMALL_STATE(926)] = 33627,
  [SMALL_STATE(927)] = 33677,
  [SMALL_STATE(928)] = 33729,
  [SMALL_STATE(929)] = 33779,
  [SMALL_STATE(930)] = 33859,
  [SMALL_STATE(931)] = 33909,
  [SMALL_STATE(932)] = 33959,
  [SMALL_STATE(933)] = 34009,
  [SMALL_STATE(934)] = 34058,
  [SMALL_STATE(935)] = 34107,
  [SMALL_STATE(936)] = 34156,
  [SMALL_STATE(937)] = 34205,
  [SMALL_STATE(938)] = 34254,
  [SMALL_STATE(939)] = 34303,
  [SMALL_STATE(940)] = 34352,
  [SMALL_STATE(941)] = 34401,
  [SMALL_STATE(942)] = 34450,
  [SMALL_STATE(943)] = 34499,
  [SMALL_STATE(944)] = 34548,
  [SMALL_STATE(945)] = 34597,
  [SMALL_STATE(946)] = 34646,
  [SMALL_STATE(947)] = 34695,
  [SMALL_STATE(948)] = 34744,
  [SMALL_STATE(949)] = 34793,
  [SMALL_STATE(950)] = 34842,
  [SMALL_STATE(951)] = 34891,
  [SMALL_STATE(952)] = 34940,
  [SMALL_STATE(953)] = 34989,
  [SMALL_STATE(954)] = 35038,
  [SMALL_STATE(955)] = 35087,
  [SMALL_STATE(956)] = 35136,
  [SMALL_STATE(957)] = 35185,
  [SMALL_STATE(958)] = 35234,
  [SMALL_STATE(959)] = 35283,
  [SMALL_STATE(960)] = 35332,
  [SMALL_STATE(961)] = 35381,
  [SMALL_STATE(962)] = 35430,
  [SMALL_STATE(963)] = 35479,
  [SMALL_STATE(964)] = 35530,
  [SMALL_STATE(965)] = 35579,
  [SMALL_STATE(966)] = 35628,
  [SMALL_STATE(967)] = 35677,
  [SMALL_STATE(968)] = 35726,
  [SMALL_STATE(969)] = 35775,
  [SMALL_STATE(970)] = 35826,
  [SMALL_STATE(971)] = 35875,
  [SMALL_STATE(972)] = 35924,
  [SMALL_STATE(973)] = 35973,
  [SMALL_STATE(974)] = 36022,
  [SMALL_STATE(975)] = 36074,
  [SMALL_STATE(976)] = 36126,
  [SMALL_STATE(977)] = 36178,
  [SMALL_STATE(978)] = 36230,
  [SMALL_STATE(979)] = 36282,
  [SMALL_STATE(980)] = 36332,
  [SMALL_STATE(981)] = 36384,
  [SMALL_STATE(982)] = 36436,
  [SMALL_STATE(983)] = 36483,
  [SMALL_STATE(984)] = 36530,
  [SMALL_STATE(985)] = 36577,
  [SMALL_STATE(986)] = 36624,
  [SMALL_STATE(987)] = 36671,
  [SMALL_STATE(988)] = 36718,
  [SMALL_STATE(989)] = 36765,
  [SMALL_STATE(990)] = 36812,
  [SMALL_STATE(991)] = 36859,
  [SMALL_STATE(992)] = 36910,
  [SMALL_STATE(993)] = 36957,
  [SMALL_STATE(994)] = 37004,
  [SMALL_STATE(995)] = 37051,
  [SMALL_STATE(996)] = 37102,
  [SMALL_STATE(997)] = 37153,
  [SMALL_STATE(998)] = 37204,
  [SMALL_STATE(999)] = 37251,
  [SMALL_STATE(1000)] = 37298,
  [SMALL_STATE(1001)] = 37345,
  [SMALL_STATE(1002)] = 37392,
  [SMALL_STATE(1003)] = 37439,
  [SMALL_STATE(1004)] = 37490,
  [SMALL_STATE(1005)] = 37537,
  [SMALL_STATE(1006)] = 37588,
  [SMALL_STATE(1007)] = 37633,
  [SMALL_STATE(1008)] = 37680,
  [SMALL_STATE(1009)] = 37727,
  [SMALL_STATE(1010)] = 37774,
  [SMALL_STATE(1011)] = 37821,
  [SMALL_STATE(1012)] = 37868,
  [SMALL_STATE(1013)] = 37915,
  [SMALL_STATE(1014)] = 37962,
  [SMALL_STATE(1015)] = 38009,
  [SMALL_STATE(1016)] = 38056,
  [SMALL_STATE(1017)] = 38103,
  [SMALL_STATE(1018)] = 38150,
  [SMALL_STATE(1019)] = 38197,
  [SMALL_STATE(1020)] = 38244,
  [SMALL_STATE(1021)] = 38295,
  [SMALL_STATE(1022)] = 38342,
  [SMALL_STATE(1023)] = 38389,
  [SMALL_STATE(1024)] = 38436,
  [SMALL_STATE(1025)] = 38483,
  [SMALL_STATE(1026)] = 38530,
  [SMALL_STATE(1027)] = 38577,
  [SMALL_STATE(1028)] = 38624,
  [SMALL_STATE(1029)] = 38671,
  [SMALL_STATE(1030)] = 38718,
  [SMALL_STATE(1031)] = 38765,
  [SMALL_STATE(1032)] = 38812,
  [SMALL_STATE(1033)] = 38859,
  [SMALL_STATE(1034)] = 38906,
  [SMALL_STATE(1035)] = 38953,
  [SMALL_STATE(1036)] = 39000,
  [SMALL_STATE(1037)] = 39059,
  [SMALL_STATE(1038)] = 39118,
  [SMALL_STATE(1039)] = 39165,
  [SMALL_STATE(1040)] = 39216,
  [SMALL_STATE(1041)] = 39262,
  [SMALL_STATE(1042)] = 39308,
  [SMALL_STATE(1043)] = 39354,
  [SMALL_STATE(1044)] = 39400,
  [SMALL_STATE(1045)] = 39446,
  [SMALL_STATE(1046)] = 39492,
  [SMALL_STATE(1047)] = 39538,
  [SMALL_STATE(1048)] = 39584,
  [SMALL_STATE(1049)] = 39630,
  [SMALL_STATE(1050)] = 39676,
  [SMALL_STATE(1051)] = 39726,
  [SMALL_STATE(1052)] = 39772,
  [SMALL_STATE(1053)] = 39818,
  [SMALL_STATE(1054)] = 39864,
  [SMALL_STATE(1055)] = 39910,
  [SMALL_STATE(1056)] = 39956,
  [SMALL_STATE(1057)] = 40002,
  [SMALL_STATE(1058)] = 40048,
  [SMALL_STATE(1059)] = 40094,
  [SMALL_STATE(1060)] = 40140,
  [SMALL_STATE(1061)] = 40186,
  [SMALL_STATE(1062)] = 40232,
  [SMALL_STATE(1063)] = 40282,
  [SMALL_STATE(1064)] = 40328,
  [SMALL_STATE(1065)] = 40374,
  [SMALL_STATE(1066)] = 40420,
  [SMALL_STATE(1067)] = 40466,
  [SMALL_STATE(1068)] = 40512,
  [SMALL_STATE(1069)] = 40558,
  [SMALL_STATE(1070)] = 40604,
  [SMALL_STATE(1071)] = 40650,
  [SMALL_STATE(1072)] = 40708,
  [SMALL_STATE(1073)] = 40766,
  [SMALL_STATE(1074)] = 40812,
  [SMALL_STATE(1075)] = 40858,
  [SMALL_STATE(1076)] = 40904,
  [SMALL_STATE(1077)] = 40950,
  [SMALL_STATE(1078)] = 40996,
  [SMALL_STATE(1079)] = 41042,
  [SMALL_STATE(1080)] = 41088,
  [SMALL_STATE(1081)] = 41134,
  [SMALL_STATE(1082)] = 41180,
  [SMALL_STATE(1083)] = 41225,
  [SMALL_STATE(1084)] = 41270,
  [SMALL_STATE(1085)] = 41315,
  [SMALL_STATE(1086)] = 41360,
  [SMALL_STATE(1087)] = 41409,
  [SMALL_STATE(1088)] = 41454,
  [SMALL_STATE(1089)] = 41499,
  [SMALL_STATE(1090)] = 41544,
  [SMALL_STATE(1091)] = 41589,
  [SMALL_STATE(1092)] = 41634,
  [SMALL_STATE(1093)] = 41679,
  [SMALL_STATE(1094)] = 41724,
  [SMALL_STATE(1095)] = 41781,
  [SMALL_STATE(1096)] = 41842,
  [SMALL_STATE(1097)] = 41899,
  [SMALL_STATE(1098)] = 41944,
  [SMALL_STATE(1099)] = 41989,
  [SMALL_STATE(1100)] = 42034,
  [SMALL_STATE(1101)] = 42083,
  [SMALL_STATE(1102)] = 42128,
  [SMALL_STATE(1103)] = 42173,
  [SMALL_STATE(1104)] = 42218,
  [SMALL_STATE(1105)] = 42263,
  [SMALL_STATE(1106)] = 42308,
  [SMALL_STATE(1107)] = 42353,
  [SMALL_STATE(1108)] = 42398,
  [SMALL_STATE(1109)] = 42443,
  [SMALL_STATE(1110)] = 42488,
  [SMALL_STATE(1111)] = 42533,
  [SMALL_STATE(1112)] = 42578,
  [SMALL_STATE(1113)] = 42623,
  [SMALL_STATE(1114)] = 42684,
  [SMALL_STATE(1115)] = 42739,
  [SMALL_STATE(1116)] = 42794,
  [SMALL_STATE(1117)] = 42849,
  [SMALL_STATE(1118)] = 42904,
  [SMALL_STATE(1119)] = 42946,
  [SMALL_STATE(1120)] = 42992,
  [SMALL_STATE(1121)] = 43038,
  [SMALL_STATE(1122)] = 43084,
  [SMALL_STATE(1123)] = 43129,
  [SMALL_STATE(1124)] = 43174,
  [SMALL_STATE(1125)] = 43219,
  [SMALL_STATE(1126)] = 43264,
  [SMALL_STATE(1127)] = 43309,
  [SMALL_STATE(1128)] = 43354,
  [SMALL_STATE(1129)] = 43393,
  [SMALL_STATE(1130)] = 43432,
  [SMALL_STATE(1131)] = 43471,
  [SMALL_STATE(1132)] = 43510,
  [SMALL_STATE(1133)] = 43549,
  [SMALL_STATE(1134)] = 43588,
  [SMALL_STATE(1135)] = 43627,
  [SMALL_STATE(1136)] = 43666,
  [SMALL_STATE(1137)] = 43705,
  [SMALL_STATE(1138)] = 43744,
  [SMALL_STATE(1139)] = 43782,
  [SMALL_STATE(1140)] = 43820,
  [SMALL_STATE(1141)] = 43854,
  [SMALL_STATE(1142)] = 43888,
  [SMALL_STATE(1143)] = 43922,
  [SMALL_STATE(1144)] = 43959,
  [SMALL_STATE(1145)] = 43996,
  [SMALL_STATE(1146)] = 44026,
  [SMALL_STATE(1147)] = 44074,
  [SMALL_STATE(1148)] = 44122,
  [SMALL_STATE(1149)] = 44170,
  [SMALL_STATE(1150)] = 44218,
  [SMALL_STATE(1151)] = 44266,
  [SMALL_STATE(1152)] = 44314,
  [SMALL_STATE(1153)] = 44362,
  [SMALL_STATE(1154)] = 44410,
  [SMALL_STATE(1155)] = 44458,
  [SMALL_STATE(1156)] = 44506,
  [SMALL_STATE(1157)] = 44554,
  [SMALL_STATE(1158)] = 44602,
  [SMALL_STATE(1159)] = 44650,
  [SMALL_STATE(1160)] = 44698,
  [SMALL_STATE(1161)] = 44746,
  [SMALL_STATE(1162)] = 44794,
  [SMALL_STATE(1163)] = 44842,
  [SMALL_STATE(1164)] = 44890,
  [SMALL_STATE(1165)] = 44938,
  [SMALL_STATE(1166)] = 44986,
  [SMALL_STATE(1167)] = 45034,
  [SMALL_STATE(1168)] = 45082,
  [SMALL_STATE(1169)] = 45130,
  [SMALL_STATE(1170)] = 45178,
  [SMALL_STATE(1171)] = 45226,
  [SMALL_STATE(1172)] = 45265,
  [SMALL_STATE(1173)] = 45304,
  [SMALL_STATE(1174)] = 45343,
  [SMALL_STATE(1175)] = 45382,
  [SMALL_STATE(1176)] = 45421,
  [SMALL_STATE(1177)] = 45459,
  [SMALL_STATE(1178)] = 45497,
  [SMALL_STATE(1179)] = 45533,
  [SMALL_STATE(1180)] = 45571,
  [SMALL_STATE(1181)] = 45607,
  [SMALL_STATE(1182)] = 45645,
  [SMALL_STATE(1183)] = 45681,
  [SMALL_STATE(1184)] = 45719,
  [SMALL_STATE(1185)] = 45755,
  [SMALL_STATE(1186)] = 45793,
  [SMALL_STATE(1187)] = 45831,
  [SMALL_STATE(1188)] = 45869,
  [SMALL_STATE(1189)] = 45905,
  [SMALL_STATE(1190)] = 45943,
  [SMALL_STATE(1191)] = 45981,
  [SMALL_STATE(1192)] = 46019,
  [SMALL_STATE(1193)] = 46057,
  [SMALL_STATE(1194)] = 46095,
  [SMALL_STATE(1195)] = 46133,
  [SMALL_STATE(1196)] = 46171,
  [SMALL_STATE(1197)] = 46209,
  [SMALL_STATE(1198)] = 46242,
  [SMALL_STATE(1199)] = 46275,
  [SMALL_STATE(1200)] = 46308,
  [SMALL_STATE(1201)] = 46339,
  [SMALL_STATE(1202)] = 46372,
  [SMALL_STATE(1203)] = 46405,
  [SMALL_STATE(1204)] = 46438,
  [SMALL_STATE(1205)] = 46471,
  [SMALL_STATE(1206)] = 46504,
  [SMALL_STATE(1207)] = 46537,
  [SMALL_STATE(1208)] = 46570,
  [SMALL_STATE(1209)] = 46603,
  [SMALL_STATE(1210)] = 46636,
  [SMALL_STATE(1211)] = 46669,
  [SMALL_STATE(1212)] = 46702,
  [SMALL_STATE(1213)] = 46735,
  [SMALL_STATE(1214)] = 46768,
  [SMALL_STATE(1215)] = 46801,
  [SMALL_STATE(1216)] = 46834,
  [SMALL_STATE(1217)] = 46874,
  [SMALL_STATE(1218)] = 46914,
  [SMALL_STATE(1219)] = 46954,
  [SMALL_STATE(1220)] = 46994,
  [SMALL_STATE(1221)] = 47034,
  [SMALL_STATE(1222)] = 47074,
  [SMALL_STATE(1223)] = 47114,
  [SMALL_STATE(1224)] = 47144,
  [SMALL_STATE(1225)] = 47168,
  [SMALL_STATE(1226)] = 47203,
  [SMALL_STATE(1227)] = 47238,
  [SMALL_STATE(1228)] = 47273,
  [SMALL_STATE(1229)] = 47299,
  [SMALL_STATE(1230)] = 47321,
  [SMALL_STATE(1231)] = 47345,
  [SMALL_STATE(1232)] = 47365,
  [SMALL_STATE(1233)] = 47391,
  [SMALL_STATE(1234)] = 47413,
  [SMALL_STATE(1235)] = 47447,
  [SMALL_STATE(1236)] = 47469,
  [SMALL_STATE(1237)] = 47489,
  [SMALL_STATE(1238)] = 47515,
  [SMALL_STATE(1239)] = 47537,
  [SMALL_STATE(1240)] = 47559,
  [SMALL_STATE(1241)] = 47579,
  [SMALL_STATE(1242)] = 47601,
  [SMALL_STATE(1243)] = 47626,
  [SMALL_STATE(1244)] = 47659,
  [SMALL_STATE(1245)] = 47692,
  [SMALL_STATE(1246)] = 47725,
  [SMALL_STATE(1247)] = 47758,
  [SMALL_STATE(1248)] = 47784,
  [SMALL_STATE(1249)] = 47810,
  [SMALL_STATE(1250)] = 47830,
  [SMALL_STATE(1251)] = 47856,
  [SMALL_STATE(1252)] = 47876,
  [SMALL_STATE(1253)] = 47902,
  [SMALL_STATE(1254)] = 47928,
  [SMALL_STATE(1255)] = 47948,
  [SMALL_STATE(1256)] = 47968,
  [SMALL_STATE(1257)] = 47988,
  [SMALL_STATE(1258)] = 48008,
  [SMALL_STATE(1259)] = 48028,
  [SMALL_STATE(1260)] = 48048,
  [SMALL_STATE(1261)] = 48075,
  [SMALL_STATE(1262)] = 48094,
  [SMALL_STATE(1263)] = 48121,
  [SMALL_STATE(1264)] = 48148,
  [SMALL_STATE(1265)] = 48167,
  [SMALL_STATE(1266)] = 48194,
  [SMALL_STATE(1267)] = 48213,
  [SMALL_STATE(1268)] = 48240,
  [SMALL_STATE(1269)] = 48267,
  [SMALL_STATE(1270)] = 48294,
  [SMALL_STATE(1271)] = 48321,
  [SMALL_STATE(1272)] = 48348,
  [SMALL_STATE(1273)] = 48366,
  [SMALL_STATE(1274)] = 48394,
  [SMALL_STATE(1275)] = 48413,
  [SMALL_STATE(1276)] = 48434,
  [SMALL_STATE(1277)] = 48455,
  [SMALL_STATE(1278)] = 48472,
  [SMALL_STATE(1279)] = 48489,
  [SMALL_STATE(1280)] = 48508,
  [SMALL_STATE(1281)] = 48525,
  [SMALL_STATE(1282)] = 48543,
  [SMALL_STATE(1283)] = 48561,
  [SMALL_STATE(1284)] = 48579,
  [SMALL_STATE(1285)] = 48599,
  [SMALL_STATE(1286)] = 48625,
  [SMALL_STATE(1287)] = 48641,
  [SMALL_STATE(1288)] = 48657,
  [SMALL_STATE(1289)] = 48677,
  [SMALL_STATE(1290)] = 48695,
  [SMALL_STATE(1291)] = 48713,
  [SMALL_STATE(1292)] = 48731,
  [SMALL_STATE(1293)] = 48747,
  [SMALL_STATE(1294)] = 48763,
  [SMALL_STATE(1295)] = 48781,
  [SMALL_STATE(1296)] = 48807,
  [SMALL_STATE(1297)] = 48825,
  [SMALL_STATE(1298)] = 48845,
  [SMALL_STATE(1299)] = 48865,
  [SMALL_STATE(1300)] = 48883,
  [SMALL_STATE(1301)] = 48901,
  [SMALL_STATE(1302)] = 48919,
  [SMALL_STATE(1303)] = 48937,
  [SMALL_STATE(1304)] = 48958,
  [SMALL_STATE(1305)] = 48979,
  [SMALL_STATE(1306)] = 49004,
  [SMALL_STATE(1307)] = 49025,
  [SMALL_STATE(1308)] = 49040,
  [SMALL_STATE(1309)] = 49063,
  [SMALL_STATE(1310)] = 49076,
  [SMALL_STATE(1311)] = 49099,
  [SMALL_STATE(1312)] = 49114,
  [SMALL_STATE(1313)] = 49139,
  [SMALL_STATE(1314)] = 49162,
  [SMALL_STATE(1315)] = 49187,
  [SMALL_STATE(1316)] = 49202,
  [SMALL_STATE(1317)] = 49225,
  [SMALL_STATE(1318)] = 49246,
  [SMALL_STATE(1319)] = 49271,
  [SMALL_STATE(1320)] = 49292,
  [SMALL_STATE(1321)] = 49307,
  [SMALL_STATE(1322)] = 49329,
  [SMALL_STATE(1323)] = 49351,
  [SMALL_STATE(1324)] = 49365,
  [SMALL_STATE(1325)] = 49387,
  [SMALL_STATE(1326)] = 49409,
  [SMALL_STATE(1327)] = 49423,
  [SMALL_STATE(1328)] = 49437,
  [SMALL_STATE(1329)] = 49459,
  [SMALL_STATE(1330)] = 49481,
  [SMALL_STATE(1331)] = 49503,
  [SMALL_STATE(1332)] = 49525,
  [SMALL_STATE(1333)] = 49542,
  [SMALL_STATE(1334)] = 49561,
  [SMALL_STATE(1335)] = 49572,
  [SMALL_STATE(1336)] = 49591,
  [SMALL_STATE(1337)] = 49610,
  [SMALL_STATE(1338)] = 49623,
  [SMALL_STATE(1339)] = 49638,
  [SMALL_STATE(1340)] = 49649,
  [SMALL_STATE(1341)] = 49668,
  [SMALL_STATE(1342)] = 49683,
  [SMALL_STATE(1343)] = 49700,
  [SMALL_STATE(1344)] = 49719,
  [SMALL_STATE(1345)] = 49730,
  [SMALL_STATE(1346)] = 49743,
  [SMALL_STATE(1347)] = 49762,
  [SMALL_STATE(1348)] = 49779,
  [SMALL_STATE(1349)] = 49796,
  [SMALL_STATE(1350)] = 49815,
  [SMALL_STATE(1351)] = 49834,
  [SMALL_STATE(1352)] = 49847,
  [SMALL_STATE(1353)] = 49858,
  [SMALL_STATE(1354)] = 49871,
  [SMALL_STATE(1355)] = 49890,
  [SMALL_STATE(1356)] = 49903,
  [SMALL_STATE(1357)] = 49922,
  [SMALL_STATE(1358)] = 49941,
  [SMALL_STATE(1359)] = 49960,
  [SMALL_STATE(1360)] = 49979,
  [SMALL_STATE(1361)] = 49996,
  [SMALL_STATE(1362)] = 50015,
  [SMALL_STATE(1363)] = 50034,
  [SMALL_STATE(1364)] = 50053,
  [SMALL_STATE(1365)] = 50072,
  [SMALL_STATE(1366)] = 50091,
  [SMALL_STATE(1367)] = 50108,
  [SMALL_STATE(1368)] = 50123,
  [SMALL_STATE(1369)] = 50138,
  [SMALL_STATE(1370)] = 50157,
  [SMALL_STATE(1371)] = 50168,
  [SMALL_STATE(1372)] = 50185,
  [SMALL_STATE(1373)] = 50200,
  [SMALL_STATE(1374)] = 50213,
  [SMALL_STATE(1375)] = 50228,
  [SMALL_STATE(1376)] = 50243,
  [SMALL_STATE(1377)] = 50256,
  [SMALL_STATE(1378)] = 50271,
  [SMALL_STATE(1379)] = 50286,
  [SMALL_STATE(1380)] = 50299,
  [SMALL_STATE(1381)] = 50314,
  [SMALL_STATE(1382)] = 50333,
  [SMALL_STATE(1383)] = 50348,
  [SMALL_STATE(1384)] = 50363,
  [SMALL_STATE(1385)] = 50378,
  [SMALL_STATE(1386)] = 50395,
  [SMALL_STATE(1387)] = 50412,
  [SMALL_STATE(1388)] = 50429,
  [SMALL_STATE(1389)] = 50448,
  [SMALL_STATE(1390)] = 50467,
  [SMALL_STATE(1391)] = 50486,
  [SMALL_STATE(1392)] = 50500,
  [SMALL_STATE(1393)] = 50512,
  [SMALL_STATE(1394)] = 50524,
  [SMALL_STATE(1395)] = 50540,
  [SMALL_STATE(1396)] = 50556,
  [SMALL_STATE(1397)] = 50572,
  [SMALL_STATE(1398)] = 50588,
  [SMALL_STATE(1399)] = 50600,
  [SMALL_STATE(1400)] = 50612,
  [SMALL_STATE(1401)] = 50626,
  [SMALL_STATE(1402)] = 50642,
  [SMALL_STATE(1403)] = 50658,
  [SMALL_STATE(1404)] = 50670,
  [SMALL_STATE(1405)] = 50684,
  [SMALL_STATE(1406)] = 50696,
  [SMALL_STATE(1407)] = 50712,
  [SMALL_STATE(1408)] = 50728,
  [SMALL_STATE(1409)] = 50744,
  [SMALL_STATE(1410)] = 50760,
  [SMALL_STATE(1411)] = 50776,
  [SMALL_STATE(1412)] = 50790,
  [SMALL_STATE(1413)] = 50806,
  [SMALL_STATE(1414)] = 50822,
  [SMALL_STATE(1415)] = 50838,
  [SMALL_STATE(1416)] = 50854,
  [SMALL_STATE(1417)] = 50870,
  [SMALL_STATE(1418)] = 50886,
  [SMALL_STATE(1419)] = 50902,
  [SMALL_STATE(1420)] = 50915,
  [SMALL_STATE(1421)] = 50928,
  [SMALL_STATE(1422)] = 50941,
  [SMALL_STATE(1423)] = 50954,
  [SMALL_STATE(1424)] = 50963,
  [SMALL_STATE(1425)] = 50976,
  [SMALL_STATE(1426)] = 50989,
  [SMALL_STATE(1427)] = 51002,
  [SMALL_STATE(1428)] = 51015,
  [SMALL_STATE(1429)] = 51028,
  [SMALL_STATE(1430)] = 51041,
  [SMALL_STATE(1431)] = 51054,
  [SMALL_STATE(1432)] = 51067,
  [SMALL_STATE(1433)] = 51080,
  [SMALL_STATE(1434)] = 51093,
  [SMALL_STATE(1435)] = 51106,
  [SMALL_STATE(1436)] = 51119,
  [SMALL_STATE(1437)] = 51132,
  [SMALL_STATE(1438)] = 51145,
  [SMALL_STATE(1439)] = 51158,
  [SMALL_STATE(1440)] = 51171,
  [SMALL_STATE(1441)] = 51184,
  [SMALL_STATE(1442)] = 51197,
  [SMALL_STATE(1443)] = 51210,
  [SMALL_STATE(1444)] = 51223,
  [SMALL_STATE(1445)] = 51236,
  [SMALL_STATE(1446)] = 51249,
  [SMALL_STATE(1447)] = 51262,
  [SMALL_STATE(1448)] = 51275,
  [SMALL_STATE(1449)] = 51284,
  [SMALL_STATE(1450)] = 51297,
  [SMALL_STATE(1451)] = 51310,
  [SMALL_STATE(1452)] = 51323,
  [SMALL_STATE(1453)] = 51336,
  [SMALL_STATE(1454)] = 51349,
  [SMALL_STATE(1455)] = 51362,
  [SMALL_STATE(1456)] = 51371,
  [SMALL_STATE(1457)] = 51384,
  [SMALL_STATE(1458)] = 51397,
  [SMALL_STATE(1459)] = 51410,
  [SMALL_STATE(1460)] = 51423,
  [SMALL_STATE(1461)] = 51436,
  [SMALL_STATE(1462)] = 51449,
  [SMALL_STATE(1463)] = 51462,
  [SMALL_STATE(1464)] = 51475,
  [SMALL_STATE(1465)] = 51488,
  [SMALL_STATE(1466)] = 51501,
  [SMALL_STATE(1467)] = 51514,
  [SMALL_STATE(1468)] = 51527,
  [SMALL_STATE(1469)] = 51540,
  [SMALL_STATE(1470)] = 51553,
  [SMALL_STATE(1471)] = 51566,
  [SMALL_STATE(1472)] = 51579,
  [SMALL_STATE(1473)] = 51592,
  [SMALL_STATE(1474)] = 51605,
  [SMALL_STATE(1475)] = 51618,
  [SMALL_STATE(1476)] = 51631,
  [SMALL_STATE(1477)] = 51640,
  [SMALL_STATE(1478)] = 51649,
  [SMALL_STATE(1479)] = 51662,
  [SMALL_STATE(1480)] = 51675,
  [SMALL_STATE(1481)] = 51686,
  [SMALL_STATE(1482)] = 51699,
  [SMALL_STATE(1483)] = 51712,
  [SMALL_STATE(1484)] = 51721,
  [SMALL_STATE(1485)] = 51734,
  [SMALL_STATE(1486)] = 51747,
  [SMALL_STATE(1487)] = 51760,
  [SMALL_STATE(1488)] = 51773,
  [SMALL_STATE(1489)] = 51786,
  [SMALL_STATE(1490)] = 51799,
  [SMALL_STATE(1491)] = 51812,
  [SMALL_STATE(1492)] = 51825,
  [SMALL_STATE(1493)] = 51838,
  [SMALL_STATE(1494)] = 51847,
  [SMALL_STATE(1495)] = 51860,
  [SMALL_STATE(1496)] = 51873,
  [SMALL_STATE(1497)] = 51886,
  [SMALL_STATE(1498)] = 51899,
  [SMALL_STATE(1499)] = 51912,
  [SMALL_STATE(1500)] = 51925,
  [SMALL_STATE(1501)] = 51938,
  [SMALL_STATE(1502)] = 51949,
  [SMALL_STATE(1503)] = 51962,
  [SMALL_STATE(1504)] = 51975,
  [SMALL_STATE(1505)] = 51984,
  [SMALL_STATE(1506)] = 51997,
  [SMALL_STATE(1507)] = 52010,
  [SMALL_STATE(1508)] = 52023,
  [SMALL_STATE(1509)] = 52036,
  [SMALL_STATE(1510)] = 52047,
  [SMALL_STATE(1511)] = 52060,
  [SMALL_STATE(1512)] = 52073,
  [SMALL_STATE(1513)] = 52086,
  [SMALL_STATE(1514)] = 52097,
  [SMALL_STATE(1515)] = 52110,
  [SMALL_STATE(1516)] = 52123,
  [SMALL_STATE(1517)] = 52136,
  [SMALL_STATE(1518)] = 52149,
  [SMALL_STATE(1519)] = 52162,
  [SMALL_STATE(1520)] = 52175,
  [SMALL_STATE(1521)] = 52183,
  [SMALL_STATE(1522)] = 52193,
  [SMALL_STATE(1523)] = 52203,
  [SMALL_STATE(1524)] = 52213,
  [SMALL_STATE(1525)] = 52223,
  [SMALL_STATE(1526)] = 52233,
  [SMALL_STATE(1527)] = 52243,
  [SMALL_STATE(1528)] = 52253,
  [SMALL_STATE(1529)] = 52263,
  [SMALL_STATE(1530)] = 52273,
  [SMALL_STATE(1531)] = 52283,
  [SMALL_STATE(1532)] = 52291,
  [SMALL_STATE(1533)] = 52301,
  [SMALL_STATE(1534)] = 52311,
  [SMALL_STATE(1535)] = 52319,
  [SMALL_STATE(1536)] = 52329,
  [SMALL_STATE(1537)] = 52339,
  [SMALL_STATE(1538)] = 52349,
  [SMALL_STATE(1539)] = 52359,
  [SMALL_STATE(1540)] = 52369,
  [SMALL_STATE(1541)] = 52379,
  [SMALL_STATE(1542)] = 52389,
  [SMALL_STATE(1543)] = 52399,
  [SMALL_STATE(1544)] = 52409,
  [SMALL_STATE(1545)] = 52419,
  [SMALL_STATE(1546)] = 52429,
  [SMALL_STATE(1547)] = 52439,
  [SMALL_STATE(1548)] = 52449,
  [SMALL_STATE(1549)] = 52459,
  [SMALL_STATE(1550)] = 52469,
  [SMALL_STATE(1551)] = 52479,
  [SMALL_STATE(1552)] = 52489,
  [SMALL_STATE(1553)] = 52499,
  [SMALL_STATE(1554)] = 52509,
  [SMALL_STATE(1555)] = 52519,
  [SMALL_STATE(1556)] = 52529,
  [SMALL_STATE(1557)] = 52539,
  [SMALL_STATE(1558)] = 52549,
  [SMALL_STATE(1559)] = 52557,
  [SMALL_STATE(1560)] = 52567,
  [SMALL_STATE(1561)] = 52577,
  [SMALL_STATE(1562)] = 52587,
  [SMALL_STATE(1563)] = 52597,
  [SMALL_STATE(1564)] = 52607,
  [SMALL_STATE(1565)] = 52617,
  [SMALL_STATE(1566)] = 52625,
  [SMALL_STATE(1567)] = 52635,
  [SMALL_STATE(1568)] = 52645,
  [SMALL_STATE(1569)] = 52653,
  [SMALL_STATE(1570)] = 52661,
  [SMALL_STATE(1571)] = 52671,
  [SMALL_STATE(1572)] = 52681,
  [SMALL_STATE(1573)] = 52691,
  [SMALL_STATE(1574)] = 52701,
  [SMALL_STATE(1575)] = 52711,
  [SMALL_STATE(1576)] = 52721,
  [SMALL_STATE(1577)] = 52731,
  [SMALL_STATE(1578)] = 52741,
  [SMALL_STATE(1579)] = 52751,
  [SMALL_STATE(1580)] = 52761,
  [SMALL_STATE(1581)] = 52771,
  [SMALL_STATE(1582)] = 52781,
  [SMALL_STATE(1583)] = 52791,
  [SMALL_STATE(1584)] = 52801,
  [SMALL_STATE(1585)] = 52811,
  [SMALL_STATE(1586)] = 52819,
  [SMALL_STATE(1587)] = 52827,
  [SMALL_STATE(1588)] = 52837,
  [SMALL_STATE(1589)] = 52847,
  [SMALL_STATE(1590)] = 52857,
  [SMALL_STATE(1591)] = 52867,
  [SMALL_STATE(1592)] = 52877,
  [SMALL_STATE(1593)] = 52887,
  [SMALL_STATE(1594)] = 52897,
  [SMALL_STATE(1595)] = 52907,
  [SMALL_STATE(1596)] = 52917,
  [SMALL_STATE(1597)] = 52927,
  [SMALL_STATE(1598)] = 52937,
  [SMALL_STATE(1599)] = 52947,
  [SMALL_STATE(1600)] = 52957,
  [SMALL_STATE(1601)] = 52967,
  [SMALL_STATE(1602)] = 52977,
  [SMALL_STATE(1603)] = 52987,
  [SMALL_STATE(1604)] = 52997,
  [SMALL_STATE(1605)] = 53007,
  [SMALL_STATE(1606)] = 53017,
  [SMALL_STATE(1607)] = 53027,
  [SMALL_STATE(1608)] = 53037,
  [SMALL_STATE(1609)] = 53047,
  [SMALL_STATE(1610)] = 53057,
  [SMALL_STATE(1611)] = 53067,
  [SMALL_STATE(1612)] = 53077,
  [SMALL_STATE(1613)] = 53087,
  [SMALL_STATE(1614)] = 53097,
  [SMALL_STATE(1615)] = 53107,
  [SMALL_STATE(1616)] = 53117,
  [SMALL_STATE(1617)] = 53127,
  [SMALL_STATE(1618)] = 53135,
  [SMALL_STATE(1619)] = 53145,
  [SMALL_STATE(1620)] = 53155,
  [SMALL_STATE(1621)] = 53165,
  [SMALL_STATE(1622)] = 53175,
  [SMALL_STATE(1623)] = 53183,
  [SMALL_STATE(1624)] = 53193,
  [SMALL_STATE(1625)] = 53201,
  [SMALL_STATE(1626)] = 53211,
  [SMALL_STATE(1627)] = 53221,
  [SMALL_STATE(1628)] = 53231,
  [SMALL_STATE(1629)] = 53239,
  [SMALL_STATE(1630)] = 53249,
  [SMALL_STATE(1631)] = 53259,
  [SMALL_STATE(1632)] = 53269,
  [SMALL_STATE(1633)] = 53277,
  [SMALL_STATE(1634)] = 53287,
  [SMALL_STATE(1635)] = 53297,
  [SMALL_STATE(1636)] = 53307,
  [SMALL_STATE(1637)] = 53317,
  [SMALL_STATE(1638)] = 53327,
  [SMALL_STATE(1639)] = 53337,
  [SMALL_STATE(1640)] = 53347,
  [SMALL_STATE(1641)] = 53357,
  [SMALL_STATE(1642)] = 53367,
  [SMALL_STATE(1643)] = 53377,
  [SMALL_STATE(1644)] = 53387,
  [SMALL_STATE(1645)] = 53397,
  [SMALL_STATE(1646)] = 53405,
  [SMALL_STATE(1647)] = 53415,
  [SMALL_STATE(1648)] = 53423,
  [SMALL_STATE(1649)] = 53433,
  [SMALL_STATE(1650)] = 53443,
  [SMALL_STATE(1651)] = 53453,
  [SMALL_STATE(1652)] = 53463,
  [SMALL_STATE(1653)] = 53473,
  [SMALL_STATE(1654)] = 53483,
  [SMALL_STATE(1655)] = 53493,
  [SMALL_STATE(1656)] = 53503,
  [SMALL_STATE(1657)] = 53513,
  [SMALL_STATE(1658)] = 53523,
  [SMALL_STATE(1659)] = 53533,
  [SMALL_STATE(1660)] = 53543,
  [SMALL_STATE(1661)] = 53553,
  [SMALL_STATE(1662)] = 53563,
  [SMALL_STATE(1663)] = 53573,
  [SMALL_STATE(1664)] = 53583,
  [SMALL_STATE(1665)] = 53593,
  [SMALL_STATE(1666)] = 53603,
  [SMALL_STATE(1667)] = 53613,
  [SMALL_STATE(1668)] = 53623,
  [SMALL_STATE(1669)] = 53633,
  [SMALL_STATE(1670)] = 53641,
  [SMALL_STATE(1671)] = 53651,
  [SMALL_STATE(1672)] = 53661,
  [SMALL_STATE(1673)] = 53669,
  [SMALL_STATE(1674)] = 53679,
  [SMALL_STATE(1675)] = 53689,
  [SMALL_STATE(1676)] = 53699,
  [SMALL_STATE(1677)] = 53709,
  [SMALL_STATE(1678)] = 53719,
  [SMALL_STATE(1679)] = 53727,
  [SMALL_STATE(1680)] = 53737,
  [SMALL_STATE(1681)] = 53747,
  [SMALL_STATE(1682)] = 53757,
  [SMALL_STATE(1683)] = 53767,
  [SMALL_STATE(1684)] = 53775,
  [SMALL_STATE(1685)] = 53785,
  [SMALL_STATE(1686)] = 53795,
  [SMALL_STATE(1687)] = 53805,
  [SMALL_STATE(1688)] = 53813,
  [SMALL_STATE(1689)] = 53823,
  [SMALL_STATE(1690)] = 53833,
  [SMALL_STATE(1691)] = 53843,
  [SMALL_STATE(1692)] = 53851,
  [SMALL_STATE(1693)] = 53859,
  [SMALL_STATE(1694)] = 53869,
  [SMALL_STATE(1695)] = 53879,
  [SMALL_STATE(1696)] = 53889,
  [SMALL_STATE(1697)] = 53899,
  [SMALL_STATE(1698)] = 53909,
  [SMALL_STATE(1699)] = 53916,
  [SMALL_STATE(1700)] = 53923,
  [SMALL_STATE(1701)] = 53930,
  [SMALL_STATE(1702)] = 53937,
  [SMALL_STATE(1703)] = 53944,
  [SMALL_STATE(1704)] = 53951,
  [SMALL_STATE(1705)] = 53958,
  [SMALL_STATE(1706)] = 53965,
  [SMALL_STATE(1707)] = 53972,
  [SMALL_STATE(1708)] = 53979,
  [SMALL_STATE(1709)] = 53986,
  [SMALL_STATE(1710)] = 53993,
  [SMALL_STATE(1711)] = 54000,
  [SMALL_STATE(1712)] = 54007,
  [SMALL_STATE(1713)] = 54014,
  [SMALL_STATE(1714)] = 54021,
  [SMALL_STATE(1715)] = 54028,
  [SMALL_STATE(1716)] = 54035,
  [SMALL_STATE(1717)] = 54042,
  [SMALL_STATE(1718)] = 54049,
  [SMALL_STATE(1719)] = 54056,
  [SMALL_STATE(1720)] = 54063,
  [SMALL_STATE(1721)] = 54070,
  [SMALL_STATE(1722)] = 54077,
  [SMALL_STATE(1723)] = 54084,
  [SMALL_STATE(1724)] = 54091,
  [SMALL_STATE(1725)] = 54098,
  [SMALL_STATE(1726)] = 54105,
  [SMALL_STATE(1727)] = 54112,
  [SMALL_STATE(1728)] = 54119,
  [SMALL_STATE(1729)] = 54126,
  [SMALL_STATE(1730)] = 54133,
  [SMALL_STATE(1731)] = 54140,
  [SMALL_STATE(1732)] = 54147,
  [SMALL_STATE(1733)] = 54154,
  [SMALL_STATE(1734)] = 54161,
  [SMALL_STATE(1735)] = 54168,
  [SMALL_STATE(1736)] = 54175,
  [SMALL_STATE(1737)] = 54182,
  [SMALL_STATE(1738)] = 54189,
  [SMALL_STATE(1739)] = 54196,
  [SMALL_STATE(1740)] = 54203,
  [SMALL_STATE(1741)] = 54210,
  [SMALL_STATE(1742)] = 54217,
  [SMALL_STATE(1743)] = 54224,
  [SMALL_STATE(1744)] = 54231,
  [SMALL_STATE(1745)] = 54238,
  [SMALL_STATE(1746)] = 54245,
  [SMALL_STATE(1747)] = 54252,
  [SMALL_STATE(1748)] = 54259,
  [SMALL_STATE(1749)] = 54266,
  [SMALL_STATE(1750)] = 54273,
  [SMALL_STATE(1751)] = 54280,
  [SMALL_STATE(1752)] = 54287,
  [SMALL_STATE(1753)] = 54294,
  [SMALL_STATE(1754)] = 54301,
  [SMALL_STATE(1755)] = 54308,
  [SMALL_STATE(1756)] = 54315,
  [SMALL_STATE(1757)] = 54322,
  [SMALL_STATE(1758)] = 54329,
  [SMALL_STATE(1759)] = 54336,
  [SMALL_STATE(1760)] = 54343,
  [SMALL_STATE(1761)] = 54350,
  [SMALL_STATE(1762)] = 54357,
  [SMALL_STATE(1763)] = 54364,
  [SMALL_STATE(1764)] = 54371,
  [SMALL_STATE(1765)] = 54378,
  [SMALL_STATE(1766)] = 54385,
  [SMALL_STATE(1767)] = 54392,
  [SMALL_STATE(1768)] = 54399,
  [SMALL_STATE(1769)] = 54406,
  [SMALL_STATE(1770)] = 54413,
  [SMALL_STATE(1771)] = 54420,
  [SMALL_STATE(1772)] = 54427,
  [SMALL_STATE(1773)] = 54434,
  [SMALL_STATE(1774)] = 54441,
  [SMALL_STATE(1775)] = 54448,
  [SMALL_STATE(1776)] = 54455,
  [SMALL_STATE(1777)] = 54462,
  [SMALL_STATE(1778)] = 54469,
  [SMALL_STATE(1779)] = 54476,
  [SMALL_STATE(1780)] = 54483,
  [SMALL_STATE(1781)] = 54490,
  [SMALL_STATE(1782)] = 54497,
  [SMALL_STATE(1783)] = 54504,
  [SMALL_STATE(1784)] = 54511,
  [SMALL_STATE(1785)] = 54518,
  [SMALL_STATE(1786)] = 54525,
  [SMALL_STATE(1787)] = 54532,
  [SMALL_STATE(1788)] = 54539,
  [SMALL_STATE(1789)] = 54546,
  [SMALL_STATE(1790)] = 54553,
  [SMALL_STATE(1791)] = 54560,
  [SMALL_STATE(1792)] = 54567,
  [SMALL_STATE(1793)] = 54574,
  [SMALL_STATE(1794)] = 54581,
  [SMALL_STATE(1795)] = 54588,
  [SMALL_STATE(1796)] = 54595,
  [SMALL_STATE(1797)] = 54602,
  [SMALL_STATE(1798)] = 54609,
  [SMALL_STATE(1799)] = 54616,
  [SMALL_STATE(1800)] = 54623,
  [SMALL_STATE(1801)] = 54630,
  [SMALL_STATE(1802)] = 54637,
  [SMALL_STATE(1803)] = 54644,
  [SMALL_STATE(1804)] = 54651,
  [SMALL_STATE(1805)] = 54658,
  [SMALL_STATE(1806)] = 54665,
  [SMALL_STATE(1807)] = 54672,
  [SMALL_STATE(1808)] = 54679,
  [SMALL_STATE(1809)] = 54686,
  [SMALL_STATE(1810)] = 54693,
  [SMALL_STATE(1811)] = 54700,
  [SMALL_STATE(1812)] = 54707,
  [SMALL_STATE(1813)] = 54714,
  [SMALL_STATE(1814)] = 54721,
  [SMALL_STATE(1815)] = 54728,
  [SMALL_STATE(1816)] = 54735,
  [SMALL_STATE(1817)] = 54742,
  [SMALL_STATE(1818)] = 54749,
  [SMALL_STATE(1819)] = 54756,
  [SMALL_STATE(1820)] = 54763,
  [SMALL_STATE(1821)] = 54770,
  [SMALL_STATE(1822)] = 54777,
  [SMALL_STATE(1823)] = 54784,
  [SMALL_STATE(1824)] = 54791,
  [SMALL_STATE(1825)] = 54798,
  [SMALL_STATE(1826)] = 54805,
  [SMALL_STATE(1827)] = 54812,
  [SMALL_STATE(1828)] = 54819,
  [SMALL_STATE(1829)] = 54826,
  [SMALL_STATE(1830)] = 54833,
  [SMALL_STATE(1831)] = 54840,
  [SMALL_STATE(1832)] = 54847,
  [SMALL_STATE(1833)] = 54854,
  [SMALL_STATE(1834)] = 54861,
  [SMALL_STATE(1835)] = 54868,
  [SMALL_STATE(1836)] = 54875,
  [SMALL_STATE(1837)] = 54882,
  [SMALL_STATE(1838)] = 54889,
  [SMALL_STATE(1839)] = 54896,
  [SMALL_STATE(1840)] = 54903,
  [SMALL_STATE(1841)] = 54910,
  [SMALL_STATE(1842)] = 54917,
  [SMALL_STATE(1843)] = 54924,
  [SMALL_STATE(1844)] = 54931,
  [SMALL_STATE(1845)] = 54938,
  [SMALL_STATE(1846)] = 54945,
  [SMALL_STATE(1847)] = 54952,
  [SMALL_STATE(1848)] = 54959,
  [SMALL_STATE(1849)] = 54966,
  [SMALL_STATE(1850)] = 54973,
  [SMALL_STATE(1851)] = 54980,
  [SMALL_STATE(1852)] = 54987,
  [SMALL_STATE(1853)] = 54994,
  [SMALL_STATE(1854)] = 55001,
  [SMALL_STATE(1855)] = 55008,
  [SMALL_STATE(1856)] = 55015,
  [SMALL_STATE(1857)] = 55022,
  [SMALL_STATE(1858)] = 55029,
  [SMALL_STATE(1859)] = 55036,
  [SMALL_STATE(1860)] = 55043,
  [SMALL_STATE(1861)] = 55050,
  [SMALL_STATE(1862)] = 55057,
  [SMALL_STATE(1863)] = 55064,
  [SMALL_STATE(1864)] = 55071,
  [SMALL_STATE(1865)] = 55078,
  [SMALL_STATE(1866)] = 55085,
  [SMALL_STATE(1867)] = 55092,
  [SMALL_STATE(1868)] = 55099,
  [SMALL_STATE(1869)] = 55106,
  [SMALL_STATE(1870)] = 55113,
  [SMALL_STATE(1871)] = 55120,
  [SMALL_STATE(1872)] = 55127,
  [SMALL_STATE(1873)] = 55134,
  [SMALL_STATE(1874)] = 55141,
  [SMALL_STATE(1875)] = 55148,
  [SMALL_STATE(1876)] = 55155,
  [SMALL_STATE(1877)] = 55162,
  [SMALL_STATE(1878)] = 55169,
  [SMALL_STATE(1879)] = 55176,
  [SMALL_STATE(1880)] = 55183,
  [SMALL_STATE(1881)] = 55190,
  [SMALL_STATE(1882)] = 55197,
  [SMALL_STATE(1883)] = 55204,
  [SMALL_STATE(1884)] = 55211,
  [SMALL_STATE(1885)] = 55218,
  [SMALL_STATE(1886)] = 55225,
  [SMALL_STATE(1887)] = 55232,
  [SMALL_STATE(1888)] = 55239,
  [SMALL_STATE(1889)] = 55246,
  [SMALL_STATE(1890)] = 55253,
  [SMALL_STATE(1891)] = 55260,
  [SMALL_STATE(1892)] = 55267,
  [SMALL_STATE(1893)] = 55274,
  [SMALL_STATE(1894)] = 55281,
  [SMALL_STATE(1895)] = 55288,
  [SMALL_STATE(1896)] = 55295,
  [SMALL_STATE(1897)] = 55302,
  [SMALL_STATE(1898)] = 55309,
  [SMALL_STATE(1899)] = 55316,
  [SMALL_STATE(1900)] = 55323,
  [SMALL_STATE(1901)] = 55330,
  [SMALL_STATE(1902)] = 55337,
  [SMALL_STATE(1903)] = 55344,
  [SMALL_STATE(1904)] = 55351,
  [SMALL_STATE(1905)] = 55358,
  [SMALL_STATE(1906)] = 55365,
  [SMALL_STATE(1907)] = 55372,
  [SMALL_STATE(1908)] = 55379,
  [SMALL_STATE(1909)] = 55386,
  [SMALL_STATE(1910)] = 55393,
  [SMALL_STATE(1911)] = 55400,
  [SMALL_STATE(1912)] = 55407,
  [SMALL_STATE(1913)] = 55414,
  [SMALL_STATE(1914)] = 55421,
  [SMALL_STATE(1915)] = 55428,
  [SMALL_STATE(1916)] = 55435,
  [SMALL_STATE(1917)] = 55442,
  [SMALL_STATE(1918)] = 55449,
  [SMALL_STATE(1919)] = 55456,
  [SMALL_STATE(1920)] = 55463,
  [SMALL_STATE(1921)] = 55470,
  [SMALL_STATE(1922)] = 55477,
  [SMALL_STATE(1923)] = 55484,
  [SMALL_STATE(1924)] = 55491,
  [SMALL_STATE(1925)] = 55498,
  [SMALL_STATE(1926)] = 55505,
  [SMALL_STATE(1927)] = 55512,
  [SMALL_STATE(1928)] = 55519,
  [SMALL_STATE(1929)] = 55526,
  [SMALL_STATE(1930)] = 55533,
  [SMALL_STATE(1931)] = 55540,
  [SMALL_STATE(1932)] = 55547,
  [SMALL_STATE(1933)] = 55554,
  [SMALL_STATE(1934)] = 55561,
  [SMALL_STATE(1935)] = 55568,
  [SMALL_STATE(1936)] = 55575,
  [SMALL_STATE(1937)] = 55582,
  [SMALL_STATE(1938)] = 55589,
  [SMALL_STATE(1939)] = 55596,
  [SMALL_STATE(1940)] = 55603,
  [SMALL_STATE(1941)] = 55610,
  [SMALL_STATE(1942)] = 55617,
  [SMALL_STATE(1943)] = 55624,
  [SMALL_STATE(1944)] = 55631,
  [SMALL_STATE(1945)] = 55638,
  [SMALL_STATE(1946)] = 55645,
  [SMALL_STATE(1947)] = 55652,
  [SMALL_STATE(1948)] = 55659,
  [SMALL_STATE(1949)] = 55666,
  [SMALL_STATE(1950)] = 55673,
  [SMALL_STATE(1951)] = 55680,
  [SMALL_STATE(1952)] = 55687,
  [SMALL_STATE(1953)] = 55694,
  [SMALL_STATE(1954)] = 55701,
  [SMALL_STATE(1955)] = 55708,
  [SMALL_STATE(1956)] = 55715,
  [SMALL_STATE(1957)] = 55722,
  [SMALL_STATE(1958)] = 55729,
  [SMALL_STATE(1959)] = 55736,
  [SMALL_STATE(1960)] = 55743,
  [SMALL_STATE(1961)] = 55750,
  [SMALL_STATE(1962)] = 55757,
  [SMALL_STATE(1963)] = 55764,
  [SMALL_STATE(1964)] = 55771,
  [SMALL_STATE(1965)] = 55778,
  [SMALL_STATE(1966)] = 55785,
  [SMALL_STATE(1967)] = 55792,
  [SMALL_STATE(1968)] = 55799,
  [SMALL_STATE(1969)] = 55806,
  [SMALL_STATE(1970)] = 55813,
  [SMALL_STATE(1971)] = 55820,
  [SMALL_STATE(1972)] = 55827,
  [SMALL_STATE(1973)] = 55834,
  [SMALL_STATE(1974)] = 55841,
  [SMALL_STATE(1975)] = 55848,
  [SMALL_STATE(1976)] = 55855,
  [SMALL_STATE(1977)] = 55862,
  [SMALL_STATE(1978)] = 55869,
  [SMALL_STATE(1979)] = 55876,
  [SMALL_STATE(1980)] = 55883,
  [SMALL_STATE(1981)] = 55890,
  [SMALL_STATE(1982)] = 55897,
  [SMALL_STATE(1983)] = 55904,
  [SMALL_STATE(1984)] = 55911,
  [SMALL_STATE(1985)] = 55918,
  [SMALL_STATE(1986)] = 55925,
  [SMALL_STATE(1987)] = 55932,
  [SMALL_STATE(1988)] = 55939,
  [SMALL_STATE(1989)] = 55946,
  [SMALL_STATE(1990)] = 55953,
  [SMALL_STATE(1991)] = 55960,
  [SMALL_STATE(1992)] = 55967,
  [SMALL_STATE(1993)] = 55974,
  [SMALL_STATE(1994)] = 55981,
  [SMALL_STATE(1995)] = 55988,
  [SMALL_STATE(1996)] = 55995,
  [SMALL_STATE(1997)] = 56002,
  [SMALL_STATE(1998)] = 56009,
  [SMALL_STATE(1999)] = 56016,
  [SMALL_STATE(2000)] = 56023,
  [SMALL_STATE(2001)] = 56030,
  [SMALL_STATE(2002)] = 56037,
  [SMALL_STATE(2003)] = 56044,
  [SMALL_STATE(2004)] = 56051,
  [SMALL_STATE(2005)] = 56058,
  [SMALL_STATE(2006)] = 56065,
  [SMALL_STATE(2007)] = 56072,
  [SMALL_STATE(2008)] = 56079,
  [SMALL_STATE(2009)] = 56086,
  [SMALL_STATE(2010)] = 56093,
  [SMALL_STATE(2011)] = 56100,
  [SMALL_STATE(2012)] = 56107,
  [SMALL_STATE(2013)] = 56114,
  [SMALL_STATE(2014)] = 56121,
  [SMALL_STATE(2015)] = 56128,
  [SMALL_STATE(2016)] = 56135,
  [SMALL_STATE(2017)] = 56142,
  [SMALL_STATE(2018)] = 56149,
  [SMALL_STATE(2019)] = 56156,
  [SMALL_STATE(2020)] = 56163,
  [SMALL_STATE(2021)] = 56170,
  [SMALL_STATE(2022)] = 56177,
  [SMALL_STATE(2023)] = 56184,
  [SMALL_STATE(2024)] = 56191,
  [SMALL_STATE(2025)] = 56198,
  [SMALL_STATE(2026)] = 56205,
  [SMALL_STATE(2027)] = 56212,
  [SMALL_STATE(2028)] = 56219,
  [SMALL_STATE(2029)] = 56226,
  [SMALL_STATE(2030)] = 56233,
  [SMALL_STATE(2031)] = 56240,
  [SMALL_STATE(2032)] = 56247,
  [SMALL_STATE(2033)] = 56254,
  [SMALL_STATE(2034)] = 56261,
  [SMALL_STATE(2035)] = 56268,
  [SMALL_STATE(2036)] = 56275,
  [SMALL_STATE(2037)] = 56282,
  [SMALL_STATE(2038)] = 56289,
  [SMALL_STATE(2039)] = 56296,
  [SMALL_STATE(2040)] = 56303,
  [SMALL_STATE(2041)] = 56310,
  [SMALL_STATE(2042)] = 56317,
  [SMALL_STATE(2043)] = 56324,
  [SMALL_STATE(2044)] = 56331,
  [SMALL_STATE(2045)] = 56338,
  [SMALL_STATE(2046)] = 56345,
  [SMALL_STATE(2047)] = 56352,
  [SMALL_STATE(2048)] = 56359,
  [SMALL_STATE(2049)] = 56366,
  [SMALL_STATE(2050)] = 56373,
  [SMALL_STATE(2051)] = 56380,
  [SMALL_STATE(2052)] = 56387,
  [SMALL_STATE(2053)] = 56394,
  [SMALL_STATE(2054)] = 56401,
  [SMALL_STATE(2055)] = 56408,
  [SMALL_STATE(2056)] = 56415,
  [SMALL_STATE(2057)] = 56422,
  [SMALL_STATE(2058)] = 56429,
  [SMALL_STATE(2059)] = 56436,
  [SMALL_STATE(2060)] = 56443,
  [SMALL_STATE(2061)] = 56450,
  [SMALL_STATE(2062)] = 56457,
  [SMALL_STATE(2063)] = 56464,
  [SMALL_STATE(2064)] = 56471,
  [SMALL_STATE(2065)] = 56478,
  [SMALL_STATE(2066)] = 56485,
  [SMALL_STATE(2067)] = 56492,
  [SMALL_STATE(2068)] = 56499,
  [SMALL_STATE(2069)] = 56506,
  [SMALL_STATE(2070)] = 56513,
  [SMALL_STATE(2071)] = 56520,
  [SMALL_STATE(2072)] = 56527,
  [SMALL_STATE(2073)] = 56534,
  [SMALL_STATE(2074)] = 56541,
  [SMALL_STATE(2075)] = 56548,
  [SMALL_STATE(2076)] = 56555,
  [SMALL_STATE(2077)] = 56562,
  [SMALL_STATE(2078)] = 56569,
  [SMALL_STATE(2079)] = 56576,
  [SMALL_STATE(2080)] = 56583,
  [SMALL_STATE(2081)] = 56590,
  [SMALL_STATE(2082)] = 56597,
  [SMALL_STATE(2083)] = 56604,
  [SMALL_STATE(2084)] = 56611,
  [SMALL_STATE(2085)] = 56618,
  [SMALL_STATE(2086)] = 56625,
  [SMALL_STATE(2087)] = 56632,
  [SMALL_STATE(2088)] = 56639,
  [SMALL_STATE(2089)] = 56646,
  [SMALL_STATE(2090)] = 56653,
  [SMALL_STATE(2091)] = 56660,
  [SMALL_STATE(2092)] = 56667,
  [SMALL_STATE(2093)] = 56674,
  [SMALL_STATE(2094)] = 56681,
  [SMALL_STATE(2095)] = 56688,
  [SMALL_STATE(2096)] = 56695,
  [SMALL_STATE(2097)] = 56702,
  [SMALL_STATE(2098)] = 56709,
  [SMALL_STATE(2099)] = 56716,
  [SMALL_STATE(2100)] = 56723,
  [SMALL_STATE(2101)] = 56730,
  [SMALL_STATE(2102)] = 56737,
  [SMALL_STATE(2103)] = 56744,
  [SMALL_STATE(2104)] = 56751,
  [SMALL_STATE(2105)] = 56758,
  [SMALL_STATE(2106)] = 56765,
  [SMALL_STATE(2107)] = 56772,
  [SMALL_STATE(2108)] = 56779,
  [SMALL_STATE(2109)] = 56786,
  [SMALL_STATE(2110)] = 56793,
  [SMALL_STATE(2111)] = 56800,
  [SMALL_STATE(2112)] = 56807,
  [SMALL_STATE(2113)] = 56814,
  [SMALL_STATE(2114)] = 56821,
  [SMALL_STATE(2115)] = 56828,
  [SMALL_STATE(2116)] = 56835,
  [SMALL_STATE(2117)] = 56842,
  [SMALL_STATE(2118)] = 56849,
  [SMALL_STATE(2119)] = 56856,
  [SMALL_STATE(2120)] = 56863,
  [SMALL_STATE(2121)] = 56870,
  [SMALL_STATE(2122)] = 56877,
  [SMALL_STATE(2123)] = 56884,
  [SMALL_STATE(2124)] = 56891,
  [SMALL_STATE(2125)] = 56898,
  [SMALL_STATE(2126)] = 56905,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT(124),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1179),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1006),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1331),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1863),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(36),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1811),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1479),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1776),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1821),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1635),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1638),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1267),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1397),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1682),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1321),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1616),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1522),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1523),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(612),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(60),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(56),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [89] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [93] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value, 1, 0, 0),
  [95] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value, 1, 0, 0),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 0),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(87),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(675),
  [143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1927),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1731),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2093),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1270),
  [183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1516),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2036),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2094),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1693),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1688),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1265),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1618),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1330),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1619),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1620),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2038),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2039),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(415),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1249),
  [219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1251),
  [221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1255),
  [225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1258),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1989),
  [233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2019),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1830),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1869),
  [241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(305),
  [245] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_script_block, 3, 0, 0),
  [247] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_script_block, 2, 0, 0),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(323),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(795),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(843),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(316),
  [257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1240),
  [259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(898),
  [263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [267] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_list, 1, 0, 0),
  [269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1018),
  [275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(326),
  [277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(124),
  [282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(160),
  [285] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1179),
  [288] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1213),
  [291] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(157),
  [294] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1006),
  [297] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1340),
  [300] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(598),
  [303] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(142),
  [306] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(349),
  [309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(258),
  [312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0),
  [314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(36),
  [317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2093),
  [320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1270),
  [323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1516),
  [326] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2036),
  [329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2094),
  [332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1693),
  [335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1688),
  [338] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(572),
  [341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1265),
  [347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1415),
  [350] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1618),
  [353] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1330),
  [356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1619),
  [359] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1620),
  [362] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1621),
  [365] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(916),
  [368] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2038),
  [371] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2039),
  [374] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(597),
  [377] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(612),
  [380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(60),
  [383] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(56),
  [386] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(539),
  [389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(353),
  [395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1238),
  [397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(99),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(950),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_list, 1, 0, 0),
  [417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0),
  [419] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(381),
  [422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1811),
  [425] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1263),
  [428] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1479),
  [431] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1776),
  [434] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1821),
  [437] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1635),
  [440] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1638),
  [443] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1267),
  [446] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1397),
  [449] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1682),
  [452] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1321),
  [455] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1616),
  [458] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1522),
  [461] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1523),
  [464] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1916),
  [467] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1808),
  [470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2099),
  [474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2056),
  [480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2100),
  [482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1694),
  [484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1689),
  [486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1322),
  [494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1527),
  [496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1528),
  [498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1529),
  [500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2058),
  [502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2059),
  [504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1755),
  [506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2105),
  [508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1271),
  [510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1518),
  [512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2075),
  [514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2106),
  [516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1695),
  [518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1690),
  [520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1260),
  [522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1573),
  [526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1574),
  [530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1575),
  [532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1576),
  [534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2078),
  [538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_literal, 3, 0, 0),
  [540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_literal, 3, 0, 0),
  [542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(609),
  [550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(656),
  [560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(634),
  [572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable, 1, 0, 0),
  [588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable, 1, 0, 0),
  [590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_access, 3, 0, 0),
  [592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_access, 3, 0, 0),
  [594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 14),
  [598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 14),
  [600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_name, 1, 0, 0),
  [602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_name, 1, 0, 0),
  [604] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__literal, 1, 0, 0), REDUCE(sym_member_name, 1, 0, 0),
  [607] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__literal, 1, 0, 0), REDUCE(sym_member_name, 1, 0, 0),
  [610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 3, 0, 7),
  [612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 3, 0, 7),
  [614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_invokation_foreach_expression, 3, 0, 0),
  [616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_invokation_foreach_expression, 3, 0, 0),
  [618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expandable_string_literal, 4, 0, 0),
  [620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expandable_string_literal, 4, 0, 0),
  [622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block_expression, 4, 0, 0),
  [624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_script_block_expression, 4, 0, 0),
  [626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_access, 4, 0, 0),
  [628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_element_access, 4, 0, 0),
  [630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_invokation_expression, 4, 0, 0),
  [632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_invokation_expression, 4, 0, 0),
  [634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 2, 0, 0),
  [636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 2, 0, 0),
  [638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sub_expression, 2, 0, 0),
  [640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sub_expression, 2, 0, 0),
  [642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sub_expression, 3, 0, 7),
  [644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sub_expression, 3, 0, 7),
  [646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expandable_string_literal, 2, 0, 0),
  [652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expandable_string_literal, 2, 0, 0),
  [654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 1, 0, 0),
  [656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 1, 0, 0),
  [658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_with_unary_operator, 1, 0, 0),
  [660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_with_unary_operator, 1, 0, 0),
  [662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block_expression, 3, 0, 0),
  [664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_script_block_expression, 3, 0, 0),
  [666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_literal_expression, 2, 0, 0),
  [668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_literal_expression, 2, 0, 0),
  [670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pre_increment_expression, 2, 0, 0),
  [672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pre_increment_expression, 2, 0, 0),
  [674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pre_decrement_expression, 2, 0, 0),
  [676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pre_decrement_expression, 2, 0, 0),
  [678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 2, 0, 0),
  [680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 2, 0, 0),
  [682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_literal_expression, 3, 0, 0),
  [684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_literal_expression, 3, 0, 0),
  [686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_post_increment_expression, 2, 0, 0),
  [688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_post_increment_expression, 2, 0, 0),
  [690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_post_decrement_expression, 2, 0, 0),
  [692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_post_decrement_expression, 2, 0, 0),
  [694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expandable_string_literal, 3, 0, 0),
  [698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expandable_string_literal, 3, 0, 0),
  [700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer_literal, 1, 0, 0),
  [702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer_literal, 1, 0, 0),
  [704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expandable_here_string_literal, 3, 0, 0),
  [706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expandable_here_string_literal, 3, 0, 0),
  [708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 1, 0, 0),
  [710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 1, 0, 0),
  [712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_invokation_expression, 1, 0, 0),
  [714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_invokation_expression, 1, 0, 0),
  [716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expandable_here_string_literal, 2, 0, 0),
  [718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expandable_here_string_literal, 2, 0, 0),
  [720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_with_unary_operator, 2, 0, 0),
  [726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_with_unary_operator, 2, 0, 0),
  [728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(696),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(706),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1640),
  [752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0),
  [762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0),
  [764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 3, 0, 0),
  [766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 3, 0, 0),
  [768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 1, 0, 0),
  [770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 1, 0, 0),
  [772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal_expression, 1, 0, 0),
  [774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal_expression, 1, 0, 0),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [778] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(655),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal_expression, 2, 0, 0),
  [785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal_expression, 2, 0, 0),
  [787] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(637),
  [790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_multiplicative_expression, 1, 0, 0),
  [792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_multiplicative_expression, 1, 0, 0),
  [794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_multiplicative_expression, 3, 0, 0),
  [798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_multiplicative_expression, 3, 0, 0),
  [800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_expression, 1, 0, 0),
  [802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_expression, 1, 0, 0),
  [804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_expression, 3, 0, 0),
  [808] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_expression, 3, 0, 0),
  [810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_additive_expression, 1, 0, 0),
  [814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_additive_expression, 1, 0, 0),
  [816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(580),
  [818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_additive_expression, 3, 0, 0),
  [822] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_additive_expression, 3, 0, 0),
  [824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(574),
  [826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_expression, 1, 0, 0),
  [830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_expression, 1, 0, 0),
  [832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(554),
  [834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(557),
  [836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_expression, 3, 0, 0),
  [838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_expression, 3, 0, 0),
  [840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1651),
  [844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitwise_expression, 1, 0, 0),
  [848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitwise_expression, 1, 0, 0),
  [850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1138),
  [852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1138),
  [854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1539),
  [858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1660),
  [864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(242),
  [868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1544),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1553),
  [876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(208),
  [880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1584),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(227),
  [886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1590),
  [888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1598),
  [894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1686),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitwise_expression, 3, 0, 0),
  [904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitwise_expression, 3, 0, 0),
  [906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(201),
  [908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1642),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(233),
  [914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1684),
  [916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(232),
  [920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1524),
  [922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(753),
  [928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1185),
  [930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(820),
  [934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1354),
  [936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(605),
  [938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(267),
  [942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(38),
  [944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(606),
  [946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(68),
  [952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(781),
  [956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1189),
  [960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(782),
  [964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(269),
  [972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(33),
  [974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(625),
  [976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(626),
  [978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(47),
  [980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(71),
  [982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(533),
  [984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(621),
  [986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(623),
  [990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(641),
  [992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(642),
  [994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(643),
  [996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1665),
  [998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1654),
  [1000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1602),
  [1002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_argument, 1, 0, 0),
  [1004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1044),
  [1006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_argument, 1, 0, 0),
  [1008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1593),
  [1010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1679),
  [1012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1613),
  [1014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1608),
  [1016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1601),
  [1018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1542),
  [1020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1614),
  [1022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1664),
  [1024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [1026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(835),
  [1028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1193),
  [1030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1207),
  [1032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(830),
  [1034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [1036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(671),
  [1038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(836),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(273),
  [1042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(41),
  [1044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(672),
  [1046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(673),
  [1048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(52),
  [1050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [1052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [1054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [1056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [1058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(859),
  [1060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1191),
  [1062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [1064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(842),
  [1066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [1068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(667),
  [1070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(832),
  [1072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(271),
  [1074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [1076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(668),
  [1078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(669),
  [1080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(51),
  [1082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(74),
  [1084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [1086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1047),
  [1088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1557),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(627),
  [1092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(628),
  [1094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(629),
  [1096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(630),
  [1098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(631),
  [1100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(632),
  [1102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1548),
  [1104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1556),
  [1106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1563),
  [1108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [1110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [1112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [1114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [1116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1381),
  [1118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(262),
  [1120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [1122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [1124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(676),
  [1126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(677),
  [1128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [1130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(66),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(531),
  [1134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1102),
  [1136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1098),
  [1138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_flow_control_statement, 1, 0, 0),
  [1140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(972),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [1144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [1150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [1152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [1154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [1156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [1158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(949),
  [1160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [1162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [1164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(649),
  [1166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [1168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [1170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [1172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [1174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [1176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1008),
  [1178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [1180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [1182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [1184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [1186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [1188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [1190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [1192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [1194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(984),
  [1196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [1198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [1200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(645),
  [1202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [1204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [1206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [1208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [1210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [1212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [1214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [1216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(663),
  [1218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [1222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [1224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [1226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [1230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [1232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1072),
  [1234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [1236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1204),
  [1238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(933),
  [1240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [1242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1128),
  [1244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [1246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [1248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(650),
  [1250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(651),
  [1252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(70),
  [1254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(63),
  [1256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [1258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1071),
  [1260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1196),
  [1262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [1264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [1266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1390),
  [1268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [1270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(264),
  [1272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(40),
  [1274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(646),
  [1276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(647),
  [1278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(67),
  [1280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(58),
  [1282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [1284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 13),
  [1286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2098),
  [1288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1656),
  [1290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 2, 0, 0),
  [1292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1336),
  [1294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1643),
  [1296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 2, 0, 0),
  [1298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1335),
  [1300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1567),
  [1302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 13),
  [1304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1792),
  [1306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1636),
  [1308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0),
  [1310] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1336),
  [1313] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elseif_clauses, 1, 0, 0),
  [1315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_param_block, 3, 0, 0),
  [1317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_param_block, 3, 0, 0),
  [1319] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_param_block, 5, 0, 0),
  [1321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_param_block, 5, 0, 0),
  [1323] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0),
  [1325] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(2098),
  [1328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0),
  [1330] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0), SHIFT_REPEAT(302),
  [1333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clauses, 1, 0, 0),
  [1335] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_param_block, 4, 0, 0),
  [1337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_param_block, 4, 0, 0),
  [1339] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_script_block_repeat1, 1, 0, 0),
  [1343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0),
  [1345] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1335),
  [1348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elseif_clauses, 1, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0),
  [1352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1792),
  [1355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clauses, 1, 0, 0),
  [1357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 3, 0, 0),
  [1359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 15),
  [1361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elseif_clause, 5, 0, 13),
  [1363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_block, 2, 0, 0),
  [1365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 0),
  [1367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_block, 3, 0, 11),
  [1369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 2, 0, 0),
  [1371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 2, 0, 0),
  [1373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 15),
  [1375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 3, 0, 0),
  [1377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_block, 3, 0, 11),
  [1379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_block, 2, 0, 0),
  [1381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 0),
  [1383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elseif_clause, 5, 0, 13),
  [1385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 0),
  [1387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 17),
  [1389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 7, 0, 0),
  [1391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 8, 0, 0),
  [1393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 21),
  [1395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 20),
  [1397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 17),
  [1399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 8, 0, 0),
  [1401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 22),
  [1403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 16),
  [1405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 23),
  [1407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 24),
  [1409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 22),
  [1411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 0),
  [1413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 12, 0, 25),
  [1415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trap_statement, 3, 0, 0),
  [1417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 5, 0, 0),
  [1419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 17),
  [1423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_data_statement, 3, 0, 0),
  [1425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 6, 0, 18),
  [1427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_statement, 6, 0, 0),
  [1429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_statement, 5, 0, 0),
  [1431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 4, 0, 0),
  [1433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_body, 2, 0, 0),
  [1435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 4, 0, 0),
  [1437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_statement, 4, 0, 0),
  [1439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1441] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 7, 0, 19),
  [1443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 2, 0, 0),
  [1445] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trap_statement, 2, 0, 0),
  [1447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 2, 0, 0),
  [1449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_statement, 5, 0, 0),
  [1451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_statement, 2, 0, 0),
  [1453] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_data_statement, 2, 0, 0),
  [1457] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_data_statement, 4, 0, 0),
  [1459] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inlinescript_statement, 2, 0, 0),
  [1461] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 4, 0, 0),
  [1463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_statement, 4, 0, 0),
  [1465] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 20),
  [1467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_body, 3, 0, 0),
  [1469] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parallel_statement, 2, 0, 0),
  [1471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 5, 0, 0),
  [1473] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 5, 0, 13),
  [1475] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 6, 0, 0),
  [1477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_statement, 2, 0, 0),
  [1479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_statement, 4, 0, 0),
  [1481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 4, 0, 0),
  [1483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_body, 3, 0, 0),
  [1487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 5, 0, 0),
  [1489] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 5, 0, 13),
  [1491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_statement, 5, 0, 0),
  [1493] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 5, 0, 0),
  [1495] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_statement, 5, 0, 0),
  [1497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 16),
  [1499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 17),
  [1503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 6, 0, 18),
  [1505] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_statement, 6, 0, 0),
  [1507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 6, 0, 0),
  [1509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 7, 0, 19),
  [1513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 20),
  [1517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 17),
  [1519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 7, 0, 0),
  [1521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 8, 0, 0),
  [1523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 21),
  [1525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 20),
  [1527] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 17),
  [1529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 8, 0, 0),
  [1531] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 22),
  [1533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 23),
  [1535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 24),
  [1537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 22),
  [1539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 12, 0, 25),
  [1541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 2, 0, 0),
  [1543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trap_statement, 2, 0, 0),
  [1545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 4, 0, 0),
  [1547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_statement, 2, 0, 0),
  [1549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_statement, 4, 0, 0),
  [1551] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 0),
  [1553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inlinescript_statement, 2, 0, 0),
  [1555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parallel_statement, 2, 0, 0),
  [1557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trap_statement, 3, 0, 0),
  [1559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_statement, 3, 0, 0),
  [1561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 2, 0, 0),
  [1563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 0),
  [1565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_statement, 4, 0, 0),
  [1567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 4, 0, 0),
  [1569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_body, 2, 0, 0),
  [1571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignement_operator, 1, 0, 0),
  [1573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(694),
  [1575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [1577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(693),
  [1579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [1581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(692),
  [1583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [1585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [1587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [1589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_argument_sep, 1, 0, 0),
  [1591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(441),
  [1593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_argument_sep, 1, 0, 0),
  [1595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0),
  [1597] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(439),
  [1600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(613),
  [1603] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [1606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0),
  [1608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [1610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1900),
  [1612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(439),
  [1614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [1616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(604),
  [1619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [1621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [1623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(459),
  [1625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [1627] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(458),
  [1630] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(459),
  [1633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [1635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [1637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [1639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [1641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [1643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [1645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [1647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [1649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [1651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [1653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1349),
  [1655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1224),
  [1657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1303),
  [1659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [1661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [1663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [1665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(494),
  [1667] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(494),
  [1670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [1672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [1674] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_argument_sep_repeat1, 2, 0, 0), SHIFT_REPEAT(505),
  [1677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_multiplicative_argument_expression, 3, 0, 0),
  [1679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_multiplicative_argument_expression, 3, 0, 0),
  [1681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_argument_expression, 1, 0, 0),
  [1683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_argument_expression, 1, 0, 0),
  [1685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [1687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(906),
  [1689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1361),
  [1691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(684),
  [1693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(921),
  [1695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [1697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_argument_expression, 3, 0, 0),
  [1699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_argument_expression, 3, 0, 0),
  [1701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_argument_expression, 3, 0, 0),
  [1703] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_argument_expression, 3, 0, 0),
  [1705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [1707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(908),
  [1709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1359),
  [1711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(685),
  [1713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [1715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [1717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [1719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1343),
  [1721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(682),
  [1723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [1725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(857),
  [1727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1346),
  [1729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(681),
  [1731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [1733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_argument_expression, 1, 0, 0),
  [1735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_argument_expression, 1, 0, 0),
  [1737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_multiplicative_argument_expression, 1, 0, 0),
  [1739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_multiplicative_argument_expression, 1, 0, 0),
  [1741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [1743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [1745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2115),
  [1747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [1749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [1751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [1753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(670),
  [1755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [1757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [1759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [1761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [1763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [1765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [1767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [1769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_additive_argument_expression, 3, 0, 0),
  [1771] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_additive_argument_expression, 3, 0, 0),
  [1773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [1775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [1777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_additive_argument_expression, 1, 0, 0),
  [1779] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_additive_argument_expression, 1, 0, 0),
  [1781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [1783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [1785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [1787] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(972),
  [1790] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(972),
  [1793] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1314),
  [1796] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1176),
  [1799] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1204),
  [1802] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(933),
  [1805] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(2115),
  [1808] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1388),
  [1811] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(670),
  [1814] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(949),
  [1817] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(949),
  [1820] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(266),
  [1823] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [1826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0),
  [1828] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(670),
  [1831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(674),
  [1834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(678),
  [1837] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(70),
  [1840] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(63),
  [1843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hash_literal_body_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [1846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_literal_body, 1, 0, 0),
  [1848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [1850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(562),
  [1852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [1854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [1856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [1858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [1860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [1862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [1864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(561),
  [1866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [1868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [1870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [1872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(602),
  [1874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [1876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [1878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [1880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [1882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(614),
  [1884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [1886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [1888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [1890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_argument_expression, 1, 0, 0),
  [1892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_argument_expression, 1, 0, 0),
  [1894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [1896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(575),
  [1898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_argument_expression, 3, 0, 0),
  [1900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_argument_expression, 3, 0, 0),
  [1902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitwise_argument_expression, 3, 0, 0),
  [1904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitwise_argument_expression, 1, 0, 0),
  [1906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [1908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [1910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [1912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [1914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [1916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [1918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [1920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [1922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [1924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [1926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [1928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [1930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [1932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [1934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [1936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [1938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [1940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [1942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [1944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [1946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [1948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [1950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [1952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [1954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [1956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [1958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [1960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [1962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [1964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [1966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [1968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [1970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [1972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [1974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [1976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [1978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [1980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [1982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [1984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [1986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [1988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [1990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [1992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [1994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [1996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [1998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [2000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [2002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [2004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [2006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [2008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [2010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [2012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [2014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [2016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [2018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [2020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [2022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [2024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [2026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [2028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [2030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [2032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [2036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [2038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [2040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [2042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [2044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [2046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [2050] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__value, 1, 0, 0), REDUCE(aux_sym_path_command_name_repeat1, 1, 0, 0),
  [2053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 1, 0, 0),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__value, 1, 0, 0), REDUCE(aux_sym_path_command_name_repeat1, 1, 0, 0),
  [2058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pipeline_chain_tail, 1, 0, 0),
  [2060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [2062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [2064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [2066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [2068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [2070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [2072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1145),
  [2074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1056),
  [2076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [2078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(268),
  [2080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [2082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 2, 0, 5),
  [2084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [2086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(448),
  [2088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 2, 0, 5),
  [2090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [2092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 1, 0, 1),
  [2094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 1, 0, 1),
  [2096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [2098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1081),
  [2100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [2102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [2104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(35),
  [2106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [2108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(449),
  [2110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_path_command_name, 1, 0, 0),
  [2112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [2114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [2116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_path_command_name, 1, 0, 0),
  [2118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [2120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(729),
  [2122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [2124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [2126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [2128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(463),
  [2130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0),
  [2132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(747),
  [2135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(715),
  [2138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0),
  [2140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [2142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [2144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1092),
  [2146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [2148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [2150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(43),
  [2152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [2154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(461),
  [2156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [2158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_elements, 1, 0, 0),
  [2160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_elements, 1, 0, 0),
  [2162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [2164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(1145),
  [2167] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(1056),
  [2170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(721),
  [2173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(268),
  [2176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(42),
  [2179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0),
  [2181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [2184] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(448),
  [2187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0),
  [2189] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(737),
  [2192] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(723),
  [2195] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(1081),
  [2198] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(724),
  [2201] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(270),
  [2204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(35),
  [2207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(445),
  [2210] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(449),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(751),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(757),
  [2219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [2221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(1091),
  [2224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(727),
  [2227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(272),
  [2230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(39),
  [2233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(457),
  [2236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(463),
  [2239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(1092),
  [2242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(728),
  [2245] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(274),
  [2248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(43),
  [2251] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(453),
  [2254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [2259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [2261] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(751),
  [2264] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(731),
  [2267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(757),
  [2270] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(732),
  [2273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1094),
  [2275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [2277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1096),
  [2281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [2283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [2285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [2287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1312),
  [2289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1860),
  [2291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1941),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [2295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(909),
  [2297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(907),
  [2299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(427),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(833),
  [2303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1982),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(355),
  [2307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(784),
  [2311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(695),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [2317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_clauses, 1, 0, 0),
  [2319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [2321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [2323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [2325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(699),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [2329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(972),
  [2332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1312),
  [2335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1176),
  [2338] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1204),
  [2341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(933),
  [2344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1349),
  [2347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(949),
  [2350] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1860),
  [2353] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(266),
  [2356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [2359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0),
  [2361] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(70),
  [2364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(63),
  [2367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_switch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [2370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirected_file_name, 1, 0, 0),
  [2372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirected_file_name, 1, 0, 0),
  [2374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(503),
  [2376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(846),
  [2378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(848),
  [2380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(710),
  [2382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1633),
  [2384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_left_assignment_expression, 1, 0, 0),
  [2386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [2388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [2390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline_chain, 1, 0, 0),
  [2392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1143),
  [2394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_expression, 1, 0, 0),
  [2396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_expression, 1, 0, 0),
  [2398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [2402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [2404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [2406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_name_expr, 1, 0, 0),
  [2408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_name_expr, 1, 0, 0),
  [2410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [2412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [2414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [2416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1626),
  [2420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_expression, 3, 0, 0),
  [2422] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_expression, 3, 0, 0),
  [2424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [2427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0),
  [2429] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [2432] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1175),
  [2435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0),
  [2437] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(911),
  [2440] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(911),
  [2443] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1174),
  [2446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [2448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_name, 1, 0, 0),
  [2450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(915),
  [2452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [2454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [2456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_name, 2, 0, 0),
  [2458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [2460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [2462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [2464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1175),
  [2466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_name, 1, 0, 0),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [2470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [2472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_name, 2, 0, 0),
  [2474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(918),
  [2477] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(918),
  [2480] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1172),
  [2483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [2485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(923),
  [2487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [2489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [2491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [2493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(922),
  [2495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1173),
  [2497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [2499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [2501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [2503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(918),
  [2505] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(924),
  [2508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(924),
  [2511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1173),
  [2514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expandable_string_literal_immediate, 3, 0, 0),
  [2516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expandable_string_literal_immediate, 3, 0, 0),
  [2518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expandable_string_literal_immediate, 2, 0, 0),
  [2520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expandable_string_literal_immediate, 2, 0, 0),
  [2522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1318),
  [2524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1874),
  [2526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [2528] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [2530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(617),
  [2536] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(617),
  [2539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(644),
  [2541] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(644),
  [2544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [2546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(662),
  [2549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(662),
  [2551] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(666),
  [2554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(666),
  [2556] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_element, 1, 0, 0),
  [2558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(479),
  [2560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_element, 1, 0, 0),
  [2562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_invokation_operator, 1, 0, 0),
  [2564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(468),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [2568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [2570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [2572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(697),
  [2574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [2582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [2584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(698),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [2590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirection, 1, 0, 0),
  [2592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirection, 1, 0, 0),
  [2594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirection, 2, 0, 0),
  [2596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirection, 2, 0, 0),
  [2598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_argument, 2, 0, 0),
  [2600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_argument, 2, 0, 0),
  [2602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_element, 2, 0, 0),
  [2604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_element, 2, 0, 0),
  [2606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(472),
  [2608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_merging_redirection_operator, 1, 0, 0),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_merging_redirection_operator, 1, 0, 0),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(475),
  [2614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(701),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [2618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [2622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirections, 1, 0, 0),
  [2624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirections, 1, 0, 0),
  [2626] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0), SHIFT_REPEAT(1145),
  [2629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0), SHIFT_REPEAT(1145),
  [2632] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0), SHIFT_REPEAT(1136),
  [2635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0),
  [2637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0),
  [2639] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirections_repeat1, 2, 0, 0), SHIFT_REPEAT(1135),
  [2642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_entry, 5, 0, 0),
  [2644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_entry, 5, 0, 0),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [2648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0),
  [2650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1120),
  [2653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_entry, 4, 0, 0),
  [2655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_entry, 4, 0, 0),
  [2657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [2659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [2661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [2663] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(648),
  [2666] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_literal_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(601),
  [2669] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_operator, 1, 0, 0),
  [2671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_operator, 1, 0, 0),
  [2673] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_operator, 1, 0, 0),
  [2675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_operator, 1, 0, 0),
  [2677] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_clause, 4, 0, 0),
  [2679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1142),
  [2681] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_clause, 3, 0, 0),
  [2683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [2685] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1142),
  [2688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirection_operator, 1, 0, 0),
  [2690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirection_operator, 1, 0, 0),
  [2692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1850),
  [2694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [2698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [2700] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0), SHIFT_REPEAT(1850),
  [2703] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0), SHIFT_REPEAT(1325),
  [2706] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0), SHIFT_REPEAT(142),
  [2709] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0), SHIFT_REPEAT(142),
  [2712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0),
  [2714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0), SHIFT_REPEAT(1264),
  [2717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [2721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [2723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [2725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [2727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2052),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2071),
  [2731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [2733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [2739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1883),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [2745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [2747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [2749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [2753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [2755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [2757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1178),
  [2759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1453),
  [2761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1327),
  [2763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(50),
  [2765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1182),
  [2767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [2769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(966),
  [2771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1188),
  [2773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1450),
  [2775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(944),
  [2777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1180),
  [2779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1515),
  [2781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [2783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1184),
  [2785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1456),
  [2787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [2789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1181),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [2793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1230),
  [2795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(967),
  [2797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1235),
  [2799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(65),
  [2801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1210),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [2805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [2807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1200),
  [2809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1499),
  [2811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1323),
  [2813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [2817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [2819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [2821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(931),
  [2823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [2825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1508),
  [2827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [2829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1187),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [2833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(103),
  [2835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1451),
  [2837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(926),
  [2839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1186),
  [2843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(755),
  [2845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(791),
  [2847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [2849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1457),
  [2851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(947),
  [2853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [2857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(797),
  [2859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [2861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [2863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [2865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(872),
  [2867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [2869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [2871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [2873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [2875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [2879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [2881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [2883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1013),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [2887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1203),
  [2889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [2891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [2893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(57),
  [2895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [2897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1211),
  [2899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [2901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [2903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1198),
  [2905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [2907] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat1, 2, 0, 0), SHIFT_REPEAT(1231),
  [2910] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat1, 2, 0, 0), SHIFT_REPEAT(1200),
  [2913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat1, 2, 0, 0),
  [2915] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat1, 2, 0, 0), SHIFT_REPEAT(50),
  [2918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [2920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [2922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [2926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1202),
  [2928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [2930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [2932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [2934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1208),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [2938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [2940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1209),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [2944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [2948] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1210),
  [2951] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1210),
  [2954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 2, 0, 0),
  [2956] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1235),
  [2959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(65),
  [2962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expandable_here_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1211),
  [2965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_here_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1211),
  [2968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expandable_here_string_literal_repeat1, 2, 0, 0),
  [2970] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_here_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1241),
  [2973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_here_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(57),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [2978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1201),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [2984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1214),
  [2986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [2990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [2992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1205),
  [2994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [2996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [3000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1907),
  [3002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [3004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [3012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_path_command_name_repeat1, 1, 0, 0),
  [3014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2011),
  [3016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [3018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [3020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1228),
  [3023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 1, 0, 0),
  [3025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 1, 0, 0),
  [3027] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat1, 1, 0, 0), REDUCE(aux_sym_expandable_string_literal_repeat2, 1, 0, 0),
  [3030] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_statement_repeat2, 2, 0, 0),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [3034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_statement_repeat2, 3, 0, 0),
  [3036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat2, 3, 0, 0),
  [3038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [3040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_property_definition_repeat1, 2, 0, 0),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_property_definition_repeat1, 2, 0, 0),
  [3044] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_property_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1264),
  [3047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [3049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [3051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [3053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [3055] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_attribute, 1, 0, 0), SHIFT(1947),
  [3058] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 0), SHIFT(142),
  [3061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_attribute, 1, 0, 0), SHIFT(142),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [3066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [3070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 5, 0, 0),
  [3072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 5, 0, 0),
  [3074] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1272),
  [3077] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1272),
  [3080] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1250),
  [3083] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 6, 0, 0),
  [3085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 6, 0, 0),
  [3087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [3089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [3091] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1329),
  [3094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [3096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [3098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 7, 0, 0),
  [3100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 7, 0, 0),
  [3102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 8, 0, 0),
  [3104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 8, 0, 0),
  [3106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 10, 0, 0),
  [3108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 10, 0, 0),
  [3110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_method_definition, 9, 0, 0),
  [3112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_definition, 9, 0, 0),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1518),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1695),
  [3122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 5, 0, 0),
  [3124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 5, 0, 0),
  [3126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [3128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [3130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1694),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [3136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [3138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [3140] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_attribute, 1, 0, 0),
  [3142] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_attribute, 1, 0, 0),
  [3144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [3146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [3148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [3150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1693),
  [3152] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 6, 0, 0),
  [3154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 6, 0, 0),
  [3156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [3158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [3160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [3164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(691),
  [3166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [3168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_parameters, 1, 0, 0),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1817),
  [3172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [3174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_parameters_repeat1, 2, 0, 0),
  [3176] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(1309),
  [3179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 1, 0, 0),
  [3181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__verbatim_command_argument_chars, 2, 0, 0), SHIFT_REPEAT(1283),
  [3184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__verbatim_command_argument_chars, 2, 0, 0),
  [3186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [3188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_verbatim_command_argument, 2, 0, 0),
  [3190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_verbatim_command_argument, 2, 0, 0),
  [3192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(705),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [3196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [3198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [3200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_named_block_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1731),
  [3203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_named_block_list_repeat1, 2, 0, 0),
  [3205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [3207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(690),
  [3209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [3211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [3213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_block_list, 1, 0, 0),
  [3215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__verbatim_command_argument_chars, 2, 0, 0), SHIFT_REPEAT(1298),
  [3218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__verbatim_command_argument_chars, 2, 0, 0),
  [3220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [3222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1317),
  [3224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [3226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [3228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(707),
  [3230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [3232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1306),
  [3235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1306),
  [3238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_name_repeat1, 2, 0, 0), SHIFT_REPEAT(1171),
  [3241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(536),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2110),
  [3245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1595),
  [3247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_parameter, 1, 0, 0),
  [3249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [3251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [3253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_clause_condition, 1, 0, 0),
  [3255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [3257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [3259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [3261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2104),
  [3263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1550),
  [3265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [3267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1306),
  [3269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_filename, 1, 0, 0),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [3273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [3277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [3279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [3283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [3285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [3287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline, 2, 0, 0),
  [3289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [3291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_argument_expression, 3, 0, 0),
  [3293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [3297] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_expression, 1, 0, 0),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [3301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline_chain, 2, 0, 0),
  [3303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_argument_expression, 1, 0, 0),
  [3305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline, 1, 0, 0),
  [3307] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 3, 0, 0),
  [3309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipeline_repeat1, 2, 0, 0), SHIFT_REPEAT(688),
  [3312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipeline_repeat1, 2, 0, 0),
  [3314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 3, 0, 9),
  [3316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 3, 0, 9),
  [3318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_block, 2, 0, 0),
  [3320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 2, 0, 6),
  [3322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 2, 0, 6),
  [3324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 1, 0, 0),
  [3326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(2110),
  [3329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__pipeline_tail, 2, 0, 0),
  [3331] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__pipeline_tail, 2, 0, 0), SHIFT_REPEAT(1144),
  [3334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline_chain, 3, 0, 0),
  [3336] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__pipeline_tail, 2, 0, 0), SHIFT_REPEAT(1143),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [3341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [3343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1501),
  [3345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_spec, 1, 0, 0),
  [3347] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1333),
  [3350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_catch_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1389),
  [3353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_elseif_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(2104),
  [3356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [3358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_parameter, 2, 0, 0),
  [3360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1290),
  [3362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__pipeline_tail, 2, 0, 0),
  [3364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [3366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2010),
  [3368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [3370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [3372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [3374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [3376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_parameter, 1, 0, 0),
  [3378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1577),
  [3381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_statement_repeat1, 2, 0, 0),
  [3383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_statement_repeat1, 3, 0, 0),
  [3387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [3389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [3391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [3393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [3395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [3397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [3399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [3401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1716),
  [3403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_arguments, 2, 0, 0),
  [3405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3407] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expandable_string_literal_repeat2, 2, 0, 0), SHIFT_REPEAT(1420),
  [3410] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expandable_string_literal_repeat2, 2, 0, 0),
  [3412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [3414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1681),
  [3416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [3418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [3420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_type_list, 1, 0, 0),
  [3422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_spec, 2, 0, 0),
  [3424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [3426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [3428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [3430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [3432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [3434] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_type_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1350),
  [3437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_type_arguments_repeat1, 2, 0, 0),
  [3439] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_expression_list, 2, 0, 0),
  [3441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [3443] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 1, 0, 0),
  [3445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [3447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_method_parameter_list_repeat1, 2, 0, 0),
  [3449] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_method_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [3454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [3456] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1879),
  [3459] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_statement_repeat1, 2, 0, 0),
  [3461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_arguments, 1, 0, 0),
  [3463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [3465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [3467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_commands_list, 2, 0, 0),
  [3469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [3471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1655),
  [3473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [3475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [3477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type_arguments, 2, 0, 0),
  [3479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [3481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(234),
  [3483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1659),
  [3485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [3487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1167),
  [3489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [3491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [3493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1666),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [3497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_data_commands_list_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [3500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_data_commands_list_repeat1, 2, 0, 0),
  [3502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_type_list, 2, 0, 0),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_spec, 3, 0, 0),
  [3506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [3512] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat2, 2, 0, 0), SHIFT_REPEAT(1452),
  [3515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__expandable_string_literal_immediate_repeat2, 2, 0, 0),
  [3517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [3519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [3521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [3523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_arguments_repeat1, 2, 0, 0),
  [3525] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(464),
  [3528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [3530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [3532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [3534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [3536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1549),
  [3538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [3540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [3542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [3544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [3546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1552),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [3550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_expression_list_repeat1, 2, 0, 0),
  [3552] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(514),
  [3555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [3557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_expression_list, 1, 0, 0),
  [3559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(244),
  [3561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1559),
  [3563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [3565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [3567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1572),
  [3569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [3571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [3573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension, 1, 0, 0),
  [3575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [3577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [3579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [3581] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_argument, 1, 0, 0),
  [3583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(776),
  [3585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_commands_list, 1, 0, 0),
  [3587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1034),
  [3589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [3591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type_arguments, 1, 0, 0),
  [3593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [3595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0),
  [3597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1234),
  [3600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [3602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1594),
  [3604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [3606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [3608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [3610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1597),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [3616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(230),
  [3618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1604),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [3622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [3624] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_catch_type_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1649),
  [3627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_catch_type_list_repeat1, 2, 0, 0),
  [3629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [3631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [3633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [3635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type_name, 2, 0, 0),
  [3637] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_name, 2, 0, 0),
  [3639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(228),
  [3641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1646),
  [3643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [3645] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_parameter_list, 1, 0, 0),
  [3647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [3649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [3651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_parameter_list, 2, 0, 0),
  [3653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(941),
  [3655] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dimension_repeat1, 2, 0, 0), SHIFT_REPEAT(1512),
  [3658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dimension_repeat1, 2, 0, 0),
  [3660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [3662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [3666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [3670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2064),
  [3672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_name, 1, 0, 0),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [3678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_parameter, 1, 0, 0),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [3682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 3, 0, 0),
  [3684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_command, 1, 0, 0),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [3688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 1, 0, 0),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_member, 1, 0, 0),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [3698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_name, 1, 0, 0),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [3702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 2, 0, 0),
  [3704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_argument, 3, 0, 0),
  [3706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_method_parameter, 2, 0, 0),
  [3708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [3710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [3712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_name, 1, 0, 0),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [3716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 4, 0, 0),
  [3718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [3720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1936),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2046),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [3734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [3736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [3738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_parameter_default, 2, 0, 0),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [3742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2083),
  [3746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_parameter, 3, 0, 0),
  [3748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [3750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [3754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_declaration, 2, 0, 0),
  [3756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [3758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1769),
  [3760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [3764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [3766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1951),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [3770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 8),
  [3772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2025),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2031),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [3786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_condition, 1, 0, 0),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [3794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [3796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_name, 1, 0, 0),
  [3798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [3804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [3806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [3808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [3812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [3814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [3816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [3818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 1, 0, 0),
  [3820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_initializer, 1, 0, 0),
  [3822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [3824] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [3828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [3830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [3832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_declaration, 3, 0, 0),
  [3834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [3836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [3838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [3840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [3842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [3844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [3846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [3848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [3850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1448),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [3858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [3860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [3862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [3864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [3866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [3870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [3872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [3874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [3880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [3888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [3890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_parameter, 1, 0, 0),
  [3892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [3894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [3896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [3898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [3900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [3904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [3906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [3908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [3910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [3912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [3914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [3916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [3918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [3920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [3922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [3924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [3926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [3928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [3930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [3932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [3934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [3936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_member, 3, 0, 0),
  [3938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1054),
  [3940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [3942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [3944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [3946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [3948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [3950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [3952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [3956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [3958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_commands_allowed, 2, 0, 0),
  [3960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [3962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [3964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [3966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [3970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [3972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [3974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [3976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [3978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [3980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1088),
  [3982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [3984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [3986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [3988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [3990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [3992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [3994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [3996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [3998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1696),
  [4000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [4002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [4004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [4006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2122),
  [4008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [4010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [4012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1676),
  [4014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [4016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [4018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [4020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [4022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [4026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [4032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [4034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1677),
  [4036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [4038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block, 1, 0, 2),
  [4040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_condition, 2, 0, 0),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [4044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [4054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1650),
  [4056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [4058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1652),
  [4060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [4062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [4064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [4066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [4068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [4070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1657),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [4074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1662),
  [4078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1639),
  [4094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [4096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_iterator, 1, 0, 0),
  [4098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 5, 0, 0),
  [4100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [4102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [4104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [4106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [4108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 2, 0, 0),
  [4110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [4112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [4120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [4124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1814),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [4138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [4140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [4142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [4144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [4146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [4150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [4152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [4156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [4158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1611),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [4168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_condition, 1, 0, 0),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2042),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [4194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [4196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [4198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [4200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [4202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [4204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block_body, 1, 0, 3),
  [4206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1609),
  [4208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [4210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [4212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block_body, 1, 0, 4),
  [4214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [4216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_property_definition, 6, 0, 0),
  [4218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [4220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [4222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1674),
  [4224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [4226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block, 4, 0, 12),
  [4228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [4230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [4232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [4234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [4238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1888),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1890),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [4244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [4246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [4248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [4250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [4252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [4254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [4256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [4258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1911),
  [4260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [4262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [4264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [4266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1658),
  [4268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1937),
  [4270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [4272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [4274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [4276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [4278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [4280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [4282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [4284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [4286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [4288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [4290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [4292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label_expression, 1, 0, 0),
  [4294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1680),
  [4296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [4298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [4300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [4302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [4304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [4306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [4308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [4310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [4312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [4314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [4316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [4318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [4320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_flow_control_statement, 2, 0, 0),
  [4322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [4324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [4326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [4328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [4332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [4336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [4338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [4342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [4346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [4348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [4350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [4352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [4354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [4356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_key_expression, 1, 0, 0),
  [4358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script_block, 3, 0, 10),
  [4360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2019),
  [4362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [4364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [4366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [4368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [4370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_condition, 3, 0, 0),
  [4372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1282),
  [4374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [4376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1281),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__statement_terminator = 0,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__statement_terminator] = sym__statement_terminator,
};

static const bool ts_external_scanner_states[2][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__statement_terminator] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_powershell_external_scanner_create(void);
void tree_sitter_powershell_external_scanner_destroy(void *);
bool tree_sitter_powershell_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_powershell_external_scanner_serialize(void *, char *);
void tree_sitter_powershell_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_powershell(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_powershell_external_scanner_create,
      tree_sitter_powershell_external_scanner_destroy,
      tree_sitter_powershell_external_scanner_scan,
      tree_sitter_powershell_external_scanner_serialize,
      tree_sitter_powershell_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "powershell",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 10,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
